/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Lang
#else
#define __SEG__ Lang
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <ctype.h>
#include "dbug.h"

#ifndef CR
#define CR 13
#endif /*CR*/
#ifndef LF
#define LF 10
#endif /*LF*/

extern Symbolhandle    Lookup();
extern Symbolhandle    RInstall();
extern void            Removesymbol();
extern char          **mygethandle();
extern char          **mygrowhandle();
extern void            mydisphandle();
extern long            symbolSize();

static char            IndexNames[MAXWDEPTH][NAMELENGTH+1];
/* setup for if, while, for */
/*
   931110  Ifsetup() now returns error code rather than set Global IfsetupError
*/
int Ifsetup(list, op)
Symbolhandle    list;
long            op;
{

	/* routine to test logic and correspondingly modify the input stream of
	the parser */
/* 
  as of 911002, the only modification of input stream is to skip ahead to
  be just before '{'  and to wipe out input stream on errors.
*/

	Symbolhandle    logic, indexVal, range;
	char            c, c1;
	char           *opname;
	unsigned char  *inputString;
	long            condition;
	long            i,i1;
	long            iftype = (op == IF || op == ELSEIF);
	long            nullList = (list == (Symbolhandle) 0);
	WHERE("Ifsetup");

	OUTSTR[0] = '\0';
	SETUPINT(errorExit);
	
	if(op != ELSE)
	{ /* IF, ELSEIF, FOR and WHILE */
		if(iftype)
		{
			opname = (op == IF) ? "if" : "elseif";
		}
		else
		{
			opname = (op == WHILE) ? "while" : "for";
		}
		
		if((!iftype && WDEPTH >= MAXWDEPTH) ||
		   (op == IF && IDEPTH >= MAXIDEPTH))
		{
			sprintf(OUTSTR,"ERROR: too many nested %s's",opname);

			goto errorExit;
		}

		if(!iftype)
		{ /* for & while */
			WDEPTH++;
		}
		else if(op == IF)
		{
			IDEPTH++;
		}
	
		if(op == WHILE || (iftype && !nullList))
		{
			if(NCOMPS(list) > 1)
			{
				sprintf(OUTSTR,"ERROR: usage is %s(LOGICAL){ ... }",opname);
			}
			else if ((logic = COMPVALUE(list,0)) == (Symbolhandle) 0)
			{
				sprintf(OUTSTR,"ERROR: no %s condition value",opname);
			}
			else if(!isDefined(logic))
			{
				sprintf(OUTSTR,"ERROR: %s condition value is UNDEFINED",
						opname);
			}
			else if(TYPE(logic) != LOGIC)
			{
				sprintf(OUTSTR,"ERROR: %s condition value not LOGICAL",opname);
			}
			else if(!isScalar(logic))
			{
				sprintf(OUTSTR,
						"ERROR: %s condition value has length > 1", opname);
			}
			else if(isMissing(DATAVALUE(logic,0)))
			{
				sprintf(OUTSTR,
						"ERROR: %s condition value is MISSING",opname);
			}
			if(*OUTSTR)
			{
				goto errorExit;
			}
			condition = (DATAVALUE(logic,0) != 0.0);
		} /* if(op == WHILE || (iftype && !nullList) */
		else if(op == FOR)
		{ /* for */
			opname = "for";
			if(FORVECTORS[WDEPTH-1] == (double **) 0)
			{
				/* first time through at this level */
				indexVal = COMPVALUE(list,0);
				COMPVALUE(list,0) = (Symbolhandle) 0;
				if(indexVal == (Symbolhandle) 0 ||
				   TYPE(indexVal) == BLTIN || isscratch(NAME(indexVal)))
				{
					sprintf(OUTSTR,
							"ERROR: %s index to %s loop",
							(indexVal==(Symbolhandle) 0) ? "missing" : "illegal",
							opname);
					goto errorExit;
				}

				if(NCOMPS(list) != 2 ||
				   (range = COMPVALUE(list,1)) == (Symbolhandle) 0 ||
				   !isVector(range) || TYPE(range) != REAL)
				{
					sprintf(OUTSTR,
							"ERROR: usage is  for(index,realVector){ .... }");
					goto errorExit;
				}
				WHILELIMITS[WDEPTH-1] = symbolSize(range);
				/* get temporary storage for range */
				FORVECTORS[WDEPTH-1] = 
					(double **) mygethandle(WHILELIMITS[WDEPTH-1]*sizeof(double));
				if(FORVECTORS[WDEPTH-1] == (double **) 0)
				{
					goto errorExit;
				}
				/* copy range to FORVECTORS in reverse order */
				i1 = WHILELIMITS[WDEPTH-1];
				i = 0;
				while(i1 > 0)
				{
					(*FORVECTORS[WDEPTH-1])[--i1] = DATAVALUE(range,i++);
				}
				/* save index name */
				strncpy(IndexNames[WDEPTH-1],NAME(indexVal),NAMELENGTH);
				IndexNames[WDEPTH-1][NAMELENGTH] = '\0';
			} /* if(FORVECTORS[WDEPTH-1] == (double **) 0) */
			else
			{ /* not first time through */
				indexVal = Lookup(IndexNames[WDEPTH-1]);
			}
			/* re-use index variable if it is a REAL scalar */
			if(indexVal != (Symbolhandle) 0 &&
				(TYPE(indexVal) != REAL || NDIMS(indexVal) != 1 ||
				 DIMVAL(indexVal,1) != 1))
			{
				Removesymbol(indexVal);
				indexVal = (Symbolhandle) 0;
			}
			if(indexVal == (Symbolhandle) 0)
			{
				indexVal = RInstall(IndexNames[WDEPTH-1],1);
				if(indexVal == (Symbolhandle) 0)
				{
					goto errorExit;
				}
			}
			DATAVALUE(indexVal,0) = 
				(*FORVECTORS[WDEPTH-1])[WHILELIMITS[WDEPTH-1]-1];
		}
 
		if(op == WHILE)
		{
			/* WHILESTARTS, WHILEBRACKETLEV, and IFBRACKETLEV set in yylex */
			if(--WHILELIMITS[WDEPTH-1] <= 0 && condition)
			{					/* safety precaution */
				/* Note: Parser also recognizes this condition and terminates
				   loop
				*/
				sprintf(OUTSTR,"ERROR: more than %ld repetitions of %s loop",
						(long) MAXWHILE-1,opname);
				putOUTSTR();
			}	
			WHILECONDITIONS[WDEPTH-1] = condition;
		} /*if(op == WHILE)*/
		else if(op == FOR)
		{/* Note: WDEPTH & WHILELIMITS[WDEPTH-1] are decremented in mainpars */
			WHILECONDITIONS[WDEPTH-1] = 1;
			if(WHILELIMITS[WDEPTH-1] == 1)
			{ /* this is the last time */
				IndexNames[WDEPTH-1][0] = '\0';
				mydisphandle((char **) FORVECTORS[WDEPTH-1]);
				FORVECTORS[WDEPTH-1] = (double **) 0;
			}
		} /*if(op == WHILE){}else if(op == FOR)*/
		else if(!nullList) /* op == IF or op == ELSEIF */
		{
			IFCONDITIONS[IDEPTH-1] = condition;
		}
	} /* if(op != ELSE) */
	else
	{
		opname = "else";
	}
	
	/* go look for starting bracket enclosing the compound statement */

	inputString = *INPUTSTRING;
	while ((c = inputString[ISTRCHAR++]))
	{
		if (c == ' ' || c == '\t')
		{
			continue;
		}
		if(c == '\\' &&
		   ((c1 = inputString[ISTRCHAR]) == CR || c1 == LF))
		{
			ISTRCHAR++;
			continue;
		}
		
		if (c == '{')
		{
			break;
		}
		sprintf(OUTSTR,"ERROR: need '{' after '%s'",opname);
		putOUTSTR();
		break;
	} /*while ((c = inputString[ISTRCHAR++]))*/

	ISTRCHAR--;	/* put '{' or other character back so yylex will see it */
	if(op == FOR)
	{
		WHILESTARTS[WDEPTH-1] = ISTRCHAR;
	}
	*OUTSTR = '\0';
	
	if(interrupted() != INTNOTSET)
	{
		goto errorExit;
	}

	RETURN (1);

  errorExit:
	putOUTSTR();

	/* report error to parser */

	RETURN (0);
	
}
