/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Lang
#else
#define __SEG__ Lang
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <math.h>
#include "blas.h"

#define RETURN2(RESULT) PRINT("RDEPTH = %d, INTERRUPT = %08x\n",RDEPTH,INTERRUPT,0,0); \
	RDEPTH -= (RDEPTH > 0) ? 1 : 0; return RESULT

#define MAX(A,B) (((A) > (B) ) ? (A) : (B))
#define illegalBitValue(X) ((X) < 0 || (X) != floor(X) || (X) > MAXBITVALUE)

extern Symbolhandle RInstall(), Install();
extern void         Removesymbol();

extern char       **mygethandle(), **mygrowhandle();

extern char        *opName(), *typeName();

extern long         symbolSize();
extern long         dupDims();
extern long         isNull();
extern long         isMatrix();
extern long         isVector();
extern long         anyMissing();
extern char        *skipStrings();

static Symbolhandle doMatMult(Symbolhandle, Symbolhandle, unsigned long);
static Symbolhandle doBinary(Symbolhandle, Symbolhandle, unsigned long, 
							 unsigned long *);
static Symbolhandle doUnary(Symbolhandle, unsigned long, unsigned long *);

Symbolhandle        Unary(Symbolhandle, int);
Symbolhandle        Arith(Symbolhandle, Symbolhandle, int);
Symbolhandle        Logic(Symbolhandle, Symbolhandle, int);

static Symbolhandle elementMatrix(Symbolhandle, Symbolhandle);
static char       **extractIt(Symbolhandle, Symbolhandle, long []);

Symbolhandle        Element(Symbolhandle, Symbolhandle);
long                buildSubscript(Symbolhandle, long, 	long, long *);
long                subscriptLength(Symbolhandle, long);

/*
  Check operands of binary operatons for type and compatibility.  They
  are already known to be non-null
*/

static unsigned long checkOperands(Symbolhandle arg1, Symbolhandle arg2,
								   int op)
{
	long            type1 = TYPE(arg1), type2 = TYPE(arg2);
	long            structype1 = type1, structype2 = type2;
	char           *symptom = (char *) 0, *opType = (char *) 0;
	unsigned long   control = (unsigned long) op;
	long            reply;
	long            matmult, logOp, arithOp, compareOp, bitOp;
	WHERE("checkOperands");
	
	*OUTSTR = '\0';
	matmult = (op == MATMULT || op == TRMATMULT || op == MATMULTTR);
	logOp = (op == AND || op == OR);
	compareOp = (op == EQ || op == NE || op == LT || op == LE ||
				 op == GT || op == GE);
	bitOp = (op == BITAND || op == BITOR || op == BITXOR);
	arithOp = !(logOp || compareOp || bitOp);

	if(type1 == STRUC)
	{
		control |= LEFTRECUR;
		structype1 = getSingleType(arg1);
		if(structype1 == 0)
		{
			sprintf(OUTSTR,
					"ERROR: structure components of left operand to %s are not all of same type",
					opName(op));
			putOUTSTR();
			goto yyerrorMsg;
		}
		
		if(((arithOp || bitOp) && structype1 != REAL &&
			!matmult && structype1 != LOGIC) ||
		   (logOp && structype1 != LOGIC) ||
		   (compareOp && structype1 != REAL && structype1 != CHAR &&
			structype1 != LOGIC))
		{
			sprintf(OUTSTR,
					"ERROR: structure components of left operand to %s are",
					opName(op));
			putOUTSTR();

			if(logOp || arithOp && matmult)
			{ /* only one type allowed */
				sprintf(OUTSTR, "not all %ss",(logOp) ? "LOGICAL" : "REAL");
			}
			else if(arithOp || bitOp)
			{ /* REAL and LOGICAL allowed */
				sprintf(OUTSTR, "not all REAL or LOGICAL");
			}
			else
			{ /* compareOP */
				sprintf(OUTSTR, "not all REAL, CHARACTER, or LOGICAL");
			}
			goto yyerrorMsg;
		}
	} /*if(type1 == STRUC)*/

	if(type2 == STRUC)
	{
		control |= RIGHTRECUR;
		structype2 = getSingleType(arg2);
		if(structype2 == 0)
		{
			sprintf(OUTSTR,
					"ERROR: structure components of right operand to %s are not all of same type",
					opName(op));
			putOUTSTR();
			goto yyerrorMsg;
		}
		
		if(((arithOp || bitOp) && structype2 != REAL &&
			!matmult && structype2 != LOGIC) ||
		   (logOp && structype2 != LOGIC) ||
		   (compareOp && structype2 != REAL && structype2 != CHAR &&
			structype2 != LOGIC))
		{
			sprintf(OUTSTR,
					"ERROR: structure components of right operand to %s are",
					opName(op));
			putOUTSTR();

			if(logOp || arithOp && matmult)
			{ /* only one type allowed */
				sprintf(OUTSTR, "not all %ss",(logOp) ? "LOGICAL" : "REAL");
			}
			else if(arithOp || bitOp)
			{ /* REAL and LOGICAL allowed */
				sprintf(OUTSTR, "not all REAL or LOGICAL");
			}
			else
			{ /* compareOP */
				sprintf(OUTSTR, "not all REAL, CHARACTER, or LOGICAL");
			}
			goto yyerrorMsg;
		}
	} /*if(type2 == STRUC)*/

	if(type1 != STRUC || type2 != STRUC)
	{
		if(type1 == UNDEF || type1 == ASSIGNED ||
		   type2 == UNDEF || type2 == ASSIGNED)
		{
			symptom = "undefined";
		}
		else if(matmult)
		{
			if(type1 != REAL && type1 != STRUC ||
			   type2 != REAL && type2 != STRUC)
			{
				symptom = "non-numeric";
			}
		}
		else if(logOp)
		{
			if(type1 != LOGIC && type1 != STRUC ||
			   type2 != LOGIC && type2 != STRUC)
			{
				symptom = "non-logical";
			}
		}
		else if(arithOp || bitOp)
		{
			if((type1 != REAL && type1 != LOGIC && type1 != STRUC) || 
			   (type2 != REAL && type2 != LOGIC && type2 != STRUC))
			{
				symptom = "non-numeric and non-logical";
			}
		}
		else
		{ /*comparOp*/
			if(type1 != REAL && type1 != CHAR && type1 != LOGIC &&
			   type1 != STRUC ||
			   type2 != REAL && type2 != CHAR && type2 != LOGIC &&
			   type2 != STRUC)
			{
				symptom = "non REAL, CHARACTER, or LOGICAL";
			}
		}
		if(symptom)
		{
			if(matmult)
			{
				opType = "matrix multiplication";
			}
			else if(arithOp)
			{
				opType = "arithmetic";
			}
			else if(bitOp)
			{
				opType = "bit operation";
			}
			else if(logOp)
			{
				opType = "logical operation";
			}
			else
			{
				opType = "comparison";
			}
			
			sprintf(OUTSTR,"ERROR: %s with %s operand",opType,symptom);
			putOUTSTR();
			sprintf(OUTSTR,"%s %s %s",
					typeName(type1),opName(op),typeName(type2));
			goto yyerrorMsg;
		}		
	}
	if(type1 == STRUC)
	{
		type1 = structype1;
	}
	if(type2 == STRUC)
	{
		type2 = structype2;
	}
	
	if(compareOp && type1 != type2 && (type1 == CHAR || type2 == CHAR))
	{
		myerrorout("ERROR: CHARACTER value compared with non-CHARACTER value");
		sprintf(OUTSTR,"%s %s %s",
				typeName(type1),opName(op),typeName(type2));
		goto yyerrorMsg;
	}
	
	/* check dimensions and tree structure for compatibility */

	reply = treecmp(arg1,arg2,control);

	if(reply == BADDIMS)
	{
		sprintf(OUTSTR,
				"ERROR: dimension mismatch %sfor %s",
				(control & (LEFTRECUR|RIGHTRECUR)) ?
				"somewhere in structure " : "",
				opName(op));
	}
	else if(reply == BADSTRUC)
	{
		sprintf(OUTSTR,
				"ERROR: structure shapes do not match for %s", opName(op));
	}
	else if(reply == HASMISSING)
	{
		sprintf(OUTSTR,"ERROR: MISSING values not allowed for %s",opName(op));
	}
	if(*OUTSTR)
	{
		goto yyerrorMsg;
	}

	return (control);

  yyerrorMsg:
	yyerror(OUTSTR);

	putOUTSTR();

	return (0);
	
} /*checkOperands()*/

static Symbolhandle doMatMult(Symbolhandle arg1, Symbolhandle arg2,
							  unsigned long control)
{
	long            i, j, ij;
	long            dim[2][2];
	long            ikincr, kjincr, ik, kj, incri, incrj;
	long            ikstart, kjstart;
	long            limi, limj, limk;
	long            ncomps;
	long            leftrecur = isStruc(arg1);
	long            rightrecur = isStruc(arg2);
	int             op = control & OPMASK;
	Symbolhandle    left, right, symh;
	Symbolhandle    result = (Symbolhandle) 0;
	double          y;
	double         *a1, *a2, *a3;
	WHERE("doMatMult");

/*  note, we already know arg1 and arg2 are either both non-null
	and compatible or both null */	

	if(leftrecur || rightrecur)
	{
		ncomps = (leftrecur) ? NCOMPS(arg1) : NCOMPS(arg2);
		result = StrucInstall(SCRATCH,ncomps);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		
		for(i=0;i<ncomps;i++)
		{
			left = (leftrecur) ? COMPVALUE(arg1,i) : arg1;
			right = (rightrecur) ? COMPVALUE(arg2,i) : arg2;
			symh = doMatMult(left,right,control);
			if(symh == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			setCompName(symh,(leftrecur) ? NAME(left) : NAME(right));
			COMPVALUE(result,i) = symh;
		}
		return (result);
	}
	else if(!isNull(arg1))
	{
		(void) isMatrix(arg1,dim[0]);
		(void) isMatrix(arg2,dim[1]);
		if(op == MATMULT)
		{
			limi = dim[0][0];
			limj = dim[1][1];
			ikincr = dim[0][0];
			kjincr = 1;
			incri = 1;
			incrj = dim[1][0];
			limk = dim[0][1];
		}
		else if(op == TRMATMULT)
		{
			limi = dim[0][1];
			limj = dim[1][1];
			ikincr = 1;
			kjincr = 1;
			incri = dim[0][0];
			incrj = dim[1][0];
			limk = dim[0][0];
		}
		else					/* MATMULTTR */
		{
			limi = dim[0][0];
			limj = dim[1][0];
			ikincr = dim[0][0];
			kjincr = dim[1][0];
			incri = 1;
			incrj = 1;
			limk = dim[0][1];
			if(limk != dim[1][1])
			{
 /* this should never happen now, since dimensions are pre-checked*/
				limk = -1;
			}
		}	
	}
	else
	{
		limi = limj = 0;
	}
	
	result = RInstall(SCRATCH,limi * limj);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	if(limi > 0)
	{
		setNDIMS(result,2);
		setDIM(result,1,limi);
		setDIM(result,2,limj);
	
		a1 = DATAPTR(arg1);
		a2 = DATAPTR(arg2);
		a3 = DATAPTR(result);

		ikstart = 0;
		for (i = 0; i < limi; i++)
		{
			kjstart = 0;
			ij = i;
			for (j = 0; j < limj; j++)
			{

				ik = ikstart;
				kj = kjstart;
				kjstart += incrj;
			
/* blas ddot should save a little on %c% because kjincr and ikincr are both 1*/

				y = DDOT(&limk,a1+ik,&ikincr,a2+kj,&kjincr);
				a3[ij] = y;
				ij += limi;
			} /*for (j = 0; j < limj; j++)*/
			ikstart += incri;
		} /*for (i = 0; i < limi; i++)*/
	} /*if(limi > 0)*/
	
	return (result);

  errorExit:
	Removesymbol(result);
	
	return (0);
} /*doMatMult()*/

static Symbolhandle doUnary(Symbolhandle arg,
							unsigned long control, unsigned long *status)
{
	long            i, ndims, tot1;
	double          x;
	double        **dest;
	long            ncomps;
	long            recur = isStruc(arg);
	int             op = control & OPMASK;
	Symbolhandle    left, symh;
	Symbolhandle    result = (Symbolhandle) 0;
	WHERE("doUnary");

/*  note, we already know arg non-null */	
	if(recur)
	{
		ncomps = NCOMPS(arg);
		if(control & REUSELEFT)
		{
			result = arg;
		}
		else
		{
			result = StrucInstall(SCRATCH,ncomps);
			if(result == (Symbolhandle) 0)
			{
				goto errorExit;
			}
		}		

		for(i=0;i<ncomps;i++)
		{
			left = (recur) ? COMPVALUE(arg,i) : arg;
			if(!isNull(left))
			{
				symh = doUnary(left,control,status);
			}
			else if(left == (Symbolhandle) 0)
			{
				symh = RInstall(SCRATCH,0);
			}
			else
			{
				symh = left;
			}
			
			if(symh == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			if(left != (Symbolhandle) 0)
			{
				setCompName(symh,NAME(left));
			}
			else
			{
				setCompName(symh,"NULL");
			}
			COMPVALUE(result,i) = symh;
		}
		return (result);
	} /*if(recur)*/

	/* Not a structure */
	/* Find total number of data values */
	tot1 = symbolSize(arg);
	ndims = NDIMS(arg);

	/* set up result structure */
	if(control & REUSELEFT)
	{ /* reuse argument if arg is scratch or unary plus */
		 result = arg;
	}
	else
	{
		result = RInstall(SCRATCH,tot1);
		if (result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNDIMS(result,ndims);
		for(i=1;i<=ndims;i++)
		{
			setDIM(result,i,DIMVAL(arg,i));
		}
	}
	if(op == NOT)
	{
		setTYPE(result,LOGIC);
	}
	else
	{
		setTYPE(result,REAL);
	}
	dest = DATA(result);
	
	for (i = 0; i < tot1; i++)
	{
		x = DATAVALUE(arg,i);
		if(isMissing(x))
		{
			*status |= FOUNDMISSING;
		}
		else if(op == NOT)
		{
			x = (double) (x == 0.0);
		}
		else if(op == '-')
		{ /* unary minus */
			x = -x;
		}
#ifdef BITOPS
		else if(op == BITNOT)
		{ /* bit reversal */
			if(illegalBitValue(x))
			{
				*status |= FOUNDDOMERR;
				setMissing(x);
			}
			else
			{
				x = (double) ~ (unsigned long) x;
			}
		}
#endif /*BITOPS*/
		/* do nothing if unary plus or MISSING */
		(*dest)[i] = x;
	}

	return (result);

  errorExit:
	Removesymbol(result);
	
	return (0);
	
} /*doUnary()*/

static Symbolhandle doBinary(Symbolhandle arg1, Symbolhandle arg2,
							 unsigned long control, unsigned long *status)
{
	long            tot1, maxtot, middle1, middle2, middle;
	long            dim1[MAXDIMS+1], dim2[MAXDIMS+1], dim[MAXDIMS+1];
	long            ndims,klim;
	long            i, j, k, jk1,jk2, ijk1,ijk2, ijk;
	long            incijk1, incijk2, incjk1, incjk2, inck1, inck2;
	long            scalar1;
	long            typeChar = (TYPE(arg1) == CHAR  || TYPE(arg1) == MACRO);
	long            arithOp = 0, logOp = 0;
	double          yijk, xijk1 , xijk2;
	char           *sijk1, *sijk2, *sjk1, *sjk2;
	char           *s1, *s2;
	double         *x1, *x2, *dest;
	long            ncomps;
	long            leftrecur = isStruc(arg1);
	long            rightrecur = isStruc(arg2);
	int             op = control & OPMASK;
	Symbolhandle    left, right, symh;
	Symbolhandle    result = (Symbolhandle) 0;
	short           addOp, subOp, multOp, divOp, modOp, expOp;
	short           eqOp, neOp, gtOp, geOp, ltOp, leOp, andOp, orOp;
#ifdef BITOPS
	short           bitandOp, bitxorOp, bitorOp;
#endif /*BITOPS*/
	WHERE("doBinary");	

/*  note, we already know arg1 and arg2 are either both non-null and
	compatible or both null */	
	if(leftrecur || rightrecur)
	{
		ncomps = (leftrecur) ? NCOMPS(arg1) : NCOMPS(arg2);
		if(control & REUSELEFT)
		{
			result = arg1;
		}
		else if(control & REUSERIGHT)
		{
			result = arg2;
		}
		else
		{
			result = StrucInstall(SCRATCH,ncomps);
			if(result == (Symbolhandle) 0)
			{
				goto errorExit;
			}
		}
		
		for(i=0;i<ncomps;i++)
		{
			left = (leftrecur) ? COMPVALUE(arg1,i) : arg1;
			right = (rightrecur) ? COMPVALUE(arg2,i) : arg2;
			symh = doBinary(left,right,control,status);
			if(symh == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			if(!(control & (REUSELEFT | REUSERIGHT)))
			{
				Cutsymbol(symh);
			}
			
			setCompName(symh,(leftrecur) ? NAME(left) : NAME(right));
			
			COMPVALUE(result,i) = symh;
		} /*for(i=0;i<ncomps;i++)*/

		return (result);
	} /*if(leftrecur || rightrecur)*/

	/* Neither argument is structure */
	/* Copy dimensions into scratch & compute total size */
	if(!isNull(arg1))
	{
		tot1 = dupDims(dim1,arg1);
		(void) dupDims(dim2,arg2);
		
		ndims = MAX(dim1[0],dim2[0]);

		/* fill out short dimension with 1's */
		for(i=dim1[0]+1;i<=ndims;i++)
		{
			dim1[i] = 1;
		}
	
		for(i=dim2[0]+1;i<=ndims;i++)
		{
			dim2[i] = 1;
		}
	
		dim1[0] = dim2[0] = ndims;
	
		middle1 = middle2 = 1;	/* product of dimensions excluding 1st and last */

		scalar1 = (tot1 == 1);

		/* note: dimensions are known to match */
		for(i=2;i<ndims;i++)
		{/* middle dimensions must match exactly unless at least 1 arg scalar*/
			dim[i] = (!scalar1) ? dim1[i] : dim2[i];
			middle1 *= dim1[i];
			middle2 *= dim2[i];
		}

		dim[1] = MAX(dim1[1],dim2[1]);
		if(ndims > 1)
		{
			dim[ndims] = MAX(dim1[ndims],dim2[ndims]);
			middle = MAX(middle1,middle2);
			klim = dim[ndims];
			maxtot = dim[1] * middle * dim[ndims];
		}
		else
		{
			klim = middle = 1;
			maxtot = dim[1];
		}

		incijk1 = (dim1[1] == 1) ? 0 : 1;
		incijk2 = (dim2[1] == 1) ? 0 : 1;

		inck1 = (dim1[ndims] == 1) ? 0 : dim1[1]*middle1;
		inck2 = (dim2[ndims] == 1) ? 0 : dim2[1]*middle2;

		incjk1 = (middle1 == 1) ? 0 : dim1[1];
		incjk2 = (middle2 == 1) ? 0 : dim2[1];

	}
	else
	{
		maxtot = 0;
	}
	
	
	/* make up output structure */

	if(control & REUSELEFT)
	{
		result = arg1;
	}
	else if(control & REUSERIGHT)
	{
		result = arg2;
	}
	else
	{
		result = RInstall(SCRATCH,maxtot);
		if (result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
	}
	
	if(maxtot > 0)
	{
		/* ensure result is not a factor */
		setNCLASS(result, -1);
		setNDIMS(result,ndims);
		for(i=1;i<=ndims;i++)
		{
			setDIM(result,i,dim[i]);
		}
	
		addOp = (op == ADD);
		subOp = (op == SUB);
		multOp = (op == MULT);
		divOp = (op == DIV);
		modOp = (op == MOD);
		expOp = (op == EXP);
		arithOp = (addOp || subOp || multOp || divOp || modOp || expOp);
		if(arithOp)
		{
			eqOp = neOp = gtOp = geOp = ltOp = leOp = andOp = orOp = 0;
#ifdef BITOPS
			bitandOp = bitxorOp = bitorOp = 0;
#endif /*BITOPS*/
		} /*if(arithOp)*/
		else
		{
			eqOp = (op == EQ);
			neOp = (op == NE);
			gtOp = (op == GT);
			geOp = (op == GE);
			ltOp = (op == LT);
			leOp = (op == LE);
			andOp = (op == AND);
			orOp = (op == OR);
#ifndef BITOPS
			logOp = 1;
#else /*BITOPS*/
			logOp = eqOp || neOp || gtOp || geOp || ltOp || leOp ||
				andOp || orOp;
			if(logOp)
			{
				bitandOp = bitxorOp = bitorOp = 0;
			}
			else
			{
				bitandOp = (op == BITAND);
				bitxorOp = (op == BITXOR);
				bitorOp = (op == BITOR);
			} /*if(!logOp)*/
#endif /*BITOPS*/
		} /*if(arithOp){}else{}*/
	
		setTYPE(result, (logOp) ? LOGIC : REAL);
		dest = DATAPTR(result);
		if(typeChar)
		{
			s1 = STRINGPTR(arg1);
			s2 = STRINGPTR(arg2);
		}
		else
		{
			x1 = DATAPTR(arg1);
			x2 = DATAPTR(arg2);
		}

		ijk = 0;
		for(k=0;k<klim;k++)
		{
		
			jk1 = k*inck1;
			jk2 = k*inck2;
			if(typeChar)
			{
				sjk1 = skipStrings(s1,jk1);
				sjk2 = skipStrings(s2,jk2);
			}
		
			for(j=0;j<middle;j++)
			{
				if(typeChar)
				{
					sijk1 = sjk1;
					sijk2 = sjk2;
				}
				else
				{
					ijk1 = jk1;
					ijk2 = jk2;
				}
				for(i=0;i<dim[1];i++)
				{
					if(!typeChar)
					{
						xijk1 = x1[ijk1];
						xijk2 = x2[ijk2];
						if (isMissing(xijk1) || isMissing(xijk2))
						{
							setMissing(yijk);
							*status |= FOUNDMISSING;
						}
						else if(arithOp)
						{
							if(addOp)
							{
								yijk = xijk1 + xijk2;
							}
							else if(subOp)
							{
								yijk = xijk1 - xijk2;
							}
							else if(multOp)
							{
								yijk = xijk1 * xijk2;
							}
							else if(divOp)
							{
								if (xijk2 == 0.0)
								{
									if(xijk1 == 0.0)
									{ /* 0/0 is 0 */
										yijk = 0.0;
									}
									else
									{
										setMissing(yijk);
										*status |= FOUNDZERODIV;
									}
								}
								else
								{
									yijk = xijk1 / xijk2;
								}
							}
							else if(modOp)
							{
								if (xijk2 == 0.0)
								{
									setMissing(yijk);
									*status |= FOUNDZERODIV;
								}
								else
								{
									yijk = floor(xijk1 / xijk2);
									yijk = xijk1 - xijk2*yijk;
								}
							}
							else if(expOp)
							{
								if (xijk2 == 0.)
								{ /* x^0 is 1, including 0^0 */
									yijk = 1.0;
								}
								else if (xijk1 == 0.)
								{ /* 0^p is 0, p > 0 or MISSING, p < 0 */
									if(xijk2 > 0)
									{
										yijk = 0;
									}
									else
									{
										setMissing(yijk);
									}
								}
								else
#ifdef USEPOW /* if defined, system pow() is efficient on integer powers */
								{ /* x != 0, p != 0 */
									yijk = pow(fabs(xijk1), xijk2);
									if(xijk1 < 0)
									{
										if (xijk2 != floor(xijk2))
										{ /* x < 0, p not integer */
											yijk = -yijk;
											*status |= FOUNDNEG;
										}
										else if((long) xijk2 % 2 != 0)
										{ /* x < 0, odd integer power */
											yijk = -yijk;
										}
									}	
								} /* if ... else */
#else  /*USEPOW*/
								if(xijk2 == floor(xijk2))
								{/* x != 0, p an integer */
									yijk = intpow(xijk1,xijk2);
								} /*if(xijk2 == floor(xijk2))*/
								else
								{/* x != 0, p not integer */
									yijk = pow(fabs(xijk1), xijk2);
									if(xijk1 < 0)
									{
										yijk = -yijk;
										*status |= FOUNDNEG;
									}
								} /*if(xijk2 == floor(xijk2)){}else{}*/
#endif /*USEPOW*/
							}
						} /* if(arithOp) */
						else if(logOp)
						{/* logical operations */
							if(eqOp)
							{
								yijk = (xijk1 == xijk2) ? 1.0 : 0.0;
							}
							else if(neOp)
							{
								yijk = (xijk1 != xijk2) ? 1.0 : 0.0;
							}
							else if(gtOp)
							{
								yijk = (xijk1 > xijk2) ? 1.0 : 0.0;
							}
							else if(geOp)
							{
								yijk = (xijk1 >= xijk2) ? 1.0 : 0.0;
							}
							else if(ltOp)
							{
								yijk = (xijk1 < xijk2) ? 1.0 : 0.0;
							}
							else if(leOp)
							{
								yijk = (xijk1 <= xijk2) ? 1.0 : 0.0;
							}
							else if(andOp)
							{
								yijk = ((xijk1!=0.0) && (xijk2!=0.0)) ?
									1.0 : 0.0;
							}
							else if(orOp)
							{
								yijk = ((xijk1!=0.0) || (xijk2!=0.0)) ?
									1.0 : 0.0;
							} /* if ... else ... else ... */
						}  /* if(arithOp){}else if(logOp){}*/
#ifdef BITOPS
						else
						{ /* bit operations */
							if (illegalBitValue(xijk1) ||
								illegalBitValue(xijk2))
							{
								setMissing(yijk);
								*status |= FOUNDDOMERR;
							}
							else if(bitandOp)
							{
								yijk = (double) ((unsigned long) xijk1 &
												 (unsigned long) xijk2);
							}
							else if(bitxorOp)
							{
								yijk = (double) ((unsigned long) xijk1 ^
												 (unsigned long) xijk2);
							}
							else if(bitorOp)
							{
								yijk = (double) ((unsigned long) xijk1 |
												 (unsigned long) xijk2);
							}
						} /* if(arithOp){}else if(logOp){}else{}*/
#endif /*BITOPS*/
					} /* if(!typeChar) */
					else
					{ /* typeChar */
						if(eqOp)
						{
							yijk = (strcmp(sijk1, sijk2) == 0) ? 1.0 : 0.0;
						}
						else if(neOp)
						{
							yijk = (strcmp(sijk1, sijk2) != 0) ? 1.0 : 0.0;
						}
						else if(gtOp)
						{
							yijk = (strcmp(sijk1, sijk2) > 0) ? 1.0 : 0.0;
						}
						else if(geOp)
						{
							yijk = (strcmp(sijk1, sijk2) >= 0) ? 1.0 : 0.0;
						}
						else if(ltOp)
						{
							yijk = (strcmp(sijk1, sijk2) < 0) ? 1.0 : 0.0;
						}
						else if(leOp)
						{
							yijk = (strcmp(sijk1, sijk2) <= 0) ? 1.0 : 0.0;
						}
					} /* if(!typeChar){}else{} */
					dest[ijk++] = yijk;

					if(i < dim[1] - 1)
					{
						if(incijk1)
						{
							if(typeChar)
							{
								sijk1 = skipStrings(sijk1,incijk1);
							}
							else
							{
								ijk1++;
							}
						}
						if(incijk2)
						{
							if(typeChar)
							{
								sijk2 = skipStrings(sijk2,incijk2);
							}
							else
							{
								ijk2++;
							}
						} /*if(incijk2)*/
					} /*if(i < dim[1] - 1)*/
				} /* for(i=0;i<dim[1];i++) */

				if(j < middle-1)
				{
					if(incjk1)
					{
						if(typeChar)
						{
							skipStrings(sjk1,incjk1);
						}
						else
						{
							jk1 += incjk1;
						}
					}
					if(incjk2)
					{
						if(typeChar)
						{
							sjk2 = skipStrings(sjk2,incjk2);
						}
						else
						{
							jk2 += incjk2;
						}
					} /*if(incjk2)*/
				} /* if(j < middle-1) */
			} /* for(j=0;j<middle;j++) */
		} /* for(k=0;k<klim;k++) */
	} /*if(maxtot > 0)*/

	return (result);

  errorExit:
	Removesymbol(result);

	return (0);
	
} /*doBinary()*/

Symbolhandle    Unary(arg,op)
Symbolhandle    arg;
int             op;
{
	/* unary minus or plus or bitwise complement for a real data set*/
	Symbolhandle    result = (Symbolhandle) 0;
	char           *symptom = (char *) 0;
	long            type;
	unsigned long   control = (long) op, status = 0;

	*OUTSTR = '\0';
	
	if(isNull(arg))
	{
		symptom = "missing or null";
	}
	else if((type = TYPE(arg)) == UNDEF || type == ASSIGNED)
	{
		symptom = "undefined argument";
	}
	else if(type != REAL && type != LOGIC)
	{
		if(type != STRUC)
		{
			symptom = "non REAL and non LOGICAL argument";
		}
		else if(!isReal(arg) && !isLogic(arg))
		{
			symptom = "structure with non REAL and non LOGICAL component";
		}
		else
		{
			control |= LEFTRECUR;
		}
	}
	if(symptom)
	{
		sprintf(OUTSTR,"ERROR: %s with %s",
				(op != BITNOT) ? "arithmetic" : "bit operation", symptom);
		goto yyerrorMsg;
	}
	if(isscratch(NAME(arg)))
	{
		control |= REUSELEFT;
	}
	
	result = doUnary(arg, control, &status);
/*
	Note: doUnary overwrites a SCRATCH arg and returns it as result.
	Hence it should not be deleted here, even if SCRATCH.
*/
	if(status & FOUNDMISSING)
	{
		sprintf(OUTSTR,"WARNING: %s%s with MISSING value(s)",
				(op == BITNOT) ? "" : "unary ",opName(op));
		putOUTSTR();
	}
#ifdef BITOPS
	if(status & FOUNDDOMERR)
	{
		sprintf(OUTSTR,
				"WARNING: bit operation with illegal value(s); operation is %s",
				opName(op));
		putOUTSTR();
	}
#endif /*BITOPS*/
	return (result);

  yyerrorMsg:
	yyerror(OUTSTR);

	return (0);
} /*Unary()*/

/*
  Function to handle matrix multiplication, elementwise binary, and
  arithmetic operations.

  It has been modified (by kb) to allow more cases than just combination
  of identical dimensioned arrays, or an array with a scalar.
  1.  the dimensions of the array with fewest dimensions is augmented with 1's
      to the number of dimensions of the other array.
  2.  If the first dimension of an array is 1, it is treated as if it had the
      same first dimension as the other array, with each row (trailing slab)
      identical
  3.  If the last dimension of an array is 1, it is treated as if it had the
      same last dimension of the other array, with each column (trailing slab)
      identical.
  4.  If all dimensions of an array are 1, it is treated as a scalar, except
      the dimensionality of the result will be at least as long as the
      dimensionality of the array.
  5.  arg1 and/or arg2 may be structures.  If they both are, then their
      structure patterns must be identical and corresponding components must
      be compatible as in 1-4.  If only one is a structure, all its components
      must be compatible with the other argument as in 1-4.
*/

Symbolhandle    Arith(arg1, arg2, op)
Symbolhandle    arg1, arg2;
int             op;
{

	Symbolhandle    result = (Symbolhandle) 0;
	unsigned long   control, control1, status = 0;
	int             bitop = (op == BITAND || op == BITOR || op == BITXOR);
	WHERE("Arith");

	*OUTSTR = '\0';
	
	if (isNull(arg1) || isNull(arg2))
	{
		sprintf(OUTSTR,
				"ERROR: %s with null argument; operation is %s",
				(!bitop) ? "arithmetic" : "bit operation",
				opName(op));
		goto yyerrorMsg;
	}
	
	control = checkOperands(arg1, arg2, op);

	if(control != 0)
	{
		if(op == MATMULT || op == TRMATMULT || op == MATMULTTR)
		{
			result = doMatMult(arg1,arg2,control);
		}
		else
		{
			control1 = (control & ~(OPMASK)) | NEARSTRICT;
			if(isscratch(NAME(arg1)) &&
			   (isScalar(arg2) || 
				(isStruc(arg1) || !isStruc(arg2)) &&
				 treecmp(arg1, arg2, control1) == 0))
			{
					control |= REUSELEFT;
			}
			else if(isscratch(NAME(arg2)) &&
			   (isScalar(arg1) || 
				(isStruc(arg2) || !isStruc(arg1)) &&
				 treecmp(arg1, arg2, control1) == 0))
			{
				control |= REUSERIGHT;
			}

			result = doBinary(arg1, arg2,control,&status);

			if(status & FOUNDMISSING)
			{
				sprintf(OUTSTR,
						"WARNING: %s with missing value(s); operation is %s",
						(!bitop) ? "arithmetic" : "bit operation",
						opName(op));
				putOUTSTR();
			}					/*if(status & FOUNDMISSING)*/
#ifdef BITOPS
			if(status & FOUNDDOMERR)
			{
				sprintf(OUTSTR,
						"WARNING: bit operation with illegal value(s); operation is %s",
						opName(op));
				putOUTSTR();
			}	/*if(status & FOUNDDOMERR)*/
#endif /*BITOPS*/

			if(status & FOUNDZERODIV)
			{
				myerrorout("WARNING: Zero divide set to MISSING");
			}					/*if(status & FOUNDZERODIV)*/

			if(status & FOUNDNEG)
			{
				myerrorout("WARNING: fractional power of negative value");
			}					/*if(status & FOUNDNEG)*/
		}
	} /*if(control != 0)*/
/* fall through */	
  yyerrorMsg:
	if(*OUTSTR)
	{
		yyerror(OUTSTR);
	}
	
 /* remove unconsumed SCRATCH arguments */
	 if(arg1 != (Symbolhandle) 0 &&
		isscratch(NAME(arg1)) && !(control & REUSELEFT))
	 {
		 Removesymbol(arg1);
	 }
	 if(arg2 != (Symbolhandle) 0 &&
		isscratch(NAME(arg2)) && !(control & REUSERIGHT))
	 {
		 Removesymbol(arg2);
	 }
	
	 return (result);
} /*Arith()*/

/*
  routine to do logical operations on arrays and structures of arrays
  operations are 
	  AND, OR, NOT (require LOGICAL operands)
	  EQ, NE, LT, LE, GT, GE (require REAL or CHAR operands)
*/

Symbolhandle    Logic(arg1, arg2, op)
Symbolhandle    arg1, arg2;
int             op;
{
	Symbolhandle  result = (Symbolhandle) 0;
	unsigned long control = 0, status = 0, control1;
	long          type = 0, logOp;
	
	*OUTSTR = '\0';
	logOp = (op == AND || op == OR || op == NOT);

	if (isNull(arg1) || op != NOT && isNull(arg2))
	{
		sprintf(OUTSTR,"ERROR: %s with null argument; operation = %s",
				(logOp) ? "logical operation" : "comparison",opName(op));
		goto yyerrorMsg;
	}
	
	type = TYPE(arg1);
	if(op != NOT)
	{
		control = checkOperands(arg1, arg2, op);
		if(!control)
		{
			goto errorExit;
		}
		control1 = (control & ~(OPMASK)) | NEARSTRICT;
		if(isscratch(NAME(arg1)) && getSingleType(arg1) != CHAR &&
		   (isScalar(arg2) || 
			(isStruc(arg1) || !isStruc(arg2)) &&
			treecmp(arg1, arg2, control1) == 0))
		{
				control |= REUSELEFT;
		}
		else if(isscratch(NAME(arg2)) && getSingleType(arg2) != CHAR &&
		   (isScalar(arg1) || 
			(isStruc(arg2) || !isStruc(arg1)) &&
			treecmp(arg1, arg2, control1) == 0))
		{
			control |= REUSERIGHT;
		}
		result = doBinary(arg1, arg2, control, &status);
	} /*if(op != NOT)*/
	else
	{
		if(type == STRUC)
		{
			if(!isLogic(arg1))
			{
				sprintf(OUTSTR,
						"ERROR: structure components for operand of %s are not all LOGICAL",
						opName(op));
				goto yyerrorMsg;
			}
		}
		else if(type != LOGIC)
		{
			sprintf(OUTSTR,
					"ERROR: nonlogical operand for logical operator");
			yyerror(OUTSTR);
			sprintf(OUTSTR,"%s(%s)",opName(op),typeName(type));
			goto errorExit;
		}
		control = op;
		if(isscratch(NAME(arg1)))
		{
			control |= REUSELEFT;
		}
		result = doUnary(arg1,control, &status);
	} /*if(op != NOT){}else{}*/
	
	if(status & FOUNDMISSING)
	{
		sprintf(OUTSTR,
				"WARNING: %s with missing value(s)",
				(logOp) ? "logical operation" : "comparison");
	} /*if(status & FOUNDMISSING)*/
	
  yyerrorMsg:
	if(*OUTSTR)
	{
		yyerror(OUTSTR);
	}
	/* fall through */
  errorExit:
	putOUTSTR();
	
 /* remove unconsumed SCRATCH arguments */
	 if(arg1 != (Symbolhandle) 0 &&
		isscratch(NAME(arg1)) && !(control & REUSELEFT))
	 {
		 Removesymbol(arg1);
	 }
	 if(op != NOT && arg2 != (Symbolhandle) 0 &&
		isscratch(NAME(arg2)) && !(control & REUSERIGHT))
	 {
		 Removesymbol(arg2);
	 }
	
	return (result);

} /*Logic()*/


static Symbolhandle elementMatrix(Symbolhandle arg,
								  Symbolhandle subscriptsymh)
{
	long            n, m, i, j, ndims;
	long            lstart, k, l, factor;
	long            subscript;
	long            foundBadSubscript = 0;
	double         *x, *y, *subscr;
	Symbolhandle    result = (Symbolhandle) 0;
	WHERE("elementMatrix");
	
	ndims = NDIMS(subscriptsymh);
	n = DIMVAL(subscriptsymh,ndims);
	m = symbolSize(subscriptsymh)/n;

	result = RInstall(SCRATCH,m);
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	setTYPE(result,TYPE(arg));
	setNDIMS(result,ndims - 1);

	for(i=1;i<ndims;i++)
	{
		setDIM(result,i,DIMVAL(subscriptsymh,i));
	}
	/* dereference handles */
	x = DATAPTR(arg);
	y = DATAPTR(result);
	subscr = DATAPTR(subscriptsymh);
	
	lstart = 0;
	for(i=0;i<m;i++)
	{
		l = lstart++;
		k = 0;
		factor = 1;
		for(j=1;j<=n;j++)
		{
			subscript = (long) subscr[l] - 1;
			l += m;
			if(subscript < 0 || subscript >= DIMVAL(arg,j))
			{
				foundBadSubscript = 1;
				break;
			}

			k += factor*subscript;
			factor *= DIMVAL(arg,j);
		}
		if(j <= n)
		{
			setMissing(y[i]);
		}
		else
		{
			y[i] = x[k];
		}
	}
	if(foundBadSubscript)
	{
		myerrorout("WARNING: illegal subscripts found.  Result set to MISSING");
	}

	return(result);

  errorExit:
	Removesymbol(result);

	return (0);
	
} /*elementMatrix()*/

#define GALLSUBS    0
#define GDH         1
#define NTRASH      2

static char   **extractIt(Symbolhandle arg, Symbolhandle list,
						  long resdim[])
{
	long       *subvals[MAXDIMS];	
	long      **allsubs = (long **) 0;
	long        ii[MAXDIMS];
	double    **dh = (double **) 0;
	char      **ch = (char **) 0;
	char       *chTail, *chPlace;
	long        lastTail, lTail, lLast = 0;
	long        maxSubscr, needed = 0;
	long        m ,n, chLength, resLength;
	long        i, j, k, l;
	long        type = TYPE(arg), ndims = NDIMS(arg);
	long        ncomps = NCOMPS(list);
	long        tmplong;
	WHERE("extractIt");
	TRASH(NTRASH,errorExit);
	
	for(i=0;i<ncomps;i++)
	{
		subvals[i] = (long *) 0;
		ii[i] = 0;
		resdim[i] = 0;
	}

	while(ncomps > ndims && COMPVALUE(list,ncomps-1) == (Symbolhandle) 0)
	{ /* ignore trailing empty subscripts */
		ncomps--;
	}
	
	for(i=0;i<ncomps;i++)
	{
		if(ncomps == 1)
		{
			maxSubscr = symbolSize(arg);
		}
		else
		{
			maxSubscr = DIMVAL(arg,i+1);
			maxSubscr = (maxSubscr > 0) ? maxSubscr : 1;
		}
		resdim[i] = subscriptLength(COMPVALUE(list,i),maxSubscr);
		needed += resdim[i];
		if(resdim[i] == 0)
		{
			goto errorExit;
		}
	}

	if(!getScratch(allsubs,GALLSUBS,needed,long))
	{
		goto errorExit;
	}
	m = n = 1;
	subvals[0] = *allsubs;
	for(i=0;i<ncomps;i++)
	{
		if(ncomps == 1)
		{
			maxSubscr = symbolSize(arg);
		}
		else
		{
			maxSubscr = DIMVAL(arg,i+1);
			maxSubscr = (maxSubscr > 0) ? maxSubscr : 1;
		}
		resdim[i] = tmplong = buildSubscript(COMPVALUE(list, i),
											 maxSubscr,resdim[i],subvals[i]);
		if(tmplong == 0)
		{
			goto errorExit;
		}

		if(i > 0)
		{
			for(j=0;j<tmplong;j++)
			{
				(subvals[i])[j] *= n;
			}
		}
		m *= tmplong;
		n *= maxSubscr;
		subvals[i+1] = subvals[i] + tmplong;
	} /* for i */

/* m is number of elements in output, n is number of elements in input */

	if(type == CHAR)
	{
		lastTail = 0;
		chPlace = chTail = STRINGPTR(arg);
	}
	else
	{
		if(!getScratch(dh,GDH,m,double))
		{
			goto errorExit;
		}
		ch = (char **) dh;
	}
	
	lTail = -1;
	for(k = 0;k < m;k++)
	{ /* loop over destination elements */
		l = (subvals[0])[ii[0]];
		if(lTail < 0)
		{ /* lTail < 0 signals change in ii[j], j > 0 */
			lTail = 0;
			for(j=1;j<ncomps;j++)
			{
				lTail += (subvals[j])[ii[j]];
			}
			
			if(type == CHAR)
			{
				if(lTail < lastTail)
				{
					lastTail = 0;
					chTail = STRINGPTR(arg);
				}

				chTail = skipStrings(chTail,lTail-lastTail);
				lastTail = lTail;
			}
		}
		l += lTail;
		if(type == CHAR)
		{
			if(l < lLast)
			{
				lLast = lTail;
				chPlace = chTail;
			}
			chPlace = skipStrings(chPlace,l-lLast);
			lLast = l;
			
			chLength = strlen(chPlace) + 1;
 			if(k == 0)
			{
				ch = mygethandle(chLength);
				resLength = 0;
			}
			else
			{
				ch = mygrowhandle(ch,resLength + chLength);
			}
			if(ch == (char **) 0)
			{
				goto errorExit;
			}
			strcpy((*ch) + resLength,chPlace);
			resLength += chLength;
		}
		else
		{
			(*dh)[k] = DATAVALUE(arg,l);
		}

		for(j=0;j<ncomps;j++)
		{ /* step odometer */
			if(++ii[j] < resdim[j])
			{
				break;
			}
			lTail = -1;
			ii[j] = 0;
		} /* for j */
	} /* for k */

	unTrash(GDH);
	emptyTrash();

	return (ch);

  errorExit:
	emptyTrash();

	return (0);
	
} /*extractIt()*/

Symbolhandle    Element(arg1, list)
Symbolhandle    arg1, list;
{

	/* routine to extract elements of arrays */

	Symbolhandle    subscriptsymh, result = (Symbolhandle) 0;
	long            argType, subscrType;
	long            ndims, ncomps = NCOMPS(list), resdim[MAXDIMS];
	long            i;
	long            matrixSubscript = 0;
	char          **ch;
	WHERE("Element");

	if (isNull(arg1))
	{
		sprintf(OUTSTR,"ERROR: attempt to use subscripts with null variable");
	}
	else if(!isDefined(arg1))
	{
		sprintf(OUTSTR,
				"ERROR: attempt to use subscripts with undefined variable %s",
				NAME(arg1));
	}
	else if ((argType = TYPE(arg1)) == BLTIN)
	{
		sprintf(OUTSTR,
				"ERROR: subscripts illegal with function name");
	}
	else if(ncomps > MAXDIMS)
	{
		sprintf(OUTSTR,"ERROR: more than %ld subscripts",(long) MAXDIMS);
	}
	else if (argType == STRUC)
	{
		return (Extract(arg1,list,0));
	}
	else if (argType != REAL && argType != LOGIC && argType != CHAR)
	{	/* subscripts are not of proper type */
			sprintf(OUTSTR,"ERROR: subscripts illegal with type %s variable",
					typeName(argType));
	}
	if(*OUTSTR)
	{
		goto yyerrorMsg;
	}
	
	ndims = NDIMS(arg1);
	
	if (ncomps != 1 && ncomps < ndims)
	{
		sprintf(OUTSTR,"ERROR: too few subscripts");
		goto yyerrorMsg;
	}
	
	subscriptsymh = COMPVALUE(list,0);
	
	if(!isNull(subscriptsymh) && isDefined(subscriptsymh) &&
	   ncomps == 1 &&  !isVector(subscriptsymh))
	{/* single subscript is matrix or array */
		if(DIMVAL(subscriptsymh,NDIMS(subscriptsymh)) != ndims)
		{
			sprintf(OUTSTR,
					"ERROR: last dimension of subscript array of wrong size");
			goto yyerrorMsg;
		}
		matrixSubscript = 1;
	}
	
	for (i = 0; i < ncomps; i++)
	{
		subscriptsymh = COMPVALUE(list, i);
		if (subscriptsymh == (Symbolhandle) 0)
		{		/* blank subscript */
			continue;
		}
		if(isNull(subscriptsymh))
		{
			sprintf(OUTSTR,"ERROR: null subscript");
		}
		else if(!isDefined(subscriptsymh))
		{
			sprintf(OUTSTR,"ERROR: undefined subscript %s", 
					NAME(subscriptsymh));
		}
		else if ((subscrType = TYPE(subscriptsymh)) != LOGIC &&
				 subscrType != REAL)
		{
			sprintf(OUTSTR,"ERROR: subscript not REAL or LOGICAL");
		}
		else if(anyMissing(subscriptsymh))
		{
			sprintf(OUTSTR,"ERROR: MISSING values not allowed as subscripts");
		}
		else if(matrixSubscript)
		{
			break;
		}
		else if(!isVector(subscriptsymh))
		{
			sprintf(OUTSTR,"ERROR: array used as marginal subscript");
		}
		if(*OUTSTR)
		{
			goto yyerrorMsg;
		}
	} /* for */

	if (ncomps == 1 && COMPVALUE(list, 0) == (Symbolhandle) 0)
	{ /* single empty subscript: simply copy entire array */
		if(!isscratch(NAME(arg1)))
		{
			result = Install(SCRATCH, (int) TYPE(arg1));
			if (result == (Symbolhandle) 0 || Copy(arg1, result) == 0)
			{
				goto errorExit;
			}
		}
		else
		{ /* already SCRATCH, so reuse it */
			result = arg1;
			arg1 = (Symbolhandle) 0;
		}
		setNAME(result,SCRATCH);
	}
	else if(matrixSubscript) /* matrix of subscripts */
	{		
		if(TYPE(arg1) == CHAR)
		{
			sprintf(OUTSTR,
					"ERROR: matrix of subscripts illegal for CHAR array");
			goto yyerrorMsg;
		}
		
		result = elementMatrix(arg1,subscriptsymh);
	}
	else
	{
		ch = extractIt(arg1, list, resdim);

		if (ch != (char **) 0 && 
			(result = Install(SCRATCH, (int) argType)) != (Symbolhandle) 0)
		{
			switch (argType)
			{
			  case LOGIC:
			  case REAL:
				setDATA(result,(double **) ch);
				break;
			  case CHAR:
				setSTRING(result,ch);
				break;
			}					/* switch */

			ndims = 0;
			for(i=0;i<ncomps;i++)
			{
				if(resdim[i] == 0)
				{
					break;
				}
				ndims++;
			}
			setNDIMS(result,ndims);
			
			for (i = 1; i <= ndims; i++)
			{
				setDIM(result,i,resdim[i - 1]);
			}					/* for */
		}
	}
	if(arg1 != (Symbolhandle) 0 && isscratch(NAME(arg1)))
	{
		Removesymbol(arg1);
	}
	
	*OUTSTR = '\0';
	return (result);

  yyerrorMsg:
	yyerror(OUTSTR);
	/* fall through */

  errorExit:
	putOUTSTR();
	Removesymbol(result);

	return (0);
} /*Element()*/

/* compute maximum space needed for scratch for subscript */

long subscriptLength(subscriptsymh,maxsubscr)
Symbolhandle     subscriptsymh;
long             maxsubscr;
{
	long         nitems;
	long         subscriptSign;
	WHERE("subScriptLength");
	
	if (subscriptsymh == (Symbolhandle) 0)
	{
		/* use everything for empty list */
		nitems = maxsubscr;
	}
	else if (TYPE(subscriptsymh) == LOGIC)
	{
		nitems = symbolSize(subscriptsymh);
		if(nitems != maxsubscr)
		{
			sprintf(OUTSTR,
					"ERROR: length of LOGICAL subscript vector must match dimension");
			goto yyerrorMsg;
		}
	}
	else
	{
		/* numeric list */
		nitems = symbolSize(subscriptsymh);
		subscriptSign = (DATAVALUE(subscriptsymh,0) >= 0) ? 1 : -1;

		if(subscriptSign < 0)
		{
			if(nitems >= maxsubscr)
			{
				sprintf(OUTSTR,"ERROR: too many negative subscripts");
				goto yyerrorMsg;
			}
			else	
			{
				nitems = maxsubscr;
			}
		}
	}
	return (nitems);

  yyerrorMsg:
	yyerror(OUTSTR);

	return (0);
} /*subscriptLength()*/

/*
  fill vector subValues with subscripts (origin 0) associated with subscriptsymh
  maxsubscr is the maximum permissible value (origin 1) and *nitems is set
  to the number returned
*/

long buildSubscript(subscriptsymh,maxsubscr,nitems,subValues)
Symbolhandle     subscriptsymh;
long             maxsubscr, nitems, *subValues;
{
	long         i, j, itmp, mitems, length = nitems;
	long         subscriptSign;
	double       tmp;
	WHERE("buildSubscript");
	
	if (subscriptsymh == (Symbolhandle) 0)
	{
		/* use everything for empty list */
		for (i = 0; i < length; i++)
		{
			subValues[i] = i;
		}
	}
	else if (TYPE(subscriptsymh) == LOGIC)
	{
		/* convert logical to an integer list */
		j = 0;
		for (i = 0; i < length; i++)
		{
			if (DATAVALUE(subscriptsymh,i) > 0.0)
			{
				if (i >= maxsubscr)
				{
					sprintf(OUTSTR,"ERROR: subscript out of range");
					goto yyerrorMsg;
				}
				subValues[j++] = i;
			}
		}
		if (j == 0)
		{
			sprintf(OUTSTR,"ERROR: all F logical subscript");
			goto yyerrorMsg;
		}
		nitems = j;
	}
	else
	{
		/* numeric list */
		subscriptSign = (DATAVALUE(subscriptsymh,0) >= 0) ? 1 : -1;

		mitems = length;
		length = symbolSize(subscriptsymh);
		if(subscriptSign < 0)	
		{
			for(i=0;i<mitems;i++)
			{
				subValues[i] = 0;
			}
		}

		for (i = 0; i < length; i++)
		{
			tmp = DATAVALUE(subscriptsymh,i);
			if (tmp != floor(tmp))
			{
				sprintf(OUTSTR,"ERROR: noninteger subscript");
				goto yyerrorMsg;
			}
			if((itmp = (long) tmp) == 0 || labs(itmp) > maxsubscr)
			{
				sprintf(OUTSTR,"ERROR: subscript out of range");
				goto yyerrorMsg;
			}
			if(itmp*subscriptSign < 0)
			{
				sprintf(OUTSTR,"ERROR: can't mix positive and negative subscripts");
				goto yyerrorMsg;
			}
			if(subscriptSign > 0)
			{
				subValues[i] = itmp-1;
			}
			else
			{
				if(++subValues[(-itmp)-1] > 1)
				{
					sprintf(OUTSTR,"ERROR: duplicate negative subscripts");
					goto yyerrorMsg;
				}
			}
		} /* for(i=0;i<length;i++) */
		if(subscriptSign < 0)
		{
			j = 0;
			for(i = 0;i<mitems;i++)
			{
				if(subValues[i] == 0)
				{
					subValues[j++] = i;
				}
			}
			nitems = j;
		}
	}

	return (nitems);
	
  yyerrorMsg:
	yyerror(OUTSTR);
	
	return (0);
	
} /*buildSubscript()*/
