/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Lang
#else
#define __SEG__ Lang
#endif /*MPW3*/
#endif /*MPW*/
#include "globals.h"
#include <ctype.h>
#include "dbug.h"

#ifndef Newline
#ifdef MACINTOSH
#define Newline 13
#else
#define Newline 10
#endif /*MACINTOSH*/
#endif /*Newline*/

/*
  Modified by kb to add expanded macro to INPUTSTRING instead of starting
  fresh.  This was necessitated by the introduction of while loops for
  which it is normally necessary to return to the start of a loop.  A while
  loop containing a macro would crash.

  '$1', '$2', ... are replaced by the text of argument 1, argument 2, ...
  '$0'  is equivalent to '$1,$2,...,$nargs'
  '$N'  is replaced by the number of arguments to the macro
  '$A'  is equivalent to 'cat("$1","$2",...)'
  '$K'  is like '$0' except only keyword arguments are included
  '$k'  is like '$N' except only keyword arguments are counted
  '$V'  is like '$0' except only non-keyword arguments are included
  '$v'  is like '$N' except only non-keyword arguments are counted
  '$S'  is replaced by the name of the macro
  '$$'  is replaced by the current value of static variable level which is
        incremented mod 100
  A '$' preceded by '\' is not interpreted specially
  All these constructs except $A are recognized insided quotes as well as
  outside.

  Macrosetup returns 0 on error, 1 on normal return
*/

long            MACROSTART; /* set by yylex.c */

#define CHUNK 64
#define HEAD   2 /* number of characters inserted before text */
#define TAIL   3 /* maximum no. characters inserted after text, including '\0' */
#define SPACEUSED(I) (SpaceUsed += (I), SpaceLeft -= (I))
#define inputString    (*INPUTSTRING)
#define Trash          NAMEFORTRASH
static Symbolhandle    Trash;

static unsigned char **TempString;
static long            SpaceUsed, SpaceLeft;
static long            CurrentLength;

static int moreSpace(long needed)
{
	if(CurrentLength < SpaceUsed)
	{ /* should never happen */
		fprintf(stderr,"PANIC: CurrentLength = %ld, SpaceUsed = %ld\n",
				CurrentLength, SpaceUsed);
	}
	if(CurrentLength == 0)
	{
		setSTRING(Trash,mygethandle(needed));
	}
	else
	{
		setSTRING(Trash,mygrowhandle(STRING(Trash),CurrentLength+needed));
	}
	
	TempString = (unsigned char **) STRING(Trash);
	if (TempString == (unsigned char **) 0)
	{
		return (0);
	}
	CurrentLength += needed;
	SpaceLeft = CurrentLength - SpaceUsed;

	return (1);
}

static long countEscapes(register char *s)
{
	register long   nq = 0;
	register char   c;
	
	while((c = *s++) != '\0')
	{
		if(c == '"' || c == '\\')
		{
			nq++;
		}
	}
	return (nq);
}

/*
   Copy string s2 to s1, inserting '\' before every '"' and '\'
   Returns the number of characters added to s1
*/

static long strEscapeCpy(register unsigned char *s1, char *s2)
{
	register unsigned char     c;
	unsigned char             *start = s1;

	while((c = *s2++) != '\0')
	{
		if(c == '"' || c == '\\')
		{
			*s1++ = '\\';
		}
		*s1++ = c;
	}
	*s1 = '\0';
	return (s1 - start);
}

/* check to see whether argument is a keyword phrase */
static long iskeyarg(char *s)
{
	char     c;
	
	if (!istempname(s) && !isnamestart(*s))
	{
		return (0);
	} /*if (!istempname(s) && !isnamestart(*s))*/
	
	while((c = *(++s), isnamechar(c) || isspace(c)))
	{
		;
	} /*while((c = *(++s), isnamechar(c)))*/
	return (c == ':');
} /*iskeyarg()*/

long Macrosetup(macrosymh, list)
Symbolhandle    macrosymh, list;
{
	/* routine to insert macro code into the input stream of the parser */

	char            c, lastc, nextc = '\0', *name;
	char            level[2];
	unsigned char **macroText;
	char           *argList;
	long            i, j, argno, macstrchar;
	long            needed, length;
	long            argListSize = 0, varListSize = 0, keyListSize = 0;
	long            inquotes, iskey, copyit, escaped = 0;
	long            nargs = NARGS(list), nKey = 0, nVar = 0, ncopied;
	Symbolhandle    symh;
	WHERE("Macrosetup");
	
	Trash = Install(SCRATCH,CHAR);
	if(Trash == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	level[0] = '\0';
	
	MDEPTH++;

	if(nargs == 1 && COMPVALUE(list,0) == (Symbolhandle) 0)
	{
		nargs = 0;
	}
	
	macroText = (unsigned char **) STRING(macrosymh);

	/* this should be a valid handle because macro() should enforce it */

	/* MACROSTART set by yylex() to mark the insertion point of the macro */
	needed = MACROSTART + HEAD + strlen((char *) *macroText) + TAIL;
	/* compute best guess at space needed */
	for(i = 0;i<nargs;i++)
	{
		/* Note all non-null arguments should be guaranteed to have type CHAR*/
		if((symh = COMPVALUE(list,i)) != (Symbolhandle) 0)
		{
			/* assume each arg used at least once */
			argList = STRINGPTR(symh);
			length = strlen(argList);
			needed += length; /* assumes each argument used at least once */
			argListSize += length + 1;
			iskey = iskeyarg(argList);
				
			if(iskey)
			{
				keyListSize += length+1;
				nKey++;
			}
			else
			{
				varListSize += length+1;
				nVar++;
			}
		}
	} /*for(i = 0;i<nargs;i++)*/

	SpaceUsed = 0;
	CurrentLength = 0;
	SpaceLeft = 0;
	if(!moreSpace(needed))
	{
		goto errorExit;
	}

	/* begin by duplicating INPUTSTRING up to start of macro */
	strncpy((char *) (*TempString), (char *) inputString, (size_t) MACROSTART);
	SPACEUSED(MACROSTART);

/* The macro text is wrapped with a leading '{' and a trailing '\n}' */
	(*TempString)[SpaceUsed] = '{';
	SPACEUSED(1);
	(*TempString)[SpaceUsed] = '\n';
	SPACEUSED(1);
	(*TempString)[SpaceUsed] = '\0';

	/*
	  Now copy macroText into TempString with substitutions.

	  TempString should have space for the complete text of the macro plus
	  one copy of each argument.  Thus we only need to check for space needed
	  when expanding something.

	  Spaces and tabs outside of quotes are not copied, nor are trailing
	  comments starting with '#'
    */

	macstrchar = 0;
	inquotes = 0;
	lastc = '\0';
	while ((c = (*macroText)[macstrchar++]))
	{
		needed = 1;
		if(!inquotes)
		{
			if(c == '#')
			{					/* don't copy comment */
				while((c = (*macroText)[macstrchar++]) && c != Newline)
				{/* skip to end of line or end of string */
					;
				}
				
				if(c == '\0')
				{
					break;
				}
			}
			else if((c == ' ' || c == '\t') && (lastc == ' ' || lastc == '\t'))
			{/* don't copy extra white space outside of quotes */
				continue;
			}
		} /*if(!inquotes)*/
		
		nextc = (*macroText)[macstrchar];
		if (SpaceLeft < needed + TAIL && !moreSpace(CHUNK))
		{
			goto errorExit;
		}
		/*
		  At this point, lastc is last character copied or '\0';
		  c is current character, not yet copied;
		  and nextc is the next character to be examined
		*/
		switch (c)
		{
		  case '$':
			/* possible start of item to expand */
			escaped = 0;
			if(lastc == '\\')
			{ /* treat '\$' as non-expanding */
				(*TempString)[SpaceUsed-1] = c;
			}
			else if (isdigit(nextc))
			{/* insert argument into (*INPUTSTRING) */
				argno = 0;
				while (isdigit(c = (*macroText)[macstrchar]))
				{
					argno = 10*argno + (c - '0');
					macstrchar++;
				}
				/* insert argument, if any, or entire list for '$0' */
				if (nargs < argno ||
					(symh = COMPVALUE(list, argno - 1)) == (Symbolhandle) 0)
				{ /* argument is missing */
					if(!inquotes)
					{ /* message will be printed if argument missing */
						sprintf(OUTSTR,
								"error(\"ERROR: Argument %ld to macro %s missing\");",
								argno,NAME(macrosymh));
					}
					else
					{ /* no expansion if missing argument is in quotes */
						*OUTSTR = '\0';
					}
					name = OUTSTR;
				}
				else if(argno > 0)
				{
					name = STRINGPTR(symh);
				}

				needed = (argno > 0) ? strlen(name) : argListSize;
				if(inquotes)
				{ /* needed to count the number of '"' && '\' */
					if(argno > 0)
					{
						needed += countEscapes(name);
					} /*if(argno > 0)*/
					else
					{
						for(j=0;j<nargs;j++)
						{
							symh = COMPVALUE(list,j);
							if(symh != (Symbolhandle) 0)
							{							
								needed += countEscapes(STRINGPTR(symh));
							}
						}
					} /*if(argno > 0){}else{}*/
				} /*if(inquotes)*/

				if (SpaceLeft < needed + TAIL && !moreSpace(needed + CHUNK))
				{
					goto errorExit;
				}
				if(argno > 0)
				{ /* copy argument */
					if(!inquotes)
					{
						strcpy((char *) *TempString + SpaceUsed, name);
						length = strlen(name);
					}
					else
					{
						length = strEscapeCpy(*TempString + SpaceUsed,name);
					}
					SPACEUSED(length);
				}
				else
				{/* copy entire argument list, separated by commas */
					for(j=0;j<nargs;j++)
					{
						symh = COMPVALUE(list,j);
						if(symh != (Symbolhandle) 0)
						{
							argList = STRINGPTR(symh);
							if(!inquotes)
							{
								strcpy((char *) *TempString + SpaceUsed,
									   argList);
								length = strlen(argList);
							}
							else
							{
								length = strEscapeCpy(*TempString + SpaceUsed,
												   argList);
							}
							SPACEUSED(length);
						}
						if(j<nargs-1)
						{
							(*TempString)[SpaceUsed] = ',';
							SPACEUSED(1);
						}
					}
				}				
				c = '\0';
			} /* else if (isdigit(nextc))*/
			else if(!isnamechar(lastc) &&
					!isnamechar((*macroText)[macstrchar+1]) &&
					(strchr("KkNSVv",nextc) ||
					 !inquotes && nextc == 'A'))
			{ /* Process $N, $A, $K, $k, $V, $v, $S */
				macstrchar++;
				switch (nextc)
				{
				  case 'N':
				  case 'k':
				  case 'v':
					 /* replace '$N' by nargs, '$k' by nKey, '$v' by nVar */
					sprintf(OUTSTR,"%ld",
							(nextc == 'N') ? nargs :
								((nextc == 'k') ? nKey :nVar));
					needed = strlen(OUTSTR);
					if(SpaceLeft < needed + TAIL && !moreSpace(needed+CHUNK))
					{
						goto errorExit;
					}
					strcpy((char *) *TempString+SpaceUsed,OUTSTR);
					SPACEUSED(needed);
					break;

				  case 'S':
					/* replace $S by NAME(macrosymh) */
					needed = strlen(NAME(macrosymh));
					if(SpaceLeft < needed + TAIL && !moreSpace(needed+CHUNK))
					{
						goto errorExit;
					}
					strcpy((char *) *TempString+SpaceUsed,NAME(macrosymh));
					SPACEUSED(needed);
					break;

				  case 'A':
					/*
					  replace '$A' by cat("arg1","arg2",...,"argn")
					*/
					needed = argListSize + 2*nargs + 5 + 2*(nargs == 0);
					for(j=0;j<nargs;j++)
					{
						symh = COMPVALUE(list,j);
						if(symh != (Symbolhandle) 0)
						{							
							needed += countEscapes(STRINGPTR(symh));
						}
					}
					
					if(SpaceLeft < needed + TAIL && !moreSpace(needed+CHUNK))
					{
						goto errorExit;
					}
					strcpy((char *) *TempString+SpaceUsed,"cat(\"");
					SPACEUSED(5);
					for(j=0;j<nargs;j++)
					{
						symh = COMPVALUE(list,j);
						if(symh != (Symbolhandle) 0)
						{							
							length = strEscapeCpy(*TempString + SpaceUsed,
											   STRINGPTR(symh));
							SPACEUSED(length);
						}
						if(j<nargs-1)
						{
							strcpy((char *) *TempString + SpaceUsed,"\",\"");
							SPACEUSED(3);
						}
					} /*for(j=0;j<nargs;j++)*/ /* for j */
					strcpy((char *) *TempString + SpaceUsed,"\")");
					SPACEUSED(2);
					break;

				  case 'K':
				  case 'V':
					needed = (nextc == 'K') ? keyListSize : varListSize;
					if(inquotes)
					{
						for(j=0;j<nargs;j++)
						{
							symh = COMPVALUE(list,j);
							if(symh != (Symbolhandle) 0)
							{							
								argList = STRINGPTR(symh);
								iskey = iskeyarg(argList);
								if(nextc == 'K' && iskey ||
								   nextc == 'V' && !iskey)
								{
									needed += countEscapes(STRINGPTR(symh));
								}
							}
						}
					} /*if(inquotes)*/					
					
					if (SpaceLeft < needed + TAIL &&
						!moreSpace(needed + CHUNK))
					{
						goto errorExit;
					}
					copyit = 0;
					ncopied = 0;
					for(j=0;j<nargs;j++)
					{
						symh = COMPVALUE(list,j);
						if(symh != (Symbolhandle) 0)
						{
							argList = STRINGPTR(symh);
							iskey = iskeyarg(argList);
							copyit = (nextc == 'K' && iskey) ||
								(nextc == 'V' && !iskey);
							if(copyit)
							{
								if(inquotes)
								{
									length = strEscapeCpy(*TempString+SpaceUsed,
														  argList);
								}
								else
								{
									strcpy((char *) *TempString + SpaceUsed,
										   argList);
									length = strlen(argList);
								}
								
								SPACEUSED(length);
								if(++ncopied < ((nextc == 'K') ? nKey : nVar))
								{
									(*TempString)[SpaceUsed] = ',';
									SPACEUSED(1);
								}
							} /*if(copyit)*/							
						} /*if(symh != (Symbolhandle) 0)*/
					} /*for(j=0;j<nargs;j++)*/
					break;

				} /*switch (nextc)*/
				c = '\0';
			}
			else if(nextc == '$' && !isdigit((*macroText)[macstrchar+1]))
			{ /* '$$' not followed by digit ('$$1' treated like '$arg1') */
				macstrchar++;
				if(level[0] == '\0')
				{/* increment value to be substituted just once per expansion*/
					if(++MLEVEL > MAXLEVEL)
					{
						myerrorout("ERROR: macros nested too deeply or too many active");
						goto errorExit;
					}
					level[0] = MLEVEL/10 + '0';
					level[1] = MLEVEL%10 + '0';
				}
				
				needed = 2;
				if(SpaceLeft < needed + TAIL && !moreSpace(needed+CHUNK))
				{
					goto errorExit;
				}
				(*TempString)[SpaceUsed] = level[0];
				(*TempString)[SpaceUsed+1] = level[1];
				SPACEUSED(2);
			}
			else
			{					/* not a macro substitution */
				goto saveChar;
			}
			break;

		  case '\\':
			if (inquotes)
			{
				escaped = !escaped;
			}
			goto saveChar;

		  case '"':
			if (!inquotes)
			{
				inquotes = 1;
			}
			else if (!escaped)
			{
				inquotes = 0;
			}
			escaped = 0;
			goto saveChar;

		  default:				/* anything else */
			escaped = 0;
			
		  saveChar:
			(*TempString)[SpaceUsed] = c;
			SPACEUSED(1);
		} /*switch (c)*/
		lastc = c;
		(*TempString)[SpaceUsed] = '\0';
	}/* while ((c = (*macroText)[macstrchar++])) */

	if ((c = (*TempString)[SpaceUsed - 1]) != '\n' && c != '\r')
	{ /*ensure expanded macro ends with new line */
		(*TempString)[SpaceUsed] = '\n';
		SPACEUSED(1);
	}
	(*TempString)[SpaceUsed] = '}';
	SPACEUSED(1);

	(*TempString)[SpaceUsed] = '\0';

	/* now copy the rest of the INPUTSTRING into the new input stream */
	length = strlen((char *) inputString+ISTRCHAR);
	needed = length + 1;
	if (SpaceLeft < needed && !moreSpace(needed-SpaceLeft))
	{
		goto errorExit;
	}

	strcpy((char *) (*TempString) + SpaceUsed,
		   (char *) inputString+ISTRCHAR);
	SPACEUSED(length);

	/* now close it off and replace INPUTSTRING */
	(*TempString)[SpaceUsed] = '\0';

	mydisphandle((char **) INPUTSTRING);
	STRING(Trash) = (char **) 0;
	INPUTSTRING = TempString;
	ISTRCHAR = MACROSTART;

	Removesymbol(Trash);

	Trash = (Symbolhandle) 0;
	*OUTSTR = '\0';
	
	return (1);

  errorExit:
	*OUTSTR = '\0';
	Removesymbol(Trash);
	Trash = (Symbolhandle) 0;
	return (0);
} /*Macrosetup()*/

