/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Lang
#else
#define __SEG__ Lang
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

extern void     Removesymbol();

extern void     mydisphandle();
extern long     symbolSize();
extern long     isVector();
extern long     anyMissing();
extern long     buildSubscript();
extern char    *skipStrings();

/*
  Modified by kb to allow the subscripts to be a single matrix, with each
  row providing a full set of subscripts.  Unlike in Element() in Lang.c, it
  requires a matrix (2 dimensional array).

  Modified by kb to allow negative subscripts, specifying positions not to
  be replaced
  arg1[list] <- arg2

  Modified 920608 by the addition of an additonal argument.  It is now
  called twice by yypars, once with checkLhs != 0 to check the l.h.s, before
  parsing the r.h.s., and a second time with checkLhs == 0 after parsing
  the r.h.s.

  Modified 920728 so that it returns an object of the expected shape, that is
  for example  'b<- a[1,] <- 3' sets the first row of a to 3 and sets
  b to a row vector of 3's with dim(b)[2] == dim(a)[2].  Previously it
  simply returned arg1.
*/

#define trash   NAMEFORTRASH
#define GITEMLIST  0
#define GALLSUBS   1
#define GCHNEWARG1 2
#define GCHRESULT  3
#define NTRASH     4

Symbolhandle    Subassign(arg1, list, arg2,checkLhs)
Symbolhandle    arg1, list, arg2;
long            checkLhs;
{

	/* routine to replace elements of arrays */

	Symbolhandle    subscrSymh, result = (Symbolhandle) 0;
	Symbolhandle    dest = (Symbolhandle) 0;
	long            i, i2, i2inc, j, k;
	long            space, resSpace, arg1tot, arg2tot, nitems, factor;
	long            jLast, length;
	long            l, lTail;
	long            maxSubscr, matrixSubscr, sameSymbol = 0;
	long            ii[MAXDIMS],resdim[MAXDIMS];
	long            ndims, ncomps;
	long          **itemlist = (long **) 0;
	long          **allsubs = (long **) 0;
	long           *subValues[MAXDIMS+1];
	long            needed = 0;
	long            position;
	long            subLimit;
	long            argType, subscrType;
	char           *ch1, *ch2;
	double          nclass;
	double          y;
	double         *subscript;
	char          **chNewArg1 = (char **) 0, **chResult = (char **) 0;
	char            namearg1[NAMELENGTH+1];
	Symbolhandle    trash = (Symbolhandle) 0;
	WHERE("Subassign");
	
	*OUTSTR = '\0';
	ncomps = NCOMPS(list);
	
	/* check to make sure subscripts are of proper type and number */
	if(checkLhs)
	{
		if (arg1 == (Symbolhandle) 0)
		{ /* should not happen */
			sprintf(OUTSTR,"ERROR: no variables in subscript assignment");
		}
		else if(!isDefined(arg1))
		{
			sprintf(OUTSTR,"ERROR: use of subscripts with undefined variable %s",
					NAME(arg1));
		}
		else if ((argType = TYPE(arg1)) == STRUC)
		{
			sprintf(OUTSTR,
					"ERROR: illegal to assign to a structure component");
		}
		else if (argType != REAL && argType != LOGIC &&
				 argType != CHAR)
		{
			sprintf(OUTSTR,"ERROR: use of subscripts with %s variable",
					typeName(argType));
		}
		else if(isscratch(NAME(arg1)))
		{
			sprintf(OUTSTR,"ERROR: assignment to subscripts of number, expression or function result");
		}
		else if (ncomps != 1 && ncomps < NDIMS(arg1))
		{
			sprintf(OUTSTR,"ERROR: too few subscripts");
		}
		else if (ncomps == 1 && COMPVALUE(list, 0) == (Symbolhandle) 0)
		{
			sprintf(OUTSTR,"ERROR: empty subscript list");
		}

		if(*OUTSTR)
		{
			yyerror(OUTSTR);
			arg1 = (Symbolhandle) 0;
		}
		return (arg1);
	} /*if(checkLhs)*/
	
	/* now carry out assignment */
	strcpy(namearg1,NAME(arg1));
	if(ncomps > MAXDIMS)
	{
		sprintf(OUTSTR,
				"ERROR: more than %ld subscripts",(long) MAXDIMS);
		goto yyerrorMsg;
	}
	
	for(i = 0;i<ncomps;i++)
	{
		ii[i] = 0;
		subValues[i] = (long *) 0;
	}
	ndims = NDIMS(arg1);
	if(isNull(arg2) || !isDefined(arg2))
	{
		sprintf(OUTSTR,
				"ERROR: replacement elements are %s",
				(isNull(arg2)) ? "null" : "undefined");
		goto yyerrorMsg;
	}
	
	if (TYPE(arg2) != (argType = TYPE(arg1)))
	{
		sprintf(OUTSTR,"ERROR: replacement elements of wrong type");
		goto yyerrorMsg;
	}
	arg1tot = symbolSize(arg1);
	arg2tot = symbolSize(arg2);

	for (i = 0; i < ncomps; i++)
	{
		subscrSymh = COMPVALUE(list, i);
		if (subscrSymh == (Symbolhandle) 0)
		{						/* blank subscript */
			continue;
		}
		else if(!isDefined(subscrSymh))
		{
			sprintf(OUTSTR,"ERROR: undefined subscript");
		}
		else if ((subscrType = TYPE(subscrSymh)) != LOGIC && subscrType != REAL)
		{
			sprintf(OUTSTR,"ERROR: subscript not REAL or LOGICAL");
		}
		else if(anyMissing(subscrSymh))
		{
			sprintf(OUTSTR,"ERROR: MISSING values not allowed as subscripts");
		}
		else if (!isVector(subscrSymh))
		{
			if(ncomps > 1)
			{
				sprintf(OUTSTR,"ERROR: array used as marginal subscript");
			}
			else if(subscrType != REAL)
			{
				sprintf(OUTSTR,"ERROR: matrix of subscripts must be REAL");
			}
			else if(NDIMS(subscrSymh) != 2 || DIMVAL(subscrSymh,2) != ndims)
			{
				sprintf(OUTSTR,
						"ERROR: matrix of subscripts has wrong dimensions");
			}
			else if(arg2tot != DIMVAL(subscrSymh,1) && arg2tot > 1)
			{
				sprintf(OUTSTR,
						"ERROR: number of rows of matrix of subscripts does not match number of values");
			}
		}						/* if */
		if(*OUTSTR)
		{
			goto yyerrorMsg;
		}
	}/* for i */

	if((trash = GarbInstall(NTRASH)) == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	matrixSubscr = (ncomps == 1 && !isVector(subscrSymh));

	if(matrixSubscr)
	{							/* matrix of subscripts */
		k = 0;
		/* check subscripts for validity */
		nitems = DIMVAL(subscrSymh,1);
		for(j=0;j<ndims;j++)
		{
			subLimit = DIMVAL(arg1,j+1);
			for(i=0;i<nitems;i++)
			{
				y = DATAVALUE(subscrSymh,k++);
				if(y <= 0 || y > subLimit)
				{
					sprintf(OUTSTR,"ERROR: subscript out of range");
				}
				else if(y != floor(y))
				{
					sprintf(OUTSTR,"ERROR: non-integral subscript");
				}
				if(*OUTSTR)
				{
					goto yyerrorMsg;
				}
			} /* for i*/
		} /* for j */

		if(argType == CHAR && !getScratch(itemlist,GITEMLIST,nitems,long))
		{
			goto errorExit;
		}
		
		resdim[0] = 1;
		for(j=1;j<ndims;j++)
		{
			resdim[j] = DIMVAL(arg1,j)*resdim[j-1];
		}

		result = (argType == CHAR) ?
			Install(SCRATCH,CHAR) : RInstall(SCRATCH,nitems);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setTYPE(result,argType);
	} /*if(matrixSubscr)*/
	else
	{ /* list of subscripts */
		if(argType == REAL && NCLASS(arg1) > 0)
		{	/* destination is factor; check values for validity */
			nclass = (double) NCLASS(arg1);
			for(i=0;i<arg2tot;i++)
			{
				y = DATAVALUE(arg2,i);
				if(!isMissing(y) && (y != floor(y) || y < 1.0 || y > nclass))
				{
					sprintf(OUTSTR,
							"ERROR: attempt to assign non-integer or out of range values to a factor");
					goto yyerrorMsg;
				}
			}
		}
		factor = nitems = 1;
		for(i=0;i<ncomps;i++)
		{
			subscrSymh = COMPVALUE(list,i);
			maxSubscr = (ncomps > 1) ? DIMVAL(arg1,i+1) : symbolSize(arg1);
			if(maxSubscr == 0)
			{
				maxSubscr = 1;
			}
			resdim[i] = subscriptLength(COMPVALUE(list,i),maxSubscr);
			needed += resdim[i];
			if(resdim[i] == 0)
			{
				goto errorExit;
			}
		} /*for(i=0;i<ncomps;i++)*/

		if(!getScratch(allsubs,GALLSUBS,needed,long))
		{
			goto errorExit;
		}

		subValues[0] = *allsubs;
		for(i=0;i<ncomps;i++)
		{
			subscrSymh = COMPVALUE(list,i);
			maxSubscr = (ncomps > 1) ? DIMVAL(arg1,i+1) : symbolSize(arg1);
			if(maxSubscr == 0)
			{
				maxSubscr = 1;
			}
			resdim[i] = buildSubscript(COMPVALUE(list, i),maxSubscr,resdim[i],
									   subValues[i]);
			if(resdim[i] == 0)
			{
				goto errorExit;
			}

			if(i > 0)
			{
				for(j=0;j<resdim[i];j++)
				{
					(subValues[i])[j] *= factor;
				}
			}
			
			factor *= maxSubscr;
			nitems *= resdim[i];
			subValues[i+1] = subValues[i] + resdim[i];
		} /* for(i=0;i<ncomps;i++) */

		/* nitems is number of elements specified by subscript(s) */
		if(arg2tot > 1 && nitems != arg2tot)
		{
			sprintf(OUTSTR,"ERROR: wrong number of replacement items");
			goto yyerrorMsg;
		}
		if(argType == CHAR)
		{
			if(!getScratch(itemlist,GITEMLIST,nitems,long))
			{
				goto errorExit;
			}
			result = Install(SCRATCH,CHAR);
		}
		else
		{
			result = RInstall(SCRATCH,nitems);
		}
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setTYPE(result,argType);
	}

/*
  dest is symbol for the subscripted symbol
  If arg1 and arg2 are the same and are REAL or LOGICAL,
  then to ensure arg2 is not corrupted, arg2 is copied to dest and is removed
  only in the case of success
*/
	sameSymbol = (arg1 == arg2) && argType != CHAR;

	if(!sameSymbol)
	{
		dest = arg1;
	}
	else
	{/* allows, e.g., a[run(length(a),1)] <- a to work correctly */
		dest = Install(namearg1,REAL);
		if(dest == (Symbolhandle) 0 || !Copy(arg1,dest))
		{
			goto errorExit;
		}
		Cutsymbol(arg1); /* move arg1 to Scratch symbol table */
		setNAME(arg1,SCRATCH);
		Addsymbol(arg1);
	}

/* For REAL or LOGIC insert values; for CHAR fill (*itemlist) */
	if(matrixSubscr)
	{							/* matrix of subscripts */
		i2 = 0;
		i2inc = (arg2tot == 1) ? 0 : 1;
		for(i=0;i<nitems;i++)
		{
			subscript = DATAPTR(subscrSymh) + i;
			position = (long) *subscript - 1;
			for(j = 1;j< ndims;j++)
			{
				subscript += nitems;
				position += resdim[j]*((long) *subscript - 1);
			}
			if(argType == CHAR)
			{
				(*itemlist)[i] = position;
			}
			else
			{
				DATAVALUE(result,i) = DATAVALUE(dest,position) = DATAVALUE(arg2,i2);
				i2 += i2inc;
			}
		} /* for i */
	} /*if(matrixSubscr)*/
	else
	{ /* list of subscripts */
		lTail = -1;
		i2 = 0;
		i2inc = (arg2tot == 1) ? 0 : 1;
		for(i=0;i<nitems;i++)
		{
			l = (subValues[0])[ii[0]];
			if(lTail < 0)
			{
				lTail = 0;
				for(j = 1;j<ncomps;j++)
				{
					lTail += (subValues[j])[ii[j]];
				}
			}
			l += lTail;
			if(argType == CHAR)
			{
				(*itemlist)[i] = l;
			}
			else
			{
				DATAVALUE(result,i) = DATAVALUE(dest,l) = DATAVALUE(arg2,i2);
				i2 += i2inc;
			}
			
			for(j=0;j<ncomps;j++)
			{ /* turn over odometer */
				if(++ii[j] < resdim[j])
				{
					break;
				}
				ii[j] = 0;
				lTail = -1;
			}
		} /* for i */
	}

	if(argType == CHAR)
	{
/*
  At this point *itemlist contains all the offsets in the destination
*/
		/* compute space needed and allocate*/
		space = 0;
		ch1 = STRINGPTR(arg1);
		jLast = -1;
		resSpace = 0;
		for (i = 0; i < arg1tot; i++)
		{ /* compute space needed */
			for(j = 0;j < nitems; j++)
			{ /* see if string i is to be replaced */
				if((*itemlist)[j] == i)
				{
					break;
				}
			}
			if(j == nitems)
			{ /* string i not to be replaced */
				length = strlen(ch1) + 1;
				ch1 += length;
			}
			else
			{ /* string i will be replaced */
				ch1 = skipStrings(ch1,1L);
				if(arg2tot == 1)
				{
					j = 0;
				}
				if(j < jLast || jLast < 0)
				{
					ch2 = STRINGPTR(arg2);
					jLast = 0;
				}
				
				ch2 = skipStrings(ch2,j-jLast);
				length = strlen(ch2) + 1;
				resSpace += length;
				jLast = j;
			}
			space += length;
		} /* for i */

		if(!getScratch(chNewArg1,GCHNEWARG1,space,char) ||
		   !getScratch(chResult,GCHRESULT,resSpace,char))
		{
			goto errorExit;
		}

		resSpace = space = 0;
		ch1 = STRINGPTR(arg1);
		jLast = -1;
		for (i = 0; i < arg1tot; i++)
		{ /* now fill it up */
			for(j = 0;j < nitems; j++)
			{
				if((*itemlist)[j] == i)
				{
					break;
				}
			}
			if(j == nitems)
			{ /* string i not to be replaced */
				strcpy(*chNewArg1 + space,ch1);
				length = strlen(ch1) + 1;
				ch1 += length;
			}
			else
			{ /* string i to be replaced */
				if(arg2tot == 1)
				{
					j = 0;
				}
				
				ch1 = skipStrings(ch1,1L); /* skip string being replaced */
				if(j < jLast || jLast < 0)
				{
					ch2 = STRINGPTR(arg2);
					jLast = 0;
				}
				
				ch2 = skipStrings(ch2,j-jLast);
				strcpy(*chNewArg1 + space, ch2);
				strcpy(*chResult + resSpace, ch2);
				length = strlen(ch2) + 1;
				resSpace += length;
				jLast = j;
			}
			space += length;
		} /* for i */
		mydisphandle((char **) STRING(dest));
		unTrash(GCHNEWARG1);
		setSTRING(dest,chNewArg1);
		unTrash(GCHRESULT);
		setSTRING(result,chResult);
	} /*if(argType == CHAR)*/
	
	if(!matrixSubscr)
	{
		setNDIMS(result,ncomps);
		for(j=0;j<ncomps;j++)
		{
			setDIM(result,j+1,resdim[j]);
		}
	}
	else
	{
		setNDIMS(result,1);
		setDIM(result,1, nitems);
	}
	
	if(sameSymbol)
	{
		Removesymbol(arg1);
	}
	else if(isscratch(NAME(arg2)))
	{
		Removesymbol(arg2);
	}
	
	emptyTrash();

	return (result);

  yyerrorMsg:
	yyerror(OUTSTR);
	/* fall through */
  errorExit:
	putOUTSTR();

	if(sameSymbol && dest != (Symbolhandle) 0)
	{
		Removesymbol(dest);
		Cutsymbol(arg1); /* preserve arg1 */
		setNAME(arg1,namearg1);
		Addsymbol(arg1);
	}
	Removesymbol(result);
	emptyTrash();
	return (0);
}


