/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#include "globals.h"

#ifdef MPW
#ifdef MPW3
#pragma segment Symbol
#else
#define __SEG__ Symbol
#endif /*MPW3*/
#endif /*MPW*/

extern Char   **mygethandle();
extern Char   **mygrowhandle();
extern void     mydisphandle();
extern Char    *isKeyword();
extern Char    *typeName();
extern long     isNull();

/* list of entry points */
Symbolhandle    Lookup();  /* search symbol table */
Symbolhandle    Makesymbol(); /* allocate storage for new symbol */
Symbolhandle    Makestruc(); /* allocate storage for STRUC */
Symbolhandle    Makereal(); /* allocate storage for REAL vector */
Symbolhandle    Makelong(); /* allocate storage for LONG scratch vector */

void            Addsymbol(); /* add symbol to symbol table */
void            Cutsymbol(); /* remove symbol from symbol table */

Symbolhandle    Install(); /* make symbol and add it to table */
Symbolhandle    RInstall(); /* make REAL symbol and add it to table */
Symbolhandle    LongInstall(); /* make LONG symbol and add it to table */
Symbolhandle    GarbInstall(); /* make GARB symbol and add it to table */
Symbolhandle    GraphInstall(); /* make PLOTINFO symbol and add it to table */
Symbolhandle    StrucInstall(); /* make STRUC symbol and add it to table */
Symbolhandle    RSInstall(); /* make STRUC w/REAL comps and add it to table */
Symbolhandle    RSInstall2();
Symbolhandle    RSInstall3();

void            Remove(); /* remove named symbol from table and delete it */
void            Removesymbol(); /* remove symbold from table and delete it */
void            Delete(); /* deallocate storage associated with symbol */
void            Removelist(); /* remove list from table after cleaning up */
void            Unscratch(); /* Remove all SCRATCH and keyword variables */
void            RemoveKeywd(); /* Remove all keyword variables */
void            RemoveUndef(); /* remove all variables with TYPE == UNDEF */
void            CleanupList(); /* remove SCRATCH & keyword vars from list */

Long            Copy(); /* duplicate symbol */

Symbolhandle    Assign(); /* assignment & subscript assignment */
Symbolhandle    Extract(); /* extract component of structure */
Symbolhandle    Makelist();
Symbolhandle    Addlist();
Symbolhandle    Growlist();
Symbolhandle    Byname();
void            Dumpsymbols(); /* for debugging */

/* start of linked list of scratch, keyword, and undefined symbols */
Symbolhandle    Scratchstart = (Symbolhandle) 0;

Long            Nscratch = 0, Maxscratch = 0;
Long            Nsymbols = 0, Maxsymbols = 0;
Long            Nfunctions = 0; /* number of installed functions */

#define inScratchTable(symh, name) \
	(name = NAME(symh),\
	 (name[0] == TEMPPREFIX && \
	  (name[1] == KEYPREFIX2 || name[1] == SCRATCHPREFIX2))\
	 || TYPE(symh) == UNDEF)

#define TABLESIZE     512
#define TABLEMASK     511

static Symbolhandle  *Symboltable;
static Long           Lastplace = -1;
/*
  Compute hash code = sum of ascii codes in name mod TABLESIZE
*/
static int  symbolHash(Char * name)
{
	register int     code = 0;
	register Char   *pc = name;
	
	while(*pc != '\0')
	{
		code += *pc++;
	}
	return (code & TABLEMASK); /* code % TABLESIZE */
}

/*
  Allocate and initialize storage for not-scratch symbol table.  Each entry
  is a starting node for a doubly linked list of symbols with the same
  hash code.  Hence the symbol table is not limited in size.
*/

Long Symbolinit()
{
	int        i;

	Symboltable = (Symbolhandle *) mygetpointer((TABLESIZE+1)*sizeof(Symbolhandle));
	if(Symboltable == (Symbolhandle *) 0)
	{
		return (0);
	}
	for(i = 0;i<TABLESIZE;i++)
	{
		Symboltable[i] = (Symbolhandle) 0;
	}
	Symboltable[TABLESIZE] = ENDPREV; /* non-zero to stop searches */

	Scratchstart = (Symbolhandle) 0; /* just for good measure */

	return (1);
}

/* find start of scratch chain or chain nearest beginning of Symboltable[] */
Symbolhandle Firstsymbol(scratch)
int             scratch;
{
	register Symbolhandle *psymh;
	WHERE("Firstsymbol");

	if(scratch)
	{
		return Scratchstart;
	}

	for(psymh = Symboltable; *psymh == (Symbolhandle) 0;psymh++)
	{
		;
	}
	Lastplace = psymh - Symboltable;
/*
  Firstsymbol should never be called when table is empty; hence don't test
  for validity of psymh
*/
	return (*psymh);
}

/*
  Find the end of the chain starting nearest the end of Symboltable[]
*/
Symbolhandle Lastsymbol(void)
{
	register Symbolhandle *psymh;
	register Symbolhandle  last;
	
	for(psymh = Symboltable + TABLESIZE-1;*psymh == (Symbolhandle) 0;psymh--)
	{/* because of functions, a non-zero *psymh will always be found */
		;
	}

	for(last = *psymh;NEXT(last) != (Symbolhandle) 0; last = NEXT(last))
	{ /* find end of list pointed to by psymh */
		;
	}

	return (last);
}

/*
  Used only when traversing the symbol table.  If it is used otherwise, it
   would need to be changed because it is remembering the last position in
   Symboltable examined.  It's most important use is by RemoveUndef().
*/

Symbolhandle Nextsymbol(symh,scratch)
Symbolhandle    symh;
int             scratch;
{
	register Symbolhandle     next = NEXT(symh);
	register Symbolhandle    *place;
	WHERE("Nextsymbol");
	
	if(next == (Symbolhandle) 0 && !scratch)
	{ /* symh is at end of chain */
		/* proceed along Symboltable looking for a chain */
		if(Lastplace < 0)
		{ /* just starting, should not normally occur */
			Lastplace = symbolHash(NAME(symh)) + 1;
		}
		else
		{
			Lastplace++;
		}
			
		for(place = Symboltable + Lastplace;*place == (Symbolhandle) 0;place++)
		{ /* guaranteed to stop because Symboltable[TABLESIZE] is ENDPREV */
			;
		}
		if((Lastplace = place - Symboltable) < TABLESIZE)
		{ /* found a symbol */
			next = *place;
		}
		else
		{ /* run out of symbols */
			next = (Symbolhandle) 0;
			Lastplace = -1;
		}
	}
	
	return (next);
}

/*
  Used by save() to traverse symbol table backward from end of Symboltable[]
*/

Symbolhandle Prevsymbol(symh)
Symbolhandle    symh;
{
	register Symbolhandle     prev = PREV(symh);
	register int              place;
	int                       scratch;
	Char                     *name;

	if(prev == ENDPREV)
	{ /* most recent added to chain */
		scratch = inScratchTable(symh,name);
		if(!scratch)
		{ /* Backdown Symboltable looking for a chain */
			for(place = symbolHash(name) - 1;
				place >= 0 && Symboltable[place] == (Symbolhandle) 0;
				place--)
			{
				;
			}
			if(place >= 0)
			{ /* found one; find the first element added */
				for(prev = Symboltable[place];NEXT(prev) != (Symbolhandle) 0;
					prev = NEXT(prev))
				{
					;
				}
			}
		}		
	}
	return (prev);
}

/* 
  Function to add symbol to symbol table.  If the symbol is scratch, UNDEF,
  or a keyword, it is added to the beginning of the scratch symbol table;
  otherwisee it is always added at the beginning of the chain starting
  at Symboltable[symbolHash(NAME(symh))] in the regular symbol table.
  Note that the table to which it is added depends on the name.  Thus the
  name should always be set *before* adding a symbol.  This is particularly
  important after a call to Copy(), since Copy() copies the name.

  No storage is allocated.  Do nothing if symh is null.
*/

void    Addsymbol(symh)
Symbolhandle    symh;
{
	Char         *name;
	int           scratch, place;
	Symbolhandle  start;
	WHERE("Addsymbol");
	
	if(symh != (Symbolhandle) 0)
	{
		if(NEXT(symh) == (Symbolhandle) 0 && PREV(symh) == (Symbolhandle) 0)
		{						/* add only if symh is not in symbol table */
			scratch = inScratchTable(symh,name);
			start = (scratch) ?
				Firstsymbol(1) : Symboltable[place = symbolHash(name)];
			if (start != (Symbolhandle) 0)
			{					/* another symbol is in list */
				setPREV(start,symh);
			}
		
			setNEXT(symh,start);
			setPREV(symh,ENDPREV);
			if(scratch)
			{
				Scratchstart = symh;
				Nscratch++;
				Maxscratch = (Maxscratch >= Nscratch) ? Maxscratch : Nscratch;
			}
			else
			{
				Symboltable[place] = symh;
				Nsymbols++;
				Maxsymbols = (Maxsymbols >= Nsymbols) ? Maxsymbols : Nsymbols;
				if(TYPE(symh) == BLTIN)
				{
					Nfunctions++;
				}
			}
		}
		else
		{						/* should not happen */
			myerrorout("ERROR: (Internal) attempt to add bad symbol");
		}
	}
}

/*
  Function to cut symbol out of symbol table; does not release any storage
  Do nothing if symh is null.
*/

void Cutsymbol(symh)
Symbolhandle       symh;
{
	Symbolhandle        next, prev;
	int                 scratch;
	Char               *name;
	WHERE("Cutsymbol");
	
/*
  Note: NEXT(symh) should be zero only if symh was the first symbol put
  in its list.  PREV(symh) should be a valid address, except when symh was
  the most recently added symbol to its list.  If symh was the most recent
  symbol, PREV(symh) should be ENDPREV.
*/
	if(symh != (Symbolhandle) 0)
	{
		prev = PREV(symh);
		next = NEXT(symh);
		if(next != (Symbolhandle) 0 || prev != (Symbolhandle) 0)
		{						/* symbol actually in a symbol table */
			scratch = inScratchTable(symh,name);
			if (next != (Symbolhandle) 0) /* if there is a following symbol */
			{
				setPREV(next,prev);
			}
		
			if (prev != ENDPREV)
			{					/* if there is a preceding symbol */
				setNEXT(prev,next);
			}
			else if(scratch)
			{
				Scratchstart = next;
			}
			else
			{
				Symboltable[symbolHash(name)] = next;
			}
			setPREV(symh,(Symbolhandle) 0);
			setNEXT(symh,(Symbolhandle) 0);
			if(scratch)
			{
				Nscratch--;
			}
			else
			{
				Nsymbols--;
			}
		} /*if(next != (Symbolhandle) 0 || prev != (Symbolhandle) 0)*/
	} /*if(symh != (Symbolhandle) 0)*/	
}

void Setname(symh,name)
Symbolhandle    symh;
char           *name;
{
	strncpy(NAME(symh),name,NAMELENGTH);
	NAME(symh)[NAMELENGTH] = '\0';
}

/*
  Search the table for fpr first occurence of name symbolName
*/
Symbolhandle    Lookup(symbolName)
Char           *symbolName;
{
	int                      place = symbolHash(symbolName);
	register Symbolhandle    symh = (Symbolhandle) 0;
	register Char            firstchar = symbolName[0], *name;
	WHERE("Lookup");
	
#ifdef MACINTOSH
	if(firstchar == 'C' && strcmp(symbolName,"CLIPBOARD") == 0)
	{
		LongInt     length;
		Handle      hDest = (Handle) 0;
		Char      **contents = (char **) 0;
		LongInt     offset;

		symh = Install(SCRATCH,CHAR);
		if(symh != (Symbolhandle) 0 && (hDest = NewHandle(0)) != (Handle) 0)
		{
			length = GetScrap (hDest, 'TEXT', &offset);
			if(length < 0)
			{
				length = 0;
			}
			if((contents = mygethandle(length + 1)) != (char **) 0)
			{
				setSTRING(symh,contents);
				HLock(hDest);
				if(length > 0)
				{
					BlockMove(*hDest, *contents, length);
				}
				(*contents)[length] = '\0';
				setNDIMS(symh,1);
				setDIM(symh,1,1);
			}
		}
		if(hDest != (Handle) 0)
		{
			DisposHandle(hDest);
		}

		if(hDest == (Handle) 0 || contents == (char **) 0)
		{
			Removesymbol(symh);
			symh = (Symbolhandle) 0;
		}
		return (symh);
	} /*if(strcmp(symbolName,"CLIPBOARD") == 0)*/
#endif /*MACINTOSH*/
	/* search regular symbol table only */
	for (symh = Symboltable[place]; symh != (Symbolhandle) 0;
		 symh = NEXT(symh))
	{
		if (firstchar == *(name = NAME(symh)) && strcmp(name, symbolName) == 0)
		{
			break;
		}
	}

	return (symh);
}

/*
  routine to create a new Symbol, not yet in symbol table
*/

Symbolhandle    Makesymbol(type)
int             type;
{
	Symbolhandle    symh;
	WHERE("Makesymbol");
	
	if(type != BLTIN)
	{
		symh = (Symbolhandle) mygethandle((Long) sizeof(Symbol));
	}
	else
	{ /* FunctionSymbol is smaller */
		symh = (Symbolhandle) mygethandle((Long) sizeof(FunctionSymbol));
	}
	
	if (symh != (Symbolhandle) 0)
	{
		setTYPE(symh,type);
		setNCLASS(symh,-1);		/* a variate by default */
#ifndef USEMEMSET
		if(type == REAL || type == LOGIC)
		{
			setDATA(symh,(Double **) 0);
		}
		else if(type == CHAR || type == MACRO)
		{
			setSTRING(symh,(Char **) 0);
		}
		else if(type == STRUC || type == LIST)
		{
			setCOMP(symh,(Symbolhandle**) 0);
		}
		else if(type == GARB)
		{
			setGARBAGE(symh,(Char ****) 0);
		}
		else if(type == BLTIN)
		{
			setFPTR(symh,(Symbol **(*) ()) 0);
		}
	
		setNDIMS(symh,0);
		if(type != BLTIN)
		{
			short      i;
			for(i=1;i<=MAXDIMS;i++)
			{
				setDIM(symh,i,0);
			}
		}
		else
		{
			setDIM(symh,1,0);
		}
#else							/*USEMEMSET*/
		setDATA(symh,(Double **) 0);
		memset((void *) DIM(symh) , 0,
			   ((type == BLTIN) ? 2 :(MAXDIMS+1))*sizeof(Long));
#endif							/*USEMEMSET*/
		setNAME(symh,SCRATCH);
		setPREV(symh,(Symbolhandle) 0);
		setNEXT(symh,(Symbolhandle) 0);
	}
	return (symh);
}

/*
  Make a structure with ncomp components, initializing all component
  handles to 0 and not installing in symbol table
*/

Symbolhandle    Makestruc(ncomp)
Long            ncomp;
{
	/* make a structure with ncomp components, not in symbol table */
	Symbolhandle       result;
	Long               i;
	WHERE("Makestruc");
	
	result = Makesymbol(STRUC);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	TMPHANDLE = mygethandle(ncomp * sizeof(Symbolhandle));
	if (TMPHANDLE == (Char **) 0)
	{
		goto errorExit;
	}
	setCOMP(result,(Symbolhandle**) TMPHANDLE);

	for (i = 0; i < ncomp; i++)
	{
		COMPVALUE(result,i) = (Symbolhandle) 0;
	}
	setNDIMS(result,1);
	setNCOMPS(result,ncomp);

	return (result);

  errorExit:
	Delete(result);
	return (0);

}

/*
  Note: Makereal(0) differs from Makesymbol(REAL) in setting NDIMS(result) 
  to 1
*/

Symbolhandle    Makereal(length)
Long            length;	
{
	/* make real vector, not in symbol table */
	Symbolhandle    result;
	WHERE("Makereal");
	
	result = Makesymbol(REAL);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	setNDIMS(result,1);
	setDIM(result,1,length);
	if (length > 0)
	{
		TMPHANDLE = mygethandle(length * sizeof(Double));
		if (TMPHANDLE == (Char **) 0)
		{
			goto errorExit;
		}
		setDATA(result,(Double **) TMPHANDLE);
	}
	return (result);

  errorExit:
	Delete (result);
	return (0);
}

Symbolhandle    Makelong(length)
Long            length;	
{
	/* make Long vector, not in symbol table */
	Symbolhandle    result;
	WHERE("Makelong");
	
	result = Makesymbol(LONG);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	setNDIMS(result,1);
	setDIM(result,1,length);
	if (length > 0)
	{
		TMPHANDLE = mygethandle(length * sizeof(Long));
		if (TMPHANDLE == (Char **) 0)
		{
			goto errorExit;
		}
		setLONGDATA(result,(Long **) TMPHANDLE);
	}
	return (result);

  errorExit:
	Delete (result);
	return (0);
}

Symbolhandle    Install(name, type)
Char           *name;
int             type;
{
	/* routine to put new symbol in table */

	Symbolhandle    symh;
	Char            tempname[NAMELENGTH+1];
	WHERE("Install");

	strncpy(tempname,name,NAMELENGTH);
	tempname[NAMELENGTH] = '\0';
	symh = Makesymbol(type);
	if(symh != (Symbolhandle) 0)
	{
		setNAME(symh, tempname);
		Addsymbol(symh);
	}
	return (symh);
}


Symbolhandle    RInstall(name, length)
Char           *name;
Long            length;
{
	/* install a REAL symbol with length units available */
	/* assums vector structure */

	Symbolhandle    symh;
	WHERE("RInstall");

	symh = Makereal(length);
	if (symh != (Symbolhandle) 0)
	{
		setNAME(symh,name);
		Addsymbol(symh);
	}	
	return (symh);
}

Symbolhandle    GraphInstall(name)
Char           *name;
{
	/* install a REAL symbol with length units available */
	/* assums vector structure */

	Symbolhandle            symh;
	whole_graph           **graph = (whole_graph **) 0;
	WHERE("GraphInstall");

	symh = Makesymbol(PLOTINFO);
	if (symh == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	setNAME(symh,name);
	Addsymbol(symh);
	graph = (whole_graph **) mygethandle(sizeof(whole_graph));
	if(graph == (whole_graph **) 0)
	{
		goto errorExit;
	}
	(*graph)->nplots = 0;
	(*graph)->first_plot = (curve_points **) 0;
	(*graph)->nstrings = 0;
	(*graph)->first_string = (plot_string **) 0;
	setGRAPH(symh,graph);
	setNDIMS(symh,1);
	setDIM(symh,1,1);

	return (symh);

  errorExit:
	Removesymbol(symh);
	return ((Symbolhandle) 0);
}

Symbolhandle    LongInstall(name, length)
Char           *name;
Long            length;
{
	/* install a LONG symbol with length units available */
	/* assumes vector structure */

	Symbolhandle    symh;
	WHERE("LongInstall");

	symh = Makelong(length);
	if (symh != (Symbolhandle) 0)
	{
		setNAME(symh,name);
		Addsymbol(symh);
	}	
	return (symh);
}

Symbolhandle    GarbInstall(length)
Long            length;
{
	/* install a GARB symbol with room for length Char **'s  */

	Symbolhandle    symh;
	Long            i;
	WHERE("GarbInstall");

	symh = Install(GARBSCRATCH,GARB);
	if (symh == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	setNDIMS(symh,1);
	setDIM(symh,1,length);

	if (length > 0)
	{
		TMPHANDLE = mygethandle(length * sizeof(Char **));
		setGARBAGE(symh,(Char ****) TMPHANDLE);
		if (TMPHANDLE == (Char **) 0)
		{
			goto errorExit;
		}
		for(i=0;i<length;i++)
		{
			GARBAGEVALUE(symh,i) = (Char **) 0;
		}
	}
	
	return (symh);

  errorExit:
	Removesymbol (symh);
	return (0);
}

Symbolhandle StrucInstall(name, ncomp)
Char       *name;
Long        ncomp;
{
	Symbolhandle      result;
	
	result = Makestruc(ncomp);
	if(result != (Symbolhandle) 0)
	{
		setNAME(result,name);
		Addsymbol(result);
	}
	return (result);
}

Symbolhandle    RSInstall(name, ncomp, compNames, length)
Char           *name, *compNames[];
Long            ncomp, length;
{
	/*  install a REAL structure with ncomp components
		component names in compNames, an array of character pointers,
		real vectors all with length length  (if zero, no assignment ) */

	Symbolhandle    result, symh;
	Long            i;

	result = StrucInstall(name,ncomp);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	for (i = 0; i < ncomp; i++)
	{
		symh = COMPVALUE(result, i) = Makereal(length);
		if (symh == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		if(compNames != (char **) 0)
		{
			setNAME(symh, compNames[i]);
		}
	}
	return (result);

  errorExit:
	Removesymbol (result);
	return (0);
	
}

Symbolhandle    RSInstall2(name, ncomp, compNames, length)
Char           *name, **compNames;
Long            ncomp, length;
{
	/*  install a REAL structure with ncomp components
		component names in compNames, a handle to CHAR data.
		real vectors all with length length  (if zero, no assignment ) */

	Symbolhandle    result, symh;
	Long            i, k;

	result = StrucInstall(name,ncomp);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	k = 0;
	for (i = 0; i < ncomp; i++)
	{
		symh = COMPVALUE(result, i) = Makereal(length);
		if (symh == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNAME(symh, *compNames + k);
		k += strlen(*compNames + k) + 1;
	}

	return (result);

  errorExit:
	Removesymbol(result);
	return (0);
	
}


Symbolhandle    RSInstall3(name, ncomp, compNames, ncomp2, compNames2, length)
Char           *name, **compNames, *compNames2[];
Long            ncomp, ncomp2, length;
{
	/*  install a structure with ncomp components
		each of which is a structure with ncomp2 REAL components
		component names in compNames
		real vectors all with length length  (if zero, no assignment ) */

	Symbolhandle    result, symh;
	Long            i, k;
	WHERE("RSInstall3");

	result = StrucInstall(name,ncomp);

	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	k = 0;
	for (i = 0; i < ncomp; i++)
	{
		symh = RSInstall(SCRATCH,ncomp2,compNames2,length);
		if(symh == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		Cutsymbol(symh); /* remove from symbol table */
		COMPVALUE(result,i) = symh;
		if(compNames != (char **) 0)
		{
			setNAME(symh, (*compNames) + k);
			k += strlen((*compNames) + k) + 1;
		}
	}
	return (result);

  errorExit:
	Removesymbol(result);
	return (0);
}


/*
  As of 10/9/92, Remove() is used only by glmcleanup() in glm.c and in
  boxplot.c and myplot.c
*/

void Remove(name)
Char           *name;
{				/* find entry in symbol table and remove it */
	Removesymbol(Lookup(name));
}

/* remove symbol from symbol table */
void Removesymbol(symh)
Symbolhandle      symh;
{
	if (symh != (Symbolhandle) 0 && TYPE(symh) != BLTIN &&
		symh != NULLSYMBOL)
	{							/* don't remove built-in functions */
		/* fix up table pointers */
		Cutsymbol(symh);
		/* free up storage (works recursively for structures) */
		Delete(symh);
	}
}

/* clean up argument list and remove it */

void Removelist(list)
Symbolhandle        list;
{
	WHERE("Removelist");

	CleanupList(list); /* delete SCRATCH and keyword entries  */
	Removesymbol(list);
}

/* 
  Get rid of a symbol handle but does not remove from symbol table if there;
  hence Cutsymbol() should always be called before Delete() if symh is in the
  symbol table.  In fact, that is what Removesymbol() does.
  Note that NULLSYMBOL cannot be deleted
 */
void Delete(symh)
Symbolhandle    symh;
{
	WHERE("Delete");

	if (symh != (Symbolhandle) 0 && TYPE(symh) != BLTIN &&
		 symh != NULLSYMBOL)
	{
		DeleteContents(symh);
	
		if(NCLASS(symh) != -32768)
		{						/* not fake Symbol handle */
			mydisphandle((Char **) symh);
		}
	}	
}

/*
  get rid of a contents of symbol but do not delete symbol
  The name is changed to SCRATCH and the type is changed to UNDEF; hence
  symh should be removed from teh symbol table before calling DeleteContents
*/
void DeleteContents(symh)
Symbolhandle     symh;
{
	Symbolhandle  **components;
	Char        ****trash;
	Long            i, ncomps;
	int             type;
	WHERE("DeleteContents");

	if (symh != (Symbolhandle) 0 && (type = TYPE(symh)) != BLTIN)
	{
		switch (type)
		{
		  case LOGIC:
		  case REAL:
			if (DATA(symh) != (Double **) 0)
			{
				mydisphandle((Char **) DATA(symh));
				setDATA(symh,(Double **) 0);
			}
			break;

		  case LONG:
			if (LONGDATA(symh) != (Long **) 0)
			{
				mydisphandle((Char **) LONGDATA(symh));
				setLONGDATA(symh,(Long **) 0);
			}
			break;

		  case LIST:
			/* 
			  DeleteContents(list) should be called only following
			  CleanupList(list).  This is done automatically if list is
			  removed by Removelist().

			  CleanupList removes scratch symbols, keyword symbols, and
			  undefined symbols that are in the list but ignores other
			  symbols.  All DeleteContents does is dispose of the list handle
			  itself.  This can be contrasted with what happens with STRUCs,
			  where all components are disposed of. 
			  */
			if (COMP(symh) != (Symbolhandle**) 0)
			{
				mydisphandle((Char **) COMP(symh));
				setCOMP(symh,(Symbolhandle **) 0);
			}

			break;

		  case GARB:
			if (GARBAGE(symh) != (Char ****) 0)
			{
				ncomps = DIMVAL(symh,1);
				trash = GARBAGE(symh);
				setGARBAGE(symh,(Char ****) 0);
				for (i = 0; i < ncomps; i++)
				{
					mydisphandle((*trash)[i]);
				}				/* delete all handles */
				mydisphandle((Char **) trash);
			}
			break;
			
		  case CHAR:
		  case MACRO:
		  case MACRO1:
			if (STRING(symh) != (Char **) 0)
			{
				mydisphandle((Char **) STRING(symh));
				setSTRING(symh,(Char **) 0);
			}
			break;

		  case STRUC:
			components = COMP(symh);
			setCOMP(symh,(Symbolhandle **) 0);
			if (components != (Symbolhandle **) 0)
			{
				ncomps = NCOMPS(symh);
				for (i = 0; i < ncomps; i++)
				{
					Delete((*components)[i]);
				}				/* delete all subcomponents */
				mydisphandle((Char **) components);
			}
			break;

		  case PLOTINFO:
			if(GRAPH(symh) != (whole_graph **) 0)
			{
				deleteGraph(GRAPH(symh)); /* in plotutil.c */
				setGRAPH(symh,(whole_graph **) 0);
			}
			break;

		  case UNDEF:
		  case ASSIGNED:
			break;

		  default:
			sprintf(OUTSTR,
					"WARNING: (internal) deleting variable %s of unknown type %ld",
					NAME(symh), type);
			putOUTSTR();
		}						/* switch */
		setNCLASS(symh,-1);		/* a variate by default */

#ifndef USEMEMSET
		setNDIMS(symh,0);
		for(i=1;i<=MAXDIMS;i++)
		{
			setDIM(symh,i,0);
		}
#else							/*USEMEMSET*/
		memset((void *) DIM(symh) , 0, (MAXDIMS+1)*sizeof(Long));
#endif							/*USEMEMSET*/
		setTYPE(symh,UNDEF);
		setNAME(symh,SCRATCH);
	}	
}
	
void Unscratch()
{
	/* remove all SCRATCH variables & keyword variables from symbol list */
	Symbolhandle       symh, next;
	WHERE("Unscratch");
	
	for (symh = Scratchstart; symh != (Symbolhandle) 0; symh = next)
	{ /* clear out entire scratch list */
		next = NEXT(symh);
		Removesymbol(symh);
	}
}

/*
  Remove all keyword & SCRATCH temporary variables in argument list
  To be called by mainpars on return from a function, to avoid buildup of
  unused symbols in symbol table during looping
  Note: Components of a LIST are either in the regular symbol table or the
  scratch symbol table.  This is one difference from a STRUC.
*/

void CleanupList(list)
Symbolhandle          list;
{
	Long              i, ncomps;
	Symbolhandle      arg;
	Char             *name;
	WHERE("CleanupList");

	if(list != (Symbolhandle) 0 && TYPE(list) == LIST)
	{
		ncomps = NCOMPS(list);
		for(i=0;i<ncomps;i++)
		{
			if((arg = COMPVALUE(list,i)) != (Symbolhandle) 0 &&
			   inScratchTable(arg,name))
			{/* remove scratch, keyword, or undefined symbols */
				Removesymbol(arg);
				COMPVALUE(list,i) = (Symbolhandle) 0;
			}			
		}
	}
}

			   
/*
  remove all typed ASSIGNED and temporary variables from regular symbol table
  Called only after Unscratch() which should clear out scratch symbol table
*/

void RemoveUndef()
{
	Symbolhandle       symh, next;
	WHERE("RemoveUndef");

	for (symh = Firstsymbol(0); symh != (Symbolhandle) 0; symh = next)
	{
		next = Nextsymbol(symh,0);
		if(istempname(NAME(symh)) || TYPE(symh) == ASSIGNED)
		{ /* either undefined or a temporary variable */
			Removesymbol(symh);
		}
	}
}

int             Nassigned = 0; /* global for communicating with cleanitup */

Symbolhandle    Assign(arg1, arg2, keyword)
Symbolhandle    arg1, arg2;
Long            keyword;
{
	/* assign contents of arg2 to variable with name as in arg1 */

	Char            name[NAMELENGTH+1];
	Char           *namearg1;
	Symbolhandle    result = (Symbolhandle) 0;
	Char          **temphandle;
	Long            type1, type2;
	Long            i, op;
	Char           *msgs, *what = "";
	WHERE("Assign");
	
	msgs = (keyword == 1) ? "keyword " : "";
	
	OUTSTR[0] = '\0';
	/* check left hand side */
	if(arg1 == (Symbolhandle) 0)
	{
		sprintf(OUTSTR,"ERROR: apparently out of space during %sassign",msgs);
	}
	else if ((type1 = TYPE(arg1), keyword) != 1 && type1 == BLTIN)
	{
		sprintf(OUTSTR,"ERROR: do not assign to builtin function names");
	}
	else if ((namearg1 = NAME(arg1) , isscratch(namearg1)) ||
			 arg1 == NULLSYMBOL)
	{
		if(type1 == LOGIC)
		{
			what = "T or F, LOGICAL expression, or function result";
		}
		else if(type1 == CHAR)
		{
			what = "string or function result";
		}
		else if(type1 == STRUC)
		{
			what = "expression or function result";
		}
		else
		{
			what = "number, expression, or function result";
		}
		sprintf(OUTSTR, "ERROR: %sassignment to %s",msgs,what);
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}

	if(keyword < 0)
	{
		if(keyword == -1 && type1 == UNDEF)
		{/* we are processing only l.h.s. of non-keyword assignment*/
			/* move from scratch table to regular table */
			Cutsymbol(arg1);
			setTYPE(arg1,ASSIGNED);
			Addsymbol(arg1);
			Nassigned++;
		}
		else if(keyword < -1 && (type1 == UNDEF || type1 == ASSIGNED))
		{/* processing l.h.s. of <-+, <--, <-*, <-/, <-^, <-% assignement */
			op = -keyword;
			switch (op)
			{
			  case ADD:
				op = ASSIGNADD;
				break;
			  case SUB:
				op = ASSIGNSUB;
				break;
			  case MULT:
				op = ASSIGNMULT;
				break;
			  case DIV:
				op = ASSIGNDIV;
				break;
			  case EXP:
				op = ASSIGNPOW;
				break;
			  case MOD:
				op = ASSIGNMOD;
				break;
			  default:
				op = -1; /* should not happen */
			}
			
			sprintf(OUTSTR,
					"ERROR: UNDEFINED %s ...",opName(op));
			goto errorExit;
		}
		return (arg1);
	} /*if(keyword < 0)*/
	

	/* now check right hand side */

	if (arg2 == (Symbolhandle) 0 || (isDefined(arg2) && DIMVAL(arg2,1) == 0))
	{
		sprintf(OUTSTR,"ERROR: %sassignment from null operation",msgs);
	}
	else if (!isDefined(arg2))
	{
		sprintf(OUTSTR, "ERROR: %sassignment from undefined variable %s",
				msgs,NAME(arg2));
	}
	else if ((type2 = TYPE(arg2)) == LIST || type2 == BLTIN)
	{						/* LIST should be impossible, but play it safe */
		sprintf(OUTSTR,"ERROR: %sassignment from %s ",
				msgs,(type2 == LIST) ? "LIST" : "built-in function");
	}

	if(*OUTSTR)
	{
		goto errorExit;
	}

	if(!keyword)
	{/* must be existing variable, delete contents unless identical to RHS*/
		if(arg1 == arg2)
		{/* assignment to self */
			return (arg1);
		}
		strcpy(name, namearg1);
		if(name[0] == 'S' && strcmp(name,"STRMODEL") == 0)
		{/* changing STRMODEL invalidates all global glm info */
			clearGlobals();
		}
		Cutsymbol(arg1);

		/* clean out arg1, but save for re-use */
		if(type1 == ASSIGNED)
		{
			Nassigned--;
		}
		else
		{
			DeleteContents(arg1);
		}
	} /*if(!keyword)*/
	else
	{/* keyword: create name of form @@name */
		if(strlen(NAME(arg1)) > NAMELENGTH-2)
		{
			sprintf(OUTSTR,
					"ERROR: keyword name '%s' longer than %ld characters",
					NAME(arg1),(long) NAMELENGTH-2);
			goto errorExit;
		}
		else
		{
			name[0] = KEYPREFIX1;
			name[1] = KEYPREFIX2;
			strcpy(name+2,namearg1);
			/* the following should stop the build up of UNDEFs in macros*/
			if(type1 == UNDEF)
			{/* save for reuse */
				Cutsymbol(arg1);
			}
			else
			{/* can't reuse, so forget about arg1 */
				arg1 = (Symbolhandle) 0;
			}
		}
	} /*if(!keyword){}else{}*/

	if(isscratch(NAME(arg2)))
	{/* if from SCRATCH, reuse or move content handles, do not copy */
		Cutsymbol(arg2);
		if(arg1 == (Symbolhandle) 0)
		{/* no l.h.s. to reuse, so reuse r.h.s symbol */
			result = arg2;
		}
		else
		{/* reuse l.h.s. symbol; move content handles, do not Copy() */
			result = arg1;
			setTYPE(result,type2);
			setNCLASS(result,NCLASS(arg2));
			setNDIMS(result,NDIMS(arg2));
			for(i=1;i<=MAXDIMS;i++)
			{
				setDIM(result,i,DIMVAL(arg2,i));
			}

			switch (type2)
			{
			  case CHAR:
			  case MACRO:
				temphandle = STRING(arg2);
				setSTRING(arg2,(Char **) 0);
				setSTRING(result,temphandle);
				break;

			  case REAL:
			  case LOGIC:
				temphandle = (Char **) DATA(arg2);
				setDATA(arg2,(Double **) 0);
				setDATA(result,(Double **) temphandle);
				break;

			  case STRUC:
				temphandle = (Char **) COMP(arg2);
				setCOMP(arg2,(Symbolhandle **) 0);
				setCOMP(result,(Symbolhandle **) temphandle);
				break;

			  default:		/* safety net */
				temphandle = (Char **) COMP(arg2);
				setCOMP(arg2,(Symbolhandle **) 0);
				setCOMP(result,(Symbolhandle **) temphandle);
			} /*switch (type2)*/
			Delete(arg2);
		}
		arg2 = (Symbolhandle) 0;
	} /*if(isscratch(NAME(arg2)))*/
	else
	{/* r.h.s. not scratch, copy to l.h.s. */
		if(arg1 == (Symbolhandle) 0)
		{/* no l.h.s. symbol remembered; must be not UNDEF keyword */
			result = Makesymbol(type2);
		}
		else
		{/* re-use l.h.s. symbol */
			result = arg1;
		}

		if (result == (Symbolhandle) 0 || Copy(arg2, result) == 0)
		{
			goto errorExit;
		}
	} /*if(isscratch(NAME(arg2))){}else{}*/
	setNAME(result, name);
	Addsymbol(result);

	return (result);

  errorExit:				/* from Assign */
	if(*OUTSTR)
	{
		if(keyword)
		{
			yyerror(OUTSTR);
		}
		else
		{
			putOUTSTR();
		}
	}
	Delete(result);
	return (0);
}
/* 
  As of 930503, it is the responsibility of the calling program to get
  rid of arg2 if the copy could not be completed.
*/

Long Copy(arg1, arg2)
Symbolhandle    arg1, arg2;
{
	Symbolhandle    symh;
	Long            i, j, tot, ndims;
	WHERE("Copy");
	
	/* recursively copy contents of arg1 into arg2 */

	if (arg2 == (Symbolhandle) 0 || arg1 == (Symbolhandle) 0)
	{
		myerrorout("ERROR: (internal) illegal copy argument(s)");
		goto errorExit;
	}

	setTYPE(arg2,TYPE(arg1));
	setNCLASS(arg2,NCLASS(arg1));
	ndims = NDIMS(arg1);
	tot = symbolSize(arg1);
	setNDIMS(arg2,ndims);
	for (i = 1; i <= MAXDIMS; i++)
	{
		if (i <= ndims)
		{
			setDIM(arg2,i,DIMVAL(arg1,i));
		}
		else
		{
			setDIM(arg2,i,0);
		}
	} /*for (i = 1; i <= MAXDIMS; i++)*/

	/* copy contents, if any */
	switch ((int) TYPE(arg1))
	{
	  case UNDEF:
		break;
		
	  case MACRO:
	  case MACRO1:
	  case CHAR:
		TMPHANDLE = STRING(arg1);
		if(TMPHANDLE != (Char **) 0)
		{
			TMPHANDLE = myduphandle(TMPHANDLE);
			if(TMPHANDLE == (Char **) 0)
			{
				goto errorExit;
			}
		}
		setSTRING(arg2,TMPHANDLE);
		break;

	  case LIST:
	  case STRUC:
		if(COMP(arg1) != (Symbolhandle **) 0)
		{
			TMPHANDLE = mygethandle(myhandlelength((Char **) COMP(arg1)));
			setCOMP(arg2,(Symbolhandle**) TMPHANDLE);
			if(TMPHANDLE == (Char **) 0)
			{
				goto errorExit;
			}
			for (i = 0; i < tot; i++)
			{
				COMPVALUE(arg2, i) = (Symbolhandle) 0;
			}
			
			for (i = 0; i < tot; i++)
			{
				symh = COMPVALUE(arg2, i) = Makesymbol(REAL);
				if (symh == (Symbolhandle) 0 ||
					Copy(COMPVALUE(arg1, i), symh) == 0)
				{
					for (j = 0; j <= i; j++)
					{
						Delete(COMPVALUE(arg2, j));
						COMPVALUE(arg2,j) = 0;
					}
					goto errorExit;
				}
			}					/* for(..i..) */
		} /*if(COMP(arg1) != (Symbolhandle **) 0)*/
		else
		{
			setCOMP(arg2,(Symbolhandle**) 0);
		}
		
		break;

	  case LOGIC:
	  case REAL:
		if(DATA(arg1) != (Double **) 0)
		{
			TMPHANDLE = myduphandle((Char **) DATA(arg1));
			if(TMPHANDLE == (Char **) 0)
			{
				goto errorExit;
			}
			setDATA(arg2,(Double **) TMPHANDLE);
		}
		else
		{
			setDATA(arg2,(Double **) 0);
		}
		break;

	  case LONG:
		if(LONGDATA(arg1) != (Long **) 0)
		{
			TMPHANDLE = myduphandle((Char **) LONGDATA(arg1));
			if(TMPHANDLE == (Char **) 0)
			{
				goto errorExit;
			}
			setLONGDATA(arg2,(Long **) TMPHANDLE);
		}
		else
		{
			setLONGDATA(arg2,(Long **) 0);
		}
		
		break;

	  case PLOTINFO:
		setGRAPH(arg2,copyGraph(GRAPH(arg1)));
		if(GRAPH(arg2) == (whole_graph **) 0)
		{
			goto errorExit;
		}
		break;

	  default:
		myerrorout("ERROR: (internal) copy of illegal variable type");
		goto errorExit;
	} /*switch ((int) TYPE(arg1))*/
/*
  Note that name is copied.  Sometimes that is not wanted.  In
  any event that is done last, so name of arg1 is not changed until it is
  complete
*/
	setNAME(arg2, NAME(arg1));

	return (1);		/* successful */

  errorExit: /* from copy */
	/* leave it to calling function to cleanup arg2 */
	
	return (0);
	
}

/*
  Called by mainpars for str$name (named == 1) or str[[expr]] (named == 0),
  when str is a structure.    It is also called indirectly for str[list]
  when str is a structure.  For str[[expr]], the parser insures there must
  be exactly index.  The parser interprets str[list] as subscripting and
  the subscript processor (Element() in Lang.c) calls Extract.  If list has
  more than one element, all but the first must be 1, T, or empty.

  In addition, expr[[1]] is legal when expr is not a structure and is
  synonymous with expr.  The use of [[...]] is no longer necessary and should
  probably be phased out.
*/

Symbolhandle    Extract(strarg, namearg,named)
Symbolhandle    strarg, namearg;
Long            named;
{
	Long            ncomps = 1, i, j, length, **subvals = (Long **) 0;
	Long            nsubscripts;
	Symbolhandle    result = (Symbolhandle) 0, symh;
	Symbolhandle    longTrash = (Symbolhandle) 0;
	
	/* routine to extract a component of a structure with same name
	   as name of namearg */

	OUTSTR[0] = '\0';
	if (isNull(strarg))
	{
		sprintf(OUTSTR,
				"ERROR: structure component extraction from null variable");
	}
	else if(!isDefined(strarg))
	{
		sprintf(OUTSTR,
				"ERROR: structure component extraction from undefined variable %s",
				NAME(strarg));
	}
	else if(isStruc(strarg))
	{
		ncomps = NCOMPS(strarg);
	}
	else if(named)
	{
		sprintf(OUTSTR,"ERROR: use of expr$name when expr is not a structure");
	}
	if(*OUTSTR)
	{
		goto yyerrorMsg;
	}

	if(!named)
	{/* str[vec] or str[[i]] */
		if(TYPE(namearg) == LIST)
		{ /* str[vec] */
			nsubscripts = NCOMPS(namearg);
		}
		else if(!isDefined(namearg))
		{ /* str[[expr]] */
			sprintf(OUTSTR,
					"ERROR: undefined structure index %s",NAME(namearg));
			goto yyerrorMsg;
		}
		else
		{ /* str[[expr]] */
			nsubscripts = 1;
		}
		
		/* 
		  Check subscripts
		  Note: if nsubscripts > 1, all extra subscripts must be empty or
		  equivalent to cat(1) or cat(T)
		*/
		for(i=0;i<nsubscripts;i++)
		{
			symh = (i>1 || TYPE(namearg) == LIST) ?
				COMPVALUE(namearg,i) : namearg;
			if(symh != (Symbolhandle) 0)
			{
				if(TYPE(symh) != REAL && TYPE(symh) != LOGIC ||
				   !isVector(symh))
				{
					sprintf(OUTSTR,
							"ERROR: structure index not REAL or LOGICAL vector");
				}
				else if(anyMissing(symh))
				{
					sprintf(OUTSTR,
							"ERROR: MISSING values not allowed in structure index");
				}
				else if(i>0 && (!isScalar(symh) || DATAVALUE(symh,0) != 1.0))
				{
					sprintf(OUTSTR,
							"ERROR: extra subscripts on structures must have value 1 or T");
				}						
				if(*OUTSTR)
				{
					goto yyerrorMsg;
				}
			}/*if(symh != (Symbolhandle) 0)*/
		}/*for(i=0;i<nsubscripts;i++)*/

		if(TYPE(namearg) == LIST)
		{ /* any extra subscripts are irrelevent */
			namearg = COMPVALUE(namearg,0);
		}
		
		if(!isStruc(strarg))
		{
			/* must be expr[[1]] which is allowed, even if expr is not STRUC*/
			if(!isScalar(namearg) || DATAVALUE(namearg,0) != 1.0)
			{ /* Note: parser does not allow expr[[]] */
				sprintf(OUTSTR,
						"ERROR: use of a[[i]] with i > 1 when a is not a structure");
				goto yyerrorMsg;
			}
			else
			{/* return entire expression */
				result = Install(SCRATCH,TYPE(strarg));
				if(result == (Symbolhandle) 0 || Copy(strarg,result) == 0)
				{
					goto errorExit;
				}
				setNAME(result, SCRATCH);
				return (result);
			}
		} /* if(!isStruc(strarg)) */
		
		if((length = subscriptLength(namearg,ncomps)) == 0 ||
		   (longTrash = LongInstall(SCRATCH,length)) == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		subvals = LONGDATA(longTrash);
		if((length = buildSubscript(namearg,ncomps,length,*subvals)) == 0)
		{
			goto errorExit;
		}

		if(length > 1 &&
		   (result = StrucInstall(SCRATCH,length)) == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		for(i=0;i<length;i++)
		{
			j = (*subvals)[i];
			symh = Makesymbol(REAL);
			if(length == 1)
			{
				Addsymbol(symh);
				result = symh;
			}
			else
			{
				COMPVALUE(result,i) = symh;
			}
			if(symh == (Symbolhandle) 0 ||
			   Copy(COMPVALUE(strarg,j),symh) == 0)
			{
				goto errorExit;
			}
			if(length == 1)
			{
				setNAME(symh,SCRATCH);
			}
		} /*for(i=0;i<length;i++)*/
		Removesymbol(longTrash);
	} /*if(!named)*/
	else 
	{ /* str$name */
		if(isscratch(NAME(namearg)))
		{
			sprintf(OUTSTR,
					"ERROR: expression or string in place of component name");
			goto yyerrorMsg;
		}
		
		for (i = 0; i < ncomps; i++)
		{
			if (strcmp(NAME(COMPVALUE(strarg, i)), NAME(namearg)) == 0)
			{
				break;
			}
		}
		if(i == ncomps)
		{
			sprintf(OUTSTR,"ERROR: no component with name %s",NAME(namearg));
			goto yyerrorMsg;
		}
		if(TYPE(namearg) == UNDEF)
		{
			Removesymbol(namearg);
		}
		result = Install(SCRATCH,TYPE(COMPVALUE(strarg, i)));
		if(result == (Symbolhandle) 0 || Copy(COMPVALUE(strarg,i),result) == 0)
		{
			goto errorExit;
		}
		setNAME(result,SCRATCH);
	} /*if(!named){}else{}*/

	return (result);

  yyerrorMsg:
	yyerror(OUTSTR);
	/* fall through */
	
  errorExit: /* in Extract() */
	Removesymbol(longTrash);
	Removesymbol(result);
	
	return (0);
}

Symbolhandle    Makelist()
/* start a new list of arguments */
{
	Symbolhandle    result;
	int             i;
	
	result = Install(LISTSCRATCH,LIST);
	
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	TMPHANDLE = mygethandle(DEFAULTLISTLENGTH * (Long) sizeof(Symbolhandle));
	setCOMP(result,(Symbolhandle**) TMPHANDLE);
	if (TMPHANDLE == (Char **) 0)
	{
		goto errorExit;
	}
	setNDIMS(result,1);
	setNCOMPS(result,1);

	for(i = 0;i<DEFAULTLISTLENGTH;i++)
	{
		COMPVALUE(result, i) = (Symbolhandle) 0;
	}
	
	return (result);

  errorExit:
	Removelist(result);
	return (0);
}


Symbolhandle    Addlist(list, arg)
Symbolhandle    list, arg;
{
	/* put arg in the last space of the list */
	if (list != (Symbolhandle) 0)
	{
		if(TYPE(list) != LIST)
		{
			myerrorout("ERROR: (internal) not an argument list");
			return (0);
		}
		COMPVALUE(list, NCOMPS(list) - 1) = arg;
	}
	
	return (list);
}

Symbolhandle    Growlist(mylist)
Symbolhandle    mylist;
{
	Long        ncomps;
	Long        maxcomps, i;
	/* make an additional slot in the mylist */

	if(mylist != (Symbolhandle) 0)
	{
		maxcomps = myhandlelength((Char **) COMP(mylist))/sizeof(Symbolhandle);
		ncomps = NCOMPS(mylist) + 1;
		if(ncomps > maxcomps)
		{
			maxcomps += DEFAULTLISTLENGTH;
			TMPHANDLE = mygrowhandle((Char **) COMP(mylist),
									 maxcomps * sizeof(Symbolhandle));
			setCOMP(mylist,(Symbolhandle**) TMPHANDLE);
			if (TMPHANDLE == (Char **) 0)
			{
				goto errorExit;;
			}
			for(i=ncomps;i<maxcomps;i++)
			{
				COMPVALUE(mylist,i) = (Symbolhandle) 0;
			}
		} /*if(ncomps > maxcomps)*/
		setNCOMPS(mylist,ncomps);
	
		COMPVALUE(mylist, ncomps - 1) = (Symbolhandle) 0;
	} /*if(mylist != (Symbolhandle) 0)*/
	
	return (mylist);

  errorExit:
	Removelist (mylist);
	return (0);
}

Symbolhandle Byname(symh,byNameType,parseMacroArgs)
Symbolhandle             symh;
Long                    *byNameType, *parseMacroArgs;
{
	long                 savePlace;
	long                 i, c, tempname = 0;
	char                 name[NAMELENGTH+1];
	Char                *string;
	WHERE("Byname");
	
	*parseMacroArgs = 0;
	*OUTSTR = '\0';
	if(TYPE(symh) != CHAR || !isScalar(symh))
	{
		sprintf(OUTSTR,
				"ERROR: byname expression is not quoted string or CHARACTER scalar");
		goto errorExit;
	}
	
	string = STRINGPTR(symh);
	if(strlen(string) > NAMELENGTH)
	{
		sprintf(OUTSTR,
				"ERROR: byname expression specifies too long a name '%s'",
				string);
		goto errorExit;
	}
	strcpy(name,string);
	tempname = (name[0] == TEMPPREFIX);
	
	for(i=0;(c = name[i])!= 0;i++)
	{
		if(i == 0 && !tempname && !isnamestart(c) ||
		   i == 1 && (!isnamechar(c) || (tempname && !isnamestart(c)) ) ||
		   i > 1 && !isnamechar(c))
		{
			goto badName;
		}
	}
	if(strcmp(name,"T") == 0 || strcmp(name,"F") == 0)
	{
		goto badName;
	}
	
	if(isscratch(NAME(symh)))
	{
		Removesymbol(symh);
	}
	
	symh = Lookup(name);
	if (symh == (Symbolhandle) 0)
	{
		symh = Install(name, UNDEF);
		if(symh == (Symbolhandle) 0)
		{
			*byNameType = FATALERROR;
			return (0);
		}
		setNDIMS(symh,1);
	} /*if (symh == (Symbolhandle) 0)*/
/*
  920623 set *byNameType to VAR for everything except BLTIN or MACRO
  followed by '('
*/
	switch ((int) TYPE(symh))
	{
	  case BLTIN:
		if(findLParen((char *) 0))
		{
			if(strcmp(NAME(symh), "batch") == 0)
			{
				*byNameType = BATCH;
			}
			else
			{
				*byNameType = BLTIN;
			}
		}
		else
		{
			*byNameType = VAR;
		}
		
		break;

	  case MACRO:
		savePlace = ISTRCHAR;
		if(findLParen((char *) 0))
		{
			*parseMacroArgs = 1;
			*byNameType = MACRO;
		}
		else
		{
			ISTRCHAR = savePlace;
			*byNameType = VAR;
		}
		break;

	  default:
		*byNameType = VAR;
		break;
	}
	return (symh);
	
  badName:
	sprintf(OUTSTR,
			"ERROR: byname expression specifies illegal name '%s'", name);
	/* fall through */

  errorExit: /* from Byname */
	yyerror(OUTSTR);

	return(0);
}

/* for debugging purposes only */

#ifdef UNDEFINED__
void Dumpsymbols(scratch,msgs,count)
int            scratch; /* 0 => regular table, 1 => scratch table*/
Char          *msgs;
Long           count;
{
	Symbolhandle     symh,start;
	Char             value[80];
	Long             type;
	Char            *name;
	WHERE("Dumpsymbols");
	
	sprintf(OUTSTR,"%s symbol table",(scratch) ? "Scratch" : "Regular");
	putOUTSTR();
	start = Firstsymbol(scratch);

	for (symh = start; count-- > 0 && symh != (Symbolhandle) 0;
		 symh = Nextsymbol(symh,scratch))
	{
		name = NAME(symh);
		if(name == (Char *) 0)
		{
			break;
		}
		type = TYPE(symh);
		if(type == REAL && DATA(symh) != (Double **) 0)
		{
			sprintf(value,"%.15g",DATAVALUE(symh,0));
		}
		if(type == LONG && LONGDATA(symh) != (Long **) 0)
		{
			sprintf(value,"%15ld",LONGDATAVALUE(symh,0));
		}
		else if(type == LOGIC && DATA(symh) != (Double **) 0)
		{
			sprintf(value,"%s",(DATAVALUE(symh,0) != 0.0) ? "T" : "F");
		}
		else if(type == CHAR && STRING(symh) != (Char **) 0)
		{
			strncpy(value,STRINGPTR(symh),79);
			value[79] = '\0';
		}
		else
		{
			strcpy(value,"none");
		}
		
		sprintf(OUTSTR,
				"%s: symh = %08x, name = %s, type = %s, value = %s",msgs,symh,
				name,typeName(type),value);
		putOUTSTR();
	}
}
#endif /*UNDEFINED__*/

/*
	function to dummy up list for MacAnova command with narg arguments
	The arguments must all be scalars (vectors of length 1) with values given
	as strings in valueList[], types in typeList[], and if a keyword argument,
	with the keyword in keywordList.
.
	For use in executing commands without going through parser.
	Calling routine should call Removelist(list) after command is executed.
*/
Symbolhandle Buildlist(nargs,typeList,valueList,keywordList)
Long         nargs;
Long         typeList[];
Char        *valueList[], *keywordList[];
{
	Symbolhandle list, arg = (Symbolhandle) 0;
	Long         i;
	Char       **pc = (Char **) 0, name[NAMELENGTH+1];
	Double       value;
	WHERE("Buildlist");
	
	*OUTSTR = '\0';
	list = StrucInstall(SCRATCH,(nargs == 0) ? 1 : nargs);
	if(list == (Symbolhandle) 0)
	{
		return (0);
	}
	setTYPE(list,LIST);
	for(i=0;i<nargs;i++)
	{
#ifdef UNDEFINED__
		PRINT("i=%d, typeList[i] = %s, valueList[i] = '%s',
			keywordList[i] = '%s'\n",(int) i,typeName(typeList[i]),valueList[i],
			keywordList[i]);
#endif /*UNDEFINED__*/
		if(typeList[i] == REAL || typeList[i] == LOGIC)
		{
			arg = RInstall(SCRATCH,1);
			if(arg == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setTYPE(arg,typeList[i]);
			if(typeList[i] == REAL)
			{
				if(sscanf(valueList[i],"%lg",&value) != 1)
				{
					myerrorout("ERROR: invalid value for Buildlist");
					goto errorExit;
				}
			}
			else
			{
				value = (valueList[i][0] == 'T') ? 1.0 : 0.0;
			}
			DATAVALUE(arg,0) = value;
		}
		else if(typeList[i] == CHAR)
		{
			arg = Install(SCRATCH,CHAR);
			pc = mygethandle((Long) strlen(valueList[i]) + 1);
			if(arg == (Symbolhandle) 0 || pc == (Char **) 0)
			{
				goto errorExit;
			}
			setSTRING(arg,pc);
			strcpy(*pc,valueList[i]);
			pc = (Char **) 0;
			setNDIMS(arg,1);
			setDIM(arg,1,1);
		}
		else if(typeList[i] != 0)
		{
			sprintf(OUTSTR,
					"ERROR: illegal type in argument list to Buildlist");
			goto errorExit;
		}
		if(typeList[i] != 0)
		{
			COMPVALUE(list,i) = arg;
			if(keywordList[i][0] != '\0')
			{
				if(strlen(keywordList[i]) > NAMELENGTH-2)
				{
					sprintf(OUTSTR,
							"ERROR: Buildlist keyword name longer than %d characters",
							(int) NAMELENGTH-2);
					goto errorExit;
				}
				name[0] = KEYPREFIX1;
				name[1] = KEYPREFIX2;
				strncpy(name+2,keywordList[i],NAMELENGTH-2);
				name[NAMELENGTH] = '\0';
				setNAME(arg,name);
			}
			arg = (Symbolhandle) 0;
		}
		else
		{	/* empty argument if typeList[i] == 0 */
			COMPVALUE(list,i) = (Symbolhandle) 0;
		}
	} /* for */
	return (list);

  errorExit: /* from Buildlist */
	putOUTSTR();
	Removesymbol(arg);
	mydisphandle(pc);
	Removelist(list);
	return (0);
} /*Buildlist()*/

/* 
   Function to return the total size of memory (in bytes) associated with
   a symbol
*/

long SizeofSymbol(Symbolhandle symh)
{
	long          size = 0;
	long          type, ncomps, i;
	char        **dataHandle;
	
	if (symh != (Symbolhandle) 0)
	{
		type = TYPE(symh);	
		size = myhandlelength((char **) symh);
		dataHandle = STRING(symh);
		
		if(dataHandle != (char **) 0)
		{
			size += myhandlelength(dataHandle);
			if(type == STRUC || type == LIST)
			{
				ncomps = NCOMPS(symh);
				for (i = 0;i < ncomps; i++)
				{
					size += SizeofSymbol(COMPVALUE(symh,i));
				}
			}
			else if (type == PLOTINFO)
			{
				size += sizeofPlot((whole_graph **) dataHandle);
			}
		} /*if(dataHandle != (char **) 0)*/
	} /*if (symh != (Symbohandle) 0)*/
	return (size);
} /*SizeofSymbol()*/
				  
