/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifndef SYMBOLH__
#define SYMBOLH__
#include "typedefs.h"
#include "plot.h"

#ifndef DEFAULTLISTLENGTH
#define DEFAULTLISTLENGTH 20 /* default length for LIST symbols  */
#endif /*DEFAULTLISTLENGTH*/

typedef struct Symbol
{				/* symbol table entry */
	Char            name[NAMELENGTH+1];
	Long            type;	/* CHAR, REAL, BLTIN, LIST, STRUC, MACRO, UNDEF, LOGIC */
	union
	{
		Double        **data;	/* for real data */
		Char          **string;	/* for a character string, or MACRO */
		Long          **longdata;  /* primarily for model parsing */
		whole_graph   **graph; /* for info on graph */
		struct Symbol **(*fptr) (struct Symbol **);	/* for a builtin function */
		struct Symbol ****Comp;
		/* for components of a structure
		    or elements of a list.  The difference
		    between a list and a structure is that
		    elements of a list are in the symbol
		    table while structure components are not
		*/
		Char        ****garbage; /* for keeping track of temporary handles */
	}               value;
	struct Symbol **next;	/* next element in symbol table */
	struct Symbol **prev;	/* previous element , ie doubly linked list */
	Long            nclass;	/* number of classes for factors */
	Long            dim[MAXDIMS + 1];/* dimensions, dim[0] is number of dims */
}               Symbol, **Symbolhandle;

/*
  type for function symbols (that do not change).  Same as Symbol except
  component dim[] is shorter saving MAXDIMS-2 longs
*/

typedef struct FunctionSymbol
{				/* symbol table entry */
	Char            name[NAMELENGTH+1];
	Long            type;	/* CHAR, REAL, BLTIN, LIST, STRUC, MACRO, UNDEF, LOGIC, PLOTINFO */
	union
	{
		Double        **data;	/* for real data */
		Char          **string;	/* for a character string, or MACRO */
		Long          **longdata;  /* primarily for model parsing */
		whole_graph   **graph; /* for info on graph */
		struct Symbol **(*fptr) (Symbol **);	/* for a builtin function */
		struct Symbol ****Comp;
		/* for components of a structure
		    or elements of a list.  The difference
		    between a list and a structure is that
		    elements of a list are in the symbol
		    table while structure components are not
		*/
		Char        ****garbage; /* for keeping track of temporary handles */

	}               value;
	struct Symbol **next;	/* next element in symbol table */
	struct Symbol **prev;	/* previous element , ie doubly linked list */
	Long            nclass;	/* number of classes for factors */
	Long            dim[2];	/* dimensions, dim[0] is number of dims */
}               FunctionSymbol, **FunctionSymbolhandle;

#define ENDPREV ((Symbolhandle) -1)
#define BADHANDLE ((void **) -1)

#define DIM(A)  ((*(A))->dim)
#define DIMVAL(A,I) DIM(A)[I]
#define TYPE(A) ((*(A))->type)
#define NAME(A) ((*(A))->name)
#define NCLASS(A) ((*(A))->nclass)
#define NEXT(A) ((*(A))->next)
#define PREV(A) ((*(A))->prev)

#define FPTR(A) ((*(A))->value.fptr)

#define COMP(A) ((*(A))->value.Comp)
#define COMPPTR(A) (*COMP(A))
#define COMPVALUE(A,I) (COMPPTR(A)[I])

#define DATA(A) ((*(A))->value.data)
#define DATAPTR(A) (*DATA(A))
#define DATAVALUE(A,I) (DATAPTR(A)[I])

#define STRING(A) ((*(A))->value.string)
#define STRINGPTR(A) (*STRING(A))
#define STRINGVALUE(A,I) (STRINGPTR(A)[I])

#define LONGDATA(A) ((*(A))->value.longdata)
#define LONGDATAPTR(A) (*LONGDATA(A))
#define LONGDATAVALUE(A,I) (LONGDATAPTR(A)[I])

#define GARBAGE(A) ((*(A))->value.garbage)
#define GARBAGEPTR(A) (*GARBAGE(A))
#define GARBAGEVALUE(A,I) (GARBAGEPTR(A)[I]) /* handle for I-th scratch var*/

#define GRAPH(A) ((*(A))->value.graph)  /* handle of graph */
#define GRAPHPTR(A) (*GRAPH(A))       /* pointer to graph */

#define NDIMS(A) (DIM(A)[0])
#define NARGS(A) (DIM(A)[1])
#define NCOMPS(A) (DIM(A)[1])

#define isReal(A)  (getSingleType(A)==REAL) /* checks entire structure */
#define isLogic(A) (getSingleType(A)==LOGIC) /* checks entire structure */
#define isChar(A)  (getSingleType(A)==CHAR) /* checks entire structure */
#define isUndef(A) ((A) != (Symbolhandle) 0 && TYPE(A) == UNDEF)
#define isMacro(A) ((A) != (Symbolhandle) 0 && TYPE(A) == MACRO)
#define isGraph(A) ((A) != (Symbolhandle) 0 && TYPE(A) == PLOTINFO)
#define isStruc(A) ((A) != (Symbolhandle) 0 && TYPE(A) == STRUC)
#define isBltin    ((A) != (Symbolhandle) 0 && TYPE(A) == BLTIN)

#define setDATA(A,H)     (DATA(A)     = (H))
#define setLONGDATA(A,H) (LONGDATA(A) = (H))
#define setCOMP(A,H)     (COMP(A)     = (H))
#define setGRAPH(A,H)    (GRAPH(A)    = (H))
#define setSTRING(A,H)   (STRING(A)   = (H))
#define setFPTR(A,H)     (FPTR(A)     = (H))
#define setGARBAGE(A,H)  (GARBAGE(A)  = (H))
#define setPREV(A,H)     (PREV(A)     = (H))
#define setNEXT(A,H)     (NEXT(A)     = (H))

#define setNDIMS(A,N)    (NDIMS(A)    = (N))
#define setNCOMPS(A,N)   (NCOMPS(A)   = (N))
#define setDIM(A,I,N)    (DIM(A)[I]   = (N))

#define setTYPE(A,T)     (TYPE(A)     = (T))
#define setNCLASS(A,N)   (NCLASS(A)   = (N))

#define setNAME(A,P)     Setname(A,P)

/*
  Macros to make it easier to allocate scratch handles that will automatically
  be destroyed in case of an interrupt or aborted output.  The handles are
  saved in GARB Symbol with name 'TRASH__' which is assumed to be in the
  scratch symbol table and hence will be released at prompt time.  Typical
  usage of these macros is as follows:

#define GTMP   0
#define GY     1
#define NTRASH 2
. . . . .
		Double    **y;
		Long      **tmp;
		WHERE("name")
		TRASH(NTRASH,errorExit);  * allocates trashbasket for 2 handles *

		if(!getScratch(tmp,GTMP,n,Long) ||
		   !getScratch(y,GY,n,Double))
		{
			goto errorExit;
		}

  Before returning, use
		emptyTrash();

  If it turns out that y is not to be discarded, but, say, is to be returned
  as part of symbol result, then use
		unTrash(GY);
		DATA(result) = y;
		emptyTrash();
		return (result);
....
	  errorExit:
		emptyTrash();
		return (0);
*/

#define NAMEFORTRASH  TRASH__

/*
  Following macro MUST immediately follow declarations, including macro WHERE,
  before any executable statements.
  It declares scratch GARB Symbol TRASH__ with space for N handles,
  transferring to LABEL in case of error.
*/

#define TRASH(N,LABEL)	Symbolhandle  NAMEFORTRASH = GarbInstall(N);\
						if(NAMEFORTRASH == (Symbolhandle) 0)goto LABEL

/*
  allocate a handle of size LENGTH*sizeof(DATATYPE) to variable NAME and put
  it in slot INDEX of TRASH__.  getScratch returns a value of zero in case
  space cannot be allocated. The handle will automatically be freed when
  TRASH__ is deleted.
*/

#define getScratch(NAME,INDEX,LENGTH,DATATYPE) \
	(\
		GARBAGEVALUE(NAMEFORTRASH,INDEX) = mygethandle((LENGTH) * sizeof(DATATYPE)),\
		(NAME = (DATATYPE **) GARBAGEVALUE(NAMEFORTRASH,INDEX)) != (DATATYPE **) 0\
	)

/* add handle to items in trash */
#define toTrash(NAME,INDEX) GARBAGEVALUE(NAMEFORTRASH,INDEX) = (Char **) (NAME)

/*
  Clear handle slot INDEX in TRASH__ so that deleting TRASH__ will not delete
  the handle in that slot
*/
#define unTrash(INDEX) 	GARBAGEVALUE(NAMEFORTRASH,INDEX) = (Char **) 0

/*
  Immediately free space associated with handle NAME in slot INDEX of TRASH__
*/
#define discardScratch(NAME,INDEX,DATATYPE) (unTrash(INDEX),\
											 mydisphandle((Char **) NAME),\
											 NAME = (DATATYPE **) 0)
/*
  Remove TRASH__ from the scratch symbol table and delete it, along with
  all the handles in its slots.  It has an argument solely so that it can
  be invoked as if it were a function.
*/
#define emptyTrash()	(Removesymbol(NAMEFORTRASH) , NAMEFORTRASH = (Symbolhandle) 0)

#endif /*SYMBOLH__*/
