/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Anovacoe
#else
#define __SEG__ Anovacoe
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <math.h>
#include "dbug.h"

#define Trash  NAMEFORTRASH

extern Symbolhandle    RSInstall(), RSInstall2(), RSInstall3(), Install();
extern void            Removesymbol();

extern long            isNull();
extern void            noData(), badNargs();
extern char           *isKeyword();
extern double          cellstat();

extern double	       ROBUSTVARK;
extern double	       CURRENTRSC;

/* 
  This routine computes the model coefficients for linear models and glms.
  It does the simple thing for balanced linear models, and does some funny
  things for unbalanced models
  
  Usage: coefs() prints linear model coefficients, secoefs() also prints their
    std errors

  	coefs() or secoefs() : Coefficients for all terms
  	coefs(Term) or secoefs(Term) : Coefficients associated with given term
	secoefs(,varno) : All coefficients, variable varno (multivariate models)
	secoefs(Term,varno) : Coefficients for Term, variable varno (multivariate
	models)

	Keyword phrases se:F or coefs:F suppress the computation of either the
	standard errors or coefficients.  To have them both F is an error.
	The default values of se and coefs are F and T on coefs() and T and T on
	secoefs().

	When no specific term is specifed, keyword phrase byterm:F on secoefs()
	changes the form of the output from a structure with one component per
	term, each with components 'coefs' and 'se' to a structure with two
	componenents 'coefs' and 'se', each with one component for each term.
	byterm is ignored if a term is specified or if only coefs or only se are
	to be returned.

	For all usages of secoefs, one can use an extra argument of the form
	error:errorTerm where errorTerm is a CHAR variable which is the name of a
	term in the model that will be used to compute the MSE needed for SE's.
	This is ignored on iterative models (poisson(), logistic(), robust(),
	for which MSE is taken as 1.

	Term and errorTerm can either be CHARACTER variables (e.g., "a.b" or 
	"ERROR1" or a term number such as 1 (for CONSTANT).

	For reasons of backwards compatibility, 'anovacoefs' is synonymous with
    'coefs'.
*/

static Symbolhandle    Trash;

#define GTMPY     0
#define GTMPMEANS 1
#define GCOEFSH   2
#define GSESH     3
#define GTMPALPHA 4
#define GTMPMARG  5
#define GCOLPTR   6
#define NTRASH    7

/* get coords for the jth value in table */
static void getoldcoord(long j, long dim, long term[], long *offsets)
{
	long            i;
	WHERE("getoldcoord");
	
	for (i = dim - 1; i >= 0; i--)
	{
		term[i] = j / offsets[i];
		j -= term[i] * offsets[i];
	}
}

/*
	Code to work with (pseudo) balanced case.  Removed from anovacoefs()
	so as to shorten that code
*/
static Symbolhandle doBalanced(long do_coefs, long do_ses, long byterm,
							   long icurrmin, long icurrmax,
							   long errorterm, Symbolhandle result)
{
	Symbolhandle    symh, coefsSH, sesSH;
	long            ncells = 1;
	long            ndata = (long) NDATA, nvars = (long) NVARS;
	long            term[MAXVARS];
	long            ncomps = do_coefs + do_ses;
	long            oneTerm = (icurrmin == icurrmax);
	long            i, j, icurr, imax, ndims;
	double          dfe = (*DF)[errorterm], mse, tmp;
	double        **tmpyH, **tmpMeansH, **coefsH, **sesH;
	double         *y, *residuals, *tmpy, *ses;
	/* for balanced or pseudo balanced data, we just work through the
	"hand" computations */

	ncells = 1;
	for (i = 0; i < nvars; i++)
	{
		ncells *= NCLASSES[i];
	}
	if(do_coefs)
	{
		if(!getScratch(tmpyH,GTMPY,ndata,double) ||
		   !getScratch(tmpMeansH,GTMPMEANS,ncells,double))
		{
			goto errorExit;
		}
		y = *Y;
		residuals = *RESIDUALS;
		tmpy = *tmpyH;
		for (i = 0; i < ndata; i++)
		{
			if (PREVMODELTYPE & IPF)
			{
				*tmpy++ = log(*y++ - *residuals++);
			}
			else
			{
				*tmpy++ = *y++;
			}
		}
		for (icurr = 0; icurr < icurrmin; icurr++)
		{					/* sweep out prior stuff */
			getterm(modelterm(MODEL,icurr), term, nvars);
			/* means */
			(void) cellstat(tmpyH, term, tmpMeansH,
							(double **) 0, (double **) 0, CELLMEANSANDSWEEP);
		}
		discardScratch(tmpMeansH,GTMPMEANS,double);
	} /*if(do_coefs)*/
	
	if(do_ses > 0)
	{
		mse = (*SS)[errorterm]/dfe;
	}
	
	for (icurr = icurrmin; icurr <= icurrmax; icurr++)
	{
		if(ncomps == 2)
		{
			if(oneTerm)
			{
				coefsSH = COMPVALUE(result, 0);
				sesSH = COMPVALUE(result, 1);
			}
			else
			{
				if(byterm)
				{
					symh = COMPVALUE(result, (icurr - icurrmin));
					coefsSH = COMPVALUE(symh, 0);
					sesSH = COMPVALUE(symh, 1);
				}
				else
				{
					symh = COMPVALUE(result,0);
					coefsSH = COMPVALUE(symh,(icurr - icurrmin));
					symh = COMPVALUE(result,1);
					sesSH = COMPVALUE(symh,(icurr - icurrmin));
				}
			}
		} /*if(ncomps == 2)*/
		else
		{
			if(oneTerm)
			{
				coefsSH = result;
			}
			else
			{
				coefsSH = COMPVALUE(result, (icurr - icurrmin));
			}
			if(!do_coefs)
			{
				sesSH = coefsSH;
				coefsSH = (Symbolhandle) 0;
			}
		} /*if(ncomps == 2){}else{}*/

		/* coefsSH has coefs, sesSH has se's, if requested */

		getterm(modelterm(MODEL,icurr), term, nvars);	/* get means */

		imax = 1;  /* maximum number of coefficients per variable */
		ndims = 0;
		for (i = 0; i < nvars; i++)
		{
			if (term[i] > 0)
			{
				ndims++;
				imax *= NCLASSES[i];
			}
		}
		if(ndims == 0)
		{
			ndims = 1;
		}
		
		/* set dimensions and get space for coefficients if needed */
		if(do_coefs)
		{
			setNDIMS(coefsSH,ndims);
			setDIM(coefsSH,1,1);
			j = 0;
			for (i = 0; i < nvars; i++)
			{
				if (term[i] > 0)
				{
					j++;
					setDIM(coefsSH,j,NCLASSES[i]); /* get dimensions for results */
				}
			}
			if(!getScratch(coefsH,GCOEFSH,imax,double))
			{
				goto errorExit;
			}
		} /*if(do_coefs)*/
		
		/* set dimensions and get space for standard errors if needed */
		if (do_ses)
		{
			setNDIMS(sesSH,ndims);
			setDIM(sesSH,1,1);
			j = 0;
			for (i = 0; i < nvars; i++)
			{
				if (term[i] > 0)
				{
					j++;
					setDIM(sesSH,j,NCLASSES[i]); /* set dimensions for results */
				}
			}
			if(!getScratch(sesH,GSESH,imax,double))
			{
				goto errorExit;
			}
		} /*if (do_ses)*/

		if(do_coefs)
		{
	/* tmpyH has residuals after sweeping out terms 0 through icurr - 1 */
			(void) cellstat(tmpyH, term, coefsH, (double **) 0,
							(double **) 0, CELLMEANSANDSWEEP);
	/* tmpyH now has residuals after sweeping out terms 0 through icurr */
			unTrash(GCOEFSH);
			setDATA(coefsSH,coefsH);
		}
		
		if (do_ses)
		{
			ses = *sesH;
			if(do_ses > 0)
			{
				tmp = sqrt((*DF)[icurr] * mse /NDATA);
				for (i = 0; i < imax; i++)
				{
					*ses++ = tmp;
				}
			}
			else
			{
				for (i = 0; i < imax; i++)
				{
					setMissing(*ses);
					ses++;
				}
			}
			unTrash(GSESH);
			setDATA(sesSH,sesH);
		} /*if (do_ses)*/
	} /* for (icurr = icurrmin; icurr <= icurrmax; icurr++) */

	return (result);

  errorExit:
	return (0);
} /*doBalanced()*/

Symbolhandle    anovacoefs(list)
Symbolhandle    list;
{
	long            i, i2, colct = 0, icurr, j, j2,k, k1;
	long            ij, ik, jk, ikk;
	long            term[MAXVARS], term2[MAXVARS];
	long            type, margs;
	long          **colptrH = (long **) 0, *colptr;
	long            do_ses, do_coefs = 1, oneTerm, addin, subout;
	long            imax;
	long            icurrmin, icurrmax;
	long            offsets[MAXVARS];
	long            colptri, colptrj;
	long            ndims, nargs, varno = 0, df;
	long            ny = (long) NY;
	long            nterms = (long) NTERMS, errorterm = -1;
	long            nvars = (long) NVARS, nregx = (long) NREGX;
	long            ncomps, dim;
	long            byterm = 1;
	modelPointer    iterm;
	char           *compNames[2], *keyword = (char *) 0, *glmname;
	char            compNames2[9];
	char           *compNames2P = compNames2, **compNames2H = &compNames2P;
	Symbolhandle    symh, symh2, symhkey, result = (Symbolhandle) 0;
	Symbolhandle    coefsSH = (Symbolhandle) 0, sesSH = (Symbolhandle) 0;
	double        **tmpmargH = (double **) 0;
	double        **tmpAlphaH = (double **) 0;
	double        **coefsH = (double **) 0, **sesH = (double **) 0;
	double         *tmpmarg, *tmpAlpha;
	double         *coefs, *ses, *xtxinv, *ss;
	double          tmp, cjj, dfe;
	double          tmpargi, tmpargj;
	double          mse = 1.0;
	double          beta = .35;
	WHERE("anovacoefs");

	Trash = GarbInstall(NTRASH);
	if(Trash == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	*OUTSTR = '\0';
	if(strcmp(FUNCNAME,"anovacoefs") == 0)
	{
		myerrorout("NOTE: anovacoefs() is obsolete; use coefs() instead");
	}

	do_ses = (strcmp(FUNCNAME, "secoefs") == 0);
	margs = 6;
	
	if (PREVMODELTYPE == 0)
	{
		sprintf(OUTSTR,"ERROR: no active model");
	}
	else if (PREVMODELTYPE & LEAPS)
	{
		sprintf(OUTSTR,"ERROR: %s cannot be used after screen()",FUNCNAME);
	}
	else if (!(PREVMODELTYPE & DOCOEFS))
	{
		sprintf(OUTSTR,
				"ERROR: %s cannot be used after GLM command with coefs:F",
				FUNCNAME);
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}
	
	nargs = NARGS(list);

	if(nargs > margs)
	{
		badNargs(FUNCNAME,-margs);
		goto errorExit;
	}
	
	for(i=0;i<nargs;i++)
	{
		symhkey = COMPVALUE(list,i);
		if(i > 0 && symhkey == (Symbolhandle) 0)
		{
			(void) argOK(symhkey,0,i+1);
			goto errorExit;
		}
		
		if((keyword = isKeyword(symhkey)))
		{
			if(strncmp(keyword,"err",3) == 0)
			{				
				if(!isScalar(symhkey) ||
				   (type = TYPE(symhkey)) != CHAR && type != REAL)
				{
					sprintf(OUTSTR,
							"ERROR: %s not CHARACTER variable or string or term number",
							keyword);
					goto errorExit;
				}
				if(type == REAL)
				{
					tmp = DATAVALUE(symhkey,0);
					if(tmp != floor(tmp) || tmp <= 0.0 || tmp > (double) (nterms+1))
					{
						sprintf(OUTSTR,
								"ERROR: %s term number is not integer between 1 and %ld",
								keyword,nterms+1);
						goto errorExit;
					}
					errorterm = (long) tmp - 1;
				}
				else
				{
					i2 = 0;
					for(errorterm = 0;errorterm <= nterms;errorterm++)
					{			/* try to match it */
						if(strcmp(*TERMNAMES + i2,STRINGPTR(symhkey)) == 0)
						{
							break;
						}
						i2 += strlen(*TERMNAMES + i2) + 1;
					}
					if(errorterm > nterms)
					{
						sprintf(OUTSTR,
								"ERROR: %s is not a term in the model %s",
								STRINGPTR(symhkey),STRMODEL);
						goto errorExit;
					}
				}
			} /*if(strncmp(keyword,"err",3) == 0)*/
			else if(strncmp(keyword,"se",2) == 0 ||
					strncmp(keyword,"coef",4) == 0)
			{
				if(!isScalar(symhkey) || TYPE(symhkey) != LOGIC)
				{
					notTorF(keyword);
					goto errorExit;
				}
				if(*keyword == 's')
				{
					do_ses = (DATAVALUE(symhkey,0) != 0.0);
				}
				else
				{
					do_coefs = (DATAVALUE(symhkey,0) != 0);
				}
			}
			else if(strcmp(keyword,"byterm") == 0)
			{
				if(!isScalar(symhkey) || TYPE(symhkey) != LOGIC)
				{
					notTorF(keyword);
					goto errorExit;
				}
				byterm = (DATAVALUE(symhkey,0) != 0.0);
			}
			else
			{
				badKeyword(FUNCNAME,keyword);
				goto errorExit;
			}
			setNAME(symhkey,USEDKEYWORD);
		}
	} /*for(i=0;i<nargs;i++)*/

	if(PREVMODELTYPE & (IPF | FASTANOVA))
	{
		glmname = (PREVMODELTYPE & IPF) ? "ipf" : "fastanova";
		if(!do_coefs)
		{
			sprintf(OUTSTR,
					"ERROR: coefs:F is illegal with %s() after %s()",
					FUNCNAME, glmname);
		}
		else if (do_ses)
		{
			sprintf(OUTSTR,
					"ERROR: standard errors not available after %s()",
					glmname);
		}		
		if(*OUTSTR)
		{
			goto errorExit;
		}
	} /*if(PREVMODELTYPE & (IPF | FASTANOVA))*/
	
	if(nargs > 1 && !isKeyword(symh2 = COMPVALUE(list,1)))
	{/* should be variable number */
		if(!argOK(symh2,REAL,2))
		{
			goto errorExit;
		}
		tmp = DATAVALUE(symh2,0);
		if(!isScalar(symh2) || tmp != floor(tmp) || tmp <= 0 || tmp > NY)
		{
			sprintf(OUTSTR,
					"ERROR: variable number must be positive integer <= %ld",
					ny);
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
		varno = (long) tmp;
		nargs--;
		ny = 1;
	} /*if(nargs > 1 && !isKeyword(symh2 = COMPVALUE(list,1)))*/
	
	ncomps = do_coefs + do_ses;
	if(ncomps == 0)
	{
		sprintf(OUTSTR,
				"ERROR: neither coefficients or standard errors requested");
		goto errorExit;
	}
	
	if(errorterm < 0)
	{
		errorterm = nterms;
	}
	else if(!do_ses)
	{
		sprintf(OUTSTR,
				"WARNING: keyword 'error' ignored when no standard errors requested by %s",FUNCNAME);
		putOUTSTR();
	}
	
	symh = COMPVALUE(list, 0);
	/* icurrmin and icurrmax are the first and last terms for which
		we want coefficients */

	if (symh == (Symbolhandle) 0 || isKeyword(symh))
	{
		icurrmin = 0;
		icurrmax = nterms - 1;	/* do all terms */
	}
	else
	{
		if(!argOK(symh,0,1))
		{
			goto errorExit;
		}

		if(!isScalar(symh) || (type = TYPE(symh)) != CHAR && type != REAL)
		{
			sprintf(OUTSTR,
					"ERROR: 1st argument to %s not CHARACTER variable, string or term number",
					FUNCNAME);
			goto errorExit;
		}
		if(type == REAL)
		{
			tmp = DATAVALUE(symh,0);
			if(tmp != floor(tmp) || tmp <= 0.0 || tmp > (double) nterms)
			{
				sprintf(OUTSTR,
						"ERROR: specified term number integer between 1 and %ld",
						nterms);
				goto errorExit;
			}
			i = (long) tmp - 1;
		}
		else
		{
			i2 = 0;
			for (i = 0; i < nterms; i++)
			{					/* try to match term */
				if (strcmp(STRINGPTR(symh), *TERMNAMES + i2) == 0)
				{
					break;
				}
				i2 += strlen(*TERMNAMES + i2) + 1;
			}
			if (i >= nterms)
			{
				sprintf(OUTSTR, "ERROR: %s is not in active anova model",
						STRINGPTR(symh));
				goto errorExit;
			}
		}
		icurrmin = i;
		icurrmax = i;
	}
	oneTerm = (icurrmin == icurrmax);
	
	compNames[0] = compNames2;
	strcpy(compNames[0],(do_coefs) ? "coefs" : "se");
	if(ncomps == 2)
	{
		compNames[1] = compNames2+6;
		strcpy(compNames[1], "se");
	}
	
	dfe = (*DF)[errorterm];

	if (do_ses)
	{
		/* if this is called by the function secoefs, we want to compute se's
		for the coefs.  do_ses is a flag that says to do se's */

		if (PREVMODELTYPE & (LOGITREG | POISSONREG))
		{
			myerrorout("NOTE: standard errors assume scale parameter is 1");
		}
		else if(dfe <= 0.0)
		{
			myerrorout("WARNING: no degrees of freedom for error; standard errors set MISSING");
			do_ses = -1;
		}
	} /*if (do_ses)*/
	
	
	/* Symbols for results; actual space not yet allocated */
	if(ncomps == 2)
	{
		if (oneTerm)
		{						/* single term with standard errors */
			result = RSInstall(SCRATCH, 2L, compNames, 0L);
		}
		else
		{						/* all terms with standard errors */
			if(byterm)
			{
				result = RSInstall3(SCRATCH,nterms, TERMNAMES, 
									2L, compNames, 0L);
			}
			else
			{
				result = RSInstall3(SCRATCH, 2, compNames2H,
									nterms, (char **) 0, 0);
				if(result != (Symbolhandle) 0)
				{
					for(i=0;i<2;i++)
					{
						k=0;
						symh = COMPVALUE(result,i);
						for(j=0;j<nterms;j++)
						{
							setNAME(COMPVALUE(symh,j), *TERMNAMES + k);
							k += strlen(*TERMNAMES + k) + 1;
						}
					}
				}
			} /*if(byterm)*/
		} /* if(oneTerm){}else{}*/
	} /*if(ncomps == 2)*/
	else
	{
		if (oneTerm)
		{/* single term with only coefs or only standard errors */
			result = Install(SCRATCH, REAL);
		}
		else
		{/* all terms with only coefs or only standard errors */
			result = RSInstall2(SCRATCH, nterms, TERMNAMES, 0L);
		}
	} /*if(ncomps == 2){}else{}*/

	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	if (PREVMODELTYPE & (BALANOVA + FASTANOVA + IPF))
	{			/* (pseudo) balanced data */
		
		result = doBalanced(do_coefs, do_ses, byterm, icurrmin, icurrmax,
							errorterm, result);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		emptyTrash();
	
		return (result);
	} /* if (PREVMODELTYPE & (BALANOVA + FASTANOVA + IPF)) */

	/* now the unbalanced case */

	/* find number of columns in on REGX matrix */
	colct = 0;
	for (i = 0; i < icurrmin; i++)
	{		/* for each term */
		colct += buildXmatrix(i, (double *) 0, MODELINFO);
	}

	for (icurr = icurrmin; icurr <= icurrmax; icurr++)
	{
		if(ncomps == 2)
		{
			if(oneTerm)
			{
				coefsSH = COMPVALUE(result, 0);
				sesSH = COMPVALUE(result, 1);
			}
			else
			{
				if(byterm)
				{
					symh = COMPVALUE(result, (icurr - icurrmin));
					coefsSH = COMPVALUE(symh, 0);
					sesSH = COMPVALUE(symh, 1);
				}
				else
				{
					symh = COMPVALUE(result,0);
					coefsSH = COMPVALUE(symh,(icurr - icurrmin));
					symh = COMPVALUE(result,1);
					sesSH = COMPVALUE(symh,(icurr - icurrmin));
				}
			}
		} /*if(ncomps == 2)*/
		else
		{
			if(oneTerm)
			{
				coefsSH = result;
			}
			else
			{
				coefsSH = COMPVALUE(result, (icurr - icurrmin));
			}
			if(!do_coefs)
			{
				sesSH = coefsSH;
				coefsSH = (Symbolhandle) 0;
			}
		} /*if(ncomps == 2){}else{}*/

		imax = 1;  /* maximum number of coefficients per variable */

		iterm = modelterm(MODEL,icurr);
		if (modeltermEqual(iterm,(modelPointer) NULLTERM))
		{ /* constant term */
			ndims = 1;
			dim = ndims + ((ny > 1) ? 1 : 0);
			if(do_coefs)
			{
				setNDIMS(coefsSH,dim);
				setDIM(coefsSH,1,1);
			}
			if(do_ses)
			{
				setNDIMS(sesSH,dim);
				setDIM(sesSH,1,1);
			}
			offsets[0] = 1;
		} /*if (modeltermEqual(iterm,(modelPointer) NULLTERM))*/
		else
		{
			ndims = 0;
			for(j = 0; j < nvars; j++)
			{
				if(inTerm(j,iterm))
				{
					ndims++;
				}
			}
			dim = ndims + ((ny > 1) ? 1 : 0);
			if(do_coefs)
			{
				setNDIMS(coefsSH,dim);
			}
			if(do_ses)
			{
				setNDIMS(sesSH,dim);
			}
			
			j2 = 0;
			for (j = 0; j < nvars; j++)
			{
				if(inTerm(j,iterm))
				{
					offsets[j2++] = imax;
					dim = labs(NCLASSES[j]);
					if (NCLASSES[j] > 0)
					{
						imax *= dim;
					}
					if(do_coefs)
					{
						setDIM(coefsSH,j2,dim);
					}
					if(do_ses)
					{
						setDIM(sesSH,j2,dim);
					}
				}
			} /*for (j = 0; j < nvars; j++)*/
		} /*if (modeltermEqual(iterm,(modelPointer) NULLTERM)){}else{}*/

		if (ny > 1)
		{
			if(do_coefs)
			{
				setDIM(coefsSH,ndims+1,ny);
			}
			if(do_ses)
			{
				setDIM(sesSH,ndims+1,ny);
			}
		} /*if (ny > 1)*/

		if(do_coefs && !getScratch(coefsH,GCOEFSH,ny*imax,double))
		{
			goto errorExit;
		}
		if(do_ses && !getScratch(sesH,GSESH,ny*imax,double))
		{
			goto errorExit;
		}

		if(!getScratch(tmpmargH,GTMPMARG,imax,double))
		{
			goto errorExit;
		}

		df = getAlpha(icurr, colct, (do_coefs) ? &tmpAlphaH : (double ***) 0,
					  &colptrH);
		if(df < 0)
		{
			goto errorExit;
		}
		if(do_coefs)
		{
			toTrash(tmpAlphaH,GTMPALPHA);
		}
		toTrash(colptrH,GCOLPTR);
		colct += df;

		/* 
		  Now rearrange and fix up marginal terms;
		  warn if pivoted term included
		  Note: (*colptrH)[0] contains ncells
		*/
		colptr = *colptrH + 1;
		for (j = 0; j < imax; j++)
		{
			if (colptr[j] < -1)
			{
				sprintf(OUTSTR,
						"WARNING: Missing df(s) in term %s",
						(icurrmax > icurrmin) ?
						NAME(COMPVALUE(result,icurr-icurrmin)) : "");
				putOUTSTR();
				if(PRINTWARNINGS)
				{
					myerrorout("Missing effects set to zero");
				}
				break;
			} /*if (colptr[j] < -1)*/
		} /*for (j = 1; j <= imax; j++)*/

		/* start by filling in body with zeros */

		if(do_coefs)
		{
			coefs = *coefsH;
		}
		if(do_ses)
		{
			ses = *sesH;
		}
		
		for (i = 0; i < imax * ny; i++)
		{
			if (do_coefs)
			{
				*coefs++ = 0.0;
			}
			if (do_ses)
			{
				*ses++ = 0.0;
			}
		} /*for (i = 0; i < imax * ny; i++)*/

		/*
			Now look at each value, add it in where it is marginal, and
			subtract off where it is internal
		*/
		if(do_coefs)
		{
			coefs = *coefsH;
		}
		if(do_ses)
		{
			ses = *sesH;
		}
		xtxinv = *XTXINV;
		ss = *SS;
		tmpmarg = *tmpmargH;
		if(do_coefs)
		{
			tmpAlpha = *tmpAlphaH;
		}
		
		for (j = 0; j < imax; j++)
		{
			getoldcoord(j, ndims, term2, offsets);
			for (k = 0; k < ndims; k++)
			{
				term2[k]++;
			}
			for (i = 0; i < imax; i++)
			{
				getoldcoord(i, ndims, term, offsets);
				
				/* add in anywhere we match nonzero old coords */
				addin = 1;
				for (k = 0; k < ndims; k++)
				{
					if (term[k] > 0 && term2[k] != term[k])
					{
						addin = 0;
						break;
					}
				}				/*for (k = 0; k < ndims; k++)*/
				tmpmarg[i] = (double) addin;
				/*
				  "sub" out where any nonzero old coords are at max
				  and others match
				  */
				subout = 0;
				addin = 1;
				/* check for max coords */
				for (k = 0; k < ndims; k++)
				{
					dim = (do_coefs) ? DIMVAL(coefsSH,k + 1) :
						DIMVAL(sesSH,k + 1);
					if (term[k] > 0 && term2[k] == dim)
					{
						subout = 1;
						addin *= -1;
					}
				} /*for (k = 0; k < ndims; k++)*/
				/* check for other matches */
				for (k = 0; k < ndims; k++)
				{
					dim = (do_coefs) ? DIMVAL(coefsSH,k + 1) :
						DIMVAL(sesSH,k + 1);
					if (term[k] > 0 && term2[k] != dim && term2[k] != term[k])
					{
						subout = 0;
						break;
					}
				} /*for (k = 0; k < ndims; k++)*/

				if (subout)
				{
					tmpmarg[i] = (double) addin;
				}
			} /*for (i = 0; i < imax; i++)*/
			
			if(do_ses > 0)
			{
				cjj = 0.;
				for (i = 0; i < imax; i++)
				{
					colptri = colptr[i];
					if (colptri >= 0)
					{
						tmpargi = tmpmarg[i];
						for (j2 = 0; j2 < imax; j2++)
						{
							colptrj = colptr[j2];
							if(colptrj >= 0)
							{
								tmpargj = tmpmarg[j2];
								ij = colptri + colptrj * nregx;			
								cjj += tmpargi * tmpargj * xtxinv[ij];
							}
						} /*for (j2 = 0; j2 < imax; j2++)*/
					} /*if (colptri >= 0)*/
				} /*for (i = 0; i < imax; i++)*/
			} /*if(do_ses > 0)*/
			
			jk = j; /* j + k*imax */
			for (k = 0; k < ny; k++)
			{
				k1 = (varno != 0) ? varno - 1 : k;
				if(do_coefs)
				{
					ik = k1 * imax; /* i+k1*imax */
					for (i = 0; i < imax; i++)
					{
						coefs[jk] += tmpAlpha[ik] * tmpmarg[i];
						ik++;
					} /*for (i = 0; i < imax; i++)*/
				} /*if(do_coefs)*/
				
				if (do_ses)
				{
					if(PREVMODELTYPE & ROBUSTREG)
					{
						ses[jk] = sqrt(fabs(cjj * beta)) *
							ROBUSTVARK * CURRENTRSC; 
						/* variance formula from page 40 of Huber 1977 */
					}
					else if (PREVMODELTYPE & ITERWEIGHTS)
					{
						ses[jk] = sqrt(fabs(cjj));
					}
					else
					{
						if(do_ses > 0)
						{
							ikk = k1*(nterms+1)*(ny+1) + errorterm;
							mse = ss[ikk]/dfe;
							ses[jk] = sqrt(fabs(cjj * mse));
						}
						else
						{
							setMissing(ses[jk]);
						}
					}
				} /*if (do_ses)*/
				jk += imax;
			} /*for (k = 0; k < ny; k++)*/
		} /*for (j = 0; j < imax; j++)*/

		if(do_coefs)
		{
			unTrash(GCOEFSH);
			setDATA(coefsSH,coefsH);
		} /*if(do_coefs)*/

		if(do_ses)
		{
			unTrash(GSESH);
			setDATA(sesSH,sesH);
		} /*if(do_ses)*/

		if(do_coefs)
		{
			discardScratch(tmpAlphaH,GTMPALPHA,double);
		}
		discardScratch(tmpmargH,GTMPMARG,double);
		discardScratch(colptrH,GCOLPTR,long);
	} /*for (icurr = icurrmin; icurr <= icurrmax; icurr++)*/
	emptyTrash();
	
	return (result);
	
  errorExit:
	putOUTSTR();
	
	emptyTrash();
	
	Removesymbol(result);

	return (0);
}

