/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Array
#else
#define __SEG__ Array
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <math.h>
#include "dbug.h"

extern long     argOK();
extern void     badType();
extern void     myerrorout();

extern long     isVector();
extern long     symbolSize();

/*
  usage: array(x, dims)
     or: array(x,n1,n2, . . .)
  routine to take a set of data an rewrite subscripts into array form
  first argument is data, second argument is vector of dimensions
  data are always entered with leftmost dimension varying fastest
*/
Symbolhandle    array(list)
Symbolhandle    list;
{

	Symbolhandle    result = (Symbolhandle) 0, arg1, arg2;
	long            tot, i, j;
	long            ndims = 0, dimi, nargs = NARGS(list);
	long            dim[MAXDIMS];
	char           *usage = "usage: array(x,dimVec) or array(x,n1,n2,...)";
	double          val;
	WHERE("array");
	
	*OUTSTR = '\0';
	if(nargs < 2)
	{
		badNargs(FUNCNAME,-(1000+2));
		goto errorExit;
	}
	arg1 = COMPVALUE(list, 0);
	if(!argOK(arg1,0,1))
	{
		goto errorExit;
	}
	if(TYPE(arg1) != REAL && TYPE(arg1) != LOGIC && TYPE(arg1) != CHAR)
	{
		badType(FUNCNAME,-TYPE(arg1),1);
		goto errorExit;
	}
		
	tot = 1;
	for(i=1;i<nargs;i++)
	{
		arg2 = COMPVALUE(list,i);
		if(!argOK(arg2,0,i+1))
		{
			goto errorExit;
		}
		if(TYPE(arg2) != REAL || !isVector(arg2))
		{
			sprintf(OUTSTR,
					"ERROR: dimension arguments for %s must be REAL scalars or vectors",
					FUNCNAME);
			goto errorExit;
		}
		dimi = symbolSize(arg2);
		for(j=0;j<dimi;j++)
		{
			if(ndims >= MAXDIMS)
			{
				sprintf(OUTSTR,
						"ERROR: more than %ld dimensions specified for %s",
						(long) MAXDIMS, FUNCNAME);
				goto errorExit;
			}
			val = DATAVALUE(arg2,j);
		
			if (val!= floor(val) || val < 1.0)
			{
				sprintf(OUTSTR,
						"ERROR: dimension %ld for %s is not positive integer",
						ndims+1, FUNCNAME);
				goto errorExit;
			} /*if (val!= floor(val) || val < 1.0)*/
			dim[ndims++] = (long) val;
			tot *= (long) val;
		} /*for(j=0;j<dimi;j++)*/		
	} /*for(i=1;i<nargs;i++)*/

	if (tot != symbolSize(arg1))
	{
		sprintf(OUTSTR,
				"ERROR: product of dimensions for %s does not match length of data",
				FUNCNAME);
		goto errorExit;
	}

	usage = (char *) 0;

	if(isscratch(NAME(arg1)))
	{ /* don't need new symbol */
		result = arg1;
		COMPVALUE(list,0) = (Symbolhandle) 0;
	}
	else
	{
		result = Install(SCRATCH,TYPE(arg1));
		if (result == (Symbolhandle) 0 || Copy(arg1, result) == 0)
		{
			goto errorExit;
		}
		setNAME(result,SCRATCH);
	}

	for (i=ndims+1;i<=NDIMS(result);i++)
	{
		setDIM(result, i, 0);
	}

	setNCLASS(result,-1);
	setNDIMS(result,ndims);
	for(i=1;i<=ndims;i++)
	{
		setDIM(result,i, dim[i-1]);
	}

	return (result);
	
  errorExit:
	putOUTSTR();
	if(usage)
	{
		myerrorout(usage);
	}

	Removesymbol(result);

	return (0);
	
}
