/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Batch
#else
#define __SEG__ Batch
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <ctype.h>

#ifdef MACINTOSH
#include "macIface.h"
#endif

#define MAXBATCHERRORS 1  /* maximum number of errors allowed on batch */

extern long     symbolSize();
extern char   **mygethandle();
extern void     mydisphandle();
extern char    *isKeyword();

Symbolhandle    batch(list)
Symbolhandle    list;
{

/*
  formerly a routine to insert batch commands into the input stream of
  the parser.  Now (kb) a routine to switch the input stream to an
  alternative file.
*/

	char           *fname, *keyword;
	unsigned char  *inputString;
	long            i, c;
	long            nargs = NARGS(list);
	long            echo; /* default value for 'echo' */
	Symbolhandle    arg1, arg2;
	char           *usage = "usage is 'batch(fileName [,echo:F])', where fileName is CHARACTER";
	WHERE("batch");

	OUTSTR[0] = '\0';
	echo = (BDEPTH > 0) ? BATCHECHO[BDEPTH] : DEFAULTBATCHECHO;
	BDEPTH++;
	if(nargs > 2)
	{
		badNargs(FUNCNAME,-2);
		goto errorExit;
	}
	
	arg1 = COMPVALUE(list,0);
	if (!argOK(arg1,CHAR,1))
	{
		sprintf(OUTSTR,usage);
	}
	else if (!isScalar(arg1))
	{
		sprintf(OUTSTR,"ERROR: %s",usage);
	}
	else if(BDEPTH == MAXBDEPTH)
	{
		sprintf(OUTSTR,"ERROR: batchfiles nested too deep");
	}

	if(*OUTSTR)
	{
		goto errorExit;
	}
	
	if(nargs == 2)
	{
		arg2 = COMPVALUE(list,1);
		if(!argOK(arg2,0,2))
		{
			sprintf(OUTSTR,"%s",usage);
		}
		else if((keyword = isKeyword(arg2)) && strcmp(keyword,"echo") != 0)
		{
			sprintf(OUTSTR,"ERROR: keyword %s not recognized by %s",
					keyword,FUNCNAME);
		}
		else if(TYPE(arg2) != LOGIC || !isScalar(arg2))
		{
			sprintf(OUTSTR,
					"ERROR: 2nd argument (echo) must be T or F");
		}
		else
		{
			echo = (DATAVALUE(arg2,0) != 0.0);
		}
	}

	if(*OUTSTR)
	{
		goto errorExit;
	}

	fname = STRINGPTR(arg1);

#ifdef MACINTOSH
	if(BatchVolume)
	{
		SetVol(0L, BatchVolume);
	}
	if(*fname == '\0')
	{ /* null string o.k. for mac */
		
		fname = macOpen("\pSpecify the batchfile", (Str255) 0, READIT,
			0, (OSType *) 0);

		if(fname == (char *) 0)
		{ /* cancelled */
			goto errorExit;
		}
		BatchVolume = Reply.vRefNum;
	} /*if(*fname == '\0')*/
#endif /*MACINTOSH*/

	if((fname = expandFilename(fname)) == (char *) 0)
	{
		goto errorExit;
	}

	if(!isfilename(fname))
	{
		goto errorExit;
	}

	for(i=1;i<BDEPTH;i++)
	{
		if(strcmp(*INPUTFILENAMES[i],fname) == 0)
		{
			sprintf(OUTSTR,"ERROR: recursive use of batch()");
			goto errorExit;
		}
	}

	if(MDEPTH > 0)
	{
		myerrorout("WARNING: all current macros terminated by use of batch()");
		MDEPTH = 0;
	} /*if(MDEPTH > 0)*/

	if(WDEPTH > 0)
	{
		myerrorout("WARNING: all current 'while' and 'for' loops terminated by use of batch()");
		while(WDEPTH > 0)
		{
			WHILECONDITIONS[WDEPTH-1] = 0;
			mydisphandle((char **) FORVECTORS[WDEPTH-1]);
			FORVECTORS[WDEPTH-1] = (double **) 0;
			WDEPTH--;
		}
	} /*if(WDEPTH > 0)*/

	mydisphandle(INPUTFILENAMES[BDEPTH]);
	if((INPUTFILENAMES[BDEPTH] = mygethandle(strlen(fname)+1)) == (char **) 0)
	{
		goto errorExit;
	}
	strcpy(*INPUTFILENAMES[BDEPTH], fname);

	INPUTFILE[BDEPTH] = fopen(fname,"r");

	if (INPUTFILE[BDEPTH] == (FILE *) 0)
	{
		mydisphandle(INPUTFILENAMES[BDEPTH]);
		INPUTFILENAMES[BDEPTH] = (char **) 0;
		sprintf(OUTSTR,"ERROR: cannot open file %s for %s input",
				fname,FUNCNAME);
		goto errorExit;
	}

	BATCHECHO[BDEPTH] = echo;

	if(INPUTSTRING != (unsigned char **) 0)
	{
		inputString = *INPUTSTRING;
		i = ISTRCHAR;
		c = inputString[i++];
		while(c != '\0')
		{ /* discard remainder of line */
			if(c != '}' && c != ')' && c != ' ' && c != '\t' &&
			   c != ';' && c != '\n' && c != '\r')
			{
				if(c != '#')
				{
					myerrorout("WARNING: remainder of input line after batch() discarded");
				}
				break;
			}
			c = inputString[i++];
		} /*while(c != '\0')*/
		mydisphandle ((char **) INPUTSTRING);
		INPUTSTRING = (unsigned char **) 0;
	} /*if(INPUTSTRING != (unsigned char **) 0)*/

	MAXERRORS1 = (MAXERRORS1 > 0) ? MAXERRORS1 : MAXBATCHERRORS ;
	
	return (NULLSYMBOL);

  errorExit:
	putOUTSTR();

	BDEPTH--;
	return (0);
	
}
