/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Stemleaf
#else
#define __SEG__ Stemleaf
#endif /*MPW3*/
#endif /*MPW*/

#ifdef MPW
#include <Memory.h>
#endif /*MPW*/

#include "globals.h"
#include "plot.h"
#ifdef MACINTOSH
#include "macIface.h"
#endif /*MACINTOSH*/
#include "dbug.h"

#define MAXBOX 40

extern char   **mygethandle();
extern void     mydisphandle();
extern long     symbolSize();

extern long     argOK();
extern long     isNull();
extern char    *typeName();

extern char    *LastGraph; /* global defined in myplot.c */

#define INNERPOINT     5  /* star */
#define OUTERPOINT     0  /* diamond */
#define INNERCHAR     "*"
#define OUTERCHAR     "o" /* for dumb terminal */

Symbolhandle    boxplot(list)
Symbolhandle    list;
{
	/* routine to do boxplot */

	Symbolhandle   symhx, symhfile;
	Symbolhandle   graphsymbol = (Symbolhandle) 0;
	whole_graph  **graph = (whole_graph **) 0;
	curve_points **this_plot = (curve_points **) 0;
	curve_points **last_plot = (curve_points **) 0;
	long           j, k1, k2, ibox, nbox;
	long           tot, notmiss, nout[4], npoints;
	long           nargs = NARGS(list), firstArg = 0, first;
	long           isStruct = 0;
	long           findMin, findMax;	
	long           logx = 0, logy = 0, xaxis = 0, yaxis = 0;
	long           dumb = 0, keep = 1, show = 1, ps = 1, windno = -1;
	long           landscape = 0;
	char          *screenDumpFile = (char *) 0;
	long           toFile = 0, new = 0, termType = STANDARDTERM;
	FILE          *plotfile = (FILE *) 0;
	double         hugedbl = HUGEDBL;
	double       **x = (double **) 0, **y = (double **) 0;
	double         xmin = hugedbl, xmax = -hugedbl;
	double         ymin = hugedbl, ymax = -hugedbl;
#ifdef MULTIPLOTWINDOWS
	long           window = HUGEINT;
#endif /*MULTIPLOTWINDOWS*/
	double         fivenum[5];	
	double         left, right, top, bottom, median, mid;
	char          *msgs = "argument";
	char          *fileName = (char *) 0;
	char           title[TITLESIZE+2];
	char           xlab[XLABELSIZE+2];
	char           ylab[YLABELSIZE+2];
#ifndef MACINTOSH
	long           pause = 1;
#else /*MACINTOSH*/
	long           pause = 0;
	OSType         fileType = 'TEXT';
#endif /*MACINTOSH*/
	WHERE("boxplot");
	TRASH(1,errorExit); /* space for 1 scratch handle */
	
	OUTSTR[0] = title[0] = xlab[0] = ylab[0] = '\0';

	if(FUNCNAME[0] == 'f')
	{
		toFile = 1;
		firstArg = 1;
		symhfile = COMPVALUE(list,0);
		if(!argOK(symhfile,CHAR,1))
		{
			goto errorExit;
		}

		if(!isScalar(symhfile))
		{
			notCharOrString("file name");
			goto errorExit;
		}
		fileName = STRINGPTR(symhfile);
	} /*if(FUNCNAME[0] == 'f')*/
	else
	{
		toFile = 0;
	}
	
	if(nargs < firstArg + 1)
	{
		badNargs(FUNCNAME,-(firstArg + 1));
		goto errorExit;
	}
	
	first = getPlotKeys(list, firstArg+1, (char *) title,
						(char *) xlab, (char *) ylab,
						&xmin, &ymin, &xmax, &ymax,
						(long *) 0, (long *) 0,
						&xaxis, &yaxis, &dumb, &keep,
						(!toFile) ? &show : (long *) 0,
						(long *) 0,
						&ps,
						&new,
						(!toFile) ? &fileName : (char **) 0,
						(long *) 0,
#ifdef MULTIPLOTWINDOWS
						&window,
#else /*MULTIPLOTWINDOWS*/
						(long *) 0,
#endif /*MULTIPLOTWINDOWS*/
						&pause,
						(long *) 0,
						(long *) 0,
						(long *) 0,
						(double *) 0,
						&landscape,
#ifdef SCREENDUMP
						&screenDumpFile
#else /*SCREENDUMP*/
						(char **) 0
#endif /*SCREENDUMP*/
						);
	if(first < 0)
	{
		goto errorExit;
	}
	
	toFile = (toFile || fileName != (char *) 0);

	if(toFile)
	{
		termType = PSTERM;

		if(new && ps)
		{
			PutProlog = 1;
		}
		else
		{
			PutProlog = 0;
		}
		if(!ps)
		{
#ifndef TEK
			termType = DUMBTERM;
#else /*TEK*/
			termType = STANDARDTERM;
#endif /*TEK*/
		}
	}
	
	if(dumb)
	{
		termType = DUMBTERM;
	}

	if (screenDumpFile != (char *) 0)
	{
		if(termType != STANDARDTERM)
		{
			sprintf(OUTSTR,
					"ERROR: screendump:fileName can only be used with high resolution plot");
			goto errorExit;
		}
		
		if((screenDumpFile = expandFilename(screenDumpFile)) == (char *) 0)
		{
			goto errorExit;
		}
	} /*if (screenDumpFile != (char *) 0)*/

	if(!keep && !show)
	{
		sprintf(OUTSTR,	"ERROR: keep:F not legal with show:F");
		goto errorExit;
	}
	
#ifdef MULTIPLOTWINDOWS
	if(window != HUGEINT)
	{
		if(window > NGRAPHS)
		{
			sprintf(OUTSTR,
					"ERROR: value for window must be integer, 0 <= window <= %ld",
					(long) NGRAPHS);
			goto errorExit;
		}
		windno = window;
	}
	else
	{
		windno = -1;
	}
#endif /*MULTIPLOTWINDOWS*/
	
	if(title[0] == '\0')
	{
		strcpy(title, "Box Plot");
	}
	if(xlab[0] == '\0')
	{
		strcpy(xlab,"Values");
	}
	if(ylab[0] == '\0')
	{
		strcpy(ylab,"Box Number");
	}


	nbox = ((first == 0) ? nargs : first-1) - firstArg;

	symhx = COMPVALUE(list, firstArg);
	if(!argOK(symhx,(long) 0, firstArg+1))
	{
		goto errorExit;
	}
	
	if(nbox == 1 && TYPE(symhx) == STRUC)
	{
		isStruct = 1;
		msgs = "structure component";
		list = symhx;	/* if structure form, set list to structure */
		nbox = NCOMPS(list);
		firstArg = 0;
	}
	
	if (nbox > MAXBOX)
	{
		sprintf(OUTSTR,"ERROR: too many boxes to plot");
		goto errorExit;
	}

	if(ymin > .5)
	{
		ymin = .5;
	}
	if(ymax < (double) nbox + 0.5)
	{
		ymax = (double) nbox + 0.5;
	}
	findMin = (xmin == hugedbl);
	findMax = (xmax == -hugedbl);
	
	for (ibox = 0; ibox < nbox; ibox++)
	{ /* check all args or components before doing anything */
		symhx = COMPVALUE(list, firstArg+ibox);
		if (isNull(symhx))
		{
			if(!isStruct)
			{
				sprintf(OUTSTR,"ERROR: %ld-th %s for %s is missing or null",
						firstArg+ibox+1, msgs, FUNCNAME);
			}
		}
		else if (!isDefined(symhx))
		{
			sprintf(OUTSTR,"ERROR: %ld-th %s (%s) for %s is undefined",
					firstArg+ibox+1, msgs, NAME(symhx),FUNCNAME);
		}
		else if (TYPE(symhx) != REAL)
		{
			sprintf(OUTSTR,"ERROR: %ld-th %s for %s is %s, not REAL",
					firstArg+ibox+1, msgs, FUNCNAME,typeName(TYPE(symhx)));
		}
		else if (!isVector(symhx))
		{
			sprintf(OUTSTR,"ERROR: %ld-th %s for %s is not vector",
					firstArg+ibox+1, msgs, FUNCNAME);
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
	} /* for ibox */
	
	if(toFile)
	{
#ifdef MACINTOSH
		if(PlotVolume)
		{
			SetVol(0L, PlotVolume);
		}
		if(strlen(fileName) == 0)
		{
		/* null filename requests dialog box on Mac */

		/* Default file name no longer specified */
			fileName = macOpen("\pFile for box plot:",
							   "\p", WRITEIT, 0, (OSType *) 0);
			if (fileName == (char *) 0)
			{ /* Canceled */
				return (0);
			}
			PlotVolume = Reply.vRefNum;
		} /*if(strlen(fileName) == 0)*/

#endif /*MACINTOSH*/
		if((fileName = expandFilename(fileName)) == (char *) 0)
		{
			goto errorExit;
		}

		if(!isfilename(fileName))
		{
			goto errorExit;
		}
		
		plotfile = openPlotFile(fileName,new);
		if(plotfile == (FILE *) 0)
		{
			goto errorExit;
		}
	}
	else
	{
		plotfile = STDOUT;
	}

	/*
	  Be very careful to keep the structure of graph up to date.
	  All handles are zeroed prior to attempting to get storage;
	  (*graph)->nplots is updated as soon a a new struct curve_points
	  is allocated, as is (*graph)->first_plot or NEXTGRAPH(last_plot);
	  NEXTGRAPH(this_plot) and GRAPHPOINTS(this_plot) are zeroed as soon
	  as this_plot is allocated.
	  This allows cleanup to be done by deleteGraph
	*/
	if((graphsymbol = GraphInstall(SCRATCH)) == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	graph = GRAPH(graphsymbol);

	for (ibox = 0; ibox < nbox; ibox++)
	{
		symhx = COMPVALUE(list, firstArg+ibox);
		if(!isNull(symhx))
		{/* null symbol is not an error */
			tot = symbolSize(symhx);
			x = DATA(symhx);
			if(!getScratch(y,0,tot,double))
			{
				goto errorExit;
			}

			notmiss = 0;
			for (j = 0; j < tot; j++)
			{
				if (!isMissing((*x)[j]))
				{
					(*y)[notmiss++] = (*x)[j];
				}
			}
			if (notmiss < tot)
			{
				sprintf(OUTSTR, "WARNING: %ld missing values in box %ld",
						tot - notmiss, ibox+1);
				putOUTSTR();
			}

			if(notmiss > 0)
			{
				boxprep(*y, notmiss, fivenum, nout);
	
				if (findMin && (*y)[0] < xmin)
				{
					xmin = (*y)[0];
				}
				if (findMax && (*y)[notmiss - 1] > xmax)
				{
					xmax = (*y)[notmiss - 1];
				}
				mid = (double) ibox + 1 + .05; /* offset from tick marks */
				bottom = mid - .33;
				top = mid + .33;
				left = fivenum[1];
				median = fivenum[2];
				right = fivenum[3];
			
				/* always draw whiskers */
				this_plot = (curve_points **)
					mygethandle(sizeof(curve_points));
				if(this_plot == (curve_points **) 0)
				{
					goto errorExit;
				}
				NEXTGRAPH(this_plot) = (curve_points **) 0;
				GRAPHPOINTS(this_plot) = (coordinate **) 0;
				(*graph)->nplots++;

				if((*graph)->first_plot == (curve_points **) 0)
				{
					(*graph)->first_plot = this_plot;
				}
				else
				{
					NEXTGRAPH(last_plot) = this_plot;
				}
				mylockhandle((char **) this_plot);
				last_plot = this_plot;
			
				/* whiskers (2 segments) */
				TMPHANDLE = mygethandle(2*3 * sizeof(coordinate));
				if (TMPHANDLE == (char **) 0)
				{
					goto errorExit;
				}
				GRAPHPOINTS(this_plot) = (coordinate **) TMPHANDLE;
				mylockhandle(TMPHANDLE); /* needed in Mac version */

				PLOTSTYLE(this_plot) = LINES;
				POINTTYPE(this_plot) = INNERPOINT; /* ignored */
				LINETYPE(this_plot) = BORDERLINETYPE;
			
				PCOUNT(this_plot) = 6;
				/* left whisker */
				set_point(this_plot,0,fivenum[0],mid,(char *) 0);
				set_point(this_plot,1,left,mid,(char *) 0);
				undef_point(this_plot,2);

				/* right whisker */
				set_point(this_plot,3,right,mid,(char *) 0);
				set_point(this_plot,4,fivenum[4],mid,(char *) 0);
				undef_point(this_plot,5);

				/* draw box with median line (5 segments) */
				this_plot = (curve_points **)
					mygethandle(sizeof(curve_points));
				if(this_plot == (curve_points **) 0)
				{
					goto errorExit;
				}
				NEXTGRAPH(this_plot) = (curve_points **) 0;
				GRAPHPOINTS(this_plot) = (coordinate **) 0;
				NEXTGRAPH(last_plot) = this_plot;
				(*graph)->nplots++;
				mylockhandle((char **) this_plot);
				last_plot = this_plot;

				TMPHANDLE = mygethandle(5*3 * sizeof(coordinate));
				if (TMPHANDLE == (char **) 0)
				{
					goto errorExit;
				}
	
				GRAPHPOINTS(this_plot) = (coordinate **) TMPHANDLE;
				mylockhandle(TMPHANDLE); /* needed in Mac version */

				PLOTSTYLE(this_plot) = LINES;
				POINTTYPE(this_plot) = INNERPOINT; /* ignored */
				LINETYPE(this_plot) = BORDERLINETYPE;

				PCOUNT(this_plot) = 15;

				/* top box */
				set_point(this_plot,0,left,top,(char *) 0);
				set_point(this_plot,1,right,top,(char *) 0);
				undef_point(this_plot,2);
				/* bottom box */
				set_point(this_plot,3,left,bottom,(char *) 0);
				set_point(this_plot,4,right,bottom,(char *) 0);
				undef_point(this_plot,5);
				/* vertical lines */
				set_point(this_plot,6,left,bottom,(char *) 0);
				set_point(this_plot,7,left,top,(char *) 0);
				undef_point(this_plot,8);
				set_point(this_plot,9,median,bottom,(char *) 0);
				set_point(this_plot,10,median,top,(char *) 0);
				undef_point(this_plot,11);
				set_point(this_plot,12,right,bottom,(char *) 0);
				set_point(this_plot,13,right,top,(char *) 0);
				undef_point(this_plot,14);

				/* outside values */
				if ((npoints = nout[1] + nout[2]) > 0)
				{
					this_plot = (curve_points **)
						mygethandle(sizeof(curve_points));
					if(this_plot == (curve_points **) 0)
					{
						goto errorExit;
					}
					NEXTGRAPH(this_plot) = (curve_points **) 0;
					GRAPHPOINTS(this_plot) = (coordinate **) 0;
					NEXTGRAPH(last_plot) = this_plot;
					(*graph)->nplots++;
					mylockhandle((char **) this_plot);
					last_plot = this_plot;

					TMPHANDLE = mygethandle(npoints*sizeof(coordinate));
					if (TMPHANDLE == (char **) 0)
					{
						goto errorExit;
					}
					GRAPHPOINTS(this_plot) = (coordinate **) TMPHANDLE;
					mylockhandle(TMPHANDLE); /* needed in Mac version */

					PLOTSTYLE(this_plot) = CHARS;
					POINTTYPE(this_plot) = INNERPOINT; /* ignored */
					LINETYPE(this_plot) = BORDERLINETYPE;

					PCOUNT(this_plot) = npoints;

					k1 = 0;
					/* low outside */
					k2 = nout[0];
					for (j = 0; j < nout[1]; j++)
					{
						set_point(this_plot,k1++,(*y)[k2++],mid,INNERCHAR);
					}
					/* hi out */
					k2 = notmiss - 1 - nout[3];
					for (j = 0; j < nout[2]; j++)
					{
						set_point(this_plot,k1++,(*y)[k2--],mid,INNERCHAR);
					}
				}
	
				/* far outside values */
				if ((npoints = nout[0] + nout[3]) > 0)
				{
					this_plot = (curve_points **)
						mygethandle(sizeof(curve_points));
					if(this_plot == (curve_points **) 0)
					{
						goto errorExit;
					}
					NEXTGRAPH(this_plot) = (curve_points **) 0;
					GRAPHPOINTS(this_plot) = (coordinate **) 0;
					NEXTGRAPH(last_plot) = this_plot;
					(*graph)->nplots++;
					mylockhandle((char **) this_plot);
					last_plot = this_plot;
					
					TMPHANDLE = mygethandle(npoints*sizeof(coordinate));
					if (TMPHANDLE == (char **) 0)
					{
						goto errorExit;
					}
					/* NOTE TMPHANDLE is indirectly disposed of by mydispnhandles */
					GRAPHPOINTS(this_plot) = (coordinate **) TMPHANDLE;
					mylockhandle(TMPHANDLE); /* needed in Mac version */
				
					PLOTSTYLE(this_plot) = CHARS;
					POINTTYPE(this_plot) = 0; /* ignored */
					LINETYPE(this_plot) = BORDERLINETYPE;

					PCOUNT(this_plot) = npoints;

					k1 = k2 = 0;
					for (j = 0; j < nout[0]; j++)
					{
						set_point(this_plot,k1++,(*y)[k2++],mid,OUTERCHAR);
					}
					k2 = notmiss - 1;
					for (j = 0; j < nout[3]; j++)
					{
						set_point(this_plot,k1++,(*y)[k2--],mid,OUTERCHAR);
					}
				}
			}					/* if notmiss > 0 */
			discardScratch(y,0,double);
		} /* if(!isNull(symhx) */
	} /* for ibox */
	if ((*graph)->nplots == 0)
	{
		sprintf(OUTSTR,"ERROR: no data!");
		goto errorExit;
	}
	strcpy((*graph)->title,title);
	strcpy((*graph)->xlab,xlab);
	strcpy((*graph)->ylab,ylab);
	(*graph)->xmin = xmin;
	(*graph)->xmax = xmax;
	(*graph)->ymin = ymin;
	(*graph)->ymax = ymax;
	(*graph)->logx = logx;
	(*graph)->logy = logy;
	(*graph)->xaxis = xaxis;
	(*graph)->yaxis = yaxis;
	
	if(keep)
	{
		Remove(LastGraph);
		Cutsymbol(graphsymbol);
		setNAME(graphsymbol,LastGraph);
		Addsymbol(graphsymbol);
	}	

	if(!do_whole_graph(graph, plotfile, termType, windno, pause,landscape,
					   screenDumpFile))
	{
		if(keep)
		{
			graph = (whole_graph **) 0;
		}
		goto errorExit;
	}

	if(plotfile != STDOUT)
	{
#ifdef MACINTOSH
		macSetInfo(PlotVolume,fileName,fileType,CREATOR);
#endif /*MACINTOSH*/
		fclose(plotfile);
	}
	
	if(!keep)
	{
		Removesymbol(graphsymbol);
	}
	emptyTrash();
	return (NULLSYMBOL);

  errorExit:
	putOUTSTR();
	
	if(plotfile != STDOUT)
	{
		fclose(plotfile);
	}
	emptyTrash();
	Removesymbol(graphsymbol);
	
	return (0);
	
}


