/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Cellstat
#else
#define __SEG__ Cellstat
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

#define x(i,j) (*X[j])[i]	/* i is case, j is variable */

/*  routine to compute cell by cell stats for a particular model term
    used for description or sweeping

	curyy     handle to response vector
	X         array of handles to class variables
	NVARS     number of variables (columns of x)
	NCLASSES  number of classes in each variable
	term      particular model term to fit    term is 1 or 0 as a particular
	          variable is included or excluded from the term
	tmeans    handle for vector of means
	tvars     handle for vector of variances
	tcounts   handle for vector of counts
	op        opcode
	  CELLSTATS  compute means, variances, counts
	  CELLSUMS   compute cell sums only assuming balance (ipf)
	  CELLCOUNTS compute cell counts only
	  CELLMEANSANDSWEEP compute cell means, sweep out from y, and return RSS\
	NOTE: tvars not referenced if op != CELLSTATS
	      tmeans is not
*/


double          cellstat(cury, term, tmeans, tvars, tcounts, op)
double        **cury, **tmeans, **tvars, **tcounts;
long            term[], op;
{

	long            i, j, k, level, icellk, ik, ncells, icell;
	long            ndata = (long) NDATA, nvars = (long) NVARS, ny = (long) NY;
	double          comct, curss = 0.0, d;
	double         *means, *counts, *vars;
	double         *y = *cury, *misswts;;

	if(tvars != (double **) 0)
	{
		vars  = *tvars;
	}
	if(tmeans != (double **) 0)
	{
		means  = *tmeans;
	}
	if(tcounts != (double **) 0)
	{
		counts  = *tcounts;
	}
	if(MODELTYPE & MISSWEIGHTS)
	{
		misswts = *MISSWTS;
	}

	ncells = 1;
	for (j = 0; j < nvars; j++)
	{
		term[j] *= ncells;
		if (term[j] > 0)
		{
			ncells *= NCLASSES[j];
		}
	}

	/*  ncells contains the number of means to be computed
	
	    term now contains the offset for indexing into array
	    of means  */

	if (op == CELLMEANSANDSWEEP || op == CELLSUMS)
	{			/* just do for balanced anova */
		for (icell = 0; icell < ncells; icell++)
		{
			means[icell] = 0.;
		}

		for (i = 0; i < ndata; i++)
		{
			icell = 0;
			for (j = 0; j < nvars; j++)
			{
				level = x(i, j);
				icell += term[j] * (level - 1);	/* find cell number */
			}
			means[icell] += y[i];	/* increment sum  */
		}

		if (op == CELLMEANSANDSWEEP)	/* just get totals for ipf */ 
		{
			comct = NDATA / (double) ncells;
			for (icell = 0; icell < ncells; icell++)
			{ /* compute means */
				means[icell] /= comct;
			}
	
			for (i = 0; i < ndata; i++)
			{ /* sweep out current model from y annd compute RSS */
				icell = 0;
				for (j = 0; j < nvars; j++)
				{
					level = x(i, j);
					icell += term[j] * (level - 1);	/* find cell number */
				}
				y[i] -= means[icell];
				curss += y[i] * y[i];
			}
		}
	}
	else if (op == CELLCOUNTS)
	{			/* just get counts */
		for (icell = 0; icell < ncells; icell++)
		{
			counts[icell] = 0.;
		}

		for (i = 0; i < ndata; i++)
		{
			if (!(MODELTYPE & MISSWEIGHTS) || misswts[i] != 0.0)
			{
				icell = 0;
				for (j = 0; j < nvars; j++)
				{
					level = x(i, j);
					icell += term[j] * (level - 1);	/* find cell number */
				}
				counts[icell]++; /* increment count */
			}			
		}
	}
	else
	{ /* CELLSTATS */
		for (icell = 0; icell < ncells * ny; icell++)
		{
			vars[icell] = 0.;	/*  initialize arrays  */
			means[icell] = 0.;
			counts[icell] = 0.;
		}
	
	
		for (i = 0; i < ndata; i++)
		{
			if (!(MODELTYPE & MISSWEIGHTS) || (*MISSWTS)[i] != 0.0)
			{	
				icell = 0;
				for (j = 0; j < nvars; j++)
				{
					level = x(i, j);
					icell += term[j] * (level - 1);	/* find cell number */
				}
				icellk = icell; /* icell + k*ncells */
				ik = i; /* i + k * ndata */
				for (k = 0; k < ny; k++)
				{
					counts[icellk]++; /* increment count */
					means[icellk] += y[ik]; /* increment sum  */
					icellk += ncells;
					ik += ndata;
				}
			}			
		}
	
		for (icell = 0; icell < ncells; icell++)
		{
			icellk = icell; /* icell + k * ncells */
			for (k = 0; k < ny; k++)
			{
				if (counts[icellk] > 0.0)
				{
					means[icellk] /= counts[icellk];
				}
				icellk += ncells;
			}
		}
	
		for (i = 0; i < ndata; i++)
		{
			if (!(MODELTYPE & MISSWEIGHTS) || (*MISSWTS)[i] != 0.0)
			{	
				icell = 0;
				for (j = 0; j < nvars; j++)
				{
					level = x(i, j);
					icell += term[j] * (level - 1);	/* find cell number */
				}
				icellk = icell; /* icell + k * ncells */
				ik = i; /* i + k*ndata */
				for (k = 0; k < ny; k++)
				{				/* get variance*/
					d = y[ik] - means[icellk];
					vars[icellk] += d*d;
					icellk += ncells;
					ik += ndata;
				}
			}			
		}
	
		for (icell = 0; icell < ncells; icell++)
		{
			icellk = icell; /* icell + k * ncells */
			for (k = 0; k < ny; k++)
			{
				if (counts[icellk] > 1.0)
				{
					vars[icellk] /= counts[icellk] - 1.0;
				}
				icellk += ncells;
			}
		}
	}
	return (curss);
}
