/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Cellstts
#else
#define __SEG__ Cellstts
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <ctype.h>

extern long           isScalar();
extern long           argOK();
extern Symbolhandle   RSInstall();

static long getstringterm(char **stringH, long term[])
{

	long            j, nvars = (long) NVARS;
	char           *string = *stringH;
	char            name[NAMELENGTH+1];

	for (j = 0; j < nvars; j++)
	{
		term[j] = 0;
	}

	name[0] = '\0';
	while (*string)
	{
		while (isspace(*string))
		{
			string++;
		}
		j = 0;
		while (isnamechar(*string) && j < NAMELENGTH)
		{
			name[j++] = *string++;
		}
		name[j] = '\0';
		for (j = 1; j <= nvars; j++)
		{
			if (strcmp(name, VARNAMES[j]) == 0)
			{
				term[j - 1] = 1;
				break;
			}
		}
		if (j > nvars)
		{
			myerrorout("ERROR: unrecognized variable in term");
			return (1);
		}
		if (isnamechar(*string))
		{
			myerrorout("ERROR: variable name too long");
			return (1);
		}
		while (isspace(*string))
		{
			string++;
		}
		if (*string != '.' && *string != '\0')
		{
			myerrorout("ERROR: illegal character or improper term specification");
			return (1);
		}
		if (*string == '.')
		{
			string++;
		}
	} /*while (*string)*/
	return (0);
}

Symbolhandle    cellstats(list)
Symbolhandle    list;
{
	Symbolhandle    result, symh;
	double        **out[3];
	long            n, i, j, j2, k, term[10];
	long            nvars = (long) NVARS, ny = (long) NY;
	char          **strterm, *cellcompname[3];

	*OUTSTR = '\0';
	cellcompname[0] = "mean";
	cellcompname[1] = "var";
	cellcompname[2] = "count";

	if (nvars < 1)
	{
		sprintf(OUTSTR,"ERROR: no current model for %s", FUNCNAME);
		goto errorExit;
	}

	if (NARGS(list) != 1)
	{
		badNargs(FUNCNAME,1);
		goto errorExit;
	}

	symh = COMPVALUE(list, 0);
	if(!argOK(symh,CHAR,1))
	{
		goto errorExit;
	}
	if (!isScalar(symh))
	{
		sprintf(OUTSTR,
				"ERROR: argument to %s must be single character variable",
				FUNCNAME);
		goto errorExit;
	}

	strterm = STRING(symh);
	if (getstringterm(strterm, term) > 0)
	{
		/* can't find proper term */
		goto errorExit;
	}

	k = 0;
	n = 1;
	for (i = 0; i < nvars; i++)
	{
		if (term[i] > 0)
		{
			if (NCLASSES[i] < 1)
			{
				sprintf(OUTSTR,
						"ERROR: variables in term for %s must be factors",
						FUNCNAME);
				goto errorExit;
			}
			n *= NCLASSES[i];
			k++;
		}
	} /*for (i = 0; i < nvars; i++)*/

	result = RSInstall(SCRATCH, 3L, cellcompname, ny * n );

	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	for (i = 0; i < 3; i++)
	{
		out[i] = DATA(COMPVALUE(result,i));
	}
	(void) cellstat(Y, term, out[0], out[1], out[2], CELLSTATS);
	if (PREVMODELTYPE & FASTANOVA)	/* for fastanova output */
	{
		myerrorout("WARNING: results include estimated missing data");
	}
	
	for (i = 0; i < 3; i++)
	{
		symh = COMPVALUE(result, i);
		setNDIMS(symh,k + ((ny > 1) ? 1 : 0));
		j2 = 1;
		for (j = 0; j < nvars; j++)
		{
			if (term[j] > 0)
			{
				setDIM(symh,j2,NCLASSES[j]);
				j2++;
			}
		}
		if (ny > 1)
		{
			setDIM(symh,k + 1,ny);
		}
	}
	
	return (result);

  errorExit:
	putOUTSTR();
	
	return (0);
}

