/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Makestr
#else
#define __SEG__ Makestr
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

extern void           Addsymbol();
extern Symbolhandle   Makesymbol();
extern Symbolhandle   Makestruc();
extern void           Delete();
extern long           Copy();
extern void           badNargs();
extern void           noData();
extern void           badType();
extern long           isScalar();
extern void           myerrorout();

/*
  changestr(struc,compName or number, value)
  replace the designated component by value
  If number == ncomps(struc) + 1, add new component with name "compxxx"
  If number < 0, the component will be deleted
  If compName does not match any component, a new component will be
  added.  If compName is specified, the name of the new or replaced
  component will be compName; otherwise it will be the name of value
  of the keyword name is value is of form name:variable
*/

Symbolhandle    changestr(list)
Symbolhandle    list;
{

	/* function to replace component of structure */

	Symbolhandle    arg[3],symh, symh1, result = (Symbolhandle) 0;
	double          tmp;
	long            i, j, ncomps, mcomps, icomp, type1;
	long            nargs = NARGS(list);
	char            compName[NAMELENGTH+1], *keyword;
	WHERE("changestr");

	*OUTSTR = '\0';
	if(nargs < 2 || nargs > 3)
	{
		badNargs(FUNCNAME,(nargs < 2) ? -(1000+2) : -3);
		goto errorExit;
	}
	for(i=0;i<nargs;i++)
	{
		if(isNull(arg[i] = COMPVALUE(list,i)))
		{
			noData(FUNCNAME,i+1);
			goto errorExit;
		}
		if(!isDefined(arg[i]))
		{
			noData(FUNCNAME,-(i+1));
			goto errorExit;
		}
	}

	if(TYPE(arg[0]) != STRUC)
	{
		badType(FUNCNAME,STRUC,1);
		goto errorExit;
	}

	type1 = TYPE(arg[1]);
	if(type1 != REAL && type1 != CHAR || !isScalar(arg[1]))
	{
		sprintf(OUTSTR,
			   "ERROR: usage is %s(struc,compName or compNumber,value)",
			   FUNCNAME);
		goto errorExit;
	}

	mcomps = ncomps = NCOMPS(arg[0]);

	compName[0] = '\0';
	if(nargs == 3)
	{
		if(!(keyword = isKeyword(arg[2])))
		{
			strcpy(compName,NAME(arg[2]));
		}
		else
		{
			strcpy(compName,keyword);
		}
	} /*if(nargs == 3)*/
	
	if(type1 == REAL)
	{
		tmp = DATAVALUE(arg[1],0);
		icomp = (long) tmp - 1;
		if(floor(tmp) != tmp || icomp == -1 || icomp > ncomps ||
		   -icomp > ncomps + 1)
		{
			sprintf(OUTSTR,"ERROR: illegal component number for %s",FUNCNAME);
			goto errorExit;
		}
		if(icomp < 0 && nargs > 2)
		{
			sprintf(OUTSTR,
					"ERROR: only two arguments to %s when compNumber < 0",
					FUNCNAME);
			goto errorExit;
		}
	} /*if(type1 == REAL)*/
	else
	{ /* TYPE(arg[1]) is CHAR */
		if(strlen(STRINGPTR(arg[1])) > NAMELENGTH)
		{
			sprintf(OUTSTR,
					"ERROR: '%s' is too long to be a component name in %s",
					STRINGPTR(arg[1]), FUNCNAME);
			goto errorExit;
		}
		for(i=0;i<ncomps;i++)
		{
			if(strcmp(STRINGPTR(arg[1]),NAME(COMPVALUE(arg[0],i))) == 0)
			{
				break;
			}
		}
		icomp = i;
		/*
		  if no such name, create new component with the name	
		  always use name supplied whether adding or replacing component
		*/
		strcpy(compName,STRINGPTR(arg[1]));
	} /*if(type1 == REAL){}else{}*/

	if(icomp < 0)
	{ /* deleting component */
		mcomps--;
		if(mcomps == 0)
		{
			sprintf(OUTSTR,
					"ERROR: you can't delete the only component of a structure");
			goto errorExit;
		}
	} /*if(icomp < 0)*/
	else if(icomp == ncomps)
	{
		mcomps++;
	}
	
	if((result = StrucInstall(SCRATCH,mcomps)) == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	j = 0;
	for(i=0;i<mcomps;i++)
	{
		COMPVALUE(result,i) = symh = Makesymbol(ASSIGNED);
		if(symh == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		
		if(icomp < 0 && -icomp - 2 == i)
		{ /* skip component to be deleted */
			j++;
		}
		symh1 = (i == icomp) ? arg[2] : COMPVALUE(arg[0],j);
		if(Copy(symh1,symh) == 0)
		{
			goto errorExit;
		}
		j++;
	} /*for(i=0;i<mcomps;i++)*/
	
	if(icomp >= 0)
	{
		setNAME(COMPVALUE(result,icomp),compName);
	}

	return (result);

  errorExit:
	putOUTSTR();

	Removesymbol(result);
	return (0);
	
}
