/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Columnop
#else
#define __SEG__ Columnop
#endif /*MPW3*/
#endif /*MPW*/

/*
  operations sum, prod, max, min
  modified 12/7/91 so that op(a,b,c,...) is equivalent to op(cat(a,b,c,...)),
  where a, b, c, ... are all scalars.
  931121 op(a,b,c,...) now allows vectors.

  Modified February, 1993 to operate recursively on structures.
  Modified March, 1993 to accept logical arguments
*/

#include <math.h>
#include "globals.h"
#include "dbug.h"

extern Symbolhandle  RInstall();
extern Symbolhandle  Makereal();
extern void          Delete();
extern char        **mygethandle();

extern long          dupDims();
extern long          matchKey();

static char       *OpNames[] =
{
	"sum",
	"prod",
	"max",
	"min",
	(char *) 0
};

#define ISUM   1
#define IPROD  2
#define IMAX   3
#define IMIN   4

static long     OpCodes[] =
{
	ISUM, IPROD, IMAX, IMIN
};

static double   Startup[] =
{
	0.0, 1.0, -HUGEDBL, HUGEDBL
};

static Symbolhandle   doOps(Symbolhandle arg,double *params,
							unsigned long control, unsigned long *status)
{
	Symbolhandle   result = (Symbolhandle) 0;
	double        **x = (double **) 0, **resultVals = (double **) 0;
	double          s, y;
	long            i, j, k, nrows, ndims, type = TYPE(arg);
	long            size, l, foundOK = 0;
	int             op = control & OPMASK;
	WHERE("doOps");
	
	x = DATA(arg);
	ndims = NDIMS(arg);
	nrows = DIMVAL(arg,1);
	size = (nrows > 0) ? symbolSize(arg)/nrows : 0;

	result = RInstall(SCRATCH,size);

	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	if(size > 0)
	{
		resultVals = DATA(result);
		if(ndims > 2)
		{
			setNDIMS(result,ndims - 1);
			for(i = 1; i < ndims; i++)
			{
				setDIM(result,i,DIMVAL(arg,i+1));
			}
		}
		else if(ndims == 2)
		{						/* make resultVals a row vector if the argument is a matrix */
			setNDIMS(result,2);
			setDIM(result,1,1);
			setDIM(result,2,DIMVAL(arg,2));
		}
		else
		{
			setNDIMS(result,1);
			setDIM(result,1,1);
		}
		l = k = 0;
		for(j = 0; j < size;j++)
		{
			s = Startup[op-1];

			for (i = 0; i < nrows; i++)
			{
				y = (*x)[l++];
				if (!isMissing(y))
				{
					foundOK = 1;
					if(type == REAL)
					{
						switch (op)
						{
						  case ISUM:
							s += y;
							break;
						  case IPROD:
							s *= y;
							break;
						  case IMIN:
							s = (y < s) ? y : s;
							break;
						  case IMAX:
							s = (y > s) ? y : s;
							break;
						} /* switch(op) */
					} /*if(type == REAL)*/
					else
					{			/* LOGIC */
						switch (op)
						{
						  case ISUM:
							if(y != 0.0)
							{
								s++;
							}
							break;
						  case IPROD:
						  case IMIN:
							s = (y != 0.0) ? s : 0.0;
							break;
						  case IMAX:
							s = (y != 0.0) ? 1.0 : s;
							break;
						} /* switch(op) */
					} /*if(type == REAL){}else{}*/
				} /*if (!isMissing(y))*/
				else
				{
					*status |= FOUNDMISSING;
				}
			} /*for (i = 0; i < nrows; i++)*/

			if((op == IMAX || op == IMIN) && s == Startup[op-1])
			{
				if(!foundOK)
				{
					setMissing((*resultVals)[k]);
				}
				else
				{				/* must be type LOGIC */
					(*resultVals)[k] = (op == IMAX) ? 0.0 : 1.0;
				}
			} /*if((op == IMAX || op == IMIN) && s == Startup[op-1])*/
			else
			{
				(*resultVals)[k] = s;
			}
			k++;
		} /*for(j = 0; j < size;j++)*/
	}
	
	return (result);
	
  errorExit:
	Removesymbol(result);
	return (0);
}

				
/*
  operations sum, prod, max, min
  930202 These now all operate on structures
  931129 Bug fix: If structure component is empty, so is result
*/

Symbolhandle    columnops(list)
Symbolhandle    list;
{
	Symbolhandle    symh, arg1 = (Symbolhandle) 0, result = (Symbolhandle) 0;
	long            i;
	unsigned long   status = 0, control = 0;
	int             op;
	long            nargs = NARGS(list), type, length;
	double         *source, *dest;
	WHERE("columnops");

	*OUTSTR = '\0';

	op = matchKey(FUNCNAME,OpNames,OpCodes);

	arg1 = COMPVALUE(list, 0);
	if(nargs == 1)
	{
		if(!argOK(arg1,0,0))
		{
			return (0);
		}
		type = TYPE(arg1);
		if(type != STRUC && type != REAL && type != LOGIC)
		{
			badType(FUNCNAME,-type,0);
			goto errorExit;
		}
		if(type == STRUC)
		{
			type = getStrucTypes(arg1);
			if(!(type & (REALTYPE1|LOGICTYPE1)) ||
			   (type & ~(REALTYPE1|LOGICTYPE1)))
			{
				sprintf(OUTSTR,
							"ERROR: components of structure argument to %s must all be REAL or LOGICAL",
						FUNCNAME);
				goto errorExit;
			}
		}
	} /*if(nargs == 1)*/
	else
	{ /* allow list of vectors as equivalent to vector */
		length = 0;
		for(i=0;i<nargs;i++)
		{
			symh = COMPVALUE(list,i);
			if(!argOK(symh,0,i+1))
			{
				goto errorExit;
			}
			if(i==0)
			{
				type = TYPE(symh);
			}
			else if(TYPE(symh) != type)
			{
				sprintf(OUTSTR,
						"ERROR: all arguments to %s must have same type",
						FUNCNAME);
				goto errorExit;
			}
			
			if(!isVector(symh))
			{
				sprintf(OUTSTR,
						"ERROR: multiple arguments to %s must all be vectors",
						FUNCNAME);
				goto errorExit;
			}
			length += symbolSize(symh);
		} /*for(i=0;i<nargs;i++)*/

		symh = COMPVALUE(list,0);
		/* put new arg1 in the list so that it will eventually be deleted */
		arg1 = COMPVALUE(list,0) = RInstall(SCRATCH,length);
		if(arg1 == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setTYPE(arg1,type);
		dest = DATAPTR(arg1);
		for(i=0;i<nargs;i++)
		{
			if(i > 0)
			{
				symh = COMPVALUE(list,i);
			}
			source = DATAPTR(symh);
			length = symbolSize(symh);
			while(length-- > 0)
			{
				*dest++ = *source++;
			}
			if(i == 0 && isscratch(NAME(symh)))
			{
				Removesymbol(symh);
			}
		} /*for(i=0;i<nargs;i++)*/
	} /*if(nargs == 1){}else{}*/

	control = (isStruc(arg1)) ? LEFTRECUR : 0;

	result = doRecur1(doOps,arg1,(double *) 0, control | op, &status);
	/* Note: result is already in symbol table */
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	if(status & FOUNDMISSING)
	{
		sprintf(OUTSTR,"WARNING: MISSING values found by %s", FUNCNAME);
		putOUTSTR();
	}

	return (result);

  errorExit:
	putOUTSTR();

	return (0);
}
