/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Handles
#else
#define __SEG__ Handles
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#ifdef MPW
#include <Memory.h>
#define entry Entry
#endif /*MPW*/

/* macros are in to ignore and then reenable SIGINT but the cost is too high */
#undef IGNOREINT
#undef CATCHINT
#define IGNOREINT
#define CATCHINT

char          **mygethandle();
char          **mygrowhandle();
void            mydisphandle();
long            myhandlelength();

extern void   **myNewHandle(), **mySetHandleSize();


long            Nhandles = 0;  /*globals for keeping count */
long            Maxhandles = 0;
long            CurrentMemory = 0;
long            MaxMemory = 0;

char          **mygethandle(n)
long            n;
/* get handle to n bytes and put into table */
{
	void          **h;
	WHERE("mygethandle");
	
	IGNOREINT;
	h = myNewHandle(n);

	if (h == (void **) 0)
	{ /* unable to allocate memory */
		if(!FatalError)
		{
			myerrorout("ERROR: not enough memory, try deleting variables");
		}
	}
	else if(h == BADHANDLE)
	{ /* myNewHandle reports impossible request and has printed error message*/
		h = (void **) 0;
	}
	else
	{
		Nhandles++;
		if(Nhandles > Maxhandles)
		{
			Maxhandles = Nhandles;
		}
	}
	CATCHINT;
	
	return ((char **) h);

}

void mydisphandle(h)
char          **h;
{
    void          **entry;
	
	WHERE("mydisphandle");
	
	if (h == (char **) 0)
	{
		return;		/* nothing to remove */
	}
	
	IGNOREINT;
	
	/* find entry in handle table */
	entry = myFindHandle((void **) h);

	if (entry != (void **) 0)
	{
		myDisposHandle(entry);
		Nhandles--;
	}
	CATCHINT;	
}

/* dispose of up to 5 handles on one call */
void mydisphandle5(h1,h2,h3,h4,h5)
char          **h1, **h2, **h3, **h4, **h5;
{
	mydisphandle(h1);
	mydisphandle(h2);
	mydisphandle(h3);
	mydisphandle(h4);
	mydisphandle(h5);
}

/* dispose of n handles on one call */
void mydispnhandles(h, n)
char         ***h;
long            n;
{
	if(h != (char ***) 0)
	{
		while(n-- > 0)
		{
			mydisphandle(*h);
			*h++ = (char **) 0;
		}
	}	
}

	
char          **mygrowhandle(h, n)
char          **h;
long            n;
/* make a handle bigger */

{
	void          **h2;
	void          **entry;
	WHERE("mygrowhandle");
	
	IGNOREINT;
	/* find handle in entry list */
	entry = myFindHandle((void **) h);

	if (entry == (void **) 0)
	{
		h2 = (void **) 0;
	}
	else
	{
		h2 = mySetHandleSize(entry, n);
		if (h2 == (void **) 0)
		{
			myerrorout("Out of memory, try deleting variables");
		}
		else if(h2 == BADHANDLE)
		{/* myNewHandle reports impossible request, has printed error message*/
			h2 = (void **) 0;
		}
	}
	CATCHINT;
	
	return ((char **) h2);
}

char          **myduphandle(h)
char          **h;
/* duplicate a handle */

{
	void          **h2;
	void          **entry;
	WHERE("myduphandle");
	
	IGNOREINT;

	/* find handle in entry list */
	entry = myFindHandle((void **) h);

	if (entry == (void **) 0)
	{
		h2 = (void **) 0;
	}
	else
	{
		h2 = myDupHandle(entry);

		if (h2 == (void **) 0)
		{
			myerrorout("Out of memory, try deleting variables");
		}
		else
		{
			Nhandles++;
		}
	}
	CATCHINT;
	
	return ((char **) h2);
}

/* currently (921012) myhandlelength() is used only by save() */
long myhandlelength(h)
char            **h;
{
	void        **entry;
	long          length;
	WHERE("myhandlelength");
	
	IGNOREINT;
	
	if (h == (char **) 0)
	{
		length = -1;
	}
	else
	{		
		/* find entry in handle table */
		entry = myFindHandle((void **) h);
		if (entry == (void **) 0)
		{
			length = -2;
		}
		else
		{
			length = myGetHandleSize(entry);
		}
	}
	CATCHINT;
	return (length);
}

void mylockhandle(char **h)
{
	void        **entry;
	
	IGNOREINT;
	
	if (h != (char **) 0 && (entry = myFindHandle((void **) h)) != (void **) 0)
	{
		myHLock(entry);
	}
	CATCHINT;
}

void myunlockhandle(char **h)
{
	void        **entry;
	
	IGNOREINT;
	if (h != (char **) 0 && (entry = myFindHandle((void **) h)) != (void **) 0)
	{
		myHUnlock(entry);
	}
	CATCHINT;
}

char *mygetpointer(long n)
{
	char           *newpointer;

	IGNOREINT;
	newpointer = (char *) myNewPtr(n);
	CATCHINT;

	return (newpointer);
}

void myfreepointer(char *p)
{
	IGNOREINT;
	myDisposePtr((void *) p);
	CATCHINT;
}
