/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Print
#else
#define __SEG__ Print
#endif /*MPW3*/
#endif /*MPW*/

#include <math.h>

#include "globals.h"
#include "plot.h"
#include "dbug.h"

#ifdef MPW
#define comp Comp /*comp is MPW  reserved word */
#endif /*MPW*/

long           symbolSize();
char          *typeName();

/*
  930301
  myprint(), myeol(), myerrorout(), myfeof(), and putOUTSTR() moved here from
  unxio.c and macio.c
*/

#ifdef MACINTOSH
#define PRINTABLE(C)  ((C) >= 0x20 && (C) != 0x7f || (C) == '\t')
#else /*MACINTOSH*/
#define PRINTABLE(C)  ((C) >= 0x20 && (C) < 0x7f || (C) == '\t')
#endif /*MACINTOSH*/

#ifdef UNDEFINED__
/* taken from globkb.h, here for reference */
EXTERN long    		SCREENWIDTH INIT(80);   /* columns on screen */
EXTERN long         FIELDWIDTH INIT(DEFAULTPRINTDEC+7+1);
	/* width of output field */
EXTERN char         DATAFMT[20] INIT(" %12.5g"); /* format for double output */
EXTERN char         MISSINGFMT[10] INIT(" %12s"); /* for MISSING */
EXTERN long         BEFOREDEC INIT(DEFAULTPRINTDEC+7);   /* e.g. 12 in 12.5g */
EXTERN long         AFTERDEC INIT(DEFAULTPRINTDEC);
	/* # after dp in format, e.g., 5 in 12.5g */
EXTERN char         FMTTYPE INIT('g');
	/* output format type, e.g., 'g' in 12.5g */
#endif /*UNDEFINED__*/

#ifndef INDENTSTEP
#define INDENTSTEP  2
#endif /*INDENTSTEP*/

static long         IndentLevel;

static void printSymbol(Symbolhandle symh, FILE *fp, long labels)
{

	long            lastdim, i, j, k, start, chstart, width;
	long            type;
	long            margin, labelWidth = 0, fieldsPerLine;
	long            ndims = NDIMS(symh), nRows, iRow;
	long            factors[MAXDIMS], coord[MAXDIMS];
#ifdef MACINTOSH
	short           deleteOnly = 0;
#endif /*MACINTOSH*/
	char          **ch;
	char           *dataFmt, missingFmt[10];
	unsigned char   c;
	double        **db;
	double          tmp;
	WHERE("printSymbol");

#ifdef MACINTOSH
	deleteOnly = 1;
#endif /*MACINTOSH*/
	lastdim = DIMVAL(symh,ndims);
	nRows  = symbolSize(symh)/lastdim;
	type = TYPE(symh);

	if(labels)
	{
		for(i=1;i<=ndims;i++)
		{
			labelWidth += (long) (log10((double) DIMVAL(symh,i) + 1e-3)) + 2;
		}
		labelWidth += 1;		/* space for closing ')' */
	}
	else
	{
		labelWidth = 0;
	}
	
	if(type == LOGIC)
	{
		/* use width no wider than LENGTHMISSING */
		width = (FIELDWIDTH <= LENGTHMISSING) ? FIELDWIDTH - 1 : LENGTHMISSING;
		fieldsPerLine = (SCREENWIDTH - labelWidth)/(width+1);
		sprintf(missingFmt," %%-%lds",width);
		dataFmt = missingFmt;
	}
	else if(type == REAL)
	{
		fieldsPerLine = (SCREENWIDTH - labelWidth)/FIELDWIDTH;
		dataFmt = DATAFMT;
		strcpy(missingFmt,MISSINGFMT);
	}
	else
	{
		fieldsPerLine = 1;
	}
	if(fieldsPerLine < 1)
	{
		fieldsPerLine = 1;
	}
	factors[0] = 1;
	if(ndims > 1)
	{
		for(i=0;i<ndims-1;i++)
		{
			coord[i] = 1;
			factors[i+1] = factors[i]*DIMVAL(symh,i+1);
		}
	}
	else
	{
		coord[0] = 1;
	}

	
	for(iRow=0;iRow<nRows;iRow++)
	{
		start = 0;
		for(i=0;i<ndims-1;i++)
		{
			start += (coord[i]-1)*factors[i];
		}
		
		switch ((int) type)
		{
		  case CHAR:
		  case MACRO:
		  case MACRO1:
			ch = STRING(symh);
			chstart = skipStrings(*ch,start) - *ch;
			for (coord[ndims - 1] = 1; coord[ndims - 1] <= lastdim;
				 coord[ndims - 1]++)
			{

				/* put out coord label */
				if(labels)
				{
					fmyprint("(",fp);
					for (j = 0; j < ndims - 1; j++)
					{
						sprintf(OUTSTR, "%ld,", coord[j]);
						fmyprint(OUTSTR,fp);
					}
					sprintf(OUTSTR, "%ld) ", coord[ndims - 1]);
					fmyprint(OUTSTR,fp);
				}
				
				OUTSTR[0] = '"';
				i = 1;
				do
				{
					c = (unsigned char) (*ch)[chstart++];
					if(c == '\0')
					{ /* terminate with quote */
						OUTSTR[i++] = '"';
						OUTSTR[i] = '\0';
						fmyprint(OUTSTR,fp);
					}
					else
					{
						if(c == '\\' || c == '"')
						{ /* escape '\' && '"' */
							OUTSTR[i++] = '\\';
						}
						if(c != '\n')
						{ /* put character in buffer */
							if(!PRINTABLE(c))
							{ /* output excaped octal */
								OUTSTR[i++] = '\\';
								OUTSTR[i++] = c/64 + '0';
								OUTSTR[i++] = ((c/8) & 0x07) + '0';
								OUTSTR[i++] = (c & 0x07) + '0';
							}
							else
							{
								OUTSTR[i++] = c;
							}
						} /*if(c != '\n')*/

						if(c == '\n' || i > BUFFERLENGTH-5)
						{
							OUTSTR[i] = '\0';
							fmyprint(OUTSTR,fp);
							i = 0;
							if(c == '\n')
							{
								fmyeol(fp);
								if(INTERRUPT != INTNOTSET)
								{
									break;
								}
							}						
						} /*if(c == '\n' || i > BUFFERLENGTH-5)*/
					} /*if(c == '\0'){...}else{...}*/
				} while(c != '\0');
				fmyeol(fp);
				if (INTERRUPT != INTNOTSET)
				{
					goto interruptExit;
				}
				if(coord[ndims-1] < lastdim)
				{
					chstart = skipStrings(*ch + chstart,nRows-1) - *ch;
				}
			} /* for (coord[ndims - 1] = 1; 
					coord[ndims - 1] <= lastdim;coord[ndims - 1]++) */
			*OUTSTR = '\0';
			break;

		  case LOGIC:
		  case REAL:
			db = DATA(symh);
			k =start;
			coord[ndims - 1] = 1;
			while(coord[ndims - 1] <= lastdim)
			{
				if(labels)
				{/* put out coord label */
					strcpy(OUTSTR,"(");
					for (j = 0; j < ndims - 1; j++)
					{
						sprintf(OUTSTR+strlen(OUTSTR), "%ld,", coord[j]);
					}
					sprintf(OUTSTR+strlen(OUTSTR), "%ld)", coord[ndims - 1]);
					margin = labelWidth - strlen(OUTSTR);
					while(margin > 0)
					{
						fmyprint(" ",fp);
						margin--;
					}
					fmyprint(OUTSTR,fp);
				}				

				j = 0;
				while (coord[ndims-1] <= lastdim && j < fieldsPerLine)
				{
					tmp = (*db)[k];
					if (type == LOGIC)
					{
						if (tmp == 1.0)
						{
							sprintf(OUTSTR, dataFmt, "T");
						}
						else if (tmp == 0.0)
						{
							sprintf(OUTSTR, dataFmt, "F");
						}
						else if(isMissing(tmp))
						{
							sprintf(OUTSTR, missingFmt, NAMEFORMISSING);
						}
						else
						{
							sprintf(OUTSTR, missingFmt, "BADVALUE");
						}
					} /*if (type == LOGIC)*/
					else
					{
						if (isMissing(tmp))
						{
							sprintf(OUTSTR, missingFmt, NAMEFORMISSING);
						}
						else
						{
							sprintf(OUTSTR, dataFmt, tmp);
						}
					} /*if (type == LOGIC){...}else{...}*/
					fmyprint(OUTSTR,fp);
					coord[ndims-1]++;
					j++;
					k += nRows;
				} /*while (coord[ndims-1] <= lastdim && j < fieldsPerLine)*/
				fmyeol(fp);
				if (INTERRUPT != INTNOTSET)
				{
					goto interruptExit;
				}
			} /*while(coord[ndims - 1] <= lastdim)*/
			break;
		  case PLOTINFO:
			do_whole_graph(GRAPH(symh), fp, DUMBTERM, -1, 1, 0, (char *) 0);
			break;
			
		} /* switch(type) */

		for(i=ndims-2;i>=0;i--)
		{
			coord[i]++;
			if(coord[i] <= DIMVAL(symh,i+1))
			{
				break;
			}
			coord[i] = 1;
		}
	} /*for(iRow=0;iRow<nRows;iRow++)*/

  interruptExit:
	*OUTSTR = '\0';
}

static void doprexpr(Symbolhandle symh, FILE *fp, long labels)
/* print out an expression */
{
	long            i, j, tmpdim;
	Symbolhandle    comp;
	char           *name;
	WHERE("doprexpr");
	
	if (symh == (Symbolhandle) 0)
	{
		fmyprint("(NULL)", fp);
		fmyeol(fp);
		
		return;
	}

	if (!isDefined(symh))
	{
		fmyprint("UNDEFINED",fp);
		fmyeol(fp);
		return;
	}
	if(DIMVAL(symh,1) == 0)
	{
		return;
	}
	
	if (TYPE(symh) == BLTIN)
	{
		fmyprint("Built-in function cannot be printed",fp);
		fmyeol(fp);
		return;
	}

	SETUPINT(interruptExit);

	IndentLevel += INDENTSTEP;

	switch ((int) TYPE(symh))
	{
	  case LIST:
	  case STRUC:
		tmpdim = DIMVAL(symh,1);
		for (i = 0; i < tmpdim; i++)
		{
			comp = COMPVALUE(symh, i);
			if(comp != (Symbolhandle) 0)
			{
				name = isscratch(NAME(comp)) ? NAME(comp) + 2 : NAME(comp);
				
				for(j=0;j<IndentLevel;j++)
				{
					OUTSTR[j] = ' ';
				}
				
				sprintf(OUTSTR+IndentLevel, "component: %s", name);
				fmyprint(OUTSTR,fp);
				fmyeol(fp);
				if(INTERRUPT != INTNOTSET)
				{
					goto interruptExit;
				}
				doprexpr(comp,fp,labels);
			} /*if(comp != (Symbolhandle) 0)*/
			else
			{
				fmyprint("component: (no data)", fp);
				fmyeol(fp);
			}
			
			if (INTERRUPT != INTNOTSET)
			{
				goto interruptExit;
			}
		} /*for (i = 0; i < tmpdim; i++)*/
		break;

	  case MACRO:
	  case MACRO1:
	  case CHAR:
	  case LOGIC:
	  case REAL:
	  case PLOTINFO:
		printSymbol(symh, fp,labels);
	} /*switch ((int) TYPE(symh))*/

	/* fall through */
  interruptExit:
	IndentLevel -= INDENTSTEP;
	RETURN;
	
} /*doprexpr()*/

void myprint(msg)
char           *msg;
{
	IndentLevel = 0;
	fmyprint(msg,STDOUT);

}

/*
  output a possible multiline string, using myerrorout for each line
  so that lines will be counted
*/

void mymultiprint(outstr)
char           *outstr;
{
	register char   *pc1 = outstr, *pc2;
	
	for(pc2 = pc1;*pc2 != '\0';pc2++)
	{
		if(*pc2 == '\n')
		{
			*pc2 = '\0';
			myerrorout(pc1);
			if(INTERRUPT != INTNOTSET)
			{
				goto interruptExit;
			}
			*pc2 = '\n';
			pc1 = pc2 + 1;
		}
	} /*for(pc2 = pc1;*pc2 != '\0';pc2++)*/
	if(pc1 < pc2)
	{
		myerrorout(pc1);
	}

	/* fall through */
  interruptExit:
	return;
} /*mymultiprint()*/

		
void myeol()
{
	fmyeol(STDOUT);
}

void myerrorout(msg)
char           *msg;
{
	if(PRINTWARNINGS || strncmp("WARNING:", msg, 8) != 0)
	{
		myprint(msg);
		myeol();
	}
} /*myerrorout()*/

/*
   function to output line with up to 4 pieces.  Particularly designed to
   use when outputting error messages concerning file names to avoid
   overflowing OUTSTR
*/
void putPieces(char * piece1, char * piece2, char * piece3, char * piece4)
{
	char      *pieces[4];
	short      i;
	
	pieces[0] = piece1;
	pieces[1] = piece2;
	pieces[2] = piece3;
	pieces[3] = piece4;

	for (i=0; i < 4 && pieces[i] != (char *) 0; i++)
	{
		myprint(pieces[i]);
		if (pieces[i] == OUTSTR)
		{
			OUTSTR[0] = '\0';
		}
	}
	myeol();
} /*myprintpieces()*/

/*
  echo contents of *LINE (which usually will terminate with '\n') to output,
  taking care that it will be counted appropriately
*/

void echoLINE(void)
{
	char      *pc = *LINE + strlen(*LINE) - 1;
	char       c = *pc;

	if(pc >= *LINE && (c == '\n' || c == '\r'))
	{
		*pc = '\0';
	}
	myerrorout(*LINE);
	if(pc >= *LINE)
	{
		*pc = c;
	}
}

void putOUTSTR(void)
{
	if(*OUTSTR)
	{
		myerrorout(OUTSTR);
		*OUTSTR = '\0';
	}
}

int myfeof(fp)
FILE          *fp;
{
	return ((fp != STDOUT) ? feof(fp) : 0);
}

void prexpr(symh)
Symbolhandle    symh;
{
	IndentLevel = -INDENTSTEP;
	doprexpr(symh,STDOUT,1L);
}

void fprexpr(symh,fp, labels)
Symbolhandle    symh;
FILE           *fp;
long            labels;
{
	IndentLevel = -INDENTSTEP;
	doprexpr(symh,fp,labels);
}

#define LINEWIDTH 132

void matWrite(arg,fp, missValue,header,separator)
Symbolhandle       arg;
FILE              *fp;
double             missValue;
long               header;
char              *separator;
{
	char          *dataFmt = DATAFMT, *missingFmt = DATAFMT;
	char          *ch, *ch1;
	char          *name = NAME(arg);
	long           m, n, i, j, k, l, size;
	long           ndims;
	long           nperline;
	long           foundMissing;
	long           incs[MAXDIMS];
	long           ii[MAXDIMS];
	long           place[MAXDIMS];
	double         y;
	double       **yhandle;
	WHERE("matWrite");

	ndims = NDIMS(arg);
	size = symbolSize(arg);
	m = DIMVAL(arg,ndims);

	if((foundMissing = anyMissing(arg)))
	{
		sprintf(OUTSTR,missingFmt,missValue);
		sscanf(OUTSTR,"%lf",&y);
		if(y != missValue)
		{
			missingFmt = (isMissing(missValue)) ? " %.4f" : " %.15g";
		}
	}
	
	nperline = (LINEWIDTH-1)/FIELDWIDTH;
	if(nperline > m)
	{
		nperline = m;
	}	
/* write header */
	if(isscratch(name) || isKeyword(arg))
	{
		name += 2;
	}
	
	if(header)
	{
		sprintf(OUTSTR,"%-12s",name);
		fmyprint(OUTSTR,fp);
		for(i=1;i<=ndims;i++)
		{
			sprintf(OUTSTR," %5ld",DIMVAL(arg,i));
			fmyprint(OUTSTR,fp);
		}
		fmyeol(fp);
		if(INTERRUPT != INTNOTSET)
		{
			goto interruptExit;
		}
	
		if(foundMissing)
		{
			fmyprint(") MISSING:",fp);
			sprintf(OUTSTR,missingFmt,missValue);
			fmyprint(OUTSTR,fp);
			fmyeol(fp);
			if(INTERRUPT != INTNOTSET)
			{
				goto interruptExit;
			}
		}
		if(TYPE(arg) == LOGIC)
		{
			fmyprint(") LOGICAL",fp);
			fmyeol(fp);
			if(INTERRUPT != INTNOTSET)
			{
				goto interruptExit;
			}
		}
		
		fmyprint(")\"%lf",fp);
		for(i=1;i<nperline;i++)
		{
			fmyprint(" %lf",fp);
		}
		fmyprint("\"",fp);
		fmyeol(fp);
		if(INTERRUPT != INTNOTSET)
		{
			goto interruptExit;
		}
	} /*if(header)*/
	incs[0] = size;
	for(j = 1;j<=ndims;j++)
	{
		ii[j-1] = 0;
		place[j-1] = 0;
		incs[j] = incs[j-1]/DIMVAL(arg,ndims-j+1);
	}

	yhandle = DATA(arg);
	
	/* print out in transposed order */

	n = 0; /* number of items already printed */
	l = 0; /* position in array of element being printed */
	k = 0; /* position in output line */
	while(n<size)
	{
		y = (*yhandle)[l];
		if(!isMissing(y))
		{
			sprintf(OUTSTR,dataFmt,y);
		}
		else
		{
			sprintf(OUTSTR,missingFmt,missValue);
		}
/* when seperator not null, trim leading and trailing blanks */
		if(*separator)
		{
			for(ch = OUTSTR;*ch == ' ';ch++)
			{
				;
			}
			for(ch1 = ch;*ch1 != ' ' && *ch1 != '\0';ch1++)
			{
				;
			}
			*ch1 = '\0';
		}
		else
		{
			ch = OUTSTR;
		}
		
		fmyprint(ch,fp);
		n++;
		k++;
		if(n % m == 0 || k % nperline == 0)
		{
			fmyeol(fp);
			if(INTERRUPT != INTNOTSET)
			{
				goto interruptExit;
			}
			k = 0;
		}
		else if(*separator)
		{
			fmyprint(separator,fp);
		}
/* the following is essentially equivalent to ndims nested for loops */
		for (j = 1; j <= ndims; j++)
		{
			if(ii[j-1] == 0)
			{
				place[j-1] = l;
			}
			if((ii[j-1] += incs[j]) < incs[j-1])
			{
				l += incs[j];
				break;
			}
			ii[j-1] = 0;
			l = place[j-1];
		} /*for (j = 1; j <= ndims; j++)*/
	} /*while(n<size)*/
	/* fall through */
  interruptExit:
	*OUTSTR = '\0';
} /*matWrite()*/

void macWrite(arg,fp,header)
Symbolhandle       arg;
FILE              *fp;
long               header;
{
	char          *name = NAME(arg);
	long           m, i;
	long           c, length, place = 0;
	WHERE("macWrite");
	
	m = 1;
	i = 0;
	while((c = STRINGVALUE(arg,i)))
	{
		if(c == '\n' || c == '\r')
		{
			m++;
		}
		i++;
	}
	
	if(header)
	{
/* write header */
		if(isscratch(name) || isKeyword(arg))
		{
			name += 2;
		}
		sprintf(OUTSTR,"%-12s %5ld %s",name,m,typeName(TYPE(arg)));
		fmyprint(OUTSTR,fp);
		fmyeol(fp);
		if(INTERRUPT != INTNOTSET)
		{
			goto interruptExit;
		}
	}
	length = strlen(STRINGPTR(arg));
	OUTSTR[BUFFERLENGTH-1] = '\0';
	while(length > 0)
	{
		strncpy(OUTSTR,STRINGPTR(arg)+place,BUFFERLENGTH-1);
		fmyprint(OUTSTR,fp);
		place += BUFFERLENGTH-1;
		length -= BUFFERLENGTH-1;
	}
	fmyeol(fp);
	/* fall through*/

  interruptExit:
	*OUTSTR = '\0';

} /*macWrite()*/
