/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Concat
#else
#define __SEG__ Concat
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

extern char          **mygethandle();
extern void            badType();
extern long            argOK();
extern long            symbolSize();

static long doConcat(Symbolhandle result, Symbolhandle symh,long place)
{
	int              type = TYPE(result);
	long             i, ncomps;
	WHERE("doConcat");
	
	if(!isNull(symh))
	{
		if(TYPE(symh) == STRUC || TYPE(symh) == LIST)
		{
			ncomps = NCOMPS(symh);
			for(i=0;i<ncomps;i++)
			{
				place = doConcat(result,COMPVALUE(symh,i),place);
			}
		} /*if(TYPE(symh) == STRUC || TYPE(symh) == LIST)*/
		else
		{			
			if(type == REAL || type == LOGIC)
			{
				register double *x = DATAPTR(symh);
				register double *y = DATAPTR(result) + place;
				register long    n = symbolSize(symh);

				place += n;
		
				while(n-- > 0)
				{
					*y++ = *x++;
				}
			} /*if(type == REAL || type == LOGIC)*/
			else
			{/* type == CHAR */
				register char   *cx = STRINGPTR(symh);
				register char   *cy = STRINGPTR(result) + place;
				register long    nc = myhandlelength(STRING(symh));

				place += nc;

				while(nc-- > 0)
				{
					*cy++ = *cx++;
				}
			} /*if(type == REAL || type == LOGIC){}else{}*/
		} /*if(TYPE(symh) == STRUC || TYPE(symh) == LIST){}else{}*/
	} /*if(!isNull(symh))*/
	
	return (place);
}

/* concatenate the elements in list */

Symbolhandle    concat(list)
Symbolhandle    list;
{
	long            type, type1;
	long            i, tot, ctot;
	long            nargs = NARGS(list);
	long            foundData = 0, foundNull = 0;
	char           *name;
	Symbolhandle    result = (Symbolhandle) 0, symh;
	WHERE("concat");

	*OUTSTR = '\0';
	
	for(i=0;i<nargs;i++)
	{
		symh = COMPVALUE(list, i);
		if(!argOK(symh,(long) 0,(nargs > 1) ? i+1 : 0))
		{
			goto errorExit;
		}
		type1 = TYPE(symh);
		if(type1 != REAL && type1 != CHAR && type1 != LOGIC && 
		   type1 != STRUC)
		{
			badType(FUNCNAME,-type1,(nargs > 1) ? i+1 : 0);
			goto errorExit;
		}
		
		if(type1 == STRUC)
		{
			type1 = getSingleType(symh);
			if(type1 == 0)
			{
				sprintf(OUTSTR,
						"ERROR: not all components of a structure argument to %s have same type",
						FUNCNAME);
				goto errorExit;
			}
			if(type1 > 0 && type1 != REAL && type1 != LOGIC && type1 != CHAR)
			{
				sprintf(OUTSTR,
						"ERROR: structure components of type %s are illegal for %s()",
						typeName(type1), FUNCNAME);
				goto errorExit;
			}
			foundNull = foundNull || strucAnyNull(symh);
		} /*if(type1 == STRUC)*/

		if(type1 > 0)
		{			
			/*
			  If symh is a structure, at this point we know that
			  all components have type type1, REAL, CHAR, or LOGIC
			  */
			if(!foundData)
			{
				foundData = 1;
				type = type1;
			} /*if(!foundData)*/
			else
			{
				if(type1 != type)
				{
					sprintf(OUTSTR,
							"ERROR: not all arguments to %s have the same type",
							FUNCNAME);
					goto errorExit;
				}
			} /*if(!foundData){}else{}*/
		} /*if(type1 > 0)*/
	} /*for(i=0;i<nargs;i++)*/
	
	if(!foundData)
	{
		sprintf(OUTSTR,
				"ERROR: all arguments to %s are structures with all null components",
				FUNCNAME);
		goto errorExit;
	}
	if(foundNull)
	{
		sprintf(OUTSTR,
				"WARNING: null structure component found in an argument to %s",
				FUNCNAME);
		putOUTSTR();
	}
	
	tot = strucSymbolSize(list);

	if (tot > MAXCOORD)
	{ /* too many items for short coordinate */
		sprintf(OUTSTR,"ERROR: output vector from %s too long", FUNCNAME);
		goto errorExit;
	}

	if (nargs == 1 && TYPE(symh) != STRUC && isscratch(NAME(symh)))
	{
		result = symh;
		setNDIMS(symh,1);
		setDIM(symh,1,tot);
		if(tot == 1)
		{
			if(type == CHAR)
			{
				name = STRINGSCRATCH;
			}
			else if(type == REAL)
			{
				name = NUMSCRATCH;
			}
			else
			{
				name = LOGICSCRATCH;
			}
		}
		else
		{
			name = VECSCRATCH;
		}
		
		setNAME(symh, name);
		COMPVALUE(list,0) = (Symbolhandle) 0;
	} /*if (nargs == 1 && TYPE(symh) != STRUC && isscratch(NAME(symh)))*/
	else
	{
		if (type == REAL || type == LOGIC)
		{
			result = RInstall(SCRATCH,tot);
			if (result == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setTYPE(result,type);
		} /*if (type == REAL || type == LOGIC)*/
		else
		{						/* type == CHAR */
			result = Install(SCRATCH,CHAR);
			if(result == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNDIMS(result,1);
			setDIM(result,1,tot);
			ctot = strucSize(list);

			TMPHANDLE = mygethandle(ctot);
			setSTRING(result,TMPHANDLE);
			if(TMPHANDLE == (char **) 0)
			{
				goto errorExit;
			}
		} /*if (type == REAL || type == LOGIC){}else{}*/

		(void) doConcat(result,list,0);
	}
	
	return (result);

  errorExit:
	putOUTSTR();

	Removesymbol(result);
	
	return (0);
	
}
