/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Contrast
#else
#define __SEG__ Contrast
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <math.h>
#include <string.h>

extern Symbolhandle    RSInstall();
extern Symbolhandle    Makereal();
extern long            isScalar();
extern long            argOK();

extern double	       ROBUSTVARK;
extern double	       CURRENTRSC;

#define Trash   NAMEFORTRASH
static Symbolhandle Trash = (Symbolhandle) 0; /* repository for scratch */

/*#define x(i,j) ((*REGX)[(long) (i) + (long) (j) * ndata ])*/

#define FUZZ     1.0e-7 /* allowed departure of sum(wts) from zero */

#define GGROUPCTS    0
#define GCONVEC      1
#define NTRASH       2

static double **Convec, **Concoef;
static long     Conlen, Convars[MAXVARS], Nconvars;

/* check columns to see how to build coefs for contrast */
static void buildConvec(long termi, long *colno)
{
	long            add, i, i2, j, term[MAXVARS], term2[MAXVARS];
	modelType       trialterm;
	modelPointer    prevterm;
	long            nvars = (long) NVARS;
	long            icombo, ncombos;
	double         *convec = *Convec, *concoef = *Concoef;

	getterm(modelterm(MODEL,termi), term, nvars);

	ncombos = setterm(term, nvars, NCLASSES);

	for(icombo = 0;icombo < ncombos;icombo++)
	{/* check to see if each combination is already in through lower order terms */
		effectiveTerm(term,(modelPointer) trialterm, nvars);
	
		/*  trialterm is now effective term for this potential column */
		for (i = 0; i < termi; i++)
		{
			prevterm = modelterm(MODEL,i);
			if(inEffectiveTerm(prevterm,(modelPointer) trialterm))
			{/* if effective term is in earlier term */
				/* make sure any variates in earlier term are in eff term */
				getterm(prevterm, term2, nvars);
				for (i2 = 0; i2 < nvars; i2++)
				{
					if (term[i2] >= 0 && term2[i2] > 0 && NCLASSES[i2] < 0)
					{
						break;
					}
					/* break out if variate in earlier term not here */
					/* will keep this effective term */
				} /*for (i2 = 0; i2 < nvars; i2++)*/
				if (i2 >= nvars)
				{
					break;
				}
			} /*if(inEffectiveTerm(prevterm,(modelPointer) trialterm))*/
		} /*for (i = 0; i < termi; i++)*/

		if(i == termi)
		{/* if here, effective term needs to be entered */
			/* 
			  Now figure which terms of Concoef go into this element of Convec
			  loop through all elements of Concoef and add in appropriate
			  coefs
			*/
			for (i = 0; i < Nconvars; i++)
			{
				term2[i] = 1;
			}
			
			for (i = 0; i < Conlen; i++)
			{
				add = 1;
				for (j = 0; j < Nconvars; j++)
				{
					if (term[Convars[j] - 1] == 1 ||
						term2[j] + 1 == term[Convars[j] - 1])
					{
						add *= 1; /* a 1 always matches  or offset 1 matches*/
					}
					else if (term2[j] == NCLASSES[Convars[j] - 1])
					{
						add *= -1; /* tops get taken off */
					}
					else
					{
						add = 0;
						break;
					}
				}
				convec[*colno] += add * concoef[i];

				/* step odometer */
				for (j = 0; j < Nconvars; j++)
				{
					term2[j]++;
					if (term2[j] <= NCLASSES[Convars[j] - 1])
					{
						break;
					}
					term2[j]  = 1;
				} /*for (j = 0; j < Nconvars; j++)*/
			} /*for (i = 0; i < Conlen; i++)*/
			(*colno)++;
		} /*if(i == termi)*/

		/* step odometer */
		for(i=nvars-1;i>=0;i--)
		{
			if(term[i] > 0)
			{
				term[i]++;
				if(term[i] <= NCLASSES[i])
				{
					break;
				}
				term[i] = 1;
			}
		} /*for(i=nvars-1;i>=0;i--)*/
	} /*for(icombo = 0;icombo < ncombos;icombo++)*/
}

static long getconentry(long i, long convars[], long nconvars)
{
	long            j, k, step;

	step = 0;
	for (j = nconvars; j > 0; j--)
	{
		k = convars[j - 1] - 1;
		step = ((long) (*X[k])[i]) - 1 + step * NCLASSES[k];
	}
	return (step);
}

static void do_balanced(Symbolhandle symhcon, Symbolhandle symhest,
						Symbolhandle symhss, Symbolhandle symhse,
						long errorterm,
						long byvar, long do_ses, long nbyvar)
{
	double             *concoef = DATAPTR(symhcon);
	double             *conest = DATAPTR(symhest);
	double             *conss = DATAPTR(symhss);
	double             *conse = DATAPTR(symhse);
	double             *sserror = *SS + errorterm;
	double              dferror = (*DF)[errorterm];
	double             *y = *Y;
	double             *byvariable;
	double              tmp, mse, con, denom;
	long                ibyvar, i, ndata = (long) NDATA; 
	WHERE("do_balanced");

	if(byvar > 0)
	{
		byvariable = *X[byvar - 1];
	}

	if(do_ses > 0)
	{
		mse = *sserror/dferror;
	}

	for (ibyvar = 0; ibyvar < nbyvar; ibyvar++)
	{
		con = 0.;
		denom = 0.;
		for (i = 0; i < ndata; i++)
		{
			if (byvar < 1 || byvariable[i] == ibyvar + 1)
			{
				tmp = concoef[getconentry(i, Convars, Nconvars)];
				con += y[i] * tmp;
				denom += tmp * tmp;
			}
		} /*for (i = 0; i < ndata; i++)*/
		tmp = (double) (Conlen * nbyvar) / NDATA;
		conest[ibyvar] = tmp * con;
		conss[ibyvar] = con * con / denom;
		if(do_ses > 0)
		{
			conse[ibyvar] = tmp*sqrt(mse*denom);
		}
		else
		{
			setMissing(conse[ibyvar]);
		}
	} /*for (ibyvar = 0; ibyvar < nbyvar; ibyvar++)*/
} /*do_balanced()*/

	/* routine to compute contrast SS for a given contrast */

Symbolhandle    contrast(list)
Symbolhandle    list;
{

	Symbolhandle    symhname, symhcon, symhby, symhkey;
	Symbolhandle    symhest = (Symbolhandle) 0, symhss = (Symbolhandle) 0;
	Symbolhandle    symhse = (Symbolhandle) 0;
	Symbolhandle    result = (Symbolhandle) 0;
	double         *conss, *conse, *conest;
	long            contermvec[MAXVARS];
	modelType       conterm;
	long            nbyvar, byvar, firstcol, lastcol, termi, type;
	long            i, j, j1, j2, k, inck, ij, ik, jk, kk, kkinc, j1j2;
	long            k1, k2, ik1, ik2, k1k2, ik1k2;
	long            ibyvar;
	long            tmpi, dims[2];
	long            nargs = NARGS(list), nvars = (long) NVARS;
	long            ny = (long) NY, ndata = (long) NDATA, nregx = (long) NREGX;
	long            nterms = (long) NTERMS, errorterm = nterms;
	long            do_ses = 1;
	double          con, denom, tmp, singchk, mse, dferror;
	double         *y, *sserror, *regcoef, *regss, *xtxinv, *wtdresiduals;
	double         *allwts, *misswts, *casewts;
	double         *convec, *conveci, *convecj, *concoef;
	double        **groupctsH, *groupcts;
	double         *byvariable, *xij;
	double          beta = .35;
	char            varname[NAMELENGTH+1], *endvarname;
	char           *name, *compname[3], *keyword, *pc;
	char           *usage =
		"usage: contrast(term, contrast [, byvar] [,error:termName])";
	WHERE("contrast");
	
	Trash = (Symbolhandle) 0;
	compname[0] = "estimate";
	compname[1] = "ss";
	compname[2] = "se";
	OUTSTR[0] = '\0';
	
	if (PREVMODELTYPE == 0 || PREVMODELTYPE & (LEAPS + OLSREG))
	{
		sprintf(OUTSTR,"ERROR: no active ANOVA model");
	}
	else if (PREVMODELTYPE & (FASTANOVA | IPF))
	{
		sprintf(OUTSTR,"ERROR: %s() does not work with %s()",
				FUNCNAME, (PREVMODELTYPE & FASTANOVA) ? "fastanova" : "ipf");
	}
	else if (!(PREVMODELTYPE & (BALANOVA | DOCOEFS)))
	{
		sprintf(OUTSTR,
				"ERROR: %s() does not work after GLM command with coefs:F",
				FUNCNAME);
	}
	else if(nargs < 2)
	{
		sprintf(OUTSTR,"ERROR: %s",usage);
	}
	else if(!argOK(symhname = COMPVALUE(list, 0),0, 1) || 
			!argOK(symhcon = COMPVALUE(list, 1),REAL, 2))
	{
		strcpy(OUTSTR,usage);
	}
	else if(!isScalar(symhname))
	{
		sprintf(OUTSTR,"ERROR: first argument must be term name or number");
		putOUTSTR();
		strcpy(OUTSTR,usage);
	}
	else if(TYPE(symhname) != CHAR && TYPE(symhname) != REAL)
	{
		badType(FUNCNAME,-TYPE(symhname),1);
		goto errorExit;
	}
	
	if(*OUTSTR)
	{
		goto errorExit;
	}
	
	if(nargs > 2)
	{ /* check for error:termName */
		symhkey = COMPVALUE(list,nargs-1);
		if(symhkey != (Symbolhandle) 0 && (keyword = isKeyword(symhkey)) &&
		   strncmp(keyword,"error",5) == 0)
		{ /* error term specified for computing standard errors */
			if(PREVMODELTYPE & ITERWEIGHTS)
			{
				sprintf(OUTSTR,
						"ERROR: keyword %s not legal for poisson(), logistic() or robust()",
						keyword);
				goto errorExit;
			}
			else
			{
				if(!isScalar(symhkey) ||
				   (type = TYPE(symhkey)) != CHAR && type != REAL)
				{
					sprintf(OUTSTR,
							"ERROR: value for %s not CHARACTER variable or string or term number",
							keyword);
					goto errorExit;
				}
				if(type == REAL)
				{
					tmp = DATAVALUE(symhkey,0);
					if(tmp != floor(tmp) || tmp <= 0.0 ||
					   tmp > (double) (nterms+1))
					{
						sprintf(OUTSTR,
								"ERROR: %s error term number must be integer between 1 and %ld",
								FUNCNAME, nterms+1);
						goto errorExit;
					}
					errorterm = (long) tmp - 1;
				}
				else					
				{
					name = *TERMNAMES;
					for(errorterm = 0;errorterm <= nterms;errorterm++)
					{
						if(strcmp(name,STRINGPTR(symhkey)) == 0)
						{
							break;
						}
						name += strlen(name) + 1;
					}

					if(errorterm > nterms)
					{
						sprintf(OUTSTR,
								"ERROR: %s is not a term in the model %s",
								STRINGPTR(symhkey),STRMODEL);
						goto errorExit;
					}
				}
			}
			nargs--;			
		}
	} /*if(nargs > 2)*/
	
	type = TYPE(symhname);
	if(type == REAL)
	{
		tmp = DATAVALUE(symhname,0);
		if(tmp != floor(tmp) || tmp <= 0.0 || tmp > (double) nterms)
		{
			sprintf(OUTSTR,
					"ERROR: term number for %s must be integer between 1 and %ls",
					FUNCNAME, nterms);
			goto errorExit;
		}
		name = skipStrings(*TERMNAMES, (long) tmp - 1);
	}
	else
	{
		name = STRINGPTR(symhname);
	}

	pc = name;
	Nconvars = 1;		/* get number of variables in contrast */
	while ((pc = strchr(pc, '.')) != (char *) 0)
	{
		Nconvars++;
		pc++;
	}

	for (j = 0; j < MAXVARS; j++)
	{
		contermvec[j] = 0;
	}
	
	zeroTerm(conterm);
	pc = name;
	for (k = 0; k < Nconvars; k++)
	{
		strncpy(varname,pc,NAMELENGTH);
		varname[NAMELENGTH] = '\0';
		endvarname = strchr(varname, '.');
		if(endvarname != (char *) 0)
		{
			*endvarname = '\0';
		}
		
		for (j = 1; j <= nvars; j++)
		{
			if (strcmp(varname, VARNAMES[j]) == 0)
			{
				if (contermvec[j - 1] != 0)
				{
					sprintf(OUTSTR,
							"ERROR: variable %s used more than once in %s",
							VARNAMES[j], name);
				}
				else if(NCLASSES[j - 1] < 0)
				{
					sprintf(OUTSTR,"ERROR: variable %s is not a factor",
							VARNAMES[j]);
				}
				if(*OUTSTR)
				{
					goto errorExit;
				}
				Convars[k] = j;
				contermvec[j - 1] = 2;
				break;
			}
		} /*for (j = 1; j <= nvars; j++)*/

		if (j > nvars)
		{
			sprintf(OUTSTR,"ERROR: contrast variable %s not in model", 
					varname);
			goto errorExit;
		} /*if (j > nvars)*/

		pc = strchr(pc, '.') + 1;

	} /*for (k = 0; k < Nconvars; k++)*/

	effectiveTerm(contermvec,(modelPointer) conterm,nvars);

	if (Nconvars == 1 && !isVector(symhcon) ||
		Nconvars == 2 && !isMatrix(symhcon,dims) ||
		Nconvars > 2 && NDIMS(symhcon) != Nconvars)
	{
		sprintf(OUTSTR,
				"ERROR: dimensions of contrast coefficients do not match number of factors");
		goto errorExit;
	}

	Conlen = 1;
	for (k = 0; k < Nconvars; k++)
	{
		if (DIMVAL(symhcon,k + 1) != NCLASSES[Convars[k] - 1])
		{
			sprintf(OUTSTR,	
				"ERROR: contrast coefficients do not match dimensions of term %s",
					name);
			goto errorExit;
		}
		Conlen *= NCLASSES[Convars[k] - 1];
	} /*for (k = 0; k < Nconvars; k++)*/

	con = 0.0;		/* check for adding to zero */
	concoef = DATAPTR(symhcon);
	for (i = 0; i < Conlen; i++)
	{
		con += *concoef++;
	}
	if (fabs(con) > FUZZ)
	{
		sprintf(OUTSTR,"ERROR: coefficients add to %g != 0");
		goto errorExit;
	}

	byvar = 0;
	nbyvar = 1;
	if (nargs > 2)
	{
		symhby = COMPVALUE(list, 2);
		if( !argOK(symhby, CHAR, 3))
		{
			strcpy(OUTSTR,usage);
		}
		else if(!isScalar(symhby))
		{
			sprintf(OUTSTR,"ERROR: %s",usage);
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
		
		name = STRINGPTR(symhby);

		for (j = 1; j <= nvars; j++)
		{
			if (strcmp(name, VARNAMES[j]) == 0)
			{
				byvar = j;
				break;
			}
		}
		if (j > nvars)
		{
			sprintf(OUTSTR,"ERROR: byvar variable %s not in model", name);
		}
		else if (contermvec[j - 1] > 0)
		{
			sprintf(OUTSTR,"ERROR: byvar variable %s is in contrast term",
					name);
		}
		else if ((nbyvar = NCLASSES[j - 1]) < 1)
		{
			sprintf(OUTSTR,"ERROR: byvar variable %s is not a factor", name);
		}
		else if (PREVMODELTYPE & (ITERWEIGHTS))
		{
			sprintf(OUTSTR,
					"ERROR: cannot use a byvar variable after poisson(), logistic(), or robust()");
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
	} /* if(nargs > 2) */

	dferror = (*DF)[errorterm];
	
	if (PREVMODELTYPE & (IPF | FASTANOVA))
	{
		do_ses = 0;
	}
	else if(PREVMODELTYPE & ITERWEIGHTS)
	{
		if(!(PREVMODELTYPE & ROBUSTREG))
		{
			myerrorout("NOTE: Deviances based on full model weights");
			myerrorout("      Standard errors assume scale parameter is 1");
		}
	}
	else if(dferror <= 0.0)
	{
		myerrorout("WARNING: no degrees of freedom for error; standard errors set to MISSING");
		do_ses = -1;
	}

	/* ny should be > 1 only for MANOVA */
	result = RSInstall(SCRATCH, (do_ses) ? 3L : 2L,
					   compname, (ny == 1) ? nbyvar : 0L);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	if(ny == 1)
	{
		symhest = COMPVALUE(result, 0);
		symhss = COMPVALUE(result, 1);
		if(do_ses)
		{
			symhse = COMPVALUE(result, 2);
		}
	} /*if(ny == 1)*/
	else
	{
		symhest = COMPVALUE(result,0) = Makereal(nbyvar * ny);
		symhss = COMPVALUE(result,1) = Makereal(nbyvar*ny*ny);
		if(symhest == (Symbolhandle) 0 || symhss == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNAME(symhest,compname[0]);
		setNDIMS(symhest,2);
		setDIM(symhest,1,nbyvar);
		setDIM(symhest,2,ny);
		
		setNAME(symhss,compname[1]);
		setNDIMS(symhss,3);
		setDIM(symhss,1,nbyvar);
		setDIM(symhss,2,ny);
		setDIM(symhss,3,ny);

		if(do_ses)
		{
			symhse = COMPVALUE(result,2) = Makereal(nbyvar * ny);
			if(symhse == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symhse,compname[2]);
			setNDIMS(symhse,2);
			setDIM(symhse,1,nbyvar);
			setDIM(symhse,2,ny);
		} /* if(do_ses) */
	} /*if(ny == 1){...}else{...}*/

	sserror = *SS + errorterm;
	if(PREVMODELTYPE & ITERWEIGHTS)
	{
		setNAME(symhss, "deviance");
	}
	
	if (PREVMODELTYPE & BALANOVA)
	{			/* balanced */
		do_balanced(symhcon,symhest,symhss,symhse,
					errorterm, byvar, do_ses, nbyvar);
		return (result);
	} /*if (PREVMODELTYPE & BALANOVA)*/

	/* unbalanced if here */
	Trash = GarbInstall(NTRASH);
	if(Trash == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	if(do_ses > 0 && ny == 1)
	{
		if(PREVMODELTYPE & ROBUSTREG)
		{
			mse = ROBUSTVARK * CURRENTRSC;
			mse = mse * mse * beta;
			sprintf(OUTSTR,
					"Robust deviance / %g approximately chi-squared on 1 d.f.",
					mse);
			putOUTSTR();
		}
		else if(!(PREVMODELTYPE & ITERWEIGHTS))
		{
			mse = *sserror/dferror;
		}
		else
		{
			mse = 1.0;
		}
	} /*if(do_ses > 0 && ny == 1)*/

	/* first determine if we do additive or subtractive SS */
	for (termi = 0; termi < nterms; termi++)
	{
		if(modeltermEqual(modelterm(MODEL,termi),(modelPointer) conterm))
		{
			break;
		}
	}

	if (termi < nterms && nbyvar == 1)
	{			/*  conterm in model, so subtractive */
		/* approach is to build up a set of coefs for the params estimated*/
		Concoef = DATA(symhcon);

		/* make sure that the contrast conforms to anova model variable order*/
		i = 1;
		while (i < Nconvars)
		{
			if (Convars[i] < Convars[i - 1])
			{
				myerrorout("ERROR:  contrast term variables not in anova model order");
				goto errorExit;
			}
			i++;
		}

		/* find first and last columns of term coefficients */
		firstcol = 0;
		i = 0;
		while (i < termi)
		{
			firstcol += buildXmatrix(i, (double *) 0, MODELINFO);
			i++;
		}
		lastcol = firstcol + buildXmatrix(termi, (double *) 0, MODELINFO);
		lastcol--;

		/* get space for modified coefs */

		if(!getScratch(Convec,GCONVEC,lastcol - firstcol + 1,double))
		{
			goto errorExit;
		}

		convec = *Convec;
		regss = *REGSS;
		regcoef = *REGCOEF;
		xtxinv = *XTXINV;
		sserror = *SS + errorterm;
		y = *Y;
		conest = DATAPTR(symhest);
		conss = DATAPTR(symhss);
		if(do_ses)
		{
			conse = DATAPTR(symhse);
		}
		
		for (j = 0; j < lastcol - firstcol + 1; j++)
		{
			convec[j] = 0.;
		}

		i = 0;
		buildConvec(termi, &i);

		/* check if missing (pivoted) column has nonzero coef */
		tmpi = 0;
		for (j = firstcol; j <= lastcol; j++)
		{
			if (regss[j] < -0.5 && convec[tmpi] != 0.0)
			{
				myerrorout("ERROR:  pivoted df has nonzero contrast coefficient");
				goto errorExit;
			}
			tmpi++;
		} /*for (j = firstcol; j <= lastcol; j++)*/

		/* N.B.: nbyvar is 1 here */
		denom = 0.;
		conveci = convec;
		for (j1 = firstcol; j1 <= lastcol; j1++)
		{
			j1j2 = j1 + firstcol * nregx; /* j1 + j2*nregx */
			convecj = convec;
			for (j2 = firstcol; j2 <= lastcol; j2++)
			{
				denom += *conveci * *convecj++ * xtxinv[j1j2];
				j1j2 += nregx;
			}
			conveci++;
		}

		if(do_ses > 0 && ny > 1 && !(PREVMODELTYPE & ITERWEIGHTS))
		{
			kk = 0;				/* k*(nterms+1)*(ny+1) */
			kkinc = (nterms+1)*(ny+1);
		}
		else
		{
			kkinc = 0;
		}
		
		for (k = 0; k < ny; k++)
		{
			con = 0.;
			tmpi = 0;
			jk = firstcol + k*nregx; /* i + k*nregx */
			for (j = firstcol; j <= lastcol; j++)
			{
				if(regss[j] > -0.5)
				{
					con += convec[tmpi] * regcoef[jk];
				}
				tmpi++;
				jk++;
			} /*for (j = firstcol; j <= lastcol; j++)*/
			conest[k] = con;
			if(do_ses)
			{
				if(kkinc != 0)
				{
					mse = sserror[kk]/dferror;
					kk += kkinc;
				}
				if(do_ses > 0)
				{
					conse[k] = sqrt(denom*mse);
				}
				else
				{
					setMissing(conse[k]);
				}
			}
		} /* for (k=0;k<ny;k++) */

		for (k1 = 0; k1 < ny; k1++)
		{
			k1k2 = k1; /* k1 + ny * k2 */
			for (k2 = 0; k2 < ny; k2++)
			{
				conss[k1k2] = conest[k1] * conest[k2] / denom;
				k1k2 += ny;
			}
		}
		emptyTrash();
		
		return (result);
	} /*if (termi < nterms && nbyvar == 1)*/

	/* termi >= NTERMS || nbyvar > 1 */
	/* conterm not in model or byvar specified, so additive */

	if(!getScratch(Convec,GCONVEC,ndata,double) ||
	   !getScratch(groupctsH,GGROUPCTS,Conlen,double))
	{
		goto errorExit;
	}
	convec = *Convec;
	sserror = *SS + errorterm;
	y = *Y;
	groupcts = *groupctsH;
	if(byvar > 0)
	{
		byvariable = *X[byvar - 1];
	}

	conest = DATAPTR(symhest);
	conss = DATAPTR(symhss);
	concoef = DATAPTR(symhcon);
	if(do_ses)
	{
		conse = DATAPTR(symhse);
	}

	if(PREVMODELTYPE & USEWEIGHTS)
	{
		allwts = *ALLWTS;
	}
	if(PREVMODELTYPE & MISSWEIGHTS)
	{
		misswts = *MISSWTS;
	}
	if(PREVMODELTYPE & CASEWEIGHTS)
	{
		casewts = *CASEWTS;
	}
	if (PREVMODELTYPE & ITERWEIGHTS)
	{
		wtdresiduals = *WTDRESIDUALS;
	}

	/* repeat for each level of by variable */
	for (ibyvar = 0; ibyvar < nbyvar; ibyvar++)
	{
		for (j = 0; j < Conlen; j++)
		{
			groupcts[j] = 0.;
		}

		/* compute cell counts for every cell */
		for (i = 0; i < ndata; i++)
		{		/* for all data, even missing */
			if (!(PREVMODELTYPE & MISSWEIGHTS && misswts[i] == 0.0) &&
				!(PREVMODELTYPE & CASEWEIGHTS && casewts[i] == 0.0))
			{
				if (byvar < 1 || byvariable[i] == ibyvar + 1)
				{
					groupcts[getconentry(i, Convars, Nconvars)] += 1.;
				}
			}
		} /*for (i = 0; i < ndata; i++)*/

/* Rescale coefs so they apply to y and compute sum of contrast coefficients */
		tmp = 0.;
		for (j = 0; j < Conlen; j++)
		{
			if (groupcts[j] > 0.)
			{
				groupcts[j] = concoef[j] / groupcts[j];
				tmp += concoef[j];
			}
		} /*for (j = 0; j < Conlen; j++)*/

		if (fabs(tmp) > FUZZ)
		{		/* not a contrast at this level of by var */
			inck = nbyvar*ny;
			ik1 = ibyvar;
			for(k1 = 0;k1 < ny;k1++)
			{ /* ik1 = ibyvar + k1*nbyvar */
				setMissing(conest[ik1]);
				if(do_ses)
				{
					setMissing(conse[ik1]);
				}
				ik1k2 = ik1;
				for(k2 = 0;k2 < ny;k2++)
				{ /* ik1k2 = ibyvar + k1*nbyvar + k2*nbyvar*ny */
					setMissing(conss[ik1k2]);
					ik1k2 += inck;
				}
				ik1 += nbyvar;
			}
		} /*if (fabs(tmp) > FUZZ)*/
		else
		{
			singchk = 0.0;
			for (i = 0; i < ndata; i++)
			{					/* for all data, even missing */
				if (!(PREVMODELTYPE & MISSWEIGHTS && misswts[i] == 0.0))
				{
					if (byvar < 1 || byvariable[i] == ibyvar + 1)
					{
						convec[i] = groupcts[getconentry(i, Convars, Nconvars)];
						singchk += convec[i] * convec[i];
					}
					else
					{
						convec[i] = 0.;
					}
				}
			} /*for (i = 0; i < ndata; i++)*/

			/* now put in weights if weights are used */

			if (PREVMODELTYPE & USEWEIGHTS)
			{
				double    weight;
				
				for (i = 0; i < ndata; i++)
				{
					if(PREVMODELTYPE & ROBUSTREG)
					{
						weight = (PREVMODELTYPE & MISSWEIGHTS) ? 
							misswts[i] : 1.0;
					}
					else
					{
						weight = allwts[i];
					}
					
					convec[i] *= weight;
				}
			}

			/* when no byvar, sweep all columns out of Convec*/

			if (nbyvar == 1)
			{
				for (j = 0; j < nregx; j++)
				{
					denom = 0.;
					con = 0.;
					xij = *REGX + j * ndata; /* *xij is x(i,j) */
					for (i = 0; i < ndata; i++)
					{
						/*
						denom += x(i, j) * x(i, j);
						con += convec[i] * x(i, j);
						*/
						denom += *xij * *xij;
						con += convec[i] * *xij++;
					}

					if (denom > 0.0)
					{
						con /= denom;
						xij = *REGX + j * ndata; /* &x(i, j) */
						for (i = 0; i < ndata; i++)
						{
							/*
							(*Convec)[i] -= con * x(i, j);
							*/
							convec[i] -= con * *xij++;
						}
					}
				}
			} /*if (nbyvar == 1)*/

			denom = 0.0;
			for(i = 0;i < ndata;i++)
			{
				denom += convec[i] * convec[i];
			}
		
			if (denom / singchk <= 1.0e-15)
			{
				myerrorout("WARNING:  zero df for contrast");
				inck = nbyvar*ny;
				ik2 = ibyvar; /*ibyvar + k2 * nbyvar*/
				for(k2 = 0;k2 < ny;k2++)
				{
					ik1k2 = ik2; /* ibyvar + k2 * nbyvar + k1*nbyvar*ny */;
					setMissing(conest[ik1k2]);
					if(do_ses)
					{
						setMissing(conse[ik1k2]);
					}
					for(k1=0;k1<ny;k1++)
					{
						setMissing(conss[ik1k2]);
						ik1k2 += inck;
					}
					ik2 += nbyvar;
				}
				continue;
			} /*if (denom / singchk <= 1.0e-15)*/

			if(ny > 1 && do_ses > 0 && !(PREVMODELTYPE & ITERWEIGHTS))
			{
				kk = 0; /*j2*(nterms+1)*(ny+1)*/
				kkinc = (nterms+1)*(ny+1);
			}
			else
			{
				kkinc = 0;
			}
			
			ij = ibyvar;				/* ibyvar + k * nbyvar */
			ik = 0; /* i+k*ndata*/
			for (k = 0; k < ny; k++)
			{
				con = 0.;
				for (i = 0; i < ndata; i++)
				{
					if (PREVMODELTYPE & ITERWEIGHTS)
					{
						con += convec[i] * wtdresiduals[ik];
					}
					else if (PREVMODELTYPE & USEWEIGHTS)
					{
						con += convec[i] * allwts[i] * y[ik];
					}
					else
					{
						con += convec[i] * y[ik];
					}
					ik++;
				} /*for (i= 0; i < ndata; i +)*/

				/*		conest[ibyvar+k*nbyvar] = con;   */
				conest[ij] = con;
				if(do_ses)
				{
					if(kkinc != 0)
					{
						mse = sserror[kk]/dferror;
						kk += kkinc;
					}
					if(do_ses > 0)
					{
						conse[ij] = sqrt(denom*mse);
					}
					else
					{
						setMissing(conse[ij]);
					}
				} /*if(do_ses)*/
				ij += nbyvar;
			} /*for (k = 0; k < ny; k++)*/

			if(ny == 1)
			{
				con = con * con / denom;
				conss[ibyvar] = con;
			} /*if(ny == 1)*/
			else
			{
				inck = nbyvar*ny;
				ik2 = ibyvar; /* ibyvar + k2*nbyvar */
				for (k2 = 0; k2 < ny; k2++)
				{
					ik1k2 = ik2; /* ibyvar + k2*nbyvar + k1*nbyvar*ny */
					ik1 = ibyvar; /* ibyvar + k1*nbyvar */
					for (k1 = 0; k1 < ny; k1++)
					{
						conss[ik1k2] = conest[ik2] * conest[ik1] / denom;
						ik1k2 += inck; 
						ik1 += nbyvar;
					} /*for (k = 0; k < ny; k1++)*/
					ik2 += nbyvar;
				} /*for (k2 = 0; k2 < ny; k2++)*/
			}  /*if(ny == 1){...}else{...}*/
		}  /*if (fabs(tmp) > 1.0e-7){...} else {...}*/
	} /*for (ibyvar = 0; ibyvar < nbyvar; ibyvar++)*/
	emptyTrash();
	return (result);

  errorExit:
	putOUTSTR();

	emptyTrash();
	Removesymbol(result);

	return (0);
}


