/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Delete
#else
#define __SEG__ Delete
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#define IREALS      0x01
#define ICHARS      0x02
#define ILOGICS     0x04
#define IMACROS     0x08
#define ISTRUCS     0x10
#define IGRAPHS     0x20
#define IALL        0x3f /*IREALS|ICHARS|ILOGICS|IMACROS|ISTRUCS|IGRAPHS*/
static char        *DeleteKeys[] = 
{
	"real", "char", "logi", "macr", "stru", "grap", "all", (char *) 0
};
static short        KeyCodes[] = 
{
	IREALS, ICHARS, ILOGICS, IMACROS, ISTRUCS, IGRAPHS, IALL
};

Symbolhandle    deleter(list)
Symbolhandle    list;

{
	/*  routine to delete a list of variables */

	Symbolhandle    symh, Install();
	Symbolhandle    next;
	long            i, j, nargs = NARGS(list);
	long            type;
	long            foundKey = 0;
	long            op = 0;
	long            ndeleted = 0;
	char           *keyword;
	WHERE("deleter");
	
	*OUTSTR = '\0';
	for(i = 0; i < nargs; i++)
	{
		symh = COMPVALUE(list,i);
		if((keyword = isKeyword(symh)))
		{
			foundKey = 1;
			for(j=0;DeleteKeys[j] != (char *) 0;j++)
			{
				if(strncmp(keyword,DeleteKeys[j],4) == 0)
				{
					break;
				}
			}
			if(DeleteKeys[j] == (char *) 0)
			{
				badKeyword(FUNCNAME,keyword);
				goto errorExit;
			}
			if(!isScalar(symh) || TYPE(symh) != LOGIC)
			{
				notTorF(keyword);
				goto errorExit;
			}

			if(DATAVALUE(symh,0) != 0.0)
			{
				op |= KeyCodes[j];
			}
			else
			{
				op &= ~KeyCodes[j];
			}
		} /*if((keyword = isKeyword(symh)))*/
		else if(foundKey)
		{
			sprintf(OUTSTR,
					"ERROR: illegal to have both keywords and variables as %s arguments",
					FUNCNAME);
			goto errorExit;
			
		} /*if((keyword = isKeyword(symh))){}else if(foundKey){}*/
	} /*for(i = 0; i < nargs; i++)*/

	if(!foundKey)
	{
		for (i = 0; i < nargs; i++)
		{
			symh = COMPVALUE(list, i);
			if (isNull(symh))
			{
				sprintf(OUTSTR,"WARNING: missing or null argument to %s",
						FUNCNAME);
			}
			else if (TYPE(symh) == BLTIN)
			{					/*don't delete a built in */
				sprintf(OUTSTR,
						"WARNING: attempt to delete built-in function %s",
						NAME(symh));
			}
			else if(isscratch(NAME(symh)))
			{
				sprintf(OUTSTR,
						"WARNING: attempt to delete string, expression, or function result");
			}
			else if(!isDefined(symh))
			{
				sprintf(OUTSTR,
						"WARNING: attempt to delete undefined variable %s",
						NAME(symh));
			}
			else if((keyword = isKeyword(symh)))
			{
				badKeyword(FUNCNAME, keyword);
			}
			else
			{					/* check for duplicate symbols in list */
				COMPVALUE(list,i) = (Symbolhandle) 0;
				for(j = i+1;j < nargs;j++)
				{
					if(COMPVALUE(list,j) == symh)
					{
						sprintf(OUTSTR,
								"WARNING: duplicate variable in list");
						break;
					}
				}
				if(*OUTSTR == '\0')
				{
					Removesymbol(symh);
				}
			}
			putOUTSTR();
		} /*for (i = 0; i < nargs; i++)*/
	} /*if(!foundKey)*/
	else
	{
		if(op != 0)
		{
			symh = Firstsymbol(0);
			while (symh != (Symbolhandle) 0)
			{					/* loop through regular symbol table*/
				next = Nextsymbol(symh,0);
				type  = TYPE(symh);
				if(type == REAL && op & IREALS ||
				   type == CHAR && op & ICHARS ||
				   type == LOGIC && op & ILOGICS ||
				   type == MACRO && op & IMACROS ||
				   type == STRUC && op & ISTRUCS ||
				   type == CHAR && op & ICHARS ||
				   type == PLOTINFO && op & IGRAPHS)
				{
					ndeleted++;
					Removesymbol(symh);
				}
				symh = next;
			} /*while (symh != (Symbolhandle) 0)*/
			if(ndeleted == 0)
			{
				myerrorout("WARNING: nothing deleted");
			}
		} /*if(op != 0)*/
		
	} /*if(!foundKey){}else{}*/
	
	
	return (NULLSYMBOL);

  errorExit:
	putOUTSTR();
	
	return (0);
	
}
