/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Columnop
#else
#define __SEG__ Columnop
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

extern void           Delete();
extern Symbolhandle   RSInstall();
extern void           Removesymbol();
extern long           argOK();
extern char          *typeName();

#define  IN      0
#define  IMIN    1
#define  IQ1     2
#define  IMED    3
#define  IQ3     4
#define  IMAX    5
#define  IMEAN   6
#define  IVAR    7
#define  NSTATS  8

#define  OPN     0x01
#define  OPMIN   0x02
#define  OPQ1    0x04
#define  OPMED   0x08
#define  OPQ3    0x10
#define  OPMAX   0x20
#define  OPMEAN  0x40
#define  OPVAR   0x80

static char          *Descompname[] = 
{
	"n", "min", "q1", "median", "q3", "max", "mean", "var"
};

static short doDescribe(Symbolhandle arg, Symbolhandle result, long op,
						long *status)
{
	Symbolhandle    symh, result1 = (Symbolhandle) 0;
	Symbolhandle    components[NSTATS], component;
	double         *x, *xscratch;
	double          mean, res, stat, value;
	long            n, nmiss, nrows, ncols, i, j, k, l, jj;
	long            ndims, dims[MAXDIMS+1];
	long            nstats = NCOMPS(result);
	long            compno, dosort;
	WHERE("doDescribe");

	if(isStruc(arg))
	{
		long            icomp, ncomps = NCOMPS(arg);
		Symbolhandle    component1 = (Symbolhandle) 0;
		
		for(i=0;i<nstats;i++)
		{
			components[i] = COMPVALUE(result, i) = Makestruc(ncomps);
			if(components[i] == (Symbolhandle) 0)
			{
				goto errorExit;
			}
		}
		
		result1 = StrucInstall(SCRATCH, nstats);
		if(result1 == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		
		for(icomp=0;icomp<ncomps;icomp++)
		{
			symh = COMPVALUE(arg, icomp);
			if(doDescribe(symh, result1, op, status) == 0)
			{
				goto errorExit;
			}
			
			for(i=0;i<nstats;i++)
			{
				component1 = COMPVALUE(result1, i);
				COMPVALUE(result1, i) = (Symbolhandle) 0;
				COMPVALUE(components[i], icomp) = component1;
				setCompName(component1, NAME(symh));
			}
		} /* for(icomp=0;i<ncomps;i++) */
		Removesymbol(result1);
		return (1);
	} /* if(isStruc(arg)) */

	ncols = 1;
	ndims = NDIMS(arg);
	if(ndims <= 1)
	{
		dims[0] = 1;
	}
	else
	{
		ndims--;
		for(i = 0;i < ndims;i++)
		{
			dims[i] = DIMVAL(arg, i+2);
			ncols *= dims[i];
		}
	}
	
	nrows = DIMVAL(arg, 1);

	for(i=0;i<nstats;i++)
	{
		components[i] = COMPVALUE(result,i) = Makereal(ncols);
		if(components[i] == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNDIMS(components[i], ndims);
		for(j = 0;j < ndims; j++)
		{
			setDIM(components[i], j+1, dims[j]);
		}
	}

	if(nrows > 0)
	{
		Symbolhandle scratchH = RInstall(SCRATCH, nrows);

		if(scratchH == (Symbolhandle) 0)
		{
			goto errorExit;;
		}
		xscratch = DATAPTR(scratchH);

		x = DATAPTR(arg);
		l = 0;
		for(jj = 0;jj < ncols;jj++)
		{
			compno = 0;
			nmiss = 0;			/* missing count */
			for (i = 0; i < nrows; i++)
			{
				if (!isMissing(x[l]))
				{
					xscratch[i - nmiss] = x[l];
				}
				else
				{
					nmiss++;
				}
				l++;
			}

			if (nmiss > 0)
			{
				*status |= FOUNDMISSING;
			}

			n = nrows - nmiss;
			if(op & OPN)
			{
				DATAVALUE(components[compno++], jj) = (double) n;
			}
			
			if(n > 0)
			{
				dosort = (op & (OPQ1 | OPMED | OPQ3));
				if(dosort)
				{
					sortquick(xscratch, n);
				}
				
				if(op & OPMIN)
				{
					stat = xscratch[0];
					component = components[compno++];
					if(!dosort)
					{
						for(i=1;i<n;i++)
						{
							value = xscratch[i];
							if(value < stat)
							{
								stat = value;
							}
						}
					}
					DATAVALUE(component, jj) = stat;
				} /*if(op & OPMIN)*/
				if(op & OPQ1)
				{
					j = (n - 1)/4;
					k = (n + 1)/4;
					DATAVALUE(components[compno++], jj) =
						(xscratch[j] + xscratch[k]) / 2.0;
				} /*if(op & OPMQ1)*/
				if(op & OPMED)
				{
					j = (n - 1) / 2;
					k = n / 2;
					DATAVALUE(components[compno++], jj) =
						(xscratch[j] + xscratch[k]) / 2.0;
				} /*if(op & OPMMED)*/
				if(op & OPQ3)
				{
					j = n - 1 - (n - 1)/4;
					k = n - 1 - (n + 1)/4;
					DATAVALUE(components[compno++], jj) =
						(xscratch[j] + xscratch[k]) / 2.0;
				} /*if(op & OPMQ3)*/
				if(op & OPMAX)
				{
					stat = xscratch[n-1];
					component = components[compno++];
					if(!dosort)
					{
						for (i=n-2; i >= 0; i--)
						{
							value = xscratch[i];
							if(value > stat)
							{
								stat = value;
							}
						}
					}
					DATAVALUE(component, jj) = stat;
				} /*if(op & OPMAX)*/
					
				if(op & (OPMEAN | OPVAR))
				{
					
					mean = 0.;
					for (i = 0; i < n; i++)
					{
						mean += xscratch[i];
					}
					mean /= (double) n;
					if(op & OPMEAN)
					{
						DATAVALUE(components[compno++], jj) = mean;
					} /*if(op & OPMEAN)*/

					if(op & OPVAR)
					{
						stat = 0.0;
						for (i = 0; i < n; i++)
						{
							res = xscratch[i] - mean;
							stat += res * res;
						}
			
						DATAVALUE(components[compno], jj) = (n > 1) ?
							stat / (double) (n - 1) : 0.0;
					} /*if(op & OPVAR)*/
				} /*if(op & (OPMEAN | OPVAR))*/
			} /*if(n > 0)*/
			else
			{ /* n == 0 */
				for(i = compno;i < nstats;i++)
				{
					setMissing(DATAVALUE(components[i], jj));
				}
			}
		} /* for(jj = 0;jj < ncols;jj++) */
		Removesymbol(scratchH);
	} /* if(nrows > 0) */
	else
	{/* component is empty */
		for(i=0;i<nstats;i++)
		{
			symh = components[i];
			setMissing(DATAVALUE(symh, 0));
			setNDIMS(symh, 1);
			setDIM(symh, 1, 1);
		}
	}
	
	return (1);

  errorExit:
	Removesymbol(result1);

	return (0);
}	

Symbolhandle    describe(list)
Symbolhandle    list;
{
	Symbolhandle    result = (Symbolhandle) 0, symh, symhkey;
	long            i, j, nstats, nargs = NARGS(list);
	long            status = 0, op, stats[NSTATS];
	char           *keyword;
	WHERE("describe");

	*OUTSTR = '\0';
	
	if(nargs > NSTATS + 1)
	{
		badNargs(FUNCNAME, -(NSTATS+1));
		goto errorExit;
	}
	
	symh = COMPVALUE(list, 0);
	if(!argOK(symh, 0, 0))
	{
		goto errorExit;
	}
	if(!isReal(symh))
	{
		sprintf(OUTSTR,
				"ERROR: argument%s to %s must be REAL or structure of REALs",
				(nargs == 1) ? "" : " 1", FUNCNAME);
		goto errorExit;
	}
	
	if(nargs == 1)
	{
		op = OPN | OPMIN | OPQ1 | OPMED | OPQ3 | OPMAX | OPMEAN | OPVAR;
	}
	else
	{
		op = 0;
		for(i=1;i<nargs;i++)
		{
			symhkey = COMPVALUE(list,i);
			if(!argOK(symhkey,0,i+1))
			{
				goto errorExit;
			}
			
			if(!(keyword = isKeyword(symhkey)))
			{
				sprintf(OUTSTR,
						"ERROR: all arguments to %s except the first must be keyword phrases",
						FUNCNAME);
				goto errorExit;
			}
			for(j=0;j<NSTATS && strcmp(keyword,Descompname[j]) != 0;j++)
			{
				;
			}
			if(j == NSTATS)
			{
				badKeyword(FUNCNAME,keyword);
				goto errorExit;
			}
			if(TYPE(symhkey) != LOGIC || !isScalar(symhkey))
			{
				notTorF(keyword);
				goto errorExit;
			}
			if(op & (1L << j))
			{
				sprintf(OUTSTR,
						"ERROR: keyword %s for %s used more than once",
						keyword, FUNCNAME);
				goto errorExit;
			}
			if(DATAVALUE(symhkey,0) != 0.0)
			{
				op |= (1L << j);
			}
		} /*for(i=1;i<nargs;i++)*/
		if(op == 0)
		{
			sprintf(OUTSTR,
					"ERROR: no statistics specified for %s()", FUNCNAME);
			goto errorExit;
		}
	} /*if(nargs == 1){...}else{...}*/	

	nstats = 0;
	if(op & OPN)
	{
		stats[nstats++] = IN;
	}
	if(op & OPMIN)
	{
		stats[nstats++] = IMIN;
	}
	if(op & OPQ1)
	{
		stats[nstats++] = IQ1;
	}
	if(op & OPMED)
	{
		stats[nstats++] = IMED;
	}
	if(op & OPQ3)
	{
		stats[nstats++] = IQ3;
	}
	if(op & OPMAX)
	{
		stats[nstats++] = IMAX;
	}
	if(op & OPMEAN)
	{
		stats[nstats++] = IMEAN;
	}
	if(op & OPVAR)
	{
		stats[nstats++] = IVAR;
	}

	if((result = StrucInstall(SCRATCH, nstats)) == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	if(doDescribe(symh, result, op, &status) == 0)
	{
		goto errorExit;
	}
	
	if(nstats > 1)
	{
		for(i=0;i<nstats;i++)
		{
			setNAME(COMPVALUE(result, i), Descompname[stats[i]]);
		}
	}
	else
	{
		symh = COMPVALUE(result,0);
		COMPVALUE(result, 0) = (Symbolhandle) 0;
		setNAME(symh, SCRATCH);
		Addsymbol(symh);
		Removesymbol(result);
		result = symh;
	}
	
	if(status & FOUNDMISSING)
	{
		myerrorout("WARNING: missing values in input to describe");
		myerrorout("WARNING: output reflects nonmissing data only");
	}		
	
	return (result);
	
  errorExit:
	putOUTSTR();

	Removesymbol(result);
	
	return (0);
} /*describe()*/
