/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Dim
#else
#define __SEG__ Dim
#endif /*MPW3*/
#endif /*MPW*/

#include <math.h>
#include "globals.h"
#include "dbug.h"

extern Symbolhandle     RInstall();
extern long             isNull();
extern void             noData();
extern void             badNargs();
extern 	long            symbolSize();

#define INDIMS   0
#define INROWS   1
#define INCOLS   2
#define INCOMPS  3
#define ILENGTH  4
#define IDIM     5

static long strucMaxNdims(Symbolhandle arg)
{
	long        maxNdims = 0;
	long        i, ncomps, ndims;
	
	if(!isNull(arg))
	{
		if(isStruc(arg))
		{
			ncomps = NCOMPS(arg);
			for(i=0;i<ncomps;i++)
			{
				ndims = strucMaxNdims(COMPVALUE(arg,i));
				maxNdims = (ndims > maxNdims) ? ndims : maxNdims;
			}
		}
		else
		{
			maxNdims = NDIMS(arg);
		}
	}
	return (maxNdims);
}

static Symbolhandle doDimOps(Symbolhandle arg, long op)
{
	long           i, ncomps,dim[2], val;
	Symbolhandle   result = (Symbolhandle) 0, symh = (Symbolhandle) 0;

	if(!isNull(arg))
	{
		if(isStruc(arg))
		{
			ncomps = NCOMPS(arg);
			result = StrucInstall(SCRATCH,ncomps);
			if(result == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			
			for(i=0;i<ncomps;i++)
			{
				symh = doDimOps(COMPVALUE(arg,i),op);
				if(symh == (Symbolhandle) 0)
				{
					goto errorExit;
				}
				Cutsymbol(symh);
				COMPVALUE(result,i) = symh;
				setCompName(symh,NAME(COMPVALUE(arg,i)));
			}
		} /*if(isStruc(arg))*/
		else
		{
			result = RInstall(SCRATCH,1);
			if(result == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			if(op == INROWS || op == INCOLS)
			{
				(void) isMatrix(arg,dim);
				val = dim[(op==INROWS) ? 0 : 1];
			}
			else if(op == ILENGTH)
			{
				val = symbolSize(arg);
			}
			else if(op == INDIMS)
			{
				val = NDIMS(arg);
			}
			DATAVALUE(result,0) = (double) val;
		} /*if(isStruc(arg)){}else{}*/
	} /*if(!isNull(arg))*/
	else
	{
		result = RInstall(SCRATCH,0);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
	}
	return (result);

  errorExit:
	Removesymbol(result);
	return (0);
} /*doDimOps()*/

static long doDim(Symbolhandle arg,Symbolhandle result, long ndims)
{
	long           i, j, ncomps;
	char          *name;
	Symbolhandle   symh, dim, result1 = (Symbolhandle) 0;

	if(isStruc(arg))
	{
		ncomps = NCOMPS(arg);
		for(i=0;i<ndims;i++)
		{
			COMPVALUE(result,i) = Makestruc(ncomps);
			if(COMPVALUE(result,i) == (Symbolhandle) 0)
			{
				goto errorExit;
			}
		}
		
		result1 = StrucInstall(SCRATCH,ndims);
		if(result1 == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		
		for(i=0;i<ncomps;i++)
		{
			if(!doDim(COMPVALUE(arg,i),result1,ndims))
			{
				goto errorExit;
			}
			name = NAME(COMPVALUE(arg,i));
			for(j=0;j<ndims;j++)
			{
				dim = COMPVALUE(result1,j);
				COMPVALUE(result1,j) = (Symbolhandle) 0;
				setCompName(dim,name);
				symh = COMPVALUE(result,j);
				COMPVALUE(symh,i) = dim;
			}
		}
		Removesymbol(result1);
		return (1);
	} /*if(isStruc(arg))*/
	
	for(i=0;i<ndims;i++)
	{
		dim = COMPVALUE(result,i) = Makereal(1);
		if(dim == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		
		DATAVALUE(dim,0) = (!isNull(arg) && i < NDIMS(arg)) ?
			DIMVAL(arg,i+1) : 0;
	}
	return (1);

  errorExit:
	Removesymbol(result1);

	return (0);
}
	
/*
  dim(x), ndims(x), nrows(), ncols(), ncomps(),  length()

*/

Symbolhandle    dim(list)
Symbolhandle    list;
{
	Symbolhandle    result = (Symbolhandle) 0, symh;
	long            ndims,op;
	long            i;
	WHERE("dim");
	
	if(NARGS(list) != 1)
	{
		badNargs(FUNCNAME,1);
		goto errorExit;
	}
	
	symh = COMPVALUE(list, 0);
	if (!argOK(symh,(long) 0, 0))
	{
		goto errorExit;
	}

	if (TYPE(symh) == BLTIN)
	{
		badType(FUNCNAME,-BLTIN,0);
		goto errorExit;
	}

	if(strcmp("ncomps",FUNCNAME) == 0)
	{
		if(!isStruc(symh))
		{
			sprintf(OUTSTR,"ERROR: argument to %s not structure",FUNCNAME);
			goto errorExit;
		}
		result = RInstall(SCRATCH,1);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		DATAVALUE(result,0) = (double) NCOMPS(symh);
	}
	else if(strcmp("dim",FUNCNAME) == 0)
	{ /* dim() */
		if(!isStruc(symh))
		{
			ndims = NDIMS(symh);
			result = RInstall(SCRATCH,ndims);
			if(result == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			
			for(i = 0;i < ndims;i++)
			{
				DATAVALUE(result,i) = (double) DIMVAL(symh,i+1);
			}
		} /*if(!isStruc(symh))*/
		else
		{ /* dim(structure) */
			ndims = strucMaxNdims(symh);
			if(ndims == 0)
			{
				myerrorout("ERROR: all components of structure are null");
				goto errorExit;
			}
			if((result = StrucInstall(SCRATCH,ndims)) == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			
			if(!doDim(symh,result,ndims))
			{
				goto errorExit;
			}
			for(i=0;i<ndims;i++)
			{
				sprintf(OUTSTR,"dim%ld",i+1);
				setCompName(COMPVALUE(result,i),OUTSTR);
			}
			*OUTSTR = '\0';
		}
	}
	else
	{ /* ndims(), nrows(), ncols(), length() */
		if(strcmp(FUNCNAME,"ndims") == 0)
		{
			op = INDIMS;
		}
		else if(strcmp(FUNCNAME,"nrows") == 0)
		{
			op = INROWS;
		}
		else if(strcmp(FUNCNAME,"ncols") == 0)
		{
			op = INCOLS;
		}
		else if(strcmp(FUNCNAME,"length") == 0)
		{
			op = ILENGTH;
		}
		if((op == INROWS || op == INCOLS) && !strucIsMatrix(symh))
		{
			sprintf(OUTSTR,"ERROR: argument to %s is not a matrix%s",FUNCNAME,
					(isStruc(symh)) ? " or a structure of matrices" : "");
			goto errorExit;
		}
		result = doDimOps(symh,op);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
	}		
	
	return (result);

  errorExit:
	putOUTSTR();

	return (0);
	
} /*dim()*/
  
#define IDIAG 1
#define IDMAT 2

static Symbolhandle doDiag(Symbolhandle arg, long *dims)
{
	Symbolhandle result = (Symbolhandle) 0;
	long         incr, n, i, j;
	
	n = (dims[0] < dims[1]) ? dims[0]: dims[1];
	
	result = RInstall(SCRATCH,n);
	
	if(result != (Symbolhandle) 0)
	{
		setTYPE(result,TYPE(arg));
		incr = dims[0] + 1;
		j = 0;
		for(i = 0;i< n;i++)
		{
			DATAVALUE(result,i) = DATAVALUE(arg,j);
			j += incr;
		}
	}
	
	return (result);
} /*doDiag()*/

/*
  Construct diagonal matrix
  if there is 1 argument, it is a vector or m by 1 or 1 by m matrix
  if there are 2 argument, the first is m and the second is v and
  it returns v*iden(m)
*/

static Symbolhandle doDmat(Symbolhandle arg1,Symbolhandle arg2,long *dims)
{
	Symbolhandle result;
	double       value;
	long         n, incr, i, j, k, l;
	WHERE("doDmat");
	
	*OUTSTR = '\0';
	if(arg2 != (Symbolhandle) 0)
	{ /* dmat(n,value) */
		value = DATAVALUE(arg1,0);

		if(value != floor(value) || value <= 0)
		{
			sprintf(OUTSTR,
					"ERROR: 1st argument to %s is not positive integer",
					FUNCNAME);
			goto errorExit;
		}
		n = (long) value;
		arg1 = arg2;
		incr = 0;
	}
	else
	{ /* dmat(vector) */
		n = dims[0];
		incr = 1;
	}
	
	result = RInstall(SCRATCH,n*n);

	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	setNDIMS(result,2);
	setDIM(result,1,n);
	setDIM(result,2,n);
	setTYPE(result,TYPE(arg1));
	k = 0;
	l = 0;
	for(i = 0;i< n;i++)
	{
		for(j = 0;j < n;j++)
		{
			if(i == j)
			{
				DATAVALUE(result,l) = DATAVALUE(arg1,k);
				k += incr;
			}
			else
			{
				DATAVALUE(result,l) = 0.;
			}
			l++;
		}
	}
	
	return (result);

  errorExit:
	putOUTSTR();
	return (0);
} /*doDmat()*/

/* diag(matrix) and dmat(vector) or dmat(n,value) */

Symbolhandle    diag(list)
Symbolhandle    list;
{
	Symbolhandle    result = (Symbolhandle) 0, arg1, arg2;
	long            nargs;
	long            type, margs, dims[2];
	long            op;
	char           *dmatUsage =
		"ERROR: usage is dmat(vector) or dmat(n,value)";
	WHERE("diag");
	
	if(strcmp(FUNCNAME,"diag") == 0)
	{
		op = IDIAG;
		margs = 1;
	}
	else if(strcmp(FUNCNAME,"dmat") == 0)
	{
		op = IDMAT;
		margs = -2;
	}

	nargs = NARGS(list);
	if((margs >= 0 && nargs != margs) || (margs < 0 && nargs > -margs))
	{
		badNargs(FUNCNAME,margs);
		goto errorExit;
	}

	arg1 = COMPVALUE(list, 0);
	if(! argOK(arg1, 0, 1))
	{
		goto errorExit;
	}
	
	type = TYPE(arg1);
	if (type != REAL && type != LOGIC)
	{
		badType(FUNCNAME,-type, 1);
		goto errorExit;
	}

	if(op == IDIAG)
	{
		if(!isMatrix(arg1,dims))
		{
			myerrorout("ERROR: argument to diag must be matrix");
			goto errorExit;
		}
	}
	else if(nargs == 1)
	{ /* dmat(vector) */
		if(!isVector(arg1) && DIMVAL(arg1,2) != symbolSize(arg1))
		{ /* allow row vector as argument */
			myerrorout(dmatUsage);
			goto errorExit;
		}
		arg2 = (Symbolhandle) 0;
		dims[0] = symbolSize(arg1);
	}
	else
	{ /* nargs == 2 */
		if(!isScalar(arg1) || type != REAL)
		{
			myerrorout(dmatUsage);
			goto errorExit;
		}
		arg2 = COMPVALUE(list,1);
		if(! argOK(arg2, 0,  1))
		{
			goto errorExit;
		}

		type = TYPE(arg2);
		if(type != REAL && type != LOGIC)
		{
			badType(FUNCNAME,-type, 2);
			goto errorExit;
		}
		if(!isScalar(arg2))
		{
			myerrorout(dmatUsage);
			goto errorExit;
		}
		dims[0] = 1;
	}

	if(op == IDIAG)
	{
		result = doDiag(arg1,dims);
	}
	else if(op == IDMAT)
	{
		result = doDmat(arg1,arg2,dims);
	}
	
	return (result);

  errorExit:
	return (0);
} /*diag()*/

		











		
		
