/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef MPW
#ifdef MPW3
#pragma segment Eigen
#else
#define __SEG__ Eigen
#endif /*MPW3*/
#endif /*MPW*/
#include "globals.h"
#include "blas.h"
#include "dbug.h"
#include <math.h>

extern void         rsg(), tql2(), tqlrat();
extern void         tred1(), tred2();
extern Symbolhandle RInstall(), RSInstall(), Install();
extern void         Removesymbol();

extern long         argOK();
extern long         isMatrix(), checkSymmetry();
extern long         anyMissing();
extern void         badNargs();

static char         *OpNames[] =
{
	"eigen", "eigenvals", "releigen", "releigenvals", "trace", (char *) 0
};

#define IEIGEN        1
#define IEIGENVALS    2
#define IRELEIGEN     3
#define IRELEIGENVALS 4
#define ITRACE        5

#define GVALS         0
#define GAA           1
#define GVECS         1  /* vecs is same as aa */
#define GBB           2
#define GSUBD         3
#define GSUB2         4
#define NTRASH        5

static long          OpCodes[] =
{
	IEIGEN, IEIGENVALS, IRELEIGEN, IRELEIGENVALS, ITRACE
};

Symbolhandle    eigen(list)
Symbolhandle    list;
{

	/* eigen value calculations for a real symmetric  matrix */

	Symbolhandle    result = (Symbolhandle) 0, arg1,
	                arg2, symh;
	long            margs, nrows, tot, i;
	long            dim[2];
	long            aii, ierr;
	long            valsVecs;
	long            inc = 1;
	int             op;
	double          sum;
	double        **ax = (double **) 0, **bx = (double **) 0;
	double        **vals = (double **) 0, **vecs = (double **) 0;
	double        **subd = (double **) 0, **sub2 = (double **) 0;
	double        **aa = (double **) 0, **bb = (double **) 0;
	char           *compname[2];
	WHERE("eigen");
	TRASH(NTRASH,errorExit);
	
	op = matchKey(FUNCNAME,OpNames,OpCodes);
	if(op == IEIGEN || op == IEIGENVALS || op == ITRACE)
	{
		margs = 1;
	}
	else
	{
		margs = 2;
	}

	OUTSTR[0] = '\0';
	if(NARGS(list) != margs)
	{
		badNargs(FUNCNAME,margs);
		goto errorExit;
	}
	arg1 = COMPVALUE(list, 0);
	if(!argOK(arg1,REAL,1))
	{
		goto errorExit;
	}
	
	if(!isMatrix(arg1,dim) || dim[0] != dim[1])
	{
		sprintf(OUTSTR,"ERROR: argument to %s not REAL square matrix",
				FUNCNAME);
	}
	else if(anyMissing(arg1))
	{
		sprintf(OUTSTR,"ERROR: 1st argument to %s contains MISSING value(s)",
				FUNCNAME);
	}

	if(*OUTSTR)
	{
		goto errorExit;
	}
	
	nrows = dim[0];
	
	ax = DATA(arg1);
	if(op == ITRACE)
	{
		sum = 0.0;
		aii = 0;
		for (i = 0; i < nrows; i++)
		{
			sum += (*ax)[aii];
			aii += nrows + 1;
		}
		result = RInstall(SCRATCH, 1L);
		if (result == (Symbolhandle) 0)
		{
			return (0);
		}
		DATAVALUE(result,0) = sum;
		emptyTrash();
		return (result);
	}

/*
  operations "eigen" "eigenvals", "releigen", "releigenvals"
*/
	if(!checkSymmetry(*ax, nrows))
	{
		sprintf(OUTSTR,
				"ERROR: 1st argument to %s must be symmetric REAL matrix",
				FUNCNAME);
	}
	else if (margs == 2)
	{
		arg2 = COMPVALUE(list, 1);
		if(!argOK(arg2,REAL,2))
		{
			goto errorExit;
		}

		bx = DATA(arg2);
		if(anyMissing(arg2))
		{
			sprintf(OUTSTR,
					"ERROR: 2nd argument to %s contains MISSING value(s)",
					FUNCNAME);
		}
		else if(!isMatrix(arg2,dim) || dim[0] != dim[1])
		{
			sprintf(OUTSTR,
					"ERROR: second argument to %s is not REAL square matrix",
					FUNCNAME);
		}
		else if(dim[0] != nrows)
		{
			sprintf(OUTSTR,"ERROR: arguments to %s not same size",FUNCNAME);
		}
		else if(!checkSymmetry(*bx, nrows))
		{
			sprintf(OUTSTR,"ERROR: 2nd argument to %s not symmetric",FUNCNAME);
		}
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}

	compname[0] = "values";
	compname[1] = "vectors";

	tot = nrows * nrows;

	if(!getScratch(vals,GVALS,nrows,double) ||
	   !getScratch(vecs,GVECS,tot,double) ||
	   !getScratch(subd,GSUBD,nrows,double) ||
	   !getScratch(sub2,GSUB2,nrows,double))
	{
		goto errorExit;
	}
	aa = vecs; /* NOTE: aa and vecs coincide */

	switch (op)
	{
	  case IEIGEN:
		tred2(nrows, nrows, *ax, *vals, *subd, *vecs);
		tql2(nrows, nrows, *vals, *subd, *vecs, &ierr);

		if (ierr > 0)
		{
			myerrorout("ERROR: eigen algorithm did not converge");
			goto errorExit;
		}
		result = RSInstall(SCRATCH, 2L, compname, 0L);
		if (result == (Symbolhandle) 0)
		{
			goto errorExit;
		}

		symh = COMPVALUE(result, 0);
		unTrash(GVALS);
		setDATA(symh,vals);
		setNDIMS(symh,1);
		setDIM(symh,1,nrows);

		symh = COMPVALUE(result, 1);
		unTrash(GVECS);
		setDATA(symh,vecs);
		setNDIMS(symh,2);
		setDIM(symh,1,nrows);
		setDIM(symh,2,nrows);

		discardScratch(subd,GSUBD,double);
		discardScratch(sub2,GSUB2,double);
		vals = vecs = subd = sub2 = (double **) 0;
		
		break;

	  case IEIGENVALS:
		DCOPY(&tot, *ax, &inc, *aa, &inc); /* copy ax to aa */
		
		tred1(nrows, nrows, *aa, *vals, *subd, *sub2);
		tqlrat(nrows, *vals, *sub2, &ierr);

		if (ierr > 0)
		{
			sprintf(OUTSTR,"ERROR: eigen algorithm in %s did not converge",
					FUNCNAME);
			goto errorExit;
		}
		
		result = RInstall(SCRATCH, 0L);
		if (result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		unTrash(GVALS);
		setDATA(result,vals);
		setNDIMS(result,1);
		setDIM(result,1,nrows);

		break;
		
	  default:
		/* IRELEIGEN or IRELEIGENVALS */
		valsVecs = (op == IRELEIGEN) ? 1 : 0;
		
		if(!getScratch(bb,GBB,tot,double))
		{
			goto errorExit;
		}
		
		DCOPY(&tot, *ax, &inc, *aa, &inc); /* copy ax to aa */
		DCOPY(&tot, *bx, &inc, *bb, &inc); /* copy bx to bb */

/* modified by kb to allow return of vectors */
		rsg(nrows, nrows, *aa, *bb, *vals, valsVecs, *vecs, *subd, *sub2,
			&ierr);
		if (ierr > 0)
		{
			if(ierr == 7*nrows + 1)
			{
				sprintf(OUTSTR,
						"ERROR: 2nd argument to %s not positive definite",
						FUNCNAME);
			}
			else
			{
				sprintf(OUTSTR,
						"ERROR: eigen algorithm in %s did not converge",
						FUNCNAME);
			}
			goto errorExit;
		}

		if(valsVecs)
		{
			result = RSInstall(SCRATCH, 2L, compname, 0L);
		}
		else
		{
			result = RInstall(SCRATCH, 0L);
		}
		
		if (result == (Symbolhandle) 0)
		{
			goto errorExit;
		}

		if(valsVecs)
		{
			symh = COMPVALUE(result, 1);
			unTrash(GVECS);
			setDATA(symh,vecs);
			setNDIMS(symh,2);
			setDIM(symh,1,nrows);
			setDIM(symh,2,nrows);

			symh = COMPVALUE(result, 0);
		}
		else
		{
			symh = result;
		}
		
		unTrash(GVALS);
		setDATA(symh,vals);
		setNDIMS(symh,1);
		setDIM(symh,1,nrows);
	} /* switch (op)*/
	
	emptyTrash();

	UNLOADSEG(DCOPY);
	UNLOADSEG(rsg);

	return (result);

  errorExit:
	putOUTSTR();
	emptyTrash();
	
	UNLOADSEG(DCOPY);
	UNLOADSEG(rsg);

	return (0);
	
}




