/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.36 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Eigutils
#else
#define __SEG__ Eigutils
#endif /*MPW3*/
#endif /*MPW*/

#include <math.h>
#ifndef NOSTDLIBH
#ifdef MPW
#include <StdLib.h>
#else /*MPW*/
#include <stdlib.h>
#endif /*MPW*/
#endif /*NOSTDLIBH*/

#include "platform.h"
#include "dbug.h"

/* 
  translations of various Eispack subroutines for computing eigen values
  and vectors and singular values and vectors
*/

#define FORTRAN  /* assume Fortran storage ordering */

#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define SIGN(a,b) (((b) > 0) ? fabs((a))  :  -fabs((a)))

#ifdef FORTRAN
#define Q2a(I,J) a[(I) + nm*(J)]
#define Q2b(I,J) b[(I) + nm*(J)]
#define Q2z(I,J) z[(I) + nm*(J)]
#else
#define Q2a(I,J) a[(J) + nm*(I)]
#define Q2b(I,J) b[(J) + nm*(I)]
#define Q2z(I,J) z[(J) + nm*(I)]
#endif

extern double pythag(double /*a*/, double /*b*/);

static void     rebak(long , long, double *, double *, long, double *);
static void     reduc(long, long, double *, double *, double *, long *);
void            rsg(long, long, double *, double *, double *, long,
					double *, double *, double *, long *);
void tql2(long /*nm*/, long /*n*/, double */*d*/, double */*e*/,
		  double */*z*/, long */*ierr*/);
void tqlrat(long /*n*/, double */*d*/, double */*e2*/, long */*ierr*/);
void tred1(long /*nm*/, long /*n*/, double */*a*/, double */*d*/,
		   double */*e*/, double */*e2*/);
void tred2(long /*nm*/, long /*n*/, double */*a*/, double */*d*/,
		   double */*e*/, double */*z*/);

/*
	C. Bingham, kb@umnstat.stst.umn.edu  891213
	From Eispack
	Minor modifications and neatening up 950519
*/

/*      estimate unit roundoff in quantities of size x. */


/*
	This program should function properly on all systems
      satisfying the following two assumptions,
         1.  The base used in representing floating point
             numbers is not a power of three.
         2.  The quantity  a  in statement 10 is represented to
             the accuracy used in floating point variables
             that are stored in memory.
      The label LAB10 and the goto LAB10 are intended to
      force optimizing compilers to generate code satisfying
      assumption 2.
      Under these assumptions, it should be true that,
             a  is not exactly equal to four-thirds,
             b  has a zero for its last bit or digit,
             c  is not exactly equal to one,
             eps  measures the separation of 1.0 from
                  the next larger floating point number.
      The developers of eispack would appreciate being informed
      about any systems where these assumptions do not hold.

      This version dated 4/6/83.
*/

static double          epslon(double x)
{
	double          a, b, c, eps;

	a = 4.0 / 3.0;
/* use of label designed to out-fox optimizing compiler; see above */
  LAB10:
	b = a - 1.0;
	c = b + b + b;
	eps = fabs(c - 1.0);
	if (eps == 0)
	{
		goto LAB10;
	}
	return (eps * fabs(x));
}

/*
	Initial translation to C from Ratfor by ratfor2c
	From Eispack library
	C. Bingham, kb@umnstat.stst.umn.edu  891213
*/

/* translated from rebak.f */
/*
	Initial translation to C from Ratfor by ratfor2c
	From Eispack Library
	Compile with -DFORTRAN for Fortran storage format
	C. Bingham, kb@umnstat.stst.umn.edu  891215
*/


/*
      This subroutine is a translation of the Algol procedure rebaka,
      Num. Math. 11, 99-110(1968) by Martin and Wilkinson.
      Handbook for Auto. Comp., Vol.Ii-Linear Algebra, 303-314(1971).


      This subroutine forms the eigenvectors of a generalized
      symmetric eigensystem by back transforming those of the
      derived symmetric matrix determined by  reduc.


      On input


         nm must be set to the row dimension (column dimension ifndef
		   FORTRAN) of dimensional array parameters as declared in the
		   calling program dimension statement.


         n is the order of the matrix system.


         b contains information about the similarity transformation
           (cholesky decomposition) used in the reduction by  reduc
           in its strict lower triangle.


         dl contains further information about the transformation.


         m is the number of eigenvectors to be back transformed.


         z contains the eigenvectors to be back transformed
           in its first m columns.


      On output


         z contains the transformed eigenvectors
           in its first m columns.


      Questions and comments should be directed to Burton S. Garbow,
      Mathematics and Computer Science Div, Argonne National Laboratory


      This version dated August 1983.


      ------------------------------------------------------------------
*/

static void            rebak(long nm, long  n, double *b, double *dl, long  m,
							 double *z)
/*double b[n][nm],dl[n],z[n][nm];*/
{
	long            i, j, k;
	double          x;

	for (j = 0; j < m; j++)
	{
/*      .......... for i=n step -1 until 1 do -- .......... */
		for (i = n - 1; i >= 0; i--)
		{
			x = Q2z(i, j);
			for (k = i + 1; k < n; k++)
			{
				x -= Q2b(k, i) * Q2z(k, j);
			}
			Q2z(i, j) = x / dl[i];
		} /*for (i = n - 1; i >= 0; i--)*/
	} /*for (j = 0; j < m; j++)*/
} /*rebak()*/

/* translated from reduc.f */
/*
	Initial translation to C from Ratfor by ratfor2c
	From Eispack Library

	C. Bingham, kb@umnstat.stst.umn.edu  891215
	Compile with -DFORTRAN for fortran storage format
*/

/*
      this subroutine is a translation of the algol procedure reduc1,
      num. math. 11, 99-110(1968) by martin and wilkinson.
      handbook for auto. comp., vol.ii-linear algebra, 303-314(1971).


      This subroutine reduces the generalized symmetric eigenproblem
      ax=(hlambda)bx, where b is positive definite, to the standard
      symmetric eigenproblem using the Cholesky factorization of b.


      On input


         nm must be set to the row dimension (column dimension ifndef
		   FORTRAN) of two-dimensional array parameters as declared in the
		   calling program dimension statement.


         n is the order of the matrices a and b.  if the cholesky
           factor l of b is already available, n should be prefixed
           with a minus sign.


         a and b contain the real symmetric input matrices.  only the
           full upper triangles of the matrices need be supplied.  if
           n is negative, the strict lower triangle of b contains,
           instead, the strict lower triangle of its cholesky factor l.


         dl contains, if n is negative, the diagonal elements of l.


      On output


         a contains in its full lower triangle the full lower triangle
           of the symmetric matrix derived from the reduction to the
           standard form.  The strict upper triangle of a is unaltered.


         b contains in its strict lower triangle the strict lower
           triangle of its Cholesky factor l.  The full upper
           triangle of b is unaltered.


         dl contains the diagonal elements of l.


         ierr is set to
           zero       for normal return,
           7*n+1      if b is not positive definite.


      Questions and comments should be directed to Burton S. Garbow,
      Mathematics and Computer Science Div, Argonne National Laboratory


      This version dated August 1983.


      ------------------------------------------------------------------
*/

static void            reduc(long nm, long n, double *a, double *b, double *dl,
							 long *ierr)
/*
double a[n][nm],b[n][nm],dl[n]; for C storage mode
double a(nm,n), b(nm,n),dl(n) for Fortran storage mode
*/
{
	long            i, j, k, nn;
	double          x, y = 0.0;

	*ierr = 0;
	nn = labs(n);
	if (n >= 0)
	{
/*      .......... form l in the arrays b and dl .......... */
		for (i = 0; i < n; i++)
		{
			for (j = i; j < n; j++)
			{
				x = Q2b(i, j);
				for (k = 0; k < i; k++)
				{
					x -= Q2b(i, k) * Q2b(j, k);
				} /*for (k = 0; k < i; k++)*/
				if (j != i)
				{
					Q2b(j, i) = x / y;
				} /*if (j != i)*/
				else
				{
					if (x <= 0.0)
					{
/*      .......... set error -- b is not positive definite .......... */
						*ierr = 7 * n + 1;
						return;
					}
					y = sqrt(x);
					dl[i] = y;
				} /*if (j != i){}else*/
			} /*for (j = i; j < n; j++)*/
		} /*for (i = 0; i < n; i++)*/
	} /*if (n >= 0)*/

/*
      .......... form the transpose of the upper triangle of inv(l)*a
                 in the lower triangle of the array a ..........
*/
	for (i = 0; i < nn; i++)
	{
		y = dl[i];
		for (j = i; j < nn; j++)
		{
			x = Q2a(i, j);
			for (k = 0; k < i; k++)
			{
				x -= Q2b(i, k) * Q2a(j, k);
			} /*for (k = 0; k < i; k++)*/
			Q2a(j, i) = x / y;
		} /*for (j = i; j < nn; j++)*/
	} /*for (i = 0; i < nn; i++)*/

/*      .......... pre-multiply by inv(l) and overwrite .......... */
	for (j = 0; j < nn; j++)
	{
		for (i = j; i < nn; i++)
		{
			x = Q2a(i, j);
			for (k = j; k < i; k++)
			{
				x -= Q2a(k, j) * Q2b(i, k);
			} /*for (k = j; k < i; k++)*/

			for (k = 0; k < j; k++)
			{
				x -= Q2a(j, k) * Q2b(i, k);
			} /*for (k = 0; k < j; k++)*/
			Q2a(i, j) = x / dl[i];
		} /*for (i = j; i < nn; i++)*/
	} /*for (j = 0; j < nn; j++)*/
} /*reduc()*/

/* translated from rsg.f */
/*
	Initial translation to C from Ratfor by ratfor2c 891213
	From Eispack Library

	C. Bingham,  kb@umnstat.stat.umn.edu  891215
*/


/*
      This subroutine calls the recommended sequence of
      subroutines from the eigensystem subroutine package (eispack)
      to find the eigenvalues and eigenvectors (if desired)
      for the real symmetric generalized eigenproblem  ax = (lambda)bx.


      On input


         nm must be set to the row dimension (column dimension ifndef
		 FORTRAN) of two-dimensional array parameters as declared in the
		 calling program dimension statement.


         n  is the order of the matrices  a  and  b.


         a  contains a real symmetric matrix.


         b  contains a positive definite real symmetric matrix.


         matz  is an integer variable set equal to zero if
         only eigenvalues are desired.  Otherwise it is set to
         any non-zero integer for both eigenvalues and eigenvectors.


      On output


         w  contains the eigenvalues in ascending order
		 (changed to descending by c. bingham)


         z  contains the eigenvectors if matz is not zero.


         ierr  is a pointer to an integer output variable set equal to an error
            completion code described in the documentation for tqlrat
            and tql2.  the normal completion code is zero.


         fv1  and  fv2  are temporary storage arrays.


      Questions and comments should be directed to Burton S. Garbow,
      Mathematics and Computer Science Div, Argonne National Laboratory


      This version dated August 1983.


      ------------------------------------------------------------------
*/

void            rsg(long nm, long n, double *a, double *b, double *w,
					 long matz, double *z, double *fv1, double *fv2,
					 long *ierr)
/*
double a[n][nm],b[n][nm],w[n],z[n][nm],fv1[n],fv2[n]; (C-storage mode)
double a(nm,n),b(nm,n),w(n),z(nm,n),fv1(n),fv2(n)  (Fortran storage mode)
*/
{
	if (n > nm)
	{
		*ierr = 10 * n;
	} /*if (n > nm)*/
	else
	{
		reduc(nm, n, a, b, fv2, ierr);
		if (*ierr == 0)
		{
			if (matz == 0)
			{
/*      .......... find eigenvalues only .......... */
				tred1(nm, n, a, w, fv1, fv2);
				tqlrat(n, w, fv2, ierr);
			} /*if (matz == 0)*/
			else
			{
/*      .......... find both eigenvalues and eigenvectors .......... */
				tred2(nm, n, a, w, fv1, z);
				tql2(nm, n, w, fv1, z, ierr);
				if (*ierr == 0)
				{
					rebak(nm, n, b, fv2, n, z);
				}
			} /*if (matz == 0){}else{}*/
		} /*if (*ierr == 0)*/
	} /*if (n > nm){}else{}*/

} /*rsg()*/


/* translated from tql2.f */

/*
	Initial translation to C from Ratfor by ratfor2c
	From Eispack Library

	C. Bingham,  kb@umnstat.stat.umn.edu  891215
	Compile with -DFORTRAN for fortran storage format
*/

/*
      This subroutine is a translation of the Algol procedure tql2,
      Num. Math. 11, 293-306(1968) by Bowdler, Martin, Reinsch, and
      Wilkinson.
      Handbook for Auto. Comp., Vol.Ii-Linear Algebra, 227-240(1971).


      This subroutine finds the eigenvalues and eigenvectors
      of a symmetric tridiagonal matrix by the ql method.
      The eigenvectors of a full symmetric matrix can also
      be found if  tred2  has been used to reduce this
      full matrix to tridiagonal form.


      On input


         nm must be set to the row dimension (column dimension ifndef
		   FORTRAN) of two-dimensional array parameters as declared in the
		   calling program dimension statement.


         n is the order of the matrix.


         d contains the diagonal elements of the input matrix.


         e contains the subdiagonal elements of the input matrix
           in its last n-1 positions.  e[0] is arbitrary.


         z contains the transformation matrix produced in the
           reduction by  tred2, if performed.  If the eigenvectors
           of the tridiagonal matrix are desired, z must contain
           the identity matrix.


       On output


         d contains the eigenvalues in ascending order.  If an
           error exit is made, the eigenvalues are correct but
           unordered for indices 1,2,...,ierr-1.


         e has been destroyed.


         z contains orthonormal eigenvectors of the symmetric
           tridiagonal (or full) matrix.  If an error exit is made,
           z contains the eigenvectors associated with the stored
           eigenvalues.


         ierr is a pointer to an integer error number.  *ierr is set to
           zero       for normal return,
           j          if the j-th eigenvalue has not been
                      determined after 30 iterations.


      Calls pythag for dsqrt(a*a + b*b) .


      Questions and comments should be directed to Burton S. Garbow,
      Mathematics and Computer Science Div, Argonne National Laboratory


      This version dated August 1983.


      ------------------------------------------------------------------
*/

void            tql2(long nm, long n, double *d, double *e, double *z,
					 long *ierr)
/*
double d[n],e[n],z[n][nm]; (C storage mode)
double d(n),e(n),z(nm,n) (Fortran storage mode)
*/
{
	long            i, j, k, l, m, ii, l1, l2;
	double          c, c2, c3, dl1, el1, f, g, h, p, r, s, s2, tst1,
	                tst2;

	*ierr = 0;
	if (n != 1)
	{
		for (i = 1; i < n; i++)
		{
			e[i - 1] = e[i];
		}
		f = 0.0;
		tst1 = 0.0;
		e[n - 1] = 0.0;

		for (l = 0; l < n; l++)
		{
			j = 0;
			h = fabs(d[l]) + fabs(e[l]);
			if (tst1 < h)
			{
				tst1 = h;
			}
/*      .......... look for small sub-diagonal element .......... */
			for (m = l; m < n; m++)
			{
				tst2 = tst1 + fabs(e[m]);
				if (tst2 == tst1)
				{
					break;
				}
/*      .......... e(n-1) is always zero, so there is no exit */
/*                  through the bottom of the loop........... */
			} /*for (m = l; m < n; m++)*/

			if (m != l)
			{
				do
				{
					if (j == 30)
					{
						goto errorExit;
					}
					j++;
/*      .......... form shift .......... */
					l1 = l + 1;
					l2 = l1 + 1;
					g = d[l];
					p = (d[l1] - g) / (2.0 * e[l]);
					r = pythag(p, 1.0);
					d[l] = e[l] / (p + SIGN(r, p));
					d[l1] = e[l] * (p + SIGN(r, p));
					dl1 = d[l1];
					h = g - d[l];
					for (i = l2; i < n; i++)
					{
						d[i] -= h;
					}
					f += h;
/*      .......... ql transformation .......... */
					p = d[m];
					c = 1.0;
					c2 = c;
					el1 = e[l1];
					s = 0.0;
/*      .......... for i=m-1 step -1 until l do -- .......... */
					for (i = m - 1; i >= l; i--)
					{
						c3 = c2;
						c2 = c;
						s2 = s;
						g = c * e[i];
						h = c * p;
						r = pythag(p, e[i]);
						e[i + 1] = s * r;
						s = e[i] / r;
						c = p / r;
						p = c * d[i] - s * g;
						d[i + 1] = h + s * (c * g + s * d[i]);
/*      .......... form vector .......... */
						for (k = 0; k < n; k++)
						{
							h = Q2z(k, i + 1);
							Q2z(k, i + 1) = s * Q2z(k, i) + c * h;
							Q2z(k, i) = c * Q2z(k, i) - s * h;
						}
					}
					p = -s * s2 * c3 * el1 * e[l] / dl1;
					e[l] = s * p;
					d[l] = c * p;
					tst2 = tst1 + fabs(e[l]);
				} while (tst2 > tst1);
			} /*if (m != l)*/
			d[l] += f;
		} /*for (l = 0; l < n; l++)*/

/*      .......... order eigenvalues and eigenvectors .......... */
		for (ii = 1; ii < n; ii++)
		{
			i = ii - 1;
			k = i;
			p = d[i];
			for (j = ii; j < n; j++)
			{
				if (d[j] > p) /* changed '<' to '>' (C. Bingham) */
				{
					k = j;
					p = d[j];
				}
			} /*for (j = ii; j < n; j++)*/
			if (k != i)
			{
				d[k] = d[i];
				d[i] = p;
				for (j = 0; j < n; j++)
				{
					p = Q2z(j, i);
					Q2z(j, i) = Q2z(j, k);
					Q2z(j, k) = p;
				}
			} /*if (k != i)*/
		} /*for (ii = 1; ii < n; ii++)*/
	} /*if (n != 1)*/
	return;

/*      .......... set error -- no convergence to an */
/*                 eigenvalue after 30 iterations .......... */
  errorExit:
	*ierr = l;
} /*tql2()*/

/* translated from tqlrat.f */
/*
	Initial translation to C from Ratfor by ratfor2c
	From Eispack Library

	C. Bingham,  kb@umnstat.stat.umn.edu  891215
*/

/*
      This subroutine is a translation of the Algol procedure tqlrat,
      Algorithm 464, Comm. Acm 16, 689(1973) by Reinsch.


      This subroutine finds the eigenvalues of a symmetric
      tridiagonal matrix by the rational ql method.


      On input


         n is the order of the matrix.


         d contains the diagonal elements of the input matrix.


         e2 contains the squares of the subdiagonal elements of the
           input matrix in its last n-1 positions.  e2(1) is arbitrary.


       On output


         d contains the eigenvalues in ascending order.  If an
           error exit is made, the eigenvalues are correct and
           ordered for indices 1,2,...ierr-1, but may not be
           the smallest eigenvalues.


         e2 has been destroyed.


         ierr is a pointer to an integer error number.  *ierr is set to
           zero       for normal return,
           j          if the j-th eigenvalue has not been
                      determined after 30 iterations.


      Calls pythag for dsqrt(a*a + b*b) .


      Questions and comments should be directed to Burton S. Garbow,
      Mathematics and Computer Science Div, Argonne National Laboratory


      This version dated August 1983.


      ------------------------------------------------------------------
*/

void            tqlrat(long n, double *d, double *e2, long *ierr)
/*double d[n],e2[n];*/
{
	long            i, j, l, m, l1;
	double          b, c, f, g, h, p, r, s, t;

	*ierr = 0;
	if (n != 1)
	{
		for (i = 1; i < n; i++)
		{
			e2[i - 1] = e2[i];
		}
		f = 0.0;
		t = 0.0;
		e2[n - 1] = 0.0;

		for (l = 0; l < n; l++)
		{
			j = 0;
			h = fabs(d[l]) + sqrt(e2[l]);
			if (t <= h)
			{
				t = h;
				b = epslon(t);
				c = b * b;
			} /*if (t <= h)*/
/*      .......... look for small squared sub-diagonal element .......... */
			for (m = l; m < n; m++)
			{
				if (e2[m] <= c)
				{
					break;
				}
			} /*for (m = l; m < n; m++)*/

			if (m != l)
			{
				do
				{
					if (j == 30)
					{
						goto errorExit;
					}
					j++;
/*      .......... form shift .......... */
					l1 = l + 1;
					s = sqrt(e2[l]);
					g = d[l];
					p = (d[l1] - g) / (2.0 * s);
					r = pythag(p, 1.0);
					d[l] = s / (p + SIGN(r, p));
					h = g - d[l];
					for (i = l1; i < n; i++)
					{
						d[i] -= h;
					}
					f += h;
/*      .......... rational ql transformation .......... */
					g = d[m];
					if (g == 0.0)
					{
						g = b;
					}
					h = g;
					s = 0.0;
/*      .......... for i=m-1 step -1 until l do -- .......... */
					for (i = m - 1; i >= l; i--)
					{
						p = g * h;
						r = p + e2[i];
						e2[i + 1] = s * r;
						s = e2[i] / r;
						d[i + 1] = h + s * (h + d[i]);
						g = d[i] - e2[i] / g;
						if (g == 0.0)
						{
							g = b;
						}
						h = g * p / r;
					}
					e2[l] = s * g;
					d[l] = h;
/*      .......... guard against underflow in convergence test .......... */
					if (h == 0.0)
					{
						break;
					}
					if (fabs(e2[l]) <= fabs(c / h))
					{
						break;
					}
					e2[l] = h * e2[l];
				} while (e2[l] != 0.0);
			} /*if (m != l)*/
			p = d[l] + f;
/*      .......... order eigenvalues .......... */
/*      .......... for i=l step -1 until 2 do -- .......... */
			for (i = l; i >= 1; i--)
			{
				if (p <= d[i - 1]) /* changed '>=' to '<=' (c. bingham) */
				{
					break;
				}
				d[i] = d[i - 1];
			} /*for (i = l; i >= 1; i--)*/
			d[i] = p;
		} /*for (l = 0; l < n; l++)*/
	} /*if (n != 1)*/
	return;
	
  errorExit:
/*      .......... set error -- no convergence to an */
/*                 eigenvalue after 30 iterations .......... */
	*ierr = l;
} /*tqlrat()*/

/* translated from tred1.f */

/*
	Initial translation to C from Ratfor by ratfor2c
	From Eispack Library

	C. Bingham,  kb@umnstat.stat.umn.edu  891215
	Compile with -DFORTRAN for fortran storage format
*/


/*
      This subroutine is a translation of the Algol procedure tred1,
      Num. Math. 11, 181-195(1968) by Martin, Reinsch, and Wilkinson.
      Handbook for Auto. Comp., Vol.Ii-Linear Algebra, 212-226(1971).


      This subroutine reduces a real symmetric matrix
      to a symmetric tridiagonal matrix using
      orthogonal similarity transformations.


      On input


         nm must be set to the row dimension (column dimension ifndef
		   FORTRAN) of two-dimensional array parameters as declared in the
		   calling program dimension statement.


         n is the order of the matrix.


         a contains the real symmetric input matrix.  Only the
           lower triangle of the matrix need be supplied.


      On output


         a contains information about the orthogonal trans-
           formations used in the reduction in its strict lower
           triangle.  The full upper triangle of a is unaltered.


         d contains the diagonal elements of the tridiagonal matrix.


         e contains the subdiagonal elements of the tridiagonal
           matrix in its last n-1 positions.  e(1) is set to zero.


         e2 contains the squares of the corresponding elements of e.
           e2 may coincide with e if the squares are not needed.


      Questions and comments should be directed to Burton S. Garbow,
      Mathematics and Computer Science Div, Argonne National Laboratory


      This version dated August 1983.


      ------------------------------------------------------------------
*/

void            tred1(long nm, long n, double *a, double *d, double *e,
					  double *e2)
/*double a[n][nm],d[n],e[n],e2[n];*/
{
	long            i, j, k, l, jp1;
	double          f, g, h, scale;

	for (i = 0; i < n; i++)
	{
		d[i] = Q2a(n - 1, i);
		Q2a(n - 1, i) = Q2a(i, i);
	} /*for (i = 0; i < n; i++)*/

/*      .......... for i=n step -1 until 1 do -- .......... */
	for (i = n - 1; i >= 0; i--)
	{
		l = i - 1;
		h = 0.0;
		scale = 0.0;
		if (l >= 0)
		{
/*      .......... scale row (algol tol then not needed) .......... */
			for (k = 0; k <= l; k++)
			{
				scale += fabs(d[k]);
			}
			
			if (scale == 0.0)
			{
				for (j = 0; j <= l; j++)
				{
					d[j] = Q2a(l, j);
					Q2a(l, j) = Q2a(i, j);
					Q2a(i, j) = 0.0;
				}
			} /*if (scale == 0.0)*/
			else
			{
				for (k = 0; k <= l; k++)
				{
					d[k] /= scale;
					h += d[k] * d[k];
				}
				e2[i] = scale * scale * h;
				f = d[l];
				g = -SIGN(sqrt(h), f);
				e[i] = scale * g;
				h -= f * g;
				d[l] = f - g;
				if (l != 0)
				{
/*      .......... form a*u .......... */
					for (j = 0; j <= l; j++)
					{
						e[j] = 0.0;
					}
					for (j = 0; j <= l; j++)
					{
						f = d[j];
						g = e[j] + Q2a(j, j) * f;
						jp1 = j + 1;
						if (l >= jp1)
						{
							for (k = jp1; k <= l; k++)
							{
								g += Q2a(k, j) * d[k];
								e[k] += Q2a(k, j) * f;
							}
						}
						e[j] = g;
					}
/*      .......... form p .......... */
					f = 0.0;
					for (j = 0; j <= l; j++)
					{
						e[j] /= h;
						f += e[j] * d[j];
					}
					h = f / (h + h);
/*      .......... form q .......... */
					for (j = 0; j <= l; j++)
					{
						e[j] -= h * d[j];
					}
/*      .......... form reduced a .......... */
					for (j = 0; j <= l; j++)
					{
						f = d[j];
						g = e[j];
						for (k = j; k <= l; k++)
						{
							Q2a(k, j) -= f * e[k] + g * d[k];
						}
					}
				}
				for (j = 0; j <= l; j++)
				{
					f = d[j];
					d[j] = Q2a(l, j);
					Q2a(l, j) = Q2a(i, j);
					Q2a(i, j) = f * scale;
				}
				continue;
			} /*if (scale == 0.0){}else{}*/
		} /*if (l >= 0)*/
		e[i] = 0.0;
		e2[i] = 0.0;
	} /*for (i = n - 1; i >= 0; i--)*/
} /*tred1()*/

/* translated from tred2.f */

/*
	Initial translation to C from Ratfor by ratfor2c
	From Eispack Library

	C. Bingham,  kb@umnstat.stat.umn.edu  891215
	Compile with -DFORTRAN for fortran storage format
*/


/*
      This subroutine is a translation of the Algol procedure tred2,
      Num. Math. 11, 181-195(1968) by Martin, Reinsch, and Wilkinson.
      Handbook for Auto. Comp., Vol.Ii-Linear Algebra, 212-226(1971).


      This subroutine reduces a real symmetric matrix to a
      symmetric tridiagonal matrix using and accumulating
      orthogonal similarity transformations.


      On input


         nm must be set to the row dimension (column dimension ifndef
		   FORTRAN) of two-dimensional array parameters as declared in the
		   calling program dimension statement.


         n is the order of the matrix.


         a contains the real symmetric input matrix.  Only the
           lower triangle of the matrix need be supplied.


      On output


         d contains the diagonal elements of the tridiagonal matrix.


         e contains the subdiagonal elements of the tridiagonal
           matrix in its last n-1 positions.  e(1) is set to zero.


         z contains the orthogonal transformation matrix
           produced in the reduction.


         a and z may coincide.  If distinct, a is unaltered.


      Questions and comments should be directed to Burton S. Garbow,
      Mathematics and Computer Science Div, Argonne National Laboratory


      This version dated August 1983.


      ------------------------------------------------------------------
*/

void            tred2(long nm, long n, double *a, double *d, double *e,
					  double *z)
/*double a[n][nm],d[n],e[n],z[n][nm];*/
{
	long            i, j, k, l, jp1;
	double          f, g, h, hh, scale;

	for (i = 0; i < n; i++)
	{
		for (j = i; j < n; j++)
		{
			Q2z(j, i) = Q2a(j, i);
		}
		d[i] = Q2a(n - 1, i);
	} /*for (i = 0; i < n; i++)*/

	if (n != 1)
	{
/*      .......... for i=n step -1 until 2 do -- .......... */
		for (i = n - 1; i >= 1; i--)
		{
			l = i - 1;
			h = 0.0;
			scale = 0.0;
			if (l >= 1)
			{
/*      .......... scale row (algol tol then not needed) .......... */
				for (k = 0; k <= l; k++)
				{
					scale += fabs(d[k]);
				}
				if (scale != 0.0)
				{
					for (k = 0; k <= l; k++)
					{
						d[k] /= scale;
						h += d[k] * d[k];
					}
					f = d[l];
					g = -SIGN(sqrt(h), f);
					e[i] = scale * g;
					h -= f * g;
					d[l] = f - g;
/*      .......... form a*u .......... */
					for (j = 0; j <= l; j++)
					{
						e[j] = 0.0;
					}
					for (j = 0; j <= l; j++)
					{
						f = d[j];
						Q2z(j, i) = f;
						g = e[j] + Q2z(j, j) * f;
						jp1 = j + 1;
						if (l >= jp1)
						{
							for (k = jp1; k <= l; k++)
							{
								g += Q2z(k, j) * d[k];
								e[k] += Q2z(k, j) * f;
							}
						}
						e[j] = g;
					}
/*      .......... form p .......... */
					f = 0.0;
					for (j = 0; j <= l; j++)
					{
						e[j] /= h;
						f += e[j] * d[j];
					}
					hh = f / (h + h);
/*      .......... form q .......... */
					for (j = 0; j <= l; j++)
					{
						e[j] -= hh * d[j];
					}
/*      .......... form reduced a .......... */
					for (j = 0; j <= l; j++)
					{
						f = d[j];
						g = e[j];
						for (k = j; k <= l; k++)
						{
							Q2z(k, j) -= f * e[k] + g * d[k];
						}
						d[j] = Q2z(l, j);
						Q2z(i, j) = 0.0;
					} /*for (j = 0; j <= l; j++)*/
					d[i] = h;
					continue;
				} /*if (scale != 0.0)*/
			} /*if (l >= 1)*/
			e[i] = d[l];
			for (j = 0; j <= l; j++)
			{
				d[j] = Q2z(l, j);
				Q2z(i, j) = 0.0;
				Q2z(j, i) = 0.0;
			}
			d[i] = h;
		} /*for (i = n - 1; i >= 1; i--)*/

/*      .......... accumulation of transformation matrices .......... */
		for (i = 1; i < n; i++)
		{
			l = i - 1;
			Q2z(n - 1, l) = Q2z(l, l);
			Q2z(l, l) = 1.0;
			h = d[i];
			if (h != 0.0)
			{
				for (k = 0; k <= l; k++)
				{
					d[k] = Q2z(k, i) / h;
				} /*for (k = 0; k <= l; k++)*/
				for (j = 0; j <= l; j++)
				{
					g = 0.0;
					for (k = 0; k <= l; k++)
					{
						g += Q2z(k, i) * Q2z(k, j);
					}
					for (k = 0; k <= l; k++)
					{
						Q2z(k, j) -= g * d[k];
					} /*for (k = 0; k <= l; k++)*/
				} /*for (j = 0; j <= l; j++)*/
			} /*if (h != 0.0)*/
			for (k = 0; k <= l; k++)
			{
				Q2z(k, i) = 0.0;
			}
		} /*for (i = 1; i < n; i++)*/
	} /*if (n != 1)*/

	for (i = 0; i < n; i++)
	{
		d[i] = Q2z(n - 1, i);
		Q2z(n - 1, i) = 0.0;
	} /*for (i = 0; i < n; i++)*/

	Q2z(n - 1, n - 1) = 1.0;
	e[0] = 0.0;

} /*tred2()*/

#ifdef MPW
#ifdef MPW3
#pragma segment Svd
#else
#define __SEG__ Svd
#endif /*MPW3*/
#endif /*MPW*/

#ifdef FORTRAN
#define Q2u(I,J) u[(I) + nm*(J)]
#define Q2v(I,J) v[(I) + n*(J)]
#else
#define Q2u(I,J) u[(J) + nm*(I)]
#define Q2v(I,J) v[(J) + n*(I)]
#endif

/*
      This subroutine is a translation of the Algol procedure svd,
      Num. Math. 14, 403-420(1970) by Golub and Reinsch.
      Handbook for Auto. Comp., Vol Ii-Linear Algebra, 134-151(1971).

      This subroutine determines the singular value decomposition
           t
      a=usv  of a real m by n rectangular matrix.  Householder
      bidiagonalization and a variant of the QR algorithm are used.

      On input

         nm must be set to the row dimension of two-dimensional
           array parameters as declared in the calling program
           dimension statement.  Note that nm must be at least
           as large as the maximum of m and n.

         m is the number of rows of a (and u).

         n is the number of columns of a (and u) and the order of v.

         a contains the rectangular input matrix to be decomposed.

         matu should be set to .true. if the u matrix in the
           decomposition is desired, and to .false. otherwise.

         matv should be set to .true. if the v matrix in the
           decomposition is desired, and to .false. otherwise.

      On output

         a is unaltered (unless overwritten by u or v).

         w contains the n (non-negative) singular values of a (the
           diagonal elements of s).  They are unordered.  If an
           error exit is made, the singular values should be correct
           for indices ierr+1,ierr+2,...,n.

         u contains the matrix u (orthogonal column vectors) of the
           decomposition if matu has been set to .true.  Otherwise
           u is used as a temporary array.  u may coincide with a.
           If an error exit is made, the columns of u corresponding
           to indices of correct singular values should be correct.

         v contains the matrix v (orthogonal) of the decomposition if
           matv has been set to .true.  Otherwise v is not referenced.
           v may also coincide with a if u is not needed.  If an error
           exit is made, the columns of v corresponding to indices of
           correct singular values should be correct.

         ierr is set to
           zero       for normal return,
           k          if the k-th singular value has not been
                      determined after 30 iterations.

         rv1 is a temporary storage array.

      Calls pythag for  sqrt(a*a + b*b).

      Questions and comments should be directed to Burton S. Garbow,
      Mathematics and Computer Science Div, Argonne National Laboratory

      This version dated August 1983.

      ------------------------------------------------------------------
*/

/*
  Modified so that v is assumed to have to have row dimension n rather than
  nm.  Name changed from svd to eispsvd
  Note that nm is used by macro Q2u(I,j).
*/

void eispsvd(long nm, long m, long n, double *a, double *w,
			 long matu, double *u, long matv, double *v, double *rv1,
			 long *ierr)
/* a(nm,n), w(n), u(nm,n), v(nm,n), rv1(n)  in original Fortran */
{
	int ii, i, j, k, l, i1, k1, l1, mn, its;
	double c, f, g, h, s, x, y, z, tst1, tst2, scale;
	WHERE("eispsvd");
	
	for (i = 0;i < m ;i++)
	{
		for (j = 0;j < n ;j++)
		{
			Q2u(i,j) = Q2a(i,j);
		}
	} /*for (i = 0;i < m ;i++)*/
/*      .......... Householder reduction to bidiagonal form .......... */
	g = 0.0;
	scale = 0.0;
	x = 0.0;

	for (i = 0;i < n ;i++)
	{
		l = i + 1;
		rv1[i] = scale*g;
		g = 0.0;
		s = 0.0;
		scale = 0.0;
		if (i < m) 
		{
			for (k = i;k < m ;k++)
			{
				scale += fabs(Q2u(k,i));
			} /*for (k = i;k < m ;k++)*/
			
			if (scale != 0.0) 
			{
				for (k = i;k < m ;k++)
				{
					Q2u(k,i) /= scale;
					s += Q2u(k,i)*Q2u(k,i);
				}

				f = Q2u(i,i);
				g = -SIGN(sqrt(s),f);
				h = f*g - s;
				Q2u(i,i) = f - g;
				for (j = l;j < n ;j++)
				{
					s = 0.0;
					for (k = i;k < m ;k++)
					{
						s += Q2u(k,i)*Q2u(k,j);
					}
					f = s/h;

					for (k = i;k < m ;k++)
					{
						Q2u(k,j) += f*Q2u(k,i);
					}
				} /*for (j = l;j < n ;j++)*/

				for (k = i;k < m ;k++)
				{
					Q2u(k,i) *= scale;
				}
			} /*if (scale != 0.0) */
		} /*if (i < m) */

		w[i] = scale*g;
		g = 0.0;
		s = 0.0;
		scale = 0.0;
		if (i < m && i != n - 1) 
		{
			for (k = l;k < n ;k++)
			{
				scale += fabs(Q2u(i,k));
			} /*for (k = l;k < n ;k++)*/
			if (scale != 0.0) 
			{
				for (k = l;k < n ;k++)
				{
					Q2u(i,k) /= scale;
					s += Q2u(i,k)*Q2u(i,k);
				} /*for (k = l;k < n ;k++)*/

				f = Q2u(i,l);
				g = -SIGN(sqrt(s),f);
				h = f*g - s;
				Q2u(i,l) = f - g;

				for (k = l;k < n ;k++)
				{
					rv1[k] = Q2u(i,k)/h;
				} /*for (k = l;k < n ;k++)*/
				for (j = l;j < m ;j++)
				{
					s = 0.0;

					for (k = l;k < n ;k++)
					{
						s += Q2u(j,k)*Q2u(i,k);
					} /*for (k = l;k < n ;k++)*/
					
					for (k = l;k < n ;k++)
					{
						Q2u(j,k) += s*rv1[k];
					} /*for (k = l;k < n ;k++)*/
				} /*for (j = l;j < m ;j++)*/

				for (k = l;k < n ;k++)
				{
					Q2u(i,k) *= scale;
				} /*for (k = l;k < n ;k++)*/
			} /*if (scale != 0.0) */
		} /*if (i < m && i != n - 1) */

		x = MAX(x,fabs(w[i])+fabs(rv1[i]));
	} /*for (i = 0;i < n ;i++)*/

/*      .......... accumulation of right-hand transformations .......... */
	if (matv)
	{
/*      .......... for i=n step -1 until 1 do -- .......... */
		for (i = n - 1;i >= 0;i--)
		{
			l = i + 1;
			if (i != n - 1)
			{
				if (g != 0.0) 
				{
					for (j = l;j < n ;j++)
					{
/*      .......... double division avoids possible underflow .......... */
						Q2v(j,i) = (Q2u(i,j)/Q2u(i,l))/g;
					} /*for (j = l;j < n ;j++)*/
					for (j = l;j < n ;j++)
					{ 
						s = 0.0;

						for (k = l;k < n ;k++)
						{
							s += Q2u(i,k)*Q2v(k,j);
						}
						
						for (k = l;k < n ;k++)
						{
							Q2v(k,j) += s*Q2v(k,i);
						}
					} /*for (j = l;j < n ;j++)*/
				} /*if (g != 0.0) */

				for (j = l;j < n ;j++)
				{
					Q2v(i,j) = 0.0;
					Q2v(j,i) = 0.0;
				}
			} /*if (i != n - 1) */

			Q2v(i,i) = 1.0;
			g = rv1[i];
		} /*for (i = n - 1;i >= 0;i--)*/
	} /*if (matv)*/
	
/*      .......... accumulation of left-hand transformations .......... */
	if (matu) 
	{
/*      ..........for i = min(m,n) step -1 until 1 do -- .......... */
		mn = (m < n) ? m : n;
		for (i = mn - 1;i >= 0;i--)
		{
			l = i + 1;
			g = w[i];
			for (j = l;j < n ;j++)
			{
				Q2u(i,j) = 0.0;
			}
			
			if (g == 0.0)
			{
				for (j = i;j < m ;j++) 
				{
					Q2u(j,i) = 0.0;
				}
			} /*if (g == 0.0)*/
			else 
			{
				for (j = l;j < n ;j++)
				{ 
					s = 0.0;

					for (k = l;k < m ;k++)
					{
						s += Q2u(k,i)*Q2u(k,j);
					} /*for (k = l;k < m ;k++)*/
/*      .......... double division avoids possible underflow .......... */
					f = (s/Q2u(i,i))/g;

					for (k = i;k < m ;k++)
					{
						Q2u(k,j) += f*Q2u(k,i);
					} /*for (k = i;k < m ;k++)*/
				} /*for (j = l;j < n ;j++)*/

				for (j = i;j < m ;j++)
				{
					Q2u(j,i) /= g;
				}
			} /*if (g == 0.0){}else{}*/
			Q2u(i,i) += 1.0;
		} /*for (i = mn - 1;i >= 0;i--)*/
	} /*if (matu) */
/*      .......... diagonalization of the bidiagonal form .......... */
	tst1 = x;
/*      .......... for k=n step -1 until 1 do -- .......... */
	for (k = n - 1;k >= 0;k--)
	{
		k1 = k - 1;
		its = 0;
		while (1)
		{/*      .......... test for splitting. */
/*                 for l=k step -1 until 1 do -- .......... */
			for (l = k;l >= 0;l--)
			{
				l1 = l - 1;
				tst2 = tst1 + fabs(rv1[l]);
				if (tst2 == tst1)
				{
					goto LAB10;
				}
/*      .......... rv1[0] is always zero, so there is no exit */
/*                 through the bottom of the loop .......... */
				tst2 = tst1 + fabs(w[l1]);
				if (tst2 == tst1)
				{
					break;
				}
			} /*for (l = k;l >= 0;l--)*/

/*      .......... cancellation of rv1[l] if l greater than 1 .......... */
			c = 0.0;
			s = 1.0;

			for (i = l;i <= k ;i++)
			{
				f = s*rv1[i];
				rv1[i] = c*rv1[i];
				tst2 = tst1 + fabs(f);
				if (tst2 == tst1)
				{
					break;
				}
				g = w[i];
				h = pythag(f,g);
				w[i] = h;
				c = g/h;
				s = -f/h;
				if (matu)
				{
					for (j = 0;j < m ;j++)
					{
						y = Q2u(j,l1);
						z = Q2u(j,i);
						Q2u(j,l1) = y*c + z*s;
						Q2u(j,i) = -y*s + z*c;
					} /*for (j = 0;j < m ;j++)*/
				} /*if (matu)*/
			} /*for (i = l;i <= k ;i++)*/
/*      .......... test for convergence .......... */
		  LAB10:
			z = w[k];
			if (l == k)
			{
				break;
			}
/*      .......... shift from bottom 2 by 2 minor .......... */
			if (its == 30)
			{
				goto errorExit;
			}
			its++;
			x = w[l];
			y = w[k1];
			g = rv1[k1];
			h = rv1[k];
			f = 0.5*(((g + z)/h)*((g - z)/y) + y/h - h/y);
			g = pythag(f,1.0);
			f = x - (z/x)*z + (h/x)*(y/(f + SIGN(g,f)) - h);
/*      .......... next qr transformation .......... */
			c = 1.0;
			s = 1.0;

			for (i1 = l;i1 <= k1 ;i1++)
			{
				i = i1 + 1;
				g = rv1[i];
				y = w[i];
				h = s*g;
				g = c*g;
				z = pythag(f,h);
				rv1[i1] = z;
				c = f/z;
				s = h/z;
				f = x*c + g*s;
				g =  -x*s + g*c;
				h = y*s;
				y = y*c;
				if (matv)
				{
					for (j = 0;j < n ;j++)
					{
						x = Q2v(j,i1);
						z = Q2v(j,i);
						Q2v(j,i1) = x*c + z*s;
						Q2v(j,i) = -x*s + z*c;
					}
				}
				
				z = pythag(f,h);
				w[i1] = z;
/*      .......... rotation can be arbitrary if z is zero .......... */
				if (z != 0.0) 
				{
					c = f/z;
					s = h/z;
				}
				f = c*g + s*y;
				x = -s*g + c*y;
				if (matu)
				{
					for (j = 0;j < m ;j++)
					{
						y = Q2u(j,i1);
						z = Q2u(j,i);
						Q2u(j,i1) = y*c + z*s;
						Q2u(j,i) = -y*s + z*c;
					} /*for (j = 0;j < m ;j++)*/
				} /*if (matu)*/
			}/*for (i1 = l;i1 <= k1 ;i1++)*/

			rv1[l] = 0.0;
			rv1[k] = f;
			w[k] = x;
		} /*while (1)*/
/*      .......... convergence .......... */
		if (z < 0.0) 
		{
/*      .......... w[k] is made non-negative .......... */
			w[k] = -z;
			if (matv)
			{
				for (j = 0;j < n ;j++) 
				{
					Q2v(j,k) = -Q2v(j,k);
				}
			}
		} /*if (z < 0.0) */
	} /*for (k = n - 1;k >= 0;k--)*/

/*      .......... order singular values and singular vectors (added by kb) */
	for (ii = 1; ii < n; ii++)
	{
		i = ii - 1;
		k = i;
		c = w[i];
		for (j = ii; j < n; j++)
		{
			if (w[j] > c)
			{
				k = j;
				c = w[j];
			}
		} /*for (j = ii; j < n; j++)*/

		if (k != i)
		{
			w[k] = w[i];
			w[i] = c;
			if (matu)
			{
				for (j = 0; j < m; j++)
				{
					c = Q2u(j, i);
					Q2u(j, i) = Q2u(j, k);
					Q2u(j, k) = c;
				}
			} /*if (matu)*/
			
			if (matv)
			{
				for (j = 0; j < n; j++)
				{
					c = Q2v(j, i);
					Q2v(j, i) = Q2v(j, k);
					Q2v(j, k) = c;
				} /*for (j = 0; j < n; j++)*/
			} /*if (matv)*/
			
		} /*if (k != i)*/
	} /*for (ii = 1; ii < n; ii++)*/
	*ierr = 0;
	return;
	
/*
      .......... set error -- no convergence to a
                 singular value after 30 iterations ..........
*/
  errorExit:
	*ierr = k;
	
} /*eispsvd()*/

#ifdef MPW
#ifdef MPW3
#pragma segment Trideige
#else
#define __SEG__ Trideige
#endif /*MPW3*/
#endif /*MPW*/

/* Routines used by trideigen() */

/*
	Translated to C by C. Bingham (kb@stat.umn.edu) 950516
*/

/*
 ***Begin prologue  tridib
 ***Date written   760101   (yymmdd)
 ***Revision date  830518   (yymmdd)
 ***Category No.  D4A5,D4C2A
 ***Keywords  eigenvalues,eigenvectors,eispack
 ***Author  Smith, B. T., et al.
 ***Purpose  computes eigenvalues of symmetric tridiagonal matrix
             given interval using Sturm sequencing.
 ***Description

      This subroutine is a translation of the ALGOL procedure bisect,
      Num. Math. 9, 386-393(1967) by Barth, Martin, and Wilkinson.
      Handbook for Auto. Comp., Vol.II-Linear Algebra, 249-256(1971).

      This subroutine finds those eigenvalues of a tridiagonal
      symmetric matrix between specified boundary indices,
      using bisection.

      On input

         n is the order of the matrix.

         eps1 is an absolute error tolerance for the computed
           eigenvalues.  if the input eps1 is non-positive,
           it is reset for each submatrix to a default value,
           namely, minus the product of the relative machine
           precision and the 1-norm of the submatrix.

         d contains the diagonal elements of the input matrix.

         e contains the subdiagonal elements of the input matrix
           in its last n-1 positions.  e(1) is arbitrary.

         e2 contains the squares of the corresponding elements of e.
           e2(1) is arbitrary.

         m11 specifies the lower boundary index for the desired
           eigenvalues.

         m specifies the number of eigenvalues desired.  the upper
           boundary index m22 is then obtained as m22=m11+m-1.

      on output

         eps1 is unaltered unless it has been reset to its
           (last) default value.

         d and e are unaltered.

         elements of e2, corresponding to elements of e regarded
           as negligible, have been replaced by zero causing the
           matrix to split into a direct sum of submatrices.
           e2(1) is also set to zero.

         lb and ub define an interval containing exactly the desired
           eigenvalues.

         w contains, in its first m positions, the eigenvalues
           between indices m11 and m22 in ascending order.

         ind contains in its first m positions the submatrix indices
           associated with the corresponding eigenvalues in w --
           1 for eigenvalues belonging to the first submatrix from
           the top, 2 for those belonging to the second submatrix, etc.

         ierr is set to
           zero       for normal return,
           3*n+1      if multiple eigenvalues at index m11 make
                      unique selection impossible,
           3*n+2      if multiple eigenvalues at index m22 make
                      unique selection impossible.

         rv4 and rv5 are temporary storage arrays.

      Note that subroutine tql1, imtql1, or tqlrat is generally faster
      than tridib, if more than n/4 eigenvalues are to be found.

      Questions and comments should be directed to B. S. Garbow,
      Applied Mathematics Division, Argonne National Laboratory
      ------------------------------------------------------------------
 ***References  B. T. Smith, J. M. Boyle, J. J. Dongarra, B. S. Garbow,
                  Y. Ikebe, V. C. Klema, C. B. Moler, *Matrix Eigen-
                  System Routines - EISPACK Guide*, Springer-Verlag,
                  1976.
 ***routines called  (none)
 ***end prologue  tridib
*/

/*
  Comments on C-version:
   All indices now have 0 origin.  This includes i, j, p, q, r, s, m1, m2

   Also argument m11 is assumed relative to 0 origin

   Many of the original line numbers are present in comments of the form
   #nnn at the end of a line

*/

void tridib(long n, double *eps1, double *d, double *e, double *e2,
			double *lb,double *ub, long m11, long m, double *w,
			long *ind, long *ierr, double *rv4, double *rv5)
/*double d[n],e[n],e2[n],w[m],rv4[n],rv5[n];*/
/*int    ind[m]*/
{
	long            i, j, k, l, p, q, r, s, m1, m2, m22, tag, isturm = 0;
	int            breaker = 0, continuer = 0;
	double         u, v, t1, t2, xu, x0, x1, s1, s2, kluge;
	static double  machep = 1.0;
	WHERE("tridib");

	/* ***first executable statement  tridib */
	if (machep == 1.0) 
	{
/*
   Note: kluge is used to avoid problem of register arithmetic have more
   significant digits that double can hold, as on Macintosh
*/
#ifdef DJGPP
		machep = epslon(1.0);
#else /*DJGPP*/
		do
		{
			machep *= 0.5; /*#5*/
			kluge = 1.0 + machep;
		} while (kluge > 1.0);
#endif /*DJGPP*/
		machep *= machep;
	} /*if (machep == 1.0) */
	*ierr = 0; /*#10*/
	tag = 0;
	xu = d[0];
	x0 = d[0];
	u = 0.0;

/*
      .......... Look for small sub-diagonal entries and determine an
                 interval containing all the eigenvalues ..........
*/
	for (i=0;i < n ;i++)
	{
		x1 = u;
		u = (i == n - 1) ? 0.0 : fabs(e[i + 1]);
		xu = MIN(d[i] - (x1 + u), xu);
		x0 = MAX(d[i] + (x1 + u), x0);
		if (i > 0) 
		{
			s1 = fabs(d[i]) + fabs(d[i - 1]);
			s2 = s1 + fabs(e[i]);
			if (s2 > s1)
			{
				continue;
			}
		} /*if (i > 0) */
		e2[i] = 0.0; /*#20*/
	} /*for (i=1;i <= n ;i++)*/ /*#40*/

	x1 = MAX(fabs(xu), fabs(x0))*machep*(double) n;
	xu -= x1;
	t1 = xu;
	x0 += x1;
	t2 = x0;
/*
      .......... Determine an interval containing exactly
                 the desired eigenvalues ..........
*/
	p = 0;
	q = n-1;
	m1 = m11 - 1;
	if (m1 >= 0) 
	{
		isturm = 1;
		goto errorCheck;
	}

	while (1)
	{
		m22 = m1 + m; /*#75*/
		if (m22 != n - 1) 
		{
			x0 = t2;
			isturm = 2;
			goto errorCheck;
		} /*if (m22 != n) */

		while (2)
		{
			q = -1; /*#90*/
			r = -1;
			while (3)
			{
/*
      .......... Establish and process next submatrix, refining
                 interval by the Gerschgorin bounds ..........
*/
				if (r == m-1) /*#100*/
				{
					goto exitPoint;
				}
				tag++;
				p = q + 1;
				xu = d[p];
				x0 = d[p];
				u = 0.0;

				for (q=p;q < n ;q++)
				{
					x1 = u;
					u = 0.0;
					v = 0.0;
					if (q < n - 1) 
					{
						u = fabs(e[q + 1]);
						v = e2[q + 1];
					}
					xu = MIN(d[q] - (x1 + u), xu); /*#110*/
					x0 = MAX(d[q] + (x1 + u), x0);
					if (v == 0.0)
					{
						break;
					}
				} /*for (q=p;q < n ;q++)*/ /*#120*/

				x1 = MAX(fabs(xu), fabs(x0))*machep; /*#140*/
				if (*eps1 <= 0.0)
				{
					*eps1 = -x1;
				}
				if (p != q) 
				{
					x1 *= (double) (q - p + 1); /*#180*/
					*lb = MAX(t1, xu - x1);
					*ub = MIN(t2, x0 + x1);
					x1 = *lb;
					isturm = 3;

					while (4)
					{
/*      .......... In-line procedure for Sturm sequence .......... */
						s = p - 1; /*#320*/
						u = 1.0;

						for (i=p;i <= q ;i++)
						{
							if (u != 0.0)
							{
								v = e2[i]/u; /*#325*/
							} /*if (u != 0.0)*/
							else 
							{
								v = fabs(e[i])/machep;
								if (e2[i] == 0.0)
								{
									v = 0.0;
								}
							} /*if (u != 0.0){}else{}*/
							u = d[i] - x1 - v; /*#330*/
							if (u < 0.0)
							{
								s++;
							}
						} /*for (i=p;i <= q ;i++)*/ /*#340*/

						switch(isturm)
						{
						  case 1:
							if (s == m1) /*#60*/
							{
								breaker = 4;
								break;
							} /*if (s == m1)*/
							else
							{
								if (s < m1)
								{
									xu = x1; /*#65*/
								}
								else
								{
									x0 = x1; /*#70*/
								}
								goto errorCheck;
							} /*if (s == m1){}else{}*/ 

						  case 2:
							if (s == m22) /*#80*/
							{
								breaker = 3;
								break;
							}
							else
							{
								if (s < m22)
								{
									xu = x1; /*#65*/
								}
								else
								{
									x0 = x1; /*#70*/
								}
								goto errorCheck;
							}

						  case 3:
							m1 = s + 1; /*#200*/
							x1 = *ub;
							isturm = 4;
							continue;

						  case 4:
							m2 = s; /*#220*/
							if (m1 > m2)
							{
								if (q >= n - 1) /*#940*/
								{
									goto exitPoint;
								}
								continuer = 1;
								breaker = 2;
								break;
							}
/* .......... Find roots by bisection .......... */
							x0 = *ub;
							isturm = 5;

							for (i=m1;i <= m2 ;i++)
							{
								rv5[i] = *ub;
								rv4[i] = *lb;
							} /*#240*/
/*
      .......... Loop for k-th eigenvalue
                 for k=m2 step -1 until m1 do --
                 (-do- not used to legalize -computed go to-) ..........
*/
							k = m2;
							break;
							
						  case 5:
/*      .......... Refine intervals .......... */
							if (s >= k) /*#360*/
							{
								x0 = x1; /*#400*/
							} /*if (s >= k)*/
							else 
							{
								xu = x1;
								if (s < m1)
								{
									rv4[m1] = x1;
								} /*if (s < m1)*/
								else 
								{
									rv4[s + 1] = x1; /*#380*/
									if (rv5[s] > x1)
									{
										rv5[s] = x1;
									}
								} /*if (s < m1){}else{}*/
							} /*if (s >= k){}else{}*/
							goto bisection;
						} /*switch(isturm)*/
						if (breaker-- > 1)
						{
							break;
						}

						while (5)
						{
							xu = *lb; /*#250*/
/*      .......... For i=k step -1 until m1 do -- .......... */
							for (i=k;i >= m1;i--)
							{
								if (xu < rv4[i])
								{
									xu = rv4[i];
									break;
								}
							} /*#260*/		
							if (x0 > rv5[k]) /*#280*/
							{
								x0 = rv5[k];
							}

/*      .......... Next bisection step .......... */
						  bisection: /*#300*/
							x1 = (xu + x0)*0.5;
							s1 = fabs(xu) + fabs(x0) + fabs(*eps1);
#ifdef DJGPP
			/* use epslon() to outfox compiler so convergence is noticed */
							if (fabs(x0 - xu)/2.0 > epslon(s1))
							{
								continuer = 1;
								breaker = 1;
								break;
							}
#else /*DJGPP*/
							s2 = s1 + fabs(x0 - xu)/2.0;
							if (s2 != s1)
							{
								continuer = 1;
								breaker = 1;
								break;
							}
#endif /*DJGPP*/

/*      .......... k-th eigenvalue found .......... */
							rv5[k] = x1; /*#420*/
							k--;
							if (k < m1)
							{
								breaker = 2;
								break;
							}
						} /*while (5)*/

						if (breaker-- > 1)
						{
							break;
						}
						if (continuer > 0)
						{
							continuer = 0;
							continue;
						}

					  errorCheck: /*#50*/
						v = x1;
						x1 = xu + (x0 - xu)*0.5;
						if (x1 == v)
						{
/*
      .......... Set error -- interval cannot be found containing
                 exactly the desired eigenvalues ..........
*/
							*ierr = 3*n + isturm; /*#980*/
							goto exitPoint;
						}
					} /*while (4)*/
					if (breaker-- > 1)
					{
						break;
					}
					if (continuer > 0)
					{
						continuer = 0;
						continue;
					}
				} /*if (p != q) */
				else 
				{
/*      .......... Check for isolated root within interval .......... */
					if (t1 > d[p] || d[p] >= t2)
					{
						if (q >= n - 1) /*#940*/
						{
							goto exitPoint;
						}
						continue;
					}
					m1 = p;
					m2 = p;
					rv5[p] = d[p];
				} /*if (p != q){}else{}*/
/*
      .......... Order eigenvalues tagged with their
                 submatrix associations ..........
*/
				s = r; /*#900*/
				r += m2 - m1 + 1;
				j = 0;
				k = m1;
				for (l=0;l <= r ;l++)
				{
					if (j <= s) 
					{
						if (k > m2)
						{
							break;
						}
						if (rv5[k] >= w[l]) 
						{
							j++; /*#915*/
							continue;
						} /*if (rv5[k] >= w[l]) */
						else
						{
							for (i=l+s-j;i >= l;i--)
							{
								w[i + 1] = w[i];
								ind[i + 1] = ind[i];
							} /*#905*/
						} /*if (rv5[k] >= w[l]){}else{} */
					} /*if (j <= s) */

					w[l] = rv5[k]; /*#910*/
					ind[l] = tag;
					k++;
				} /*for (l=0;l <= r ;l++)*/ /*#920*/

				if (q >= n - 1) /*#940*/
				{
					goto exitPoint;
				}
			} /*while (3)*/
			if (breaker-- > 1)
			{
				break;
			}
			if (continuer > 0)
			{
				continuer = 0;
				continue;
			}
			t2 = x1; /*#85*/
		} /*while (2)*/

		if (breaker-- > 1)
		{
			break;
		}

		xu = x1; /*#73*/
		t1 = x1;
	} /*while (1)*/

  exitPoint:
	*lb = t1; /*#1001*/
	*ub = t2;
} /*tridib()*/


#define MAXIT 5

/*
 ***begin prologue  tinvit
 ***date written   760101   (yymmdd)
 ***revision date  830518   (yymmdd)
 ***category no.  d4c3
 ***keywords  eigenvalues,eigenvectors,eispack
 ***author  smith, b. t., et al.
 ***purpose  eigenvectors of symmetric tridiagonal matrix corresponding
             to some specified eigenvalues, using inverse iteration
 ***description

      This subroutine is a translation of the inverse iteration tech-
      nique in the ALGOL procedure tristurm by Peters and Wilkinson.
      Handbook for Auto. Comp., Vol.II-Linear Algebra, 418-439(1971).

      This subroutine finds those eigenvectors of a tridiagonal
      symmetric matrix corresponding to specified eigenvalues,
      using inverse iteration.

      On input

         nm must be set to the row dimension of two-dimensional
           array parameters as declared in the calling program
           dimension statement.

         n is the order of the matrix.

         d contains the diagonal elements of the input matrix.

         e contains the subdiagonal elements of the input matrix
           in its last n-1 positions.  e(1) is arbitrary.

         e2 contains the squares of the corresponding elements of e,
           with zeros corresponding to negligible elements of e.
           e(i) is considered negligible if it is not larger than
           the product of the relative machine precision and the sum
           of the magnitudes of d(i) and d(i-1).  e2(1) must contain
           0.0e0 if the eigenvalues are in ascending order, or 2.0e0
           if the eigenvalues are in descending order.  if  bisect,
           tridib, or  imtqlv  has been used to find the eigenvalues,
           their output e2 array is exactly what is expected here.

         m is the number of specified eigenvalues.

         w contains the m eigenvalues in ascending or descending order.

         ind contains in its first m positions the submatrix indices
           associated with the corresponding eigenvalues in w --
           1 for eigenvalues belonging to the first submatrix from
           the top, 2 for those belonging to the second submatrix, etc.

      On output

        ** all input arrays are unaltered.**

         z contains the associated set of orthonormal eigenvectors.
           any vector which fails to converge is set to zero.

         ierr is set to
           zero       for normal return,
           -r         if the eigenvector corresponding to the r-th
                      eigenvalue fails to converge in 5 iterations.

         rv1, rv2, rv3, rv4, and rv6 are temporary storage arrays.

      Questions and comments should be directed to B. S. Garbow,
      Applied Mathematics Division, Argonne National Laboratory
      ------------------------------------------------------------------
 ***References  B. T. Smith, J. M. Boyle, J. J. Dongarra, B. S. Garbow,
                  Y. Ikebe, V. C. Klema, C. B. Moler, *Matrix Eigen-
                  System Routines - EISPACK Guide*, Springer-Verlag,
                  1976.
 ***Routines called  (none)
 ***end prologue  tinvit
*/
/*
  Comments on C-version:
   All indices now have 0 origin.  This includes i, j, p, q, r, s, jj

   In case of error on the r-th root, *ierr is set to -(r+1), r = 0,...,m-1

   Many of the original line numbers are present in comments of the form
   #nnn at the end of a line

   Fortran storage convention for z (1-st subscript changing fastest is
   assumed) and is not affected by definition or lack thereof of macro
   FORTRAN.  All reference to z is through pointer z.j
*/

/* NB: This works only using Fortran like storage */

void tinvit(long nm, long n, double *d, double *e, double *e2,
			long m, double *w,long *ind, double *z, long *ierr,
			double *rv1, double *rv2, double *rv3,double *rv4, double *rv6)
{
	/*
	   long ind[m];
	   double d[n], e[n], e2[n], w[m], z[nm][m];
	   double rv1[n], rv2[n], rv3[n], rv4[n], rv6[n];
	   */

	long           i, j, p, q, r, s, jj, its, tag, group = 0;
	double        u, v, uk, xu, x0, x1, eps2, eps3, eps4, norm, order;
	double        kluge, *zj;
	WHERE("tinvit");

	/* ***first executable statement  tinvit */
	*ierr = 0;

	if (m != 0)
	{
		tag = 0;
		order = 1.0 - e2[0];
		q = -1;
		do /*  while(q < n - 1);*/
		{
/*      .......... establish and process next submatrix .......... */
			p = q + 1; /*#100*/

			for (q=p;q < n - 1 && e2[q+1] != 0.0 ;q++)
			{
				;
			} /*for (q=p;q < n - 1 && e2[q] != 0.0 ;q++)*/ /*#120*/
/*      .......... find vectors by inverse iteration .......... */
			tag++; /*#140*/
			s = -1;

			for (r=0;r < m ;r++)
			{
				if (ind[r] == tag)
				{
					its = 1;
					x1 = w[r];
					if (s < 0)
					{
/*      .......... check for isolated root .......... */
						xu = 1.0;
						if (p == q)
						{
							rv6[p] = 1.0;
							goto setZir;
						} /*if (p == q) */
						norm = fabs(d[p]); /*#490*/

						for (i=p+1;i <= q ;i++)
						{
							norm = MAX(norm, fabs(d[i]) + fabs(e[i]));
						} /*for (i=p+1;i <= q ;i++)*/ /*#500*/
/*
  .......... eps2 is the criterion for grouping,
			 eps3 replaces zero pivots and equal
			 roots are modified by eps3,
			 eps4 is taken very small to avoid overflow ..........
*/
						eps2 = .001*norm;
#ifdef DJGPP
						eps3 = epslon(norm);
#else /*DJGPP*/
						eps3 = norm;
						do
						{
							eps3 *= 0.5; /*#502*/
							kluge = norm + eps3;
						} while (kluge > norm);
#endif /*DJGPP*/
						uk = sqrt((double)(q - p + 5));
						eps3 = uk*eps3;
						eps4 = uk*eps3;
						uk = eps4/uk;
						s = p;
						group = 0; /*#505*/
					} /*if (s < 0) */
					else
					{
/*      .......... look for close or coincident roots .......... */
						
						if (fabs(x1 - x0) < eps2) /*#510*/
						{
							group++;
							if (order*(x1 - x0) <= 0.0)
							{
								x1 = x0 + order*eps3;
							}
						} /*if (fabs(x1 - x0) < eps2) */
						else
						{
							group = 0; /*#505*/
						}
					} /*if (s < 0) {}else{}*/
/*
      .......... elimination with interchanges and
                 initialization of vector ..........
*/
					v = 0.0; /*#520*/

					for (i=p;i <= q ;i++)
					{
						rv6[i] = uk;
						if (i > p)
						{
							if (fabs(e[i]) < fabs(u))
							{
								xu = e[i]/u; /*#540*/
								rv4[i] = xu;
								rv1[i - 1] = u;
								rv2[i - 1] = v;
								rv3[i - 1] = 0.0;
							} /*if (fabs(e[i]) < fabs(u)) */
							else
							{
/*
      .......... warning -- a divide check may occur here if
                 e2 array has not been specified correctly ..........
*/
								xu = u/e[i];
								rv4[i] = xu;
								rv1[i - 1] = e[i];
								rv2[i - 1] = d[i] - x1;
								rv3[i - 1] = 0.0;
								if (i != q)
								{
									rv3[i - 1] = e[i + 1];
								}
								u = v - xu*rv2[i - 1];
								v = -xu*rv3[i - 1];
								continue;
							} /*if (fabs(e[i]) < fabs(u)){}else{} */
						} /*if (i > p)*/
						u = d[i] - x1 - xu*v; /*#560*/
						if (i != q)
						{
							v = e[i + 1];
						}
					} /*for (i=p;i <= q ;i++)*/ /*#580*/

					if (u == 0.0)
					{
						u = eps3;
					}
					rv1[q] = u;
					rv2[q] = 0.0;
					rv3[q] = 0.0;

					while (1)
					{
/*
      .......... back substitution
                 for i=q step -1 until p do -- ..........
*/

						for (i=q; i >= p; i--) /*#600*/
						{
							rv6[i] = (rv6[i] - u*rv2[i] - v*rv3[i])/rv1[i];
							v = u;
							u = rv6[i];
						} /*for (i=q; i >= p; i--)*/ /*#620*/
/*
      .......... orthogonalize with respect to previous
                 members of group ..........
*/
						j = r;

						for (jj=0;jj < group ;jj++)
						{
							do
							{
								j--; /*#630*/
							} while (ind[j] != tag);
							xu = 0.0;

							zj = z + j*nm; /*zj[i] is Qz(i,j)*/
							for (i=p;i <= q ;i++)
							{
								xu += rv6[i]*zj[i];
							} /*for (i=p;i <= q ;i++)*/ /*#640*/

							for (i=p;i <= q ;i++)
							{
								rv6[i] -= xu*zj[i];
							} /*for (i=p;i <= q ;i++)*/
						} /*for (jj=0;jj < group ;jj++)*/ /*#680*/

						norm = 0.0; /*#700*/

						for (i=p;i <= q ;i++)
						{
							norm += fabs(rv6[i]);
						} /*for (i=p;i <= q ;i++)*/ /*#720*/

						if (norm >= 1.0)
						{
							break;
						} /*if (norm >= 1.0)*/

/*      .......... forward substitution .......... */
						if (its == MAXIT)
						{
/*      .......... set error -- non-converged eigenvector .......... */
							*ierr = -(r+1); /*#830*/
							xu = 0.0;
						    break;
						} /*if (its == MAXIT)*/

						if (norm != 0.0)
						{
							xu = eps4/norm; /*#740*/

							for (i=p;i <= q ;i++)
							{
								rv6[i] = rv6[i]*xu;
							} /*#760*/
						} /*if (norm != 0.0) */
						else
						{
							rv6[s] = eps4;
							s++;
							if (s > q)
							{
								s = p;
							}
						} /*if (norm != 0.0) {}else{}*/
/*
      .......... elimination operations on next vector
                 iterate ..........
*/
						for (i=p+1;i <= q ;i++) /*#780*/
						{
							u = rv6[i];
/*
      .......... if rv1[i-1] .eq. e(i), a row interchange
                 was performed earlier in the
                 triangularization process ..........
*/
							if (rv1[i - 1] == e[i])
							{
								u = rv6[i - 1];
								rv6[i - 1] = rv6[i];
							}
							rv6[i] = u - rv4[i]*rv6[i - 1]; /*#800*/
						} /*for (i=p;i <= q ;i++)*/ /*#820*/
						if (*ierr)break;

						its++;
					} /*while (1)*/

					if (*ierr == 0)
					{
/*
      .......... normalize so that sum of squares is
                 1 and expand to full order ..........
*/

						u = 0.0; /*#840*/

						for (i=p;i <= q ;i++)
						{
							u += rv6[i]*rv6[i];
						} /*for (i=p;i <= q ;i++)*/ /*#860*/
						xu = 1.0/sqrt(u);
					} /*if (*ierr == 0)*/

				  setZir: /*#870*/
					zj = z + r*nm;
					for (i=0;i < n ;i++)
					{
						zj[i] = 0.0;
					} /*for (i=0;i < n ;i++)*/ /*#880*/

					for (i=p;i <= q ;i++)
					{
						zj[i] = rv6[i]*xu;
					} /*for (i=p;i <= q ;i++)*/ /*#900*/

					x0 = x1;
				} /*if (ind[r] == tag) */
			} /*for (r=0;r < m ;r++)*/ /*#920*/
		} while(q < n - 1);
	} /*if (m != 0)*/
} /*tinvit()*/
