/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Utils
#else
#define __SEG__ Utils
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

void           noData();
void           badType();
void           badDims();
void           badNargs();
void           undefArg();

extern char   *typeName();


void badType(who,type,argno)
char         *who;
long          type;
long          argno;
{
	if(type > 0)
	{
		if(argno > 0)
		{
			sprintf(OUTSTR,"ERROR: argument %ld to %s must have type %s",
					argno, who, typeName(type));
		}
		else
		{
			sprintf(OUTSTR,"ERROR: argument to %s must have type %s",
					who, typeName(type));
		}
		
	}
	else if(type == -UNDEF || type == -ASSIGNED)
	{
		if(argno > 0)
		{
			sprintf(OUTSTR,"ERROR: argument %ld to %s is undefined",
					argno, who);
		}
		else
		{
			sprintf(OUTSTR,"ERROR: argument to %s is undefined",
					who);
		}
	}
	else
	{
		if(argno > 0)
		{
			sprintf(OUTSTR,"ERROR: argument %ld to %s must not have type %s",
					argno, who, typeName(-type));
		}
		else
		{
			sprintf(OUTSTR,"ERROR: argument to %s must not have type %s",
					who, typeName(-type));
		}
	}
	putOUTSTR();
}

void badDims(who, ndims,argno)
char *who;
long ndims;
long argno;
{
	if(ndims > 0)
	{
		if(argno > 0)
		{
			sprintf(OUTSTR,
					"ERROR: argument %ld to %s does not have %ld dimensions",
					argno, who, ndims);
		}
		else
		{
			sprintf(OUTSTR,
					"ERROR: argument to %s does not have %ld dimensions",
					who, ndims);
		}
	}
	else if(ndims < 0)
	{
		if(argno > 0)
		{
			sprintf(OUTSTR,
					"ERROR: argument %ld to %s has more than %ld dimensions",
					argno, who, -ndims);
		}
		else
		{
			sprintf(OUTSTR,
					"ERROR: argument to %s has more than %ld dimensions",
					who, -ndims);
		}
	}
	else
	{
		if(argno > 0)
		{
			sprintf(OUTSTR,
					"ERROR: argument %ld to %s has wrong number of dimensions",
					argno,who);
		}
		else
		{
			sprintf(OUTSTR,
					"ERROR: argument to %s has wrong number of dimensions",
					who);
		}			
	}
		
	putOUTSTR();
}

void badNargs(who, nargs)
char    *who;
long     nargs;
{
	if(nargs == 0)
	{
		sprintf(OUTSTR,"ERROR: no argument(s) given for %s",who);
	}
	else if(nargs > 0)
	{
		sprintf(OUTSTR,"ERROR: %s requires %ld argument%s",who, nargs,
				(nargs > 1) ? "s" : "");
	}
	else if(nargs > -1000)
	{
		nargs = -nargs;
		sprintf(OUTSTR,"ERROR: %s may have at most %ld argument%s",who, nargs,
				(nargs > 1) ? "s" : "");
	}
	else if(nargs == -1000)
	{
		sprintf(OUTSTR,"ERROR: %s should have no arguments", FUNCNAME);
	}
	else
	{
		nargs = -nargs - 1000;
		sprintf(OUTSTR,"ERROR: %s must have at least %ld argument%s",who,
				nargs,
				(nargs > 1) ? "s" : "");
	}
		
	putOUTSTR();
}

void noData(who, argno)
char *who;
long argno;
{
	if(argno > 0)
	{
		sprintf(OUTSTR,"ERROR: argument %ld to %s is missing or null",argno,who);
	}
	else if(argno == 0)
	{
		sprintf(OUTSTR,"ERROR: argument to %s is missing or null",who);
	}
	else
	{
		undefArg(who, (Symbolhandle) 0, labs(argno));
	}
	putOUTSTR();
}

void undefArg(who,arg,argno)
char             *who;
Symbolhandle      arg;
long              argno;
{
	if(arg == (Symbolhandle) 0)
	{
		if(argno != 0)
		{
			sprintf(OUTSTR,"ERROR: argument %ld to %s is undefined",
					(long) abs((int) argno),who);
		}
		else
		{
			sprintf(OUTSTR,"ERROR: argument to %s is undefined", who);
		}
	}
	else 
	{
		if(argno != 0)
		{
			sprintf(OUTSTR,"ERROR: argument %ld (%s) to %s is undefined",
					(long) abs((int) argno), NAME(arg),who);
		}
		else
		{
			sprintf(OUTSTR,"ERROR: argument (%s) to %s is undefined",
					NAME(arg), who);
		}
	}
	putOUTSTR();
}

void badKeyword(who,keyword)
char            *who, *keyword;
{
	sprintf(OUTSTR,"ERROR: %s is not valid keyword for %s",keyword,who);
	putOUTSTR();
}

void notTorF(char *keyword)
{
	sprintf(OUTSTR,
			"ERROR: value%s%s must be T or F",
			(keyword) ? " for " : "", (keyword) ? keyword : "");
	putOUTSTR();
}

void notCharOrString(char *keyword)
{
	sprintf(OUTSTR,
			"ERROR: value%s%s must be quoted string or CHARACTER scalar",
			(keyword) ? " for " : "", (keyword) ? keyword : "");
	putOUTSTR();
}

void notNumberOrReal(char *keyword)
{
	sprintf(OUTSTR,
			"ERROR: value%s%s must be number or REAL scalar",
			(keyword) ? " for " : "", (keyword) ? keyword : "");
	putOUTSTR();
}

void notImplemented(who)
char           *who;
{
	sprintf(OUTSTR,"ERROR: '%s' is not implemented in this version",who);
	putOUTSTR();
}
