/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Pvalsub1
#else
#define __SEG__ Pvalsub1
#endif /*MPW3*/
#endif /*MPW*/

#include <math.h>
#ifndef HUGEDBL
#define HUGEDBL 1e38
#endif


#define TRUE 1
#define FALSE 0

extern double   lgamma(double);
extern double   ppnd(double,long *);
extern void     normbase(double */*x*/, double */*phi*/);
extern void      gammabase(double */*x*/, double */*a*/, double */*p*/);

#ifndef NOPROTOTYPES
static double gammp(double /*x*/, double /*a*/);
static double gnorm(double /*x*/, double /*a*/);
static double gser(double /*x*/, double /*a*/, double /*gln*/);
static double gcf(double /*x*/, double /*a*/, double /*gln*/);
static double ppchi2(double */*p*/, double */*v*/,double * /*g*/,long * /*ifault*/);
static double gammad(double * /*x*/,double * /*a*/,long * /*iflag*/);
#else /*NOPROTOTYPES*/
static double gammp(/*double x, double a*/);
static double gnorm(/*double x, double a*/);
static double gser(/*double x, double a, double gln*/);
static double gcf(/*double x, double a, double gln*/);
static double ppchi2(/*double *p, double *v,double *g,long *ifault*/);
static double gammad(/*double *x,double *a,long *iflag*/);
#endif /*NOPROTOTYPES*/

void gammabase(x, a, p)
double         *x, *a, *p;
{
	*p = gammp(*a, *x);
}

double          ppgamma(p, a, ifault)
double          p, a;
long           *ifault;
{
	double          x, v, g;

	v = 2.0 * a;
	g = lgamma(a);
	x = ppchi2(&p, &v, &g, ifault);
	return (x / 2.0);
}

/*
  Static Routines
*/

/*
  From Numerical Recipes, with normal approximation from Appl. Stat. 239
*/

#define EPSILON 1.0e-14
#define LARGE_A 10000.0
#define ITMAX 1000

static double   gammp(a, x)
double          a, x;
{
	double          gln, p;

	if (x <= 0.0 || a <= 0.0)
	{
		p = 0.0;
	}
	else if (a > LARGE_A)
	{
		p = gnorm(a, x);
	}
	else
	{
		gln = lgamma(a);
		if (x < (a + 1.0))
		{
			p = gser(a, x, gln);
		}
		else
		{
			p = 1.0 - gcf(a, x, gln);
		}
	}
	return (p);
}

/* compute gamma cdf by a normal approximation */
static double   gnorm(a, x)
double          a, x;
{
	double          p, sx;

	if (x <= 0.0 || a <= 0.0)
	{
		p = 0.0;
	}
	else
	{
		sx = sqrt(a) * 3.0 * (pow(x / a, 1.0 / 3.0) + 1.0 / (a * 9.0) - 1.0);
		normbase(&sx, &p);
	}
	return (p);
}

/* compute gamma cdf by its series representation */
static double   gser(a, x, gln)
double          a, x, gln;
{
	double          p, sum, del, ap;
	long            n, done = FALSE;

	if (x <= 0.0 || a <= 0.0)
	{
		p = 0.0;
	}
	else
	{
		ap = a;
		del = sum = 1.0 / a;
		for (n = 1; !done && n < ITMAX; n++)
		{
			ap += 1.0;
			del *= x / ap;
			sum += del;
			if (fabs(del) < EPSILON)
			{
				done = TRUE;
			}
		}
		p = sum * exp(-x + a * log(x) - gln);
	}
	return (p);
}

/* compute complementary gamma cdf by its continued fraction expansion */
static double   gcf(a, x, gln)
double          a, x, gln;
{
	double          gold = 0.0, g, fac = 1.0, b1 = 1.0;
	double          b0 = 0.0, anf, ana, an, a1, a0 = 1.0;
	double          p;
	long            done = FALSE;

	a1 = x;
	p = 0.0;
	for (an = 1.0; !done && an <= ITMAX; an += 1.0)
	{
		ana = an - a;
		a0 = (a1 + a0 * ana) * fac;
		b0 = (b1 + b0 * ana) * fac;
		anf = an * fac;
		a1 = x * a0 + anf * a1;
		b1 = x * b0 + anf * b1;
		if (a1 != 0.0)
		{
			fac = 1.0 / a1;
			g = b1 * fac;
			if (fabs((g - gold) / g) < EPSILON)
			{
				p = exp(-x + a * log(x) - gln) * g;
				done = TRUE;
			}
			gold = g;
		}
	}
	return (p);
}

static double   gammad(double *x,double *a,long *iflag)
{
	double          cdf;

	gammabase(x, a, &cdf);
	*iflag = 0;
	return (cdf);
}

/*
  ppchi2.f -- translated by f2c and modified

  Algorithm AS 91   Appl. Statist. (1975) Vol.24, P.35
  To evaluate the percentage points of the chi-squared
  probability distribution function.

  p must lie in the range 0.000002 to 0.999998,
    (but I am using it for 0 < p < 1 - seems to work)
  v must be positive,
  g must be supplied and should be equal to ln(gamma(v/2.0))

  Auxiliary routines required: ppnd = AS 111 (or AS 241) and gammad.
*/

static double   ppchi2(p, v, g, ifault)
double         *p, *v, *g;
long           *ifault;
{
	/* Initialized data */
	/* coefficients are not built in to code */
	/*
	  static double pmin = 2e-6;
	  static double pmax = .999998;
	*/
	double   pmin = 0.0;
	double   pmax = 1.0;

	/* System generated locals */
	double          ret_val, d_1, d_2;

	/* Local variables */
	double   a, b, c, q, t, x, p1, p2, s1, s2, s3, s4, s5, s6, ch;
	double   xx;
	long     if1;


	/* test arguments and initialise */
	ret_val = -1.0;
	*ifault = 1;
	if (*p <= pmin || *p >= pmax)
	{
		return ret_val;
	}
	*ifault = 2;
	if (*v <= 0.0)
	{
		return ret_val;
	}
	*ifault = 0;
	xx = 0.5 * *v;
	c = xx - 1.0;

	if (*v < -1.24 * log(*p))
	{
		/* starting approximation for small chi-squared */
		ch = pow(*p * xx * exp(*g + xx * .6931471806), 1.0 / xx);
		if (ch < 5e-7)
		{
			ret_val = ch;
			return ret_val;
		}
	}
	else if (*v > .32)
	{
		/* call to algorithm AS 111 - note that p has been tested above. */
		/* AS 241 could be used as an alternative. */
		x = ppnd(*p, &if1);

		/* starting approximation using Wilson and Hilferty estimate */
		p1 = .222222 / *v;
		/* Computing 3rd power */
		d_1 = x * sqrt(p1) + 1.0 - p1;
		ch = *v * (d_1 * d_1 * d_1);

		/* starting approximation for p tending to 1 */
		if (ch > 2.2 * *v + 6.0)
		{
			ch = -2.0 * (log(1.0 - *p) - c * log(0.5 * ch) + *g);
		}
	}
	else
	{
		/* starting approximation for v less than or equal to 0.32 */
		ch = .4;
		a = log(1.0 - *p);
		do
		{
			q = ch;
			p1 = 1.0 + ch * (4.67 + ch);
			p2 = ch * (6.73 + ch * (6.66 + ch));
			d_1 = -0.5 + (4.67 + 2.0 * ch) / p1;
			d_2 = (6.73 + ch * (13.32 + 3.0 * ch)) / p2;
			t = d_1 - d_2;
			ch -= (1.0 - exp(a + *g + 0.5 * ch + c * .6931471806) * p2 / p1) / t;
		} while (fabs(q / ch - 1.0) > .01);
	}

	do
	{
		/* call to gammad and calculation of seven term Taylor series */
		q = ch;
		p1 = 0.5 * ch;
		p2 = *p - gammad(&p1, &xx, &if1);
		if (if1 != 0)
		{
			*ifault = 3;
			return ret_val;
		}
		t = p2 * exp(xx * .6931471806 + *g + p1 - c * log(ch));
		b = t / ch;
		a = 0.5 * t - b * c;
		s1 = (210. + a * (140. + a * (105. + a * (84. + a * (70. + 60. * a))))) / 420.;
		s2 = (420. + a * (735. + a * (966. + a * (1141. + 1278. * a)))) / 2520.;
		s3 = (210. + a * (462. + a * (707. + 932. * a))) / 2520.;
		s4 = (252. + a * (672. + 1182. * a) + c * (294. + a * (889. + 1740. * a))) / 5040.;
		s5 = (84. + 2264. * a + c * (1175. + 606. * a)) / 2520.;
		s6 = (120. + c * (346. + 127. * c)) / 5040.;
		d_1 = (s3 - b * (s4 - b * (s5 - b * s6)));
		d_1 = (s1 - b * (s2 - b * d_1));
		ch += t * (1.0 + 0.5 * t * s1 - b * c * d_1);
	} while (fabs(q / ch - 1.0) > 5e-7);

	ret_val = ch;
	return ret_val;
}
