/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Getmodel
#else
#define __SEG__ Getmodel
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include "blas.h"

extern long         Copy();
extern long         symbolSize();
extern char       **mygethandle();
extern void         mydisphandle();
extern Symbolhandle Install(), Lookup();
extern Symbolhandle Makesymbol();
extern double       cellstat();
extern int          modelparse();
extern long         isVector();

#define GCTS     0
#define NTRASH   1

long getmodel(list,nargs,silent)
Symbolhandle    list;
long            nargs, silent;
{
	char          **chtmp = (char **) 0;
	/* procedure which gets model string and converts it
	   to proper format returns  0 if ok 1 if error */

	long            i, j, k, i2, term[MAXVARS], inc;
	long            ny, ndata, nvars, nfactors, nterms, nerrorterms;
	long            ncells, needed, balanced = 0, length;
	long            keyindex;
	long            intercept;
	long            nmissing, nzerowt;
	long            dim[2];
	modelPointer    kterm;
	Symbolhandle    symh, symTERMN = (Symbolhandle) 0;
	Symbolhandle    symDEPVN = (Symbolhandle) 0;
	double          tmp;
	double        **cts = (double **) 0, *pcts;
	double         *glmoffset, *casewts, *misswts;
	double          yi, logitni, *y, *logitn;
	long            caseweights = MODELTYPE & CASEWEIGHTS;
	modelInfoPointer info = (modelInfoPointer) 0;
	WHERE("getmodel");
	TRASH(NTRASH,errorExit);

	*OUTSTR = '\0';
	
	initmodelparse(STRMODEL);
	
	if (modelparse() != 0)
	{							/* parse it */
		goto errorExit;
	}

	info = getModelInfo(); /* retrieve model information */
	if((ndata = checkVariables(info)) < 0)
	{
		goto errorExit;
	}
	if(!(MODELTYPE & MANOVA) && !isVector(info->modelvars[0]))
	{
		sprintf(OUTSTR,
				"ERROR: non-vector response variable for %s", FUNCNAME);
		goto errorExit;
	}
	
	NVARS = nvars = MODELINFO->nvars = info->nvars;
	NFACTORS = nfactors = MODELINFO->nfactors = info->nfactors;
	NTERMS = nterms = MODELINFO->nterms = info->nterms;
	
	NERRORTERMS = nerrorterms = MODELINFO->nerrorterms = info->nerrorterms;
	for(i=0;i<nerrorterms;i++)
	{
		modeltermAssign((modelPointer) info->errorterms[i],
						(modelPointer) MODELINFO->errorterms[i]);
	}
/* Note: As of 930601, INTMODEL has been eliminated */
	MODEL = info->model;
	info->model = (modelHandle) 0;
	
	if(MODELTYPE & (OLSREG|LEAPS))
	{
		intercept = 0;
		for(j=0;j<nterms;j++)
		{
			kterm = modelterm(MODEL,j);
			if(nvInTerm(kterm) > 1)
			{
				sprintf(OUTSTR,
						"ERROR: crossed variates and/or factors are illegal for %s",
						FUNCNAME);
				goto errorExit;
			} /*if(nvInTerm(modelterm(MODEL,j)) > 1)*/
			if(modeltermEqual(kterm,(modelPointer) NULLTERM))
			{
				intercept = 1;
			}
		} /*for(j=0;j<nterms;j++)*/
	} /*if(MODELTYPE & (OLSREG|LEAPS))*/
	if(MODELTYPE & LEAPS && !intercept)
	{
		sprintf(OUTSTR,"ERROR: a model without an intercept is illegal for %s",
				FUNCNAME);
		goto errorExit;
	}
	
	/* set up MODELVARS */

	for (i = 0; i <= nvars; i++)
	{
		if(MODELVARS[i] != (Symbolhandle) 0)
		{
			Delete(MODELVARS[i]);
		}
		
		MODELVARS[i] = Makesymbol(REAL);
		if (MODELVARS[i] == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		
		if (Copy(info->modelvars[i], MODELVARS[i]) == 0)
		{
			goto errorExit;
		}
		info->modelvars[i] = (Symbolhandle) 0;
#ifdef UNDEFINED__
 /* MODELVARS is now a macro expanding to MODELINFO->modelvars */;
		MODELINFO->modelvars[i] = MODELVARS[i];
#endif /*UNDEFINED__*/
		strcpy(VARNAMES[i], NAME(MODELVARS[i]));
		if(i < nvars)
		{
			NCLASSES[i] = info->nclasses[i];
		}			
	} /*for (i = 0; i <= nvars; i++)*/

	/* set up Y and X */
	/* response and factors are known to be matrix or vector */
	for (i = 0; i <= nvars; i++)
	{
		symh = MODELVARS[i];
		(void) isMatrix(symh,dim); /* get dimensions */
		setNDIMS(symh,2);
		setDIM(symh,1,dim[0]);
		setDIM(symh,2,dim[1]);
		if(i == 0)
		{
			Y = DATA(symh);
			ny = dim[1];
			NDATA = (double) ndata;
			NY = (double) ny;
			/* make sure dimensions are correct, always 2 */
		} /*if(i == 0)*/
		else
		{/* i > 0 */
			X[i - 1] = DATA(symh);

			if (NCLASSES[i - 1] < 2 && (MODELTYPE & BALANOVA))
			{
				MODELTYPE &= ~BALANOVA;
				MODELTYPE |= UNBALANO;
			}
			if (NCLASSES[i - 1] < 2 && (MODELTYPE & IPF))
			{
				if(!silent)
				{
					myerrorout("WARNING: ipf cannot be used, poisson substituted");
				}
				MODELTYPE &= ~IPF;
				MODELTYPE |= POISSONREG;
			}
		}  /*if(i == 0){}else{}*/
		if(*OUTSTR)
		{
			goto errorExit;
		}
	} /*for (i = 0; i <= nvars; i++)*/

	/* check for crossed variates */

	for (i = 0; i < nvars-1; i++)
	{
		if (NCLASSES[i] <= 1)
		{
			/* variable i is a variate */
			for (j = 0; j < nterms; j++)
			{					/* 
				find the terms it is in and check to see it is not crossed
				with another variate
			  */
				kterm = modelterm(MODEL,j);
				
				if(inTerm(i,kterm) && nvInTerm(kterm) > 1)
				{
					/* 
					  j-th term has variable i and there are at least 2
					  variables or factors in the term.
					  Check to see if j-th term has another variate
					*/
					for (i2 = i+1; i2 < nvars; i2++)
					{
						if (NCLASSES[i2] <= 1 && inTerm(i2,kterm))
						{
							sprintf(OUTSTR,
									"ERROR: crossed variates in model are illegal for %s",
									FUNCNAME);
							goto errorExit;
						}
					} /*for (i2 = 0; i2 < nvars; i2++)*/
				} /*if(inTerm(i,kterm))*/
			} /*for (j = 0; j < nterms; j++)*/
		} /*if (NCLASSES[i] <= 1)*/
	} /*for (i = 0; i < nvars; i++)*/

	/* check for case weights, i.e., user specified weights */

	if (caseweights)
	{
		if (nargs < 2 || (symh = COMPVALUE(list, 1)) == (Symbolhandle) 0)
		{
			sprintf(OUTSTR,
					"ERROR: no weights given for %s",FUNCNAME);
		}
		else if(TYPE(symh) != REAL || !isVector(symh))
		{
			sprintf(OUTSTR,
					"ERROR: weights for %s not REAL vector", FUNCNAME);
		}
		else if (symbolSize(symh) != ndata)
		{
			sprintf(OUTSTR,"ERROR:  weights for %s not the right length",
					FUNCNAME);
		}
		else
		{
			CASEWTS = (double **) mygethandle(ndata * sizeof(double));
			if (CASEWTS == (double **) 0)
			{
				goto errorExit;
			}
			casewts = *CASEWTS;
			inc = 1;
			DCOPY(&ndata, DATAPTR(symh), &inc, casewts, &inc);
			for (i = 0; i < ndata; i++)
			{
				if (!isMissing(casewts[i]) && casewts[i] < 0)
				{
					sprintf(OUTSTR,"ERROR: negative case weights for %s",
							FUNCNAME);
					break;
				}
			} /*for (i = 0; i < ndata; i++)*/
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
	} /* if(caseweights) */

	/* check for logistic N */
	if (MODELTYPE & LOGITREG)
	{
		if (nargs < 2 || (symh = COMPVALUE(list, 1)) == (Symbolhandle) 0)
		{
			sprintf(OUTSTR,"ERROR: no N variable given for %s",FUNCNAME);
		}
		else if(!isVector(symh) || TYPE(symh) != REAL || 
				symbolSize(symh) != ndata && symbolSize(symh) != 1)
		{
			sprintf(OUTSTR,
					"ERROR: %s N not REAL scalar or vector of same length as response",
					FUNCNAME);
		}
		else
		{
			LOGITN = (double **) mygethandle(ndata * sizeof(double));
			if (LOGITN == (double **) 0)
			{
				goto errorExit;
			}
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
		inc = (symbolSize(symh) == 1) ? 0 : 1;
		i2 = 0;
		y = *Y;
		logitn = *LOGITN;
		for (i = 0; i < ndata; i++)
		{
			yi = y[i];
			logitni = logitn[i] = DATAVALUE(symh,i2);
			if (!isMissing(yi) && !isMissing(logitni))
			{
				tmp = floor(logitni);
				if (tmp <= 0.0 || logitni != tmp)
				{
					sprintf(OUTSTR,
							"ERROR:  N values must be nonnegative integers for %s",
							FUNCNAME);
				}
				else if (yi < 0.0 || yi > logitni || yi != floor(yi))
				{
					sprintf(OUTSTR,
							"ERROR: response value not integer or out of range for %s",
							FUNCNAME);
				}
				if(*OUTSTR)
				{
					goto errorExit;
				}
			}
			i2 += inc;
		}/* for (i = 0; i < ndata; i++) */
	} /* if (MODELTYPE & LOGITREG) */

	/* check for nonnegative Y in POISSONREG */
	if (MODELTYPE & POISSONREG)
	{
		y = *Y;
		for (i = 0; i < ndata; i++)
		{
			if (!isMissing(yi  = y[i]) && (yi < 0. || floor(yi) != yi))
			{
				sprintf(OUTSTR,
						"ERROR: response variable negative or not integer for %s",
						FUNCNAME);
				goto errorExit;
			}
		}
	} /* if (MODELTYPE & POISSONREG) */

	/* check for user defined glmoffset */
	USEGLMOFFSET = 0;

	if( MODELTYPE & (POISSONREG | IPF | LOGITREG) ) 
	{
		if((keyindex = findKeyword(list,"offset",1)) >= 0)
		{
			symh = COMPVALUE(list,keyindex);
			if(MODELTYPE & IPF)
			{
				sprintf(OUTSTR,
						"ERROR: 'offset' cannot be used with %s; use poisson instead",
						FUNCNAME);
				goto errorExit;
			}

			OUTSTR[0] = '\0';
			if(TYPE(symh) != REAL)
			{
				sprintf(OUTSTR,"ERROR: offsets for %s of type %s",
						FUNCNAME,typeName(TYPE(symh)));
			}
			else if(!isVector(symh) || symbolSize(symh) != ndata)
			{
				sprintf(OUTSTR,
						"ERROR: offsets for %s must be vector of same length as response vector",
						FUNCNAME);
			}
			else
			{
				USEGLMOFFSET = 1;
				GLMOFFSET = (double **) mygethandle(ndata * sizeof(double));
				if (GLMOFFSET == (double **) 0)
				{
					goto errorExit;
				}
				inc = 1;
				DCOPY(&ndata, DATAPTR(symh), &inc, *GLMOFFSET, &inc);
			}
			if(*OUTSTR)
			{
				goto errorExit;
			}
			setNAME(symh,USEDKEYWORD);
		} /*if((keyindex = findKeyword(list,"offset",1)) >= 0)*/
	} /* if( MODELTYPE & (POISSONREG | IPF | LOGITREG) ) */

	/* check for missing */
	nmissing = countMissing(MODELINFO,&MISSWTS);
	if(nmissing < 0)
	{
		goto errorExit;
	}
	
	if(caseweights || MODELTYPE & LOGITREG || USEGLMOFFSET)
	{
		casewts = (caseweights) ? *CASEWTS : (double *) 0;
		logitn = (MODELTYPE & LOGITREG) ? *LOGITN : (double *) 0;
		glmoffset = (USEGLMOFFSET) ? *GLMOFFSET : (double *) 0;
		misswts = (nmissing > 0) ? *MISSWTS : (double *) 0;
		for (i = 0; i < ndata; i++)
		{
			if (caseweights && isMissing(casewts[i]) ||
				(MODELTYPE & LOGITREG) && isMissing(logitn[i]) ||
				USEGLMOFFSET && isMissing(glmoffset[i]))
			{
				if(nmissing == 0)
				{
					MISSWTS = (double **) mygethandle(ndata * sizeof(double));
					if (MISSWTS == (double **) 0)
					{
						goto errorExit;
					}
					misswts = *MISSWTS;
					doubleFill(misswts, 1.0, ndata);
					casewts = (caseweights) ? *CASEWTS : (double *) 0;
					logitn = (MODELTYPE & LOGITREG) ? *LOGITN : (double *) 0;
					glmoffset = (USEGLMOFFSET) ? *GLMOFFSET : (double *) 0;
				}
				if (misswts[i] != 0.0)
				{
					misswts[i] = 0.0;
					nmissing++;
				}
			}
		} /*for (i = 0; i < ndata; i++)*/
	} /*if(caseweights || MODELTYPE & LOGITREG || USEGLMOFFSET)*/
	
	if(nmissing > 0)
	{
		MODELTYPE |= MISSWEIGHTS;
		if (MODELTYPE & BALANOVA)
		{
			MODELTYPE &= ~BALANOVA;
			MODELTYPE |= UNBALANO;
		}
		else if (MODELTYPE & IPF)
		{
			if(!silent)
			{
				myerrorout("WARNING: ipf cannot be used, poisson substituted");
			}
			MODELTYPE &= ~IPF;
			MODELTYPE |= POISSONREG;
		}
	} /*if(nmissing > 0)*/

	if (nmissing > 0 && !silent)
	{
		myerrorout("WARNING: cases with missing values deleted");
	}

	/* check for zero case weights, i.e., user specified weights */
	if (caseweights)
	{
		casewts = *CASEWTS;
		misswts = (MODELTYPE & MISSWEIGHTS) ? *MISSWTS : (double *) 0;
		nzerowt = 0;
		for (i = 0; i < ndata; i++)
		{
			if (casewts[i] == 0 &&
				(!(MODELTYPE & MISSWEIGHTS) || misswts[i] > 0.0))
			{
				nzerowt++;
			}
		}
		if (nzerowt != 0 && !silent)
		{
			myerrorout("WARNING: cases with zero weight completely removed");
		}
		nmissing += nzerowt;
	} /*if (caseweights)*/
	NOTMISSING = NDATA - nmissing;

	if(NOTMISSING == 0)
	{
		sprintf(OUTSTR,
				"ERROR: no non-missing cases with non-zero weight");
		goto errorExit;
	}
	
	/*
	  In the future, we need code to recognize fractional factorials,
	  confounded factorials, etc
	*/

	/* Check for possible Latin Square or higher order orthogonal square */
	if (MODELTYPE & BALANOVA && nfactors > 1 && 
		(nterms == nfactors + 1 || nterms == nfactors))
	{ 
		balanced = 1;
		for(i = 0;i < nterms; i++)
		{
			if(i < nfactors && ndata % NCLASSES[i] != 0)
			{
				balanced = 0;
				break;
			}
			
			if(nvInTerm(modelterm(MODEL, i)) > 1)
			{
				balanced = -1;
				break;
			}
		} /*for(i = 0;i < nterms; i++)*/

		if(i == nterms)
		{/* only main effects; check for (hyper) Latin Square */
			ncells = 0;
			for (i = 1; i < nvars; i++)
			{
				for ( j = 0; j < i; j++)
				{
					length = NCLASSES[i] * NCLASSES[j];
					ncells = (length > ncells) ? length : ncells;
				}
			}
			if(!getScratch(cts,GCTS,ncells,double))
			{
				sprintf(OUTSTR, "Try anova(,unbalanced:T)");
				goto errorExit;
			}
			
			/* check all two-way marginals for balance */
			pcts = *cts;
			for (i = 1; i < nvars; i++)
			{
				for ( j = 0; j < i; j++)
				{
					for(k = 0; k < nvars; k++)
					{
						term[k] = 0;
					}
					term[i] = term[j] = 1;
					cellstat(Y, term, (double **) 0, (double **) 0,cts,
							 CELLCOUNTS);
					ncells = NCLASSES[i]*NCLASSES[j];
					for(k = 1; k < ncells;k++)
					{
						if(pcts[k] != pcts[0])
						{
							balanced = 0;
							break;
						}
					} /*for(k = 0; k < ncells;k++)*/
					if(!balanced)
					{
						break;
					}
				} /*for ( j = 0; j < i; j++)*/
				if(!balanced)
				{
					break;
				}
			} /*for (i = 1; i < nvars; i++)*/
			if(!balanced)
			{
				MODELTYPE &= ~BALANOVA;
				MODELTYPE |= UNBALANO;
			}			
		} /*if(i == nterms)*/
	}
	
	if(balanced < 0)
	{
		balanced = 0;
	}
	
	if (!balanced && MODELTYPE & (BALANOVA | IPF))
	{			/* check for complete balance with all variables */
		ncells = 1;
		for(i=0;i<nvars;i++)
		{
			term[i] = 1;
			ncells *= NCLASSES[i];
		}

		if(ndata >= ncells && ndata % ncells == 0)
		{
			if(!getScratch(cts,GCTS,ncells,double))
			{
				if(MODELTYPE & BALANOVA)
				{
					sprintf(OUTSTR, "Try anova(,unbalanced:T)");
				}
				else
				{
					sprintf(OUTSTR, "Try poisson()");
				}
				goto errorExit;
			}

			/* compute cell counts */
			cellstat(Y, term, (double **) 0, (double **) 0, cts, CELLCOUNTS);
			pcts = *cts;
			for (i = 1; i < ncells; i++)	
			{
				if (pcts[i] != pcts[0])
				{
					break;
				}
			} /*for (i = 1; i < ncells; i++) */
			discardScratch(cts,GCTS,double);
			if(i == ncells)
			{
				balanced = 1;
			}
		} /*if(ndata >= ncells && ndata % ncells == 0)*/

		if(!balanced)
		{
			if (MODELTYPE & (BALANOVA | UNBALANO))
			{
				MODELTYPE &= ~BALANOVA;
				MODELTYPE |= UNBALANO | caseweights;
			}
			else
			{
				if(!silent)
				{
					myerrorout("WARNING: ipf cannot be used, poisson substituted");
				}
				MODELTYPE = POISSONREG;
			}
		} /*if(!balanced)*/
	} /*if (MODELTYPE & (BALANOVA + IPF))*/

	/* check for incremental deviances */
	INCREMENTAL = 0;
	if (MODELTYPE & (POISSONREG | IPF | ROBUSTREG) && nargs > 1 ||
		MODELTYPE & LOGITREG && nargs > 2)
	{
		if((keyindex = findKeyword(list,"inc",1)) >= 0)
		{
			symh = COMPVALUE(list,keyindex);
			OUTSTR[0] = '\0';
			if( !isScalar(symh) || TYPE(symh) != LOGIC)
			{
				sprintf(OUTSTR,
						"ERROR: value for keyword inc for %s must be T or F",
						FUNCNAME);
				goto errorExit;
			}
			INCREMENTAL = (long) DATAVALUE(symh,0);
			setNAME(symh,USEDKEYWORD);
		} /*if((keyindex = findKeyword(list,"inc",1)) >= 0)*/
		else if(!isKeyword(COMPVALUE(list,nargs-1)))
		{
			INCREMENTAL = 1;
		}
	}

	/* save dependent variable name */
	needed = strlen(NAME(MODELVARS[0])) + 1;
	DEPVNAME = mygethandle(needed);
	if(DEPVNAME == (char **) 0)
	{
		goto errorExit;
	}
	strcpy(*DEPVNAME, VARNAMES[0]);
	
	symDEPVN = Install("DEPVNAME", CHAR);
	if(symDEPVN == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	chtmp = myduphandle(DEPVNAME);
	setSTRING(symDEPVN,chtmp);
	if(chtmp == (char **) 0)
	{
		Removesymbol(symDEPVN);
		goto errorExit;
	}
	setNDIMS(symDEPVN,1);
	setDIM(symDEPVN,1,1);

	/* use needed for character count , first get total needed */

	needed = 0;
	for (i = 0; i < nterms; i++)
	{			/* for each term in model */
		kterm = modelterm(MODEL,i);
		if (modeltermEqual(kterm,(modelPointer) NULLTERM))	/* constant */
		{
			needed += strlen("CONSTANT") + 1;
		}
		else
		{			
			for (k = 0; k < nerrorterms - 1; k++)
			{					/* an error term ? */
				if(modeltermEqual(kterm, MODELINFO->errorterms + k))
				{
					needed += strlen("ERROR") + ((k >= 9) ? 3 : 2);
					break;
				}
			}

			if (k == nerrorterms - 1)
			{
				/* usual (not error) term */
				for (k = 0; k < nvars; k++)
				{
					if(inTerm(k,kterm))
					{
						needed += strlen(VARNAMES[k + 1]) + 
							((VARNAMES[k + 1][0] == TEMPPREFIX) ? 0 : 1);
					}
				}
			} /*if (k == nerrorterms - 1)*/
		}
	} /*for (i = 0; i < nterms; i++)*/
	needed += strlen("ERROR") + ((nerrorterms > 9) ? 3 : 2);
	/* now that we know how much space we need, go back and fill in */

	TERMNAMES = mygethandle((long) needed);
	if (TERMNAMES == (char **) 0)
	{
		goto errorExit;
	}

	j = 0;
	for (i = 0; i < nterms; i++)
	{
		kterm = modelterm(MODEL,i);
		if (modeltermEqual(kterm,(modelPointer) NULLTERM))
		{
			strcpy(*TERMNAMES + j, "CONSTANT");
			j += strlen("CONSTANT");
			(*TERMNAMES)[j++] = '\0';
		}
		else
		{
			for (k = 0; k < nerrorterms - 1; k++)
			{
				if (modeltermEqual(kterm, MODELINFO->errorterms + k))
				{
					strcpy(*TERMNAMES + j, "ERROR");
					j += strlen("ERROR");
					if(k >= 9)
					{
						(*TERMNAMES)[j++] = (k+1)/10 + '0';
						(*TERMNAMES)[j++] = (k+1)%10 + '0';
					}
					else
					{
						(*TERMNAMES)[j++] = k+1 + '0';
					}
					(*TERMNAMES)[j++] = '\0';
					break;
				} /*if (modeltermEqual(kterm, MODELINFO->errorterms + k))*/
			} /*for (k = 0; k < nerrorterms - 1; k++)*/

			if (k == nerrorterms - 1)
			{/* regular (not error) term */
				for (k = 0; k < nvars; k++)
				{
					if (inTerm(k,kterm))
					{
						if(istempname(VARNAMES[k+1]))
						{		/* strip off TEMPPREFIX ('@') if temporary name */
							strcpy(*TERMNAMES + j, VARNAMES[k + 1] + 1);
						}
						else
						{
							strcpy(*TERMNAMES + j, VARNAMES[k + 1]);
						}
						j += strlen(*TERMNAMES + j);
						if(moreInTerm(k,kterm,nvars))
						{
							(*TERMNAMES)[j++] = '.';
						}
					}
				} /*for (k = 0; k < nvars; k++)*/
				(*TERMNAMES)[j++] = '\0';
			} /*if (k == nerrorterms - 1)*/
		} /*if (modeltermEqual(kterm,(modelPointer) NULLTERM)){...}else{...}*/
	} /*for (i = 0; i < nterms; i++)*/

	strcpy(*TERMNAMES + j, "ERROR");
	j += strlen("ERROR");
	if(nerrorterms > 9)
	{
		(*TERMNAMES)[j++] = nerrorterms/10 + '0';
		(*TERMNAMES)[j++] = nerrorterms%10 + '0';
	}
	else
	{
		(*TERMNAMES)[j++] = nerrorterms + '0';
	}
	(*TERMNAMES)[j++] = '\0';
	
	TERMNAMES = mygrowhandle(TERMNAMES,j); /* make sure it's the right size */
	if(TERMNAMES == (char **) 0)
	{
		goto errorExit;
	}

	symTERMN = Install("TERMNAMES", CHAR);
	if (symTERMN == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	chtmp = myduphandle(TERMNAMES);
	setSTRING(symTERMN,chtmp);
	if(chtmp == (char **) 0)
	{
		Removesymbol(symTERMN);
		goto errorExit;
	}
	setNDIMS(symTERMN,1);
	setDIM(symTERMN,1,nterms + 1);

	emptyTrash();
	
	return (0); /* normal (no error) return */

/* NOTE: cleanup of globals is done in glm after error return */
errorExit:
	putOUTSTR();
	emptyTrash();
	if(info != (modelInfoPointer) 0)
	{
		mydisphandle((char **) info->model);
		MODEL = info->model = (modelHandle) 0;
	}
	
	return (1);

}
