/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifndef GLMH__
#define GLMH__
/*
  defines and global declarations related to glm computations
  930811 Modified so that large globals such as MODELINFO are allocated
	dynamically rather than taking up part of the global segment;  this was
	because the Mac version ran over 32K globals
  Appropriate changes where made in the glm modules.
*/

#include "typedefs.h"

#undef EXTERN
#undef INIT
#undef INITDIM
#undef INITARRAY
#undef COMMA

#ifdef MAIN__
#define EXTERN
#define INIT(VAL) = VAL
#define INITDIM(N) N
#define INITARRAY(VALS) = { VALS }
#else
#define EXTERN extern
#define INIT(VAL)
#define INITDIM(N)
#define INITARRAY(VALS)
#endif /*MAIN__*/

#define COMMA ,

/* Limits (formerly in globdefs.h) */
#ifndef WORDSPERTERM
#define WORDSPERTERM  3       /* number of words in a term */
#endif /*WORDSPERTERM*/

#ifndef VARSPERWORD
#define VARSPERWORD  32     /* number of variables per word */
#endif /*VARSPERWORD*/

#define OLDMAXVARS   31

#ifndef REGTOL
#define REGTOL     1.0e-12  /* used in singularity check in gramschmidt */
#endif /*REGTOL*/

#if     WORDSPERTERM == 1
#define MAXVARS    VARSPERWORD		/* max number of variables in model */
#else
#define MAXVARS    (WORDSPERTERM*VARSPERWORD)		/* max number of variables in model */
#endif  /*WORDSPERTERM == 1 */

#define MAXERRTRMS 10		/* max number of error terms */

/* stuff relating to representation of models */
typedef unsigned long termWord;
typedef termWord      modelType[WORDSPERTERM];
typedef modelType    *modelPointer, **modelHandle;

typedef struct       modelInfo
{
	long             nvars;
	long             nfactors;
	long             nterms;
	long             nerrorterms;
	modelHandle      model;
	modelType        errorterms[MAXERRTRMS+1];
	Symbolhandle     modelvars[MAXVARS+1];
	long             nclasses[MAXVARS];
} modelInfo, *modelInfoPointer, **modelInfoHandle;

#define modelterm(M,I) (((modelPointer) *(M)) + (I)) /* type modelPointer */
#define zeroTerm(T)  modeltermAssign((modelPointer) NULLTERM, (modelPointer) T)


/* constants related to glm's */

/* special cases, different algorithms */
#define NOMODEL      0x000L  /* 0 no current model */
#define LEAPS        0x001L  /* 1 leaps screening */
#define BALANOVA     0x002L  /* 2 balanced anova */
#define FASTANOVA    0x004L  /* 4 iterative fast anova via EM */

/* single pass gram schmidt models */
#define MANOVA       0x008L  /* 8 manova */
#define UNBALANO     0x010L  /* 16 unbalanced anova */
#define OLSREG       0x020L  /* 32 standard regression */

/* multipass iterative gram schmidt models */
#define ROBUSTREG    0x040L  /* 64 robust regression */
#define LOGITREG     0x080L  /* 128 logistic regression */
#define POISSONREG   0x100L  /* 256 poisson regression */
#define PROBITREG    0x200L  /* 512 probit regression */
#define ITERWEIGHTS  0x3c0L  /* ROBUSTREG+LOGITREG+POISSONREG+PROBITREG */
/* weight indicators */

#define CASEWEIGHTS  0x400L  /* 1024 user specified case weights */
#define MISSWEIGHTS  0x800L  /* 2048 missing values, so weights needed */
#define USEWEIGHTS   0xfc0L  /* ITERMWEIGHTS+CASEWEIGHTS+MISSWEIGHTS */

#define IPF          0x1000L  /* 4096 iterative proportional fitting */

/* op codes for cellstat() */
#define CELLSTATS         1
#define CELLSUMS          2
#define CELLCOUNTS        3
#define CELLMEANSANDSWEEP 4

/* opcodes for gramschmidt() and/or itergs() */
#define DONTNOTIFY   0x000000
#define DONTDOSWEEP  0x000000
#define DONTDOCOEFS  0x000000
#define DONTDORINV   0x000000
#define DONTDOSS     0x000000

#define DONOTIFY     0x010000  /* warn if singular column found */
#define DOSWEEP      0x020000  /* actually do orthogonalization */
#define DOCOEFS      0x040000  /* compute coefficients */
#define DORINV       0x080000  /* compute R^(-1) in QR decomposition */
#define DOSS         0x100000  /* compute SS or SSCP */

/* opcodes for iterfns.c functions */
#define INITIAL      1L
#define STEP         2L
#define FINAL        4L

/* global variables related to doing glm */

#if    WORDSPERTERM == 1
EXTERN modelType       NULLTERM INITARRAY(0);
#elif  WORDSPERTERM == 2
EXTERN modelType       NULLTERM INITARRAY(0 COMMA 0);
#else  /*WORDSPERTERM == 3*/
EXTERN modelType       NULLTERM INITARRAY(0 COMMA 0 COMMA 0);
#endif /*WORDSPERTERM == 3*/

/* MODELINFO is initialized in glmInit() */
EXTERN modelInfoPointer MODELINFO; /* previously global space was allocated */

#undef UNDEFINED__
#ifdef UNDEFINED__ /* INTMODEL no longer referenced */
EXTERN modelHandle     INTMODEL INIT((modelHandle) 0);	/*  like MODEL except integers */
#endif /*UNDEFINED__*/

/* VARNAMES is initialized in glmInit() */
/* matrix of variable names in model*/
EXTERN Char           (*VARNAMES)[NAMELENGTH+1];
#ifdef UNDEFINED__   /* previously global space was allocated */
EXTERN Char            VARNAMES[INITDIM(MAXVARS+1)][NAMELENGTH+1];
#endif /*UNDEFINED__*/

#undef UNDEFINED__
#ifdef UNDEFINED__ /* MODEL now a macro */
EXTERN modelHandle     MODEL INIT((modelHandle) 0);
#endif /*UNDEFINED__*/

#define MODEL          MODELINFO->model
	/* handle to modelTypes indicating which order to enter terms in model */

EXTERN long            INCREMENTAL INIT(0);/*  incremental deviance flag */
EXTERN long	           USEGLMOFFSET INIT(0); /* flag for using glm offsets */

#define MODELVARS      MODELINFO->modelvars

#undef UNDEFINED__
#ifdef UNDEFINED__ 
EXTERN Symbolhandle    MODELVARS[INITDIM(MAXVARS+1)];
#endif /*UNDEFINED__*/

	/* handles to variables used in model, 0 is y */
EXTERN Double          NVARS INIT(-1);	/* number of variables in model */
EXTERN Double          NFACTORS INIT(-1); /* number of factors in model */
EXTERN Double          NDATA INIT(0);	/* number of data values */
EXTERN Double          NOTMISSING INIT(0);	/* number of nonmissing data */
EXTERN Double          NTERMS INIT(0);	/* number of terms in model */

#define NCLASSES       MODELINFO->nclasses
#undef UNDEFINED__
#ifdef UNDEFINED__ 
EXTERN Long            NCLASSES[INITDIM(MAXVARS)];
#endif /*UNDEFINED__*/
	/* number of classes in each model variable */

EXTERN Double          NREGX; /* number of columns in regression matrix */
EXTERN Double        **RESIDUALS INIT(0);	/* handle to residuals */
EXTERN Double        **WTDRESIDUALS INIT(0);	/* handle to residuals */
EXTERN Double        **HII INIT(0);	/* handle to leverages */
/* Note: Y should always be identical to DATA(MODELVARS[0]) */
EXTERN Double        **Y INIT(0);		/* handle to y data */
EXTERN Double          NY INIT(0);		/* number of Y's (for manova) */
/* Note: X[i] should always be identical to DATA(MODELVARS[i+1]) */
EXTERN Double       ***X; /*initialized in glmInit*/
#ifdef UNDEFINED__
EXTERN Double        **X[INITDIM(MAXVARS)];	/* array of handles to x columns */
#endif /* UNDEFINED__ */
EXTERN Double        **SS INIT(0);	/* handle for model term sums of squares */
EXTERN Double        **DF INIT(0);		/* handle for model term dfs */
EXTERN Double        **MISSWTS INIT(0);	/* handle for missing value weights */
EXTERN Double        **CASEWTS INIT(0);
	/* handle for user weights in linear models */
EXTERN Double        **ITERWTS INIT(0);	/* handle for iterative weights */
EXTERN Double        **ALLWTS INIT(0);	/* combined weights */
EXTERN Double        **LOGITN INIT(0);	/* counts for logistic N */
EXTERN Double	     **GLMOFFSET INIT(0); 	/* offsets in a glm model */
EXTERN Long            MODELTYPE INIT(NOMODEL);	/* indicator for model type */
EXTERN Long            PREVMODELTYPE INIT(NOMODEL); /* previous model type */
/*
				0  means no current model (NOMODEL)
				1  means leaps screening (LEAPS)
				2  means balanced anova (BALANOVA)
				4  means fast anova (FASTANOVA) 
				8  means manova (MANOVA)
			   16  means unbalanced anova (UNBALANO)
               32  means standard regression (OLSREG)
			   64  means robust regression (ROBUSTREG)
			  128  means logistic regression (LOGITREG)
			  256  means Poisson regression (POISSONREG)
              512  means Probit regression (PROBITREG)
			 1024  means case weights (CASEWEIGHTS)
			 2048  means missing value weights (MISSWEIGHTS)
*/

EXTERN Char          **STRMODEL INIT(0);	/* string form of inputted model */

/* ERRORTERMS is initialized in glmInit() */
EXTERN modelPointer    ERRORTERMS INIT((modelPointer) 0);
#ifdef UNDEFINED__
EXTERN modelPointer    ERRORTERMS INIT((modelPointer) MODELINFO.errorterms);
#endif /*UNDEFINED__*/

		/* terms used as errors */
EXTERN Double          NERRORTERMS;	/* number of error terms */
EXTERN Char          **DEPVNAME INIT(0);    /* name of dependent variable */
EXTERN Char          **TERMNAMES INIT(0);	/* names of terms */
EXTERN Double        **REGX INIT(0);	/* x matrix for regression */
EXTERN Double        **REGX2 INIT(0);	/* x matrix for regression */
EXTERN Double        **XTXINV INIT(0);	/* xtxinv matrix for regression */
EXTERN Double        **REGCOEF INIT(0);	/* regression coefficients */
EXTERN Double        **REGSS INIT(0);	/* sequential ss for regression */

#endif /*GLMH_*/
