/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Glm
#else
#define __SEG__ Glm
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

extern modelInfoPointer MreaderModelInfoPointer; /* defined in mreader.y */

/*
  Below are public functions related to glm

  long glmInit(void)
      Allocate glm-related globals (called by initiali.c)
  void getterm(long n, long term[], nvars)
	  Break out bits of jterm into vector term[]
  long setterm(long term[], nvars, nclasses)
	  Compute number of cells in marginal specified term and nclasses
	  and set term[i] < 0 for any variate (nclasses[i] < 0) in term
	  (term[i] != 0).
  void effectiveTerm(long term[], long *effTerm, long nvars)
	  Set effTerm to reflect variates or the factors with level > 1 in term[]
	  If nvars == 0 it sets effTerm to 0
  long inEffectiveTerm(long term,long effTerm) (may be implemented as macro)
	  Check whether previous term prevterm is in current effective term
	  effTerm
  long inTerm(long jvar, modelPointer term) (may be implemented as macro)
	  Check whether model variable jvar is in term; jvar counts from 0
  long nvInTerm(modelPointer term)
	  Count the number of variables and factors in term.
  long moreInTerm(long jvar, modelPointer term) (may be implemented as macro)
	  Check whether there is a variable kvar > jvar in term, where
	  jvar is a variable in term;  jvar counts from 0
  long modeltermEqual(modelPointer term1,modelPointer term2)
	  Check two model terms for equality; return 1 if equal, 0 otherwise
  void modeltermAssign(modelPointer from, modelPointer to)
	  Copy contents of *from to *to
  char **setStrmodel(char **modelString, long clean)
	  Add '1+' to start of r.h.s. of modelString if not there, returning
	  handle for possibly modified model.  If clean != 0 also strip
	  '1[ \t]\+' from modelString if there, thereby possibly modifying
	  an argument.
  long getAlpha(long termi, long colstart, double ***alphaH, long ***colptrH)
	  Extract alphas associated with term termi from REGCOEF, returning results
	  in handle *alphaH to which is allocated ncells*ny doubles, where ncells
	  is the product of the dimensions of the factors in term termi.
	  (**colptr)[i]  is -1 if the  i-th column in the term is intrinsically
	  aliassed, -1 if it has been swept out, and the column number, otherwise.
	  It seems it will only make sense to call getAlpha() after fitting a glm;
	  hence it uses the global information rather than obtaining it through
	  its argument list.
  void xtxinvInit(void)
	  Initialize upper triangle of (*XTXINV) to identity
  long Xsetup(void)
	  Routine to initialize DF, NREGX, and allocate storage for REGX, XTXINV,
	  REGCOEF, & REGSS, and compute the matrix of X-variables.
  long buildXmatrix(termi, regx, info)
	  If regx != 0 set up columns of x matrix corresponding to the termi-th
	  term in the model specified by info; otherwise just count cols needed.
	  Return number of columns needed or generated.  This is the maximum number
	  of degrees of freedom associated with the term (actual d.f. may be less)
	  depending on confounding
  long checkVariables(modelInfoPointer info)
	  Check to see that all variables are REAL and that their lengths match.
	  also check that dependent variable is a vector or matrix and that each
	  r.h.s variable is a vector.  If variates can ever be matrices, this
	  will need to be modified. If all o.k. return the number of cases (ndata);
	  otherwise return -1
  long countMissing(modelInfoPointer info, double ***misswts)
	  Return the number of cases that have missing values in either the
	  dependent variable or any of the model variables.
	  If misswts != (double ***) 0 and there are missing values, allocate a
	  handle for the missing weights and set (**misswts)[i] to 0 or 1
	  depending on whether case i contains missing values.  If there are
	  no missing values no handle is allocated.
*/

/* allocate glm related globals */
long glmInit(void)
{
	char        *tempPointer;
	long         length;
	long         i;

	length = (MAXVARS+1) * (NAMELENGTH+1) * sizeof(char);
	tempPointer = mygetpointer(length);
	if(tempPointer == (char *) 0)
	{
		return (0);
	}
	VARNAMES = (char (*)[NAMELENGTH+1]) tempPointer;
	for(i=0;i<length;i++)
	{
		*tempPointer++ = '\0';
	}
	length = MAXVARS * sizeof(double **);
	tempPointer = mygetpointer(length);
	if(tempPointer == (char *) 0)
	{
		return (0);
	}
	X = (double ***) tempPointer;
	for(i=0;i<length;i++)
	{
		*tempPointer++ = '\0';
	}

	length = sizeof(modelInfo);
	tempPointer = mygetpointer(length);
	if(tempPointer == (char *) 0)
	{
		return (0);
	}
	MODELINFO = (modelInfoPointer) tempPointer;
	for(i=0;i<length;i++)
	{
		*tempPointer++ = '\0';
	}
	ERRORTERMS = MODELINFO->errorterms;

	tempPointer = mygetpointer(length);
	if(tempPointer == (char *) 0)
	{
		return (0);
	}
	MreaderModelInfoPointer = (modelInfoPointer) tempPointer;
	for(i=0;i<length;i++)
	{
		*tempPointer++ = '\0';
	}

	return (1);
}

/*
  Break out bits of jterm into vector term[]
*/

void getterm(jterm, term, nvars)
modelPointer    jterm;
long            term[], nvars;
{
	termWord        thisword = (*jterm)[0];
	long            i = 0, j;
	long            limit = VARSPERWORD;
	WHERE("getterm");

	for(j=0; j < nvars; j++)
	{
		if(j == limit)
		{
			thisword = (*jterm)[++i];
			limit += VARSPERWORD;
		}
		term[j] = thisword & 1L;
		thisword >>= 1;
	}
}

/*
  Compute number of cells in marginal specified term and nclasses
  and set term[i] < 0 for any variate (nclasses[i] < 0) in term (term[i] != 0).
*/

long setterm(long term[], long nvars, long nclasses[])
{
	long         i;
	long         ncells = 1;

	for(i=0;i<nvars;i++)
	{
		if(term[i] != 0)
		{
			if(nclasses[i] > 0)
			{
				ncells *= nclasses[i];
			}
			else
			{
				term[i] = -1;
			}
		}
	}
	return (ncells);
}

/*
  Set effTerm to reflect variates or the factors with level > 1 in term[]
  If nvars == 0 it sets effTerm to 0
*/
void effectiveTerm(long term[], modelPointer effTerm, long nvars)
{
	register termWord    thisword = 0;
	register long        i = 0, j, bit = 1;
	long                 limit = VARSPERWORD;

	zeroTerm(effTerm);

	for (j = 0; j < nvars; j++)
	{
		if(j == limit)
		{
			(*effTerm) [i++] = thisword;
			limit += VARSPERWORD;
			thisword = 0;
			bit = 1;
		}
		if (term[j] < 0 || term[j] > 1)
		{	/* live var or non-1 factor */
			thisword |= bit;
		}
		bit <<= 1;
	}
	(*effTerm)[i] = thisword;
}


/*
  Check whether previous term prevterm is in current effective term effTerm
*/
long inEffectiveTerm(modelPointer prevterm,modelPointer effterm)
{
	register int     i;
	termWord         prev, eff;

	if(modeltermEqual(prevterm, effterm))
	{
		return (1);
	}
	
	for(i=0;i<WORDSPERTERM;i++)
	{
		prev = (*prevterm)[i];
		eff = (*effterm)[i];
		if(prev|eff && ((prev & eff) ^ eff))
		{
			return (0);
		}
	}
	return (1);
}

/*
  Check whether model variable jvar is in term
  jvar counts from 0
*/
long inTerm(long jvar, modelPointer term)
{
	long       i = jvar/VARSPERWORD;
	
	jvar -= i*VARSPERWORD;
	return ((((*term)[i] >> jvar) & 1L) != 0);
}

/*
   Count the number of variables and factors in term
*/
long nvInTerm(modelPointer term)
{
	long       i, nv = 0;
	termWord   piece;
	
	for(i=0;i<WORDSPERTERM;i++)
	{
		for(piece = (*term)[i];piece != 0;piece >>= 1)
		{
			nv += (piece & 1L);
		}
	}
	return (nv);
}

/*
  Check whether there is a later variable than jvar in term, where jvar is in
  term;  jvar counts from 0
*/
long moreInTerm(long jvar, modelPointer term, long nvars)
{
	long           i, imax;
		
	if(nvars <= VARSPERWORD)
	{
		return ((((*term)[0] >> jvar) ^ 1) != 0);
	}

	i = jvar/VARSPERWORD;
	jvar %= VARSPERWORD;
	if((((*term)[i] >> jvar) ^ 1) != 0)
	{
		return (1);
	}
	
	imax = (nvars-1)/VARSPERWORD;
	for(++i;i <= imax; i++)
	{
		if((*term)[i] != (termWord) 0)
		{
			return (1);
		}
	}
	return (0);
}

/*
  Check two model terms for equality; return 1 if equal, 0 otherwise
*/
long modeltermEqual(modelPointer term1,modelPointer term2)
{
	register long   i;

	for(i=0;i<WORDSPERTERM;i++)
	{
		if((*term1)[i] != (*term2)[i])
		{
			return (0);
		}
	}
	return (1);
}

/*
  Copy contents of *from to *to
*/
void        modeltermAssign(modelPointer from, modelPointer to)
{
	register long    i;

	for(i=0;i<WORDSPERTERM;i++)
	{
		(*to)[i] = (*from)[i];
	}
}


/*
  Add '1+' to start of r.h.s. of modelString if not there, returning
  pointer to possibly modified model.  If clean != 0 also strip
  '1[ \t]\+' from modelString if there, thereby possibly modifying
  an argument.
*/

char **setStrmodel(char **modelString, long clean)
{
	char          **strmodel;
	long            i, i1, j;
	char            c;
	
	strmodel = mygethandle(strlen(*modelString) + 3);

	if (strmodel != (char **) 0)
	{
		j = 0;

		for (i=0;(c = (*modelString)[i]) != '\0' && c != '=';i++)
		{ /* copy to just before '=' */
			(*strmodel)[j++] = c;
		}
		(*strmodel)[j++] = c; /* '=' or '\0' */
	
		if(c != '\0')
		{
			for(i++;(c = (*modelString)[i]) == ' ' || c == '\t';i++)
			{/* copy whitespace after '=' */
				(*strmodel)[j++] = c;
			}
			
			if(c != '1')
			{ /* insert '1+' */
				(*strmodel)[j++] = '1';
				(*strmodel)[j++] = '+';
			}
			/* copy remainder of model */
			strcpy(*strmodel + j, *modelString + i);

			if(c == '1' && clean)
			{/* if '1[ \t]*+' in string, delete it */
				for(i1 = i+1;
					(c = (*modelString)[i1]) == ' ' || c == '\t';i1++)
				{/* skip whitespace after '1' */
					;
				}
				if(c == '+')
				{/* squeeze out '1[ \t]*\+' in modelString */
					strcpy(*modelString+i,*modelString+i1+1);
				}
			} /*if(c == '1' && clean)*/
		} /*if(c != '\0')*/

		/* shrink strmodel to actual size */
		strmodel = mygrowhandle(strmodel,strlen(*strmodel)+1);
	}
	
	return (strmodel);
}

/*
  Extract alphas associated with term termi from REGCOEF, returning results
  in alpha;   alpha is allocated with length ncells*ny, where ncells
  is the product of the dimensions of the factors in term termi.
  *colptr is allocated space for ncells+1 longs.  colptr[0] is set to ncells
  colptr[i+1]  is -1 if the  i-th column in the term is intrinsically
  aliassed, -1 if it has been swept out, and the column number, otherwise.
  It seems it will only make sense to call getAlpha() after fitting a glm;
  hence it uses the global information rather than obtaining it through
  its argument list.
  If alphH is NULL, the alpha's are not computed, but colptrH is.
*/
#define GTMPALPHA    0
#define GTMPCOLPTR   1
#define NTRASH       2

long getAlpha(long termi, long colstart, double ***alphaH, long ***colptrH)
{
	long            i, i2, offset, j, k, term[MAXVARS], term2[MAXVARS];
	modelType       trialterm;
	modelPointer    prevterm, thisterm;
	long            itmp1, itmp2;
	long            nvars = (long) NVARS, ny = (long) NY, nregx = (long) NREGX;
	long            combo, ncells, imax, colno = colstart;
	long            do_coefs = (alphaH != (double ***) 0);
	double         *regss = *REGSS, *regcoef = *REGCOEF;
	double        **tmpalpha = (double **) 0, *alpha;
	long          **tmpcolptr = (long **) 0, *colptr;
	TRASH(NTRASH,errorExit);
	
	thisterm = modelterm(MODEL,termi);
	getterm(thisterm,term, nvars);
	imax = ncells = setterm(term,nvars, NCLASSES);
	if((do_coefs && !getScratch(tmpalpha,GTMPALPHA,imax*ny,double)) ||
	   !getScratch(tmpcolptr,GTMPCOLPTR,imax+1,long))
	{
		goto errorExit;
	}
	
	if(do_coefs)
	{
		*alphaH = tmpalpha;
		alpha = *tmpalpha;
	}
	
	*colptrH = tmpcolptr;
	colptr = *tmpcolptr;
	*colptr++ = ncells;

	for(combo = 0;combo < ncells;combo++)
	{
		/* check to see if each combination is already in through
		   lower order terms */
		offset = 0;
		k = 1;
		for (j = 0; j < nvars; j++)
		{
			if (term[j] > 0)
			{
				offset += (term[j] - 1) * k;
				k *= NCLASSES[j];
			}
		}
		/*
		  offset is position in table of coefficients, traversed in
		  transposed order
		*/
		effectiveTerm(term, (modelPointer) trialterm, nvars);

		/* trialterm is now effective term for this column */
		for (k = 0; k < termi; k++)
		{
			prevterm = modelterm(MODEL,k);
			if(inEffectiveTerm(prevterm,(modelPointer) trialterm))
			{
				/* make sure any variates in earlier term are in eff term */
				getterm(prevterm, term2, nvars);
				for (j = 0; j < nvars; j++)
				{
					if (term[j] >= 0 && term2[j] > 0 && NCLASSES[j] < 0)
					{
						break;
					}
					/* break out if variate in earlier term not here */
					/* will keep this effective term */
				} /*for (j = 0; j < nvars; j++)*/

				if (j == nvars)
				{/* effective term is in earlier term */
					if(do_coefs)
					{
						itmp1 = offset; /* offset + i2 * imax */
						for (i2 = 0; i2 < ny; i2++)
						{
							alpha[itmp1] =  0.0;
							itmp1 += imax;
						}		/*for (i2 = 0; i2 < ny; i2++)*/
					} /*if(do_coefs)*/
					
					/* flag for intrinsic aliasing */
					colptr[offset] = -1;
					break;
				} /*if (j == nvars)*/
			} /*if(inEffectiveTerm(prevterm,(modelPointer) trialterm))*/
		} /*for (k = 0; k < termi; k++)*/

		if(k == termi)
		{/* if to here, we have a term that is not obviously pivoted out */
			if (regss[colno] < -0.5)
			{/* here column was pivoted out */
				if(do_coefs)
				{
					itmp1 = offset; /* offset + i2*imax */
					for (i2 = 0; i2 < ny; i2++)
					{
						alpha[itmp1] = 0.0;
						itmp1 += imax;
					}
				}
				
				/* flag for pivoted out */
				colptr[offset] = -2;
			} /*if (regss[colno] < -0.5)*/
			else
			{/* here we have a coefficient */
				if(do_coefs)
				{
					itmp1 = offset;	/* offset + i2*imax */
					itmp2 = colno; /* colno + i2*nregx */
					for (i2 = 0; i2 < ny; i2++)
					{
						alpha[itmp1] = regcoef[itmp2];
						itmp1 += imax;
						itmp2 += nregx;
					}
				}				
				colptr[offset] = colno;
			}
			colno++;
		}/*if(k == termi)*/

		for(i=nvars-1;i>=0;i--)
		{
			if(term[i] > 0)
			{ /* it's a factor, so step it */
				term[i]++;
				if(term[i] <= NCLASSES[i])
				{
					break;
				}
				term[i] = 1;
			}
		} /*for(i=nvars-1;i>=0;i--)*/
	} /*for(combo = 0;combo < ncells;combo++)*/
	if(do_coefs)
	{
		unTrash(GTMPALPHA);
	}
	unTrash(GTMPCOLPTR);
	emptyTrash();
	
	return (colno - colstart);

  errorExit:
	emptyTrash();
	if(do_coefs)
	{
		*alphaH = (double **) 0;
	}
	*colptrH = (long **) 0;
	return (-1);
}

/*
  Routine to initialize DF, NREGX, and allocate storage for REGX, XTXINV,
  REGCOEF, & REGSS, and compute the matrix of X-variables
  Called only by glm().
*/

long             Xsetup(void)
{
	long            i;
	long            nterms = (long) NTERMS, ndata = (long) NDATA;
	long            nregx, ny = (long) NY;
	WHERE("Xsetup");
	
	/* first compute maximum possible df's */
	/* sum will be NREGX */

	nregx = 0;
	for (i = 0; i < nterms; i++)
	{			/* for each term */
		(*DF)[i] = (double) buildXmatrix(i, (double *) 0, MODELINFO);
		nregx += (long) (*DF)[i];
	}
	NREGX = (double) nregx;
	
	/* now get space */
	REGX = (double **) mygethandle((nregx * ndata) * sizeof(double));
	REGSS = (double **) mygethandle(((nregx + 1) * ny * ny) * sizeof(double));
	if (REGX == (double **) 0 || REGSS == (double **) 0)
	{
		return (1);
	}

	if(MODELTYPE & DOCOEFS)
	{
		REGCOEF = (double **) mygethandle((nregx * ny) * sizeof(double));
		XTXINV = (double **) mygethandle((nregx * nregx) * sizeof(double));
		if(REGCOEF == (double **) 0 || XTXINV == (double **) 0)
		{
			return (1);
		}
	}
	
	/* now setup x */

	nregx = 0;
	for (i = 0; i < nterms; i++)
	{			/* for each term */
		nregx += buildXmatrix(i, *REGX + nregx*ndata, MODELINFO);
	}

	return (0);

} /*Xsetup()*/

#ifdef UNDEFINED__ /* see new code below */
/* Initialize upper triangle of (*XTXINV) to identity */
void xtxinvInit(void)
{
	long         i, j, ij;
	long         nregx = (long) NREGX;
	double      *xtxinv = *XTXINV;

	for (i = 0; i < nregx; i++)
	{
		ij = i*(nregx + 1); /* i + j*nregx */
		xtxinv[ij] = 1.0; /* xtx(i, i) = 1.0 */
		for (j = i + 1; j < nregx; j++)
		{
			ij += nregx;
			xtxinv[ij] = 0.0;
		}
	}
}
#else /*UNDEFINED__*/
/* Initialize all of XTXINV using memset */
void xtxinvInit(void)
{
	long         i;
	long         nregx = (long) NREGX;
	double      *xtxinv = *XTXINV;

	(void) memset((void *) xtxinv, 0, (size_t) nregx*nregx*sizeof(double));
	for (i = 0; i < nregx; i++)
	{
		*xtxinv = 1.0;
		xtxinv += nregx + 1;
	}
}
#endif /*UNDEFINED__*/

/*
  If regx != 0 set up columns of x matrix corresponding to the termi-th term
  in the model specified by info; otherwise just count cols needed.
  Return number of columns needed or generated.  This is the maximum number
  of degrees of freedom associated with the term (actual d.f. may be less)
*/

long buildXmatrix(termi, regx, info)
long             termi;
double          *regx;
modelInfoPointer info;
{
	Symbolhandle    symh;
	long            i, j, k, level, term[MAXVARS],term2[MAXVARS];
	modelHandle     model = info->model;
	modelPointer    prevterm, thisterm;
	modelType       trialterm;
	long            ndata;
	long            nvars = info->nvars;
	double         *xvar, **x[MAXVARS];
	long            combo, ncells, ncols = 0;
	long           *nclasses;
	long            dim[2];
	WHERE("buildXmatrix");

	(void) isMatrix(info->modelvars[0],dim);
	ndata = dim[0];

	for(j=0;j<nvars;j++)
	{
		symh = info->modelvars[j+1];
		x[j] = DATA(symh);
	}
	nclasses = info->nclasses;
	
	thisterm = modelterm(model,termi);

	getterm(thisterm,term,nvars);

	ncells = setterm(term, nvars, nclasses);

	for(combo = 0;combo < ncells;combo++)
	{
		/* check to see if each combination is already in through
		   lower order terms */
		effectiveTerm(term, (modelPointer) trialterm, nvars);
		/*  trialterm is now effective term for this potential column */
		for (k = 0; k < termi; k++)
		{
			prevterm = modelterm(model,k);
			if (inEffectiveTerm(prevterm,(modelPointer) trialterm))
			{	/* if effective term is in earlier term */
				/* make sure any variates in earlier term are in eff term */
				getterm(prevterm, term2, nvars);
				for (j = 0; j < nvars; j++)
				{
					if (term[j] >= 0 && term2[j] > 0 && nclasses[j] < 0)
					{
						break;
					}
					/* break out if variate in earlier term not here */
					/* will keep this effective term */
				} /*for (j = 0; j < nvars; j++)*/
				if (j == nvars)
				{ /* no such variate; discard */
					break;
				}
			} /* if (inEffectiveTerm(prevterm,trialterm)) */
		} /* for (k = 0; k < termi; k++) */
		if (k == termi)
		{		/* if here, effective term needs to be entered */
			if(regx != (double *) 0)
			{/* set columns */
				for (i = 0; i < ndata; i++)
				{
					regx[i] = 1.0;
				}

				if(!modeltermEqual(thisterm,(modelPointer) NULLTERM))
				{
					for (j = 0; j < nvars; j++)
					{
						if(term[j] != 0)
						{/* variable in the term */
							xvar = *x[j];
							if (term[j] < 0)
							{	/* live variate */
								for (i = 0; i < ndata; i++)
								{
									regx[i] *= xvar[i];
								}
							}
							else if (term[j] > 1)
							{	/* live factor */
								for (i = 0; i < ndata; i++)
								{
									level = (long) xvar[i];	/* category */
									if (level == nclasses[j])
									{ /* last group */
										regx[i] *= -1.;
									}
									else if (level != term[j] - 1)
									{
										regx[i] = 0.;
									}
								} /*for (i = 0; i < ndata; i++)*/
							}
						}
					}/*for (j = 0; j < nvars; j++)*/
				} /*if(!modeltermEqual(thisterm,(modelPointer) NULLTERM))*/
				regx += ndata;
			} /* if(regx != (double *) 0)*/
			ncols++;
		} /*if (k == termi)*/
		
		/* step odometer */
		for(j=nvars-1;j>=0;j--)
		{
			if(term[j] > 0)
			{ /* it's a factor, so step it */
				term[j]++;
				if(term[j] <= nclasses[j])
				{
					break;
				}
				term[j] = 1;
			}
		} /*for(j=nvars-1;j>=0;j--)*/
	}/* for(combo = 0;combo < ncells;combo++) */

	return (ncols);
}

/*
  Check to see that all variables are REAL and that their lengths match.
  also check that dependent variable is a vector or matrix and that each
  r.h.s variable is a vector.  If variates can ever be matrices, this
  will need to be modified. If all o.k. return the number of cases (ndata);
  otherwise return -1
*/

long checkVariables(modelInfoPointer info)
{
	Symbolhandle    symh;
	long            i, ndata = -1;
	long            dim[2];
	char           *name;

	*OUTSTR = '\0';
	for(i=0;i<=info->nvars;i++)
	{
		symh = info->modelvars[i];
		name = NAME(symh);
		if(TYPE(symh) != REAL)
		{
			sprintf(OUTSTR,"ERROR: non REAL variable %s in model", name);
		}
		else if(!isMatrix(symh,dim))
		{
			sprintf(OUTSTR,
					"ERROR: %s %s is not a vector%s",
					(i == 0) ? "dependent variable" : "variate or factor",
					name,(i == 0) ? " or matrix" : "");
		}
		else if(i > 0 && dim[1] > 1)
		{
			sprintf(OUTSTR,"ERROR: non-vector %s on r.h.s of model", name);
		}
		else if(i > 0 && dim[0] != ndata)
		{
			sprintf(OUTSTR,
					"ERROR: variable %s on r.h.s of model has length different from response",
					name);
		}
		else if(i == 0)
		{
			ndata = dim[0];
		}

		if(*OUTSTR)
		{
			goto errorExit;
		}
	}							/*for(i=0;i<=nvars;i++)*/
	return (ndata);

  errorExit:
	putOUTSTR();
		
	return (-1);
}
	
/*
  Return the number of cases that have missing values in either the
  dependent variable or any of the model variables.
  If misswts != (double ***) 0 and there are missing values, allocate a
  handle for the missing weights and set (**misswts)[i] to 0 or 1
  depending on whether case i contains missing values.  If there are
  no missing values no handle is allocated.
*/

long countMissing(modelInfoPointer info, double ***misswts)
{
	long            ndata, ny, nmissing = 0, dim[2];
	double        **y, **x[MAXVARS];
	long            i, j1, j2, k, ij;
	long            nvars = info->nvars;
	WHERE("countMissing");
	
	(void) isMatrix(info->modelvars[0],dim);
	ndata = dim[0];
	ny = dim[1];

	if(misswts != (double ***) 0)
	{
		*misswts = (double **) 0;
	}

	y = DATA(info->modelvars[0]);
	for(i = 0; i < nvars;i++)
	{
		x[i] = DATA(info->modelvars[i+1]);
	}
	
	for (i = 0; i < ndata; i++)
	{
		ij = i;					/* i + j1*ndata */
		for (j1 = 0; j1 < ny; j1++)
		{
			if (isMissing((*y)[ij]))
			{
				break;
			}
			ij += ndata;
		} /* for (j1 = 0; j1 < ny; j1++)*/
	
		if(j1 == ny)
		{
			for (j2 = 0; j2 < nvars; j2++)
			{
				if (isMissing((*x[j2])[i]))
				{
					break;
				}
			}					/* for j */
		} /*if(j1 == ny)*/

		if(j1 < ny || j2 < nvars)
		{						/* found missing */
			if(misswts != (double ***) 0)
			{
				if(nmissing == 0)
				{
					*misswts = (double **) mygethandle(ndata * sizeof(double));
					if (*misswts == (double **) 0)
					{
						goto errorExit;
					}
					for(k=0;k<ndata;k++)
					{
						(**misswts)[k] = 1.0;
					}
				}
				(**misswts)[i] = 0.0;
			}			
			nmissing++;
		}
	} /* for (i = 0; i < ndata; i++) */
	if(nmissing == ndata)
	{
		sprintf(OUTSTR,
				"ERROR: no cases without missing data");
		goto errorExit;
	}
	
	return (nmissing);

  errorExit:
	mydisphandle((char **) *misswts);
	*misswts = (double **) 0;
	
	return (-1);
}


