/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifndef GLOBDEFSH__
#define GLOBDEFSH__

/* miscellaneous globally defined macros */
#ifndef NAMEFORWHERE
#define NAMEFORWHERE where__
#endif /*NAMEFORWHERE*/

#ifndef WHERE
/* NB: macro WHERE should be followed by ';' among declarations */
#define WHERE(name) char *NAMEFORWHERE = name
#endif /*WHERE*/


#ifndef HUGEDBL
#define HUGEDBL 1.79769313486e+308
#endif /*HUGEDBL*/

#ifndef HUGEINT
#define HUGEINT     2146727027   /*2^31 - 756621*/
#endif /*HUGEINT*/

/* various globally defined limits */
/* NOTE: As of 930518, MAXVARS & MAXERRTRMS are defined in glm.h */
#define MAXIDEPTH       100 /* maximum number of nested if statements*/
#define MAXWDEPTH        10 /* maximum number of nested while or for loops*/
#define MAXBDEPTH        10 /* maximum number of nested batch() commands*/
#define MAXLEVEL         99 /* maximum permissible value for '$$' in macros*/
#define MAXSTRING      1999 /* maximum number of characters in quoted string*/
#define MAXLINE         500 /* maximum length of input line */
#define MAXWHILE       1001 /* maximum number of repetitions of while loop */
#define MAXDIMS          31 /* maximum of dimensions in symbol */
#define MAXCOORD 2000000000	/* max size in one coord direction */
#define NAMELENGTH       12 /* maximum length of symbol names */
#define BUFFERLENGTH    200 /* length of certain string buffers */
/* following is largest permissible value to bit operations */
#define MAXBITVALUE    4294967295.0 /*2^32-1 = ffffffff */

#ifndef PATHSIZE       /* length of buffers to hold complete path */
#ifdef MACINTOSH
#define PATHSIZE  64
#else /*MACINTOSH*/
#define PATHSIZE 256
#endif /*MACINTOSH*/
#endif /*PATHSIZE*/

/* Variously globally defined constants */
/* for use in getKeyword() and perhaps elsewhere */
#define OPMASK         0x00000fffUL
#define TYPE1MASK      0x000ff000UL
#define TYPE2MASK      0x0ff00000UL
#define CONTROLMASK    0xf0000000UL

#define REALTYPE1      0x00001000UL
#define LOGICTYPE1     0x00002000UL
#define CHARTYPE1      0x00004000UL
#define STRUCTYPE1     0x00008000UL
#define PLOTINFOTYPE1  0x00010000UL
#define LISTTYPE1      0x00020000UL
#define NULLTYPE1      0x00040000UL
#define UNKNOWNTYPE1   0x00080000UL
#define REALTYPE2      0x00100000UL
#define LOGICTYPE2     0x00200000UL
#define CHARTYPE2      0x00400000UL
#define STRUCTYPE2     0x00800000UL
#define PLOTINFOTYPE2  0x01000000UL
#define LISTTYPE2      0x02000000UL
#define NULLTYPE2      0x04000000UL

#define UNKNOWNTYPE2   0x08000000UL

#define SCALAR1        0x00001000UL /* total length 1 */
#define VECTOR1        0x00002000UL /* dim[1] == length */
#define MATRIX1        0x00004000UL /* no more than 2 dims > 1 */
#define ARRAY1         0x00008000UL /* no restrictions on dims */
#define SCALAR2        0x00100000UL /* total length 1 */
#define VECTOR2        0x00200000UL /* dim[1] == length */
#define MATRIX2        0x00400000UL /* no more than 2 dims > 1 */
#define ARRAY2         0x00800000UL /* no restrictions on dims */

/* miscellaneous control codes for recursion and elsewhere */
#define LEFTRECUR      0x10000000UL /*1st argument or left operand is STRUC*/
#define RIGHTRECUR     0x20000000UL /*2nd argument or right operand is STRUC*/
#define REUSELEFT      0x40000000UL /*reuse 1st argument or left operand */
#define REUSERIGHT     0x80000000UL /*reuse 2nd argument or right operand */

/* miscellaneous status codes */
#define FOUNDMISSING    0x00100000UL
#define FOUNDNONPOS     0x00200000UL
#define FOUNDNEG        0x00400000UL
#define FOUNDDOMERR     0x00800000UL
#define FOUNDOVERFLOW   0x01000000UL
#define FOUNDZERODIV    0x02000000UL

/* Various op codes and return values */
/* op codes for dimcmp() */
#define STRICT          0x00000001UL
#define NEARSTRICT      0x00000002UL

/*return values from dimcmp() and treecmp()*/
#define HASMISSING    2
#define BADDIMS       1
#define BADSTRUC     -1

/*
  Values for ASCII, defining save operations; <=0 are binary, >0 are ascii
  SAVEBINARY1 & SAVEASCII1 refer to original binary format, with MAXVAR==31
  SAVEBINARY2 & SAVEASCII2 refer to first modified format, with MAXVAR == 31
  SAVEBINARY3 & SAVEASCII3 refer to modified format allowing MAXVAR == 96
  SAVEBINARY4 & SAVEASCII4 non-printable characters escaped on asciisave

  Types 2 and 3 differ only in the structure of linear model related globals
  MacAnova originally allowed only 31 variables in models, thus allowing each
  term to be described in a single 32 bit long, although in fact the
  information was kept in doubles.  When the number of variables allowed was
  increased, a term became expressible by an array of unsigned longs.

  The original ascii save format had long lines and imbedded '\0's terminating
  strings and macros.  Later formats have short lines and have character counts
  for each string and macro.  The only non-standard embedded characters would
  be in strings defined by escape sequences such as "\1".  Also, only
  NDIMS(symh) dimensions are put in file.

  Type 4 is the same as type 3 for save(), byt Type 4 escapes '\', '\t',
  and all non-printable characters (except '\n').

*/

/* It is important that SAVEBINARYi == -SAVEASCIIi */
#define SAVEBINARY1      -1   /* MAXVAR == 31 binary save */
#define SAVEBINARY2      -2   /* same */
#define SAVEBINARY3      -3   /* MAXVAR == 96 binary save */
#define SAVEBINARY4      -4   /* same */
#define SAVEASCII1        1   /* original ascii save, MAXVAR == 31 */
#define SAVEASCII2        2   /* revised ascii save, MAXVAR == 31 */
#define SAVEASCII3        3   /* revised ascii save, MAXVAR == 96 */
#define SAVEASCII4        4   /* same, except escaped nonprintables */

#define BINARYHEADER1   "////MacAnovaSaveFile////"
#define BINARYHEADER2   "////MacAnovaSaveFile////"
#define BINARYHEADER3   "////MacAnovaSaveFile3///"
#define BINARYHEADER4   "////MacAnovaSaveFile3///"
#define ASCIIHEADER1    "////MacAnovaASCIISave///"
#define ASCIIHEADER2    "////MacAnovaASCIISave2//"
#define ASCIIHEADER3    "////MacAnovaASCIISave3//"
#define ASCIIHEADER4    "////MacAnovaASCIISave4//"

/* Type defines from mainpars.h as of 930621 */
/* 
  They are here in case different version of yacc generates different codes
*/
# define CHAR_V31     257
# define REAL_V31     258
# define BLTIN_V31    259
# define LIST_V31     260
# define STRUC_V31    261
# define MACRO_V31    262
# define UNDEF_V31    263
# define LOGIC_V31    266
# define MACRO1_V31   267
# define PLOTINFO_V31 278
# define GARB_V31     280
# define LONG_V31     281

#endif /*GLOBDEFSH__*/
