/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifndef GLOBKBH__
#define GLOBKBH__

#include "typedefs.h"

#undef UNDEFINED__

#undef EXTERN
#undef INIT
#undef INITDIM
#undef INITARRAY
#undef COMMA

#ifdef MAIN__
#define EXTERN
#define INIT(VAL) = VAL
#define INITDIM(N) N
#define INITARRAY(VALS) = { VALS }
#else
#define EXTERN extern
#define INIT(VAL)
#define INITDIM(N)
#define INITARRAY(VALS)
#endif /*MAIN__*/

#define COMMA ,

#ifndef MISSINGVALUE
#define MISSINGVALUE -99999.9999
#endif

#ifndef FILE
#include <stdio.h>
#endif /*FILE*/

#ifdef USECBREAK
#undef TILDE           /* ioctl.h redefines TILDE */
#include <sys/ioctl.h>
#undef TILDE
#define TILDE  '~'
extern void                    ioctl(int, int, char *);
EXTERN struct sgttyb           TerminalInfo;
#define saveTerminalInfo()     ioctl(0, TIOCGETP, (char *) &TerminalInfo)
#define restoreTerminalInfo()  ioctl(0, TIOCSETP, (char *) &TerminalInfo)
#define cbreak()               ioctl(0, CBREAK, (char *) 0)
#define nocbreak()             restoreTerminalInfo()
#endif /*USECBREAK*/

#ifdef ACTIVEUPDATE
EXTERN short ActiveUpdate INIT(ACTIVEUPDATE);
#else /*ACTIVEUPDATE*/
EXTERN short ActiveUpdate INIT(0);
#endif /*ACTIVEUPDATE*/

/* internal representation of missing value */
EXTERN Double MISSING INIT(MISSINGVALUE);
/* name for MISSING in output */
#define LENGTHMISSING   7
EXTERN Char   NAMEFORMISSING[LENGTHMISSING+1] INIT("MISSING");

/* macros related to MISSING */

#define isMissing(X)  ((X) == MISSING)
#define setMissing(X) (X = MISSING)

/* global variables related to input processing */
EXTERN uChar **INPUTSTRING INIT(0);
EXTERN Long            ISTRCHAR INIT(0);

/* 
  MDEPTH is the current depth to which macros are nested
  MLEVEL is the number of macros using '$$' since last prompt, used
  as value for $$ in a macro
*/
EXTERN Long            MDEPTH INIT(0);
EXTERN Long            MLEVEL INIT(-1);

EXTERN Long            BRACKETLEV INIT(0); 
EXTERN Long            PARENLEV INIT(0); 

EXTERN Long            WDEPTH INIT(0);
EXTERN Long            WHILESTARTS[INITDIM(MAXWDEPTH)];
EXTERN Long            WHILECONDITIONS[INITDIM(MAXWDEPTH)];
EXTERN Long            WHILELIMITS[INITDIM(MAXWDEPTH)];
EXTERN Long            WHILEBRACKETLEV[INITDIM(MAXWDEPTH)];
EXTERN Double        **FORVECTORS[INITDIM(MAXWDEPTH)];

EXTERN Long            IDEPTH INIT(0);
EXTERN Long            IFCONDITIONS[INITDIM(MAXIDEPTH)];
EXTERN Long            IFBRACKETLEV[INITDIM(MAXIDEPTH)];

EXTERN Char            FUNCNAME[INITDIM(NAMELENGTH+1)];
EXTERN Symbolhandle    NULLSYMBOL;

#define TEMPPREFIX     '@'        /* 1st character of temporary names */

#define KEYPREFIX1     TEMPPREFIX /* 1st character of keyword names */
#define KEYPREFIX2     TEMPPREFIX /* 2nd character of keyword names */

#define SCRATCHPREFIX1 TEMPPREFIX /* 1st character of scratch names */
#define SCRATCHPREFIX2 '$'        /* 2nd character of scratch names */

/* standard name for used up keyword, must start with KEYPREFIXs */
EXTERN Char             USEDKEYWORD[] INIT("@@@USED");

/* standard names for scratch variables.  Must start with SCRATCHPREFIXs */
#define LENGTHSCRATCH  9
EXTERN Char             SCRATCH[LENGTHSCRATCH+1] INIT("@$SCRATCH");

#define LENGTHLOGSCRATCH  9
EXTERN Char             LOGICSCRATCH[LENGTHLOGSCRATCH+1] INIT("@$LOGICAL");

#define LENGTHNUMSCRATCH  8
EXTERN Char             NUMSCRATCH[LENGTHNUMSCRATCH+1] INIT("@$NUMBER");

#define LENGTHEXPRSCRATCH  12
EXTERN Char             EXPRSCRATCH[LENGTHEXPRSCRATCH+1] INIT("@$EXPRESSION");

#define LENGTHSTRINGSCRATCH  8
EXTERN Char             STRINGSCRATCH[LENGTHSTRINGSCRATCH+1] INIT("@$STRING");

#define LENGTHSCALSCRATCH  8
EXTERN Char             SCALSCRATCH[LENGTHSCALSCRATCH+1] INIT("@$SCALAR");

#define LENGTHVECSCRATCH  8
EXTERN Char             VECSCRATCH[LENGTHVECSCRATCH+1] INIT("@$VECTOR");

#define LENGTHMATSCRATCH  8
EXTERN Char             MATSCRATCH[LENGTHMATSCRATCH+1] INIT("@$MATRIX");

#define LENGTHARRAYSCRATCH  7
EXTERN Char             ARRAYSCRATCH[LENGTHARRAYSCRATCH+1] INIT("@$ARRAY");

#define LENGTHSTRUCSCRATCH  11
EXTERN Char             STRUCSCRATCH[LENGTHSTRUCSCRATCH+1] INIT("@$STRUCTURE");

#define LENGTHCOMPSCRATCH  11
EXTERN Char             COMPSCRATCH[LENGTHCOMPSCRATCH+1] INIT("@$COMPONENT");

#define LENGTHGARBSCRATCH  9
EXTERN Char             GARBSCRATCH[LENGTHGARBSCRATCH+1] INIT("@$GARBAGE");

#define LENGTHMACROSCRATCH  7
EXTERN Char             MACROSCRATCH[LENGTHMACROSCRATCH+1] INIT("@$MACRO");

#define LENGTHLISTSCRATCH  6
EXTERN Char             LISTSCRATCH[LENGTHLISTSCRATCH+1] INIT("@$LIST");

#define LENGTHGRAPHSCRATCH  7
EXTERN Char             GRAPHSCRATCH[LENGTHGRAPHSCRATCH+1] INIT("@$GRAPH");

#define SPECIALPREFIX1 SCRATCHPREFIX1 /* 1st character of special save names */
#define SPECIALPREFIX2 SCRATCHPREFIX2 /* 2nd character of special save names */
#define SPECIALPREFIX3 '$'            /* 3rd character of special save names */

/* names for special objects saved by save(). Must start with SPECIALPREFIXs*/
#define LENGTHOPTIONS      12
EXTERN Char             OPTIONS[LENGTHOPTIONS+1] INIT("@$$$$OPTIONS");

#define LENGTHGLOBALS      12
EXTERN Char             GLOBALS[LENGTHGLOBALS+1] INIT("@$$$$GLOBALS");

#define LENGTHSAVETIME     12
EXTERN Char             SAVETIME[LENGTHSAVETIME+1] INIT("@$$$$SAVETIM");

#define LENGTHMACHINE      12
EXTERN Char             MACHINE[LENGTHMACHINE+1] INIT("@$$$$MACHINE");

#define isnamestart(C) isalpha(C) /* check for legal start to name */
#define isnamechar(C)  (isalnum(C) || (C) == '_')
						/* check for legal characters in name */
#define iskeyname(A) ((A)[0] == KEYPREFIX1 && (A)[1] == KEYPREFIX2)
#define istempname(A) ((A)[0] == TEMPPREFIX)
#define isscratch(A)  ((A)[0] == SCRATCHPREFIX1 && (A)[1] == SCRATCHPREFIX2)

#ifndef STDOUT
#define STDOUT (FILE *) 0
#endif /*STDOUT*/

#ifndef STDIN
#ifdef MACINTOSH
#define STDIN (FILE *) 0
#else /*MACINTOSH*/
#define STDIN stdin
#endif /*MACINTOSH*/
#endif /*STDIN*/

#ifndef TILDE
#define TILDE   '~'
#endif /*TILDE*/

#if defined(MACINTOSH)
EXTERN char * DIRSEPARATOR INIT(":");
#elif defined(MSDOS)
EXTERN char * DIRSEPARATOR INIT("\\");
#else
EXTERN char * DIRSEPARATOR INIT("/");
#endif /*MACINTOSH*/

#ifndef UNLOADSEG
#ifdef MACINTOSH
#include <SegLoad.h>
#define UNLOADSEG(F)    UnloadSeg((Ptr) (F))
#else /*MACINTOSH*/
#define UNLOADSEG
#endif /*MACINTOSH*/
#endif /*UNLOADSEG*/


/* define initialization file name */
#ifndef MACANOVAINI
#if defined(MSDOS)
#define MACANOVAINI "MACANOVA.INI"
#elif defined(MACINTOSH)
#define MACANOVAINI "MacAnova.ini"
#else
#define MACANOVAINI ".macanova.ini"
#endif
#endif /*MACANOVAINI*/

#ifndef HELPNAME
#if defined(MSDOS)
#define HELPNAME "MACANOVA.HLP"
#elif defined(MACINTOSH)
#define HELPNAME "MacAnova.hlp"
#else
#define HELPNAME "macanova.hlp"
#endif
#endif /*HELPNAME*/

/* global SPOOL file */
EXTERN FILE           *SPOOLFILE INIT(0);
EXTERN Char          **SPOOLFILENAME INIT(0);

/* global HELP file (kb) */
EXTERN FILE           *HELPFILE  INIT(0);
EXTERN Char          **HELPFILENAME INIT(0);

/* global plot file (kb) */
EXTERN FILE           *PLOTFILE  INIT(0);

/* global restore file (kb) */
EXTERN FILE           *RESTOREFILE  INIT(0);

/* global save file info (kb) */
EXTERN int             KEEPFILE  INIT(-1); /* file descriptor for save file */
EXTERN Char          **KEEPFILENAME  INIT(0); /* handle */
EXTERN Long            ASCII INIT(0); /* flag to indicate save or asciisave */
/*
  ASCII == -3 or -4 means modified binary save format with 96 variables
  ASCII == -1 or -2 means original binary save format with 31 variables
  ASCII == 1 means old asciisave format, with long lines and imbedded '\0's
  ASCII == 2 means new asciisave format, with short lines, no imbedded '\0's
  ASCII == 3 same as 2 except no superfluous dimension information, and
             allowing 96 variables in models
  ASCII == 4 same as 3 except non-printable characters in strings escaped w '\'
*/

/* stack of input files, updated by batch() (kb) */
EXTERN Long            BDEPTH INIT(0);
EXTERN Long            BATCHECHO[INITDIM(MAXBDEPTH)] INITARRAY(1);
EXTERN FILE           *INPUTFILE[INITDIM(MAXBDEPTH)]  INITARRAY(stdin);
EXTERN Char          **INPUTFILENAMES[INITDIM(MAXBDEPTH)];

/* Descriptor of version, e.g., "MacAnova 3.32 of 11/01/93" */
EXTERN Char          **VERSION_ID INIT(0);

/* resettable by option batchecho */
EXTERN Long            DEFAULTBATCHECHO INIT(1);

/* resettable by option nowarning */
EXTERN Long            PRINTWARNINGS INIT(1);

#ifdef CONSOLENAME
EXTERN Char           *CONSOLE INIT(CONSOLENAME);
#endif /*CONSOLENAME*/

/* global variables related to error handling */
/*
  MAXERRORS is settable by setoptions() and is a global error limit.
  MAXERRORS1 is the operative limit.  At the stdin level is should be the
  same as MAXERRORS.  For batch files, MAXERRORS1 will be 1, if MAXERRORS == 0,
  or MAXERRORS, otherwise.

  if MAXERRORS1 != 0 NERRORS gets incremented by myprint() for every line
  starting with "ERROR:"  If MAXERRORS == 0, it is reset to zero before
  the prompt when BDEPTH == 0
*/
EXTERN Long            NERRORS INIT(0); /* cumulative number of errors */
EXTERN Long            MAXERRORS INIT(0);
	/* settable error limit, 0 means none */
EXTERN Long            MAXERRORS1 INIT(0);/* operative error limit */
EXTERN Long            WARNINGLEVEL INIT(0);/* 1 or 2 suppress WARNINGs or ERRORs*/
EXTERN Long            FatalError INIT(0); /* != 0 if Fatal Error occurred */

/* global buffer for building output strings */
EXTERN Char            OUTSTR[INITDIM(BUFFERLENGTH)];

/* global handle for line input buffer (was in matdat.h) */
EXTERN Char          **LINE INIT((Char **) 0);

EXTERN Symbolhandle    PLOT INIT(0);    /* for global info on last plot */

EXTERN Char          **TMPHANDLE INIT(0);	/* generic temporary handle */

/* global variables related to algorithms */
EXTERN Double          ANGLES2RADIANS INIT(1.0);
	/* either 1 (radians), 2*pi (cycles), or 2*pi/360 (degrees)*/

/* variables related to random number generation */
EXTERN Long         RANDS1 INIT(0) ; /* random number generator seed */
EXTERN Long         RANDS2 INIT(0) ; /* random number generator seed */

/* origin for subscripts (for possible future use) */
EXTERN Long         ORIGIN INIT(1);

/* hidden option for debugging */
EXTERN Long         GUBED  INIT(0);

/* global variables related to output formatting */
EXTERN Long         NLINES INIT(0);
	/* number of output lines since start of command*/
#define DEFAULTPRINTDEC  5
#define DEFAULTWRITEDEC  9
EXTERN Long         SCREENHEIGHT INIT(0); /* number of lines on screen */
EXTERN Long    		SCREENWIDTH INIT(80);   /* columns on screen */
EXTERN Long         FIELDWIDTH INIT(DEFAULTPRINTDEC+7+1);  /* width of output field */
EXTERN Char         DATAFMT[20] INIT(" %12.5g"); /* format for double output */
EXTERN Char         MISSINGFMT[10] INIT(" %12s"); /* for MISSING */
EXTERN Long         BEFOREDEC INIT(DEFAULTPRINTDEC+7);   /* e.g. 12 in 12.5g */
EXTERN Long         AFTERDEC INIT(DEFAULTPRINTDEC);
	/* # after dp in format, e.g., 5 in 12.5g */
EXTERN Char         FMTTYPE INIT('g');
	/* output format type, e.g., 'g' in 12.5g */
EXTERN Long         PRINTFORMAT[]
	INITARRAY(DEFAULTPRINTDEC+7 COMMA DEFAULTPRINTDEC COMMA (Long) 'g');
EXTERN Long         WRITEFORMAT[]
	INITARRAY(DEFAULTWRITEDEC+7 COMMA DEFAULTWRITEDEC COMMA (Long) 'g');

/* Global promps */
EXTERN Char         PROMPT[21] INIT("Cmd> ");/* resetable by setoptions() */
EXTERN Char         MOREPROMPT[7] INIT("More> ");

/* globals related to environment */
EXTERN Long         ISATTY INIT(0); /* is input from keyboard */
EXTERN Char        *TERMINAL INIT(0) ; /* on Unix set to $TERM */

/* macro for emergency error messages */
#ifdef MACINTOSH
#define PUTALERT(MSGS) (myAlert(MSGS), myerrorout(MSGS))
#else /*MACINTOSH*/
#define PUTALERT(MSGS) (fprintf(stderr,"%s\n",(MSGS)), myerrorout(MSGS))
#endif /*MACINTOSH*/

#undef	EXTERN
#undef	INIT
#undef	INITDIM
#undef	INITARRAY
#undef  COMMA

#ifdef MAIN__
/* copyright notice */
static char     copyright1[] =
	"MacAnova, (C) Copyright 1994 Gary W. Oehlert and Christopher Bingham";
#ifdef MACINTOSH
#ifndef MPW
/* copyright notice */
static char     copyright2[] = "(C) Copyright 1986 THINK Technologies, Inc.  \n\
Certain portions of the Macintosh version of this software are copyrighted\n\
by THINK Technologies, Inc.";
#endif /*MPW*/
#endif /*MACINTOSH*/
#endif /* MAIN__ */

#endif /*GLOBKBH__*/
