/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Graphics
#else
#define __SEG__ Graphics
#endif /*MPW3*/
#endif /*MPW*/

/*
 *
 *    G N U P L O T  --  graphics.c
 *
 *  Copyright (C) 1986, 1987  Thomas Williams, Colin Kelley
 *
 *  You may use this code as you wish if credit is given and this message
 *  is retained.
 *
 *  Please e-mail any useful additions to vu-vlsi!plot so they may be
 *  included in later releases.
 *
 *  This file should be edited with 4-column tabs!  (:set ts=4 sw=4 in vi)
 */

#include <math.h>
#undef  GRAPHICSC__
#define GRAPHICSC__ /* this should be defined only here */
#include "plot.h"
#include "dbug.h"
#include <string.h>
#include <stdlib.h>
char           *strcpy(), *strncpy(), *strcat();
extern long GUBED;
/*BOOLEAN log_x = 0, log_y = 0; *//* set to false */
long            log_x = 0, log_y = 0;

#ifdef UNDEFINED__ /* following is now defined in plot.h */
extern long     term;
#endif /*UNDEFINED__*/

/*extern BOOLEAN term_init;*/
extern long     term_init;

extern struct termentry term_tbl[];

#ifdef UNDEFINED__
#define map_x(x) (long)((x-xmin)*xscale+xoffset)	/* maps floating point x to screen */
#define map_y(y) (long)((y-ymin)*yscale+yoffset)	/* same for y */
#endif /*UNDEFINED__*/

#define map_x(x) mapXToScreen(x) /* defined in plot.h */
#define map_y(x) mapYToScreen(x)

static double   raise(double x, long y)
{
	long            i;
	double          val;

	val = 1.0;
	for (i = 0; i < abs((int) y); i++)
	{
		val *= x;
	}
	if (y < 0)
	{
		return (1.0 / val);
	}
	return (val);
}


static double          make_tics(double tmin, double tmax, long logscale)
{
	double          xr, xnorm, tics, tic, l10;

	xr = fabs(tmin - tmax);

	l10 = log10(xr);
	if (logscale)
	{
		tic = raise(10.0, (l10 >= 0.0) ? (long) l10 : (long) floor(l10));
		if (tic < 1.0)
		{
			tic = 1.0;
		}
	}
	else
	{
		xnorm = pow(10.0, l10 -
					(double) ((l10 >= 0.0) ? (long) l10 : (long) floor(l10)));
		if (xnorm <= 2)
		{
			tics = 0.2;
		}
		else if (xnorm <= 5)
		{
			tics = 0.5;
		}
		else
		{
			tics = 1.0;
		}
		tic = tics * raise(10.0,
						   (l10 >= 0.0) ? (long) l10 : (long) floor(l10));
#ifdef UNDEFINED__
/* fossil code */
		tmin = tic * floor(tmin/tic);
		tmax = tic * ceil(tmax/tic);
		count = (tmax-tmin)/tic;
		if( count > 6 && tics == 0.2 )
		{
			tics = 0.5;
			tic = tics * raise(10.0,(l10 >= 0.0 ) ? (long)l10 : (long)floor(l10));
		}
		else if (count > 6 && tics == 0.5)
		{
			tics = 1.0;
			tic = tics * raise(10.0,(l10 >= 0.0 ) ? (long)l10 : (long)floor(l10));
		}
#endif /*UNDEFINED__*/
	}
	return (tic);
}

static char           *idx(char *a, char b)
{
	do
	{
		if (*a == b)
		{
			return (a);
		}
	} while (*a++);
	return (0);
}


static void num2str(double num, char str[])
{
	static char     temp[80];
	double          d;
	char           *a, *b;

	if ((d = fabs(num)) > 9999.0 || d < 0.001 && d > 1.e-10)
	{
		(void) sprintf(temp, "%-.3e", num);
	}
	else
	{
		(void) sprintf(temp, "%-.3f", num);
	}
	if ((b = idx(temp, 'e')))
	{
		a = b;
		while (*(--a) == '0')	/* trailing zeros */
		{
			;
		}
		if (*a == '.')
		{
			a--;
		}
		(void) strncpy(str, temp, (size_t) (a - temp) + 1);
		str[(long) (a - temp) + 1] = '\0';
		a = b + 1;	/* point to 1 after 'e' */
		(void) strcat(str, "e");
		if (*a == '-')
		{
			(void) strcat(str, "-");
		}
		a++;		/* advance a past '+' or '-' */
		while (*a == '0' && *(a + 1) != '\0')	/* leading zeroes */
		{
			a++;
		}
		(void) strcat(str, a);	/* copy rest of string */
	}
	else
	{
		(void) strcpy(str, temp);
		a = str + strlen(str) - 1;
		while (*a == '0')
		{
			*a = 0;
			a--;
		}
		if (*a == '.')
		{
			*a = 0;
		}
	}
}

#define trimx(X) ((Tmp__ = ((X) >= Ixmin) ? (X) : Ixmin) , \
				  ((Tmp__ <= Ixmax) ? Tmp__ : Ixmax))
#define trimy(Y) ((Tmp__ = ((Y) >= Iymin) ? (Y) : Iymin) , \
				  ((Tmp__ <= Iymax) ? Tmp__ : Iymax))

static long Ixmin, Ixmax, Iymin, Iymax, Tmp__;

static void findEdge(long xin,long yin,long *xout,long *yout)
{
	long      xEdge, yEdge, dx, dy;
	double    slope;

	dx = *xout - xin;
	dy = *yout - yin;
	xEdge = trimx(*xout);
	yEdge = trimy(*yout);
	if(dx != 0 && dy != 0)
	{
		slope = (double) dy / (double) dx;
		xEdge = (dx > 0) ? Ixmax : Ixmin;
		yEdge = yin + slope * (xEdge - xin);
		if(trimy(yEdge) != yEdge)
		{
			yEdge = (dy > 0) ? Iymax : Iymin;
			xEdge = xin + (yEdge - yin)/slope;
		}
	}
	*xout = xEdge;
	*yout = yEdge;
}

/* 
  findEdges() is called to find the visible portion if any of the line
  connecting (*x1,*y1) and (*x2,*y2).  If the line does not intersect the
  frame, findEdges() returns 0 without modifying its arguments.  Otherwise it
  returns 1 and the points are modified so as to lie on the boundary of the
  frame.
 */

static long findEdges(long *x1,long *y1,long *x2,long *y2)
{
	long      xEdge1, xEdge2, yEdge1, yEdge2, dx, dy;
	long      trialx, trialy;
	double    slope;

	xEdge1 = trimx(*x1);
	yEdge1 = trimy(*y1);
	xEdge2 = trimx(*x2);
	yEdge2 = trimy(*y2);

	if((dx = *x2 - *x1) == 0)
	{  /* vertical line */
		if(*x1 != xEdge1 || yEdge1 == yEdge2)
		{
			return (0);
		}
		*y1 = yEdge1;
		*y2 = yEdge2;
	}
	else if((dy = *y2 - *y1) == 0)
	{ /* horizontal line */
		if(*y1 != yEdge1 || xEdge1 == xEdge2)
		{
			return (0);
		}
		*x1 = xEdge1;
		*x2 = xEdge2;
	}
	else
	{ /* line not horizontal or vertical */
		if(xEdge1 == xEdge2 || yEdge1 == yEdge2)
		{ /* both points on same side of frame */
			return (0);
		}
		slope = (double) dy / (double) dx;
		trialx = xEdge1;
		trialy = *y1 + slope * (trialx - *x1);
		if(trialy != trimy(trialy))
		{
			trialy = yEdge1;
			trialx = *x1 + (trialy - *y1) / slope;
			if(trialx != trimx(trialx))
			{
				return (0);
			}
		}
		
		*x1 = trialx;
		*y1 = trialy;
		findEdge(trialx,trialy,x2,y2);
	}
	return (1);
}

long do_plot(plots, pcount, strings, scount,
			 xmin, xmax, ymin, ymax, logx, logy, xaxis, yaxis,
			 xlab, ylab, title)
#ifdef USEHANDLES
curve_points **plots;
plot_string  **strings;
#else /*USEHANDLES*/
curve_points *plots;
plot_string  *strings;
#endif /*USEHANDLES*/
long            pcount;		/* count of plots in linked list */
long            scount;   /* count of strings in linked list */
double          xmin, xmax;
double          ymin, ymax;
long            logx, logy;
long            xaxis, yaxis;
/* NB: xlab, ylab, and title used to be in struct curve_points */
char           *xlab, *ylab, *title;
{
	long                 i, n, x, y, xt, yt;
	long                 xlast, ylast;
	struct termentry *t = &term_tbl[term];
	/*BOOLEAN prev_undef;*/
	long                 prev_undef;
	long                 curve, string, xaxis_y, yaxis_x;
	int                  style;
#ifdef USEHANDLES
	curve_points        **this_plot;
	plot_string         **this_string;
#else /*USEHANDLES*/
	curve_points *this_plot;
	plot_string         *this_string;
#endif /*USEHANDLES*/
	double               xscale, yscale, xoffset, yoffset;
	double               ytic, xtic, ticplace, ticplace1;
	unsigned long        c;
	static char          label[80];
	WHERE("do_plot");

/*
	for MAC & conceivably for other terminals this initializes t->xmax,
	t->xmin, etc.  Hence it must precede their use.
*/
	if((*t->init) () > 0)
	{
		return (1);
	}
	
	if ((*t->graphics) () > 0)
	{
		return (1);
	}

	GraphMap.logx = log_x = logx;
	GraphMap.logy = log_y = logy;
	GraphMap.xmin = xmin;
	GraphMap.ymin = ymin;
	GraphMap.screenxmax = t->xmax;
	GraphMap.screenymax = t->ymax;

	xtic = make_tics(xmin, xmax, log_x);
	ytic = make_tics(ymin, ymax, log_y);

	if(t->h_char > 1)
	{
		GraphMap.xoffset = xoffset = (t->xmax - 2) * 0.15;		
		GraphMap.xscale = xscale = (t->xmax - 2) / (xmax - xmin) * 0.85;
		Ixmin = map_x(xmin);
		Ixmax = map_x(xmax);
	}
	else
	{
		GraphMap.xoffset = xoffset = 11;
		GraphMap.xscale = xscale = (t->xmax - 1 - xoffset)/(xmax - xmin);
		Ixmin = xoffset;
		Ixmax = t->xmax - 1;
	}
	if(t->v_char > 1)
	{
		GraphMap.yoffset = yoffset = (t->ymax - 2*t->v_char - 2) * 0.1;
		GraphMap.yscale = yscale = (t->ymax - 2*t->v_char - 2) / (ymax - ymin) * 0.9;
		Iymin = map_y(ymin);
		Iymax = map_y(ymax);
	}
	else
	{
		GraphMap.yoffset = yoffset = 2;
		GraphMap.yscale = yscale = (t->ymax - 2 - yoffset)/(ymax - ymin);
		Iymin = yoffset;
		Iymax = t->ymax - 2;
	}
	(*t->linetype) ((long) BORDERLINETYPE);	/* border linetype */

	/* draw plot border */
	(*t->move) (Ixmin, Iymin);
	(*t->vector) (Ixmax, Iymin);
	(*t->vector) (Ixmax, Iymax);
	(*t->vector) (Ixmin, Iymax);
	(*t->vector) (Ixmin, Iymin);

/*kb addition*/
	ticplace = ytic * floor(ymin / ytic);

	while(ticplace < ymin)
	{
		ticplace1 = ticplace + ytic;
		if(ticplace == ticplace1)
		{
			break;
		}
		ticplace = ticplace1;
	}

	/* Draw and label y-axis ticks*/
	while(ticplace <= ymax)
	{
		(*t->move) (Ixmin, map_y(ticplace));
		(*t->vector) (Ixmin + t->h_tic, map_y(ticplace));
		(*t->move) (Ixmax, map_y(ticplace));
		(*t->vector) (Ixmax - t->h_tic, map_y(ticplace));

		num2str(ticplace, label);
		/*
		vertstring((long)(0.6*xoffset),map_y(ticplace),label);
		horstring((long) xoffset - (strlen(label) * (long) t->h_char) / 2 - 2,
			map_y(ticplace), label);
		*/
		if(t->h_char > 1)
		{
			horstring(Ixmin - (strlen(label)*(long) t->h_char) / 2 - 1,
					  map_y(ticplace), label, CENTRE);
		}
		else
		{
			horstring(Ixmin - (strlen(label)*(long) t->h_char - 1) / 2 - 1,
					  map_y(ticplace), label, CENTRE);
		}
		ticplace1 = ticplace + ytic;
		if(ticplace == ticplace1)
		{ /* make sure we don't get stuck*/
			break;
		}
		ticplace = ticplace1;
	} /*while(ticplace <= ymax)*/

	ticplace = xtic * floor(xmin / xtic);
	while(ticplace < xmin)
	{
		ticplace1 = ticplace + xtic;
		if(ticplace == ticplace1)
		{ /* make sure we don't get stuck */
			break;
		}
		ticplace = ticplace1;
	} /*while(ticplace < xmin)*/

	/* Draw and label x-axis ticks*/
	while(ticplace <= xmax)
	{
		(*t->move) (map_x(ticplace), Iymin);
		(*t->vector) (map_x(ticplace), Iymin + t->v_tic);
		(*t->move) (map_x(ticplace), Iymax);
		(*t->vector) (map_x(ticplace), Iymax - t->v_tic);
		num2str(ticplace, label);
/*
		horstring(map_x(ticplace), (long) (0.6 * yoffset) + t->v_char/2, label);
*/
		if(t->v_char > 1)
		{
			horstring(map_x(ticplace),
					  Iymin - t->v_char/2, label, CENTRE);
		}
		else
		{
			horstring(map_x(ticplace), (long) 1, label, CENTRE);
		}
		ticplace1 = ticplace + xtic;
		if(ticplace == ticplace1)
		{ /* make sure we don't get stuck */
			break;
		}
		ticplace = ticplace1;
	} /*	while(ticplace <= xmax)*/

	/* DRAW AXES */
	xaxis_y = map_y(0.0);
	yaxis_x = map_x(0.0);

	if (logy || xaxis_y <= Iymin)
	{
		xaxis_y = Iymin;	/* save for impulse plotting */
	}
	else if (xaxis_y >= Iymax)
	{
		xaxis_y = Iymax;
	}
	else if (xaxis)
	{
		(*t->linetype) ((long) XAXISLINETYPE);	/* x-axis line type */
		(*t->move) (Ixmin, xaxis_y);
		(*t->vector) (Ixmax, xaxis_y);
	}

	if (yaxis && !log_x && yaxis_x > Ixmin && yaxis_x < Ixmax)
	{
		(*t->linetype) ((long) YAXISLINETYPE);	/* y-axis line type */
		(*t->move) (yaxis_x, Iymin);
		(*t->vector) (yaxis_x, Iymax);
	}

	this_plot = plots;
	/* DRAW TITLE (kb) */

	if (title != '\0')
	{
		horstring(map_x((xmax + xmin)/2.),Iymax + t->v_char, title, CENTRE);
	}

	if(t->v_char > 1)
	{
		horstring(map_x((xmax + xmin) / 2.), Iymin - 3*t->v_char/2,xlab,
				  CENTRE);
	}
	else
	{
		horstring(map_x((xmax + xmin) / 2.), (long) 0, xlab, CENTRE);	
	}

	if(t->h_char > 1)
	{
		vertstring((long) (0.2 * xoffset), map_y((ymax + ymin) / 2.), ylab);
	}
	else
	{
		vertstring(1,map_y((ymax + ymin) / 2.), ylab);
	}

	/* DRAW CURVES */

	for (curve = 0; curve < pcount; this_plot = NEXTGRAPH(this_plot), curve++)
	{
		n = PCOUNT(this_plot);
		style = PLOTSTYLE(this_plot);
		/* draw lines first */
		if(style & LINES)
		{
			(*t->linetype) (LINETYPE(this_plot));

			prev_undef = 1;
			for (i = 0; i < n; i++)
			{
				if (!POINTUNDEF(this_plot,i))
				{
					x = map_x(XCOORD(this_plot,i));
					y = map_y(YCOORD(this_plot,i));

					xt = trimx(x);
					yt = trimy(y);
					if(x == xt && y == yt)
					{ /* new point inside rectangle */
						if (prev_undef == 1)
						{
							(*t->move) (x,y);
						}
						else
						{
							if(prev_undef == 2)
							{	/* last point outside rectangle */
								findEdge(x,y,&xlast, &ylast);
							}
							(*t->move) (xlast,ylast);
							(*t->vector) (x,y);
						}
						prev_undef = 0;
					} /* if(x == xt && y == yt) */
					else
					{ /* current point outside rectangle */
						xt = x;
						yt = y;
						if(prev_undef == 0)
						{ /* last point in rectangle */
							findEdge(xlast,ylast,&xt,&yt);
							(*t->vector) (xt,yt);
						}
						else if(prev_undef == 2 &&
								findEdges(&xlast,&ylast,&xt, &yt))
						{ /* last point outside && edge/line intersection */
							(*t->move) (xlast,ylast);
							(*t->vector) (xt,yt);
						}
						prev_undef = 2;
					}  /* if(x == xt && y == yt){}else{} */
					xlast = x;
					ylast = y;
				} /* if (!POINTUNDEF(this_plot,i)) */
				else
				{
					prev_undef = 1;
				}
			} /*for (i = 0; i < n; i++)*/
		} /*if(style & LINES)*/
		
		if(style & (POINTS | IMPULSES | CHARS))
		{
			for(i = 0; i < n; i++)
			{
				if (!POINTUNDEF(this_plot,i))
				{
					x = map_x(XCOORD(this_plot,i));
					y = map_y(YCOORD(this_plot,i));
					xt = trimx(x);
					yt = trimy(y);

					if(style & IMPULSES && x == xt)
					{
						(*t->linetype) (LINETYPE(this_plot));
						(*t->move) (x, xaxis_y);
						(*t->vector) (x, yt);
					}

					if((style & (POINTS | CHARS)) && x == xt && y == yt)
					{ /* new point inside rectangle */
						(*t->move) (x,y);
						if(style & CHARS)
						{
							if((c = POINTSTRING(this_plot,i)[0]) != '\0' &&
							   c < ' ')
							{	/* drawn character */
								(*t->linetype) ((long) BORDERLINETYPE);
								(*t->point) (x, y, c-1);
							}
							else
							{	/* printable character(s) */
								horstring(x, y, POINTSTRING(this_plot,i),
										  CENTRE);
							}
						} /*if(style & CHARS)*/

						if(style & POINTS)
						{
							(*t->linetype) ((long) BORDERLINETYPE);
							(*t->point) (x, y, POINTTYPE(this_plot));
						}
					} /*if((style & (POINTS | CHARS)) && x == xt && y == yt)*/
				} /*if (!POINTUNDEF(this_plot,i))*/
			} /*for(i = 0; i < n; i++)*/
		} /*if(style & (POINTS | IMPULSES | CHARS))*/
	}/*for(curve=0;curve<pcount;this_plot=NEXTGRAPH(this_plot),curve++)*/

	/* DRAW STRINGS */
	this_string = strings;
	for (string = 0; string < scount; string++)
	{ /* Note: (x,y) may be outside frame */
		x = map_x(STRINGX(this_string));
		y = map_y(STRINGY(this_string));
#ifdef USEHANDLES
		horstring(x, y, *PLOTSTRING(this_string), JUSTIFYSTR(this_string));
#else /*USEHANDLES*/
		horstring(x, y, PLOTSTRING(this_string), JUSTIFYSTR(this_string));
#endif /*USEHANDLES*/
		this_string = NEXTSTRING(this_string);
	} /*for (string = 0; string < scount; string++)*/
	
	(*t->linetype) ((long) BORDERLINETYPE);
	(*t->text) ();

	return (0);
}

