/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Hconcat
#else
#define __SEG__ Hconcat
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <math.h>

extern long             argOK();
extern long             symbolSize();

Symbolhandle    hconcat(list)
Symbolhandle    list;
{
	long            i, j, k, l, length, tot, subtot;
	long            k1, k2;
	long            ncols, nrows, nrows1, nargs = NARGS(list);
	long            type;
	long            hcat;
	long            dims[2];
	Symbolhandle    result = (Symbolhandle) 0, symh;
	double          **resmat;
	WHERE("hconcat");
	
/* concatenate argument matrices sideways (hconcat) or vertically (vconcat) */

	OUTSTR[0] = '\0';
	if(strcmp(FUNCNAME,"hconcat") == 0)
	{
		hcat = 1;
	}
	else
	{
		hcat = 0;
	}
	symh = COMPVALUE(list,0);
	if(!argOK(symh, (long) 0, 1))
	{
		goto errorExit;
	}
	
	type = TYPE(symh);
	if(type != REAL && type != LOGIC)
	{
		badType(FUNCNAME,-type,1);
		goto errorExit;
	}
	
	tot = 0;

/* check all arguments before proceeding && found size of output */	
	for (i = 0; i < nargs; i++)
	{ /* for each component in list */

		symh = COMPVALUE(list, i);

		if(!argOK(symh,(long) 0,i+1))
		{
			goto errorExit;
		}
		else if (TYPE(symh) != type)
		{
			sprintf(OUTSTR,"ERROR: arguments to %s must all have same type",
					FUNCNAME);
		}
		else if(!isMatrix(symh,dims))
		{
			sprintf(OUTSTR,"ERROR: argument %ld to %s not matrix",i+1,FUNCNAME);
		}
		else if(i == 0)
		{
			length = dims[(hcat) ? 0 : 1];
		}
		else if (dims[(hcat) ? 0 : 1] != length)
		{
			sprintf(OUTSTR,
					"ERROR: all arguments to %s must have same number of %s",
					FUNCNAME, (hcat) ? "rows" : "columns");
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
		
		tot += dims[0]*dims[1];
	}

	if (tot / length > MAXCOORD)
	{			/* too many items for short coordinate */
		sprintf(OUTSTR,"ERROR: output of %s to large",FUNCNAME);
		goto errorExit;
	}

	if(hcat)
	{
		nrows = length;
		ncols = tot / nrows;
	}
	else
	{
		ncols = length;
		nrows = tot / ncols;
	}
	
	result = RInstall(SCRATCH,tot);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	setTYPE(result,type);
	setNDIMS(result,2);
	setDIM(result,1,nrows);
	setDIM(result,2,ncols);	/* number of columns */

	resmat = DATA(result);

	l = 0;
	for (i = 0; i < nargs; i++)
	{
		/* for each component in list */

		symh = COMPVALUE(list, i);

		subtot = symbolSize(symh);
		
		if(hcat)
		{
			for (j = 0; j < subtot; j++)
			{
				(*resmat)[l++] = DATAVALUE(symh,j);
			}
		}
		else
		{
			(void) isMatrix(symh,dims);
			nrows1 = dims[0];
			for(j = 0;j<nrows1;j++)
			{
				k1 = l;
				k2 = j;
				for(k = 0;k<ncols;k++)
				{
					(*resmat)[k1] = DATAVALUE(symh,k2);
					k1 += nrows;
					k2 += nrows1;
				}
				l++;
			}
		}
	}
		
	return (result);

  errorExit:
	putOUTSTR();

	return (0);
}
