/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef MPW
#ifdef MPW3
#pragma segment Help
#else
#define __SEG__ Help
#endif /*MPW3*/
#endif /*MPW*/

/*
  Modified by kb so as to use external help file
*/
  
#include "globals.h"
#include <ctype.h>
#include "dbug.h"

#ifdef MACINTOSH
#include "macIface.h"
#ifdef MPW
#include <StdLib.h>
#endif /*MPW*/
#endif /*MACINTOSH*/

#define LF   10
#define CR   13

#ifdef MACINTOSH
#define NEWLINE   CR
#else
#define NEWLINE   LF
#endif /*MACINTOSH*/

#define MARKER '='
#define NMARKERS 4   /* help topics start with NMARKERS MARKER's */
/* Both of the following strings should be NMARKERS long */
#define MARKERSTRING "===="
#define KEYSTRING    "????"
#define STARTKEYS    '#'
#define SEPKEYS      ','

#define EOFSTRINGLENGTH 7
#define EOFSTRING    "_E_O_F_"
#define EOFSTRINGSTART '_'

#define HELPROOM      2000 /* initial space for list of topics */
#define KEYROOM        400  /* initial space for list of keys */
#define NTOPICS        250 /* initial number of topics */
#define KEYLENGTH       50
#define TOPICNAMELENGTH 20
#define TOPICCHUNK      10
#define HELPCHUNK   (7*TOPICCHUNK)

#define STARTMATCH    1
#define ENDMATCH      2
#define MIDDLEMATCH   (STARTMATCH | ENDMATCH)
#define EXACTMATCH	  4
#define ANYMATCH      8
#define MAXKEYS      32 /* maximum number of keys allowed per file */
extern long            isNull();

static short           NHelpTopics = -1; /* < 0 means initHelp not called */
static char          **Topics = (char **) 0;
static short         **TopicOffsets = (short **) 0; /* offsets in Topics */
static unsigned long **KeyTable = (unsigned long **) 0; /*Help key info */
static char          **KeyList = (char **) 0; /* list of help keys */
static short           KeyOffsets[MAXKEYS], NKeys;
#ifdef USEFSEEK
static long          **FileOffsets = (long **) 0; /* positions in file */
static long            LineOffset = 0;
#endif /*USEFSEEK*/
static char          **Buffer = (char **) 0; /* line buffer */
static short           CurrentTopic = 0;
static short           BufferFull = 0; /* != 0 means a line is in buffer */
static char          **HelpFileName = (char **) 0;
static FILE           *HelpFile = (FILE *) 0;
static short           HelpEofHit = 0;

static void rewindHelp(void)
{
	rewind(HelpFile);
#ifdef USEFSEEK
	if(LineOffset > 0)
	{
		LineOffset = 0;
	}
#endif /*USEFSEEK*/
	BufferFull = 0;
	CurrentTopic = 0;
	HelpEofHit = 0;
} /*rewindHelp()*/

/* 
  Read upto n characters from fp into s[].  return 0 is EOF hit before
  any characters read
  A line starting with EOFSTRING is considered to be an internal EOF marker
  Line is put in *Buffer, without trailing NEWLINE
*/
static int getHelpLine(void)
{
	int         c, i, n = BUFFERLENGTH;
	char       *buffer = *Buffer;
	WHERE("getHelpLine");
	
#ifdef USEFSEEK
	if(LineOffset >= 0)
	{ /* need file position only during the first scann */
		LineOffset = ftell(HelpFile);
	}
#endif /*USEFSEEK*/
	for(i=0;i<n;i++)
	{
		c = getc(HelpFile);
#ifdef BINARYHELPFILE
		if(c == CR)
		{ /* discard CR when help file is opened in binary mode*/
			c = getc(HelpFile);
		}
#endif /*BINARYHELPFILE*/
		if(c == NEWLINE || c == EOF)
		{
			break;
		}
		else
		{
			buffer[i] = c;
		}
	} /*for(i=0;i<n;i++)*/

	if(buffer[0] == EOFSTRINGSTART &&
	   strncmp(buffer, EOFSTRING, EOFSTRINGLENGTH) == 0)
	{
		i = 0;
		c = EOF;
	}
	buffer[i] = '\0';
	
	HelpEofHit = (c == EOF);

	if(i == 0 && HelpEofHit)
	{
		return (0);
	}
	else
	{
		return (1);
	}
} /*getHelpLine()*/

/*
  read help file until line is found starting with ====topicName
  return pointer to topicName in buffer
  return NULL on EOF
*/

static int nextTopic(char *topic)
{
	int         i;
	char        *buffer;
	WHERE("nextTopic");

	if(HelpEofHit)
	{
		rewindHelp();
	}
	
	topic[0] = '\0';
	
	while(1)
	{
		if(!BufferFull)
		{
			if(getHelpLine() == 0)
			{
				return (0);
			}
		}
		else
		{
			BufferFull = 0;
		}
		buffer = *Buffer;
		if(strncmp(buffer, MARKERSTRING, NMARKERS) == 0)
		{
			BufferFull = 1;
			buffer += NMARKERS;
			for(i=0;i<TOPICNAMELENGTH;i++)
			{
				if(!isnamechar(buffer[i]))
				{
					break;
				}
				topic[i] = buffer[i];
			}
			topic[i] = '\0';
			BufferFull = 1;
			return (1);
		} /*if(strncmp(buffer, MARKERSTRING, NMARKERS) == 0)*/
	} /* while(1) */
} /*nextTopic()*/

/*
   Function to read list of help keys from file, allocate storage, and
   save them in *KeyList.  Offsets to the keys go in array KeyOffsets
   When called, the Key markers should be in the buffer.  Comma- and/or
   newline-separated keys start on the next line, terminated by a new
   set of key markers
*/

static void initKeys(void)
{
	int      c, nleft = KEYROOM, length, place = 0, ibuf;
	char     key[KEYLENGTH], *pc;
	char    *buffer;
	WHERE("initKeys");
	
	NKeys = 0;
	mydisphandle(KeyList);
	KeyList = mygethandle(nleft*sizeof(char));
	if(KeyList == (char **) 0)
	{
		goto errorExit;
	}
	
	while (getHelpLine() != 0)
	{ /* loop over lines in help file */
		if(strncmp(*Buffer, KEYSTRING, NMARKERS) == 0 ||
		   strncmp(*Buffer, MARKERSTRING, NMARKERS) == 0)
		{
			if((*Buffer)[0] == MARKER)
			{
				BufferFull = 1;
			}
			break;
		}
		ibuf = 0;

		do /* while(c != '\0') */
		{ /* collect keys on current line */
			key[0] = '\0';
			pc = key;
			buffer = *Buffer;
			while((c = buffer[ibuf++]) != SEPKEYS && c != '\0')
			{
				if(c != ' ' || key[0] != '\0')
				{ /* skip leading spaces*/
					*pc++ = c;
				}
			}
			pc--;
			while(*pc == ' ' && pc >= key)
			{ /* trim off trailing spaces */
				pc--;
			}
			pc[1] = '\0';
			if(key[0] != '\0')
			{ /* we got a key */
				length = strlen(key);
				if(length + 1 > nleft)
				{ /* we need more room */
					nleft = KEYROOM;
					KeyList = mygrowhandle(KeyList, place + nleft);
					if(KeyList == (char **) 0)
					{
						goto errorExit;
					}
				}
				strcpy(*KeyList + place, key);
				KeyOffsets[NKeys++] = place;
				place += length + 1;
				nleft -= length+1;
				if(NKeys == MAXKEYS)
				{
					break;
				}
			} /*if(key[0] != '\0')*/
		} while(c != '\0');
	} /*while (getHelpLine() != 0)*/
	if(NKeys == 0)
	{
		mydisphandle(KeyList);
		KeyList = (char **) 0;
	}
	else if((KeyList = mygrowhandle(KeyList, place)) == (char **) 0)
	{
		goto errorExit;
	}
	
	return;
	
  errorExit:
	mydisphandle(KeyList);
	KeyList = (char **) 0;
	NKeys = 0;
} /*initKeys()*/

/* comparison of strings ignoring case */
static long keycmp(long iKey, char * string)
{
	char     *key = *KeyList + KeyOffsets[iKey];
	int       c1, c2;
	int       i;
	int       n = strlen(key);
	WHERE("keycmp");

	for(i=0;i < n;i++)
	{
		c1 = key[i];
		c2 = string[i];
		c1 = (isupper(c1)) ? tolower(c1) : c1;
		c2 = (isupper(c2)) ? tolower(c2) : c2;
		if(c1 != c2)
		{ /* match if run out of string before we run out of key */
			return ((c2 == '\0') ? 0 : 1);
		}
		else if(c1 == '\0')
		{
			return 0;
		}
	}
	return (0);
} /*keycmp()*/

static void listKeys(void)
{
	short       iKey;
	short       maxlength = 0, length;
	short       nperline, nlines, line;
	char       *outstr;
		
	for(iKey=0;iKey<NKeys;iKey++)
	{
		length = strlen(*KeyList + KeyOffsets[iKey]);
		maxlength = (length > maxlength) ? length : maxlength;
	}
	maxlength++;
	nperline = SCREENWIDTH/maxlength;
	nperline = (nperline > 0) ? nperline : 1;
	nlines = (NKeys - 1)/nperline + 1;
	sprintf(OUTSTR,
		"Type 'help(key:\"heading\")', where heading is in following list:");
	putOUTSTR();
	for(line = 0;line < nlines;line++)
	{
		outstr = OUTSTR;
		for(iKey = line; iKey < NKeys; iKey += nlines)
		{
			strcpy(outstr, *KeyList + KeyOffsets[iKey]);
			length = strlen(outstr);
			outstr += length;

			if(iKey + nlines <= NKeys)
			{
				while(length++ <= maxlength)
				{
					*outstr++ = ' ';
				}
			}
			*outstr = '\0';
		} /*for(iKey = line; iKey < NKeys; iKey += nlines)*/
		putOUTSTR();
	} /*for(iKey=0;iKey<NKeys;)*/
} /*listKeys()*/

static void setKeyTable(char *buffer, long n)
{
	unsigned long   *keyTable = *KeyTable + n - 1;
	short            i, length;
	char             key[KEYLENGTH];
	char             c;
	
	*keyTable = 0;
	do /* while(c != '\0')*/
	{ /* loop over keys on line after '#'*/
		while(*buffer == ' ')
		{ /* skip spaces */
			buffer++;
		}
		length = 0;
		while((c = *buffer++) != SEPKEYS && c != '\0')
		{
			key[length++] = c;
		}
		while(length > 0 && key[length-1] == ' ')
		{ /* trim off trailing blanks */
			length--;
		}
		key[length] = '\0';
		
		if(key[0] != '\0')
		{
			for(i=0;i<NKeys;i++)
			{
				if(keycmp(i, key) == 0)
				{
					*keyTable |= (1L << i);
				}
			}
		} /*if(key[0] != '\0')*/		
	} while(c != '\0');
} /*setKeyTable()*/

/* Called by help every time, normally does anything only the first time */

static void initHelp(void)
{
	char       topic[TOPICNAMELENGTH+1], *keystart;
	long       room = 0, spaceleft = 0, topicsleft = 0;
	long       length, needed;
	WHERE("initHelp");
	
	if(NHelpTopics < 0)
	{
		if(HelpFileName == (char **) 0)
		{
			HelpFileName = mygethandle(strlen(*HELPFILENAME) + 1);
			if(HelpFileName == (char **) 0)
			{
				return;
			}
			strcpy(*HelpFileName,*HELPFILENAME);
			HelpFile = HELPFILE;
		} /*if(HelpFileName == (char **) 0)*/
		
		if(HelpFile == (FILE *) 0)
		{
#ifndef BINARYHELPFILE
			HelpFile = fopen(*HelpFileName,"r");
#else /*BINARYHELPFILE*/
			HelpFile = fopen(*HelpFileName,"rb"); /* binary mode for djgpp */
#endif /*BINARYHELPFILE*/
		}

		if(HelpFile == (FILE *) 0)
		{
			myerrorout("WARNING: unable to open help file");
			goto errorExit;
		}
		else
		{
#ifdef USEFSEEK
			LineOffset = 0;
#endif /*USEFSEEK*/
			rewindHelp();
		}
		
		mydisphandle5((char **) TopicOffsets, Topics, Buffer,
					  (char **) KeyTable, (char **) 0);		

		KeyTable = (unsigned long **) 0;

		TopicOffsets = (short **) mygethandle(NTOPICS * sizeof(short));
		Topics = mygethandle(HELPROOM * sizeof(char));
		Buffer = mygethandle(BUFFERLENGTH * sizeof(char));
		if(TopicOffsets == (short **) 0 ||
		   Topics == (char **) 0 || Buffer == (char **) 0)
		{
			goto noroom;
		}
#ifdef USEFSEEK
		mydisphandle((char **) FileOffsets);
		FileOffsets = (long **) mygethandle(NTOPICS * sizeof(long));
		if(FileOffsets == (long **) 0)
		{
			goto noroom;
		}
#endif /*USEFSEEK*/

		while(getHelpLine() != 0)
		{
			if(strncmp(*Buffer, KEYSTRING, NMARKERS) == 0)
			{
				initKeys();
				if(NKeys > 0)
				{ /* create table of keys associated with topics*/
					KeyTable = (unsigned long **)
						mygethandle(NTOPICS*sizeof(unsigned long));
					if(KeyTable == (unsigned long **) 0)
					{
						goto noroom;
					}
				}
				break;
			}
			else if(strncmp(*Buffer, MARKERSTRING, NMARKERS) == 0)
			{
				BufferFull = 1;
				break;
			}
		} /*while(getHelpLine() != 0)*/
		topicsleft = NTOPICS;
		spaceleft = HELPROOM;
		if(!BufferFull)
		{
			(*Buffer)[0] = '\0';
		}
		
	    /* find all help topics available */
		NHelpTopics = 0;
		while(nextTopic(topic) != 0)
		{
			length = strlen(topic);
			if(length + 1 > spaceleft)
			{
				spaceleft = HELPCHUNK;
				Topics = mygrowhandle(Topics,(room + spaceleft)*sizeof(char));
				if(Topics == (char **) 0)
				{
					goto noroom;
				}
			}
					
			if(topicsleft <= 0)
			{
				topicsleft = TOPICCHUNK;
				needed = NHelpTopics+topicsleft;
				TopicOffsets = (short **)
					mygrowhandle((char **) TopicOffsets, needed*sizeof(short));
				if(TopicOffsets == (short **) 0)
				{
					goto noroom;
				}
#ifdef USEFSEEK
				FileOffsets = (long **)
					mygrowhandle((char **) FileOffsets, needed*sizeof(long));
				if(FileOffsets == (long **) 0)
				{
					goto noroom;
				}
#endif /*USEFSEEK*/
				if(NKeys > 0)
				{
					KeyTable = (unsigned long **)
						mygrowhandle((char **) KeyTable,
									 needed*sizeof(unsigned long));
					if(KeyTable == (unsigned long **) 0)
					{
						goto noroom;
					}
				}
				
			} /*if(topicsleft <= 0)*/
					
			strcpy((*Topics) + room, topic);
			(*TopicOffsets)[NHelpTopics] = room;
#ifdef USEFSEEK
			(*FileOffsets)[NHelpTopics] = LineOffset;
#endif /*USEFSEEK*/
			
			NHelpTopics++;
			room += length+1;
			spaceleft -= length+1;
			topicsleft--;
			if(NKeys > 0 &&
			   (keystart = strchr(*Buffer, STARTKEYS)) != (char *) 0)
			{ /* retrieve keys for this entry */
				setKeyTable(keystart+1, NHelpTopics);
			}
			BufferFull = 0;
		} /*while(nextTopic(topic) != 0)*/
		rewindHelp();
#ifdef USEFSEEK
		LineOffset = -1; /* signals no need to use ftell()*/
#endif /*USEFSEEK*/
	} /*if(NHelpTopics < 0)*/
	return;
	
  noroom:
	myerrorout("WARNING: unable to allocate memory needed for help");

  errorExit:
	mydisphandle5((char **) TopicOffsets, Topics, Buffer,
				  (char **) 0, (char **) 0);
	TopicOffsets = (short **) 0;
	Topics = Buffer = (char **) 0;
#ifdef USEFSEEK
	mydisphandle((char **) FileOffsets);
	FileOffsets = (long **) 0;
#endif /*USEFSEEK*/
	NHelpTopics = -1;
	
} /*initHelp() */

static long       fillTopiclist(char *topic, short topiclist[], long matchType,
								unsigned long keyMask, short *maxlength)
{
	short           itopic, ntopics = 0, length;
	
	*maxlength = 0;
	for(itopic=0;itopic<NHelpTopics;itopic++)
	{ /* count topics to be listed */
		if(NKeys > 0 && (*KeyTable)[itopic] & keyMask ||
		   matchName(*Topics + (*TopicOffsets)[itopic], matchType,topic))
		{
			length = strlen(*Topics + (*TopicOffsets)[itopic]);
			*maxlength = (length > *maxlength) ? length : *maxlength;
			topiclist[ntopics++] = itopic;
		}
	}
	return (ntopics);
} /*fillTopiclist()*/

static void       listTopics(char *topic, long matchType)
{
	short          nperline, itopic, nlines, ntopics = 0, line;
	short          length, topiclist[500], maxlength = 0;
	unsigned long  keyMask = 0;
	char           *name, *outstr;
	WHERE("listTopics");

	SETUPINT(errorExit);
	
	if(matchType < 0)
	{
		keyMask = (-matchType <= MAXKEYS) ? 1L << -(matchType+1) : 0xffffffff;
	}
	
	initHelp();
	
	if(NHelpTopics > 0)
	{
		if(matchType == ANYMATCH || matchType == -(MAXKEYS + 1))
		{
			myerrorout("Help is available on the following topics: ");
		}
		else if(matchType < 0)
		{
			sprintf(OUTSTR,
					"The following help topics concern %s",
					*KeyList + KeyOffsets[-(matchType+1)]);
			putOUTSTR();
		}
	
		ntopics = fillTopiclist(topic, topiclist, matchType,
								keyMask, &maxlength);
		maxlength++;
		nperline = SCREENWIDTH/maxlength;
		if(nperline < 1)
		{
			nperline = 1;
		}
		nlines = (ntopics - 1)/nperline + 1;
			
		for(line = 0;line < nlines;line++)
		{
			outstr = OUTSTR;
			for(itopic = line;itopic < ntopics;itopic += nlines)
			{
				name = *Topics + (*TopicOffsets)[topiclist[itopic]];
				strcpy(outstr, name);
				length = strlen(name);
				outstr += length;
				if(itopic + nlines < ntopics)
				{
					while(length++ < maxlength)
					{
						*outstr++ = ' ';
					}
				}
				*outstr = '\0';
			} /*for(itopic = line;itopic < NHelpTopics;itopic += nlines)*/
			putOUTSTR();
			if (INTERRUPT != INTNOTSET)
			{
				goto errorExit;
			}
		} /*for(line = 0;line < nlines;line++)*/

		if(matchType == ANYMATCH)
		{
			myerrorout("For information on help, enter help(help)");
		}
		myerrorout("For help on topic foo, enter help(foo) or help(\"foo\")");
	} /*if(NHelpTopics > 0)*/
	
	/* fall through */

  errorExit:
	RETURN;
	
} /*listTopics()*/

static char      *SpecialNames[] =
{
	"break", "breakall", "else", "elseif", "for",
	"help", "Help", "if", "while",
	(char *) 0
};


/*
  Note: atan() and round() are omitted from this list because they have
  their own entries
*/

static char           *TransformationNames[] = 
{
	"abs"    ,"acos"   ,"asin"   ,"cos"    ,"exp"    ,"lgamma" ,"log"    ,
	"log10"  ,"sin"    ,"sqrt"   ,"tan"    ,"cosh"   ,"sinh"   ,"tanh"   ,
	"atanh"  , (char *) 0
};

#define isSpecial(NAME) matchKey(NAME, SpecialNames, (long *) 0)
#define isTransformation(NAME) matchKey(NAME, TransformationNames, (long *) 0)

Symbolhandle    help(list)
Symbolhandle    list;
{
	Symbolhandle    symh;
	long            iTopic, nTopics = 0, i, k;
	long            type, done;
	char            date[7];
	char           *tempName = (char *) 0;
	char           *name;
	double          val;
	long            nargs = NARGS(list), start = 0, nameLength;
	long            matchType = 0;
	char           *argTopic, *what, *keyword = (char *) 0;
	char            target[TOPICNAMELENGTH+1], topic[TOPICNAMELENGTH+1];
	char           *buffer;
	long            doneTransformation = 0;
#ifdef MACINTOSH
	OSType          types[1];
#endif /*MACINTOSH*/
	WHERE("help");
	
	/* NHelpTopics and TopicOffsets are initialized by initHelp() */

	SETUPINT(errorExit);

	if((symh = COMPVALUE(list,0)) != (Symbolhandle) 0 && nargs == 1 &&
	   TYPE(symh) == CHAR && strcmp(STRINGPTR(symh),"__teaser__") == 0)
	{
		myerrorout("Type 'help()' for list of topics or 'help(topic)' for help on topic");
		return (NULLSYMBOL);
	}
	
#ifdef UNDEFINED__
	if(HelpFileName == (char **) 0)
	{
		HelpFileName = mygethandle(strlen(*HELPFILENAME) + 1);
		if(HelpFileName == (char **) 0)
		{
			goto errorExit;
		}
		strcpy(*HelpFileName,*HELPFILENAME);
		HelpFile = HELPFILE;
		NHelpTopics = -1;
		rewindHelp();
	}
#endif /*UNDEFINED__*/
	keyword = isKeyword(symh);
	if(keyword)
	{
		if(strncmp(keyword,"key",3) == 0)
		{
			if(nargs != 1)
			{
				sprintf(OUTSTR,
						"ERROR: '%s' cannot be used with any other arguments",
						keyword);
				goto errorExit;
			}
			if(!isScalar(symh) || TYPE(symh) != CHAR)
			{
				notCharOrString(keyword);
				goto errorExit;
			}

			initHelp();
			if(NHelpTopics <= 0)
			{
				goto notAvailable;
			}
			if(NKeys <= 0)
			{
				sprintf(OUTSTR,
						"ERROR: no keys used in help file");
				goto errorExit;
			}
			
			if(strcmp("all", STRINGPTR(symh)) == 0)
			{
				matchType = -(MAXKEYS + 1);
				i = 0;
			}
			else if(STRINGVALUE(symh,0) == '?')
			{
				i = NKeys + 1;
			}
			else
			{
				for(i=0;i<NKeys;i++)
				{
					if(keycmp(i,STRINGPTR(symh)) == 0)
					{
						matchType = -(i + 1);
						break;
					}
				}
			}
			
			if(i >= NKeys)
			{ /* list keys */
				if(i == NKeys)
				{
					myprint("ERROR: Key ");
					myprint(STRINGPTR(symh));
					sprintf(OUTSTR,
							" does not match any key in the help file");
					putOUTSTR();
				}
				listKeys();
				if(i == NKeys)
				{
					goto errorExit;
				}
				goto normalExit;
			} /*if(i >= NKeys)*/
		}
		else if(strcmp(keyword,"file") == 0 || strcmp(keyword,"orig") == 0)
		{
			start = 2;
		}
		else if((keyword = isKeyword(symh = COMPVALUE(list,nargs-1))) &&
				(strcmp(keyword,"file") == 0  || strcmp(keyword,"orig") == 0))
		{
			start = 1;
			nargs--;
		}
	} /*if(keyword)*/
	
	if(start != 0)
	{ /* keywords 'file' or 'orig' */
		start--;
		if(strcmp(keyword,"orig") == 0)
		{
			if(TYPE(symh) != LOGIC || !isScalar(symh))
			{
				notTorF(keyword);
				goto errorExit;
			}
			if(DATAVALUE(symh,0) != 0.0)
			{
				NHelpTopics = -1;
				if(HelpFile != HELPFILE)
				{
					if(HelpFile != (FILE *) 0)
					{
						fclose(HelpFile);
					}
					HelpFile = HELPFILE;

					mydisphandle(HelpFileName);
					HelpFileName = mygethandle(strlen(*HELPFILENAME) + 1);
					if(HelpFileName == (char **) 0)
					{
						goto errorExit;
					}
					strcpy(*HelpFileName,*HELPFILENAME);
				} /*if(HelpFile != HELPFILE)*/				
				rewindHelp();
			} /*if(DATAVALUE(symh,0) != 0.0)*/
		} /*if(strcmp(keyword,"orig") == 0)*/
		else
		{ /* keyword 'file' */
			NHelpTopics = -1;
			if(TYPE(symh) != CHAR || !isScalar(symh))
			{
#ifdef MACINTOSH
				sprintf(OUTSTR,
						"ERROR: usage is %s(%s:\"fileName\") or %s(%s:\"\")",
						FUNCNAME,keyword,FUNCNAME,keyword);
#else
				sprintf(OUTSTR,
						"ERROR: usage is %s(%s:\"fileName\")",FUNCNAME,keyword);
#endif /*MACINTOSH*/
				goto errorExit;
			}
			tempName = STRINGPTR(symh);
			nameLength = strlen(tempName);
			if(nameLength >= PATHSIZE)
			{
				sprintf(OUTSTR,
						"ERROR: help file name too long");
				goto errorExit;
			}
#ifdef MACINTOSH
			if(HelpVolume)
			{
				SetVol(0L, HelpVolume);
			}
			if (nameLength == 0)
			{  /* On Mac, file:""  means use dialog box on mac */
				types[0] = 'TEXT';
				tempName = macOpen("\pSpecify the Help File", (Str255) 0,
								   READIT, 1, types);
				if(tempName == (char *) 0)
				{
					goto errorExit;
				}
				HelpVolume = Reply.vRefNum;
				nameLength = strlen(tempName);
			} /* if (nameLength == 0)*/
#endif /*MACINTOSH*/

/* Note: HELPFILE is never closed, even when we are using an alternative */
			if(HelpFile == HELPFILE)
			{
				rewindHelp();
			}
			else if(HelpFile != (FILE *) 0)
			{
				fclose(HelpFile);
			}
			HelpFile = (FILE *) 0;
			mydisphandle(HelpFileName);
			HelpFileName = (char **) 0;

			tempName = expandFilename(tempName);
			if(tempName == (char *) 0)
			{
				goto errorExit;
			}
			if(!isfilename(tempName))
			{
				goto errorExit;
			}
			nameLength = strlen(tempName);
			HelpFileName = mygethandle(nameLength + 1);
			if(HelpFileName == (char **) 0)
			{
				goto errorExit;
			}
			strcpy(*HelpFileName, tempName);
		} /*if(strcmp(keyword,"orig") == 0){}else{}*/
	} /*if(start != 0)*/

	initHelp();

	if(NHelpTopics <= 0)
	{
		goto notAvailable;
	}
	
	if (isNull(symh = COMPVALUE(list,0)) ||
		nargs == 1 && matchType == 0 &&
		!(keyword = isKeyword(symh)) &&
		TYPE(symh) == CHAR && isScalar(symh) && isscratch(NAME(symh)) &&
		strlen(name = STRINGPTR(symh)) <= NAMELENGTH)
	{ /* if single quoted character string as argument*/
		if(symh != (Symbolhandle) 0)
		{
			if(matchType >= 0)
			{
				if(!isSpecial(name))
				{/* not syntax element such as 'break', or 'help' */
					if(!scanPat(name,&matchType,target))
					{
						sprintf(OUTSTR,
								"ERROR: improper help topic '%s'",STRINGPTR(symh));
						goto errorExit;
					}

					for(i=0;i<NHelpTopics;i++)
					{			/* count the number of matches */
						name = *Topics + (*TopicOffsets)[i];
						if(matchName(name,matchType,target))
						{
							nTopics++;
						}
					}
				} /*if(!isSpecial(name))*/
				else
				{
					matchType = EXACTMATCH;
					nTopics = 1;
				}
			} /*if(matchType >= 0)*/			
		} /*if(symh != (Symbolhandle) 0)*/
		else
		{ /* null argument list */
			target[0] = '\0';
			matchType = ANYMATCH;
			nTopics = NHelpTopics;
		}
	}  /* if(single quoted character string as argument)*/

	if(nTopics > 1 || matchType < 0)
	{ /* 1 argument and more than one topic matched, list them */
		listTopics(target, matchType);
		if(INTERRUPT != INTNOTSET)
		{
			goto errorExit;
		}
	} /*if(nTopics > 1)*/
	else
	{ /* otherwise, print help */
		for(k=start;k<nargs;k++)
		{
			date[0] = '\0';
			
			symh = COMPVALUE(list,k);
			if(symh == (Symbolhandle) 0)
			{
				myerrorout("WARNING: empty argument to help()");
				continue;
			}
			type = TYPE(symh);
			name = NAME(symh);
			
			matchType = 0;

			if((keyword = isKeyword(symh)) && strcmp(keyword,"news") == 0)
			{
				if(type != REAL || !isScalar(symh) || 
				   (val = DATAVALUE(symh,0)) < 910000 ||
				   val > 999999 || val != floor(val))
				{
					sprintf(OUTSTR,
							"WARNING: value for keyword '%s' must be date of form yymmdd",
							keyword);
					putOUTSTR();
				}
				else
				{
					sprintf(date,"%6ld",(long) val);
				}
				argTopic = keyword;
			} /*if((keyword=isKeyword(symh)) && strcmp(keyword,"news") == 0)*/
			else if(keyword != (char *) 0 && !GUBED)
			{
				sprintf(OUTSTR,
						"WARNING: unrecognized keyword %s",keyword);
				putOUTSTR();
				continue;
			}
			else if (type == CHAR && isScalar(symh) && isscratch(name))
			{
				argTopic = STRINGPTR(symh);
				if(!isSpecial(argTopic) && scanPat(argTopic,&matchType,target))
				{ /* nTopics < 0 means 1st arg is syntax element */
					argTopic = target;
				}
				else
				{
					matchType = EXACTMATCH;
				}
			}
			else if(isscratch(name))
			{
				myerrorout("ERROR: improper argument to help");
				goto errorExit;
			}
			else
			{
				argTopic = NAME(symh);
			}

			if(isTransformation(argTopic))
			{
				argTopic = "transformations";
			}
			
			matchType = (matchType == 0) ? EXACTMATCH : matchType;
			if(matchType != EXACTMATCH)
			{
				rewindHelp();
			}
			
			for (iTopic = 0; iTopic < NHelpTopics; iTopic++)
			{ /* look for first matching topic name */
				if (matchName(*Topics + (*TopicOffsets)[iTopic], matchType,
							  argTopic))
				{
					strcpy(target, *Topics + (*TopicOffsets)[iTopic]);
					break;
				}
			} /*for (iTopic = 0; iTopic < NHelpTopics; iTopic++)*/

			if(strcmp(argTopic,"transformations") != 0 ||!doneTransformation++)
			{
				if (iTopic >= NHelpTopics)
				{
					if(type == BLTIN)
					{
						what = "function";
					}
					else if(type == MACRO)
					{
						what = "macro";
					}
					else
					{
						what = "topic";
					}
				
					sprintf(OUTSTR,
							"WARNING: no help on %s '%s'",what,argTopic);
					putOUTSTR();
				} /*if (iTopic >= NHelpTopics)*/
				else
				{
#ifdef USEFSEEK
					fseek(HelpFile, (*FileOffsets)[iTopic], 0);
					BufferFull = 0;
					HelpEofHit = 0;
#else /*USEFSEEK*/
					if(iTopic <= CurrentTopic)
					{
						rewindHelp();
					}
#endif /*USEFSEEK*/
					buffer = *Buffer;
					while(nextTopic(topic) != 0)
					{
						if(strcmp(topic,target) == 0)
						{
							BufferFull = 0;
							if(nargs - start > 1)
							{
								sprintf(OUTSTR,"Help on %s:",argTopic);
								putOUTSTR();
							}
						
							done = 0;
							while(!done)
							{
								if (INTERRUPT != INTNOTSET)
								{
									*OUTSTR = '\0';
									goto errorExit;
								}
								buffer[0] = '\0';
								if(getHelpLine() == 0 ||
								   (buffer[0] == MARKER && buffer[1] == MARKER
									&& buffer[2] == MARKER && buffer[3] == MARKER))
								{
									done = 1;
								}
								else
								{
									if(*date)
									{
										for(i=0;i<6;i++)
										{
											if(!isdigit(buffer[i]))
											{
												break;
											}
										}
										if(i == 6 && strncmp(buffer,date,6) < 0)
										{
											done = 1;
											break;
										}
									} /* if(*date) */
									myerrorout(buffer);
								}
							}	/* while(!done) */
							if(*buffer)
							{
								BufferFull = 1;
							}
							break;
						}
						BufferFull = 0;
					} /*while(nextTopic(topic) != 0)*/
					if(topic[0] == '\0')
					{
						rewindHelp();
					}
					else
					{
						CurrentTopic = iTopic;
					}
				}/*if (iTopic >= NHelpTopics){}else{}*/
			}/*if(strcmp(argTopic,"transformations") != 0 ||!doneTransformation++)*/
		} /*for(k=start;k<nargs;k++)*/
	} /*if(nTopics > 1){}else{}*/

  normalExit:
	RETURN (NULLSYMBOL);

  notAvailable:
	sprintf(OUTSTR,"WARNING: Help is not available");
	/* fall through*/

  errorExit:
	putOUTSTR();

	RETURN (0);
} /*help()*/



