/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Initialize
#else /*MPW3*/
#define __SEG__ Initialize
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <stdlib.h>

#ifdef MACINTOSH
#include "macIface.h"
#undef LF
#undef CR
#define LF      10
#define CR      13

#define MACROTYPE   'wstr'
#define ENDNAME     "_END_"
#define IDBASE      10000
#define NAMEPREFIX  "_MV_"
#endif /*MACINTOSH*/

/*
   Initialize pre-defined macros

*/

/*
   This file must be in a particular format to allow creation of a MPW rez
   file using script iniMacro.awk .  If additional macros are added they
   should follow this format.

   The script copies the copyright information and exctracts the values of
   macros MACROTYPE, ENDNAME, IDBASE and NAMEPREFIX (see above) and then
   ignores everything up through the line '/@STARTMACROS@/'.  It ignores all
   other  preprocessor commands.

   The quoted name of each macro followed by a comma with no intervening
   whitespace must be the second field of any line line whose first field
   is '/@MNAME@/'.

   The quoted text of a macro must start with a line whose first field is
   '/@MTEXT@/'. All lines except the last should end in '\\'

   The final line of a macro must terminate with '", /@MEND@/', where '"' is
   the terminating double quote for the macro and there must be no intervening
   whitespace before the comma.  For a one line macro this would be on the
   same line with '/@MTEXT@/'

   The script skips any lines bracketed by lines whose first fields are
   '/@NOTFORMACSTART@/' and '/@NOTFORMACEND@/'.

   ('@' is used here instead of '#' because the MPW C preprocessor got
   confused otherwise.)

   Preprocessor lines are not copied

   Scanning terminates when the line containing '/@ENDMACROS@/' is found

*/

#ifndef MACINTOSH /* macros are loaded from resources on Mac */
static struct
{
	char           *name;
	char           *text;
}               macros[] =
{
/*STARTMACROS*/
/*MNAME*/ "redo", /* redo() or redo(charVariable) or redo(macrovariable) */
/*MTEXT*/ "REDO<-macro({if($N==0){LASTLINE}else{$1}})\n\
REDO()", /*MEND*/
	
/*MNAME*/ "readcols",  /* readcols(filename,name1,...,namek [,echo:T or F]) */
/*MTEXT*/ "# readcols(filename,name1,...,namek [,echo:T or F]), only filename quoted\n\
if($v < 2){\n\
\terror(\"ERROR: too few arguments for $S\")\n\
}\n\
@file$$<-$1\n\
if(!ischar(@file$$) || !isscalar(@file$$)){\n\
\terror(\"ERROR: $1 not quoted string or CHARACTER scalar\")\n\
}\n\
@names$$ <- if($v == 2){\n\
\tif(ischar($2)){$2}else{\"$2\"}\n\
}else{$A[-1]}\n\
@n$$ <- if($v == 2){length(@names$$)}else{$v-1}\n\
@data$$<-if($k==0){vecread($1)}else{vecread($1,$K)}\n\
@data$$<-matrix(@data$$,@n$$)\n\
for(@i$$,run(@n$$)){\n\
\t<<@names$$[@i$$]>> <- cat(@data$$[@i$$,]);;\n\
}", /*MEND*/

/*MNAME*/ "twotailt",
/*MTEXT*/ "@tval$$<-abs($1);@df$$<-$2\n\
if(!isscalar(@tval$$) && !isscalar(@df$$) && length(@tval$$)!=length(@df$$)){\n\
\terror(\"ERROR: lengths of $S arguments don't match\");\n\
}\n\
2*(1-cumstu(@tval$$,@df$$))", /*MEND*/

/*MNAME*/ "resvsrankits",
/*MTEXT*/ "#resvsrankits([jvar] [,title:\"Your title\")\n\
if(!isdefined(DF) || !isdefined(SS) ||\\\n\
!isdefined(RESIDUALS) || !isdefined(HII)){\n\
\terror(\"ERROR: Apparently no linear model has been fitted\")\n\
}\n\
@m$$<-length(DF);@i$$<-if($v==0){1}else{$1}\n\
@r$$ <- RESIDUALS[,@i$$]/sqrt((SS[@m$$,@i$$,@i$$]/DF[@m$$])*(1 - HII))\n\
chplot(Rankits:rankits(@r$$),@r$$,\"*\",DD:0,$K,ylab:\"Studentized Resids\")", /*MEND*/

/*MNAME*/ "resvsyhat",
/*MTEXT*/ "#resvsyhat([jvar [, y]] [,title:\"Your title\")\n\
if(!isdefined(DF) || !isdefined(SS) ||\\\n\
!isdefined(RESIDUALS) || !isdefined(HII)){\n\
\terror(\"ERROR: Apparently no linear model has been fitted\")\n\
}\n\
@m$$<-length(DF)\n\
@i$$<-if($v==0){1}else{$1}\n\
@r$$ <- RESIDUALS[,@i$$]/sqrt((SS[@m$$,@i$$,@i$$]/DF[@m$$])*(1 - HII))\n\
@y$$ <- {if($v < 2){modelvars(0)[,@i$$]}else{($2)[,@i$$]}}\n\
chplot(Yhat:@y$$-RESIDUALS[,@i$$],@r$$,\"*\",DD:0,$K,\\\n\
ylab:\"Studentized Resids\")", /*MEND*/

/*MNAME*/ "resvsindex",
/*MTEXT*/ "#resvsindex([jvar] [,title:\"Your title\")\n\
if(!isdefined(DF) || !isdefined(SS) ||\\\n\
!isdefined(RESIDUALS) || !isdefined(HII)){\n\
\terror(\"ERROR: Apparently no linear model has been fitted\")\n\
}\n\
@m$$<-length(DF)\n\
@i$$<-if($v==0){1}else{$1}\n\
@r$$ <- RESIDUALS[,@i$$]/sqrt((SS[@m$$,@i$$,@i$$]/DF[@m$$])*(1 - HII))\n\
chplot(1,@r$$,\"*\",DD:0,$K,xlab:\"Case Number\",\\\n\
ylab:\"Studentized Resids\")", /*MEND*/

/*MNAME*/ "hist",
/*MTEXT*/ "# usage: hist(x), hist(x,nbins), or hist(x,binvec)\n\
if($v < 1 || $v > 2){\n\
\terror(\"ERROR: usage is $S(x), $S(x,nbins) or $S(x,binvec)\")\n\
}\n\
@x$$<-$1\n\
if(!isvector(@x$$)||!isreal(@x$$)){error(\"ERROR: $1 is not a REAL vector\")}\n\
if(anymissing(@x$$)){@x$$<-@x$$[!ismissing(@x$$)]}\n\
if($v == 2){\n\
\t@bins$$<-$2\n\
\tif(!isvector(@bins$$)||!isreal(@bins$$)){\n\
\t\terror(\"ERROR: $2 is not a REAL scalar or vector\")\n\
\t}\n\
}else{\n\
\t@bins$$ <- floor(log(length(@x$$))/log(2)) + 1\n\
}\n\
@nbins$$ <- if(isscalar(@bins$$)){@bins$$}else{length(@bins$$)-1}\n\
if(@nbins$$ <= 0 || @nbins$$ != floor(@nbins$$)){\n\
\terror(\"ERROR: $2 is not positive integer\")\n\
}\n\
if(isscalar(@bins$$)){\n\
\t@min$$ <- 1.05*min(@x$$) - .05*max(@x$$)\n\
\t@max$$ <- 1.05*max(@x$$) - .05*min(@x$$)\n\
\t@widths$$ <- (@max$$ - @min$$)/@nbins$$\n\
\t@bins$$ <- run(@min$$,@max$$+.1*@widths$$,@widths$$)\n\
} else {\n\
\tif(min(@x$$) <=min(@bins$$) || max(@x$$) > max(@bins$$)){\n\
\t\terror(\"ERROR: minimum or maximum of data outside of bins\")\n\
\t}\n\
}\n\
@widths$$ <- movavg(1,@bins$$)[-1]\n\
if(min(@widths$$) < 0){\n\
\terror(\"ERROR: bin boundaries are not monotonic\")\n\
}\n\
@hts$$ <- bin(@x$$, @bins$$)$counts/@widths$$\n\
@xpts$$ <- cat(hconcat(@bins$$,@bins$$,@bins$$)')[-1]\n\
@ypts$$ <- cat(0,hconcat(@hts$$,@hts$$,rep(0,length(@hts$$)))',0)/length(@x$$)\n\
lineplot(@xpts$$,@ypts$$,$K,xlab:\"$1\",ylab:\"Density\",\\\n\
\tyaxis:F,title:\"Histogram of $1 with total area 1\")\n\
delete(@x$$,@bins$$,@nbins$$,@hts$$,@xpts$$,@ypts$$,@widths$$)", /*MEND*/

/*MNAME*/ "boxcox",
/*MTEXT*/ "@x$$<-$1;@power$$<-$2\n\
if(!isscalar(@power$$)){\n\
\terror(\"ERROR: $S power not a scalar\");\n\
}\n\
@gm$$ <- exp(describe(log(@x$$),mean:T))'\n\
if ( @power$$ == 0 ) {\n\
\t@ans$$ <- @gm$$ * log(@x$$)\n\
} else {\n\
\t@ans$$ <- (@x$$^@power$$ - 1)/(@power$$*@gm$$^(@power$$-1))\n\
}\n\
array(@ans$$,dim(@x$$))", /*MEND*/

/*MNAME*/ "makefactor",
/*MTEXT*/ "# makefactor(values) where values either REAL or CHAR\n\
if($N != 1){\n\
\terror(\"ERROR: $S requires one argument\")\n\
}\n\
@values$$ <- $1\n\
if(!isvector(@values$$)){\n\
\terror(\"ERROR: argument to $S must be vector\")\n\
}\n\
if(ischar(@values$$)){\n\
\tfactor(match(@values$$,unique(@values$$)))\n\
}else{\n\
\tfactor(match(@values$$,unique(sort(@values$$))))\n\
}", /*MEND*/

/* Note: white space squeezed out of following to allow MPW compilation */
/*MNAME*/ "resid",
/*MTEXT*/ "# resid() or resid(model), model is a linear model\n\
if($N > 1){\n\
error(\"ERROR: usage is $S() or $S(model)\"\n\
}\n\
if($N == 1){\n\
manova($1,silent:T)\n\
}\n\
if(!isdefined(RESIDUALS)||!isdefined(SS)||!isdefined(HII)||!isdefined(DF)){\n\
error(\"ERROR: no active model; try $S(model)\")\n\
}\n\
@m$$<-dim(SS)[1]\n\
@mse$$<-diag(SS[@m$$,,])'/DF[@m$$]\n\
@r$$<-RESIDUALS/sqrt((1-HII)*@mse$$)\n\
hconcat(\\\n\
modelvars(0),\\\n\
@r$$,\\\n\
HII,\\\n\
@r$$^2*HII/((1-HII)*sum(DF[-@m$$])),\\\n\
@r$$*sqrt((DF[@m$$]-1)/(DF[@m$$]-@r$$^2))\\\n\
)", /*MEND*/

/*MNAME*/ "yhat",
/*MTEXT*/ "# yhat() or yhat(model), where model is a linear model\n\
if($N > 1){\n\
\terror(\"ERROR: usage is $S() or $S(model)\")\n\
}\n\
if($N == 1){\n\
\tmanova($1,silent:T)\n\
}\n\
if(!isdefined(RESIDUALS) || !isdefined(SS) ||\\\n\
\t!isdefined(HII) || !isdefined(DF)){\n\
\terror(\"ERROR: apparently no active linear model; try $S(model)\")\n\
}\n\
@m$$<-dim(SS)[1]\n\
@mse$$<-diag(SS[@m$$,,])'/DF[@m$$]\n\
@depvar$$<-modelvars(0)\n\
hconcat(\\\n\
\t@depvar$$ ,\\\n\
\t@depvar$$-RESIDUALS,\\\n\
\tRESIDUALS/(1-HII),\\\n\
\tsqrt(HII*@mse$$),\\\n\
\tsqrt((1+HII)*@mse$$)\\\n\
)", /*MEND*/

/*MNAME*/ "anovapred",
/*MTEXT*/ "# anovapred(a,b,...) , where a, b, ..., are all the factors in the model\n\
if($v == 0){\n\
\terror(\"ERROR: usage is $S(a,b,...) where a,b,... are all the factors in model\")\n\
}\n\
if(!isdefined(DEPVNAME) || !isdefined(RESIDUALS)||\\\n\
\t !isdefined(HII) || !isdefined(DF)){\n\
\terror(\"ERROR: $S: apparently no active linear model\")\n\
}\n\
@error$$<-dim(SS)[1]\n\
@mse$$<-diag(SS[@error$$,,])'/DF[@error$$]\n\
makestr(estimate:tabs(<<DEPVNAME>>-RESIDUALS,$V,mean:T),\\\n\
SEest:sqrt(tabs(@mse$$*HII,$V,mean:T)),\\\n\
SEpred:sqrt(tabs(@mse$$*(1+HII),$V,mean:T)))", /*MEND*/

/*MNAME*/ "makecols",
/*MTEXT*/ "# makecols(y,name1,name2,...,namek), y a matrix, name1,... unquoted names\n\
if($v < 2){\n\
\terror(\"ERROR: too few arguments for $S\")\n\
}\n\
@data$$<-$1\n\
if(!isreal(@data$$) || !ismatrix(@data$$)){\n\
\terror(\"ERROR: $1 not REAL matrix scalar\")\n\
}\n\
@data$$<-matrix(@data$$,nrows(@data$$))\n\
@names$$ <- if($v == 2){\n\
\tif(ischar($2)){$2}else{\"$2\"}\n\
}else{$A[-1]}\n\
@n$$ <- min(length(@names$$),ncols(@data$$))\n\
for(@i$$,run(@n$$)){\n\
\t<<@names$$[@i$$]>> <- cat(@data$$[,@i$$]);;\n\
}", /*MEND*/

/*MNAME*/ "colplot",
/*MTEXT*/ "#colplot(y [,title:\"Title of your choice\")\n\
if($N<1){error(\"ERROR: $S needs at least 1 argument\")}\n\
chplot(1,$1,lines:T,$K,xlab:\"Row Number\")", /*MEND*/

/*MNAME*/ "rowplot",
/*MTEXT*/ "#rowplot(y [,title:\"Title of your choice\")\n\
if($N<1){error(\"ERROR: $S needs at least 1 argument\")}\n\
chplot(1,y:($1)',lines:T,$K,xlab:\"Column Number\")", /*MEND*/

/*MNAME*/ "fcolplot",
/*MTEXT*/ "#fcolplot(fileName, y [,title:\"Title of your choice\")\n\
if($N<2){error(\"ERROR: $S needs at least 2 arguments\")}\n\
chplot(1,$2,lines:T,file:$1,$K,xlab:\"Row Number\")", /*MEND*/

/*MNAME*/ "frowplot",
/*MTEXT*/ "#frowplot(fileName, y [,title:\"Title of your choice\")\n\
if($N<2){error(\"ERROR: $S needs at least 2 arguments\")}\n\
chplot(1,y:($2)',lines:T,file:$1,$K,xlab:\"Column Number\")", /*MEND*/

/*MNAME*/ "regs",
/*MTEXT*/ "#regs(x,y),matrix or vector y, matrix x\n\
@Xvars$$<-$1\n\
@Y<-$2\n\
if(!ismatrix(@Xvars$$)||!ismatrix(@Y)){\n\
error(\"ERROR: usage: $S(x,y), matrix x, vector or matrix y\")\n\
}\n\
@p$$<-length(@Xvars$$)/dim(@Xvars$$)[1]\n\
@X1<-@Xvars$$[,1]\n\
STRMODEL<-\"@Y=@X1\"\n\
if(@p$$>1){\n\
for(@i$$,run(2,@p$$)){\n\
<<paste(\"@X\",@i$$,sep:\"\")>><-@Xvars$$[,@i$$]\n\
STRMODEL<-paste(STRMODEL,\"+@X\",@i$$,sep:\"\")\n\
}\n\
}\n\
if(length(@Y)==dim(@Y)[1]){\n\
regress()\n\
}else{\n\
manova()\n\
print(\"NOTE: use secoefs() to get coefficients and standard errors.\")\n\
}", /*MEND*/

/*MNAME*/ "breakif",
/*MTEXT*/ "if($1){\n\
\t@depth$$ <- if($N==1){1}else{$2}\n\
\t@break$$ <- macro(paste(\"delete(@depth$$);break \",@depth$$,\";\"))\n\
\t@break$$()\n\
}", /*MEND*/

/*NOTFORMACSTART*/ /*macros up to NOTFORMACEND are not defined on Macintosh*/
#ifdef DEFINEEDIT
#ifdef DJGPP

/*MNAME*/ "edit",
/*MTEXT*/ "# edit(realVar), edit(macro), edit(), or edit(0)\n\
@arg$$ <- if($N == 0){macro(\"\")}else{$1}\n\
if($N > 2 || (!ismacro(@arg$$) && !isreal(@arg$$))){\n\
error(\"ERROR: usage: $S(realVar), $S(macro), $S(), or $S(0)\")\n\
}\n\
@save$$<-if($N == 2){$2}else{F}\n\
if(!islogic(@save$$) || !isscalar(@save$$)){\n\
\terror(\"ERROR: 2nd argument to $S must be T or F\")\n\
}\n\
@tmpfile$$<-paste(\"m\",round(10000000*runi(1)),sep:\"\")\n\
@editor$$<-if(isdefined(EDITOR) && ischar(EDITOR) && isscalar(EDITOR)){\n\
\tEDITOR\n\
}else{\n\
\t\"edit\"\n\
}\n\
if(ismacro(@arg$$)){\n\
\tmacrowrite(@tmpfile$$,edited:@arg$$[[1]])\n\
}else{\n\
\t@vector$$ <- isscalar(@arg$$) && @arg$$[1] == 0\n\
\tif(!@vector$$){\n\
\t\tmatwrite(@tmpfile$$,edited:@arg$$)\n\
\t}\n\
}\n\
shell(paste(@editor$$,@tmpfile$$))\n\
if(ismacro(@arg$$)){\n\
\t@arg$$<-macroread(@tmpfile$$)\n\
}else{\n\
\tif(!@vector$$){\n\
\t\t@arg$$<-matread(@tmpfile$$)\n\
\t}else{\n\
\t\t@arg$$<-vecread(@tmpfile$$)\n\
\t}\n\
}\n\
shell(paste(\"erase\",@tmpfile$$))\n\
if(@save$$){\n\
\t$1 <- @arg$$\n\
\tdelete(@arg$$)\n\
}else{\n\
\t@arg$$\n\
}", /*MEND*/

#elif defined(UNIX) /* DJGPP */

/*MNAME*/ "edit",
/*MTEXT*/ "# edit(realVar), edit(macro), edit(), or edit(0)\n\
@arg$$ <- if($N == 0){macro(\"\")}else{$1}\n\
if($N > 2 || (!ismacro(@arg$$) && !isreal(@arg$$))){\n\
error(\"ERROR: usage: $S(realVar), $S(macro), $S(), or $S(0)\")\n\
}\n\
@save$$<-if($N == 2){$2}else{F}\n\
if(!islogic(@save$$) || !isscalar(@save$$)){\n\
\terror(\"ERROR: 2nd argument to $S must be T or F\")\n\
}\n\
@tmpfile$$<-paste(\"/tmp/macanova.\",round(100000*runi(1)),sep:\"\")\n\
@editor$$<-if(isdefined(EDITOR) && ischar(EDITOR) && isscalar(EDITOR)){\n\
\tEDITOR\n\
}else{\n\
\t\"vi\"\n\
}\n\
if(ismacro(@arg$$)){\n\
\tmacrowrite(@tmpfile$$,edited:@arg$$[[1]])\n\
}else{\n\
\t@vector$$ <- isscalar(@arg$$) && @arg$$[1] == 0\n\
\tif(!@vector$$){\n\
\t\tmatwrite(@tmpfile$$,edited:@arg$$)\n\
\t}\n\
}\n\
shell(paste(@editor$$,@tmpfile$$))\n\
if(ismacro(@arg$$)){\n\
\t@arg$$<-macroread(@tmpfile$$)\n\
}else{\n\
\tif(!@vector$$){\n\
\t\t@arg$$<-matread(@tmpfile$$)\n\
\t}else{\n\
\t\t@arg$$<-vecread(@tmpfile$$)\n\
\t}\n\
}\n\
shell(paste(\"rm\",@tmpfile$$))\n\
if(@save$$){\n\
\t$1 <- @arg$$\n\
\tdelete(@arg$$)\n\
}else{\n\
\t@arg$$\n\
}", /*MEND*/

#endif /*DJGPP*/
#endif /*DEFINEEDIT*/
/*NOTFORMACEND*/

/*MNAME*/ "console",
/*MTEXT*/ "#y <- console()\n\
if($k==0){vecread(CONSOLE)}else{vecread(CONSOLE,$K)}", /*MEND*/

/*MNAME*/ "model",
/*MTEXT*/ "#model(y=a+b), e.g., sets STRMODEL to \"y=a+b\"\n\
STRMODEL <- \"$1\";;", /*MEND*/

/*MNAME*/ "getmacros",
/*MTEXT*/ "#getmacros(macro1,macro2,...[,quiet:T])\n\
if($v==0){error(\"ERROR: must specify at least one macro name for $S\")}\n\
if(!isdefined(MACROFILE)){\n\
\terror(\"ERROR: CHARACTER variable MACROFILE does not exist\")}\n\
@args$$<-$A\n\
for(@i$$,run($v)){\n\
\t@macname$$<-@args$$[@i$$]\n\
\t<<@macname$$>> <- if($k==0){\n\
\t\tmacroread(MACROFILE,@macname$$)\n\
\t}else{\n\
\t\tmacroread(MACROFILE,@macname$$,$K)\n\
\t}\n\
};;", /*MEND*/

/*MNAME*/ "getdata",
/*MTEXT*/ "# y <- getdata(datasetName [,quiet:T])\n\
if($v>1){\n\
\terror(\"ERROR: too many arguments for $S\")\n\
}\n\
if($v==1){\n\
\tif($k==0){matread(DATAFILE,\"$1\")}else{matread(DATAFILE,\"$1\",$K)}\n\
}else{\n\
\tif($k==0){matread(DATAFILE)}else{matread(DATAFILE,$K)}\n\
}", /*MEND*/

/*NOTFORMACSTART*/ /*macros up to NOTFORMACEND are not defined on Macintosh*/
#ifdef TEK
/* on some terminals, in particular Versaterm, puts in Tektronix 40xx mode */
/*MNAME*/ "tek",
/*MTEXT*/ "putascii(cat(29,27,56))", /*MEND*/

/* puts xterm in Tek4014 mode */
/*MNAME*/ "tekx",
/*MTEXT*/ "putascii(cat(27,91,63,51,56,104,27,56))", /*MEND*/

/* on some terminals, in particular Versaterm, puts in VT100 mode */
/*MNAME*/ "vt",
/*MTEXT*/ "putascii(cat(27,50))", /*MEND*/

/* puts xterm in vt100 mode */
/*MNAME*/ "vtx",
/*MTEXT*/ "putascii(cat(27,3))", /*MEND*/

#endif /*TEK*/
/*NOTFORMACEND*/

/*ENDMACROS*/
	(char *) 0, (char *) 0
};
#endif /*MACINTOSH*/

#ifdef UNDEFINED__ /* following no longer needed */
#ifdef MACINTOSH

/*
   Clean up macros for Macintosh
   This involves replacing LF by CR and correcting an MPW compiler bug
   on its treatment of '\t'
*/

static void      cleanMacro(char *stringval)
{
	char        *pc1, *pc2;

	for (pc1 = pc2 = stringval;*pc1;pc1++, pc2++)
	{
		if (*pc1 == LF)
		{
			*pc2 = CR;
		}
		else if(pc1[0] == '\\' && pc1[1] == 't')
		{/* to compensate for compiler bug */
			pc1++;
			*pc2 = '\t';
		}
		else
		{
			*pc2 = *pc1;
		}
	} /*for (pc1 = pc2 = stringval;*pc1;pc1++, pc2++)*/
	*pc2 = '\0';
} /*cleanMacro()*/
#endif /*MACINTOSH*/
#endif /*UNDEFINED__*/

/*
	Install text in symbol table as MACRO name
	There are length characters in text which may not be null-terminated
*/
static long installMacro(char * text, char *name, int length)
{
	Symbolhandle       symh;
	
	symh = Install(name, MACRO);

	TMPHANDLE = mygethandle(length + 1);
	if(symh == (Symbolhandle) 0 || TMPHANDLE == (char **) 0)
	{
		goto errorExit;
	}
	setSTRING(symh,TMPHANDLE);
	strncpy(*TMPHANDLE, text, length);
	(*TMPHANDLE)[length] = '\0';

#ifdef UNDEFINED__
#ifdef MACINTOSH
	cleanMacro(*TMPHANDLE);
	if (strlen(*TMPHANDLE) != length)
	{
		TMPHANDLE = mygrowhandle(TMPHANDLE, strlen(*TMPHANDLE) + 1);
		if (TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
	}
#endif /*MACINTOSH*/
#endif /*UNDEFINED__*/
	
	setNDIMS(symh,1);
	setDIM(symh,1,1);
	
	return (1);

  errorExit:
	return (0);
} /*installMacro()*/

/*
	Initialize pre-defined macros
	On Macintosh they are loaded from resources of type 'wstr' with id's
	IDBASE+1, IDBASE+2, ...,  The resource name is the macro name, prefixed
	by NAMEPREFIX.  The last resource has name NAMEPREFIXENDNAME ("_MV__END_")
	and is not loaded.
	If resedit is used to edit macros or add macros, it would be best if no
	gaps were left in the id numbers.
*/

long iniMacros(void)
{
#ifndef MACINTOSH
	int        i;

	for (i = 0; macros[i].name; i++)
	{
		if (!installMacro(macros[i].text, macros[i].name,
						  strlen(macros[i].text)))
		{
			goto errorExit;
		}
	}
#else /*MACINTOSH*/
	Integer          id = IDBASE, id1;
	Integer          i, count = CountResources(MACROTYPE);
	char            *prefix = NAMEPREFIX;
	Integer          prefixLength = strlen(prefix);
	Str255           resName;
	char            *name = (char *) resName;
	ResType          type;
	unsigned char  **text;
	
	/*
	   examine all resources of the wanted type
	   If the name starts with prefix, install it as long as it is not
	   the name of the dummy last macro
	*/
	for(i=1; i<= count; i++)
	{
		text = (unsigned char **) GetIndResource(MACROTYPE, i);
		if (ResError() != noErr)
		{
			break;
		}
		GetResInfo((Handle) text, &id1, &type, resName);
		if (ResError() != noErr)
		{
			goto errorExit;
		}
		PtoCstr(name);
		if (strncmp(prefix, name, prefixLength) == 0 &&
			strcmp(name + prefixLength, ENDNAME) != 0)
		{
			HLock((Handle) text);
		/* first two characters of *text give number of characters */
			if (!installMacro((*text) + 2, name + prefixLength,
							  (*text)[0] * 256 + (*text)[1]))
			{
				goto errorExit;
			}
			HUnlock((Handle) text);
		}
		ReleaseResource((Handle) text);
	} /*while(1)*/
#endif /*MACINTOSH*/
	return (1);

  errorExit:
	return (0);
} /*iniMacros()*/ 
