/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Initialize
#else
#define __SEG__ Initialize
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <stdlib.h>

#ifdef MACINTOSH
#undef LF
#undef CR
#define LF      10
#define CR      13
#endif /*MACINTOSH*/

extern Symbolhandle Transform();
extern Symbolhandle anovacoefs();
extern Symbolhandle array();
extern Symbolhandle batch();
extern Symbolhandle bin();
extern Symbolhandle boxplot();
extern Symbolhandle cellstats();
extern Symbolhandle changestr(); /*kb*/
extern Symbolhandle cholesky(); /*kb*/
extern Symbolhandle cluster(); /*kb*/
extern Symbolhandle columnops(); /*kb*/
extern Symbolhandle compnames();
extern Symbolhandle concat();
extern Symbolhandle contrast();
extern Symbolhandle cor();
extern Symbolhandle cumcdf();
extern Symbolhandle deleter();
extern Symbolhandle describe();
extern Symbolhandle diag(); /*kb*/
extern Symbolhandle dim(); /*kb*/
extern Symbolhandle eigen();
extern Symbolhandle fft();
extern Symbolhandle getseed();
extern Symbolhandle glm();
extern Symbolhandle hconcat(); /*kb*/
extern Symbolhandle help();
extern Symbolhandle iswhat(); /*kb*/
extern Symbolhandle inforead(); /*kb*/
extern Symbolhandle listbrief();
extern Symbolhandle macro();
extern Symbolhandle macroNargs(); /*kb*/
extern Symbolhandle makefactor();
extern Symbolhandle makestr();
extern Symbolhandle matrix();
extern Symbolhandle modelvar();
extern Symbolhandle movavg();
extern Symbolhandle myplot();
extern Symbolhandle nameOf(); /*kb*/
extern Symbolhandle ndims(); /*kb*/
extern Symbolhandle outer();
extern Symbolhandle partacf();
extern Symbolhandle paste();
extern Symbolhandle polyroot();
extern Symbolhandle power1();
extern Symbolhandle predtable();
extern Symbolhandle print();
extern Symbolhandle putascii();
extern Symbolhandle qr();
extern Symbolhandle rangen();
extern Symbolhandle rational();
extern Symbolhandle readdata();
extern Symbolhandle regpred();
extern Symbolhandle rep();
extern Symbolhandle restore();
extern Symbolhandle run();
extern Symbolhandle samplesize();
extern Symbolhandle save();
extern Symbolhandle setoptions(); /*kb*/
extern Symbolhandle setseed();
extern Symbolhandle sort();
extern Symbolhandle shell();
extern Symbolhandle solve();
extern Symbolhandle split();
extern Symbolhandle spool();
extern Symbolhandle stemleaf();
extern Symbolhandle svd(); /*kb*/
extern Symbolhandle swp(); /*kb*/
extern Symbolhandle t2val();
extern Symbolhandle tabs();
extern Symbolhandle toeplitz();
extern Symbolhandle transpose();
extern Symbolhandle trideigen();
extern Symbolhandle tserops();
extern Symbolhandle tval();
extern Symbolhandle varnames(); /*kb*/
extern Symbolhandle xvariables(); /*kb*/
extern Symbolhandle yates();


static struct
{                               /* constants */
	char           *name;
	double          value;
}                   constants[] =
{
	"DELTAT",    1,                      
	"PI",        3.14159265358979323846, 
	"E",         2.71828182845904523536, 
	"DEGPERRAD", 57.29577951308232087680,
	0,           0                       
};

static struct
{
	char           *name;
	char           *string;
}               stringvars[] =

{
#ifdef CONSOLENAME
	"CONSOLE", CONSOLENAME, /* "/dev/tty" or "Dev:Console" */
#endif /*CONSOLENAME*/
	(char *) 0, (char *) 0
};

static struct
{/* built in functions */
	char           *name;
    Symbolhandle  (*function) ();
}               builtins[] =

{
	"cat",          concat,
	"macro",        macro,
	"makestr",      makestr,
	"strconcat",    makestr,     /*kb*/
	"compnames",    compnames,   /*kb*/
	"varnames",     varnames, /*kb*/
	"modelvars",    modelvars, /*kb*/
	"modelinfo",    xvariables, /*kb*/
	"xvariables",   xvariables, /*kb*/
	"anova",        glm,
	"factor",       makefactor,  /*kb*/
	"match",        makefactor,  /*kb*/
	"unique",       makefactor,  /*kb*/
	"matrix",       matrix,
	"delete",       deleter,
	"list",         listbrief,
	"listbrief",    listbrief,
	"hconcat",      hconcat,     /*kb*/
	"vconcat",      hconcat,     /*kb*/
	"paste",        paste,       /*kb*/
	"shell",        shell,       /*kb*/
	"array",        array,
	"print",        print,
	"fprint",       print,       /*kb*/
	"matprint",     print,       /*kb*/
	"write",        print,       /*kb*/
	"fwrite",       print,       /*kb*/
	"matwrite",     print,       /*kb*/
	"macrowrite",   print,       /*kb*/
	"error",        print,       /*kb*/
	"putascii",     putascii,    /*kb*/
	"contrast",     contrast,
	"anovacoefs",   anovacoefs,
	"log",          Transform,
	"log10",        Transform,
	"exp",          Transform,
	"abs",          Transform,
	"cellstats",    cellstats,
	"describe",     describe,
	"halfnorm",     sort,        /*kb*/
	"rankits",      sort,
	"regress",      glm,
	"setoptions",   setoptions,
	"getoptions",   setoptions,
	"sort",         sort,
	"split",        split,
	"tval",         tval,
	"t2val",        t2val,
	"tint",         tval,
	"t2int",        t2val,
	"bin",          bin,
	"stemleaf",     stemleaf,
	"regpred",      regpred,
	"yates",        yates,
	"rational",     rational,    /*kb*/
	"partacf",      partacf,     /*kb*/
	"yulewalker",   partacf,     /*kb*/
	"movavg",       movavg,      /*kb*/
	"autoreg",      movavg,      /*kb*/
	"toeplitz",     toeplitz,      /*kb*/
	"polyroot",     polyroot,      /*kb*/
	"rft",          fft,         /*kb*/
	"hft",          fft,         /*kb*/
	"cft",          fft,         /*kb*/
	"hconj",        tserops,     /*kb*/
	"cconj",        tserops,     /*kb*/
	"hreal",        tserops,     /*kb*/
	"creal",        tserops,     /*kb*/
	"himag",        tserops,     /*kb*/
	"cimag",        tserops,     /*kb*/
	"hcopyc",       tserops,     /*kb*/
	"ccopyh",       tserops,     /*kb*/
	"htoc",         tserops,     /*kb*/
	"ctoh",         tserops,     /*kb*/
	"hprdh",        tserops,     /*kb*/
	"hprdhj",       tserops,     /*kb*/
	"cprdc",        tserops,     /*kb*/
	"cprdcj",       tserops,     /*kb*/
	"cmplx",        tserops,     /*kb*/
	"hpolar",       tserops,     /*kb*/
	"cpolar",       tserops,     /*kb*/
	"hrect",        tserops,     /*kb*/
	"crect",        tserops,     /*kb*/
	"unwind",       tserops,     /*kb*/
	"convolve",     tserops,     /*kb*/
	"rotate",       tserops,     /*kb*/
	"rotation",     rotatefac,   /*kb*/
	"reverse",      tserops,     /*kb*/
	"padto",        tserops,     /*kb*/
	"sin",          Transform,
	"cos",          Transform,
	"atan",         Transform,
	"atan2",        Transform,   /*kb*/
	"hypot",        Transform,   /*kb*/
	"sqrt",         Transform,
	"lgamma",       Transform,
	"asin",         Transform,
	"acos",         Transform,
	"tan",          Transform,
	"sinh",         Transform,
	"cosh",         Transform,
	"tanh",         Transform,
	"atanh",        Transform,
	"floor",        Transform,
	"ceiling",      Transform,
	"round",        Transform,
	"nbits",        Transform,
	"predtable",    predtable,
	"rep",          rep,
	"run",          run,
	"fastanova",    glm,
	"screen",       glm,
	"runi",         rangen,
	"setseeds",     setseed,
	"getseeds",     getseed,
	"rnorm",        rangen,
	"ismissing",    sort,
	"anymissing",   sort,        /*kb*/
	"isreal",       iswhat,      /*kb*/
	"ischar",       iswhat,      /*kb*/
	"islogic",      iswhat,      /*kb*/
	"isvector",     iswhat,      /*kb*/
	"ismatrix",     iswhat,      /*kb*/
	"isscalar",     iswhat,      /*kb*/
	"isfactor",     iswhat,      /*kb*/
	"isnull",       iswhat,      /*kb*/
	"isdefined",    iswhat,      /*kb*/
	"isstruc",      iswhat,      /*kb*/
	"ismacro",      iswhat,      /*kb*/
	"isgraph",      iswhat,      /*kb*/
	"nameof",       nameOf,      /*kb*/
	"wtregress",    glm,
	"wtanova",      glm,
	"manova",       glm,
	"wtmanova",     glm,
	"grade",        sort,        /*kb*/
	"rank",         sort,
	"sum",          columnops,   /*kb*/
	"prod",         columnops,   /*kb*/
	"max",          columnops,   /*kb*/
	"min",          columnops,   /*kb*/
	"dim",          dim,         /*kb*/
	"nrows",        dim,         /*kb*/
	"ncols",        dim,         /*kb*/
	"ndims",        dim,         /*kb*/
	"ncomps",       dim,         /*kb*/
	"length",       dim,         /*kb*/
	"diag",         diag,        /*kb*/
	"dmat",         diag,        /*kb*/
	"swp",          swp,         /*kb*/
	"bcprd",        swp,         /*kb*/
	"det",          solve,       /*kb*/
	"solve",        solve,       /*kb*/
	"cholesky",     cholesky,    /*kb*/
	"trilower",     qr,          /*kb*/
	"triupper",     qr,          /*kb*/
	"triunpack",    qr,          /*kb*/
	"qr",           qr,          /*kb*/
	"regsqr",       qr,          /*kb*/
	"robust",       glm,
	"poisson",      glm,
	"ipf",          glm,
	"tabs",         tabs,
	"logistic",     glm,
	"secoefs",      anovacoefs,
	"coefs",        anovacoefs,
	"cor",          cor,
	"trace",        eigen,
	"plot",         myplot,
	"fplot",        myplot,
	"addpoints",    myplot,
	"chplot",       myplot,
	"fchplot",      myplot,
	"addchars",     myplot,
	"lineplot",     myplot,
	"flineplot",    myplot,
	"addlines",     myplot,
	"addstrings",   myplot,
	"showplot",     myplot,
	"eigen",        eigen,
	"eigenvals",    eigen,
	"releigen",     eigen,
	"releigenvals", eigen,       /*kb*/
	"trideigen",	trideigen,   /*kb*/
	"svd",          svd,         /*kb*/
	"cluster",      cluster,     /*kb*/
	"changestr",    changestr,   /*kb*/
	"t",            transpose,
	"outer",        outer,
	"boxplot",      boxplot,
	"fboxplot",     boxplot,
	"save",         save,
	"asciisave",    save,
	"restore",      restore,
	"batch",        batch,
	"cumnor",       cumcdf,
	"invnor",       cumcdf,      /*kb*/
	"cumchi",       cumcdf,
	"invchi",       cumcdf,      /*kb*/
	"cumstu",       cumcdf,
	"invstu",       cumcdf,      /*kb*/
	"cumF",         cumcdf,
	"invF",         cumcdf,
	"cumgamma",     cumcdf,      /*kb*/
	"invgamma",     cumcdf,      /*kb*/
	"cumbeta",      cumcdf,      /*kb*/
	"invbeta",      cumcdf,      /*kb*/
	"cumbin",       cumcdf,
	"cumpoi",       cumcdf,
	"vecread",      readdata,    /*kb*/
	"matread",      readdata,    /*kb*/
	"macroread",    readdata,    /*kb*/
	"inforead",     readdata,    /*kb*/
	"spool",        spool,
	"power",        power1,
	"power2",       power1,
	"samplesize",   samplesize,
	"Help",         help,
	"help",         help,
	(char *) 0,     (Symbolhandle (*) ()) 0
};


extern Symbolhandle         Install(), RInstall(), Makesymbol();
extern Symbolhandle         Makestruc();
extern Symbolhandle         batch();
extern char               **mygethandle();
extern void                 clearGlobals();

char     *StdinName = "standard input";

#ifdef MACINTOSH
static void      cleanStrings()
{
	int          i;
	char        *pc;

	/* change \n to \r in macros and strings */
	for (i = 0; stringvars[i].name; i++)
	{
		for (pc = stringvars[i].string; *pc; pc++)
		{
			if (*pc == LF)
			{
				*pc = CR;
			}
		}
	} /*for (i = 0; stringvars[i].name; i++)*/
} /*cleanStrings()*/
#endif /*MACINTOSH*/

#ifndef DATAPATHNAME
#ifdef MACINTOSH
#define DATAPATHNAME ""
#endif /*MACINTOSH*/
#ifdef UNIX
#define DATAPATHNAME "./"
#endif /*UNIX*/
#ifdef MSDOS
#define DATAPATHNAME ".\\"
#endif /*MSDOS*/
#endif /*DATAPATHNAME*/

#ifndef MACROPATHNAME
#define MACROPATHNAME DATAPATHNAME
#endif /*MACROPATHNAME*/

#ifndef HELPPATHNAME
#ifdef MACINTOSH
#define HELPPATHNAME ""
#endif /*MACINTOSH*/
#ifdef UNIX
#define HELPPATHNAME "./"
#endif /*UNIX*/
#ifdef MSDOS
#define HELPPATHNAME ".\\"
#endif /*MSDOS*/
#endif /*HELPPATHNAME*/

#ifndef DATAFILENAME
#ifdef MACINTOSH
#define DATAFILENAME "MacAnova.dat"
#endif /*MACINTOSH*/

#if defined(UNIX) && !defined(MSDOS)
#define DATAFILENAME "macanova.dat"
#endif /*UNIX && !defined(MSDOS)*/

#ifdef MSDOS
#define DATAFILENAME "MACANOVA.DAT"
#endif /*MSDOS*/
#endif /*DATAFILENAME*/

#ifndef MACROFILENAME
#ifdef MACINTOSH
#define MACROFILENAME "MacAnova.mac"
#endif /*MACINTOSH*/

#if UNIX && !defined(MSDOS)
#define MACROFILENAME "macanova.mac"
#endif /*UNIX && !defined(MSDOS)*/

#ifdef MSDOS
#define MACROFILENAME "MACANOVA.MAC"
#endif /*MSDOS*/
#endif /*MACROFILENAME*/

/* use fake handles for builtin functions */
static Symbol             **FunctionPointers; /* array of pointers to builtins */
static FunctionSymbol      *FunctionSymbols; /* array of Function Symbols */
extern long                 Nfunctions;

void initialize(helpFileName, dataFileName, macroFileName) /* install these */
char               *helpFileName, *dataFileName, *macroFileName;
{
	long            i;
	Symbolhandle    symh;
	long            length;
	long            nBuiltins = 0;
	char           *dataPathName, *macroPathName;
#ifdef MSDOS
	char            homePathName[PATHSIZE];
#endif /*MSDOS*/
	char           *pathName;
	double          value;
	WHERE("initialize");
	
#ifdef MSDOS
	macroPathName = dataPathName = get_dataPath();
#else /*MSDOS*/
	dataPathName = DATAPATHNAME;
	macroPathName = MACROPATHNAME;
#endif /*MSDOS*/

	if(!Symbolinit())
	{ /* create size TABLESIZE symbol table and zero it out*/
		goto fatalError;
	}

	if(!glmInit())
	{ /* allocate glm related globals */
		goto fatalError;
	}
	
	/* create fake handles for function symbols */
	for (i = 0; builtins[i].name; i++)
	{ /* count builtins */
		;
	}
	nBuiltins = i;
	FunctionPointers = (Symbol **) mygetpointer(nBuiltins * sizeof(Symbol *));
	FunctionSymbols = (FunctionSymbol *) mygetpointer(nBuiltins*
													  sizeof(FunctionSymbol));
	if(FunctionPointers == (Symbol **) 0 ||
	   FunctionSymbols == (FunctionSymbol *) 0)
	{
		goto fatalError;
	}
	
	for (i = 0;i<nBuiltins; i++)
	{ /* make fake Symbolhandle for each builtin function */
		symh = FunctionPointers + i;
		*symh = (Symbol *) (FunctionSymbols+i);
		setTYPE(symh,BLTIN);
		setNCLASS(symh,-32768); /* indicates it's fake if we ever need to know */
		setNDIMS(symh,1);
		setDIM(symh,1,1);
		setPREV(symh,(Symbolhandle) 0);
		setNEXT(symh,(Symbolhandle) 0);
		setFPTR(symh,builtins[i].function);
		setNAME(symh,builtins[i].name);
		Addsymbol(symh);
	}

	/* now install pre-defined constants */
	for (i = 0; constants[i].name; i++)
	{
		symh = RInstall(constants[i].name, 1L);
		if(symh == (Symbolhandle) 0)
		{
			goto fatalError;
		}
		if(strcmp(NAME(symh),"PI") == 0)
		{
			value = 4.0*atan(1.0);
		}
		else if(strcmp(NAME(symh),"E") == 0)
		{
			value = exp(1.0);
		}
		else if(strcmp(NAME(symh),"DEGPERRAD") == 0)
		{
			value = 45.0/atan(1.0);
		}
		else
		{
			value = constants[i].value;
		}
		DATAVALUE(symh,0) = value;
	}

	/* 
	  Create NULLSYMBOL (value returned by print(), etc.
	  Note that it is *not* installed in the symbol table
	*/
	NULLSYMBOL = Makereal(0);
	if(NULLSYMBOL == (Symbolhandle) 0)
	{
		goto fatalError;
	}
	setNAME(NULLSYMBOL, "NULLSYMBOL");

#ifdef MACINTOSH
	cleanStrings();
#endif /*MACINTOSH*/

	/* install pre-defined macros */
	if (!iniMacros())
	{
		goto fatalError;
	}
	
	/* install pre-defined string variables (they must all be scalar) */
	for (i = 0; stringvars[i].name; i++)
	{
		symh = Install(stringvars[i].name, CHAR);
		TMPHANDLE = mygethandle((long) strlen(stringvars[i].string) + 1);
		if(symh == (Symbolhandle) 0 || TMPHANDLE == (char **) 0)
		{
			goto fatalError;
		}
		setSTRING(symh,TMPHANDLE);
		setNDIMS(symh,1);
		setDIM(symh,1,1);
		strcpy(*TMPHANDLE,stringvars[i].string);
	}

#ifdef VERSION
	symh = Install("VERSION", CHAR);
	sprintf(OUTSTR,"MacAnova %s ",VERSION);
#ifdef TODAY
	strcat(OUTSTR,TODAY);
#endif /*TODAY*/
	TMPHANDLE = mygethandle((long) strlen(OUTSTR) + 1);
	if(symh == (Symbolhandle) 0 || TMPHANDLE == (char **) 0)
	{
		goto fatalError;
	}
	setSTRING(symh,TMPHANDLE);
	setNDIMS(symh,1);
	setDIM(symh,1,1);
	strcpy(*TMPHANDLE,OUTSTR);
	*OUTSTR = '\0';
	VERSION_ID = myduphandle(TMPHANDLE);
	if(VERSION_ID == (char **) 0)
	{
		goto fatalError;
	}
#endif /*VERSION*/

	if(helpFileName == (char *) 0)
	{
		helpFileName = HELPNAME;
#ifdef MSDOS
		pathName = get_dataPath();
#else /*MSDOS*/
		pathName = HELPPATHNAME;
#endif /*MSDOS*/
	}
	else
	{
		pathName = "";
	}
	length = strlen(pathName) + strlen(helpFileName);
	HELPFILENAME = mygethandle(length + 1);
	if(HELPFILENAME == (char **) 0)
	{
		goto fatalError;
	}
	strcpy(*HELPFILENAME,pathName);
	strcat(*HELPFILENAME,helpFileName);

	length = strlen(dataPathName);
	TMPHANDLE = mygethandle(length + 1);
	symh = Install("DATAPATH",CHAR);
	if(TMPHANDLE == (char **) 0 || symh == (Symbolhandle) 0)
	{
		goto fatalError;
	}
	setSTRING(symh,TMPHANDLE);
	setNDIMS(symh,1);
	setDIM(symh,1,1);
	strcpy(*TMPHANDLE,dataPathName);

#if defined(UNIX) || defined(MSDOS)
	/* pre-install CHARACTER variable HOME */
#ifdef MSDOS
	strcpy(homePathName, dataPathName);
	length = strlen(homePathName);
	if(homePathName[length] == '\\' || homePathName[length] == '/')
	{
		homePathName[length--] = '\0';
	}
	pathName = homePathName;
#elif defined(UNIX)
	pathName = getenv("HOME");
	length = (pathName != (char *) 0) ? strlen(pathName) : 0;
#endif

	if(length > 0)
	{
		TMPHANDLE = mygethandle(length + 1);
		symh = Install("HOME",CHAR);
		if(TMPHANDLE == (char **) 0 || symh == (Symbolhandle) 0)
		{
			goto fatalError;
		}
		setSTRING(symh,TMPHANDLE);
		setNDIMS(symh,1);
		setDIM(symh,1,1);
		strcpy(*TMPHANDLE,pathName);
	}
#endif /* defined(UNIX) || defined(MSDOS)*/
	
	if(dataFileName == (char *) 0)
	{
		dataFileName = DATAFILENAME;
		pathName = dataPathName;
	}
	else
	{
		pathName = "";
	}
	length = strlen(pathName) + strlen(dataFileName);
	TMPHANDLE = mygethandle(length + 1);
	symh = Install("DATAFILE",CHAR);
	if(TMPHANDLE == (char **) 0 || symh == (Symbolhandle) 0)
	{
		goto fatalError;
	}
	setSTRING(symh,TMPHANDLE);
	setNDIMS(symh,1);
	setDIM(symh,1,1);
	strcpy(*TMPHANDLE,pathName);
	strcat(*TMPHANDLE,dataFileName);

	if(macroFileName == (char *) 0)
	{
		macroFileName = MACROFILENAME;
		pathName = macroPathName;
	}
	else
	{
		pathName = "";
	}
	length = strlen(pathName) + strlen(macroFileName);
	TMPHANDLE = mygethandle(length + 1);
	symh = Install("MACROFILE",CHAR);
	if(TMPHANDLE == (char **) 0 || symh == (Symbolhandle) 0)
	{
		goto fatalError;
	}
	setSTRING(symh,TMPHANDLE);
	setNDIMS(symh,1);
	setDIM(symh,1,1);
	strcpy(*TMPHANDLE,pathName);
	strcat(*TMPHANDLE,macroFileName);
	
	for(i = 0;i < MAXWDEPTH;i++)
	{
		WHILELIMITS[i] = MAXWHILE;
		FORVECTORS[i] = (double **) 0;
	}

	clearGlobals();
	
	saveFormat(); /* save copy of default format */
	INPUTSTRING = (unsigned char **) 0;
	ISTRCHAR = 0;

	INPUTFILE[0] = STDIN;
	INPUTFILENAMES[0] = &StdinName;

	for(i=1;i<MAXBDEPTH;i++)
	{
		INPUTFILE[i] = (FILE *) 0;
		INPUTFILENAMES[i] = (char **) 0;
	}

	TMPHANDLE = mygethandle(1L);
	if(TMPHANDLE == (char **) 0)
	{
		goto fatalError;
	}
	INPUTSTRING = (unsigned char **) TMPHANDLE;
	(*INPUTSTRING)[0] = '\0';

	GUBED = 0; /* this is here to make it easy to change for debugging */

#ifdef DJGPP
	INTERRUPT = INTSET;
	(void) interrupted(); /* to get DJGPP started */
	INTERRUPT = INTNOTSET;
#endif /*DJGPP*/

	return ;

  fatalError:
	PUTALERT("FATAL ERROR: unable to allocate space during inialization sequence");
	exit(-1);
	
}



