/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Ipf
#else
#define __SEG__ Ipf
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

#define residuals(i) ((*RESIDUALS)[i])

extern void            getterm();
extern double          cellstat();
extern long            buildXmatrix();

/*

 	routine for computing ss and df for balanced designs

	RESIDUALS	handle to response vector (which will be changed)
	residuals	local define for residuals
	X		array of handles to class variables
	NVARS	numberof variables (columns of x)
	NCLASSES	number of classes in each variable
	NDATA	number of data values
	MODEL	handle giving terms in order to be used in model
		term i contains the variables in the binary expansion of i
		0 is constant; 3 is AB; 5 is AC etc
	SS	handle to term sums of squares
	DF	handle of term df
*/

#define GMEANS   0
#define GFITS    0
#define GOMARG   1
#define GFMARG   2
#define NTRASH   3

long ipf(maxiter,epsilon)
long              maxiter;
double            epsilon;
{
	long            i, j, k, term[MAXVARS];
	modelPointer    iterm;
	double          curdev, newdev, olddev, curss, newss;
	double        **means = (double **) 0;
	double        **fits = (double **) 0, **omarg = (double **) 0;
	double        **fmarg = (double **) 0;
	double          yi;
	long            istart, ii, icell;
	long            ndata = (long) NDATA, nterms = (long) NTERMS;
	long            nvars = (long) NVARS;
	long            nfactors = (long) NFACTORS, ncells, space = 0;
	long            iter;
	long            returnValue = 0;
	WHERE("ipf");
	TRASH(NTRASH,errorExit);
	
	/* first compute maximum possible df's */

	SETUPINT(errorExit);
	
	k = 0;			/* total of cols before this term */

	for (i = 0; i < nterms; i++)
	{			/* for each term */
		(*DF)[i] = (double) buildXmatrix(i, (double *) 0, MODELINFO);
	}

	(*DF)[nterms] = ndata;
	for (i = 0; i < nterms; i++)
	{
		(*DF)[nterms] -= (*DF)[i];
	}

	ncells = 1;
	
	for(i = 0; i < nfactors; i++)
	{
		ncells *= NCLASSES[i];
	}
	
	/* zero out SS */
	doubleFill(*SS, 0.0, nterms+1);

	/* now go through terms in model getting residual ss for each term
	   and sweeping out effects */

	
   	if (MODELTYPE & BALANOVA)
	{
		/* find maximum marginal table size */
		for(i=0;i<nterms;i++)
		{
			getterm(modelterm(MODEL,i), term, nvars);
			ncells = setterm(term, nvars, NCLASSES);
			space = (ncells > space) ? ncells : space;
		}
		
		if(!getScratch(means,GMEANS,space,double))
		{
			goto errorExit;
		}

		curss = 0.0;
		for (i = 0; i < ndata; i++)
		{
			curss += residuals(i) * residuals(i);	/*  get total SS */
		}
		
		for (i = 0; i < nterms; i++)
		{
			if (interrupted() != INTNOTSET)
			{
				goto errorExit;
			}
			 /* get means and sweep out term*/
			getterm(modelterm(MODEL,i), term, nvars);
			newss = cellstat(RESIDUALS, term, means,
							 (double **) 0, (double **) 0, CELLMEANSANDSWEEP);
			/* set ss for term */
			(*SS)[i] = (curss > newss) ? curss - newss : 0.0;
			curss = newss;
		} /*for (i = 0; i < nterms; i++)*/

		(*SS)[nterms] = ((*DF)[nterms] > 0 ) ? curss : 0.0;	/* error ss */

	} /*if (MODELTYPE & BALANOVA)*/
	else
	{			/* IPF */
		if(!getScratch(fits,GFITS,ndata,double) ||
		   !getScratch(omarg,GOMARG,ndata,double) ||
		   !getScratch(fmarg,GFMARG,ndata,double))
		{
			goto errorExit;
		}

		curdev = 0.0;
		for (i = 0; i < ndata; i++)
		{
			yi = (*Y)[i];
			if (yi > 0.0)
			{
				curdev += yi * log(yi) - yi + 1.0;
			}
			else
			{
				curdev += 1.0;
			}
		}
		curdev = 2.0 * curdev;
		for (i = 0; i < ndata; i++)
		{
			(*fits)[i] = 1.0;
		}
		
		if (INCREMENTAL)
		{
			istart = 1;
		}
		else
		{
			istart = nterms;
		}
		for (ii = istart; ii <= nterms; ii++)
		{
			newdev = 0.;
			olddev = -1.;
			iter = 0;
			while (fabs(olddev - newdev) / fabs(olddev + newdev) > epsilon)
			{
				if(++iter > maxiter)
				{
					returnValue = 1;
					break;
				}
				
				olddev = newdev;
				for (i = 0; i < ii; i++)
				{
					if (interrupted() != INTNOTSET)
					{
						goto errorExit;
					}
					iterm = modelterm(MODEL,i);
					getterm(iterm, term, nvars);
					/*get marginal sums of data for term*/
					(void) cellstat(Y, term, omarg,
									(double **) 0, (double **) 0, CELLSUMS);
					/* get marginal sums of current fits*/
					getterm(iterm, term, nvars);
					(void) cellstat(fits, term, fmarg,
									(double **) 0, (double **) 0, CELLSUMS);
					/* note cellstat() modifies term to contain offsets */
					for (j = 0; j < ndata; j++)
					{
						icell = 0;
						for (k = 0; k < nvars; k++)
						{
							icell += term[k] * ((*X[k])[j] - 1);
						}
						if ((*fmarg)[icell] > 0.)
						{ /* adjust fits */
							(*fits)[j] *= (*omarg)[icell] / (*fmarg)[icell];
						}
						else
						{
							(*fits)[j] = 0.;
						}
					}
				}
				newdev = 0.;
				for (i = 0; i < ndata; i++)
				{
					yi = (*Y)[i];
					if (yi == 0.)
					{
						newdev += (*fits)[i];
					}
					else
					{
						newdev += yi * log(yi/(*fits)[i]) -
							yi + (*fits)[i];
					}
				}
				newdev *= 2.0;
			} /*while (fabs(olddev - newdev) / fabs(olddev + newdev) > epsilon)*/
			(*SS)[ii - 1] = curdev - newdev;
			curdev = newdev;
		}
		(*SS)[nterms] = curdev;

		for (i = 0; i < ndata; i++)
		{
			yi = (*fits)[i];
			(*RESIDUALS)[i] = (*Y)[i] - yi;
			if (yi > 0.)
			{
				(*WTDRESIDUALS)[i] = (*RESIDUALS)[i] / sqrt(yi);
			}
			else
			{
				(*WTDRESIDUALS)[i] = 0.;
			}
		}
		/* dispose of workspace */
	}

	/* compute HII */
	curss = (NDATA - (*DF)[nterms]) / NDATA;
	for (i = 0; i < ndata; i++)
	{
		(*HII)[i] = curss;
	}

	returnValue = (returnValue != 0) ? returnValue : 2;
	/* fall through */

  errorExit:
	emptyTrash(); /* cleanup scratch */
	RETURN (returnValue);

}

