/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Iswhat
#else
#define __SEG__ Iswhat
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

/*
  operations isreal(), ischar(), islogic(), isvector(), ismatrix(),
  isscalar(), isfactor(), isnull(), ismacro()
  Each returns a logical vector of length NARGS(list)
*/
Symbolhandle iswhat(list)
Symbolhandle           list;
{
	long               nargs = NARGS(list);
	long               i, type, value;
	Symbolhandle       symh, result = (Symbolhandle) 0;
	long               dims[2];
	char              *what = FUNCNAME + 2;
	WHERE("iswhat");

	result = Makereal(nargs);
	if(result == (Symbolhandle) 0)
	{
		return (0);
	}
	setTYPE(result,LOGIC);
	
	for(i=0;i<nargs;i++)
	{
		symh = COMPVALUE(list,i);
		type = (symh != (Symbolhandle) 0) ? TYPE(symh) : 0;
		
		if(strcmp(what,"real") == 0)
		{
			value = (type == REAL);
		}
		else if(strcmp(what,"char") == 0)
		{
			value = (type == CHAR);
		}
		else if(strcmp(what,"logic") == 0)
		{
			value = (type == LOGIC);
		}
		else if(strcmp(what,"struc") == 0)
		{
			value = (type == STRUC);
		}
		else if(strcmp(what,"macro") == 0)
		{
			value = (type == MACRO);
		}
		else if(strcmp(what,"graph") == 0)
		{
			value = (type == PLOTINFO);
		}
		else if(strcmp(what,"vector") == 0)
		{
			value = (isVector(symh) != 0);
		}
		else if(strcmp(what,"scalar") == 0)
		{
			value = (isScalar(symh) != 0);
		}
		else if(strcmp(what,"matrix") == 0)
		{
			value = (isMatrix(symh,dims) != 0);
		}
		else if(strcmp(what,"factor") == 0)
		{
			value = (isFactor(symh) > 0);
		}
		else if(strcmp(what,"null") == 0)
		{
			value = (isNull(symh) != 0);
		}
		else if(strcmp(what,"defined") == 0)
		{
			value = (isDefined(symh) != 0);
		}
		else
		{
			sprintf(OUTSTR,"ERROR: ****programming error for operation %s",
					   FUNCNAME);
			Delete(result);
			return (0);
		}
		DATAVALUE(result,i) = (double) value;
	} /*for(i=0;i<nargs;i++)*/
	Addsymbol(result);
	
	return (result);
}

#ifdef UNDEFINED__ /* code is unfinished */
/*
   usage:
     checkarg("$1",option1 [,option2,...])
	 where options are
	    string:T or F    argument is or is not apparently quoted string
		variable:T or F  argument is or is not variable name
		tempvar:T or F   argument is or is not temporary, variable name
		keyword:T or F   argument is or is not keyword phrase
		expression:T or F argument is or is not apparent expression
		TorF:T or F      argument is either T or F
		number:T or F    argument is a number

   It returns a LOGICAL scalar which is the anding of all the conditions.
   Thus
       checkarg("\"quoted string\"",string:T,keyword:T)
   is False because it is not a keyword phrase
       checkarg("\"quoted string\"",string:T,keyword:F)
   is True because it is both a quoted string and not a keyword phrase
       checkarg("a:\"quoted string\"",string:T,keyword:T)
   is True because it is both a quoted string and a keyword phrase
   	   checkarg("@x",variable:T,tempvar:T)
   is True because '@x' is a temporary variable name
   	   checkarg("x",variable:T,tempvar:F)
   is False because 'x' is not a temporary variable name
*/

Symbolhandle checkarg(list)
Symbolhandle      list;
{
	Symbolhandle        result = (Symbolhandle) 0, symhKey;
	Symbolhandle        arg1;
	long                nargs = NARGS(list);
	long                i, retval = 1;
	char               *keyword;
	
	*OUTSTR = '\0';
	
	if(nargs <= 1)
	{
		badNargs(FUNCNAME,-1002);
		goto errorExit;
	}
	
	arg1 = COMPVALUE(list,0);
	if(!argOK(arg,0,1))
	{
		goto errorExit;
	}
	if(TYPE(arg1) != CHAR || !isscalar(arg1))
	{
		sprintf(OUTSTR,
				"ERROR: 1st argument to %s must be quoted string or CHARACTER scalar",
				FUNCNAME);
		goto errorExit;
	}
	
	for(i=0,i<nargs;i++)
	{
		symhKey = COMPVALUE(list,i);
		if(!argOK(symhKey,0,(nargs > 1) ? i+1 : 0))
		{
			goto errorExit;
		}
		if(!(keyword = isKeyword(symhKey)) )
		{
			sprintf(OUTSTR,
					"ERROR: %ld-th argument to %s is not keyword phrase",
					i+1, FUNCNAME);
			goto errorExit;
		}
	}
  errorExit:
	putOUTSTR();
	return (0);
}
#endif /*UNDEFINED__*/
	
/*
  vec <- nameof(arg1, arg2, ..., argk) returns vector of argument names
*/

Symbolhandle nameOf(list)
Symbolhandle          list;
{
	Symbolhandle      arg, result = (Symbolhandle) 0;
	long              nargs = NARGS(list), i;
	long              needed = 0, place = 0;
	char             *name;
	WHERE("nameOf");
	
	for(i = 0;i<nargs;i++)
	{
		if((arg = COMPVALUE(list,i)) == (Symbolhandle) 0)
		{
			noData(FUNCNAME,i+1);
			goto errorExit;
		}
		name = NAME(arg);
		needed += strlen(name) + 1;
		if(isscratch(name))
		{
			needed -= 2;
		}
	}
	
	if((result = Install(SCRATCH,CHAR)) == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	TMPHANDLE = mygethandle(needed);
	setSTRING(result,TMPHANDLE);
	if(TMPHANDLE == (char **) 0)
	{
		goto errorExit;
	}
	for(i = 0;i < nargs;i++)
	{
		name = NAME(COMPVALUE(list,i));
		if(isscratch(name))
		{
			name += 2;
		}
		strcpy(*TMPHANDLE + place,name);
		place += strlen(name) + 1;
	}
	setNDIMS(result,1);
	setDIM(result,1,nargs);

	return (result);

  errorExit:
	Removesymbol(result);
	return (0);
	
}

