/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Iterglm
#else
#define __SEG__ Iterglm
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <math.h>
#include "dbug.h"

#define xtx(I,J) ((*XTXINV)[ (I) + (J) *  nregx ])
#define x(I,J) ((*REGX)[ (I) + (J) * ndata ])

#define newresidH     RESIDUALS

extern double   CURRENTRSC;

/*
  routine for computing ss and df for unbalanced designs

  RESIDUALS	handle to response vector (which will be changed)
  newresH defined to be RESIDUALS
  newres[i] (*RESIDUALS)[i]
  X		array of handles to class variables
  NVARS	numberof variables (columns of x)
  NCLASSES	number of classes in each variable
  NDATA	number of data values
  MODEL	handle giving terms in order to be used in model
	  term i contains the variables in the binary expansion of i
	  0 is constant; 3 is AB; 5 is AC etc
  SS	handle to term sums of squares
  DF	handle of term df
*/


long iterglm(maxiter,epsilon,silent)
long                maxiter,silent;
double              epsilon;
{
	long            i, icurr, j, k, k2, jk, ifit;
	long            ndata = (long) NDATA;
	long            nregx = (long) NREGX, nterms = (long) NTERMS;
	long            nfits = 1, dfi, maxdfi;
	long            ny = (long) NY;
	long            robustreg = (MODELTYPE & ROBUSTREG);
	long            doCoefs = MODELTYPE & DOCOEFS;
	long            doRinv = (doCoefs) ? DORINV : DONTDORINV;
	double         *newresid;
	double          weight, xval;
	double          curdev, olddev, newdev;
	double         *regx, *xik, *regx2, *allwts, *misswts, *casewts;
	double         *regss, *df, *hii;
	long            returnValue = 2;
	double          (*devfn) (long);
	void            (*residfn) (void), (*varfn) (long), (*linkfn) (long);
	char           *chPtr1, *chPtr2;
	WHERE("iterglm");
	
	SETUPINT(errorExit);
	
	ALLWTS = (double **) mygethandle(ndata * sizeof(double));
	if (ALLWTS == (double **) 0)
	{
		goto errorExit;
	}

	newresid = *newresidH;
	
	if (robustreg)
	{
		varfn = robustvar;
		linkfn = robustlink;
		devfn = robustdev;
		residfn = robustresid;

		regx = *REGX;
		if(MODELTYPE & MISSWEIGHTS)
		{
			misswts = *MISSWTS;
		}
		if(MODELTYPE & CASEWEIGHTS)
		{
			casewts = *CASEWTS;
		}
		
		for(i=0;i<ndata;i++)
		{
			weight = 1.0;
			if (MODELTYPE & MISSWEIGHTS)
			{
				weight *= misswts[i];
			}
			if (MODELTYPE & CASEWEIGHTS)
			{
				weight *= sqrt(casewts[i]);
			}

			xik = regx++; /* *xik is x(i,k) */
			for (k = 0; k < nregx; k++)
			{
				*xik *= weight;
				xik += ndata;
			}
		} /*for(i=0;i<ndata;i++)*/

		/*
		  Do Gramschmidt for full model once and for all with DOSWEEP
		  Use DONTDOSWEEP in later calls to gramschmidt()
		*/
		if(doCoefs)
		{
			xtxinvInit();
		}
		
		gramschmidt(DONTNOTIFY | DOSWEEP | DONTDOSS | doRinv |
					DONTDOCOEFS);
		if(INTERRUPT != INTNOTSET)
		{
			goto errorExit;
		}		
	} /*if (robustreg)*/
	else if (MODELTYPE & LOGITREG)
	{
		varfn = logitvar;
		linkfn = logitlink;
		devfn = logitdev;
		residfn = logitresid;
	} /*if (robustreg){}else if (MODELTYPE & LOGITREG){}*/
	else if (MODELTYPE & POISSONREG)
	{
		varfn = poissonvar;
		linkfn = poissonlink;
		devfn = poissondev;
		residfn = poissonresid;
	}

	/* zero out SS */
	doubleFill(*SS, 0.0, nterms + 1);

	if (INCREMENTAL)
	{
		nregx = 0;
		nfits = nterms;
	}

	df = *DF;
	for (ifit = 0; ifit < nfits; ifit++)
	{
		if(INCREMENTAL)
		{
			NREGX = (double) (nregx += df[ifit]);
		}
			
		olddev = -10.0e30;
#ifdef UNDEFINED__
		if(!robustreg)
		{
			xtxinvInit();
		}
#endif /*UNDEFINED__*/
		(*linkfn) (INITIAL);
		if(INTERRUPT != INTNOTSET)
		{
			goto errorExit;
		}		
		(*varfn) (INITIAL);
		if(ifit == 0)
		{
			curdev = (*devfn) (INITIAL);
		}
			
		allwts = *ALLWTS;
		regx = *REGX;
		if(!robustreg)
		{
			regx2 = *REGX2;
		}
		
		for (k = 0; k < maxiter; k++)
		{
			for (j = 0; j < ndata; j++)
			{
				weight = allwts[j];
				if(!robustreg)
				{
					jk = j;		/* j + k2*ndata */
					for (k2 = 0; k2 < nregx; k2++)
					{
						regx[jk] = weight*regx2[jk];
						jk += ndata;
					}
				}
					
				jk = j;			/* j + k2*ndata */
				for (k2 = 0; k2 < ny; k2++)
				{
					newresid[jk] *= weight;
					jk += ndata;
				}
			} /*for (j = 0; j < ndata; j++)*/

#ifdef UNDEFINED__
			/* now setup xtxinv */
			if(!robustreg)
			{
				xtxinvInit();
			}
#endif /*UNDEFINED__*/
			gramschmidt(DONTNOTIFY | ((robustreg) ? DONTDOSWEEP : DOSWEEP) |
						DONTDOSS | DONTDORINV | DONTDOCOEFS);
			
			if (INTERRUPT != INTNOTSET)
			{
				goto errorExit;
			}
			newdev = (*devfn) (STEP);
			if (fabs(olddev - newdev) <= epsilon * fabs(olddev + newdev))
			{
				break;
			}
			olddev = newdev;
			if (k < maxiter - 1)
			{
				(*linkfn) (STEP);
				if(INTERRUPT != INTNOTSET)
				{
					goto errorExit;
				}		
				(*varfn) (STEP);
			}
		} /*for (k = 0; k < maxiter; k++)*/

		if(k == maxiter)
		{
			if(INCREMENTAL && ifit < nterms-1)
			{
				sprintf(OUTSTR,
						"WARNING: term %ld not fully converged in %ld iterations",
						ifit + 1, maxiter);
			}
			else
			{
				sprintf(OUTSTR,
						"WARNING: final deviance not fully converged in %ld iterations",
						maxiter);
			}
			
			putOUTSTR();
			returnValue = 1;
		} /*if(k == maxiter)*/
			
		if(INCREMENTAL)
		{
			(*SS)[ifit] = curdev - newdev;	/* set dev for term */
			curdev = newdev;
		}
		else
		{
			(*SS)[nterms-1] = curdev - newdev;
		}
	} /* for(ifit=0;ifit<nfits;ifit++)*/
	(*SS)[nterms] = newdev;


	/* now go all the way through and get final coefficients */

	/* now setup xtxinv */

	if(!robustreg && doCoefs)
	{
		xtxinvInit();
	}


	(*linkfn) (STEP);
	if(INTERRUPT != INTNOTSET)
	{
		goto errorExit;
	}		
	(*varfn) (STEP);

	regx = *REGX;
	if(!robustreg)
	{
		regx2 = *REGX2;
	}
	
	allwts = *ALLWTS;
	for (j = 0; j < ndata; j++)
	{
		weight = allwts[j];
		if(!robustreg)
		{
			jk = j;				/* j + k2*ndata */
			for (k2 = 0; k2 < nregx; k2++)
			{
				regx[jk] = weight*regx2[jk];
				jk += ndata;
			}
		}
		
		jk = j; /* j + k2*ndata */
		for (k2 = 0; k2 < ny; k2++)
		{
			newresid[jk] *= weight;
			jk += ndata;
		}
	}

	gramschmidt(DONTNOTIFY | ((robustreg) ? DONTDOSWEEP : DOSWEEP) |
				DONTDOSS | ((robustreg) ? DONTDORINV : doRinv) | doCoefs);
	if (INTERRUPT != INTNOTSET)
	{
		goto errorExit;
	}

	(*residfn) ();

	if (robustreg && !silent)
	{
		myeol();
		strcpy(OUTSTR, "Robust estimate of scale: ");
		chPtr1 = chPtr2 = OUTSTR + strlen(OUTSTR);
		sprintf(chPtr1, DATAFMT, CURRENTRSC);
		while(*chPtr2 == ' ')
		{ /* trim off leading blanks */
			chPtr2++;
		}
		strcpy(chPtr1, chPtr2);
		putOUTSTR();
		myeol();
	}

	/* fix up DF */
	regss = *REGSS;
	df = *DF;

	icurr = 0;
	for (i = 0; i < nterms; i++)
	{
		dfi = 0;
		maxdfi = df[i];
		for (j = 0; j < maxdfi; j++)
		{		/* go through the cols for each term */
			if (regss[icurr++] > -0.5)
			{
				dfi++;
			}
		} /*for (j = 0; j < maxdfi; j++)*/
		df[i] = dfi;
	} /*for (i = 0; i < nterms; i++)*/

	curdev = NOTMISSING;
	for (i = 0; i < nterms; i++)
	{
		curdev -= df[i];
	}
	df[nterms] = curdev;
	if(curdev == 0.0)
	{
		(*SS)[nterms] = 0.0;
	}
	
	/* fix up HII */
	regx = *REGX;
	hii = *HII;
	misswts = (MODELTYPE & MISSWEIGHTS) ? *MISSWTS : (double *) 0;
	for (i = 0;i < ndata;i++)
	{
		if (misswts == (double *) 0 || misswts[i] != 0.0)
		{
			hii[i] = 0.0;
		}
		else
		{
			setMissing(hii[i]);
		}
	} /*for (i = 0;i < ndata;i++)*/

	for (j = 0; j < nregx; j++)
	{
		if(regss[j] > -0.5)
		{
			for (i = 0;i < ndata;i++)
			{
				if (!isMissing(hii[i]))
				{
					xval = regx[i];
					hii[i] += xval*xval;
				}
			} /*for (i = 0;i < ndata;i++)*/			
		}
		regx += ndata;			
	} /*for (j = 0; j < nregx; j++)*/

	*OUTSTR = '\0';

	UNLOADSEG(robustvar);
	
	RETURN (returnValue);

  errorExit:
	*OUTSTR = '\0';

	UNLOADSEG(robustvar);
	
	RETURN (0);
	
}
