/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment List
#else
#define __SEG__ List
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <ctype.h>

extern char           *typeName();
extern char           *isKeyword();

extern long            Nsymbols, Nfunctions, Nscratch; /*defined in Symbol.c*/

static long            NOut, NperLine;
static char            Fmt[10];
static long            NObjects; /*number of items to list*/

#define Trash  NAMEFORTRASH
static Symbolhandle    Trash; /* will contain handles of items to list */

/*
  return 1 if name1 > name2, -1 if name1 < name2, 0 if name1 == name2
  when name1 is folded to lower case letters.  name2 is assumed already in
  lower case
 */
static short compareNames(char *name1, char *name2)
{
	register char           c1, c2, *nm1 = name1, *nm2 = name2;
	
	do
	{
		c1 = *nm1++;
		c2 = *nm2++;
		if (isupper(c1))
		{
			c1 = tolower(c1);
		}
		
		if (c1 != c2)
		{
			return( (c1 > c2) ? 1 : -1);
		}
	} while (c1 != '\0');

	return (0);
}

static char     Target[NAMELENGTH+1];
static long     MatchType;

#define STARTMATCH    1
#define ENDMATCH      2
#define MIDDLEMATCH   (STARTMATCH | ENDMATCH)
#define EXACTMATCH	  4
#define ANYMATCH      8

#define IREALS      0x01
#define ICHARS      0x02
#define ILOGICS     0x04
#define IMACROS     0x08
#define ISTRUCS     0x10
#define IGRAPHS     0x20
#define IFACTORS    0x40
#define IALL        0x7f /*IREALS|ICHARS|ILOGICS|IMACROS|ISTRUCS|IGRAPHS|IFACTORS*/
static char        *TypeKeys[] = 
{
	"real", "char", "logi", "macr", "stru", "grap", "fact", "all", (char *) 0
};
static short        KeyCodes[] = 
{
	IREALS, ICHARS, ILOGICS, IMACROS, ISTRUCS, IGRAPHS, IFACTORS, IALL
};

static void sortObjects(void)
{
	long            n = NObjects;
	Symbolhandle  **v = (Symbolhandle **) GARBAGEVALUE(Trash,0);
	Symbolhandle    z;
	char            pivot[NAMELENGTH+1], *pc1, *pc2;
	long            i, j, p, lv[20], uv[20];

	lv[0] = 0;
	uv[0] = n - 1;
	p = 0;
	while (p > -1)
	{
		if (lv[p] >= uv[p])
		{
			p = p - 1;
		} /*if (lv[p] >= uv[p])*/
		else
		{
			i = lv[p] - 1;
			j = uv[p];
			pc1 = NAME((*v)[j]);
			pc2 = pivot;
			while (*pc1)
			{
				*pc2 = *pc1++;
				if (isupper(*pc2))
				{
					*pc2 = tolower(*pc2);
				}
				pc2++;
			}
			*pc2 = '\0';
			while (i < j)
			{
				i = i + 1;
				while (compareNames(NAME((*v)[i]),pivot) < 0)
				{
					i = i + 1;
				}
				j = j - 1;
				while (j > i)
				{
					if (compareNames(NAME((*v)[j]), pivot) <= 0)
					{
						break;
					}
					j = j - 1;
				}
				if (i < j)
				{
					z = (*v)[i];
					(*v)[i] = (*v)[j];
					(*v)[j] = z;
				}
			}
			j = uv[p];
			z = (*v)[i];
			(*v)[i] = (*v)[j];
			(*v)[j] = z;
			if (i - lv[p] < uv[p] - i)
			{
				lv[p + 1] = lv[p];
				uv[p + 1] = i - 1;
				lv[p] = i + 1;
			}
			else
			{
				lv[p + 1] = i + 1;
				uv[p + 1] = uv[p];
				uv[p] = i - 1;
			}
			p++;
		} /*if (lv[p] >= uv[p]){}else{}*/
	} /*while (p > -1)*/
} /*sortObjects()*/

static void addToList(Symbolhandle symh)
{
	Symbolhandle     **objects = (Symbolhandle **) GARBAGEVALUE(Trash,0);
	
	if (matchName(NAME(symh),MatchType,Target))
	{
		(*objects)[NObjects++] = symh;
	}
}

static void listObject(Symbolhandle symh,long brief)
{
	char            *name;
	int              type;
	long             i, nclass;
	WHERE("listObject");
	
	if (symh != (Symbolhandle) 0)
	{
		name = NAME(symh);
		type = TYPE(symh);
	}
	
	if(brief < 0)
	{ /* list()*/
		if (symh != (Symbolhandle) 0)
		{
			sprintf(OUTSTR,"%-15s %-6s",name,typeName(type));
			if(brief < -1)
			{
				sprintf(OUTSTR + strlen(OUTSTR), " %7ld   ",
						SizeofSymbol(symh));
			}
			myprint(OUTSTR);

			if (type == REAL || type == LOGIC || type == CHAR || type == STRUC)
			{
				for (i = 1; i <= NDIMS(symh); i++)
				{
					sprintf(OUTSTR, " %-5ld", DIMVAL(symh,i));
					myprint(OUTSTR);
				}
				if ((nclass = NCLASS(symh)) > 0)
				{
					sprintf(OUTSTR, " FACTOR with %ld levels",nclass);
					myprint(OUTSTR);
				}
			}
			myeol();
		}
	} /*if (brief < 0)*/
	else
	{ /* listbrief() */
		if (symh != (Symbolhandle) 0)
		{
			NOut++;
			if (NOut % NperLine != 0)
			{
				sprintf(OUTSTR,Fmt,name);
				myprint(OUTSTR);
			}
			else
			{
				myerrorout(name);
			}
		} /*if (symh != (Symbolhandle) 0)*/
		else
		{
			if (NOut % NperLine != 0)
			{
				myeol();
			}
		} /*if (symh != (Symbolhandle) 0){}else{}*/
	} /*if (brief < 0){}else{}*/
	*OUTSTR = '\0';
} /*listObject()*/

static char    *keyNames[2] = {"keep", "size"};

Symbolhandle    listbrief(list)
Symbolhandle        list;
{
	Symbolhandle    symh, result = (Symbolhandle) 0;
	Symbolhandle  **objects;
	long            type, i, j, badarg = 0, startArgs = 0;
	long            keyindex;
	long            brief = (strcmp(FUNCNAME,"listbrief") == 0) ? 1 : -1;
	long            foundUndef = 0;
	long            foundMissing = 0;
	char           *keyword, *string, *name;
	char            foundKeyword = 0;
	long            keyValue, nargs = NARGS(list);
	long            op = 0, keep = 0;
	long            nosort = 0;
	long            maxNobj = Nsymbols - Nfunctions;
	long            needed, nlines;
	WHERE("list");
	
	*OUTSTR = '\0';
	Trash = GarbInstall(1);
	if (Trash == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	
	/* 
	  unadvertised feature:  Allow last argument 'nosort:T' to suppress sort
	*/
	if ((symh = COMPVALUE(list,nargs-1)) != (Symbolhandle) 0 &&
	   (keyword = isKeyword(symh)) && strcmp(keyword,"nosort") == 0 &&
	   isScalar(symh) && TYPE(symh) == LOGIC)
	{
		nosort = (DATAVALUE(symh,0) != 0.0) ? 1 : 0;
		if (nargs == 1)
		{
			Removesymbol(symh);
			COMPVALUE(list,0) = (Symbolhandle) 0;
		}
		else
		{
			nargs--;
		}
	}
	
	if (GUBED & 2)
	{
		maxNobj += Nscratch;
	}
	
	if (!getScratch(objects,0,maxNobj,Symbolhandle))
	{
		goto errorExit;
	}
	NObjects = 0;
	MatchType = ANYMATCH;
	

	if (!(GUBED & 2))
	{
		for (i = 0; i < 2; i++)
		{			
			if ((keyindex = findKeyword(list, keyNames[i], 0)) >= 0)
			{
				symh = COMPVALUE(list, keyindex);
				keyword = isKeyword(symh);
				if (!isScalar(symh) || TYPE(symh) != LOGIC)
				{
					notTorF(keyword);
					goto errorExit;
				}
				keyValue = (DATAVALUE(symh,0) != 0.0);
				if (symh == COMPVALUE(list,nargs-1))
				{
					Removesymbol(symh);
					COMPVALUE(list,nargs-1) = (Symbolhandle) 0;
					nargs = (nargs > 1) ? nargs - 1 : 1;
				}
				else
				{
					setNAME(symh,USEDKEYWORD);
				}
				if (i == 0)
				{
					keep = keyValue;
				} /*if (i == 0)*/
				else
				{
					if (brief > 0)
					{
						sprintf(OUTSTR,
								"WARNING: %s not legal keyword with %s(); ignored",
								keyword, FUNCNAME);
						putOUTSTR();
					}
					else
					{
						brief = (keyValue) ? -2 : -1;
					}
				} /*if (i == 0){}else{}*/
			} /*if ((keyindex = findKeyword(list, keyNames[i], 0)) >= 0)*/
		} /*for (i = 0; i < 2; i++)*/
	} /*if (!(GUBED & 2))*/
	
	symh = COMPVALUE(list,0);
	if (symh != (Symbolhandle) 0 && isScalar(symh) && TYPE(symh) == CHAR)
	{ /* first arg may be a pattern */
		if (isscratch(NAME(symh)) ||
		   (keyword = isKeyword(symh)) && strncmp(keyword,"pat",3) == 0)
		{
			string = STRINGPTR(symh);
			if (strlen(string) > NAMELENGTH)
			{
				sprintf(OUTSTR,"ERROR: %s is too long a pattern for %s",
						string, FUNCNAME);
				goto errorExit;
			}
			if (!scanPat(string, (long *) &MatchType, Target))
			{
				sprintf(OUTSTR,"ERROR: '%s' is not valid pattern for %s",
						string, FUNCNAME);
				goto errorExit;
			}
			
			startArgs = 1;
			symh = (Symbolhandle) 0;
		}
	} /*if (symh != (Symbolhandle) 0 && isScalar(symh) && TYPE(symh) == CHAR)*/

	if (nargs == 1 && symh == (Symbolhandle) 0)
	{  /* no arguments or just pattern, list everything that matches */
		for (symh = Firstsymbol(0); symh != (Symbolhandle) 0 ;
			 symh = Nextsymbol(symh,0))
		{ /* omit builtins, undefined, temporary, and scratch variables */
			type = TYPE(symh);
			if (type != BLTIN &&
				(type != ASSIGNED && !istempname(NAME(symh)) || GUBED & 2))
			{
				addToList(symh);
			}
		}
		if (GUBED & 2)
		{ /* list scratch table */
			for (symh = Firstsymbol(1); symh != (Symbolhandle) 0; 
				symh = Nextsymbol(symh,1))
			{
				addToList(symh);
			}
		}
	} /*if (nargs == 1 && symh == (Symbolhandle) 0)*/
	else   /* one or more arguments  */
	{
		for (i = startArgs;i < nargs;i++)
		{
			symh = COMPVALUE(list,i);
			if (!(GUBED&2) && (keyword = isKeyword(symh)))
			{
				if (!isScalar(symh) || TYPE(symh) != LOGIC)
				{
					notTorF(keyword);
					goto errorExit;
				}
				keyValue = (DATAVALUE(symh,0) != 0.0);
				
				if (strcmp(keyword,USEDKEYWORD+2) != 0)
				{
					foundKeyword = 1;
					for (j=0; (string = TypeKeys[j]) != (char *) 0;j++)
					{
						if (strncmp(keyword,string,4) == 0)
						{
							if (keyValue)
							{
								op |= KeyCodes[j];
							}
							else
							{
								op &= ~KeyCodes[j];
							}
							break;
						}
					} /*for (j=0; (string = DeleteKeys[j]) != (char *) 0;j++)*/
					if (string == (char *) 0)
					{
						badKeyword(FUNCNAME, keyword);
						goto errorExit;
					}
				} /*if (strcmp(keyword,USEDKEYWORD+2) != 0)*/
			} /*if (!(GUBED&2) && (keyword = isKeyword(symh)))*/
			else if (foundKeyword)
			{
				sprintf(OUTSTR,
						"ERROR: illegal combination of keywords and variables in %s",
						FUNCNAME);
				goto errorExit;
			}
		} /*for (i = startArgs;i < nargs;i++)*/

		if (foundKeyword)
		{ /* foundKeyword not T if only found 'keep' */
			if (op == 0)
			{
				sprintf(OUTSTR,
						"WARNING: no object types specified for %s",FUNCNAME);
				putOUTSTR();
			}
			else
			{
				for (symh=Firstsymbol(0);symh!= (Symbolhandle) 0;
					symh=Nextsymbol(symh,0))
				{/* omit scratch, builtins, undefined, temporary variables */
					type = TYPE(symh);
					if ((isDefined(symh) &&	!istempname(NAME(symh))) || GUBED & 2)
					{
						if (op & IREALS && type == REAL ||
						   op & IFACTORS && NCLASS(symh) > 0 && type == REAL ||
						   op & ICHARS && type == CHAR ||
						   op & IMACROS && type == MACRO ||
						   op & ILOGICS && type == LOGIC ||
						   op & IGRAPHS && type == PLOTINFO ||
						   op & ISTRUCS && type == STRUC)
						{
							addToList(symh);
						}
					}
				}
			}
		} /* if (foundKeyword) */
		else
		{
			for (i = startArgs;i < nargs; i++)
			{
				symh = COMPVALUE(list,i);
				if (symh == (Symbolhandle) 0)
				{
					foundMissing = 1;
				}
				else if (GUBED & 2)
				{
					addToList(symh);
				}
				else if (isscratch(NAME(symh)))
				{
					badarg = 1;
				}
				else if (!isDefined(symh))
				{
					foundUndef = 1;
				}
				else if (!isKeyword(symh))
				{
					addToList(symh);
				}
			} /*for (i = startArgs;i < nargs; i++)*/
		} /* if (foundKeyword){}else{}*/
		
		if (badarg)
		{
			myerrorout("WARNING: arguments to list() must not be expressions or strings");
		}
		if (foundMissing)
		{
			sprintf(OUTSTR,
					"WARNING: missing argument(s) to %s", FUNCNAME);
			putOUTSTR();
		}

		if (foundUndef)
		{
			for (i = 0;i < nargs; i++)
			{
				symh = COMPVALUE(list,i);
				if (symh != (Symbolhandle) 0 && !isDefined(symh))
				{
					sprintf(OUTSTR,"WARNING: %s is not defined",NAME(symh));
					putOUTSTR();
				}
			}
		} /*if (foundUndef)*/
	} /*if (nargs == 1 && symh == (Symbolhandle) 0){}else{}*/

	if (NObjects != 0)
	{
		if (!nosort)
		{
			sortObjects();
		}
	
		if (!keep)
		{
			if (brief > 0)
			{
				NperLine = SCREENWIDTH/(NAMELENGTH+1);
				if (NperLine < 1)
				{
					NperLine = 1;
				}
				NOut = 0;
				sprintf(Fmt,"%%-%lds",(long) NAMELENGTH+1);
			} /*if (brief > 0)*/

			if (brief < 0)
			{
				for (i=0;i<NObjects;i++)
				{
					listObject((*objects)[i], brief);
				}
			} /*if (brief < 0)*/
			else
			{
				nlines = (NObjects - 1)/NperLine + 1;
				for (j = 0;j < nlines; j++)
				{
					NOut = 0;
					for (i = j;i < NObjects;i += nlines)
					{
						listObject((*objects)[i], brief);
					}
					listObject((Symbolhandle) 0, brief);
				}
			} /*if (brief < 0){}else{}*/
			result = NULLSYMBOL;
		} /*if (!keep)*/
		else
		{
			needed = 0;
			for (i=0;i<NObjects;i++)
			{
				needed += strlen(NAME((*objects)[i])) + 1;
			}
			if ((result = Install(SCRATCH,CHAR)) == (Symbolhandle) 0 ||
			   (TMPHANDLE = mygethandle(needed)) == (char **) 0)
			{
				goto errorExit;
			}
			setSTRING(result,TMPHANDLE);
			setNDIMS(result,1);
			setDIM(result,1,NObjects);
			string = *TMPHANDLE;
			for (i=0;i<NObjects;i++)
			{
				name = NAME((*objects)[i]);
				strcpy(string,name);
				string += strlen(name) + 1;
			}
		} /*if (!keep){}else{}*/
	} /*if (NObjects != 0)*/	
	else if (keep)
	{
		sprintf(OUTSTR,
				"ERROR: no variables found whose names to keep");
		goto errorExit;
	}
	
	emptyTrash();
	
	return (result);

  errorExit:
	putOUTSTR();

	Removesymbol(result);

	emptyTrash();
	
	return (0);
}


