/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#include <math.h>
#include "globals.h"

#define n (*N)
#define ml (*ML)
#define ls (*LS)
#define lx (*LX)
#define cn (*CN)
#define co (*CO)
#define iq (*IQ)
#define nd (*ND)
#define nn (*NN)
#define ky (*KY)
#define ll (*LL)
#define ly (*LY)
#define xn(i,j) ( XN[(i-1)+ *ND * (j-1)] )
#define dt(i) ( DT[i-1] )
#define ind(i) ( IND[i-1] )
#define myint(i) ( INT[i-1] )
/*
#                             partial sweep
*/
void leapspivot(N, ML, LS, LX, XN, INT, IND, CN, CO, DT, IQ, ND,
				NN, KY, LL, LY)
double         *XN, *DT, *CN, *CO;
long           *N, *ML, *LS, *LX, *INT, *IND, *IQ, *ND, *NN, *KY, *LL,
               *LY;
/*
dimension xn(nd,nd),dt(nd),ind(nd),int(nd)
*/
{
	double          b;
	long            j, l, k, m;
	nn = lx;
	if (iq != 3)
	{
		cn = cn + co;
		xn(n, n) = -xn(n, n);
		xn(ky, ky) = xn(ky, ky) + xn(n, ky) * xn(n, ky) / xn(n, n);
	}
	if (ml != 0)
	{
		/*
		#                            gauss-jordan
		*/
		for (l = 1; l <= ml; l++)
		{
			j = myint(l);
			b = xn(n, j) / xn(n, n);
			if (iq == 0)
			{
				xn(j, j) = xn(j, j) + b * xn(n, j);
			}
			if (iq == 1)
			{
				xn(j, j) = dt(l);
			}
			for (m = ll; m <= ly; m++)
			{
				k = ind(m);
				xn(j, k) = xn(j, k) + b * xn(n, k);
				xn(k, j) = xn(j, k);
			}
		}
	}

	if (ll <= lx)
	{
		/*
		#                                gauss
		*/
		for (l = ll; l <= lx; l++)
		{
			j = ind(l);
			b = xn(n, j) / xn(n, n);
			for (m = ls; m <= l; m++)
			{
				k = ind(m);
				xn(j, k) = xn(j, k) + b * xn(n, k);
				xn(k, j) = xn(j, k);
			}
			xn(j, ky) = xn(j, ky) + b * xn(n, ky);
			xn(ky, j) = xn(j, ky);
		}
	}
}


#define wt (*WT)
#define mn (*MN)
#define xi(i,j) ( XI[ (i-1) + nd * (j-1) ] )
#define d(i) ( D[i-1] )
#define dd(i) ( DD[i-1] )
#define sq(i) ( SQ[i-1] )
/*
#                           check tolerances
*/
void leapstest(WT, N, XI, INT, D, DD, SQ, DT, ML, MN, ND)
double         *WT, *XI, *D, *DD, *SQ, *DT;
long           *N, *INT, *ML, *MN, *ND;
{
	long            j, l;

	d(n) = xi(n, n);
	wt = d(n) / sq(n);
	dd(mn) = wt * d(n);
	myint(mn) = n;
	if (ml != 0)
	{
		for (l = 1; l <= ml; l++)
		{
			j = myint(l);
			/*
			#*****the next two lines of code are modified for cdc (?)
			*/
			dt(l) = xi(j, j);
			if (xi(n, n) != 0.)
			{
				dt(l) = dt(l) - xi(n, j) * xi(n, j) / xi(n, n);
			}
			if (-dd(l) / dt(l) < wt)
			{
				wt = -dd(l) / dt(l);
			}
		}
	}
}


#define rs (*RS)
#define cl(i) ( CL[i-1] )
#ifdef MPW
#ifdef MPW3
#pragma segment Screen
#else
#define __SEG__ Screen
#endif /*MPW3*/
#endif /*MPW*/

#define rm(i) ( RM[i-1] )
#define mb (*MB)
#define df (*DF)
#define sig (*SIG)
#define pen (*PEN)
#define ib (*IB)
#define ns (*NS)
/*
#                   labels and saves best regressions
*/
void leapsqstore(RS, CN, CL, RM, MB, DF, MN, SIG, PEN, IB, NS)
double         *RS, *CN, *CL, *RM, *DF, *SIG, *PEN;
long           *MB, *MN, *IB, *NS;
{
	double          tr;
	long            l, m, mr;

	leapstrans(&tr, &mr, RS, DF, MN, SIG, PEN, IB, MB);
	if (rm(mr) > tr)
	{
		for (m = 1; m <= mb; m++)
		{
			l = mr - mb + m;
			if (cl(l) == cn)
			{
				return;
			}
		}
		if (mb != 1)
		{
			for (m = 2; m <= mb; m++)
			{
				if (tr >= rm(l - 1))
				{
					break;
				}
				rm(l) = rm(l - 1);
				cl(l) = cl(l - 1);
				l--;
			}
		}

		rm(l) = tr;
		cl(l) = cn;
	}
}


#define mt (*MT)
#define mp (*MP)
#define ci (*CI)
#define sc(i) ( SC[i-1] )
/*
#                              matrix copy
*/
void leapscopy(XN, XI, INT, IND, ML, MT, MP, ND, SC, CN, CI, IQ)
double         *XN, *XI, *SC, *CN, *CI;
long           *IND, *INT, *ML, *MT, *MP, *ND, *IQ;
{
	long            j, k, l, m;
	for (l = 1; l <= mp; l++)
	{
		if (l > ml)
		{
			m = mt + l - ml - 1;
			myint(l) = ind(m);
		}
		k = myint(l);
		for (m = 1; m <= l; m++)
		{
			j = myint(m);
			if (iq == 0)
			{
				xn(j, k) = xi(j, k);
			}
			if (iq == 1)
			{
				xn(j, k) = xi(j, k) / (sc(j) * sc(k));
			}
			xn(k, j) = xn(j, k);
		}
	}
	cn = ci;
}


#undef myint

#define tr (*TR)
#define mr (*MR)
/*
#             transform of criterion-smaller must be better
*/
void leapstrans(TR, MR, RS, DF, MN, SIG, PEN, IB, MB)
double         *TR, *RS, *DF, *SIG, *PEN;
long           *MR, *MN, *IB, *MB;
{
	mr = mb;

	switch ((int) ib)
	{
	  case 1:
		tr = rs;
		mr = mn * mb;
		break;
	  case 2:
		tr = rs / (df - (double) mn);
		break;
	  case 3:
		tr = rs + pen * (mn) * sig;
		break;
	}
}
