/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Grafwind
#else
#define __SEG__ Grafwind
#endif /*MPW3*/
#endif /*MPW*/

/*
	MacAnova graph window handler.  Modelled after TransSkel windows
	940429 Modified to work with TransSkel version 3.12
*/


#include	<stdio.h>

#include	"macIface.h"
#include	"globals.h"
#include	"dbug.h"

extern void     PictWindPrint();
Rect            PanelRect;
/* return whether the window is zoomed in or out */
static Integer GraphWindZoomStatus(Integer windno)
{
	Rect            r, stdState;
	WindowPtr       theWindow;
	WindowPeek      theWindowPk;
	WStateData    **wStateHandle;
	WHERE("GraphWindZoomStatus");

	theWindow = (windno < NGRAPHS) ? GraphWind[windno] : PanelWind;
	theWindowPk = (WindowPeek) theWindow;

	wStateHandle = (WStateData **) theWindowPk->dataHandle;
	HLock((Handle) wStateHandle);
	stdState = (**wStateHandle).stdState;
	HUnlock((Handle) wStateHandle);

	r = theWindow->portRect;
	if(r.top - r.bottom == stdState.top - stdState.bottom &&
	   r.left - r.right == stdState.left - stdState.right)
	{
		return (inZoomOut);
	}
	else
	{
		return (inZoomIn);
	}
}

/*
	window handler procedures
*/


/*
	Dispose of graphWindow;  do nothing if graphWindow == 0
*/
static void GraphWindDispose(WindowPtr graphWindow)
{
	Integer         windno;
	WindowPeek      theWindow;
	WindowPtr       nextWindow;
	WStateData    **wStateHandle, *wStatePtr;
	WHERE("GraphWindDispose");

	if(graphWindow != (WindowPtr) 0)
	{
		windno = whichGraphWindow(graphWindow);
		if(0 <= windno && windno < NGRAPHS)
		{
		/*
			Save zoom info for non-panel windows
		*/
			theWindow = (WindowPeek) graphWindow;
			wStateHandle = (WStateData **) theWindow->dataHandle;
			HLock((Handle) wStateHandle);
			wStatePtr = GraphWState + windno;
			wStatePtr->userState = (*wStateHandle)->userState;
			wStatePtr->stdState = (*wStateHandle)->stdState;
			HUnlock((Handle) wStateHandle);
		} /*if(0 <= windno && windno < NGRAPHS)*/

		SkelRmveWind(graphWindow); /* calls GraphWindClobber */
		
		if(GraphWind[0] == (PicHandle) 0 && GraphWind[1] == (PicHandle) 0 &&
		   GraphWind[2] == (PicHandle) 0 && GraphWind[3] == (PicHandle) 0 &&
		   PanelWind != (WindowPtr) 0)
		{
			SkelRmveWind(PanelWind);
		}
		nextWindow = FrontWindow();
		if(nextWindow != (WindowPtr) 0)
		{
			SelectWindow(nextWindow);
		}
	} /*if(graphWindow != (WindowPtr) 0)*/
} /*GraphWindDispose()*/

/*
	Set the frame for the panel window to the minimum size that will hold
	all the current panels.
	At entry, r should be the port rectangle of the window
*/
static void setPanelRect(Rect *r)
{
	Integer        powtwo = 1, windowPattern = 0;
	Integer        halfHeight, halfWidth;
	Integer        k;
	
	powtwo = 1;
	for(k=0;k<4;k++)
	{ /* set a bit for each window present */
		if(myPic[k] != (PicHandle) 0)
		{
			windowPattern |= powtwo;
		}
		powtwo *= 2;
	}

	r->right -= RIGHTOFFSET;
	halfHeight = (r->bottom - r->top)/2;
	halfWidth = (r->right - r->left)/2;

	if(!(windowPattern & 0x03))
	{ /* no panels in top half */
		r->top += halfHeight;
	}
	if(!(windowPattern & 0x0c))
	{ /* no panels in bottom half */
		r->bottom -= halfHeight;
	}
	if(!(windowPattern & 0x05))
	{ /* no panels in left half*/
		r->left += halfWidth;
	}
	if(!(windowPattern & 0x0a))
	{ /* no panels in right half */
		r->right -= halfWidth;
	}
} /*setPanelRect()*/

/*
	Draw copies of each of the basic graph windows in the corners of
	the pane window
*/ 
static void drawPanelPicture(void)
{
	Rect           r, r1;
	Integer        i, j, k;
	Integer        halfHeight, halfWidth;
	WindowPtr      theWindow;

	SetRect(&r, -200, -200, 1483, 1275);
	ClipRect(&r);

	r = PanelRect;
	setPanelRect(&r);
	PanelPic = OpenPicture(&r);

	r = PanelRect;
	r.right -= RIGHTOFFSET;
	halfHeight = (r.bottom - r.top) / 2;
	halfWidth = (r.right - r.left) / 2;

	k = 0;
	for(i = 0;i<2;i++)
	{
		r1.top = r.top;
		if(i == 1)
		{
			r1.top += halfHeight;
		}
		r1.bottom = r1.top + halfHeight;
		for(j = 0;j<2;j++)
		{
			r1.left = r.left;
			if(j == 1)
			{
				r1.left += halfWidth;
			}
			r1.right = r1.left + halfWidth;
			if((theWindow = GraphWind[k]) != (WindowPtr) 0)
			{
				DrawPicture(myPic[k], &r1);
			}
			k++;
		}
	} /*for(i = 0;i<2;i++)*/
	ClosePicture();
	EnableItem(WindowMenu, panel);
} /*drawPanelPicture()*/

/*
	Called from macMain.c with windno == -1 to zero
	GraphWState[.].userState and GraphWState[.].stdState

	Called from GraphWindInit() to initialize GraphWState[windno].userState
	and GraphWState[windno].stdState
*/

void GraphRectInit(windno)
Integer         windno;
{
	WindowPeek  theWindow;
	Integer     i;
	WStateData **wStateHandle;
	Rect       *r, strucRgnBox, contRgnBox;
	Integer     top[4], left[4], bottom[4], right[4];
	Integer     strucHeight, contHeight, height, width, barHeight;
	
	if(windno < 0)
	{ /* zero GraphWindState[i].{user,std}State, i = 0,...,3 */
		for(i=0;i<NGRAPHS;i++)
		{
			r = &GraphWState[i].userState;
			r->top = r->left = r->bottom = r->right = 0;
			r = &GraphWState[i].stdState;
			r->top = r->left = r->bottom = r->right = 0;		
			GraphWInZoom[i] = inZoomOut;
		}
	} /*if(windno < 0)*/
	else
	{
		theWindow = (WindowPeek) GraphWind[windno];
	/*
		setup GraphWindState[windno].{user,std}State, using
		information in theWindow->dataHandle, theWindow->strucRgn and
		theWindow->contRgn
	*/
		wStateHandle = (WStateData **) (theWindow->dataHandle);
		HLock((Handle) wStateHandle);
		left[0] = left[2] = (*wStateHandle)->stdState.left;
		right[1] = right[3] = (*wStateHandle)->stdState.right;
		width = (right[1] - left[0] + 1)/2;
		right[0] = right[2] = left[0] + width;
		left[1] = left[3] = right[1] - width;

		strucRgnBox = (*(theWindow->strucRgn))->rgnBBox;
		contRgnBox = (*(theWindow->contRgn))->rgnBBox;
		strucHeight = strucRgnBox.bottom-strucRgnBox.top;
		contHeight = contRgnBox.bottom-contRgnBox.top;
		barHeight = strucHeight - contHeight;

		top[0] = top[1] = (*wStateHandle)->stdState.top;
		bottom[2] = bottom[3] = (*wStateHandle)->stdState.bottom;
		height = (bottom[2] - top[0] - barHeight + 1)/2;
		bottom[0] = bottom[1] = top[0] + height;
		top[2] = top[3] = bottom[2] - height;

		r = &GraphWState[windno].userState;
		r->top = top[windno % NGRAPHS];
		r->left = left[windno % NGRAPHS];
		r->bottom = bottom[windno % NGRAPHS];
		r->right = right[windno % NGRAPHS];
		(*wStateHandle)->userState = *r;
		HUnlock((Handle) wStateHandle);
	} /*if(windno < 0){}else{}*/
} /*GraphRectInit()*/

/*
	Initialize a specific graphics window, 0, 1, 2, or 3 = NGRAPHS - 1
*/
Boolean GraphWindInit(Integer windno)
{
	Rect            r;
	Str15           str;
	Integer         initialize, offset;
	WindowPeek      theWindow;
	WStateData    **wStateHandle, *wStatePtr;
	GrafPtr         thePort;
	WHERE("GraphWindInit");

	GetPort(&thePort);

	GraphWindDispose(GraphWind[windno]);
	GraphWind[windno] = GetNewWindow(GRAPHWINDOW, (Ptr) nil, (WindowPtr) -1L);
	if(GraphWind[windno] == (WindowPtr) 0 || !SkelWindow(GraphWind[windno],
				   (SkelWindMouseProcPtr) GraphWindMouse,
				   (SkelWindKeyProcPtr) GraphWindKey,
				   (SkelWindUpdateProcPtr) GraphWindUpdate,
				   (SkelWindActivateProcPtr) GraphWindActivate,
				   (SkelWindCloseProcPtr) GraphWindClose,
				   (SkelWindClobberProcPtr) GraphWindClobber,
				   (SkelWindIdleProcPtr) nil, true))
	{
		goto errorExit;
	}

	sprintf((char *) str, "Graph %d",windno+1);
	CtoPstr((char *) str);
	SetWTitle(GraphWind[windno], str);

	theWindow = (WindowPeek) GraphWind[windno];

	wStateHandle = (WStateData **) theWindow->dataHandle;
	HLock((Handle) wStateHandle);
	(*wStateHandle)->stdState = (*wStateHandle)->userState;

	SetPort((GrafPtr) GraphWind[windno]);

	wStatePtr = GraphWState + windno;

	initialize = (wStatePtr->userState.top == 0 &&
				  wStatePtr->userState.left == 0 &&
				  wStatePtr->userState.bottom == 0 &&
				  wStatePtr->userState.right == 0);
	if(initialize)
	{/* first time, so set up info for zooming */
		GraphRectInit(windno);
	}
	else
	{/* restore zoom information */
		(*wStateHandle)->userState = wStatePtr->userState;
		(*wStateHandle)->stdState = wStatePtr->stdState;
	}
	HUnlock((Handle) wStateHandle);
	/* always create picture with window zoomed out */
	ZoomWindow(GraphWind[windno], inZoomOut, true);

	setCmdFileItems(false);
	setGraphEditItems();
	EnableItem(WindowMenu, hideit);
	EnableItem(WindowMenu, closeit);
	offset = MAXWINDOWS - Nwindows;
	DisableItem(WindowMenu, gototop - offset);
	DisableItem(WindowMenu, gotobottom - offset);
	DisableItem(WindowMenu, pageup - offset);
	DisableItem(WindowMenu, pagedown - offset);
	
	TextFont(GraphFont);
	TextSize(GraphFontSize);

	SetRect(&r, -200, -200, 1483, 1275);
	ClipRect(&r);
	r = GraphWind[windno]->portRect;
	EraseRect(&r);

	r.right -= RIGHTOFFSET;

	myPic[windno] = OpenPicture(&r);

	ShowPen();

	return (true);

  errorExit:
	if(GraphWind[windno] != (WindowPtr) 0)
	{
		DisposeWindow(GraphWind[windno]);
	}
	return (false);
} /*GraphWindInit()*/

/*
	Return the index for an available graph window, if any.
	If global ThisWindow >= 0, return ThisWindow + 1 as the number of the
	graph window to be used.
	Otherwise, if there is a graph window that is not in use, RETURN its index
	(from 1 to NGRAPHS).
	If there is no available graph window, return 0.

	ThisWindow is set in do_whole_graph() to -1 (if window:windno not a
	plotting command argument) or to windno-1 (if windno > 0), or is
	left unchanged (if windno == 0).
*/
Integer GraphWindSetup()
{
	int             windno;
	WHERE("GraphWindSetup");

	if(ThisWindow < 0)
	{ /* find unused window */
		for (windno = 0; windno < NGRAPHS; windno++)
		{
			if (GraphWind[windno] == (WindowPtr) 0)
			{
				ThisWindow = windno;
				break;
			}
		}
		if(windno == NGRAPHS)
		{ /* no unused windows */
			return (0);
		}
	} /*if(ThisWindow < 0)*/
/* at this point, 0 <= ThisWindow < NGRAPHS */
	return ((GraphWindInit(ThisWindow)) ? ThisWindow + 1 : -1);
} /*GraphWindSetup()*/

void GraphOver()
{
	Rect            r;
	GrafPtr         thePort;
	Integer         windno;
	WHERE("GraphOver");

	GetPort(&thePort);
	windno = whichGraphWindow((WindowPtr) thePort);

	ClosePicture();
	HidePen();

	ZoomWindow(GraphWind[windno], GraphWInZoom[windno], true);
	MyShowWindow(GraphWind[windno]);
	DrawGrowBox(thePort);

	r = thePort->portRect;
	EraseRect(&r);
	r.right -= RIGHTOFFSET;
	DrawPicture(myPic[windno], &r);

	EnableItem(WindowMenu, graph1 + windno);
	setGraphFileItems(true);
	setCommandM(CurrentWindow);

	if(PanelWind == (WindowPtr) 0)
	{
		PanelWind = GetNewWindow(GRAPHWINDOW, (Ptr) nil, (WindowPtr) thePort);
		if(PanelWind == (WindowPtr) 0 ||
		   !SkelWindow(PanelWind,
					   (SkelWindMouseProcPtr) GraphWindMouse,
					   (SkelWindKeyProcPtr) GraphWindKey,
					   (SkelWindUpdateProcPtr) GraphWindUpdate,
					   (SkelWindActivateProcPtr) GraphWindActivate,
					   (SkelWindCloseProcPtr) GraphWindClose,
					   (SkelWindClobberProcPtr) GraphWindClobber,
					   (SkelWindIdleProcPtr) GraphWindIdle, true))
		{
			goto errorExit;
		}
		PanelRect = PanelWind->portRect;

		SkelSetZoom(PanelWind, (SkelWindZoomProcPtr) GraphWindZoom);
		SetWTitle(PanelWind, PanelTitle);
	} /*if(PanelWind == (WindowPtr) 0)*/

	if(PanelPic != (PicHandle) 0)
	{
		KillPicture(PanelPic);
		PanelPic = (PicHandle) 0;
	}

	HideWindow(PanelWind);

	drawPanelPicture();
	ShowWindow(PanelWind);

	SetPort(thePort);
	ValidRect(&thePort->portRect);

	return;

  errorExit:
	if(PanelWind != (WindowPtr) 0)
	{
		DisposeWindow(PanelWind);
	}
	myAlert("Cannot create Panel of Graphs window; out of memory (?)");
} /*GraphOver()*/


/* Should be entered only when a graph window is in fron */
pascal void GraphWindIdle(void)
{
	Integer      windno;

	windno = whichGraphWindow(FrontWindow());
	if(0 <= windno && windno <= NGRAPHS)
	{ /* be paranoid*/
		setGraphFileItems(true);
	}
} /*GraphWindIdle()*/

/*
	get rid of graph window and any associated picture.
	The picHandle and WindowPtr are set to nil
*/
pascal void
GraphWindClobber(void)
{
	GrafPtr         thePort;
	Integer         windno;
	WHERE("GraphWindClobber");

	GetPort(&thePort);
	windno = whichGraphWindow((WindowPtr) thePort);
	
	if(windno >= 0)
	{
		if(windno < NGRAPHS && myPic[windno] != (PicHandle) 0)
		{
			KillPicture(myPic[windno]);
			myPic[windno] = (PicHandle) 0;
			GraphWind[windno] = (WindowPtr) 0;
		}
		else if(windno == NGRAPHS)
		{
			if(PanelPic != (PicHandle) 0)
			{
				KillPicture(PanelPic);
				PanelPic = (PicHandle) 0;
			}
			PanelWind = (WindowPtr) 0;
		}
		GraphWindPaused = false;
		DisableItem(WindowMenu, graph1+windno);
	}
	DisposeWindow((WindowPtr) thePort);
}/*GraphWindClobber()*/


/*
	Close box clicked or close on the Window menu
*/
pascal void GraphWindClose(void)
{
	WindowPtr      theWindow = FrontWindow();
	Integer        windno = whichGraphWindow(theWindow);
	WHERE("GraphWindClose");

	if(windno >= 0)
	{
		GraphWindDispose(theWindow);
		/* throw away panel picture if we are closing non-panel graph window */
		if(windno < NGRAPHS && PanelPic != (PicHandle) 0)
		{
			KillPicture(PanelPic);
			PanelPic = (PicHandle) 0;
			SetPort((GrafPtr) PanelWind);
			InvalRect(&PanelWind->portRect);
		}
	}
} /*GraphWindClose()*/

pascal void GraphWindZoom(WindowPtr theWindow, Integer zoomDir)
{
	Integer         windno = whichGraphWindow(theWindow);

	if (0 <= windno && windno < NGRAPHS)
	{
		GraphWInZoom[windno] = zoomDir;
	} EraseRect(&theWindow->portRect);
	ZoomWindow(theWindow, zoomDir, false);
} /*GraphWindZoom()*/

pascal void GraphWindActivate(active)
Boolean         active;
{
	GrafPtr         thePort;
	Integer         offset = MAXWINDOWS - Nwindows;
	WHERE("GraphWindActivate");

	GetPort(&thePort);
	DrawGrowBox(thePort);
	if (active)
	{
		InitCursor();

		setGraphFileItems(true);
		EnableItem(FileMenu, (Integer) 0);

		setGraphEditItems();
		EnableItem(EditMenu, (Integer) 0);

		if(UseWindows > 0)
		{
			setCommandM(CurrentWindow);
		}
		if (!GraphWindPaused)
		{
			EnableItem(WindowMenu, (Integer) 0);
			EnableItem(WindowMenu, hideit);
			EnableItem(WindowMenu, closeit);
			DisableItem(WindowMenu, newwind);
			DisableItem(WindowMenu, gototop - offset);
			DisableItem(WindowMenu, gotobottom - offset);
			DisableItem(WindowMenu, pageup - offset);
			DisableItem(WindowMenu, pagedown - offset);

			EnableItem(CommandMenu, (Integer) 0);
			EnableItem(OptionsMenu, (Integer) 0);
		}		/*if(!GraphWindPaused)*/
	} /*if(active)*/ 
	else
	{
		setGraphFileItems(false);

		DisableItem(FileMenu, (Integer) 0);
		DisableItem(WindowMenu, (Integer) 0);

		DisableItem(CommandMenu, (Integer) 0);
		DisableItem(OptionsMenu, (Integer) 0);

		EnableItem(EditMenu, undoit);
		EnableItem(EditMenu, cutit);
		EnableItem(EditMenu, copyit);
		EnableItem(EditMenu, pastit);
	} /*if(active){...}else{...}*/
	DrawMenuBar();
}/*GraphWindActivate()*/


pascal void GraphWindMouse(thePt, when, mods)
Point           thePt;
LongInt         when;
Integer         mods;
{
	GrafPtr         thePort;
	WindowPtr       graphWindow;
	Integer         whichGraph = 0;
	Integer         halfHeight, halfWidth;
	Rect            r;
	Point           thePt2 = thePt;

	if(!GraphWindPaused)
	{
		when;			/*just to defeat compiler message */
		mods;			/*just to defeat compiler message */
		InitCursor();
		GetPort(&thePort);
		graphWindow = (WindowPtr) thePort;
		if (graphWindow == PanelWind)
		{
			r = PanelWind->portRect;
			halfHeight = (r.bottom - r.top) / 2;
			halfWidth = (r.right - r.left) / 2;
			if (thePt.v > r.top + halfHeight)
			{
				whichGraph += 2;
			}
			if (thePt.h > r.left + halfWidth)
			{
				whichGraph++;
			}
			if ((graphWindow = GraphWind[whichGraph]) != (WindowPtr) 0)
			{
				MyShowWindow(graphWindow);
			}
		} /*if (graphWindow == PanelWind)*/
	} /*if(!GraphWindPaused)*/
} /*GraphWindMouse()*/

/* Process key strokes in Graph Window Ignore everything except Command Period
*/

pascal void GraphWindKey(Integer ch1, Integer code, Integer mods)
{
	unsigned char   ch = (unsigned char) ch1;
	Integer         cmdNo = -1;

	HLock((Handle) CommandMenu);
	if(ch1 == FUNCTIONKEY)
	{
		cmdNo = functionKeyToNumber(code);
		if(((*CommandMenu)->enableFlags & 1) &&
		   5 <= cmdNo && cmdNo < 5 + NCOMMANDS)
		{
			DoCommand(command1+cmdNo-5); /*simulate menu selection */
		}
		else if(cmdNo == 3) /*copy*/
		{
			DoEdit(copyit);
		}
	}
	else if (mods & cmdKey)
	{
		if (ch == Period)
		{
			DoFile(interrupt);
		}
		else if(!GraphWindPaused && ch == OptionQ)
		{
			SaveOnQuit = false;
			quitIt();
		}
		else if (((*CommandMenu)->enableFlags & 1) &&
				 (cmdNo = optionKeyToNumber(ch)) > 0 && cmdNo <= NCOMMANDS)
		{
			DoCommand(command1+cmdNo-1); /*simulate menu selection */
		}
	} /*if(mods & cmdKey)*/ 
	else
	{
		if (ch == Return || ch == Enter)
		{
			if(GraphWindPaused)
			{ /* terminate pause */
				GraphWindPaused = false;
			}
			else
			{ /* switch to command window */
				DoWind((CmdWindows + CurrentWindow)->menuItem);
			}
		} /*if (ch == Return || ch == Enter)*/
	}/*if(mods & cmdKey){}else{}*/
	HUnlock((Handle) CommandMenu);
} /*GraphWindKey()*/

/*
	Update Graph window.
*/

pascal void GraphWindUpdate(resized)
Boolean         resized;
{
	Rect            r;
	GrafPtr         thePort;
	PicHandle       picture;
	int             windno;
	WHERE("GraphWindUpdate");

	GetPort(&thePort);
	windno = whichGraphWindow((WindowPtr) thePort);
	if(windno >= 0)
	{
		r = thePort->portRect;
		EraseRect(&r);
		r.right -= RIGHTOFFSET;
	
		if (0 <= windno && windno < NGRAPHS)
		{
			if (resized)
			{		/* save current zoom state */
				GraphWInZoom[windno] = GraphWindZoomStatus(windno);
			} picture =	myPic[windno];
		} /*if (0 <= windno && windno < NGRAPHS)*/
		else
		{ /* must be panel */
			if(PanelPic == (PicHandle) 0)
			{ /* create picture for panel */
				drawPanelPicture();
			} /*if(PanelPic == (PicHandle) 0)*/
			picture = PanelPic;
			r = PanelWind->portRect;
			setPanelRect(&r);
		}
		DrawPicture(picture, &r);
		DrawGrowBox(thePort);
	} /*if(windno >= 0)*/
}/*GraphWindUpdate()*/

static pascal Boolean GraphWindEventHook(EventRecord *theEvent)
{
	Integer     evtPart;
	GrafPtr     evtPort;
	Point       evtPt;

	if(theEvent->what == mouseDown)
	{
		/* Don't allow user to switch windows, but allow access to menu bar */
		evtPt = theEvent->where;
		evtPart = FindWindow(evtPt, &evtPort);
		switch(evtPart)
		{
		  case inSysWindow:
		  case inGoAway:
		  case inMenuBar:
			return (false);
		  default:
			SysBeep(30);
			return (true);
		} /*switch(evtPart)*/
	} /*if(TheEvent->what == mouseDown)*/
	return (false);
} /*GraphWindEventHook()*/

void GraphWindPause(void)
{
	SkelEventHookProcPtr   oldHook = SkelGetEventHook();
	WHERE("GraphWindPause");

	SkelSetEventHook((SkelEventHookProcPtr) GraphWindEventHook);
	GraphWindPaused = true;
	setGraphFileItems(true);
	DisableItem(WindowMenu, (Integer) 0);

	DisableItem(CommandMenu, (Integer) 0);
	DisableItem(OptionsMenu, (Integer) 0);

	DrawMenuBar();

	while(GraphWindPaused)
	{
		SkelDoEvents(everyEvent);
	}
	SkelSetEventHook(oldHook);

/* Note: we only get here by pausing in the middle of a plot command */
	Running = 0;
	setGraphFileItems(true);
	Running = 1;
	EnableItem(WindowMenu, (Integer) 0);

	EnableItem(CommandMenu, (Integer) 0);
	EnableItem(OptionsMenu, (Integer) 0);
	DrawMenuBar();

} /*GraphWindPause()*/

void GraphWindPrint(windno)
Integer             windno;
{
	GrafPtr     thePort;
	Rect        r;
	WHERE("GraphWindPrint");
	
	if(0 <= windno && windno <= NGRAPHS)
	{
		GetPort(&thePort);
		PictWindPrint(windno);
		r = thePort->portRect;
		InvalRect(&r);
		BeginUpdate((WindowPtr) thePort);
		GraphWindUpdate((Boolean) 0);
		EndUpdate((WindowPtr) thePort);
	}
} /*GraphWindPrint()*/

void GraphWindEditMenu(item, whichWindow)
Integer             item;
WindowPtr           whichWindow;
{
	PicHandle          thisPicture = (PicHandle) 0;
	Integer            windno;
#ifdef MPW
	unsigned LongInt  *pack;
#else
	unsigned long     *pack;
#endif
	char              *tmp;
	WHERE("GraphWindEditMenu");

	switch (item)
	{
	  case undoit:			/* only works in cmd window */
		break;
		
	  case cutit:				/* only works in cmd window */
		break;

	  case copyit:
#ifdef MPW
		pack = (unsigned LongInt *) "PICT";
#else
		pack = (unsigned long *) "PICT";
#endif /* MPW */

		windno = whichGraphWindow(whichWindow);
		if(windno == NGRAPHS)
		{
			thisPicture = PanelPic;
		}
		else if(windno >= 0)
		{
			thisPicture = myPic[windno];
		}

		if(thisPicture != (PicHandle) 0)
		{
			LongInt      answer;

			(void) ZeroScrap();
			HLock((Handle) thisPicture);
			tmp = (char *) *thisPicture;
			answer = PutScrap((LongInt) ((*thisPicture)->picSize), *pack, tmp);
			HUnlock((Handle) thisPicture);
		}
		break;

	  case pastit:				/* only works in cmd window */
		break;

	  case copytoend:			/* only works in cmd window */
		break;

	 } /*switch (item)*/
} /*GraphWindEditMenu()*/

/* routine to set File menu items for graph window */
void setGraphFileItems(Boolean active)
{
	if(active)
	{
		EnableItem(FileMenu, pagesetup);
		EnableItem(FileMenu, printit);
		SetItem(FileMenu, printit, "\pPrint Graph");
		EnableItem(FileMenu, interrupt);
		if (!GraphWindPaused)
		{
			if (!Running && BDEPTH <= 0)
			{
				EnableItem(AppleMenu, helpit);
				EnableItem(FileMenu, keepit);
				EnableItem(FileMenu, keepitas);
				EnableItem(FileMenu, quit);
			}
			DisableItem(FileMenu, go_on);
		}	/*if(!GraphWindPaused)*/
		else
		{
			DisableItem(AppleMenu, helpit);
			EnableItem(FileMenu, go_on);
			DisableItem(FileMenu, keepit);
			DisableItem(FileMenu, keepitas);
			DisableItem(FileMenu, quit);
		} /*if(!GraphWindPaused){}else{}*/
	} /*if(active)*/
	else
	{
		DisableItem(FileMenu, go_on);
	} /*if(active){}else{}*/
} /*setGraphFileItems()*/

void setGraphEditItems()
{
	DisableItem(EditMenu, undoit);
	DisableItem(EditMenu, cutit);
	DisableItem(EditMenu, pastit);
	DisableItem(EditMenu, copytoend);
	EnableItem(EditMenu, copyit);
} /*setGraphEditItems()*/
