/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Handles
#else
#define __SEG__ Handles
#endif /*MPW3*/
#endif /*MPW*/

#ifdef MPW
#include <Memory.h>
#include <Errors.h>
#include <Stdlib.h>

#ifdef MPW1
void ResrvMem();
OSErr MemError();
#endif /*MPW1*/

#define MemErr MemError()
#else /*MPW*/
#include <MemoryMgr.h>
#endif /*MPW*/

#ifdef USEMAXBLOCK
#define isRoom(N) (N < MaxBlock())
#else
#define isRoom(N) (1)
#endif USEMAXBLOCK

#include "globals.h"

/*
  New version of 930125
  931015  infoblock structures no longer have a component 'next'

  A static array of infoblock structures is now used so as to reduce the number
  of ROM references.  The first long word of the data contains the true
  handle.

  Handlestart is the most recent ("youngest") handle allocated.

  infoblock(h).next is next "older" handle, or 0 if h is the "oldest"

  Version of 950804 implemented limits on warehouse lengths

*/

#define MEMINC(N) \
				(\
					CurrentMemory += (N) + DATAOFFSET, \
					MaxMemory = (CurrentMemory > MaxMemory) ?\
						CurrentMemory : MaxMemory\
				)
#define MEMDEC(N) (CurrentMemory -= (N) + DATAOFFSET)

typedef struct infoblock
{
	char       *data;         /* pointer to data */
	char      **realHandle;   /* true system handle */
	long        length;       /* number of bytes of data*/
	short       next;         /* index of next handle in chain */
} infoblock, *infoblockptr;

#define DOUBLESIZE    sizeof(double)
#define SYMBOLSIZE    sizeof(Symbol)

#define LISTSIZE (DEFAULTLISTLENGTH * sizeof(Symbolhandle))

#define POW2          3  /*log2 DOUBLESIZE*/
#define MASK          0x07 /*DOUBLESIZE - 1*/

#ifdef POW2
#define MODDOUBLE(A)  ((A) & MASK)
#define DIVDOUBLE(A)  ((A) >> POW2)
#else /*POW2*/
#define MODDOUBLE(A)  ((A) % DOUBLESIZE)
#define DIVDOUBLE(A)  ((A) / DOUBLESIZE)
#endif /*POW2*/

#define DATAOFFSET    DOUBLESIZE /* DOUBLESIZE to retain alignment */

extern long            CurrentMemory, MaxMemory; /* defined in comhandl.c */
long                   UseResrvMem = 0;
static infoblockptr    Handles = (infoblockptr) 0;

/*
	Warehouse handles of sizes {1,2,...,NWAREHOUSES}*DOUBLESIZE in
	Warehouse[0],...,Warehouse[NWAREHOUSES-1], LISTSIZE
	in Warehouse[NWAREHOUSES], and SYMBOLSIZE in Warehouse[NWAREHOUSES+1]

  950804: Implemented maximum warehouse sizes.  Currently all warehouse have
          length of MAXWAREHOUSELENGTH (settable in Makefile).  Another
          possibility would be to have an array MaxWareHouseLength[]
          initialized in myInitHandles() with differing lengths depending
          on the warehouse
*/

#ifndef NWAREHOUSES
#define NWAREHOUSES    10 /* number of basic warehouses */
#endif /*NWAREHOUSES*/
#define EXTRAWAREHOUSES 2 /* number of additional warehouses */

#ifndef MAXWAREHOUSELENGTH /* maximum number of handles in any warehouse */
#define MAXWAREHOUSELENGTH      16
#endif /*MAXWAREHOUSELENGTH*/

#define LISTWAREHOUSE   NWAREHOUSES
#define SYMBOLWAREHOUSE (NWAREHOUSES+1)

/*
  Warehouse is a static array (allocated by NewPtr), each element of which is
  the start of a singly linked list of allocated (by the operating system)
  handles of certain fixed sizes, which have been de-allocated by
  mydisphandle() but are not given up to the system in the expectation that
  they will be needed again.  There are warehouses for handles of
  j*sizeof(double), j = 1,...,NWAREHOUSES and for handles whose size is
  sizeof(Symbol) and for handles whose size is that of the default size for a
  LIST. Warehouse[i] contains the index (in array Handles) of the most
  recently disposed of handle in that warehouse.  Handles[Warehouse[i]].next
  is the index of the next in the warehouse, etc.  Warehouse[i] == -1 means
  the warehouse is empty, and the component 'next' of the first handle put in
  a warehouse (the end of the chain) is -1.
*/

static short       *Warehouse; /*Warehouse[NWAREHOUSES+EXTRAWAREHOUSES]*/

/* The following are globals for debugging purposes; see mainpars.y */
short               WarehouseLength[NWAREHOUSES+EXTRAWAREHOUSES];
short               Nwarehouses = NWAREHOUSES+EXTRAWAREHOUSES;

/* Maximum sizes for each warehouse, set by setWarehouseLimits()*/

static unsigned short  WarehouseLimits[NWAREHOUSES+EXTRAWAREHOUSES];

#ifdef COUNTMISSES
long                WarehouseMisses[NWAREHOUSES+EXTRAWAREHOUSES];
long                WarehouseHits[NWAREHOUSES+EXTRAWAREHOUSES];
#endif /*COUNTMISSES*/

#define maxWarehouseLength(I)  WarehouseLimits[I]

/*
  FreeList is the start of a singly linked list of "slots" in array Handles
  which do not have actual handles assigned.  Specifically, if FreeList == i,
  Handles[i] has not been allocated, and Handles[i].next is the index of the
  the next unallocated slot.  The last unallocated slot has 'next' component
  -1, and FreeList == -1 means there are no more unallocated slots.
*/
static short           FreeList = -1;

/* LastHandle is the highest index in Handles that has been used */
static short           LastHandle = -1;

/* MaxHandles is the number of handles that can be used */
static short           MaxHandles = -1;

static void    freeList(short list)
{
	register short  i, next;
	infoblockptr    info;
	char          **h;

	for(i = list;i >= 0;i = next)
	{
		info = Handles + i;
		next = info->next;
		MEMDEC(info->length);
		h = info->realHandle;
		DisposHandle((Handle) h);
		info->realHandle = (char **) 0;
		info->data = (char *) 0;

		/* add to free list */
		info->next = FreeList;
		FreeList = i;
	}
}

/*
   950805 added code to zero WarehouseLength[]
*/

void    collectGarbage()
{
	register short  i;
	infoblockptr    info = Handles;
	char          **h;

	/* discard warehoused handles */
	for(i = 0;i < Nwarehouses;i++)
	{
		freeList(Warehouse[i]);
		Warehouse[i] = -1;
		WarehouseLength[i] = 0;
	}

	/* at this point there should be no warehoused handles */

	for(i = 0;i<LastHandle;i++)
	{
		if((h = (char **) info->realHandle) != (char **) 0)
		{/* unlock all handles in use */
			HUnlock(h);
		}
		info++;
	}

	info = Handles;
	for(i = 0;i<LastHandle;i++)
	{
		if((h = (char **) info->realHandle) != (char **) 0)
		{ /* MoveHHi all handles in use and relock */
			MoveHHi(h);
			HLock(h);
			info->data = (char *) (*h + DATAOFFSET);
		}
		info++;
	}
}

void **myFindHandle(void ** h)
{
	infoblockptr   info = (infoblockptr) h;
	char        ***self = (char ***) (info->data - DATAOFFSET);

	if(info->realHandle != *self)
	{
		myerrorout("WARNING: Probable memory corruption");
		h = (void **) 0;
	}

	return (h);
}

/* get handle to needed bytes */
void          **myNewHandle(n)
long            n;
{
	long            needed = n + DATAOFFSET;
	char          **h = (char **) 0;
	char         ***self;
	infoblockptr    info;
	short           handleIndex, whichWarehouse, next;
	OSErr           err = memFullErr;

	if(n == SYMBOLSIZE)
	{
		whichWarehouse = SYMBOLWAREHOUSE;
	}
	else if(n == LISTSIZE)
	{
		whichWarehouse = LISTWAREHOUSE;
	}
	else if(MODDOUBLE(n) == 0 &&
			(whichWarehouse = DIVDOUBLE(n)) <= NWAREHOUSES)
	{
		whichWarehouse--;
	}
	else
	{
		whichWarehouse = -1;
	}
	if(whichWarehouse < 0 || (handleIndex = Warehouse[whichWarehouse]) < 0)
	{ /* none warehoused of right size */
		handleIndex = FreeList;
		whichWarehouse = -1;
	}
	if(handleIndex<0)
	{
		myerrorout("Compacting memory, please stand by");
		collectGarbage();
		handleIndex = FreeList;
		if(handleIndex < 0)
		{
			myAlert("Cannot allocate more memory; continuation impossible");
			FatalError = 1;
			return ((void **) 0);
		}
	}
	if(handleIndex == LastHandle)
	{
		LastHandle++;
	}
	info = Handles + handleIndex;
	if(whichWarehouse<0)
	{ /* was not warehoused handle */
		if (isRoom(needed))
		{
			if(!UseResrvMem || (ResrvMem(needed),err = MemErr) == noErr)
			{
				h = (char **) NewHandle(needed);
				err = MemErr;
			}
		}
		if(err != noErr)
		{ /* try again, after allowing system to compact relocatable blocks */
			myerrorout("Compacting memory, please stand by");
			collectGarbage();
			if(!UseResrvMem || (ResrvMem(needed),err = MemErr) == noErr)
			{
				h = (char **) NewHandle(needed);
				err = MemErr;
			}
		}
		if(err == noErr)
		{
			MEMINC(n);
		}
	} /*if(whichWarehouse<0)*/
	else
	{ /* was warehoused handle */
		err = noErr;
		h = info->realHandle;
		WarehouseLength[whichWarehouse]--;
	}

	if(err == noErr)
	{ /* got handle; now initialize */
		HNoPurge(h);
		HLock(h);
		info->data = *h + DATAOFFSET;
		info->realHandle = h;
		info->length = n;
		self = (char ***) *h;
		*self = h; /* used to validate handle */
		h = (char **) info;
		next = info->next;
		if(whichWarehouse >= 0)
		{/* was warehoused handle */
			Warehouse[whichWarehouse] = next;
		}
		else
		{
			FreeList = next;
		}
		info->next = 0;
	} /* if(err == noErr) */
	else
	{
		h = (char **) 0;
	}

	return ((void **) h);
}

/*
   dispose of a handle

   950804: added code to check and increment WarehouseLength[whichWarehouse]
*/

void myDisposHandle(h1)
void          **h1;
{
	infoblockptr info = (infoblockptr) h1;
	short        handleIndex = info - Handles;
	short        whichWarehouse, k;
	long         length;
	Handle       h;
	char      ***self;

	if(h1 != (void **) 0)
	{
		h = (Handle) info->realHandle;
		/* make sure myFindHandle would report not found */
		self = (char ***) (*h);
		*self = (char **) 0;
		info->data = (char *) 0;
		length = info->length;

		/* warehouse Symbols, standard lists and small REAL vectors */
		if(length == SYMBOLSIZE)
		{
			whichWarehouse = SYMBOLWAREHOUSE;
		}
		else if(length == LISTSIZE)
		{
			whichWarehouse = LISTWAREHOUSE;
		}
		else if(MODDOUBLE(length) == 0 &&
				(whichWarehouse = DIVDOUBLE(length)) <= NWAREHOUSES)
		{
			whichWarehouse--;
		}
		else
		{
			whichWarehouse = -1;
		}
		if(whichWarehouse >= 0 &&
		   WarehouseLength[whichWarehouse] < maxWarehouseLength(whichWarehouse))
		{
			if((k = Warehouse[whichWarehouse]) >= 0)
			{ /* warehouse is not empty */
				info->next = k;
			}
			else
			{
				info->next = -1;
			}
			Warehouse[whichWarehouse] = handleIndex;
			WarehouseLength[whichWarehouse]++;
			HUnlock(h);
		}
		else
		{ /* dispose of everything else */
			MEMDEC(info->length);
			info->length = 0;
			info->realHandle = (char **) 0;
			info->next = FreeList;
			FreeList = handleIndex;
			DisposHandle(h);
		}
	}
}

/* make a handle bigger */

void          **mySetHandleSize(h1, newn)
void          **h1;
long            newn;
{
	long            needed = newn + DATAOFFSET;
	long            oldn = myGetHandleSize(h1);
	infoblockptr    info = (infoblockptr) h1;
	Handle          h, newh;
	OSErr           err = memFullErr;
	WHERE("mySetHandleSize");

	if(newn != oldn)
	{ /* only do something if desired size differs */
		h = (Handle) info->realHandle;
		HUnlock(h);
#ifdef _UNDEFINED_
		if (isRoom(needed))
		{
			if(!UseResrvMem || (ResrvMem(needed),err = MemErr) == noErr)
			{
				SetHandleSize(h, needed);
				err = MemErr;
			}
		}
		if(err != noErr)
		{/* try again, after allowing system to compact relocatable blocks */
			myerrorout("Compacting memory, please stand by");
			collectGarbage();
			if(!UseResrvMem || (ResrvMem(needed),err = MemErr) == noErr)
			{
				SetHandleSize(h, needed);
				err = MemErr;
			}
		} /*if(err == noErr)*/
		if(err == noErr)
		{
			HNoPurge(h);
			HLock(h);
			MEMDEC(oldn);
			MEMINC(newn);
			/* info->realHandle should still be correct */
			info->data = (char *) *h + DATAOFFSET;
	 		info->length = newn; /* set new length */
			h = (Handle) info;
		} /*if(err == noErr)*/
#else /*_UNDEFINED_*/
		newh = (Handle) myNewHandle(newn);
		if(newh != (Handle) 0)
		{
			BlockMove((Ptr) *h1, (Ptr) *newh, (oldn < newn) ? oldn : newn);
			myDisposHandle(h1);
			h = newh;
		}
#endif /*_UNDEFINED_*/
		else
		{	/* get rid of handle */
			myDisposHandle(h1);
			h = (Handle) 0;
		}
	} /*if(newn != oldn)*/
	else
	{
		h = (Handle) h1;
	}

	return ((void **) h);

}

void          **myDupHandle(h)
void          **h;
{
	long        length = myGetHandleSize(h);
	char      **h1 = (char **) h;
	char      **h2 = (char **) myNewHandle(length);
	char       *new, *old;

	if(h2 != (void **) 0)
	{
		old = *h1;
		new = *h2;

		while(length-- > 0)
		{
			*new++ = *old++;
		}
	}

	return ((void **) h2);
}


long myGetHandleSize(void ** h)
{
	infoblockptr    info = (infoblockptr) h;

	return (info->length);
}

void myHLock(void **h)
{
	infoblockptr   info = (infoblockptr) h;

	HLock((Handle) info->realHandle);
}

void myHUnlock(void **h)
{
	infoblockptr   info = (infoblockptr) h;

	HUnlock((Handle) info->realHandle);
}

/* Note: memory allocated by myNewPtr is not counted */
void *myNewPtr(long n)
{
	Ptr         ptr = NewPtr(n);

	if(MemErr != noErr)
	{
		ptr = (Ptr) 0;
	}
	return ((void *) ptr);
}

void myDisposePtr(void *p)
{
	DisposPtr((Ptr) p);
}

/*
   Added 950809
*/

void setWarehouseLimits(long standard)
{
	int         i;
	
	for (i=0;i < Nwarehouses;i++)
	{
		WarehouseLimits[i] = standard;
#ifdef COUNTMISSES
		WarehouseHits[i] = WarehouseMisses[i] = 0;
#endif /*COUNTMISSES*/
	} /*for (i=0;i < Nwarehouses;i++)*/
	WarehouseLimits[0] = 10*standard;
	WarehouseLimits[SYMBOLWAREHOUSE] = 4*standard;
} /*setWarehouseLimits()*/

/*
   950805 added code to initialize WarehouseLength[]
*/

long myInitHandles(short maxhandles)
{
	infoblockptr       info;
	short              i;

	if(Handles == (infoblockptr) 0)
	{ /* test should be unnecessary but play it safe */
		MaxHandles = maxhandles;
		Handles = (infoblockptr) myNewPtr((size_t) maxhandles * sizeof(infoblock));
		if(Handles == (infoblockptr) 0)
		{
			FatalError = 1;
			return (0);
		}
		info = Handles;
		for(i = 0;i<maxhandles;i++, info++)
		{
			info->data = (char *) 0;
			info->length = 0;
			info->realHandle = (char **) 0;
#ifdef _UNDEFINED_
			if(i == 0)
			{
				info->next = -1;
			}
			else
			{
				(info-1)->next = i;
			}
#else /*_UNDEFINED_*/
			info->next = (i+1 < maxhandles) ? i+1 : -1;
#endif /*_UNDEFINED_*/
		}
		FreeList = 0; /* index of first free handle */
		LastHandle = 0; /* index of last handle allocated */
		Warehouse = (short *) NewPtr((size_t)(Nwarehouses * sizeof(short)));
		if(MemErr != noErr)
		{
			FatalError = 1;
			return (0);
		}

		for(i = 0;i < Nwarehouses;i++)
		{
			Warehouse[i] = -1;
			WarehouseLength[i] = 0;
		}
		setWarehouseLimits(MAXWAREHOUSELENGTH);
 	} /*if(Handles == (infoblockptr) 0)*/
	return (1);
} /*myInitHandles()*/

/* dispose of all handles */
void myDispAll(void)
{
	infoblockptr       info = Handles;
	short              i;

	if(info != (infoblockptr) 0)
	{
		for(i=0;i<MaxHandles;i++, info++)
		{
			if(info->realHandle != (char **) 0)
			{
				DisposHandle((Handle) info->realHandle);
			}
		}
		myDisposePtr((void *) Handles);
		Handles = (infoblockptr) 0;
	}
}
