/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

/*
  Header unifying Mac interface
*/
#ifndef MACIFACEH__
#define MACIFACEH__

#undef EXTERN
#undef COMMA
#define COMMA ,

#ifdef MAIN__
#define EXTERN
#define INIT(VAL) = VAL
#define INITDIM(N) N
#define INITARRAY(VALS) = { VALS }
#else
#define EXTERN extern
#define INIT(VAL)
#define INITDIM(N)
#define INITARRAY(VALS)
#endif /*MAIN__*/

#ifdef MPW
#define PtoCstr p2cstr
#define CtoPstr c2pstr
#include <Dialogs.h>	/* pulls in Windows.h, Types.h, QuickDraw.h */
#include <Events.h>
#include <Desk.h>
#include <Scrap.h>
#include <SegLoad.h>
#include <Memory.h>
#include <DiskInit.h>
#include <OSEvents.h>
#include <Traps.h>
#include <Errors.h>
#include <Resources.h>
#include <Fonts.h>
#include <ToolUtils.h>
#include <Controls.h>
#include <Packages.h>
#include <Files.h>
#include <Menus.h>
#include <strings.h>
#include <string.h>
#include <PrintTraps.h>
#include <TextEdit.h>
#else
#include <WindowMgr.h>	/* pulls in MacTypes.h, QuickDraw.h */
#include <ControlMgr.h>
#include <StdFilePkg.h>
#include <FileMgr.h>
#include <MenuMgr.h>
#include <string.h>
#include <TextEdit.h>
#include <OSUtil.h>
#include <PrintTraps.h>
#endif

#ifndef MACTYPEDEF

#ifndef SIZEINTEGER
#undef  SIZELONGINT
#ifdef MPW
#define SIZEINTEGER short
#define SIZELONGINT long
#else
#define SIZEINTEGER int
#define SIZELONGINT long
#endif /*MPW*/
#endif /*SIZEINTEGER*/

typedef SIZEINTEGER       Integer;
typedef SIZELONGINT       LongInt;

typedef LongInt Size;

#define MACTYPEDEF /* indicates preceding typedefs done */
#endif /*MACTYPEDEF*/

#include "TransSkel.h"
#ifndef CMDSLEEP
#define CMDSLEEP   3L
#endif /*CMDSLEEP*/

#ifndef BATCHSLEEP
#define BATCHSLEEP 0L  /* minimizes time lost by WaitNextEvent() */
#endif /*BATCHSLEEP*/

#ifndef INTERRUPTSLEEP
#define INTERRUPTSLEEP 0L  /* minimizes time lost by WaitNextEvent() */
#endif /*INTERRUPTSLEEP*/

#ifndef BACKGROUNDSLEEP
#define BACKGROUNDSLEEP 120L  /* wait time while in background */
#endif /*BACKGROUNDSLEEP*/

EXTERN LongInt     Sleep INIT(CMDSLEEP);
EXTERN LongInt     CmdSleep INIT(CMDSLEEP);
EXTERN LongInt     BatchSleep INIT(BATCHSLEEP);
EXTERN LongInt     InterruptSleep INIT(INTERRUPTSLEEP);
EXTERN LongInt     BackgroundSleep INIT(BACKGROUNDSLEEP);

#ifndef nil
# define	nil		0L
#endif /*nil*/

/*
	Globals related to non-interactive mode
	UseWindows indicates the status of non-interactive mode
	If non-zero all calles to print to windows are short circuited
	All graphics will be DUMB
	Value < 0 means during startup sequence
	Value > 0 means use windows
	Value of 0 means don't use windows
*/

EXTERN Integer UseWindows INIT(0);
EXTERN Boolean SaveOnQuit INIT(true);

/*
	Menu handles.
*/

EXTERN MenuHandle      AppleMenu;
EXTERN MenuHandle      FileMenu;
EXTERN MenuHandle      EditMenu;
EXTERN MenuHandle      WindowMenu;
EXTERN MenuHandle      CommandMenu;
EXTERN MenuHandle      OptionsMenu;
EXTERN MenuHandle      FontMenu;
EXTERN MenuHandle      FontSizeMenu;

EXTERN SFReply         Reply;

/*
	Most recent volume IDs
*/

EXTERN Integer         KeepVolume INIT(0); /* goes with save() */
EXTERN Integer         RestoreVolume INIT(0);
EXTERN Integer         SpoolVolume INIT(0);
EXTERN Integer         BatchVolume INIT(0);
EXTERN Integer         ReadVolume INIT(0);
EXTERN Integer         PrintVolume INIT(0);
EXTERN Integer         SaveVolume INIT(0); /* goes with Save Window */
EXTERN Integer         HelpVolume INIT(0);
EXTERN Integer         PlotVolume INIT(0);

EXTERN Integer         HomeVolume INIT(0);
EXTERN LongInt         HomeDirectory INIT(0);

/*
	Global print record handle
*/
#define TEXTPRINT      1
#define GRAPHPRINT     2
EXTERN THPrint         PrintHdl[2] INITARRAY(0 COMMA 0);

/*
	Graph windows
*/
#define NGRAPHS  4

#ifdef MPW
EXTERN WindowPtr       GraphWind[INITDIM(NGRAPHS)] INITARRAY((WindowPtr) 0 COMMA\
		(WindowPtr) 0 COMMA (WindowPtr) 0 COMMA (WindowPtr) 0);
#else
EXTERN WindowPtr       GraphWind[INITDIM(NGRAPHS)] INITARRAY((WindowPtr) 0 COMMA (WindowPtr) 0 COMMA (WindowPtr) 0 COMMA (WindowPtr) 0);
#endif
EXTERN char           *PanelTitle INIT("\pPanel of Graphs");

EXTERN WindowPtr       PanelWind INIT(0);
EXTERN WStateData      GraphWState[INITDIM(NGRAPHS)];
EXTERN Integer         GraphWInZoom[INITDIM(NGRAPHS)] INITARRAY(inZoomOut COMMA inZoomOut COMMA inZoomOut COMMA inZoomOut);

EXTERN PicHandle       myPic[INITDIM(NGRAPHS)] INITARRAY(0 COMMA 0 COMMA 0 COMMA 0);
EXTERN PicHandle       PanelPic INIT(0);

EXTERN Integer         ThisWindow INIT(-1);

EXTERN Boolean         GraphWindPaused INIT(false);

/*
	The following are related to the introduction of multiple command windows
	Most of the items are just copies of the analogous globals when the window
	is in front.
*/
typedef struct cmdWindowInfo
{
	WindowPtr          cmdWind;
	TEHandle           teCmd;
	Integer            menuItem;    /* item in Windows menu */
	Integer            cmdVRefNum;  /* vRefNum for associated file */
	Boolean            cmdDirty;    /* window used flag */
	Boolean            cmdEditable; /* user can change window flag */
	ControlHandle      cmdScroll;
	Integer            cmdLine;
	Integer            cmdStrPos;
	Integer            undoStatus;
	Integer            undoPlace;
	Integer            pasteLength;
	LongInt            oldCmdStrPos;
	LongInt            lastTypingPos;
	LongInt            lastLength;
	LongInt            lastCmdStrPos;
	Handle             cmdScrap;
} cmdWindowInfo, *cmdWindowInfoPtr, **cmdWindowInfoHandle;

#ifndef CMDWINDOWLIMIT
#define CMDWINDOWLIMIT   32000  /* threshhold for window to be full */
#endif /*CMDWINDOWLIMIT*/

#ifndef CMDWINDOWNEWSIZE
#define CMDWINDOWNEWSIZE 25000  /* target size after trimming */
#endif /*CMDWINDOWNEWSIZE*/

#define MAXWINDOWS     9

EXTERN Integer         CurrentWindow INIT(-1);
EXTERN Integer         Nwindows INIT(-1);
EXTERN cmdWindowInfo   CmdWindows[INITDIM(MAXWINDOWS)];


/*
	Globals related to the Command Window - a simple text editing window
*/

EXTERN WindowPtr       CmdWind;     /* Pointer to active command window */
EXTERN TEHandle        teCmd;		/* handle to text window TextEdit record */
EXTERN Boolean         CmdDirty;    /* window changed after first prompt */
EXTERN Boolean         CmdEditable; /* is window editable */
EXTERN ControlHandle   CmdScroll;	/* command window scroll bar */
EXTERN Integer         CmdLine;	/* line currently at top of window */
EXTERN Integer         CmdScrollLines;	/* number of lines not visible */
EXTERN Integer         HalfPage;	/* number of lines in half a window */
EXTERN Integer         CmdInsertLine;	/* line of current insertion */
EXTERN Integer         ViewLines;	/* number of lines on screen */
EXTERN LongInt         CmdStrPos;	/* position to begin looking for commands*/

/*
	Globals related to the Batch (non interactive mode) window
*/

EXTERN WindowPtr       BatchWind INIT(0);
EXTERN DialogPtr       BatchDialog INIT(0);

EXTERN Integer         Running INIT(0); /* !0 means under control of yyparse */

/*
	Globals and defines related to undoing and redoing
*/
EXTERN Integer         UndoStatus INIT(0); /* current undo status */

/* possible values for UndoStatus */
#define DELETEUNDO    1
#define OUTPUTUNDO    2
#define PASTEUNDO     3
#define CUTUNDO       4
#define TYPINGUNDO    5
#define CANTUNDO      0
#define DELETEREDO   -1 /* not currently used */
#define OUTPUTREDO   -2
#define PASTEREDO    -3 /* not currently used */
#define CUTREDO      -4 /* not currently used */
#define TYPINGREDO   -5

/*
	When UndoStatus == OUTPUTUNDO, LastLength is the length of the buffer,
	less 1 for the trailing newline, just before a command line is executed,
	and LastCmdStrPos is the value of CmdStrPos at that time.  
*/
EXTERN LongInt         LastLength INIT(-1);
EXTERN LongInt         LastCmdStrPos INIT(-1);
EXTERN Integer         UndoPlace INIT(-1); /* start of change */
EXTERN Integer         PasteLength INIT(-1); /* length of paste */
EXTERN Handle          CmdScrap = (Handle) 0; /* handle to save stuff*/

/* OldCmdStrPos is CmdStrPos at time of a deletion, replacement, cut, or paste*/
EXTERN LongInt         OldCmdStrPos INIT(-1);
EXTERN LongInt         LastTypingPos INIT(-1);

/* Defines and globals related to fonts */
#define MACANOVAFONT     16000
#define MACANOVAFONTNAME "\pMcAOVMonaco"

#define REGULARFONT      monaco 
#define REGULARFONTNAME  "\pMonaco"

#ifndef CMDFONT
#define CMDFONT MACANOVAFONT
#endif /*CMDFONT*/

#ifndef CMDFONTSIZE
#define CMDFONTSIZE 9
#endif /*CMDFONTSIZE*/

#ifndef CMDFONTNAME
#define CMDFONTNAME MACANOVAFONTNAME
#endif /*CMDFONTNAME*/

#ifndef GRAPHFONT
#define GRAPHFONT MACANOVAFONT
#endif /*GRAPHFONT*/

#ifndef GRAPHFONTSIZE
#define GRAPHFONTSIZE 9
#endif /*GRAPHFONTSIZE*/

#ifndef GRAPHFONTNAME
#define GRAPHFONTNAME MACANOVAFONTNAME
#endif /*GRAPHFONTNAME*/

#ifndef PRINTFONT
#define PRINTFONT REGULARFONT
#endif /*PRINTFONT*/

#ifndef PRINTFONTSIZE
#define PRINTFONTSIZE 9
#endif /*PRINTFONTSIZE*/

#ifndef PRINTFONTNAME
#define PRINTFONTNAME REGULARFONTNAME
#endif /*PRINTFONTNAME*/

EXTERN Integer         CmdFont INIT(CMDFONT);
EXTERN Integer         CmdFontSize INIT(CMDFONTSIZE);
EXTERN Str27           CmdFontName INIT(CMDFONTNAME);

EXTERN Integer         GraphFont INIT(GRAPHFONT);
EXTERN Integer         GraphFontSize INIT(GRAPHFONTSIZE);
EXTERN Str27           GraphFontName INIT(GRAPHFONTNAME);

EXTERN Integer         PrintFont INIT(PRINTFONT);
EXTERN Integer         PrintFontSize INIT(PRINTFONTSIZE);
EXTERN Str27           PrintFontName INIT(PRINTFONTNAME);

/* Defines and Globals related to files */
#ifndef KEEPTYPE
#define KEEPTYPE 'S000'  /* File type for save() file */
#endif /*KEEPTYPE*/

#ifndef ASCIIKEEPTYPE
#define ASCIIKEEPTYPE 'Sasc' /*File type for asciisave() file */
#endif /*KEEPTYPE*/

#ifndef CREATOR
#define CREATOR  'mat2' /*Creator for save(), asciisave(), Cmd window files*/
#endif /*CREATOR*/

#ifndef TEXTCREATOR
#define TEXTCREATOR  'ttxt'  /* Creator for output files */
#endif /*TEXTCREATOR*/

#ifndef MACANOVAINI
#define MACANOVAINI "Macanova.ini"  /*Startup initialization file*/
#endif /*MACANOVAINI*/

#ifndef HELPNAME
#define HELPNAME "Macanova.hlp"  /*Default Help file name*/
#endif  /*HELPNAME*/

EXTERN OSType          ASCIIKEEPFILETYPE INIT(ASCIIKEEPTYPE);
EXTERN OSType          KEEPFILETYPE INIT(KEEPTYPE);
EXTERN OSType          KEEPCREATOR INIT(CREATOR);

/* Prototypes removed because they were all in macProto.h */

/* Cursors */
EXTERN Cursor          IBEAM;
EXTERN Cursor          WATCH;
EXTERN Cursor          CROSS;

/* Globals used by myDialogFilter */
EXTERN Boolean         DialogEdit INIT(false);
EXTERN Integer         NDialogButtons INIT(0);
EXTERN unsigned char   ButtonChars[4] INITARRAY(0 COMMA 0 COMMA 0 COMMA 0);

/*
	Handles for Commands associated with Command Menu
*/

#define NCOMMANDS        8  /* number of commands in menu*/
EXTERN char          **CommandItems[INITDIM(NCOMMANDS)];

#include "macResource.h"  /* resource ID's */

/* defines and enum declarations only after this point */

#define LF            10
#define CR            13

/* key codes */
#define HomeKey        1
#define Enter          3
#define EndKey         4
#define HelpKey        5
#define Backspace      8
#define TabKey         9
#define Pageup        11
#define Pagedown      12
#define Return        13
#define ClearKey      27
#define EscapeKey     27
#define LeftArrow     28
#define RightArrow    29
#define UpArrow       30
#define DownArrow     31
#define Space         32  /*' '*/
#define Period        46  /*'.'*/
#define Zero          48  /*'0'*/
#define DeleteKey    127

#define LineStart     15  /* MacAnova code to move to start of line */

#define Newline       CR
/*
	Standard option number key codes
*/

#define Option0     0xbc
#define Option1     0xc1
#define Option2     0xaa
#define Option3     0xa3
#define Option4     0xa2
#define Option5     0xb0
#define Option6     0xa4
#define Option7     0xa6
#define Option8     0xa5
#define Option9     0xbb

/*
	Standard lower case option key character codes
	Dead key codes are coded as the code of the character
	produced by the sequence Option letter, letter (e.g. option e, e)
*/

#define OptionA     0x8c
#define OptionB     0xba
#define OptionC     0x8d
#define OptionD     0xb6
#define OptionE     0x8e
#define OptionF     0xc4
#define OptionG     0xa9
#define OptionH     0xfa
#define OptionI     0x94
#define OptionJ     0xc6
#define OptionK     0xfb
#define OptionL     0xc2
#define OptionM     0xb5
#define OptionN     0x96
#define OptionO     0xbf
#define OptionP     0xb9
#define OptionQ     0xcf
#define OptionR     0xa8
#define OptionS     0xa7
#define OptionT     0xa0
#define OptionU     0x9f
#define OptionV     0xc3
#define OptionW     0xb7
#define OptionX     0xc5
#define OptionY     0xb4
#define OptionZ     0xbd


#define COMMANDKEYCODE   0x37 /*Virtual key code for Command key */
#define OPTIONKEYCODE    0x3a /*Virtual key code for Option key */
#define BKEYCODE         0x0b /*Virtual key code for 'B' key */
#define NOTCOMMAND       0xff /*Not a virtual key code */

#define FUNCTIONKEY 0x10 /* character code for F1 through F15 */

/*
	The following constants are the virtual key codes for the
	function keys on the Apple Extended keyboard as pictured on page
	2-43 of IM, Macintosh Toolbox Essentials
*/

#define F1          0x7a
#define F2          0x78
#define F3          0x63
#define F4          0x76
#define F5          0x60
#define F6          0x61
#define F7          0x62
#define F8          0x64
#define F9          0x65
#define F10         0x6d
#define F11         0x67
#define F12         0x6F
#define F13         0x69
#define F14         0x6b
#define F15         0x71

#define UNDOKEY     F1
#define CUTKEY      F2
#define COPYKEY     F3
#define PASTEKEY    F4

/*
	Information related to windows
*/

/* offsets for view rectangle in command windows */
#define TOPOFFSET      2
#define LEFTOFFSET     4
#define BOTTOMOFFSET   2
#define RIGHTOFFSET   19

/* offset of top of window is windno * STACKOFFSET */
#define STACKOFFSET   10

/*
	information related to menus
*/

typedef enum
{ /* item numbers for Apple Menu */
	aboutmacanova = 1,
	helpit
}	appleItems;

typedef enum
{ /* Item numbers for File Menu */
	openit = 1,
	saveit,
	saveitas,
	/* --- */
	pagesetup = 5,
	printit,
	/* --- */
	interrupt = 8,
	go_on,
	/* --- */
	restoreit = 11,
	keepit,
	keepitas,
	/* --- */
	quit = 15
}               fileItems;

typedef enum
{ /* item numbers for Edit Menu */
	undoit = 1,
	/* --- */
	cutit = 3,
	copyit,
	pastit,
	/* --- */
	copytoend = 7
}               editItems;

typedef enum
{/* item numbers for Window Menu */
	hideit = 1,
	closeit,
	newwind,
	/* --- */
	graph1 = 5,
	graph2,
	graph3,
	graph4,
	panel,
	/* --- */
	cmd1 = 11,
	cmd2,
	cmd3,
	cmd4,
	cmd5,
	cmd6,
	cmd7,
	cmd8,
	cmd9, /* must be exactly MAXWINDOWS of these */
	/* --- */
	gototop = 21, /* cmd1 + MAXWINDOWS + 1 */
	gotobottom,
	pageup,
	pagedown
} windItems;

typedef enum
{/* item numbers for Command Menu */
	editcommands = 1,
	/* --- */
	command1 = 3,
	command2,
	command3,
	command4,
	command5,
	command6,
	command7,
	command8,   /* Currently only NCOMMANDS == 8 are used */
	command9,
	command10,
	command11,
	command12,
	command13,
	command14,
	command15
}               commandItems;

typedef enum
{/* item numbers for Options Menu */
	insig = 1,
	iformats,
	iseeds,
	iangles,
	iboptions
}               optionsItems;

/* constants for MacOpen */
#define READIT  1
#define WRITEIT 2

/* end of globals and definitions */

#undef	INIT
#undef	INITDIM
#undef	INITARRAY
#undef  COMMA
#undef  EXTERN

#endif /*MACIFACEH__*/
