/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#include "globals.h"
#include "matdat.h"
#include "macIface.h"

extern void     myprint();
extern void     myerrorout();
extern void     myeol();

void            fmyprint();
void            fmyeol();
void            putprompt();
long            isfilename();

#define BEL '\007'
#define TILDE '~'

/*
   Reads a line, including the terminating '\n', from fp into *LINE.
   If the line is longer than MAXLINE, it reads MAXLINE-1 characters,
   adds a '\n'  and skips to just past the next '\n'.

   If an EOF is encountered during the read, it returns EOF; otherwise
   it returns 1.  This does not mean that no characters were read, since
   the file might not be terminated by '\n'.

   The calling function can determine how many characters were read by
   strlen(*LINE).
*/
static int fileToLINE(FILE * fp)
{
	int         c;
	long        n = 0, nc = MAXLINE;
	char       *buffer = *LINE;
	WHERE("fileToLINE");

	*buffer = '\0';
	c = getc(fp);
	if(c != EOF)
	{
		do
		{
			*buffer++ = c;
			nc--;
			n++;
		} while(nc > 0 && c != CR && c != NL && (c = getc(fp)) != EOF);

		if(c != EOF)
		{
			buffer--;
		}
		*buffer++ = '\n';
		*buffer = '\0';

		while(c != CR && c != NL && c != EOF && n++)
		{						/* read remainder of the line */
			c = getc(fp);
		}
	} /*if(c != EOF)*/

	if(n > MAXLINE)
	{
		sprintf(OUTSTR,
				"WARNING: line longer than %ld characters; last %ld discarded",
				(long) MAXLINE, n - MAXLINE);
		putOUTSTR();
	}
	return ((c == EOF) ? EOF : 1);
}

#ifndef OUTLINESIZE
#define OUTLINESIZE   400  /*size of output line buffer*/
#endif /*OUTLINESIZE*/

/* OutLineBuf is allocated by mygetpointer in main() in macMain.c */

extern char      *OutLineBuf; /* buffer for building output lines */
static long       OutLinePlace = 0;

/*
  Write line of output to file fp which may specify command window
  If writing to console, interpret leading '\007's as BEL's
  If spooling is active, write line to spool file
  If line starts with "ERROR:", ring bell and count error
*/


static void lineToWindow(char * msg,  long msgLength)
{
	long            nLines, cutLength;
	long            teLength = (**teCmd).teLength;
	long            windno;
	long            isOutLine = (msg == OutLineBuf);
	Str255          wTitle;
	char           *outstr = (char *) wTitle;
	char           *dashes = "\p\n----------------------------------\n";
	WHERE("lineToWindow");

	while(*msg == BEL)
	{						/* turn leading BEL's into beeps */
		if(UseWindows > 0)
		{
			mybeep(1);
		}
		msg++;
	}

	if (SPOOLFILE != (FILE *) 0)
	{
		fputs(msg, SPOOLFILE);
	}

	if(UseWindows > 0 && msgLength > 0)
	{
		if(UndoStatus != CANTUNDO && UndoStatus != OUTPUTUNDO)
		{
			clearUndoInfo();
		}

		if(msgLength > CMDWINDOWLIMIT)
		{
			sprintf(outstr,
					"WARNING: output line longer than %ld characters is truncated",
					(long) CMDWINDOWLIMIT);
			myAlert(outstr);
			msgLength = CMDWINDOWNEWSIZE;
		}

		if (teLength + msgLength > CMDWINDOWLIMIT)
		{
			myAlert("Output window contents are too long; saving window");

			DoFile(saveitas);

			for(windno=0;windno<MAXWINDOWS;windno++)
			{
				if(CmdWindows[windno].cmdWind == (WindowPtr) 0)
				{
					break;
				}
			}

			if(windno < MAXWINDOWS)
			{ /* we have room for a new window */
				TESetSelect(teLength, teLength, teCmd);
				TEInsert(dashes+1, dashes[0], teCmd);
				outstr = "\pContinued in new window\n";
				TEInsert(outstr+1, outstr[0], teCmd);
				CmdStrPos = (**teCmd).teLength;
				ScrollToInsertPt();
				GetWTitle(CmdWind, wTitle);

				clearUndoInfo();

				DoWind(-newwind); /* create new window without prompt*/

				TESetSelect(0, 0, teCmd);
				outstr = "\pOutput continued from ";
				TEInsert(outstr+1, outstr[0], teCmd);
				TEInsert(wTitle+1, wTitle[0], teCmd);
				outstr = dashes;
				TEInsert(outstr+1, outstr[0], teCmd);
			} /*if(windno < MAXWINDOWS)*/
			else
			{ /* no room, trim off top */
				CmdInsertLine = 1;
				TESetSelect(0L, 0L, teCmd);
				ScrollToInsertPt();
				
				cutLength = msgLength + teLength - CMDWINDOWNEWSIZE;
				sprintf(outstr,
						"Cannot open a new window\n%ld characters will be trimmed off the top",
						cutLength);
				myAlert(outstr);

				TESetSelect(0L, cutLength, teCmd);
				TECut(teCmd);
				macUpdate(CmdWind);
				if(UndoStatus == OUTPUTUNDO)
				{
					LastLength -= cutLength;
					LastCmdStrPos -= cutLength;
					if(LastCmdStrPos < 0)
					{
						clearUndoInfo();
					}
				} /*if(UndoStatus == OUTPUTUNDO)*/
			} /*if(windno < MAXWINDOWS){}else{}*/
		} /*if ((**teCmd).teLength > CMDWINDOWLIMIT)*/

		TESetSelect((**teCmd).teLength, (**teCmd).teLength, teCmd);
		teLength = (**teCmd).teLength;
		TEInsert(msg, msgLength, teCmd);
		nLines = (**teCmd).nLines;
		CmdInsertLine = nLines;
		CmdStrPos = (**teCmd).teLength;
		CmdScrollLines = nLines + 1 - ViewLines;
		SetCtlMax(CmdScroll, (CmdScrollLines > 0 ? CmdScrollLines : 0));
		CmdWindows[CurrentWindow].cmdDirty = CmdDirty = true;
	} /*if(UseWindows > 0 && msgLength > 0)*/

	if(strncmp("ERROR:",msg, (int) 6) == 0)
	{
		if(MAXERRORS1)
		{
			NERRORS++;
		}

		mybeep(1);
		if(!UseWindows && strncmp(" too many errors", msg+6, 16) != 0)
		{
			if (!twoChoiceAlert("Go On", "Quit", "gq", msg))
			{
				if (isOutLine)
				{
					OutLinePlace = 0;
					OutLineBuf[0] = '\0';
				}
				if(!quitAlert('e'))
				{
					INTERRUPT = INTSET;
				}
			}
		}
	} /*if(strncmp("ERROR:",msg, (int) 6) == 0)*/

	if (isOutLine)
	{
		OutLinePlace = 0;
		OutLineBuf[0] = '\0';
	}
} /*lineToWindow()*/

/*
	If fp != STDOUT, put msg to file fp; otherwise add msg to OutLineBuf and flush
	*OutLineBuf if last character is CR (Mac end of line)
*/

void fmyprint(char *msg,FILE *fp)
{
	long            i, msgLength;
	char           *outline, c;
	WHERE("fmyprint");

	if(fp == STDOUT)
	{
		/* collect line in OutLineBuf */
		msgLength = strlen(msg);
		if (OutLinePlace + msgLength > OUTLINESIZE)
		{
			lineToWindow(OutLineBuf, OutLinePlace);
			if(msgLength > OUTLINESIZE)
			{
				lineToWindow(msg, msgLength);
				msgLength = 0;
			}
		}
		outline = OutLineBuf + OutLinePlace;
		for (i = 0; i < msgLength; i++)
		{
			c = *msg++;
			*outline++ = (c == NL) ? CR : c;
		}
		*outline = '\0';
		OutLinePlace += msgLength;
		if (OutLinePlace > 0 && OutLineBuf[OutLinePlace-1] == CR)
		{
			lineToWindow(OutLineBuf, OutLinePlace);
		}
	} /*if(fp == STDOUT)*/
	else
	{
		fputs(msg, fp);
	} /*if(fp == STDOUT){}else{}*/
} /*fmyprint()*/

/*
  terminate output line and pause if writing the console and screen is full
  (no pausing on Mac)
*/

void fmyeol(fp)
FILE          *fp;
{
	if(fp == STDOUT)
	{
		if (OutLinePlace > 0)
		{ /* flush OutLineBuf */
			lineToWindow(OutLineBuf, OutLinePlace);
		}
		if(UseWindows > 0)
		{
			if(UndoStatus != CANTUNDO && UndoStatus != OUTPUTUNDO)
			{
				clearUndoInfo();
			}
			TESetSelect((**teCmd).teLength, (**teCmd).teLength, teCmd);
			TEKey(Return, teCmd);
			CmdWindows[CurrentWindow].cmdDirty = CmdDirty = true;
			CmdInsertLine = (**teCmd).nLines;
			CmdStrPos = (**teCmd).teLength;
			CmdScrollLines = (**teCmd).nLines + 1 - ViewLines;
			SetCtlMax(CmdScroll, (CmdScrollLines > 0 ? CmdScrollLines : 0));
			if (CmdInsertLine - CmdLine > ViewLines - 1)
			{
				DoScroll(CmdInsertLine - ViewLines - CmdLine + 5);
			}
		} /*if(UseWindows > 0)*/
		if (SPOOLFILE != (FILE *) 0)
		{
			putc(Newline, SPOOLFILE);
		}
		SkelChkOneMaskedEvent (everyEvent, Sleep);
		if(INTERRUPT != INTNOTSET)
		{
			INTERRUPT = PRINTABORT;
		}
	} /*if(fp == STDOUT)*/
	else
	{
		putc(Newline, fp);
	}
} /*fmyeol()*/

void fmyflush(fp)
FILE    *fp;
{
	if(fp != STDOUT)
	{
		fflush(fp);
	}
	else
	{
		lineToWindow(OutLineBuf, OutLinePlace);

		if(SPOOLFILE != (FILE *) 0)
		{
			fflush(SPOOLFILE);
		}
	}
} /*fmyflush()*/

/*
  sound n beeps, but only if not reading batch file and output is to console
*/

void mybeep(n)
long        n;
{
	if(BDEPTH == 0 && ISATTY)
	{
		while(n-- > 0)
		{
			SysBeep(30);
		}
	}
} /*mybeep()*/

/*
	draw title in Console dialog box using system font
*/
static pascal consoleUserItem(WindowPtr theWindow, Integer item)
{
	Rect           r;
	Integer        oldFont = theWindow->txFont;
	Integer        oldSize = theWindow->txSize;
	Integer        itemType;
	Handle         itemHandle;
	char          *msg = "\pEnter up to 500 characters.  Click Done to end";
	Integer        h, v;
	FontInfo       info;
	WHERE("editCommandUserItem");

	TextFont(systemFont);
	TextSize(12);
	GetFontInfo(&info);

	GetDItem((DialogPtr) theWindow, item, &itemType, &itemHandle, &r);

	h = (r.left + r.right - StringWidth((Str255) msg))/2;
	v = r.bottom - info.descent;
	MoveTo(h, v);
	DrawString((Str255) msg);

	TextFont(oldFont);
	TextSize(oldSize);
}

/*
  Fill global LINE with line from FILE *fn
  On Dos and Unix, fn should always be non-null.
  On Mac, STDIN is (FILE *) 0, and LINE is filled from dialog window
*/

static Boolean        HitDone = false;
static unsigned char  LineBuffer[MAXLINE+1];
static unsigned char *Place = (unsigned char *) LineBuffer;

int fillLINE(fn)
FILE        *fn;
{
	GrafPtr               thePort;
	DialogPtr             theDialog;
	Integer               length;
	Rect                  r;
	Handle                itemHandle;
	TEHandle              textH;
	Integer               itemType, itemHit, nButtons = 2;
	Integer               doneItem = cancel;
	Integer               textItem = nButtons + 2, titleItem = textItem+1;
	unsigned char        *pc;
	Integer               teLength;
	SignedByte            flags;
	WHERE("fillLINE");

	if(fn != STDIN)
	{
		return (fileToLINE(fn));
	} /*if(fn != STDIN)*/
	else
	{ /*
		Use dialog window or what was read previously
		Always terminate (*LINE) with '\n'
		Return only through first '\n'.
		*/
		if((*LINE)[0] == '\0')
		{ /* first time in this trip */
			Place = LineBuffer;
			HitDone = false;
		}

		(*LINE)[0] = '\0';
		if(Place == LineBuffer)
		{
			if(HitDone)
			{
				HitDone = false;
				return (EOF);
			}
			/* need to read from Dialog box */
			GetPort(&thePort);
			SetDAFont(CmdFont);
			theDialog = getDlog(CONSOLEDLOG, true, nButtons, "nd");
			SetPort((GrafPtr) theDialog);
			TextFont(CmdFont);
			TextSize(CmdFontSize);
			GetDItem(theDialog, titleItem, &itemType, &itemHandle, &r);
			SetDItem(theDialog, titleItem, itemType,
					 (Handle) consoleUserItem, &r);

			ShowWindow((WindowPtr) theDialog);

			do
			{
				ModalDialog((ModalFilterProcPtr) myDialogFilter,&itemHit);
			} while(itemHit > nButtons);

			HitDone = (itemHit == doneItem);

		/* directly access the text edit handle rather than use GetDItem() */
			textH = ((DialogPeek) theDialog)->textH;
			flags = HGetState((Handle) (**textH).hText);
			HLock((Handle) (**textH).hText);
			teLength = (**textH).teLength;
			if(teLength > MAXLINE)
			{
				SetDAFont(systemFont);
				sprintf(OUTSTR,
						"WARNING: more than %ld characters entered; last %d ignored",
						MAXLINE, teLength - MAXLINE);
				myAlert(OUTSTR);
				teLength = MAXLINE;
			}
			if(teLength > 0)
			{
				BlockMove((Ptr) *(**textH).hText, (Ptr) LineBuffer, teLength);
			}
			HSetState((Handle) (**textH).hText, flags);

			LineBuffer[teLength] = '\0';
			strcpy(*LINE, (char *) LineBuffer);

			SetPort(thePort);
			SetCursor(&WATCH);
			DisposDialog(theDialog);
			SetDAFont(systemFont);
			if(LineBuffer[0] == '\0' && HitDone)
			{
				HitDone = false;
				return (EOF);
			}
		} /*if(Place == LineBuffer)*/
		else
		{
			itemHit = (HitDone) ? doneItem : ok;
			strcpy(*LINE, (char *) Place);
		}

		pc = (unsigned char *) strchr((char *) Place, '\n');
		if(pc != (unsigned char *) 0 && pc[1] != '\0')
		{ /* still something not seen */
			pc++;
			length = pc - Place;
			Place = pc;
			itemHit = ok;
		} /*if(pc != (unsigned char *) 0)*/
		else
		{ /* this is the last of it */
			length = strlen((char *) Place);
			if(pc == (unsigned char *) 0)
			{
				(*LINE)[length++] = '\n';
			}
			Place = LineBuffer;
			HitDone = false;
		} /*if(Place == LineBuffer){}else{}*/
		(*LINE)[length] = '\0';
		return ((itemHit == doneItem) ? EOF : 1);
	} /*if(fn != STDIN){}else{}*/
} /*fillLINE()*/

#ifdef TIMEANDMEMORY
#include <time.h>
#endif /*TIMEANDMEMORY*/

/*
  If argument prompt is NULL, use global PROMPT (BDEPTH == 0)
  or batch file name (BDEPTH > 0), preceded by EOL.  This should only
  happen at first prompt for a command line.
  If prompt != NULL (probably "more> "), print it with no preceding EOL
*/

void putprompt(prompt)
char        *prompt;
{
	char           *prmpt;
	WHERE("putprompt");

	prmpt = (prompt == (char *) 0) ? PROMPT : prompt;

	if(BDEPTH == 0)
	{
		if(prompt == (char *) 0)
		{
			myeol();
		}
		myprint(prmpt);
		fmyflush(STDOUT);
		if(UseWindows > 0)
		{
			TEKey(Return, teCmd);
			TEKey(Backspace, teCmd);
			CmdStrPos = (**teCmd).teLength;
		}
	}
	else if(BATCHECHO[BDEPTH])
	{
		if(prompt == (char *) 0)
		{
			myeol();
			myprint(*INPUTFILENAMES[BDEPTH]);
			myprint("> ");
		}
		else
		{
			myprint(prmpt);
		}
	}
} /*putprompt()*/

static char    FullPathName[PATHSIZE];

char *expandFilename(fname)
char         *fname;
{
	Symbolhandle    home;
	char           *pc;
	char           *userdir = (char *) 0;
	char           *sep1 = DIRSEPARATOR;
	char           *sep2 = "/";
	long            dirNameLength;

	*OUTSTR = '\0';
	if(fname[0] == TILDE)
	{
		if(fname[1] == *sep1 || fname[1] == *sep2)
		{/* treat ~:... or ~/... specially */
			home = Lookup("HOME"); /* should be initialized at startup */
			if(home == (Symbolhandle) 0 || TYPE(home) != CHAR ||
			   !isScalar(home))
			{
				sprintf(OUTSTR,
						"ERROR: cannot expand %c%c unless CHARACTER variable HOME is defined",
						fname[0], fname[1]);
				goto errorExit;
			}
			userdir = STRINGPTR(home);
			pc = fname+2;
		}
		else
		{
/* can't expand ~name on a Macintosh */
			userdir = (char *) 0;
		}

		if(userdir == (char *) 0)
		{
			putPieces("ERROR: unable to expand ",fname,
					  (char *) 0, (char *) 0);
			goto errorExit;
		}

		dirNameLength = strlen(userdir);
/* allow for the possibility that the HOME ends in ":" */
		if(userdir[dirNameLength-1] == *sep1)
		{
			dirNameLength--;
		}
		if(dirNameLength + strlen(pc) + 1 >= PATHSIZE)
		{
			putPieces("ERROR: expanded file name ", userdir, pc,
					  " is too long");
			goto errorExit;
		}

		strcpy(FullPathName,userdir);
		FullPathName[dirNameLength++] = *sep1;
		FullPathName[dirNameLength] = '\0';
		strcat(FullPathName,pc);
	} /*if(fname[0] == TILDE)*/
	else
	{
		if(strlen(fname) > PATHSIZE)
		{
			putPieces("ERROR: filename ", fname, " is too long", (char *) 0);
			goto errorExit;
		}

		strcpy(FullPathName, fname);
	}

	return ((char *) FullPathName);

  errorExit:
	putOUTSTR();

	return ((char *) 0);

} /*expandFilename()*/

/*
  Check file names for legality on platform
  On the Mac, essentially any name is legal as long as it is not too
  long and does not start with '.' (names starting with '.' may be interpreted
  as driver names by the OS)
*/

long isfilename(fname)
char         *fname;
{
	*OUTSTR = '\0';
	if(*fname == '\0')
	{
		sprintf(OUTSTR,"ERROR: zero length file name");
	}
	else if(strlen(fname) >= PATHSIZE)
	{
		putPieces("ERROR: file name ", fname, " is too long", (char *) 0);
		goto errorExit;
	}
	else if(*fname == '.')
	{
		sprintf(OUTSTR, "ERROR: file names must not start with '.'");
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}
	return (1);

  errorExit:
	putOUTSTR();
	return (0);
} /*isfilename()*/
