/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Macmain
#else
#define __SEG__ Macmain
#endif /*MPW3*/
#endif /*MPW*/

/*
  Main driver for MacAnova, originally consisting primarily of modules from
  TransSkel
*/

/*
	This file was originally modified from the main module of a
	TransSkel multiple-window demonstration written by Paul DuBois, the
	author of TransSkel 24 June 1986.  Until April, 1994 it relied on
	TransSkel, version 1.1, as modified for the needs of MacAnova.

	940429 Made modifications necessary to run using TransSkel 3.11
*/


#define MAIN__

#ifdef PERFORMANCE
#include "profile.h"
#endif /*PERFORMANCE*/

#include "macIface.h" /* pulls in TransSkel.h */
#include "globals.h"

#include <Files.h>

#ifndef MASTERBLOCKS
#define MASTERBLOCKS 20
#endif /*MASTERBLOCKS*/
#define MAXHANDLES   (MASTERBLOCKS * 64)
#ifndef EXTRABLOCKS
#define EXTRABLOCKS   4 /* handles for system use, e.g., windows, scrap */
#endif /*EXTRABLOCKS*/

#ifndef OUTLINESIZE
#define OUTLINESIZE   400  /*size of output line buffer*/
#endif /*OUTLINESIZE*/

char        *OutLineBuf;

#undef UNDEFINED__

#ifdef MPW
#ifdef MPW3
#pragma segment Getyacc
#else
#define __SEG__ Getyacc
#endif /*MPW3*/
#endif /*MPW*/

/* resource numbers for parser arrays*/
#define YYEXCA 1001
#define YYACT  1002
#define YYPACT 1003
#define YYPGO  1004
#define YYR1   1005
#define YYR2   1006
#define YYCHK  1007
#define YYDEF  1008
/* Global pointers to parser arrays */
short *Pyyexca;
short *Pyyact;
short *Pyypact;
short *Pyypgo;
short *Pyyr1;
short *Pyyr2;
short *Pyychk;
short *Pyydef;

static short * getYaccTable(Integer id)
{
	Handle      yaccTable;
	short      *table;
	LongInt     rsrcSize;

	yaccTable = GetResource('YTAB', id);
	if(yaccTable != (Handle) 0)
	{
		HNoPurge(yaccTable);
		MaxMem(&rsrcSize);
		HLock(yaccTable);
		table = (short *) *yaccTable;
	}
	else
	{
		table = (short *) 0;
	}
	return (table);

} /*getYaccTable()*/

static int  setupYaccTables(void)
{
	if((Pyyexca = getYaccTable(YYEXCA)) == (short *) 0)
	{
		goto resourceError;
	}

	if((Pyyact = getYaccTable(YYACT)) == (short *) 0)
	{
		goto resourceError;
	}

	if((Pyypact = getYaccTable(YYPACT)) == (short *) 0)
	{
		goto resourceError;
	}

	if((Pyypgo = getYaccTable(YYPGO)) == (short *) 0)
	{
		goto resourceError;
	}

	if((Pyyr1 = getYaccTable(YYR1)) == (short *) 0)
	{
		goto resourceError;
	}

	if((Pyyr2 = getYaccTable(YYR2)) == (short *) 0)
	{
		goto resourceError;
	}

	if((Pyychk = getYaccTable(YYCHK)) == (short *) 0)
	{
		goto resourceError;
	}

	if((Pyydef = getYaccTable(YYDEF)) == (short *) 0)
	{
		goto resourceError;
	}

	return (1);

  resourceError:
	myAlert("FATAL ERROR reading resource");
	return (0);


} /*setupYaccTables()*/

#ifdef MPW
#ifdef MPW3
#pragma segment Macmain
#else
#define __SEG__ Macmain
#endif /*MPW3*/
#endif /*MPW*/

#ifndef NKEYCHECKS
#define NKEYCHECKS   5 /* number of times to check for events */
#endif /*NKEYCHECKS*/
#ifndef NTICKS
#define NTICKS   30 /* half a second */
#endif /*NTICKS*/

/*	Note that byte order of integers is from most significant to least, that
	is, in the order of the usual hex representation.  This is relevent
	since a KeyMap variable (long [4]) represents a packed array of
	Boolean in Pascal.
*/

static KeyMap        TheKeys;
static unsigned long Tickcount;

static Boolean isKeyPressed (Integer keyCode)
{
	Integer       word = keyCode >> 5, bit = keyCode & 0x1f;
	Integer       byte = bit >> 3;
	unsigned long ul;
	unsigned char uc;

	bit &= 0x07;
	ul = TheKeys[word];
	uc = (0xff & (ul >> (8*(3 - byte))));
	return (((uc >> bit) & 1) != 0);
}

/* Check keys at start up */
static unsigned char checkKey(void)
{
	Integer        i;
	unsigned char  ch = NOTCOMMAND;
	WHERE("checkKey");

	Tickcount = TickCount();
	do
	{
		GetKeys (TheKeys);
		if(isKeyPressed(COMMANDKEYCODE))
		{
			ch = COMMANDKEYCODE;
			break;
		}
	} while(TickCount() - Tickcount < NTICKS);

	while(ch == COMMANDKEYCODE && TickCount() - Tickcount < NTICKS)
	{
		for(i=0;i<127;i++)
		{
			if(i != COMMANDKEYCODE && isKeyPressed(i))
			{
				ch = i;
				break;
			}
		}
	}
	FlushEvents(everyEvent, 0);

	return (ch);
} /*checkKey()*/


void main(void)
{
	Integer         OporPr, nAppFiles;
	AppFile         startup_info;
	FILE           *iniFile = (FILE *) 0;
	char           *iniFileName = MACANOVAINI;
	CursHandle      cursor;
	Integer         i, loaded = 0, restored = 0, batched = 0, errorFlag = 0;
	Boolean         didbatch = false, assumeBatch;
	Symbolhandle    list = (Symbolhandle) 0;
	char           *valueList[2];
	long            typeList[2];
	char           *keywordList[2];
	unsigned char   key;
	long            nargs;
	SkelInitParams  initParams;
	WHERE("main");

#ifdef MPW
	_DataInit();
	UnloadSeg((Ptr)_DataInit);
#endif /*MPW*/

	initParams.skelMoreMasters = MASTERBLOCKS+EXTRABLOCKS;
	initParams.skelGzProc = (GrowZoneProcPtr) nil;
	initParams.skelResumeProc = (ResumeProcPtr) nil;
	initParams.skelStackAdjust = (Size) 0;

/* Initialize TransSkel */
	SkelInit(&initParams);
	SkelSetWaitTimes(CmdSleep, InterruptSleep);

#ifdef PERFORMANCE
	if(!InitPerf(&ThePGlobals,TIMERCOUNT,CODEBUCKETSIZE,DOROM,DOAPPCODE,
				 "\pCODE",ROMID,ROMNAME,DONTDORAM,0,0,0))
	{
		myAlert("InitPerf failure");
		exit(1);
	}
#endif /*PERFORMANCE*/

	if(!setupYaccTables())
	{
		exit (0);
	}
	UNLOADSEG(setupYaccTables);

/* Initialize MacAnova handles and output line buffer (see macIo.c) */
	if(!myInitHandles(MAXHANDLES) ||
	   (OutLineBuf = mygetpointer(OUTLINESIZE+1)) == (char *) 0)
	{
		myAlert("Out of memory.  Execution terminated");
		exit(1);
	}
	OutLineBuf[0] = '\0';

	{ /* get and save information about startup directory */
		LongInt   dirID;
		Integer   vRefNum;

		errorFlag = HGetVol((StringPtr) 0, &vRefNum, &dirID);

		if(errorFlag == noErr)
		{
			HomeVolume = vRefNum;
			HomeDirectory = dirID;
		}
		else
		{
			HomeVolume = HomeDirectory = 0;
		}
	}

	cursor = GetCursor(iBeamCursor);
	IBEAM = **cursor;
	cursor = GetCursor(crossCursor);
	CROSS = **cursor;
	cursor = GetCursor(watchCursor);
	WATCH = **cursor;

	ISATTY = 1; /* just for safety */


	if(!SetUpMenus())		/* install menu handlers */
	{
		myAlert("FATAL ERROR: Cannot build menus");
		exit(1);
	}

	DisableItem(FileMenu, (Integer) 0);
	DisableItem(EditMenu, (Integer) 0);
	DisableItem(WindowMenu, (Integer) 0);
	DisableItem(CommandMenu, (Integer) 0);
	DisableItem(OptionsMenu, (Integer) 0);
	DisableItem(FontMenu, (Integer) 0);
	DisableItem(FontSizeMenu, (Integer) 0);

	/* initialize symbol table */
	initialize((char *) 0,(char *) 0,(char *) 0);
	UNLOADSEG(initialize);

	GraphRectInit(-1); /* initialize default zoom rectangles */

	key = checkKey();

	UseWindows = (key == COMMANDKEYCODE || key == BKEYCODE) ? -1 : 1;

	assumeBatch = (UseWindows <= 0);

	/* try to open initialization and help files */
	iniFile = fopen(iniFileName, "r");
	HELPFILE = fopen(*HELPFILENAME,"r");

/*	Initialize Font Information */
	GetFNum(CmdFontName, &CmdFont);
	if(CmdFont == 0)
	{
		GetFNum(REGULARFONTNAME, &CmdFont);
	}
	GetFNum(GraphFontName, &GraphFont);
	if(GraphFont == 0)
	{
		GetFNum(REGULARFONTNAME, &GraphFont);
	}
	GetFNum(PrintFontName, &PrintFont);
	if(PrintFont == 0)
	{
		GetFNum(REGULARFONTNAME, &PrintFont);
	}

	if(assumeBatch)
	{ /* open Spool file for output */
		nargs = 1;
		typeList[0] = CHAR;
		valueList[0] = keywordList[0] = "";
		list = Buildlist(nargs,typeList,valueList,keywordList);

		if(list == (Symbolhandle) 0 || spool(list) == (Symbolhandle) 0)
		{
			myAlert("WARNING: cannot use non-interactive mode without a spool file");
			UseWindows = 1;
		}
		else
		{
			BatchModeInit();
			UseWindows = 0;
		}
		UnloadSeg((Ptr) spool);
		Removelist(list);
	} /*if(assumeBatch)*/

	if(UseWindows > 0)
	{
	/* install window handlers  */
		CmdWindInit();
	}
	if(FatalError)
	{
		myAlert("INITIALIZATION ERROR: cannot continue");
		(void) quitIt();
		exit (0);
	}
/*
	Make sure help file is found if there; once found, it is never closed
	so that it won't be lost if the directory changes
*/
	if(HELPFILE != (FILE *) 0 &&
	   (LINE = mygethandle(MAXLINE)) != (char **) 0)
	{
		/* look for message of the day preceded and followed by "!!!!"*/
		long      printit = 0;

		while(fillLINE(HELPFILE) != EOF)
		{
			if(strncmp(*LINE,"!!!!",4) == 0)
			{
				if(printit)
				{
					break;
				}
				else
				{
					printit = 1;
					continue;
				}
			}

			if(printit)
			{
				if(printit == 1)
				{
					myeol();
					printit = 2;
				}
				myprint(*LINE);
			}
		} /*while(fillLINE(HELPFILE) != EOF)*/
		rewind(HELPFILE);
		mydisphandle(LINE);
		LINE = (char **) 0;
	}

	CountAppFiles(&OporPr, &nAppFiles);

	for(i = 1; i<= nAppFiles; i++)
	{
		GetAppFiles(i, &startup_info);
		if(startup_info.fType == 'TEXT')
		{
	/* assume it is saved output window or batch file; only use first one */
			if(!assumeBatch)
			{ /* take it as window to be read in */
				if(!loaded)
				{
					loaded = i;
				}
			}
			else if(!batched)
			{ /* assume it is batch file to execute */
				batched = i;
			}
		}
		else if(startup_info.fType == KEEPTYPE ||
				startup_info.fType == ASCIIKEEPTYPE)
		{
			if(!restored)
			{
				restored = i;
			}
		}
		else
		{
			PtoCstr((char *) startup_info.fName);
			sprintf(OUTSTR,
					"WARNING: startup file %s is not text file or save file",
					(char *) startup_info.fName);
			myAlert(OUTSTR);
			if(UseWindows > 0)
			{
				myerrorout(OUTSTR);
			}
		}
		ClrAppFiles(i);
	} /*for(i = 1; i<= nAppFiles; i++)*/

	if(UseWindows == 0 && batched == 0)
	{ /* Non-interactive but no batch file found yet */
		batched = -1;
	}

	if(restored)
	{ /* now do actual restoration */
		GetAppFiles(restored, &startup_info);
		PtoCstr((char *) startup_info.fName);
		if((errorFlag = doRestore((char *) startup_info.fName,
								   startup_info.vRefNum)) != noErr)
		{
			restored = 0;
		}
	} /*if(restored)*/

	if(batched != 0)
	{ /* Note, batch file will be executed after MacAnova.ini */
		BDEPTH = 0;
		if(batched > 0)
		{ /* file provided */
			GetAppFiles(batched, &startup_info);
			BatchVolume = startup_info.vRefNum;
			PtoCstr((char *) startup_info.fName);
		}
		else
		{ /* ask for file */
			startup_info.fName[0] = '\0';
		}
		typeList[0] = CHAR;
		valueList[0] = (char *) startup_info.fName;
		keywordList[0] = "";
		typeList[1] = LOGIC;
		valueList[1] = "T";
		keywordList[1] = "echo";
		list = Buildlist(nargs, typeList, valueList, keywordList);

		if(list == (Symbolhandle) 0 || batch(list) == (Symbolhandle) 0)
		{
			if(!UseWindows)
			{
				myAlert("No command file supplied. MacAnova terminating");
				return;
			}
			BatchVolume = 0;
		}
		Removelist(list);
		if(BatchVolume != 0 && !UseWindows)
		{ /* make file the level zero command file */
			INPUTFILE[0] = INPUTFILE[1];
/*
	Note: If UseWindows == 0, then we must have found a command file and hence
	INPUTFILE[0] != STDIN
*/
			INPUTFILENAMES[0] = INPUTFILENAMES[1];
			BATCHECHO[0] = 1;
			INPUTFILE[1] = (FILE *) 0;
			INPUTFILENAMES[1] = (char **) 0;
			BATCHECHO[1] = 0;
			BDEPTH = 0;
			ISATTY = 0;
			sprintf(OUTSTR,"Input taken from file '%s'", *INPUTFILENAMES[0]);
			putOUTSTR();
		}
		UNLOADSEG(batch);
	} /*if(batched)*/

	if(iniFile != (FILE *) 0 && restored)
	{ /* don't use initialization file if restoring workspace */
		fclose(iniFile);
		iniFile = (FILE *) 0;
	}

	if(iniFile != (FILE *) 0)
	{ /* We were able previously to open the file */
		/* set up execution of initialization file */
		TMPHANDLE = mygethandle(strlen(iniFileName) + 1);
		INPUTFILENAMES[BDEPTH+1] = TMPHANDLE;
		if(TMPHANDLE != (char **) 0)
		{
			strcpy(*TMPHANDLE, iniFileName);
			INPUTFILE[++BDEPTH] = iniFile;
			didbatch = true;
		}
		else
		{
			fclose(iniFile);
		}
	} /*if(iniFile != (FILE *) 0)*/

	if(loaded)
	{
		GetAppFiles(loaded, &startup_info);
		PtoCstr((char *) startup_info.fName);
		errorFlag = loadWindow((char *) startup_info.fName,
							   startup_info.vRefNum, false);
		CtoPstr((char *) startup_info.fName);
		if(errorFlag != noErr)
		{
			myAlert("Unable to read saved command window");
			loaded = 0;
		}
	} /*if(loaded)*/

	if(restored)
	{ /* make sure message appears */
		GetAppFiles(restored, &startup_info);
		PtoCstr((char *) startup_info.fName);
		sprintf(OUTSTR,"Workspace restored from file %s",
				(char *) startup_info.fName);
		putOUTSTR();
	} /*if(restored)*/

	if(!UseWindows || !didbatch && (restored || !loaded))
	{
		putprompt((char *) 0);
	}

	if(UseWindows > 0)
	{
		CmdDirty = CmdWindows[CurrentWindow].cmdDirty = false;
		EnableItem(FileMenu, (Integer) 0);
		EnableItem(EditMenu, (Integer) 0);
		EnableItem(WindowMenu, (Integer) 0);
		EnableItem(CommandMenu, (Integer) 0);
		EnableItem(OptionsMenu, (Integer) 0);
		EnableItem(FontMenu, (Integer) 0);
		EnableItem(FontSizeMenu, (Integer) 0);
	}

	DrawMenuBar();

	SkelEventLoop();

	SkelCleanup();		/* throw away windows and menus */

} /*main()*/


/*
	Initialize menus.  Tell Skel to process the Apple menu automatically,
	and associate the proper procedures with the File and Edit menus.
*/
#define MAXITEMLENGTH   32

static Integer  LegalFontSizes[] = {9, 10, 12, 14, 18, 20, 24, 36, 0};

#define SUBMENU     true
#define DRAWBAR     true

Boolean SetUpMenus()
{
	Integer       i, length, size;
	Integer       markChar;
	Str255        itemString;

	SkelApple((StringPtr) "\pAbout MacAnova;Help/H",
			  (SkelMenuSelectProcPtr) DoApple);
	AppleMenu = SkelGetAppleMenu();
	FileMenu = GetMenu(FILEMENU);
	EditMenu = GetMenu(EDITMENU);
	WindowMenu = GetMenu(WINDOWMENU);
	CommandMenu = GetMenu(COMMANDMENU);
	OptionsMenu = GetMenu(OPTIONSMENU);
	FontMenu = GetMenu(FONTMENU);
	FontSizeMenu = GetMenu(FONTSIZEMENU);
	if(FileMenu == (MenuHandle) 0 || EditMenu == (MenuHandle) 0 ||
	   WindowMenu == (MenuHandle) 0 || CommandMenu == (MenuHandle) 0 ||
	   OptionsMenu == (MenuHandle) 0 || FontMenu == (MenuHandle) 0 ||
	   FontSizeMenu == (MenuHandle) 0 ||
	   !SkelMenu(FileMenu, (SkelMenuSelectProcPtr) DoFile,
				 (SkelMenuClobberProcPtr) nil, !SUBMENU, !DRAWBAR) ||
	   !SkelMenu(EditMenu, (SkelMenuSelectProcPtr) DoEdit,
				 (SkelMenuClobberProcPtr) nil, !SUBMENU, !DRAWBAR) ||
	   !SkelMenu(WindowMenu, (SkelMenuSelectProcPtr) DoWind,
				 (SkelMenuClobberProcPtr) nil, !SUBMENU, !DRAWBAR) ||
	   !SkelMenu(CommandMenu, (SkelMenuSelectProcPtr) DoCommand,
				 (SkelMenuClobberProcPtr) nil, !SUBMENU,!DRAWBAR) ||
	   !SkelMenu(OptionsMenu, (SkelMenuSelectProcPtr) DoOptions,
				 (SkelMenuClobberProcPtr) nil, !SUBMENU, !DRAWBAR) ||
	   !SkelMenu(FontMenu, (SkelMenuSelectProcPtr) DoFont,
				 (SkelMenuClobberProcPtr) nil, !SUBMENU, !DRAWBAR) ||
	   !SkelMenu(FontSizeMenu, (SkelMenuSelectProcPtr) DoFontSize,
				 (SkelMenuClobberProcPtr) nil, SUBMENU, DRAWBAR))
	{
		return (false);
	}

	/* Save pre-defined commands */
	for(i=0;i<NCOMMANDS;i++)
	{
		GetItem(CommandMenu, command1 + i, itemString);
		length = itemString[0];
		PtoCstr(itemString);
		GetItemMark(CommandMenu, command1 + i, &markChar);
		if(markChar == diamondMark && itemString[length-1] != Newline)
		{
			length++;
		}
		CommandItems[i] = TMPHANDLE = mygethandle(length + 1);
		if(TMPHANDLE == (char **) 0)
		{
			return (false);
		}
		strcpy(*TMPHANDLE, itemString);

		if(markChar == diamondMark)
		{
			(*TMPHANDLE)[length--] = '\0';
			(*TMPHANDLE)[length] = Newline;
		} /*if(markChar == diamondMark)*/

		if(length > MAXITEMLENGTH)
		{
			itemString[MAXITEMLENGTH] = 0xc9; /* '...' */

			itemString[MAXITEMLENGTH+1] = '\0';
		}
		CtoPstr(itemString);
		SetItem(CommandMenu, command1+i, itemString);
	} /*for(i=0;i<NCOMMANDS,i++)*/

	AddResMenu(FontMenu, 'FONT');

	for (i = 0; size = LegalFontSizes[i]; i++)
	{
		Str27      pointSize;

		sprintf((char *) pointSize, (size < 10) ? "%3d Point" : "%2d Point", size);
		CtoPstr((char *) pointSize);
		InsMenuItem(FontSizeMenu, pointSize, 100);
	}
	decorateFontMenus(CmdFont, CmdFontSize);
	return (true);

} /*SetUpMenus()*/


/*
	Handle selection of About MacAnova item from Apple menu
*/

extern char       *Welcome[];

#define ABOUTTICKS         240L
#define AboutFont          systemFont
#define AboutFontSize      12
#define AboutFrameItem     ok+1
#define AboutMessageItem1  (AboutFrameItem+1)
#define AboutCurvature     32
#define NAboutLines        7

/* Returns ok for any key or time expired */

pascal Boolean aboutDialogFilter (DialogPtr theDialog, EventRecord *theEvent,
							   Integer *itemHit)
{
	Integer         evtMods = theEvent->modifiers;
	char            evtChar = theEvent->message & charCodeMask;
	WHERE("aboutDialogFilter");

	if (theEvent->what == keyDown || theEvent->what == autoKey)
	{
		*itemHit = ok;
		return (true);
	}

	if(Tickcount != 0 && TickCount() - Tickcount > ABOUTTICKS)
	{
		*itemHit = ok;
		return (true);
	}
	return (false);
} /*aboutDialogFilter*/

static pascal aboutUserItem(WindowPtr theWindow, Integer itemNo)
{
	Rect            r;
	Integer         itemType, lineNo = itemNo - AboutMessageItem1;
	Handle          itemHandle;
	char           *start, *end;
	Integer         length, h, v;
	Str255          msg;
	FontInfo       info;
	WHERE("aboutUserItem");

	GetDItem((DialogPtr) theWindow, itemNo, &itemType, &itemHandle, &r);

	if(itemNo == AboutFrameItem)
	{
		PenSize(3,3);
		InsetRect(&r,2,2);
#ifdef UNDEFINED__
		EraseOval(&r);
		FrameOval(&r);
		PenSize(1,1);
		InsetRect(&r,4,4);
		FrameOval(&r);
#endif /*UNDEFINED__*/
		EraseRect(&r);
		FrameRoundRect(&r, AboutCurvature, AboutCurvature);
		PenSize(1,1);
		InsetRect(&r,4,4);
		FrameRoundRect(&r, AboutCurvature-4, AboutCurvature-4);
	} /*if(itemNo == AboutFrameItem)*/
	else if(lineNo < NAboutLines)
	{
		start = Welcome[lineNo];
		while(*start == ' ')
		{
			start++;
		}
		if(start[0] == '\0')
		{
			return;
		}
		GetFontInfo(&info);
		end = start + strlen(start) - 1;
		while(end > start && *end == ' ')
		{
			end--;
		}
		length = end - start + 1;
		strncpy((char *) msg + 1,start,length);
		msg[0] = length;
		h = (r.left + r.right - StringWidth((Str255) msg))/2;
		v = r.bottom - info.descent;
		MoveTo(h, v);
		DrawString((Str255) msg);
	} /*if(itemNo == AboutFrameItem){}else{}*/
} /*aboutUserItem()*/


pascal void DoApple(Integer item)
{
	GrafPtr            thePort;
	DialogPtr          theDialog;
	Integer            i, j;
	Rect               r;
	Handle             itemHandle;
	Integer            itemType, itemHit;
	Symbolhandle       list;
	Boolean            brief;
	LongInt            selStart = (**teCmd).selStart, selEnd = (**teCmd).selEnd;
	char               topic[21], c;
	char              *valueList[2];
	long               typeList[2];
	char              *keywordList[2];
	WHERE("DoApple");

	GetPort(&thePort);
	switch (abs(item))
	{
	  case aboutmacanova:

		theDialog = getDlog(ABOUTMACANOVA, false, 1, "y");
		SetPort((GrafPtr) theDialog);
		
		TextFont(AboutFont);
		TextSize(AboutFontSize);
		GetDItem(theDialog, AboutFrameItem, &itemType, &itemHandle, &r);
		r = theDialog->portRect;
		InsetRect(&r,2,2);
		SetDItem(theDialog, AboutFrameItem, userItem, (Handle) aboutUserItem,
				 &r);
		for(i=0; i < NAboutLines && Welcome[i] != (char *) 0;i++)
		{
			GetDItem(theDialog, AboutMessageItem1 + i, &itemType, &itemHandle,
					 &r);
			SetDItem(theDialog, AboutMessageItem1 + i, userItem,
					 (Handle) aboutUserItem, &r);
		}
		ShowWindow((WindowPtr) theDialog);

		Tickcount = (item < 0) ? TickCount() : 0;
		ModalDialog((ModalFilterProcPtr) aboutDialogFilter, &itemHit);
		DisposDialog(theDialog);
		
		break;

	  case helpit:
		brief = (FrontWindow() != CmdWind);
		MyShowWindow(CmdWind);

		/* dummy up help("selection") */
		j = 0;
		if(!brief)
		{
			for(i=0;i<selEnd - selStart;i++)
			{ /* trim off leading blanks && non alphas while copying */
				c = (*((**teCmd).hText))[selStart+i];
				if(j > 0 || isnamestart(c))
				{ /* stop when topic is full or hit not alphanumeric*/
					if(!isnamechar(c) || j == sizeof(topic) -1)
					{
						break;
					}
					topic[j++] = c;
				}
			}
			topic[j] = '\0';
		} /*if(!brief)*/
		else
		{ /* "__teaser__" tells help() to print short message */
			strcpy(topic,"__teaser__");
		} /*if(!brief){}else{}*/

		if(topic[0] != '\0')
		{
			typeList[0] = CHAR;
			valueList[0] = topic;
			if(!brief)
			{
				sprintf(OUTSTR,"help(\"%s\")\n",topic);
			}
			else
			{
				sprintf(OUTSTR,"help\n");
			}
		}
		else
		{/* must have been all blanks or non alphas; fake 'help()' */
			typeList[0] = 0;
			valueList[0] = (char *) 0;
			sprintf(OUTSTR,"help()\n");
		}

		insertCmd(OUTSTR, true);

		/* save info in case needed for undoing output  */
		clearUndoInfo();
		LastCmdStrPos = CmdStrPos;
		LastLength = (**teCmd).teLength;
		UndoStatus = OUTPUTUNDO;

		keywordList[0] = "";
		list = Buildlist(1,typeList,valueList,keywordList);
		if(list != (Symbolhandle) 0)
		{
			DisableItem(EditMenu, undoit);
			DisableItem(CommandMenu, 0);
			DisableItem(OptionsMenu, (Integer) 0);
			DisableItem(FontMenu, (Integer) 0);
			DisableItem(FontSizeMenu, (Integer) 0);
			DrawMenuBar();
			setCmdFileItems(false);
			HiliteMenu (0);		/* turn off menu hiliting */

			(void) help(list);
			UnloadSeg((Ptr) help);
			Removelist(list);

			EnableItem(CommandMenu, 0);
			EnableItem(OptionsMenu, (Integer) 0);
			EnableItem(FontMenu, (Integer) 0);
			EnableItem(FontSizeMenu, (Integer) 0);
			DrawMenuBar();
		} /*if(list != (Symbolhandle) 0)*/
		setCmdEditItems();
		EnableItem(AppleMenu, helpit);

		if(INTERRUPT == INTNOTSET)
		{
			putprompt((char *) 0);
		}
		*OUTSTR = '\0';
		break;
	} /*switch (item)*/
	
} /*DoApple()*/

/*
	Process selection from File menu.

	Help    List all help options
	Print {window,selection,graph}
	Keep    Equivalent to save("")
	Restore Equivalent to restore("")
	Save    Save command window as text file
	Quit	Request a halt by calling SkelStopEventLoop().  This makes
			SkelEventLoop() return.
	Interrupt   Set INTERRUPT non zero
*/

pascal void DoFile(item)
Integer             item;
{
	char           *valueList[2];
	long            typeList[2];
	char           *keywordList[2];
	long            nargs;
	Symbolhandle    list;
	LongInt         teLength = (**teCmd).teLength;
	WindowPtr       theWindow = FrontWindow();
	Integer         answer;
	Integer         running = Running;
	Integer         windno;
	WHERE("DoFile");

	if(!UseWindows && item != interrupt && item != keepitas)
	{
		return;
	}
/*
	Running == 2 indicates not running but processing a menu command
	Recognized by CmdWindActivate
*/
	Running = (Running) ? 1 : 2;
	switch (item)
	{
	  case pagesetup:
		if (theWindow == CmdWind)
		{
			(void) doPageSetup(TEXTPRINT);
		}
		else
		{
		  	(void) doPageSetup(GRAPHPRINT);
		}
		break;

	  case printit:
		if((windno = whichGraphWindow(theWindow)) >= 0)
		{
			GraphWindPrint(windno);
		}
		else if((windno = whichCmdWindow(theWindow)) >= 0)
		{
			CmdWindPrint(windno);
		}
		break;

	  case openit:
		answer = loadWindow((char *) 0, (Integer) 0, true);
		if(answer == noErr && CmdEditable)
		{
			putprompt((char *) 0);
			CmdDirty = CmdWindows[CurrentWindow].cmdDirty = false;
		}
		break;

	  case saveit:
	  case saveitas:
		doSave(CmdWindows + CurrentWindow, item);
		break;

	  case keepit:
	  case keepitas:
	  	nargs = 2;
		keywordList[0] = "";
		if(item == keepit && KeepVolume != 0)
		{ /* Save Workspace and there has been previous save() or asciisave()*/
			typeList[0] = 0; /* simulate 'save(, options:T)'*/
		}
		else
		{ /* Save Workspace As... or no previous save() or asciisave() */
			typeList[0] = CHAR; /* simulate 'save("", options:T)'*/
			valueList[0] = "";
		}
		typeList[1] = LOGIC;
		valueList[1] = "T";
		keywordList[1] = "options"; /* always save options on menu save */

		list = Buildlist(nargs,typeList,valueList,keywordList);

		if(list != (Symbolhandle) 0)
		{
			strcpy(FUNCNAME, (ASCII > 0) ? "asciisave" : "save");
			(void) save(list);
			UnloadSeg((Ptr) save);
			Removelist(list);
		}
		if (UseWindows > 0)
		{
			putprompt((char *) 0);
		}
		*OUTSTR = '\0';
		break;

	  case restoreit:
		DisableItem(FileMenu, (Integer) 0);
		(void) doRestore((char *) 0, (Integer) 0);
		EnableItem(FileMenu, (Integer) 0);
		putprompt((char *) 0);
		*OUTSTR = '\0';
		break;

	  case quit:
		(void) quitIt();			/* close files and request halt */
		break;

	  case interrupt:
		if(UseWindows > 0)
		{
			MyShowWindow(CmdWind);
			myeol();
			myerrorout("*****  Interrupt  *****");
			myeol();

			if(!running)
			{
				clearUndoInfo();
			}
		}
		INTERRUPT = INTSET; /* set interrupt flag */
		GraphWindPaused = false;
		break;

	  case go_on:
		GraphWindPaused = false;
		break;

	} /*switch (item)*/

	Running = running;

} /*DoFile()*/

/*
	Process selection from Window menu

	Hide         Hide front window
	Command      Display command window
	Graph n      Display graph window n, where n = 1,2,3 or 4
	Close        Hide the frontmost window.  If it belongs to a desk accessory,
	             close the accessory.

*/
pascal void DoWind(item)
Integer             item;
{
	WindowPeek      wPeek;
	WindowPtr       theWindow = FrontWindow();
	LongInt         selStart = (**teCmd).selStart, selEnd = (**teCmd).selEnd;
	Integer         insertLine = CmdInsertLine;
	Integer         windno = whichGraphWindow(theWindow);
	Integer         lastWind = CurrentWindow;
	Integer         running = Running;
	Str255          wTitle;
	unsigned char   key = '\0';
	Boolean         fromMenu = (item > 0);
	WHERE("DoWind");

	item = abs(item);
	if(!UseWindows)
	{
		return;
	}
	/*
		Running == 2 indicates not running but processing a menu command
		Recognized by CmdWindActivate
	*/
	Running = (Running) ? 1 : 2;
	if(item > cmd1 + Nwindows-1)
	{
		item += MAXWINDOWS - Nwindows;
	}

	switch (item)
	{
	  case hideit:
		if (windno >= 0 || nVisible() > 1)
		{
			HideWindow(theWindow);
		}
		break;

	/*
	  Close the front window.  Take into account whether it belongs
	  to a desk accessory or not.
	  */
	  case closeit:
		wPeek = (WindowPeek) theWindow;
		if (wPeek->windowKind < 0)
		{
			CloseDeskAcc(wPeek->windowKind);
		}
		else if (windno >= 0)
		{ /* must be graph window */
			GraphWindClose();
		}
		else
		{
			CmdWindClose();
		}
		break;

	  case newwind:
		DisableItem(WindowMenu, newwind);
	    windno = createWindow((Str255) 0);
		if(windno < 0)
		{
			myAlert("Cannot create new window");
			FatalError = 1;
			(void) quitIt();
		}
		else if(windno >= MAXWINDOWS)
		{
			myAlert("Too many windows already");
		}
		else
		{
			saveWindowInfo(CurrentWindow);
			GetWTitle(CmdWindows[windno].cmdWind, wTitle);
			item = cmd1 + Nwindows - 1;
			InsMenuItem(WindowMenu,wTitle,item - 1);
			CmdWindows[windno].menuItem = item;
			setCommandM(CurrentWindow); /* set menu key to toggle to old window */
			restoreWindowInfo(windno);
			clearUndoInfo();
			if(fromMenu)
			{
				putprompt((char *) 0);
			}
			CmdDirty = CmdWindows[CurrentWindow].cmdDirty = false;
		}
		break;

	  case cmd1:
	  case cmd2:
	  case cmd3:
	  case cmd4:
	  case cmd5:
	  case cmd6:
	  case cmd7:
	  case cmd8:
	  case cmd9:
/*	  case cmd10:*/ /*there should be exactly MAXWINDOWS of these*/
		for(windno = 0;windno < MAXWINDOWS; windno++)
		{
			if(item == CmdWindows[windno].menuItem)
			{
				break;
			}
		}
		if(windno != CurrentWindow)
		{
			saveWindowInfo(CurrentWindow);
			setCommandM(CurrentWindow);
			restoreWindowInfo(windno);
		}
		MyShowWindow(CmdWind);
		break;

	  case graph1:
	  case graph2:
	  case graph3:
	  case graph4: /* There should be exactly NGRAPHS of these */
	  case panel:
		item -= graph1;
		theWindow = (item < NGRAPHS) ? GraphWind[item] : PanelWind;
		MyShowWindow(theWindow);
		break;

	  case gototop:
		key = HomeKey;
		break;

	  case gotobottom:
		key = EndKey;
		break;

	  case pageup:
		key = Pageup;
		break;

	  case pagedown:
		key = Pagedown;
		break;

	} /*switch (item)*/
	if(key != '\0')
	{
		(void) nonTextKey(key, '\0', 0);
	}
	Running = running;
} /*DoWind()*/

#define TITLEITEM  20  /* item number for user item drawing title */
#define NLITEM     21  /* item number for user item drawing "NL?" */

static pascal editCommandUserItem(WindowPtr theWindow, Integer item)
{
	Rect           r;
	Integer        oldFont = theWindow->txFont;
	Integer        oldSize = theWindow->txSize;
	Integer        itemType;
	Handle         itemHandle;
	char          *msg = (item == TITLEITEM) ?
						"\pMake Changes, Then Click OK" : "\pNL?";
	Integer        h, v;
	FontInfo       info;
	WHERE("editCommandUserItem");

	TextFont(systemFont);
	TextSize(12);
	GetFontInfo(&info);
	
	GetDItem((DialogPtr) theWindow, item, &itemType, &itemHandle, &r);

	h = (r.left + r.right - StringWidth((Str255) msg))/2;
	v = r.bottom - info.descent;
	MoveTo(h, v);
	DrawString((Str255) msg);

	TextFont(oldFont);
	TextSize(oldSize);
}

/* Handler for Command Menu */
pascal void DoCommand(item)
Integer             item;
{
	Str255          command;
	DialogPtr       theDialog;
	Integer         i, length;
	Integer         textItem1 = 4;
	Integer         checkBox1 = textItem1 + NCOMMANDS;
	Rect            r;
	Handle          itemHandle;
	Integer         itemType, itemHit, itemValue, checks = 0;
	WHERE("DoCommand");

	if(!UseWindows)
	{
		return;
	}

	switch (item)
	{
	  case editcommands:
		SetDAFont(CmdFont);

		theDialog = getDlog(EDITCOMMANDS, true, 2, "yn");

		for(i = 0;i < NCOMMANDS;i++)
		{
			strcpy((char *) command, *(CommandItems[i]));
			CtoPstr((char *) command);
			itemValue = (command[(int) command[0]] == Newline) ? 1 : 0;
			GetDItem(theDialog, textItem1 + i, &itemType, &itemHandle, &r);
			command[0] -= itemValue;
			SetIText(itemHandle,command);
			checks |= itemValue << i;
		} /*for(i = 0;i < NCOMMANDS;i++)*/
		setDlogItemValues(theDialog, checkBox1, NCOMMANDS, checks);

		GetDItem(theDialog, TITLEITEM, &itemType, &itemHandle, &r);
		if((itemType & 0x7f) == userItem)
		{
			SetDItem(theDialog, TITLEITEM, userItem,
					 (Handle) editCommandUserItem, &r);
		}
		GetDItem(theDialog, NLITEM, &itemType, &itemHandle, &r);

		if((itemType & 0x7f) == userItem)
		{
			SetDItem(theDialog, NLITEM, userItem,
					 (Handle) editCommandUserItem, &r);
		}
		ShowWindow((WindowPtr) theDialog);

		do /* while(itemHit > cancel)*/
		{
			ModalDialog((ModalFilterProcPtr) myDialogFilter, &itemHit);
			if(itemHit > cancel)
			{
				GetDItem(theDialog, itemHit, &itemType, &itemHandle, &r);
				itemValue = GetCtlValue((ControlHandle) itemHandle);
				SetCtlValue((ControlHandle) itemHandle,
					(itemValue == 0) ? 1 : 0);
			}
		} while(itemHit > cancel);

		if(itemHit == ok)
		{ /* check for zero length commands */
			for(i = 0;i < NCOMMANDS;i++)
			{
				GetDItem(theDialog, textItem1 + i, &itemType, &itemHandle,
						 &r);
				GetIText(itemHandle, command);
				if(command[0] == '\0')
				{
					myAlert("ERROR: You cannot leave any command blank!");
					break;
				}
			} /*for(i = 0;i < NCOMMANDS;i++)*/

			if(i == NCOMMANDS)
			{
				checks = getDlogItemValues(theDialog, checkBox1, NCOMMANDS);
				for(i = 0;i < NCOMMANDS;i++)
				{
					GetDItem(theDialog, textItem1 + i, &itemType,
							 &itemHandle, &r);
					GetIText(itemHandle, command);
					itemValue = (checks >> i) & 1;
					if(itemValue)
					{
						command[0]++;
						command[(int) command[0]] = Newline;
					}
					length = command[0];
					if(myhandlelength(CommandItems[i]) != length + 1)
					{
						TMPHANDLE = mygethandle(length + 1);
						if(TMPHANDLE != (char **) 0)
						{
							mydisphandle(CommandItems[i]);
						}
					}
					else
					{
						TMPHANDLE = CommandItems[i];
					}

					if(TMPHANDLE != (char **) 0)
					{
						CommandItems[i] = TMPHANDLE;
						PtoCstr(command);
						strcpy(*TMPHANDLE, (char *) command);
						if(itemValue)
						{
							length--;
						}
						if(length > MAXITEMLENGTH)
						{
							length = MAXITEMLENGTH;
							command[length++] = 0xc9; /* '...' */
						}
						command[length] = '\0';
						CtoPstr((char *) command);
						SetItem(CommandMenu,command1+i,command);
						SetItemMark(CommandMenu, command1+i,
							(itemValue) ? diamondMark : noMark);
					} /*if(TMPHANDLE != (char **) 0)*/
				} /*for(i = 0;i < NCOMMANDS;i++)*/
			} /*if(i == NCOMMANDS)*/
		} /*if(itemHit == ok)*/
		DisposDialog(theDialog);
		SetDAFont(systemFont);
		break;

	  case command1:
	  case command2:
	  case command3:
	  case command4:
	  case command5:
	  case command6:
	  case command7:
	  case command8:
		/* insert command in Command Window */
		insertCmd(*(CommandItems[item - command1]), false);
		break;

	} /*switch (item)*/

}/*DoCommand()*/

/*
	Process item selected from Options sub menu
*/


static void setNsig(char * command)
{
	DialogPtr       theDialog;
	Rect            r;
	Handle          itemHandle;
	long            nsig;
	Str255          string;
	Integer         itemType, itemHit;
	Integer         textItem = 5;
	WHERE("setNsig");

	theDialog = getDlog(NSIGDLOG, true, 2, "yn");
	
	nsig = (PRINTFORMAT[2] == 'g') ? PRINTFORMAT[1] : DEFAULTPRINTDEC;

	setDlogNumber(theDialog, textItem, (double) nsig, "%g", true);

	ShowWindow((WindowPtr) theDialog);

	do
	{
		ModalDialog((ModalFilterProcPtr) myDialogFilter, &itemHit);
	} while(itemHit > cancel);

	if(itemHit == ok)
	{
		GetDItem(theDialog, textItem, &itemType, &itemHandle, &r);
		GetIText(itemHandle, string);
		PtoCstr((char *) string);
		sprintf(command,"setoptions(nsig:%s)\n", (char *) string);
	}
	else
	{
		command[0] = '\0';
	}
	DisposDialog(theDialog);
	InitCursor();
} /*setNsig()*/

#define WFORMAT    0
#define PFORMAT    1

#define FLOATFMT   'g'
#define FIXEDFMT   'f'

static void setFormats(char * command)
{
	DialogPtr       theDialog;
	Rect            r;
	Handle          itemHandle;
	Integer         defaultButton = 3;
	Integer         radioButton1 = 6, radioButton3 = 9;
	Integer         textItem1 = 12, textItem2 = 14;
	Integer         value, digits;
	Str255          widthStr[2];
	Str255          digitsStr[2];
	Integer         itemType, itemHit, whichRadioButton;
	Integer         currentFormat, newFormat;
	char            modes[2];
	char           *floatlabel = "\pSignificant Digits:";
	char           *fixedlabel = "\pNumber of Decimals:";
	char           *fmt = "%g";
	WHERE("setFormats");

	theDialog = getDlog(FORMATSDLOG, true, 3, "ynd");

	sprintf((char *) widthStr[PFORMAT],"%ld",PRINTFORMAT[0]);
	CtoPstr((char *) widthStr[PFORMAT]);
	sprintf((char *) digitsStr[PFORMAT],"%ld",PRINTFORMAT[1]);
	CtoPstr((char *) digitsStr[PFORMAT]);
	modes[PFORMAT] = PRINTFORMAT[2];

	sprintf((char *) widthStr[WFORMAT],"%ld",WRITEFORMAT[0]);
	CtoPstr((char *) widthStr[WFORMAT]);
	sprintf((char *) digitsStr[WFORMAT],"%ld",WRITEFORMAT[1]);
	CtoPstr((char *) digitsStr[WFORMAT]);
	modes[WFORMAT] = WRITEFORMAT[2];

	GetDItem(theDialog, textItem1, &itemType, &itemHandle, &r);
	SetIText(itemHandle, widthStr[PFORMAT]);
	GetDItem(theDialog, textItem2, &itemType, &itemHandle, &r);
	SetIText(itemHandle, digitsStr[PFORMAT]);
	SelIText(theDialog, textItem1, 0, 32767);

	setDlogItemValues(theDialog, radioButton1, 2, 1<<PFORMAT);

	value = (modes[PFORMAT] == FLOATFMT) ? 0x01 : 0x02;
	setDlogItemValues(theDialog, radioButton3, 2, value);

	GetDItem(theDialog, textItem2-1, &itemType, &itemHandle, &r);
	SetIText(itemHandle, (value & 1) ? floatlabel : fixedlabel);

	ShowWindow((WindowPtr) theDialog);

	do /* while(itemHit > cancel) */
	{
		ModalDialog((ModalFilterProcPtr) myDialogFilter, &itemHit);

		currentFormat = (getDlogItemValues(theDialog, radioButton1, 2) == 0x01) ?
			WFORMAT : PFORMAT;

		GetDItem(theDialog, textItem1, &itemType, &itemHandle, &r);
		GetIText(itemHandle, widthStr[currentFormat]);
		GetDItem(theDialog, textItem2, &itemType, &itemHandle, &r);
		GetIText(itemHandle, digitsStr[currentFormat]);

		if(itemHit == defaultButton)
		{ /* set default value */
			digits = (currentFormat == PFORMAT) ?
				DEFAULTPRINTDEC : DEFAULTWRITEDEC;
			sprintf((char *) widthStr[currentFormat],"%ld",digits+7);
			CtoPstr((char *) widthStr[currentFormat]);
			GetDItem(theDialog, textItem1, &itemType, &itemHandle, &r);
			SetIText(itemHandle, widthStr[currentFormat]);

			sprintf((char *) digitsStr[currentFormat],"%ld",digits);
			CtoPstr((char *) digitsStr[currentFormat]);
			GetDItem(theDialog, textItem2, &itemType, &itemHandle, &r);
			SetIText(itemHandle, digitsStr[currentFormat]);

			SelIText(theDialog, textItem1, 0, 32767);

			modes[currentFormat] = 'g';
			setDlogItemValues(theDialog, radioButton3, 2, 0x01);
			GetDItem(theDialog, textItem2-1, &itemType, &itemHandle, &r);
			SetIText(itemHandle, floatlabel);
		}
		else if(itemHit > cancel)
		{ /* radio button hit */
			whichRadioButton = itemHit;
			if(whichRadioButton == radioButton1 ||
			   whichRadioButton == radioButton1 + 1)
			{ /* format or wformat */
				whichRadioButton -= radioButton1;
				newFormat = (whichRadioButton) ? PFORMAT : WFORMAT;
				setDlogItemValues(theDialog, radioButton1, 2, 1 << whichRadioButton);
				if(newFormat != currentFormat)
				{ /* value has changed */
					value = (modes[newFormat] == FLOATFMT) ? 0x01 : 0x02;
					setDlogItemValues(theDialog, radioButton3, 2, value);

					GetDItem(theDialog, textItem1, &itemType, &itemHandle, &r);
					SetIText(itemHandle, widthStr[newFormat]);

					GetDItem(theDialog, textItem2, &itemType, &itemHandle, &r);
					SetIText(itemHandle, digitsStr[newFormat]);

					SelIText(theDialog, textItem1, 0, 32767);

					GetDItem(theDialog, textItem2-1, &itemType, &itemHandle, &r);
					SetIText(itemHandle, (modes[newFormat] == FLOATFMT) ?
							 floatlabel : fixedlabel);
				} /*if(newFormat != currentFormat)*/
			} /*if(whichRadioButton == radioButton1 or radioButton1 + 1)*/
			else if(whichRadioButton == radioButton3 ||
					whichRadioButton == radioButton3 + 1)
			{ /* fixed or floating */
				whichRadioButton -= radioButton3;
				setDlogItemValues(theDialog, radioButton3, 2,
								  1 << whichRadioButton);
				modes[currentFormat] = (whichRadioButton == 0) ?
					FLOATFMT : FIXEDFMT;
				GetDItem(theDialog, textItem2-1, &itemType, &itemHandle, &r);
				SetIText(itemHandle, (whichRadioButton==0) ?
						 floatlabel : fixedlabel);
			}
		} /*if(itemHit > cancel)*/
	} while(itemHit > cancel);

	if(itemHit == ok)
	{
		PtoCstr(widthStr[PFORMAT]);
		PtoCstr(digitsStr[PFORMAT]);
		PtoCstr(widthStr[WFORMAT]);
		PtoCstr(digitsStr[WFORMAT]);
		sprintf(command,"setoptions(format:\"%s.%s%c\",wformat:\"%s.%s%c\")\n",
				widthStr[PFORMAT], digitsStr[PFORMAT], modes[PFORMAT],
				widthStr[WFORMAT], digitsStr[WFORMAT], modes[WFORMAT]);
	} /*if(itemHit == ok)*/
	else
	{ /* cancel */
		command[0] = '\0';
	}
	DisposDialog(theDialog);
	InitCursor();
} /*setFormats()*/

static void setSeeds(char * command)
{
	DialogPtr       theDialog;
	Rect            r;
	Handle          itemHandle;
	Str255          string1, string2;
	Integer         itemType, itemHit;
	Integer         textItem1 = 5, textItem2 = 7;
	char           *fmt = "%.15g";
	WHERE("setSeeds");

	theDialog = getDlog(SEEDSDLOG, true, 2, "yn");
	
	setDlogNumber(theDialog, textItem2, (double) RANDS2, fmt, false);
	setDlogNumber(theDialog, textItem1, (double) RANDS1, fmt, true);
	ShowWindow((WindowPtr) theDialog);

	do
	{
		ModalDialog((ModalFilterProcPtr) myDialogFilter, &itemHit);
	} while(itemHit > cancel);

	if(itemHit == ok)
	{
		GetDItem(theDialog, textItem1, &itemType, &itemHandle, &r);
		GetIText(itemHandle, string1);
		PtoCstr((char *) string1);
		GetDItem(theDialog, textItem2, &itemType, &itemHandle, &r);
		GetIText(itemHandle, string2);
		PtoCstr((char *) string2);
		sprintf(command,"setoptions(seeds:cat(%s,%s))\n",
				(char *) string1, (char *) string2);
	}
	else
	{
		command[0] = '\0';
	}
	DisposDialog(theDialog);
	InitCursor();
} /*setSeeds()*/

static void setAngles(char * command)
{
	DialogPtr       theDialog;
	Rect            r;
	Handle          itemHandle;
	Integer         radioButton1 = 5;
	Integer         value;
	Integer         itemType, itemHit;
	Integer         i;
	double          angles2cycles = 8.0*atan(1.0);
	double          angles2degrees = atan(1.0)/45.0;
	Str255          string;
	WHERE("setAngles");

	theDialog = getDlog(ANGLESDLOG, false, 2, "yn");

	value = (ANGLES2RADIANS == angles2degrees) ? 0x01 :
		((ANGLES2RADIANS == angles2cycles) ? 0x02 : 0x04);
	setDlogItemValues(theDialog, radioButton1, 3, value);

	ShowWindow((WindowPtr) theDialog);

	do
	{
		ModalDialog((ModalFilterProcPtr) myDialogFilter, &itemHit);
		if(itemHit > cancel)
		{
			i = itemHit - radioButton1;
			setDlogItemValues(theDialog, radioButton1, 3, 1 << i);
		} /*if(itemHit > cancel)*/
	} while(itemHit > cancel);

	if(itemHit == ok)
	{
		value = getDlogItemValues(theDialog, radioButton1, 3);
		itemHit = radioButton1 + ((value & 1) ? 0 : ((value & 2) ? 1 : 2));
		GetDItem(theDialog, itemHit, &itemType, &itemHandle, &r);
		GetCTitle((ControlHandle) itemHandle, string);
		PtoCstr(string);
		string[0] = tolower(string[0]);
		sprintf(command,"setoptions(angles:\"%s\")\n", (char *) string);
	} /*if(itemHit == ok)*/
	else
	{
		command[0] = '\0';
	}
	DisposDialog(theDialog);
	InitCursor();

}

static void setBatchOptions(char * command)
{
	DialogPtr       theDialog;
	Rect            r;
	Handle          itemHandle;
	Integer         radioButton1 = 4;
	Integer         textItem = 7;
	Integer         value;
	Str255          string;
	Integer         itemType, itemHit;
	Integer         i;
	WHERE("setBatchOptions");

	theDialog = getDlog(BOPTIONSDLOG, true, 2, "yn");

	setDlogNumber(theDialog, textItem, (double) MAXERRORS, "%g", true);

	setDlogItemValues(theDialog, radioButton1, 2,
					  (DEFAULTBATCHECHO) ? 0x01 :  0x02);

	ShowWindow((WindowPtr) theDialog);

	do
	{
		ModalDialog((ModalFilterProcPtr) myDialogFilter, &itemHit);
		if(itemHit > cancel)
		{
			i = itemHit - radioButton1;
			setDlogItemValues(theDialog, radioButton1, 2, 1 << i);
		} /*if(itemHit > cancel)*/
	} while(itemHit > cancel);

	if(itemHit == ok)
	{
		value = getDlogItemValues(theDialog, radioButton1, 2);
		GetDItem(theDialog, textItem, &itemType, &itemHandle, &r);
		GetIText(itemHandle, string);
		PtoCstr((char *) string);
		sprintf(command,"setoptions(batchecho:%c,errors:%s)\n",
				(value & 1) ? 'T' : 'F', string);
	} /*if(itemHit == ok)*/
	else
	{
		command[0] = '\0';
	}
	DisposDialog(theDialog);
	InitCursor();

} /*setBatchOptions()*/

pascal void DoOptions(Integer item)
{
	char       command[50];
	WHERE("DoOptions");
	
	switch (item)
	{
	  case insig:
		setNsig(command);
		break;

	  case iformats:
		setFormats(command);
		break;

	  case iseeds:
		setSeeds(command);
		break;

	  case iangles:
		setAngles(command);
		break;

	  case iboptions:
		setBatchOptions(command);
		break;

	} /*switch (item)*/
	if(command[0] != '\0')
	{
		insertCmd(command, true);
	}
}

pascal void DoFont(Integer item)
{
	Str255      fontName;
	Integer     fontNumber;
	WHERE("DoFont");
	
	GetItem(FontMenu, item, fontName);
	GetFNum(fontName, &fontNumber);
	setCmdWindFont(fontNumber, CmdFontSize);
} /*DoFont()*/

pascal void DoFontSize(Integer item)
{
	Str255      pointSize;
	int         fontSize;
	WHERE("DoFontSize");
	
	GetItem(FontSizeMenu, item, pointSize);
	PtoCstr((char *) pointSize);
	sscanf((char *) pointSize, "%d", &fontSize);
	setCmdWindFont(CmdFont, (Integer) fontSize);
} /*DoFontSize()*/

/*
	Process item selected from Edit menu.  First check whether it should
	get routed to a desk accessory or not.  If not, then for route the
	selection to the text editing window, as that is the only one for
	this application to which edit commands are valid.

	(The test of FrontWindow is not strictly necessary, as the Edit
	menu is disabled when any of the other windows is frontmost, and so
	this Proc couldn't be called.)
*/

pascal void DoEdit(item)
Integer             item;
{
	WindowPtr          theWindow;
	WHERE("doEdit");

	if (!SystemEdit(item - 1))	/* check DA edit choice */
	{
		theWindow = FrontWindow();
		if (UseWindows > 0 && theWindow == CmdWind)
		{
			CmdWindEditMenu(item);
		}
		else if(whichGraphWindow(theWindow) >= 0 )
		{
			GraphWindEditMenu(item,theWindow);
		}
	} /*if (!SystemEdit(item - 1)) */
} /*DoEdit()*/


/*
	Miscellaneous routines
	These take care of drawing the grow box and the line along
	the right edge of the window, and of setting and resetting the clip
	region to disallow drawing in that right edge by the other drawing
	routines.
*/


void DrawGrowBox(wind)
WindowPtr       wind;
{
	Rect            r;
	RgnHandle       oldClip;

	r = wind->portRect;
	r.left = r.right - 15;	/* draw only along right edge */
	oldClip = NewRgn();
	GetClip(oldClip); /* save clipping region of current grafPort */
	ClipRect(&r); /* make r the clipping region */
	DrawGrowIcon(wind);
	SetClip(oldClip); /* restore previous clipping region */
	DisposeRgn(oldClip);
} /*DrawGrowBox()*/

#ifdef UNDEFINED__ /* following two functions not currently referenced */

/* shift right boundary of clipping region for window in by 15*/
void SetWindClip(wind)
WindowPtr       wind;
{
	Rect            r;

	r = wind->portRect;
	r.right -= 15;		/* don't draw along right edge */
	oldClip = NewRgn();
	GetClip(oldClip);  /* save clipping region of current grafPort in global */
	ClipRect(&r);
} /*SetWindClip()*/


void ResetWindClip()
{
	SetClip(oldClip);   /* restore clipping region previously saved in global */
	DisposeRgn(oldClip);
} /*ResetWindClip()*/
#endif /*UNDEFINED__*/

/*
	Function to terminate MacAnova
	Unless SaveOnQuit is false, the user is given the opportunity to save
	the workspace and any open text windows.
	Unless FatalError is non-zero, at any of the stages, quitting may be
	aborted by clicking Cancel in a dialog box.
*/

Integer quitIt()
{
	Integer          i;
	Integer          windno, button = ok;
	Str255           wTitle;
	Boolean          optionkey;
	cmdWindowInfoPtr wp;
	WHERE("quitIt");

	optionkey = (SkelGetCurrentEvent()->modifiers & optionKey) != 0;
	GraphWindPaused = false;
/*
	The following is supposed to copy the last few lines of the command window
	to the Clipboard in case of a fatal error, but it does not seem always
	to work.  Unless a copy was done in MacAnova, on exit, the Clipboard
	holds what it did before invoking MacAnova
*/
	if(FatalError && UseWindows > 0)
	{ /* should copy last few lines of command window to Clip Board */
		onFatalError();
	}

	/* give use a chance to save the workspace */
	if(SaveOnQuit && !optionkey)
	{
		if((button = saveAlert((Str255) 0, false)) == cancel && !FatalError)
		{
			return (cancel);
		}
		if(button == ok)
		{
			DoFile(keepitas);
			if(UseWindows > 0 && !FatalError && !Reply.good)
			{
				return (cancel);
			}
		}
	} /*if(SaveOnQuit && !optionkey)*/

	/* Give use a change to save dirty windows before quitting */
	if(UseWindows > 0  && SaveOnQuit && !optionkey)
	{
		for(windno = 0, wp = CmdWindows;windno < MAXWINDOWS; windno++, wp++)
		{
			if(wp->cmdWind != (WindowPtr) 0 && wp->cmdEditable && wp->cmdDirty)
			{
				GetWTitle(wp->cmdWind, wTitle);
				if((button = saveAlert(wTitle, true)) == cancel && !FatalError)
				{
					return (button);
				}
				if(button == ok)
				{
					doSave(wp, quit);
					if(!FatalError && !Reply.good)
					{
						return (cancel);
					}
				} /*if(button == ok)*/
			} /*if(wp->cmdWind!=(WindowPtr) 0 && wp->cmdEditable && wp->cmdDirty)*/
		} /*for(windno = 0, wp = CmdWindows;windno < MAXWINDOWS; windno++, wp++)*/
	} /*if(UseWindows > 0  && SaveOnQuit && !optionkey)*/

	if(!UseWindows)
	{
		myAlert("End of non-interactive MacAnova run");
	}

#ifdef PERFORMANCE
	{
		OSErr    err;

		if(ProfileStatus > 0)
		{
			err = PerfDump(ThePGlobals,PERFORMOUT,DOHISTOGRAM,
						   RPTFILECOLUMNS);
			if(err)
			{
				ALERT("PerfDump reports err = %d",err,0,0,0);
			}
		}
		TermPerf(ThePGlobals);
	}
#endif /*PERFORMANCE*/

	SkelStopEventLoop();

/*	Close all files */
	if(INPUTFILE[0] != STDIN)
	{
		fclose(INPUTFILE[0]);
	}
	if(SPOOLFILE != (FILE *) 0)
	{
		fclose(SPOOLFILE);
	}
	if(HELPFILE != (FILE *) 0)
	{
		fclose(HELPFILE);
	}
	closeBatch(1);  /* shut down all batch files*/
	for(i=0;i<2;i++)
	{
		if(PrintHdl[i] != (THPrint) 0)
		{
			DisposHandle((Handle) PrintHdl[i]);
		}
	}
	myDispAll(); /* release all MacAnova handles */

	return (ok);

} /*quitIt()*/

/* Write length characters starting at linePtr to file */
static OSErr saveLine(Integer refNum, char *linePtr, LongInt length)
{
	return (FSWrite(refNum,&length,linePtr));
}
/*
	Save contents of command window in text file
*/

void doSave(wp, item)
cmdWindowInfoPtr   wp;
Integer            item;
{
	OSErr          errorFlag;
	FInfo          fndrInfo;
	Integer        refNum, vRefNum;
	Integer        ask;
	Str255         wTitle;
	char          *fileName, *msgs;
	WHERE("doSave");

	if(!UseWindows)
	{
		return;
	}
	vRefNum = wp->cmdVRefNum;
	if(vRefNum == 0 && wp->cmdDirty)
	{
		item = saveitas;
	}
	if(item == saveit && !wp->cmdDirty)
	{
		Reply.good = true;
		return;
	}

	ask = (item == saveitas || item == quit);

	GetWTitle(wp->cmdWind, wTitle);

	if(!ask)
	{
		SetVol(0L, vRefNum);
	} /*if(!ask)*/
	else
	{
		fileName = macOpen("\pFile for output window:", wTitle, WRITEIT,
			0, (OSType *) 0);
		if (fileName != (char *) 0)
		{
			vRefNum = Reply.vRefNum;
			strcpy((char *) wTitle, fileName);
			CtoPstr((char *) wTitle);
		}
		else
		{
			return;
		}
	} /*if(!ask){}else{}*/

	errorFlag = GetFInfo(wTitle, vRefNum, &fndrInfo);

	if(errorFlag == noErr)
	{
	 /* if file exists, delete it */
		errorFlag = FSDelete(wTitle, vRefNum);
		if(errorFlag != noErr)
		{
			goto errorExit;
		}
	}
 /* file not found or deleted, create it */
	errorFlag = Create(wTitle, vRefNum, CREATOR, 'TEXT');
	if(errorFlag != noErr)
	{
		goto errorExit;
	}

	errorFlag = FSOpen(wTitle, vRefNum,&refNum);
	if(errorFlag != noErr)
	{
		goto errorExit;
	}

	HLock((**wp->teCmd).hText);
	errorFlag = saveLine(refNum, *(**wp->teCmd).hText, (**wp->teCmd).teLength);
	HUnlock((**wp->teCmd).hText);
	if(errorFlag != noErr)
	{
		goto errorExit;
	}

/* Write a couple of lines at end with date */
	msgs = "\n#Output Window saved ";
	errorFlag = saveLine(refNum, msgs, strlen(msgs));
	if(errorFlag != noErr)
	{
		goto errorExit;
	}
	msgs = getTimeAndDate();
	errorFlag = saveLine(refNum, msgs, strlen(msgs));
	if(errorFlag != noErr)
	{
		goto errorExit;
	}
	msgs = "\n#--------------------------------------------\n";
	errorFlag = saveLine(refNum, msgs, strlen(msgs));
	if(errorFlag != noErr)
	{
		goto errorExit;
	}
	wp->cmdVRefNum = vRefNum;
	SetWTitle(wp->cmdWind, wTitle);
	SetItem(WindowMenu, wp->menuItem, wTitle);

	FSClose(refNum);
	FlushVol ((StringPtr) 0, vRefNum);
	wp->cmdDirty = false;
	if(wp - CmdWindows == CurrentWindow)
	{
		CmdDirty = false;
	}
	PtoCstr(wTitle);
	macSetInfo(vRefNum,(char *) wTitle,'TEXT',CREATOR);

	return;

  errorExit:
	PtoCstr(wTitle);
	fileErrorMsg(errorFlag, wTitle);
	Reply.good = false;
} /*doSave()*/

/*
	Replace contents of command window with contents of text file
*/
Integer loadWindow(fileName, vRefNum, new)
char           *fileName;
Integer         vRefNum;
Boolean         new;
{
	LongInt		logEOF;
	LongInt     length;
	Integer     refNum;
	OSErr       errorFlag = 0;
	Handle      hText = (Handle) 0;
	Integer     truncated = 0, windno, item;
	Str255      wTitle;
	WHERE("loadWindow");

	if(new && Nwindows == MAXWINDOWS)
	{
		myAlert("Too many windows already open");
		return (1);
	}
	if(fileName == (char *) 0)
	{
		fileName = macOpen("\pFile for command window",(Str255) 0, READIT,
			0, (OSType *) 0); /* macOpen default type is TEXT */
		if(fileName == (char *) 0)
		{
			return (1);
		}
		vRefNum = Reply.vRefNum; /*global set by macOpen() */
	}
	SetVol(0L, vRefNum);
	strcpy((char *) wTitle, fileName);
	CtoPstr((char *) wTitle);
	errorFlag = FSOpen(wTitle,vRefNum, &refNum);

	if(errorFlag != noErr)
	{
		myprint("WARNING: unable to open file ");
		myerrorout(fileName);
	} /*if(errorFlag != noErr)*/
	else
	{
		errorFlag = GetEOF(refNum,&logEOF);
		if(errorFlag == noErr)
		{
			if(logEOF > CMDWINDOWNEWSIZE)
			{
				errorFlag = SetFPos(refNum, fsFromStart, logEOF-CMDWINDOWNEWSIZE);
				truncated = logEOF - CMDWINDOWNEWSIZE;
				logEOF = CMDWINDOWNEWSIZE;
			}
			if(errorFlag == noErr)
			{
				if((hText = NewHandle (logEOF)) == (Handle) 0)
				{
					myerrorout("ERROR: insufficient free memory");
					return(2);
				}
				length = logEOF;
				HLock(hText);
				errorFlag = FSRead(refNum, &length, (Ptr) (*hText));
				HUnlock(hText);
				if(errorFlag == noErr && length != logEOF)
				{
					errorFlag = eofErr;
				}

				if(errorFlag != noErr)
				{
					DisposHandle(hText);
				}
			} /*if(errorFlag == noErr)*/

			if(errorFlag == noErr)
			{
				if(new)
				{
					windno = createWindow(wTitle);
					if(windno < 0)
					{
						myAlert("Cannot create new window");
						FatalError = 1;
						quitIt();
					}
					DisposHandle((Handle) (**CmdWindows[windno].teCmd).hText);
					item = cmd1 + Nwindows - 1;
					InsMenuItem(WindowMenu,wTitle,item - 1);
					CmdWindows[windno].menuItem = item;
					setCommandM(CurrentWindow); /* set menu key to toggle back*/
					restoreWindowInfo(windno);
				} /*if(new)*/
				else
				{
					DisposHandle((Handle) (**teCmd).hText);
					SetWTitle(CmdWind, wTitle);
					SetItem(WindowMenu,
							CmdWindows[CurrentWindow].menuItem,	wTitle);
				} /*if(new){}else{}*/
				(**teCmd).hText = hText;
				(**teCmd).teLength = length;
				TESetSelect(length, length, teCmd);
				TECalText(teCmd);

				CmdInsertLine = GetCmdIL();
				ScrollToInsertPt();
				CmdWindows[CurrentWindow].cmdVRefNum = vRefNum;
				CmdWindows[CurrentWindow].cmdDirty = CmdDirty = false;
				CmdEditable = true;
				CmdStrPos = length;
				saveWindowInfo(CurrentWindow);

				InvalRect(&(CmdWind->portRect));
				macUpdate(CmdWind);

				if(truncated)
				{
					sprintf(OUTSTR,
							"WARNING: file too long; only last %ld characters loaded",
							(long) CMDWINDOWNEWSIZE);
					myAlert(OUTSTR);
					putOUTSTR();
				}
			} /*if(errorFlag == noErr)*/
			FSClose(refNum);
		} /*if(errorFlag == noErr)*/
		if(errorFlag != noErr)
		{
			myprint("WARNING: problems loading file ");
			myerrorout(fileName);
		}
	} /*if(errorFlag != noErr){}else{}*/
	fileErrorMsg(errorFlag, fileName);
	return (errorFlag);
} /*loadWindow()*/

/*
	Dummy up restore(fileName) command and execute it.

	doRestore appears in main.c in non-mac versions; the ifdefs are included
	to make the code identical
*/
#ifdef MACINTOSH
int doRestore(fileName, vRefNum)
char            *fileName;
Integer          vRefNum;
#else
int     doRestore(fileName)
char            *fileName;
#endif /*MACINTOSH*/
{
	Symbolhandle list;
	long         typeList[1];
	char        *valueList[1];
	char        *keywordList[1];
	FILE        *restoreFile;
	int          errorFlag = 0;
#ifdef MACINTOSH
	OSType       types[2];

	if(fileName == (char *) 0)
	{
		types[0] = KEEPFILETYPE;
		types[1] = ASCIIKEEPFILETYPE;
		fileName = macOpen("\pMacAnova workspace file",(Str255) 0, READIT,
			2, types);
		if(fileName == (char *) 0)
		{
			return (1);
		}
		vRefNum = Reply.vRefNum; /*global set by macOpen() */
	}
	SetVol(0L, vRefNum);
#endif /*MACINTOSH*/

	restoreFile = fopen(fileName,"r");
	if(restoreFile == (FILE *) 0)
	{
		myprint("WARNING: unable to open file ");
		myerrorout(fileName);
		errorFlag = 1;
	}
	else
	{
		fclose(restoreFile);

		/* dummy up restore(filename) */
		typeList[0] = CHAR;
		valueList[0] = fileName;
		keywordList[0] = "";
		list = Buildlist(1,typeList,valueList,keywordList);
		if(list == (Symbolhandle) 0)
		{
			errorFlag = 1;
		}
		else
		{
			if(restore(list) == (Symbolhandle) 0)
			{
				errorFlag = 1;
			}
#ifdef MACINTOSH
			UnloadSeg((Ptr) restore);
#endif /*MACINTOSH*/
			Removelist(list);
		}
		if(errorFlag)
		{
			myprint("WARNING: unable to restore workspace from file ");
			myerrorout(fileName);
		}
	}
	return (errorFlag);
} /*doRestore()*/


