
/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Mainpars
#else
#define __SEG__ Mainpars
#endif /*MPW3*/
#endif /*MPW*/

/*
  main parser for macanova by G. Oehlert, as modifed by C. Bingham for use
  with MacAnova 3.0
*/

/*
	Version of 920623 made many changes, here and in yylex.c and Symbol.c
	All names are returned by yylex as VAR unless they are BLTINs or MACROs
	followed by '('.  Thus the parser should never see STRUC, UNDEF, and should
	never see MACRO or BLTIN or BATCH unless followed by '('
	The number of shift/reduce conflicts has been reduced to 13

	930201 Added 'break n' functionality which also is associated with token
	BREAK, and modified yylex so that 'breakall' is equivalent to
	'break WDEPTH'.  When n is a literal positive integer, 'break n' results
	in macanova	breaking out of exactly n enclosing loops.

	930208 Added a[..][..] capability, now 17 shift/reduce errors

	930804 Added operators <-+ <-- <-*, <-/, <-^, <-%%; 23 shift/reduce errors

	931007 Added lines to make use of batch() as an expression illegal.

	931103 Added error checking for left side of <-+, <--, etc; 47 S/R errors

	931110 Communication with Ifsetup() now by return value rather than global

	931228 Found that mainpars.c as directly produced was too large for the
           Mac MPW compiler.  Accordingly, rearranged various cases and used
		   more gotos, to maximize the length of sequences of case statements
           with identical code that could be folded into one another by
           editing of mainpars.c for use on the Macintosh.  The first case
           of a set of, say 3, cases to be folded has a comment of the
		   form 'Mac: fold 3 cases'. Script mainpars.awk for awk uses this
		   to create mainpars.c.mac.new.  This also has the parser tables
		   edited out and put into file mainpars.r.mac.new in a form readable
		   by MPW resource compiler Rez.

		   Note: Some of the folded cases are not identical in mainpars.y, but
		   are translated into identical code by Yacc.  Conversely, some
		   cases that appear to be identical in mainpars.y are translated
		   by Yacc in different ways.
	
	940110 Fixed bug in byname BATCH parenlist

	940119 Found and fixed bug.  yyparse() was telling yylex() to send an
	unnecessary (and harmful) ';' after if(l){..}else{...} and after
	while(l){...} .

	940331 Added bitwise operations BITAND, BITXOR, BITOR, and BITNOT

	940411 Added line adding  component ASSIGN  to definition of 'assign',
	in order to generate useful error message.
	Now 48 shift/reduce conflicts

	940630 Implemented if{...}elseif{...}...elseif{...}else{...}
	Now 49 shift/reduce

	950411 Fixed bug in elseif ({...} not always skipped on False
	950710 Added parserCopy() to fix bug interpreting statements like
    {b<-1}+{b<-2}

	950722 Restricted use of parserCopy() to eliminate buildup of scratch
	variables.  {b<-1}+{b<-2} again will give the wrong answer.  However,
	if the {...} is really an expanded macro it will give the right answer
	This can still result in the buildup of scratch is the value of a
	macro is not assigned.  Hence, guidelines should suggest that
	macros not returning a value should end with ';;'

	950807 Added check for BLTIN and LIST to parserCopy()

VERSION: 950807   This date is used by mainpars.awk in making macMainpars.r

*/

#include "globals.h"

#ifdef MACINTOSH

#define CHECKINTERRUPT	if(INTERRUPT != INTNOTSET) YYERROR

#else /*MACINTOSH*/

#define CHECKINTERRUPT

#endif /*MACINTOSH*/

/* these must be defined identically in yylex.c */
#define SKIPBRACKETBLOCK  3
#define SKIPPARENBLOCK    4

#define YYSTYPE Symbolhandle 	/* handle to symbol */
#define YYDEBUG 1

extern Symbolhandle Install(), Extract();
extern Symbolhandle Makelist(), Addlist(), Growlist();
extern Symbolhandle Assign(), Subassign(), Arith();
extern Symbolhandle Unary(), Logic(), Element();
extern Symbolhandle Byname();
extern void         Removelist();
extern void         Unscratch();
extern void         RemoveUndef();

void                cleanitup();

#ifdef MACINTOSH
extern short *Pyyexca;
extern short *Pyyact;
extern short *Pyypact;
extern short *Pyypgo;
extern short *Pyyr1;
extern short *Pyyr2;
extern short *Pyychk;
extern short *Pyydef;
#endif /*MACINTOSH*/

static void         setupfakefor();
static long         BynameType;
static long         ParseMacroArgs;
static int          Op, AssignOp;
static int          SeenError = 0;
static int          TargetLevel;

#ifdef UNDEFINED__
static WHERE("yyparse"); /*now set by make*/
#endif /*UNDEFINED__*/

long    PARSEMACROARGS = 0;  /* for communication with yylex */
long    YYLEXMESSAGE = 0;  /* for communication with yylex */
long    LASTCH = '\0';  /* for communication with yylex */
int     ALTINPUTTOKENS[5*MAXWDEPTH];  /* for communication with yylex */
int    *ALTTOKENPTR;  /* for communication with yylex */
/*
   End of MacAnova Code before parser instructions
*/
#ifdef __cplusplus
#  include <stdio.h>
#  include <yacc.h>
#endif	/* __cplusplus */ 
# define CHAR 257
# define REAL 258
# define BLTIN 259
# define LIST 260
# define STRUC 261
# define MACRO 262
# define UNDEF 263
# define VAR 264
# define END 265
# define LOGIC 266
# define MACRO1 267
# define NUMBER 268
# define IF 269
# define ELSE 270
# define ELSEIF 271
# define WHILE 272
# define FOR 273
# define BREAK 274
# define BREAKALL 275
# define BATCH 276
# define ERROR 277
# define FATALERROR 278
# define PLOTINFO 279
# define ASSIGNED 280
# define GARB 281
# define LONG 282
# define BADTOKEN 283
# define LEFTANGLE 284
# define RIGHTANGLE 285
# define LEFTBRACK2 286
# define RIGHTBRACK2 287
# define MULT 288
# define ADD 289
# define DIV 290
# define MOD 291
# define SUB 292
# define EXP 293
# define GE 294
# define LE 295
# define GT 296
# define LT 297
# define EQ 298
# define NE 299
# define AND 300
# define OR 301
# define NOT 302
# define ENDOFL 303
# define POW 304
# define MATMULT 305
# define TRMATMULT 306
# define MATMULTTR 307
# define BITAND 308
# define BITXOR 309
# define BITOR 310
# define BITNOT 311
# define ASSIGN 312
# define ASSIGN1 313
# define ASSIGNMULT 314
# define ASSIGNADD 315
# define ASSIGNDIV 316
# define ASSIGNSUB 317
# define ASSIGNPOW 318
# define ASSIGNMOD 319
# define UNARYMINUS 320
#define yyclearin yychar = -1
#define yyerrok yyerrflag = 0
extern int yychar;
#ifndef YYMAXDEPTH
#define YYMAXDEPTH 150
#endif

/* __YYSCLASS defines the scoping/storage class for global objects
 * that are NOT renamed by the -p option.  By default these names
 * are going to be 'static' so that multi-definition errors
 * will not occur with multiple parsers.
 * If you want (unsupported) access to internal names you need
 * to define this to be null so it implies 'extern' scope.
 * This should not be used in conjunction with -p.
 */
#ifndef __YYSCLASS
# define __YYSCLASS static
#endif
#ifndef YYSTYPE
#define YYSTYPE int
#endif
YYSTYPE yylval;
__YYSCLASS YYSTYPE yyval;
typedef short yytabelem;
# define YYERRCODE 256


void yyerror(s)
char           *s;
{
	/*  output error message */
	unsigned char   c = '\0', *inputString = *INPUTSTRING;
	int             start = 0;
	int             istrchar = ISTRCHAR;
	int             iswarning = (strncmp("WARNING:", s, 8) == 0);
	WHERE("yyerror");

	if (!SeenError && (PRINTWARNINGS || !iswarning))
	{
		while(istrchar >= 0)
		{
			if((c = inputString[istrchar]) == '#' || c == '"')
			{
				break;
			}
			istrchar--;
		}
		if(c == '#')
		{
			ISTRCHAR = istrchar;
		}
		
		myprint(s);

		myprint(" near ");
		if(strlen(s) + 6 + ISTRCHAR > SCREENWIDTH)
		{
			myeol();
		}
		
		if(ISTRCHAR > SCREENWIDTH && !(GUBED & 2))
		{
			start = ISTRCHAR - (SCREENWIDTH - 5);
			start = (start >= 0) ? start : 0;
			myprint(". . .");
		}

		c = inputString[ISTRCHAR];
		inputString[ISTRCHAR] = '\0';
		mymultiprint((char *) inputString + start);
		inputString[ISTRCHAR] = c;
		SeenError = !iswarning;
	} /*if (!SeenError && (PRINTWARNINGS || !iswarning))*/
	
	if(s == OUTSTR)
	{
		*OUTSTR = '\0';
	}
} /*yyerror()*/

/* tell yylex to report  ';for()' to the parser */
static int fakefortokens[] = {(int) ';', FOR, (int) '(', (int) ')', 0 };
static void setupfakefor(void)
{
	int   i;

	for(i=0;fakefortokens[i];i++)
	{
		ALTINPUTTOKENS[i] = fakefortokens[i];
	 }
	ALTINPUTTOKENS[i] = 0;
} /*setupfakefor()*/

static void notfunction(Symbolhandle symh)
{
	if(isscratch(NAME(symh)))
	{
		yyerror("ERROR: problem with input");
	}
	else
	{
		sprintf(OUTSTR,"ERROR: %s is not function or macro",NAME(symh));
		yyerror(OUTSTR);
		*OUTSTR = '\0';
	}
} /*notfunction()*/

/*
   Clean up looping related globals
*/
static void cleanLoops(long level1, long level2)
{
	int      i;
	
	for(i=level1;i<level2;i++)
	{
		WHILECONDITIONS[i] = 0;
		WHILELIMITS[i] = MAXWHILE;
		mydisphandle((char **) FORVECTORS[i]);
		FORVECTORS[i] = (double **) 0;
	}
} /*cleanLoops()*/

static Symbolhandle runCommand(Symbolhandle func, Symbolhandle list)
{
	Symbolhandle      result;
	
	strcpy(FUNCNAME,NAME(func));
	result = (*(FPTR(func)))(list);
#ifdef MACINTOSH
	UNLOADSEG(*(FPTR(func)));
#endif /*MACINTOSH*/
	Removelist(list);
	return (result);
} /*runCommand()*/

/*
   If symh is a scratch symbol, return it; otherwise return a scratch
   copy of symh
*/

static Symbolhandle parserCopy(Symbolhandle symh)
{
	Symbolhandle     symhCopy;
	long             type;

	if (symh == (Symbolhandle) 0 || isNull(symh) ||
		isscratch(NAME(symh)) || (type = TYPE(symh)) == UNDEF ||
		type == ASSIGNED || type == BLTIN || type == LIST)
	{
		return (symh);
	}
	symhCopy = Makesymbol(TYPE(symh));
	if (symhCopy != (Symbolhandle) 0 && Copy(symh, symhCopy))
	{
		setNAME(symhCopy, SCRATCH);
		Addsymbol(symhCopy);
	} /*if (symhCopy != (Symbolhandle) 0 && Copy(symh, symhCopy))*/
	else
	{
		Delete(symhCopy);
		symhCopy = (Symbolhandle) 0;
	}
	return (symhCopy);
} /*parserCopy()*/

/*
  Called from main before every invocation of getinput.
  It needs to be here because it resets PARSEMACROARGS, ALTTOKENPTR, and
  ALTINPUTTOKENS which are defined only here and in yylex.c
  It also resets static variable SeenError
*/

/* global for communication with Assign() in Symbol.c */
extern int     Nassigned;

 /* globals maintained by Addsymbol & Cutsymbol*/
extern long    Nscratch, Maxscratch;
extern long    Nsymbols, Maxsymbols;

 /* globals defined in comhandl.c */
extern long    Nhandles, Maxhandles;
extern long    CurrentMemory, MaxMemory;

/*
   globals for communication with handle manager in unxhandl.c, macHandle.c,...
*/
extern short   Nwarehouses;          /* number of warehouses (chaches)*/
extern short   WarehouseLength[];    /* number of handles in warehouses */

#ifdef COUNTMISSES
extern short   WarehouseMisses[];    /* number of failures to find handle */
extern short   WarehouseHits[];    /* number of successes to find handle */
#endif /*COUNTMISSES*/

void memoryUsage(char *NAMEFORWHERE)
{
	int       i;

	PRINT("Nscratch = %ld, Maxscratch = %ld, Nsymbols = %ld, Maxsymbols = %ld\n",
		  Nscratch, Maxscratch, Nsymbols, Maxsymbols);
	PRINT("WarehouseLength[] = %d, %d, %d, %d",
		  (int) WarehouseLength[0], (int) WarehouseLength[1],
		  (int) WarehouseLength[2], (int) WarehouseLength[3]);
	PRINTMORE(", %d, %d, %d, %d",
		  (int) WarehouseLength[4], (int) WarehouseLength[5],
		  (int) WarehouseLength[6], (int) WarehouseLength[7]);
	for (i = 8; i < Nwarehouses; i++)
	{
		PRINTMORE(", %d",(int) WarehouseLength[i],0,0,0);
	}
	PRINTMORE("\n",0,0,0,0);

#ifdef COUNTMISSES
	PRINT("WarehouseMisses[] = %d, %d, %d, %d",
		  (int) WarehouseMisses[0], (int) WarehouseMisses[1],
		  (int) WarehouseMisses[2], (int) WarehouseMisses[3]);
	PRINTMORE(", %d, %d, %d, %d",
		  (int) WarehouseMisses[4], (int) WarehouseMisses[5],
		  (int) WarehouseMisses[6], (int) WarehouseMisses[7]);
	for (i = 8; i < Nwarehouses; i++)
	{
		PRINTMORE(", %d",(int) WarehouseMisses[i],0,0,0);
	}
	PRINTMORE("\n",0,0,0,0);

	PRINT("WarehouseHits[] = %d, %d, %d, %d",
		  (int) WarehouseHits[0], (int) WarehouseHits[1],
		  (int) WarehouseHits[2], (int) WarehouseHits[3]);
	PRINTMORE(", %d, %d, %d, %d",
		  (int) WarehouseHits[4], (int) WarehouseHits[5],
		  (int) WarehouseHits[6], (int) WarehouseHits[7]);
	for (i = 8; i < Nwarehouses; i++)
	{
		PRINTMORE(", %d",(int) WarehouseHits[i],0,0,0);
	}
	PRINTMORE("\n",0,0,0,0);
#endif /*COUNTMISSES*/

	PRINT("Nhandles = %ld, Maxhandles = %ld, Nassigned = %ld\n",
		  Nhandles, Maxhandles, Nassigned, 0);
	PRINT("CurrentMemory = %ld, MaxMemory = %ld\n",
		  CurrentMemory, MaxMemory, 0, 0);
} /*void memoryUsage()*/

void cleanitup(void)
{
	WHERE("cleanitup");

	if (GUBED & 4096)
	{
		memoryUsage(NAMEFORWHERE);
	}
	
	SeenError = 0;

	Unscratch(); /* clears scratch symbol table (SCRATCH, keywords, UNDEF's)*/
	RemoveUndef(); /* clears ASSIGNED && temp variables from symbol table */
	Nassigned = 0;

	INTERRUPT = INTNOTSET;
	*OUTSTR = '\0';

	if(BDEPTH == 0)
	{
		NLINES = 0;
	}

	if(ISATTY && BDEPTH == 0 && MAXERRORS == 0)
	{
		NERRORS = 0;
	}

	restoreFormat(); /* restore current default format */

	BRACKETLEV = 0;
	PARENLEV = 0;
	MDEPTH = 0;
	WDEPTH = 0;
	IDEPTH = 0;
	MLEVEL = -1;
	PARSEMACROARGS = 0;
	ALTTOKENPTR = (int *) 0;
	ALTINPUTTOKENS[0] = 0;
	cleanLoops(0, MAXWDEPTH);
} /*cleanitup()*/
/* Deleted definition of __YYSCLASS yytabelem yyexca[] */
# define YYNPROD 207
# define YYLAST 1384
/* Deleted definition of __YYSCLASS yytabelem yyact[] */
/* Deleted definition of __YYSCLASS yytabelem yypact[] */
/* Deleted definition of __YYSCLASS yytabelem yypgo[] */
/* Deleted definition of __YYSCLASS yytabelem yyr1[] */
/* Deleted definition of __YYSCLASS yytabelem yyr2[] */
/* Deleted definition of __YYSCLASS yytabelem yychk[] */
/* Deleted definition of __YYSCLASS yytabelem yydef[] */
typedef struct { char *t_name; int t_val; } yytoktype;
#ifndef YYDEBUG
#	define YYDEBUG	0	/* don't allow debugging */
#endif

#if YYDEBUG

__YYSCLASS yytoktype yytoks[] =
{
	"CHAR",	257,
	"REAL",	258,
	"BLTIN",	259,
	"LIST",	260,
	"STRUC",	261,
	"MACRO",	262,
	"UNDEF",	263,
	"VAR",	264,
	"END",	265,
	"LOGIC",	266,
	"MACRO1",	267,
	"NUMBER",	268,
	"IF",	269,
	"ELSE",	270,
	"ELSEIF",	271,
	"WHILE",	272,
	"FOR",	273,
	"BREAK",	274,
	"BREAKALL",	275,
	"BATCH",	276,
	"ERROR",	277,
	"FATALERROR",	278,
	"PLOTINFO",	279,
	"ASSIGNED",	280,
	"GARB",	281,
	"LONG",	282,
	"BADTOKEN",	283,
	"LEFTANGLE",	284,
	"RIGHTANGLE",	285,
	"LEFTBRACK2",	286,
	"RIGHTBRACK2",	287,
	"MULT",	288,
	"ADD",	289,
	"DIV",	290,
	"MOD",	291,
	"SUB",	292,
	"EXP",	293,
	"GE",	294,
	"LE",	295,
	"GT",	296,
	"LT",	297,
	"EQ",	298,
	"NE",	299,
	"AND",	300,
	"OR",	301,
	"NOT",	302,
	"ENDOFL",	303,
	"POW",	304,
	"MATMULT",	305,
	"TRMATMULT",	306,
	"MATMULTTR",	307,
	"BITAND",	308,
	"BITXOR",	309,
	"BITOR",	310,
	"BITNOT",	311,
	"ASSIGN",	312,
	"ASSIGN1",	313,
	"ASSIGNMULT",	314,
	"ASSIGNADD",	315,
	"ASSIGNDIV",	316,
	"ASSIGNSUB",	317,
	"ASSIGNPOW",	318,
	"ASSIGNMOD",	319,
	",",	44,
	"|",	124,
	"&",	38,
	"=",	61,
	"<",	60,
	">",	62,
	"!",	33,
	"+",	43,
	"-",	45,
	"*",	42,
	"/",	47,
	"UNARYMINUS",	320,
	"$",	36,
	"[",	91,
	"(",	40,
	"\'",	39,
	"-unknown-",	-1	/* ends search */
};

__YYSCLASS char * yyreds[] =
{
	"-no such reduction-",
	"cmpdline : cmpdnobreak",
	"cmpdline : cmpdbreak",
	"cmpdnobreak : line",
	"cmpdnobreak : cmpdnobreak line",
	"cmpdbreak : break",
	"cmpdbreak : cmpdnobreak break",
	"line : ENDOFL",
	"line : ';'",
	"line : assgn ENDOFL",
	"line : assgn ';'",
	"line : batch ENDOFL",
	"line : batch ';'",
	"line : expr ENDOFL",
	"line : expr ';'",
	"line : badbatch",
	"line : expr ERROR",
	"line : ERROR",
	"line : error",
	"break : BREAK",
	"break : BREAK ';'",
	"assgn : expr1 ASSIGN",
	"assgn : expr1 ASSIGN expr",
	"assgn : component ASSIGN",
	"assgn : VAR bracklist ASSIGN",
	"assgn : VAR bracklist ASSIGN expr",
	"assgn : '(' expr ')' bracklist ASSIGN",
	"assgn : '(' expr ')' bracklist ASSIGN expr",
	"assgn : functionvalue bracklist ASSIGN",
	"assgn : functionvalue bracklist ASSIGN expr",
	"assgn : byname VAR bracklist ASSIGN",
	"assgn : byname VAR bracklist ASSIGN expr",
	"assgn : expr1 ASSIGNADD expr",
	"assgn : expr1 ASSIGNSUB expr",
	"assgn : expr1 ASSIGNMULT expr",
	"assgn : expr1 ASSIGNDIV expr",
	"assgn : expr1 ASSIGNMOD expr",
	"assgn : expr1 ASSIGNPOW expr",
	"assgn : '(' expr ')' bracklist assignop",
	"assgn : functionvalue bracklist assignop",
	"assgn : byname VAR bracklist assignop",
	"assgn : VAR bracklist assignop",
	"assignop : ASSIGNADD",
	"assignop : ASSIGNSUB",
	"assignop : ASSIGNMULT",
	"assignop : ASSIGNDIV",
	"assignop : ASSIGNMOD",
	"assignop : ASSIGNPOW",
	"batch : BATCH parenlist",
	"batch : byname BATCH parenlist",
	"keyword : compname ASSIGN1 expr",
	"functionvalue : BLTIN parenlist",
	"functionvalue : byname BLTIN parenlist",
	"component : expr '$' compname",
	"component : expr brack2list",
	"expr1 : functionvalue",
	"expr1 : missingrparen",
	"expr1 : VAR",
	"expr1 : byname VAR",
	"expr1 : expr '+' expr",
	"expr1 : expr '-' expr",
	"expr1 : expr '*' expr",
	"expr1 : expr MATMULT expr",
	"expr1 : expr TRMATMULT expr",
	"expr1 : expr MATMULTTR expr",
	"expr1 : expr '/' expr",
	"expr1 : expr MOD expr",
	"expr1 : expr POW expr",
	"expr1 : expr BITOR expr",
	"expr1 : expr BITXOR expr",
	"expr1 : expr BITAND expr",
	"expr1 : '(' expr ')'",
	"expr1 : '-' expr",
	"expr1 : '+' expr",
	"expr1 : BITNOT expr",
	"expr1 : expr '=' expr",
	"expr1 : expr EQ expr",
	"expr1 : expr NE expr",
	"expr1 : expr '<' expr",
	"expr1 : expr LE expr",
	"expr1 : expr '>' expr",
	"expr1 : expr GE expr",
	"expr1 : expr '&' expr",
	"expr1 : expr AND expr",
	"expr1 : expr '|' expr",
	"expr1 : expr OR expr",
	"expr1 : '!' expr",
	"expr1 : expr '\''",
	"expr1 : expandedmacro",
	"expandedmacro : macro parenlist",
	"expandedmacro : macro parenlist '{' cmpdline '}'",
	"missingrparen : '(' ENDOFL",
	"missingrparen : '(' expr1 ENDOFL",
	"missingrparen : '(' ';'",
	"missingrparen : '(' expr1 ';'",
	"expr : expr1",
	"expr : component",
	"expr : assgn",
	"expr : ifstart",
	"expr : ifstart ELSE",
	"expr : ifstart ELSE '{' cmpdline '}'",
	"expr : WHILE parenlist",
	"expr : WHILE parenlist '{' cmpdline '}'",
	"expr : FOR parenlist",
	"expr : FOR parenlist '{' cmpdline '}'",
	"expr : '{' cmpdline '}'",
	"expr : '{' expr '}'",
	"expr : subscripted",
	"expr : VAR '('",
	"expr : END",
	"expr : FATALERROR",
	"badbatch : expr '+' batch",
	"badbatch : expr '-' batch",
	"badbatch : expr '*' batch",
	"badbatch : expr MATMULT batch",
	"badbatch : expr TRMATMULT batch",
	"badbatch : expr MATMULTTR batch",
	"badbatch : expr '/' batch",
	"badbatch : expr MOD batch",
	"badbatch : expr POW batch",
	"badbatch : batch '+'",
	"badbatch : batch '-'",
	"badbatch : batch '*'",
	"badbatch : batch MATMULT",
	"badbatch : batch TRMATMULT",
	"badbatch : batch MATMULTTR",
	"badbatch : batch '/'",
	"badbatch : batch MOD",
	"badbatch : batch POW",
	"badbatch : '-' batch",
	"badbatch : '+' batch",
	"badbatch : expr EQ batch",
	"badbatch : expr NE batch",
	"badbatch : expr '<' batch",
	"badbatch : expr LE batch",
	"badbatch : expr '>' batch",
	"badbatch : expr GE batch",
	"badbatch : expr '&' batch",
	"badbatch : expr AND batch",
	"badbatch : expr '|' batch",
	"badbatch : expr OR batch",
	"badbatch : batch EQ",
	"badbatch : batch NE",
	"badbatch : batch '<'",
	"badbatch : batch LE",
	"badbatch : batch '>'",
	"badbatch : batch GE",
	"badbatch : batch '&'",
	"badbatch : batch AND",
	"badbatch : batch '|'",
	"badbatch : batch OR",
	"badbatch : '!' batch",
	"badbatch : '(' batch",
	"badbatch : batch '('",
	"badbatch : batch '['",
	"badbatch : batch '$'",
	"badbatch : batch RIGHTBRACK2",
	"badbatch : batch ASSIGN",
	"subscripted : '{' expr '}' bracklist",
	"subscripted : '{' cmpdline '}' bracklist",
	"subscripted : '(' expr ')' bracklist",
	"subscripted : functionvalue bracklist",
	"subscripted : component bracklist",
	"subscripted : expandedmacro bracklist",
	"subscripted : VAR bracklist",
	"subscripted : subscripted bracklist",
	"subscripted : byname VAR bracklist",
	"ifstart : IF parenlist",
	"ifstart : IF parenlist '{' cmpdline '}'",
	"ifstart : ifstart ELSEIF",
	"ifstart : ifstart ELSEIF parenlist",
	"ifstart : ifstart ELSEIF parenlist '{' cmpdline '}'",
	"ifstart : ifstart ERROR",
	"macro : MACRO",
	"macro : byname MACRO",
	"compname : VAR",
	"compname : byname VAR",
	"byname : LEFTANGLE expr RIGHTANGLE",
	"byname : LEFTANGLE expr",
	"list : ','",
	"list : expr ','",
	"list : keyword ','",
	"list : list expr ','",
	"list : list keyword ','",
	"list : list ','",
	"completelist : expr",
	"completelist : keyword",
	"completelist : list",
	"completelist : list expr",
	"completelist : list keyword",
	"parenlist : '(' ')'",
	"parenlist : '(' completelist ')'",
	"parenlist : '(' completelist ENDOFL",
	"parenlist : '(' completelist ';'",
	"bracklist : '[' ']'",
	"bracklist : '[' RIGHTBRACK2",
	"bracklist : '[' completelist ']'",
	"bracklist : '[' completelist RIGHTBRACK2",
	"bracklist : '[' ';'",
	"bracklist : '[' ENDOFL",
	"bracklist : '[' completelist",
	"bracklist : '[' completelist ERROR",
	"brack2list : LEFTBRACK2 '[' expr RIGHTBRACK2 ']'",
	"brack2list : LEFTBRACK2 '[' expr RIGHTBRACK2 RIGHTBRACK2",
	"brack2list : LEFTBRACK2 '[' ';'",
	"brack2list : LEFTBRACK2 '[' ENDOFL",
	"brack2list : LEFTBRACK2 '[' expr",
};
#endif /* YYDEBUG */
#define YYFLAG  (-3000)
/* @(#) $Revision: 70.7 $ */    

/*
** Skeleton parser driver for yacc output
*/

#if defined(NLS) && !defined(NL_SETN)
#include <msgbuf.h>
#endif

#ifndef nl_msg
#define nl_msg(i,s) (s)
#endif

/*
** yacc user known macros and defines
*/
#define YYERROR		goto yyerrlab

#ifndef __RUNTIME_YYMAXDEPTH
#define YYACCEPT	return(0)
#define YYABORT		return(1)
#else
#define YYACCEPT	{free_stacks(); return(0);}
#define YYABORT		{free_stacks(); return(1);}
#endif

#define YYBACKUP( newtoken, newvalue )\
{\
	if ( yychar >= 0 || ( yyr2[ yytmp ] >> 1 ) != 1 )\
	{\
		yyerror( (nl_msg(30001,"syntax error - cannot backup")) );\
		goto yyerrlab;\
	}\
	yychar = newtoken;\
	yystate = *yyps;\
	yylval = newvalue;\
	goto yynewstate;\
}
#define YYRECOVERING()	(!!yyerrflag)
#ifndef YYDEBUG
#	define YYDEBUG	1	/* make debugging available */
#endif

/*
** user known globals
*/
int yydebug;			/* set to 1 to get debugging */

/*
** driver internal defines
*/
/* define for YYFLAG now generated by yacc program. */
/*#define YYFLAG		(FLAGVAL)*/

/*
** global variables used by the parser
*/
# ifndef __RUNTIME_YYMAXDEPTH
__YYSCLASS YYSTYPE yyv[ YYMAXDEPTH ];	/* value stack */
__YYSCLASS int yys[ YYMAXDEPTH ];		/* state stack */
# else
__YYSCLASS YYSTYPE *yyv;			/* pointer to malloc'ed value stack */
__YYSCLASS int *yys;			/* pointer to malloc'ed stack stack */

#if defined(__STDC__) || defined (__cplusplus)
#include <stdlib.h>
#else
	extern char *malloc();
	extern char *realloc();
	extern void free();
#endif /* __STDC__ or __cplusplus */


static int allocate_stacks(); 
static void free_stacks();
# ifndef YYINCREMENT
# define YYINCREMENT (YYMAXDEPTH/2) + 10
# endif
# endif	/* __RUNTIME_YYMAXDEPTH */
static long  yymaxdepth = YYMAXDEPTH;

__YYSCLASS YYSTYPE *yypv;			/* top of value stack */
__YYSCLASS int *yyps;			/* top of state stack */

__YYSCLASS int yystate;			/* current state */
__YYSCLASS int yytmp;			/* extra var (lasts between blocks) */

int yynerrs;			/* number of errors */
__YYSCLASS int yyerrflag;			/* error recovery flag */
int yychar;			/* current input token number */



/*
** yyparse - return 0 if worked, 1 if syntax error not recovered from
*/
int
yyparse()
{
	register YYSTYPE *yypvt;	/* top of value stack for $vars */
#ifdef MACINTOSH
	register yytabelem *yyexca = Pyyexca;
	register yytabelem *yyact = Pyyact;
	register yytabelem *yypact = Pyypact;
	register yytabelem *yypgo = Pyypgo;
	register yytabelem *yyr1 = Pyyr1;
	register yytabelem *yyr2 = Pyyr2;
	register yytabelem *yychk = Pyychk;
	register yytabelem *yydef = Pyydef;
#endif /*MACINTOSH*/
	WHERE("yyparse");

	/*
	** Initialize externals - yyparse may be called more than once
	*/
# ifdef __RUNTIME_YYMAXDEPTH
	if (allocate_stacks()) YYABORT;
# endif
	yypv = &yyv[-1];
	yyps = &yys[-1];
	yystate = 0;
	yytmp = 0;
	yynerrs = 0;
	yyerrflag = 0;
	yychar = -1;

	goto yystack;
	{
		register YYSTYPE *yy_pv;	/* top of value stack */
		register int *yy_ps;		/* top of state stack */
		register int yy_state;		/* current state */
		register int  yy_n;		/* internal state number info */

		/*
		** get globals into registers.
		** branch to here only if YYBACKUP was called.
		*/
	yynewstate:
		yy_pv = yypv;
		yy_ps = yyps;
		yy_state = yystate;
		goto yy_newstate;

		/*
		** get globals into registers.
		** either we just started, or we just finished a reduction
		*/
	yystack:
		yy_pv = yypv;
		yy_ps = yyps;
		yy_state = yystate;

		/*
		** top of for (;;) loop while no reductions done
		*/
	yy_stack:
		/*
		** put a state and value onto the stacks
		*/
#if YYDEBUG
		/*
		** if debugging, look up token value in list of value vs.
		** name pairs.  0 and negative (-1) are special values.
		** Note: linear search is used since time is not a real
		** consideration while debugging.
		*/
		if ( yydebug )
		{
			register int yy_i;

			printf( "State %d, token ", yy_state );
			if ( yychar == 0 )
				printf( "end-of-file\n" );
			else if ( yychar < 0 )
				printf( "-none-\n" );
			else
			{
				for ( yy_i = 0; yytoks[yy_i].t_val >= 0;
					yy_i++ )
				{
					if ( yytoks[yy_i].t_val == yychar )
						break;
				}
				printf( "%s\n", yytoks[yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ++yy_ps >= &yys[ yymaxdepth ] )	/* room on stack? */
		{
# ifndef __RUNTIME_YYMAXDEPTH
			yyerror( (nl_msg(30002,"ERROR: Parser stack overflow; probably too deep macro recursion")) );
			YYABORT;
# else
			/* save old stack bases to recalculate pointers */
			YYSTYPE * yyv_old = yyv;
			int * yys_old = yys;
			yymaxdepth += YYINCREMENT;
			yys = (int *) realloc(yys, yymaxdepth * sizeof(int));
			yyv = (YYSTYPE *) realloc(yyv, yymaxdepth * sizeof(YYSTYPE));
			if (yys==0 || yyv==0) {
			    yyerror( (nl_msg(30002,"ERROR: Parser stack overflow; probably too deep macro recursion")) );
			    YYABORT;
			    }
			/* Reset pointers into stack */
			yy_ps = (yy_ps - yys_old) + yys;
			yyps = (yyps - yys_old) + yys;
			yy_pv = (yy_pv - yyv_old) + yyv;
			yypv = (yypv - yyv_old) + yyv;
# endif

		}
		*yy_ps = yy_state;
		*++yy_pv = yyval;

		/*
		** we have a new state - find out what to do
		*/
	yy_newstate:
		if ( ( yy_n = yypact[ yy_state ] ) <= YYFLAG )
			goto yydefault;		/* simple state */
#if YYDEBUG
		/*
		** if debugging, need to mark whether new token grabbed
		*/
		yytmp = yychar < 0;
#endif
		if ( ( yychar < 0 ) && ( ( yychar = yylex() ) < 0 ) )
			yychar = 0;		/* reached EOF */
#if YYDEBUG
		if ( yydebug && yytmp )
		{
			register int yy_i;

			printf( "Received token " );
			if ( yychar == 0 )
				printf( "end-of-file\n" );
			else if ( yychar < 0 )
				printf( "-none-\n" );
			else
			{
				for ( yy_i = 0; yytoks[yy_i].t_val >= 0;
					yy_i++ )
				{
					if ( yytoks[yy_i].t_val == yychar )
						break;
				}
				printf( "%s\n", yytoks[yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ( ( yy_n += yychar ) < 0 ) || ( yy_n >= YYLAST ) )
			goto yydefault;
		if ( yychk[ yy_n = yyact[ yy_n ] ] == yychar )	/*valid shift*/
		{
			yychar = -1;
			yyval = yylval;
			yy_state = yy_n;
			if ( yyerrflag > 0 )
				yyerrflag--;
			goto yy_stack;
		}

	yydefault:
		if ( ( yy_n = yydef[ yy_state ] ) == -2 )
		{
#if YYDEBUG
			yytmp = yychar < 0;
#endif
			if ( ( yychar < 0 ) && ( ( yychar = yylex() ) < 0 ) )
				yychar = 0;		/* reached EOF */
#if YYDEBUG
			if ( yydebug && yytmp )
			{
				register int yy_i;

				printf( "Received token " );
				if ( yychar == 0 )
					printf( "end-of-file\n" );
				else if ( yychar < 0 )
					printf( "-none-\n" );
				else
				{
					for ( yy_i = 0;
						yytoks[yy_i].t_val >= 0;
						yy_i++ )
					{
						if ( yytoks[yy_i].t_val
							== yychar )
						{
							break;
						}
					}
					printf( "%s\n", yytoks[yy_i].t_name );
				}
			}
#endif /* YYDEBUG */
			/*
			** look through exception table
			*/
			{
				register yytabelem *yyxi = yyexca;

				while ( ( *yyxi != -1 ) ||
					( yyxi[1] != yy_state ) )
				{
					yyxi += 2;
				}
				while ( ( *(yyxi += 2) >= 0 ) &&
					( *yyxi != yychar ) )
					;
				if ( ( yy_n = yyxi[1] ) < 0 )
					YYACCEPT;
			}
		}

		/*
		** check for syntax error
		*/
		if ( yy_n == 0 )	/* have an error */
		{
			/* no worry about speed here! */
			switch ( yyerrflag )
			{
			case 0:		/* new error */
				yyerror( (nl_msg(30003,"ERROR: problem with input")) );
				yynerrs++;
				goto skip_init;
			yyerrlab:
				/*
				** get globals into registers.
				** we have a user generated syntax type error
				*/
				yy_pv = yypv;
				yy_ps = yyps;
				yy_state = yystate;
				yynerrs++;
			skip_init:
			case 1:
			case 2:		/* incompletely recovered error */
					/* try again... */
				yyerrflag = 3;
				/*
				** find state where "error" is a legal
				** shift action
				*/
				while ( yy_ps >= yys )
				{
					yy_n = yypact[ *yy_ps ] + YYERRCODE;
					if ( yy_n >= 0 && yy_n < YYLAST &&
						yychk[yyact[yy_n]] == YYERRCODE)					{
						/*
						** simulate shift of "error"
						*/
						yy_state = yyact[ yy_n ];
						goto yy_stack;
					}
					/*
					** current state has no shift on
					** "error", pop stack
					*/
#if YYDEBUG
#	define _POP_ "Error recovery pops state %d, uncovers state %d\n"
					if ( yydebug )
						printf( _POP_, *yy_ps,
							yy_ps[-1] );
#	undef _POP_
#endif
					yy_ps--;
					yy_pv--;
				}
				/*
				** there is no state on stack with "error" as
				** a valid shift.  give up.
				*/
				YYABORT;
			case 3:		/* no shift yet; eat a token */
#if YYDEBUG
				/*
				** if debugging, look up token in list of
				** pairs.  0 and negative shouldn't occur,
				** but since timing doesn't matter when
				** debugging, it doesn't hurt to leave the
				** tests here.
				*/
				if ( yydebug )
				{
					register int yy_i;

					printf( "Error recovery discards " );
					if ( yychar == 0 )
						printf( "token end-of-file\n" );
					else if ( yychar < 0 )
						printf( "token -none-\n" );
					else
					{
						for ( yy_i = 0;
							yytoks[yy_i].t_val >= 0;
							yy_i++ )
						{
							if ( yytoks[yy_i].t_val
								== yychar )
							{
								break;
							}
						}
						printf( "token %s\n",
							yytoks[yy_i].t_name );
					}
				}
#endif /* YYDEBUG */
				if ( yychar == 0 )	/* reached EOF. quit */
					YYABORT;
				yychar = -1;
				goto yy_newstate;
			}
		}/* end if ( yy_n == 0 ) */
		/*
		** reduction by production yy_n
		** put stack tops, etc. so things right after switch
		*/
#if YYDEBUG
		/*
		** if debugging, print the string that is the user's
		** specification of the reduction which is just about
		** to be done.
		*/
		if ( yydebug )
			printf( "Reduce by (%d) \"%s\"\n",
				yy_n, yyreds[ yy_n ] );
#endif
		yytmp = yy_n;			/* value to switch over */
		yypvt = yy_pv;			/* $vars top of value stack */
		/*
		** Look in goto table for next state
		** Sorry about using yy_state here as temporary
		** register variable, but why not, if it works...
		** If yyr2[ yy_n ] doesn't have the low order bit
		** set, then there is no action to be done for
		** this reduction.  So, no saving & unsaving of
		** registers done.  The only difference between the
		** code just after the if and the body of the if is
		** the goto yy_stack in the body.  This way the test
		** can be made before the choice of what to do is needed.
		*/
		{
			/* length of production doubled with extra bit */
			register int yy_len = yyr2[ yy_n ];

			if ( !( yy_len & 01 ) )
			{
				yy_len >>= 1;
				yyval = ( yy_pv -= yy_len )[1];	/* $$ = $1 */
				yy_state = yypgo[ yy_n = yyr1[ yy_n ] ] +
					*( yy_ps -= yy_len ) + 1;
				if ( yy_state >= YYLAST ||
					yychk[ yy_state =
					yyact[ yy_state ] ] != -yy_n )
				{
					yy_state = yyact[ yypgo[ yy_n ] ];
				}
				goto yy_stack;
			}
			yy_len >>= 1;
			yyval = ( yy_pv -= yy_len )[1];	/* $$ = $1 */
			yy_state = yypgo[ yy_n = yyr1[ yy_n ] ] +
				*( yy_ps -= yy_len ) + 1;
			if ( yy_state >= YYLAST ||
				yychk[ yy_state = yyact[ yy_state ] ] != -yy_n )
			{
				yy_state = yyact[ yypgo[ yy_n ] ];
			}
		}
					/* save until reenter driver code */
		yystate = yy_state;
		yyps = yy_ps;
		yypv = yy_pv;
	}
	/*
	** code supplied by user is placed in this switch
	*/
	switch( yytmp )
	{
		
case 1:
		/*Folding 6 cases into 1*/
case 2:
case 3:
case 4:
case 5:
case 6:
{yyval = yypvt[-0];} break;
case 7:
{
						yyval = NULLSYMBOL;
					  lev0Return:
						if(BRACKETLEV == 0)
						{
							return(0);
						}
					} break;
case 8:
{
					yyval = NULLSYMBOL;
				  lev0Unscratch:
					
					if(BRACKETLEV == 0)
					{
						Unscratch();
					}
				} break;
case 9:
{
						yyval = yypvt[-1];
						goto lev0Return;
					} break;
case 10:
{
						yyval = yypvt[-1];
						goto lev0Unscratch;
					} break;
case 11:
		/*Folding 2 cases into 1*/
case 12:
{
					  dobatch:
						(void) runCommand(Lookup("batch"), yypvt[-1]);
		  				return (0); /* terminate parse */
					} break;
case 13:
{
						yyval = yypvt[-1];
						if(BRACKETLEV == 0)
						{
							prexpr(yypvt[-1]);
							return(0);
						}
					} break;
case 14:
{
						yyval = yypvt[-1];
						if(BRACKETLEV == 0)
						{
							prexpr(yypvt[-1]);
							Unscratch();
						}
					} break;
case 15:
{
						yyerror("ERROR: illegal use of batch()");
						Removelist(yypvt[-0]);
		  				return (0); /* terminate parse */
					} break;
case 16:
		/*Folding 2 cases into 1*/
case 17:
{yyval = (Symbolhandle) 0;YYERROR;} break;
case 18:
{yyval = (Symbolhandle) 0;yyerrok; return(0);} break;
case 19:
{  /* break, break n, breakall */
					if(yypvt[-0] == (Symbolhandle) 0)
					{
						YYERROR;
					}
					TargetLevel = WDEPTH - (int) DATAVALUE(yypvt[-0],0);
					Removesymbol(yypvt[-0]);
					if(TargetLevel < 0)
					{
						YYERROR;
					}
					{
						YYLEXMESSAGE = MAXWDEPTH +
							WHILEBRACKETLEV[TargetLevel] + 1;
						cleanLoops(TargetLevel,WDEPTH);
					}
				} break;
case 20:
{
					yyval = NULLSYMBOL;
				} break;
case 21:
{
							yyval = Assign(yypvt[-1],(Symbolhandle) 0,-1);
							if(yyval == (Symbolhandle) 0)
							{
								YYERROR;
							}
						} break;
case 22:
{
							yyval = Assign(yypvt[-3],yypvt[-0],0);
						  checkResult:
							if(yyval == (Symbolhandle) 0)
							{
								YYERROR;
							}
						} break;
case 23:
{
							yyerror("ERROR: illegal assignment to structure component");
							YYERROR;
						} break;
case 24:
{
							  subassigncheck2:
								yyval = Subassign(yypvt[-2],yypvt[-1],(Symbolhandle) 0,1);
								if(yyval == (Symbolhandle) 0)
								{
									YYERROR;
								}
							} break;
case 25:
{
							  subassign5:
								yyval = Subassign(yypvt[-4],yypvt[-3],yypvt[-0],0);
								Removelist(yypvt[-3]);
								if(yyval == (Symbolhandle) 0)
								{
									YYERROR;
								}
							} break;
case 26:
{
								yyval = Subassign(yypvt[-3],yypvt[-1],(Symbolhandle) 0,1);
								if(yyval == (Symbolhandle) 0)
								{
									YYERROR;
								}
							} break;
case 27:
{
								yyval = Subassign(yypvt[-5],yypvt[-3],yypvt[-0],0);
								Removelist(yypvt[-3]);
								if(yyval == (Symbolhandle) 0)
								{
									YYERROR;
								}
							} break;
case 28:
{
								goto subassigncheck2;
							} break;
case 29:
{
								goto subassign5;
							} break;
case 30:
{
							  subassigncheck3:
								yyval = Subassign(yypvt[-3],yypvt[-1],(Symbolhandle) 0,1);
								if(yyval == (Symbolhandle) 0)
								{
									YYERROR;
								}
							} break;
case 31:
{
							  subassign6:
								yyval = Subassign(yypvt[-5],yypvt[-3],yypvt[-0],0);
								Removelist(yypvt[-3]);
								if(yyval == (Symbolhandle) 0)
								{
									YYERROR;
								}
							} break;
case 32:
{
								Op = ADD;
							  doassignop:
								yyval = Assign(yypvt[-2],(Symbolhandle) 0,-Op);
								if(yyval == (Symbolhandle) 0)
								{
									YYERROR;
								}
								yyval = Arith(yypvt[-2], yypvt[-0], Op);
								UNLOADSEG(Arith);
								if(yyval != (Symbolhandle) 0)
								{
									yyval = Assign(yypvt[-2],yyval,0);
								}
								goto checkResult;
							} break;
case 33:
{
								Op = SUB;
								goto doassignop;
							} break;
case 34:
{
								Op = MULT;
								goto doassignop;
							} break;
case 35:
{
								Op = DIV ;
								goto doassignop;
							} break;
case 36:
{
								Op = MOD;
								goto doassignop;
							} break;
case 37:
{
								Op = EXP;
								goto doassignop;
							} break;
case 38:
		/*Folding 4 cases into 1*/
case 39:
case 40:
case 41:
{
									AssignOp = LONGDATAVALUE(yypvt[-0],1);
								  assignoperror:
									sprintf(OUTSTR,
											"ERROR: illegal left side of %s",
											opName(AssignOp));
									yyerror(OUTSTR);
									YYERROR;
								} break;
case 42:
{
	 						Op = ADD;
							AssignOp = ASSIGNADD;
						  setassignOp:
							{
								Symbolhandle symhOp = LongInstall(SCRATCH, 2);
								if(symhOp == (Symbolhandle) 0)
								{
									YYERROR;
								}
								LONGDATAVALUE(symhOp,0) = (long) Op;
								LONGDATAVALUE(symhOp,1) = (long) AssignOp;
								yyval = symhOp;
							}
						} break;
case 43:
{Op = SUB;AssignOp = ASSIGNSUB;goto setassignOp;} break;
case 44:
{Op = MULT;AssignOp = ASSIGNMULT;goto setassignOp;} break;
case 45:
{Op = DIV;AssignOp = ASSIGNDIV;goto setassignOp;} break;
case 46:
{Op = MOD;AssignOp = ASSIGNMOD;goto setassignOp;} break;
case 47:
{Op = EXP;AssignOp = ASSIGNPOW;goto setassignOp;} break;
case 48:
{
							yyval = yypvt[-0];
						} break;
case 49:
{
									yyval = yypvt[-0];
				 				} break;
case 50:
{
								yyval = Assign(yypvt[-2],yypvt[-0], 1);

								goto checkResult;
							} break;
case 51:
{
									yyval = runCommand(yypvt[-1], yypvt[-0]);
								  setScrName1:
									if(yyval == (Symbolhandle) 0)
									{
										YYERROR;
									}
									setScratchName(yyval);
									CHECKINTERRUPT;
								} break;
case 52:
{
									yyval = runCommand(yypvt[-2],yypvt[-0]);
									goto setScrName1;
								} break;
case 53:
{
									yyval = Extract(yypvt[-2],yypvt[-0],1);
									
								  setScrName2:
									if(yyval == (Symbolhandle) 0)
									{
										YYERROR;
									}
									setScratchName(yyval);
								} break;
case 54:
{
							yyval = Extract(yypvt[-1],yypvt[-0],0);
							goto setScrName2;
						} break;
case 55:
{yyval = yypvt[-0];} break;
case 56:
{
						YYERROR;
					} break;
case 57:
{yyval = yypvt[-0];} break;
case 58:
{yyval = yypvt[-1];} break;
case 59:
{
							Op = ADD;
						  arith:
							yyval = Arith(yypvt[-2],yypvt[-0],Op);
							/* Arith() removes SCRATCH unconsumed args */
							UNLOADSEG(Arith);
							goto setScrName1;
						 } break;
case 60:
{ Op = SUB;goto arith; } break;
case 61:
{ Op = MULT;goto arith; } break;
case 62:
{ Op = MATMULT;goto arith; } break;
case 63:
{ Op = TRMATMULT;goto arith; } break;
case 64:
{ Op = MATMULTTR;goto arith; } break;
case 65:
{ Op = DIV;goto arith; } break;
case 66:
{ Op = MOD;goto arith; } break;
case 67:
{ Op = EXP;goto arith; } break;
case 68:
{ Op = BITOR;goto arith; } break;
case 69:
{ Op = BITXOR;goto arith; } break;
case 70:
{ Op = BITAND;goto arith; } break;
case 71:
{yyval = parserCopy(yypvt[-1]);} break;
case 72:
{
									Op = '-';
								  unary:
									yyval = Unary(yypvt[-0],Op);
									/*Unary() reuses space if arg is SCRATCH */
									UNLOADSEG(Unary);
									goto setScrName1;
								} break;
case 73:
{
									Op = '+';
									goto unary;
								} break;
case 74:
{
									Op = BITNOT;
									goto unary;
								} break;
case 75:
{
						Op = EQ;
					  logic:
						yyval = Logic(yypvt[-2],yypvt[-0],Op);
						/*Logic() removes unconsumed scratch args */
						UNLOADSEG(Logic);
						goto setScrName1;
					} break;
case 76:
{ Op = EQ; goto logic; } break;
case 77:
{ Op = NE; goto logic; } break;
case 78:
{ Op = LT; goto logic; } break;
case 79:
{ Op = LE; goto logic; } break;
case 80:
{ Op = GT; goto logic; } break;
case 81:
{ Op = GE; goto logic; } break;
case 82:
{ Op = AND; goto logic; } break;
case 83:
{ Op = AND; goto logic; } break;
case 84:
{ Op = OR; goto logic; } break;
case 85:
{ Op = OR; goto logic; } break;
case 86:
{
							yyval = Logic(yypvt[-0],yypvt[-0],NOT);
							UNLOADSEG(Unary);
							goto setScrName1;
						} break;
case 87:
{
							Symbolhandle list = Addlist( Makelist(), yypvt[-1] );
							if(list == (Symbolhandle) 0)
							{
								YYERROR;
							}
							strcpy(FUNCNAME,"prime");
							yyval = transpose(list);
							Removelist(list);
							UNLOADSEG(transpose);
							goto setScrName1;
				  		} break;
case 88:
{yyval = yypvt[-0];} break;
case 89:
{
							long reply = Macrosetup(yypvt[-1],yypvt[-0]);

							Removelist(yypvt[-0]);
							if(!reply)
							{
								YYERROR;
							}
						} break;
case 90:
{
							yyval = parserCopy(yypvt[-1]);
							MDEPTH--;
						} break;
case 91:
		/*Folding 4 cases into 1*/
case 92:
case 93:
case 94:
{
								  missingRparen:
									yyval = 0;
									yyerror("ERROR: missing ')'");
									YYERROR;
								} break;
case 95:
		/*Folding 3 cases into 1*/
case 96:
case 97:
{yyval = yypvt[-0];} break;
case 98:
{
					IDEPTH--;
					yyval = yypvt[-0];
				} break;
case 99:
{
						if(IFCONDITIONS[IDEPTH-1] > 0)
						{
							 yyval = yypvt[-1];
							 YYLEXMESSAGE = SKIPBRACKETBLOCK;
						}
						CHECKINTERRUPT;
						if(!Ifsetup((Symbolhandle) 0, ELSE))
						{
							yyval = 0;
							YYERROR;
						}
					} break;
case 100:
{
						if(IFCONDITIONS[IDEPTH-1] == 0)
						{
							 yyval = yypvt[-1];
						}
						IDEPTH--;
					} break;
case 101:
{
							int    ifOK;

							CHECKINTERRUPT;
							ifOK = Ifsetup(yypvt[-0], WHILE);
							Removelist(yypvt[-0]);
							if(!ifOK)
							{
								yyval = 0;
								YYERROR;
							}
							if(WHILECONDITIONS[WDEPTH-1] == 0)
							{
								YYLEXMESSAGE = SKIPBRACKETBLOCK;
							}
							else if(WHILELIMITS[WDEPTH-1] <= 0)
							{ /* too many reps */
								yyerrok ;
								return(0) ;
							}
						} break;
case 102:
{
							if(WHILECONDITIONS[WDEPTH-1])
							{
								ISTRCHAR = WHILESTARTS[WDEPTH-1];
								if(yypvt[-1] != (Symbolhandle) 0 && BRACKETLEV > 0 &&
									 isscratch(NAME(yypvt[-1])))
								{
									Removesymbol(yypvt[-1]);
									yypvt[-1] = NULLSYMBOL;
								}
								YYLEXMESSAGE = -((long) ';');
							}
							else
							{
								WHILELIMITS[WDEPTH-1] = MAXWHILE;
							}
							WDEPTH--;
							yyval = yypvt[-1];
						} break;
case 103:
{
							int   ifOK;

							CHECKINTERRUPT;
							ifOK = Ifsetup(yypvt[-0], FOR);
							Removelist(yypvt[-0]);
							if(!ifOK)
							{
								yyval = 0;
								YYERROR;
							}
						 } break;
case 104:
{
							if(WHILECONDITIONS[WDEPTH-1] &&
								 --WHILELIMITS[WDEPTH-1] > 0)
							{
								/* go back for another round */
								ISTRCHAR = WHILESTARTS[WDEPTH-1];
								setupfakefor();
								ALTTOKENPTR = ALTINPUTTOKENS;
								if(yypvt[-1] != (Symbolhandle) 0 && BRACKETLEV > 0 &&
									 isscratch(NAME(yypvt[-1])))
								{
									Removesymbol(yypvt[-1]);
									yypvt[-1] = NULLSYMBOL;
								}
							}
							else
							{ /* done with loop */
								WHILELIMITS[WDEPTH-1] = MAXWHILE;
							}
							WDEPTH--;
							yyval = yypvt[-1];
						} break;
case 105:
		/*Folding 2 cases into 1*/
case 106:
{yyval = yypvt[-1];} break;
case 107:
{yyval = yypvt[-0];} break;
case 108:
{
						notfunction(yypvt[-1]);
						yyval = 0;
						YYERROR;
					} break;
case 109:
{ return(END); } break;
case 110:
{
						return (FATALERROR);
					} break;
case 111:
		/*Folding 9 cases into 1*/
case 112:
case 113:
case 114:
case 115:
case 116:
case 117:
case 118:
case 119:
{yyval = yypvt[-0];} break;
case 120:
		/*Folding 9 cases into 1*/
case 121:
case 122:
case 123:
case 124:
case 125:
case 126:
case 127:
case 128:
{yyval = yypvt[-1];} break;
case 129:
		/*Folding 12 cases into 1*/
case 130:
case 131:
case 132:
case 133:
case 134:
case 135:
case 136:
case 137:
case 138:
case 139:
case 140:
{yyval = yypvt[-0];} break;
case 141:
		/*Folding 10 cases into 1*/
case 142:
case 143:
case 144:
case 145:
case 146:
case 147:
case 148:
case 149:
case 150:
{yyval = yypvt[-1];} break;
case 151:
		/*Folding 2 cases into 1*/
case 152:
{yyval = yypvt[-0];} break;
case 153:
		/*Folding 5 cases into 1*/
case 154:
case 155:
case 156:
case 157:
{yyval = yypvt[-1];} break;
case 158:
		/*Folding 3 cases into 1*/
case 159:
case 160:
{
								  bracklist4:
									yyval = Element(yypvt[-2],yypvt[-0]);
									Removelist(yypvt[-0]);
									goto setScrName2;
								} break;
case 161:
		/*Folding 5 cases into 1*/
case 162:
case 163:
case 164:
case 165:
{
							  bracklist2:
								yyval = Element(yypvt[-1],yypvt[-0]);
								Removelist(yypvt[-0]);
								goto setScrName2;
							} break;
case 166:
{
								yyval = Element(yypvt[-2],yypvt[-0]);
								Removelist(yypvt[-0]);
								goto setScrName2;
							} break;
case 167:
{
							int    ifOK;

							CHECKINTERRUPT;
							ifOK = Ifsetup(yypvt[-0], IF);
							Removelist(yypvt[-0]);
							if(!ifOK)
							{
								yyval = 0;
								YYERROR;
							}
							if(IFCONDITIONS[IDEPTH-1] == 0)
							{
								YYLEXMESSAGE = SKIPBRACKETBLOCK;
							}
						} break;
case 168:
{
							yyval = (IFCONDITIONS[IDEPTH-1] ) ? yypvt[-1] : NULLSYMBOL;
						} break;
case 169:
{
							if(IFCONDITIONS[IDEPTH-1] > 0)
							{
								 YYLEXMESSAGE = SKIPPARENBLOCK;
							}
						} break;
case 170:
{
							int    ifOK;

							if(IFCONDITIONS[IDEPTH-1] > 0)
							{ /* already found True condition*/
								YYLEXMESSAGE = SKIPBRACKETBLOCK;
								Removelist(yypvt[-0]);
								yypvt[-0] = (Symbolhandle) 0;
							}
							CHECKINTERRUPT;
							ifOK = Ifsetup(yypvt[-0], ELSEIF);
							if(yypvt[-0] != (Symbolhandle) 0)
							{
								Removelist(yypvt[-0]);
							}
							if(!ifOK)
							{
								yyval = 0;
								YYERROR;
							}
							if(IFCONDITIONS[IDEPTH-1] == 0)
							{
								YYLEXMESSAGE = SKIPBRACKETBLOCK;
							}
						} break;
case 171:
{
							if(yypvt[-4] == (Symbolhandle) 0)
							{
								yyval = yypvt[-7];
							}
							else if(IFCONDITIONS[IDEPTH-1] > 0)
							{
								yyval = yypvt[-1];
							}
							else
							{
								yyval = NULLSYMBOL;
							}
						} break;
case 172:
{ yyval = 0 ; YYERROR;} break;
case 173:
{yyval = yypvt[-0];} break;
case 174:
{
							yyval = yypvt[-1];
							PARSEMACROARGS = ParseMacroArgs;
						} break;
case 175:
{yyval = yypvt[-0];} break;
case 176:
{yyval = yypvt[-1];} break;
case 177:
{
						yyval = Byname(yypvt[-1],&BynameType,&ParseMacroArgs);
						if(yyval == (Symbolhandle) 0)
						{
							YYERROR;
						}
						else
						{
							YYLEXMESSAGE = -BynameType;
						}
					} break;
case 178:
{
							yyval = 0;
							yyerror("ERROR: missing '>>'");
							YYERROR;
						} break;
case 179:
{
							  growMake:
								yyval = Growlist(Makelist());
								goto checkResult;
							} break;
case 180:
		/*Folding 2 cases into 1*/
case 181:
{
							  growAddMake:
								yyval = Growlist(Addlist(Makelist(),yypvt[-1]));
								goto checkResult;
							} break;
case 182:
		/*Folding 2 cases into 1*/
case 183:
{
							  growAdd:
								yyval = Growlist(Addlist(yypvt[-2],yypvt[-1]));
								goto checkResult;
							} break;
case 184:
{
								yyval = Growlist(yypvt[-1]);
								goto checkResult;
							} break;
case 185:
		/*Folding 2 cases into 1*/
case 186:
{
							  addMake:
								yyval = Addlist( Makelist(), yypvt[-0] );
								goto checkResult;
							} break;
case 187:
{
								yyval = yypvt[-0];
								goto checkResult;
							} break;
case 188:
		/*Folding 2 cases into 1*/
case 189:
{
							  addlist:
								yyval = Addlist( yypvt[-1], yypvt[-0]);
								goto checkResult;
							} break;
case 190:
{
									  makelist:
										yyval = Makelist();
										goto checkResult;
									} break;
case 191:
{
									yyval = yypvt[-1];
								} break;
case 192:
		/*Folding 2 cases into 1*/
case 193:
{goto missingRparen;} break;
case 194:
		/*Folding 2 cases into 1*/
case 195:
{goto makelist;} break;
case 196:
		/*Folding 2 cases into 1*/
case 197:
{ yyval = yypvt[-1]; goto checkResult;} break;
case 198:
		/*Folding 3 cases into 1*/
case 199:
case 200:
{
					  missingRbracket:
						yyerror("ERROR: missing ']'");
						yyval = 0;
						YYERROR;
					} break;
case 202:
		/*Folding 2 cases into 1*/
case 203:
{ yyval = yypvt[-2]; goto checkResult;} break;
case 204:
		/*Folding 3 cases into 1*/
case 205:
case 206:
{
					  missingRbrack2:
						yyerror("ERROR: missing ']]'");
						yyval = 0;
						YYERROR;
					} break;
	}
	goto yystack;		/* reset registers in driver code */
}

# ifdef __RUNTIME_YYMAXDEPTH

static int allocate_stacks() {
	/* allocate the yys and yyv stacks */
	yys = (int *) malloc(yymaxdepth * sizeof(int));
	yyv = (YYSTYPE *) malloc(yymaxdepth * sizeof(YYSTYPE));

	if (yys==0 || yyv==0) {
	   yyerror( (nl_msg(30004,"unable to allocate space for yacc stacks")) );
	   return(1);
	   }
	else return(0);

}


static void free_stacks() {
	if (yys!=0) free((char *) yys);
	if (yyv!=0) free((char *) yyv);
}

# endif  /* defined(__RUNTIME_YYMAXDEPTH) */

