/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


/* main program and driver */

#define MAIN__
#include "globals.h"
#include "dbug.h"

#include <ctype.h>
#include <stdlib.h>

#ifdef BCPP
#include <io.h> /* for isatty */
#else /*BCPP*/
#ifndef EPX
#include <unistd.h> /* isatty for ultrix on DECStation */
#else /*EPX*/
int   isatty(int);
#endif /*EPX*/
#endif /*BCPP*/

#define CHUNK (10*sizeof(double)) /* number of bytes to increment INPUTSTRING */

#ifdef BCPP
unsigned          _ovrbuffer = 0x1000;
extern unsigned   _stklen = 32000;
#endif /* BCPP */

Symbolhandle    RInstall();
extern int      yydebug; /* set 1 for yacc debugging output */
extern void     cleanitup();
extern char   **mygethandle(), **mygrowhandle();
extern void     mydisphandle();
extern void     startupMsg();
extern void     myerrorout();
extern void     myprint(), fmyprint();
extern void     putprompt();
extern void     initialize();

extern char         *getenv();
extern void         *malloc();

#ifdef HASISATTY
extern int           isatty();
#endif /*HASISATTY*/

#ifdef COMMANDLINE
#ifdef MSDOS
#define SEP '\\'
#else /*MSDOS*/
#define SEP '/'
#endif /*MSDOS*/

#ifndef DEFAULTSCREENHEIGHT
#ifdef MSDOS
#define DEFAULTSCREENHEIGHT 25
#else /*MSDOS*/
#define DEFAULTSCREENHEIGHT 24
#endif /*MSDOS*/
#endif /*DEFAULTSCREENHEIGHT*/

#define SHIFT (++argv, --argc)

#define MEMGUBED   (4096)
#ifdef MONITOR
#include <mon.h>
static short   MonitorStarted = 0;
static WORD   *MonitorBuffer = (WORD *) 0;
extern etext;
extern int     endofmv(void); /*loaded last before libraries*/

#define MONGUBED   (16384)
#define MONSTART (void(*)()) 2
#define MONEND (void(*)()) endofmv /* instead of (void(*)()) &etext*/
#define MONEND1 ((void(*)()) &etext) /* yields garbage */
#define MONEND2 ((void (*)()) 2000000) /* seems to work best */
#ifndef MONNFUNC
#define MONNFUNC   (GUBED - MONGUBED)
#endif /*MONNFUNC*/
#ifndef MONBUFSIZE
#define MONBUFSIZE  (1000000)
#endif /*MONBUFSIZE*/

static startMonitor(void)
{
	long        i;
	WHERE("startMonitor");
	
	MonitorBuffer = (MonitorBuffer == (WORD *) 0) ?
		(WORD *) malloc(MONBUFSIZE * sizeof(WORD)) : MonitorBuffer;
	
	if (MonitorBuffer == (WORD *) 0)
	{
		return;
	} /*if (MonitorBuffer == (WORD *) 0)*/
	for (i=0; i < MONBUFSIZE; i++)
	{
		MonitorBuffer[i] = (WORD) 0;
	}
	monitor(MONSTART, MONEND2, MonitorBuffer, MONBUFSIZE, MONNFUNC);
	MonitorStarted = 1;
} /*startMonitor()*/

static void stopMonitor(void)
{
	monitor((void (*)()) 0, (void (*)()) 0, (WORD *) 0, 0, 0);
	MonitorStarted = 0;
}

#endif /*MONITOR*/

static char    *ProgName;
static char     *UsageMsg[] = 
{
	"Options are:",
	"    -q             Suppress printing banner",
	"    -b batchFile   Simulate batch(\"batchFile\")",
	"    -f initFile    Use file iniFile instead of .macanova.ini",
	"    -r restoreFile Simulate restore(\"restoreFile\")",
	"    -h helpFile    Use helpFile as default file for help information",
	"    -m macroFile   Set variable MACROFILE to \"macroFile\"",
	"    -d dataFile    Set variable DATAFILE to \"dataFile\"",
	"    -l nlines      Set screen height to nlines (>= 6)",
	"    -w ncols       Set screen width to ncols (>= 20)",
	"    -f and -r should not both be specified",
	(char *) 0	
};

static void usage(void)
{
	char    **msgsLine;

	sprintf(OUTSTR,	"usage: %s [options]", ProgName);
	putOUTSTR();

	for(msgsLine = UsageMsg;*msgsLine;msgsLine++)
	{
		myerrorout(*msgsLine);
	}
	exit(1);
} /*usage()*/
#endif /*COMMANDLINE*/

#ifdef HASISATTY
#define TTY       isatty(0)
#else /*HASISATTY*/
#define TTY      1   /* always true for non-Unix */
#endif /*HASISATTY*/

int             getinput();
void            closeBatch();
void            globalFpeRoutine(), globalIntRoutine();
static short    DoInit = 1;
void            doMacanova(void);
extern long     Nscratch, Maxscratch;
extern long     Nsymbols, Maxsymbols;
extern long     Nhandles, Maxhandles;
extern long     CurrentMemory, MaxMemory; /* defined in unxhandl.c */

#ifndef COMMANDLINE
main()
#else /*COMMANDLINE*/

#ifdef MSDOS
char	      **ArgV;   /* set equal to argv */
#endif /*MSDOS*/

int main(int argc,char ** argv)
#endif /*COMMANDLINE*/
{
#ifdef COMMANDLINE
	char            *home, **tempHandle = (char **) 0, *cmd;
	char            *rcFileName;
	char            *initFileName = (char *) 0;
	char            *batchFileName = (char *) 0;
	char            *restoreFileName = (char *) 0;
	char            *helpFileName = (char *) 0;
	char            *dataFileName = (char *) 0;
	char            *macroFileName = (char *) 0;
	char            *nlines = (char *) 0, *ncols = (char *) 0, *tmp;
	long             value;
	long             quiet = 0;
	long             i;
	FILE            *batchFile;
#endif /*COMMANDLINE*/
	WHERE("main");
	
	ISATTY = TTY;

#ifdef MACANOVAINI
	rcFileName  = MACANOVAINI;
#else /*MACANOVAINI*/
	rcFileName = (char *) 0;
#endif /*MACANOVAINI*/

#ifdef COMMANDLINE
#ifdef MSDOS
	ArgV = argv;
#endif /*MSDOS*/

	for(ProgName = *argv;*ProgName;ProgName++)
		;
#ifdef MSDOS
	while(--ProgName >= *argv && *ProgName != SEP && *ProgName != ':')
		;
#else /*MSDOS*/
	while(--ProgName >= *argv && *ProgName != SEP)
		;
#endif /*MSDOS*/
	ProgName++;

	while(SHIFT && **argv == '-')
	{
		cmd = *argv;
		if(strcmp(cmd,"-q") == 0)
		{
			quiet = 1;
		} /*if(strcmp(cmd,"-q") == 0)*/
		else
		{
			SHIFT;
			if(argc == 0)
			{
				usage();
			}
			
			if(strcmp(cmd,"-f") == 0)
			{
				initFileName = *argv;
			}
			else if(strcmp(cmd,"-r") == 0)
			{
				restoreFileName = *argv;
			}
			else if(strcmp(cmd,"-b") == 0)
			{
				batchFileName = *argv;
			}
			else if(strcmp(cmd,"-h") == 0)
			{
				helpFileName = *argv;
			}
			else if(strcmp(cmd,"-d") == 0)
			{
				dataFileName = *argv;
			}
			else if(strcmp(cmd,"-m") == 0)
			{
				macroFileName = *argv;
			}
			else if(strcmp(cmd,"-l") == 0)
			{
				nlines = *argv;
			}
			else if(strcmp(cmd,"-w") == 0)
			{
				ncols = *argv;
			}
			else
			{
				usage();
			}
		} /*if(strcmp(cmd,"-q") == 0){}else{}*/
	} /*while(SHIFT && **argv == '-')*/
	if(argc > 0)
	{
		usage();
	}
	TERMINAL = getenv("TERM");
	
	if(ISATTY)
	{
		SCREENHEIGHT = DEFAULTSCREENHEIGHT;
		BATCHECHO[0] = 0;
	}
	else
	{
		SCREENHEIGHT = 0;
		BATCHECHO[0] = 1;
	}
#ifdef USECBREAK
	saveTerminalInfo();
#endif /*USECBREAK*/
	myInitHandles(); /* initialize handles */
#endif /*COMMANDLINE*/

#ifdef ACTIVEUPDATE
	if (ISATTY)
	{
		activeUpdateInit();
	}
	
#endif /*ACTIVEUPDATE*/

	if(!quiet)
	{
		startupMsg();
	}	

	RDEPTH = 0;	
	if(setjmp(RestartBuf[RDEPTH]))
	{
		if(!ISATTY)
		{
			myerrorout("WARNING: MacAnova terminated by interrupt in non-interactive mode");
			goto endit;
		}
	}

	INTERRUPT = INTNOTSET;

	signal(SIGFPE, fpeRoutine);
	signal(SIGINT, intRoutine);
	if(DoInit)
	{ /* make sure initialize is only called once */
		DoInit = 0;

		INTWHERE[0] = NAMEFORWHERE;
		for(i=1;i<MAXRDEPTH;i++)
		{
			INTWHERE[i] = (char *) 0;
		}

		if(helpFileName != (char *) 0 && !isfilename(helpFileName))
		{
			sprintf(OUTSTR,
					"WARNING: %s is not legal name for help file.  Default will be used",
					helpFileName);
			putOUTSTR();
			helpFileName = (char *) 0;
		}
		if(dataFileName != (char *) 0 && !isfilename(dataFileName))
		{
			sprintf(OUTSTR,
					"WARNING: %s is not legal name for data file.  Default will be used",
					dataFileName);
			putOUTSTR();
			dataFileName = (char *) 0;
		}
		if(macroFileName != (char *) 0 && !isfilename(macroFileName))
		{
			sprintf(OUTSTR,
					"WARNING: %s is not legal name for macro file.  Default will be used",
					macroFileName);
			putOUTSTR();
			macroFileName = (char *) 0;
		}

		initialize(helpFileName,dataFileName,macroFileName);
		 /* make sure help file is found if there */
#ifndef BINARYHELPFILE
		HELPFILE = fopen(*HELPFILENAME,"r");
#else /*BINARYHELPFILE*/
		HELPFILE = fopen(*HELPFILENAME,"rb"); /* open in binary mode */
#endif /*BINARYHELPFILE*/
		if(!quiet && HELPFILE != (FILE *) 0 &&
		   (LINE = mygethandle(MAXLINE)) != (char **) 0)
		{
			/* look for message of the day preceded and followed by "!!!!"*/
			long      printit = 0;

			while(fillLINE(HELPFILE) != EOF)
			{
				if(strncmp(*LINE,"!!!!",4) == 0)
				{
					if(printit)
					{
						break;
					}
					else
					{
						printit = 1;
						continue;
					}
				}
				
				if(printit)
				{
					if(printit == 1)
					{
						myeol();
						printit = 2;
					}
					myprint(*LINE);
				}
			} /*while(fillLINE(HELPFILE) != EOF)*/
			rewind(HELPFILE);
			mydisphandle(LINE);
			LINE = (char **) 0;
		}

#ifdef COMMANDLINE
		if(restoreFileName != (char *) 0)
		{
			if(doRestore(restoreFileName))
			{
				if(PRINTWARNINGS)
				{
					myprint("WARNING: cannot restore workspace from file ");
					myerrorout(restoreFileName);
				}				
				restoreFileName = (char *) 0;
			}
			else
			{
				if(initFileName != (char *) 0)
				{
					myerrorout("WARNING: cannot use -r and -f options together; -f ignored");
				}
				initFileName = rcFileName = (char *) 0;
			}
		}
		if(nlines != (char *) 0)
		{
			value = 0;
			tmp = nlines;
			while(*tmp == ' ')
			{
				tmp++;
			}
			while(*tmp && isdigit(*tmp))
			{
				value = 10*value + (*tmp++) - '0';
			}
			if(*tmp != '\0' || tmp == nlines || (value < 6 && value != 0))
			{
				sprintf(OUTSTR,
						"ERROR: non-zero value '%s' for -l not integer >= 6; ignored", 
						nlines);
				putOUTSTR();
			}
			else
			{
				SCREENHEIGHT = value;
			}
		}
		if(ncols != (char *) 0)
		{
			value = 0;
			tmp = ncols;
			while(*tmp == ' ')
			{
				tmp++;
			}
			while(*tmp && isdigit(*tmp))
			{
				value = 10*value + (*tmp++) - '0';
			}
			if(*tmp != '\0' || tmp == ncols || value < 20)
			{
				sprintf(OUTSTR,
						"ERROR: value '%s' for -w not integer >= 20; ignored",
						ncols);
				putOUTSTR();
			}
			else
			{
				SCREENWIDTH = value;
			}
		}

/*
  If -f initFileName not specified use ~/rcFileName, if it exists
*/
		if(initFileName == (char *) 0 && rcFileName != (char *) 0)
		{
			if((home = getenv("HOME")))
			{
				tempHandle = mygethandle(strlen(home) + strlen(rcFileName) + 2);
				if(tempHandle == (char **) 0)
				{
					goto fatalError;
				}
				strcpy(*tempHandle,home);
				strcat(*tempHandle,"/");
				strcat(*tempHandle,rcFileName);
				initFileName = *tempHandle;
			}
			else
			{
				initFileName = rcFileName;
			}
		}
/*
  NOTE: if batchFileName has been specified, it is actually executed *after*
        initFileName, even though it appears to be invoked first.  It is as
		if batch(initFileName,echo:F) were the first line of batchFileName.
*/
		if(batchFileName != (char *) 0)
		{
			batchFile = fopen(batchFileName,"r");
			if(batchFile != (FILE *) 0)
			{
				fclose(batchFile);
				(void) doBatch(batchFileName,1);
			}
			else if(PRINTWARNINGS)
			{
				myprint("WARNING: Unable to open batch file ");
				myerrorout(batchFileName);
			}
		}
		if(initFileName != (char *) 0)
		{ /* no warning needed here since macanova.ini file is optional */
			(void) doBatch(initFileName,0);
		}
		if(tempHandle != (char **) 0)
		{
			mydisphandle(tempHandle);
		}
#endif /*COMMANDLINE*/
	}
	doMacanova();
	
	if (GUBED & MEMGUBED)
	{
		memoryUsage(NAMEFORWHERE);
	}
	
  endit:
	
	if(HELPFILE != (FILE *) 0)
	{
		fclose(HELPFILE);
	}
	if(BDEPTH > 0)
	{
		closeBatch(1);
	}
#ifdef USECBREAK
	restoreTerminalInfo();
#endif /*USECBREAK*/
	return (0);

  fatalError:
	PUTALERT("FATAL ERROR: unable to allocate space during inialization sequence");
	exit(-1);
} /*main()*/

void doMacanova(void)
{
	int               done = 0;
	long              lastchar;
#ifdef HASSYSTEM
	unsigned int      errorCode;
#endif /*HASSYSTEM*/	
	WHERE("doMacanova");
	
	while( !done)
	{
		cleanitup();

		putprompt((char *) NULL);
		yydebug = (GUBED & 2048) ? 1 : 0;
#ifdef COUNTMISSES
/*
   code to allow run-time tuning of sizes of warehouses (handle caches)
   activated by setoptions(GUBED:4096+size)
*/
		if (GUBED & MEMGUBED && GUBED > MEMGUBED)
		{
			setWarehouseLimits(GUBED - MEMGUBED);
			GUBED = MEMGUBED;
		} /*if (GUBED & MEMGUBED && GUBED > MEMGUBED)*/
#endif /*COUNTMISSES*/

#ifdef MONITOR
		if (GUBED & MONGUBED && !MonitorStarted)
		{
			startMonitor();
			if (!MonitorStarted)
			{
				PRINT("Cannot allocate MonitorBuffer\n",0,0,0,0);
				GUBED = 0;
			}
			else
			{
				GUBED = MONGUBED;
			}
		}
		else if (MonitorStarted && !(GUBED & MONGUBED))
		{
			stopMonitor();
		}
#endif /*MONITOR*/

		if(getinput() <= 0 )
		{
			done = 1;
		}
#ifdef HASSYSTEM
		else if((*INPUTSTRING)[0] == '!')
		{
			lastchar = strlen((char *) *INPUTSTRING) - 1;
			if ((*INPUTSTRING)[lastchar] == '\n')
			{ /* trim off trailing '\n'*/
				(*INPUTSTRING)[lastchar] = '\0';
			}
			
			errorCode = system((char *) *INPUTSTRING + 1);
			if(errorCode)
			{
				sprintf(OUTSTR,"WARNING: ");
#ifdef UNIX
				errorCode >>= 8;
#endif /*UNIX*/
				sprintf(OUTSTR,"WARNING: ");
#ifdef BCPP
				strcat(OUTSTR,strerror(errno));
#else /*BCPP*/
				sprintf(OUTSTR+strlen(OUTSTR),"Shell return code = %ld",
						errorCode);
#endif /*BCPP*/
				putOUTSTR();
			}
		}
#endif /*HASSYSTEM*/
#ifndef DJGPP
		else if((done = yyparse()) != 0)
#else /*DJGPP*/
		else if((_go32_was_ctrl_break_hit(),done = yyparse()) != 0)
#endif /*DJGPP*/
		{
			if(MAXERRORS1)
			{
				NERRORS++;
			}
			
			FatalError = (done == FATALERROR) ? 1 : FatalError;
			
			done = (done == END || FatalError != 0);
		}
		if(MAXERRORS1 && NERRORS >= MAXERRORS1)
		{
			if(BDEPTH > 0)
			{
				sprintf(OUTSTR,"WARNING: too many errors on batch file %s",
						*INPUTFILENAMES[BDEPTH]);
				putOUTSTR();
				closeBatch(1);
			}
			if(MAXERRORS && NERRORS >= MAXERRORS)
			{
				myerrorout("ERROR: too many errors");
				FatalError = 1;
				done = 1;
			}
		}
	} /* while(!done) */

#ifdef MONITOR
	if (MonitorStarted)
	{
		stopMonitor();
	} /*if (MonitorStarted)*/
#endif /*MONITOR*/

	if(FatalError)
	{
		myerrorout("FATAL ERROR: Execution terminated");
	}
	
} /*doMacanova() */
	
#ifdef COMMANDLINE
int doBatch(fileName, echo)
char         *fileName;
int           echo;
{
	FILE           *batchFile;
	long            nargs;
	long            typeList[2];
	char           *valueList[2];
	char           *keywordList[2];
	Symbolhandle    list;
	WHERE("doBatch");


	batchFile = fopen(fileName,"r");
	if(batchFile == (FILE *) 0)
	{
		return (0);
	}
 /* dummy up MacAnova 'batch(filename,echo)' command */
	fclose(batchFile);
	nargs = 2;
	typeList[0] = CHAR;
	valueList[0] = fileName;
	keywordList[0] = "";
	typeList[1] = LOGIC;
	valueList[1] = (echo) ? "T" : "F";
	keywordList[1] = "echo";
	list = Buildlist(nargs,typeList,valueList,keywordList);
	if(list != (Symbolhandle) 0)
	{
		(void) batch(list);
		Removelist(list);
	}
	return (1);
} /*doBatch()*/

/*
	Dummy up restore(fileName) command and execute it.

	doRestore appears in macMain.c in mac version; the ifdefs are included
	to make the code identical
*/

#ifdef MACINTOSH
int doRestore(fileName, vRefNum)
char            *fileName;
Integer          vRefNum;
#else /*MACINTOSH*/
int     doRestore(fileName)
char            *fileName;
#endif /*MACINTOSH*/
{
	Symbolhandle list;
	long         typeList[1];
	char        *valueList[1];
	char        *keywordList[1];
	FILE        *restoreFile;
	int          errorFlag = 0;
#ifdef MACINTOSH
	OSType       types[2];
	
	if(fileName == (char *) 0)
	{
		types[0] = KEEPFILETYPE;
		types[1] = ASCIIKEEPFILETYPE;
		fileName = macOpen("\pMacAnova workspace file",(Str255) 0, READIT,
			2, types);
		if(fileName == (char *) 0)
		{
			return (1);
		}
		vRefNum = Reply.vRefNum; /*global set by macOpen() */
	}
	SetVol(0L, vRefNum);
#endif /*MACINTOSH*/

	restoreFile = fopen(fileName,"r");
	if(restoreFile == (FILE *) 0)
	{
		if(PRINTWARNINGS)
		{
			myprint("WARNING: unable to open file ");
			myerrorout(fileName);
		}
		errorFlag = 1;
	}
	else
	{
		fclose(restoreFile);

		/* dummy up restore(filename) */
		typeList[0] = CHAR;
		valueList[0] = fileName;
		keywordList[0] = "";
		list = Buildlist(1,typeList,valueList,keywordList);
		if(list == (Symbolhandle) 0)
		{
			errorFlag = 1;
		}
		else
		{
			if(restore(list) == (Symbolhandle) 0)
			{
				errorFlag = 1;
			}
#ifdef MACINTOSH
			UnloadSeg((Ptr) restore);
#endif /*MACINTOSH*/
			Removelist(list);
		}
		if(errorFlag && PRINTWARNINGS)
		{
			myprint("WARNING: unable to restore workspace from file ");
			myerrorout(fileName);
		}
	}
	return (errorFlag);
} /*doRestore()*/
#endif /*COMMANDLINE*/

void closeBatch(all)
long        all;
{

	while(BDEPTH > 0)
	{
		mydisphandle(INPUTFILENAMES[BDEPTH]);
		INPUTFILENAMES[BDEPTH] = (char **) 0;
		fclose(INPUTFILE[BDEPTH]);
		INPUTFILE[BDEPTH--] = (FILE *) 0;
		if(!all)
		{
			break;
		}
	}
	if(BDEPTH <= 0)
	{
		MAXERRORS1 = MAXERRORS;
	}
}

#define BS    8
#define TAB   9
#define LF   10
#define VT   11
#define FF   12
#define CR   13
#define DEL  127

static char      **LastInputString = (char **) 0;

int getinput()
{

	/* routine to get a "line" of input from stream and put
	   it in a place for the parser to use
	   
	   "line" is general, because we have to worry about carry
	   across literal quotes and compound lines
	   
	   */
	Symbolhandle    lastline;
	unsigned int    c, c1;
	int             first = 1;
	int             cc, cc1;
#ifdef ACTIVEUPDATE
	long            linestart = 0;
#endif /*ACTIVEUPDATE*/
	long            spaceused, spaceleft, quotelvl, bracklvl;
	long            eofFound = 0;
	long            foundComment = 0;
	int             foundbang = 0;
	char           *helpCall = "help(\"__teaser__\")\n";
	FILE           *inputFile;
	WHERE("getinput");
	
	inputFile = INPUTFILE[BDEPTH];
	spaceleft = CHUNK;

	Remove("LASTLINE");
	if(LastInputString != (char **) 0 &&
		(lastline = Install("LASTLINE", MACRO)) != (Symbolhandle) 0)
	{
		setSTRING(lastline,LastInputString);
		setNDIMS(lastline,1);
		setDIM(lastline,1,1);
	}
	LastInputString = (char **) 0;

	mydisphandle((char **) INPUTSTRING);
	
	INPUTSTRING = (unsigned char **) mygethandle(spaceleft);
	if(INPUTSTRING == (unsigned char **) 0)
	{
		goto errorExit;
	}
	
	(*INPUTSTRING)[0] = (unsigned char) '\0';
	ISTRCHAR = 0;
	quotelvl = 0;
	bracklvl = 0;
	spaceused = 0;

	do
	{ /* skip completely blank lines on batch file */
		cc = getc(inputFile);
	} while(BDEPTH > 0 && (cc == LF || cc == CR));

	if(cc == EOF)
	{
		eofFound = 1;
	}
	else
	{
		c = (unsigned int) cc;
#ifdef HASSYSTEM
		foundbang = (cc == '!');
#endif /*HASSYSTEM*/
	}
	
	while (!eofFound)
	{/* loop exited on EOF or by return on EOL outside of {...} or "..." */
		if (spaceleft < 5)
		{/* make  more space */
			spaceleft = CHUNK;
			INPUTSTRING = (unsigned char **) mygrowhandle((char **)INPUTSTRING,
									   (long) spaceused + spaceleft);
			if (INPUTSTRING == (unsigned char **) 0)
			{
				goto errorExit;
			}
			/* should have a panic space saver here */
		} /*if (spaceleft < 5)*/

		if(first)
		{
			first = 0;
		}
		else
		{
			c = (unsigned int) (cc = getc(inputFile));
		}
		
		if(c >= ' ' || cc == EOF || cc == BS ||
		   cc != VT && cc != FF && isspace(cc))
		{ /* ignore most controls */
			if(cc == EOF)
			{
				eofFound = 1;
			}
			else
			{
				(*INPUTSTRING)[spaceused] = c;
				spaceused++;
				spaceleft--;
				if (foundbang && cc != '\n' && cc != '\\')
				{	
					continue;
				}
				
				switch (cc)
				{
				  case '"':
					if(!foundComment)
					{
						quotelvl = 1 - quotelvl;
					}
					break;
				  case '{':
					if (quotelvl == 0 && !foundComment)
					{
						bracklvl++;
					}
					break;
				  case '}':
					if (quotelvl == 0 && !foundComment)
					{
						bracklvl--;
					}
					break;
				  case '#':
					if(quotelvl == 0)
					{
						foundComment = 1;
					}
					break;
				  case '\\': /* escape out a character */
					/* 
					   On '\', copy '\' and next character to INPUTLINE.
					   If next character is EOL, give prompt if at prompt
					   level.
					   */
					/* get character following '\' */
					c1 = (unsigned int) (cc1 = getc(inputFile));
					if(cc1 == EOF)
					{
						eofFound = 1;
					}
					else
					{
						if (foundbang && (cc1 == '\n' || cc1 == '\r'))
						{ /* delete '\' before new line */
							spaceused--;
							spaceleft++;
						}
						
						(*INPUTSTRING)[spaceused] = c1;
						spaceused++;
						spaceleft--;
						if (foundbang)
						{
							continue;
						}
						
						if((cc1 == '\n' || cc1 == '\r') &&
						   (BDEPTH == 0 && !BATCHECHO[0]))
						{
#ifdef ACTIVEUPDATE
							(*INPUTSTRING)[spaceused-1] = '\0';
							saveForUpdate((char *) *INPUTSTRING + linestart);
							saveForUpdate("\n");
							if (NLINES < SCREENHEIGHT - 2)
							{
								incrementNLINES("\n");
							}
							linestart += strlen((char *) *INPUTSTRING + linestart) + 1;
							(*INPUTSTRING)[spaceused-1] = c1;
#endif /*ACTIVEUPDATE*/
							putprompt(MOREPROMPT);
						}
					}
					break;

				  case LF:
				  case CR:
					/* Note character already added to line */
					foundComment = 0;
					if(BDEPTH == 0 && !BATCHECHO[0])
					{
						if (NLINES < SCREENHEIGHT - 2)
						{	/* be careful not to force pause */
							incrementNLINES("\n");
						}
#ifdef ACTIVEUPDATE
						(*INPUTSTRING)[spaceused-1] = '\0';
						saveForUpdate((char *) *INPUTSTRING + linestart);
						(*INPUTSTRING)[spaceused-1] = c;
						saveForUpdate("\n");
						linestart += strlen((char *) *INPUTSTRING + linestart) + 1;
#endif /*ACTIVEUPDATE*/
					}
					if (foundbang || quotelvl == 0 && bracklvl <= 0)
					{		/* command line is complete */
						(*INPUTSTRING)[spaceused] = '\0';
						if(BATCHECHO[BDEPTH])
						{
							mymultiprint((char *) *INPUTSTRING);
						}
						else if(BDEPTH == 0)
						{
							if(SPOOLFILE != (FILE *) 0)
							{
								fprintf(SPOOLFILE, "%s", *INPUTSTRING);
							}
						}

						/* treat naive call for help as just that */
						if(spaceused == 5 &&
						   (strncmp((char *) *INPUTSTRING,"help",4) == 0 ||
							strncmp((char *) *INPUTSTRING,"Help",4) == 0))
						{
							strcpy((char *) *INPUTSTRING,helpCall);
							spaceused = strlen(helpCall);
							(*INPUTSTRING)[spaceused] = '\0';
						}

						goto normalExit;
					} /*if (quotelvl == 0 && bracklvl <= 0)*/
					if(BDEPTH == 0 && !BATCHECHO[0])
					{
						putprompt(MOREPROMPT);
					}
					break;

				  case BS:
				  case DEL:
					spaceused--;
					spaceleft++;
					if (spaceused > 0)
					{
						spaceused--;
						spaceleft++;
					}
					break;

				  default:
					;
				} /*switch (cc)*/
			} /*if(cc == EOF){...}else{...}*/
		} /*if(c >= ' ' || cc == EOF || cc == BS || isspace(cc))*/
		else
		{
			; /* ignore anything else */
		}		
	} /* while(!eofFound) */

	/* you only get here if EOF has been found */
	(*INPUTSTRING)[spaceused] = '\0';

	if(BDEPTH == 0 && spaceused == 0)
	{ /* fake quit if endfile at start of line */
		strcpy((char *) *INPUTSTRING,"quit\n");
		spaceused = strlen((char *) *INPUTSTRING);
		SCREENHEIGHT = 0;
		myeol();
		myerrorout("Normal termination by end of file on input");
	} /*if(BDEPTH == 0 && spaceused == 0)*/

	if(BDEPTH > 0)
	{
		if(BATCHECHO[BDEPTH])
		{
			if(spaceused == 0)
			{
				putPieces("(end of file on ",
						  (char *) *INPUTFILENAMES[BDEPTH], ")", (char *) 0);
			}
			else
			{
				mymultiprint((char *) *INPUTSTRING);
			}
		}
		closeBatch(0); /* close batch file and clean up */
	} /*if(BDEPTH > 0)*/
	
  normalExit:
	if((*INPUTSTRING)[0] != '\0' &&
	   (LastInputString = myduphandle((char **) INPUTSTRING)) != (char **) 0)
	{
		if(spaceused > 0 &&
		   ((cc = (*LastInputString)[spaceused-1]) == LF || cc == CR))
		{
			(*LastInputString)[--spaceused] = '\0';
		}
		LastInputString = mygrowhandle(LastInputString, spaceused+1);
	}

	return (1);

  errorExit:
	myerrorout("REALLY out of space! Sorry!");
	FatalError = 1;
	
	return (0);
}
