%{
/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Mainpars
#else
#define __SEG__ Mainpars
#endif /*MPW3*/
#endif /*MPW*/

/*
  main parser for macanova by G. Oehlert, as modifed by C. Bingham for use
  with MacAnova 3.0
*/

/*
	Version of 920623 made many changes, here and in yylex.c and Symbol.c
	All names are returned by yylex as VAR unless they are BLTINs or MACROs
	followed by '('.  Thus the parser should never see STRUC, UNDEF, and should
	never see MACRO or BLTIN or BATCH unless followed by '('
	The number of shift/reduce conflicts has been reduced to 13

	930201 Added 'break n' functionality which also is associated with token
	BREAK, and modified yylex so that 'breakall' is equivalent to
	'break WDEPTH'.  When n is a literal positive integer, 'break n' results
	in macanova	breaking out of exactly n enclosing loops.

	930208 Added a[..][..] capability, now 17 shift/reduce errors

	930804 Added operators <-+ <-- <-*, <-/, <-^, <-%%; 23 shift/reduce errors

	931007 Added lines to make use of batch() as an expression illegal.

	931103 Added error checking for left side of <-+, <--, etc; 47 S/R errors

	931110 Communication with Ifsetup() now by return value rather than global

	931228 Found that mainpars.c as directly produced was too large for the
           Mac MPW compiler.  Accordingly, rearranged various cases and used
		   more gotos, to maximize the length of sequences of case statements
           with identical code that could be folded into one another by
           editing of mainpars.c for use on the Macintosh.  The first case
           of a set of, say 3, cases to be folded has a comment of the
		   form 'Mac: fold 3 cases'. Script mainpars.awk for awk uses this
		   to create mainpars.c.mac.new.  This also has the parser tables
		   edited out and put into file mainpars.r.mac.new in a form readable
		   by MPW resource compiler Rez.

		   Note: Some of the folded cases are not identical in mainpars.y, but
		   are translated into identical code by Yacc.  Conversely, some
		   cases that appear to be identical in mainpars.y are translated
		   by Yacc in different ways.
	
	940110 Fixed bug in byname BATCH parenlist

	940119 Found and fixed bug.  yyparse() was telling yylex() to send an
	unnecessary (and harmful) ';' after if(l){..}else{...} and after
	while(l){...} .

	940331 Added bitwise operations BITAND, BITXOR, BITOR, and BITNOT

	940411 Added line adding  component ASSIGN  to definition of 'assign',
	in order to generate useful error message.
	Now 48 shift/reduce conflicts

	940630 Implemented if{...}elseif{...}...elseif{...}else{...}
	Now 49 shift/reduce

	950411 Fixed bug in elseif ({...} not always skipped on False
	950710 Added parserCopy() to fix bug interpreting statements like
    {b<-1}+{b<-2}

	950722 Restricted use of parserCopy() to eliminate buildup of scratch
	variables.  {b<-1}+{b<-2} again will give the wrong answer.  However,
	if the {...} is really an expanded macro it will give the right answer
	This can still result in the buildup of scratch is the value of a
	macro is not assigned.  Hence, guidelines should suggest that
	macros not returning a value should end with ';;'

	950807 Added check for BLTIN and LIST to parserCopy()

VERSION: 950807   This date is used by mainpars.awk in making macMainpars.r

*/

#include "globals.h"

#ifdef MACINTOSH

#define CHECKINTERRUPT	if(INTERRUPT != INTNOTSET) YYERROR

#else /*MACINTOSH*/

#define CHECKINTERRUPT

#endif /*MACINTOSH*/

/* these must be defined identically in yylex.c */
#define SKIPBRACKETBLOCK  3
#define SKIPPARENBLOCK    4

#define YYSTYPE Symbolhandle 	/* handle to symbol */
#define YYDEBUG 1

extern Symbolhandle Install(), Extract();
extern Symbolhandle Makelist(), Addlist(), Growlist();
extern Symbolhandle Assign(), Subassign(), Arith();
extern Symbolhandle Unary(), Logic(), Element();
extern Symbolhandle Byname();
extern void         Removelist();
extern void         Unscratch();
extern void         RemoveUndef();

void                cleanitup();

#ifdef MACINTOSH
extern short *Pyyexca;
extern short *Pyyact;
extern short *Pyypact;
extern short *Pyypgo;
extern short *Pyyr1;
extern short *Pyyr2;
extern short *Pyychk;
extern short *Pyydef;
#endif /*MACINTOSH*/

static void         setupfakefor();
static long         BynameType;
static long         ParseMacroArgs;
static int          Op, AssignOp;
static int          SeenError = 0;
static int          TargetLevel;

#ifdef UNDEFINED__
static WHERE("yyparse"); /*now set by make*/
#endif /*UNDEFINED__*/

long    PARSEMACROARGS = 0;  /* for communication with yylex */
long    YYLEXMESSAGE = 0;  /* for communication with yylex */
long    LASTCH = '\0';  /* for communication with yylex */
int     ALTINPUTTOKENS[5*MAXWDEPTH];  /* for communication with yylex */
int    *ALTTOKENPTR;  /* for communication with yylex */
/*
   End of MacAnova Code before parser instructions
*/
%}
/* 
  Be very careful about deleting tokens so that the values of data types
  CHAR, REAL, STRUC, MACRO, LOGIC, and PLOTINFO do not change since the
  values are used in save().  If they were changed, they would render
  save files unreadable, although the current version of asciisave files
  would remain readable

  The previous comment is no longer operative.  The current version of
  save(), producing files headed "////MacAnovaSaveFileN///", where N >= 3
  uses symbolic names for variable types instead of the numerical values of
  the tokens.  The codes necessary to restore save files from previous versions
  are defined in globdefs.h (CHAR_V31, REAL_V31, etc.).
*/
%token CHAR REAL BLTIN LIST STRUC MACRO UNDEF VAR END
%token LOGIC MACRO1 NUMBER
%token IF ELSE ELSEIF WHILE FOR BREAK BREAKALL BATCH ERROR FATALERROR
%token PLOTINFO ASSIGNED GARB LONG BADTOKEN
%token LEFTANGLE RIGHTANGLE
%token LEFTBRACK2 RIGHTBRACK2
%token MULT ADD DIV MOD SUB EXP GE LE GT LT EQ NE AND OR NOT ENDOFL POW
%token MATMULT TRMATMULT MATMULTTR
%token BITAND BITXOR BITOR BITNOT
%token 
%nonassoc ASSIGN ASSIGN1
%nonassoc ASSIGNMULT ASSIGNADD ASSIGNDIV ASSIGNSUB ASSIGNPOW ASSIGNMOD
%left ','
%left BITOR
%left BITXOR
%left BITAND
%nonassoc BITNOT
%left '|' OR
%left '&' AND
%nonassoc GE LE GT LT EQ NE '=' '<' '>'
%nonassoc '!'
%left '+' '-'
%left '*' '/' MOD MATMULT TRMATMULT MATMULTTR
%left UNARYMINUS
%left POW
%left '$'
%left LEFTBRACK2
%left '['
%left '('
%left '\''

%%
cmpdline: cmpdnobreak   {$$ = $1;/*Mac: fold 6 cases*/}
	| cmpdbreak 		{$$ = $1;}
	;

cmpdnobreak: line		{$$ = $1;}
	| cmpdnobreak line	{$$ = $2;}
	;

cmpdbreak: break		{$$ = $1;}
	| cmpdnobreak break {$$ = $2;}
	;

line:	  ENDOFL	{
						$$ = NULLSYMBOL;
					  lev0Return:
						if(BRACKETLEV == 0)
						{
							return(0);
						}
					}
	| ';'		{
					$$ = NULLSYMBOL;
				  lev0Unscratch:
					
					if(BRACKETLEV == 0)
					{
						Unscratch();
					}
				}
	| assgn ENDOFL	{
						$$ = $1;
						goto lev0Return;
					}
	| assgn ';'		{
						$$ = $1;
						goto lev0Unscratch;
					}
	| batch ENDOFL	{goto dobatch;/*Mac: fold 2 cases*/}

	| batch ';'		{
					  dobatch:
						(void) runCommand(Lookup("batch"), $1);
		  				return (0); /* terminate parse */
					}

	| expr ENDOFL	{
						$$ = $1;
						if(BRACKETLEV == 0)
						{
							prexpr($1);
							return(0);
						}
					}
	| expr ';'		{
						$$ = $1;
						if(BRACKETLEV == 0)
						{
							prexpr($1);
							Unscratch();
						}
					}
	| badbatch		{
						yyerror("ERROR: illegal use of batch()");
						Removelist($1);
		  				return (0); /* terminate parse */
					}
	| expr ERROR	{$$ = (Symbolhandle) 0;YYERROR;/*Mac: fold 2 cases*/}
	| ERROR			{$$ = (Symbolhandle) 0;YYERROR;}
	| error			{$$ = (Symbolhandle) 0;yyerrok; return(0);}
	;

break:	BREAK	{  /* break, break n, breakall */
					if($1 == (Symbolhandle) 0)
					{
						YYERROR;
					}
					TargetLevel = WDEPTH - (int) DATAVALUE($1,0);
					Removesymbol($1);
					if(TargetLevel < 0)
					{
						YYERROR;
					}
					{
						YYLEXMESSAGE = MAXWDEPTH +
							WHILEBRACKETLEV[TargetLevel] + 1;
						cleanLoops(TargetLevel,WDEPTH);
					}
				}
			';'  /* yylex will ensure the ';' is found */
				{
					$$ = NULLSYMBOL;
				}
	;

/*
	Assignment to scratch variable (e.g., (1+3) <- a, is grammatically
	correct, but Assign() diagnoses error
*/

assgn:	expr1 ASSIGN    {
							$$ = Assign($1,(Symbolhandle) 0,-1);
							if($$ == (Symbolhandle) 0)
							{
								YYERROR;
							}
						}
					 	expr
						{
							$$ = Assign($1,$4,0);
						  checkResult:
							if($$ == (Symbolhandle) 0)
							{
								YYERROR;
							}
						}
	| component ASSIGN	{
							yyerror("ERROR: illegal assignment to structure component");
							YYERROR;
						}

	| VAR bracklist ASSIGN 	{
							  subassigncheck2:
								$$ = Subassign($1,$2,(Symbolhandle) 0,1);
								if($$ == (Symbolhandle) 0)
								{
									YYERROR;
								}
							}
						expr
							{
							  subassign5:
								$$ = Subassign($1,$2,$5,0);
								Removelist($2);
								if($$ == (Symbolhandle) 0)
								{
									YYERROR;
								}
							}
	| '(' expr ')' bracklist ASSIGN	
							{
								$$ = Subassign($2,$4,(Symbolhandle) 0,1);
								if($$ == (Symbolhandle) 0)
								{
									YYERROR;
								}
							}
						expr
							{
								$$ = Subassign($2,$4,$7,0);
								Removelist($4);
								if($$ == (Symbolhandle) 0)
								{
									YYERROR;
								}
							}
	| functionvalue bracklist ASSIGN	
							{
								goto subassigncheck2;
							}
						expr
							{
								goto subassign5;
							}
	| byname VAR bracklist ASSIGN
							{
							  subassigncheck3:
								$$ = Subassign($1,$3,(Symbolhandle) 0,1);
								if($$ == (Symbolhandle) 0)
								{
									YYERROR;
								}
							}
						expr
							{
							  subassign6:
								$$ = Subassign($1,$3,$6,0);
								Removelist($3);
								if($$ == (Symbolhandle) 0)
								{
									YYERROR;
								}
							}

	| expr1 ASSIGNADD expr	{
								Op = ADD;
							  doassignop:
								$$ = Assign($1,(Symbolhandle) 0,-Op);
								if($$ == (Symbolhandle) 0)
								{
									YYERROR;
								}
								$$ = Arith($1, $3, Op);
								UNLOADSEG(Arith);
								if($$ != (Symbolhandle) 0)
								{
									$$ = Assign($1,$$,0);
								}
								goto checkResult;
							}
	| expr1 ASSIGNSUB expr	{
								Op = SUB;
								goto doassignop;
							}
	| expr1 ASSIGNMULT expr	{
								Op = MULT;
								goto doassignop;
							}
	| expr1 ASSIGNDIV  expr	{
								Op = DIV ;
								goto doassignop;
							}
	| expr1 ASSIGNMOD expr	{
								Op = MOD;
								goto doassignop;
							}
	| expr1 ASSIGNPOW expr	{
								Op = EXP;
								goto doassignop;
							}


	| '(' expr ')' bracklist assignop
								{ /* Mac: fold 4 cases */
									AssignOp = LONGDATAVALUE($5,1);
									goto assignoperror;
								}
	| functionvalue bracklist assignop
								{
									AssignOp = LONGDATAVALUE($3,1);
									goto assignoperror;
								}
	| byname VAR bracklist assignop
								{
									AssignOp = LONGDATAVALUE($4,1);
									goto assignoperror;
								}
	| VAR bracklist assignop 	{
									AssignOp = LONGDATAVALUE($3,1);
								  assignoperror:
									sprintf(OUTSTR,
											"ERROR: illegal left side of %s",
											opName(AssignOp));
									yyerror(OUTSTR);
									YYERROR;
								}
	;


assignop:  ASSIGNADD	{
	 						Op = ADD;
							AssignOp = ASSIGNADD;
						  setassignOp:
							{
								Symbolhandle symhOp = LongInstall(SCRATCH, 2);
								if(symhOp == (Symbolhandle) 0)
								{
									YYERROR;
								}
								LONGDATAVALUE(symhOp,0) = (long) Op;
								LONGDATAVALUE(symhOp,1) = (long) AssignOp;
								$$ = symhOp;
							}
						}
	|		ASSIGNSUB   {Op = SUB;AssignOp = ASSIGNSUB;goto setassignOp;}
	|		ASSIGNMULT	{Op = MULT;AssignOp = ASSIGNMULT;goto setassignOp;}
	|		ASSIGNDIV	{Op = DIV;AssignOp = ASSIGNDIV;goto setassignOp;}
	|		ASSIGNMOD	{Op = MOD;AssignOp = ASSIGNMOD;goto setassignOp;}
	|		ASSIGNPOW	{Op = EXP;AssignOp = ASSIGNPOW;goto setassignOp;}
	;

batch:	BATCH parenlist	{
							$$ = $2;
						}
	| byname BATCH parenlist	{
									$$ = $3;
				 				}
keyword:  compname ASSIGN1 expr 	
							{
								$$ = Assign($1,$3, 1);

								goto checkResult;
							}
	;

/* yylex will never return BLTIN unless followed by '(' */

functionvalue: BLTIN parenlist	{
									$$ = runCommand($1, $2);
								  setScrName1:
									if($$ == (Symbolhandle) 0)
									{
										YYERROR;
									}
									setScratchName($$);
									CHECKINTERRUPT;
								}

	| byname BLTIN parenlist	{
									$$ = runCommand($1,$3);
									goto setScrName1;
								}
	;

component:	expr '$' compname	{
									$$ = Extract($1,$3,1);
									
								  setScrName2:
									if($$ == (Symbolhandle) 0)
									{
										YYERROR;
									}
									setScratchName($$);
								}
	| expr brack2list	{
							$$ = Extract($1,$2,0);
							goto setScrName2;
						}
	;

expr1:	functionvalue	{$$ = $1;}
	| missingrparen	{
						YYERROR;
					}

	| VAR				{$$ = $1;}
	| byname VAR		{$$ = $1;}
	| expr '+' expr		{
							Op = ADD;
						  arith:
							$$ = Arith($1,$3,Op);
							/* Arith() removes SCRATCH unconsumed args */
							UNLOADSEG(Arith);
							goto setScrName1;
						 }
	| expr '-' expr			{ Op = SUB;goto arith; }
	| expr '*' expr			{ Op = MULT;goto arith; }
	| expr MATMULT expr		{ Op = MATMULT;goto arith; }
	| expr TRMATMULT expr	{ Op = TRMATMULT;goto arith; }
	| expr MATMULTTR expr	{ Op = MATMULTTR;goto arith; }
	| expr '/' expr			{ Op = DIV;goto arith; }
	| expr MOD expr			{ Op = MOD;goto arith; }
	| expr POW expr			{ Op = EXP;goto arith; }
/*#ifdef BITOPS*/
	| expr BITOR expr		{ Op = BITOR;goto arith; }
	| expr BITXOR expr		{ Op = BITXOR;goto arith; }
	| expr BITAND expr		{ Op = BITAND;goto arith; }
/*#endif */
	| '(' expr ')'			{$$ = parserCopy($2);}

	| '-' expr %prec UNARYMINUS	{
									Op = '-';
								  unary:
									$$ = Unary($2,Op);
									/*Unary() reuses space if arg is SCRATCH */
									UNLOADSEG(Unary);
									goto setScrName1;
								}
	| '+' expr %prec UNARYMINUS	{
									Op = '+';
									goto unary;
								}
/*#ifdef BITOPS*/
	| BITNOT expr	 			{
									Op = BITNOT;
									goto unary;
								}
/*#endif BITOPS*/
	| expr '=' expr	{
						Op = EQ;
					  logic:
						$$ = Logic($1,$3,Op);
						/*Logic() removes unconsumed scratch args */
						UNLOADSEG(Logic);
						goto setScrName1;
					}
	| expr EQ  expr		{ Op = EQ; goto logic; }
	| expr NE  expr		{ Op = NE; goto logic; }
	| expr '<' expr		{ Op = LT; goto logic; }
	| expr LE  expr		{ Op = LE; goto logic; }
	| expr '>' expr		{ Op = GT; goto logic; }
	| expr GE  expr		{ Op = GE; goto logic; }
	| expr '&' expr		{ Op = AND; goto logic; }
	| expr AND expr		{ Op = AND; goto logic; }
	| expr '|' expr		{ Op = OR; goto logic; }
	| expr OR  expr		{ Op = OR; goto logic; }
	| '!' expr			{
							$$ = Logic($2,$2,NOT);
							UNLOADSEG(Unary);
							goto setScrName1;
						}
	| expr '\''			{
							Symbolhandle list = Addlist( Makelist(), $1 );
							if(list == (Symbolhandle) 0)
							{
								YYERROR;
							}
							strcpy(FUNCNAME,"prime");
							$$ = transpose(list);
							Removelist(list);
							UNLOADSEG(transpose);
							goto setScrName1;
				  		}
/* Macrosetup brackets expanded macro with '{' and '\n }' */
	| expandedmacro	{$$ = $1;}
	;

 expandedmacro:
	 macro parenlist	{
							long reply = Macrosetup($1,$2);

							Removelist($2);
							if(!reply)
							{
								YYERROR;
							}
						}
		 '{' cmpdline '}'  /* Macrosetup() ensures '{' and '}' are there */
						{
							$$ = parserCopy($5);
							MDEPTH--;
						}
	;

missingrparen:	'('   ENDOFL	{goto missingRparen; /*Mac: fold 4 cases*/}
	|	'('  expr1  ENDOFL		{goto missingRparen;}
	|	'('   ';'				{goto missingRparen;}
	|	'('  expr1  ';'			{
								  missingRparen:
									$$ = 0;
									yyerror("ERROR: missing ')'");
									YYERROR;
								}
	;

expr:	expr1		{$$ = $1;/*Mac: fold 3 cases*/}
	|	component	{$$ = $1;}
	| 	assgn		{$$ = $1;}
	|	ifstart
				{
					IDEPTH--;
					$$ = $1;
				}

	| ifstart ELSE	{
						if(IFCONDITIONS[IDEPTH-1] > 0)
						{
							 $$ = $1;
							 YYLEXMESSAGE = SKIPBRACKETBLOCK;
						}
						CHECKINTERRUPT;
						if(!Ifsetup((Symbolhandle) 0, ELSE))
						{
							$$ = 0;
							YYERROR;
						}
					}
		'{' cmpdline '}'
					{
						if(IFCONDITIONS[IDEPTH-1] == 0)
						{
							 $$ = $5;
						}
						IDEPTH--;
					}

	| WHILE parenlist	{
							int    ifOK;

							CHECKINTERRUPT;
							ifOK = Ifsetup($2, WHILE);
							Removelist($2);
							if(!ifOK)
							{
								$$ = 0;
								YYERROR;
							}
							if(WHILECONDITIONS[WDEPTH-1] == 0)
							{
								YYLEXMESSAGE = SKIPBRACKETBLOCK;
							}
							else if(WHILELIMITS[WDEPTH-1] <= 0)
							{ /* too many reps */
								yyerrok ;
								return(0) ;
							}
						}
		'{' cmpdline '}'
						{
							if(WHILECONDITIONS[WDEPTH-1])
							{
								ISTRCHAR = WHILESTARTS[WDEPTH-1];
								if($5 != (Symbolhandle) 0 && BRACKETLEV > 0 &&
									 isscratch(NAME($5)))
								{
									Removesymbol($5);
									$5 = NULLSYMBOL;
								}
								YYLEXMESSAGE = -((long) ';');
							}
							else
							{
								WHILELIMITS[WDEPTH-1] = MAXWHILE;
							}
							WDEPTH--;
							$$ = $5;
						}

	| FOR parenlist		{
							int   ifOK;

							CHECKINTERRUPT;
							ifOK = Ifsetup($2, FOR);
							Removelist($2);
							if(!ifOK)
							{
								$$ = 0;
								YYERROR;
							}
						 }
			'{' cmpdline '}'
						{
							if(WHILECONDITIONS[WDEPTH-1] &&
								 --WHILELIMITS[WDEPTH-1] > 0)
							{
								/* go back for another round */
								ISTRCHAR = WHILESTARTS[WDEPTH-1];
								setupfakefor();
								ALTTOKENPTR = ALTINPUTTOKENS;
								if($5 != (Symbolhandle) 0 && BRACKETLEV > 0 &&
									 isscratch(NAME($5)))
								{
									Removesymbol($5);
									$5 = NULLSYMBOL;
								}
							}
							else
							{ /* done with loop */
								WHILELIMITS[WDEPTH-1] = MAXWHILE;
							}
							WDEPTH--;
							$$ = $5;
						}

	| '{' cmpdline '}'	{$$ = $2;/*Mac: fold 2 cases*/}
	| '{' expr '}'		{$$ = $2;}
	| subscripted	{$$ = $1;}
	| VAR  '('		{
						notfunction($1);
						$$ = 0;
						YYERROR;
					}

	| END			{ return(END); }

	| FATALERROR	{
						return (FATALERROR);
					}
	;

badbatch: expr '+' batch	{$$ = $3;/*Mac: fold 9 cases*/}
	| expr '-' batch		{$$ = $3;}
	| expr '*' batch		{$$ = $3;}
	| expr MATMULT batch	{$$ = $3;}
	| expr TRMATMULT batch	{$$ = $3;}
	| expr MATMULTTR batch	{$$ = $3;}
	| expr '/' batch		{$$ = $3;}
	| expr MOD batch		{$$ = $3;}
	| expr POW batch		{$$ = $3;}
	| batch '+'				{$$ = $1;/*Mac: fold 9 cases*/}
	| batch '-'				{$$ = $1;}
	| batch '*'				{$$ = $1;}
	| batch MATMULT			{$$ = $1;}
	| batch TRMATMULT		{$$ = $1;}
	| batch MATMULTTR		{$$ = $1;}
	| batch '/'				{$$ = $1;}
	| batch MOD				{$$ = $1;}
	| batch POW				{$$ = $1;}
	| '-' batch %prec UNARYMINUS	{$$ = $2;/*Mac: fold 12 cases*/}
	| '+' batch %prec UNARYMINUS	{$$ = $2;}
	| expr EQ  batch	{$$ = $3;}
	| expr NE  batch	{$$ = $3;}
	| expr '<' batch	{$$ = $3;}
	| expr LE  batch	{$$ = $3;}
	| expr '>' batch	{$$ = $3;}
	| expr GE  batch	{$$ = $3;}
	| expr '&' batch	{$$ = $3;}
	| expr AND batch	{$$ = $3;}
	| expr '|' batch	{$$ = $3;}
	| expr OR  batch	{$$ = $3;}
	| batch EQ 			{$$ = $1;/*Mac: fold 10 cases*/}
	| batch NE 			{$$ = $1;}
	| batch '<'			{$$ = $1;}
	| batch LE 			{$$ = $1;}
	| batch '>'			{$$ = $1;}
	| batch GE 			{$$ = $1;}
	| batch '&'			{$$ = $1;}
	| batch AND			{$$ = $1;}
	| batch '|'			{$$ = $1;}
	| batch OR 			{$$ = $1;}
	| '!' batch			{$$ = $2;/*Mac: fold 2 cases*/}
	| '(' batch			{$$ = $2;}
	| batch '('			{$$ = $1;/*Mac: fold 5 cases*/}
	| batch '['			{$$ = $1;}
	| batch '$'			{$$ = $1;}
	| batch RIGHTBRACK2	{$$ = $1;}
	| batch ASSIGN		{$$ = $1;}
	;

subscripted: '{' expr '}' bracklist	{ goto bracklist4; /*Mac: fold 3 cases */}
	| '{' cmpdline '}' bracklist	{ goto bracklist4; }
	| '(' expr ')' bracklist	{
								  bracklist4:
									$$ = Element($2,$4);
									Removelist($4);
									goto setScrName2;
								}
	| functionvalue bracklist	{ goto bracklist2; /*Mac: fold 5 cases */}
	| component bracklist 		{ goto bracklist2; }
	| expandedmacro  bracklist	{ goto bracklist2; }
	| VAR  bracklist 			{ goto bracklist2; }
	| subscripted bracklist	{
							  bracklist2:
								$$ = Element($1,$2);
								Removelist($2);
								goto setScrName2;
							}
	| byname VAR bracklist 	{
								$$ = Element($1,$3);
								Removelist($3);
								goto setScrName2;
							}
	;
		
ifstart: IF parenlist	{
							int    ifOK;

							CHECKINTERRUPT;
							ifOK = Ifsetup($2, IF);
							Removelist($2);
							if(!ifOK)
							{
								$$ = 0;
								YYERROR;
							}
							if(IFCONDITIONS[IDEPTH-1] == 0)
							{
								YYLEXMESSAGE = SKIPBRACKETBLOCK;
							}
						}
		 '{' cmpdline '}'
		 				{
							$$ = (IFCONDITIONS[IDEPTH-1] ) ? $5 : NULLSYMBOL;
						}

	|	ifstart ELSEIF 
						{
							if(IFCONDITIONS[IDEPTH-1] > 0)
							{
								 YYLEXMESSAGE = SKIPPARENBLOCK;
							}
						}
						parenlist
						{
							int    ifOK;

							if(IFCONDITIONS[IDEPTH-1] > 0)
							{ /* already found True condition*/
								YYLEXMESSAGE = SKIPBRACKETBLOCK;
								Removelist($4);
								$4 = (Symbolhandle) 0;
							}
							CHECKINTERRUPT;
							ifOK = Ifsetup($4, ELSEIF);
							if($4 != (Symbolhandle) 0)
							{
								Removelist($4);
							}
							if(!ifOK)
							{
								$$ = 0;
								YYERROR;
							}
							if(IFCONDITIONS[IDEPTH-1] == 0)
							{
								YYLEXMESSAGE = SKIPBRACKETBLOCK;
							}
						}
		 '{' cmpdline '}'
		 				{
							if($4 == (Symbolhandle) 0)
							{
								$$ = $1;
							}
							else if(IFCONDITIONS[IDEPTH-1] > 0)
							{
								$$ = $7;
							}
							else
							{
								$$ = NULLSYMBOL;
							}
						}
	| ifstart ERROR		{ $$ = 0 ; YYERROR;}
		 ;

macro:	MACRO	{$$ = $1;}
	|   byname MACRO	{
							$$ = $1;
							PARSEMACROARGS = ParseMacroArgs;
						}
	;


/* WARNING: Do not fold next 2 cases */
compname:	VAR		{$$ = $1;}
	|	byname VAR	{$$ = $1;}
	;

byname:	LEFTANGLE expr RIGHTANGLE
					{
						$$ = Byname($2,&BynameType,&ParseMacroArgs);
						if($$ == (Symbolhandle) 0)
						{
							YYERROR;
						}
						else
						{
							YYLEXMESSAGE = -BynameType;
						}
					}
	|	LEFTANGLE expr {
							$$ = 0;
							yyerror("ERROR: missing '>>'");
							YYERROR;
						}
	;

/* note that a 'list' always terminates with a comma */
list:	 /* nothing */ ','	{
							  growMake:
								$$ = Growlist(Makelist());
								goto checkResult;
							}
	| expr ','	 			{goto growAddMake;/*Mac: fold 2 cases*/}
	| keyword ',' 			{
							  growAddMake:
								$$ = Growlist(Addlist(Makelist(),$1));
								goto checkResult;
							}
	| list expr ','			{goto growAdd; /*Mac: fold 2 cases*/}
	| list keyword ','		{
							  growAdd:
								$$ = Growlist(Addlist($1,$2));
								goto checkResult;
							}
	| list ',' 				{
								$$ = Growlist($1);
								goto checkResult;
							}
	;


completelist:  expr 		{goto addMake; /*Mac: fold 2 cases*/}
	|  keyword 				{
							  addMake:
								$$ = Addlist( Makelist(), $1 );
								goto checkResult;
							}
	|  list 				{
								$$ = $1;
								goto checkResult;
							}
	|  list expr 			{goto addlist;/*Mac: fold 2 cases*/}
	|  list keyword			{
							  addlist:
								$$ = Addlist( $1, $2);
								goto checkResult;
							}
	;

parenlist:	'(' /* nothing */ ')'	{
									  makelist:
										$$ = Makelist();
										goto checkResult;
									}
	|	'(' completelist ')'	{
									$$ = $2;
								}
	|	'('	completelist ENDOFL	{goto missingRparen; /*Mac: fold 2 cases*/}
	|	'('	completelist ';'	{goto missingRparen;}
	;

bracklist:   '[' /* nothing */ ']'
				  	{goto makelist;/*Mac: fold 2 cases*/}
	|   '[' RIGHTBRACK2
				  	{goto makelist;}
	|	'[' completelist ']'
				  	{ $$ = $2; goto checkResult;/*Mac: fold 2 cases*/}
	|	'[' completelist RIGHTBRACK2
				  	{ $$ = $2; goto checkResult;}
	|	'['  ';'  	{goto missingRbracket;/*Mac: fold 3 cases*/}
	|	'['  ENDOFL {goto missingRbracket;}
	|   '[' completelist
					{
					  missingRbracket:
						yyerror("ERROR: missing ']'");
						$$ = 0;
						YYERROR;
					}
					ERROR
	;
/*
	LEFTBRACK2 & RIGHTBRACK2 are '[' & ']' but signal that the next character
	is the same.  A subscript list can never start with LEFTBRACK2
*/
brack2list:   LEFTBRACK2 '[' expr RIGHTBRACK2 ']'
				  	{ $$ = $3; goto checkResult;/*Mac: fold 2 cases*/}
	|	LEFTBRACK2 '[' expr RIGHTBRACK2 RIGHTBRACK2
				  	{ $$ = $3; goto checkResult;}
	|	LEFTBRACK2 '[' ';'		{goto missingRbrack2; /*Mac: fold 3 cases*/}
	|	LEFTBRACK2 '[' ENDOFL	{goto missingRbrack2;}
	|	LEFTBRACK2 '[' expr
					{
					  missingRbrack2:
						yyerror("ERROR: missing ']]'");
						$$ = 0;
						YYERROR;
					}
	;

%%
void yyerror(s)
char           *s;
{
	/*  output error message */
	unsigned char   c = '\0', *inputString = *INPUTSTRING;
	int             start = 0;
	int             istrchar = ISTRCHAR;
	int             iswarning = (strncmp("WARNING:", s, 8) == 0);
	WHERE("yyerror");

	if (!SeenError && (PRINTWARNINGS || !iswarning))
	{
		while(istrchar >= 0)
		{
			if((c = inputString[istrchar]) == '#' || c == '"')
			{
				break;
			}
			istrchar--;
		}
		if(c == '#')
		{
			ISTRCHAR = istrchar;
		}
		
		myprint(s);

		myprint(" near ");
		if(strlen(s) + 6 + ISTRCHAR > SCREENWIDTH)
		{
			myeol();
		}
		
		if(ISTRCHAR > SCREENWIDTH && !(GUBED & 2))
		{
			start = ISTRCHAR - (SCREENWIDTH - 5);
			start = (start >= 0) ? start : 0;
			myprint(". . .");
		}

		c = inputString[ISTRCHAR];
		inputString[ISTRCHAR] = '\0';
		mymultiprint((char *) inputString + start);
		inputString[ISTRCHAR] = c;
		SeenError = !iswarning;
	} /*if (!SeenError && (PRINTWARNINGS || !iswarning))*/
	
	if(s == OUTSTR)
	{
		*OUTSTR = '\0';
	}
} /*yyerror()*/

/* tell yylex to report  ';for()' to the parser */
static int fakefortokens[] = {(int) ';', FOR, (int) '(', (int) ')', 0 };
static void setupfakefor(void)
{
	int   i;

	for(i=0;fakefortokens[i];i++)
	{
		ALTINPUTTOKENS[i] = fakefortokens[i];
	 }
	ALTINPUTTOKENS[i] = 0;
} /*setupfakefor()*/

static void notfunction(Symbolhandle symh)
{
	if(isscratch(NAME(symh)))
	{
		yyerror("syntax error");
	}
	else
	{
		sprintf(OUTSTR,"ERROR: %s is not function or macro",NAME(symh));
		yyerror(OUTSTR);
		*OUTSTR = '\0';
	}
} /*notfunction()*/

/*
   Clean up looping related globals
*/
static void cleanLoops(long level1, long level2)
{
	int      i;
	
	for(i=level1;i<level2;i++)
	{
		WHILECONDITIONS[i] = 0;
		WHILELIMITS[i] = MAXWHILE;
		mydisphandle((char **) FORVECTORS[i]);
		FORVECTORS[i] = (double **) 0;
	}
} /*cleanLoops()*/

static Symbolhandle runCommand(Symbolhandle func, Symbolhandle list)
{
	Symbolhandle      result;
	
	strcpy(FUNCNAME,NAME(func));
	result = (*(FPTR(func)))(list);
#ifdef MACINTOSH
	UNLOADSEG(*(FPTR(func)));
#endif /*MACINTOSH*/
	Removelist(list);
	return (result);
} /*runCommand()*/

/*
   If symh is a scratch symbol, return it; otherwise return a scratch
   copy of symh
*/

static Symbolhandle parserCopy(Symbolhandle symh)
{
	Symbolhandle     symhCopy;
	long             type;

	if (symh == (Symbolhandle) 0 || isNull(symh) ||
		isscratch(NAME(symh)) || (type = TYPE(symh)) == UNDEF ||
		type == ASSIGNED || type == BLTIN || type == LIST)
	{
		return (symh);
	}
	symhCopy = Makesymbol(TYPE(symh));
	if (symhCopy != (Symbolhandle) 0 && Copy(symh, symhCopy))
	{
		setNAME(symhCopy, SCRATCH);
		Addsymbol(symhCopy);
	} /*if (symhCopy != (Symbolhandle) 0 && Copy(symh, symhCopy))*/
	else
	{
		Delete(symhCopy);
		symhCopy = (Symbolhandle) 0;
	}
	return (symhCopy);
} /*parserCopy()*/

/*
  Called from main before every invocation of getinput.
  It needs to be here because it resets PARSEMACROARGS, ALTTOKENPTR, and
  ALTINPUTTOKENS which are defined only here and in yylex.c
  It also resets static variable SeenError
*/

/* global for communication with Assign() in Symbol.c */
extern int     Nassigned;

 /* globals maintained by Addsymbol & Cutsymbol*/
extern long    Nscratch, Maxscratch;
extern long    Nsymbols, Maxsymbols;

 /* globals defined in comhandl.c */
extern long    Nhandles, Maxhandles;
extern long    CurrentMemory, MaxMemory;

/*
   globals for communication with handle manager in unxhandl.c, macHandle.c,...
*/
extern short   Nwarehouses;          /* number of warehouses (chaches)*/
extern short   WarehouseLength[];    /* number of handles in warehouses */

#ifdef COUNTMISSES
extern short   WarehouseMisses[];    /* number of failures to find handle */
extern short   WarehouseHits[];    /* number of successes to find handle */
#endif /*COUNTMISSES*/

void memoryUsage(char *NAMEFORWHERE)
{
	int       i;

	PRINT("Nscratch = %ld, Maxscratch = %ld, Nsymbols = %ld, Maxsymbols = %ld\n",
		  Nscratch, Maxscratch, Nsymbols, Maxsymbols);
	PRINT("WarehouseLength[] = %d, %d, %d, %d",
		  (int) WarehouseLength[0], (int) WarehouseLength[1],
		  (int) WarehouseLength[2], (int) WarehouseLength[3]);
	PRINTMORE(", %d, %d, %d, %d",
		  (int) WarehouseLength[4], (int) WarehouseLength[5],
		  (int) WarehouseLength[6], (int) WarehouseLength[7]);
	for (i = 8; i < Nwarehouses; i++)
	{
		PRINTMORE(", %d",(int) WarehouseLength[i],0,0,0);
	}
	PRINTMORE("\n",0,0,0,0);

#ifdef COUNTMISSES
	PRINT("WarehouseMisses[] = %d, %d, %d, %d",
		  (int) WarehouseMisses[0], (int) WarehouseMisses[1],
		  (int) WarehouseMisses[2], (int) WarehouseMisses[3]);
	PRINTMORE(", %d, %d, %d, %d",
		  (int) WarehouseMisses[4], (int) WarehouseMisses[5],
		  (int) WarehouseMisses[6], (int) WarehouseMisses[7]);
	for (i = 8; i < Nwarehouses; i++)
	{
		PRINTMORE(", %d",(int) WarehouseMisses[i],0,0,0);
	}
	PRINTMORE("\n",0,0,0,0);

	PRINT("WarehouseHits[] = %d, %d, %d, %d",
		  (int) WarehouseHits[0], (int) WarehouseHits[1],
		  (int) WarehouseHits[2], (int) WarehouseHits[3]);
	PRINTMORE(", %d, %d, %d, %d",
		  (int) WarehouseHits[4], (int) WarehouseHits[5],
		  (int) WarehouseHits[6], (int) WarehouseHits[7]);
	for (i = 8; i < Nwarehouses; i++)
	{
		PRINTMORE(", %d",(int) WarehouseHits[i],0,0,0);
	}
	PRINTMORE("\n",0,0,0,0);
#endif /*COUNTMISSES*/

	PRINT("Nhandles = %ld, Maxhandles = %ld, Nassigned = %ld\n",
		  Nhandles, Maxhandles, Nassigned, 0);
	PRINT("CurrentMemory = %ld, MaxMemory = %ld\n",
		  CurrentMemory, MaxMemory, 0, 0);
} /*void memoryUsage()*/

void cleanitup(void)
{
	WHERE("cleanitup");

	if (GUBED & 4096)
	{
		memoryUsage(NAMEFORWHERE);
	}
	
	SeenError = 0;

	Unscratch(); /* clears scratch symbol table (SCRATCH, keywords, UNDEF's)*/
	RemoveUndef(); /* clears ASSIGNED && temp variables from symbol table */
	Nassigned = 0;

	INTERRUPT = INTNOTSET;
	*OUTSTR = '\0';

	if(BDEPTH == 0)
	{
		NLINES = 0;
	}

	if(ISATTY && BDEPTH == 0 && MAXERRORS == 0)
	{
		NERRORS = 0;
	}

	restoreFormat(); /* restore current default format */

	BRACKETLEV = 0;
	PARENLEV = 0;
	MDEPTH = 0;
	WDEPTH = 0;
	IDEPTH = 0;
	MLEVEL = -1;
	PARSEMACROARGS = 0;
	ALTTOKENPTR = (int *) 0;
	ALTINPUTTOKENS[0] = 0;
	cleanLoops(0, MAXWDEPTH);
} /*cleanitup()*/
