/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment MakeFact
#else
#define __SEG__ MakeFact
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

#define IFACTOR  1
#define IMATCH   2
#define IUNIQUE  3
/*
	factor(arg)
		If arg legal for factor (all posInts) return factor equivalent
		If arg already a factor, the number of levels is recomputed from
		the actual values in arg, not taken from NCLASS(arg).
	match(arg,vals [,noMatch])
		result[i] = (arg[i] == vals[j]) ? j+1 : noMatch
		arg and vals must be both REAL or both CHAR
		vals must contain no missing values
		The default value for noMatch is length(vals) + 1
	unique(arg)
		Produces vector of unique values in arg. It is an error for arg to
		be all MISSING.  arg must be REAL or CHAR

  950823    Fixed bug with scratch LOGICAL argument to factor().
*/

Symbolhandle    makefactor(Symbolhandle list)
{
	long            i, j, nclass, n, m, op;
	long            notMatched = 0, quiet = 0;
	long            nargs = NARGS(list), margs;
	long            type, cplace = 0;
	long            chOutSize = 0, chLength;
	long            ival;
	Symbolhandle    args[3], result = (Symbolhandle) 0;
	double          val = 0.0, noMatch;
	double        **outH;
	double         *in, *out, *values;
	char          **chInH, **chOutH = (char **) 0, **chValuesH = (char **) 0;
	char           *chVal, *chVal1;
	WHERE("makefactor");

	*OUTSTR = '\0';
	if(strcmp(FUNCNAME,"factor") == 0)
	{
		op = IFACTOR;
		margs = 1;
	}
	else if(strcmp(FUNCNAME,"match") == 0)
	{
		op = IMATCH;
		margs = -3;
	}
	else
	{
		op = IUNIQUE;
		margs = 1;
	}
		
	if(margs > 0 && nargs != margs || margs < 0 && nargs > labs(margs))
	{
		badNargs(FUNCNAME,margs);
		goto errorExit;
	}
	if(op == IMATCH && nargs < 2)
	{
		badNargs(FUNCNAME,-(1000+2));
		goto errorExit;
	}
	
	for(i=0;i<nargs;i++)
	{
		args[i] = COMPVALUE(list,i);
		if(!argOK(args[i],0,(nargs > 1) ? i+1 : 0))
		{
			goto errorExit;
		}
	} /*for(i=0;i<nargs;i++)*/
	type = TYPE(args[0]);
	if(op == IFACTOR && type != REAL && type != LOGIC)
	{
		badType(FUNCNAME,-type,0);
		goto errorExit;
	}
	if(op != IFACTOR && type != REAL && type != CHAR)
	{
		badType(FUNCNAME,-type,(op == IMATCH) ? 1 : 0);
		goto errorExit;
	}
	if(op == IMATCH)
	{
		if(TYPE(args[1]) != type)
		{
			sprintf(OUTSTR,
					"ERROR: %s arguments to %s must have same type",
					(nargs == 2) ? "both" : "first two", FUNCNAME);
		}			
		else if(nargs == 3)
		{
			quiet = 1;
			if(TYPE(args[2]) != REAL || !isScalar(args[2]))
			{
				notNumberOrReal("match() argument 3");
			}
			else
			{
				noMatch = DATAVALUE(args[2],0);
			}
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
	} /*if(op == IMATCH)*/

	if(op == IFACTOR)
	{
		if(!isVector(args[0]))
		{
			sprintf(OUTSTR,"ERROR: argument to %s not a vector",FUNCNAME);
		}
		else if(!isFactor(args[0]))
		{
			sprintf(OUTSTR,
			"ERROR: argument for %s with non-integers or nonpositive integers or all MISSING",
				FUNCNAME);
		}
	}
	else if(op == IMATCH && type == REAL && anyMissing(args[1]))
	{
		sprintf(OUTSTR,
				"ERROR: argument 2 to %s contains missing values",FUNCNAME);
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}

	n = symbolSize(args[0]);
	if ((op == IFACTOR || op == IMATCH && type == REAL) &&
		isscratch(NAME(args[0])))
	{ /* reuse argument */
		result = args[0];
		if (type == LOGIC)
		{
			setTYPE(result, REAL);
		}
		COMPVALUE(list,0) = (Symbolhandle) 0;
	}
	else if(type == REAL || op != IUNIQUE)
	{
		result = RInstall(SCRATCH,n);
	}
	else
	{
		result = Install(SCRATCH,CHAR);
	}
	
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	if(type == REAL || type == LOGIC)
	{
		in = DATAPTR(args[0]);
	}
	else
	{
		chInH = STRING(args[0]);
	}

	if(TYPE(result) == REAL)
	{
		outH = DATA(result);
		out = *outH;
	}
		
	if(op == IMATCH)
	{
		m = symbolSize(args[1]);
		if(!quiet)
		{
			noMatch = (double) (m+1);
		}
		
		if(type == REAL)
		{
			values = DATAPTR(args[1]);
		}
		else
		{
			chValuesH = STRING(args[1]);
			chVal = *chInH;
		}
		
		for(i = 0; i < n;i++)
		{
			if(type == REAL)
			{
				val = in[i];
				if(!isMissing(val))
				{
					for(j=0;j<m;j++)
					{
						if(val == values[j])
						{
							break;
						}
					} /*for(j=0;j<m;j++)*/
				} /*if(!isMissing(val))*/
				else
				{
					j = -1;
				}
			} /*if(type == REAL)*/
			else
			{ /*CHAR*/
				chVal1 = *chValuesH;
				for(j=0;j<m;j++)
				{
					if(strcmp(chVal,chVal1) == 0)
					{
						break;
					}
					chVal1 = skipStrings(chVal1, 1);
				} /*for(j=0;j<m;j++)*/
				chVal = skipStrings(chVal, 1);
			} /*if(type == REAL){}else{}*/

			if(j == m)
			{ /* not matched */
				notMatched++;
				out[i] = noMatch;
			}
			else if(j >= 0)
			{ /* matched non missing*/
				out[i] = (double) (j+1);
			}
			else
			{
				setMissing(out[i]);
			}
		} /*for(i = 0; i< n;i++)*/

		if(notMatched != 0 && !quiet)
		{
			sprintf(OUTSTR,"WARNING: %ld values not matched coded as %ld",
					notMatched, m+1);
			putOUTSTR();
		}
		if (result != args[0])
		{
			setNDIMS(result,NDIMS(args[0]));
			for(i=1;i<=NDIMS(args[0]);i++)
			{
				setDIM(result,i,DIMVAL(args[0],i));
			} /*for(i=1;i<=NDIMS(args[0]);i++)*/
		}		
	} /*if(op == IMATCH)*/
	else if(op == IFACTOR)
	{ /* just copy and set NCLASS to number of categories */
		nclass = 1;
		for(i = 0;i < n; i++)
		{
			val = in[i];
			if(isMissing(val))
			{
				setMissing(out[i]);
			}
			else
			{
				if (type == LOGIC)
				{
					val = (val != 0.0) ? 2.0 : 1.0;
				}
				out[i] = val;
				ival = (long) val;
				if (ival > nclass)
				{
					nclass = ival;
				}
			}
		} /*for(i = 0;i < n; i++)*/
		setNCLASS(result,nclass);
		setNDIMS(result, 1);
		setDIM(result,1,n);
	} /*if(op == IMATCH){}else if(op == IFACTOR)*/
	else
	{ /* find unique non-missing values */
		m = 0;
		for(i=0;i<n;i++)
		{
			if(type == REAL)
			{
				val = in[i];
				if(!isMissing(val))
				{
					for(j=0;j<m;j++)
					{
						if(val == out[j])
						{
							break;
						}
					} /*for(j=0;j<m;j++)*/
					if(j == m)
					{
						out[m++] = val;
					}
				} /*if(!isMissing(val))*/
			} /*if(type == REAL)*/
			else
			{ /* type == CHAR */
				if(i == 0)
				{
					chOutSize = chLength = strlen(*chInH) + 1;
					chOutH = mygethandle(chOutSize);
					setSTRING(result,chOutH);
					if(chOutH == (char **) 0)
					{
						goto errorExit;
					}
					strcpy(*chOutH,*chInH);
					m = 1;
				} /*if(i == 0)*/
				else
				{
					chVal = *chInH + cplace;
					chVal1 = *chOutH;
					chLength = strlen(chVal) + 1;
					for(j=0;j<m;j++)
					{
						if(strcmp(chVal,chVal1) == 0)
						{
							break;
						}
						chVal1 = skipStrings(chVal1, 1);
					} /*for(j=0;j<m;j++)*/

					if(j == m)
					{/* not matched */
						chOutH = mygrowhandle(chOutH,chOutSize + chLength);
						setSTRING(result,chOutH);
						if(chOutH == (char **) 0)
						{
							goto errorExit;
						}
						strcpy(*chOutH + chOutSize, *chInH + cplace);
						chOutSize += chLength;
						m++;
					} /*if(j == m)*/
				} /*if(i == 0){}else{}*/
				cplace += chLength;
			} /*if(type == REAL){}else{}*/
		} /*for(i=0;i<n;i++)*/

		if(m == 0)
		{
			sprintf(OUTSTR,"ERROR: all values in argument to %s are MISSING",
					FUNCNAME);
			goto errorExit;
		}

		if(type == REAL && m < n)
		{ /* resize outH */
			outH = (double **) mygrowhandle((char **) outH, m * sizeof(double));
			setDATA(result,outH);
			if(outH == (double **) 0)
			{
				goto errorExit;
			}
		}
		setNDIMS(result,1);
		setDIM(result,1,m);
	}  /*if(op == IMATCH){}else if(op == IFACTOR){}else{}*/

	return (result);

  errorExit:
	putOUTSTR();
	Removesymbol(result);

	return (0);
} /*makefactor()*/

