/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Makestr
#else
#define __SEG__ Makestr
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

/* 
  makestr(arg1,...,) make a new structure from items in list
  components can be assigned names by keywords, e.g., a:cat(10,20)
  If an argument has a temporary name starting with TEMPPREFIX ('@'), it
  is stripped it off to make component name

  strconcat(arg1,...) is like makestr, except if argj is itself a structure,
  the top level components of argj are top level components of the
  result.  This is particularly useful for merging structures or for adding
  several components to a structure.

  Both accept a final keyword argument of the form compnames:CharVec
  which supplies names for the top level components.  If CharVec is
  a CHARACTER scalar, say, "var", it is taken as a root for constructing
  names "var1", "var2", ... .

  Any null arguments, are turned into REAL components of length 0.
*/


Symbolhandle    makestr(list)
Symbolhandle    list;
{
	Symbolhandle    result = (Symbolhandle) 0, arg, component, symh;
	Symbolhandle    names = (Symbolhandle) 0;
	long            strconcat = (strcmp(FUNCNAME,"strconcat") == 0);
	long            i, j, nNames = 0, maxname;
	long            nargs = NARGS(list), ncomps = 0, ncomps1;
	long            foundEmpty = 0, foundLong = 0, reuse = 0;
	char           *keyword, *nameptr;
	char            name[NAMELENGTH+1], namestart[NAMELENGTH];
	WHERE("makestr");
	
	*OUTSTR = '\0';
	
	if(nargs > 1 && (keyword = isKeyword(COMPVALUE(list,nargs-1))) &&
	   strncmp(keyword,"compname",8) == 0)
	{
		names = COMPVALUE(list,nargs-1);
		nargs--;
		if(TYPE(names) != CHAR || !isVector(names))
		{
			sprintf(OUTSTR,
					"ERROR: keyword %s for %s must be a CHARACTER vector or scalar",
					keyword, FUNCNAME);
			goto errorExit;
		}
		nNames = symbolSize(names);
	}
	for (i = 0; i < nargs; i++)
	{
		arg = COMPVALUE(list, i);
		if (!isDefined(arg))
		{
			/* 
			   Replace null arguments in list by REAL symbols of length 0
			   If they are UNDEF or ASSIGNED and have a non-temporary name
			   make it a temporary name
			*/
			foundEmpty++;
			ncomps++;
			COMPVALUE(list,i) = symh = Makereal(0);
			if(symh == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			if(arg != (Symbolhandle) 0)
			{
				nameptr = NAME(arg);
				if(!istempname(nameptr))
				{
					for(j=NAMELENGTH-1;j >= 0;j--)
					{
						nameptr[j] = nameptr[j-1];
					}
					nameptr[0] = TEMPPREFIX;
					nameptr[NAMELENGTH] = '\0';
				}
			}
			arg = symh;
			Addsymbol(arg); /* make sure elements of list are in symbol table*/
		}
		else if(strconcat)
		{ /* count the number of output components for strconcat() */
			ncomps += ((isStruc(arg)) ? NCOMPS(arg) : 1);
		}
		else
		{
			ncomps++;
		}
	} /*for (i = 0; i < nargs; i++)*/
	if(foundEmpty > 0)
	{
		myerrorout("WARNING: empty component(s) in structure");
	}
	
	if(nNames > 1 && nNames != ncomps)
	{
		sprintf(OUTSTR,
				"ERROR: incorrect number of component names for %s", FUNCNAME);
		goto errorExit;
	}

	if(nNames == 1 && ncomps > 1)
	{
		strncpy(namestart,STRINGPTR(names),NAMELENGTH);
		if(namestart[0] == '\0')
		{
			strcpy(namestart,"component");
		}
		maxname = NAMELENGTH - (long)(log10(ncomps+.5)) - 1;
		namestart[maxname] = '\0';
	}
	else
	{
		namestart[0] = '\0';
	}
	
	result = StrucInstall(SCRATCH,ncomps);

	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	if(!strconcat)
	{ /* makestr() */
		for(i=0;i<nargs;i++)
		{
			arg = COMPVALUE(list,i);
			if(!isKeyword(arg) && !isscratch(NAME(arg)))
			{ /* make copy */
				symh = COMPVALUE(result,i) = Makesymbol(REAL);
				if(symh == (Symbolhandle) 0)
				{
					goto errorExit;
				}

				if(Copy(arg,symh) == 0)
				{
					goto errorExit;
				}
			}
			else
			{ /* keyword or scratch: move symbol to result */
				symh = arg;
				COMPVALUE(list,i) = (Symbolhandle) 0;
				Cutsymbol(symh);
				COMPVALUE(result,i) = symh;
			}
			setCompName(symh,NAME(symh));
		}
	} /* if(!strconcat) */
	else
	{ /* strconcat() */
		ncomps = 0;
		for(i=0;i<nargs;i++)
		{
			arg = COMPVALUE(list,i);
			reuse = (isscratch(NAME(arg)) || isKeyword(arg));
			if(isStruc(arg))
			{
				ncomps1 = NCOMPS(arg);
				for(j=0;j<ncomps1;j++)
				{
					component = COMPVALUE(arg,j);
					if(reuse)
					{
						symh = component;
						COMPVALUE(arg,j) = (Symbolhandle) 0;
						if(symh == (Symbolhandle) 0)
						{ /* make sure symh is not null handle */
							symh = Makereal(0);
							if(symh == (Symbolhandle) 0)
							{
								goto errorExit;
							}
						}
						COMPVALUE(result,ncomps++) = symh;
					}
					else
					{
						symh = COMPVALUE(result,ncomps++) = Makesymbol(REAL);
						if(symh == (Symbolhandle) 0)
						{
							goto errorExit;
						}
						if(component != (Symbolhandle) 0 &&
						   !Copy(component,symh))
						{
							goto errorExit;
						}
					}					
					setCompName(symh,NAME(symh));
				} /*for(j=0;j<ncomps1;j++)*/
			} /* if(isStruc(arg)) */
			else
			{
				if(reuse)
				{
					symh = arg;
					COMPVALUE(list,i) = (Symbolhandle) 0;
					Cutsymbol(symh);
					COMPVALUE(result,ncomps++) = symh;
				}
				else
				{
					symh = COMPVALUE(result,ncomps++) = Makesymbol(REAL);
					if(symh == (Symbolhandle) 0)
					{
						goto errorExit;
					}
					arg = COMPVALUE(list,i);
					if(!Copy(arg,symh))
					{
						goto errorExit;
					}
				}
				setCompName(symh,NAME(symh));
			} /*if(isStruc(arg)) {...} else {...} */
		}/* for(i=0;i<nargs;i++)*/
	} /* if(!strconcat){...}else{...} */	
	
	if(names != (Symbolhandle) 0)
	{
		nameptr = STRINGPTR(names);
		for(i=0;i<ncomps;i++)
		{
			if(namestart[0] != '\0')
			{
				sprintf(name,"%s%ld",namestart,i+1);
			}
			else
			{
				if(strlen(nameptr) > NAMELENGTH)
				{
					foundLong++;
				}
				strncpy(name,nameptr,NAMELENGTH);
				name[NAMELENGTH] = '\0';
				nameptr = skipStrings(nameptr,1);
			}
			
			if(name[0] != '\0')
			{
				setCompName(COMPVALUE(result,i),name);
			}
		}
		if(foundLong)
		{
			sprintf(OUTSTR,
					"WARNING: name%s longer than %ld characters truncated by %s",
					(foundLong == 1) ? "" : "s", (long) NAMELENGTH, FUNCNAME);
			putOUTSTR();
		}		
	}
	
	return (result);

  errorExit:
	putOUTSTR();
	Removesymbol(result);
	
	return (0);
	
}

Symbolhandle compnames(list)
Symbolhandle      list;
{
	Symbolhandle        result = (Symbolhandle) 0, arg;
	char              **names = (char **) 0;
	long                i, needed = 0, ncomps, cplace = 0;
	
	if(NARGS(list) > 1)
	{
		badNargs(FUNCNAME,1);
		goto errorExit;
	}
	if(!argOK((arg = COMPVALUE(list,0)),STRUC,0))
	{
		goto errorExit;
	}
	ncomps = NCOMPS(arg);

	for(i=0;i<ncomps;i++)
	{
		needed += strlen(NAME(COMPVALUE(arg,i))) + 1;
	}
	result = Install(SCRATCH,CHAR);
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	names = mygethandle(needed);
	setSTRING(result,names);
	if(names == (char **) 0)
	{
		goto errorExit;
	}

	for(i=0;i<ncomps;i++)
	{
		strcpy(*names + cplace, NAME(COMPVALUE(arg,i)));
		cplace += strlen (*names + cplace) + 1;
	}
	setNDIMS(result,1);
	setDIM(result,1,ncomps);

	return (result);
	
  errorExit:
	Removesymbol(result);
	
	return (0);
}

