/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Utils
#else
#define __SEG__ Utils
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "plot.h"
#include <ctype.h>

/*
   contains substitutes for various functions that may not be available
   on all systems.
*/

/*NOSTRTOD defined implies no working strtod in standard library*/
#ifdef NOSTRTOD

/*
  version of strtod not using sscanf

  Note:  This is non-conforming since it always expects '.' as "radix
  character", regardless of locale

  C. Bingham, 911012
*/

#define MAXPOW 20
static double powers[MAXPOW+1] =
{
	1e0,1e1,1e2,1e3,1e4,1e5,1e6,1e7,1e8,1e9,1e10,
	1e11,1e12,1e13,1e14,1e15,1e16,1e17,1e18,1e19,1e20
};

double mystrtod(nptr, eptr)
char          *nptr;
char         **eptr;
{
	long       ssign = 1, esign = 1, decPlaces = 0;
	double     mantissa = 0.0, expon = 0.0;
	char      *endptr;
	long       i;
	short      c;

	if(eptr != (char **) NULL)
	{
		*eptr = nptr;
	}
	if(*nptr == '\0')
	{
		return (0.0);
	}
	/* expected pattern is [-+]?[0-9]*\.?[0-9]*\([eE][ -+]?[0-9][0-9]*\)? */

	i = 0;
	while( isspace( nptr[i] ))
	{
		i++;					/* skip leading nonnumeric */
	}

	c = nptr[i];
	if(c == '+' || c == '-')
	{
		if(c == '-')
		{
			ssign = -1;
		}
		c = nptr[++i];
	}

	if(isdigit(c))
	{							/* decode integral part of mantissa */
		do
		{
			mantissa = 10.0*mantissa + (c - '0');
			c = nptr[++i];
		} while(isdigit(c));
	}
	else if(c != '.' || !isdigit(nptr[i+1]))
	{							/* not a number */
		return (0.0);
	}

	if(c == '.')
	{							/* decode fractional part of mantissa */
		c = nptr[++i];
		if(isdigit(c))
		{
			do
			{
				mantissa = 10.*mantissa + (c - '0');
				decPlaces++;
				c = nptr[++i];
			}while(isdigit(c));
		}
	}
	endptr = nptr + i;
	i++;
	if(ssign < 0)
	{
		mantissa = -mantissa;
	}
	if(c == 'e' || c == 'E')
	{ /* decode exponent */
		if((c = nptr[i]) == '-' || c == '+' || c == ' ')
		{
			if(c == '-')
			{
				esign = -1;
			}
			c = nptr[++i];
		}
		if(isdigit(c))
		{
			do
			{ /* decode values of exponent */
				expon = 10.0*expon + (c - '0');
				c = nptr[++i];
			} while(isdigit(c));
			endptr = nptr + i;
			if(esign < 0)
			{
				expon = -expon;
			}
		}
	}
	expon -= (double) decPlaces;
	if(fabs(expon) > (double) MAXPOW)
	{
		mantissa *= pow(10.,expon);
	}
	else
	{
		if(expon >= 0.0)
		{
			mantissa *= powers[(long) expon];
		}
		else
		{
			mantissa /= powers[-(long) expon];
		}
	}
	if(eptr != (char **) NULL)
	{
		*eptr = endptr;
	}
	return (mantissa);
}
#else /*NOSTRTOD*/
/* system library has strtod */
double mystrtod(nptr, eptr)
char          *nptr;
char         **eptr;
{
	return (strtod(nptr, eptr));
}

#endif /*NOSTRTOD*/

#ifdef NOHYPOT
#define MAX(A,B) (((A) > (B)) ? (A) : (B))
#define MIN(A,B) (((A) < (B)) ? (A) : (B))

/*      finds dsqrt(a**2+b**2) without overflow or destructive underflow */
double pythag(a,b)
double a,b;
{
	double p,r,s,t,u;
	double temp;

	p = MAX(fabs(a),fabs(b));
	if (p!=0.0)
	{
		r = MIN(fabs(a),fabs(b))/p;
		r *= r;
		while(1)
		{
			t = 4.0+r;
			if (t==4.0)
				break ;
			s = r/t;
			u = 1.0+2.0*s;
			p = u*p;
			temp = s/u;
			r = temp*temp*r;
		}
	}
	return p;
}
#else /*NOHYPOT*/
#ifndef MPW /*following is inconsistant with MPW*/
extern double hypot(double, double);
#endif /*MPW*/

double pythag(a,b)
double a,b;
{
	return hypot(a,b);
}
#endif /*NOHYPOT*/

#ifndef USEPOW
/*
  To be used when system pow() is inefficient with integer exponents as in MPW
*/

double intpow(register double x, double p)
{
	register double         y;
	register long           iexp = labs((long) p);
	
	y = (iexp & 1) ? x : 1.0;
	
	for(iexp >>= 1 ;iexp != 0;iexp >>= 1)
	{
		x *= x;
		if(iexp & 1)
		{
			y *= x;
		}
	}
	
	return ((p < 0.0) ? 1.0/y : y);
}
#else /*USEPOW*/
double intpow(double x, double p)
{
	double       y = pow(fabs(x), p);

	return ((x < 0 && ((long) p) % 2 != 0) ? -y : y);
}
#endif /*USEPOW*/

#ifdef NOLGAMMA
#ifdef MPW
#ifdef MPW3
#pragma segment Trans
#else
#define __SEG__ Trans
#endif /*MPW3*/
#endif /*MPW*/

static double   ln2pio2 = 0.9189385332046727417803297;

static double   clocal[] =
{
	-0.577191652,
	0.988205891,
	-0.897056937,
	0.918206857,
	-0.756704078,
	0.482199394,
	-0.193527818,
	0.035868343,
};
static double   casymp[] =
{
	12.,
	-360.,
	1260.,
	-1680.,
};


double          lgamma(double x)
{
	long            i;
	double          out, z, z2, value;

	if (x > 8.)
	{			/* asymptotic, 6.1.41 of Abramowitz and Stegun */
		z = 1 / x;
		z2 = z * z;
		out = 0.0;
		for (i = 0; i < 4; i++)
		{
			out += z / casymp[i];
			z *= z2;
		}
		return ((x - .5) * log(x) - x + ln2pio2 + out);
	}

	/* small x, 6.1.36 of Abramowitz and Stegun */
	value = 0.0;
	while (x < 1.0)
	{
		value -= log(x++);
	}
	while(x > 2.0)
	{
		value += log(--x);
	}
	
	z = x - 1.;
	z2 = z;
	out = 1.;
	for (i = 0; i < 8; i++)
	{
		out += clocal[i] * z2;
		z2 *= z;
	}
	return (log(out) + value);
}

#endif /*NOLGAMMA*/

