/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Matrix
#else
#define __SEG__ Matrix
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <math.h>

extern long            Copy();
extern long            symbolSize();
extern Symbolhandle    Install();
extern void            Removesymbol();
/*
  usage: matrix(x, dims)
  routine to take a set of data an rewrite subscripts into matrix form
  first argument is data, second argument is number of rows
  data are always entered down columns first
*/

Symbolhandle    matrix(list)
Symbolhandle    list;
{
	Symbolhandle    result = (Symbolhandle) 0, arg1, arg2;
	double          value;
	long            nrows, tot, type;
	long            i;
	char           *usage = "usage: matrix(data,nrows)";
	WHERE("matrix");
	
	*OUTSTR = '\0';
	if (NARGS(list) != 2)
	{
		badNargs(FUNCNAME,2);
		goto errorExit;
	}

	arg1 = COMPVALUE(list, 0);
	arg2 = COMPVALUE(list, 1);
	if (!argOK(arg1,(long) 0, 1)|| !argOK(arg2,REAL,2))
	{
		goto errorExit;
	}

	type = TYPE(arg1);
	if(type != REAL && type != LOGIC && type != CHAR)
	{
		badType(FUNCNAME,-type,1);
		goto errorExit;
	}
	
	tot = symbolSize(arg1);

	if (! isScalar(arg2) || 
		(value = DATAVALUE(arg2,0)) <= 0 || value != floor(value))
	{
		sprintf(OUTSTR,"ERROR: nrows must be single positive integer");
	}
	else if (tot % (nrows = (long) value) != 0)
	{
		sprintf(OUTSTR,"ERROR: nrows must divide length of data");
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}
	usage = (char *) 0;
	
	if(isscratch(NAME(arg1)))
	{ /* reuse symbol, removing it from list */
		COMPVALUE(list,0) = (Symbolhandle) 0;
		result = arg1;
	}
	else
	{
		result = Install(SCRATCH,TYPE(arg1));
		if (result == (Symbolhandle) 0 || Copy(arg1, result) == 0)
		{
			goto errorExit;
		}
	}
	setNAME(result,SCRATCH);

	for (i=3;i<=NDIMS(result);i++)
	{
		setDIM(result, i, 0);
	}
	setNCLASS(result,-1);

	setNDIMS(result,2);
	setDIM(result,1,nrows);
	setDIM(result,2,tot / nrows);

	return (result);

  errorExit:
	putOUTSTR();
	if(usage)
	{
		myerrorout(usage);
	}

	Removesymbol(result);
	
	return (0);
}
