
/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Glm
#else /*MPW3*/
#define __SEG__ Glm
#endif /*MPW3*/
#endif /*MPW*/

/*
  model parser written in yacc
  Version of 940224
  yacc message
	conflicts: 7 shift/reduce
  930426 Changed yylval type to Symbolhandle, allowing retrieval of space
  in case of interrupt.
  940224 Added checks to ensure that an error term is not deleted and that an
  error term is not constant

  950510 Changed '1 << var' to '1L << var' in MakeTerm() to make fix in
  Borland DOS version for which int means short.
*/

#include <stdio.h>

#include "globals.h"
#include "dbug.h"
#include <ctype.h>
#undef BADTOKEN

#define YYSTYPE Symbolhandle
#define YYDEBUG

typedef YYSTYPE      Model;

/*
  As of 930517, all the code depends on the representation of a model only
  through the following macros.  If a new representation is chosen
  only these will need to be changed
*/

#define RemoveModel(M)       Removesymbol(M)

#define GetModel(A) (modelHandle) LONGDATA((Symbolhandle) A)
#define SetModel(A,B) setLONGDATA((Symbolhandle) A,(long **) (B))

/* in following, A is a modelHandle, B is a modelPointer, I and L are longs */
/* in mreader, term numbers run from 1 to nterms */
#define GetLength(A) DIMVAL((Symbolhandle) A,2)
#define SetLength(A,L) setDIM((Symbolhandle) A,2,L)
#define TermI(A,I) modelterm(*(A),(I) - 1)  /* value is modelPointer */
#define SetTermI(A,I,B) modeltermAssign(B,(*(A) + (I) - 1)) /*(*A)[I-1] = B*/
#define ModelSize(L) ((L) * sizeof(modelType))
#define SetModelDims(A,L)	{\
								setNDIMS((Symbolhandle) A,2);\
								setDIM((Symbolhandle) A,1,WORDSPERTERM);\
								SetLength(A,L);\
							}

/* in following, A and B are modelPointers's & I is a variable number */
#define EqualTerms(A,B) modeltermEqual(A,B)
#define ContainTerm(A,B) inEffectiveTerm(A,B)
#define SetErrorTerm(A) modeltermAssign(A,(modelPointer) Info->errorterms[Info->nerrorterms-1])
#define GetErrorTerm(I) ((modelPointer) Info->errorterms[(I)-1])

#define yylex yylex_mreader
#define yyerror yyerror_mreader

#define   TAB       '\t'
#define   NL        10
#define   CR        13
#define   SP        ' '

#define Info        MreaderModelInfoPointer

modelInfoPointer    Info;

static char       **Strmodel; /* handle of model being scanned */
static long         Imodchar; /* position in the model */
#ifdef __cplusplus
#  include <stdio.h>
#  include <yacc.h>
#endif	/* __cplusplus */ 
# define MODELVAR 257
# define MODELEND 258
# define BADTOKEN 259
#define yyclearin yychar = -1
#define yyerrok yyerrflag = 0
static int yychar;
#ifndef YYMAXDEPTH
#define YYMAXDEPTH 150
#endif

/* __YYSCLASS defines the scoping/storage class for global objects
 * that are NOT renamed by the -p option.  By default these names
 * are going to be 'static' so that multi-definition errors
 * will not occur with multiple parsers.
 * If you want (unsupported) access to internal names you need
 * to define this to be null so it implies 'extern' scope.
 * This should not be used in conjunction with -p.
 */
#ifndef __YYSCLASS
# define __YYSCLASS static
#endif
#ifndef YYSTYPE
#define YYSTYPE int
#endif
static YYSTYPE yylval;
__YYSCLASS YYSTYPE yyval;
typedef int yytabelem;
# define YYERRCODE 256


/* start of private functions */
static void yyerror(char *s)
{
	/*  output error message */
	char            c;
	WHERE("yyerror");

	if(strcmp(s,"syntax error") == 0)
	{
		myprint("ERROR: Model syntax error");
	}
	else
	{
		myprint(s);
	}

	myprint(" near ");
	c = (*Strmodel)[Imodchar];
	(*Strmodel)[Imodchar] = '\0';
	mymultiprint(*Strmodel);
	(*Strmodel)[Imodchar] = c;
}

static Model InstallModel(long nterms)
{
	Model             modelH = (Model) Install(SCRATCH,LONG);
	modelHandle       model;

	if(modelH != (Model) 0)
	{ /* will be structured as WORDSPERTERM by nterms matrix */
		model = (modelHandle) mygethandle(ModelSize(nterms));
		SetModel(modelH,model);
		if(model != (modelHandle) 0)
		{
			SetModelDims(modelH,nterms);
		}
		else
		{
			RemoveModel(modelH);
			modelH = (Model) 0;
		}
	}
	return (modelH);
}

static modelPointer MakeTerm(long var)
{
	static modelType     term;
	long                 i;

	zeroTerm(term);

	if(var != 0)
	{
		i = (var-1) / VARSPERWORD;
		var = (var - 1) % VARSPERWORD;
		term[i] = (termWord) (1L << var);
	}
	return ((modelPointer) term);
}

static modelPointer   OrTerms(modelPointer termA, modelPointer termB)
{
	static modelType     termC;
	long                 i;

	for(i=0;i<WORDSPERTERM;i++)
	{
		termC[i] = (termWord) ((*termA)[i] | (*termB)[i]);
	}
	return ((modelPointer) termC);
}

#ifdef UNDEFINED__  /* currently not used */
static modelPointer   AndTerms(modelPointer termA, modelPointer termB)
{
	static modelType     termC;
	long                 i;

	for(i=0;i<WORDSPERTERM;i++)
	{
		termC[i] = (termWord) ((*termA)[i] & (*termB)[i]);
	}
	return ((modelPointer) termC);
}
#endif /*UNDEFINED__*/

static Model unique(Model modela)
{
	/*  remove repeated model terms from list */
	modelHandle    a = GetModel(modela);
	long           alength = GetLength(modela);
	long           i,j,k;
	WHERE("unique");

	for(i=1;i< alength; i++)
	{
		for(j=i+1;j<=alength;j++)
		{
			if(EqualTerms(TermI(a,i),TermI(a,j)))
			{ /* a repeat */
				for(k=j+1;k<=alength;k++)
				{
					SetTermI(a,k-1,TermI(a,k));
				}
				alength--;
			}
		}
	}
 	SetLength(modela,alength);

	return (modela);
}

/*
  Routine to join (add) two models together
  preserve left precedence
*/

static Model mjoin(Model modela, Model modelb)
{
	modelHandle    a = GetModel(modela), b = GetModel(modelb), c;
	long           alength = GetLength(modela), blength = GetLength(modelb);
	long           clength = alength + blength;
	Model          modelc = InstallModel(clength);
	long           i, j, k;
	WHERE("mjoin");

	if(modelc != (Model) 0)
	{
		c = GetModel(modelc);
		k = 1;
		for(i = 1; i <= alength; i++)
		{
			SetTermI(c,k,TermI(a,i));
			k++;
		}
		for(j = 1; j <= blength; j++)
		{
			SetTermI(c,k,TermI(b,j));
			k++;
		}
	}
	RemoveModel(modela);
	RemoveModel(modelb);

	return(modelc);
}

/*
  Routine to nest two models together
  c = a/b
 */

static Model  mnest(Model modela, Model modelb)
{
	modelHandle     a = GetModel(modela), b = GetModel(modelb), c;
	long            alength = GetLength(modela), blength = GetLength(modelb);
	long            clength = alength + blength;
	Model           modelc = InstallModel(clength);
	modelType       merge;
	modelPointer    pMerge = (modelPointer) merge;
	long            i, j, k;

	if(modelc != (Model) 0)
	{
		c = GetModel(modelc);
		zeroTerm(merge);
		k = 1;
		for(i = 1; i <= alength; i++)
		{
			SetTermI(c,k,TermI(a,i));
			k++;
			modeltermAssign(OrTerms(TermI(a,i), pMerge), pMerge);
		}
		for(j = 1; j <= blength; j++)
		{
			SetTermI(c,k,OrTerms(TermI(b,j),pMerge));
			k++;
		}
	}
	RemoveModel(modela);
	RemoveModel(modelb);

	return(modelc);
}

/*
  c = a.b
*/
static Model mdot(Model modela, Model modelb)
{
	/*  routine to dot a model with a model */
	modelHandle  a = GetModel(modela), b = GetModel(modelb), c;
	long         alength = GetLength(modela), blength = GetLength(modelb);
	long         clength = alength * blength;
	Model        modelc = InstallModel(clength);
	long         i, j, k;

	if(modelc != (Model) 0)
	{
		c = GetModel(modelc);
		k = 1;
		for(i = 1; i <= blength; i++)
		{
			for(j = 1; j <= alength; j++)
			{
				SetTermI(c,k,OrTerms(TermI(a,j),TermI(b,i)));
				k++;
			}
		}
	}
	RemoveModel(modela);
	RemoveModel(modelb);

	return(modelc);
}

/*  c = a*b = a + b + a.b   a la GLIM  */

static Model mstar(Model modela, Model modelb)
{
	modelHandle   a = GetModel(modela), b = GetModel(modelb), c;
	long          alength = GetLength(modela), blength = GetLength(modelb);
	long          clength = alength + blength + alength * blength;
	Model         modelc = InstallModel(clength);
	long          i, j, k;

	if(modelc != (Model) 0)
	{
		c = GetModel(modelc);
		k = 1;
		for(i = 1; i <= alength; i++)
		{
			SetTermI(c,k, TermI(a,i));
			k++;
		}
		for(i = 1; i <= blength; i++)
		{
			SetTermI(c,k, TermI(b,i));
			k++;
		}
		for(i = 1; i <= blength; i++)
		{
			for(j = 1; j <= alength; j++)
			{
				SetTermI(c,k, OrTerms(TermI(a,j),TermI(b,i)));
				k++;
			}
		}
	}
	RemoveModel(modela);
	RemoveModel(modelb);

	return(modelc);
}

/* remove from a terms containing terms in b: c = a -* b*/

static Model mmstar(Model modela, Model modelb)
{
	modelHandle    a = GetModel(modela), b = GetModel(modelb), c = a;
	Model          modelc = modela;
	modelPointer   termbj, termci;
	long           blength = GetLength(modelb), clength = GetLength(modelc);
	long           i, j, k, deleted = 0;

	for(j=1; j <= blength; j++)
	{
		termbj = TermI(b,j);
		for(i = 1; i <= clength; i++)
		{
			while(i <= clength && ContainTerm(termci = TermI(c,i),termbj))
			{ /* drop out i-th term of c */
				for(k=1;k<Info->nerrorterms;k++)
				{ /* check to see whether it is error term */
					if(EqualTerms(GetErrorTerm(k), termci))
					{
						yyerror("ERROR: attempt to remove error term from model");
						return ( (Model) 0);
					}
				} /*for(k=1;k<Info->nerrorterms;k++)*/

				for( k = i; k < clength ; k++)
				{ /* squeeze term out of model */
					SetTermI(c,k,TermI(c,k+1));
					deleted++;
				} /*for( k = i; k < clength ; k++)*/
				clength--;
			}/*while(i <= clength && ContainTerm(termci = TermI(c,i),termbj))*/
		} /*for(i = 1; i <= clength; i++)*/
	} /*for(j=1; j <= blength; j++)*/
	SetLength(modelc,clength);
	RemoveModel(modelb);
	if(deleted == 0)
	{
		yyerror("ERROR: attempt to subtract term(s) not in the model");
		return ((Model) 0);
	}

	return(modelc);
}

/* remove terms in b from a: c = a - b */
static Model mminus(Model modela, Model modelb)
{
	modelHandle  a = GetModel(modela), b = GetModel(modelb), c = a;
	Model        modelc = modela;
	modelPointer termbj;
	long         blength = GetLength(modelb), clength = GetLength(modelc);
	long         i, j, k, deleted = 0;

	for(j=1; j <= blength; j++)
	{
		termbj = TermI(b,j);
		for(k=1;k<Info->nerrorterms;k++)
		{
			if(EqualTerms(GetErrorTerm(k), termbj))
			{
				yyerror("ERROR: attempt to remove error term from model");
				return ( (Model) 0);
			}
		}
		
		for(i = 1; i <= clength; i++)
		{
			while( EqualTerms(TermI(c,i),termbj)  && i <= clength )
			{ /* squeeze out term i from c */
				for( k = i; k < clength ; k++)
				{
					SetTermI(c,k,TermI(c,k+1));
				}
				deleted++;
				clength--;
			}
		} /*for(i = 1; i <= clength; i++)*/
	} /*for(j=1; j <= blength; j++)*/
	if(deleted == 0)
	{
		yyerror("ERROR: attempt to subtract term(s) not in the model");
		return ((Model) 0);
	}
	SetLength(modelc,clength);
	RemoveModel(modelb);

	return(modelc);
} /*mminus()*/

static Model madder(Model modela)
{
	/* add term to list of error terms */
	modelHandle   a = GetModel(modela);
	modelPointer  term = TermI(a,1);
	long          k;
	
	if(EqualTerms(term, (modelPointer) NULLTERM))
	{ /* E(1) is illegal */
		yyerror("ERROR: error term is CONSTANT");
		return ( (Model) 0);
	}

	for(k=1;k<Info->nerrorterms;k++)
	{ /* make sure term isn't already an error term */
		if(EqualTerms(term, GetErrorTerm(k)))
		{
			yyerror("ERROR: duplicate error term in model");
			return ( (Model) 0);
		}
	}
	
	SetErrorTerm(TermI(a,1));
	Info->nerrorterms++;

	return (modela);
}

static long setyylval(long var)
{
	if((yylval = InstallModel(1)) == (Model) 0)
	{
		return (BADTOKEN);
	}
	SetTermI(GetModel(yylval),1,MakeTerm(var));

	return (MODELVAR);
}

/* Makefile must change references to yylex generated by yacc to modelyylex */
static int modelyylex(void)
{  /* lexical analyzer for model parser */

	char          *inputString = *Strmodel;
	char           token[NAMELENGTH+1];
	char          *place;
	long           c;
	long           ivar, i;
	Symbolhandle   symh;
	WHERE("modelyylex");

	for(;;)
	{
		c = inputString[Imodchar++];
		if(c == '\\')
		{
			c = inputString[Imodchar++];
		}
		switch(c)
		{

		  case CR:  /* treat end of lines as white space */
		  case NL:
		  case SP:
		  case TAB:
			break;				/* skip white space */

		  case '+':
		  case '-':
		  case '/':
		  case '*':
		  case '.':
		  case '=':
		  case '(':
		  case ')':
			return(c);
			/* simply pass through model symbols */

		  case '\0':
		  case '#':
			return(MODELEND);			/* end of string */

			/* treat 1 specially for constant */
		  case '1':
			if(!isnamechar(inputString[Imodchar]) )
			{ /* set yylval to single zero term & return MODELVAR */
				return (setyylval(0));
			}

		  default:				/* get next variable name in token */

			/* treat "E\W*(" specially; it begins an error term */
			if(c  == 'E')
			{
				for(place = inputString + Imodchar;*place;place++)
				{
					if(*place != ' ' && *place != '\t' &&
						*place != CR && *place != NL)
					{
						break;
					}
				}
				if(*place == '(')
				{
					Imodchar = place - inputString;
					return((int) 'E');
				}
			}

			if(isnamestart(c) || c == TEMPPREFIX)
			{
				token[0] = c;
				for(i=1;i<NAMELENGTH;i++)
				{
					c = inputString[Imodchar];
					if(!isnamechar(c))
					{
						break;
					}
					token[i] = c;
					Imodchar++;
				}
			}
			else
			{
				/* we have illegal character */
				yyerror("ERROR: illegal character in model");
				return(BADTOKEN);
			}
			token[i] = '\0';

			if(isnamechar(inputString[Imodchar]))
			{
				/* variable name too long */
				Imodchar++;
				yyerror("ERROR: variable name in model too long");
				return(BADTOKEN);
			}

			symh = Lookup(token);

			if(symh == (Symbolhandle)  0)
			{
				/* unrecognized variable */
				yyerror("ERROR: unrecognized variable in model");
				return(BADTOKEN);
			}

			if(Info->nvars == -1)
			{	/* new Y variable */
				Info->modelvars[0] = symh;
				Info->nvars = 0;
				return (setyylval(0));
			}

			/* check to see if it is already in model */
			for(ivar=0;ivar <= Info->nvars;ivar++)
			{
				if(strcmp(token,NAME(Info->modelvars[ivar])) == 0)
				{
					break;
				}
			}

			if(ivar == 0)
			{		/* Y variable appearing in model terms */
				yyerror("ERROR: dependent varable on right hand side of model");
				return(BADTOKEN);
			}

			if(ivar > Info->nvars)
			{ /* must be new variable */
				Info->nvars++;

				if(Info->nvars > MAXVARS)
				{	/* too many variables */
					sprintf(OUTSTR,
							"ERROR: more than %ld variables in model",
							(long) MAXVARS);
					putOUTSTR();
					return(BADTOKEN);
				}
				if(NCLASS(symh) > 0)
				{
					Info->nfactors++;
				}
				if(Info->nfactors > MAXDIMS)
				{	/* too many factors */
					sprintf(OUTSTR,
							"ERROR: more than %ld factors in model",
							(long) MAXDIMS);
					putOUTSTR();
					return(BADTOKEN);
				}

				Info->modelvars[ivar] = symh;
				if(ivar > 0)
				{
					Info->nclasses[ivar-1] = NCLASS(symh);
				}
			} /*if(ivar > Info->nvars)*/

			return (setyylval(ivar));
		} /*switch(c)*/
	} /* for(;;) */
}

/* start of externally accessible (public) functions */

void initmodelparse(char **model)
{
	int      i;

	Info->nvars = -1;
	Info->nfactors = 0;
	Info->nerrorterms = 1;
	Imodchar = 0;
	for(i=0;i<=MAXVARS;i++)
	{
		Info->modelvars[i] = (Symbolhandle) 0;
	}
	for(i=0;i<=MAXERRTRMS;i++)
	{
		zeroTerm(Info->errorterms[i]);
	}
	Strmodel = model;
	mydisphandle((char **) Info->model);
	Info->model = (modelHandle) 0;
}

modelInfoPointer getModelInfo(void)
{
	return (Info);
}



__YYSCLASS yytabelem yyexca[] ={
-1, 1,
	0, -1,
	-2, 0,
	};
# define YYNPROD 20
# define YYLAST 240
__YYSCLASS yytabelem yyact[]={

    27,    15,    13,     8,    17,    14,    16,    15,    13,    32,
    17,    14,    16,     8,    15,    26,    15,     5,    14,    16,
    14,     4,     2,    29,     5,     5,    14,     6,    11,     3,
    34,    20,     9,    28,     1,     0,    18,     0,     0,     0,
     0,     0,     9,     0,     0,     0,     0,     0,    18,    30,
    18,    18,    18,    18,    18,     0,     7,    33,     0,     0,
    18,     0,     0,    30,    33,    19,     0,     0,    35,     0,
    21,    22,    23,    24,    25,     0,     0,     0,     0,     0,
     0,     0,     0,    31,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     5,     0,
    10,     0,     5,    12,     5,     0,     0,     5,     0,     0,
    10,     5,     5,     5,     0,     0,     0,     0,     0,     5 };
__YYSCLASS yytabelem yypact[]={

  -235, -3000,   -34, -3000, -3000, -3000,   -37,   -35,   -37,    -9,
 -3000, -3000, -3000,   -37,   -37,   -37,   -37,   -27, -3000,   -41,
  -234,   -28,  -242,   -20,   -26,   -28,   -37, -3000,   -32,   -16,
 -3000,   -20, -3000, -3000,  -234,  -242 };
__YYSCLASS yytabelem yypgo[]={

     0,    34,    56,    28,    33 };
__YYSCLASS yytabelem yyr1[]={

     0,     1,     1,     1,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     4,     4,     4,     4,     3 };
__YYSCLASS yytabelem yyr2[]={

     0,     9,     3,     3,     7,     7,     7,     7,     7,     7,
     9,     9,     3,     5,     3,     3,     7,     5,     3,     3 };
__YYSCLASS yytabelem yychk[]={

 -3000,    -1,   257,    -3,   256,   259,    61,    -2,    40,    69,
   257,    -3,   258,    43,    46,    42,    47,    45,    -3,    -2,
    40,    -2,    -2,    -2,    -2,    -2,    42,    41,    -4,   257,
    -3,    -2,    41,    -3,    46,    -4 };
__YYSCLASS yytabelem yydef[]={

     0,    -2,     0,     2,     3,    19,     0,     0,     0,     0,
    12,    14,     1,     0,     0,     0,     0,     0,    13,     0,
     0,     4,     5,     6,     7,     9,     0,     8,     0,    15,
    18,    10,    11,    17,     0,    16 };
typedef struct { char *t_name; int t_val; } yytoktype;
#ifndef YYDEBUG
#	define YYDEBUG	0	/* don't allow debugging */
#endif

#ifdef YYDEBUG

__YYSCLASS yytoktype yytoks[] =
{
	"MODELVAR",	257,
	"MODELEND",	258,
	"BADTOKEN",	259,
	"+",	43,
	"-",	45,
	"/",	47,
	"*",	42,
	".",	46,
	"-unknown-",	-1	/* ends search */
};

__YYSCLASS char * yyreds[] =
{
	"-no such reduction-",
	"assgn : MODELVAR '=' model MODELEND",
	"assgn : lexerror",
	"assgn : error",
	"model : model '+' model",
	"model : model '.' model",
	"model : model '*' model",
	"model : model '/' model",
	"model : '(' model ')'",
	"model : model '-' model",
	"model : model '-' '*' model",
	"model : 'E' '(' term ')'",
	"model : MODELVAR",
	"model : model lexerror",
	"model : lexerror",
	"term : MODELVAR",
	"term : MODELVAR '.' term",
	"term : term lexerror",
	"term : lexerror",
	"lexerror : BADTOKEN",
};
#endif /* YYDEBUG */
#define YYFLAG  (-3000)
/* @(#) $Revision: 70.7 $ */    

/*
** Skeleton parser driver for yacc output
*/

#if defined(NLS) && !defined(NL_SETN)
#include <msgbuf.h>
#endif

#ifndef nl_msg
#define nl_msg(i,s) (s)
#endif

/*
** yacc user known macros and defines
*/
#define YYERROR		goto yyerrlab

#ifndef __RUNTIME_YYMAXDEPTH
#define YYACCEPT	return(0)
#define YYABORT		return(1)
#else
#define YYACCEPT	{free_stacks(); return(0);}
#define YYABORT		{free_stacks(); return(1);}
#endif

#define YYBACKUP( newtoken, newvalue )\
{\
	if ( yychar >= 0 || ( yyr2[ yytmp ] >> 1 ) != 1 )\
	{\
		yyerror( (nl_msg(30001,"syntax error - cannot backup")) );\
		goto yyerrlab;\
	}\
	yychar = newtoken;\
	yystate = *yyps;\
	yylval = newvalue;\
	goto yynewstate;\
}
#define YYRECOVERING()	(!!yyerrflag)
#ifndef YYDEBUG
#	define YYDEBUG	1	/* make debugging available */
#endif

/*
** user known globals
*/
static int yydebug = 0;			/* set to 1 to get debugging */

/*
** driver internal defines
*/
/* define for YYFLAG now generated by yacc program. */
/*#define YYFLAG		(FLAGVAL)*/

/*
** global variables used by the parser
*/
# ifndef __RUNTIME_YYMAXDEPTH
__YYSCLASS YYSTYPE yyv[ YYMAXDEPTH ];	/* value stack */
__YYSCLASS int yys[ YYMAXDEPTH ];		/* state stack */
# else
__YYSCLASS YYSTYPE *yyv;			/* pointer to malloc'ed value stack */
__YYSCLASS int *yys;			/* pointer to malloc'ed stack stack */

#if defined(__STDC__) || defined (__cplusplus)
#include <stdlib.h>
#else
	extern char *malloc();
	extern char *realloc();
	extern void free();
#endif /* __STDC__ or __cplusplus */


static int allocate_stacks(); 
static void free_stacks();
# ifndef YYINCREMENT
# define YYINCREMENT (YYMAXDEPTH/2) + 10
# endif
# endif	/* __RUNTIME_YYMAXDEPTH */
static long  yymaxdepth = YYMAXDEPTH;

__YYSCLASS YYSTYPE *yypv;			/* top of value stack */
__YYSCLASS int *yyps;			/* top of state stack */

__YYSCLASS int yystate;			/* current state */
__YYSCLASS int yytmp;			/* extra var (lasts between blocks) */

static int yynerrs;			/* number of errors */
__YYSCLASS int yyerrflag;			/* error recovery flag */
static int yychar;			/* current input token number */



/*
** modelparse - return 0 if worked, 1 if syntax error not recovered from
*/
int
modelparse()
{
	register YYSTYPE *yypvt;	/* top of value stack for $vars */
	WHERE("modelparse");

	/*
	** Initialize externals - modelparse may be called more than once
	*/
# ifdef __RUNTIME_YYMAXDEPTH
	if (allocate_stacks()) YYABORT;
# endif
	yypv = &yyv[-1];
	yyps = &yys[-1];
	yystate = 0;
	yytmp = 0;
	yynerrs = 0;
	yyerrflag = 0;
	yychar = -1;

	goto yystack;
	{
		register YYSTYPE *yy_pv;	/* top of value stack */
		register int *yy_ps;		/* top of state stack */
		register int yy_state;		/* current state */
		register int  yy_n;		/* internal state number info */

		/*
		** get globals into registers.
		** branch to here only if YYBACKUP was called.
		*/
	yynewstate:
		yy_pv = yypv;
		yy_ps = yyps;
		yy_state = yystate;
		goto yy_newstate;

		/*
		** get globals into registers.
		** either we just started, or we just finished a reduction
		*/
	yystack:
		yy_pv = yypv;
		yy_ps = yyps;
		yy_state = yystate;

		/*
		** top of for (;;) loop while no reductions done
		*/
	yy_stack:
		/*
		** put a state and value onto the stacks
		*/
#ifdef YYDEBUG
		/*
		** if debugging, look up token value in list of value vs.
		** name pairs.  0 and negative (-1) are special values.
		** Note: linear search is used since time is not a real
		** consideration while debugging.
		*/
		if ( yydebug )
		{
			register int yy_i;

			printf( "State %d, token ", yy_state );
			if ( yychar == 0 )
				printf( "end-of-file\n" );
			else if ( yychar < 0 )
				printf( "-none-\n" );
			else
			{
				for ( yy_i = 0; yytoks[yy_i].t_val >= 0;
					yy_i++ )
				{
					if ( yytoks[yy_i].t_val == yychar )
						break;
				}
				printf( "%s\n", yytoks[yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ++yy_ps >= &yys[ yymaxdepth ] )	/* room on stack? */
		{
# ifndef __RUNTIME_YYMAXDEPTH
			yyerror( (nl_msg(30002,"yacc stack overflow")) );
			YYABORT;
# else
			/* save old stack bases to recalculate pointers */
			YYSTYPE * yyv_old = yyv;
			int * yys_old = yys;
			yymaxdepth += YYINCREMENT;
			yys = (int *) realloc(yys, yymaxdepth * sizeof(int));
			yyv = (YYSTYPE *) realloc(yyv, yymaxdepth * sizeof(YYSTYPE));
			if (yys==0 || yyv==0) {
			    yyerror( (nl_msg(30002,"yacc stack overflow")) );
			    YYABORT;
			    }
			/* Reset pointers into stack */
			yy_ps = (yy_ps - yys_old) + yys;
			yyps = (yyps - yys_old) + yys;
			yy_pv = (yy_pv - yyv_old) + yyv;
			yypv = (yypv - yyv_old) + yyv;
# endif

		}
		*yy_ps = yy_state;
		*++yy_pv = yyval;

		/*
		** we have a new state - find out what to do
		*/
	yy_newstate:
		if ( ( yy_n = yypact[ yy_state ] ) <= YYFLAG )
			goto yydefault;		/* simple state */
#ifdef YYDEBUG
		/*
		** if debugging, need to mark whether new token grabbed
		*/
		yytmp = yychar < 0;
#endif
		if ( ( yychar < 0 ) && ( ( yychar = modelyylex() ) < 0 ) )
			yychar = 0;		/* reached EOF */
#ifdef YYDEBUG
		if ( yydebug && yytmp )
		{
			register int yy_i;

			printf( "Received token " );
			if ( yychar == 0 )
				printf( "end-of-file\n" );
			else if ( yychar < 0 )
				printf( "-none-\n" );
			else
			{
				for ( yy_i = 0; yytoks[yy_i].t_val >= 0;
					yy_i++ )
				{
					if ( yytoks[yy_i].t_val == yychar )
						break;
				}
				printf( "%s\n", yytoks[yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ( ( yy_n += yychar ) < 0 ) || ( yy_n >= YYLAST ) )
			goto yydefault;
		if ( yychk[ yy_n = yyact[ yy_n ] ] == yychar )	/*valid shift*/
		{
			yychar = -1;
			yyval = yylval;
			yy_state = yy_n;
			if ( yyerrflag > 0 )
				yyerrflag--;
			goto yy_stack;
		}

	yydefault:
		if ( ( yy_n = yydef[ yy_state ] ) == -2 )
		{
#ifdef YYDEBUG
			yytmp = yychar < 0;
#endif
			if ( ( yychar < 0 ) && ( ( yychar = modelyylex() ) < 0 ) )
				yychar = 0;		/* reached EOF */
#ifdef YYDEBUG
			if ( yydebug && yytmp )
			{
				register int yy_i;

				printf( "Received token " );
				if ( yychar == 0 )
					printf( "end-of-file\n" );
				else if ( yychar < 0 )
					printf( "-none-\n" );
				else
				{
					for ( yy_i = 0;
						yytoks[yy_i].t_val >= 0;
						yy_i++ )
					{
						if ( yytoks[yy_i].t_val
							== yychar )
						{
							break;
						}
					}
					printf( "%s\n", yytoks[yy_i].t_name );
				}
			}
#endif /* YYDEBUG */
			/*
			** look through exception table
			*/
			{
				register int *yyxi = yyexca;

				while ( ( *yyxi != -1 ) ||
					( yyxi[1] != yy_state ) )
				{
					yyxi += 2;
				}
				while ( ( *(yyxi += 2) >= 0 ) &&
					( *yyxi != yychar ) )
					;
				if ( ( yy_n = yyxi[1] ) < 0 )
					YYACCEPT;
			}
		}

		/*
		** check for syntax error
		*/
		if ( yy_n == 0 )	/* have an error */
		{
			/* no worry about speed here! */
			switch ( yyerrflag )
			{
			case 0:		/* new error */
				yyerror( (nl_msg(30003,"syntax error")) );
				yynerrs++;
				goto skip_init;
			yyerrlab:
				/*
				** get globals into registers.
				** we have a user generated syntax type error
				*/
				yy_pv = yypv;
				yy_ps = yyps;
				yy_state = yystate;
				yynerrs++;
			skip_init:
			case 1:
			case 2:		/* incompletely recovered error */
					/* try again... */
				yyerrflag = 3;
				/*
				** find state where "error" is a legal
				** shift action
				*/
				while ( yy_ps >= yys )
				{
					yy_n = yypact[ *yy_ps ] + YYERRCODE;
					if ( yy_n >= 0 && yy_n < YYLAST &&
						yychk[yyact[yy_n]] == YYERRCODE)					{
						/*
						** simulate shift of "error"
						*/
						yy_state = yyact[ yy_n ];
						goto yy_stack;
					}
					/*
					** current state has no shift on
					** "error", pop stack
					*/
#ifdef YYDEBUG
#	define _POP_ "Error recovery pops state %d, uncovers state %d\n"
					if ( yydebug )
						printf( _POP_, *yy_ps,
							yy_ps[-1] );
#	undef _POP_
#endif
					yy_ps--;
					yy_pv--;
				}
				/*
				** there is no state on stack with "error" as
				** a valid shift.  give up.
				*/
				YYABORT;
			case 3:		/* no shift yet; eat a token */
#ifdef YYDEBUG
				/*
				** if debugging, look up token in list of
				** pairs.  0 and negative shouldn't occur,
				** but since timing doesn't matter when
				** debugging, it doesn't hurt to leave the
				** tests here.
				*/
				if ( yydebug )
				{
					register int yy_i;

					printf( "Error recovery discards " );
					if ( yychar == 0 )
						printf( "token end-of-file\n" );
					else if ( yychar < 0 )
						printf( "token -none-\n" );
					else
					{
						for ( yy_i = 0;
							yytoks[yy_i].t_val >= 0;
							yy_i++ )
						{
							if ( yytoks[yy_i].t_val
								== yychar )
							{
								break;
							}
						}
						printf( "token %s\n",
							yytoks[yy_i].t_name );
					}
				}
#endif /* YYDEBUG */
				if ( yychar == 0 )	/* reached EOF. quit */
					YYABORT;
				yychar = -1;
				goto yy_newstate;
			}
		}/* end if ( yy_n == 0 ) */
		/*
		** reduction by production yy_n
		** put stack tops, etc. so things right after switch
		*/
#ifdef YYDEBUG
		/*
		** if debugging, print the string that is the user's
		** specification of the reduction which is just about
		** to be done.
		*/
		if ( yydebug )
			printf( "Reduce by (%d) \"%s\"\n",
				yy_n, yyreds[ yy_n ] );
#endif
		yytmp = yy_n;			/* value to switch over */
		yypvt = yy_pv;			/* $vars top of value stack */
		/*
		** Look in goto table for next state
		** Sorry about using yy_state here as temporary
		** register variable, but why not, if it works...
		** If yyr2[ yy_n ] doesn't have the low order bit
		** set, then there is no action to be done for
		** this reduction.  So, no saving & unsaving of
		** registers done.  The only difference between the
		** code just after the if and the body of the if is
		** the goto yy_stack in the body.  This way the test
		** can be made before the choice of what to do is needed.
		*/
		{
			/* length of production doubled with extra bit */
			register int yy_len = yyr2[ yy_n ];

			if ( !( yy_len & 01 ) )
			{
				yy_len >>= 1;
				yyval = ( yy_pv -= yy_len )[1];	/* $$ = $1 */
				yy_state = yypgo[ yy_n = yyr1[ yy_n ] ] +
					*( yy_ps -= yy_len ) + 1;
				if ( yy_state >= YYLAST ||
					yychk[ yy_state =
					yyact[ yy_state ] ] != -yy_n )
				{
					yy_state = yyact[ yypgo[ yy_n ] ];
				}
				goto yy_stack;
			}
			yy_len >>= 1;
			yyval = ( yy_pv -= yy_len )[1];	/* $$ = $1 */
			yy_state = yypgo[ yy_n = yyr1[ yy_n ] ] +
				*( yy_ps -= yy_len ) + 1;
			if ( yy_state >= YYLAST ||
				yychk[ yy_state = yyact[ yy_state ] ] != -yy_n )
			{
				yy_state = yyact[ yypgo[ yy_n ] ];
			}
		}
					/* save until reenter driver code */
		yystate = yy_state;
		yyps = yy_ps;
		yypv = yy_pv;
	}
	/*
	** code supplied by user is placed in this switch
	*/
	switch( yytmp )
	{
		
case 1:
{
										yyval = unique(yypvt[-1]);
										Info->nterms = GetLength(yyval);
										Info->model = GetModel(yyval);
										SetModel(yyval,0);
										RemoveModel(yyval);
										/* shrink to correct size */
										Info->model = (modelHandle)
											mygrowhandle((char **) Info->model,
														 ModelSize(Info->nterms));
										return(Info->model == (modelHandle) 0);
									} break;
case 2:
{ return(1); } break;
case 3:
{ return(1); } break;
case 4:
{
								yyval = mjoin(yypvt[-2],yypvt[-0]);
							  checkit:
								if(yyval == (Model) 0) return(1);
							} break;
case 5:
{ yyval = mdot(yypvt[-2],yypvt[-0]); goto checkit; } break;
case 6:
{ yyval = mstar(yypvt[-2],yypvt[-0]); goto checkit; } break;
case 7:
{ yyval = mnest(yypvt[-2],yypvt[-0]); goto checkit; } break;
case 8:
{ yyval = yypvt[-1]; goto checkit; } break;
case 9:
{ yyval = mminus(yypvt[-2],yypvt[-0]); goto checkit; } break;
case 10:
{ yyval = mmstar(yypvt[-3],yypvt[-0]); goto checkit; } break;
case 11:
{ yyval = madder(yypvt[-1]); goto checkit; } break;
case 12:
{ yyval = yypvt[-0] ; goto checkit; } break;
case 13:
{ return (1); } break;
case 14:
{ return (1); } break;
case 15:
{ yyval = yypvt[-0]; goto checkit; } break;
case 16:
{ yyval = mdot(yypvt[-2],yypvt[-0]); goto checkit; } break;
case 17:
{ return(1); } break;
case 18:
{ return(1); } break;
case 19:
{ return(1); } break;
	}
	goto yystack;		/* reset registers in driver code */
}

# ifdef __RUNTIME_YYMAXDEPTH

static int allocate_stacks() {
	/* allocate the yys and yyv stacks */
	yys = (int *) malloc(yymaxdepth * sizeof(int));
	yyv = (YYSTYPE *) malloc(yymaxdepth * sizeof(YYSTYPE));

	if (yys==0 || yyv==0) {
	   yyerror( (nl_msg(30004,"unable to allocate space for yacc stacks")) );
	   return(1);
	   }
	else return(0);

}


static void free_stacks() {
	if (yys!=0) free((char *) yys);
	if (yyv!=0) free((char *) yyv);
}

# endif  /* defined(__RUNTIME_YYMAXDEPTH) */

