%{
/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Glm
#else /*MPW3*/
#define __SEG__ Glm
#endif /*MPW3*/
#endif /*MPW*/

/*
  model parser written in yacc
  Version of 940224
  yacc message
	conflicts: 7 shift/reduce
  930426 Changed yylval type to Symbolhandle, allowing retrieval of space
  in case of interrupt.
  940224 Added checks to ensure that an error term is not deleted and that an
  error term is not constant

  950510 Changed '1 << var' to '1L << var' in MakeTerm() to make fix in
  Borland DOS version for which int means short.
*/

#include <stdio.h>

#include "globals.h"
#include "dbug.h"
#include <ctype.h>
#undef BADTOKEN

#define YYSTYPE Symbolhandle
#define YYDEBUG

typedef YYSTYPE      Model;

/*
  As of 930517, all the code depends on the representation of a model only
  through the following macros.  If a new representation is chosen
  only these will need to be changed
*/

#define RemoveModel(M)       Removesymbol(M)

#define GetModel(A) (modelHandle) LONGDATA((Symbolhandle) A)
#define SetModel(A,B) setLONGDATA((Symbolhandle) A,(long **) (B))

/* in following, A is a modelHandle, B is a modelPointer, I and L are longs */
/* in mreader, term numbers run from 1 to nterms */
#define GetLength(A) DIMVAL((Symbolhandle) A,2)
#define SetLength(A,L) setDIM((Symbolhandle) A,2,L)
#define TermI(A,I) modelterm(*(A),(I) - 1)  /* value is modelPointer */
#define SetTermI(A,I,B) modeltermAssign(B,(*(A) + (I) - 1)) /*(*A)[I-1] = B*/
#define ModelSize(L) ((L) * sizeof(modelType))
#define SetModelDims(A,L)	{\
								setNDIMS((Symbolhandle) A,2);\
								setDIM((Symbolhandle) A,1,WORDSPERTERM);\
								SetLength(A,L);\
							}

/* in following, A and B are modelPointers's & I is a variable number */
#define EqualTerms(A,B) modeltermEqual(A,B)
#define ContainTerm(A,B) inEffectiveTerm(A,B)
#define SetErrorTerm(A) modeltermAssign(A,(modelPointer) Info->errorterms[Info->nerrorterms-1])
#define GetErrorTerm(I) ((modelPointer) Info->errorterms[(I)-1])

#define yylex yylex_mreader
#define yyerror yyerror_mreader

#define   TAB       '\t'
#define   NL        10
#define   CR        13
#define   SP        ' '

#define Info        MreaderModelInfoPointer

modelInfoPointer    Info;

static char       **Strmodel; /* handle of model being scanned */
static long         Imodchar; /* position in the model */
%}

%token MODELVAR MODELEND BADTOKEN
%left '+' '-'
%left '/'
%left '*'
%left '.'

%%
assgn: 	MODELVAR '=' model MODELEND	{
										$$ = unique($3);
										Info->nterms = GetLength($$);
										Info->model = GetModel($$);
										SetModel($$,0);
										RemoveModel($$);
										/* shrink to correct size */
										Info->model = (modelHandle)
											mygrowhandle((char **) Info->model,
														 ModelSize(Info->nterms));
										return(Info->model == (modelHandle) 0);
									}
	| lexerror				{ return(1); }
	| error					{ return(1); }
	;

model: 	  model '+' model	{
								$$ = mjoin($1,$3);
							  checkit:
								if($$ == (Model) 0) return(1);
							}
	| model '.' model		{ $$ = mdot($1,$3); goto checkit; }
	| model '*' model		{ $$ = mstar($1,$3); goto checkit; }
	| model '/' model		{ $$ = mnest($1,$3); goto checkit; }
	| '(' model ')'			{ $$ = $2; goto checkit; }
	| model '-' model		{ $$ = mminus($1,$3); goto checkit; }
	| model '-' '*' model	{ $$ = mmstar($1,$4); goto checkit; }
	| 'E' '(' term ')'		{ $$ = madder($3); goto checkit; }
	| MODELVAR				{ $$ = $1 ; goto checkit; }
	| model lexerror		{ return (1); }
	| lexerror				{ return (1); }
	;


term:	  MODELVAR			{ $$ = $1; goto checkit; }
	| MODELVAR '.' term		{ $$ = mdot($1,$3); goto checkit; }
	| term lexerror			{ return(1); }
	| lexerror				{ return(1); }
	;

lexerror: BADTOKEN  		{ return(1); }
	;
%%
/* start of private functions */
static void yyerror(char *s)
{
	/*  output error message */
	char            c;
	WHERE("yyerror");

	if(strcmp(s,"syntax error") == 0)
	{
		myprint("ERROR: Model syntax error");
	}
	else
	{
		myprint(s);
	}

	myprint(" near ");
	c = (*Strmodel)[Imodchar];
	(*Strmodel)[Imodchar] = '\0';
	mymultiprint(*Strmodel);
	(*Strmodel)[Imodchar] = c;
}

static Model InstallModel(long nterms)
{
	Model             modelH = (Model) Install(SCRATCH,LONG);
	modelHandle       model;

	if(modelH != (Model) 0)
	{ /* will be structured as WORDSPERTERM by nterms matrix */
		model = (modelHandle) mygethandle(ModelSize(nterms));
		SetModel(modelH,model);
		if(model != (modelHandle) 0)
		{
			SetModelDims(modelH,nterms);
		}
		else
		{
			RemoveModel(modelH);
			modelH = (Model) 0;
		}
	}
	return (modelH);
}

static modelPointer MakeTerm(long var)
{
	static modelType     term;
	long                 i;

	zeroTerm(term);

	if(var != 0)
	{
		i = (var-1) / VARSPERWORD;
		var = (var - 1) % VARSPERWORD;
		term[i] = (termWord) (1L << var);
	}
	return ((modelPointer) term);
}

static modelPointer   OrTerms(modelPointer termA, modelPointer termB)
{
	static modelType     termC;
	long                 i;

	for(i=0;i<WORDSPERTERM;i++)
	{
		termC[i] = (termWord) ((*termA)[i] | (*termB)[i]);
	}
	return ((modelPointer) termC);
}

#ifdef UNDEFINED__  /* currently not used */
static modelPointer   AndTerms(modelPointer termA, modelPointer termB)
{
	static modelType     termC;
	long                 i;

	for(i=0;i<WORDSPERTERM;i++)
	{
		termC[i] = (termWord) ((*termA)[i] & (*termB)[i]);
	}
	return ((modelPointer) termC);
}
#endif /*UNDEFINED__*/

static Model unique(Model modela)
{
	/*  remove repeated model terms from list */
	modelHandle    a = GetModel(modela);
	long           alength = GetLength(modela);
	long           i,j,k;
	WHERE("unique");

	for(i=1;i< alength; i++)
	{
		for(j=i+1;j<=alength;j++)
		{
			if(EqualTerms(TermI(a,i),TermI(a,j)))
			{ /* a repeat */
				for(k=j+1;k<=alength;k++)
				{
					SetTermI(a,k-1,TermI(a,k));
				}
				alength--;
			}
		}
	}
 	SetLength(modela,alength);

	return (modela);
}

/*
  Routine to join (add) two models together
  preserve left precedence
*/

static Model mjoin(Model modela, Model modelb)
{
	modelHandle    a = GetModel(modela), b = GetModel(modelb), c;
	long           alength = GetLength(modela), blength = GetLength(modelb);
	long           clength = alength + blength;
	Model          modelc = InstallModel(clength);
	long           i, j, k;
	WHERE("mjoin");

	if(modelc != (Model) 0)
	{
		c = GetModel(modelc);
		k = 1;
		for(i = 1; i <= alength; i++)
		{
			SetTermI(c,k,TermI(a,i));
			k++;
		}
		for(j = 1; j <= blength; j++)
		{
			SetTermI(c,k,TermI(b,j));
			k++;
		}
	}
	RemoveModel(modela);
	RemoveModel(modelb);

	return(modelc);
}

/*
  Routine to nest two models together
  c = a/b
 */

static Model  mnest(Model modela, Model modelb)
{
	modelHandle     a = GetModel(modela), b = GetModel(modelb), c;
	long            alength = GetLength(modela), blength = GetLength(modelb);
	long            clength = alength + blength;
	Model           modelc = InstallModel(clength);
	modelType       merge;
	modelPointer    pMerge = (modelPointer) merge;
	long            i, j, k;

	if(modelc != (Model) 0)
	{
		c = GetModel(modelc);
		zeroTerm(merge);
		k = 1;
		for(i = 1; i <= alength; i++)
		{
			SetTermI(c,k,TermI(a,i));
			k++;
			modeltermAssign(OrTerms(TermI(a,i), pMerge), pMerge);
		}
		for(j = 1; j <= blength; j++)
		{
			SetTermI(c,k,OrTerms(TermI(b,j),pMerge));
			k++;
		}
	}
	RemoveModel(modela);
	RemoveModel(modelb);

	return(modelc);
}

/*
  c = a.b
*/
static Model mdot(Model modela, Model modelb)
{
	/*  routine to dot a model with a model */
	modelHandle  a = GetModel(modela), b = GetModel(modelb), c;
	long         alength = GetLength(modela), blength = GetLength(modelb);
	long         clength = alength * blength;
	Model        modelc = InstallModel(clength);
	long         i, j, k;

	if(modelc != (Model) 0)
	{
		c = GetModel(modelc);
		k = 1;
		for(i = 1; i <= blength; i++)
		{
			for(j = 1; j <= alength; j++)
			{
				SetTermI(c,k,OrTerms(TermI(a,j),TermI(b,i)));
				k++;
			}
		}
	}
	RemoveModel(modela);
	RemoveModel(modelb);

	return(modelc);
}

/*  c = a*b = a + b + a.b   a la GLIM  */

static Model mstar(Model modela, Model modelb)
{
	modelHandle   a = GetModel(modela), b = GetModel(modelb), c;
	long          alength = GetLength(modela), blength = GetLength(modelb);
	long          clength = alength + blength + alength * blength;
	Model         modelc = InstallModel(clength);
	long          i, j, k;

	if(modelc != (Model) 0)
	{
		c = GetModel(modelc);
		k = 1;
		for(i = 1; i <= alength; i++)
		{
			SetTermI(c,k, TermI(a,i));
			k++;
		}
		for(i = 1; i <= blength; i++)
		{
			SetTermI(c,k, TermI(b,i));
			k++;
		}
		for(i = 1; i <= blength; i++)
		{
			for(j = 1; j <= alength; j++)
			{
				SetTermI(c,k, OrTerms(TermI(a,j),TermI(b,i)));
				k++;
			}
		}
	}
	RemoveModel(modela);
	RemoveModel(modelb);

	return(modelc);
}

/* remove from a terms containing terms in b: c = a -* b*/

static Model mmstar(Model modela, Model modelb)
{
	modelHandle    a = GetModel(modela), b = GetModel(modelb), c = a;
	Model          modelc = modela;
	modelPointer   termbj, termci;
	long           blength = GetLength(modelb), clength = GetLength(modelc);
	long           i, j, k, deleted = 0;

	for(j=1; j <= blength; j++)
	{
		termbj = TermI(b,j);
		for(i = 1; i <= clength; i++)
		{
			while(i <= clength && ContainTerm(termci = TermI(c,i),termbj))
			{ /* drop out i-th term of c */
				for(k=1;k<Info->nerrorterms;k++)
				{ /* check to see whether it is error term */
					if(EqualTerms(GetErrorTerm(k), termci))
					{
						yyerror("ERROR: attempt to remove error term from model");
						return ( (Model) 0);
					}
				} /*for(k=1;k<Info->nerrorterms;k++)*/

				for( k = i; k < clength ; k++)
				{ /* squeeze term out of model */
					SetTermI(c,k,TermI(c,k+1));
					deleted++;
				} /*for( k = i; k < clength ; k++)*/
				clength--;
			}/*while(i <= clength && ContainTerm(termci = TermI(c,i),termbj))*/
		} /*for(i = 1; i <= clength; i++)*/
	} /*for(j=1; j <= blength; j++)*/
	SetLength(modelc,clength);
	RemoveModel(modelb);
	if(deleted == 0)
	{
		yyerror("ERROR: attempt to subtract term(s) not in the model");
		return ((Model) 0);
	}

	return(modelc);
}

/* remove terms in b from a: c = a - b */
static Model mminus(Model modela, Model modelb)
{
	modelHandle  a = GetModel(modela), b = GetModel(modelb), c = a;
	Model        modelc = modela;
	modelPointer termbj;
	long         blength = GetLength(modelb), clength = GetLength(modelc);
	long         i, j, k, deleted = 0;

	for(j=1; j <= blength; j++)
	{
		termbj = TermI(b,j);
		for(k=1;k<Info->nerrorterms;k++)
		{
			if(EqualTerms(GetErrorTerm(k), termbj))
			{
				yyerror("ERROR: attempt to remove error term from model");
				return ( (Model) 0);
			}
		}
		
		for(i = 1; i <= clength; i++)
		{
			while( EqualTerms(TermI(c,i),termbj)  && i <= clength )
			{ /* squeeze out term i from c */
				for( k = i; k < clength ; k++)
				{
					SetTermI(c,k,TermI(c,k+1));
				}
				deleted++;
				clength--;
			}
		} /*for(i = 1; i <= clength; i++)*/
	} /*for(j=1; j <= blength; j++)*/
	if(deleted == 0)
	{
		yyerror("ERROR: attempt to subtract term(s) not in the model");
		return ((Model) 0);
	}
	SetLength(modelc,clength);
	RemoveModel(modelb);

	return(modelc);
} /*mminus()*/

static Model madder(Model modela)
{
	/* add term to list of error terms */
	modelHandle   a = GetModel(modela);
	modelPointer  term = TermI(a,1);
	long          k;
	
	if(EqualTerms(term, (modelPointer) NULLTERM))
	{ /* E(1) is illegal */
		yyerror("ERROR: error term is CONSTANT");
		return ( (Model) 0);
	}

	for(k=1;k<Info->nerrorterms;k++)
	{ /* make sure term isn't already an error term */
		if(EqualTerms(term, GetErrorTerm(k)))
		{
			yyerror("ERROR: duplicate error term in model");
			return ( (Model) 0);
		}
	}
	
	SetErrorTerm(TermI(a,1));
	Info->nerrorterms++;

	return (modela);
}

static long setyylval(long var)
{
	if((yylval = InstallModel(1)) == (Model) 0)
	{
		return (BADTOKEN);
	}
	SetTermI(GetModel(yylval),1,MakeTerm(var));

	return (MODELVAR);
}

/* Makefile must change references to yylex generated by yacc to modelyylex */
static int modelyylex(void)
{  /* lexical analyzer for model parser */

	char          *inputString = *Strmodel;
	char           token[NAMELENGTH+1];
	char          *place;
	long           c;
	long           ivar, i;
	Symbolhandle   symh;
	WHERE("modelyylex");

	for(;;)
	{
		c = inputString[Imodchar++];
		if(c == '\\')
		{
			c = inputString[Imodchar++];
		}
		switch(c)
		{

		  case CR:  /* treat end of lines as white space */
		  case NL:
		  case SP:
		  case TAB:
			break;				/* skip white space */

		  case '+':
		  case '-':
		  case '/':
		  case '*':
		  case '.':
		  case '=':
		  case '(':
		  case ')':
			return(c);
			/* simply pass through model symbols */

		  case '\0':
		  case '#':
			return(MODELEND);			/* end of string */

			/* treat 1 specially for constant */
		  case '1':
			if(!isnamechar(inputString[Imodchar]) )
			{ /* set yylval to single zero term & return MODELVAR */
				return (setyylval(0));
			}

		  default:				/* get next variable name in token */

			/* treat "E\W*(" specially; it begins an error term */
			if(c  == 'E')
			{
				for(place = inputString + Imodchar;*place;place++)
				{
					if(*place != ' ' && *place != '\t' &&
						*place != CR && *place != NL)
					{
						break;
					}
				}
				if(*place == '(')
				{
					Imodchar = place - inputString;
					return((int) 'E');
				}
			}

			if(isnamestart(c) || c == TEMPPREFIX)
			{
				token[0] = c;
				for(i=1;i<NAMELENGTH;i++)
				{
					c = inputString[Imodchar];
					if(!isnamechar(c))
					{
						break;
					}
					token[i] = c;
					Imodchar++;
				}
			}
			else
			{
				/* we have illegal character */
				yyerror("ERROR: illegal character in model");
				return(BADTOKEN);
			}
			token[i] = '\0';

			if(isnamechar(inputString[Imodchar]))
			{
				/* variable name too long */
				Imodchar++;
				yyerror("ERROR: variable name in model too long");
				return(BADTOKEN);
			}

			symh = Lookup(token);

			if(symh == (Symbolhandle)  0)
			{
				/* unrecognized variable */
				yyerror("ERROR: unrecognized variable in model");
				return(BADTOKEN);
			}

			if(Info->nvars == -1)
			{	/* new Y variable */
				Info->modelvars[0] = symh;
				Info->nvars = 0;
				return (setyylval(0));
			}

			/* check to see if it is already in model */
			for(ivar=0;ivar <= Info->nvars;ivar++)
			{
				if(strcmp(token,NAME(Info->modelvars[ivar])) == 0)
				{
					break;
				}
			}

			if(ivar == 0)
			{		/* Y variable appearing in model terms */
				yyerror("ERROR: dependent varable on right hand side of model");
				return(BADTOKEN);
			}

			if(ivar > Info->nvars)
			{ /* must be new variable */
				Info->nvars++;

				if(Info->nvars > MAXVARS)
				{	/* too many variables */
					sprintf(OUTSTR,
							"ERROR: more than %ld variables in model",
							(long) MAXVARS);
					putOUTSTR();
					return(BADTOKEN);
				}
				if(NCLASS(symh) > 0)
				{
					Info->nfactors++;
				}
				if(Info->nfactors > MAXDIMS)
				{	/* too many factors */
					sprintf(OUTSTR,
							"ERROR: more than %ld factors in model",
							(long) MAXDIMS);
					putOUTSTR();
					return(BADTOKEN);
				}

				Info->modelvars[ivar] = symh;
				if(ivar > 0)
				{
					Info->nclasses[ivar-1] = NCLASS(symh);
				}
			} /*if(ivar > Info->nvars)*/

			return (setyylval(ivar));
		} /*switch(c)*/
	} /* for(;;) */
}

/* start of externally accessible (public) functions */

void initmodelparse(char **model)
{
	int      i;

	Info->nvars = -1;
	Info->nfactors = 0;
	Info->nerrorterms = 1;
	Imodchar = 0;
	for(i=0;i<=MAXVARS;i++)
	{
		Info->modelvars[i] = (Symbolhandle) 0;
	}
	for(i=0;i<=MAXERRTRMS;i++)
	{
		zeroTerm(Info->errorterms[i]);
	}
	Strmodel = model;
	mydisphandle((char **) Info->model);
	Info->model = (modelHandle) 0;
}

modelInfoPointer getModelInfo(void)
{
	return (Info);
}



