/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Myplot
#else
#define __SEG__ Myplot
#endif /*MPW3*/
#endif /*MPW*/

#include <math.h>
#include "globals.h"
#ifdef MACINTOSH
#include "macIface.h" /* defines SFReply Reply as global*/
#endif /*MACINTOSH*/
#include "plot.h"
#include "dbug.h"

#ifdef MPW
#include <Memory.h>
#endif

#define TRUE 1
#define FALSE 0

#ifndef LASTGRAPH
#define LASTGRAPH "LASTPLOT"
#endif /*LASTGRAPH*/

char               *LastGraph = LASTGRAPH; /* external variable */

extern Symbolhandle Install();
extern long         symbolSize();
extern char       **mygethandle();
extern void         mydisphandle();
extern long         isfilename();
extern long         isVector(), isScalar();
extern char        *isKeyword();

#define NEEDXY      0x01
#define PTPLOT      0x02
#define CHARPLOT    0x04
#define LINEPLOT    0x08
#define STRINGPLOT  0x10
#define TOFILE      0x20
#define SHOW        0x40
#define ADDTOPLOT   0x80

#define IPLOT       PTPLOT    | NEEDXY
#define ICHPLOT     CHARPLOT  | NEEDXY
#define ILINEPLOT   LINEPLOT  | NEEDXY
#define IADDPOINTS  IPLOT     | ADDTOPLOT
#define IADDCHARS   ICHPLOT   | ADDTOPLOT
#define IADDLINES   ILINEPLOT | ADDTOPLOT
#define IADDSTRINGS STRINGPLOT| ADDTOPLOT | NEEDXY
#define ISHOWPLOT   SHOW
#define IFPLOT      IPLOT     | TOFILE
#define IFCHPLOT    ICHPLOT   | TOFILE
#define IFLINEPLOT  ILINEPLOT | TOFILE

typedef struct opEntry
{
	char *name;
	short iop;
} opEntry;

static opEntry Ops[] =
{ /* name        iop  */
	"plot",      IPLOT,
	"chplot",    ICHPLOT,
	"lineplot",  ILINEPLOT,
	"addpoints", IADDPOINTS,
	"addchars",  IADDCHARS,
	"addlines",  IADDLINES,
	"showplot",  ISHOWPLOT,
	"fplot",     IFPLOT,
	"fchplot",   IFCHPLOT,
	"flineplot", IFLINEPLOT,
	"addstrings",IADDSTRINGS,
	(char *) 0,  0
} ;

#define generateChars (cmax < 0)

static long buildPlot(long op, whole_graph **graph,
			long add, long style, long justify, long linetype,
			Symbolhandle symhx, Symbolhandle  symhy, Symbolhandle symhc,
			long nx, long ny, long nchars, long ncols, long cRows, long cCols,
			double xmin, double xmax, double ymin, double ymax,
			double cmax)
{
	Symbolhandle          stringsymbol = symhc;
	curve_points        **this_plot = (curve_points **) 0;
	curve_points        **last_plot = (curve_points **) 0;
	plot_string         **this_string = (plot_string  **) 0;
	plot_string         **last_string = (plot_string  **) 0;
	double              **x = DATA(symhx), **y = DATA(symhy);
	double                value;
	double                hugedbl = HUGEDBL;
	long                  cval;
	long                  nvals = ncols*ny;
	long                  findMin, findMax;
	long                  notMissing = 0, nch;
	long                  i, j, k, ij, iplot, cstart = 0, cplace, length;
	long                  nplots = (*graph)->nplots;
	long                  nstrings = (*graph)->nstrings;
	char                **plotc = (char **) 0;
	WHERE("buildPlot");
	TRASH(1,errorExit);

	findMin = (xmin == hugedbl || xmax != -hugedbl && xmin >= xmax);
	findMax = (xmax == -hugedbl || xmin != hugedbl && xmin >= xmax);

	if(op & ADDTOPLOT)
	{
		if(findMin)
		{
			xmin = (*graph)->xmin;
		}
		if(findMax)
		{
			xmax = (*graph)->xmax;
		}
	} /*if(op & ADDTOPLOT)*/

	for(i=0;i<nx;i++)
	{
		value = (*x)[i];
		if(!isMissing(value))
		{
			notMissing++;
			xmin = (findMin && value < xmin) ? value : xmin;
			xmax = (findMax && value > xmax) ? value : xmax;
		}
	} /*for(i=0;i<nx;i++)*/
	if(notMissing == 0)
	{
		sprintf(OUTSTR,"ERROR: no non-missing x values for %s to plot",
				FUNCNAME);
		goto errorExit;
	}

	notMissing = 0;
	findMin = (ymin == hugedbl || ymax != -hugedbl && ymin >= ymax);
	findMax = (ymax == -hugedbl || ymin != hugedbl && ymin >= ymax);
	if(op & ADDTOPLOT)
	{
		if(findMin)
		{
			ymin = (*graph)->ymin;
		}
		if(findMax)
		{
			ymax = (*graph)->ymax;
		}
	}

	for(i=0;i<nvals;i++)
	{
		value = (*y)[i];
		if(!isMissing(value))
		{
			notMissing++;
			ymin = (findMin && value < ymin) ? value : ymin;
			ymax = (findMax && value > ymax) ? value : ymax;
		}
	}
	if(notMissing == 0)
	{
		sprintf(OUTSTR,"ERROR: no non-missing y values for %s to plot",
				FUNCNAME);
		goto errorExit;
	}

	if(op & CHARPLOT)
	{
		if(generateChars || TYPE(symhc) == REAL)
		{
			nch = (long) log10(fabs(cmax) + .5) + 1;

			if(!getScratch(plotc,0,nchars*(nch+1),char))
			{
				goto errorExit;
			}
			k = 0;
			for(i=0;i<nchars;i++)
			{
				cval = (generateChars) ? (i%1000)+1 : (long) DATAVALUE(symhc,i);
				sprintf(OUTSTR,"%ld", cval);
				strcpy((*plotc)+k,OUTSTR);
				k += strlen(OUTSTR) + 1;
			}
			*OUTSTR = '\0';
		}
		else
		{
			plotc = STRING(symhc);
		}
	} /*if(op & CHARPLOT)*/

	if(add)
	{							/* use existing struct whole_graph */
		nplots = (*graph)->nplots;
		last_plot = (*graph)->first_plot;
		for(i=1;i<nplots;i++)
		{						/* find final struct curve_points */
			last_plot = NEXTGRAPH(last_plot);
		}

		nstrings = (*graph)->nstrings;
		last_string = (*graph)->first_string;
		for(i=1;i<nstrings;i++)
		{						/* find final struct plot_string */
			last_string = NEXTSTRING(last_string);
		}
	}
	else
	{
		nplots = nstrings = 0;
	}

	if(!(op & STRINGPLOT))
	{
		/* loop over columns of y, adding new struct curve_point for each col*/
		for(iplot = 0;iplot < ncols; iplot++)
		{
			TMPHANDLE = mygethandle(sizeof(curve_points));
			this_plot = (curve_points **) TMPHANDLE;
			if(nplots == 0)
			{
				(*graph)->first_plot = this_plot;
			}
			else
			{
				NEXTGRAPH(last_plot) = this_plot;
			}
			if (TMPHANDLE == (char **) 0)
			{
				goto errorExit;
			}
			mylockhandle(TMPHANDLE);

			last_plot = this_plot;
			nplots++;
			(*graph)->nplots = nplots;
			NEXTGRAPH(this_plot) = (curve_points **) 0;
			GRAPHPOINTS(this_plot) = (coordinate **) 0;

			TMPHANDLE = mygethandle(ny * sizeof(coordinate));
			GRAPHPOINTS(this_plot) = (coordinate **) TMPHANDLE;
			if(TMPHANDLE == (char **) 0)
			{
				goto errorExit;
			}
			mylockhandle(TMPHANDLE);

			PLOTSTYLE(this_plot) = style;

			POINTTYPE(this_plot) = (iplot == 0) ? 5 : iplot-1;
			LINETYPE(this_plot) = iplot + linetype;

			PCOUNT(this_plot) = nx;

			if(op & CHARPLOT)
			{
				cstart = (cCols > 1 && iplot > 0) ?
					skipStrings(*plotc+cstart,cRows) - *plotc  : 0;
				cplace = cstart;
				j = 0;
			}

			ij = iplot*nx;		/* iplot*nx + i */
			for (i = 0; i < nx; i++, ij++)
			{
				if (isMissing((*x)[i]) || isMissing((*y)[ij]))
				{
					undef_point(this_plot,i);
				}
				else
				{
					set_point(this_plot,i,(*x)[i],(*y)[ij],
							  (op & CHARPLOT) ? *plotc + cplace : (char *) 0);
				}

				if(op & CHARPLOT)
				{
					j++;
					/* characters are wrapped around cyclically (kb) */
					if(j >= cRows)
					{
						j = 0;
						cplace = cstart;
					}
					else
					{
						cplace += strlen(*plotc + cplace) + 1;
					}
				} /*if(op & CHARPLOT)*/
			} /*for (i = 0; i < nx; i++, ij++)*/
		} /*for(iplot = 0;iplot < ncols; iplot++)*/
	} /*if(!(op & STRINGPLOT))*/
	else
	{
		cplace = 0;
		for(i=0;i<ny;i++)
		{
			length = strlen(STRINGPTR(stringsymbol) + cplace);
			if(length > 0 && !isMissing((*x)[i]) && !isMissing((*y)[i]))
			{
				TMPHANDLE = mygethandle(sizeof(plot_string));
				this_string = (plot_string **) TMPHANDLE;
				if(nstrings == 0)
				{
					(*graph)->first_string = this_string;
				}
				else
				{
					NEXTSTRING(last_string) = this_string;
				}
				if(TMPHANDLE == (char **) 0)
				{
					goto errorExit;
				}

				nstrings++;
				(*graph)->nstrings = nstrings;
				STRINGX(this_string) = (*x)[i];
				STRINGY(this_string) = (*y)[i];
				NEXTSTRING(this_string) = (plot_string **) 0;
				PLOTSTRING(this_string) = (char **) 0;
				last_string = this_string;

				TMPHANDLE = mygethandle(length + 1);
				PLOTSTRING(this_string) = TMPHANDLE;
				if(TMPHANDLE == (char **) 0)
				{
					goto errorExit;
				}
				strcpy(*TMPHANDLE,STRINGPTR(stringsymbol) + cplace);
				JUSTIFYSTR(this_string) = justify;
			}
			cplace += length + 1;
		} /* for i */
	}
	(*graph)->xmin = xmin;
	(*graph)->xmax = xmax;
	(*graph)->ymin = ymin;
	(*graph)->ymax = ymax;

	emptyTrash();
	
	return (1);

  errorExit:
	putOUTSTR();
	
	emptyTrash();

	return (0);

} /* buildPlot()*/

Symbolhandle    myplot(list)
Symbolhandle    list;
{
	/* commands plot, chplot, lineplot, fplot fchplot, flineplot */

	Symbolhandle          symhx = (Symbolhandle) 0;
	Symbolhandle          symhy = (Symbolhandle) 0;
	Symbolhandle          symhc = (Symbolhandle) 0;
	Symbolhandle          symhstruc = (Symbolhandle) 0;
	Symbolhandle          symhfile = (Symbolhandle) 0;
	Symbolhandle          graphsymbol = (Symbolhandle) 0;
	Symbolhandle          lastgraph = (Symbolhandle) 0;
	Symbolhandle          oldplot = (Symbolhandle) 0;
	Symbolhandle          stringsymbol = (Symbolhandle) 0;
	whole_graph         **graph = (whole_graph **) 0;
	char                 *fileName = (char *) 0;
	char                 *screenDumpFile = (char *) 0;
	FILE                 *plotfile = STDOUT;
	long                  termType = STANDARDTERM;
	long                  xisIndex = 0;
	long                  minNargs, nargs, startArg, startKey, first;
	long                  i, nchars;
	long                  dims[2], nx, ny, cRows, cCols;
	long                  op, toFile = 0, add, showPlot, strucArg = 0;
	long                  ncols = 0;
	long                  style = POINTS;
	long                  dumb = 0, keep = 1, show = 1, ps = 1, new = 0;
	long                  landscape = 0;
	long                  logx = 0, logy = 0, xaxis = 1, yaxis = 1;
	long                  justify = CENTRE, windno = -1;
	long                  lines = 0, impulses = 0;
	long                  newlogx = -1, newlogy = -1;
	long                  newxaxis = -1, newyaxis = -1;
	double                hugedbl = HUGEDBL;
	double                xmin = hugedbl, xmax = -hugedbl;
	double                ymin = hugedbl, ymax = -hugedbl;
	double                newxmin = hugedbl, newxmax = hugedbl;
	double                newymin = hugedbl, newymax = hugedbl;
#ifdef MULTIPLOTWINDOWS
	long                  window = HUGEINT;
#endif /*MULTIPLOTWINDOWS*/
	long                  linetype = HUGEINT;
	double                thickness = hugedbl;
	double                x0, incx;
	double                cmax = 0.0, value;
	char                 *keyword, *name;
	char                  xlab[XLABELSIZE+2], ylab[YLABELSIZE+2];
	char                  title[TITLESIZE+2];
#ifndef MACINTOSH
	long                  pause = 1;
#else /*MACINTOSH*/
	long                  pause = 0;
	OSType                fileType = 'TEXT';
#endif /*MACINTOSH*/
	WHERE("myplot");

	OUTSTR[0] = title[0] = xlab[0] = ylab[0] = '\0';

	for(i=0; Ops[i].name != (char *) 0;i++)
	{
		if(strcmp(FUNCNAME,Ops[i].name) == 0)
		{
			op = Ops[i].iop;
			break;
		}
	}
	if(Ops[i].name == (char *) 0)
	{ /* should never happen */
		sprintf(OUTSTR,"ERROR: (internal) %s not recognized in myplot()",
				FUNCNAME);
		goto errorExit;
	}

	showPlot = !(op & NEEDXY);
	add = (op & ADDTOPLOT);

	nargs = NARGS(list);
	minNargs = (showPlot) ? 0 : 1;
	minNargs = (op & STRINGPLOT) ? 2 : minNargs;
	startKey = minNargs;
	startArg = 0;

	lines = (op & LINEPLOT);

	symhfile = COMPVALUE(list,0);

	if(op & TOFILE ||
	   symhfile != (Symbolhandle) 0 && (keyword = isKeyword(symhfile)) &&
			strcmp(keyword,"file") == 0)
	{
		toFile = 1;
		minNargs++;
		startKey++;
		startArg++;
		if(!argOK(symhfile,CHAR,1))
		{
			goto errorExit;
		}
		if(!isScalar(symhfile))
		{
			sprintf(OUTSTR,"ERROR: file name for %s must be single name",
					FUNCNAME);
			goto errorExit;
		}
		fileName = STRINGPTR(symhfile);
	}

	if(!showPlot)
	{
		if(nargs > startArg &&
		   (symhstruc = COMPVALUE(list,startArg)) != (Symbolhandle) 0 &&
		   TYPE(symhstruc) == STRUC)
		{
			strucArg = 1;
			if(NCOMPS(symhstruc) != 2)
			{
				sprintf(OUTSTR,
						"ERROR: STRUCTURE argument to %s must have 2 REAL components",
						FUNCNAME);
				goto errorExit;
			}
			symhx = COMPVALUE(symhstruc,0);
			symhy = COMPVALUE(symhstruc,1);
			if(symhx != (Symbolhandle) 0 &&
			   (TYPE(symhx) != REAL || !isVector(symhx)))
			{
				sprintf(OUTSTR,
						"ERROR: component 1 of STRUCTURE argument to %s not a REAL vector",
						FUNCNAME);
			}
			else if(symhy != (Symbolhandle) 0 &&
					(TYPE(symhy) != REAL || !isMatrix(symhy,(long *) 0)))
			{
				sprintf(OUTSTR,
						"ERROR: component 2 of STRUCTURE argument to %s not a REAL vector or matrix",
						FUNCNAME);
			}
			if(*OUTSTR)
			{
				goto errorExit;
			}
		}
		else
		{
			minNargs++;
			startKey++;
		}
	}
	
	
	first = getPlotKeys(list, startKey, (char *) title,
					   (char *) xlab, (char *) ylab,
					   &newxmin, &newymin, &newxmax, &newymax,
					   &newlogx, &newlogy, &newxaxis, &newyaxis, &dumb, &keep,
					   (!toFile) ? &show : (long *) 0,
					   (!add) ? &add : (long *) 0,
					   &ps,
					   &new,
					   &fileName,
					   (op & STRINGPLOT) ? &justify : (long *) 0,
#ifdef MULTIPLOTWINDOWS
					   &window,
#else /*MULTIPLOTWINDOWS*/
					   (long *) 0,
#endif /*MULTIPLOTWINDOWS*/
					   &pause,
					   (showPlot) ? (long *) 0 : &impulses,
					   (showPlot) ? (long *) 0 : &lines,
					   (showPlot) ? (long *) 0 : &linetype,
					   (showPlot) ? (double *) 0 : &thickness,
						&landscape,
#ifdef SCREENDUMP
						&screenDumpFile
#else /*SCREENDUMP*/
						(char **) 0
#endif /*SCREENDUMP*/
					   );
	if(first < 0)
	{
		goto errorExit;
	}
	if(add)
	{
		op |= ADDTOPLOT;
	}

	if(op & LINEPLOT && !lines)
	{
		sprintf(OUTSTR,
				"ERROR: lines:F illegal with %s",FUNCNAME);
		goto errorExit;
	}
	
	if(linetype != HUGEINT)
	{
		linetype--; /* now between 0 and MAXLINETYPE-1 */
		if(!lines)
		{
			myerrorout("WARNING: keyword 'linetype' ignored when not drawing lines");
		}
	}
	else
	{
		linetype = 0;
	}

	if(thickness != hugedbl)
	{
		if(!lines)
		{
			myerrorout("WARNING: keyword 'thickness' ignored when not drawing lines");
		}
		else
		{
	/* code thickness into linetype so actual type is linetype % MAXLINETYPE */
			linetype += MAXLINETYPE *
				(long) ((double) UNITWIDTH * thickness / (double) MAXLINETYPE);
		}
	} /*if(thickness != hugedbl)*/
	
	toFile = fileName != (char *) 0;
	if(toFile)
	{
		if(!show)
		{
			myerrorout("WARNING: show:F ignored when writing plot to file");
			show = 1;
		}

/* default is PostScript unless ps:F or dumb:T is argument */
		if(ps)
		{
			termType = PSTERM;
		}
		else
		{
#ifndef TEK
			termType = DUMBTERM;
#else	/*TEK*/
			termType = STANDARDTERM;
#endif	/*TEK*/
		}

		if((fileName = expandFilename(fileName)) == (char *) 0)
		{
			goto errorExit;
		}
		PutProlog = (new && ps) ? 1 : 0;
	} /*if(toFile)*/
	else
	{
		if(new)
		{
			myerrorout("WARNING: new:T ignored when not writing plot to file");
			new = 0;
		}
		if(!ps)
		{
			myerrorout("WARNING: ps:F ignored when not writing plot to file");
			ps = 1;
		}
	} /*if(toFile){}else{}*/

	if(dumb)
	{
		termType = DUMBTERM;
	}

	if (screenDumpFile != (char *) 0)
	{
		if(termType != STANDARDTERM)
		{
			sprintf(OUTSTR,
					"ERROR: screendump:fileName can only be used with high resolution plot");
			goto errorExit;
		}
		
		if((screenDumpFile = expandFilename(screenDumpFile)) == (char *) 0)
		{
			goto errorExit;
		}
	} /*if (screenDumpFile != (char *) 0)*/

	if(nargs < minNargs)
	{
		badNargs(FUNCNAME,-(minNargs + 1000));
		goto errorExit;
	}
	lastgraph = Lookup(LastGraph);

	graphsymbol = (add) ? Install(SCRATCH,PLOTINFO) : GraphInstall(SCRATCH);
	if(!add)
	{
		graph = GRAPH(graphsymbol);
	}

	if(add || showPlot)
	{
		if(nargs > minNargs)
		{
			oldplot = COMPVALUE(list,startArg);
			if(oldplot == lastgraph)
			{
				COMPVALUE(list,startArg) = (Symbolhandle) 0;
			}

			if(oldplot == (Symbolhandle) 0 || TYPE(oldplot) == PLOTINFO)
			{
				minNargs++;
				startKey++;
				startArg++;
			}
		}

		if(oldplot == (Symbolhandle) 0 || TYPE(oldplot) != PLOTINFO)
		{ /* use LASTPLOT if no PLOTINFO variable provided */
			if(lastgraph == (Symbolhandle) 0)
			{
				sprintf(OUTSTR,"ERROR: %s not found by %s",LastGraph,FUNCNAME);
			}
			else if(TYPE(lastgraph) != PLOTINFO)
			{
				sprintf(OUTSTR,"ERROR: variable %s not of type %s",
						LastGraph, typeName(PLOTINFO));
			}
			if(*OUTSTR)
			{
				goto errorExit;
			}
			oldplot = lastgraph;
		}
		/* make copy of GRAPH(oldplot) */
		if((graph = copyGraph(GRAPH(oldplot))) ==
		   (whole_graph **) 0)
		{
			goto errorExit;
		}
		setGRAPH(graphsymbol,graph);
		setNDIMS(graphsymbol,1);
		setDIM(graphsymbol,1,1);
		/* save info from graph only if not replaced by keywords */

		xmin = (newxmin != hugedbl) ? newxmin : (*graph)->xmin;
		xmax = (newxmax != hugedbl) ? newxmax : (*graph)->xmax;
		ymin = (newymin != hugedbl) ? newymin : (*graph)->ymin;
		ymax = (newymax != hugedbl) ? newymax : (*graph)->ymax;

		if(showPlot)
		{ /* extremes cannot be recomputed */
			if(xmin >= xmax)
			{
				sprintf(OUTSTR,
						"WARNING: xmax <= xmin on %s() ignored", FUNCNAME);
				putOUTSTR();
				xmin = (*graph)->xmin;
				xmax = (*graph)->xmax;
			}
			if(ymin >= ymax)
			{
				sprintf(OUTSTR,
						"WARNING: ymax <= ymin on %s() ignored", FUNCNAME);
				putOUTSTR();
				ymin = (*graph)->ymin;
				ymax = (*graph)->ymax;
			}
		}

		logx = (newlogx >= 0) ? newlogx : (*graph)->logx;
		logy = (newlogy >= 0) ? newlogy : (*graph)->logy;
		xaxis = (newxaxis >= 0) ? newxaxis : (*graph)->xaxis;
		yaxis = (newyaxis >= 0) ? newyaxis : (*graph)->yaxis;
		if(xlab[0] == '\0')
		{
			strcpy(xlab,(*graph)->xlab);
		}
		if(ylab[0] == '\0')
		{
			strcpy(ylab,(*graph)->ylab);
		}
		if(title[0] == '\0')
		{
			strcpy(title,(*graph)->title);
		}
	} /* if(add || showPlot)*/
	else
	{
		xmin = (newxmin != hugedbl) ? newxmin : xmin;
		xmax = (newxmax != hugedbl) ? newxmax : xmax;
		ymin = (newymin != hugedbl) ? newymin : ymin;
		ymax = (newymax != hugedbl) ? newymax : ymax;
		logx = (newlogx >= 0) ? newlogx : logx;
		logy = (newlogy >= 0) ? newlogy : logy;
		xaxis = (newxaxis >= 0) ? newxaxis : xaxis;
		yaxis = (newyaxis >= 0) ? newyaxis : yaxis;
	} /* if(add || showPlot){}else{}*/

	if(!showPlot)
	{ /* get x and y */
		if(symhx == (Symbolhandle) 0 &&
		   !argOK(symhx = COMPVALUE(list,startArg ),REAL,startArg+1))
		{
			goto errorExit;
		}

		if(symhy == (Symbolhandle) 0 &&
		   !argOK(symhy = COMPVALUE(list, startArg+1),REAL,startArg+2))
		{
			goto errorExit;
		}

		nx = symbolSize(symhx);
		xisIndex = (!(op & STRINGPLOT) && nx <= 2 && isVector(symhx)) ? 1 : 0;
		if(!isVector(symhx))
		{
			sprintf(OUTSTR,
					"ERROR: x must be a vector for %s",FUNCNAME);
		}
		else if (op & STRINGPLOT && !isMatrix(symhy,dims) && dims[1] > 1)
		{
			sprintf(OUTSTR,
					"ERROR: y must be a vector for %s",FUNCNAME);
		}
		else if (!(op & STRINGPLOT) && !isMatrix(symhy,dims))
		{
			sprintf(OUTSTR,
					"ERROR: y must be a vector or matrix for %s",FUNCNAME);
		}
		else if(nx != (ny = dims[0]) && !xisIndex)
		{
			sprintf(OUTSTR,
					"ERROR: x and y do not have same number of rows for %s",
					FUNCNAME);
		}
		else if(!add)
		{
			/*
				use names or keywords as axis labels only when not adding info
				and when they have not been set by keywords 'xlab' & 'ylab'
			*/
			if(xlab[0] == '\0')
			{
				name = NAME(symhx);
				if((keyword = isKeyword(symhx)))
				{
					strncpy(xlab,keyword,XLABELSIZE);
				}
				else if(!isscratch(name) && strcmp(name,NUMSCRATCH+2) != 0 &&
						strcmp(name,VECSCRATCH+2) != 0 &&
						strcmp(name,MATSCRATCH+2) != 0)
				{
					strncpy(xlab,name + ((istempname(name)) ? 1 : 0),
							XLABELSIZE);
				}
				xlab[XLABELSIZE] = '\0';
			} /*if(xlab[0] == '\0')*/
			if(ylab[0] == '\0')
			{
				name = NAME(symhy);
				if((keyword = isKeyword(symhy)))
				{
					strncpy(ylab,keyword,YLABELSIZE);
				}
				else if(!isscratch(name) && strcmp(name,NUMSCRATCH+2) != 0 &&
						strcmp(name,VECSCRATCH+2) != 0 &&
						strcmp(name,MATSCRATCH+2) != 0)
				{
					strncpy(ylab,name + ((istempname(name)) ? 1 : 0),
							YLABELSIZE);
				}
				ylab[YLABELSIZE] = '\0';
			} /*if(ylab[0] == '\0')*/
		} /*if(!add)*/
		ncols = dims[1];	/* number of columns of y */
	} /* if(!(showPlot) */

	if(*OUTSTR)
	{
		goto errorExit;
	}

	if(toFile)
	{
#ifdef MACINTOSH
		if(PlotVolume)
		{
			SetVol(0L, PlotVolume);
		}
		if(strlen(fileName) == 0)
		{
			/* null filename requests dialog box on Mac */

			/* no default file file name specified any more */
			fileName = macOpen("\pFile for graphs:",
							   "\p", WRITEIT, 0, (OSType *) 0);
			if (fileName == (char *) 0)
			{
				return (0);
			}
			PlotVolume = Reply.vRefNum;
		} /*if(strlen(fileName) == 0)*/
#endif /*MACINTOSH*/
		if(!isfilename(fileName))
		{
			goto errorExit;
		}
		/* should have filename */
	} /* if(toFile) */

	if(!showPlot && xisIndex)
	{ /* create equally spaced x-variable if needed */
		xisIndex = (nx != ny) ? 1 : 0;
		if(xisIndex)
		{
			x0 = DATAVALUE(symhx,0);
			incx = (nx > 1) ? DATAVALUE(symhx,1) : 1.0;
			nx = ny;
			/* at this point we are done with original symhx; make a new one */
			symhx = RInstall(SCRATCH,nx);
			if(symhx == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			xisIndex = 2;		/* indicates symhx needs to be disposed of */
			for(i=0;i<nx;i++)
			{
				DATAVALUE(symhx,i) = x0 + incx * (double) i;
			}
		}
	} /* if(!showPlot && xisIndex) */

	if(op & CHARPLOT)
	{							/* find out what characters to plot */
		style = CHARS;
		if(impulses)
		{
			style |= IMPULSES;
		}
		
		if(nargs > minNargs &&
		   !isKeyword(symhc = COMPVALUE(list, startArg+(!strucArg) ? 2 : 1)))
		{ /* characters or integers are supplied */
			startKey++;
			if(symhc == (Symbolhandle) 0)
			{
				sprintf(OUTSTR,"ERROR: %s(%s%schars)", FUNCNAME,
						(toFile) ? "filename," : "",
						(strucArg) ? "structure," : "x,y,");
			}
			else if(!isMatrix(symhc,dims))
			{
				sprintf(OUTSTR,
						"ERROR: chars must be a vector%s for %s",
						(ncols > 0) ? " or matrix" : "", FUNCNAME);
			}
			else if((cCols = dims[1]) != 1 && cCols != ncols)
			{
				sprintf(OUTSTR,"ERROR: ncols(chars) must be 1 or match ncols(y)");
			}
			else if(TYPE(symhc) == REAL)
			{ /* check to see values are integers */
				cRows = dims[0];
				if(cCols == 1 && cRows == ncols)
				{
					cCols = ncols;
					cRows = 1;
				}

				nchars = cRows*cCols;

				for(i=0;i<nchars;i++)
				{
					if((value = DATAVALUE(symhc,i)) < 0.0 || value > 999.0 ||
					   value != floor(value))
					{
						sprintf(OUTSTR,
								"ERROR: numerical chars for %s must be non-negative integers < 1000",
								FUNCNAME);
						break;
					}
					cmax = (value > cmax) ? value : cmax;
				} /*for(i=0;i<nchars;i++)*/
			}
			else if(TYPE(symhc) == CHAR)
			{ /* characters */
				cRows = dims[0];
				if(cCols == 1 && cRows == ncols)
				{
					cCols = ncols;
					cRows = 1;
				}
			}
			else
			{
				sprintf(OUTSTR,
						"ERROR: argument %ld to %s must be characters or positive integers < 1000 ",
						startArg+1,FUNCNAME);
			}
		}
		else
		{ /* no labelling information supplied */
			symhc = (Symbolhandle) 0;
			if(ncols > 1)
			{
				cRows = 1;
				cCols = ncols;
			}
			else
			{
				cRows = nx;
				cCols = 1;
			}
			nchars = cRows*cCols;
			/* negative cmax indicates labels will need to be built */
			cmax = -((nchars < 999) ? nchars : 999);
		}

		if(lines)
		{
			style |= LINES;
		}
	} /* if(op & CHARPLOT) */
	else if(op & PTPLOT)
	{
		style = (impulses) ? IMPULSES : POINTS;
		if(lines)
		{
			style |= LINES;
		}
	}
	else if(op & LINEPLOT)
	{
		style = LINES;
		if(impulses)
		{
			style |= IMPULSES;
		}
	}
	else if(op & STRINGPLOT)
	{
		stringsymbol = COMPVALUE(list,startArg+2);
		if(!isVector(stringsymbol) || TYPE(stringsymbol) != CHAR ||
		   symbolSize(stringsymbol) != ny)
		{
			sprintf(OUTSTR,
					"ERROR: strings must be a CHARACTER vector of the same length as x for %s",
					FUNCNAME);
			goto errorExit;
		}
	}

	if(*OUTSTR)
	{
		goto errorExit;
	}

#ifdef MULTIPLOTWINDOWS
	windno = (window != HUGEINT) ? window : -1;
#endif /*MULTIPLOTWINDOWS*/

	if(!keep && !show)
	{
		sprintf(OUTSTR,	"ERROR: keep:F not legal with show:F");
		goto errorExit;
	}

	if(toFile)
	{
		plotfile = openPlotFile(fileName,new);
		if(plotfile == (FILE *) NULL)
		{
			goto errorExit;
		}
	}
	else
	{
		plotfile = STDOUT;
	}

	if(!showPlot)
	{
		if(!buildPlot(op, graph, add, style, justify, linetype,
					  symhx, symhy,
					  (op & STRINGPLOT) ? stringsymbol : symhc,
					  nx, ny, nchars, ncols, cRows, cCols,
					  xmin, xmax, ymin, ymax, cmax))
		{
			goto errorExit;
		}
	}
	else
	{
		(*graph)->xmin = xmin;
		(*graph)->xmax = xmax;
		(*graph)->ymin = ymin;
		(*graph)->ymax = ymax;
	}

	(*graph)->logx = logx;
	(*graph)->logy = logy;
	(*graph)->xaxis = xaxis;
	(*graph)->yaxis = yaxis;
	strcpy((*graph)->xlab,xlab);
	strcpy((*graph)->ylab,ylab);
	strcpy((*graph)->title,title);

	if(keep)
	{/* save plotting info in LASTPLOT */
		Removesymbol(lastgraph); /* make sure LASTGRAPH is removed */
		Cutsymbol(graphsymbol);
		setNAME(graphsymbol,LastGraph);
		Addsymbol(graphsymbol);
		graphsymbol = (Symbolhandle) 0;
	}

	if(show && !do_whole_graph(graph, plotfile, termType, windno, pause,
							   landscape, screenDumpFile))
	{
		goto errorExit;
	}

	if(plotfile != STDOUT)
	{
#ifdef MACINTOSH
		macSetInfo(PlotVolume,fileName,fileType,CREATOR);
#endif /*MACINTOSH*/
		fclose(plotfile);
	}

	Removesymbol(graphsymbol);

	if(xisIndex == 2)
	{
		Removesymbol(symhx);
	}

	*OUTSTR = '\0';

	return (NULLSYMBOL);

  errorExit:
	putOUTSTR();

	if(plotfile != STDOUT)
	{
		fclose(plotfile);
	}
	Removesymbol(graphsymbol);

	if(xisIndex == 2)
	{
		Removesymbol(symhx);
	}

	return (0);
}

