/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Outer
#else
#define __SEG__ Outer
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

extern Symbolhandle     RInstall();
extern long             argOK();
extern void             badNargs();
extern long             symbolSize();

Symbolhandle    outer(list)
Symbolhandle    list;
{

	/* routine to form the outer product of two arrays */
	/* dimension vector is the concatenation of the two dimension vectors */

	Symbolhandle    result = (Symbolhandle) 0, arg1, arg2;
	long            tot1, tot2, totd, i, j;
	long            ndims1, ndims2;
	long            foundMissing = 0, nMissing = 0;
	double         *prods, *x1, *x2;
	double          tmp1, tmp2;
	WHERE("outer");

	if(NARGS(list) != 2)
	{
		badNargs(FUNCNAME,2);
		goto errorExit;
	}
	
	arg1 = COMPVALUE(list, 0);
	arg2 = COMPVALUE(list, 1);
	if(!argOK(arg1, REAL, 1) || !argOK(arg2, REAL, 2))
	{
		goto errorExit;
	}
	
	ndims1 = NDIMS(arg1);
	ndims2 = NDIMS(arg2);

	if ((totd = ndims1 + ndims2) > MAXDIMS)
	{
		myerrorout("ERROR:  too many dimensions");
		goto errorExit;
	}

	tot1 = symbolSize(arg1);
	tot2 = symbolSize(arg2);

	result = RInstall(SCRATCH, tot1 * tot2);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	setNDIMS(result,totd);
	for (i = 1; i <= ndims1; i++)
	{
		setDIM(result,i,DIMVAL(arg1,i));
	}
	
	for (i = 1; i <= ndims2; i++)
	{
		setDIM(result,i + ndims1,DIMVAL(arg2,i));
	}
	
	x2 = DATAPTR(arg2);
	prods = DATAPTR(result);

	for (j = 0; j < tot2; j++)
	{
		tmp2 = *x2++;
		if(isMissing(tmp2) || nMissing >= tot1)
		{
			for(i = 0; i < tot1; i++)
			{
				setMissing(*prods++);
			}
			foundMissing = 1;
		}
		else
		{
			x1 = DATAPTR(arg1);
			nMissing = 0;
			for (i = 0; i < tot1; i++)
			{
				tmp1 = *x1++;
				if (!isMissing(tmp1))
				{
					*prods++ = tmp1 * tmp2;
				}
				else
				{
					setMissing(*prods++);
					nMissing++;
				}
			} /* for i */
		}
	} /* for j */

	if(foundMissing || nMissing > 0)
	{
		sprintf(OUTSTR, "WARNING: missing values in argument(s) to %s",
				FUNCNAME);
		putOUTSTR();
	}
	
	return (result);

  errorExit:
	return (0);
}
