/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Paste
#else
#define __SEG__ Paste
#endif /*MPW3*/
#endif /*MPW*/

#define CHUNK    (5*sizeof(double))

#include "globals.h"
#include <ctype.h>

#define ISEP        1
#define IFORMAT     2
#define IIWIDTH     3
#define IAWIDTH     4
#define MAXIWIDTH 100
#define MAXFWIDTH 100

static char       *LegalKeywords[] =
{
	"sep", "format", "intwidth", "charwidth", (char *) 0
};

	
Symbolhandle paste(list)
Symbolhandle      list;
{
	long          nargs = NARGS(list);
	Symbolhandle  symh, result = (Symbolhandle) 0;
	char        **resultString = (char **) 0;
	char         *space = " ";
	char        **separator = (char **) 0;
	char          format[20];
	char          intformat[20];
	char         *keyword, *outstr;
	int           i,j;
	long          nchar = 0, nitems, nleft = CHUNK, length, place=0, needed;
	long          iwidth = -1, fwidth = -1, awidth = -1;
	long          type, key;
	long          fmt[3];
	double        value;
	WHERE("paste");
	
	intformat[0] = format[0] = OUTSTR[0] = '\0';
	fmt[0] = -1;
	for(i=0;i<nargs;i++)
	{
		if(!argOK(symh = COMPVALUE(list,i),0,i+1))
		{
			return (0);
		}
		if((type = TYPE(symh)) == STRUC || type == PLOTINFO || type == BLTIN)
		{
			badType(FUNCNAME, -type, i+1);
			return (0);
		}
		
		if((keyword = isKeyword(symh)))
		{
			key = matchKey(keyword,LegalKeywords, (long *) 0);
			if(key == 0)
			{
				badKeyword(FUNCNAME,keyword);
				return (0);
			}
			else
			{
				if((key == ISEP || key == IFORMAT) &&
				   (!isScalar(symh) || TYPE(symh) != CHAR))
				{
					sprintf(OUTSTR,
							"ERROR: keyword '%s' must be CHARACTER variable or string",
							keyword);
					goto errorExit;
				}
				else if((key == IIWIDTH || key == IAWIDTH) &&
				   (!isScalar(symh) || TYPE(symh) != REAL || 
					(value = DATAVALUE(symh,0)) <= 0 ||
					value != floor(value)))
				{
					sprintf(OUTSTR,
							"ERROR: keyword '%s' have positive integer as value",
							keyword);
					goto errorExit;
				}
			}
		}
	} /*for(i=0;i<nargs;i++)*/

	for(i=nargs-1;i>=0;i--)
	{
		symh = COMPVALUE(list,i);
		if(!(keyword = isKeyword(symh)))
		{
			break;
		}
		nargs--;
		key = matchKey(keyword,LegalKeywords, (long *) 0);
		
		/* trailing separators or formats become the default */
		if(key == ISEP && separator == (char **) 0)
		{
			separator = STRING(symh);
		}
		else if(key == IFORMAT && *format == '\0')
		{
			if(!setFormat(STRINGPTR(symh),fmt))
			{
				sprintf(OUTSTR,
						"ERROR: %s is not legal form for format for %s",
						STRINGPTR(symh),FUNCNAME);
				goto errorExit;
			}
			if(isdigit(STRINGVALUE(symh,0)))
			{
				fwidth = fmt[0];
				if(fwidth > MAXFWIDTH)
				{
					fwidth = fmt[0] = MAXFWIDTH;
				}
			}
		}
		else if(key == IIWIDTH && iwidth < 0)
		{
			iwidth = (long) DATAVALUE(symh,0) + 1000;
			if(iwidth > 1000 + MAXIWIDTH)
			{
				iwidth = 1000 + MAXIWIDTH;
			}
			sprintf(intformat,"%%%ldld", iwidth - 1000);
		}
		else if(key == IAWIDTH && awidth < 0)
		{
			awidth = (long) DATAVALUE(symh,0);
		}
		if(separator != (char **) 0 && fmt[0] > 0 &&
		   iwidth > 0 && awidth > 0)
		{
			break;
		}
	} /*for(i=nargs-1;i>=0;i++)*/

	if(separator == (char **) 0)
	{
		separator = &space;
	}
	if(iwidth < 0 && fwidth >= 0)
	{
		iwidth = fwidth;
		sprintf(intformat,"%%%ldld", iwidth);
	}
	
	if(fmt[0] < 0)
	{
		fmt[0] = PRINTFORMAT[0];
		fmt[1] = PRINTFORMAT[1];
		fmt[2] = PRINTFORMAT[2];
	}
	sprintf(format,"%%%ld.%ld%c",fmt[0],fmt[1],(char) fmt[2]);
	if(iwidth < 0)
	{
		strcpy(intformat,"%ld");
	}
	
	result = Install(SCRATCH,CHAR);
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	resultString = mygethandle(nleft);
	setSTRING(result,resultString);
	if(resultString == (char **) 0)
	{
		goto errorExit;
	}
	setNDIMS(result,1);
	setDIM(result,1,1);

	**resultString = '\0';
	nchar++;
	nleft--;
		
	for(i=0;i<nargs;i++)
	{
		symh = COMPVALUE(list,i);
		if((keyword = isKeyword(symh)))
		{
			key = matchKey(keyword,LegalKeywords, (long *) 0);
			if(key == ISEP)
			{
				separator = STRING(symh);
			}
			else if(key == IFORMAT)
			{
				if(!setFormat(STRINGPTR(symh),fmt))
				{
					sprintf(OUTSTR,
							"ERROR: %s is not legal form for format for %s",
							STRINGPTR(symh),FUNCNAME);
					goto errorExit;
				}
				if(isdigit(STRINGVALUE(symh,0)))
				{
					fwidth = fmt[0];
					if(fwidth > MAXFWIDTH)
					{
						fwidth = fmt[0] = MAXFWIDTH;
					}
					if(iwidth < 1000)
					{
						iwidth = fwidth;
						sprintf(intformat,"%%%ldld", iwidth);
					}
				}
				else
				{
					fwidth = -1;
					if(iwidth < 1000)
					{
						iwidth = -1;
					}
				}
			}
			else if(key == IIWIDTH)
			{
				iwidth = (long) DATAVALUE(symh,0) + 1000;
				if(iwidth > 1000 + MAXIWIDTH)
				{
					iwidth = 1000 + MAXIWIDTH;
				}
				sprintf(intformat,"%%%ldld", iwidth - 1000);
			}
			else if(key == IAWIDTH)
			{
				awidth = (long) DATAVALUE(symh,0);
			}
			sprintf(format,"%%%ld.%ld%c",fmt[0],fmt[1],(char) fmt[2]);
		} /*if((keyword = isKeyword(symh)))*/
		else
		{
			nitems = symbolSize(symh);
			place = 0;
			for(j=0;j<nitems;j++)
			{
				if(nchar > 1)
				{
					strcpy(OUTSTR,*separator);
				}
				else
				{
					*OUTSTR = '\0';
				}
				
				outstr = OUTSTR + strlen(OUTSTR);
				type = TYPE(symh);
				switch ((int) type)
				{
				  case LOGIC:
					value = DATAVALUE(symh,j);
					sprintf(outstr,"%s", (value != 0.0) ? "T" : "F");
					break;
					
				  case REAL:
					value = DATAVALUE(symh,j);
					if(fabs(value) < 2147483647.5 && value == floor(value))
					{
						sprintf(outstr,intformat, (long) value);
					}
					else
					{
						sprintf(outstr,format,value);
						if(fwidth < 0)
						{ /* no width specified */
							while(*outstr == *space)
							{	/* trim off leading blanks by shifting left */
								strcpy(outstr,outstr+1);
							}
							length = strlen(outstr);
							while(outstr[length-1] == *space)
							{	/* trim off trailing blanks */
								length--;
							}
							outstr[length] = '\0';
						} /*if(fwidth < 0)*/
					}
					break;
				  case CHAR:
				  case MACRO:
					/* do not copy directly to OUTSTR; may be too long */
					needed = strlen(STRINGPTR(symh)+place);
					if(awidth > needed)
					{
						needed = awidth;
					}
					break;
					
				} /*switch ((int) type)*/

				length = strlen(OUTSTR);
				if(type == CHAR || type == MACRO)
				{
					length += needed;
				}
				
				if(length >= nleft)
				{
					nleft = CHUNK*((length - 1)/CHUNK + 1);
					resultString = mygrowhandle(resultString,nchar+nleft);
					setSTRING(result,resultString);
					if(resultString == (char **) 0)
					{
						goto errorExit;
					}
				}
				strcat(*resultString,OUTSTR);
				if(type == CHAR || type == MACRO)
				{ /* copy directly from source */
					outstr = *resultString + strlen(*resultString);
					strcpy(outstr,STRINGPTR(symh)+place);
					needed = strlen(STRINGPTR(symh) + place);
					outstr += needed;
					while(awidth > needed)
					{
						*outstr++ = ' ';
						needed++;
					}
					*outstr = '\0';
					place += strlen(STRINGPTR(symh)+place) + 1;
				}
				
				nleft -= length;
				nchar += length;
				*OUTSTR = '\0';
			} /*for(j=0;j<nitems;j++)*/
		} /*if((keyword = isKeyword(symh))){}else{}*/
	} /*for(i=0;i<nargs;i++)*/

/* trim resultString to exact size */
	resultString = mygrowhandle(resultString, nchar);
	setSTRING(result,resultString);
	if(resultString == (char **) 0)
	{
		goto errorExit;
	}
	

	return (result);
	
  errorExit:
	putOUTSTR();
	Removesymbol(result);
	return (0);
}
