/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


/*
   for defines that differ between platforms
   Normally not in dependency list in Makefile
*/

#ifndef PLATFORMH__
#define PLATFORMH__
#undef UNDEFINED__

/*
   No compilation tried with THINK_C in years.  Many changes would
   have to be made
*/
#ifdef THINK_C
#define MACINTOSH
#define COPYRIGHT3 	"(C) Copyright 1986 THINK Technologies, Inc."
#define COPYRIGHT4  "Certain portions of this software are copyrighted by THINK Technologies, Inc."
#endif /*THINK_C*/

#ifdef MPW
#define MPW3
#define MPW3_2
#define MACINTOSH
#ifdef UNDEFINED__         /* MPW 3.2 has strtod() */
#define NOSTRTOD           /* does not have strtod() */
#endif /*UNDEFINED__*/
#define CANSETFONTS        /*options 'font' and 'fontsize' meaningful*/
#define NOLGAMMA           /* no function lgamma() */
#define MULTIPLOTWINDOWS   /* several plotting windows available */
#define SCROLLABLEWINDOW   /* output window is scrollable */
#define PLATFORM     "Macintosh (MPW)"
#ifndef MACANOVAINI
#define MACANOVAINI  "MacAnova.ini"
#endif /*MACANOVAINI*/
#endif /*MPW*/

#ifdef BCPP  /* Borland C++ compiler; used to be DOS */
#define MSDOS           /* operating system of PC compatible */
#define MAXHANDLESIZE   65000L  /* maximum size of handles allocated */
#define ACTIVEUPDATE 1 /* update console after plotting commands */
#define HASISATTY	/*has isatty()*/
#define COMMANDLINE	/*has command line arguments*/
#define HASSYSTEM	/*has system()*/
#define NOLGAMMA           /* no function lgamma() */
#define USEPOW		/*use pow() for integer exponents*/
#define PLATFORM    "Generic MSDOS (BCPP)"
#ifndef MACANOVAINI
#define MACANOVAINI  "MACANOVA.INI"
#endif /*MACANOVAINI*/
#endif /*BCPP*/

#ifdef DJGPP  /* dj gpp compiler for MSDOS */
#ifdef UNDEFINED__  /*MSDOS is pre-defined in djgpp*/
#define MSDOS           /* operating system of PC compatible */
#endif /*UNDEFINED__*/
#define SCREENDUMP	/*screendump implemented for plotting commands*/
#define BINARYHELPFILE  /* open help file as binary file */
#define ACTIVEUPDATE 1 /* update console after plotting commands */
#define HASISATTY	/*has isatty()*/
#define COMMANDLINE	/*has command line arguments*/
#define HASSYSTEM	/*has system()*/
#define NOLGAMMA    /* no function lgamma() */
#define USEPOW		/*use pow() for integer exponents*/
#define DEFINEEDIT /*edit a pre-defined macro*/
#define PLATFORM    "32 bit MSDOS protected mode (DJGPP)"
#ifndef MACANOVAINI
#define MACANOVAINI  "MACANOVA.INI"
#endif /*MACANOVAINI*/
#endif /*DJGPP*/

#ifdef UNIX   /* generic unix */
#define HASPWDH		/*has <pwd.h>*/
#ifndef ACTIVEUPDATE /* allow it to be set in Makefile */
#define ACTIVEUPDATE 0 /* update console after plotting commands */
#endif /*ACTIVEUPDATE*/
#define HASISATTY	/*has isatty()*/
#define COMMANDLINE	/*has command line arguments*/
#define HASSYSTEM	/*has system()*/
#ifndef HPGCC /*gcc gave inaccurate integer powers on HP/UX*/
#define USEPOW		/*use pow() for integer exponents*/
#endif /*HPGCC*/
#define DEFINEEDIT /*edit a pre-defined macro*/

#if defined(EPX)
#define PLATFORM       "CDC EP/IX"
#define NOSTRTOD	/* epx does not seem to have strtod() */
#define labs(A)  abs(A)  /* epx does not seem to have labs() */
#elif defined(DEC3100)
#define PLATFORM       "DECStation 3100 Ultrix"
#elif defined(HPPA)
#define PLATFORM       "Hewlett-Packard HPUX on HP-PA"
#define _INCLUDE_POSIX_SOURCE
#define SIGNALARG /*signal handler of form void handler(int) */
#elif defined(HPGCC)
#define PLATFORM       "Hewlett-Packard HPUX on HP-PA"
#define _INCLUDE_POSIX_SOURCE
#define SIGNALARG /*signal handler of form void handler(int) */
#elif defined(LINUX)
#define PLATFORM       "Linux"
#else
#define PLATFORM       "Unknown Unix"
#endif

#ifndef TEK
#define TEK
#endif

#ifndef MACANOVAINI
#define MACANOVAINI  ".macanova.ini"
#endif /*MACANOVAINI*/
#endif /*UNIX*/

/* following are (and probably should be) defined on all platforms */
#define DUMB		/* "dumb" plots an option on plotting commands */
#define BITOPS		/* bit (%&, %|, etc.) operations will be defined */
#define POSTSCRIPT	/* postscript output an option on plotting commands */
#define USEMEMSET 	/* use memset() to initialize symbols */
#define USEMEMCPY   /* use memcpy() to copy handles */
#define USEFSEEK   /* use ftell() & fseek() to locate help items */
#ifndef CONSOLENAME
#define CONSOLENAME "CONSOLE"	/* CONSOLE recognized as pseudo file name */
#endif /*CONSOLENAME*/
#define HASTIMEH	/* has <time.h> */
#define HASSTDLIBH	/* has <stdlib.h> */

/* 1st two lines of copyright information for all platforms */
#define COPYRIGHT1 "Type 'help(copyright)' for copyright and warranty info"
#define COPYRIGHT2 "Copyright 1994, 1995, Gary W. Oehlert and Christopher Bingham"

#endif /*PLATFORMH__*/
