/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

/*
 *
 *    G N U P L O T  --  plot.h
 *
 *  Copyright (C) 1986, 1987  Thomas Williams, Colin Kelley
 *
 *  You may use this code as you wish if credit is given and this message
 *  is retained.
 *
 *  Please e-mail any useful additions to vu-vlsi!plot so they may be
 *  included in later releases.
 *
 *  This file should be edited with 4-column tabs!  (:set ts=4 sw=4 in vi)
 */

#include "platform.h"  /*kb*/

#ifndef PLOTH__
#define PLOTH__

#include <math.h>
#include "typedefs.h"

#define ZERO	1e-8		/* default for 'zero' set option */

/*
#define TRUE 1
#define FALSE 0
*/



/*
 * note about HUGE:  this number is just used as a flag for really
 *   big numbers, so it doesn't have to be the absolutely biggest number
 *   on the machine.
 *
 * define HUGE here if your compiler doesn't define it in <math.h>
 * example:
 */
#ifndef HUGE
#define HUGE 1e38
#endif /*HUGE*/

#define USEHANDLES /* MacAnova assumes handles used*/

#define xsize   1.0  /* global defined in setshow.c in gnuplot 3 */
#define ysize   1.0  /* global defined in setshow.c in gnuplot 3 */

#define XLABELSIZE     50 /*kb*/
#define YLABELSIZE     20 /*kb*/
#define TITLESIZE      75 /*kb*/
#define PLOTSTRINGSIZE  3 /* kb */

#define STANDARDTERM 1 /* default terminal type for platform */
#define DUMBTERM     2 /* dumb terminal output to screen or file */
#define PSTERM       3 /* postscript output to file */

#define XAXISLINETYPE   (-3)
#define YAXISLINETYPE   (-2)
#define BORDERLINETYPE  (-1)

#ifdef TERMC__ /* should be defined only in term.c */
Long            PutProlog = 1;
Long            term = STANDARDTERM;
#ifdef MACINTOSH
Long            PauseAfterPlot = 0;
#else /*MACINTOSH*/
Long            PauseAfterPlot = 1;
#endif /*MACINTOSH*/
#else /*TERMC__*/
extern Long     PutProlog;
extern Long     term;
extern Long     PauseAfterPlot;
#endif /*TERMC__*/

/* map floating point x and y to screen */
#define mapXToScreen(x) \
   (Long)(GraphMap.xoffset + (x - GraphMap.xmin)*GraphMap.xscale)
#define mapYToScreen(y) \
   (Long)(GraphMap.yoffset + (y - GraphMap.ymin)*GraphMap.yscale)

/* map screen coordinates to floating point x and y */
#define mapScreenToX(ix) \
	(GraphMap.xmin + ((Double) ix - GraphMap.xoffset)/GraphMap.xscale)
#define mapScreenToY(iy) \
	(GraphMap.ymin + ((Double) iy - GraphMap.yoffset)/GraphMap.yscale)

typedef struct graph_map
{
	Double                xmin;
	Double                xscale;
	Double                xoffset;
	Double                ymin;
	Double                yscale;
	Double                yoffset;
	Long                  logx;
	Long                  logy;
	Long                  screenxmax;
	Long                  screenymax;
} graph_map;

#ifdef GRAPHICSC__    /* defined only in graphics.c */
graph_map                 GraphMap;
#else /*GRAPHICSC__*/
extern graph_map          GraphMap;
#endif /*GRAPHICSC__*/

/*
typedef Long BOOLEAN;
*/

/* prototypes for public functions in term.c */
void horstring(Long /*x*/, Long /*y*/, Char * /*str*/, Long /*justify*/);
void vertstring(Long /*x*/, Long /*y*/, Char * /*str*/);

typedef int     (*FUNC_PTR) ();

enum PLOT_STYLE
{ /* Plot style constants used in Version 3.1x */
	LINES_V31, POINTS_V31, IMPULSES_V31, CHARS_V31,
	BOTHLANDP_V31, BOTHLANDC_V31
};

/* Plot style constants used in version 3.33 and later */	
#define LINES          0x01
#define POINTS         0x02
#define IMPULSES       0x04
#define CHARS          0x08
#define BOTHLANDP      0x03 /*(LINES | POINTS)*/
#define BOTHLANDC      0x09 /*(LINES | CHARS)*/
#define BOTHLANDI      0x05 /*(LINES | IMPULSES)*/

/* Line type is coded as
   (long) (MAXLINETYPE*(fraction*(UNITWIDTH/MAXLINETYPE)))+linetype
   where fraction is between .1 and 10 and specified the width of the line
   and linetype is between 0 and MAXLINETYPE-1 and specifies the patter of
   the line, e.g., solid, dashed, dotted, etc.
*/

#define UNITWIDTH      1000000 /* code line width as (long) (width*UNITWIDTH)*/
#define MAXLINETYPE    100     /* linetype must be < MAXLINETYPE */

enum JUSTIFY
{
	LEFT, CENTRE, RIGHT
};

typedef struct plot_string 
{
	Double                x, y;
	Long                  justify;
#ifdef USEHANDLES
	Char                **string;
	struct plot_string  **next_string;
#else /*USEHANDLES*/
	Char                 *string;
	struct plot_string   *next_string;
#endif /*USEHANDLES*/
} plot_string;

typedef struct coordinate
{
	/*	BOOLEAN undefined;	*//* TRUE if value off screen */
	Long            undefined;
	Double          x, y;
	Char            string[PLOTSTRINGSIZE+1]; /* changed from 3 by kb */
} coordinate;

/*
  If USEHANDLES is defined, the argument should be a (struct curve_points **),
  e.g., PLOTSTYLE(GRAPH(symh)), PLOTSTYLE(NEXTGRAPH(GRAPH(symh)))
  and the macros pre-suppose that all pointers in curve_point are
  changed to be handles 

  If USEHANDLES is not defined, the argument should be a (struct curve_points *)
  and the original definitions (points) in curve_point are assumed
*/

#ifdef USEHANDLES
/* A should be (struct curve-point **), B should be (struct plot_string  **) */
#define PLOTSTYLE(A) ((*(A))->plot_style)
#define NEXTGRAPH(A) ((*(A))->next_cp) /* handle to next graph */
#define POINTTYPE(A) ((*(A))->point_type)
#define LINETYPE(A) ((*(A))->line_type)
#define PLOTTITLE(A) ((*(A))->title)
#define PCOUNT(A) ((*(A))->p_count)
#define GRAPHPOINTS(A) ((*(A))->points) /* handle to data */
#define GRAPHPOINT(A,I) ((*GRAPHPOINTS(A))[I])
#define STRINGX(B)  ((*(B))->x)
#define STRINGY(B)  ((*(B))->y)
#define JUSTIFYSTR(B) ((*(B))->justify)
#define PLOTSTRING(B)   ((*(B))->string) /* handle */
#define NEXTSTRING(B)  ((*(B))->next_string)
#else /*USEHANDLES*/

/* following are for A of type (struct curve-point *) */
#define PLOTSTYLE(A) ((A)->plot_style)
#define NEXTGRAPH(A) ((A)->next_cp) /* pointer to next graph */
#define POINTTYPE(A) ((A)->point_type)
#define LINETYPE(A) ((A)->line_type)
#define PLOTTITLE(A) ((A)->title)
#define PCOUNT(A) ((A)->p_count)
#define GRAPHPOINTS(A) ((A)->points) /* pointer to data */
#define GRAPHPOINT(A,I) (GRAPHPOINTS(A)[I])
#define STRINGX(B)  ((B)->x)
#define STRINGY(B)  ((B)->y)
#define JUSTIFYSTR(B) ((B)->justify)
#define PLOTSTRING(B)   ((B)->string) /* pointer */
#define NEXTSTRING(B)  ((B)->next_string)
#endif /*USEHANDLES*/

/* the following are independent of status of USEHANDLES */
#define POINTUNDEF(A,I) (GRAPHPOINT(A,I).undefined)
#define XCOORD(A,I) (GRAPHPOINT(A,I).x)
#define YCOORD(A,I) (GRAPHPOINT(A,I).y)
#define POINTSTRING(A,I) (GRAPHPOINT(A,I).string)

typedef struct curve_points
{
#ifdef USEHANDLES
	struct curve_points **next_cp;	/* handle for next plot in linked list */
#else /*USEHANDLES*/
	struct curve_points *next_cp;	/* pointer to next plot in linked list */
#endif /*USEHANDLES*/
#ifdef UNDEFINED__
	enum PLOT_STYLE plot_style;
#else /*UNDEFINED__*/
	Long            plot_style;
#endif /*UNDEFINED__*/
	Long            point_type;
	Long            line_type;
	Long            p_count;/* count of points in .points[] */

#ifdef USEHANDLES
	coordinate **points; /* handle for data */
#else /*USEHANDLES*/
	coordinate *points; /* pointer to data */
#endif /*USEHANDLES*/
	/* will usually be less in malloc()'d copy */
} curve_points;

typedef struct whole_graph
{
	Long                  nplots;
	Long                  nstrings;
	Double                xmin;
	Double                xmax;
	Double                ymin;
	Double                ymax;
	Long                  logx;
	Long                  logy;
	Long                  xaxis;
	Long                  yaxis;
/* NB: title, xlab and ylab used to be in struct curve_points */
	Char                  title[TITLESIZE+2]; /*kb*/
	Char                  xlab[XLABELSIZE+2]; /*kb*/
	Char                  ylab[YLABELSIZE+2]; /*kb*/
#ifdef USEHANDLES
	curve_points **first_plot;
	plot_string  **first_string;
#else /*USEHANDLES*/
	curve_points *first_plot;
	plot_string  *first_string;
#endif /*USEHANDLES*/
} whole_graph ;

/* 
  Note: Ultrix 4.2 version of CC cannot handle the termentry declaration with
  function prototypes with more than 1 argument whose type is defined by
  typedef.  DEFINETYPES should be defined in the Makefile if needed
*/

#ifdef DEFINETYPES
#ifdef BCPP
#define Void                void huge
#define Double              double huge
#define Long                long huge
#define uLong               unsigned long huge
#define Char                char huge
#define uChar               unsigned char huge
#define Int                 int        
#define Short               short
#else /*BCPP*/
#define Void                void
#define Double              double     
#define Long                long       
#define uLong               unsigned long
#define Char                char       
#define uChar               unsigned char
#define Int                 int        
#define Short               short      
#endif /*BCPP*/
#endif /*DEFINETYPES*/

typedef struct termentry
{
	Char           *name;
	uLong           xmax, ymax, v_char, h_char, v_tic, h_tic;
	int             (*init)(void);
	void            (*reset)(void);
	void            (*text)(void);
	int             (*graphics)(void);
	void            (*move)(uLong, uLong);
	void            (*vector)(uLong, uLong);
	void            (*linetype)(Long);
	void            (*chr_text)(Char);
	void            (*str_text)(Char *);
	void            (*point)(Long, Long, Long);
} termentry;

#ifdef DEFINETYPES
#undef Void
#undef Double
#undef Long
#undef uLong
#undef Char
#undef uChar
#undef Int
#undef Short
#endif /*DEFINETYPES*/

/* original form of declaration 
struct termentry
{
	Char           *name;
	uLong           xmax, ymax, v_char, h_char, v_tic, h_tic;
	FUNC_PTR        init, reset, text, graphics, move, vector, linetype,
	                chr_text, str_text, point;
};
*/
#endif /*PLOTH__*/
