/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Myplot
#else
#define __SEG__ Myplot
#endif /*MPW3*/
#endif /*MPW*/

#include <math.h>
#include "globals.h"

#ifdef MACINTOSH
#include "macIface.h"
#endif /*MACINTOSH*/

#include "plot.h"
#include "dbug.h"

#ifdef MPW
#include <Memory.h>
#endif

#define Binary (ASCII < 0)

extern long        TermCount; /* defined at end of term.c */
extern long        ps_portrait; /* defined in term.c */
void set_point(curve_points **plot, long i, double x, double y,
			   char *string)
{
	WHERE("set_point");

	XCOORD(plot,i) = x;
	YCOORD(plot,i) = y;
	POINTUNDEF(plot,i) = 0;
	if(string == (char *) 0)
	{
		POINTSTRING(plot,i)[0] = '\0';
	}
	else
	{
		strncpy(POINTSTRING(plot,i), string, PLOTSTRINGSIZE);
		POINTSTRING(plot,i)[PLOTSTRINGSIZE] = '\0';
	}

} /*set_point()*/

void undef_point(curve_points **plot,long i)
{
	XCOORD(plot,i) = 0.0;
	YCOORD(plot,i) = 0.0;
	POINTUNDEF(plot,i) = 1;
	POINTSTRING(plot,i)[0] = '\0';
} /*undef_point()*/

#define SCREENDUMPOK        1
#define SCREENDUMPUNDEFINED 0
#define BADSCREENMODE      -1
#define BADCOLORS          -2
#define BADWIDTH           -3
#define CANTOPEN           -4
#define TOOMANYDUMPS       -5
#define FILEERROR          -6

short  ScreenDumpError = SCREENDUMPOK;
char  *ScreenDumpFile = (char *) 0;

static void screenDumpMsg(void)
{
	switch (ScreenDumpError)
	{
	  case SCREENDUMPOK:
		return;

	  case SCREENDUMPUNDEFINED:
		sprintf(OUTSTR,
				"ERROR: screen dump is not available in this plotting mode");
		break;

	  case BADSCREENMODE:
	  case BADCOLORS:
	  case BADWIDTH:
		sprintf(OUTSTR,
				"ERROR: screen characteristics make screen dump impossible");
		break;

	  case CANTOPEN:
	  case FILEERROR:

		myprint((ScreenDumpError == CANTOPEN) ? "ERROR: can't open file " :
				"ERROR: error writing file ");
		myprint(ScreenDumpFile);
		OUTSTR[0] = '.';
		OUTSTR[1] = '\0';
		break;

	  default:
		;
	}
	putOUTSTR();
} /*screenDumpMsg()*/

long do_whole_graph(graph, fp, termType, windno, pause, landscape,
					screendumpfile)
whole_graph    **graph;
FILE                   *fp;
long                    termType, windno, pause, landscape;
char                   *screendumpfile;
{
	double              value, offset;
	long                size;
	double              xmin = (*graph)->xmin, xmax = (*graph)->xmax;
	double              ymin = (*graph)->ymin, ymax = (*graph)->ymax;
	long                plotOK;
#ifdef MULTIPLOTWINDOWS
	long                prevWindow;
#endif /*MULTIPLOTWINDOWS*/
	WHERE("do_whole_graph");

	if(landscape && termType != PSTERM)
	{
		myerrorout("WARNING: landscape mode ignored except for PostScript output");
		landscape = 0;
	}
	if(termType == PSTERM)
	{
		ps_portrait = !landscape;
	}

	term = termType;
	PLOTFILE = fp;
	if(termType != DUMBTERM)
	{
		value = .02 * (xmax - xmin);
		if(xmin == (xmin - value) || xmax == xmax + value)
		{
			value = 1.0;
		}
		xmin -= value;
		xmax += value;
		value = .02 * (ymax - ymin);
		if(ymin == (ymin - value) || ymax == ymax + value)
		{
			value = 1.0;
		}
		ymin -= value;
		ymax += value;
	}
	else
	{
		size = (SCREENWIDTH > 15) ? SCREENWIDTH - 1 : 79;
		offset = 11.0;
		value = (xmax - xmin) / (size - 3.0 - offset);
		if(value != 0 && fabs(xmin)/value < 1e10 && fabs(xmax)/value < 1e10)
		{
			xmin -= 1.00001*value;
			xmax += 0.99999*value;
		}
		else
		{
			xmin -= 1.0;
			xmax += 1.0;
		}

		size = ((SCREENHEIGHT > 6) ? SCREENHEIGHT : 24) - 1;
		offset = 2.0;
		value = (ymax - ymin) / (size - 3.0 - offset);
		if(value != 0 && fabs(ymin)/value < 1e10 && fabs(ymax)/value < 1e10)
		{
			ymin -= 1.00001*value;
			ymax += 0.99999*value;
		}
		else
		{
			ymin -= 1.0;
			ymax += 1.0;
		}
#ifndef SCROLLABLEWINDOW
		if (BDEPTH > 0 && SCREENHEIGHT > 0)
		{
			NLINES = SCREENHEIGHT;
			incrementNLINES("\n");
		}
#endif /*SCROLLABLEWINDOW*/
	}

#ifdef MULTIPLOTWINDOWS
	prevWindow = ThisWindow;
	if(windno < 0)
	{
		ThisWindow = -1;
	}
	else if(windno > 0)
	{
		ThisWindow = windno - 1;
	}
/* if windno == 0, do not change ThisWindow */
#endif /*MULTIPLOTWINDOWS*/
	PauseAfterPlot = pause;
	
	ScreenDumpError = SCREENDUMPOK;
	ScreenDumpFile = screendumpfile;

	plotOK = !do_plot((*graph)->first_plot, (*graph)->nplots,
					(*graph)->first_string, (*graph)->nstrings,
					xmin, xmax, ymin, ymax,
					(*graph)->logx, (*graph)->logy,
					(*graph)->xaxis, (*graph)->yaxis,
					(*graph)->xlab, (*graph)->ylab, (*graph)->title);
	
#ifdef ACTIVEUPDATE
	if (ActiveUpdate && pause && termType == STANDARDTERM)
	{
		updateConsole();
	}
#endif /*ACTIVEUPDATE*/

	if (screendumpfile != (char *) 0)
	{
		screenDumpMsg();
	}

#ifdef MULTIPLOTWINDOWS
	if(!plotOK)
	{
		ThisWindow = prevWindow;
	}
#endif /*MULTIPLOTWINDOWS*/

	PLOTFILE = (FILE *) 0;
	return (plotOK);

} /*do_whole_graph()*/

/*
  for this to work, as a struct whole_graph is being constructed, it should
  be kept "up-to-date", incrementing (*graphs)->nplots every time a
  (struct curve_points **) plot is added and zeroing NEXTGRAPH(plot) and
  GRAPHPOINTS(plots).  That way, all handles are either zero or are the values
  of valid handles.
*/
void deleteGraph(graph)
whole_graph        **graph;
{
	curve_points       **this_plot, **next_plot;
	plot_string        **this_string, **next_string;
	long                        i, nplots, nstrings;
	WHERE("deleteGraph");

	if(graph != (whole_graph **) 0)
	{
		next_plot = (*graph)->first_plot;
		nplots = (*graph)->nplots;
		for(i = 0;i<nplots;i++)
		{
			this_plot = next_plot;
			next_plot = NEXTGRAPH(this_plot);
			mydisphandle((char **) GRAPHPOINTS(this_plot));
			mydisphandle((char **) this_plot);
		}

		nstrings = (*graph)->nstrings;
		next_string = (*graph)->first_string;
		for(i = 0;i<nstrings;i++)
		{
			this_string = next_string;
			next_string = NEXTSTRING(this_string);
			mydisphandle(PLOTSTRING(this_string));
			mydisphandle((char **) this_string);
		}
		mydisphandle((char **) graph);
	}
} /*deleteGraph()*/

whole_graph **copyGraph(graphIn)
whole_graph        **graphIn;
{
	Symbolhandle         graphTrash = (Symbolhandle) 0;
	whole_graph        **graphOut;
	curve_points       **next_plot = (*graphIn)->first_plot;
	curve_points       **this_plot = (curve_points **) 0;
	curve_points       **out_plot = (curve_points **) 0;
	curve_points       **last_plot = (curve_points **) 0;
	plot_string        **next_string = (*graphIn)->first_string;
	plot_string        **this_string = (plot_string **) 0;
	plot_string        **out_string = (plot_string **) 0;
	plot_string        **last_string = (plot_string **) 0;
	long                 i, nplots = (*graphIn)->nplots;
	long                 nstrings = (*graphIn)->nstrings;
	WHERE("copyGraph");

	graphTrash = Install(SCRATCH,PLOTINFO);
	TMPHANDLE = myduphandle((char **) graphIn);
	if(graphTrash == (Symbolhandle) 0 || TMPHANDLE == (char **) 0)
	{
		goto errorExit;
	}

	graphOut = (whole_graph **) TMPHANDLE;
	(*graphOut)->nplots = 0;
	(*graphOut)->first_plot = (curve_points **) 0;
	(*graphOut)->nstrings = 0;
	(*graphOut)->first_string = (plot_string **) 0;
	setGRAPH(graphTrash,graphOut);
	for(i = 0;i< nplots;i++)
	{
		this_plot = next_plot;
		next_plot = NEXTGRAPH(this_plot);
		TMPHANDLE = myduphandle((char **) this_plot);
		out_plot = (curve_points **) TMPHANDLE;
		if(TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
		if(i == 0)
		{
			(*graphOut)->first_plot = out_plot;
		}
		else
		{
			NEXTGRAPH(last_plot) = out_plot;
		}
		(*graphOut)->nplots++;
		GRAPHPOINTS(out_plot) = (coordinate **) 0;
		NEXTGRAPH(out_plot) = (curve_points **) 0;
		last_plot = out_plot;

		TMPHANDLE = myduphandle((char **) GRAPHPOINTS(this_plot));
		GRAPHPOINTS(out_plot) = (coordinate **) TMPHANDLE;
		if(TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
	} /* for i */

	for(i=0;i<nstrings;i++)
	{
		this_string = next_string;
		next_string = NEXTSTRING(this_string);
		TMPHANDLE = myduphandle((char **) this_string);
		out_string = (plot_string **) TMPHANDLE;
		if(TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
		PLOTSTRING(out_string) = (char **) 0;
		NEXTSTRING(out_string) = (plot_string **) 0;
		(*graphOut)->nstrings++;
		if(i == 0)
		{
			(*graphOut)->first_string = out_string;
		}
		else
		{
			NEXTSTRING(last_string) = out_string;
		}
		last_string = out_string;
		TMPHANDLE = myduphandle((char **) PLOTSTRING(this_string));
		PLOTSTRING(out_string) = TMPHANDLE;
		if(TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
	}

	setGRAPH(graphTrash,(whole_graph **) 0);
	Removesymbol(graphTrash);

	return (graphOut);

  errorExit:
	Removesymbol(graphTrash);

	return (0);
} /*copyGraph()*/


/*
  Check argument list for plotting parameters returning their values.
  StartKey (will always be > 0) is where to start looking.
  The return value is the index of the earliest keyword matched, or 0
  if none is matched.
  -1 is returned if an error is found.
*/

#define ITITLE    0
#define IXLAB     1
#define IYLAB     2
#define IXMIN     3
#define IYMIN     4
#define IXMAX     5
#define IYMAX     6
#define ILOGX     7
#define ILOGY     8
#define IXAXIS    9
#define IYAXIS   10
#define IDUMB    11 /* character plot */
#define IKEEP    12 /* save plot in variable LASTPLOT */
#define ISHOW    13 /* don't display plot if F */
#define IADD     14 /* add to existing plot */
#define IPS      15 /* write PostScript file if T */
#define INEW     16 /* rewind output file if T */
#define IFILE    17 /* write to file */
#define IJUSTIFY 18 /* justification code */
#define IWINDOW  19 /* window number */
#define IPAUSE   20 /* pause after plot if T */
#define IIMPULSE 21 /* draw lines from x-axis or edge to points */
#define ILINES   22 /* connect points with lines */
#define ILINETYPE 23 /* line type, integer */
#define ITHICKNESS 24 /* line thickness, REAL >= 1.0 */
#define ILANDSCAPE 25 /* landscape mode */
#define ISCREENDUMP 26 /* save the screen */
static char      *PlotKeys[] =
{
	"title", "xlab", "ylab", "xmin", "ymin", "xmax", "ymax",
	"logx", "logy", "xaxis", "yaxis", "dumb", "keep", "show", "add",
	"ps", "new", "file", "just", "wind", "pause", "impulse",
	"lines", "linet", "thick", "land","screend",
(char *) 0
};

static long       PlotKeyTypes[] =
{
	CHAR, CHAR, CHAR, REAL, REAL, REAL, REAL,
	LOGIC, LOGIC, LOGIC, LOGIC, LOGIC, LOGIC, LOGIC, LOGIC,
	LOGIC, LOGIC, CHAR, CHAR, REAL, LOGIC, LOGIC,
	LOGIC, REAL, REAL, LOGIC, CHAR,
	0
};

/*
  getPlotKeys returns -1 if an error is found
		       0 if none of the keywords are found
					   j if first keyword is in COMPVALUE(list,j)
*/
long getPlotKeys(list, startKey,  title, xlab, ylab, xmin, ymin, xmax, ymax,
				 logx, logy, xaxis, yaxis, dumb, keep, show, add,
				 ps, new, file,  justify, window, pause,
				 impulse, lines, linetype, thickness, landscape, screendump)
Symbolhandle       list;
long               startKey;
char              *title, *xlab, *ylab, **file;
double            *xmin, *xmax, *ymin, *ymax;
long              *logx, *logy, *xaxis, *yaxis;
long              *dumb, *keep, *show, *add, *ps, *new, *justify, *pause;
long              *window, *linetype;
long              *impulse, *lines;
double            *thickness;
long              *landscape;
char              **screendump;
{
	long           first = 0, keycode, j, type, nargs = NARGS(list);
	double         value;
	Symbolhandle   symhkey;
	char          *keyword, *keyname, c;
	char           legalRange[40];
	WHERE("getPlotKeys");

	for(keycode=0;(keyname = PlotKeys[keycode]) != (char *) 0; keycode++)
	{
		type = PlotKeyTypes[keycode];
		for(j = startKey;j < nargs; j++)
		{
			symhkey = COMPVALUE(list,j);
			if((keyword = isKeyword(symhkey)) &&
			   strncmp(keyname, keyword,strlen(keyname)) == 0)
			{
				if(TYPE(symhkey) != type || !isScalar(symhkey))
				{
					sprintf(OUTSTR,
							"ERROR: value for %s must be %s", keyword,
							(type == CHAR) ?
							"character variable or string" :
							((type == REAL) ? "number or scalar REAL variable":
							 "T or F"));
					goto errorExit;
				}
				if(first == 0 || j < first - 1)
				{
					first = j + 1;
				}

				if(type == REAL || type == LOGIC)
				{
					value = DATAVALUE(symhkey,0);
					if(isMissing(value))
					{
						sprintf(OUTSTR,
								"ERROR: MISSING value for keyword %s on %s",
								keyword, FUNCNAME);
						goto errorExit;
					}
				}

				switch(keycode)
				{
				  case ITITLE:
					if(title == (char *) 0)
					{
						goto illegalKey;
					}
					strncpy(title, STRINGPTR(symhkey),TITLESIZE);
					title[TITLESIZE] = '\0';
					if(strlen(STRINGPTR(symhkey)) > TITLESIZE)
					{
#ifdef MACINTOSH
						title[TITLESIZE] = (unsigned char) 0xc9; /* '...' */
#else /*MACINTOSH*/
						title[TITLESIZE] = (unsigned char) '+';
#endif /*MACINTOSH*/
						title[TITLESIZE+1] = '\0';
					}
					break;

				  case IXLAB:
					/* xlab should have length at least XLABELSIZE+2*/
					if(xlab == (char *) 0)
					{
						goto illegalKey;
					}
					strncpy(xlab, STRINGPTR(symhkey),XLABELSIZE+1);
					xlab[XLABELSIZE+1] = '\0';
					if(strlen(STRINGPTR(symhkey)) > XLABELSIZE+1)
					{ /* add character to show label is incomplete */
#ifdef MACINTOSH
						xlab[XLABELSIZE] = (unsigned char) 0xc9; /* '...' */
#else /*MACINTOSH*/
						xlab[XLABELSIZE] = (unsigned char) '+';
#endif /*MACINTOSH*/
					}
					break;

				  case IYLAB:
					/* ylab should have length at least YLABELSIZE+2*/
					if(ylab == (char *) 0)
					{
						goto illegalKey;
					}
					strncpy(ylab, STRINGPTR(symhkey),YLABELSIZE+1);
					ylab[YLABELSIZE+1] = '\0';
					if(strlen(STRINGPTR(symhkey)) > YLABELSIZE+1)
					{ /* add character to show label is incomplete */
#ifdef MACINTOSH
						ylab[YLABELSIZE] = (unsigned char) 0xc9; /* '...' */
#else /*MACINTOSH*/
						ylab[YLABELSIZE] = (unsigned char) '+';
#endif /*MACINTOSH*/
					}
					break;

				  case IXMIN:
					if(xmin == (double *) 0)
					{
						goto illegalKey;
					}
					*xmin = value;
					break;

				  case IXMAX:
					if(xmax == (double *) 0)
					{
						goto illegalKey;
					}
					*xmax = value;
					break;

				  case IYMIN:
					if(ymin == (double *) 0)
					{
						goto illegalKey;
					}
					*ymin = value;
					break;

				  case IYMAX:
					if(ymax == (double *) 0)
					{
						goto illegalKey;
					}
					*ymax = value;
					break;

				  case ILOGX:
					if(logx == (long *) 0)
					{
						goto illegalKey;
					}
					*logx = (value != 0.0);
					break;

				  case ILOGY:
					if(logy == (long *) 0)
					{
						goto illegalKey;
					}
					*logy = (value != 0.0);
					break;

				  case IXAXIS:
					if(xaxis == (long *) 0)
					{
						goto illegalKey;
					}
					*xaxis = (value != 0.0);
					break;

				  case IYAXIS:
					if(yaxis == (long *) 0)
					{
						goto illegalKey;
					}
					*yaxis = (value != 0.0);
					break;

				  case IDUMB:
					if(dumb == (long *) 0)
					{
						goto illegalKey;
					}
					*dumb = (value != 0.0);
					break;

				  case IKEEP:
					if(keep == (long *) 0)
					{
						goto illegalKey;
					}
					*keep = (value != 0.0);
					break;

				  case ISHOW:
					if(show == (long *) 0)
					{
						goto illegalKey;
					}
					*show = (value != 0.0);
					break;

				  case IADD:
					if(add == (long *) 0)
					{
						goto illegalKey;
					}
					*add = (value != 0.0);
					break;

				  case IPS:
					if(ps == (long *) 0)
					{
						goto illegalKey;
					}
					*ps = (value != 0.0);
					break;

				  case INEW:
					if(new == (long *) 0)
					{
						goto illegalKey;
					}
					*new = (value != 0.0);
					break;

				  case IFILE:
					if(file == (char **) 0)
					{
						goto illegalKey;
					}
					if(*file != (char *) 0)
					{
						goto alreadySet;
					}
					*file = STRINGPTR(symhkey);
					break;

				  case IJUSTIFY:
					if(justify == (long *) 0)
					{
						goto illegalKey;
					}
					c = STRINGVALUE(symhkey,0);

					switch ((int) c)
					{
					  case 'c':
					  case 'C':
						*justify = (long) CENTRE;
						break;
					  case 'l':
					  case 'L':
						*justify = LEFT;
						break;
					  case 'r':
					  case 'R':
						*justify = (long) RIGHT;
						break;
					  default:
						goto badString;
					}
					break;

				  case IWINDOW:
					if(window == (long *) 0)
					{
						goto illegalKey;
					}
#ifdef MULTIPLOTWINDOWS
					if(value != floor(value) || value < 0 || value > NGRAPHS)
					{
						sprintf(legalRange,"integer between 0 and %ld",
								(long) NGRAPHS);
						goto badNumber;
					}
					*window = (long) value;
#endif /*MULTIPLOTWINDOWS*/
					break;

				  case IPAUSE:
					if(pause == (long *) 0)
					{
						goto illegalKey;
					}
					*pause = (value != 0.0);
					break;

				  case IIMPULSE:
					if(impulse == (long *) 0)
					{
						goto illegalKey;
					}
					*impulse = (value != 0.0);
					break;

				  case ILINES:
					if(lines == (long *) 0)
					{
						goto illegalKey;
					}
					*lines = (value != 0.0);
					break;

				  case ILINETYPE:
					if(linetype == (long *) 0)
					{
						goto illegalKey;
					}
					if(value != floor(value) || value <= 0 ||
					   value > (double) MAXLINETYPE)
					{
						sprintf(legalRange,"positive integer <= %ld",
								(long) MAXLINETYPE);
						goto badNumber;
					}

					*linetype = (long) value;
					break;

				  case ITHICKNESS:
					if(thickness == (double *) 0)
					{
						goto illegalKey;
					}
					*thickness = value;
					if(value < .1 || value > 10.0)
					{
						sprintf(legalRange,"between 0.1 and 10.0");
						goto badNumber;
					}
					break;

				  case ILANDSCAPE:
					if(landscape == (long *) 0)
					{
						goto illegalKey;
					}
					*landscape = (value != 0.0);
					break;

				  case ISCREENDUMP:
					if(screendump == (char **) 0)
					{
						goto illegalKey;
					}
					if(*screendump != (char *) 0)
					{
						goto alreadySet;
					}
					*screendump = STRINGPTR(symhkey);
					break;

				} /*switch(keycode)*/
				setNAME(symhkey,USEDKEYWORD);
				break;
			} /* if keyword matches keyname*/
		} /*for(j = startKey;j < nargs; j++)*/
	} /*for(keycode=0;(keyname = PlotKeys[keycode]) != (char *) 0; keycode++)*/

	return (first);

  illegalKey:
	sprintf(OUTSTR,"ERROR: '%s' is not legal keyword for %s",keyword,FUNCNAME);
	goto errorExit;

  alreadySet:
	sprintf(OUTSTR,"ERROR: %s specified more than once for %s",keyword,
			FUNCNAME);
	goto errorExit;

  badString:
	sprintf(OUTSTR,"ERROR: '%s' is not legal value for %s",STRINGPTR(symhkey),
			keyword);
	goto errorExit;

  badNumber:

	sprintf(OUTSTR,"ERROR: value for %s must be %s",keyword,legalRange);
	/* fall through */
  errorExit:
	putOUTSTR();

	return (-1);
} /*getPlotKeys()*/

static long saveGraphHandle(whole_graph **graph)
{
/*
  whole_graph info is always saved in ascii form
*/
	if(!saveLong((*graph)->nplots)   ||
	   !saveLong((*graph)->nstrings) ||
	   !saveDouble((*graph)->xmin)   ||
	   !saveDouble((*graph)->xmax)   ||
	   !saveDouble((*graph)->ymin)   ||
	   !saveDouble((*graph)->ymax)   ||
	   !saveLong((*graph)->logx)     ||
	   !saveLong((*graph)->logy)     ||
	   !saveLong((*graph)->xaxis)    ||
	   !saveLong((*graph)->yaxis)    ||
	   !saveString((*graph)->xlab)   ||
	   !saveString((*graph)->ylab)   ||
	   !saveString((*graph)->title))
	{
		return(0);
	}
	return (1);
} /*saveGraphHandle()*/

/*
  non-handle components of curve_points are always saved in ascii form
  with each component name followed by a value.  Labels and titles are
  saved only if non-null, in which case there are two values, the string
  length and the string.
*/

static long saveCurveHandle(curve_points **curve)
{
	long      plotStyle = PLOTSTYLE(curve);
	long      lineType = LINETYPE(curve);

	if(ASCII < SAVEASCII4 && ASCII > SAVEBINARY4)
	{
		if(plotStyle == POINTS)
		{
			plotStyle = POINTS_V31;
		}
		else if(plotStyle == CHARS)
		{
			plotStyle = CHARS_V31;
		}
		else if(plotStyle == LINES)
		{
			plotStyle = LINES_V31;
		}
		else if((plotStyle & IMPULSES) != 0)
		{
			plotStyle = IMPULSES_V31;
		}
		else if(plotStyle == BOTHLANDP)
		{
			plotStyle = BOTHLANDP_V31;
		}
		else if(plotStyle == BOTHLANDC)
		{
			plotStyle = BOTHLANDC_V31;
		}
		else
		{
			plotStyle = POINTS_V31;
		}
		lineType %= MAXLINETYPE;
	}

	if(!saveLong(plotStyle) ||
	   !saveLong(POINTTYPE(curve)) ||
	   !saveLong(lineType)  ||
	   !saveLong(PCOUNT(curve)))
	{
		return(0);
	}

	return (1);
} /*saveCurveHandle()*/

/* these are also defined in save.c */
#define LONGHANDLE    1
#define DOUBLEHANDLE  2
#define CHARHANDLE    3

static long savePoints(curve_points **curve)
{
	long             npoints = PCOUNT(curve), i;

	if(Binary)
	{ /* binary format */
		if(saveHandle((char **) GRAPHPOINTS(curve), CHARHANDLE) == 0)
		{
			return (0);
		}
	} /*if(Binary)*/
	else
	{
		for(i=0;i<npoints;i++)
		{
			if(!saveLong(POINTUNDEF(curve,i)))
			{
				return(0);
			}
			if(!saveDouble(XCOORD(curve,i)))
			{
				return(0);
			}
			if(!saveDouble(YCOORD(curve,i)))
			{
				return(0);
			}

			if(!saveString(POINTSTRING(curve,i)))
			{
				return (0);
			}
		} /*for(i=0;i<npoints;i++)*/
	} /*if(Binary){}else{}*/
	return (1);
} /*savePoints()*/

static long saveStringsHandle(plot_string **string)
{
	if(!saveLong(JUSTIFYSTR(string)) ||
	   !saveDouble(STRINGX(string))  ||
	   !saveDouble(STRINGY(string))  ||
	   !saveString(*PLOTSTRING(string)))
	{
		return(0);
	}
	return (1);
} /*saveStringsHandle()*/

/* save to KEEPFILE using global flag ASCII */
long saveGraph(whole_graph **graph)
{
	long           i, nplots = (*graph)->nplots;
	long           nstrings = (*graph)->nstrings;
	curve_points **next_plot;
	curve_points **this_plot;
	plot_string  **next_string;
	plot_string  **this_string;

	if(!saveGraphHandle(graph))
	{
		return (0);
	}
	next_plot = (*graph)->first_plot;
	for(i=0;i<nplots;i++)
	{
		this_plot = next_plot;
		next_plot = NEXTGRAPH(this_plot);

		if(!saveCurveHandle(this_plot) || !savePoints(this_plot))
		{
			return (0);
		}
	}
	next_string = (*graph)->first_string;
	for(i=0;i<nstrings;i++)
	{
		this_string = next_string;
		next_string = NEXTSTRING(this_string);
		if(!saveStringsHandle(this_string))
		{
			return (0);
		}
	}

	return (1);
} /*saveGraph()*/

static long restoreGraphHandle(whole_graph ** graph)
{
	WHERE("restoreGraphHandle");

	(*graph)->first_plot = (curve_points **) 0;
	if(!restoreLong(&(*graph)->nplots)   ||
	   !restoreLong(&(*graph)->nstrings) ||
	   !restoreDouble(&(*graph)->xmin)   ||
	   !restoreDouble(&(*graph)->xmax)   ||
	   !restoreDouble(&(*graph)->ymin)   ||
	   !restoreDouble(&(*graph)->ymax)   ||
	   !restoreLong(&(*graph)->logx)     ||
	   !restoreLong(&(*graph)->logy)     ||
	   !restoreLong(&(*graph)->xaxis)    ||
	   !restoreLong(&(*graph)->yaxis)    ||
	   !restoreString((*graph)->xlab)    ||
	   !restoreString((*graph)->ylab)    ||
	   !restoreString((*graph)->title))
	{
		return (0);
	}

	(*graph)->xlab[XLABELSIZE+1] = '\0';
	(*graph)->ylab[YLABELSIZE+1] = '\0';
	(*graph)->title[TITLESIZE+1] = '\0';

	return (1);
} /*restoreGraphHandle()*/

static long restoreCurveHandle(curve_points ** curve)
{
	long             style;

	if(!restoreLong(&style) ||
	   !restoreLong(&POINTTYPE(curve)) ||
	   !restoreLong(&LINETYPE(curve))  ||
	   !restoreLong(&PCOUNT(curve)))
	{
		return(0);
	}

	if(ASCII < SAVEASCII4 && ASCII > SAVEBINARY4)
	{
		if(style == LINES_V31)
		{
			style = LINES;
		}
		else if(style == IMPULSES_V31)
		{
			style = IMPULSES;
		}
		else if(style == CHARS_V31)
		{
			style = CHARS;
		}
		else if(style == BOTHLANDP_V31)
		{
			style = BOTHLANDP;
		}
		else if(style == BOTHLANDC_V31)
		{
			style = BOTHLANDC;
		}
		else
		{ /* case of doubt */
			style = POINTS;
		}
	} /*if(ASCII < SAVEASCII4 && ASCII > SAVEBINARY4)*/

	PLOTSTYLE(curve) = style;

	return (1);
} /*restoreCurveHandle()*/

static long restorePoints(curve_points ** curve)
{
	long             npoints = PCOUNT(curve), i, n;
	int              c;
	char             string[100];
	char           **hand;
	WHERE("restorePoints");
	
	if(Binary)
	{
		n = npoints * sizeof(coordinate);
		hand = (char **) GRAPHPOINTS(curve);
		for(i=0;i<n;i++)
		{
			if((c = fgetc(RESTOREFILE)) == EOF)
			{
				return (0);
			}
			(*hand)[i] = (char) c;
		}
	} /*if(Binary)*/
	else
	{ /* ASCII */
		for(i = 0;i<npoints;i++)
		{
			if(!restoreLong(&POINTUNDEF(curve,i)) ||
			   !restoreDouble(&XCOORD(curve,i))   ||
			   !restoreDouble(&YCOORD(curve,i))   ||
			   !restoreString(string) || strlen(string) > PLOTSTRINGSIZE)
			{
				return (0);
			}
			strcpy(POINTSTRING(curve,i),string);
		} /* for i */
	} /*if(Binary){...}else{...}*/
	return (1);
} /*restorePoints()*/

static long restoreStringsHandle(plot_string ** string)
{
	long             length;
	WHERE("restoreStringsHandle");

	if(!restoreLong(&JUSTIFYSTR(string)) ||
	   !restoreDouble(&STRINGX(string))  ||
	   !restoreDouble(&STRINGY(string))  ||
	   !restoreLong(&length))
	{
		return(0);
	}

	TMPHANDLE = mygethandle(length+1);
	PLOTSTRING(string) = TMPHANDLE;
	if(TMPHANDLE == 0 || length > 0 && !restoreNString(length,*TMPHANDLE))
	{
		return (0);
	}
	(*TMPHANDLE)[length] = '\0';

	return (1);
} /*restoreStringsHandle()*/

/* restore from RESTOREFILE using global flag ASCII */
char **restoreGraph(void)
{
	Symbolhandle       symhgraph = (Symbolhandle) 0;
	whole_graph      **graph = (whole_graph **) 0;
	curve_points     **last_plot = (curve_points **) 0;
	curve_points     **this_plot = (curve_points **) 0;
	plot_string      **last_string = (plot_string **) 0;
	plot_string      **this_string = (plot_string **) 0;
	long               i, nplots, npoints, nstrings;
	WHERE("restoreGraph");

	symhgraph = Install(SCRATCH,PLOTINFO); /* temporary holder for graph */
	if(symhgraph == (Symbolhandle) 0 ||
	   (TMPHANDLE = mygethandle(sizeof(whole_graph))) == (char **) 0)
	{
		goto errorExit;
	}
	graph = (whole_graph **) TMPHANDLE;
	(*graph)->first_plot = (curve_points **) 0;
	(*graph)->nplots = 0;
	(*graph)->first_string = (plot_string **) 0;
	(*graph)->nstrings = 0;
	setGRAPH(symhgraph,graph);
	if(!restoreGraphHandle(graph))
	{
		(*graph)->nplots = (*graph)->nstrings = 0;
		goto errorExit;
	}

	nplots = (*graph)->nplots;
	nstrings = (*graph)->nstrings;
	(*graph)->nstrings = (*graph)->nplots = 0;

	for (i=0;i<nplots;i++)
	{
		if((TMPHANDLE = mygethandle(sizeof(curve_points))) ==
		   (char **) 0)
		{
			goto errorExit;
		}
 		this_plot = (curve_points **) TMPHANDLE;
		NEXTGRAPH(this_plot) = (curve_points **) 0;
		GRAPHPOINTS(this_plot) = (coordinate **) 0;
		(*graph)->nplots++;
		if(i == 0)
		{
			(*graph)->first_plot = this_plot;
		}
		else
		{
			NEXTGRAPH(last_plot) = this_plot;
		}
		if(!restoreCurveHandle(this_plot))
		{
			goto errorExit;
		}
		npoints = PCOUNT(this_plot);
		TMPHANDLE = mygethandle(npoints * sizeof(coordinate));
		GRAPHPOINTS(this_plot) = (coordinate **) TMPHANDLE;
		if(TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
		if(!restorePoints(this_plot))
		{
			goto errorExit;
		}
		last_plot = this_plot;
	} /*for (i=0;i<nplots;i++)*/

	for(i=0;i<nstrings;i++)
	{
		if((TMPHANDLE = mygethandle(sizeof(plot_string))) ==
		   (char **) 0)
		{
			goto errorExit;
		}
		this_string = (plot_string **) TMPHANDLE;
		NEXTSTRING(this_string) = (plot_string **) 0;
		PLOTSTRING(this_string) = (char **) 0;
		(*graph)->nstrings++;
		if(i == 0)
		{
			(*graph)->first_string = this_string;
		}
		else
		{
			NEXTSTRING(last_string) = this_string;
		}
		if(!restoreStringsHandle(this_string))
		{
			goto errorExit;
		}
		last_string = this_string;
	}

	/* cut link from temporary symbol and remove temporary symbol */
	setGRAPH(symhgraph,(whole_graph **) 0);
	Removesymbol(symhgraph);
	return ((char **) graph);

  errorExit:
	Removesymbol(symhgraph);

	return (0);
} /*restoreGraph()*/

FILE *openPlotFile(fileName,new)
char          *fileName;
long           new;
{
	FILE      *plotfile;
	char      *mode;

	mode = (new) ? "w" : "a";
	if((plotfile = fopen(fileName,mode)) == (FILE *) 0)
	{
		sprintf(OUTSTR,"ERROR: %s cannot open file %s",FUNCNAME,fileName);
		putOUTSTR();
	}
	return (plotfile);
} /*openPlotFile()*/

/*
   Function to return the total number of bytes used in a curve_points
   structure
*/

static long sizeofC_P(curve_points ** points)
{
	long          size = 0;
	
	
	if (points != (curve_points **) 0)
	{
		size = sizeof(curve_points) + (*points)->p_count * sizeof(coordinate);
	}
	return (size);
} /*sizeofC_P()*/

/*
   Function to return the total number of bytes used in a plot_string
   structure
*/

static long sizeofP_S(plot_string ** strings)
{
	long          size = 0;
	
	if (strings != (plot_string **) 0)
	{
		size = sizeof(plot_string) +
			((*strings)->string != (Char **) 0) ?
				myhandlelength((*strings)->string) : 0;
	}
	return (size);
} /*sizeofP_S()*/

/*
   Function to return the total number of bytes used in a whole_graph
   structure
*/

long sizeofPlot(whole_graph ** graph)
{
	long           size = 0;
	curve_points **this_plot;
	plot_string  **this_string;

	if (graph != (whole_graph **) 0)
	{
		size = sizeof(whole_graph);
		for (this_plot = (*graph)->first_plot;
			 this_plot != (curve_points **) 0;
			 this_plot = (*this_plot)->next_cp)
		{
			size += sizeofC_P(this_plot);
		}
		for (this_string = (*graph)->first_string;
			 this_string != (plot_string **) 0;
			 this_string = (*this_string)->next_string)
		{
			size += sizeofP_S(this_string);
		}
	} /*if (graph != (whole_graph **) 0)*/
	return (size);
} /*sizeofPlot()*/

#ifdef DJGPP
/* Function to writeout screen in PCX format */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <grx.h>
#include <mousex.h>
#include <io.h>

typedef unsigned char uchar;
typedef unsigned short ushort;

typedef struct
{
	uchar           maker;
	uchar           version;
	uchar           code;
	uchar           bpp;    /* bits per pixel */
	ushort          x1, y1, x2, y2; /* image position */
	ushort          hres, vres;     /* image size */
	struct
	{
		uchar           R, G, B;
	}               cmap[16];       /* palette */
	uchar           vmode;  /* video mode to display it */
	uchar           nplanes;/* number of planes */
	ushort          bpl;    /* bytes per scan line */
	char            filler[128 - 68];
}               pcxhdr;

#if     defined(__GNUC__) && defined(__MSDOS__)
#define MAX_COLORS     32768    /* maximum color graphics mode supported */
#else
#define MAX_COLORS     256      /* maximum color graphics mode supported */
#endif
#define MAX_OUTCOLORS  16       /* max number of simultaneous colors */
#define MAX_WIDTH      1280     /* max horizontal image size */


static int      DumpCount = 0;

short      screenDump(char *fname)
{
	GrContext       save, xfer;
	pcxhdr          hdr;
	uchar           rows[4][MAX_WIDTH / 8];
	char            colortable[MAX_COLORS];
	char            pcxfname[100];
	FILE           *pcxfile;
	int             numcolors, block, fast;
	int             ii, xx, yy, wdt, hgt;

	if (GrCurrentMode() < GR_320_200_graphics)
	{
		return (BADSCREENMODE);
	}
	if (GrNumColors() > MAX_COLORS)
	{
		return (BADCOLORS);
	}
	wdt = GrScreenX();
	hgt = GrScreenY();
	if (wdt > MAX_WIDTH)
	{
		return (BADWIDTH);
	}
	if (fname == NULL)
	{
		for (;;)
		{
			/* find a new file name */
			sprintf(pcxfname, "dump%04d.pcx", DumpCount);
			if (access(pcxfname, 0) != 0)
			{
				pcxfile = fopen(pcxfname, "wb");
				if (pcxfile == NULL)
				{
					return (CANTOPEN);
				}
				break;
			}
			if (++DumpCount >= 10000)
			{
				return (TOOMANYDUMPS);
			}
		}
	}
	else
	{
		strcpy(pcxfname, fname);
		pcxfile = fopen(pcxfname, "wb");
		if (pcxfile == NULL)
		{
			return (CANTOPEN);
		}
	}
	GrSaveContext(&save);
	GrCreateContext(wdt, 1, NULL, &xfer);
	GrSetContext(&xfer);
	memset(&hdr, 0, sizeof(pcxhdr));
	memset(colortable, -1, sizeof(colortable));
	hdr.maker = 10;
	hdr.version = 5;
	hdr.code = 1;
	hdr.x1 = 0;
	hdr.y1 = 0;
	hdr.x2 = wdt - 1;
	hdr.y2 = hgt - 1;
	hdr.hres = wdt;
	hdr.vres = hgt;
	hdr.bpp = 1;
	hdr.bpl = (wdt + 7) / 8;
	hdr.vmode = 0x12;       /* standard 640x480 16 color VGA ! */
	hdr.nplanes = 4;
	numcolors = 0;
	fast = (GrNumColors() == 256) ? 1 : 0;
	block = MouseCursorIsDisplayed();
	if (block)
	{
		MouseEraseCursor();
	}
	fwrite(&hdr, sizeof(pcxhdr), 1, pcxfile);
	for (yy = 0; yy < hgt; yy++)
	{
		uchar           mask = 0x80;
		int             bytepos = 0;
		int             numbits = 0;
		GrBitBlt(&xfer, 0, 0, &save, 0, yy, wdt - 1, yy, GrWRITE);
		memset(rows, 0, sizeof(rows));
		for (xx = 0; xx < wdt; xx++)
		{
			int             color = fast ? xfer.gc_baseaddr[xx] : GrPixel(xx, 0);
			char            outcolor = colortable[color];
			if (outcolor < 0)
			{       /* new color */
				int             r, g, b;
				GrQueryColor(color, &r, &g, &b);
				for (ii = 0; ii < numcolors; ii++)
				{
					if (hdr.cmap[ii].R != r)
					{
						continue;
					}
					if (hdr.cmap[ii].G != g)
					{
						continue;
					}
					if (hdr.cmap[ii].B != b)
					{
						continue;
					}
					outcolor = ii;
					break;
				}
				if (outcolor < 0)
				{
					if (numcolors >= MAX_OUTCOLORS)
					{
						goto error;
					}
					outcolor = ii = numcolors++;
					hdr.cmap[ii].R = r;
					hdr.cmap[ii].G = g;
					hdr.cmap[ii].B = b;
				}
				colortable[color] = outcolor;
			}
			if (outcolor & 1)
			{
				rows[0][bytepos] |= mask;
			}
			if (outcolor & 2)
			{
				rows[1][bytepos] |= mask;
			}
			if (outcolor & 4)
			{
				rows[2][bytepos] |= mask;
			}
			if (outcolor & 8)
			{
				rows[3][bytepos] |= mask;
			}
			mask >>= 1;
			if (++numbits == 8)
			{
				mask = 0x80;
				numbits = 0;
				bytepos++;
			}
		}
		for (ii = 0; ii < 4; ii++)
		{
			uchar          *ptr = rows[ii];
			uchar           pixval = *ptr;
			numbits = 0;
			for (bytepos = hdr.bpl; --bytepos >= 0; ptr++)
			{
				if (pixval != *ptr)
				{
					while (numbits > 0x3f)
					{
						putc((0xc0 | 0x3f), pcxfile);
						putc(pixval, pcxfile);
						numbits -= 0x3f;
					}
					if (numbits > 0)
					{
						if ((numbits > 1) || (pixval >= 0xc0))
						{
							putc((0xc0 | numbits), pcxfile);
						}
						putc(pixval, pcxfile);
					}
					numbits = 0;
				}
				pixval = *ptr;
				numbits++;
			}
			while (numbits > 0x3f)
			{
				putc((0xc0 | 0x3f), pcxfile);
				putc(pixval, pcxfile);
				numbits -= 0x3f;
			}
			if (numbits > 0)
			{
				if ((numbits > 1) || (pixval >= 0xc0))
				{
					putc((0xc0 | numbits), pcxfile);
				}
				putc(pixval, pcxfile);
			}
		}
	}
	if (ferror(pcxfile))
	{
		goto error;
	}
	if (block)
	{
		MouseDisplayCursor();
	}
	GrSetContext(&save);
	GrDestroyContext(&xfer);
	fseek(pcxfile, 0L, 0);
	fwrite(&hdr, sizeof(pcxhdr), 1, pcxfile);
	fclose(pcxfile);
	return (SCREENDUMPOK);

  error:
	if (block)
	{
		MouseDisplayCursor();
	}
	GrSetContext(&save);
	GrDestroyContext(&xfer);
	fclose(pcxfile);
	unlink(pcxfname);
	return (FILEERROR);
} /*pcxdump()*/
#endif /*DJGPP*/

