/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Power
#else
#define __SEG__ Power
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

extern Symbolhandle  RInstall();
extern void          badNargs();
extern long          argOK();
extern long          isScalar();

#define CRD     1
#define RBD     2

/*
  samplesize(noncen,ngrp,alpha,pwr [,design:"crb")
*/

#define         MAXN 127.0
Symbolhandle    samplesize(list)
Symbolhandle    list;
{
	Symbolhandle    result, args[5];
	long            i;
	double          tmp, left, right, noncen, alpha, powr,trialPower;
	double          numDf, denomDf;
	double          Qsne(), noncentF();
	double          ngrp;
	double          critval;
	long            design = CRD;
	long            nargs = NARGS(list);
	char           *keyword, *designName = "crd";
	char           *usage = "usage: samplesize(noncen,ngrp,alpha,pwr [,design:\"rbd\"])";
	WHERE("samplesize");

	SETUPINT(interruptExit);

	if(nargs < 4)
	{
		badNargs(FUNCNAME,-1000 - 4);
		goto errorExit;
	}
	else if(nargs > 5)
	{
		badNargs(FUNCNAME,-5);
		goto errorExit;
	}
	for(i = 0;i< nargs;i++)
	{
		args[i] = COMPVALUE(list,i);
		if(!argOK(args[i],(i < 4) ? REAL : CHAR,i+1))
		{
			goto errorExit;
		}
		if(i == 4)
		{
			break;
		}
		
		if(!isScalar(args[i]))
		{
			sprintf(OUTSTR,"ERROR: argument %ld to %s not single REAL value",
					i+1,FUNCNAME);
			goto errorExit;
		}
	}
	if(nargs == 5)
	{
		if((keyword = isKeyword(args[4])) && strcmp(keyword,"design") != 0)
	{
			badKeyword(FUNCNAME,keyword);
			goto errorExit;
		}
		if(!isScalar(args[4]))
		{
			sprintf(OUTSTR,
					"ERROR: design name must be single character variable or string");
			goto errorExit;
		}
		designName = STRINGPTR(args[4]);
		if(strcmp(designName,"crd") == 0 || strcmp(designName,"cr") == 0)
		{
			design = CRD;
		}
		else if(strcmp(designName,"rbd") == 0 || strcmp(designName,"crd") == 0)
		{
			design = RBD;
		}
		else
		{
			sprintf(OUTSTR,"ERROR: design name %s not recognized by %s",
					designName,FUNCNAME);
			goto errorExit;
		}
	}
	
	if ((noncen = DATAVALUE(args[0],0)) <= 0.0)
	{
		myerrorout("ERROR: noncentrality parameter must be positive");
		goto errorExit;
	}

	if ((ngrp = DATAVALUE(args[1],0)) < ((design == CRD) ? 1.0 : 2.0) || 
		ngrp != floor(ngrp))
	{
		sprintf(OUTSTR,"ERROR: number of groups for %s must be integer >= %ld",
				FUNCNAME,(design == CRD) ? 1L : 2L);
		goto errorExit;
	}
	
	if ((alpha = DATAVALUE(args[2],0)) <= 0.0 || alpha >= 1.0)
	{
		myerrorout("ERROR: significance level must be between 0 and 1");
		goto errorExit;
	}

	if ((powr = DATAVALUE(args[3],0)) <= 0.0 || powr >= 1.0)
	{
		myerrorout("ERROR: power must be between 0 and 1");
		goto errorExit;
	}

	right = 1.;
	numDf = (ngrp > 1.0) ? ngrp - 1.0 : 1.0;

	do
	{
		if (interrupted() != INTNOTSET)
		{
			goto interruptExit;
		}
		if (right >= MAXN)
		{
			break;
		}
		right *= 2.0;
		denomDf = (right - 1.0) * ((design == CRD) ? ngrp : (ngrp - 1.0));
		critval = Qsne(1 - alpha, numDf, denomDf);
		trialPower = 1.0 - noncentF(critval, right * noncen, numDf, denomDf);
	}while (trialPower < powr);


	if(trialPower > powr)
	{
		left = floor(right / 2. + .1);

		while (right - left > 1.1)
		{
			if (interrupted() != INTNOTSET)
			{
				goto interruptExit;
			}
			tmp = floor((right + left) / 2.0 + .1);
			denomDf = (tmp - 1.0) * ((design == CRD) ? ngrp : (ngrp - 1.0));
			critval = Qsne(1 - alpha, numDf, denomDf);
			trialPower = noncentF(critval,tmp * noncen, numDf, denomDf);
			if(trialPower > 1.0)
			{ /* not fully converged, ignore */
				trialPower -= floor(trialPower);
			}
			trialPower = 1.0 - trialPower;
			if (trialPower < powr)
			{
				left = tmp;
			}
			else
			{
				right = tmp;
			}
		} /* while */
	}
	else
	{
		sprintf(OUTSTR,"WARNING: samplesize truncated at %4.0f",right);
		putOUTSTR();
	}

	result = RInstall(SCRATCH, 1L);

	if (result != (Symbolhandle) 0)
	{
		DATAVALUE(result,0) = right;
	}
	
	RETURN (result);

  errorExit:
	putOUTSTR();

	myerrorout(usage);

  interruptExit:
	
	RETURN (0);

}

#define IPOWER  1
#define IPOWER2 2

/* 
  power(noncen,ngrp,alpha,groupsize)
  power2(noncen2,ndf,alpha,ddf)
*/
/* name changed for use in MPW where pow is equivalenced to power*/
Symbolhandle    power1(list)
Symbolhandle    list;
{
	Symbolhandle    args[5], result = (Symbolhandle) 0;
	double          noncen, ngrp, alpha, groupSize, powr;
	double          Qsne(), noncentF();
	double          numDf, denomDf, critval;
	long            nargs = NARGS(list);
	long            op, design = CRD;
	long            nvals = 1, inc[4], ii[4];
	char           *keyword, *designName = "crd";
	char           *usage;
	long            i, j, length;
	WHERE("power1");
	
	*OUTSTR = '\0';
	if(strcmp(FUNCNAME,"power") == 0)
	{
		op = IPOWER;
		usage = "power(noncen,ngroups,alpha,groupsize [,design:\"rbd\"])";
		if(nargs < 4 || nargs > 5)
		{
			badNargs(FUNCNAME,(nargs < 4) ? -1004 : -5);
			goto errorExit;
		}
	}
	else
	{
		op = IPOWER2;
		usage = "power2(noncen2,numDf,alpha,denomDf)";
		if(nargs != 4)
		{
			badNargs(FUNCNAME,4);
			goto errorExit;
		}
	}

	for(j=0;j<nargs;j++)
	{
		args[j] = COMPVALUE(list,j);
		if(!argOK(args[j],(j < 4) ? REAL : CHAR ,j+1))
		{
			goto errorExit;
		}
		if(j < 4)
		{
			ii[j] = 0;
			inc[j] = 0;
			if(!isVector(args[j]))
			{
				sprintf(OUTSTR,
						"ERROR: argument %ld to %s must be scalar or vector",
						j+1,FUNCNAME);
				goto errorExit;
			}
			length = DIMVAL(args[j],1);
			if(j == 0)
			{
				nvals = length;
			}
			else if(length > 1)
			{
				if(nvals > 1 && length != nvals)
				{
					sprintf(OUTSTR,
							"ERROR:  argument %s to %s is vector with non-matchich length",
							j+1,FUNCNAME);
					goto errorExit;
				}
				nvals = length;
			}
			inc[j] = (length == 1) ? 0 : 1;
		}
		else
		{ /* must be power() */
			if((keyword = isKeyword(args[4])) && 
			   strcmp(keyword,"design") != 0)
			{
				badKeyword(FUNCNAME,keyword);
				goto errorExit;
			}
			if(!isScalar(args[4]))
			{
				sprintf(OUTSTR,
						"ERROR: design name must be single character variable or string");
				goto errorExit;
			}
			designName = STRINGPTR(args[4]);
			if(strcmp(designName,"crd") == 0 || strcmp(designName,"cr") == 0)
			{
				design = CRD;
			}
			else if(strcmp(designName,"rbd") == 0 || strcmp(designName,"rb") == 0)
			{
				design = RBD;
			}
			else
			{
				sprintf(OUTSTR,"ERROR: design name %s not recognized by %s",
						designName,FUNCNAME);
				goto errorExit;
			}
		}			
	} /* for j */

	length = (inc[0] > 0) ? nvals : 1;
	for(i=0;i<length;i++)
	{
		if ((noncen = DATAVALUE(args[0],i)) < 0.0)
		{
			sprintf(OUTSTR,
					"ERROR: noncentrality parameter%s must be non-negative",
					(length > 1) ? "s": "");
			goto errorExit;
		}
	} /* for i */

	length = (inc[1] > 0) ? nvals : 1;
	for(i=0;i<length;i++)
	{
		if ((numDf = DATAVALUE(args[1],i)) < 1.0 ||
			numDf != floor(numDf))
		{
			sprintf(OUTSTR,
					"ERROR: %s is not positive integer",
					(op == IPOWER) ? "number of groups" :"numerator df");
			goto errorExit;
		}
	}						/* for i */

	length = (inc[2] > 0) ? nvals : 1;
	for(i=0;i<length;i++)
	{
		if ((alpha = DATAVALUE(args[2],i)) <= 0.0 || alpha >= 1.0)
		{
			sprintf(OUTSTR,
					"ERROR: significance level%s must be between 0 and 1",
					(length > 1) ? "s": "");
			goto errorExit;
		}
	} /* for i */

	length = (inc[3] > 0) ? nvals : 1;
	for(i=0;i<length;i++)
	{
		if ((denomDf = DATAVALUE(args[3],i)) < 1.0 ||
			denomDf != floor(denomDf) ||
			(op == IPOWER && denomDf < 2.0))
		{
			sprintf(OUTSTR,
					"ERROR: %s is not positive integer%s",
					(op == IPOWER) ? "number of groups or treatments" :
					"denominator df",
					(op == IPOWER) ? " > 1" : "");
			goto errorExit;
		}
	}						/* for i */


	groupSize = 1;

	result = RInstall(SCRATCH, nvals);

	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	for(i=0;i<nvals;i++)
	{
		noncen = DATAVALUE(args[0],ii[0]);
		ii[0] += inc[0];
		numDf = DATAVALUE(args[1],ii[1]);
		ii[1] += inc[1];
		alpha = DATAVALUE(args[2],ii[2]);
		ii[2] += inc[2];
		denomDf = DATAVALUE(args[3],ii[3]);
		ii[3] += inc[3];
		if(op == IPOWER)
		{
			ngrp = numDf;
			groupSize = denomDf;
			numDf = (ngrp > 1.0) ? ngrp - 1.0 : 1.0;
			denomDf = (groupSize - 1.0) * 
				((design == CRD || ngrp == 1) ? ngrp : (ngrp - 1.0));
			noncen *= groupSize;
		}

		if(noncen == 0.0)
		{
			powr = alpha;
		}
		else
		{
			critval = Qsne(1 - alpha, numDf, denomDf);
			powr = noncentF(critval, noncen, numDf, denomDf);
			if(powr > 1.0)
			{
				sprintf(OUTSTR,
						"WARNING: non-central F not fully converged after %g terms",
						floor(powr));
				putOUTSTR();
				powr -= floor(powr);
			}
			powr = 1.0 - powr;
		}
		DATAVALUE(result,i) = powr;
	}
	
	return (result);

  errorExit:
	putOUTSTR();
	
	myerrorout(usage);
	return (0);
}

#undef UNDEFINED__
#ifdef UNDEFINED__
/* 
  This has been obsoleted by Qsne in pvalsub.c which calls ppbeta in
  ppbeta() in betabase.c
*/

double          Fquant(p, n1, n2)
double          p, n1, n2;

{
	/* find F quantile by interval bisection */

	/*  first get starting values via Cornish-Fisher approx
	    eqn 17 p. 82 of Johnson and Kotz */

	double          left, right, center, tmp, delta, sigma;
	extern double   Csne(), ppnd();
	long            fault;

	tmp = ppnd(p, &fault);
	sigma = sqrt(0.5 * (1.0 / n1 + 1.0 / n2));
	delta = 0.5 * (1. / n2 - 1. / n1);
	center = delta * (1.0 + (tmp * tmp - 1.) / 3.) +
		tmp * sigma / sqrt(1.0 - sigma * sigma);

	if (Csne(center, n1, n2) > p)
	{
		right = center;
		left = right * 0.9;
		while (Csne(left, n1, n2) > p)
		{
			left *= 0.9;
		}
	}
	else
	{
		left = center;
		right = left * 1.1;
		while (Csne(right, n1, n2) < p)
		{
			right *= 1.1;
		}
	}

	while ((right - left) * 2 / (right + left) > 0.001)
	{
		center = (left + right) / 2.;
		if (Csne(center, n1, n2) < p)
		{
			left = center;
		}
		else
		{
			right = center;
		}
	}

	return (center);
}

/* This has been replaced by a new version in pvalsub.c calling betanc */
double          noncentF(x, lam, n1, n2)
double          x, lam, n1, n2;

{
	/* approximate cumulative for noncentral F using Tiku's approx
	   or Severo and Zelen approx if Tiku fails
	   p 195 of Johnson and kotz */

	double          H, K, n1p, c, b, n1pluslam, tmp;
	extern double   Csne(), Cnor();
	WHERE("noncentF");
	
	n1pluslam = n1 + lam;
	if (n2 > 2.0)
	{
		H = 2.*n1pluslam*n1pluslam*n1pluslam +
			3.*n1pluslam*(n1 + 2.*lam)*(n2 - 2.) +
			(n1 + 3.*lam)*(n2 - 2.)*(n2 - 2.);
		K = n1pluslam*n1pluslam + (n2 - 2.)*(n1 + 2*lam);

		n1p = 0.5*(n2 - 2.)*(sqrt(H*H/(H*H - 4.*K*K*K)) - 1.);
		c = (n1p/n1)*H/K/(2*n1p + n2 - 2.);
		b = -n2*(c - 1. - lam/n1)/(n2 - 2.);

		return (Csne((x - b)/c, n1p, n2));
	}

	tmp = pow(n1*x/(n1 + lam), 1./3.);
	b = sqrt((2./9.)*((n1 + 2.*lam)/(n1pluslam*n1pluslam) + tmp*tmp/n2));

	c = (1. - (2./9.)/n2)*tmp - 1. + 
		(2./9.)*(n1 + 2.*lam)/(n1pluslam*n1pluslam);
	return (Cnor(c/b));
}
#endif /*UNDEFINED__*/
