/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Ppnd
#else
#define __SEG__ Ppnd
#endif /*MPW3*/
#endif /*MPW*/

#include <math.h>

#define SPLIT 0.42
#define A0 2.50662823884
#define A1 -18.61500062529
#define A2 41.39119773534
#define A3 -25.44106049637

#define B1 -8.47351093090
#define B2 23.08336743743
#define B3 -21.06224101826
#define B4 3.13082909833

#define C0 -2.78718931138
#define C1 -2.29796479134
#define C2 4.85014127135
#define C3 2.32121276858

#define D1 3.54388924762
#define D2 1.63706781897
/*
  
  Algorithm as 111 Applied Statistics (1977), Vol 26 No 1 Page 121
  Produces normal deviate corresponding to lower tail area of p.
  The hash sums are the sums of the moduli of the coefficients;
  they have no inherent meanings but are incuded for use in
  checking transcriptions.  Functions fabs, log and sqrt are used.  

  Derived from AS111 Fortran version
*/


double          ppnd(p, ifault)
double          p;
long           *ifault;

{
	double          q, r, ppn;

	*ifault = 0;
	q = p - 0.5;
	if (fabs(q) <= SPLIT)
	{
		r = q * q;
		ppn = q * (((A3 * r + A2) * r + A1) * r + A0) /
				((((B4 * r + B3) * r + B2) * r + B1) * r + 1.0);
	}
	else
	{
		r = p;
		if (q > 0.0)
		{
			r = 1.0 - p;
		}
		if (r <= 0.0)
		{
			*ifault = 1;
			ppn = 0.0;
		}
		else
		{			
			r = sqrt(-log(r));
			ppn = (((C3 * r + C2) * r + C1) * r + C0) /
				((D2 * r + D1) * r + 1.0);
			if (q < 0.0)
			{
				ppn = -ppn;
			}
		}
	}
	
	return (ppn);
}

