/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Predtabl
#else
#define __SEG__ Predtabl
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <math.h>
/*#define x(i,j) ((*REGX)[(long) i + (long) j * (long) NDATA ])*/

static double **Predvec, VarMeans[MAXVARS];
static long     Tablen, FacVars[MAXVARS], NFacVars;

/* recursively check columns to get coefs to add into pred values */

static void buildPredvec(long termi, long  *colno)
{
	long            add, i, i2, j, term[MAXVARS], term2[MAXVARS];
	long            nregx = (long) NREGX, nvars = (long) NVARS;
	long            ny = (long) NY;
	long            ij1, ij2;
	long            icell, ncells;
	modelType       trialterm;
	modelPointer    prevterm;
	double          vadd;

	getterm(modelterm(MODEL,termi),term, nvars);
	ncells = setterm(term,nvars,NCLASSES);

	for(icell = 0;icell < ncells;icell++)
	{
		/* check to see if this combination is already in through
		   lower order terms */
		effectiveTerm(term,(modelPointer) trialterm, nvars);
		/*  trialterm is now effective term for this potential column */
		for (i = 0; i < termi; i++)
		{
			prevterm = modelterm(MODEL,i);
			if(inEffectiveTerm(prevterm,(modelPointer) trialterm))
			{	/* if effective term is in earlier term */
				/* make sure any variates in earlier term are in eff term */
				getterm(prevterm, term2, nvars);
				for (i2 = 0; i2 < nvars; i2++)
				{
					if (term[i2] >= 0 && term2[i2] > 0 && NCLASSES[i2] < 0)
					{
						break;
					}
					/* break out if variate in earlier term not here */
					/* will keep this effective term */
				} /*for (i2 = 0; i2 < nvars; i2++)*/
				if (i2 >= nvars)
				{
					break;
				}
			} /*if(inEffectiveTerm(prevterm,(modelPointer) trialterm))*/
		} /*for (i = 0; i < termi; i++)*/

		if(i == termi)
		{
			/*
			  If here, effective term needs to be entered
			  now figure which cells of Predvec get this coefficient
			  loop through all elements of Predvec and add in where
			  appropriate
			*/

			for (i = 1; i < NFacVars; i++)
			{
				term2[i] = 1;
			}
			term2[0] = 0;

			for (i = 0; i < Tablen; i++)
			{
				term2[0]++;
				for (j = 0; j < NFacVars; j++)
				{
					if (term2[j] > NCLASSES[FacVars[j] - 1])
					{
						term2[j] -= NCLASSES[FacVars[j] - 1];
						term2[j + 1]++;
					}
				}

				vadd = 1.;
				for (j = 0; j < nvars; j++)
				{
					if (term[j] < 0)
					{
						vadd *= VarMeans[j];
					}
				}

				add = 1;
				for (j = 0; j < NFacVars; j++)
				{
					if (term[FacVars[j] - 1] <= 1 ||
						term2[j] + 1 == term[FacVars[j] - 1])
					{ /* a 1 or 0 always matches  or offset 1 matches*/
						add *= 1;
					}
					else if (term2[j] == NCLASSES[FacVars[j] - 1])
					{ /* tops get taken off */
						add *= -1;
					}
					else
					{
						add = 0;
						break;
					}
				}
				vadd *= (double) add;
				if (add != 0 && (*REGSS)[*colno] > -.5)
				{
					ij1 = i;	/* i + j * Tablen */
					ij2 = *colno; /* *colno + j * nregx */
					for (j = 0; j < ny; j++)
					{
						(*Predvec)[ij1] += vadd * (*REGCOEF)[ij2];
						ij1 += Tablen;
						ij2 += nregx;
					}
				} /*if (add != 0 && (*REGSS)[*colno] > -.5)*/
			} /*for (i = 0; i < Tablen; i++)*/
			(*colno)++;
		} /*if(i == termi)*/

		for(i=nvars-1;i>=0;i--)
		{
			if(term[i] > 0)
			{
				term[i]++;
				if(term[i] <= NCLASSES[i])
				{
					break;
				}
				term[i] = 1;
			}
		} /*for(i=nvars-1;i>=0;i--)*/
	} /*for(icell = 0;icell < ncells;icell++)*/
} /*buildPredvec()*/

Symbolhandle    predtable(list)
Symbolhandle    list;
{

	Symbolhandle    Install(), result;
	long            firstcol, i, j, j2, k;
	long            ndata = (long) NDATA, ny = (long) NY;
	long            nterms = (long) NTERMS, nvars = (long) NVARS;
	long            nvals;

	*OUTSTR = '\0';
	/* routine to compute a table of predicted values */
    if(NARGS(list) > 1 || COMPVALUE(list,0) != (Symbolhandle) 0)
	{
		sprintf(OUTSTR,"ERROR: usage is %s()",FUNCNAME);
	}
	else if (PREVMODELTYPE == 0 || PREVMODELTYPE & (LEAPS | OLSREG))
	{
		sprintf(OUTSTR,"ERROR: no active non-regression model");
	}
	else if (!(PREVMODELTYPE & DOCOEFS))
	{
		sprintf(OUTSTR,
				"ERROR: %s() does not work after GLM command with coefs:F",
				FUNCNAME);
	}
	
	if(*OUTSTR)
	{
		goto errorExit;
	}
	
	NFacVars = 0;		/* get number of factors and placement in model */
	Tablen = 1;
	for (i = 0; i < nvars; i++)
	{
		if (NCLASSES[i] > 0)
		{
			FacVars[NFacVars] = i + 1;
			NFacVars++;
			Tablen *= NCLASSES[i];
		}
	}

	if (Tablen == 1)
	{
		sprintf(OUTSTR,"ERROR:  no factors in anova model");
		goto errorExit;
	}

	if(USEGLMOFFSET && PREVMODELTYPE & (POISSONREG | LOGITREG))
	{
		myerrorout("WARNING: Table contains fitted deviations from offset, not fitted values");
		
	}
	nvals = Tablen * ny;
	result = RInstall(SCRATCH, nvals);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	Predvec = DATA(result);
	setNDIMS(result,NFacVars + ((ny > 1) ? 1 : 0));
	for (i = 1; i <= NFacVars; i++)
	{
		setDIM(result,i,NCLASSES[FacVars[i - 1] - 1]);
	}

	if (ny > 1)
	{
		setDIM(result,NFacVars + 1,ny);
	}

	if (PREVMODELTYPE & (BALANOVA + FASTANOVA + IPF))
	{			/* balanced */
		for (j = 0; j < ndata; j++)
		{
			j2 = 0;
			for (k = nvars; k > 0; k--)
			{
				j2 *= NCLASSES[k - 1];
				j2 += (*X[k - 1])[j] - 1;
			}
			(*Predvec)[j2] = (*Y)[j] - (*RESIDUALS)[j];
		}
	} /*if (PREVMODELTYPE & (BALANOVA + FASTANOVA + IPF))*/
	else
	{	/* unbalanced case */
		/* compute variate means */
		for (i = 0; i < nvars; i++)
		{
			VarMeans[i] = 0.;
			if (NCLASSES[i] <= 0)
			{					/* variate, not factor */
				for (j = 0; j < ndata; j++)
				{				/* for all data, even missing */
					if (!(PREVMODELTYPE & MISSWEIGHTS && (*MISSWTS)[j] == 0.0))
					{
						VarMeans[i] += (*X[i])[j];
					}
				}
				VarMeans[i] /=  NOTMISSING;
			}
		}

		for (i = 0; i < nvals; i++)
		{
			(*Predvec)[i] = 0.0;
		}
	
		firstcol = 0;
		for (i = 0; i < nterms; i++)
		{
			buildPredvec(i, &firstcol);
		}

		if (PREVMODELTYPE & POISSONREG)
		{
			for (i = 0; i < nvals; i++)
			{
				(*Predvec)[i] = exp((*Predvec)[i]);
			}
		}
		else if (PREVMODELTYPE & LOGITREG)
		{
			for (i = 0; i < nvals; i++)
			{
				(*Predvec)[i] = exp((*Predvec)[i]) / (1. + exp((*Predvec)[i]));
			}
		}
	}
	
	return (result);

  errorExit:
	putOUTSTR();

	return (0);
	
}

