/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Print
#else
#define __SEG__ Print
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

#ifdef MACINTOSH
#include "macIface.h"
#endif

extern char   **getKeyword();
extern long     symbolSize();
extern char    *isKeyword();
extern long     isfilename();
extern long     isVector();
extern long     argOK();
extern void     noData();
extern void     badType();
extern void     badNargs();
extern long     setFormat();

extern void     installFormat();
extern void     saveFormat();
extern void     restoreFormat();

#ifdef CONSOLENAME
#define ISCONSOLE(NAME) (strcmp(NAME,CONSOLENAME) == 0)
#else /*CONSOLENAME*/
#define ISCONSOLE(NAME)  0
#endif /*CONSOLENAME*/

#define         IPRINT     0x01
#define         IWRITE     0x02
#define         IMACRO     0x04
#define         IMAT       0x08
#define         IFILE      0x10
#define         IERR       0x20
#define         IFPRINT       (IPRINT | IFILE)
#define         IFWRITE       (IWRITE | IFILE)
#define         IMATPRINT     (IPRINT | IMAT | IFILE)
#define         IMATWRITE     (IWRITE | IMAT | IFILE)
#define         IMACROWRITE   (IMACRO | IFILE)
#define         IERROR        (IERR   | IPRINT)

	/* formerly a routine to print a single macro */
/*
  Operations print(), write(), fprint(), fwrite(), matprint(), matwrite(),
  macrowrite(), error()

  The first argument for fprint(), fwrite(), matprint(), matwrite(),
  and macrowrite(), must be a file name.

  If the last argument is the CHAR string "new" then the file is overwritten;
  otherwise, stuff is written at the end.

  For print(), write(), fprint(), and fwrite() the remaining arguments may
  be any printable object (types REAL, LOGIC, CHAR, MACRO, STRUC, PLOTINFO).

  For matprint() and matwrite(), arguments must be REAL OR LOGIC, and are
  currently limited to ndim  <= 2.  LOGIC values T and F are translated to
  1.0 and 0.0 and a comment of the form ') LOGICAL' is prepended.

  For macrowrite(), they must be MACRO.

  write(), fwrite(), and matwrite() differ from print(), fprint(), and
  matprint(), only in printing more significant digits.

  If keyword phrase file:fileName appears on print() or write(), it
  operates like fprint() or fwrite() with that fileName.  In that case,
  new:T is acceptable.

  All operations accept keywords nsig (e.g., nsig:8) and format (e.g.,
  format:"10'4f"), although these are meaningless for macrowrite.  They
  control the printing of any REAL or LOGIC objects later in the argument
  list.  These keywords may be repeated to change the significance.

  nsig:m is is equivalent to a format of %(m+7).mg  .  The last nsig or
  format after all non-keywords or unrecognized keywords is taken as a
  starting default.  Thus print(a, b, c, nsig:9) is equivalent to
  print(nsig:9, a, b c) or write(a,b,c)

  Key word new (e.g., new:T or new:F) can appear immediately after the file
  name.

  error() is identical to print() except it will cause immediate termination
  of the statement in which it occurs.  It is designed for printing fatal
  error messages in macros.
*/

typedef struct opEntry
{
	char *name;
	short iop;
} opEntry;

static opEntry Ops[] =
{ /* name          iop */
	"write",      IWRITE,
	"fwrite",     IFWRITE,
	"matwrite",   IMATWRITE,
	"print",      IPRINT,
	"fprint",     IFPRINT,
	"matprint",   IMATPRINT,
	"macrowrite", IMACROWRITE,
	"error",      IERROR,
	(char *) 0,   0
};

typedef struct keylist
{
	char        *name;
	long         type;
} keylist;

static keylist LegalKeys[] =
{
	"new",       LOGIC,
	"file",      CHAR,
	"sep",       CHAR,
	"header",    LOGIC,
	"missing",   REAL,
	"nsig",      REAL,
	"format",    CHAR,
	"labels",    LOGIC,
	(char *) 0,  0
};

Symbolhandle    print(list)
Symbolhandle    list;
{
	Symbolhandle    arg1 = COMPVALUE(list,0);
	Symbolhandle    argi, filearg = (Symbolhandle) 0;
	Symbolhandle    result = (Symbolhandle) 0;
	FILE           *outFile = STDOUT;
	char           *fileMode = (char *) 0;
	long            nargs = NARGS(list), i, type, ifile;
	long            unusedArgs = nargs;
	long            op;
	long            toFile = 0, useConsole = 0;
	long            fmt[3];
	long            header = -1, labels = -1;
	struct opEntry *who;
	struct keylist *key;
	char            separator[2];
	char           *name;
	char           *keyword;
	char           *fileName = (char *) 0;
	double          dsig, missValue;
	long            nsig = -1;
	WHERE("print");
	
/* save current format values in case they are changed */
	saveFormat();
	setMissing(missValue);
	SETUPINT(cleanup);
	
	*OUTSTR = '\0';
	*separator = '\0';

	for(who = Ops;who->iop != 0;who++)
	{
		if(strcmp(who->name,FUNCNAME) == 0)
		{
			op = who->iop;
			break;
		}
	} /*for(who = Ops;who->iop != 0;who++)*/

	if(!(op & IERR))
	{
		for(i=0;i<nargs;i++)
		{
			argi = COMPVALUE(list,i);
		
			if((keyword = isKeyword(argi)) && strcmp(keyword,"file") == 0)
			{
				if (filearg != (Symbolhandle) 0)
				{
					sprintf(OUTSTR,
							"ERROR: repeated use of '%s' as argument to %s",
							keyword, FUNCNAME);
					goto errorExit;
				}
				
				op |= IFILE;
				ifile = i + 1;
				filearg = argi;
				if((TYPE(filearg) != CHAR || !isScalar(filearg)) &&
				   !ISCONSOLE(NAME(filearg)))
				{
					notCharOrString(keyword);
					goto errorExit;
				}
				setNAME(argi,USEDKEYWORD);
			}
		} /*for(i=0;i<nargs;i++)*/
	} /*if(!(op & IERR))*/
	
	if(op & IFILE)
	{
		if(filearg == (Symbolhandle) 0)
		{ /* no file:fileName ; file must be first argument */
			ifile = toFile = 1;
			filearg = arg1;
			if(!argOK(filearg,CHAR,1))
			{
				goto errorExit;
			}
			unusedArgs--;
		}
		else
		{
			toFile = 2;
		}
		
		useConsole = ISCONSOLE(NAME(filearg)) ||
			isScalar(filearg) && TYPE(filearg) == CHAR &&
				ISCONSOLE(STRINGPTR(filearg));

		if(useConsole)
		{
			fileName = CONSOLENAME;
		}
	} /*if(op & IFILE)*/
	
	if(toFile && !useConsole)
	{ /* writing to a file */
		if(!argOK(filearg,CHAR, ifile))
		{
			goto errorExit;
		}
		
		if(!isScalar(filearg))
		{
			notCharOrString("file name");
			goto errorExit;
		}
		fileName = STRINGPTR(filearg);
	} /*if(toFile && !useConsole)*/

	for(i=(toFile == 1) ? 1 : 0;i<nargs;i++)
	{ /* check arguments and keywords, saving trailing formatting info */
		argi = COMPVALUE(list,i);
		if(!argOK(argi,0,i+1))
		{
			goto errorExit;
		}
		type = TYPE(argi);

		if((keyword = isKeyword(argi)))
		{
			for(key = LegalKeys;key->type != 0;key++)
			{
				if(strcmp(keyword,key->name) == 0)
				{
					break;
				}
			}
			if(key->type != 0)
			{
				if(!isScalar(argi) || type != key->type)
				{
					if(key->type == CHAR)
					{
						notCharOrString(keyword);
					}
					else if(key->type == LOGIC)
					{
						notTorF(keyword);
					}
					else
					{
						notNumberOrReal(keyword);
					}
					goto errorExit;
				}
				if(strcmp(keyword,"new") == 0)
				{
					if(!toFile)
					{
						sprintf(OUTSTR,
								"ERROR: keyword %s illegal unless a file is specified",
								keyword);
						goto errorExit;
					}
					if(fileMode)
					{
						sprintf(OUTSTR,"ERROR: repeated use of keyword '%s'",
								keyword);
					}
					else
					{
						fileMode = (DATAVALUE(argi,0) == 0.0) ? "a" : "w";
						setNAME(argi,USEDKEYWORD);
					}				
				}
				else if(op & IMAT && strcmp(keyword,"sep") == 0)
				{
					if(*separator)
					{
						sprintf(OUTSTR,"ERROR: repeated use of keyword '%s'",
								keyword);
					}
					else if(strlen(STRINGPTR(argi)) > 1)
					{
						sprintf(OUTSTR,
								"ERROR: value for '%s' more than 1 character",
								keyword);
					}
					else
					{
						strcpy(separator, STRINGPTR(argi));
						setNAME(argi,USEDKEYWORD);
					}
				}
				else if(strcmp(keyword,"header") == 0)
				{
					if(header >= 0)
					{
						sprintf(OUTSTR,"ERROR: repeated use of keyword '%s'",
								keyword);
					}
					else
					{
						header = (DATAVALUE(argi,0) == 0.0) ? 0 : 1;
						setNAME(argi,USEDKEYWORD);
					}				
				}
				else if((op & IPRINT || op & IWRITE) &&
						strcmp(keyword,"labels") == 0)
				{
					if(labels >= 0)
					{
						sprintf(OUTSTR,"ERROR: repeated use of keyword '%s'",
								keyword);
					}
					else
					{
						labels = (DATAVALUE(argi,0) == 0.0) ? 0 : 1;
						setNAME(argi,USEDKEYWORD);
					}				
				}
				else if(strcmp(keyword,"nsig") == 0)
				{
					if((dsig  = DATAVALUE(argi,0)) <= 0.0 ||
					   dsig != floor(dsig))
					{
						sprintf(OUTSTR,
								"WARNING: illegal value %g for 'nsig'; not used",
								dsig);
						putOUTSTR();
					}
					else
					{
						nsig = (long) dsig;
						fmt[0] = nsig+7;
						fmt[1] = nsig;
						fmt[2] = (long) 'g';
					}
				}
				else if(strcmp(keyword,"format") == 0)
				{
					if(!setFormat(STRINGPTR(argi),fmt))
					{
						sprintf(OUTSTR,
								"WARNING: bad format specifier '%s'; not used",
								STRINGPTR(argi));
						putOUTSTR();
					}
					else
					{
						nsig = fmt[1];
					}
				}
				else if(op & IMAT && strcmp(keyword,"missing") == 0)
				{
					missValue = DATAVALUE(argi,0);
				}
				if(*OUTSTR)
				{
					goto errorExit;
				}
				unusedArgs--;
			} /*if(key->type != 0)*/
			else if(strcmp(keyword, USEDKEYWORD+2) != 0)
			{  /* not a recognized keyword */
				setMissing(missValue); /* wipes out previous missValue*/
				nsig = -1; /*wipes out previous format or nsig keywords*/
			}
		} /*if((keyword = isKeyword(argi)))*/
		else /* not a keyword */
		{
			setMissing(missValue); /* wipes out previous missValue*/
			nsig = -1; /*wipes out previous format or nsig keywords*/
		}
	} /*for(i=(toFile == 1) ? 1 : 0;i<nargs;i++)*/

	if(toFile)
	{
		if(header < 0)
		{/* default when header for sep:"c" supplied is header:F */
			header = (*separator) ? 0 : 1;
		}
		if(fileMode == (char *) 0)
		{ /* default mode is append (new:T) */
			fileMode = "a";
		}
		
#ifdef MACINTOSH
		if(PrintVolume)
		{
			SetVol(0L, PrintVolume);
		}
		if(strlen(fileName) == 0)
		{
			fileName = macOpen("\pFile for output:", "\p",WRITEIT,
							0, (OSType *) 0);
			if(fileName == (char *) 0)
			{
				goto cleanup;
			}
			PrintVolume = Reply.vRefNum;
		}
#endif /*MACINTOSH */

		if((fileName = expandFilename(fileName)) == (char *) 0)
		{
			goto cleanup;
		}

		if(!isfilename(fileName))
		{
			goto cleanup;
		}
	} /* if(toFile) */

	if(unusedArgs == 0)
	{
		sprintf(OUTSTR,"ERROR: nothing specified to print or write");
		goto errorExit;
	}
	
	if(toFile && !useConsole)
	{
		outFile = fopen(fileName,fileMode);
		if(outFile == (FILE *) 0)
		{
			sprintf(OUTSTR,"ERROR: %s cannot open %s for writing",
					FUNCNAME,STRINGPTR(arg1));
			putOUTSTR();
			goto cleanup;
		}
#ifdef MACINTOSH
		macSetInfo(PrintVolume,fileName,'TEXT',TEXTCREATOR);
#endif /*MACINTOSH*/
	} /*if(toFile && !useConsole)*/

	if(nsig < 0)
	{ /* no format has been set */
		if(op & IWRITE)
		{
			for(i=0;i<3;i++)
			{
				fmt[i] = WRITEFORMAT[i];
			}
		}
		else
		{
			for(i=0;i<3;i++)
			{
				fmt[i] = PRINTFORMAT[i];
			}
		}
	} /*if(nsig < 0)*/
	installFormat(fmt[0],fmt[1],fmt[2]);

	if((op & (IWRITE|IPRINT)) != 0 && (op & IMAT) == 0 && unusedArgs == 1)
	{ /* only one thing to write or print */
		for(i=(toFile==1) ? 1 : 0;i<nargs;i++)
		{
			argi = COMPVALUE(list,i);
			if(!argOK(argi,0,i+1))
			{
				goto cleanup;
			}
			if(!isKeyword(argi) && 
			   (type = TYPE(argi)) == CHAR && isScalar(argi))
			{/* if single character variable, print with no header */
				if(outFile == STDOUT)
				{
					if(*STRINGPTR(argi) == '\0')
					{
						myeol();
					}
					else
					{
						mymultiprint(STRINGPTR(argi));
					}
				}
				else
				{
					fmyprint(STRINGPTR(argi),outFile);
					fmyeol(outFile);
				}
				
				unusedArgs = 0;
				break;
			}
		} /*for(i=(toFile==1) ? 1 : 0;i<nargs;i++)*/
	} /*if((op&(IWRITE|IPRINT)) != 0 && (op&IMAT) == 0 && unusedArgs == 1)*/

	for(i = (toFile == 1) ? 1 : 0; i < nargs && unusedArgs > 0; i++)
	{
		argi = COMPVALUE(list, i);
		if(!argOK(argi,0,i+1))
		{
			goto cleanup;
		}
		type = TYPE(argi);
		if((keyword = isKeyword(argi)))
		{
			if(strcmp(keyword,"format") == 0)
			{
				if(!setFormat(STRINGPTR(argi),fmt))
				{
					sprintf(OUTSTR,
							"WARNING: bad format specifier '%s'; not used",
							STRINGPTR(argi));
					putOUTSTR();
				}
				else
				{
					installFormat(fmt[0],fmt[1],fmt[2]);
				}
				argi = (Symbolhandle) 0;
			} /*if(strcmp(keyword,"format") == 0)*/
			else if(strcmp(keyword,"nsig") == 0)
			{
				if((dsig  = DATAVALUE(argi,0)) <= 0.0 || dsig != floor(dsig))
				{
					sprintf(OUTSTR,
							"WARNING: illegal value %g for 'nsig'; not used",
							dsig);
					putOUTSTR();
				}
				else
				{
					nsig = (long) dsig;
					installFormat(nsig+7,nsig,(long)'g');
				}
				argi = (Symbolhandle) 0;
			}
			else if(op & IMAT && strcmp(keyword,"missing") == 0)
			{
				missValue = DATAVALUE(argi,0);
				argi = (Symbolhandle) 0;
			}
			else if(strcmp(NAME(argi), USEDKEYWORD) == 0)
			{
				argi = (Symbolhandle) 0;
			}
		} /*if((keyword = isKeyword(argi)))*/

		if(argi != (Symbolhandle) 0)
		{
			unusedArgs--;
			if(op & (IMAT + IMACRO))
			{
				if(type == REAL || type == LOGIC)
				{
					if(op & IMACRO)
					{
						sprintf(OUTSTR,
								"WARNING: %s variable written by %s",
								typeName(type),FUNCNAME);
						putOUTSTR();
					}
					matWrite(argi,outFile, missValue,header,separator);
				}
				else if(type == MACRO)
				{
					if(op & IMAT)
					{
						sprintf(OUTSTR,
								"WARNING: %s written by %s",
								typeName(type),FUNCNAME);
						putOUTSTR();
					}
					macWrite(argi,outFile,header);
				}
				else
				{ /* not REAL, LOGIC or MACRO */
					sprintf(OUTSTR,
							"ERROR: %s cannot write variable of type %s",
							FUNCNAME, typeName(type));
					putOUTSTR();
					goto cleanup;
				}
			}					/*if(op & (IMAT + IMACRO))*/
			else
			{					/* not matwrite, matprint, or macrowrite */
				if(header)
				{
					name = NAME(argi);
					if(isKeyword(argi) || isscratch(name))
					{
						name += 2;
					}
					sprintf(OUTSTR,"%s:", name);
					fmyprint(OUTSTR,outFile);
					*OUTSTR = '\0';
					fmyeol(outFile);
				}	
				fprexpr(argi,outFile,labels);
			}					/*if(op & (IMAT + IMACRO)){...}else{...}*/

			if(INTERRUPT != INTNOTSET)
			{
				if(INTERRUPT != PRINTABORT)
				{
					goto cleanup;
				}
				break;
			}
		} /*if(argi != (Symbolhandle) 0)*/
	} /*for(i = (toFile == 1) ? 1 : 0; i < nargs; i++)*/

	result = (op != IERROR) ? NULLSYMBOL : (Symbolhandle) 0;

	goto cleanup;
	
  errorExit:
	putOUTSTR();

  cleanup:
	if(outFile != (FILE *) 0 && outFile != STDOUT)
	{
		fclose(outFile);
		outFile = (FILE *) 0;
	}
	restoreFormat();
	*OUTSTR = '\0';
	
	RETURN (result);
}

/*
  putascii(vector), where vector[i] > 0 && vector[i] < 256
  routine to output raw ascii codes
*/
#define GBUFFER  0

Symbolhandle    putascii(list)
Symbolhandle    list;
{

	Symbolhandle    arg1, symh;
	Symbolhandle    result = (Symbolhandle) 0;
	long            i, j, n = 0, length;
	long            nargs = NARGS(list);
	long            new = 0, toFile = 0, keep = 0;
	double          val;
	char           *keyword, *fileName = (char *) 0, *fileMode = "a";
	unsigned char **buffer = (unsigned char **) 0;
	FILE           *fp = STDOUT;
	WHERE("putascii");
	TRASH(1,errorExit);
	
	SETUPINT(errorExit);

	if((arg1 = COMPVALUE(list,0)) == (Symbolhandle) 0 && nargs == 1)
	{
		nargs = 0;
	}
	
	for(i=nargs-1;i>=0;i--)
	{
		symh = COMPVALUE(list,i);
		if(!(keyword = isKeyword(symh)))
		{
			break;
		}
		if(strcmp(keyword,"file") == 0)
		{
			if(!isScalar(symh) || TYPE(symh) != CHAR)
			{
				notCharOrString("'file'");
				goto errorExit;
			}
			toFile = 1;
			fileName = STRINGPTR(symh);
		}
		else if(strcmp(keyword,"new") == 0 || strcmp(keyword,"keep") == 0)
		{
			if(!isScalar(symh) || TYPE(symh) != LOGIC)
			{
				notTorF(keyword);
				goto errorExit;
			}
			if(keyword[0] == 'n')
			{
				new = (DATAVALUE(symh,0) != 0.0);
			}
			else
			{
				keep = (DATAVALUE(symh,0) != 0.0);
			}
		}
		else
		{
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
		nargs--;
	} /*for(i=nargs-1;i>=0;i--)*/

	if(new && !toFile)
	{
		sprintf(OUTSTR,
				"ERROR: 'new:T' illegal without 'file:fileName'");
		goto errorExit;
	}
	if(keep && toFile)
	{
		sprintf(OUTSTR,
				"ERROR: 'keep:T' is illegal with 'file:fileName'");
		goto errorExit;
	}
	
	if(nargs == 0)
	{
		sprintf(OUTSTR,
				"ERROR: no data supplied for %s",FUNCNAME);
		goto errorExit;
	}

	for(i=0;i<nargs;i++)
	{
		symh = COMPVALUE(list,i);
		if(!argOK(arg1,0,i+1))
		{
			goto errorExit;
		}
		if(!isVector(symh) || TYPE(symh) != REAL)
		{
			sprintf(OUTSTR,
					"ERROR: all data arguments to %s must be REAL scalars or vectors",
					FUNCNAME);
			goto errorExit;
		}
		n += symbolSize(symh);
	} /*for(i=0;i<nargs;i++)*/
	
	if(toFile)
	{
		if(new)
		{ /* default mode is "a" (new:F) */
			fileMode = "w";
		}

#ifdef MACINTOSH
		if(strlen(fileName) == 0)
		{
			if(PrintVolume)
			{
				SetVol(0L, PrintVolume);
			}
			fileName = macOpen("\pFile for putascii():", "\p", WRITEIT,
							0, (OSType *) 0);
			if(fileName == (char *) 0)
			{
				goto errorExit;
			}
			PrintVolume = Reply.vRefNum;
		}
#endif /*MACINTOSH */

		if((fileName = expandFilename(fileName)) == (char *) 0)
		{
			goto errorExit;
		}

		if(!isfilename(fileName))
		{
			goto errorExit;
		}
		fp = fopen(fileName,fileMode);
		if(fp == (FILE *) 0)
		{
			sprintf("ERROR: cannot open file %s",fileName);
			goto errorExit;
		}
	} /* if(toFile) */
		
	if(!getScratch(buffer,GBUFFER,n+1,unsigned char))
	{
		goto errorExit;
	}

	n = 0;
	for(i=0;i<nargs;i++)
	{
		symh = COMPVALUE(list,i);
		length = symbolSize(symh);
		for(j=0;j<length;j++)
		{
			val = DATAVALUE(symh,j);
			if(floor(val) != val || val <= 0.0 || val > 255.0)
			{
				sprintf(OUTSTR,
						"ERROR: all values for %s must be integers between 1 and 255",
						FUNCNAME);
				goto errorExit;
			}
			(*buffer)[n++] = (unsigned char) val;
		} /*for(j=0;j<length;j++)*/
	} /*for(i=0;i<n;i++)*/
	(*buffer)[n] = '\0';

	if(!keep)
	{
		fmyprint((char *) *buffer, fp);
		fmyflush(fp);
		result = NULLSYMBOL;
	}
	else
	{
		if((result = Install(SCRATCH,CHAR)) == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		unTrash(GBUFFER);
		setSTRING(result,(char **) buffer);
		setNDIMS(result,1);
		setDIM(result,1,1);
	}
	
	/* fall through */

  errorExit:
	putOUTSTR();

	if(fp != STDOUT)
	{
		fclose(fp);
	}
	emptyTrash();
	
	RETURN (result);
}
