/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Printano
#else
#define __SEG__ Printano
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

extern Symbolhandle    RInstall();

void printglm()
{

	/* print out anova output */

	long            i, chpt, k, k2;
	long            ny = (long) NY, nterms = (long) NTERMS;
	long            printTerms;
	char           *colLabels[3];
	Symbolhandle    symh = (Symbolhandle) 0;
	double          modeldf = 0.0;
	double         *tmp, *ss, *df = *DF;
	WHERE("printglm");

	SETUPINT(errorExit);
	
	chpt = 0;
	colLabels[0] = "DF";
	if (MODELTYPE & ROBUSTREG)
	{
		colLabels[1] = "Criterion";
		colLabels[2] = "Crit/DF";
	}
	else if (MODELTYPE & (LOGITREG | POISSONREG | IPF))
	{
		colLabels[1] = "Deviance";
		colLabels[2] = "MDev";
	}
	else if (MODELTYPE & MANOVA)
	{
		/* Note: following '\n' gets explicitly counted below */
		colLabels[1] = "\n          SS and SP Matrices";
		colLabels[2] = "";
		symh = RInstall(SCRATCH,ny * ny);
		if (symh == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNDIMS(symh,2);
		setDIM(symh,1,ny);
		setDIM(symh,2,ny);
		tmp = DATAPTR(symh);
	}
	else
	{
		colLabels[1] = "SS";
		colLabels[2] = "MS";
	}
	ss = *SS;
	df = *DF;
	sprintf(OUTSTR, "%30s %4s","",colLabels[0]);
	myprint(OUTSTR);
	if(colLabels[1][0] == '\n')
	{
		NLINES++;
	}
	sprintf(OUTSTR,MISSINGFMT,colLabels[1]);
	
	myprint(OUTSTR);
	if(colLabels[2][0] != '\0')
	{
		sprintf(OUTSTR,MISSINGFMT,colLabels[2]);
		putOUTSTR();
	}
	else
	{
		myeol();
	}
	if(INTERRUPT != INTNOTSET)
	{
		goto errorExit;
	}
	
	printTerms = (INCREMENTAL ||
		!(MODELTYPE & (POISSONREG | LOGITREG | IPF | ROBUSTREG)));
		   
	for (i = 0; i <= nterms; i++)
	{
		modeldf += df[i];
		if (MODELTYPE & MANOVA)
		{
			sprintf(OUTSTR, "%-30s %4.0lf", (*TERMNAMES + chpt), df[i]);
			putOUTSTR();
			for (k = 0; k < ny; k++)
			{
				for (k2 = 0; k2 < ny; k2++)
				{
					tmp[k + k2*ny] =
						ss[i + (nterms + 1)*(k + k2*ny)];
				}
			}
			prexpr(symh);
			if(INTERRUPT != INTNOTSET)
			{
				goto errorExit;
			}
		} /* if (MODELTYPE & MANOVA) */
		else
		{
			if(printTerms || i == nterms)
			{
				sprintf(OUTSTR, "%-30s %4.0lf", (*TERMNAMES + chpt), df[i]);
				myprint(OUTSTR);
				sprintf(OUTSTR, DATAFMT,  ss[i] );
				myprint(OUTSTR);
				sprintf(OUTSTR, DATAFMT,
						ss[i] / (df[i] > 0 ? df[i] : 1));
				putOUTSTR();
			}
			else if(i == nterms - 1)
			{
				sprintf(OUTSTR, "%-30s %4.0lf", "Overall model", modeldf);
				myprint(OUTSTR);
				sprintf(OUTSTR, DATAFMT,  ss[i] );
				myprint(OUTSTR);
				sprintf(OUTSTR, DATAFMT,
						ss[i] / (modeldf > 0 ? modeldf : 1));
				putOUTSTR();
			}
			if(INTERRUPT != INTNOTSET)
			{
				goto errorExit;
			}
		} /*if (MODELTYPE & MANOVA){...}else{...} */
		chpt += strlen((*TERMNAMES + chpt)) + 1;
	} /*for (i = 0; i <= nterms; i++)*/

	if(MODELTYPE & ROBUSTREG)
	{
		myerrorout("NOTE: Criterion should not be interpreted as chi-squared");
	}
	/* fall through */
	
  errorExit:
	Removesymbol(symh);

	*OUTSTR = '\0';
	RETURN;
}
