/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Printreg
#else
#define __SEG__ Printreg
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <math.h>
extern void         myeol(), myerrorout(), myprint();

void printregress()
{
	double         *df = *DF, *ss = *SS, *allwts;
	long            ndata = (long) NDATA, nregx = (long) NREGX;
	long            i, modeldf = 0, errordf = (long) df[nregx];
	long            iprev = -1;
	long            nterms = (long) NTERMS;
	long            weighted = MODELTYPE & (MISSWEIGHTS | CASEWEIGHTS);
	double          dw1, dw2, tmp, sd, mse, r2, fstat = 0.0;
	double         *residuals, *xtxinv = *XTXINV;
	double         *regcoef = *REGCOEF;
	char           *termnames = *TERMNAMES, *undefined = "undefined";
	char           *outstr;
	
	residuals = (MODELTYPE & CASEWEIGHTS) ? *WTDRESIDUALS : *RESIDUALS;

	if(weighted)
	{
		allwts = *ALLWTS;
	}

	mse = (errordf > 0) ? mse = ss[nregx] / (double) errordf : 0.0;

	r2 = 0.0;
	for (i = 0; i < nregx; i++)
	{
		if (df[i] > 0.0 && !modeltermEqual(modelterm(MODEL,i),
											 (modelPointer) NULLTERM))
		{
			r2 += ss[i];
			modeldf += (long) df[i];
		}
	} /*for (i = 0; i < nregx; i++)*/

	if(mse != 0.0)
	{
		fstat = (r2/(double) modeldf) / mse;
	}

	r2 /= r2 + ss[nregx];

	/* Compute Durbin-Watson statistic */
	if(mse > 0)
	{
		dw1 = 0.0;
		dw2 = 0.0;
		for (i = 0; i < ndata; i++)
		{
			if (!weighted || allwts[i] > 0.0)
			{
				tmp = residuals[i];
				dw2 += tmp*tmp;
				if(iprev >= 0)
				{
					tmp -= residuals[iprev];
					dw1 += tmp*tmp;
				}
				iprev = i;
			} /*if (!weighted || allwts[i] > 0.0)*/
		} /*for (i = 0; i < ndata; i++)*/
		if(dw2 > 0.0)
		{
			dw1 /= dw2;
		}
	} /*if(mse > 0)*/
	
	/*  print out the regression  results */

	outstr = OUTSTR;
	sprintf(outstr,"%18s","");
	outstr += strlen(outstr);
	sprintf(outstr,MISSINGFMT,"Coef");
	outstr += strlen(outstr);  
	sprintf(outstr,MISSINGFMT,"StdErr");
	outstr += strlen(outstr);
	sprintf(outstr,MISSINGFMT,"t");
	putOUTSTR();

	for (i = 0; i < nterms; i++)
	{
		if (df[i] > 0.0)
		{
			sd = sqrt(xtxinv[i * nregx + i] * mse);
			outstr = OUTSTR;
			sprintf(outstr, "%-18s ",termnames);
			outstr += strlen(outstr);
			sprintf(outstr,DATAFMT,regcoef[i]);
			outstr += strlen(outstr);
			if(errordf > 0)
			{
				sprintf(outstr,DATAFMT,sd);
			}
			else
			{
				sprintf(outstr, MISSINGFMT, undefined);
			}
			outstr += strlen(outstr);
			if(sd > 0.0)
			{
				sprintf(outstr,DATAFMT,regcoef[i] / sd);
			}
			else
			{
				sprintf(outstr, MISSINGFMT, undefined);
			}
		} /*if (df[i] > 0.0)*/
		else
		{
			sprintf(OUTSTR, "%-18s  deleted", termnames);
		}
		putOUTSTR();

		termnames += strlen(termnames) + 1;
		if(INTERRUPT != INTNOTSET)
		{
			goto interruptExit;
		}
	} /*for (i = 0; i < nterms; i++)*/

	myeol();

	outstr = OUTSTR;
	sprintf(outstr, "N: %ld   MSE:",ndata);
	outstr += strlen(outstr);
	if(errordf > 0)
	{
		sprintf(outstr, DATAFMT, mse);
	}
	else
	{
		sprintf(outstr,undefined);
	}
	outstr += strlen(outstr);
	sprintf(outstr," DF: %ld  R^2: %7.5lf",errordf, r2);
	putOUTSTR();

	outstr = OUTSTR;
	sprintf(OUTSTR,"Regression F(%ld,%ld): ",modeldf,errordf);
	outstr += strlen(outstr);

	if(mse > 0)
	{
		sprintf(outstr, DATAFMT, fstat);
	}
	else
	{
		strcpy(outstr, undefined);
	}
	outstr += strlen(outstr);
	
	strcpy(outstr, " Durbin-Watson: ");
	outstr += strlen(outstr);
	if(mse > 0)
	{
		sprintf(outstr, "%7.5lf", dw1);
	}
	else
	{
		strcpy(outstr, undefined);
	}
	putOUTSTR();

	strcpy(OUTSTR,"To see the ANOVA table type 'anova()'");
	putOUTSTR();
	/* fall through */
  interruptExit:
	;
} /*printregress()*/
