/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Pvals
#else
#define __SEG__ Pvals
#endif /*MPW3*/
#endif /*MPW*/

#define POSREAL(X)    ((X) > 0)
#define POSINT(X) (POSREAL(X) && floor(X) == (X))
#define NONNEG(X) ((X) >= 0)
#define INBOUNDS(X,A,B) ((X) >= (A) && (X) <= (B))

#include "globals.h"

extern Symbolhandle      RInstall();

extern double   Cnor(), Cstu(), Cchi(), Csne(), Cgam(), Cbet(), Cpoi();
extern double   Cbin(), Chyp();
extern double   Qnor(), Qstu(), Qchi(), Qsne(), Qgam(), Qbet();
extern long	    dupDims();

extern long matchKey();
extern long argOK();
extern void noData();
extern void badNargs();
extern void badType();

#define MAXBINOMN 100000  /* max allow n for binomial */
#define ICUMNOR   301
#define ICUMCHI   302
#define ICUMSTU   303
#define ICUMF     304
#define ICUMPOI   305
#define ICUMBIN   306
#define IINVNOR   307
#define IINVCHI   308
#define IINVSTU   309
#define IINVF     310
#define ICUMGAMMA 311
#define IINVGAMMA 312
#define ICUMBETA  313
#define IINVBETA  314

typedef struct opEntry
{
	char *name;
	short iop;
	short nargs;
} opEntry;

static opEntry Ops[] =
{ /* name        iop      nargs */
	"cumnor",    ICUMNOR,     1, /* cumnor(x) */
	"cumchi",    ICUMCHI,     2, /* cumchi(x,df) */
	"cumstu",    ICUMSTU,     2, /* cumstu(x,df) */
	"cumF",      ICUMF,      -4, /* cumF(x,ndf,ddf[,lam]) */
	"cumpoi",    ICUMPOI,     2, /* cumpoi(x,lambda) */
	"cumbin",    ICUMBIN,     3, /* cumbin(x,n,p) */
	"invnor",    IINVNOR,     1, /* invnor(p) */
	"invchi",    IINVCHI,     2, /* invchi(p,df) */
	"invstu",    IINVSTU,     2, /* invstu(p,df) */
	"invF",      IINVF,       3, /* invF(p,ndf,ddf) */
	"cumgamma",  ICUMGAMMA,   2, /* cumgamma(u,a) */
	"invgamma",  IINVGAMMA,   2, /* invgamma(u,a) */
	"cumbeta",   ICUMBETA,   -4, /* cumbeta(u,a,b[,lam]) */
	"invbeta",   IINVBETA,    3, /* invbeta(u,a,b) */
	(char *) 0,      0,       0
};

Symbolhandle    cumcdf(list)
Symbolhandle    list;
{

	Symbolhandle    arg[4], result = (Symbolhandle) 0;
	double        **y;
	double          x, param[3];
	long            ndims, dims[4][MAXDIMS+1],sizes[4], ii[3];
	long            dim[MAXDIMS+1], size = 1;
	long            minargs, margs, nargs = NARGS(list);
	long            checkedParam[3], foundBadParam[3], foundOutOfRange=0;
	long            misMatch = 0, nonConverge = 0;
	long            i, j;
	long            arithMissingFound = 0;
	int             iop;
	opEntry        *op;
	char           *parName[3];
	WHERE("cumnor");
	
	*OUTSTR = '\0';
	
	for(i=0;i<4;i++)
	{
		sizes[i] = 0;
	}
	
	for(op=Ops; op->iop; op++)
	{
		if(strcmp(op->name,FUNCNAME) == 0)
		{
			break;
		}
	}

	iop = op->iop;
	margs = op->nargs;
	minargs = (margs > 0) ? margs : -margs - 1; /* allow for 1 optional arg */
	
	if(iop == 0)
	{ /* should never occur */
		sprintf(OUTSTR,"ERROR: unrecognized pvalue or inv cdf name %s",
				FUNCNAME);
		goto errorExit;
	}
	
	if(margs > 0 && nargs != margs || margs < 0 && nargs > -margs)
	{
		badNargs(FUNCNAME,margs);
		goto errorExit;
	}
	else if(margs < 0 && nargs < minargs)
	{
		badNargs(FUNCNAME,-(minargs+1000));
	}
	
	switch(iop)
	{
	  case ICUMCHI:
	  case ICUMSTU:
	  case IINVCHI:
	  case IINVSTU:
		parName[0] = "degrees of freedom";
		break;
	  case ICUMF:
	  case IINVF:
		parName[0] = "numerator. d.f.";
		parName[1] = "denominator. d.f.";
		parName[2] = "noncentrality";
		break;
	  case ICUMPOI:
		parName[0] = "mean";
		break;
	  case ICUMBIN:
		parName[0] = "N";
		parName[1] = "P";
		break;
	  case ICUMGAMMA:
	  case IINVGAMMA:
		parName[0] = "alpha";
		break;
	  case ICUMBETA:
	  case IINVBETA:
		parName[0] = "alpha";
		parName[1] = "beta";
		parName[2] = "noncentrality";
		break;
	  default:
		;
	}
	
	for(i=0;i<nargs;i++)
	{
		arg[i] = COMPVALUE(list, i);

		if(!argOK(arg[i],REAL,i+1))
		{
			goto errorExit;
		}

		sizes[i] = dupDims(dims[i],arg[i]);
		if(i == 0 || size == 1 && sizes[i] > 1)
		{
			size = sizes[i];
			for(j = 0;j<=NDIMS(arg[i]);j++)
			{
				dim[j] = dims[i][j];
			}
		}
		
		if(i >= 1)
		{
			if(sizes[i] != 1)
			{
				if(sizes[i] != size)
				{
					misMatch = 1;
				}
				else
				{
					for(j=1;j<=dims[i][0];j++)
					{
						if((j<=dim[0] && dim[j] != dim[j]) ||
						   (j > dim[0] && dim[j] > 1))
						{
							misMatch = 1;
							break;
						}
					}
				}
				if(misMatch)
				{
					sprintf(OUTSTR,
							"ERROR: %s for %s not scalar and has wrong dimensions",
							parName[i-1],FUNCNAME);
					goto errorExit;
				} /*if(misMatch)*/
			} /*if(sizes[i] != 1)*/
		} /*if(i >= 1)*/
	} /*for(i=0;i<nargs;i++)*/
	
	ndims = dim[0];

	result = RInstall(SCRATCH,size);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	y = DATA(result);
	setNDIMS(result,ndims);
	for(i = 1; i <= ndims; i++)
	{
		setDIM(result,i,dim[i]);
	}
/*
   restrictions on parameter values:
   cumnor(x)
   cumstu(x,PosReal)
   cumchi(x,PosReal)
   cumF(x,PosReal,PosReal[,PosReal])
   cumpois(x,PosReal)
   cumbin(x,PosInt,p)
   invnor(p)
   invstu(p,PosReal)
   invchi(p,PosReal)
   invF(p,PosReal,PosReal)
   cumgamma(x,PosReal)
   invgamma(p,PosReal)
   cumbeta(p,PosReal,PosReal[,PosReal])
   invbeta(p,PosReal,PosReal)
   where x is real and 0 <= p <= 1 is real
*/

	for(i=0;i<3;i++)
	{
		param[i] = 0.0;
		ii[i] = foundBadParam[i] = checkedParam[i] = 0;
	}
	
	for(i = 0;i < size;i++)
	{
#ifdef MACINTOSH
		if(timetocheck(i) && interrupted() != INTNOTSET)
		{
			goto intrrupt;
		}
#endif /*MACINTOSH*/
	
		if(sizes[0] == 1)
		{
			x = DATAVALUE(arg[0],0);
		}
		else
		{
			x = DATAVALUE(arg[0],i);
		}
		if(isMissing(x))
		{
			arithMissingFound = 1;
		}
		
		for(j=1;j<nargs;j++)
		{
			if(!checkedParam[j-1])
			{
				param[j-1] = DATAVALUE(arg[j],ii[j-1]++);
			}
		}

		switch (iop)
		{
		  case ICUMNOR:
			if(!isMissing(x))
			{
				x = Cnor(x);
			}
			break;
		  case ICUMCHI:
		  case ICUMGAMMA:
			if(!checkedParam[0] && !POSREAL(param[0]))
			{
				foundBadParam[0] = 1;
				setMissing(x);
			}
			if(!isMissing(x))
			{
				x = (iop == ICUMCHI) ? Cchi(x,param[0]) : Cgam(x,param[0],1.0);
			}
			break;
		  case ICUMSTU:
			if(!checkedParam[0] && !POSREAL(param[0]))
			{
				foundBadParam[0] = 1;
				setMissing(x);
			}
			if(!isMissing(x))
			{
				x = Cstu(x,param[0]);
			}
			break;
		  case ICUMF:
		  case ICUMBETA:
			if(!checkedParam[0] && !POSREAL(param[0]))
			{
				foundBadParam[0] = 1;
				setMissing(x);
			}
			if(!checkedParam[1] && !POSREAL(param[1]))
			{
				foundBadParam[1] = 1;
				setMissing(x);
			}
			if(nargs == 4 && !checkedParam[2] && !NONNEG(param[2]))
			{
				foundBadParam[2] = 1;
				setMissing(x);
			}
			if(!isMissing(x))
			{
				x = (iop == ICUMF) ? noncentF(x,param[2],param[0],param[1]) :
					noncentBeta(x,param[2],param[0],param[1]);
				if(nargs == 4 && x > 1.0)
				{
					nonConverge = 1;
					x = x - floor(x);
				}
			}
			break;
		  case ICUMPOI:
			if(!checkedParam[0] && !NONNEG(param[0]))
			{
				foundBadParam[0] = 1;
				setMissing(x);
			}
			if(!isMissing(x))
			{
				x = Cpoi((long) floor(x),param[0]);
			}
			break;
		  case ICUMBIN:
			if(!checkedParam[0] && (!POSINT(param[0]) || param[0] > MAXBINOMN))
			{
				foundBadParam[0] = 1;
				setMissing(x);
			}
			if(!checkedParam[1] && (!NONNEG(param[1]) ||param[1] > 1.0))
			{
				foundBadParam[1] = 1;
				setMissing(x);
			}
			if(!isMissing(x))
			{
				x = Cbin((long) floor(x),(long) param[0], param[1]);
				x = (x < 0.) ? 0.0 : ((x > 1.0) ? 1.0 : x);
			}
			break;
			
		  default:
			if(!isMissing(x) && (x <= 0 || x >= 1.))
			{
				foundOutOfRange = 1;
				setMissing(x);
			}
			switch (iop)
			{ /* all the inverses */
			  case IINVNOR:
				if(!isMissing(x))
				{
					x = Qnor(x);
				}
				break;
			  case IINVCHI:
			  case IINVGAMMA:
				if(!checkedParam[0] && !POSREAL(param[0]))
				{
					foundBadParam[0] = 1;
					setMissing(x);
				}
				if(!isMissing(x))
				{
					x = (iop == IINVCHI) ? 
						Qchi(x,param[0]) : Qgam(x,param[0],1.0);
				}
				break;
			  case IINVSTU:
				if(!checkedParam[0] && !POSREAL(param[0]))
				{
					foundBadParam[0] = 1;
					setMissing(x);
				}
				if(!isMissing(x))
				{
					x = Qstu(x,param[0]);
				}
				break;
			  case IINVF:
			  case IINVBETA:
				if(!checkedParam[0] && !POSREAL(param[0]))
				{
					foundBadParam[0] = 1;
					setMissing(x);
				}
				if(!checkedParam[1] && !POSREAL(param[1]))
				{
					foundBadParam[1] = 1;
					setMissing(x);
				}
				if(!isMissing(x))
				{
					x = (iop == IINVF) ? Qsne(x,param[0],param[1]) :
						Qbeta(x,param[0],param[1]);
				}
				break;
			} /*switch (iop)*/
		} /*switch (iop)*/

		if(sizes[1] == 1)
		{ /* nargs > 1 */
			if(foundBadParam[0])
			{
				break;
			}
			else
			{
				checkedParam[0] = 1;
			}
		}

		if(sizes[2] == 1)
		{ /* nargs > 2 */
			if(foundBadParam[1])
			{
				break;	
			}
			else
			{
				checkedParam[1] = 1;
			}
		}

		if(sizes[3] == 1)
		{ /* nargs > 3 */
			if(foundBadParam[2])
			{
				break;	
			}
			else
			{
				checkedParam[2] = 1;
			}
		}
		(*y)[i] = x;
	} /*for(i = 0;i < size;i++)*/
	UNLOADSEG(Cnor);
	UNLOADSEG(Cchi);
	UNLOADSEG(Cstu);
	UNLOADSEG(noncentF);
	UNLOADSEG(noncentBeta);
	UNLOADSEG(Cpoi);
	UNLOADSEG(Cbin);
	UNLOADSEG(Qnor);
	UNLOADSEG(Qchi);
	UNLOADSEG(Qstu);
	UNLOADSEG(Qchi);
	UNLOADSEG(Qsne);
	UNLOADSEG(Qbeta);

	
	if(nonConverge)
	{
		sprintf(OUTSTR,"WARNING: series for %s%s not fully converged",
				(size > 1) ? "at least one value of " : "", FUNCNAME);
		putOUTSTR();
	}
	
	for(j=1;j<nargs;j++)
	{
		if(foundBadParam[j-1])
		{
			if(sizes[j] == 1)
			{
				sprintf(OUTSTR,
						"ERROR: illegal value of %s for %s",
						parName[j-1],FUNCNAME);
				goto errorExit;
			}
			else
			{
				sprintf(OUTSTR,
						"WARNING: found illegal value of %s for %s",
						parName[j-1],FUNCNAME);
				putOUTSTR();
			}
		}
	}
	*OUTSTR = '\0';

	if(foundBadParam[0] || foundBadParam[1])
	{
		myerrorout("WARNING: result set to MISSING");
	}
	
	if(foundOutOfRange)
	{
		sprintf(OUTSTR,
				"WARNING: %s has probability <= 0 or >= 1. Result set to MISSING",
				FUNCNAME
		);
		putOUTSTR();
	}
	if(arithMissingFound)
	{
		myerrorout("WARNING: arithmetic with MISSING values.  Result set to MISSING");
	}

	return (result);

  errorExit:
	putOUTSTR();
	/* fall through */

#ifdef MACINTOSH
  intrrupt:
#endif /*MACINTOSH*/
	Removesymbol(result);
	return (0);
	
}

