/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Qr
#else
#define __SEG__ Qr
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include "linpack.h"

#include <math.h>

#define ITRILOWER      1
#define ITRIUPPER      2
#define ITRIUNPACK     3
#define IQR            4

static void triangle(double **hIn, double **hOut, long nrowsIn, long ncolsIn,
					 long iop, char *keyword)
{
	long           i, j;
	double        *inj, *outij, *outji;
	long           nc;
	WHERE("triangle");
	
	outij = outji = *hOut;
	if(iop != ITRIUNPACK)
	{ /* triupper or trilower */
		if(*keyword == '\0')
		{						/* full triangle (triupper(x) or trilower (x)*/
			inj = *hIn;
		
			for(j = 0;j<ncolsIn;j++)
			{
				for(i=0;i<nrowsIn;i++)
				{
					if(i > j)
					{
						*outij++ = (iop == ITRILOWER) ? *inj : 0.0;
					}
					else if(i < j)
					{
						*outij++ = (iop == ITRIUPPER) ? *inj : 0.0;
					}
					else
					{
						*outij++ = *inj;
					}
					inj++;
				} /* for i */
			} /* for j */
		}
		else if(strcmp(keyword,"pack") == 0)
		{						/* packed form */
			if(iop == ITRIUPPER)
			{ /* triupper(x,T) */
				for(j=0;j<ncolsIn;j++)
				{
					inj = *hIn + j*nrowsIn;
					for(i=0;i<=j;i++)
					{
						*outij++ = *inj++;
					} /* for i */
				} /* for j */
			}
			else
			{ /* trilower(x,T) */
				for(i=0;i<nrowsIn;i++)
				{
					inj = *hIn + i;
					for(j=0;j<=i;j++)
					{
						*outij++ = *inj;
						inj += nrowsIn;
					} /* for i */
				} /* for i */
			}
		}
		else if(strcmp(keyword,"square") == 0)
		{ /* make output ncols by ncols (only for triupper) */
			nc = (ncolsIn < nrowsIn) ? ncolsIn : nrowsIn;
			for(j = 0;j<nc;j++)
			{
				inj = *hIn + j*nrowsIn;
				for(i=0;i<nc;i++)
				{
					if(i > j)
					{
						*outij++ = 0.0;
					}
					else
					{
						*outij++ = *inj;
					}
					inj++;
				} /* for i */
			} /* for j */
		}
	}
	else
	{ /* triunpack() */
		inj = *hIn;
		if(strcmp(keyword,"lower") == 0)
		{ /* make lower triangle */
			for(i=0;i<nrowsIn;i++)
			{
				outij = *hOut + i;
				for(j=0;j<nrowsIn;j++)
				{
					*outij = (j <= i) ? *inj++ : 0.0;
					outij += nrowsIn;
				} /* for j */
			} /* for i*/
		}
		else if(strcmp(keyword,"upper") == 0)
		{ /* make upper triangular matrix */
			for(j=0;j<nrowsIn;j++)
			{
				outij = *hOut + j*nrowsIn;
				for(i=0;i<nrowsIn;i++)
				{
					*outij++ = (j >= i) ? *inj++ : 0.0;
				} /* for i */
			} /* for j */
		}
		else
		{ /* make fully symmetric */
			for(j=0;j<nrowsIn;j++)
			{
				outij = *hOut + j*nrowsIn;
				outji = *hOut + j;
				for(i = 0;i<j;i++)
				{
					*outij++ = *outji = *inj++;
					outji += nrowsIn;
				} /* for i */
				*outij = *inj++;
			} /* for j */
		}
	}
}

static void doQr(double **hx, double **hqraux, double **hr, long **hpivot,
				 double **hwork, long nrows, long ncols)
{
	long            nvals = nrows*ncols;
	double         *x = *hx, *r = *hr;
	long            pivotit = (hpivot != (long **) 0);
	long            i;
	WHERE("doQr");
	
	while(nvals-- > 0)
	{
		*r++ = *x++;
	}
	if(pivotit)
	{
		for(i=0;i<ncols;i++)
		{
			(*hpivot)[i] = 0;
		}
	}
	
	DQRDC(*hr, &nrows, &nrows, &ncols, *hqraux, 
		   (pivotit) ? *hpivot : (long *) 0,
		   (pivotit) ? *hwork : (double *) 0, &pivotit);

	UNLOADSEG(DDOT);
	UNLOADSEG(DQRDC);

}

typedef struct opEntry
{
	char *name;
	short iop;
	short nargs;
} opEntry;

static opEntry Ops[] =
{ /* name        iop             nargs */
	"trilower",  ITRILOWER,       -2, /* trilower(a [,pack:T]) */
	"triupper",  ITRIUPPER,       -2, /* triupper(a [,pack:T]) */
	"triunpack", ITRIUNPACK,      -2, /* triunpack(a [,upper|lower:T) */
	"qr",        IQR,             -2, /* qr(x [,ronly]:T) */
	(char *) 0,  0,                0
};

#define GHWORK         0
#define GHPIVOT        1
#define NTRASH         2

Symbolhandle qr(list)
Symbolhandle          list;
{
	Symbolhandle      arg[2];
	Symbolhandle      r = (Symbolhandle) 0, qraux = (Symbolhandle) 0;
	Symbolhandle      pivot = (Symbolhandle) 0;
	Symbolhandle      result = (Symbolhandle) 0;
	long            **hpivot = (long **) 0;
	double          **hwork = (double **) 0;
	long              dims[2], nrows, ncols;
	long              margs, iop, nargs = NCOMPS(list), nvals;
	long              i;
	long              logValue = 0, square = 0, ronly = 0,pivotit = 0;
	char             *keyword = "", *legalKeywords[3], **kw;
	double            tmp;
	opEntry          *op;
	char             *compNames[3];
	WHERE("qr");
	TRASH(NTRASH,errorExit);
	
	*OUTSTR = '\0';
	compNames[0] = "qr";
	compNames[1] = "qraux";
	compNames[2] = "pivot";
	
	for(op=Ops;op->iop;op++)
	{
		if(strcmp(op->name,FUNCNAME) == 0)
		{
			break;
		}
	}
	if(op->iop == 0)
	{
		sprintf(OUTSTR,"ERROR: %s not implemented",FUNCNAME);
		goto errorExit;
	}
	iop = op->iop;
	margs = op->nargs;
	if(margs > 0 && nargs != margs || margs < 0 && nargs > labs(margs))
	{
		badNargs(FUNCNAME,margs);
		goto errorExit;
	}
	for(i=0;i<nargs;i++)
	{
		arg[i] = COMPVALUE(list,i);
	}
	
	if(!argOK(arg[0],REAL,1))
	{
		goto errorExit;
	}

	if(!isMatrix(arg[0],dims))
	{
		sprintf(OUTSTR,"ERROR: argument 1 to %s is not %s",FUNCNAME,
				(iop != ITRIUNPACK) ? "matrix" : "vector");
		goto errorExit;
	}
	nrows = dims[0];
	ncols = dims[1];
	if(iop == ITRIUNPACK)
	{
		tmp = sqrt(1.0 + 8.0*(double) nrows);
		if(!isVector(arg[0]) || 
		   floor(tmp + 1e-12) != ceil(tmp - 1e-12))
		{
			sprintf(OUTSTR,
					"ERROR: 1st argument to %s is not vector with length of form n*(n+1)/2",
					FUNCNAME);
			goto errorExit;
		}
		else
		{
			nrows = ((long) floor(tmp + 1e-12) - 1) / 2;
		}
	}
	
	legalKeywords[0] = "";
	legalKeywords[1] = legalKeywords[2] = (char *) 0;
	switch(iop)
	{
	  case ITRIUPPER:
		legalKeywords[1] = "square";
	  case ITRILOWER:
		legalKeywords[0] = "pack";
		break;
	  case IQR:
		legalKeywords[0] = "pivot";
		legalKeywords[1] = "ronly";
		break;
	  case ITRIUNPACK:
		legalKeywords[0] = "lower";
		legalKeywords[1] = "upper";
		break;
	}

	if(nargs == 2)
	{
		if((keyword = isKeyword(arg[1])))
		{
			for(kw = legalKeywords;*kw && strcmp(keyword,*kw) != 0;kw++)
				;
			if(*kw == (char *) 0)
			{
				sprintf(OUTSTR,
						"ERROR: %s is not valid keyword for %s",keyword,
						FUNCNAME);
				goto errorExit;
			}
		}
		else
		{ /* use of keyword is optional */
			keyword = legalKeywords[0];
		}
		
		if(!argOK(arg[1],LOGIC,2))
		{
			goto errorExit;
		}
		if(!isScalar(arg[1]))
		{
			sprintf(OUTSTR,"ERROR: % must be T or F",keyword);
			goto errorExit;
		}
		logValue = DATAVALUE(arg[1],0);
		
		if(iop == ITRILOWER || iop == ITRIUPPER)
		{
			square = (strcmp(keyword,"square") == 0);
			if(square && !logValue)
			{
				square = 0;
				keyword = "";
			}
			
			if(logValue && nrows != ncols && !square)
			{
				sprintf(OUTSTR,
						"ERROR: input matrix must be square when %s is T",
						keyword);
				goto errorExit;
			}
		}
		else if(iop == IQR)
		{
			pivotit = (strcmp(keyword,"pivot") == 0);
			if(pivotit)
			{
				if(!logValue)
				{
					pivotit = 0;
				}
				else
				{
					logValue = 0;
				}
			}
			ronly = logValue;
		}			
	} /* if(nargs == 2) */
	
	/* nvals is space needed for (principal) output matrix */
	if(iop == ITRILOWER || iop == ITRIUPPER)
	{
		if(!logValue)
		{
			nvals = nrows*ncols;
		}
		else if(!square)
		{
			nvals = (nrows*(nrows+1))/2;
		}
		else
		{
			nvals = (ncols < nrows) ? ncols : nrows;
			nvals = nvals*nvals;
		}
	}
	else if(iop == IQR)
	{
		nvals = nrows*ncols;
	}
	else if(iop == ITRIUNPACK)
	{
		if(!logValue && *keyword)
		{ /* swap keywords if F */
			keyword = (strcmp(keyword,legalKeywords[0]) == 0) ?
				legalKeywords[1] : legalKeywords[0];
		}
		nvals = nrows * nrows;
	}
	if(iop != IQR || ronly)
	{
		result = RInstall(SCRATCH,nvals);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
	}
	else
	{ /* IQR && !ronly, save complete results in structure */
		result = StrucInstall(SCRATCH,(pivotit) ? 3L : 2L);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
	}
		
	if(iop == IQR)
	{
		/* create symbols r and qraux */
		r = RInstall(SCRATCH,nvals);
		qraux = RInstall(SCRATCH,ncols);
		if(r == (Symbolhandle) 0 || qraux == (Symbolhandle) 0)
		{
			goto errorExit;
		}			
		if(!ronly)
		{
			Cutsymbol(r);
			COMPVALUE(result,0) = r;
			Cutsymbol(qraux);
			COMPVALUE(result,1) = qraux;
			setNAME(r,compNames[0]);
			setNAME(qraux,compNames[1]);
		}
		else
		{
			setNDIMS(result,2);
			setDIM(result,1,ncols);
			setDIM(result,2,ncols);
		}
		
		setNDIMS(r,2);
		setDIM(r,1,nrows);
		setDIM(r,2,ncols);

		if(pivotit)
		{
			if(!getScratch(hpivot,GHPIVOT,ncols,long) ||
			   !getScratch(hwork,GHWORK,ncols,double) ||
			   (pivot = RInstall(SCRATCH,ncols)) == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			if(!ronly)
			{
				Cutsymbol(pivot);
				COMPVALUE(result,2) = pivot;
				setNAME(pivot,compNames[2]);
			}
		} /*if(pivotit)*/
	} /*if(iop == IQR)*/
	else if((iop == ITRIUPPER || iop == ITRILOWER) && (!logValue || square))
	{
		setNDIMS(result,2);
		setDIM(result,1,(square) ? (long) sqrt(.1+(double) nvals) : nrows);
		setDIM(result,2,(square) ? DIMVAL(result,1) : ncols);
	}
	else if(iop == ITRIUNPACK)
	{
		setNDIMS(result,2);
		setDIM(result,1,nrows);
		setDIM(result,2,nrows);
	}
	
	switch (iop)
	{
	  case ITRILOWER:
	  case ITRIUPPER:
	  case ITRIUNPACK:
		triangle(DATA(arg[0]),DATA(result),nrows,ncols, iop, keyword);
		break;
	  case IQR:
		doQr(DATA(arg[0]),DATA(qraux),DATA(r), 
			 (pivotit) ? hpivot : (long **) 0,
			 (pivotit) ? hwork : (double **) 0, nrows, ncols);
		if(pivotit)
		{
			for(i=0;i<ncols;i++)
			{
				DATAVALUE(pivot,i) = (double) (*hpivot)[i];
			}
		}
		
		if(ronly)
		{
			triangle(DATA(r),DATA(result),nrows,ncols,ITRIUPPER,"square");
			Removesymbol(qraux);
			Removesymbol(r);
		}
		break;
	}

	emptyTrash();
	return(result);
	
  errorExit:
	putOUTSTR();
	emptyTrash();
	if(iop == IQR && ronly)
	{
		Removesymbol(r);
		Removesymbol(qraux);
		if(pivotit)
		{
			Removesymbol(pivot);
		}
	}
	
	Removesymbol(result);
	
	return (0);
}



