/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Random
#else
#define __SEG__ Random
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

#define RANDM1 2147483563
#define RANDA1 40014
#define RANDQ1 53668
#define RANDR1 12211

#define RANDM2 2147483399
#define RANDA2 40692
#define RANDQ2 52774
#define RANDR2 3791

#ifdef MACINTOSH
#ifdef MPW
#include <OSUtils.h>
#endif /*MPW*/
#else /*MACINTOSH*/
#ifdef HASTIMEH
#include	<time.h>
#endif /*HASTIMEH*/
#endif /*MACINTOSH*/

#ifdef EPX
typedef long time_t;
#endif /*EPX*/

extern Symbolhandle    RInstall();
extern long            argOK();
extern void            badNargs();
extern long            isScalar();

Symbolhandle           rangen();
Symbolhandle           setseed();
Symbolhandle           getseed();

double                 uni();
void                   vuni();
void                   vnorm();

/*
static long            rands1 = 0;
static long            rands2 = 0;
*/

#define rands1 RANDS1  /* global long */
#define rands2 RANDS2  /* global long */
static char              *OpNames[] =
{
	"rnorm", "runi", (char *) 0
};

#define IRNORM   1
#define IRUNI    2

static void randomSeed(long verbose)
{
#ifdef  MACINTOSH
	GetDateTime((unsigned long *) &rands1);
#else /*MACINTOSH*/
#ifdef HASTIMEH
	rands1 = (long) time((time_t *) 0);
#else /*HASTIMEH*/
	rands1 = 1891756027;
#endif /*HASTIMEH*/
#endif /*MACINTOSH*/

	rands2 = rands1 & 1360626473;
/* throw away first two numbers */
	(void) uni();
	(void) uni();
	if (verbose)
	{
		sprintf(OUTSTR,"NOTE: random number seeds set to %ld and %ld",
				rands1, rands2);
		putOUTSTR();
	}	
} /*randomSeed()*/

static long                OpCodes[] =
{
	IRNORM, IRUNI
};

/* These are for the combined generator suggested for 32 bit machines
   in P. L'Ecuyer 1988 Comm. ACM  */

Symbolhandle    rangen(list)
Symbolhandle    list;
{
	Symbolhandle    result = (Symbolhandle) 0, symhf;
	double          sampleSize;
	long            op, verbose = 1;
	WHERE("rangen");
	
	/* generate a list of random numbers */
	op = matchKey(FUNCNAME,OpNames,OpCodes);
	OUTSTR[0] = '\0';
	
	if(NARGS(list) != 1)
	{
		badNargs(FUNCNAME,1);
		goto errorExit;
	}
	
	symhf = COMPVALUE(list,0);
	if (!argOK(symhf,REAL,1))
	{
		goto errorExit;
	}

	if(!isScalar(symhf) || (sampleSize = floor(DATAVALUE(symhf,0))) <= 0 ||
	   sampleSize != DATAVALUE(symhf,0))
	{
		sprintf(OUTSTR,
				"ERROR: argument to %s must be single positive integer",FUNCNAME);
		goto errorExit;
	}
	if(sampleSize > (double) MAXCOORD)
	{
		sprintf(OUTSTR,
				"ERROR: sample size for %s is too large",FUNCNAME);
		goto errorExit;
	}
	
	if (rands1 == 0 && rands2 == 0)
	{
		randomSeed(verbose);
	}

	result = RInstall(SCRATCH, (long) sampleSize);
	if (result != (Symbolhandle) 0)
	{
		switch (op)
		{
		  case IRUNI:
			vuni((long) sampleSize,DATAPTR(result));
			break;
		  case IRNORM:
			vnorm((long) sampleSize,DATAPTR(result));
			break;
		}
	}
	return (result);

  errorExit:
	putOUTSTR();

	return (0);
	
}
/*
   Seed the random number generator
   Usage: setseeds(seed1, seed2) or setseeds(cat(seed1,seed2))
*/

Symbolhandle    setseed(list)
Symbolhandle    list;
{
	Symbolhandle    symhf, symhl, symhKey;
	double          first, last;
	long            nargs = NARGS(list);
	long            margs = nargs;
	long            verbose = 1;
	char           *keyword;
	
	OUTSTR[0] = '\0';

	if (nargs > 3)
	{
		badNargs(FUNCNAME, -3);
		goto errorExit;
	}
	if (nargs > 1)
	{ /* check for quiet:T or simply logical scalar 3rd argument */
		symhKey = COMPVALUE(list,nargs-1);
		if (!argOK(symhKey, 0, nargs))
		{
			goto errorExit;
		}
		if ((keyword = isKeyword(symhKey)) && strcmp(keyword, "quiet") != 0)
		{
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
		if (isScalar(symhKey) && TYPE(symhKey) == LOGIC)
		{
			verbose = (DATAVALUE(symhKey,0) == 0);
			nargs--;
		}
		else if(keyword || nargs == 3)
		{
			notTorF((keyword) ? keyword : "argument 3");
			goto errorExit;
		}
	} /*if (nargs > 1)*/
	
	symhf = COMPVALUE(list,0);
	if(!argOK(symhf, REAL, (margs > 1) ? 1 : 0))
	{
		goto errorExit;
	}

	if(nargs == 2)
	{
		if(!argOK(symhl = COMPVALUE(list,1), REAL,2))
		{
			goto errorExit;
		}
		if(!isScalar(symhf) || !isScalar(symhl))
		{
			sprintf(OUTSTR,
					"ERROR: when seeds are separate arguments, both must be REAL scalars");
		}
		else
		{
			first = DATAVALUE(symhf,0);
			last = DATAVALUE(symhl,0);
		}		
	} /*if(nargs == 2)*/
	else
	{
		if(!isVector(symhf) || symbolSize(symhf) != 2)
		{
			sprintf(OUTSTR,
					"ERROR: when seeds are a vector argument to %s it must have length 2",
					FUNCNAME);
		}
		else
		{
			first = DATAVALUE(symhf,0);
			last = DATAVALUE(symhf,1);
		}
	} /*if(nargs == 2){}else{}*/
	if(*OUTSTR)
	{
		goto errorExit;
	}
	
	if(isMissing(first) || isMissing(last))
	{
		sprintf(OUTSTR,"ERROR: missing values not allowed as seeds");
	}
	else if(first != floor(first) || first < 0 ||
			last != floor(last) || last < 0 ||
			first > RANDM1 || last > RANDM2)
	{
		sprintf(OUTSTR,
				"ERROR: seeds must non-negative integers, seed1 <= %ld and seed2 <= %ld",
				RANDM1, RANDM2);
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}
	

	if(first == 0 || last == 0)
	{
		randomSeed(verbose);
	}
	else
	{
		rands1 = first;
		rands2 = last;
	}	

	return (NULLSYMBOL);

  errorExit:
	
	putOUTSTR();
	return (0);

}

Symbolhandle    getseed(Symbolhandle list)
{
	Symbolhandle    result = (Symbolhandle) 0;
	Symbolhandle    symhKey;
	char           *keyword;
	long            verbose = 1;
	long            nargs = NARGS(list);
	
	if(nargs > 1)
	{
		badNargs(FUNCNAME, -1);
		goto errorExit;
	}
	
	symhKey = COMPVALUE(list,0);
	if (symhKey != (Symbolhandle) 0)
	{
		if (!argOK(symhKey, 0, 0))
		{
			goto errorExit;
		}
		if ((keyword = isKeyword(symhKey)) && strcmp(keyword, "quiet") != 0)
		{
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
		if (!isScalar(symhKey) || TYPE(symhKey) != LOGIC)
		{
			notTorF("argument");
			goto errorExit;
		}
		verbose = (DATAVALUE(symhKey,0) == 0);
	}
	if (verbose)
	{
		sprintf(OUTSTR, "Seeds are %ld and %ld", rands1, rands2);
		putOUTSTR();
	}
	
	result = RInstall(SCRATCH, 2);
	
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	DATAVALUE(result,0) = rands1;
	DATAVALUE(result,1) = rands2;

	return (result);

  errorExit:
	return ((Symbolhandle) 0);

}

double          uni()
{
	long            z, k;

	k = rands1 / RANDQ1;
	rands1 = RANDA1 * (rands1 - k * RANDQ1) - k * RANDR1;
	if (rands1 < 0)
	{
		rands1 += RANDM1;
	}
	
	k = rands2 / RANDQ2;
	rands2 = RANDA2 * (rands2 - k * RANDQ2) - k * RANDR2;
	if (rands2 < 0)
	{
		rands2 += RANDM2;
	}
	
	z = rands1 - rands2;
	if (z < 1)
	{
		z += RANDM1 - 1;
	}
	
	return (z / (double) RANDM1);
}

void vuni(count, rvec)
long            count;
double         *rvec;

{
	long            i;
	for (i = 0; i < count; i++)
	{
		*(rvec + i) = uni();
	}
}

void vnorm(count, rvec)
long            count;
double         *rvec;
{
	long            i;
	double          u, v, x, z;
	for (i = 0; i < count; i++)
	{
		for (;;)
		{
			u = uni();
			v = 1.715528 * (uni() - 0.5);
			x = v / u;
			z = x * x / 4.;
			if (z < 1 - u)
			{
				break;
			}
			if (z > .25924 / u + 0.35)
			{
				continue;
			}
			if (z < -log(u))
			{
				break;
			}
		}
		rvec[i] = x;
	}
}
