/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Readdata
#else
#define __SEG__ Readdata
#endif /*MPW3*/
#endif /*MPW*/

#include <stdio.h>
#include "matdat.h"
#include "globals.h"
#include "dbug.h"

#define MIN(a, b) (((a) < (b)) ? (a) : (b) )
#define GETLINE1(F) (status != EOF && (status = fillLINE(F), (*LINE)[0]))

extern long readLine();

/* readFile can handle arbitrary dimensioned arrays */

long readFile(file, dims, nFields, fmt, nSkip, byRows, symhy, echo)
FILE *file;
long dims[], nFields, nSkip, byRows, echo;
char *fmt;
Symbolhandle symhy;
{
	double        *y = DATAPTR(symhy);
	long           i, j, k, l;
	long           nRead, nExpected, mExpected, nLeft;
	long           nToRead, readSome;
	long           m, n;
	long           ndims, size, howmany, status = 1;
	long           incs[MAXDIMS];
	long           ii[MAXDIMS];
	long           place[MAXDIMS];
	static         double x[MAXN];
	WHERE("readFile");

	/*
	  it is assumed that first line of data is already in LINE
	  readLine scans LINE, fillLINE refills it
	  */
/* 
  mExpected is number possibly multiline 'lines' expected in the file
  nExpected is number items expected per 'line'
*/
	
	ndims = dims[0];
	size = 1;
	for(i = 1;i <= ndims;i++)
	{
		size *= dims[i];
	}
	if(byRows)
	{
		incs[0] = size;
		for(j = 1;j <= ndims;j++)
		{
			ii[j-1] = 0;
			place[j-1] = 0;
			incs[j] = incs[j-1]/dims[ndims-j+1];
		}
	}
	
	n = dims[ndims];
	m = (ndims > 1) ? size/n : 1;
	
	nExpected = (byRows) ? n : m;
	mExpected = (byRows) ? m : n;
	
	for(i = 0;i < size;i++)
	{
		setMissing(y[i]);
	}
	
	l = howmany = 0;
	for(i = 0;i < mExpected;i++)
	{
		if(i > 0 && !GETLINE1(file))
		{
			break;
		}
		if(echo)
		{
			echoLINE();
		}
		nLeft = nExpected;
		readSome = 0;
		while(nLeft > 0)
		{
			/*
			  stop reading on empty line or line with non-numeric first field
			  */
			nToRead = MIN(nLeft, nFields);
			nRead = readLine(fmt, nSkip, x);
			
			if((nRead > 0 && nRead < nToRead) || (nRead <= 0 && readSome))
			{
				sprintf(OUTSTR, "WARNING: unreadable or missing item in %s %ld",
						(byRows)? "case" : "column",
						(ndims == 1) ? howmany + nRead + 1 :i+1);
				putOUTSTR();
			}
			if(nRead <= 0)
			{
				break;
			}
			for(j = nRead;j < nToRead;j++)
			{
				setMissing(x[j]);
			}
			howmany += nToRead;
			for(k = 0; k < nToRead; k++)
			{
				if(byRows)
				{
		/* the following is essentially equivalent to ndims nested for loops */
					y[l] = x[k];
					for (j = 1; j <= ndims; j++)
					{
						if(ii[j-1] == 0)
						{
							place[j-1] = l;
						}
						if((ii[j-1] += incs[j]) < incs[j-1])
						{
							l += incs[j];
							break;
						}
						ii[j-1] = 0;
						l = place[j-1];
					} /*for (j = 1; j <= ndims; j++)*/
				} /*if(byRows)*/
				else
				{
					y[l++] = x[k];
				}
			} /*for(k = 0; k < nToRead; k++)*/
			
			readSome += nToRead;
			nLeft -= nToRead;
			if(nLeft > 0)
			{
				if(!GETLINE1(file))
				{
					break;
				}
				nRead = strlen(*LINE);
				if(echo)
				{
					echoLINE();
				}
			} /*if(nLeft > 0)*/
		} /* while(nLeft > 0) */
		if(nRead <= 0)
		{
			break;
		}
	} /*for(i = 0;i < mExpected;i++)*/

	/* complete partially read multiline row or column */

	if(i < mExpected)
	{
			
		sprintf(OUTSTR,
				"WARNING: %ld %s read are fewer than %ld specified on header",
				i, (byRows) ? "rows" : "columns", m);
		putOUTSTR();
	}

	*OUTSTR = '\0';
	return (howmany > 0) ? howmany : NODATA;
} /* readFile*/

#define CHUNK   (10*sizeof(double))

long readMacro(file, nlines, symhtext, echo)
FILE            *file;
long             nlines, echo;
Symbolhandle     symhtext;
{
	long          c = 0, nchar = 0, error = -2, nMax = CHUNK;
	char        **text = (char **) 0;
	long          jline = 1, iplace = 0, status = 1;
	WHERE("readMacro");
	
/* it is assumed that *LINE already has something in it */
	if(echo)
	{
		echoLINE();
	}
	
	text = mygethandle(nMax+1);
	setSTRING(symhtext, text);
	if(text == (char **) 0)
	{
		goto errorExit;
	}

	while(jline <= nlines)
	{
		c = (*LINE)[iplace++];
		if(c == '\0')
		{
			if(jline == nlines || !GETLINE1(file))
			{
				break;
			}
			jline++;
			if(echo)
			{
				echoLINE();
			}
			iplace = 0;
			c = (*LINE)[iplace++];
		}
		(*text)[nchar++] = c;
		if(nchar > nMax)
		{
			nMax += CHUNK;
			text = mygrowhandle(text, nMax+1);
			setSTRING(symhtext, text);
			if(text == (char **) 0)
			{
				goto errorExit;
			}
		}
	}	/*while(jline <= nlines) */
	(*text)[--nchar] = '\0'; /* replace last newline with '\0' */
	
	if((*LINE)[0] == '\0')
	{
		sprintf(OUTSTR,
				"WARNING: only %ld lines of macro found, < %ld specified on header",
				jline, nlines);
		putOUTSTR();
	}
	if(myhandlelength(text) > nchar + 1)
	{ /* trim handle down to correct size */
		text = mygrowhandle(text, nchar+1);
		setSTRING(symhtext, text);
		if(text == (char **) 0)
		{
			goto errorExit;
		}
	}

	return (nchar);
	
  errorExit:
	
	return (error);

} /* readMacro */

		
