/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Readdata
#else
#define __SEG__ Readdata
#endif /*MPW3*/
#endif /*MPW*/

#include <stdio.h>
#include <ctype.h>
#include "globals.h"
#include "matdat.h"
#include "dbug.h"

#define MAXNF 20
#define FIELDSIZE MAXLINE


/* macro GETLINE is false if fillLINE() obtains no data */

#define GETLINE(F) (fillLINE(F), (*LINE)[0])

/*
   function to seek data set or Macro matName on FILE * file
   Input
   	file		should be opened FILE pointer
	matName		name of object; if "", then object starts in first
				non-blank line in the file
	verbose		If > 0, echo header to stdout
	 			If < 0, call must be to read the header lines only
				If == 0, skip comment lines
   Output
	dims		Dimensions of object
	nFields		The maximum number of items per line
	fmt			Address of a format
	byRows		0 only if COLUMNS or COLS specified on header; otherwise 1
	type		type of object read, e.g., REAL, LOGICAL, MACRO
	missValue	Missing value code, if any.

  Return value	NOERROR, NOTFOUND, BADHEADER, TWOFORMATS
				On NOERROR, *LINE contains first line after header (verbose >= 0)
				or the first line of the header (verbose < 0)


*/
long         readHeader(file, dims, nFields, fmt, nSkip, byRows, type, missValue,
						matName, verbose)
FILE        *file;
long         dims[], *nFields, *byRows, *nSkip, *type, verbose;
double      *missValue;
char       **fmt, *matName;
{
	char     matName1[MAXNAME+1];
	char     field[FIELDSIZE+1];
	char    *pch, ch;
	long     fields[MAXNF];
	long     nf;
	long     fmtFound = 0;
	long     nameFound = 0;
	long     formatFound = 0;
	long     done = 0;
	long     error;
	long     i, j, c, clast, tot, place = 0;
	double   tmp;
	WHERE("readHeader");

	(*fmt)[0] = '\0';
	*nFields = 0;
	matName1[0] = '\0';
	field[FIELDSIZE] = '\0';

	if(*matName != '\0')
	{
		/* case is ignored in all comparisons */
		while(!nameFound)
		{
			if(fillLINE(file) == EOF)
			{
				return NOTFOUND;
			}
			if(sscanf(*LINE, "%20s", matName1) == 1 && 
			   mystrncmp(matName, matName1, MAXNAME) == 0)
			{
				nameFound = 1;
			}
		} /*while(!nameFound)*/
	} /*if(*matName != '\0')*/
/* 1st non-blank line of file is assumed to be header if no name is provided */		
	else
	{
		/* skip empty lines at start of file */
		do /*while (ch == NL || ch == CR || ch == '\0');*/
		{
			if(fillLINE(file) == EOF)
			{
				return NOTFOUND;
			}

			for(pch = *LINE;(ch = *pch) == ' ' || ch == '\t';pch++)
			{ /* skip leading white space, if any */
				;
			}
		} while (ch == NL || ch == CR || ch == '\0');
	}  /*if(*matName != '\0'){}else{}*/
	
	if(verbose < 0)
	{ /* must be inforead() */
		return (NOERROR);
	}
	
	nf = 0;
	clast = ' ';

	/* find start of fields on header lines */
	for(i=0;i < MAXLINE;i++)
	{
		c = (*LINE)[i];
		if(c == NL || c == CR || c == '\0')
		{
			break;
		}
		
		if((c != ' ' && c != '\t') && (clast == ' ' || clast == '\t'))
		{
			fields[nf++] = i;
			if(nf >= MAXNF)
			{
				break;
			}
		} /*if((c != ' ' && c != '\t') && (clast == ' ' || clast == '\t'))*/
		clast = c;
	} /*for(i=0;i < MAXLINE;i++)*/
	dims[0] = 0;
	tot = 1; /* product of numerical fields */
/*
  Read non-negative integer fields if any after name
  All dimensions == 0 is legal; some 0 and some non-zero is not
*/
	for(i = 1;i < nf;i++)
	{
		tmp = decodeString(*LINE + fields[i], '\0', &error);
		if(error)
		{
			break;
		}
		if(tmp != floor(tmp) || tmp < 0 || tot == 0 && tmp > 0)
		{
			return (BADHEADER);
		}
		dims[i] = (long) tmp;
		dims[0]++; /* ndims */
		tot *= dims[i];
	}
	if(dims[0] == 0)
	{ /* no dimensions found */
		return (BADHEADER);
	}
	*type = 0;
	*byRows = -1;
	if(i == nf)
	{ /* only name and dimension info */
		*type = REAL;
	}
	else
	{ /* look for keywords */
		for(j = i;j < nf;j++)
		{
			strncpy(field, *LINE + fields[j], FIELDSIZE);
			field[FIELDSIZE] = '\0';
			if(mystrncmp("MACRO", field, 5) == 0)
			{
				if(*type)
				{
					return (BADHEADER);
				}
				*type = MACRO;
			}
			else if(mystrncmp("COLUMNS", field, 7) == 0 ||
					mystrncmp("COLS", field, 4) == 0)
			{
				if(*byRows >= 0 || (*type != 0 && *type != REAL))
				{
					return (BADHEADER);
				}
				*type = REAL;
				*byRows = 0;
			}
			else if(mystrncmp("ROWS", field, 4) == 0)
			{
				if(*byRows >= 0 || (*type != 0 && *type != REAL))
				{
					return (BADHEADER);
				}
				*type = REAL;
				*byRows = 1;
			}
			else if(mystrncmp("FORMAT", field, 6) == 0)
			{
				if((*type != 0 && *type != REAL) || formatFound)
				{
					return (BADHEADER);
				}
				*type = REAL;
				formatFound = 1;
			}
			else
			{

				myerrorout("ERROR: Unrecognized keyword on header");
				return (BADHEADER);
			}
		} /* for */
	} /*if(i == nf){}else{}*/
	if(*byRows < 0)
	{
		*byRows = 1;
	}
	
	if(*type == MACRO && dims[0] > 1)
	{
		return (BADHEADER);
	}
	if(*type == REAL)
	{
		if(*byRows && dims[1] > 0)
		{
			*nFields = tot/dims[1];
		}
		else
		{
			*nFields = dims[1];
		}
	}
	
	if(verbose)
	{ /* echo header line to output */
		(*LINE)[strlen(*LINE)-1] = '\0'; /* trim off '\n' */
		myerrorout(*LINE); /* use myerrorout to make sure line is counted */
	}

	*nSkip = 0;
	
	/* now look for lines starting with ')' or '(' */
	do /* while(!done) */
	{
		if(!GETLINE(file))
		{
			return (NODATA);
		}
		place = 0;
		switch ((*LINE)[0])
		{
		  case FORMATSTART: /*'('*/
/*
  if reading matrix this is probably the start of a Fortran format, and 
  presumably is the last line before data and hence the next line should be
  read in to the buffer.  If reading a macro this is presumably the first
  line and should be left in the buffer
*/
			done = 1;
			if(*type == REAL || *type == LOGIC)
			{
				/* get following line and don't print Fortran format */
				if(!GETLINE(file) && dims[1] > 0)
				{
					return (NODATA);
				}
				place = 0;
			}
			continue;

		  case COMMENTSTART: /*')'*/
/*
  Descriptive comment, may contain sscanf format or MISSING   value or LOGIC
*/
			if(sscanf(*LINE+1, "%s %lf", field, &tmp) == 2 && 
			   strncmp(field, "MISSING", 7) == 0)
			{/* ') MISSING value' specifies value to be translated as MISSING*/
				*missValue = tmp;
			}
			else if(sscanf(*LINE+1, "%s", field) == 1 &&
					strncmp(field, "LOGIC", 5) == 0)
			{/* ') LOGIC' specifies data is LOGICAL */
				*type = LOGIC;
			}
			else if((*type == REAL || *type == LOGIC) && (*LINE)[1] == '"')
			{/* comment starting ')"' contains sscanf format */
				if(fmtFound)
				{
					return TWOFORMATS;
				}
				/* determine columns to ignore and number of fields per line*/
				place = 2;
				*nSkip = 0;

	/* recognize mX or mx (e.g., '7x' or '7X') as indicating columns to skip */
				while(isdigit(c = (*LINE)[place]))
				{
					*nSkip = 10*(*nSkip) + (c - '0');
					place++;
				}
				
				if(*nSkip > 0)
				{
					if(c != 'x' && c != 'X')
					{
						return BADFORMAT;
					}
					else
					{
						place++;
					}
				}
				i = *nFields = 0;

				while(i < MAXFMT) /* count "%"'s */
				{
					c = (*LINE)[place++];
					if(c == '%')
					{
						(*nFields)++;
					}
					else if(c == '"' || c == NL || c == CR || c == '\0')
					{
						break;
					}
					else if(isupper(c))
					{
						c = tolower((int) c);
					}
					(*fmt)[i++] = c;
				}
				(*fmt)[i] = '\0';
				fmtFound = 1;
			}
			if(verbose && (*LINE)[1] != '"')
			{
				(*LINE)[strlen(*LINE)-1] = '\0'; /* trim off '\n' */
				myerrorout(*LINE);
			}
			break;

		  default:
			done = 1;
		} /*switch ((*LINE)[0])*/
	} while(!done);

/*
	If *nfields is greater than dimensions on header, assume it specifies
	correct dimension.  If it is less, this is taken to indicate multi-line
	row or column, all with same format
*/
	if (dims[1] == 0 || tot != 0)
	{ /* empty data set is legal on read */
		return (NOERROR);
	}
	return NOERROR;
} /*readHeader()*/

