/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Regpred
#else
#define __SEG__ Regpred
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <math.h>

#define v(i,j) (*XTXINV)[(i)+(j)*(nvars+1)]
#define vn(i,j) (*XTXINV)[(i)+(j)*nvars]
#define xval(i,k) (*x)[(i)*npoints+(k)]

Symbolhandle    regpred(list)
Symbolhandle    list;

{
	Symbolhandle    result, symh;
	double        **x;
	double        **out[3];
	double          pred, mse, var;
	double          xik;
	long            dims[2];
	long            i, j, k;
	long            npoints;
	long            nterms = (long) NTERMS, nvars = (long) NVARS;
	char            *compname[3];


	compname[0] = "estimate";
	compname[1] = "SEest";
	compname[2] = "SEpred";

	symh = COMPVALUE(list, 0);
	if(!argOK(symh,REAL,0))
	{
		goto errorExit;
	}
	if(!isMatrix(symh,dims))
	{
		myerrorout("ERROR: argument to regpred must be REAL vector or matrix");
		goto errorExit;
	}

	if (!(PREVMODELTYPE & OLSREG))
	{
		myerrorout("ERROR: no current regression model");
		goto errorExit;
	}

	if(dims[1] > 1 || nvars == 1)
	{
		k = dims[1];
		npoints = dims[0];
	}
	else
	{ /* x is vector, NVARS > 1 */
		 k = dims[0];
		 npoints = 1;
	}
		
	if (k != nvars)
	{
		myerrorout("ERROR: number of predictor values does not match model");
		goto errorExit;
	}

	result = RSInstall(SCRATCH, 3L, compname, npoints);
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	for(i=0;i<3;i++)
	{
		out[i] = DATA(COMPVALUE(result,i));
	}

	x = DATA(symh);

	mse = (*SS)[nterms];
	if ((*DF)[nterms] > 0.)
	{
		mse /= (*DF)[nterms];
	}


	for(k=0;k<npoints;k++)
	{
		if (modeltermEqual(modelterm(MODEL,0),(modelPointer) NULLTERM))
		{						/* with intercept */
			pred = (*REGCOEF)[0];
			var = v(0,0);
			for (i = 0; i < nvars; i++)
			{
				xik = xval(i,k);
				pred += xik * (*REGCOEF)[i+1];
				var += v(i+1, i+1) * xik * xik;
				var += 2.0 * v(0, i+1) * xik;
				for (j = i + 1; j < nvars; j++)
				{
					var += 2.0 * v(i+1, j+1) * xik * xval(j,k);
				}
			}
		}
		else
		{						/* without intercept */
			pred = 0.0;
			var = 0.0;
			for (i = 0; i < nvars; i++)
			{
				xik = xval(i,k);
				pred += xik * (*REGCOEF)[i];
				var += vn(i, i) * xik * xik;
				for (j = i + 1; j < nvars; j++)
				{
					var += 2.0 * vn(i, j) * xik * xval(j,k);
				}
			}
		}
		(*out[0])[k] = pred;
		(*out[1])[k] = sqrt(mse*var);
		(*out[2])[k] = sqrt(mse*(1.0 + var));
	}
	
	return (result);

  errorExit:
	return (0);
}
