/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Rep
#else
#define __SEG__ Rep
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <math.h>

extern long            argOK();
extern void            badType();
extern void            badNargs();
extern long            isVector();
extern long            symbolSize();
extern char          **mygethandle();
extern Symbolhandle    Install();


Symbolhandle    rep(list)
Symbolhandle    list;
{
	long            type;
	long            i, j, k, tot, ctot, nrepterms, nelements;
	long            ii, incr, jlim;
	long            nargs = NARGS(list);
	char           *tmpstr, *tmpstr2;
	Symbolhandle    result = (Symbolhandle) 0, symhx, symhr;
	double        **repfac, **tmpreal;
	char           *usage =
		"usage: rep(x,nrep), x a vector, nrep a scalar or a vector";
	WHERE("rep");
	
	*OUTSTR = '\0';
	
	if(nargs != 2)
	{
		badNargs(FUNCNAME,2);
		goto errorExit;
	}
	
	symhx = COMPVALUE(list, 0);
	symhr = COMPVALUE(list, 1);
	if(!argOK(symhx,(long) 0,1) || !argOK(symhr,REAL,2))
	{
		goto errorExit;
	}

	type = TYPE(symhx);
	if (type != REAL && type != CHAR && type != LOGIC)
	{
		badType(FUNCNAME,-type,1);
		goto errorExit;
	}

	if(!isVector(symhr))
	{
		myerrorout("ERROR: argument 2 to rep must be scalar or vector");
		strcpy(OUTSTR,usage);
		goto errorExit;
	}
	
	nelements = symbolSize(symhx);
	nrepterms = symbolSize(symhr);

	if (nrepterms != 1 && nelements != nrepterms)
	{
		sprintf(OUTSTR,
				"ERROR: replication factor must match length of x or be length 1");
		goto errorExit;
	}

	repfac = DATA(symhr);
	tot = 0;
	for (i = 0; i < nrepterms; i++)
	{
		if ((*repfac)[i] < 0 || (*repfac)[i] != floor((*repfac)[i]))
		{
			sprintf(OUTSTR,
					"ERROR: replication factors must be non-negative integers");
			goto errorExit;
		}
		tot += (*repfac)[i];
	}
	if(tot == 0)
	{
		sprintf(OUTSTR,"ERROR: %srepetition factor must be positive",
				(nrepterms > 1) ? "at least 1 " : "");
		goto errorExit;
	}
	
	if(nrepterms == 1)
	{
		incr = 0;
		tot *= nelements;
	}
	else
	{
		incr = 1;
	}
	
	if (tot > MAXCOORD)
	{
		sprintf(OUTSTR,"ERROR:  output vector from rep too long");
		goto errorExit;
	}

	result = Install(SCRATCH,type);

	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	if (type == REAL || type == LOGIC)
	{
		TMPHANDLE = mygethandle((long) tot * sizeof(double));
		setDATA(result,(double **) TMPHANDLE);
		if (TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
	}
	else
	{ /* CHAR */
		ctot = 0;
		tmpstr = STRINGPTR(symhx);
		ii = 0;
		for (i = 0; i < nelements; i++)
		{
			ctot += (strlen(tmpstr) + 1) * (long) (*repfac)[ii];
			ii += incr;
		}
		TMPHANDLE = mygethandle((long) ctot);
		setSTRING(result,TMPHANDLE);
		if (TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
	}
	setNDIMS(result,1);
	setDIM(result,1,tot);

	switch ((int) type)
	{
	  case REAL:
	  case LOGIC:
		tmpreal = DATA(symhx);
		k = 0;
		if(nrepterms == 1)
		{
			jlim = (long) (*repfac) [0];
			for(j = 0;j < jlim;j++)
			{
				for(i = 0;i < nelements; i++)
				{
					DATAVALUE(result,k++) = (*tmpreal)[i];
				}
			}
		}
		else
		{
			for (i = 0; i < nelements; i++)
			{
				jlim = (long) (*repfac) [i];
				for(j=0;j < jlim; j++)
				{
					DATAVALUE(result,k++) = (*tmpreal)[i];
				}
			}
		}

		break;

	  case CHAR:
		tmpstr = STRINGPTR(result);
		if (nrepterms == 1)
		{
			jlim = (*repfac)[0];
			for (j = 0; j < jlim; j++)
			{
				tmpstr2 = STRINGPTR(symhx);
				for (i = 0; i < nelements; i++)
				{
					strcpy(tmpstr, tmpstr2);
					tmpstr += strlen(tmpstr2) + 1;
					tmpstr2 += strlen(tmpstr2) + 1;
				}
			}
		}
		else
		{
			tmpstr2 = STRINGPTR(symhx);
			for (i = 0; i < nelements; i++)
			{
				jlim = (*repfac)[i];
				for(j =0;j < jlim; j++)
				{
					strcpy(tmpstr, tmpstr2);
					tmpstr += strlen(tmpstr2) + 1;
				}
				tmpstr2 += strlen(tmpstr2) + 1;
			}
		}
	} /* switch */

	return (result);

  errorExit:
	putOUTSTR();

	Removesymbol(result);
	
	return (0);
	
}
