/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Save
#else
#define __SEG__ Save
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

#ifdef MACINTOSH
#include "macIface.h"
#endif

#include <ctype.h>

#define CHUNK         (5*sizeof(double)) /* units to ask for more room */
#define CR            13
#define LF            10
#define TIMELENGTH    30 /* length of buffer for time information */
#define HEADERLENGTH  24 /* strlen("////MacAnovaSaveFile////") */
extern void           Cutsymbol();
extern void           Removesymbol();
extern Symbolhandle   Install();
extern long           symbolSize();
extern long           isfilename();
extern long           isOption();
extern long           argOK();

extern char         **mygethandle();
extern char         **mygrowhandle();
extern void           mydisphandle();
extern void           clearGlobals();

Symbolhandle          restore();
long                  restoreLong();

static char           PadChar;

#define Binary  (ASCII < 0)

/*
  Test for a name match.  If a match is found, the data part of symh
  is copied to pointer dPointer and 1 is returned.  Otherwise 0 is returned.
*/
static long symbol2Double(double *dPointer,char *name,long size,
						  Symbolhandle symh)
{
	long                i;
	WHERE("symbol2Double");
	
	if(strcmp(NAME(symh),name) != 0)
	{
		return (0);
	}
	size = (size <= DIMVAL(symh,1)) ? size : DIMVAL(symh,1);
	for(i=0;i<size;i++)
	{
		*dPointer++ = DATAVALUE(symh,i);
	}
	return (1);
}

/*
  Test for a name match.  If a match is found, the data part of symh
  is copied to pointer lPointer and 1 is returned.  Otherwise 0 is returned.
*/
static long symbol2Long(long *lPointer,char *name,long size, Symbolhandle symh)
{
	long                i;
	WHERE("symbol2Long");
	
	if(strcmp(NAME(symh),name) != 0)
	{
		return (0);
	}
	size = (size <= DIMVAL(symh,1)) ? size : DIMVAL(symh,1);
	for(i=0;i<size;i++)
	{
		*lPointer++ = LONGDATAVALUE(symh,i);
	}
	return (1);
}

/*
  Test for a name match.  If a match is found, the current handle is 
  disposed of, it is reset to DATA(symh), DATA(symh) is cleared, and 1
  is returned.  Otherwise 0 is returned.
*/

static long symbol2Handle(double ***handle,char *name,Symbolhandle symh)
{
	if(strcmp(NAME(symh),name) != 0)
	{
		return (0);
	}
	
	if(*handle != (double **) 0)
	{
		mydisphandle((char **) *handle);
	}

	*handle = DATA(symh);
	setDATA(symh,(double **) 0);
	return (1);
}

/*
  read symbol from file.  It is immediately put in symbol table and returned
  as an installed symbol.  This means a structure component must be cut from
  the symbol table before being attached.
*/
static Symbolhandle restoresym(void)
{
	long            i, j, tot, spaceleft, ch;
	long            type, dimi, nclass, length, nchar = 0;
	long            ndims, mdims;
	long          **longdata;
	char          **hand = (char **) 0, name[NAMELENGTH+1];
	char            typename[NAMELENGTH+1];
	Symbolhandle    symh = (Symbolhandle) 0, component;
	double        **data;
	WHERE("restoresym");
	
	/* get symbol name */
	if (!restoreName(name))
	{
		goto errorExit;;
	}

	/* get symbol type */
	if(ASCII == SAVEBINARY1 || ASCII == SAVEASCII1)
	{ /* numeric type of non-ascii and old ascii format */
		if (!restoreLong(&type))
		{
			goto errorExit;
		}
		/* make sure codes are restore correctly even if future codes change */
		switch (type)
		{
		  case CHAR_V31:
			type = CHAR;
			break;
		  case REAL_V31:
			type = REAL;
			break;
		  case LOGIC_V31:
			type = LOGIC;
			break;
		  case STRUC_V31:
			type = STRUC;
			break;
		  case MACRO_V31:
			type = MACRO;
			break;
		  case PLOTINFO_V31:
			type = PLOTINFO;
			break;
		  case LONG_V31:
			type = LONG;
			break;
		  default:
			;
		} /*switch (type)*/
	}
	else if(!restoreName(typename))
	{/* character specified type for new ascii formats and new binary format */
		goto errorExit;
	}
	else if(strcmp(typename,"REAL") == 0)
	{
		type = REAL;
	}
	else if(strcmp(typename,"LOGIC") == 0)
	{
		type = LOGIC;
	}
	else if(strcmp(typename,"LONG") == 0)
	{
		type = LONG;
	}
	else if(strcmp(typename,"CHAR") == 0)
	{
		type = CHAR;
	}
	else if(strcmp(typename,"STRUC") == 0)
	{
		type = STRUC;
	}
	else if(strcmp(typename,"MACRO") == 0)
	{
		type = MACRO;
	}
	else if(strcmp(typename,"GRAPH") == 0)
	{
		type = PLOTINFO;
	}
	else
	{ /* not recognized */
		goto errorExit;;
	}

	if(type == MACRO1)
	{
		type = MACRO;
	}

	symh = Install(name,type); /* immediately add to symbol table */

	if (symh == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	/* get NCLASS */
	if (!restoreLong(&nclass))
	{
		goto errorExit;
	}
	setNCLASS(symh,nclass);

	/* get NDIMS */
	if (!restoreLong(&ndims))
	{
		goto errorExit;
	}
	setNDIMS(symh,ndims);

	/* get dimensions */
	mdims = (ASCII >= SAVEASCII3 || ASCII <= SAVEBINARY3) ? ndims : MAXDIMS;
	for (i = 1; i <= mdims; i++)
	{
		if (!restoreLong(&dimi))
		{
			goto errorExit;
		}
		if(i <= ndims)
		{
			setDIM(symh,i,dimi);
		}
	}

	tot = symbolSize(symh);

	switch (type)
	{
	  case STRUC:
		TMPHANDLE = mygethandle((long) tot * sizeof(Symbolhandle));
		setCOMP(symh,(Symbolhandle**) TMPHANDLE);
		if (COMP(symh) == (Symbolhandle**) 0)
		{
			goto errorExit;
		}
		for (i = 0; i < tot; i++)
		{
			COMPVALUE(symh, i) = (Symbolhandle) 0;
		}
		for (i = 0; i < tot; i++)
		{
			component = restoresym();
			Cutsymbol(component);
			COMPVALUE(symh, i) = component;
			if (component == (Symbolhandle) 0)
			{
				goto errorExit;
			}
		}
		break;

	  case MACRO:
	  case CHAR:
		if(Binary || ASCII == SAVEASCII1)
		{
			hand = mygethandle((long) CHUNK);
			setSTRING(symh,hand);
			spaceleft = CHUNK;
			if (hand == (char **) 0)
			{
				goto errorExit;
			}
			j = 0;
			i = 0;
			while (i < tot && (ch = fgetc(RESTOREFILE)) != EOF)
			{
				if (spaceleft < 2)
				{
					hand = mygrowhandle((char **) hand, (long) j + CHUNK);
					setSTRING(symh,hand);
					spaceleft = CHUNK;
					if (hand == (char **) 0)
					{
						goto errorExit;
					}
				}
				(*hand)[j++] = (char) ch;
				spaceleft--;
				if (ch == '\0')
				{
					i++;
				}
			} /*while (i < tot && (ch = fgetc(RESTOREFILE)) != EOF)*/
		} /*if(Binary || ASCII == SAVEASCII1)*/
		else
		{ /* ASCII > SAVEASCII1 */
			for(i=0;i<tot;i++)
			{
				if(!restoreLong(&length))
				{
					goto errorExit;
				}
				if(i == 0)
				{
					setSTRING(symh,mygethandle(length + 1));
				}
				else
				{
					setSTRING(symh,mygrowhandle(hand, nchar + length + 1));
				}
				hand = STRING(symh);
				if(hand == (char **) 0)
				{
					goto errorExit;
				}
				if(!restoreNString(length,*hand + nchar))
				{
					goto errorExit;
				}
				nchar += length + 1;
			}
		}
		
		/* if necessary, shrink to exact size */
		if((Binary || ASCII == SAVEASCII1) && spaceleft > 0)
		{
			setSTRING(symh,(char **) 0);
			hand = mygrowhandle(hand, j);
			setSTRING(symh,hand);
			if(hand == (char **) 0)
			{
				goto errorExit;
			}
		}
		break;
		
	  case PLOTINFO:
		hand = restoreGraph();   /* restoreGraph() is in plotutil.c */
		if(hand == (char **) 0)
		{
			goto errorExit;
		}
		setGRAPH(symh,(struct whole_graph **) hand);
		break;

	  case REAL:
	  case LOGIC:
		
		if(tot > 0)
		{ /* allow for possibility of null symbol on save file */
			data = (double **) mygethandle(tot * sizeof(double));
			setDATA(symh,data);
			if (data == (double **) 0)
			{
				goto errorExit;
			}
			hand = (char **) data;
			if (!Binary)
			{/* ascii file */
				for (i = 0; i < tot; i++)
				{
					if(!restoreDouble(*data + i))
					{
						goto errorExit;
					}
				}
			}
			else 
			{                                       /* binary file */
				if(fread((void *) *hand, sizeof(double),
						 (size_t) tot, RESTOREFILE) != tot)
				{
					goto errorExit;
				}
#undef UNDEFINED__
#ifdef UNDEFINED__
				for (i = 0; i < sizeof(double) * tot; i++)
				{
					if ((ch = fgetc(RESTOREFILE)) == EOF)
					{
						goto errorExit;
					}
					(*hand)[i] = (char) ch;
				}
#endif /*UNDEFINED__*/
			}
		}
		
		break;
		
	  case LONG:
		longdata = (long **) mygethandle((long) tot * sizeof(long));
		setLONGDATA(symh,longdata);
		if (longdata == (long **) 0)
		{
			goto errorExit;
		}
		hand = (char **) longdata;
		if (!Binary)
		{ /* ascii file */
			for (i = 0; i < tot; i++)
			{
				if(!restoreLong(*longdata + i))
				{
					goto errorExit;
				}
			}
		}
		else 
		{ /* binary file */
			if(fread((void *) *hand, sizeof(long), (size_t) tot, RESTOREFILE) != tot)
			{
				goto errorExit;
			}
		}
		break;
	} /* switch(type) */
	

	return (symh);

  errorExit:

	Removesymbol(symh);
	
	return (0);
	
}

static void restoreGlobals(Symbolhandle globals)

{
	long                icomp, i, j, n, k, nvars;
	Symbolhandle        symh, symh1;
	double              dTemp[MAXVARS];
	double            **model = (double **) 0;
	double              errorterms[MAXERRTRMS + 1];
	long                maxvars = MAXVARS;
	WHERE("restoreGlobals");
	
	nvars = -1;
	for(icomp=0;icomp<NCOMPS(globals);icomp++)
	{
		symh = COMPVALUE(globals,icomp);

		if(symbol2Double(&NVARS,"NVARS",1,symh))
		{
			nvars = MODELINFO->nvars = NVARS;
		}
		else if(symbol2Long(&maxvars,"MAXVARS",1,symh))
		{
			if(maxvars > MAXVARS)
			{
				sprintf(OUTSTR,
						"ERROR: Model information saved by version allowing %ld variables in models",
						maxvars);
				putOUTSTR();
				sprintf(OUTSTR,
						"       This version allows only %ld.  Model information not restored",
						(long) MAXVARS);
				putOUTSTR();
				NVARS = MODELINFO->nvars = -1;
				break;
			}
		}
		else if(symbol2Double(&NFACTORS,"NFACTORS",1,symh))
		{
			MODELINFO->nfactors = NFACTORS;
		}
		else if(symbol2Double(&NDATA,"NDATA",1,symh))
			;
		else if(symbol2Double(&NOTMISSING,"NOTMISSING",1,symh))
			;
		else if(symbol2Double(&NTERMS,"NTERMS",1,symh))
		{
			MODELINFO->nterms = NTERMS;
		}
		else if(symbol2Double(&NREGX,"NREGX",1,symh))
			;
		else if(symbol2Double(&NY,"NY",1,symh))
			;
		else if(symbol2Double(&NERRORTERMS,"NERRORTERMS",1,symh))
		{
			MODELINFO->nerrorterms = NERRORTERMS;
		}
		else if(symbol2Double(dTemp,"PREVMODELTYP",1,symh))
		{
			PREVMODELTYPE = (long) dTemp[0];
		}
		else if(symbol2Double(dTemp,"INCREMENTAL",1,symh))
		{
			INCREMENTAL = (long) dTemp[0];
		}
		else if(symbol2Double(dTemp,"USEGLMOFFSET",1,symh))
		{
			USEGLMOFFSET = (long) dTemp[0];
		}
		else if(symbol2Double(dTemp,"NCLASSES",MAXVARS,symh))
		{
			for(j=0;j<maxvars;j++)
			{
				NCLASSES[j] = (j < nvars) ? (long) dTemp[j] : 0;
			}
		}
		else if(symbol2Double(errorterms,"ERRORTERMS",MAXERRTRMS+1,symh))
		{
			for(j=0;j<=MAXERRTRMS;j++)
			{
				zeroTerm(MODELINFO->errorterms[j]);
				MODELINFO->errorterms[j][0] = (long) errorterms[j];
			}
		}
		else if(symbol2Long((long *) MODELINFO->errorterms,"ERRORTERMS1",
							(WORDSPERTERM)*(MAXERRTRMS+1),symh))
			;
		else if(symbol2Handle((double ***) &DEPVNAME,"DEPVNAME",symh))
			;
		else if(strcmp(NAME(symh),"VARNAMES") == 0)
		{
			n = (DIMVAL(symh,1) <= maxvars+1) ? DIMVAL(symh,1) : maxvars+1;
			k = 0;
			for(j=0;j<=maxvars;j++)
			{
				if(j < n)
				{
					strcpy(VARNAMES[j],STRINGPTR(symh) + k);
					k += strlen(VARNAMES[j]) + 1;
				}
				else
				{
					VARNAMES[j][0] = '\0';
				}
			}
		}
		else if(symbol2Handle(&model,"MODEL",symh))
		{
			long      length = myhandlelength((char **) model)/sizeof(double);
			
			TMPHANDLE = mygethandle(length * sizeof(modelType)) ;
			if(TMPHANDLE == (char **) 0)
			{
				NVARS = MODELINFO->nvars = -1;
				myerrorout("       Model information not restored");
				break;
			}
			MODEL = (modelHandle) TMPHANDLE;
			for(i=0;i<length;i++)
			{
				zeroTerm((*MODEL)[i]);
				(*MODEL)[i][0] = (termWord) (*model)[i];
			}
		}
		else if(symbol2Handle((double ***) &MODEL,"MODEL1",symh))
		{
			;
		}
		/*
		  Under current version, Y should always be the same as 
		  DATA(MODELVARS[0]) so we no longer restore it;  however, we
		  do recognize it
		*/
		else if(strcmp(NAME(symh),"Y") == 0)
			;
		else if(symbol2Handle(&RESIDUALS,"RESIDUALS",symh))
			;
		else if(symbol2Handle(&WTDRESIDUALS,"WTDRESIDUALS",symh))
			;
		else if(symbol2Handle(&GLMOFFSET,"GLMOFFSET",symh))
			;
		else if(symbol2Handle(&HII,"HII",symh))
			;
		else if(symbol2Handle(&SS,"SS",symh))
			;
		else if(symbol2Handle(&DF,"DF",symh))
			;
		else if(symbol2Handle(&MISSWTS,"MISSWTS",symh))
			;
		else if(symbol2Handle(&CASEWTS,"CASEWTS",symh))
			;
		else if(symbol2Handle(&ITERWTS,"ITERWTS",symh))
			;
		else if(symbol2Handle(&ALLWTS,"ALLWTS",symh))
			;
		else if(symbol2Handle(&LOGITN,"LOGITN",symh))
			;
		else if(strcmp(NAME(symh),"STRMODEL") == 0)
		{
			if(STRMODEL != (char **) 0)
			{
				mydisphandle(STRMODEL);
			}
			STRMODEL = STRING(symh);
			setSTRING(symh,(char **) 0);
		}
		else if(strcmp(NAME(symh),"TERMNAMES") == 0)
		{
			if(TERMNAMES != (char **) 0)
			{
				mydisphandle(TERMNAMES);
			}
			TERMNAMES = STRING(symh);
			setSTRING(symh,(char **) 0);
		}
		else if(symbol2Handle(&REGX,"REGX",symh))
			;
		else if(symbol2Handle(&REGX2,"REGX2",symh))
			;
		else if(symbol2Handle(&XTXINV,"XTXINV",symh))
			;
		else if(symbol2Handle(&REGCOEF,"REGCOEF",symh))
			;
		else if(symbol2Handle(&REGSS,"REGSS",symh))
			;

		/*
		  Under current version, X[i] should always be the same as 
		  DATA(MODELVARS[i+1]) so we no longer restore it.  We do
		  recognize it, however
		*/
		else if(strcmp(NAME(symh),"X") == 0)
			;
		else if(strcmp(NAME(symh),"MODELVARS") == 0)
		{
			n = NCOMPS(symh);
			for(j=0;j<=maxvars;j++)
			{
				symh1 = (j < n) ? COMPVALUE(symh,j) : (Symbolhandle) 0;
				MODELVARS[j] = symh1;
				if(j < n)
				{
					COMPVALUE(symh,j) = (Symbolhandle) 0;
					if(j==0)
					{
						Y = DATA(MODELVARS[0]);
					}
					else
					{
						X[j-1] = DATA(MODELVARS[j]);
					}
				} /*if(j < n)*/
			} /*for(j=0;j<=maxvars;j++)*/
		}
		else
		{
			myerrorout("WARNING: unrecognized global variable on file. Globals not restored");
			NVARS = MODELINFO->nvars = -1;
			break;
		}
	} /*for(icomp=0;icomp<NCOMPS(globals);icomp++)*/
}

#ifdef PLATFORM
static long checkMachine(Symbolhandle globals)
{
	long                icomp;
	Symbolhandle        symh;
	long                maxdims, maxvars, wordsperterm, varsperword;
	long                limits[4];
	long                retvalue = 0;
	char              **platform = (char **) 0;
	WHERE("checkMachine");
	
	for(icomp=0;icomp<NCOMPS(globals);icomp++)
	{
		symh = COMPVALUE(globals,icomp);

		if(symbol2Long(limits,"LIMITS",4,symh))
		{
			maxdims = limits[0];
			maxvars = limits[1];
			wordsperterm = limits[2];
			varsperword = limits[3];
			if(maxdims != MAXDIMS)
			{
				sprintf(OUTSTR,
						"ERROR: file was saved by incompatible version of MacAnova");
				retvalue = -1;
				goto errorExit;
			}
			else if(wordsperterm != WORDSPERTERM || 
					varsperword != VARSPERWORD)
			{
				retvalue = 1;
			}
		}
		else if(symbol2Handle((double ***) &platform,"PLATFORM",symh))
		{
#ifdef MPW
			if(strcmp(*platform,PLATFORM) != 0 &&
			   strcmp(*platform,"Macintosh") != 0)
#else /*MPW*/
			if(strcmp(*platform,PLATFORM) != 0)
#endif /*MPW*/
			{
				if(Binary)
				{
					sprintf(OUTSTR,
						"ERROR: Binary file saved by version of Macanova compiled for %s",
							*platform);
					goto errorExit;
				}
				sprintf(OUTSTR,
					"WARNING: File saved by version of Macanova compiled for %s",
						*platform);
				putOUTSTR();
			} /*if(strcmp(*platform,PLATFORM) != 0)*/
		}
	} /*for(icomp=0;icomp<NCOMPS(globals);icomp++)*/
	mydisphandle(platform);
	return (retvalue);
	
  errorExit:
	putOUTSTR();
	myerrorout("       Restore aborted");
	
	mydisphandle(platform);
	return (-1);
} /*checkMachine()*/
#endif /*PLATFORM*/
/*

/* Modified from listObject in list.c */
static void listRestoredSymbol(Symbolhandle symh)
{
	char            *name;
	int              type;
	long             i, nclass;
	WHERE("listRestoredSymbol");
	
	name = NAME(symh);
	type = TYPE(symh);
	
	sprintf(OUTSTR,"Restoring %-15s %-6s",name,typeName(type));
	myprint(OUTSTR);

	if (type == REAL || type == LOGIC || type == CHAR || type == STRUC)
	{
		for (i = 1; i <= NDIMS(symh); i++)
		{
			sprintf(OUTSTR, " %-5ld", DIMVAL(symh,i));
			myprint(OUTSTR);
		}
		if ((nclass = NCLASS(symh)) > 0)
		{
			sprintf(OUTSTR, " FACTOR with %ld levels",nclass);
			myprint(OUTSTR);
		}
	}
	myeol();
	*OUTSTR = '\0';
} /*listRestoredSymbol()*/

/*
   Usage: restore(fileName [,delete:T] [,list:T])
	delete:T suppresses deletion of symbols that are not being replaced
	list:T lists symbols as they are restored.  Primarily for debugging
	purposes
*/

Symbolhandle    restore(list)
Symbolhandle    list;
{
	Symbolhandle    arg1, argKey, symh, next, symh1;
	char            instuff[HEADERLENGTH+1];
	long            i, nOptions;
	long            ch;
	long            deleteCurrent = 1, restoreGlobs = 1, listRestored = 0;
	long            nargs = NARGS(list);
	char           *keyword;
	char            whenSaved[TIMELENGTH];
	char           *fname, name[NAMELENGTH+1];
#ifdef MACINTOSH
	OSType          types[3];
#endif
	WHERE("restore");
	
	*OUTSTR = '\0';
	RESTOREFILE = (FILE *) 0;
	whenSaved[0] = '\0';
	
	arg1 = COMPVALUE(list, 0);

	if(nargs > 3)
	{
		badNargs(FUNCNAME,-3);
		goto errorExit;
	}
	
	if (!argOK(arg1, 0, 0))
	{
		goto errorExit;
	}
	if (TYPE(arg1) != CHAR || !isScalar(arg1))
	{
		sprintf(OUTSTR,
				"ERROR: file name must be quoted string or CHARACTER scalar");
		goto errorExit;
	}		
	for (i = 1; i < nargs; i++)
	{ /* must be either delete:(T|F) or simply T or F */
		argKey = COMPVALUE(list,i);
		if (!argOK(argKey, 0, i+1))
		{
			goto errorExit;
		}
		if ((keyword = isKeyword(argKey)))
		{
			if (strncmp(keyword, "del", 3) != 0 &&
				strncmp(keyword, "lis", 3) != 0 )
			{
				badKeyword(FUNCNAME, keyword);
				goto errorExit;
			}
		}
		if (TYPE(argKey) != LOGIC || !isScalar(argKey))
		{
			notTorF((keyword) ? keyword : (char *) 0);
			goto errorExit;
		}
		if (!keyword || *keyword == 'd')
		{
			deleteCurrent = (DATAVALUE(argKey,0) != 0);
		}
		else
		{
			listRestored = (DATAVALUE(argKey,0) != 0);
		}
	} /*for (i = 1; i < nargs; i++)*/
		
	fname = STRINGPTR(arg1);

#ifdef MACINTOSH
	if(RestoreVolume)
	{
		SetVol(0L, RestoreVolume);
	}

	if (fname[0] == '\0')
	{
		types[0] = KEEPFILETYPE;
		types[1] = ASCIIKEEPFILETYPE;
		types[2] = 'TEXT'; /* allow for type having been changed */
		fname = macOpen("\pSpecify the restorefile", (Str255) 0, READIT,
			3, types);
		if(fname == (char *) 0)
		{
			return (0);
		}
		RestoreVolume = Reply.vRefNum;
	} /*if (fname[0] == '\0')*/
#endif /*MACINTOSH*/

	if((fname = expandFilename(fname)) == (char *) 0)
	{
		goto errorExit;
	}
			   
	if(!isfilename(fname))
	{
		goto errorExit;
	}

#ifdef MACINTOSH
	/* save the restore file name as save file name if none is defined */
	if(KEEPFILENAME == (char **) 0 &&
	   (KEEPFILENAME = mygethandle(strlen(fname)+1)) != (char **) 0)
	{
		strcpy(*KEEPFILENAME, fname);
	}
#endif /*MACINTOSH*/

#ifdef MSDOS
	RESTOREFILE = fopen(fname, "rb");
#else /*MSDOS*/
	RESTOREFILE = fopen(fname, "r");
#endif /*MSDOS*/

	if (RESTOREFILE == (FILE *) 0 )
	{
		sprintf(OUTSTR,"ERROR: cannot open file %s for restore", fname);
		goto errorExit;
	}

	PadChar = ' ';
	if(!restoreNString(HEADERLENGTH, instuff))
	{
		sprintf(OUTSTR,
				"ERROR: Does not appear to be save file;header = %s",instuff,0,0,0);
		goto errorExit;
	}

	if (strcmp(instuff, BINARYHEADER1) == 0)
	{
		ASCII = SAVEBINARY1;
	}
	else if (strcmp(instuff, BINARYHEADER4) == 0)
	{
		ASCII = SAVEBINARY4;
	}
	else if (strcmp(instuff, ASCIIHEADER1) == 0)
	{
		ASCII = SAVEASCII1;
	}
	else if (strcmp(instuff, ASCIIHEADER2) == 0)
	{
		ASCII = SAVEASCII2;
	}
	else if (strcmp(instuff, ASCIIHEADER3) == 0)
	{
		ASCII = SAVEASCII3;
	}
	else if (strcmp(instuff, ASCIIHEADER4) == 0)
	{
		ASCII = SAVEASCII4;
	}
	else if (strncmp(instuff, ASCIIHEADER4, HEADERLENGTH-3) == 0 ||
			 strncmp(instuff, BINARYHEADER4, HEADERLENGTH-4) == 0)
	{
		sprintf(OUTSTR,
				"ERROR: MacAnova save file cannot be restored by this version of MacAnova");
		goto errorExit;
	}
	else
	{
		sprintf(OUTSTR,"ERROR: file is not a MacAnova save file");
		goto errorExit;
	}

	if(!Binary && ASCII != SAVEASCII1)
	{
		/* close file and reopen in text mode */
		fclose(RESTOREFILE);
		RESTOREFILE = fopen(fname, "rt");
		if(!restoreNString(HEADERLENGTH,instuff))
		{
			return(0);
		}
		PadChar = '\n';
	}

	myprint("Restoring workspace from file ");
	myerrorout(fname);
	if(ASCII == SAVEASCII1)
	{
		sprintf(OUTSTR,
				"WARNING: MacAnova 2.4 or earlier asciisave file ; may include old style macros");
		putOUTSTR();
	}
	
	/* delete current workspace */
	if(deleteCurrent)
	{               
		for(i=0;i<nargs;i++)
		{/* first remove items in the argument list */
			Removesymbol(COMPVALUE(list,i));
			COMPVALUE(list,i) = (Symbolhandle) 0;
		}
	
		symh = Firstsymbol(0);
		while (symh != (Symbolhandle) 0)
		{ /* clear regular symbol table*/
			next = Nextsymbol(symh,0);
			Removesymbol(symh); /* will not remove builtins */
			symh = next;
		}
		symh = Firstsymbol(1);
		while (symh != (Symbolhandle) 0)
		{ /* clear scratch, UNDEF, and keyword symbol table */
			if(symh == list)
			{/* do not remove list; that is done in mainpars.y */
				symh = Nextsymbol(symh,1);
			}
			else
			{
				next = Nextsymbol(symh,1);
				Removesymbol(symh); /* will not remove builtins */
				symh = next;
			}
		}
		clearGlobals();
	} /*if(deleteCurrent)*/
/* Make sure correct info is in VERSION */
	if(VERSION_ID != (char **) 0)
	{
		Remove("VERSION");
		symh = Install("VERSION", CHAR);
		TMPHANDLE = myduphandle(VERSION_ID);
		if(symh == (Symbolhandle) 0 || TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
		setSTRING(symh, TMPHANDLE);
		setNDIMS(symh,1);
		setDIM(symh,1,1);
	}  /*if(deleteCurrent){}else{}*/

	while(1)
	{
		while ((ch = fgetc(RESTOREFILE)) == ' ' || ch == PadChar)
		{
			;
		}
		if (ch == EOF)
		{
			break;
		}
		ungetc((char) ch, RESTOREFILE);
		symh = restoresym(); /* is already in Symbol table */
		if (symh == (Symbolhandle) 0)
		{
			myerrorout("ERROR: incomplete restore");
			fclose(RESTOREFILE);
			return (0);
		}

		strcpy(name,NAME(symh));
		if(strcmp(name,"T") == 0 || strcmp(name,"F") == 0 ||
		   strcmp(name,"HELP") == 0 || strcmp(name,"help") == 0 ||
		   strcmp(name,"VERSION") == 0)
		{ /* all but the last might be saved by MacAnova2.4 */
			Removesymbol(symh);
		}
		else if(strcmp(name,OPTIONS) == 0)
		{
			nOptions = NCOMPS(symh);                        
			for(i=0;i<nOptions;i++)
			{
				keyword = NAME(COMPVALUE(symh,i));
				if(!isOption(keyword))
				{
					break;
				}
			}
			if(i < nOptions)
			{
				sprintf(OUTSTR,
						"WARNING: unrecognized option name %s found during restore",
						keyword);
				putOUTSTR();
			}
			
			if (listRestored)
			{
				myerrorout("Restoring option values");
			}
			
			setOpt(symh, 1);
			Removesymbol(symh);
		}
#ifdef PLATFORM
		else if(strcmp(name,MACHINE) == 0)
		{
			restoreGlobs = checkMachine(symh);
			Removesymbol(symh);
			if(restoreGlobs < 0)
			{
				goto errorExit;
			}
			restoreGlobs = !restoreGlobs;
		}
#endif /*PLATFORM*/
		else if(strcmp(name, GLOBALS) == 0)
		{
			if(restoreGlobs && !deleteCurrent)
			{
				clearGlobals();
			}
			
			if(restoreGlobs)
			{
				restoreGlobals(symh);
				if (listRestored)
				{
					myerrorout("Restoring private glm values");
				}
				
				if(NVARS < 0)
				{
					clearGlobals();
				}
			}
			else if(PRINTWARNINGS)
			{
				myerrorout("WARNING: incompatible internal representation of model on restore file");
				myerrorout("         Model information not restored");
			}
			Removesymbol(symh);
		}
		else if(strcmp(name, SAVETIME) == 0)
		{
			strncpy(whenSaved,STRINGPTR(symh),TIMELENGTH-1);
			whenSaved[TIMELENGTH-1] = '\0';
			Removesymbol(symh);
		}
		else
		{ /* ordinary symbol */
			Cutsymbol(symh); /* make sure it won't be found by Lookup */
			if(!deleteCurrent)
			{ /* delete symbol if it exists */
				symh1 = Lookup(name);
				if(symh1 != (Symbolhandle) 0 && TYPE(symh1) == BLTIN)
				{
					/* This should never happen ; this is just insurance */
					myerrorout("ERROR: attempt to restore object with same name as built in function");
					break;
				}
				Removesymbol(symh1);
			}
			
			if (listRestored)
			{
				listRestoredSymbol(symh);
			}
			Addsymbol(symh); /* put back in symbol table */
		}
	} /* while(1) */
	fclose(RESTOREFILE);
	if(*whenSaved)
	{
		sprintf(OUTSTR, "Workspace saved %s", whenSaved);
		putOUTSTR();
	}
	
	return (NULLSYMBOL);

  errorExit:
	putOUTSTR();
	
	return (0);
	
}


/*
  scans for integer followed by a PadChar; if the PadChar is not present, it is
  an error (return 0)
*/

long restoreLong(longValue)
long             *longValue;
{
	long          c;
	long          sign = 1, value = 0;
	
/*
	while((c = fgetc(RESTOREFILE)) == ' ' || c == PadChar)
	{
		;
	}
*/
	c = fgetc(RESTOREFILE);
	
	if(c == EOF || (!isdigit(c) && c != '-'))
	{
		return (0);
	}
	if(c == '-')
	{
		sign = -1;
	}
	else
	{
		value = c - '0';
	}
	while(isdigit((char) (c = fgetc(RESTOREFILE))))
	{
		value = 10*value + (c - '0');
	}
	*longValue = sign*value;
	return (c == PadChar);
}

long restoreName(name)
char           *name;
{
	int         i, c;

/*
	while((c = fgetc(RESTOREFILE)) == ' ' || c == PadChar)
	{
		;
	}
	if(c == EOF)
	{
		return (0);
	}
	*name++ = c;
*/
	for(i=0;i<NAMELENGTH;i++)
	{
		if((c = fgetc(RESTOREFILE)) == EOF)
		{
			return (0);
		}

		if(c == PadChar)
		{
			break;
		}
		*name++ = (char) c;
	} /* for i */
	*name++ = '\0';
	return (i < NAMELENGTH || fgetc(RESTOREFILE) == PadChar);
}

/*
  restoreString() should be used only when one is sure the string to be
  read will fit in string[].  Otherwise, use restoreNString()
*/
long restoreString(string)
char           *string;
{
	long        length;
	
	*string = '\0';
	if(!restoreLong(&length) || !restoreNString(length,string))
	{
		return (0);
	}
	
	return (1);
		
}

long restoreNString(length,string)
long            length;
char           *string;
{
	int            ch, nc, s;
	
	while(length-- > 0)
	{
		if((ch = fgetc(RESTOREFILE)) == EOF)
		{
			return (0);
		}
		if(ASCII > 3 && ch == '\\')
		{
			if((ch = fgetc(RESTOREFILE)) == EOF)
			{
				return (0);
			}
			if(ch == 't')
			{
				ch = '\t';
			}
			else if(isdigit(ch) && ch <= '3')
			{
				s = ch - '0';
				for(nc = 1; nc < 3; nc++)
				{
					if((ch = fgetc(RESTOREFILE)) == EOF || 
					   !isdigit(ch) || ch > '7')
					{
						goto errorExit;
					}
					s = 8*s + (ch - '0');
				}
				ch = s;
			}
		} /*if(ch == '\\' && ASCII > 3)*/
		*string++ = ch;
	} /*while(length-- > 0)*/
	*string = '\0';

	return (fgetc(RESTOREFILE) == PadChar);

  errorExit:
	return (0);
}

long restoreDouble(value)
double         *value;
{
	int         c, length = 0;
	char       *end;
	char        buffer[100];
	WHERE("restoreDouble");
	
	while((c = fgetc(RESTOREFILE)) == ' ' || c == PadChar)
	{
		;
	}
	if(c == EOF)
	{
		return (0);
	}
	buffer[0] = c;

	length = 1;
	while((c = fgetc(RESTOREFILE)) != EOF && c != PadChar)
	{
		buffer[length++] = c;
	}

	if(c == EOF)
	{
		return (0);
	}
	buffer[length] = '\0';
	*value = mystrtod(buffer, &end);
	return (end - buffer == length);
}

		
		
