/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Run
#else
#define __SEG__ Run
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#define FUZZ 1e-15

Symbolhandle    run(arglist)
Symbolhandle    arglist;
{
	long            i, tot, length;
	long            nargs;
	Symbolhandle    result = (Symbolhandle) 0, symh[3];
	double         *tmpreal;
	double          first, last, by = 1.0, fi, last1, val, range;
	WHERE("run");
	
	/* generate a run of numbers */

	*OUTSTR = '\0';

	if ((nargs = NARGS(arglist)) > 3)
	{
		badNargs(FUNCNAME,-3);
		goto errorExit;
	}

	for(i=0;i<nargs;i++)
	{
		symh[i] = COMPVALUE(arglist,i);
		if(!argOK(symh[i], 0, (nargs > 1) ? i+1 : 0))
		{
			goto errorExit;
		}
		if (nargs > 1 && (!isScalar(symh[i]) || TYPE(symh[i]) != REAL ))
		{
			sprintf(OUTSTR,"ERROR: argument %ld to %s not REAL scalar",
					i+1, FUNCNAME);
			goto errorExit;
		}
	} /*for(i=0;i<nargs;i++)*/

	if(nargs == 1)
	{
		if(!isVector(symh[0]) || TYPE(symh[0]) != REAL ||
		   !isScalar(symh[0]) && (length = symbolSize(symh[0])) != 3)
		{
			sprintf(OUTSTR,
					"ERROR: arg must be REAL scalar or vector of length 3 in %s(arg)",
					FUNCNAME);
			goto errorExit;
		}
		
		if(isScalar(symh[0]))
		{
			first = 1.0;
			last = DATAVALUE(symh[0],0);
			by = 1.0;
			
			if(!isMissing(last) && (last != floor(last) || last <= 0))
			{
				sprintf(OUTSTR,"ERROR: N must be positive integer in %s(N)",
						FUNCNAME);
				goto errorExit;
			}
		}
		else
		{
			first = DATAVALUE(symh[0],0);
			last = DATAVALUE(symh[0],1);
			by = DATAVALUE(symh[0],2);
		}
	} /*if(nargs == 1)*/
	else
	{
		first = DATAVALUE(symh[0],0);
		last = DATAVALUE(symh[1],0);
		if (nargs == 3)
		{
			by = DATAVALUE(symh[2],0);
		}
		else if (last >= first)
		{
			by = 1.;
		}
		else
		{
			by = -1.;
		}
	} /*if(nargs == 1){}else{}*/

	if (isMissing(first) || isMissing(last) || isMissing(by))
	{
		sprintf(OUTSTR,
				"ERROR: MISSING values illegal in arguments to %s", FUNCNAME);
		goto errorExit;
	}
	
	if (last == first)
	{
		by = 0.0;
		range = 1;
	}
	else
	{
		range = last - first;
	}
	if (range * by < 0.0 || last != first && by == 0.0)
	{
		sprintf(OUTSTR,"ERROR: incr in %s(%s) is 0 or wrong sign", 
				FUNCNAME, (nargs> 1) ? "start,end,incr" : "cat(start,end,incr)");
		goto errorExit;
	}

	tot = (by != 0.0) ? (long) floor(range / by) + 1 : 1;
/* 
   The following is intended to insure that if last is essentially and exact
   multiple of by away from start, then the last element of the result
   will be close to last, even if slightly outside the interval from
   first to last
*/

	if (by != 0.0)
	{		
		for (last1 = first + tot * by;
			 by > 0.0 && last1 <= last || by < 0.0 && last1 >= last ||
			 fabs((last1 - last)/range) <= FUZZ;
			 last1 = first + (++tot) * by)
		{
			;
		}
	}

	result = RInstall(SCRATCH,tot);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	tmpreal = DATAPTR(result);

	for (i = 0, fi = 0.0; i < tot;i++,fi++)
	{
		val = first + fi * by;
		tmpreal[i] = (fabs(val/range) < FUZZ) ? 0.0 : val;
	}

	if (tot > 1 && fabs((val - last)/range) <= FUZZ)
	{
		tmpreal[tot-1] = last;
	}
	
	return (result);

  errorExit:
	putOUTSTR();
	
	return (0);
	
}
