/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Save
#else
#define __SEG__ Save
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

#ifdef MACINTOSH
#include "macIface.h"
#ifdef MPW
#include <FCntl.h>
#else
#include <unix.h>
#include <fcntl.h>
#endif /*MPW*/
#endif /* MACINTOSH*/

#ifdef BCPP
#include <fcntl.h>
#include <sys\stat.h>
#include <io.h>
#endif /*BCPP*/

#if defined(UNIX) || defined(DJGPP)
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/file.h>
#endif /*defined(UNIX) || defined(DJGPP)*/

#define LONGHANDLE    1
#define DOUBLEHANDLE  2
#define CHARHANDLE    3

#define MAXCHUNK  16000

#define Binary  (ASCII < 0)

extern long           symbolSize();
extern long           isfilename();
extern long           argOK();
extern Symbolhandle   getOpt();
extern char         **mygethandle();
extern void           mydisphandle();
extern long           myhandlesize();
extern char          *isKeyword();
extern Symbolhandle   Makesymbol();
extern Symbolhandle   Makestruc();
extern Symbolhandle   Install();
extern void           Removesymbol();
extern long           Copy();

/* Globals ASCII, KEEPFILE, and KEEPFILENAME are declared in globkb.h */

/*
  Values for ASCII, defining save operations; <=0 are binary, >0 are ascii
  SAVEBINARY1 & SAVEASCII1 refer to original binary format, with MAXVAR==31
  SAVEBINARY2 & SAVEASCII2 refer to first modified format, with MAXVAR == 31
  SAVEBINARY3 & SAVEASCII3 refer to modified format allowing MAXVAR == 96
  SAVEBINARY4 & SAVEASCII4 same, except nonprintables escaped for asciisave

  Type 3 not only allows saving Global information related to 96 variable
  models, but also save only the necessary dimension information, rather than
  MAXDIMS every time.
  The original ascii save format had long lines and imbedded '\0's terminating
  strings and macros.  Later formats have short lines and have character counts
  for each string and macro.  The only non-standard embedded characters would
  be in strings defined by escape sequences such as "\1".

  They are defined in globdefs.h
*/

#define DEFAULTSAVEOPTIONS -1
#define DEFAULTALL         0
		
#define NGLOBALS 37
#define NMACHINE 2
static char           *CharFmt;
static char           *LongFmt;
static char           *DoubleFmt;

static char            PadChar;

static Symbolhandle    Globals;

static Symbolhandle myInstall(char **globalHandle, size_t size)
{
	Symbolhandle   symh;
	char         **handle;
	WHERE("myInstall");
	
	if((symh = Install(SCRATCH,REAL)) == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	if((handle = myduphandle(globalHandle)) == (char **) 0)
	{
		goto errorExit;
	}
	setDATA(symh,(double **) handle);
	setNDIMS(symh,1);
	/* the calling function should reset DIM 1 for type char */
	setDIM(symh,1,myhandlelength(handle) / size);
	
	return (symh);

  errorExit:
	Removesymbol(symh);
	return(0);
}

static long saveName(char *nameItem)
{
	unsigned        m;
	WHERE("saveName");
	
	m = strlen(nameItem);
	if(m > 0 && write(KEEPFILE, nameItem, (int) m) != m ||
			  write(KEEPFILE,&PadChar,1) != 1)
	{
		return(0);
	}
	return (1);
} /*saveName()*/

static long       double2Symbol(double *dPointer,char *name,long length,
								long icomp)
{
	Symbolhandle   symh;
	long           i;
	
	if((COMPVALUE(Globals,icomp) = Makereal(length)) == (Symbolhandle) 0)
	{
		return (0);
	}
	symh = COMPVALUE(Globals,icomp);
	
	setNAME(symh,name);
	for(i=0;i<length;i++)
	{
		DATAVALUE(symh,i) = dPointer[i];
	}
	return (1);
	
}

static long       long2Symbol(long *lPointer,char *name,long length,
								long icomp)
{
	Symbolhandle   symh;
	long           i;
	
	if((COMPVALUE(Globals,icomp) = Makelong(length)) == (Symbolhandle) 0)
	{
		return (0);
	}
	symh = COMPVALUE(Globals,icomp);
	
	setNAME(symh,name);
	for(i=0;i<length;i++)
	{
		LONGDATAVALUE(symh,i) = lPointer[i];
	}
	return (1);
}

static Symbolhandle bundleGlobals(void)
{
	Symbolhandle       symh,symh1;
	long               i, j, k, n, nGlobals = 0;
	long               nvars = (long) NVARS, mvars;
	long               nterms = (long) NTERMS;
	double             dTemp;
	char             **cHandle, name[NAMELENGTH+1];
	double             nclasses[MAXVARS];
	long               maxvars = MAXVARS;
	long               newstyle = (ASCII>=SAVEASCII3 || ASCII<=SAVEBINARY3);
	WHERE("bundleGlobals");
	
	if((Globals = StrucInstall(GLOBALS,NGLOBALS)) == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	mvars = (newstyle) ? nvars : OLDMAXVARS;

	if(!newstyle && nvars > OLDMAXVARS)
	{
		sprintf(OUTSTR,
				"WARNING: more than %ld model variables; cannot save in v2.4 or v3.1 format",
				(long) OLDMAXVARS);
		putOUTSTR();
		nvars = -1;
	} /*if(!newstyle && nvars > OLDMAXVARS)*/
	
	if(nvars < 0)
	{ /* no globals; only save NVARS */
		double      nvars1 = -1.0;
		if(!double2Symbol(&nvars1,"NVARS",1,nGlobals++))
		{
			goto errorExit;
		}
	} /*if(nvars < 0)*/
	else
	{
		for(i=0;i<mvars;i++)
		{
			nclasses[i] = (double) NCLASSES[i];
		}
		
		
		if(!double2Symbol(&NVARS,"NVARS",1,nGlobals++))
		{
			goto errorExit;
		}
		if(newstyle && !long2Symbol(&maxvars,"MAXVARS",1,nGlobals++))
		{
			goto errorExit;
		}
		if(newstyle && !double2Symbol(&NFACTORS,"NFACTORS",1,nGlobals++))
		{
			goto errorExit;
		}
		if(!double2Symbol(&NDATA,"NDATA",1,nGlobals++))
		{
			goto errorExit;
		}
		if(!double2Symbol(&NOTMISSING,"NOTMISSING",1, nGlobals++))
		{
			goto errorExit;
		}
		if(!double2Symbol(&NTERMS,"NTERMS",1,nGlobals++))
		{
			goto errorExit;
		}
		if(!double2Symbol(&NREGX,"NREGX",1,nGlobals++))
		{
			goto errorExit;
		}
		if(!double2Symbol(&NY,"NY",1,nGlobals++))
		{
			goto errorExit;
		}
		if(!double2Symbol(&NERRORTERMS,"NERRORTERMS",1,nGlobals++))
		{
			goto errorExit;
		}
		dTemp = (double) PREVMODELTYPE;
		/* "PREVMODELTYPE" truncated to 12 = NAMELENGTH characters */
		if(!double2Symbol(&dTemp,"PREVMODELTYP",1,nGlobals++))
		{
			goto errorExit;
		}
		dTemp = (double) INCREMENTAL;
		if(!double2Symbol(&dTemp,"INCREMENTAL",1,nGlobals++))
		{
			goto errorExit;
		}
		dTemp = (double) USEGLMOFFSET;
		if(!double2Symbol(&dTemp,"USEGLMOFFSET",1,nGlobals++))
		{
			goto errorExit;
		}
		if(mvars > 0 && !double2Symbol(nclasses,"NCLASSES",mvars, nGlobals++))
		{
			goto errorExit;
		}
		if(newstyle)
		{
			if(!long2Symbol((long *) MODELINFO->errorterms,"ERRORTERMS1",
							WORDSPERTERM*(MAXERRTRMS+1), nGlobals))
			{
				goto errorExit;
			}
			/* "ERRORTERMS1" indicates new style of ERRORTERMS */
			symh = COMPVALUE(Globals,nGlobals++);
			/* make it a WORDSPERTERM by (MAXERRTRMS+1) matrix */
			setNDIMS(symh,2);
			setDIM(symh,1,WORDSPERTERM);
			setDIM(symh,2,MAXERRTRMS+1);
		}
		else
		{
			double    *errorterms;
			
			symh = COMPVALUE(Globals,nGlobals++) = Makereal(MAXERRTRMS+1);
			if(symh == (Symbolhandle) 0)
			{
				goto errorExit;
			}

			errorterms = DATAPTR(symh);
			for(i=0;i<=MAXERRTRMS;i++)
			{
				*errorterms++ = (double) MODELINFO->errorterms[i][0];
			}
			setNAME(symh,"ERRORTERMS");
			/* "ERRORTERMS" indicates old style of ERRORTERMS */
		}                       
		
		symh = COMPVALUE(Globals,nGlobals++) = Makesymbol(CHAR);
		if(symh == (Symbolhandle) 0)
		{
			goto errorExit;
		}

		if((cHandle = mygethandle((nvars+1)*(NAMELENGTH+1))) == (char **) 0)
		{
			goto errorExit;
		}
	
		setSTRING(symh,cHandle);
		setNAME(symh,"VARNAMES");
		setNDIMS(symh,1);
		setDIM(symh,1,nvars+1);
		k = 0;
		for(j=0;j<=nvars;j++)
		{
			strcpy((*cHandle) + k, VARNAMES[j]);
			k += strlen(VARNAMES[j]) + 1;
		}
		/* shrink cHandle to proper size */
		cHandle = mygrowhandle(cHandle,k);
		setSTRING(symh,cHandle);
		if(cHandle == (char **) 0)
		{
			goto errorExit;
		}

		/* now save handle stuff */
		if(MODEL != (modelHandle) 0)
		{
			if(newstyle)
			{
				if((symh = myInstall((char **) MODEL, sizeof(termWord))) ==
				   (Symbolhandle) 0)
				{
					goto errorExit;
				}
				Cutsymbol(symh);
				COMPVALUE(Globals,nGlobals++) = symh;
				setNAME(symh,"MODEL1"); /* MODEL1 indicates new style of MODEL */
				setTYPE(symh,LONG);
				/* make it a matrix with 1 column per term */
				setNDIMS(symh,2);
				setDIM(symh,2,DIMVAL(symh,1)/WORDSPERTERM);
				setDIM(symh,1,WORDSPERTERM);
			}/*if(newstyle)*/
			else
			{
				double        *model;
				
				symh = COMPVALUE(Globals,nGlobals++) = Makereal(nterms);
				if(symh == (Symbolhandle) 0)
				{
					goto errorExit;
				}

				model = DATAPTR(symh);
				
				for(i=0;i<nterms;i++)
				{
					*model++ = (double) (*MODEL)[i][0];
				}
				setNAME(symh,"MODEL");/*"MODEL" indicates old style of MODEL*/
			}
		}
		/* 
		  Only save X and Y when saving with v31:T. It is no longer necessary
		  since now Y is always identical to DATA(MODELVARS[0]) and X[i]
		  is now always identical to DATA(MODELVARS[i+1]).  The only place
		  where this was in question was in funbalanova, where Y (and X[i])
		  were lengthened.  That code there now resets MODELVARS
			*/
		if(!newstyle && Y != (double **) 0)
		{
			if((symh = myInstall((char **) Y, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"Y");
		}

		if(RESIDUALS != (double **) 0)
		{
			if((symh = myInstall((char **) RESIDUALS, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"RESIDUALS");
		}

		if(WTDRESIDUALS != (double **) 0)
		{
			if((symh = myInstall((char **) WTDRESIDUALS, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"WTDRESIDUALS");
		}

		if(GLMOFFSET != (double **) 0)
		{
			if((symh = myInstall((char **) GLMOFFSET, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"GLMOFFSET");
		}

		if(HII != (double **) 0)
		{
			if((symh = myInstall((char **) HII, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"HII");
		}

		if(SS != (double **) 0)
		{
			if((symh = myInstall((char **) SS, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"SS");
		}
		if(DF != (double **) 0)
		{
			if((symh = myInstall((char **) DF, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"DF");
		}

		if(MISSWTS != (double **) 0)
		{
			if((symh = myInstall((char **) MISSWTS, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"MISSWTS");
		}

		if(CASEWTS != (double **) 0)
		{
			if((symh = myInstall((char **) CASEWTS, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"CASEWTS");
		}

		if(ITERWTS != (double **) 0)
		{
			if((symh = myInstall((char **) ITERWTS, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"ITERWTS");
		}

		if(ALLWTS != (double **) 0)
		{
			if((symh = myInstall((char **) ALLWTS, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"ALLWTS");
		}

		if(LOGITN != (double **) 0)
		{
			if((symh = myInstall((char **) LOGITN, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"LOGITN");
		}

		if(STRMODEL != (char **) 0)
		{
			if((symh = myInstall(STRMODEL, sizeof(char))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setTYPE(symh,CHAR);
			setDIM(symh,1,1);
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"STRMODEL");
		}

		if(DEPVNAME != (char **) 0)
		{
			if((symh = myInstall( DEPVNAME, sizeof(char))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setTYPE(symh,CHAR);
			setDIM(symh,1,1);
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"DEPVNAME");
		}

		if(TERMNAMES != (char **) 0)
		{
			if((symh = myInstall( TERMNAMES, sizeof(char))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setTYPE(symh,CHAR);
			setDIM(symh,1,(long) NTERMS + 1);
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"TERMNAMES");
		}

		if(REGX != (double **) 0)
		{
			if((symh = myInstall((char **)REGX, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"REGX");
		}

		/* Note REGX2 is not saved, since it is used only for scratch */
		if(XTXINV != (double **) 0)
		{
			if((symh = myInstall((char **)XTXINV, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"XTXINV");
		}

		if(REGCOEF != (double **) 0)
		{
			if((symh = myInstall((char **)REGCOEF, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"REGCOEF");
		}

		if(REGSS != (double **) 0)
		{
			if((symh = myInstall((char **)REGSS, sizeof(double))) ==
			   (Symbolhandle) 0)
			{
				goto errorExit;
			}
			Cutsymbol(symh);
			COMPVALUE(Globals,nGlobals++) = symh;
			setNAME(symh,"REGSS");
		}

		/* 
		  Only save X and Y when saving with v31:T. It is no longer necessary
		  since now Y is always identical to DATA(MODELVARS[0]) and X[i]
		  is now always identical to DATA(MODELVARS[i+1]).  The only place
		  where this was in question was in funbalanova, where Y (and X[i])
		  were lengthened.  That code there now resets MODELVARS
		*/
		if(!newstyle && X[0] != (double **) 0)
		{
			symh = COMPVALUE(Globals,nGlobals++) = Makestruc(MAXVARS);
			if(symh == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symh,"X");
			for(i=0;i<MAXVARS;i++)
			{
				if(X[i] == (double **) 0)
				{
					break;
				}
				else
				{
					n = myhandlelength((char **)X[i])/sizeof(double);
					if((symh1 = Makereal(n)) == (Symbolhandle) 0)
					{
						goto errorExit;
					}
					COMPVALUE(symh,i) = symh1;
					setNCOMPS(symh,i);
					sprintf(name,"X%ld",i);
					setNAME(symh1,name);
					for(j=0;j<n;j++)
					{
						DATAVALUE(symh1,j) = (*X[i])[j];
					}
				}
			} /*for(i=0;i<MAXVARS;i++)*/
		}

		if(MODELVARS[0] != (Symbolhandle) 0)
		{
			symh = COMPVALUE(Globals,nGlobals++) = Makestruc(nvars+1);
			if(symh == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symh,"MODELVARS");

			for(i=0;i<=nvars;i++)
			{
				COMPVALUE(symh,i) = symh1 = Makesymbol(REAL);
				if(symh1 == (Symbolhandle) 0 || Copy(MODELVARS[i],symh1) == 0)
				{
					goto errorExit;
				}
			}
		}
	} /*if(nvars < 0){}else{}*/
	
	setNCOMPS(Globals,nGlobals);
	return (Globals); /* note that Globals is in symbol table */

  errorExit:
	if(Globals != (Symbolhandle) 0)
	{
		setNCOMPS(Globals,nGlobals);
		Removesymbol(Globals);
		Globals = (Symbolhandle) 0;
	}
	
	return (0);
}               

#ifdef PLATFORM

static Symbolhandle bundleMachine(void)
{
	Symbolhandle       symh;
	long               limits[4];
	long               nGlobals = 0;
	char             **cHandle;
	WHERE("bundleMachine");
	
	limits[0] = MAXDIMS;
	limits[1] = MAXVARS;
	limits[2] = WORDSPERTERM;
	limits[3] = VARSPERWORD;
	if((Globals = StrucInstall(MACHINE,NMACHINE)) == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	if(!long2Symbol(limits,"LIMITS",sizeof(limits)/sizeof(long),nGlobals++))
	{
		goto errorExit;
	}
		
	symh = COMPVALUE(Globals,nGlobals++) = Makesymbol(CHAR);
	if(symh == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	if((cHandle = mygethandle(strlen(PLATFORM)+1)) == (char **) 0)
	{
		goto errorExit;
	}
	
	strcpy(*cHandle,PLATFORM);
	setSTRING(symh,cHandle);
	setNAME(symh,"PLATFORM");
	setNDIMS(symh,1);
	setDIM(symh,1,1);

	setNCOMPS(Globals,nGlobals);
	return (Globals); /* note that Globals is in symbol table */

  errorExit:
	if(Globals != (Symbolhandle) 0)
	{
		setNCOMPS(Globals,nGlobals);
		Removesymbol(Globals);
		Globals = (Symbolhandle) 0;
	}
	
	return (0);
}
#endif /*PLATFORM*/             


static int savesym(Symbolhandle symh)
{
	long            i, ncomps, ndims, mdims;
	long            type = TYPE(symh);
	long            newstyle = (ASCII>=SAVEASCII3 || ASCII<=SAVEBINARY3);
	char           *name, *keyword;
	WHERE("savesym");
	
	if (symh == (Symbolhandle) 0)
	{
		return (1);
	}

	name = NAME(symh);
	type = TYPE(symh);

	/* don't save builtins, list, scratch, temps, keywords, T or F */
	if(type == BLTIN || type == UNDEF || type == ASSIGNED || type == LIST ||
	   type == GARB ||
	   isscratch(name) && name[2] != SPECIALPREFIX3)
	{
		return (1);
	}

	if((keyword = isKeyword(symh)))
	{
		/* caller should screen for keywords if they are not to be saved */
		name = keyword;
	}
	
	if(!saveName(name))
	{
		return (0);
	}

	if(!newstyle && ASCII != SAVEASCII2)
	{
		/* make sure type codes saved are right even if they change in future*/
		switch (type)
		{
		  case CHAR:
			type = CHAR_V31;
			break;
		  case REAL:
			type = REAL_V31;
			break;
		  case LOGIC:
			type = LOGIC_V31;
			break;
		  case STRUC:
			type = STRUC_V31;
			break;
		  case MACRO:
			type = MACRO_V31;
			break;
		  case PLOTINFO:
			type = PLOTINFO_V31;
			break;
		  case LONG:
			type = LONG_V31;
			break;
		  default:
			;
		} /*switch (type)*/
		if(!saveLong(type))
		{
			return (0);
		}
	} /*if(!newstyle)*/
	else if(!saveName(typeName(type)))
	{
		return (0);
	}

	if(!saveLong(NCLASS(symh)))
	{
		return (0);
	}

	ndims = NDIMS(symh);
	if(!saveLong(ndims))
	{
		return (0);
	}

	mdims = (newstyle) ? ndims : MAXDIMS;
	for (i = 1; i <= mdims; i++)
	{
		if(!saveLong(DIMVAL(symh,i)))
		{
			return (0);
		}
	}
	
	ncomps = symbolSize(symh);
	
	switch (type)
	{
	  case STRUC:
		for (i = 0; i < ncomps; i++)
		{
			if (!savesym(COMPVALUE(symh, i)))
			{
				return (0);
			}
		}
		return (1);

	  case MACRO:
	  case CHAR:
		return (saveHandle(STRING(symh), CHARHANDLE));
		
	  case REAL:
	  case LOGIC:
		return (saveHandle((char **) DATA(symh), DOUBLEHANDLE));

	  case LONG:
		return (saveHandle((char **) LONGDATA(symh), LONGHANDLE));

	  case PLOTINFO:
		return (saveGraph(GRAPH(symh))); /* saveGraph() is in plotutil.c */

	} /*switch (type)*/
	return (0);
}

Symbolhandle    save(list)
Symbolhandle    list;
{
	Symbolhandle    symh;
	Symbolhandle    symhFile = COMPVALUE(list,0);
	Symbolhandle    symhV24 = (Symbolhandle) 0;
	Symbolhandle    symhTorF = (Symbolhandle) 0;
	Symbolhandle    symhHelp = (Symbolhandle) 0;
	long            saveAll = DEFAULTALL;
	long            saveOptions = DEFAULTSAVEOPTIONS;
	long            saveSome = 0;
	long            v31 = -1, v24 = -1, logicvalue;
	long            ascii = (strcmp(FUNCNAME,"asciisave") == 0) ? 1 : -1;
	long            isFilename = (symhFile != (Symbolhandle) 0);
	char           *whenSaved;
	long            i, nKeywords = 0, nargs = NARGS(list);
	char           *keyword, *chptr;
	char           *fileName, *header;
#ifdef MACINTOSH
	OSType          fileType;
#endif
	WHERE("save");
	
	KEEPFILE = -1;

	OUTSTR[0] = '\0';

	for(i=1;i<nargs;i++)
	{ /* check for keywords */
		symh = COMPVALUE(list,i);
		if(!argOK(symh,(long) 0,i+1))
		{
			goto errorExit;
		}
		
		if((keyword = isKeyword(symh)))
		{
			if(strcmp(keyword,"all") == 0 ||
			   strcmp(keyword,"options") == 0)
			{
				nKeywords++;
				setNAME(symh, USEDKEYWORD);
				if(TYPE(symh) != LOGIC || symbolSize(symh) != 1)
				{
					notTorF(keyword);
					goto errorExit;
				}
				else
				{
					logicvalue = (DATAVALUE(symh,0) != 0.0);
					if(strcmp(keyword,"all") == 0)
					{
						saveOptions = saveAll = logicvalue;
					}
					else
					{
						saveOptions = logicvalue;
					}
				}
			}
			else if((strcmp(keyword,"old") == 0) ||
					strcmp(keyword,"v24") == 0 ||
					strcmp(keyword,"v31") == 0 ||
				   strcmp(keyword,"ascii") == 0)
			{/* old:T means use old save format */
				nKeywords++;
				setNAME(symh, USEDKEYWORD);
				if(TYPE(symh) != LOGIC || symbolSize(symh) != 1)
				{
					notTorF(keyword);
					goto errorExit;
				}
				else
				{
					logicvalue = (DATAVALUE(symh,0) != 0.0);
					if(strcmp(keyword,"old") == 0 ||
						strcmp(keyword,"v24") == 0)
					{
						v24 = logicvalue;
					}
					else if(strcmp(keyword,"v31") == 0)
					{
						v31 = logicvalue;
					}
					else
					{
						ascii = logicvalue;
					}
				}
			}
		} /*if((keyword = isKeyword(symh)))*/
		else if(isscratch(NAME(symh)))
		{
			sprintf(OUTSTR,
					"ERROR: attempt to save expression or function result");
		}
		else if(TYPE(symh) == BLTIN)
		{
			sprintf(OUTSTR,
					"ERROR: attempt to save built in function");
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
	} /*for(i=1;i<nargs;i++)*/
	
/*
	On the Mac, KeepVolume != 0 signals previous save() or asciisave()
	Elsewhere, KEEPFILENAME != 0 signals this.

	On Mac, a restore before a save sets KEEPFILENAME to the restore file
	name, but does not set KeepVolume

	If there has been a previous save() or asciisave(), a simple 'save()'
	without specifying a value for ascii, v31, v24, or old, uses the previous
	value of ASCII.  Otherwise, unless v31, v24, or old is set, it sets
	ASCII to the appropriate value based on previous save.  Thus, a simple
	save() continues to use the previously set mode, whether ascii or binary,
	or compatible with one of the previous version.
*/

	if(!isFilename)
	{ /* no file name */
#ifdef MACINTOSH
		if(KeepVolume == 0)
#else /*MACINTOSH*/
		if(KEEPFILENAME == (char **) 0)
#endif /*MACINTOSH*/
		{
			sprintf(OUTSTR,"ERROR: no save file name specified");
			goto errorExit;
		}
		/* Has been previous save() or asciisave()*/
		/* Use ascii if it has been set */
		if(ascii > 0)
		{ /* asciisave() or ascii:T*/
			ASCII = (ASCII == SAVEBINARY1) ? SAVEASCII1 : ASCII;
			ASCII = (ASCII == SAVEBINARY2) ? SAVEASCII2 : ASCII;
			ASCII = (ASCII == SAVEBINARY3) ? SAVEASCII3 : ASCII;
			ASCII = (ASCII == SAVEBINARY4) ? SAVEASCII4 : ASCII;
		}
		else if(ascii == 0)
		{ /* ascii:F */
			ASCII = (ASCII == SAVEASCII1) ? SAVEBINARY1 : ASCII;
			ASCII = (ASCII == SAVEASCII2) ? SAVEBINARY2 : ASCII;
			ASCII = (ASCII == SAVEASCII3) ? SAVEBINARY3 : ASCII;
			ASCII = (ASCII == SAVEASCII4) ? SAVEBINARY4 : ASCII;
		}

		if(v24 < 0 && v31 < 0)
		{ /* none of v24, v31, or old were keywords */
			v24 = (ASCII == SAVEASCII1 || ASCII == SAVEBINARY1);
			v31 = (ASCII == SAVEASCII2 || ASCII == SAVEBINARY2);
		} /*if(v24 < 0 && v31 < 0)*/
	} /*if(!isFilename)*/
	else
	{ /* file name specified*/
		ASCII = (ascii > 0) ? SAVEASCII4 : SAVEBINARY4;
	} /*if(!isFilename){}else{}*/

	v24 = (v24 < 0) ? 0 : v24;
	v31 = (v31 < 0) ? 0 : v31;
	if(v31 && v24)
	{
		sprintf(OUTSTR,
				"ERROR: illegal to use 'v31:T' with 'old:T' or 'v24:T' on %s()",
				FUNCNAME);
		goto errorExit;
	}
	
	if(v24)
	{
#ifdef MACINTOSH
		if(Binary)
		{
			sprintf(OUTSTR,
					"ERROR: Cannot write version 2.4x binary save files");
			goto errorExit;
		}
#endif /*MACINTOSH*/
		if(saveAll)
		{
			myerrorout("WARNING: all:T ignored with old:T and v24:T");
			saveAll = 0;
		}
		ASCII = (Binary) ? SAVEBINARY1 : SAVEASCII1;
	}
	else if(v31)
	{
		ASCII = (Binary) ? SAVEBINARY2 : SAVEASCII2;
	}

	if(nargs != nKeywords+1)
	{
		if(saveOptions >= 0 || saveAll > 0)
		{
			myerrorout("WARNING: 'options' and/or 'all' ignored on partial save");
		}
		saveOptions = saveAll = 0;
		saveSome = 1;
	} /*if(nargs != nKeywords+1)*/

	saveOptions = labs(saveOptions);

	if(!isFilename)
	{ /* use previous filename */
#ifdef MACINTOSH
		SetVol(0L, KeepVolume);
#endif
		sprintf(OUTSTR,"Saving on file %s",*KEEPFILENAME);
		putOUTSTR();
	} /*if(!isFilename)*/
	else
	{ /* get new file name */
		if(TYPE(symhFile) != CHAR || !isScalar(symhFile))
		{
			notCharOrString("file name");
			goto errorExit;
		}
		fileName = STRINGPTR(symhFile);

#ifdef MACINTOSH
		if(KeepVolume)
		{
			SetVol(0L, KeepVolume);
		}
		if(fileName[0] == '\0')
		{ /* null filename requests dialog box on Mac */
			Str255      fName;

			if(KEEPFILENAME != (char **) 0)
			{
				strcpy((char *) fName, *KEEPFILENAME);
				CtoPstr((char *) fName);
			}
			else
			{
				fName[0] = '\0';
			}
			fileName = macOpen("\pSave workspace as:", fName, 
							   WRITEIT, 0, (OSType *) 0);
			if (fileName == (char *) 0)
			{
				goto errorExit;
			}
			KeepVolume = Reply.vRefNum;
		} /*if(fileName[0] == '\0')*/
		
		if(RestoreVolume == 0)
		{
			RestoreVolume = KeepVolume;
		}
#endif /*MACINTOSH*/
			
		if((fileName = expandFilename(fileName)) == (char *) 0 ||
		   !isfilename(fileName))
		{
			goto errorExit;
		}
		mydisphandle(KEEPFILENAME);
		KEEPFILENAME = mygethandle(strlen(fileName) + 1);
		if(KEEPFILENAME == (char **) 0)
		{
			goto errorExit;
		}
		strcpy(*KEEPFILENAME,fileName);
	} /*if(!isFilename)*/

#ifdef MACINTOSH
	KEEPFILE = open(*KEEPFILENAME, O_WRONLY | O_TRUNC | O_CREAT);
#else /*MACINTOSH*/
#ifdef MSDOS
	if(Binary || ASCII == SAVEASCII1)
	{ /* file format must be binary */
		KEEPFILE = open(*KEEPFILENAME,
						O_WRONLY | O_TRUNC | O_CREAT | O_BINARY,
						S_IWRITE | S_IREAD);
	}
	else
	{
		KEEPFILE = open(*KEEPFILENAME, 
						O_WRONLY | O_TRUNC | O_CREAT | O_TEXT,
						S_IWRITE | S_IREAD);
	}
#else /*MSDOS*/
	KEEPFILE = creat(*KEEPFILENAME, 0644); /*Unix*/
#endif /*MSDOS*/
#endif /*MACINTOSH*/

	if (KEEPFILE < 0)
	{
		sprintf(OUTSTR,"ERROR: cannot open %s for saving", *KEEPFILENAME);
		goto errorExit;
	}

	CharFmt = "%s ";
	LongFmt = "%ld ";
	DoubleFmt = "%23.15e ";
	PadChar = ' ';
	switch (ASCII)
	{
	  case SAVEBINARY1:
	  case SAVEBINARY2:
		header = BINARYHEADER1;
		break;
	
	  case SAVEBINARY3:
	  case SAVEBINARY4:
		header = BINARYHEADER3;
		break;
	
	  case SAVEASCII1:
		header = ASCIIHEADER1;
		break;

	  case SAVEASCII2:
		header = ASCIIHEADER2;
		break;

	  case SAVEASCII3:
		header = ASCIIHEADER3;
		break;

	  case SAVEASCII4:
		header = ASCIIHEADER4;
		break;

	} /*switch (ASCII)*/

/* NOTE, PadChar should not be changed to '\n' until after saveName(header) */
	if(!saveName(header))
	{
		goto diskFull;
	}

	if(!Binary && ASCII != SAVEASCII1)
	{
		CharFmt = "%s\n";
		LongFmt = "%ld\n";
		DoubleFmt = "%.20g\n";
		PadChar = '\n';
	}
	
/* First things saved is time and date info */
	if(!v24)
	{ /* do not save time and date when old:T */
		whenSaved = getTimeAndDate();
		
		if((symh = Install(SAVETIME,CHAR)) == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		if((TMPHANDLE = mygethandle(strlen(whenSaved)+1)) == (char **) 0)
		{
			goto errorExit;
		}
		setSTRING(symh,TMPHANDLE);
		setNDIMS(symh,1);
		setDIM(symh,1,1);
		strcpy(STRINGPTR(symh),whenSaved);
		if (!savesym(symh))
		{
			goto diskFull;
		}
		Removesymbol(symh);
		
#ifdef PLATFORM
/*
   Next thing saved is information about the machine and version
   implementation
*/
		if(!v31)
		{ /* do not save machine info when v31:T is used */
			if((symh = bundleMachine()) == (Symbolhandle) 0)
			{
				goto errorExit;
			}

			if (!savesym(symh))
			{
				goto diskFull;
			}
			Removesymbol(symh);
		} /*if(!v31)*/
#endif /*PLATFORM*/
	} /*if(!v24)*/
	else
	{ /*v24:T*/
/*
   Or special things for version 2.4x such as T and F
*/
		char      *msgs = "Type Help() followed by a return to get help";
		char     **cHelp = (char **) 0;
		
		symhV24 = StrucInstall(SCRATCH,2);
		if(symhV24 == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		symhTorF = COMPVALUE(symhV24,0) = Makereal(1);
		if(symhTorF == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setTYPE(symhTorF,LOGIC);
		setNAME(symhTorF,"T");
		DATAVALUE(symhTorF,0) = 1.0;
		if(!savesym(symhTorF))
		{
			goto diskFull;
		}
		setNAME(symhTorF,"F");
		DATAVALUE(symhTorF,0) = 0.0;
		if(!savesym(symhTorF))
		{
			goto diskFull;
		}

		symhHelp = COMPVALUE(symhV24,1) = Makesymbol(CHAR);
		if(symhHelp == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		
		setSTRING(symhHelp,cHelp = mygethandle(strlen(msgs) + 1));
		if(cHelp == (char **) 0)
		{
			goto errorExit;
		}
		setNDIMS(symhHelp,1);
		setDIM(symhHelp,1,1);
		strcpy(*cHelp,msgs);
		setNAME(symhHelp,"help");
		if(!savesym(symhHelp))
		{
			goto diskFull;
		}
		setNAME(symhHelp,"HELP");
		if(!savesym(symhHelp))
		{
			goto diskFull;
		}
		Removesymbol(symhV24);
		symhV24 = (Symbolhandle) 0;
	} /*if(!v24){}else{}*/

/*
   Now save some or all of variables
*/
	
	if(!saveSome)
	{
	/* save things backwards so that they will be restored in right order*/
		for(symh = Lastsymbol(); symh != ENDPREV; symh = Prevsymbol(symh))
		{
			if (!savesym(symh))
			{
				goto diskFull;
			}
		}
	}
	else
	{/* partial save */
		for(i=1;i<nargs;i++)
		{
			symh = COMPVALUE(list,i);
			if(!(keyword = isKeyword(symh)) ||
			   strcmp(NAME(symh),USEDKEYWORD) != 0)
			{
				if (!savesym(symh))
				{
					goto diskFull;
				}
			}
		}
	}
		
	if(!v24 && saveOptions)
	{ /* do not save options with old:T */
		symh = getOpt((short *) 0);
		if(symh == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNAME(symh,OPTIONS);

		if (!savesym(symh))
		{
			goto diskFull;
		}
		Removesymbol(symh);
	}

	if(!v24 && saveAll)
	{
		if((symh = bundleGlobals()) == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		
		if (!savesym(symh))
		{
			goto diskFull;
		}
		Removesymbol(symh);
	}

#ifdef MACINTOSH
	fileType = (Binary) ? KEEPFILETYPE : ASCIIKEEPFILETYPE;
	macSetInfo(KeepVolume,*KEEPFILENAME,fileType,KEEPCREATOR);
#endif  /*MACINTOSH*/

	close(KEEPFILE);
	KEEPFILE = -1;
	
	chptr = (!Binary) ? "ascii" : "";
	if(!saveSome)
	{
		sprintf(OUTSTR,"Workspace %ssaved on file ", chptr);
	}
	else
	{
		sprintf(OUTSTR,"Specified variables %ssaved on file ", chptr);
	}
	myprint(OUTSTR);
	*OUTSTR = '\0';
	myerrorout(*KEEPFILENAME);
	
	return (NULLSYMBOL);

  diskFull:
	sprintf(OUTSTR,"ERROR: save file not complete.  Out of disk space?");

  errorExit:
	putOUTSTR();
	if(KEEPFILE >= 0)
	{
		close(KEEPFILE);
	}
	
	KEEPFILE = -1;
	if(v24)
	{
		Removesymbol(symhV24);
	}
	
	return (0);
	
} /*Symbolhandle save()*/

long saveHandle(h, type)
char    **h;
long      type;
{
	long                n, m, place = 0, i;
	long              **hlong;
	double            **hdouble;
	WHERE("saveHandle");
	
	if(h != (char **) 0)
	{
		if((n = myhandlelength(h)) < 0)
		{
			return (0);
		}
		if(Binary || ASCII == SAVEASCII1 && type == CHARHANDLE)
		{
			while (n > 0)
			{
				m = (n <= MAXCHUNK) ? n : MAXCHUNK;
				if (write(KEEPFILE, *h + place, (int) m) < m)
				{
					return (0);
				}
				n -= m;
				place += m;
			}
		} /*if(Binary || ASCII == SAVEASCII1 && type == CHARHANDLE)*/
		else
		{
			switch (type)
			{
			  case LONGHANDLE:
				hlong = (long **) h;
				n /= sizeof(long);
				for(i=0;i<n;i++)
				{
					if(!saveLong((*hlong)[i]))
					{
						return (0);
					}
				}
				break;

			  case DOUBLEHANDLE:
				hdouble  = (double **) h;
				n /= sizeof(double);
				for(i=0;i<n;i++)
				{
					if(!saveDouble((*hdouble)[i]))
					{
						return (0);
					}
				}
				break;

			  case CHARHANDLE:
				/* only reached for revised asciisave format */
				while(place < n)
				{
					if(!saveString(*h + place))
					{
						return (0);
					}
					place += strlen(*h + place) + 1;
				}
				break;

			} /*switch (type)*/
		} /*if(Binary || ASCII == SAVEASCII1 && type == CHARHANDLE){}else{}*/
	} /*if(h != (char **) 0)*/
	return (1);
} /*saveHandle()*/

long saveLong(longItem)
long           longItem;
{
	unsigned   m;
	int        result;
	
	sprintf(OUTSTR,LongFmt,longItem);
	m = strlen(OUTSTR);
	
	result = (write(KEEPFILE, OUTSTR, (int) m) == m);
	*OUTSTR = '\0';
	return (result);
} /*saveLong()*/

long saveDouble(doubleItem)
double         doubleItem;
{
	unsigned    m;
	long       result;
	
	sprintf(OUTSTR,DoubleFmt,doubleItem);
	m = strlen(OUTSTR);
	
	result = (write(KEEPFILE, OUTSTR, (int) m) == m);
	*OUTSTR = '\0';
	return (result);
} /*saveDouble()*/

/*
  Note this definition of PRINTABLE differs slightly from a similar one
  in commonou.c by including '\n' but not '\t' as a printable character
  Also all characters > '~' are escaped, even on a Macintosh.  The intent
  is that an asciisave() file should consist completely of printable
  ascii characters <= '~', plus newlines.
*/

#define PRINTABLE(C)  ((C) == '\n' || (C) >= 0x20 && (C) < 0x7f && (C) != '\\' )

long saveString(stringItem)
char          *stringItem;
{
	long           length = strlen(stringItem);
	long           nChars;
	unsigned char  c, stringBuffer[BUFFERLENGTH+3];
	unsigned char *pc = (unsigned char *) stringItem;
	WHERE("saveString");

	if(!saveLong(length))
	{
		goto errorExit;
	}

	while(length > 0)
	{
		nChars = 0;
		do
		{
			c = *pc++;
			length--;
			if(ASCII <= 3 || PRINTABLE(c))
			{
				stringBuffer[nChars++] = c;
			}
			else
			{
				stringBuffer[nChars++] = '\\';
				if(c == '\\')
				{
					stringBuffer[nChars++] = c;
				}
				else if(c == '\t')
				{
					stringBuffer[nChars++] = 't';
				}
				else
				{
					stringBuffer[nChars++] = c/64 + '0';
					stringBuffer[nChars++] = ((c/8) & 0x07) + '0';
					stringBuffer[nChars++] = (c & 0x07) + '0';
				}
			}
		} while(length > 0 && nChars < BUFFERLENGTH);
		if(write(KEEPFILE, (char *) stringBuffer, (int) nChars) != nChars)
		{
			goto errorExit;
		}
	} /*while(length > 0)*/

	if(write(KEEPFILE,&PadChar,1) != 1)
	{
		goto errorExit;
	}

	return (1);     

  errorExit:
	return (0);
} /*saveString()*/


