/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Screen
#else
#define __SEG__ Screen
#endif /*MPW3*/
#endif /*MPW*/

#include <ctype.h>
#include "globals.h"
#include "dbug.h"

extern char      **mygrowhandle(), **mygethandle();

#define GIW    0
#define NTRASH 1

/* control codes for keep */
#define KEEPP      0x01
#define KEEPCP     0x02
#define KEEPADJRSQ 0x04
#define KEEPRSQ    0x08
#define KEEPMODEL  0x10
#define KEEPALL    0x1f /*KEEPP | KEEPCP | KEEPADJRSQ | KEEPRSQ | KEEPMODEL*/

/* methods codes */
#define RSQMETHOD       1
#define ADJRSQMETHOD    2
#define CPMETHOD        3

/* positions for non-keyword controls, for compatibility with MacAnova2.4x */
#define IMBEST     1
#define IFORCED    2
#define IMETHOD    3
#define IS2        4
#define IPENALTY   5

static long         Keep;    /* specifies what to return if anything */
static long         Verbose; /* if 0, suppress printed output */
static long         Imodel;  /* sequence number of model */

static Symbolhandle Result;
static char         *Compnames[] = 
{
	"p","cp","adjrsq","rsq","model"
};
static long         Compops[] = 
{
	KEEPP, KEEPCP, KEEPADJRSQ, KEEPRSQ, KEEPMODEL
};

/* set up symbol for values to return */
static long initKeep(long imethod, long mbest, long nforced)
{
	Symbolhandle    symh;
	long            nmodels = 0, ncomps = 0, icomp;
	long            binomcoef;
	long            i, nvars = (long) NVARS - nforced;
	
	if(imethod == RSQMETHOD)
	{
		binomcoef = 1;

		for(i=1;i<=nvars;i++)
		{
			binomcoef = (binomcoef*(nvars-i+1))/i;
			nmodels += (binomcoef < mbest) ? binomcoef : mbest;
		}
	}
	else
	{
		nmodels = (nvars < 31) ? intpow(2.0,(double) nvars)-1 : 2147483647;
		nmodels = (nmodels < mbest) ? nmodels : mbest;
	}

	ncomps = 0; /* number of items requested to keep*/
	for(icomp=0;icomp<5;icomp++)
	{
		if(Keep & Compops[icomp])
		{
			ncomps++;
		}
	} /*for(icomp=0;icomp<5;icomp++)*/

	if(ncomps == 1)
	{ /* keep as vector */
		if(!(Keep & KEEPMODEL))
		{ /* REAL vector */
			if((Result = RInstall(SCRATCH,nmodels)) == (Symbolhandle) 0)
			{
				goto errorExit;
			}
		}
		else
		{ /*CHAR vector; will still need to allocate handle for string */
			if((Result = Install(SCRATCH,CHAR)) == (Symbolhandle) 0)
			{
				goto errorExit;
			}

			setNDIMS(Result,1);
		}
	} /*if(ncomps == 1)*/
	else
	{
		if((Result = StrucInstall(SCRATCH,ncomps)) == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		ncomps = 0;
		for(icomp = 0;icomp < 5;icomp++)
		{
			if(Keep & Compops[icomp])
			{
				if(Compops[icomp] != KEEPMODEL)
				{
					symh = COMPVALUE(Result,ncomps++) = Makereal(nmodels);
					if(symh == (Symbolhandle) 0)
					{
						goto errorExit;
					}
				} /*if(Compops[icomp] != KEEPMODEL)*/
				else
				{
					symh = COMPVALUE(Result,ncomps++) = Makesymbol(CHAR);
					if(symh == (Symbolhandle) 0)
					{
						goto errorExit;
					}
					setNDIMS(symh,1);
				} /*if(Compops[icomp] != KEEPMODEL){}else{}*/
				setCompName(symh,Compnames[icomp]);
			} /*if(Keep & Compops[icomp])*/
		} /*for(icomp = 0,icomp < 5;icomp++)*/
	} /*if(ncomps == 1){}else{}*/
	return (1);
	
  errorExit:
	return (0);
}

static long screenDoKeep(double p, double cp, double r2, double adjr2,
				   long *ind, long nvar)
{
	Symbolhandle   symh;
	long           ncomps = (TYPE(Result) == STRUC) ? NCOMPS(Result) : 1;
	long           jcomp = 0, icomp, i, length;
	char          *model;

	for(icomp=0;icomp<5;icomp++)
	{
		if(Keep & Compops[icomp])
		{
			symh = (ncomps > 1) ? COMPVALUE(Result,jcomp++) : Result;
			if(Compops[icomp] == KEEPP)
			{
				DATAVALUE(symh,Imodel) = p;
			}
			else if(Compops[icomp] == KEEPCP)
			{
				DATAVALUE(symh,Imodel) = cp;
			}
			else if(Compops[icomp] == KEEPRSQ)
			{
				DATAVALUE(symh,Imodel) = r2;
			}
			else if(Compops[icomp] == KEEPADJRSQ)
			{
				DATAVALUE(symh,Imodel) = adjr2;
			}
			else if(Compops[icomp] == KEEPMODEL)
			{
				length = strlen(*DEPVNAME);
				for (i = 0; i < nvar; i++)
				{
					length += strlen(VARNAMES[ind[i]]) + 1;
				}

				if(Imodel == 0)
				{ /* first model */
					TMPHANDLE = mygethandle(length+1);
					setSTRING(symh,TMPHANDLE);
					if(TMPHANDLE == (char **) 0)
					{
						goto errorExit;
					}
					model = *TMPHANDLE;
				} /*if(Imodel == 0)*/
				else
				{
					TMPHANDLE = STRING(symh);
					TMPHANDLE = mygrowhandle(TMPHANDLE,
											 myhandlelength(TMPHANDLE) + length+1);
					setSTRING(symh,TMPHANDLE);
					if(TMPHANDLE == (char **) 0)
					{
						goto errorExit;
					}
					model = skipStrings(*TMPHANDLE,Imodel);
				} /*if(Imodel == 0){}else{}*/
				strcpy(model,*DEPVNAME);
				model += strlen(model);
				for (i = 0; i < nvar; i++)
				{
					*model++ = (i == 0) ? '=' : '+';
					strcpy(model,VARNAMES[ind[i]]);
					model += strlen(model);
				}					
				setDIM(symh,1,Imodel+1);
			}
		} /*if(Keep & Compops[icomp])*/
	} /*for(icomp=0;icomp<5;icomp++)*/
	Imodel++;
	return (1);
	
  errorExit:
	return (0);
	
} /*screenDoKeep()*/

/*
   Front end for leaps and bounds screening of all possible regressions.
   The model has been decoded into the globals, and the various controlling
   keywords are in COMPVALUE(list,i),i=1,...,nargs
*/

Symbolhandle    screen(list)
Symbolhandle    list;
{
	/* control  routine to do leaps and bounds */

	Symbolhandle    symh, Install();
	Symbolhandle    symhMbest = (Symbolhandle) 0,symhForced = (Symbolhandle) 0;
	Symbolhandle    symhMethod = (Symbolhandle) 0, symhS2 = (Symbolhandle) 0;
	Symbolhandle    symhPenalty = (Symbolhandle) 0;
	Symbolhandle    symhPrint = (Symbolhandle) 0;
	Symbolhandle    symhKeep = (Symbolhandle) 0;
	char           *keyword;
	char           *tmps;
	long            keyindex;
	long            nargs = NARGS(list);
	long            i, j, k, i2, keystart;
	long            nv, it, nforced, imethod, mbest;
	long            nd, nc, nw, nr, ne, iv;
	long            ij, ji;
	long            nkeep;
	long          **iw = (long **) 0;
	long            ndata = (long) NDATA, nvars = (long) NVARS;
	long            notmissing = (long) NOTMISSING;
	char           *keepvec;
	double          penalty, tl, s2;
	double        **rt, **rw, tmpd, **tmpdh;
	double         *regx1, *regx2;
	WHERE("screen");
	TRASH(NTRASH,errorExit);
	
	*OUTSTR = '\0';
	Verbose = 1;
	Keep = Imodel = 0;
	Result = (Symbolhandle) 0;
	
	for(keystart=1;keystart<nargs;keystart++)
	{
		if(isKeyword(COMPVALUE(list,keystart)))
		{
			break;
		}
	}
/* keystart is now index of first keyword argument */
	/*
	  check for keywords
	  for compatibility, positional arguments are still allowed before 1st
	  keyword, if any
	*/
	if((keyindex = findKeyword(list,"mbest",keystart)) >= 0)
	{
		symhMbest = COMPVALUE(list, keyindex);
		setNAME(symhMbest,USEDKEYWORD);
	}
	if((keyindex = findKeyword(list,"force",keystart)) >= 0)
	{
		symhForced = COMPVALUE(list, keyindex);
		setNAME(symhForced,USEDKEYWORD);
	}
	if((keyindex = findKeyword(list,"meth",keystart)) >= 0)
	{
		symhMethod = COMPVALUE(list, keyindex);
		setNAME(symhMethod,USEDKEYWORD);
	}
	if((keyindex = findKeyword(list,"s2",keystart)) >= 0 ||
	   (keyindex = findKeyword(list,"ssq",keystart)) >= 0)
	{
		symhS2 = COMPVALUE(list, keyindex);
		setNAME(symhS2,USEDKEYWORD);
	}
	if((keyindex = findKeyword(list,"pen",keystart)) >= 0)
	{
		symhPenalty = COMPVALUE(list, keyindex);
		setNAME(symhPenalty,USEDKEYWORD);
	}
	if((keyindex = findKeyword(list,"print",keystart)) >= 0)
	{
		symhPrint = COMPVALUE(list, keyindex);
		setNAME(symhPrint,USEDKEYWORD);
	}
	if((keyindex = findKeyword(list,"keep",keystart)) >= 0)
	{
		symhKeep = COMPVALUE(list, keyindex);
		setNAME(symhKeep,USEDKEYWORD);
	}

/* look for unused keyword and non-keywords after keystart */
	
	while(nargs > keystart)
	{
		symh = COMPVALUE(list,nargs-1);
		if(!(keyword = isKeyword(symh)))
		{
			sprintf(OUTSTR,
					"ERROR: illegal mixture of keyword and non-keyword arguments");
			goto errorExit;
		}
		else if(strcmp(USEDKEYWORD,NAME(symh)) != 0)
		{
			sprintf(OUTSTR,
					"ERROR: illegal keyword %s to %s",keyword,FUNCNAME);
			goto errorExit;
		}
		nargs--;
	} /*while(nargs > keystart)*/

	if(symhKeep != (Symbolhandle) 0)
	{
		if(TYPE(symhKeep) != CHAR || !isVector(symhKeep))
		{
			sprintf(OUTSTR,
					"ERROR: value for keyword 'keep' for %s must be CHARACTER vector",
					FUNCNAME);
			goto errorExit;
		} /*if(TYPE(symhKeep) != CHAR || !isVector(symhKeep))*/

		nkeep = symbolSize(symhKeep);
		keepvec = STRINGPTR(symhKeep);
		for(i=0;i<nkeep;i++)
		{
			if(strcmp("all",keepvec) == 0)
			{
				Keep |= KEEPALL;
			}
			else if(strcmp("p",keepvec) == 0)
			{
				Keep |= KEEPP;
			}
			else if(strcmp("cp",keepvec) == 0)
			{
				Keep |= KEEPCP;
			}
			else if(strncmp("adj",keepvec,3) == 0)
			{
				Keep |= KEEPADJRSQ;
			}
			else if(strcmp("rsq",keepvec) == 0)
			{
				Keep |= KEEPRSQ;
			}
			else if(strncmp("model",keepvec,5) == 0)
			{
				Keep |= KEEPMODEL;
			}
			else
			{
				sprintf(OUTSTR,
						"ERROR: %s not legal value for 'keep'",keepvec);
				goto errorExit;
			}
			keepvec = skipStrings(keepvec,1);
		}
		Verbose = 0;
	} /*if(symhKeep != (Symbolhandle) 0)*/

	if(symhPrint != (Symbolhandle) 0)
	{
		if(TYPE(symhPrint) != LOGIC || !isScalar(symhPrint))
		{
			notTorF("print");
			goto errorExit;
		}
		Verbose = (DATAVALUE(symhPrint,0) != 0.0);
	} /*if(symhPrint != (Symbolhandle) 0)*/

	if(!Verbose && !Keep)
	{
		sprintf(OUTSTR,
				"ERROR: print:F illegal for %s unless keyword 'keep' is used",
				FUNCNAME);
		goto errorExit;
	} /*if(!Verbose && !Keep)*/

	if(Verbose)
	{
		sprintf(OUTSTR,"Model used is %s=",*DEPVNAME);
		myprint(OUTSTR);
		tmps = *STRMODEL;
		/* *STRMODEL should always be of the form "depv=1+..." */
		while(*tmps++ != '+')
		{
			;
		}
		myerrorout(tmps);
		*OUTSTR = '\0';
	} /*if(Verbose)*/
		
	/* set up screen parameters */
	nv = nvars + 1;
	it = 1;
	nd = nv;
	nc = 3 * nd;
	nw = 4 * nv;
	tl = 10.;
	ne = 0;
	iv = 0;
	

	if(!getScratch(iw,GIW,nw,long))
	{
		goto errorExit;
	}
	for (i = 0; i < nw; i++)
	{
		(*iw)[i] = 0;
	}

	for (i = 0; i < nvars; i++)
	{
		(*iw)[i] = i + 1;
	}
	(*iw)[nvars] = nv;

	keyword = (char *) 0;
	mbest = 5;
	if (nargs > IMBEST &&
		(symh = COMPVALUE(list,IMBEST)) != (Symbolhandle) 0 &&
		!isKeyword(symh))
	{
		if(symhMbest != (Symbolhandle) 0)
		{
			keyword = "mbest";
		}
		else
		{
			symhMbest = symh;
		}
	}

	if(symhMbest != (Symbolhandle) 0 && !keyword)
	{
		if (TYPE(symhMbest) != REAL || !isScalar(symhMbest) ||
			(tmpd = DATAVALUE(symhMbest,0)) <= 0 || tmpd != floor(tmpd))
		{
			sprintf(OUTSTR,"ERROR: mbest must be positive integer");
			goto errorExit;
		}
		mbest = (long) tmpd;
	}

	nforced = 0;
	if (!keyword && nargs > IFORCED &&
		(symh = COMPVALUE(list,IFORCED)) != (Symbolhandle) 0 &&
		!isKeyword(symh))
	{
		if(symhForced != (Symbolhandle) 0)
		{
			keyword = "forced";
		}
		else
		{
			symhForced = symh;
		}
	}

	if(!keyword && symhForced != (Symbolhandle) 0)
	{
		if(!isVector(symhForced) ||
		   (TYPE(symhForced) != REAL && TYPE(symhForced) != CHAR))
		{
			sprintf(OUTSTR,
					"ERROR: List of forced variables must be REAL or CHARACTER vector");
			goto errorExit;
		}

		switch ((int) TYPE(symhForced))
		{
		  case REAL:
			tmpdh = DATA(symhForced);
			for (i = 0; i < DIMVAL(symhForced,1); i++)
			{
				tmpd = (*tmpdh)[i];
				if(tmpd < 1.0 || tmpd != floor(tmpd))
				{
					sprintf(OUTSTR,
							   "ERROR: forced variable subscripts must be positive integers");
					goto errorExit;
				}
				else if((k = (long) tmpd) > nvars)
				{
					sprintf(OUTSTR,
							"ERROR: forced variable subscript out of range");
					goto errorExit;
				}
				(*iw)[k + nvars] = 1;
			}
			break;

		  case CHAR:
			tmps = STRINGPTR(symhForced);
			for (i = 0; i < DIMVAL(symhForced,1); i++)
			{
				for (k = 1; k <= nvars; k++)
				{
					if (!strcmp(tmps, VARNAMES[k]))
					{
						break;
					}
				}
				tmps += strlen(tmps) + 1;
				if (k < 1 || k > nvars)
				{
					sprintf(OUTSTR,"ERROR: forced variable not in model");
					goto errorExit;
				}
				(*iw)[k + nvars] = 1;
			}
			break;
		  default:
			/* should not occur now */
			goto errorExit;
		} /*switch ((int) TYPE(symhForced))*/

		for (i = 0; i < nvars; i++)
		{
			if ((*iw)[nvars + i + 1])
			{
				(*iw)[nforced] = i + 1;
				nforced++;
			}
			else
			{
				(*iw)[nvars - 1 - i + nforced] = i + 1;
			}
		} /*for (i = 0; i < nvars; i++)*/
		(*iw)[nvars] = nv;
	} /*if(!keyword && symhForced != (Symbolhandle) 0)*/

	if(nforced == nvars)
	{
		sprintf(OUTSTR,
				"ERROR: all variables forced in, no screening possible");
		goto errorExit;
	}
	
	imethod = CPMETHOD; /* default is Mallow's Cp */
	if (!keyword && nargs > IMETHOD &&
		(symh = COMPVALUE(list,IMETHOD)) != (Symbolhandle) 0 &&
		!isKeyword(symh))
	{
		if(symhMethod != (Symbolhandle) 0)
		{
			keyword = "method";
		}
		else
		{
			symhMethod = symh;
		}
	}

	if(!keyword && symhMethod != (Symbolhandle) 0)
	{
		if (TYPE(symhMethod) != CHAR || !isScalar(symhMethod))
		{
			sprintf(OUTSTR,
					"ERROR: method must be quoted string or CHARACTER scalar");
			goto errorExit;
		}
		strncpy(OUTSTR,STRINGPTR(symhMethod),BUFFERLENGTH-1);
		OUTSTR[BUFFERLENGTH-1] = '\0';
		for(tmps=OUTSTR;*tmps;tmps++)
		{
			if(isupper(*tmps))
			{
				*tmps = tolower(*tmps);
			}
		}
		if(strcmp(OUTSTR,"r2") == 0 || strcmp(OUTSTR,"r") == 0 ||
		   strcmp(OUTSTR,"rsq") == 0 )
		{
			imethod = RSQMETHOD;
		}
		else if(strncmp(OUTSTR,"adj",3) == 0 || 
				strcmp(OUTSTR,"a") == 0)
		{
			imethod = ADJRSQMETHOD;
		}
		else if(strcmp(OUTSTR,"cp") == 0 || 
				strcmp(OUTSTR,"c") == 0)
		{
			imethod = CPMETHOD;
		}
		else
		{
			sprintf(OUTSTR,"ERROR: illegal method %s for %s",
					STRINGPTR(symhMethod), FUNCNAME);
			goto errorExit;
		}
		*OUTSTR = '\0';
	} /*if(!keyword && symhMethod != (Symbolhandle) 0)*/
	
	s2 = 0.0;
	if (!keyword && nargs > IS2 &&
		(symh = COMPVALUE(list,IS2)) != (Symbolhandle) 0)
	{
		if(symhS2 != (Symbolhandle) 0)
		{
			keyword = "s2";
		}
		else
		{
			symhS2 = symh;
		}
	}

	if(!keyword && symhS2 != (Symbolhandle) 0)
	{
		if(imethod != CPMETHOD)
		{
			myerrorout("WARNING: s2 specified when method is not 'cp'; ignored");
		}
		else if(!keyword)
		{
			if (TYPE(symhS2) != REAL || !isScalar(symhS2) ||
				(s2 = DATAVALUE(symhS2,0)) <= 0.0)
			{
				sprintf(OUTSTR,
						"ERROR: s2 must be single positive REAL number");
				goto errorExit;
			}
		}
	} /*if(!keyword && symhS2 != (Symbolhandle) 0)*/
	
	penalty = 2.0;
	if (!keyword && nargs > IPENALTY &&
		(symh = COMPVALUE(list,IPENALTY)) != (Symbolhandle) 0)
	{
		if(symhPenalty != (Symbolhandle) 0)
		{
			keyword = "penalty";
		}
		else
		{
			symhPenalty = symh;
		}
	}

	if(!keyword && symhPenalty != (Symbolhandle) 0)
	{
		if (TYPE(symhPenalty) != REAL || !isScalar(symhPenalty) ||
			(penalty = DATAVALUE(symhPenalty,0)) <= 0.0)
		{
			sprintf(OUTSTR,"ERROR: penalty must be single positive REAL");
			goto errorExit;
		}
	}

	if(keyword)
	{
		sprintf(OUTSTR,
				"ERROR: cannot specify %s by both keyword and positional arguments",
				keyword);
		goto errorExit;
	}
	
	if(Keep != 0 && !initKeep(imethod, mbest, nforced))
	{
		goto errorExit;
	}
	

	/* set up REGX */
	REGX = (double **) mygethandle((ndata * (nvars + 1)) * sizeof(double));
	if (REGX == (double **) 0)
	{
		goto errorExit;
	}
	j = 0;
	for (i = 0; i < ndata; i++)
	{
		if (MODELTYPE & MISSWEIGHTS && (*MISSWTS)[i] == 0.0)
		{
			continue;
		}
		ij = j; /* j + i2*notmissing */
		for (i2 = 0; i2 < nvars; i2++)
		{
			(*REGX)[ij] = (*X[i2])[i];
			ij += notmissing;
		}
		(*REGX)[ij] = (*Y)[i];
		j++;
	} /*for (i = 0; i < ndata; i++)*/

	nr = 2 * mbest * nvars + 7 * nd;
	SS = (double **) mygethandle(nr * sizeof(double));
	if (SS == (double **) 0)
	{
		goto errorExit;
	}
	rw = SS;
	/* take out means */
	for (i = 0; i <= nvars; i++)
	{
		tmpd = 0.0;
		regx1 = *REGX + i*notmissing; /* (*REGX)[k+i*notmissing] */
		for (k = 0; k < notmissing; k++)
		{
			tmpd += *regx1++;
		}
		tmpd = tmpd / notmissing;
		(*rw)[i] = tmpd;
		regx1 = *REGX + i*notmissing; /* (*REGX)[k+i*notmissing] */
		for (k = 0; k < notmissing; k++)
		{
			*regx1++ -= tmpd;
		}
	}

	/* set up centered cross products matrix */
	/* should work, but doesn't on PC
	   XTXINV = (double **) mygethandle( (3 * (nvars+1) * (nvars+1)) * sizeof(double));
	   */
	i = (3 * (nvars + 1) * (nvars + 1)) * sizeof(double);
	XTXINV = (double **) mygethandle(i);
	if(XTXINV == (double **) 0)
	{
		goto errorExit;
	}
	rt = XTXINV;
	
	for (i = 0; i <= nvars; i++)
	{
		ij = i + (nvars + 1) * i; /* i + (nvars+1) * j */
		ji = ij ; /* j + (nvars+1)*i */
		for (j = i; j <= nvars; j++)
		{
			tmpd = 0.0;
			regx1 = *REGX + i*notmissing; /* (*REGX)[k + i*notmissing] */
			regx2 = *REGX + j*notmissing; /* (*REGX)[k + j*notmissing] */
			for (k = 0; k < notmissing; k++)
			{
				tmpd += (*regx1++) * (*regx2++);
			}
			(*rt)[ji] = (*rt)[ij] = tmpd;
			ij += nvars + 1;
			ji++;
		}
	} /*for (i = 0; i <= nvars; i++)*/

	if(interrupted() != INTNOTSET)
	{
		goto errorExit;
	}
	
	if(Verbose)
	{
		if (nforced > 0)
		{
			sprintf(OUTSTR, "%ld variables were forced:  ", nforced);
			myprint(OUTSTR);
			for (i = 0; i < nforced; i++)
			{
				sprintf(OUTSTR, "%s  ", VARNAMES[(*iw)[i]]);
				myprint(OUTSTR);
			}
			myeol();
		}
		if (imethod == CPMETHOD)
		{
			if (s2 > 0.0)
			{
				sprintf(OUTSTR,	"Error variance set to %11.5g",s2);
			}
			else
			{
				sprintf(OUTSTR,	"Error variance set to full model mse");
			}
			myprint(OUTSTR);
			sprintf(OUTSTR,", penalty factor is %11.5g",penalty);
			putOUTSTR();
		}
		myerrorout("  p    C(p)  Adj R^2  R^2    Model");
	} /*if(Verbose)*/
	
	if (!leapsscreen(&nv, &it, &nvars, &nforced, &notmissing,
					&imethod, &penalty,
					&mbest, *rt, &nd, &nc,
					*iw, &nw, *rw, &nr, &tl, &s2, &ne, &iv))
	{
		*OUTSTR = '\0';
		goto errorExit;
	}

	*OUTSTR = '\0';
	emptyTrash();
	glmcleanup(0);

	return ((Result != (Symbolhandle) 0) ? Result : NULLSYMBOL);

  errorExit:
	putOUTSTR();
	
	emptyTrash();
	Removesymbol(Result);
	Result = (Symbolhandle) 0;
	glmcleanup(0);
	return (0);
}

/*
   ind[]       Vector of indices of independent followed by dependent variable
               in current model
   *mn         number of variables in model
   *rm         full model MSE
   *ib         criterion: 1: r2, 2: adjr2, 3: cp
   *pen        penalty for cp
   *tss        total sum of squares
   *df         d.f. in *tss
*/
long leapsqprint(ind, mn, rm, tss, s2, scale, df, it, ib, pen)
long           *ind, *mn, *it, *ib;
double         *rm, *tss, *s2, *scale, *df, *pen;
{
	long            i, p;
	double          rss, adjr2, r2, cp;
	long            starter, length;
	WHERE("leapsqprint");
	

	if(*ib == RSQMETHOD)
	{
		rss = *rm;
	}
	else if(*ib == ADJRSQMETHOD)
	{
		rss = (*df - *mn) * *rm;
	}
	else /*CPMETHOD*/
	{
		rss = *rm - *pen * *mn * *s2;
	}
	r2 = 1. - rss / *tss;
	adjr2 = 1. - *df * rss / (*tss * (*df - *mn));
	p = *mn + *it;
	cp = rss / *s2 + *pen * p - (*df + *it);

	if(Verbose)
	{		
		sprintf(OUTSTR,"%3ld %8.3f %6.4f %6.4f",p,cp,adjr2,r2);
		myprint(OUTSTR);
	
		length = starter = strlen(OUTSTR);
	
		for (i = 0; i < *mn; i++)
		{
			sprintf(OUTSTR, " %s", VARNAMES[ind[i]]);
			myprint(OUTSTR);
			if(i < *mn - 1 &&
			   (length += strlen(OUTSTR)) >= SCREENWIDTH - NAMELENGTH)
			{
				myeol();
				length = starter;
				sprintf(OUTSTR,"%*s",starter," "); /* NOTE use of %*s */
				myprint(OUTSTR);
			}
		}
		myeol();
	} /*if(Verbose)*/

	if(Keep && !screenDoKeep((double) p, cp, r2, adjr2, ind, *mn))
	{
		goto errorExit;
	}
	
	*OUTSTR = '\0';
	return (1);

  errorExit:
	*OUTSTR = '\0';
	return (0);
	
}


#define nv (*NV)
#define it (*IT)
#define kx (*KX)
#define nf (*NF)
#define no (*NO)
#define ib (*IB)
#define fd (*FD)
#define mb (*MB)
#define nd (*ND)
#define nc (*NC)
#define nw (*NW)
#define nr (*NR)
#define tl (*TL)
#define s2 (*S2)
#define ne (*leapsNE)
#define iv (*IV)
#define rt(i,j) ( RT[(i-1) + nd*(j-1)] )
#define rw(i)  ( RW[i-1] )
#define iw(i) ( IW[i-1] )

/* converted fortran, send proper pointer for array element in function call */
#define prt(i,j) ( RT + (i-1) + nd*(j-1) )
#define prw(i)  ( RW + i-1 )
#define piw(i) ( IW + i-1 )

long leapsscreen(NV, IT, KX, NF, NO, IB, FD, MB, RT, ND, NC, IW,
				 NW, RW, NR, TL, S2, leapsNE, IV)
long           *NV, *IT, *KX, *NF, *NO, *IB, *MB, *ND, *NC, *IW,
               *NW, *NR, *leapsNE, *IV;
double         *FD, *RT, *RW, *TL, *S2;
{
	long            nrm, ns, longparms[15];
	WHERE("leapsscreen");
	/*
	  
	  dimension rt(nd,nc),iw(nw),rw(nr)
	  * ******************************************************************** *
	  *                                                                      *
	  *                   Regressions by Leaps and Bounds                    *
	  *                     G.M.Furnival and R.W.Wilson                      *
	  *                           Version 8/10/81                            *
	  *                                                                      *
	  * call screen(nv,it,kx,nf,no,ib,fd,mb,rt,nd,nc,iw,nw,rw,nr,tl,s2,ne,iv)*
	  *                                                                      *
	  *  nv = number of rows and columns in input matrix (nv.le.m where m    *
	  *       is the number of bits in the fractional part of the real       *
	  *       numbers employed in the computations)                          *
	  *  it = type of input matrix (0=products about 0, 1=products about     *
	  *       means, 2=correlation matrix)                                   *
	  *  kx = number of independent variables (kx.lt.nv)                     *
	  *  nf = number of indepent variables to be forced i.e. included in     *
	  *       every subset (nf.lt.kx)                                        *
	  *  no = number of observations (no.gt.kx when it=0, no.gt kx+1         *
	  *       otherwise)                                                     *
	  *  ib = selection criterion code (1=r**2, 2=adjusted r**2, 3=mallows   *
	  *       c(p) with frane's variable penalty)                            *
	  *  fd = variable penalty or f-to-delete for c(p). a value of 2.0       *
	  *       gives mallows original c(p). (a value of fd .le.0 is replaced  *
	  *       by 2.0)                                                        *
	  *  mb = number of best regressions desired-for each size subset when   *
	  *       ib=1, in total when ib.gt.1                                    *
	  *  rt = real two-dimensional array. first nv columns contain input     *
	  *       matrix of products or correlations. rest of array is working   *
	  *       storage.                                                       *
	  *  nd = first dimension of rt(nd.ge.nv)
	  *  nc = second dimension of rt (nc.ge.3*nd)                            *
	  *  iw = integer one-dimensional array. first kx+1 elements contain     *
	  *       subscripts of variables to be used in problem with subscripts  *
	  *       of forced variables in first nf elements, subscripts of free   *
	  *       variables in next kx-nf elements and subscript of dependent    *
	  *       variable in element kx+1. rest of array is working storage.    *
	  *  nw = dimension of iw (nw.ge.4*nd)                                   *
	  *  rw = real one dimensional array. when it=1,first nv elements must   *
	  *       contain vector of means. rest of array is working storage.     *
	  *  nr = dimension of rw (nr.ge.2*mb*kx+7*nd if ib=1, nr.ge.2*mb+7*nd   *
	  *       if ib.gt.1)                                                    *
	  *  tl = minimum allowable tolerance (tl.lt.kx*kx/2**(m/2) or.ge.1      *
	  *       reset to kx*kx/2**(m/2))                                       *
	  *  s2 = estimated variance of residuals for c(p) (value.le.0 replaced  *
	  *       by mean square residual from full model)                       *
	  *  ne = error code (1=input parameter out of bounds, 2=invalid list    *
	  *       of subscripts in iw, 3=zero or negative diagonal in input      *
	  *       matrix rt, 4=set of forced variables nearly collinear,         *
	  *       5=full model nearly collinear-s2 must be supplied for c(p),    *
	  *       6=r**2 so close to 1 (r**2.gt.1-2**(-m/2)) that selection of   *
	  *       best subsets is uncertain.                                     *
	  *  iv = version of algorithm (0=partial reordering, 2=full reordering) *
	  *                                                                      *
	  * ******************************************************************** *
	  *  changes in screen specific to multreg are indicated
	  *   by the presence of the logical word 'uofm', which is
	  *   always set to 'true'.  most code changes are in
	  *   format statements, usually e-format changes to g-format,
	  *   indicated by comment statements.
	  *************************************************************************
	  */

	ne = 0;
	ns = (ib == RSQMETHOD) ? mb * kx : mb;
	if (0 > nf ||
		nf >= kx ||
		kx >= nv ||
		nv > nd ||
		kx + (1 > it ? it : 1) >= no ||
		nc < 3 * nd ||
		nw < 4 * nd ||
		nr < 2 * ns + 7 * nd ||
		ib < RSQMETHOD ||
		ib > CPMETHOD ||
		it < 0 ||
		it > 2 ||
		mb <= 0 ||
		iv < 0 ||
		iv > 1
		)
	{
		ne = 1;					/* parameter out of bounds */
	}
	else
	{
		nrm = 7 * nd + ns + 1;
		longparms[0] = ns;
		longparms[1] = *ND;
		longparms[2] = *KX;
		longparms[3] = *NO;
		longparms[4] = *IT;
		longparms[5] = *NF;
		longparms[6] = *IB;
		longparms[7] = *MB;
		longparms[8] = *NV;
		if(!leapsforwd(longparms,
					   prt(1, 1), prt(1, nd + 1), prt(1, 2 * nd + 1),
					   prw(1), prw(nd + 1),	prw(2 * nd + 1), prw(3 * nd + 1),
					   prw(4 * nd + 1), prw(5 * nd + 1), prw(6 * nd + 1),
					   prw(7 * nd + 1), prw(nrm),
					   piw(1), piw(nd + 1), piw(2 * nd + 1), piw(3 * nd + 1),
					   S2, TL, FD, leapsNE, IV))
		{
			goto errorExit;
		}
		
	}

	if (ne > 0)
	{
		switch ((int) ne)
		{
		  case 1:
			myerrorout("ERROR: (internal to screen) screen input parameters out of bounds");
			break;
		  case 2:
			myerrorout("ERROR: (internal to screen) invalid list of iw subscripts");
			break;
		  case 3:
			myerrorout("ERROR: (internal to screen) negative diagonal");
			break;
		  case 4:
			myerrorout("ERROR: set of forced variables nearly collinear");
			break;
		  case 5:
			myerrorout("ERROR: full model nearly collinear");
			break;
		  case 6:
			myerrorout("ERROR: r^2 so close to 1 that subset selection is uncertain");
			break;
		  default:
			;
		}
	}

	return (1);
	
  errorExit:
	return (0);
}
