/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Setoptio
#else
#define __SEG__ Setoptio
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

#ifdef MACINTOSH
#include "macIface.h"
#endif /*MACINTOSH*/

#define MAXSEED 0x7fffffffL

#define ISEEDS     1
#define INSIG      2
#define IFORMAT    3
#define IWFORMAT   4
#define IANGLES    5
#define ILINES     6
#define IWIDTH     7
#define IERRORS    8
#define IPROMPT    9
#define IBATCHECHO 10
#define IUPDATE    11
#define IWARNINGS  12
#define IFONTSIZE  13
#define IFONT      14
#define NOPTIONS   IFONT /* number of retrievable options */
/* the remainder are not retrievable */
#define IMISSING   15 /*(NOPTIONS+1)*/
#define IGUBED     16 /*(NOPTIONS+2)*/
#ifdef MACINTOSH
#define IPEELS     17 /*(NOPTIONS+3)*/
#endif /*MACINTOSH*/

typedef struct optionInfo
{
	char   *name;
	short  length;
	short  code;
} optionInfo;

/* option names must not exceed 10 characters and in same order as codes */
static optionInfo   Options[] = 
{
	"seeds",     4, ISEEDS, /* random number seeds (cat(posInt1,PosInt2)) */
	"nsig",      4, INSIG,  /* sets default format to %{nsig}+7).{nsig}g */
	"format",    6, IFORMAT,/* default format and default format for print */
	"wformat",   7, IWFORMAT,/* default format for write */
	"angles",    5, IANGLES, /* value "radians","cycles", "degrees" */
	"lines",     4, ILINES,  /* lines of text on screen */
	"width",     5, IWIDTH,  /* number of characters per output line */
	"errors",    5, IERRORS, /* limit on number of errors allowed */
	"prompt",    6, IPROMPT, /* default prompt */
	"batchecho", 9, IBATCHECHO, /* should commands be echoed in batch mode */
	"update",    6, IUPDATE, /*enables active updating of screen */
	"warnings",  7, IWARNINGS,  /* should WARNING: be printed */
	"fontsize",  7, IFONTSIZE,  /*"fontName Size"*/
	"font",      4, IFONT,  /*"fontName Size"*/
	"missing",   7, -IMISSING,  /*internal representation for MISSING*/
	"GUBED",     5, -IGUBED,    /* DEBUG backwards, for debugging */
#ifdef MACINTOSH
	"PEELS",     5, -IPEELS,    /* SLEEP backwards for testing waitNextEvent */
#endif /*MACINTOSH*/
	(char *) NULL, 0, 0
};

Symbolhandle          getOpt();
void                  setOpt();

/*
  functions to set and retrieve various options using keywords
  usage:
  	setoptions(option1,...,optionk)
  where optionj is of the form keywordj:valuej, e.g., 'nsig:5'

    setoptions(str)
	where str is structure whose component names are option names

	getoptions() or getoptions(all:T)
	  returns a structure with all current values of options

	getoptions(keyword1:T [,keyword2:T...])
	  returns the specified options.  If there are more than one specified,
	  a structure is returned.

	getoptions(all:T, keyword1:F, ...)
	  returns all except the indicated options

  For backward compatibility, for getoptions, you can use, for example,
  "nsig" instead of nsig:T

  In addition to the public options (id numbers > 0) there are non-public
  options that may be set but not retrieved.  Chief among these is GUBED
  (DEBUG backwards) which is used to control various features useful in
  debugging
*/

Symbolhandle    setoptions(list)
Symbolhandle     list;
{
	Symbolhandle     option, arg;
	long             isStructure = 0, nOptions = 0;
	long             set, i, j, nargs = NARGS(list), ncomps = nargs;
	long             optionNo, getit;
	short            whichOptions[NOPTIONS];
	char            *keyword;
	WHERE("setoptions");

	OUTSTR[0] = '\0';
	set = (strcmp(FUNCNAME,"setoptions") == 0);

	if(nargs == 1)
	{
		arg = COMPVALUE(list,0);
		if(!set && arg == (Symbolhandle) 0)
		{/* empty argument list to getoptions(); get everything */
			option = getOpt((short *) 0);
			return (option);
		}
		else if(set && arg != (Symbolhandle) 0 && TYPE(arg) == STRUC)
		{ /* single STRUCT argument to setoptions() */
			isStructure = 1;
			list = arg;
			ncomps = NCOMPS(list);
		}
	} /*if(nargs == 1)*/

	if(!set)
	{
		for(i=0;i<NOPTIONS;i++)
		{
			whichOptions[i] = 0;
		}
	}
	
	for(i=0;i<ncomps;i++)
	{
		arg = COMPVALUE(list,i);
		if(arg == (Symbolhandle) 0)
		{
			sprintf(OUTSTR,"ERROR: missing argument to %s",FUNCNAME);
		}
		else if(isStructure || (keyword = isKeyword(arg)) != (char *) 0)
		{
			if(isStructure)
			{
				keyword = NAME(arg);
			} /*if(isStructure)*/
			else
			{
				if(!set && strcmp(keyword, "all") == 0)
				{
					if(!isScalar(arg) || TYPE(arg) != LOGIC ||
					   DATAVALUE(arg, 0) == 0.0)
					{
						sprintf(OUTSTR,
								"ERROR: %s must have value T", keyword);
						goto errorExit;
					}
					
					for(j=0;j<NOPTIONS;j++)
					{ /* set everything T*/
						whichOptions[j] = 1;
					}
					nOptions = NOPTIONS;
					continue;
				} /*if(!set && strcmp(keyword, "all") == 0)*/
			} /*if(isStructure){}else{}*/
			
			if(!(optionNo = isOption(keyword)) || !set && optionNo < 0)
			{
				sprintf(OUTSTR,"ERROR: %s is unrecognized %s %s", keyword,
						(isStructure) ? "component name in argument to" : 
							"keyword on",
						FUNCNAME);
			}
			else if(!set)
			{
				keyword = Options[optionNo-1].name;
				if(!isScalar(arg) || TYPE(arg) != LOGIC)
				{
					notTorF(keyword);
					goto errorExit;
				}

				if(whichOptions[optionNo - 1])
				{
					nOptions--;
				}
				getit = (DATAVALUE(arg, 0) != 0.0);
				whichOptions[optionNo - 1] = getit;
				if(getit)
				{
					nOptions++;
				}
			}			
		}
		else if(set)
		{
			sprintf(OUTSTR,
					"ERROR: %ld-th argument to %s not keyword phrase",
					i+1, FUNCNAME);
		}
		else
		{ /*must be CHAR arg to getoptions()*/
			if(TYPE(arg) != CHAR || !isScalar(arg) ||
				(optionNo = isOption(STRINGPTR(arg))) <= 0)
			{ /* can get values of public options only */
				sprintf(OUTSTR,
						"ERROR: non-keyword argument to %s must be CHAR option name",
						FUNCNAME);
			}
			else
			{
				whichOptions[optionNo - 1] = 1;
				nOptions++;
			}
		}
		
		if(*OUTSTR)
		{
			goto errorExit;
		}
	} /*for(i=0;i<ncomps;i++)*/

	if(!set && nOptions == 0)
	{
		sprintf(OUTSTR,
				"ERROR: no options specified for %s", FUNCNAME);
		goto errorExit;
	}

	if(!set)
	{ /* getoptions */
		option = getOpt(whichOptions);
	} /*if(!set)*/
	else
	{ /* setoptions */
		setOpt(list, 0);
		option = NULLSYMBOL;
	} /*if(!set){...}else{...} */
	
	return (option);

  errorExit:
	putOUTSTR();
	
	return (0);
	
} /*setoptions()*/

#define TEMPSTRINGSIZE 100
Symbolhandle getOpt(optionlist)
short     *optionlist;
{
	Symbolhandle       optionHandle = (Symbolhandle) 0;
	Symbolhandle       symh = (Symbolhandle) 0;
	char              *optionName;
	char               tempstring[TEMPSTRINGSIZE+1];
	long               i, jcomp, all = (optionlist == (short *) 0);
	long               nOptions, forgeSymbol = 1;
	long               option;
	WHERE("getOpt");

	if(!all)
	{
		nOptions = 0;
		for(i = 0;i<NOPTIONS;i++)
		{
#ifndef CANSETFONTS
			option = abs((int) Options[i].code);
			if (option == IFONT || option == IFONTSIZE)
			{
				continue;
			}
#endif /*CANSETFONTS*/
			if(optionlist[i])
			{
				nOptions++;
			}
		}
	} /*if(!all)*/
	else
	{
#ifdef CANSETFONTS
		nOptions = NOPTIONS;
#else /*CANSETFONTS*/
		nOptions = NOPTIONS - 2;
#endif /*CANSETFONTS*/
	} /*if(!all){}else{}*/

	if(nOptions > 1)
	{ /* create structure for output */
		optionHandle = StrucInstall(SCRATCH,nOptions);
		if(optionHandle == (Symbolhandle) 0)
		{
			goto errorExit;;
		}
	} /*if(nOptions > 1)*/
	else if(nOptions == 1)
	{ /* create SCRATCH variable for output */
		symh = Install(SCRATCH,REAL); /* may be CHAR or LOGIC */
		if(symh == (Symbolhandle) 0)
		{
			goto errorExit;;
		}
		setNDIMS(symh,1);
		setDIM(symh,1,1);
	} /*if(nOptions > 1){...}else if(nOptions == 1){...} */
	else
	{
		symh = NULLSYMBOL;
	}
	
	jcomp = 0;
	
	for(i=0;i<NOPTIONS;i++)
	{
		OUTSTR[0] = '\0';
		if(all || optionlist[i])
		{
			optionName = Options[i].name;
			option = abs((int) Options[i].code);
#ifndef CANSETFONTS
			forgeSymbol = (option != IFONT && option != IFONTSIZE);
#endif /*CANSETFONTS*/
			if(nOptions > 1 && forgeSymbol)
			{
				symh = COMPVALUE(optionHandle,jcomp++) = Makesymbol(REAL);
				/* Note symh may end up CHAR or LOGIC so don't use Makereal*/
				if(symh == (Symbolhandle) 0)
				{
					goto errorExit;
				}
				setNDIMS(symh,1);
				if(nOptions > 1)
				{
					setNAME(symh,optionName);
				}
			} /*if(nOptions > 1 && forgeSymbol)*/

			switch (option)
			{
			  case ISEEDS:
				TMPHANDLE = mygethandle(2 * sizeof(double));
				setDATA(symh,(double **) TMPHANDLE);
				if(TMPHANDLE == (char **) NULL)
				{
					goto errorExit;
				}
				setTYPE(symh,REAL);
				setDIM(symh,1,2);
				DATAVALUE(symh,0) = (double) RANDS1;
				DATAVALUE(symh,1) = (double) RANDS2;
				break;

				/* single double value */
			  case INSIG:
			  case IMISSING:
				/* single LOGICAL values */
			  case IBATCHECHO:
			  case IUPDATE:
			  case IWARNINGS:
				TMPHANDLE = mygethandle(sizeof(double));
				setDATA(symh,(double **) TMPHANDLE);
				if(TMPHANDLE == (char **) NULL)
				{
					goto errorExit;
				}	
				setDIM(symh,1,1);
				setTYPE(symh,(option == INSIG || option == IMISSING) ?
						REAL : LOGIC);
				if(option == INSIG)
				{
					DATAVALUE(symh,0) = (double) PRINTFORMAT[1];
				}
				else if(option == IBATCHECHO)
				{
					DATAVALUE(symh,0) = (double) DEFAULTBATCHECHO;
				}
				else if(option == IUPDATE)
				{
					DATAVALUE(symh,0) = (double) ActiveUpdate;
				}
				else if(option == IWARNINGS)
				{
					DATAVALUE(symh,0) = (double) PRINTWARNINGS;
				}
				else
				{
					setMissing(DATAVALUE(symh,0));
				}
				break;
				
				/* single string values */
			  case IFORMAT:
			  case IWFORMAT:
			  case IANGLES:
			  case IPROMPT:
			  case IFONT:
				if(option == IFORMAT)
				{
					sprintf(tempstring,"%ld.%ld%c",
							PRINTFORMAT[0],PRINTFORMAT[1],(char) PRINTFORMAT[2]);
				}
				else if(option == IWFORMAT)
				{
					sprintf(tempstring,"%ld.%ld%c",
							WRITEFORMAT[0],WRITEFORMAT[1],(char) WRITEFORMAT[2]);
				}
				else if(option == IANGLES)
				{
					double angles2cycles = 8.0*atan(1.0);
					double angles2degrees = atan(1.0)/45.0;
					if(ANGLES2RADIANS == 1.0)
					{
						strcpy(tempstring,"radians");
					}
					else if(ANGLES2RADIANS == angles2cycles)
					{
						strcpy(tempstring,"cycles");
					}
					else if(ANGLES2RADIANS == angles2degrees)
					{
						strcpy(tempstring,"degrees");
					}
					else		/* should not happen */
					{
						strcpy(tempstring,"radians");
						ANGLES2RADIANS = 1.0;
					}
				}
				else if(option == IPROMPT)
				{
					strncpy(tempstring,PROMPT,TEMPSTRINGSIZE);
				}
				else if(option == IFONT)
				{
#ifdef CANSETFONTS
#ifdef MACINTOSH
					str255cpy((Str255) tempstring, CmdFontName);
					PtoCstr(tempstring);
#endif /*MACINTOSH*/
#else /*CANSETFONTS*/
					if (!all)
					{
						sprintf(OUTSTR,
								"WARNING: option %s not available in this version",
								optionName);
					}					
#endif /*CANSETFONTS*/
				}
				tempstring[TEMPSTRINGSIZE] = '\0';
				if (forgeSymbol)
				{
					TMPHANDLE = mygethandle((strlen(tempstring)+1)*sizeof(char));
					setSTRING(symh,TMPHANDLE);
					if(TMPHANDLE == (char **) NULL)	
					{
						goto errorExit;
					}	
					setTYPE(symh,CHAR);
					setDIM(symh,1,1);
					strcpy(STRINGPTR(symh),tempstring);
				} /*if (forgeSymbol)*/
				break;

				/* single integer values */
			  case ILINES:
			  case IWIDTH:
			  case IERRORS:
			  case IFONTSIZE:
			  case IGUBED:
#ifdef MACINTOSH
			  case IPEELS:
#endif /*MACINTOSH*/
				if (forgeSymbol)
				{
					TMPHANDLE = mygethandle(sizeof(double));
					setDATA(symh,(double **) TMPHANDLE);
					if(TMPHANDLE == (char **) NULL)
					{
						goto errorExit;
					}	
					setTYPE(symh,REAL);
					setDIM(symh,1,1);
				} /*if (forgeSymbol)*/
				
				if(option == ILINES)
				{
					DATAVALUE(symh,0) = (double) SCREENHEIGHT;
				}
				else if(option == IWIDTH)
				{
					DATAVALUE(symh,0) = (double) SCREENWIDTH;
				}
				else if(option == IERRORS)
				{
					DATAVALUE(symh,0) = (double) MAXERRORS;
				}
				else if(option == IFONTSIZE)
				{
#ifdef CANSETFONTS
#ifdef MACINTOSH
					DATAVALUE(symh,0) = (double) CmdFontSize;
#endif /*MACINTOSH*/
#else /*CANSETFONTS*/
					if (!all)
					{
						sprintf(OUTSTR,
								"WARNING: option %s not available in this version",
								optionName);
					}					
#endif /*CANSETFONTS*/
				}
				else if(option == IGUBED)
				{
					DATAVALUE(symh,0) = (double) GUBED;
				}
#ifdef MACINTOSH
				else if(option == IPEELS)
				{
					DATAVALUE(symh,0) = (double) CmdSleep+100*BatchSleep+
						10000*InterruptSleep;
				}
#endif /*MACINTOSH*/
				break;

			  default:
				/* should not get here */
				sprintf(OUTSTR,"WARNING: unrecognized option '%s'",optionName);
			} /*switch (option)*/
		} /*if(all || optionlist[i])*/
		putOUTSTR();
	} /*for(i=0;i<NOPTIONS;i++)*/

	return ((nOptions > 1) ? optionHandle : symh);

  errorExit:
	
	if (nOptions > 1)
	{
		Removesymbol(optionHandle);
	}
	else
	{
		Removesymbol(symh);
	}
	
	return  (0);
} /*getOpt()*/

/*
  Function to do the actual work of setting options, called from setoptions()
  or restore().  If called from setoptions() list should either be an
  argument list (type LIST) of keyword arguments, in which case their names
  have already been checked for validity, or a structure (type STRUC), the
  names of whose components are option names which have also been checked
  for validity.  If called from restore(), list is a structure whose names
  have not been checked.  In any case, if an unrecognized name is found and
  quiet == 0, a warning message is printed. 
*/  

#ifdef CANSETFONTS
#define DEFAULTFONT "default"

#ifdef MACINTOSH
#define FONTNAMESIZE  27
#define MINFONTSIZE    9  
#define MAXFONTSIZE   64
#endif /*MACINTOSH*/

#define OKFONTSPEC       0
#define BADFONTSIZE     -1
#define FONTNAMETOOLONG -2
#define BADFONTSPEC     -3

/*
	Function to decode a font of the form "FontName size"
	If font is of form "FontName", set *fontsize to -1
*/
static int decodeFont(char *font, char **fontName, long *fontSize)
{
	static char     name[FONTNAMESIZE+1];
	long            length = strlen(font);
	char           *sizePtr = font + length - 1;
	WHERE("decodeFont");
	
	*fontSize = -1;
	while(sizePtr > font && *sizePtr == ' ')
	{ /* skip backwara over trailing spaces */
		length--;
		sizePtr--;
	}
	if (isdigit(*sizePtr))
	{ /* may be of form "Name Size" */
		while(sizePtr > font && isdigit(*sizePtr))
		{
			sizePtr--;
		}
		if (*sizePtr == ' ' || *sizePtr == '-' || *sizePtr == '+')
		{
			if (*sizePtr == ' ')
			{
				sizePtr++;
			}
			/* sizePtr should point to number to be decoded */
			if (*(sizePtr-1) != ' ')
			{ /* not of form "Name Number" */
				*fontSize = -1;
				length = strlen(font);
			}
			else
			{
				if (sscanf(sizePtr, "%ld", fontSize) != 1 ||
					*fontSize < MINFONTSIZE || *fontSize > MAXFONTSIZE)
				{
					return (BADFONTSIZE);
				}
				sizePtr--;
				while(sizePtr > font && *sizePtr == ' ')
				{
					sizePtr--;
				}
				sizePtr++;
				length = sizePtr - font;
			}
		} /*if (*sizePtr == ' ' || *sizePtr == '-' || *sizePtr == '+')*/
	} /*if (isdigit(*sizePtr))*/
	if (length > FONTNAMESIZE)
	{
		return (FONTNAMETOOLONG);
	}
	if (length == 0)
	{
		return (BADFONTSPEC);
	}
	strncpy(name, font, length);
	name[length] = '\0';
	*fontName = name;

	return (OKFONTSPEC);
} /*decodeFont()*/

static void setNewFont(char *newFontName, short newFontSize)
{
#ifdef MACINTOSH
	Str27      fontName;
	Integer    fontSize = newFontSize;
	Integer    fontNumber = -1;
	WHERE("setNewFont");
	
	if (strcmp(newFontName, DEFAULTFONT) == 0 ||
		strncmp(newFontName, "orig", 4) == 0)
	{
		str255cpy(fontName, MACANOVAFONTNAME);
	}
	else
	{
		strcpy((char *) fontName, newFontName);
		CtoPstr((char *) fontName);
	}

	if (fontSize < 0)
	{
		fontSize = CmdFontSize;
	}
	if (fontSize != CmdFontSize || str255cmp(fontName, CmdFontName) != 0)
	{ /* font size or name has been changed */
		GetFNum(fontName, &fontNumber);
		if (fontNumber == 0 && str255cmp(fontName, "\pChicago") != 0)
		{
			PtoCstr((char *) fontName);
			sprintf(OUTSTR,
					"WARNING: font \"%s\" is not available; not changed",
					fontName);
		}
		else
		{
			setCmdWindFont(fontNumber,fontSize);
		}
	}
	else
	{ /* reset SCREENWIDTH and SCREENHEIGHT even if no font change */
		setCmdScreen();
	}
	putOUTSTR();
#endif /*MACINTOSH*/
} /*setNewFont()*/
#endif /*CANSETFONTS*/

void           setOpt(list, quiet)
Symbolhandle   list;
long           quiet;
{
	long            icomp, isleep, k, ncomps, length;
	double          value[3];
	long            isStructure;
	long            fmt[3];
	long            option, logValue;
	Symbolhandle    arg;
	char           *keyword;
#ifdef CANSETFONTS
	long            newFontSize = -1, fontSize;
	char           *newFontName = (char *) 0;
	char            fontName[MAXFONTSIZE+1];
	int             reply;
#endif /*CANSETFONTS*/
	WHERE("setOpt");
	
/* is should be the case that all arguments should be valid option names */
	isStructure = (TYPE(list) == STRUC);
	ncomps = NCOMPS(list);
	for(icomp=0;icomp<ncomps;icomp++)
	{
		OUTSTR[0] = '\0';
		arg = COMPVALUE(list,icomp);
		length = symbolSize(arg);
		if(isStructure)
		{
			keyword = NAME(arg);
		}
		else
		{
			keyword = isKeyword(arg);
			/* should not need to check that non-null */
		}
		option = labs(isOption(keyword));
		keyword = Options[option-1].name;
		switch (option)
		{
		  case ISEEDS:
			if(TYPE(arg) != REAL || length != 2)
			{
				sprintf(OUTSTR,
						"WARNING: %s must be vector of two integer values",
						keyword);
			}
			else if((value[0] = DATAVALUE(arg,0)) < 0 || 
					(value[1] = DATAVALUE(arg,1)) < 0 ||
					value[0] > MAXSEED ||
					value[1] > MAXSEED ||
					value[0] != floor(value[0]) ||
					value[1] != floor(value[1]))
			{
				sprintf(OUTSTR,
						"WARNING: %s must be positive integers <= %ld",
						keyword,MAXSEED);
			}
			else
			{
				RANDS1 = (long) value[0];
				RANDS2 = (long) value[1];
			}
			break;
			
		  case INSIG:
		  case IMISSING:
			if(TYPE(arg) != REAL || length != 1)
			{
				sprintf(OUTSTR,"WARNING: %s must be single %s value",
						keyword,
						(option == INSIG) ? "integer" : "real");
			}
			else
			{
				value[0] = DATAVALUE(arg,0);
				if(option == INSIG)
				{
					if(value[0] < 0 ||value[0] > 20 ||
					   value[0] != floor(value[0]))
					{
						sprintf(OUTSTR,
								"WARNING: %s must be integer between 0 and 20",
								keyword);
					}
					else
					{
						PRINTFORMAT[0] = (long) value[0] + 7;
						PRINTFORMAT[1] = (long) value[0];
						PRINTFORMAT[2] = (long) 'g';
						installFormat(PRINTFORMAT[0],PRINTFORMAT[1],
									  PRINTFORMAT[2]);
						saveFormat(); /* save copy */
					}
				}
				else
				{
					MISSING = value[0];
				}
			}
			break;

		  case IBATCHECHO:
		  case IUPDATE:
		  case IWARNINGS:
			if(TYPE(arg) != LOGIC || length != 1)
			{
				sprintf(OUTSTR,"WARNING: %s must be T or F",
						keyword);
			}
			else
			{
				logValue = DATAVALUE(arg,0) != 0.0;
				if(option == IBATCHECHO)
				{
					DEFAULTBATCHECHO = logValue;
					if(BDEPTH > 0)
					{
						BATCHECHO[BDEPTH] = DEFAULTBATCHECHO;
					}
				}
				else if(option == IUPDATE)
				{
					ActiveUpdate = logValue;
#ifndef ACTIVEUPDATE
					sprintf(OUTSTR,
							"WARNING: option 'update' has no effect in this version");
#endif /*ACTIVEUPDATE*/
				}
				else
				{
					PRINTWARNINGS = logValue;
				}
			}
			
			break;
			
		  case IFORMAT:
		  case IWFORMAT:
			if(TYPE(arg) != CHAR || length != 1)
			{
				sprintf(OUTSTR,"WARNING: %s must be single CHARACTER value",
						keyword);
			}
			else if(!setFormat(STRINGPTR(arg),fmt))
			{
				sprintf(OUTSTR,"WARNING: '%s' is not proper value for %s",
						STRINGPTR(arg),keyword);
			}
			else
			{
				for(k = 0;k<3;k++)
				{
					if(option == IFORMAT)
					{
						PRINTFORMAT[k] = fmt[k];
					}
					else
					{
						WRITEFORMAT[k] = fmt[k];
					}
				}
				if(option == IFORMAT)
				{
					installFormat(PRINTFORMAT[0],PRINTFORMAT[1],
								  PRINTFORMAT[2]);
					saveFormat(); /* save copy */
				}
			}
			break;

		  case IANGLES:
			if(TYPE(arg) != CHAR && length != 1)
			{
				sprintf(OUTSTR,"WARNING: %s must be single CHARACTER variable",
						keyword);
			}
			else if(strncmp(STRINGPTR(arg),"radian",6) == 0)
			{
				ANGLES2RADIANS = 1.0;
			}
			else if(strncmp(STRINGPTR(arg),"cycle",5) == 0)
			{
				ANGLES2RADIANS = 8.0*atan(1.0);
			}
			else if(strncmp(STRINGPTR(arg),"degree",6) == 0)
			{
				ANGLES2RADIANS = atan(1.0)/45.0;
			}
			else
			{
				sprintf(OUTSTR,	"WARNING: %s must have value \"radians\", \"cycles\", or \"degrees\"",
						keyword);
			}
			break;

		  case IFONT:
			if(TYPE(arg) != CHAR || length != 1)
			{
				sprintf(OUTSTR,"WARNING: %s must be single CHARACTER variable",
						keyword);
			}
			else
			{
#ifdef CANSETFONTS
				reply = decodeFont(STRINGPTR(arg), &newFontName, &fontSize);
				if (reply == BADFONTSIZE)
				{
					sprintf(OUTSTR,
							"WARNING: %s specifies a bad font size",
							STRINGPTR(arg));
				}
				else if (reply == FONTNAMETOOLONG)
				{
					sprintf(OUTSTR,
							"WARNING: %s specifies too long a font name",
							STRINGPTR(arg));
				}
				else if (reply == BADFONTSPEC)
				{
					sprintf(OUTSTR,
							"WARNING: '%s' not a valid font specification",
							STRINGPTR(arg));
				}
				if (fontSize > 0)
				{
					newFontSize = fontSize;
				}
#else /*CANSETFONTS*/
				sprintf(OUTSTR,
						"WARNING: %s not valid option on this version of MacAnova",
						keyword);
#endif /*CANSETFONTS*/
			}
			break;
			
		  case IPROMPT:
			if(TYPE(arg) != CHAR && length != 1)
			{
				sprintf(OUTSTR,"WARNING: %s must be single CHARACTER variable",
						keyword);
			}
			else if(strlen(STRINGPTR(arg)) > sizeof(PROMPT) - 1)
			{
				sprintf(OUTSTR,
						"WARNING: %s must have no more than %ld characters",
						keyword, (long) sizeof(PROMPT) - 1);
			}
			else
			{
				strcpy(PROMPT,STRINGPTR(arg));
			}
			break;

		  case ILINES:
		  case IWIDTH:
		  case IERRORS:
		  case IFONTSIZE:
		  case IGUBED:
#ifdef MACINTOSH
		  case IPEELS:
#endif /*MACINTOSH*/
			value[0] = DATAVALUE(arg,0);
			if(TYPE(arg) != REAL || length != 1 || value[0] != floor(value[0]))
			{
				sprintf(OUTSTR,"WARNING: %s must be single integer value",
						keyword);
			}
			else if(option == ILINES)
			{
				if(value[0] < 0)
				{
					sprintf(OUTSTR,	"WARNING: %s must be a positive integer",
							keyword);
				}
				else
				{
					SCREENHEIGHT = (long) value[0];
#ifdef ACTIVEUPDATE
					activeUpdateInit();
#endif /*ACTIVEUPDATE*/
				}
			}
			else if(option == IWIDTH)
			{
				if(value[0] < 20)
				{
					sprintf(OUTSTR,	"WARNING: %s must be integer >= 20",
							keyword);
				}
				else
				{
					SCREENWIDTH = (long) value[0];
				}
			}
			else if(option == IERRORS)
			{
				if(value[0] < 0)
				{
					sprintf(OUTSTR,	"WARNING: %s must be integer >= 0",
							keyword);
				}
				else
				{
					if(BDEPTH == 0)
					{
						MAXERRORS1 = MAXERRORS = (long) value[0];
					}
					else
					{
						MAXERRORS1 = (long) value[0];
					}
				}
			}
			else if (option == IFONTSIZE)
			{
#ifdef CANSETFONTS
				newFontSize = (long) value[0];
				if (newFontSize < MINFONTSIZE || newFontSize > MAXFONTSIZE)
				{
					sprintf(OUTSTR,
							"WARNING: illegal font size %ld", newFontSize);
				}
				else
				{
					if (newFontName == (char *) 0)
					{
						str255cpy((Str255) fontName, CmdFontName);
						PtoCstr(fontName);
						newFontName = fontName;
					}
				}
#else /*CANSETFONTS*/
				sprintf(OUTSTR,
						"WARNING: %s not valid option on this version of MacAnova",
						keyword);
#endif /*CANSETFONTS*/
			}
			else if(option == IGUBED)
			{
				GUBED = (long) value[0];
			}

#ifdef MACINTOSH
			else if(option == IPEELS && value[0] >= 0)
			{
				/* CmdSleep + 100 * BatchSleep + 10000 * InterruptSleep */
				isleep = (long) value[0];
				CmdSleep = isleep % 100;
				isleep /= 100;
				BatchSleep = isleep % 100;
				InterruptSleep = isleep / 100;
			}
#endif /*MACINTOSH*/
			break;

		  default:
			/* should not happen */
			sprintf(OUTSTR,"WARNING: unrecognized option '%s'",keyword);
			
		} /*switch (option)*/

		if(*OUTSTR)
		{
			if(!quiet)
			{
				sprintf(OUTSTR + strlen(OUTSTR),"; %s not set",keyword);
				putOUTSTR();
			}
			else
			{
				*OUTSTR = '\0';
			}
		}
	} /*for(icomp=0;icomp<ncomps;icomp++)*/

#ifdef CANSETFONTS
	if (newFontName != (char *) 0)
	{
		setNewFont(newFontName, newFontSize);
	}
#endif /*CANSETFONTS*/
} /*setOpt()*/

/*
  Check keyword name to see if it is valid option name
*/

long isOption(char *keyword)
{
	optionInfo    *option;
	
	for(option = Options;option->length;option++)
	{
		if(strncmp(option->name, keyword, (int) option->length) == 0)
		{
			return (option->code);
		}
	} /*for(option = Options;option->length;option++)*/
	return (0);
} /*isOption()*/

