/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Columnop
#else
#define __SEG__ Columnop
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

extern Symbolhandle      RInstall();
extern long              dupDims();
extern long              argOK();
extern long              symbolSize();

typedef struct opEntry
{
	char *name;
	short iop;
	short nargs;
} opEntry;

#define          IISMISSING   0x01
#define          IANYMISSING  0x02
#define          IRANKITS     0x03
#define          IHALFNORM    0x04
#define          IRANK        0x05
#define          IGRADE       0x06
#define          ISORT        0x07

#define          UP           0x08
#define          GRADEONLY    0x00
#define          TIESIGNORE   0x10
#define          TIESMEAN     0x20
#define          TIESMIN      0x30

#define          SORTOPMASK   0x07
#define          TIEMASK      0x30

static opEntry Ops[] =
{ /* name        iop      nargs */
	"ismissing",  IISMISSING,     1, /* ismissing(x) */
	"anymissing", IANYMISSING,    1, /* anymissing(x) */
	"rankits",    IRANKITS,       1, /* rankits(x) */
	"halfnorm",   IHALFNORM,      1, /* halfnorm(x) */
	"rank",       IRANK,         -3, /* rank(x [,down:T] [,ties:"i|a|m"]) */
	"grade",      IGRADE,        -2, /* grade(x [,down:T]) */
	"sort",       ISORT,         -2, /* sort(x [,down:T) */
	(char *)0,   0,              0
};

#define GX2    0
#define GRANKS 1
#define GSCR   2
#define NTRASH 3

static Symbolhandle doOps(Symbolhandle arg1, double *params,
						  unsigned long control, unsigned long *status)
{
	Symbolhandle    result = (Symbolhandle) 0;
	double        **x2H = (double **) 0;
	double        **ranksH = (double **) 0;
	double        **scrH = (double **) 0;
	double         *x, *x2, *scr, *ranks, *answer;
	double          y, missVal = HUGEDBL;
	double          denom;
	long            i, ierr, n, n2, ndims;
	long            size, j;
	long            iop = control & SORTOPMASK;
	long            up = control & UP;
	long            rankOp = control & TIEMASK;
	long            ncols;
	long            ischar = (TYPE(arg1) == CHAR);
	char           *string;
	WHERE("doOps");
	TRASH(NTRASH,errorExit); /* allocate trash basket for scratch */

	size = symbolSize(arg1);

	if (control & REUSELEFT)
	{
		result = arg1;
	}
	else
	{
		result = RInstall(SCRATCH,(iop == IANYMISSING) ? 1 : size);
		if (result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		if (iop == IANYMISSING)
		{
			setTYPE(result, LOGIC);
			DATAVALUE(result, 0) = 0.0;
		}
	}

	setNCLASS(result, -1);

	if (size > 0)
	{
		if (iop != IANYMISSING)
		{
			ndims = NDIMS(arg1);
			n = DIMVAL(arg1,1);
			ncols = size/n;
			if (!(control & REUSELEFT))
			{
				setNDIMS(result,ndims);
				for (i = 1; i <= ndims; i++)
				{
					setDIM(result,i,DIMVAL(arg1,i));
				}
			}		
		} /*if (iop != IANYMISSING)*/
	
		if (iop == IISMISSING)
		{
			setTYPE(result,LOGIC);

			answer = DATAPTR(result);
			if (!ischar)
			{
				x = DATAPTR(arg1);
				for (i = 0; i < size; i++)
				{
					answer[i] = (isMissing(x[i])) ? 1.0 : 0.0;
				}
			}
			else
			{
				string = STRINGPTR(arg1);
				for (i = 0; i < size; i++)
				{
					answer[i] = (*string == '\0') ? 1.0 : 0.0;
					string += strlen(string) + 1;
				}
			}
		} /*if (iop == IISMISSING)*/
		else if (iop == IANYMISSING)
		{
			DATAVALUE(result,0) = (anyMissing(arg1)) ? 1.0 : 0.0;
		}
		else
		{						/* IRANKITS, IHALFNORM, IRANK, IGRADE, ISORT */
			if (iop != ISORT)
			{
				if (!getScratch(x2H,GX2,n,double) ||
					!getScratch(scrH,GSCR,n,double) ||
					!getScratch(ranksH,GRANKS,n,double))
				{
					goto errorExit;
				}
				x2 = *x2H;
				scr = *scrH;
				ranks = *ranksH;
			} /*if (iop != ISORT)*/
			x = DATAPTR(arg1);	/* if reusing arg1, x and answer may be the same */
			answer = DATAPTR(result);
		
			/*
			   Treatment of MISSING:
			   For rankits() and halfnorm(), the values computed use the
			   number of non-missing values as sample size, with MISSING
			   whereever the input is MISSING.
			   
			   For rank(), the ranks(), the rank of a MISSING value is set
			   to MISSING, the maximum rank of non-missing values being the
			   number of such values, regardless of the direction of the
			   sort.
			   
			   For sort(), MISSING values are always sorted to the end,
			   regardless of the direction of the sort.
			   
			   For grade(), if there are k MISSING values, the last k elements
			   in a column of the result are the row numbers of the MISSING
			   values; the first nrows-k values are the row numbers of the
			   order statistics of the non-missing values.  Thus if x is a
			   vector, x[grade(x,down)] should be the same as sort(x,down)
			   whether down is T or F.
			   */
			for (j = 0; j < ncols;j++)
			{
				if (iop == ISORT)
				{
					x2 = answer;
				}
			
				n2 = 0;
				for (i = 0; i < n; i++)
				{				/* transform MISSING to HUGEDBL */
					y = x[i];
					if (isMissing(y))
					{			/* should always sort to the end */
						y = missVal;
						*status |= FOUNDMISSING;
					}
					else
					{
						n2++;	/* number of non-missing values */
						y = (up) ? y : -y;
					}
					x2[i] = y;
				}
			
				switch (iop)
				{
				  case IRANKITS:
					rankquick(x2,ranks,scr,n,rankOp);
					denom = (double) n2 + 0.25;
					for (i = 0; i < n; i++)
					{
						if (isMissing(x[i]))
						{
							setMissing(answer[i]);
						}
						else
						{
							answer[i] = ppnd((ranks[i] - 0.375) / denom, &ierr);
						}
					} /*for (i = 0; i < n; i++)*/
					break;

				  case IHALFNORM:
					for (i = 0; i < n; i++)
					{
						if (!isMissing(x[i]))
						{
							x2[i] = fabs(x2[i]);
						}
					} /*for (i = 0; i < n; i++)*/

					rankquick(x2, ranks, scr, n, rankOp);

					denom  = 2.0*n2 + 0.5;
					for (i = 0; i < n; i++)
					{
						if (isMissing(x[i]))
						{
							setMissing(answer[i]);
						}
						else
						{
							answer[i] = ppnd(0.5 + (ranks[i] - 0.375) / denom,
											 &ierr);
						}
					} /*for (i = 0; i < n; i++)*/
					break;

				  case IRANK:
				  case IGRADE:
					rankquick(x2, ranks, scr, n, rankOp);

					if (iop == IRANK)
					{
						for (i = 0; i < n; i++)
						{
							if (isMissing(x[i]))
							{
								setMissing(answer[i]);
							}
							else
							{
								answer[i] = ranks[i];
							}
						} /*for (i = 0; i < n; i++)*/
					} /*if (iop == IRANK)*/
					else
					{			/*IGRADE*/
						for (i=0;i<n;i++)
						{
							answer[i] = ranks[i];
						}
					} /*if (iop == IRANK){...}else{...}*/
				
					break;

				  case ISORT:
					sortquick(answer, n);

					if (!up)
					{
						for (i = 0; i < n2; i++)
						{
							answer[i] = -answer[i];
						}
					}
				
					for (i = n2; i < n; i++)
					{
						setMissing(answer[i]);
					}
					break;
				} /* switch (iop) */
				x += n;
				answer += n;
			} /*for (j = 0; j < ncols;j++)*/
		} /* if ... else */
	} /*if (size > 0)*/
	
	emptyTrash();
	
	return (result);
	
  errorExit:
	Removesymbol(result);
	emptyTrash();
	return (0);
}

Symbolhandle    sort(list)
Symbolhandle    list;
{

	Symbolhandle    arg1, symhkey, result;
	long            up = 1;
	long            iop, margs, nargs = NARGS(list);
	long            missingOp, rankOp;
	unsigned long   control = 0, status = 0;
	long            type, type1;
	long            jarg;
	char           *keyword, c;
	opEntry        *op;
	WHERE("sort");
	
	*OUTSTR = '\0';
	for (op = Ops; op->iop != 0 && strcmp(op->name,FUNCNAME) != 0; op++)
	{
		;
	}
	
	if (op->iop == 0)
	{ /* should never occur unless programming error */
		sprintf(OUTSTR,"ERROR: unrecognized operation %s",FUNCNAME);
		goto errorExit;
	}

	margs = op->nargs;
	iop = op->iop;
	if (iop == IGRADE)
	{
		rankOp = GRADEONLY;
	}
	else if (iop == IRANKITS || iop == IHALFNORM)
	{
		rankOp = TIESIGNORE;
	}
	else
	{
		rankOp = TIESMEAN;
	}
	
	missingOp = (iop == IANYMISSING || iop == IISMISSING);
	
	if (margs > 0 && nargs != margs || margs < 0 && nargs > -margs)
	{
		badNargs(FUNCNAME,margs);
		goto errorExit;
	}
	
	arg1 = COMPVALUE(list, 0);
	if (!argOK(arg1,0,0))
	{
		goto errorExit;
	}
	type = TYPE(arg1);
	
	if (type == STRUC)
	{
		type1 = getStrucTypes(arg1);
		if (!missingOp && type1 != REALTYPE1 ||
		   missingOp && type1 & ~(REALTYPE1 | LOGICTYPE1 | CHARTYPE1))
		{
			sprintf(OUTSTR,
					"ERROR: argument to %s has component that is not REAL%s",
					FUNCNAME,(missingOp) ? " LOGICAL or CHARACTER" : "");
			goto errorExit;
		}
	}
	else if (!missingOp && type != REAL)
	{
		badType(FUNCNAME,REAL,0);
		goto errorExit;
	}
	else if (missingOp && type != REAL && type != LOGIC && type != CHAR)
	{
		badType(FUNCNAME,-type,0);
		goto errorExit;
	}
	
	for (jarg = 1;jarg < nargs;jarg++)
	{ /* only here for sort(), rank(), and grade() */
		symhkey = COMPVALUE(list,jarg);
		if (!argOK(symhkey,0,jarg+1))
		{
			goto errorExit;
		}

		if ((keyword = isKeyword(symhkey)))
		{
			if (iop == IRANK && strcmp(keyword,"ties") == 0)
			{
				rankOp = -1;
				if (TYPE(symhkey) == CHAR && isScalar(symhkey))
				{
					c = STRINGVALUE(symhkey,0);
					if (c == 'i' || c == 'I')
					{			/* "ignore" */
						rankOp = TIESIGNORE;
					}
					else if (c == 'a' || c == 'A')
					{			/* "average" */
						rankOp = TIESMEAN;
					}
					else if (c == 'm' || c == 'M')
					{			/* "minimum" */
						rankOp = TIESMIN;
					}
				}
				if (rankOp < 0)
				{
					sprintf(OUTSTR,
							"ERROR: Value for '%s' must be \"ignore\", \"average\", or \"minimum\"",
							keyword);
					goto errorExit;
				}
			} /*if (strcmp(keyword,"ties") == 0)*/
			else if (strcmp(keyword,"down") == 0 || strcmp(keyword,"up") == 0)
			{
				if (!isScalar(symhkey) || TYPE(symhkey) != LOGIC)
				{
					sprintf(OUTSTR,"ERROR: value for %s must be T or F",
							keyword);
					goto errorExit;
				}
				up = (DATAVALUE(symhkey,0) != 0.0);
				if (strcmp(keyword,"down") == 0)
				{
					up = !up;
				}
			}
			else
			{
				badKeyword(FUNCNAME,keyword);
				goto errorExit;
			}
		} /*if ((keyword = isKeyword(symhkey)))*/
		else if (jarg == nargs-1)
		{
			if (TYPE(symhkey) == LOGIC && isScalar(symhkey))
			{
				up = (DATAVALUE(symhkey,0) == 0.0);
			}
			else
			{
				sprintf(OUTSTR,
						"ERROR: value for non-keyword final argument to %s must be T of F",
						FUNCNAME);
				goto errorExit;
			}
		}
		else
		{
			sprintf(OUTSTR,
					"ERROR: 2nd argument out of 3 not a keyword for %s",
					FUNCNAME);
			goto errorExit;
		}
	} /*for (jarg = 1;jarg < nargs;jarg++)*/
	
	control = iop | rankOp;

	if (up)
	{
		control |= UP;
	}
	if (isscratch(NAME(arg1)) &&
	   (iop == ISORT ||
		iop == IISMISSING &&
		(type == REAL || type == LOGIC ||
		 (type == STRUC && !(type1 & CHARTYPE1)))))
	{ /* only cases when we can reuse argument as result */
		control |= REUSELEFT;
		COMPVALUE(list, 0) = (Symbolhandle) 0;
	}
	
	result = doRecur1(doOps, arg1, (double *) 0,control, &status);
	if (status & FOUNDMISSING)
	{
		sprintf(OUTSTR, "WARNING:  MISSING values in argument to %s", FUNCNAME);
		putOUTSTR();
	}

	return (result);

  errorExit:
	putOUTSTR();

	return (0);
}
