/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Batch
#else
#define __SEG__ Batch
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

#ifdef MACINTOSH
#include "macIface.h"
#endif /*MACINTOSH*/

/*
  Routine to spool output to a file or suspend spooling
  Modified by kb to accept second logical argment:
  	spool(fileName [,new:T])
  If new:T is present, the file is opened with fopen(fileName,"r") instead of
  fopen(fileName,"a")
    spool()  toggles spooling on and off once started.
	spool(,new:T) restarts spooling on last specified file.

  SPOOLFILE should be non-zero if and only if spooling is in effect.
  SpoolFileName should be non-zero if and only if spooling is in effect
  or the last attempt to spool was successful.
*/

Symbolhandle    spool(list)
Symbolhandle    list;
{
	Symbolhandle    arg1, arg2;
	long            rewindit = 0;
	long            doSpool = 0;
	long            nargs = NARGS(list);
	char           *keyword;
	char           *fileName;
	WHERE("spool");
	
#ifdef MACINTOSH
	if (!UseWindows)
	{ /*  */
		sprintf(OUTSTR,"WARNING: %s() disabled in non-interactive mode",
				FUNCNAME);
		goto errorExit;
	}
#endif /*MACINTOSH*/
	
	*OUTSTR = '\0';

	if(nargs > 2)
	{
		badNargs(FUNCNAME,-2);
		goto errorExit;
	}

	arg1 = COMPVALUE(list,0);
	if(arg1 != (Symbolhandle) 0)
	{
		if (TYPE(arg1) != CHAR || !isScalar(arg1))
		{
			notCharOrString("filename");
			goto errorExit;
		}
		fileName = STRINGPTR(arg1);
	} /*if(arg1 != (Symbolhandle) 0)*/		
	
	/* 2 arguments */
	if(nargs == 2)
	{ /* check to see if we need to rewind */
		arg2 = COMPVALUE(list,1);
		if((keyword = isKeyword(arg2)) && strcmp(keyword,"new") != 0)
		{ /* only legal keyword is "new" */
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
		if(arg2 == (Symbolhandle) 0 || TYPE(arg2) != LOGIC ||
		   !isScalar(arg2))
		{
			notTorF(keyword);
			goto errorExit;
		}
		
		rewindit = (DATAVALUE(arg2,0) != 0.0);
	}
	
/* now see about file name */
	if (arg1 == (Symbolhandle) 0)
	{ /* none given */
		/* no argument, so stop or restart (new:T) spooling */
		if (SPOOLFILE != (FILE *) 0)
		{
			fclose(SPOOLFILE);
			SPOOLFILE = (FILE *) 0;
			if(rewindit)
			{
				sprintf(OUTSTR,"Restart spooling on %s", *SPOOLFILENAME);
				doSpool = 1;
			}
			else
			{
				sprintf(OUTSTR,"Spooling on %s suspended",*SPOOLFILENAME);
				doSpool = 0;
			}
		}
		else if(SPOOLFILENAME == (char **) 0)
		{
			strcpy(OUTSTR,"WARNING: not spooling at present");
			doSpool = 0;
		}
		else
		{
			sprintf(OUTSTR,"%s spooling on %s",
					(rewindit) ? "Restart" : "Resume", *SPOOLFILENAME);
			doSpool = 1;
		}
		putOUTSTR();
	} /*if (arg1 == (Symbolhandle) 0)*/
	else if(SPOOLFILENAME == (char **) 0 || SPOOLFILE == (FILE *) 0 ||
			strcmp(*SPOOLFILENAME,fileName) != 0 || rewindit)
	{ /* new filename provided */
		if(SPOOLFILENAME != (char **) 0)
		{
			mydisphandle(SPOOLFILENAME);
			SPOOLFILENAME = (char **) 0;
		}
#ifdef MACINTOSH
		if(SpoolVolume)
		{
			SetVol(0L,SpoolVolume);
		}

		if(strlen(fileName) == 0)
		{
			/*
			   null file name; ask user to choose
			   macOpen() default file type is is TEXT
			*/
			fileName = macOpen("\pSave spooled output as:", "\p",
							   WRITEIT, 0, (OSType *) 0);
			if (fileName == (char *) 0)
			{
				sprintf(OUTSTR,"WARNING: spooling request cancelled");
				goto errorExit;
			}
			SpoolVolume = Reply.vRefNum;
		} /*if(strlen(fileName) == 0)*/
#endif /*MACINTOSH*/

		if((fileName = expandFilename(fileName)) == (char *) 0)
		{
			goto errorExit;
		}

		if(!isfilename(fileName))
		{
			goto errorExit;
		}
		if((SPOOLFILENAME = mygethandle(strlen(fileName) + 1)) ==
		   (char **) 0)
		{
			goto errorExit;
		}
		strcpy(*SPOOLFILENAME,fileName);
		doSpool = 1;
	}
	else
	{
		sprintf(OUTSTR,"WARNING: already spooling on file %s", *SPOOLFILENAME);
		putOUTSTR();
		doSpool = 0;
	}
	
	if(doSpool)
	{
		if (SPOOLFILE != (FILE *) 0)
		{
			fclose(SPOOLFILE);
		}
		SPOOLFILE = fopen(*SPOOLFILENAME, (rewindit) ? "w" : "a");
		if (SPOOLFILE == (FILE *) 0)
		{
			sprintf(OUTSTR,"ERROR: cannot open file %s for spooling",
					*SPOOLFILENAME);
			putOUTSTR();
			mydisphandle(SPOOLFILENAME);
			SPOOLFILENAME = (char **) 0;
			goto errorExit;
		}
#ifdef MACINTOSH
		macSetInfo(SpoolVolume, *SPOOLFILENAME, 'TEXT', TEXTCREATOR);
#endif /*MACINTOSH*/
	}
	return (NULLSYMBOL);

  errorExit:
	putOUTSTR();
	return (0);
	
} /*spool()*/
