/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Stemleaf
#else
#define __SEG__ Stemleaf
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <math.h>
#define STEMMAIN__
/*#include "stems.h"*/

extern Symbolhandle             Install();
extern long                     symbolSize();
extern long                     argOK();
extern void                     badNargs();
extern long                     isVector();
extern long                     isScalar();

#define GY     0
#define NTRASH 1

Symbolhandle    stemleaf(list)
Symbolhandle    list;

{
	/* routine to do stema and leaf plot */

	Symbolhandle    symhx,symhstem, symhkey;
	long            j, nstem, tot, notmiss;	
	long            verbose = 0, xtrems = 0, width = SCREENWIDTH, extra;
	long            nargs;
	char           *keyword, **title = (char **) 0;
	double        **x, **y = (double **) 0;
	double          mstem;
	WHERE("stemleaf");
	TRASH(NTRASH,errorExit);
	
	OUTSTR[0] = '\0';
	nargs = NARGS(list);
	symhx = COMPVALUE(list, 0);

	if(!argOK(symhx,REAL,(nargs == 1) ? 0 : 1))
	{
		goto errorExit;
	}

	if(!isVector(symhx))
	{
		sprintf(OUTSTR,"ERROR: argument 1 to %s must be vector", FUNCNAME);
		goto errorExit;
	}
	
	while(nargs > 1)
	{
		symhkey = COMPVALUE(list,nargs-1);
		if(!(keyword = isKeyword(symhkey)))
		{
			break;
		}
		
		if(strncmp(keyword,"out",3) == 0)
		{
			if(TYPE(symhkey) != LOGIC || !isScalar(symhkey))
			{
				sprintf(OUTSTR,
						"ERROR: value for %s must be F or T",keyword);
			}
			else
			{
				xtrems = (DATAVALUE(symhkey,0) == 0.0);
			}
		}
		else if(strcmp(keyword,"title") == 0)
		{
			if(TYPE(symhkey) != CHAR || !isScalar(symhkey))
			{
				sprintf(OUTSTR,
						"ERROR: value for %s must be single CHARACTER variable or string",
						keyword);
			}
			else
			{
				title = STRING(symhkey);
			}
		}
		else
		{
			sprintf(OUTSTR,
					"ERROR: unrecognized keyword %s for %s",
					keyword,FUNCNAME);
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
		nargs--;
	} /* while(nargs > 1) */
	
	/* get number of stems */

	if(nargs > 2)
	{
		badNargs(FUNCNAME,-2);
		goto errorExit;
	}
	
	if (nargs == 2)
	{
		symhstem = COMPVALUE(list, 1);
		
		if(!argOK(symhstem,REAL, 2))
		{
			goto errorExit;
		}
		if(!isScalar(symhstem) ||
		   (mstem = floor(DATAVALUE(symhstem, 0))) <= 0 ||
		   mstem != DATAVALUE(symhstem,0) || mstem < 2.0)
		{
			sprintf(OUTSTR,
					"ERROR: number of stems for %s must be integer >= 2",
					FUNCNAME);
			goto errorExit;
		}
		
		nstem = (long) mstem;
	}
	else
	{
		extra = (xtrems != 0) ? 6 : 4;
		nstem = -((SCREENHEIGHT > 10) ? SCREENHEIGHT - extra : 24 - extra);
	}
	
	tot = symbolSize(symhx);
	x = DATA(symhx);

	notmiss = 0;

	if(!getScratch(y,GY,tot,double))
	{
		return (0);
	}
	
	for (j = 0; j < tot; j++)
	{
		if (!isMissing((*x)[j]))
		{
			(*y)[notmiss++] = (*x)[j];
		}
	}

	if (notmiss < 2)
	{
		myerrorout("ERROR: at most 1 non-missing value");
		goto errorExit;
	}
	else if (notmiss < tot)
	{
		sprintf(OUTSTR, "WARNING: %ld missing values ", tot - notmiss);
		putOUTSTR();
	}

	if(!doStemLeaf(y, notmiss, width, nstem, xtrems, verbose, title))
	{
		goto errorExit;
	}
	emptyTrash();

	*OUTSTR = '\0';
	return (NULLSYMBOL);

  errorExit:
	putOUTSTR();

	emptyTrash();

	return (0);
	
}

