/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Stemleaf
#else
#define __SEG__ Stemleaf
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <math.h>

/*
  This file contains C programs translated and adapted from Fortran code from	 the book ABCs of EDA by David Hoaglin and Paul Velleman, 1981, Duxbury

  The code was somewhat restructured and simplified by C. Bingham,
  kb@umnstat.stat.umn.edu

  Adapted for use with MacAnova 921125
  functions in this file adapted from original stemsubs.c are
  	stmnlf(), depthp(), outlyp(), sltitl(), stemp()
  Other related functions adapted from those file utilities.c are
	cinit(), intfn(), rprnt(), putchr(), putnum(), wdthof(), yinfo(),
	nposw()
*/

/*
	Various utility functions used by both box plot and stem leaf routines
	Converted to C by c. bingham, 900114
*/
#define SIGN(a,b) (((b) > 0) ? fabs((a)) : -fabs((a)))
#define MAX(a,b) (((a) > (b)) ? (a) : (b) )
#define MIN(a,b) (((a) < (b)) ? (a) : (b) )

#undef  TRUE
#undef  FALSE
#define TRUE 1
#define FALSE 0

#define DEPTHOFFSET 6  /*determines distance of depth to left of gutter*/
#define GUTTERPOS 11
#define INDENT "         "  /* GUTTERPOS - 2 spaces */

/* from common area /chrbuf/ */

#define Trash       NAMEFORTRASH
#define GP          0
#define GISCRATCH   1
#define NSCRATCH    2

static Symbolhandle Trash;

static long         Pmin; /* first column position */
static long         Pmax; /* last column position */
static long         Outptr,Maxptr;
static char       **P = (char **) 0;

/* replace stuff in numbrs.h */
#ifndef MAXINT
#define MAXINT   2147483647
#endif /*MAXINT*/
#define FUZZ     1e-10

#define NN 4
static double Nicnos[NN] = {1.0,2.0,5.0,10.0};

static long cinit(long width)
{
	long         i;
	char         blank = ' ', *p;

	Maxptr = Outptr = Pmin = 1;
	Pmax = width;
	if(Pmax - Pmin < GUTTERPOS + 7)
	{
		Pmax = Pmin + GUTTERPOS + 7;
	}
	
	if(!getScratch(P,GP,Pmax+1,char))
	{
		return (0);
	}
	p = *P;
	for(i=0;i<Pmax ;i++)
	{
		*p++ = blank;
	}
	*p = '\0';
	return (1);

}

static long intfn(double x)
{
	long      result;

	if (fabs(x) <= MAXINT)
	{
		result = (long) ((1.0 + FUZZ)*x);
	}
	else
	{
		result = (x > 0.0) ? MAXINT : -MAXINT;
	}
	return (result);
	
}


/* output line in *P and reinitialize */
static void rprnt(void)
{
	char blank = ' ';
	char *p = *P;
	int i;

	p[Maxptr] = '\0';
	myerrorout(*P);
	
	for(i=0;i<Pmax ;i++)
	{
		*p++ = blank;
	}
	Maxptr = Outptr = Pmin;
}


static void putchr(long posn,long chr)
{
	WHERE("putchr");
	
	if (posn != 0)
	{
		Outptr = MAX(Pmin,posn);
	}
	Outptr = MIN(Outptr,Pmax);
	(*P)[Outptr-1] = 255&chr;
	Maxptr = MAX(Maxptr,Outptr);
	Outptr++;
}


static void putnum(long posn,long n,long width)
{
	long        ip,numlen, i;
	char        dstk[50];
	WHERE("putnum");
	
	sprintf(dstk,"%ld",n);
	numlen = strlen(dstk);
	ip = ((posn != 0) ? posn : Outptr) + width - numlen;
	
	for(i=0;i<numlen;i++)
	{
		putchr(ip++,dstk[i]);
	}
}

static long wdthof(long val)
{
	long       absval = labs(val),numlen;

	numlen = (val >= 0) ? 1 : 2;
	while(absval > 0)
	{
		absval /= 10;
		numlen++;
	}
	
	return numlen;
}


static void yinfo(double *y,long n,double *med,double *hl,double *hh,
		   double *adjl,double *adjh,
		   long *iadjl,long *iadjh,double *step)
{
	double       hfence,lfence;
	long         j,k;

	sortquick(y,n);

	k = n;
	j = (k/2);
	*med = (y[j] + y[n - 1 - j])/2.0;
	k = (k + 1)/2;
	j = (k/2);
	*hl = (y[j]+y[k-1-j])/2.0;
	*hh = (y[n - k + j] + y[n - 1 - j])/2.0;
	*step = (*hh - *hl)*1.5;
	hfence = *hh + *step;
	lfence = *hl - *step;
	for(*iadjl = 0;y[*iadjl] < lfence; (*iadjl)++)
	{
		;
	}
	
	for(*iadjh = n-1;y[*iadjh] > hfence; (*iadjh)--)
	{
		;
	}
	
	*adjl = y[*iadjl]; /* lowest value within inner fences */
	*adjh = y[*iadjh];  /* highest value within inner fences */
	
	return;
}

static long nposw(double *high,double *low,
		   long maxp,long mzero,
		   long *ptotl,double *fract,
		   double *unit,double *npw)
{
	long         i;
	double       lo = *low , hi = *high;
	long         jlo, jhi, trialjlo, trialjhi;
	long         trialstems;
	double       trialunit, trialfract, trialnpw, aprxw;
	WHERE("nposw");

	*ptotl = -1;

	aprxw = (hi-lo)/(double) maxp;
	trialunit = pow(10.0,floor(log10(aprxw)));
	trialfract = aprxw/(trialunit);
	for(i=0;i<NN ;i++)
	{
		if (trialfract <= Nicnos[i])
		{
			break;
		}
	}

	while(1)
	{
		trialfract = Nicnos[i];
		trialnpw = trialfract*trialunit;
		trialjlo = intfn(lo/trialnpw);
		trialjhi = intfn(hi/trialnpw);
		trialstems = trialjhi - trialjlo + 1;
		if(*high > trialjhi*trialnpw + .99999*trialnpw)
		{
			trialstems++;
		}
		if (mzero && (hi*lo<0.0 || hi == 0.0))
		{
			trialstems++;
		}

		if (trialstems <= maxp)
		{
/*
  we are trying to find the largest value for trialstems that does not
  exceed maxp, not just the first.
*/
			if(trialstems <= *ptotl)
			{
				break;
			}
			*ptotl = trialstems;
			*npw = trialnpw;
			*unit = trialunit;
			*fract = trialfract;
			jlo = trialjlo;
			jhi = trialjhi;
			if(trialstems == maxp)
			{
				break;
			}
		}

		i++;
		
		if (i>=NN)
		{
			i = 0;
			trialunit *= 10.0;
		}
	} /* while(1) */
	if(*fract == 1.0)
	{
		*fract = 10.0;
		*unit /= 10.0;
	}
	
	*low = *npw*jlo;
	if(lo < 0)
	{
		*low -= *npw;
	}
	*high = *npw*jhi;
	if(hi > 0)
	{
		*high += *npw;
	}
	return (1);
}


static void depthp(double *w,long *iw,long n,
				   long *pt1,long *pt2,long cut,long iadjh, long hi,
				   long *rank,long *medyet,long slbrk)
{
	long         lefcnt,ptz,depth,nwid,ptx;
	long         cutreached = FALSE;
	WHERE("depthp");
	
	ptx = *pt1;
	for(*pt1=ptx;*pt1<=iadjh ;(*pt1)++)
	{
		if (iw[*pt1]>cut  ||  (cut >= 0 && iw[*pt1] == cut))
		{
			cutreached = TRUE;
			break;
		}
	}
	if(cutreached)
	{
		if (					/* hi>0 && */ cut == 0)
		{
			for(ptz = *pt1;ptz<n ;ptz++)
			{
				if (w[ptz] >= 0.0)
				{
					break;
				}
			}
	
			*pt1 = ptz;
			for(ptz = *pt1;ptz<n ;ptz++)
			{
				if (w[ptz]>0.0)
				{
					break;
				}
			}					/* now ptz-*pt1 is number of exact zeros */
			*pt1 = *pt1+intfn((double)(ptz - *pt1)/2.0 );
		}
	}
	else
	{
		*pt1 = iadjh+1;
	}
	lefcnt = *pt1 - *pt2;
	*rank = *rank + lefcnt;
	if (*medyet)
	{
		depth = n - (*rank - lefcnt);
	}
	else
	{
		if ((double) *rank == .5*(double) n)
		{
			*medyet = TRUE;
		}
		else if ((double) *rank >= .5*(double)(n+1))
		{
			nwid = wdthof(lefcnt);
			putchr(slbrk - DEPTHOFFSET - nwid -1 ,'(');
			putnum(0,lefcnt,nwid);
			putchr(0,')');
			*medyet = TRUE;
			return;
		}
		depth = *rank;
	}
	nwid = wdthof(depth);
	putnum(slbrk - DEPTHOFFSET - nwid,depth,nwid);
	return;
}

static void outlyp(long *iw,long from,long to,long hiend,long slbrk)
{
	long         opos,nwid,lhmax,i;
	WHERE("outlyp");
	
	opos = slbrk-DEPTHOFFSET;
	if (!hiend)
	{
		putchr(opos-3,'L');
		putchr(0,'o');
		putchr(0,'w');
	}
	else
	{
		putchr(opos-4,'H');
		putchr(0,'i');
		putchr(0,'g');
		putchr(0,'h');
	}
	putchr(opos,' ');
	nwid = MAX(wdthof(iw[from]),wdthof(iw[to]));
	lhmax = Pmax-nwid-2;

	for(i=from;i<=to ;i++)
	{
		putnum(0,iw[i],nwid);
		if(i<to)
		{
			putchr(0,',');
			putchr(0,' ');
		}
		
		if (Outptr >= lhmax || i == to)
		{
			rprnt();
			putchr(opos,' ');
		}
	}
}

static void sltitl(double unit,long linwid)
{
	long         iexpt;
	char        *p11;
	WHERE("sltitl");
	
	p11 = (linwid == 10) ? " 1|1" : "1*|1";
	sprintf(OUTSTR,"%s%s  represents ",INDENT,p11);
	myprint(OUTSTR);
	iexpt = intfn(log10(unit));
	if (iexpt > 8 || iexpt == -1)
	{
		sprintf(OUTSTR,"%g",11.*unit);
	}
	else if (iexpt  > 0)
	{
		sprintf(OUTSTR,"%ld",11 * intfn(unit));
	}
	else
	{
		sprintf(OUTSTR,"%.2g",11*pow(10.,(double) iexpt));
	}
	myprint(OUTSTR);
	sprintf(OUTSTR,"  Leaf digit unit = %g",unit);
	putOUTSTR();
}

static void stemp(long stem,long linwid,long negnow,long slbrk)
{
	long         nstem,lefdig,nwid,opos;
	char        *ch5stm = "*tfs.";
	WHERE("stemp");
	
	nstem = stem/10;
	lefdig = labs(stem-nstem*10);
	nwid = wdthof(nstem);
	opos = (linwid == 10) ? slbrk-2 : slbrk-3;
	if (nstem == 0)
	{
		putchr(opos, (negnow) ? '-' : '+');
	}
	
	opos = (linwid == 10) ? slbrk - nwid : slbrk-nwid-1;
	putnum(opos,nstem,nwid);

	if (linwid == 2)
	{
		putchr(0,ch5stm[lefdig/2]);
	}
	else if (linwid == 5)
	{
		putchr(0, (lefdig < 5) ? '*' : '.');
	}
	else if (linwid != 10)
	{
		myerrorout("stemleaf error 12");
	}
	putchr(slbrk,'|');
	return;
}



static long stmnlf(double *y,long n,long *iw,
				   long maxlines,long xtrems,long verbose)
{
	double        med,hl,hh,adjl,adjh,step,unit,fract,npw;
	long          i,slbrk,pltwid,rank,iadjl,iadjh,nlins;
	long          nlmax,linwid;
	long          low,hi,cut,stem,pt1,pt2,j,spacnt,leaf;
	long          negnow,medyet;
	WHERE("stmnlf");

	slbrk = Pmin+GUTTERPOS;
	pltwid = Pmax-slbrk-2;

	yinfo(y,n,&med,&hl,&hh,&adjl,&adjh,&iadjl,&iadjh,&step);
	
	if(verbose)
	{
		sprintf(OUTSTR,"n = %ld  Min=%g  QL=%g  M=%g  QU=%g  Max=%g",
			n, y[0], hl, med, hh, y[n-1]);
		putOUTSTR();
	}
	if (adjh <= adjl || xtrems)
	{
		iadjl = 0;
		iadjh = n-1;
		adjl = y[iadjl];
		adjh = y[iadjh];
	}
	/* maxlines < 0 => -maxlines was default and was not reset */
	nlmax = (maxlines < 0) ?
		intfn(10.0*log10((double) (iadjh-iadjl+1))) : maxlines;
	nlmax = MIN(nlmax, labs(maxlines));
	if (adjh <= adjl)
	{
		adjh = adjl+1.0;
		nlmax = 1;
	}
	if(!nposw(&adjh,&adjl,nlmax,TRUE,&nlins,&fract,&unit,&npw))
	{
		return (0);
	}

	if(!xtrems)
	{
		while(iadjl >= 0 && y[iadjl] >= adjl)
		{
			iadjl--;
		}
		iadjl++;
		while(iadjh < n && y[iadjh] <= adjh)
		{
			iadjh++;
		}
		iadjh--;
	}

	for(i=0;i<n ;i++)
	{
		iw[i] = intfn(y[i]/unit);
	}
	if (fract != 10.0)
	{
		for(i=iadjl;i<=iadjh ;i++)
		{
			if (iw[i]%10 != 0)
			{
				break;
			}
		}
		if(i > iadjh)
		{
			fract = 10.0;
			npw = fract*unit;
			nlins = intfn(adjh/npw)-intfn(adjl/npw)+1;
			if (adjh*adjl<0.0 || adjh == 0.0)
			{
				nlins = nlins+1;
			}
		}
	}
	low = iw[iadjl];
	hi = iw[iadjh];
	linwid = intfn(fract);
	rank = iadjl;
	if (rank > 0)
	{
		outlyp(iw,0,rank-1,FALSE,slbrk);
	}
	cut = linwid * (int) floor((1.0+FUZZ)*(double) low/(double) linwid);
	if (adjl >= 0)
	{
		negnow = FALSE;
		stem = cut-linwid;
	}
	else
	{
		cut = (cut == 0) ? cut - linwid : cut;
		negnow = TRUE;
		stem = cut;
	}
	medyet = FALSE;
	pt1 = iadjl;
	pt2 = pt1;
	for(j=0;j<nlins ;j++)
	{
		if(pt1 > iadjh) /*if(cut > hi)*/
		{
			break;
		}
		cut += linwid;
		if (stem != 0 || !negnow)
		{
			stem += linwid;
		}
		else
		{
			negnow = FALSE;
		}
		spacnt = 0;

		depthp(y,iw,n,&pt1,&pt2,cut,iadjh,hi,&rank,&medyet,slbrk);

		stemp(stem,linwid,negnow,slbrk);

		while(pt2 < pt1)
		{
			leaf = labs(iw[pt2]-(stem/10)*10);
			putnum(0,leaf,1);
			if (++spacnt >= pltwid)
			{
				break;
			}
			pt2++;
		}
		if(pt2 < pt1)
		{ /* overflow */
			putchr(0,'*');
			pt2 = pt1;
		}
		rprnt();
	} /* for j */
	if (pt1 < n)
	{
		outlyp(iw,pt1,n-1,TRUE,slbrk);
	}
	else
	{
		myeol();
	}
	
	sltitl(unit,linwid);
	return (1);
	
}

long doStemLeaf(double ** data, long n, long width,
					   long maxlines, long xtrems, long verbose, char **title)
{
	long         **iscratch = (long **) 0;
	WHERE("doStemLeaf");
	
	P = (char **) 0;
	if((Trash = GarbInstall(2)) == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	if(!cinit(width))
	{
		goto errorExit;;
	}
	
	if(!getScratch(iscratch,GISCRATCH,n,long))
	{
		goto errorExit;
	}

	if(title != (char **) 0)
	{
		myerrorout(*title);
	}
	
	if(!stmnlf(*data,n,*iscratch,maxlines,xtrems,verbose))
	{
		goto errorExit;
	}
	emptyTrash();
	P = (char **) 0;
	
	return (1);
	
  errorExit:
	putOUTSTR();

	emptyTrash();
	P = (char **) 0;
	
	return (0);
	
}

