/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Svd
#else
#define __SEG__ Svd
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <math.h>

extern long         isMatrix();
extern long         argOK();
extern long         symbolSize();
extern char        *isKeyword();
extern long         isMatrix();
extern Symbolhandle Install(), RInstall(), RSInstall();
extern long         anyMissing();

#define GVALS      0
#define GU         1
#define GV         2
#define GSCRATCH   3
#define NTRASH     4

/*
  singular value decomoposition of a real matrix

  usage:  svd(x [,left:T] [,right:T])

  If neither left:T or right:T is present, svd returns a vector of
  singular values in decreasing order
  Otherwise it returns a structure with components 'values', 'leftvectors',
  and/or 'rightvectors'
*/

Symbolhandle    svd(list)
Symbolhandle    list;
{
	Symbolhandle    result = (Symbolhandle) 0, args[3], symh;
	long            ncomps = 1, nrows, ncols, i;
	long            dim[2], nargs;
	long            ierr;
	long            left = 0, right = 0;
	double        **vals = (double **) 0, **u = (double **) 0;
	double        **v = (double **) 0, **scratch = (double **) 0;
	char           *keyword;
	char           *compNames[3];
	WHERE("svd");
	TRASH(NTRASH,errorExit);
	
	OUTSTR[0] = '\0';
	nargs = NARGS(list);
	if(nargs > 3)
	{
		badNargs(FUNCNAME,-3);
		goto errorExit;
	}
	
	for(i=0;i < nargs;i++)
	{
		args[i] = COMPVALUE(list,i);
		if(!argOK(args[i],0, i+1))
		{
			goto errorExit;
		}

		if(i == 0)
		{
			if(TYPE(args[0]) != REAL || !isMatrix(args[0],dim))
			{
				sprintf(OUTSTR,"ERROR: 1st argument to %s is not REAL matrix",
						FUNCNAME);
			}
			else if(anyMissing(args[0]))
			{
				sprintf(OUTSTR,
						"ERROR: missing values not allowed in argument to %s",
						FUNCNAME);
			}
			else if(dim[1] > dim[0])
			{
				sprintf(OUTSTR,
						"ERROR: argument to %s has fewer rows than columns",
						FUNCNAME);
			}
		} /*if(i == 0)*/
		else
		{
			if(!(keyword = isKeyword(args[i])))
			{
				sprintf(OUTSTR,
						"ERROR: argument %ld to %s must be 'right:T' or 'left:T'",
						i+1,FUNCNAME);
			}
			else if(strcmp(keyword,"left") != 0  &&
					strcmp(keyword,"right") != 0)
			{
				sprintf(OUTSTR,"ERROR: unrecognized keyword %s",keyword);
			}
			else if(TYPE(args[i]) != LOGIC || !isScalar(args[i]))
			{
				notTorF(keyword);
				goto errorExit;
			}
			else if(strcmp(keyword,"left") == 0)
			{
				left = (DATAVALUE(args[i],0) != 0.0);
				if(left)
				{
					ncomps++;
				}
			}
			else
			{
				right = (DATAVALUE(args[i],0) != 0.0);
				if(right)
				{
					ncomps++;
				}
			}
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
	} /*for(i=0;i < nargs;i++)*/

	nrows = dim[0];
	ncols = dim[1];
		
	compNames[0] = "values";
	compNames[1] = "leftvectors";
	compNames[2] = "rightvectors";

	if(!getScratch(vals,GVALS,ncols,double) ||
	   !getScratch(u,GU,nrows*ncols,double) ||
	   right && !getScratch(v,GV,ncols * ncols, double) ||
	   !getScratch(scratch,GSCRATCH,ncols,double))
	{
		goto errorExit;
	}

	eispsvd(nrows,nrows,ncols,DATAPTR(args[0]),*vals, left, *u, right,
			(right) ? *v : (double *) 0, *scratch, &ierr);

	if (ierr > 0)
	{
		sprintf(OUTSTR,
				"ERROR: singular value algorithm in %s did not converge",
				FUNCNAME);
		goto errorExit;
	}

	if(ncomps == 1)
	{
		if((result = Install(SCRATCH,REAL)) == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		symh = result;
	} /*if(ncomps == 1)*/
	else
	{
		result = RSInstall(SCRATCH,ncomps,compNames,0L);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		i = 1;
		if(left)
		{
			unTrash(GU);
			symh = COMPVALUE(result,i++);
			setNAME(symh,compNames[1]);
			setDATA(symh,u);
			setNDIMS(symh,2);
			setDIM(symh,1,nrows);
			setDIM(symh,2,ncols);
		}
			
		if(right)
		{
			unTrash(GV);
			symh = COMPVALUE(result,i);
			setNAME(symh,compNames[2]);
			setDATA(symh,v);
			setNDIMS(symh,2);
			setDIM(symh,1,ncols);
			setDIM(symh,2,ncols);
		}
		symh = COMPVALUE(result,0);	
	}
	
	unTrash(GVALS);
	setDATA(symh,vals);
	setNDIMS(symh,1);
	setDIM(symh,1,ncols);

	emptyTrash();

	return (result);

  errorExit:
	putOUTSTR();

	emptyTrash();

	return (0);
	
}


