/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Swp
#else
#define __SEG__ Swp
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

#define GCOLSH  0
#define GDIAGSH 1
#define GHD     1  /* GDIAGSH and GHD are mutually exclusive*/
#define NTRASH  2

#define Trash NAMEFORTRASH

static Symbolhandle   Trash;

#define ISWP       1
#define IBCPRD     2

#define NOMEM     -1
#define SINGULAR  -2
#define TOOBIGCOL -3

#ifdef UNDEFINED__ /* swpError() no longer referenced */
static void swpError(char *where,long errCode,long cols[],long n)
{
	long j;
	
	fprintf(stderr,"%s: error in swp:",where);
	switch(errCode)
	{
	  case NOMEM:
		fprintf(stderr," no memory for scratch");
		break;
	  case SINGULAR:
		j = 0;
		while(j < n && cols[j] >= 0)
			j++;
		if(j < n)
			j = -cols[j] - 1;
		fprintf(stderr,"pivot in column %ld too small",j);
		break;
	  case TOOBIGCOL:
		fprintf(stderr,"attempt to sweep on non-existent column");
		break;
	}
} /*swpError()*/
#endif /*UNDEFINED__*/

/*
	this version of swp does a non-symmetric of the	m by n  matrix cp.
	The pivot row/column is specified by k.  The original k-th diagonal
	element is cpdiag

	An error code is returned.
	
*/

static long doSwp(double **arg,long m,long n,long k,double cpdiag)
{
	long             i,j;
	register double  beta;
	double           cpkk;
	double          *cp = *arg;
	long             cpkkOffset, cpkjOffset, cpijOffset, cpikOffset;
	WHERE("doSwp");

	cpkkOffset = k*(m+1);
	cpkk = cp[cpkkOffset];	/* cp[k,k] */
	if(fabs(cpkk) <= REGTOL*cpdiag)
	{
		return (SINGULAR);
	}
	cpkjOffset = k;		/* offset for cp[k,0] */
	for(j=0;j<n;j++)
	{
		if(j != k)
		{
			beta =  cp[cpkjOffset]/cpkk; /* cp[k,j]/c[k,k] */
			cpijOffset = j*m;		/* offset for cp[0,j] */
			cpikOffset = k*m;		/* offset for cp[0,k] */
			for(i=0;i<m;i++)
			{
				if(i!= k)
				{
					cp[cpijOffset] -= cp[cpikOffset]*beta;
					/* cp[i,j]-=cp[j,k]*beta */
				}
				cpijOffset++;
				cpikOffset++;
			} /*for(i=0;i<m;i++)*/
			cp[cpkjOffset] = -beta;		/* cp[k,j] = -beta; */
		} /*if(j != k)*/
		cpkjOffset += m;
	} /*for(j=0;j<n;j++)*/

	cpikOffset = k*m;				/* cp[0,k] */
	for(i=0;i<m;i++)
	{
		if(i != k)
		{
			cp[cpikOffset] /= cpkk; /* cp[i,k] /= c[k,k];*/
		} /*if (i != k)*/
		cpikOffset++;
	} /*for(i=0;i<m;i++)*/

	cp[cpkkOffset] = 1.0/cpkk;	/* cp[k,k] = 1/cp[k,k]; */
	return (0);
} /*doSwp()*/

static long swpIt(double **yH, double **colsH, long nRows, long nCols,
				  long nswp)
{
	double         **diagsH = (double **) 0;
	double          *diags, *y, *cols;
	long             i, iswp, minRowCol;
	long             retValue = 0;
	short            solve = (colsH == (double **) 0);
	WHERE("swpIt");
	
	*OUTSTR = '\0';
	minRowCol = (nRows > nCols) ? nCols : nRows;
	if(!getScratch(diagsH,GDIAGSH,minRowCol, double))
	{
		goto errorExit;
	}
	diags = *diagsH;
	y = *yH;
	for(i=0;i<minRowCol;i++)
	{ /* save diagonals */
		diags[i] = y[i*(nRows+1)];
	}

	cols = *colsH;
	for(i = 0;i<nswp;i++)
	{
		iswp = (solve) ? i : (long) cols[i] - 1;
		if(doSwp(yH, nRows, nCols, iswp, diags[iswp]))
		{
			sprintf(OUTSTR, (solve) ? 
					"ERROR: singularity indicated by tolerance failure pivoting column %ld" :
					"WARNING: tolerance failure pivoting column %ld; not swept",
					iswp+1);
			if (solve)
			{
				goto errorExit;
			}
			putOUTSTR();
		} /*if(doSwp(yH, nRows, nCols, iswp, diags[iswp]))*/
	} /*for(i = 0;i<nswp;i++)*/
	retValue = 1;

/* fall through */

  errorExit:
	putOUTSTR();
/*	discardScratch(diagsH,GDIAGSH,double);*/
	return (retValue);
}

static long doBcprd(Symbolhandle argx, Symbolhandle argcp)
{
	double             *cp, *x;
	double             *xbar, *d, **hd;
	double              factor, fn;
	long                n1, n2, n2sq, p, dim[2];
	long                i, retValue = 0;
	long                j1, j2, xij, cpij, cpji;
	WHERE("doBcprd");
	
	(void) isMatrix(argx,dim);
	n1 = dim[0];
	p = dim[1];
	n2 = p + 1;
	n2sq = n2*n2;

	if(!getScratch(hd,GHD,p,double))
	{
		goto errorExit;
	}

	x = DATAPTR(argx);
	cp = DATAPTR(argcp);
	xbar = cp;
	d = (*hd) - 1;

	cp[0] = 1./(double) n1;
	for(i = 1;i < n2sq;i++)
	{
		cp[i] = 0;
	}
	fn = 0;
	for(i = 0;i < n1;i++)
	{
		fn++;
		xij = i;
		for(j1 = 1;j1 <= p;j1++)
		{
			d[j1] = x[xij] - xbar[j1];
			xij += n1;
			xbar[j1] += d[j1]/fn;
			factor = 1.0 - 1.0/fn;
			cpij = n2 + j1;
			for(j2 = 1;j2 <= j1; j2++)
			{
				cp[cpij] += factor*d[j1]*d[j2];
				cpij += n2;
			} /*for(j2 = 1;j2 <= j1; j2++)*/
		} /*for(j1 = 1;j1 <= p;j1++)*/
	} /*for(i = 0;i < n1;i++)*/
	
	for(j1 = 1;j1<=p;j1++)
	{
		cp[j1*n2] = -xbar[j1];
		cpij = n2 + j1;
		cpji = j1*n2 + 1;
		for(j2 = 1;j2 < j1;j2++)
		{
			cp[cpji] = cp[cpij];
			cpij += n2;
			cpji++;
		} /*for(j2 = 1;j2 < j1;j2++)*/
	} /*for(j1 = 1;j1<=p;j1++)*/
	
	retValue = 1;
	/* fall through */

  errorExit:
/*	discardScratch(hd,GHD,double);*/
	return (retValue);
	
} /*doBcprd()*/

typedef struct opEntry
{
	char *name;
	short iop;
	short nargs;
} opEntry;

static opEntry Ops[] =
{ /* name		iop		nargs */
	"swp",		ISWP,	 -1002,
	"bcprd",	IBCPRD,	 1,
	(char *) 0,	0,		 0
};


/* 
   usage:
	   bcprd(x)
	   swp(x,cols) or swp(x,n1,n2,...)

  modified 931103 to accept swp(a, vec1, vec2, ...) with same meaning as
  swp(a,cat(vec1,vec2,...))
*/

Symbolhandle    swp(list)
Symbolhandle         list;
{
	Symbolhandle        arg1, argi, result = (Symbolhandle) 0;
	long                margs, nargs, nRows, nCols, nswp = 0, i, iarg;
	long                dim[2], minRowCol;
	long                length;
	long                iop;
	opEntry            *op;
	double              colNo, *coli, *colsP;
	double            **y = (double **) 0, **colsH = (double **) 0;
	WHERE("swp");
	
	if ((Trash = GarbInstall(NTRASH)) == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	OUTSTR[0] = '\0';

	dim[0] = dim[1] = 0;

	nargs = NARGS(list);
	for(op = Ops; op->iop != 0; op++)
	{
		if(strcmp(op->name,FUNCNAME) == 0)
		{
			break;
		}
	}
	if(op->iop == 0)
	{ /* should not occur */
		sprintf(OUTSTR,"ERROR: (internal) unrecognized funcname '%s' in swp()",
				FUNCNAME);
		goto errorExit;
	}
	margs = op->nargs;
	iop = op->iop;
	
	if(margs > 0 && nargs != margs || margs < 0 && nargs < labs(margs) - 1000)
	{
		badNargs(FUNCNAME, margs);
		goto errorExit;
	}
	
	arg1 = COMPVALUE(list,0);
	if(!argOK(arg1,REAL,(iop == ISWP) ? 0 : 1))
	{
		goto errorExit;
	}
	if(anyMissing(arg1))
	{
		sprintf(OUTSTR,"ERROR: missing values in %sargument to %s",
				(iop == ISWP) ? "1st " : "", FUNCNAME);
	}
	else if(!isMatrix(arg1,dim))
	{
		sprintf(OUTSTR,"ERROR: 1st argument to %s must be matrix",FUNCNAME);
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}

	nRows = dim[0];
	nCols = dim[1];
	minRowCol = (nRows < nCols) ? nRows : nCols;
	
	for(iarg=1;iarg<nargs;iarg++)
	{
		if(!argOK(argi = COMPVALUE(list,iarg),REAL,iarg+1))
		{
			goto errorExit;
		}
		length = symbolSize(argi);
		if(!isVector(argi) && length != DIMVAL(argi,2))
		{/* neither column or row vector */
			sprintf(OUTSTR,
					"ERROR: arguments to %s must be vectors except for first",
					FUNCNAME);
		}
		else if(anyMissing(argi))
		{
			sprintf(OUTSTR,"ERROR: missing values in argument %ld to %s",
					iarg + 1, FUNCNAME);
		}
		if (*OUTSTR)
		{
			goto errorExit;
		}
		colsP = DATAPTR(argi);
		for(i=0;i<length;i++)
		{
			colNo = colsP[i];
			if(colNo <= 0 || colNo != floor(colNo))
			{
				sprintf(OUTSTR,
						"ERROR: %ld-th argument to %s not all positive integers",
						iarg, FUNCNAME);
			}
			else if(colNo > minRowCol)
			{
				sprintf(OUTSTR,
						"ERROR: attempt to sweep a non-existent row or column");
			}
			if(*OUTSTR)
			{
				goto errorExit;
			}
		} /*for(i=0;i<length;i++)*/		
		nswp += length;
	} /*for(iarg=1;i<nargs;i++)*/
	
	if(iop == ISWP)
	{
		if(nargs == 2)
		{
			colsH = DATA(COMPVALUE(list,1));
		}
		else if(!getScratch(colsH, 0, nswp, double))
		{
			goto errorExit;
		}
		else
		{
			colsP = *colsH;
			for(iarg = 1;iarg < nargs; iarg++)
			{
				argi = COMPVALUE(list,iarg);
				coli = DATAPTR(argi);
				length = symbolSize(argi);
				for(i=0;i<length;i++)
				{
					colsP[i] = coli[i];
				}
				colsP += length;
			}
		}
		
		if(isscratch(NAME(arg1)))
		{ /* re-use symbol */
			COMPVALUE(list,0) = (Symbolhandle) 0;
			result = arg1;
		}
		else
		{/* copy arg1 to result */
			result = RInstall(SCRATCH, 0);
			if(result == (Symbolhandle) 0 || Copy(arg1,result) == 0)
			{
				goto errorExit;
			}
			setNAME(result,SCRATCH);
		} /*if(isscratch(NAME(arg1))){}else{}*/
		setNDIMS(result,2);
		setDIM(result,1,nRows);
		setDIM(result,2,nCols);

		y = DATA(result);
		if(!swpIt(y, colsH, nRows, nCols, nswp))
		{
			goto errorExit;
		}
	} /*if(iop == ISWP)*/
	else
	{ /* iop == IBCPRD */
		result = RInstall(SCRATCH,(nCols + 1)*(nCols + 1));
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNDIMS(result,2);
		setDIM(result,1,nCols + 1);
		setDIM(result,2,nCols + 1);
		if(doBcprd(arg1,result) == 0)
		{
			goto errorExit;
		}
	} /*if(iop == ISWP){...}else{...}*/
	
	emptyTrash();
	
	*OUTSTR = '\0';

	return (result);

  errorExit:
	putOUTSTR();
	
	emptyTrash();
	Removesymbol(result);
	
	return (0);
	
} /*swp()*/

