/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Tabs
#else
#define __SEG__ Tabs
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <ctype.h>

#ifdef UNDEFINED__
#define means(I)   ((*tmeans)[(I)+jj])
#define counts(I)  ((*tcounts)[(I)+jj])
#define vars(I)    ((*tvars)[(I)+jj])
#endif /*UNDEFINED__*/

#define coord(I,J) ((*(ClassVars[J]))[I])

#define IMEAN         0x01
#define IVAR          0x02
#define ICOUNT        0x04

static double       **ClassVars[MAXDIMS];
static long           Nvars, Ndata, Nclass[MAXDIMS], Types[MAXDIMS];

/* modified by kb to allow for multivariate data */

static void     celltabs(double **ty, double **tmeans,
						 double **tvars, double **tcounts, long ny,
						 unsigned long control)
{
	long            i, j, k, ncells, icell;
	long            ivar;
	long            term[MAXDIMS];
	long            needVars = (control & IVAR);
	long            needMeans = (control & (IMEAN | IVAR));
	long            length;
	short           nmissing = 0;
	double          d,yy;
	double         *y = *ty;
	double         *vars = (double *) 0, *means = (double *) 0, *counts;

	ncells = 1;
	for (i = 0; i < Nvars; i++)
	{
		term[i] = ncells;
		ncells *= (long) Nclass[i];
	}
	length = ncells * ny;
	/*  ncells contains the number of means to be computed
	
	    term now contains the offsets for indexing into array
	    of means  */

	counts = *tcounts;
	doubleFill(counts, 0.0, length);
	if(needVars)
	{
		vars = *tvars;
		doubleFill(vars, 0.0, length);
	}
	if(needMeans)
	{
		means = *tmeans;
		doubleFill(means, 0.0, length);
	}
	
	for(ivar = 0;ivar < ny;ivar++)
	{
		for (i = 0; i < Ndata; i++)
		{
			yy = y[i];
			if (!isMissing(yy))
			{
				for (j = 0; j < Nvars; j++)
				{
					if (isMissing(coord(i,j)))
					{
						nmissing++;
						break;
					}
				}
				if (j == Nvars)
				{
					icell = 0;
					for (j = 0; j < Nvars; j++)
					{
						k = (Types[j] == REAL) ? ((long) coord(i,j) - 1) :
							((coord(i,j) != 0.0)) ? 1 : 0;
						icell += term[j] * k;	/* find cell number */
					}
					counts[icell]++; /* increment count */
					if(needMeans)
					{
						means[icell] += yy; /* increment sum  */
					}
				} /*if (j == Nvars)*/
			} /*if (!isMissing(yy))*/
		} /*for (i = 0; i < Ndata; i++)*/

		if(needMeans)
		{
			for (i = 0; i < ncells; i++)
			{
				if (counts[i] > 0.0)
				{
					means[i] /= counts[i];
				}
			} /*for (i = 0; i < ncells; i++)*/
		}
		
		if(needVars)
		{
			for (i = 0; i < Ndata; i++)
			{
				yy = y[i];
				if (!isMissing(yy))
				{
					if(nmissing)
					{
						for (j = 0; j < Nvars; j++)
						{
							if (isMissing(coord(i,j)))
							{
								nmissing--;
								break;
							}
						} /*for (j = 0; j < Nvars; j++)*/
					} /*if(nmissing)*/
					else
					{
						j = Nvars;
					}
					
					if (j == Nvars)
					{
						icell = 0;
						for (j = 0; j < Nvars; j++)
						{
							k = (Types[j] == REAL) ? ((long) coord(i,j) - 1) :
							((coord(i,j) != 0.0)) ? 1 : 0;
							icell += term[j] * k; /* find cell number */
						}
						d = yy - means[icell];
						vars[icell] += d * d; /* get variance*/
					} /*if (j == Nvars)*/
				} /*if (!isMissing(yy))*/
			} /*for (i = 0; i < Ndata; i++)*/
			for (i = 0; i < ncells; i++)
			{
				if (counts[i] > 1.0)
				{
					vars[i] /= (counts[i] - 1.);
				}
			}
		} /*if(needVars)*/
		counts += ncells;
		means += ncells;
		vars += ncells;
		y += Ndata;
	} /*for(ivar = 0;ivar < ny;ivar++)*/
} /*static void celltabs()*/

#define GMEANS    0
#define GVARS     1
#define GCOUNTS   2
#define NTRASH    3

/*
  modified by kb to allow for the variable tabulated being a matrix
*/

Symbolhandle    tabs(list)
Symbolhandle    list;
{
	Symbolhandle    result = (Symbolhandle) 0, symh;
	double        **y, **out[3];
	long            dim[2], ny;
	long            ncells, i, j, k;
	long            ncomps = 0;
	long            nargs = NARGS(list);
	short           all = 1;
	short           doit[3];
	unsigned long   control = ICOUNT | IMEAN | IVAR, which, which1;
	char            *compname[3], *keyword;
	WHERE("tabs");
	TRASH(NTRASH, errorExit);
	
	compname[0] = "mean";
	compname[1] = "var";
	compname[2] = "count";

	for(i=0;i<3;i++)
	{
		out[i] = (double **) 0;
		doit[i] = 1;
	}
	
	*OUTSTR = '\0';

	while(nargs > 0)
	{
		if((symh = COMPVALUE(list,nargs-1)) == (Symbolhandle) 0 ||
		   (keyword = isKeyword(symh)) == (char *) 0)
		{
			break;
		}
		if(all)
		{
			all = 0;
			for(i=0;i<3;i++)
			{
				doit[i] = 0;
			}
		}
		if(strncmp(keyword, "mean", 4) == 0)
		{
			which = IMEAN;
			which1 = 0;
		}
		else if(strncmp(keyword, "var",3)  == 0)
		{
			which = IVAR;
			which1 = 1;
		}
		else if(strncmp(keyword, "count", 5) == 0)
		{
			which = ICOUNT;
			which1 = 2;
		}
		else
		{
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
		if(!isScalar(symh) || TYPE(symh) != LOGIC)
		{
			notTorF(keyword);
			goto errorExit;
		}
		if(DATAVALUE(symh,0) != 0.0)
		{
			control |= which;
			doit[which1] = 1;
		}
		else
		{
			control &= ~which;
			doit[which1] = 0;
		}
		nargs--;
	} /*while(nargs > 0)*/
	
	ncomps = doit[0] + doit[1] + doit[2];
	if(ncomps == 0)
	{
		sprintf(OUTSTR, "ERROR: nothing specified to compute");
		goto errorExit;
	}
	
	if(nargs < 2)
	{
		badNargs(FUNCNAME,-1002);
		return(0);
	}
	
	symh = COMPVALUE(list, 0);
	if(!argOK(symh,REAL,0))
	{
		goto errorExit;
	}

	if (!isMatrix(symh,dim))
	{
		sprintf(OUTSTR,
				"ERROR: variable tabulated must be REAL vector or matrix");
		goto errorExit;
	}

	y = DATA(symh);
	Ndata = dim[0];
	ny = dim[1];

	Nvars = nargs - 1;
	if(Nvars > MAXDIMS)
	{
		sprintf(OUTSTR,"ERROR: more than %ld factors in %s",(long) MAXDIMS,
				FUNCNAME);
		goto errorExit;
	}
	
	ncells = 1;
	for (i = 0; i < Nvars; i++)
	{
		symh = COMPVALUE(list, (i + 1));
		if(!argOK(symh,0,i+2))
		{
			goto errorExit;

		}
		Types[i] = TYPE(symh);
		if(Types[i] != REAL && Types[i] != LOGIC)
		{
			badType(FUNCNAME,-Types[i],i+2);
			goto errorExit;
		}
		
		if((Nclass[i] = labs(isFactor(symh))) == 0 ||
		   symbolSize(symh) != Ndata)
		{
			sprintf(OUTSTR,
					"ERROR: class variables must be factors with same length as tabulated variable");
			goto errorExit;
		}
		ClassVars[i] = DATA(symh);
		ncells *= Nclass[i];
	} /*for (i = 0; i < Nvars; i++)*/

	if((control & (IMEAN | IVAR)) &&
	   !getScratch(out[0], GMEANS, ncells*ny, double) ||
	   (control & IVAR) &&
	   !getScratch(out[1], GVARS, ncells*ny, double) ||
	   !getScratch(out[2], GCOUNTS, ncells*ny, double))
	{
		goto errorExit;
	}
	
	result = (ncomps == 1) ? RInstall(SCRATCH, 0) :
		RSInstall(SCRATCH, ncomps, (char **) 0, 0);
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	if(ncomps == 1)
	{
		for(i=0;i<3;i++)
		{
			if(doit[i])
			{
				unTrash(i);
				setDATA(result, out[i]);
				break;
			}
		}
	} /*if(ncomps == 1)*/
	else
	{
		k = 0;
		for(i=0;i<3;i++)
		{
			if(doit[i])
			{
				symh = COMPVALUE(result,k++);
				unTrash(i);
				setDATA(symh, out[i]);
				setNAME(symh, compname[i]);
			} /*if(doit[i])*/
		} /*for(i=0;i<3;i++)*/
	} /*ifncomps == 1){}else{}*/

	for(i=0;i<ncomps;i++)
	{
		symh = (ncomps == 1) ? result : COMPVALUE(result, i);
		setNDIMS(symh,Nvars + ((ny > 1) ? 1 : 0));
		for (j = 1; j <= Nvars; j++)
		{
			setDIM(symh,j,Nclass[j-1]);
		}
		if(ny > 1)
		{
			setDIM(symh,Nvars+1,ny);
		}
	} /*for(i=0;i<ncomps;i++)*/
	
	celltabs(y, out[0], out[1], out[2], ny, control);

	emptyTrash();

	return (result);

  errorExit:
	putOUTSTR();
	emptyTrash();

	return(0);
	
} /*Symbolhandle tabs()*/


