/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef MPW
#ifdef MPW3
#pragma segment Graphics
#else
#define __SEG__ Graphics
#endif /*MPW3*/
#endif /*MPW*/

/*
 *
 *    G N U P L O T  --  term.c
 *
 *  Copyright (C) 1986, 1987  Colin Kelley, Thomas Williams
 *
 *  You may use this code as you wish if credit is given and this message
 *  is retained.
 *
 *  Please e-mail any useful additions to vu-vlsi!plot so they may be
 *  included in later releases.
 *
 *  This file should be edited with 4-column tabs!  (:set ts=4 sw=4 in vi)
 */

/*
 *  Modified for MacAnova by Gary W. Oehlert
 *  Further modified by C. Bingham
 *  if t is term_tbl[term], (*t->init)() should return 0 if o.k., 1, if not
 */

#undef TERMC__
#define TERMC__ /* should be defined only here */

#include "globals.h"
#include "plot.h"
#include "dbug.h"
#include <stdio.h>

#undef assert
#define assert(A)

#ifdef MACINTOSH
#include "macIface.h"
#endif /*MACINTOSH*/

void do_point(long /*x*/, long /*y*/, long /*number*/);

#define SCREENDUMPUNDEFINED 0

extern FILE   *PLOTFILE; /*kb*/
extern char   *TERMINAL;
extern char   *ScreenDumpFile;
extern short   ScreenDumpError;

extern struct termentry term_tbl[];

#undef UNDEFINED__
#ifdef UNDEFINED__ /* following is now in plot.h */
long            term = 1;	/* gives tek, Macintosh stuff commented out*/
#endif /*UNDEFINED__*/

#define POINT_TYPES		7
#define sign(X)    (((X) >= 0) ? 1 : -1)

#ifdef MACINTOSH


#define MAC_XMAX 465
#define MAC_YMAX 263 /*273*/
#define MAC_XLAST (MAC_XMAX - 1)
#define MAC_YLAST (MAC_YMAX - 1)

#define MAC_VCHAR	11
#define MAC_HCHAR	6
#define MAC_VTIC	8
#define MAC_HTIC	7

#define HAIRLINESCALE   2

#define SetLineWidth  182  /* PicComment code from ISM */
typedef Point         TLineWidth, *TLineWidthPtr, **TLineWidthHdl;

static TLineWidthHdl LineWidthH = (TLineWidthHdl) 0;
static Integer       MacThickness;
static Integer       MacFrameTop;

int            MAC_init(void);
int            MAC_graphics(void);
void           MAC_text(void);
void           MAC_reset(void);
void           MAC_linetype(long /*linetype*/);
void           MAC_move(unsigned long /*x*/, unsigned long /*y*/);
void           MAC_vector(unsigned long /*x*/, unsigned long /*y*/);
void           MAC_str_text(char * /*str*/);
void           MAC_chr_text(char /*c*/);

static setLineScale(Integer scale)
{
	TLineWidth           oldLineWidth = **LineWidthH;

	/* undo previous width setting */
	(*LineWidthH)->v = oldLineWidth.h;
	(*LineWidthH)->h = oldLineWidth.v;
	PicComment(SetLineWidth, sizeof(TLineWidth), (Handle) LineWidthH);

	(*LineWidthH)->v = 1;
	(*LineWidthH)->h = scale;
	PicComment(SetLineWidth, sizeof(TLineWidth), (Handle) LineWidthH);

}

int MAC_init()
{
	FontInfo          fontInfo;
	Integer           count;
	Rect              r;
	PicHandle         ph;
	struct termentry *t = term_tbl + term;
	WHERE("MAC_init");

	if ((count = GraphWindSetup()) == 0)
	{
		myerrorout("ERROR: graph windows full - delete a graph and type showplot()");
		return (1);
	}
	else if(count < 0)
	{
		myerrorout("ERROR: cannot create graph window. Out of memory (?)");
		return (1);
	}
	ph = myPic[count-1];
	r = (*ph)->picFrame;

	GetFontInfo(&fontInfo);
	fontInfo.leading = 0; /* use 0 leading */
	fontInfo.widMax = CharWidth((Integer) 'M');
	t->v_char = fontInfo.ascent + fontInfo.descent +
		fontInfo.leading;
	t->h_char = fontInfo.widMax;
	t->v_tic = t->h_tic = fontInfo.widMax;
	t->xmax = r.right - r.left;
	t->ymax = r.bottom - r.top;

	MacFrameTop = t->ymax;
	MacThickness = 1;

	if(LineWidthH == (TLineWidthHdl) 0)
	{
		LineWidthH = (TLineWidthHdl) NewHandle(sizeof(TLineWidth));
		(*LineWidthH)->v = 1;
		(*LineWidthH)->h = 1;
		PicComment(SetLineWidth, sizeof(TLineWidth), (Handle) LineWidthH);
	}

	return (0);
}


int MAC_graphics()
{
	return (0);
}


void MAC_text()
{
	if (ScreenDumpFile != (char *) 0)
	{
		ScreenDumpError = SCREENDUMPUNDEFINED;
	}
	GraphOver();
	if(PauseAfterPlot)
	{
		GraphWindPause();
	}
}

static Pattern LinePatterns[8] =
{ /* all slant strip patterns */
	0xee, 0xee, 0xee, 0xee, 0xee, 0xee, 0xee, 0xee, /* x-axis (dashed)*/
	0xff, 0xff, 0xff, 0x00, 0xff, 0xff, 0xff, 0x00, /* y-axis (dashed) */
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, /* border (solid) */
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, /* solid */
	0xee, 0x77, 0xbb, 0xdd, 0xee, 0x77, 0xbb, 0xdd, /* slant 11101110 */
	0xaa, 0x66, 0x33, 0x99, 0xaa, 0x66, 0x33, 0x99, /* slant 10101010 */
	0x88, 0x44, 0x22, 0x11, 0x88, 0x44, 0x22, 0x11, /* slant 10001000 */
	0xe8, 0x74, 0x3a, 0x1d, 0x8e, 0x47, 0xa3, 0xd1  /* slant 11101000 */
};

void MAC_linetype(linetype)
long            linetype;
{
	Integer    heavy = (10*linetype >= UNITWIDTH);
	double     thickness = (heavy) ?
		(double) (MAXLINETYPE * (linetype/MAXLINETYPE)) / (double) UNITWIDTH : 1.0;
	WHERE("MAC_linetype");

	MacThickness = (Integer) ceil(thickness);
	if(heavy)
	{
		linetype %= MAXLINETYPE;
	}
	PenSize(MacThickness, MacThickness);
	if(linetype < XAXISLINETYPE)
	{
		linetype = BORDERLINETYPE;
	}
	linetype = ((linetype < 0) ? linetype: linetype % 5) - XAXISLINETYPE;

#ifdef MPW3_2
	PenPat((ConstPatternParam) LinePatterns + linetype);
#else /*MPW3_2*/
	PenPat((Pattern) LinePatterns + linetype);
#endif /*MPW3_2*/

	setLineScale(HAIRLINESCALE);
}

void MAC_move(x, y)
unsigned long            x, y;
{
	MoveTo(x, MacFrameTop - y);
}


void MAC_vector(x, y)
unsigned long       x, y;
{
	Integer    adjust = (MacThickness-1)/2;

	if(adjust)
	{  /* makesure line is centered */
		Move(-adjust,-adjust);
		LineTo(x-adjust, MacFrameTop - y - adjust);
		Move(adjust,adjust);
	}
	else
	{
		LineTo(x, MacFrameTop - y);
	}
}

void MAC_point(x, y, number)
long            x, y;
long            number;
{
	if(number < 0 || number % POINT_TYPES == 6)
	{
		setLineScale(1); /* make sure dot is big enough to see */
	}
	do_point(x, y, number);
}

void MAC_str_text(str)		/* write string to screen while still in graphics mode */
char            str[];
{
	DrawText(str, 0, strlen(str));
}


 /* write character to screen while still in graphics mode */
void MAC_chr_text(char chr)
{
	DrawText(&chr, 0, 1);
}


void MAC_reset()
{
}

#endif /*MACINTOSH*/

#ifdef DJGPP
/* For djgpp compiler for 32 bit protected mode MSDOS */
#include <grx.h>
#include <pc.h>
#include <osfcn.h>
#include "pcxdump.h"
#define FALSE 0
#define TRUE 1
int            DJG_init(void);
int            DJG_graphics(void);
void           DJG_text(void);
void           DJG_reset(void);
void           DJG_linetype(long /*linetype*/);
void           DJG_move(unsigned long /*x*/, unsigned long /*y*/);
void           DJG_vector(unsigned long /*x*/, unsigned long /*y*/);

#ifdef GNUPLOT3
int            DJG_text_angle(int /*ang*/);
int            DJG_justify_text(enum JUSTIFY /*mode*/);
void           DJG_put_text(long /*x*/,long /*y*/,char * /*str*/);
#else /*GNUPLOT3*/
void           DJG_str_text(char * /*str*/);
void           DJG_chr_text(char /*c*/);
#endif /*GNUPLOT3*/

#ifdef UNDEFINED__
#define NLINETYPES 5
#else /*UNDEFINED__*/
#define NLINETYPES 2 /*temporary fix*/
#endif /*UNDEFINED__*/
/*
	These patterns are defined, but currently (940811) do not get used by
	the grx library
*/
static unsigned char pattern[5][2] =
{
/*        (0, 0),
	(10, 10),
	(20, 10),
	(20, 20),
	(20, 15)   */
	(0, 0),
	(10, 10),
	(10, 10),
	(10, 10),
	(10, 10)
};
	/* 0xffff, 0xaaaa, 0x3333, 0x3f3f, 0x0f0f */
static int patternLength[] =
{
	0, 2, 2, 2, 2
};

static int graphics_on = FALSE;
int currentx, currenty;
int pc_angle;

#define RIGHTOFFSET 5

/*
   These defines are essentially ignored since info comes from calls
   to functions
*/

#define DJG_XMAX 640
#define DJG_YMAX 480

#define DJG_VCHAR 12
#define DJG_HCHAR 8
#define DJG_VTIC 8
#define DJG_HTIC 8


GrFont           *FontInfo = (GrFont *) 0;
int               TextGrMode = GR_80_25_text, GraphicGrMode = -1, pc_ymax;
GrLineOption      LineOptions;
GrTextOption      TextOptions;



void DJG_text(void)
{
	if (graphics_on)
	{
		if (ScreenDumpFile != (char *) 0)
		{
			ScreenDumpError = pcxdump(ScreenDumpFile);
		}
		graphics_on = FALSE;
		if(ISATTY && PauseAfterPlot)
		{
			(void) getkey(); /* wait for CR */
		}
	}
	if( TextGrMode >= 0)
	{
		GrSetMode( TextGrMode );
	}
	else
	{
		GrSetMode(GR_80_25_text);
	}
}

void DJG_reset()
{
}

int DJG_init()
{
	char             *path;
	struct termentry *t = term_tbl + term;

	if( !graphics_on && TextGrMode < 0)
	{
		TextGrMode = GrCurrentMode();
	}

	GrSetMode( GR_default_graphics );
	GrClearScreen( GrWhite() );

	if( GraphicGrMode < 0)
	{
		t->xmax = GrSizeX() - RIGHTOFFSET;
		t->ymax = GrSizeY();
		pc_ymax = t->ymax;

		GrSetFontPath( get_dataPath() );
		FontInfo = GrLoadFont("cour14.fnt");
		if( FontInfo == (GrFont *) 0) 
		{
			myerrorout("ERROR: graphics font file not found");
			return(1);
		}
		t->v_char = FontInfo->fnt_height;
		t->h_char = FontInfo->fnt_width;
		t->v_tic = FontInfo->fnt_width;
		t->h_tic = FontInfo->fnt_width;

		GraphicGrMode =  GrCurrentMode();

		TextOptions.txo_font = FontInfo;
		TextOptions.txo_xmag = 1;
		TextOptions.txo_ymag = 1;
		TextOptions.txo_fgcolor.v = GrBlack();
		TextOptions.txo_bgcolor.v = GrNOCOLOR;
		TextOptions.txo_direct = GR_TEXT_RIGHT;
		TextOptions.txo_xalign = GR_ALIGN_LEFT;
		TextOptions.txo_yalign = GR_ALIGN_BOTTOM;
		TextOptions.txo_chrtype = GR_BYTE_TEXT;
	}

	graphics_on = TRUE;

	LineOptions.lno_color = GrBlack();
	LineOptions.lno_width = 1;
	LineOptions.lno_pattlen = 0; /* continuous lines by default */


	return(0);
}


int DJG_graphics()
{
	return(0);
}


void DJG_linetype(linetype)
long linetype;
{
	long       heavy = (10*linetype >= UNITWIDTH);
	double     thickness = (heavy) ?
		(double) (MAXLINETYPE * (linetype/MAXLINETYPE)) / (double) UNITWIDTH : 1.0;

	if(heavy)
	{
		linetype %= UNITWIDTH;
	}
	if(thickness <=1.0)
	{
		heavy = 0;
	}
	if(linetype < 0)
	{
		if(linetype == XAXISLINETYPE || linetype == YAXISLINETYPE)
		{
			linetype = 2;
		}
		else
		{
			linetype = 0;
		}
	}
	else
	{
		linetype %= NLINETYPES;
	}

	LineOptions.lno_width = (heavy)? 3 : 1;
	LineOptions.lno_pattlen = 0;
	if(linetype)
	{
		LineOptions.lno_pattlen = patternLength[linetype];
		LineOptions.lno_dashpat = pattern[linetype];
	}
}

#ifdef GNUPLOT3
void DJG_put_text(x,y,str)
long  x, y;
char *str;
{
	struct termentry *t = term_tbl + term;

	strcpy(buf,str);
	outtextxy((int) x,(int) (t->ymax-1-y),buf);
}
#else /*GNUPLOT3*/
void DJG_str_text(str)
char *str;
{
	GrDrawString( str, strlen(str), currentx, pc_ymax-1-currenty, &TextOptions);
}

void DJG_chr_text(char c)
{
	char cbuf[2];
	cbuf[0] = c;
	cbuf[1] = '\0';
	DJG_str_text(cbuf);
}
#endif /*GNUPLOT3*/

void DJG_move(x,y)
unsigned long x,y;
{
	currentx = x;
	currenty = y;
}

void DJG_vector(x,y)
unsigned long x,y;
{
	GrCustomLine(currentx, pc_ymax-1-currenty, x, pc_ymax-1-y, &LineOptions);
	currentx = x;
	currenty = y;
}


#ifdef GNUPLOT3
int DJG_text_angle(ang)
int ang;
{
	int size = getmaxy() > 599 ? 2 : 1;

	pc_angle = ang;

	switch (ang)
	{
	  case 0:
		settextstyle(DEFAULT_FONT,HORIZ_DIR,size);
		break;
	  case 1:
		settextstyle(DEFAULT_FONT,VERT_DIR,size);
		break;
	}
	return TRUE;
}

int DJG_justify_text(mode)
enum JUSTIFY mode;
{
	switch(mode)
	{
	  case LEFT :
		settextjustify(LEFT_TEXT,CENTER_TEXT);
		break;
	  case CENTRE :
		settextjustify(CENTER_TEXT,CENTER_TEXT);
		break;
	  case RIGHT:
		settextjustify(RIGHT_TEXT,CENTER_TEXT);
		break;
	}
	return TRUE;
}
#endif /*GNUPLOT3*/

#endif /* DJGPP */

#ifdef BCPP
/* For Borland C++ compiler */


/* GNUPLOT - pc.trm */
/*
 * Copyright (C) 1990
 *
 * Permission to use, copy, and distribute this software and its
 * documentation for any purpose with or without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Permission to modify the software is granted, but not the right to
 * distribute the modified code.  Modifications are to be distributed
 * as patches to released version.
 *
 * This software  is provided "as is" without express or implied warranty.
 *
 * This file is included by ../term.c.
 *
 * This terminal driver supports:
 *  Under Microsoft C
 *      cga, egabios, egalib, vgabios, hercules, corona325, att
 *  Under Turboc C
 *      egalib, vgalib, vgamono, svga, mcga, cga, hercules, att
 *
 * AUTHORS
 *  Colin Kelley, Thomas Williams, William Wilson, Russell Lang
 *
 * send your comments or suggestions to (pixar!info-gnuplot@sun.com).
 *
 */

#include <graphics.h>
#include <conio.h>
#include <dos.h>

char *getenv();

#define FALSE 0
#define TRUE 1
int            BC_init(void);
int            BC_graphics(void);
void           BC_text(void);
void           BC_reset(void);
void           BC_linetype(long /*linetype*/);
void           BC_move(unsigned long /*x*/, unsigned long /*y*/);
void           BC_vector(unsigned long /*x*/, unsigned long /*y*/);

#ifdef GNUPLOT3
int            BC_text_angle(int /*ang*/);
int            BC_justify_text(enum JUSTIFY /*mode*/);
void           BC_put_text(long /*x*/,long /*y*/,char * /*str*/);
#else /*GNUPLOT3*/
void           BC_str_text(char * /*str*/);
void           BC_chr_text(char /*c*/);
#endif /*GNUPLOT3*/

static int              g_driver = DETECT, g_mode, g_error;

static struct text_info tinfo; /* So we can restore starting text mode. */

static char buf[80];       /* kludge since EGA.LIB is compiled SMALL */

/*
static int pattern[] =
{
	0xffff, 0x0f0f, 0xffff, 0xaaaa, 0x3333, 0x3f3f, 0x0f0f
};
*/
static int pattern[] =
{
	0xffff, 0xaaaa, 0x3333, 0x3f3f, 0x0f0f
};

static int graphics_on = FALSE;
int startx, starty;

int pc_angle;


void BC_text()
{
	if (ScreenDumpFile != (char *) 0)
	{
		ScreenDumpError = SCREENDUMPUNDEFINED;
	}
	if (graphics_on)
	{
		graphics_on = FALSE;
		if(ISATTY && PauseAfterPlot)
		{
			(void) getch(); /* wait for CR */
		}
	}
	restorecrtmode();

	/*
	textmode(tinfo.currmode);
	clrscr();
	*/
}

void BC_reset()
{
/*
	closegraph();
	textmode(tinfo.currmode);
    clrscr();
*/
}

#define BC_XMAX 640
#define BC_YMAX 480

#define BC_VCHAR 12
#define BC_HCHAR 8
#define BC_VTIC 8
#define BC_HTIC 8

#ifdef GNUPLOT3
int BC_text_angle(ang)
int ang;
{
	int size = getmaxy() > 599 ? 2 : 1;

	pc_angle = ang;

	switch (ang)
	{
	  case 0:
		settextstyle(DEFAULT_FONT,HORIZ_DIR,size);
		break;
	  case 1:
		settextstyle(DEFAULT_FONT,VERT_DIR,size);
		break;
	}
	return TRUE;
}

int BC_justify_text(mode)
enum JUSTIFY mode;
{
	switch(mode)
	{
	  case LEFT :
		settextjustify(LEFT_TEXT,CENTER_TEXT);
		break;
	  case CENTRE :
		settextjustify(CENTER_TEXT,CENTER_TEXT);
		break;
	  case RIGHT:
		settextjustify(RIGHT_TEXT,CENTER_TEXT);
		break;
	}
	return TRUE;
}
#endif /*GNUPLOT3*/

int BC_init()
{
	char             *path;
	struct termentry *t = term_tbl + term;

	if( g_driver == DETECT)
	{
		gettextinfo(&tinfo);
		path = get_dataPath();
		path[strlen(path)-1] = '\0'; /* wipe out trailing '\' */
		initgraph(&g_driver,&g_mode,(char far *) path);
		path[strlen(path)] = '\\';
	}

	if(g_driver<0)
	{
		switch (g_driver)
		{
		  case -2:
			myerrorout("ERROR: Graphics card not detected.");
			break;
		  case -3:
			myerrorout("ERROR: appropriate graphics driver cannot be found.");
			break;
		  case -5:
			myerrorout("ERROR: Insufficient memory to load graphics driver.");
			break;
		  default:
			myerrorout("ERROR: problem in initializing graphics");
		}
		g_driver = DETECT;
		return (1);
	} /*if(g_driver<0)*/

	t->xmax = getmaxx() + 1;
	t->ymax = getmaxy() + 1;
	if( getmaxy() > 599)
	{
		t->v_char = 2*BC_VCHAR;
		t->h_char = 2*BC_HCHAR;
		t->v_tic = 2*BC_VTIC;
		t->h_tic = 2*BC_HTIC;
	}
	else {
		t->v_char = BC_VCHAR;
		t->h_char = BC_HCHAR;
		t->v_tic = BC_VTIC;
		t->h_tic = BC_HTIC;
	}
	restorecrtmode();
	textmode(tinfo.currmode);

	return(0);
}


int BC_graphics()
{
	graphics_on = TRUE;
	gettextinfo(&tinfo);
	setgraphmode(g_mode);
	settextjustify(LEFT_TEXT,BOTTOM_TEXT);

	if (getmaxy() > 599)		/* Double the tic/font sizes. */
	{
		settextstyle(DEFAULT_FONT,HORIZ_DIR,2);
	}
	else
	{
		settextstyle(DEFAULT_FONT,HORIZ_DIR,1);
	}
	return(0);
}


void BC_linetype(linetype)
long linetype;
{
	long       heavy = (10*linetype >= UNITWIDTH);

	if(heavy)
	{
		linetype %= UNITWIDTH;
	}

	if(linetype < 0)
	{
		if(linetype == XAXISLINETYPE || linetype == YAXISLINETYPE)
		{
			linetype = 2;
		}
		else
		{
			linetype = 0;
		}
	}
	else
	{
		linetype %= 5;
	}

	setlinestyle(USERBIT_LINE,pattern[linetype],
				 (heavy) ? THICK_WIDTH : NORM_WIDTH);
}

#ifdef GNUPLOT3
void BC_put_text(x,y,str)
long  x, y;
char *str;
{
	struct termentry *t = term_tbl + term;

	strcpy(buf,str);
	outtextxy((int) x,(int) (t->ymax-1-y),buf);
}
#else /*GNUPLOT3*/
void BC_str_text(str)
char *str;
{
	outtext(str);
}

void BC_chr_text(char c)
{
	char cbuf[2];
	cbuf[0] = c;
	cbuf[1] = '\0';
	BC_str_text(cbuf);
}
#endif /*GNUPLOT3*/

void BC_move(x,y)
unsigned long x,y;
{
	moveto((int) x,(int)(getmaxy()-y));
}

void BC_vector(x,y)
unsigned long x,y;
{
	lineto((int) x,(int) (getmaxy()-y));
}


#endif /*BCPP*/

#ifdef POSTSCRIPT
/* GNUPLOT - post.trm */
/*
 * Copyright (C) 1990
 *
 * Permission to use, copy, and distribute this software and its
 * documentation for any purpose with or without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Permission to modify the software is granted, but not the right to
 * distribute the modified code.  Modifications are to be distributed
 * as patches to released version.
 *
 * This software  is provided "as is" without express or implied warranty.
 *
 * This file is included by ../term.c.
 *
 * This terminal driver supports:
 *     postscript
 *
 * AUTHORS
 *  Russell Lang
 *
 * send your comments or suggestions to (pixar!info-gnuplot@sun.com).
 *
 * The 'postscript' driver produces landscape output 10" wide and 7" high.
 * To get a smaller epsf output use 'set size 0.5,0.5',
 * 'set term postscript portrait', make only one plot per file
 * and change the first line of the postscript file from
 * '%!PS-Adobe-2.0' to '%!PS-Adobe-2.0 EPSF-2.0'
 * To change font to Times-Roman and font size to 20pts use
 * 'set term postscript "Times-Roman" 20'.
 */

/* PostScript driver by Russell Lang, rjl@monu1.cc.monash.edu.au */
#define PS_FONTSIZE  12
#define PS_PORTRAIT  1

void           PS_options(void);
int            PS_init(void);
int            PS_graphics(void);
void           PS_text(void);
void           PS_reset(void);
void           PS_linetype(long /*linetype*/);
void           PS_move(unsigned long /*x*/, unsigned long /*y*/);
void           PS_point(long /*x*/, long /*y*/, long /*point*/);
void           PS_vector(unsigned long /*x*/, unsigned long /*y*/);
void           PS_str_text(char * /*str*/);
void           PS_chr_text(char /*c*/);

char           ps_font[] = "Courier" ; /* name of font */
int            ps_fontsize = PS_FONTSIZE; /* size of font in pts */
long           ps_portrait = PS_PORTRAIT;				 /* vertical page */
long           ps_color = 0;
int            ps_page=0;			/* page count */
int            ps_path_count=0; 	/* count of lines in path */
int            ps_ang=0;			/* text angle */
enum JUSTIFY   ps_justify=LEFT;	/* text is flush left */

FILE *outfile = stdout;

char *PS_header[] =
{
	"/vpt2 vpt 2 mul def\n",
	"/hpt2 hpt 2 mul def\n",
	/* flush left show */
	"/Lshow { currentpoint stroke moveto\n",
	"  0 vshift rmoveto show } def\n",
	/* flush right show */
	"/Rshow { currentpoint stroke moveto\n",
	"  dup stringwidth pop neg vshift rmoveto show } def\n",
	/* centred show */
	"/Cshow { currentpoint stroke moveto\n",
	"  dup stringwidth pop -2 div vshift rmoveto show } def\n",
	/* Dash or Color Line */
	"/DL { Color {setrgbcolor [] 0 setdash pop}\n",
	" {pop pop pop 0 setdash} ifelse } def\n",
	/* set line width */
	"/LW	{ gnulinewidth mul /mylinewidth exch def } def\n",
	/* Border Lines */
/*	"/BL { stroke mylinewidth 2 mul setlinewidth } def\n", */
	"/BL { stroke mylinewidth setlinewidth } def\n",
	/* Axes Lines */
	"/AL { stroke mylinewidth 2 div setlinewidth } def\n",
	/* Plot Lines */
	"/PL { stroke mylinewidth setlinewidth } def\n",
	/* Line Types */
	"/LTa { AL [1 dl 2 dl] 0 setdash 0 0 0 setrgbcolor } def\n", /* axes */
	"/LTb { BL [] 0 0 0 DL } def\n", /* border */
	"/LT0 { PL [] 0 1 0 DL } def\n",
	"/LT1 { PL [4 dl 2 dl] 0 0 1 DL } def\n",
	"/LT2 { PL [2 dl 3 dl] 1 0 0 DL } def\n",
	"/LT3 { PL [1 dl 1.5 dl] 1 0 1 DL } def\n",
	"/LT4 { PL [5 dl 2 dl 1 dl 2 dl] 0 1 1 DL } def\n",
	"/LT5 { PL [4 dl 3 dl 1 dl 3 dl] 1 1 0 DL } def\n",
	"/LT6 { PL [2 dl 2 dl 2 dl 4 dl] 0 0 0 DL } def\n",
	"/LT7 { PL [2 dl 2 dl 2 dl 2 dl 2 dl 4 dl] 1 0.3 0 DL } def\n",
	"/LT8 { PL [2 dl 2 dl 2 dl 2 dl 2 dl 2 dl 2 dl 4 dl] 0.5 0.5 0.5 DL } def\n",
	"/M {moveto} def\n",
	"/L {lineto} def\n",
	"/P { stroke [] 0 setdash\n", /* Point */
	"  currentlinewidth 2 div sub moveto\n",
	"  0 currentlinewidth rlineto  stroke } def\n",
	"/D { stroke [] 0 setdash  2 copy  vpt add moveto\n", /* Diamond */
	"  hpt neg vpt neg rlineto  hpt vpt neg rlineto\n",
	"  hpt vpt rlineto  hpt neg vpt rlineto  closepath  stroke\n",
	"  P  } def\n",
	"/A { stroke [] 0 setdash  vpt sub moveto  0 vpt2 rlineto\n", /* Plus (Add) */
	"  currentpoint stroke moveto\n",
	"  hpt neg vpt neg rmoveto  hpt2 0 rlineto stroke\n",
	"  } def\n",
	"/B { stroke [] 0 setdash  2 copy  exch hpt sub exch vpt add moveto\n", /* Box */
	"  0 vpt2 neg rlineto  hpt2 0 rlineto  0 vpt2 rlineto\n",
	"  hpt2 neg 0 rlineto  closepath  stroke\n",
	"  P  } def\n",
	"/C { stroke [] 0 setdash  exch hpt sub exch vpt add moveto\n", /* Cross */
	"  hpt2 vpt2 neg rlineto  currentpoint  stroke  moveto\n",
	"  hpt2 neg 0 rmoveto  hpt2 vpt2 rlineto stroke  } def\n",
	"/T { stroke [] 0 setdash  2 copy  vpt 1.12 mul add moveto\n", /* Triangle */
	"  hpt neg vpt -1.62 mul rlineto\n",
	"  hpt 2 mul 0 rlineto\n",
	"  hpt neg vpt 1.62 mul rlineto  closepath  stroke\n",
	"  P  } def\n",
	"/S { 2 copy A C} def\n",	/* Star */
	NULL
};

#define PS_XOFF	50	/* page offset in pts */
#define PS_YOFF	250


#define PS_XMAX 5760  /* 8" */
#define PS_YMAX 7200  /* 10" */

#define PS_XMAXP (PS_XMAX-2*PS_SC*PS_XOFF) /* for portrait mode */
#define PS_YMAXP 3600 /* 5" for portrait mode */

#define PS_XMAXL PS_YMAX /* for landscape mode */
#define PS_YMAXL PS_XMAX /* for landscape mode */

#define PS_XLAST (PS_XMAX - 1)
#define PS_YLAST (PS_YMAX - 1)

#define PS_VTIC (PS_XMAX/80)
#define PS_HTIC (PS_XMAX/80)

#define PS_SC (10)				/* scale is 1pt = 10 units */
#define	PS_LW (0.5*PS_SC)		/* linewidth = 0.5 pts */

#define PS_VCHAR (PS_FONTSIZE*PS_SC)		/* default is PS_FONTSIZE point characters */
#define PS_HCHAR (PS_FONTSIZE*PS_SC*6/10)


#ifdef DO_PS_OPTIONS

void PS_options()
{
	extern struct value *const_express();
	extern double real();

	if (!END_OF_COMMAND)
	{
		if (almost_equals(c_token,"p$ortrait"))
		{
			ps_portrait=TRUE;
			c_token++;
		}
		else if (almost_equals(c_token,"l$andscape"))
		{
			ps_portrait=FALSE;
			c_token++;
		}
		else if (almost_equals(c_token,"d$efault"))
		{
			ps_portrait=FALSE;
			ps_color=FALSE;
			strcpy(ps_font,"Courier");
			ps_fontsize = PS_FONTSIZE;
			c_token++;
		}
	}

	if (!END_OF_COMMAND)
	{
		if (almost_equals(c_token,"m$onochrome"))
		{
			ps_color=FALSE;
			c_token++;
		}
		else if (almost_equals(c_token,"c$olor"))
		{
			ps_color=TRUE;
			c_token++;
		}
	}

	if (!END_OF_COMMAND && isstring(c_token))
	{
		quote_str(ps_font,c_token);
		c_token++;
	}

	if (!END_OF_COMMAND)
	{
		/* We have font size specified */
		struct value a;
		ps_fontsize = (int)real(const_express(&a));
		c_token++;
		term_tbl[term].v_char = (unsigned int)(ps_fontsize*PS_SC);
		term_tbl[term].h_char = (unsigned int)(ps_fontsize*PS_SC*6/10);
	}

	sprintf(term_options,"%s %s \"%s\" %d",
		ps_portrait ? "portrait" : "landscape",
		ps_color ? "color" : "monochrome",ps_font,ps_fontsize);
}
#endif /*DO_PS_OPTIONS*/


int PS_init()
{
	struct termentry *t = &term_tbl[term];
	int i;
	double      xscale = xsize/PS_SC, yscale = ysize/PS_SC;

	t->v_char = (unsigned int)(ps_fontsize*PS_SC);
	t->h_char = (unsigned int)(ps_fontsize*PS_SC*6/10);
	t->v_tic  = PS_VTIC;
	t->h_tic  = PS_HTIC;
	t->xmax   = (ps_portrait) ? PS_XMAXP: PS_XMAXL;
	t->ymax   = (ps_portrait) ? PS_YMAXP : PS_YMAXL;

	outfile = (PLOTFILE == STDOUT) ? stdout : PLOTFILE;
	if(PutProlog)
	{
		ps_page = 0;
		fprintf(outfile,"%%!PS-Adobe-2.0\n");
		fprintf(outfile,"%%%%Creator: gnuplot\n");
		fprintf(outfile,"%%%%DocumentFonts: %s\n", ps_font);
		fprintf(outfile,"%%%%BoundingBox: %d %d ", PS_XOFF,PS_YOFF);

		if (ps_portrait)
		{
			fprintf(outfile,"%d %d\n",
					(int)(xscale*t->xmax+0.5+PS_XOFF),
					(int)(yscale*t->ymax+0.5+PS_YOFF) );
		}
		else
		{
			fprintf(outfile,"%d %d\n",
					(int)(yscale*t->ymax+0.5+PS_XOFF),
					(int)(xscale*t->xmax+0.5+PS_YOFF) );
		}

		fprintf(outfile,"%%%%Pages: (atend)\n");
		fprintf(outfile,"%%%%EndComments\n");
		fprintf(outfile,"/gnudict 40 dict def\ngnudict begin\n");
		fprintf(outfile,"/Color %s def\n",ps_color ? "true" : "false");
		fprintf(outfile,"/gnulinewidth %.3f def\n",PS_LW);
		fprintf(outfile,"/vshift %d def\n", (int)(t->v_char)/(-3));
		fprintf(outfile,"/dl {%d mul} def\n",PS_SC); /* dash length */
		fprintf(outfile,"/hpt %.1f def\n",PS_HTIC/2.0);
		fprintf(outfile,"/vpt %.1f def\n",PS_VTIC/2.0);
		for ( i=0; PS_header[i] != NULL; i++)
		{
			fprintf(outfile,"%s",PS_header[i]);
		}
		fprintf(outfile,"end\n");
		fprintf(outfile,"%%%%EndProlog\n");
	}
	return(0);
}


int PS_graphics()
{
	struct termentry *t = &term_tbl[term];
	double       xscale = xsize/PS_SC, yscale = ysize/PS_SC;

	ps_page++;
	fprintf(outfile,"%%%%Page: %d %d\n",ps_page,ps_page);
	fprintf(outfile,"gnudict begin\n");
	fprintf(outfile,"gsave\n");
	if (ps_portrait)
	{
		fprintf(outfile,"%d %d translate\n",PS_XOFF,PS_YOFF);
	    fprintf(outfile,"%.3f %.3f scale\n", xscale, yscale);
	}
	else
	{
	    fprintf(outfile,"%.3f %.3f scale\n", yscale, xscale);
	    fprintf(outfile,"90 rotate\n");
		fprintf(outfile,"%d %d translate\n", 0 ,-(PS_XMAX));
	}
	fprintf(outfile,"0 setgray\n");
	fprintf(outfile,"/%s findfont %d ", ps_font, (t->v_char) );
	fprintf(outfile,"scalefont setfont\n");
	fprintf(outfile,"newpath\n");
	ps_path_count = 0;

	return(0);
}


void PS_text()
{
	if (ScreenDumpFile != (char *) 0)
	{
		ScreenDumpError = SCREENDUMPUNDEFINED;
	}
	ps_path_count = 0;
	fprintf(outfile,"stroke\ngrestore\nend\nshowpage\n");
}


void PS_reset()
{
	fprintf(outfile,"%%%%Trailer\n");
	fprintf(outfile,"%%%%Pages: %d\n",ps_page);
}


void PS_linetype(linetype)
long linetype;
{
	char *line = "aab012345678";
	long         heavy = (10*linetype >= UNITWIDTH);
	double       thickness = (heavy) ? (double) linetype / (double) UNITWIDTH: 1.0;

	if(heavy)
	{
		linetype %= MAXLINETYPE;
	}

	if(linetype < 0)
	{
		if(linetype < XAXISLINETYPE)
		{
			linetype = BORDERLINETYPE;
		}
	}
	else
	{
		linetype %= 9;
	}
	linetype -= XAXISLINETYPE;
	fprintf(outfile,"%5.3f LW\n",thickness);
	fprintf(outfile,"LT%c\n", line[linetype]);
	ps_path_count = 0;
}


void PS_move(x,y)
unsigned long x,y;
{
	fprintf(outfile,"%ld %ld M\n", x, y);
	ps_path_count += 1;
}


void PS_vector(x,y)
unsigned long x,y;
{
	fprintf(outfile,"%ld %ld L\n", x, y);
	ps_path_count += 1;
	if (ps_path_count >= 400)
	{
		fprintf(outfile,"currentpoint stroke moveto\n");
		ps_path_count = 0;
	}
}


void PS_str_text(str)
char str[];
{
	char ch;

	putc('(',outfile);
	ch = *str++;
	while(ch!='\0')
	{
		if ( (ch=='(') || (ch==')') || (ch=='\\') )
		{
			putc('\\',outfile);
		}
		putc(ch,outfile);
		ch = *str++;
	}
	switch(ps_justify)
	{
	  case LEFT :
		fprintf(outfile,") Lshow\n");
		break;
	  case CENTRE :
		fprintf(outfile,") Cshow\n");
		break;
	  case RIGHT :
		fprintf(outfile,") Rshow\n");
		break;
	}
	ps_path_count = 0;
}

void PS_chr_text(char ch)
{
	char str[2];

	str[0] = ch;
	str[1] = '\0';
	PS_str_text(str);
}

#ifdef GNUPLOT3
PS_put_text(x,y,str)
unsigned int x, y;
char *str;
{
	char ch;
	PS_move(x,y);
	if (ps_ang != 0)
	{
		fprintf(outfile,"currentpoint gsave translate %d rotate 0 0 moveto\n"
				,ps_ang*90);
	}
	putc('(',outfile);
	ch = *str++;
	while(ch!='\0')
	{
		if ( (ch=='(') || (ch==')') || (ch=='\\') )
			putc('\\',outfile);
		putc(ch,outfile);
		ch = *str++;
	}
	switch(ps_justify)
	{
	  case LEFT :
		fprintf(outfile,") Lshow\n");
		break;
	  case CENTRE :
		fprintf(outfile,") Cshow\n");
		break;
	  case RIGHT :
		fprintf(outfile,") Rshow\n");
		break;
	}
	if (ps_ang != 0)
	{
		fprintf(outfile,"grestore\n");
	}
	ps_path_count = 0;
}

int PS_text_angle(ang)
int ang;
{
	ps_ang=ang;
	return TRUE;
}

int PS_justify_text(mode)
enum JUSTIFY mode;
{
	ps_justify=mode;
	return TRUE;
}


#endif /*GNUPLOT3*/
/* postscript point routines */
void PS_point(x,y,number)
long          x,y;
long          number;
{
	char *point = "PDABCTSP";
	number %= POINT_TYPES;
 	if (number < -1)
	{
		number = -1;			/* negative types are all 'dot' */
	}
	fprintf(outfile,"%ld %ld %c\n", x, y, point[number+1]);
	ps_path_count = 0;
}

#endif /*POSTSCRIPT*/


#ifdef TEK
/*	For Tektronix 4014 terminal */
#ifdef UNDEFINED__
#define TEK40XMAX 1024
#define TEK40YMAX 780
#endif /*UNDEFINED__*/

#define TEK40XMAX 974
#define TEK40YMAX 780

#define TEK40XLAST (TEK40XMAX - 1)
#define TEK40YLAST (TEK40YMAX - 1)

#define TEK40VCHAR		25
#define TEK40HCHAR		14
#define TEK40VTIC		14 /* was 11, kb*/
#define TEK40HTIC		14 /* was 11, kb*/

#define HX 0x20			/* bit pattern to OR over 5-bit data */
#define HY 0x20
#define LX 0x40
#define LY 0x60

#define LOWER5 31
#define UPPER5 (31<<5)

int            TEK40init(void);
int            TEK40graphics(void);
void           TEK40text(void);
void           TEK40reset(void);
void           TEK40linetype(long /*linetype*/);
void           TEK40move(unsigned long /*x*/, unsigned long /*y*/);
void           TEK40vector(unsigned long /*x*/, unsigned long /*y*/);
void           TEK40str_text(char * /*str*/);
void           TEK40chr_text(char /*c*/);

static char *Terminal = (char *) 0;

int TEK40init()
{
	struct termentry *t = term_tbl + term;

	t->v_char = TEK40VCHAR;
	t->h_char = TEK40HCHAR;
	t->v_tic  = TEK40VTIC;
	t->h_tic  = TEK40HTIC;
	t->xmax   = TEK40XMAX;
	t->ymax   = TEK40YMAX;

	if(Terminal == '\0')
	{
		Terminal = TERMINAL;
	}

	return (0);
}


int TEK40graphics()
{
	WHERE("TEK40graphics");

	if(PLOTFILE == STDOUT && Terminal != (char *) 0 &&
	   strcmp(Terminal,"xterm") == 0)
	{ /* switch on xterm tek window */
		fprintf(stdout,"\033\133\077\063\070\150\033\070");
	}

	fprintf((PLOTFILE == STDOUT) ? stdout : PLOTFILE,"\033\014");
	return (0);
	/*                   1
		1. clear screen
	*/
}


void TEK40text()
{
	TEK40move(0L, 24L);
	fprintf((PLOTFILE == STDOUT) ? stdout : PLOTFILE,"\037");
	if (ScreenDumpFile != (char *) 0)
	{
		ScreenDumpError = SCREENDUMPUNDEFINED;
	}
	if(PLOTFILE == STDOUT)
	{
		if(ISATTY && PauseAfterPlot)
		{
			getchar(); /* wait for CR */
		}
		if(strcmp(Terminal,"xterm") == 0)
		{/* switch back to normal vt100 window */
			fprintf((PLOTFILE == STDOUT) ? stdout : PLOTFILE,"\033\003");
		}
	}

	/*                   1
		1. into alphanumerics
	*/
}


void TEK40linetype(linetype)
long            linetype;
{
	linetype = (linetype < 0) ? 0 : (linetype % UNITWIDTH) % 5;
	fprintf((PLOTFILE == STDOUT) ? stdout : PLOTFILE,
			"\035\033%c", '\140' + linetype);
}


void TEK40move(x, y)
unsigned long   x, y;
{
	putc('\035', (PLOTFILE == STDOUT) ? stdout : PLOTFILE);	/* into graphics */
	TEK40vector(x, y);
}


void TEK40vector(x, y)
unsigned long   x, y;
{
	FILE     *plotfile = (PLOTFILE == STDOUT) ? stdout : PLOTFILE;

	putc((HY | (y & UPPER5) >> 5), plotfile);
	putc((LY | (y & LOWER5)), plotfile);
	putc((HX | (x & UPPER5) >> 5), plotfile);
	putc((LX | (x & LOWER5)), plotfile);
}


void TEK40str_text(str)
char            str[];
{
	fprintf((PLOTFILE == STDOUT) ? stdout : PLOTFILE,"\037%s", str);
}


void TEK40chr_text(char chr)
{
	fprintf((PLOTFILE == STDOUT) ? stdout : PLOTFILE,"\037%c", chr);
}

void TEK40reset()
{
}

#endif /*TEK*/

#ifdef DUMB
/* For character based printer plot */
/* Modified version of code from GNUPLOT - dumb.trm */
/*
 * Copyright (C) 1991
 *
 * Permission to use, copy, and distribute this software and its
 * documentation for any purpose with or without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Permission to modify the software is granted, but not the right to
 * distribute the modified code.  Modifications are to be distributed
 * as patches to released version.
 *
 * This software  is provided "as is" without express or implied warranty.
 *
 * This file is included by ../term.c.
 *
 * This terminal driver supports:
 *   DUMB terminals
 *
 * AUTHORS
 *   Francois Pinard, 91-04-03
 *           INTERNET: pinard@iro.umontreal.ca
 * Modified for use in MacAnova by Christopher Bingham, kb@umnstat.stat.umn.edu
 * send your comments or suggestions to (pixar!info-gnuplot@sun.com).
 *
 */

#define DUMB_AXIS_CONST '\1'
#define DUMB_BORDER_CONST '\2'

#define DUMB_XMAX 79
#define DUMB_YMAX 24
#define DUMB_VCHAR 1
#define DUMB_HCHAR 1
#define DUMB_VTIC  0
#define DUMB_HTIC  0

void           dumb_set_pixel(long /*x*/,long /*y*/,long /*v*/,long /*p*/);

int            DUMB_init(void);
int            DUMB_graphics(void);
void           DUMB_text(void);
void           DUMB_reset(void);
void           DUMB_linetype(long /*linetype*/);
void           DUMB_move(unsigned long /*x*/, unsigned long /*y*/);
void           DUMB_point(long /*x*/, long /*y*/, long /*point*/);
void           DUMB_vector(unsigned long /*x*/, unsigned long /*y*/);
void           DUMB_str_text(char * /*str*/);
void           DUMB_chr_text(char /*c*/);

#define Trash  NAMEFORTRASH
#define GDUMBMATRIX   0
#define GDUMBPRIORITY 1
#define NTRASH        2

static Symbolhandle Trash = (Symbolhandle) 0;
static char **dumb_matrix = (char **) 0;    /* matrix of characters */
static char **dumb_priority = (char **) 0;/* matrix of priority at each position */

static char dumb_pen;                /* current character used to draw */
static unsigned long dumb_x;                    /* current X position */
static unsigned long dumb_y;                    /* current Y position */
static unsigned long dumb_xmax = DUMB_XMAX;
static unsigned long dumb_ymax = DUMB_YMAX;

#define DUMB_PIXEL(x,y) (*dumb_matrix)[dumb_xmax*(y)+(x)]

void dumb_set_pixel(x,y,v,p)
long x,y,v,p;
{
	long      j = dumb_xmax*y+x;

	assert(x >= 0);
	assert(x < dumb_xmax);
	assert(y >= 0);
	assert(y < dumb_ymax);
	if (p >= (*dumb_priority)[j]) /* kb changed from > */
    {
		(*dumb_matrix)[j] = v;
		(*dumb_priority)[j] = p;
    }
}

int DUMB_init()
{
	struct termentry *t = term_tbl + term;
	long npixels;
	WHERE("DUMB_init");

	Trash = GarbInstall(NTRASH);
	if(Trash == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	t->v_char = DUMB_VCHAR;
	t->h_char = DUMB_HCHAR;
	t->v_tic  = DUMB_VTIC;
	t->h_tic  = DUMB_HTIC;
	t->xmax = dumb_xmax = (SCREENWIDTH > 15) ? (unsigned long) SCREENWIDTH - 1:
		(unsigned long) 79;

	t->ymax = dumb_ymax = (SCREENHEIGHT > 6) ? (unsigned long) SCREENHEIGHT :
		(unsigned long) 24;

	npixels = dumb_xmax * dumb_ymax;
	if(!getScratch(dumb_matrix,GDUMBMATRIX,npixels + 1,char) ||
	   !getScratch(dumb_priority,GDUMBPRIORITY,npixels,char))
	{
		goto errorExit;
	}
	mylockhandle(dumb_matrix);
	mylockhandle(dumb_priority);

	return (0);

  errorExit:
	emptyTrash();
	Trash = (Symbolhandle) 0;
	return (1);

}


int DUMB_graphics ()
{
	long npixels = dumb_xmax * dumb_ymax, i;

	for(i=0;i<npixels;i++)
	{
		(*dumb_matrix)[i] = ' ';
		(*dumb_priority)[i] = 0;
	}
	(*dumb_matrix)[npixels] = '\0';
	return (0);
}


void DUMB_text ()
{
	long          y, l;
	char          ctemp;

	NLINES = -2;
	for (y = dumb_ymax - 1; y >= 0; y--)
    {
		for (l = dumb_xmax; l > 0 && DUMB_PIXEL (l - 1, y) == ' '; l--)
		{
			;
		}
		assert(l >= 0);
		assert(l <= dumb_xmax);
		assert(y >= 0);
		assert(y < dumb_ymax);
		ctemp = DUMB_PIXEL(l,y);
		DUMB_PIXEL(l,y) = '\0'; /* close off line */
		fmyprint((*dumb_matrix) + y*dumb_xmax,PLOTFILE);
		DUMB_PIXEL(l,y) = ctemp;

		if(PLOTFILE != STDOUT || y > 0)
		{
			fmyeol(PLOTFILE);
		}
		if(INTERRUPT != INTNOTSET)
		{
			break;
		}
    } /*for (y = dumb_ymax - 1; y >= 0; y--)*/
	if(PLOTFILE == STDOUT) /* standard output in MacAnova */
	{
#ifndef MACINTOSH
		if(ISATTY)
		{
			if(INTERRUPT == INTNOTSET && PauseAfterPlot)
			{
				(void) getchar(); /* wait for CR */
			}
			else
			{
				myeol();
			}
			if (SPOOLFILE != (FILE *) 0)
			{ /* make sure spool file gets new line */
				fmyeol(SPOOLFILE);
			}
		}
		else
		{
			myeol();
		}
#else /*MACINTOSH*/
		myeol();
#endif /*MACINTOSH*/
		NLINES = 0;
	}
	DUMB_reset();
}


void DUMB_reset()
{
	WHERE("DUMB_reset");

	emptyTrash();
	Trash = (Symbolhandle) 0;
}


#define NLINETYPE 5
static char pen_type[NLINETYPE+3] =
{
	DUMB_AXIS_CONST, DUMB_AXIS_CONST, DUMB_BORDER_CONST, '.', '-', '=', 'o', ':'
};

void DUMB_linetype(linetype)
long linetype;
{
	if(linetype < XAXISLINETYPE)
	{
		linetype = BORDERLINETYPE;
	}
	linetype = ((linetype < 0) ?
				linetype : (linetype % MAXLINETYPE) % NLINETYPE) -
				XAXISLINETYPE;
	dumb_pen = pen_type[linetype];
}


void DUMB_move(x, y)
unsigned long x, y;
{
	WHERE("DUMB_move");

	dumb_x = x;
	dumb_y = y;
}


#define DOT     '.'

/* modified by kb.  Original always used point % 26 + 'A' unless point = -1 */
static    char point_chars[POINT_TYPES] = { '+', 'O', 'X', '%', '#', '*', '.'};

void DUMB_point(x,y,point)
long          x, y, point;
{
	long      c;

	if(point == -1)
	{
		c = DOT;
	}
	else if(point <= 26)
	{
		c = point_chars[point % POINT_TYPES];
	}
	else
	{
		c  = point % 26 + 'A';
	}
	assert(x >= 0);
	assert(x < dumb_xmax);
	assert(y >= 0);
	assert(y < dumb_ymax);
	dumb_set_pixel (x, y, c, 4L);
}

#define CORNER       '+'
#define HORIZBORDER  '-'
#define VERTBORDER   '|'
#define HORIZAXIS    '.'
#define VERTAXIS     ':'

void DUMB_vector(x,y)
unsigned long x,y;
{
	char      pen, pen1;
	long      priority;
	long      xtmp,ytmp;
	long      x1 = (long) dumb_x, y1 = (long) dumb_y; /* start position */
	long      x2 = (long) x, y2 = (long) y; /*end position */
	long      delta, dx = x2 - x1, dy = y2 - y1;
	double    ratio;
	WHERE("DUMB_vector");

	if (labs (dy) > labs (dx))
    {
		switch (dumb_pen)
		{
		  case DUMB_AXIS_CONST:
			pen = VERTAXIS;
			pen1 = CORNER;
			priority = 1;
			break;

		  case DUMB_BORDER_CONST:
			pen = VERTBORDER;
			pen1 = CORNER;
			priority = 2;
			break;

		  default:
			pen = dumb_pen;
			pen1 = pen;
			priority = 3;
			break;

		} /*switch (dumb_pen)*/
		assert(x1 >= 0);
		assert(x1 < dumb_xmax);
		assert(y1 >= 0);
		assert(y1 < dumb_ymax);
		dumb_set_pixel (x1, y1, pen1, priority);
		ratio = (double) dx/(double) labs(dy);
		for (delta = 1; delta < labs (dy); delta++)
		{
			xtmp = (long) (x1 + delta*ratio + 0.5);
			ytmp = y1 + delta * ((y2 > y1) ? 1 : -1);
			assert(xtmp >= 0);
			assert(xtmp < dumb_xmax);
			assert(ytmp >= 0);
			assert(ytmp < dumb_ymax);
			dumb_set_pixel (xtmp,ytmp,pen, priority);
		}
		assert(x2 >= 0);
		assert(x2 < dumb_xmax);
		assert(y2 >= 0);
		assert(y2 < dumb_ymax);
		dumb_set_pixel (x2, y2, pen1, priority);
    } /*if (labs (dy) > labs (dx))*/
	else if (labs (dx) > labs (dy))
    {
		switch (dumb_pen)
		{
		  case DUMB_AXIS_CONST:
			pen = HORIZAXIS;
			pen1 = CORNER;
			priority = 1;
			break;

		  case DUMB_BORDER_CONST:
			pen = HORIZBORDER;
			pen1 = CORNER;
			priority = 2;
			break;

		  default:
			pen = dumb_pen;
			pen1 = pen;
			priority = 3;
			break;
		} /*switch (dumb_pen)*/
		assert(x1 >= 0);
		assert(x1 < dumb_xmax);
		assert(y1 >= 0);
		assert(y1 < dumb_ymax);
		dumb_set_pixel (x1, y1, pen1, priority);
		ratio = (double) dy/(double) labs(dx);
		for (delta = 1; delta < labs (dx); delta++)
		{
			xtmp = x1 + delta * ((x2 > x1) ? 1 : -1);
			ytmp = (long) (y1 + delta*ratio + 0.5);
			assert(xtmp >= 0);
			assert(xtmp < dumb_xmax);
			assert(ytmp >= 0);
			assert(ytmp < dumb_ymax);
			dumb_set_pixel (xtmp, ytmp,	pen, priority);
		} /*for (delta = 1; delta < labs (dx); delta++)*/
		assert(x2 >= 0);
		assert(x2 < dumb_xmax);
		assert(y2 >= 0);
		assert(y2 < dumb_ymax);
		dumb_set_pixel (x2, y2, pen1, priority);
	} /*if (labs (dy) > labs (dx)){}else if (labs (dx) > labs (dy))*/
	else
    { /* labs (dx) == labs (dy) */
		switch (dumb_pen)
		{
		  case DUMB_AXIS_CONST:	/* zero length axis */
			pen = CORNER;
			priority = 1;
			break;

		  case DUMB_BORDER_CONST: /* zero length border */
			pen = CORNER;
			priority = 2;
			break;

		  default:
			pen = dumb_pen;
			pen1 = pen;
			priority = 3;
			break;
		} /*switch (dumb_pen)*/

		for (delta = 0; delta <= labs (dx); delta++)
		{
			xtmp = (long) (x1 + delta * sign (dx) + 0.5);
			ytmp = (long) (y1 + delta * sign (dy) + 0.5);
			assert(xtmp >= 0);
			assert(xtmp < dumb_xmax);
			assert(ytmp >= 0);
			assert(ytmp < dumb_ymax);
			dumb_set_pixel (xtmp, ytmp,	pen, priority);
		}/*for (delta = 0; delta <= labs (dx); delta++)*/
    }
	dumb_x = x;
	dumb_y = y;
}


void DUMB_str_text(str)
char *str;
{
	long length;
	long x = (long) dumb_x, y = (long) dumb_y;/* correct vertical position */
	WHERE("DUMB_str_text");

	if (ScreenDumpFile != (char *) 0)
	{
		ScreenDumpError = SCREENDUMPUNDEFINED;
	}
	length = strlen(str);
	if (x + length > dumb_xmax)
	{
		x = (dumb_xmax > length) ? dumb_xmax - length : 0 ;
	}

	for (; x < dumb_xmax && *str; x++, str++)
	{
		assert(x >= 0);
		assert(x < dumb_xmax);
		assert(y >= 0);
		assert(y < dumb_ymax);
		dumb_set_pixel (x, y, *str, 5L);
	}
}


void DUMB_chr_text (char c)
{
	long     x = (long) dumb_x, y = (long) dumb_y;

	assert(x >= 0);
	assert(x < dumb_xmax);
	assert(y >= 0);
	assert(y < dumb_ymax);
	dumb_set_pixel (x, y, c, 5L);
}

#endif /*DUMB*/

static void UNKNOWN(void)
{
}

#undef UNDEFINED__
#ifdef UNDEFINED__
int UNKNOWN_nullint()
{
	return (0);
}

void UNKNOWN_nullvoid()
{

}
#endif /*UNDEFINED__*/

void do_point(x, y, number)
long            x, y;
long            number;
{
	register long   htic, vtic;
	register struct termentry *t;

	number %= POINT_TYPES;
 	if (number < -1)
	{
		number = -1;			/* negative types are all 'dot' */
	}
	t = &term_tbl[term];
	htic = (t->h_tic / 2);		/* should be in term_tbl[] in later version */
	vtic = (t->v_tic / 2);

	/*
	  if ( x < t->h_tic || y < t->v_tic || x >= t->xmax-t->h_tic ||
	  	y >= t->ymax-t->v_tic )
	  return;		*/		/* add clipping in later version maybe */

	switch ((int) number)
	{
	  case -1: /* dot */
	  case 6:
		(*t->move) (x, y);
		(*t->vector) (x, y);
		break;

	  case 0:					/* do diamond */
		(*t->move) (x - htic, y);
		(*t->vector) (x, y - vtic);
		(*t->vector) (x + htic, y);
		(*t->vector) (x, y + vtic);
		(*t->vector) (x - htic, y);
		(*t->move) (x, y);
		(*t->vector) (x, y);
		break;
	  case 1:					/* do plus */
		(*t->move) (x - htic, y);
		(*t->vector) (x - htic, y);
		(*t->vector) (x + htic, y);
		(*t->move) (x, y - vtic);
		(*t->vector) (x, y - vtic);
		(*t->vector) (x, y + vtic);
		break;
	  case 2:					/* do box */
		(*t->move) (x - htic, y - vtic);
		(*t->vector) (x + htic, y - vtic);
		(*t->vector) (x + htic, y + vtic);
		(*t->vector) (x - htic, y + vtic);
		(*t->vector) (x - htic, y - vtic);
		(*t->move) (x, y);
		(*t->vector) (x, y);
		break;
	  case 3:					/* do X */
		(*t->move) (x - htic, y - vtic);
		(*t->vector) (x - htic, y - vtic);
		(*t->vector) (x + htic, y + vtic);
		(*t->move) (x - htic, y + vtic);
		(*t->vector) (x - htic, y + vtic);
		(*t->vector) (x + htic, y - vtic);
		break;
	  case 4:					/* do triangle */
		(*t->move) (x, y + (4 * vtic / 3));
		(*t->vector) (x - (4 * htic / 3), y - (2 * vtic / 3));
		(*t->vector) (x + (4 * htic / 3), y - (2 * vtic / 3));
		(*t->vector) (x, y + (4 * vtic / 3));
		(*t->move) (x, y);
		(*t->vector) (x, y);
		break;
	  case 5:					/* do star */
		(*t->move) (x - htic, y);
		(*t->vector) (x - htic, y);
		(*t->vector) (x + htic, y);
		(*t->move) (x, y - vtic);
		(*t->vector) (x, y - vtic);
		(*t->vector) (x, y + vtic);
		(*t->move) (x - htic, y - vtic);
		(*t->vector) (x - htic, y - vtic);
		(*t->vector) (x + htic, y + vtic);
		(*t->move) (x - htic, y + vtic);
		(*t->vector) (x - htic, y + vtic);
		(*t->vector) (x + htic, y - vtic);
		break;
	}
}


void horstring(x, y, str, justify)
long            x, y;
char           *str;
long            justify;
{				/* print a horizontal string centered at x,y */

	struct termentry *t;
	long            left;

	if(*str)
	{
		t = &term_tbl[term];
		/* attempt to center character vertically on y */
		if(term == PSTERM)
		{
			y += PS_SC; /* move up 1 point */
		}
		else if(term == STANDARDTERM)
		{
#ifdef TEK
			y -= (t->v_char)/2 - 4;
#else
			y -= (t->v_char)/2 - 1;
#endif
		}

		if(y >= 0 && y < t->ymax)
		{
			if(justify == CENTRE)
			{
				left = x - (strlen(str) * (t->h_char)) / 2;
			}
			else if(justify == LEFT)
			{
				left = x;
			}
			else
			{
				left = x - strlen(str)*t->h_char;
			}

			if (left < 1)
			{
				left = 1;
			}
			if (left + strlen(str) * t->h_char > t->xmax - 1)
			{
				left = t->xmax - 1 - strlen(str) * t->h_char;
			}
			if (left < 1)
			{
				left = 1;
			}
			(*t->move) (left, y);
			(*t->str_text) (str);
		}
	}
}

void vertstring(x, y, str)
long            x, y;
char           *str;
{				/* print a vertical string centered at x.y */

	struct termentry *t;
	long            i, base;

	if(*str)
	{
		t = &term_tbl[term];
		base = y + ((strlen(str) - 1) * (long) (t->v_char)) / (long) 2;
		for (i = 0; i < strlen(str); i++)
		{
			(*t->move) (x, base - i * (t->v_char));
			(*t->chr_text) (str[i]);
		}
	}
}

/*
 * term_tbl[] contains an entry for each terminal.  "unknown" must be the
 *   first, since term is initialized to 0.
 *   Then (for MacAnova) must come the standard device entry
 *   Then must come the PostScript entry
 */
struct termentry term_tbl[] =
{
	{
		"unknown",
		100, 100, 1, 1,	1, 1,
		(int (*) (void)) UNKNOWN, /*UNKNOWN_init()*/
		(void (*) (void)) UNKNOWN, /*UNKNOWN_reset()*/
		(void (*) (void)) UNKNOWN, /*UNKNOWN_text()*/
		(int (*) (void)) UNKNOWN, /*UNKNOWN_graphics()*/
		(void (*) (unsigned long, unsigned long)) UNKNOWN, /*UNKNOWN_move()*/
		(void (*) (unsigned long, unsigned long)) UNKNOWN, /*UNKNOWN_vector()*/
		(void (*) (long)) UNKNOWN, /*UNKNOWN_linetype()*/
		(void (*) (char)) UNKNOWN, /*UNKNOWN_chr_text()*/
		(void (*) (char *)) UNKNOWN, /*UNKNOWN_str_text()*/
		(void (*) (long, long, long)) UNKNOWN /*UNKNOWN_point()*/
	}

#ifdef MACINTOSH
	,
	{
		"Macintosh", MAC_XMAX, MAC_YMAX, MAC_VCHAR, MAC_HCHAR,
		MAC_VTIC, MAC_HTIC, MAC_init, MAC_reset,
		MAC_text, MAC_graphics, MAC_move, MAC_vector,
		MAC_linetype, MAC_chr_text, MAC_str_text, MAC_point
	}
#endif /*MACINTOSH*/

#ifdef TEK
	,
	{
		"tek40xx", TEK40XMAX, TEK40YMAX, TEK40VCHAR, TEK40HCHAR,
		TEK40VTIC, TEK40HTIC, TEK40init, TEK40reset,
		TEK40text, TEK40graphics, TEK40move, TEK40vector,
		TEK40linetype, TEK40chr_text, TEK40str_text, do_point
	}
#endif  /*TEK*/

#ifdef BCPP
	,
	{
		"BC", BC_XMAX, BC_YMAX, BC_VCHAR, BC_HCHAR,
		BC_VTIC, BC_HTIC, BC_init, BC_reset,
		BC_text, BC_graphics, BC_move, BC_vector,
		BC_linetype, BC_chr_text, BC_str_text, do_point
	}
#endif /*BCPP*/

#ifdef DJGPP
	,
	{
		"DJG", DJG_XMAX, DJG_YMAX, DJG_VCHAR, DJG_HCHAR,
		DJG_VTIC, DJG_HTIC, DJG_init, DJG_reset,
		DJG_text, DJG_graphics, DJG_move, DJG_vector,
		DJG_linetype, DJG_chr_text, DJG_str_text, do_point
	}
#endif /*DJGPP*/

#ifdef DUMB
    ,
	{
		"dumb", DUMB_XMAX, DUMB_YMAX, DUMB_VCHAR,DUMB_HCHAR,
		DUMB_VTIC, DUMB_HTIC, DUMB_init, DUMB_reset,
		DUMB_text, DUMB_graphics, DUMB_move, DUMB_vector,
		DUMB_linetype, DUMB_chr_text, DUMB_str_text, DUMB_point
	}
#endif /* DUMB*/
#ifdef POSTSCRIPT
    ,
	{
		"postscript",PS_XMAX,PS_YMAX,PS_VCHAR,PS_HCHAR,
		PS_VTIC, PS_HTIC, PS_init, PS_reset,
		PS_text, PS_graphics, PS_move, PS_vector,
		PS_linetype, PS_chr_text, PS_str_text, PS_point
	}
#endif /*POSTSCRIPT*/
};

long       TermCount =  sizeof(term_tbl)/sizeof(struct termentry);
