/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Trans
#else
#define __SEG__ Trans
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <math.h>

#ifndef	HUGEDBL
#define HUGEDBL 1e38
#endif /*HUGEDBL*/


extern Symbolhandle RInstall();
extern double   lgamma();
extern long	    symbolSize();
extern long     dupDims();
extern long     argOK();
extern void     noData();
extern long     matchKey();
double          atanh1();

/* 
  These codes should be consecutive integers in the same order as operations
  in opEntry1
*/
#define ISQRT    101
#define IEXP     102
#define ILOG     103
#define ILOG10   104
#define IABS     105
#define ISIN     106
#define ICOS     107
#define ITAN     108
#define ISINH    109
#define ICOSH    110
#define ITANH    111
#define IATANH   112
#define IATAN    113
#define IATAN2   114
#define IASIN    115
#define IACOS    116
#define IHYPOT   117
#define ILGAMMA  118
#define IFLOOR   119
#define ICEILING 120
#define IROUND   121
#define INBITS   122
#define OPBASE   ISQRT /*code of first function in list*/

typedef struct opEntry1
{
	char *name;
	short iop;
	short nargs;
	double (*f)(double);
} opEntry1;

static opEntry1  Ops[] =
{
	"sqrt",    ISQRT,    1,    sqrt,
	"exp",     IEXP,     1,    exp,
	"log",     ILOG,     1,    log,
	"log10",   ILOG10,   1,    commonLog,
	"abs",     IABS,     1,    fabs,
	"sin",     ISIN,     1,    sin,
	"cos",     ICOS,     1,    cos,
	"tan",     ITAN,     1,    tan,
	"sinh",    ISINH,    1,    sinh,
	"cosh",    ICOSH,    1,    cosh,
	"tanh",    ITANH,    1,    tanh,
	"atanh",   IATANH,   1,    atanh1,
	"atan",    IATAN,   -2,    atan,
	"atan2",   IATAN2,   2,    (double (*) (double)) atan2,
	"asin",    IASIN,    1,    asin,
	"acos",    IACOS,    1,    acos,
	"hypot",   IHYPOT,   2,    (double (*) (double)) pythag,
	"lgamma",  ILGAMMA,  1,    lgamma,
	"floor",   IFLOOR,   1,    floor,
	"ceiling", ICEILING, 1,    ceil,
	"round",   IROUND,  -2,    (double (*) (double)) round,
	"nbits",   INBITS,   1,    nbits,
	(char *) 0, 0,       0,    (double (*) (double)) 0
};

static Symbolhandle doTrans(Symbolhandle arg1,Symbolhandle arg2,
							double *params, unsigned long control,
							unsigned long *status)
{
	Symbolhandle     result = (Symbolhandle) 0;
	long            i, tot1;
	long            nargs = 1;
	long            iop;
	long            noGood;
	long            toAngles, fromAngles;
	long            ndims;
	double        (*f1)(double), (*f2)(double,double);
	double          x1, x2, z, hugeExp;
	WHERE("doTrans");
	
	hugeExp = log(HUGEDBL);
	
	iop = control & OPMASK;

	f1 = Ops[iop-OPBASE].f;
	f2 = (double (*) (double, double)) f1;
	
	toAngles = (iop == IACOS) || (iop == IASIN) ||
		(iop == IATAN) || (iop == IATAN2);
	fromAngles = (iop == ICOS) || (iop == ISIN) || (iop == ITAN);

	if(iop == IROUND)
	{
		x2 = pow(10.0,params[0]);
		nargs = 2;
	}
	else if(iop == IATAN2 || iop == IHYPOT)
	{
		nargs = 2;
	}
	
	/* find total number of data values & copy dimensions */
	tot1 = symbolSize(arg1);

	/* set up result array */
	if(control & REUSELEFT)
	{
		result = arg1;
	}
	else if(control & REUSERIGHT)
	{
		result = arg2;
	}
	else
	{
		result = RInstall(SCRATCH,tot1);
	}
	
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	setNCLASS(result, -1); /* make sure it's a variate */

	ndims = NDIMS(arg1);
	if(nargs == 2 && iop != IROUND && ndims > NDIMS(arg2))
	{ /* atan2 or hypot */
		ndims = NDIMS(arg2);
	}
	
	setNDIMS(result,ndims);
	for (i = 1; i <= ndims; i++)
	{
		setDIM(result,i,DIMVAL(arg1,i));
	}
	
	for (i = 0; i < tot1; i++)
	{
#ifdef MACINTOSH
		if(timetocheck(i) && interrupted() != INTNOTSET)
		{
			goto errorExit;
		}
#endif /*MACINTOSH*/
		x1 = DATAVALUE(arg1,i);
		noGood = 0;
		if (isMissing(x1) ||
			(iop != IROUND && nargs == 2 &&
			 isMissing(x2 = DATAVALUE(arg2,i))))
		{
			*status |= FOUNDMISSING;
			noGood = 1;
		}
		else
		{
			switch (iop)
			{
			  case ISQRT:
				if (x1 < 0.0)
				{
					*status |= FOUNDNEG;
					noGood = 1;
				}
				break;
			  case ILOG:
			  case ILOG10:
			  case ILGAMMA:
				if (x1 <= 0.0)
				{
					*status |= FOUNDNONPOS;
					noGood = 1;
				}
				break;
			  case IEXP:
			  case ISINH:
			  case ICOSH:
				if (x1 > hugeExp)
				{
					*status |= FOUNDOVERFLOW;
				}
				break;
			  case IATANH:
				if(fabs(x1) >= 1.0)
				{
					*status |= FOUNDDOMERR;
					noGood = 1;
				}
				break;
			  case IASIN:
			  case IACOS:
				if (fabs(x1) > 1.0)
				{
					*status |= FOUNDDOMERR;
					noGood = 1;
				}
				break;
				
			  case INBITS:
				if(x1 < 0 || x1 > MAXBITVALUE || x1 != floor(x1))
				{
					*status |= FOUNDDOMERR;
					noGood = 1;
				}
				break;
				
			  default:
				break;
			} /*switch (iop)*/
		}
		if(noGood)
		{
			setMissing(z);
		} /*if(noGood)*/
		else
		{
			if(fromAngles)
			{
				x1 *= ANGLES2RADIANS;
			}
			z = (nargs == 1) ? (*f1)(x1) : (*f2)(x1,x2);
			if(toAngles)
			{
				z /= ANGLES2RADIANS;
			}
		} /*if(noGood){}else{}*/
		
		DATAVALUE(result,i) = z;
	} /*for (i = 0; i < tot1; i++)*/

	return (result);
	
  errorExit:
	Removesymbol(result);
	return (0);
}

Symbolhandle    Transform(list)
Symbolhandle    list;
{
	long            nargs, margs;
	struct opEntry1 *op;
	int             iop;
	unsigned long   status, control = 0;
	long            reply;
	Symbolhandle    arg1, arg2, result = (Symbolhandle) 0;
	double          param[1];
	WHERE("Transform");
	
	*OUTSTR = '\0';

	for(op = Ops;op->iop != 0;op++)
	{
		if(strcmp(FUNCNAME,op->name) == 0)
		{
			break;
		}
	}
	iop = op->iop;
	if(iop == 0)
	{ /* should never happen */
		sprintf(OUTSTR,"ERROR: unrecognized transformation %s()", FUNCNAME);
		goto errorExit;
	}
	
	margs = op->nargs;

	nargs = NARGS(list);
/* atan and round may have 1 or 2 arguments, hypot and atan2 require 2 */
	if(margs > 0 && nargs != margs || margs < 0 && nargs > labs(margs))
	{
		badNargs(FUNCNAME,margs);
		return (0);
	}

	arg1 = COMPVALUE(list, 0);
	if(!argOK(arg1, 0, (nargs > 1) ? 1 : 0))
	{
		goto errorExit;
	}

	if(!isReal(arg1))
	{
		sprintf(OUTSTR,
				"ERROR: argument to %s must be REAL or structure of REALs",
				FUNCNAME);
		goto errorExit;
	}
	if(isStruc(arg1))
	{
		control |= LEFTRECUR;
	}

	if(nargs == 2)
	{ /* must be atan, atan2, hypot, or round */
		if(iop == IATAN)
		{
			iop = IATAN2;
		}
		
		arg2 = COMPVALUE(list,1);
		if(!argOK(arg2,(iop != IROUND) ? 0 : REAL,2))
		{
			goto errorExit;
		}
		if(iop != IROUND && !isReal(arg2))
		{
			sprintf(OUTSTR,
					"ERROR: argument 2 to %s must be REAL or structure of REALs",
					FUNCNAME);
			goto errorExit;
		}

		if(iop == IATAN2 || iop == IHYPOT)
		{
			if(isStruc(arg2))
			{
				control |= RIGHTRECUR;
			}
			
			reply = treecmp(arg1,arg2,control|NEARSTRICT);
			if(reply != 0)
			{
				sprintf(OUTSTR,
						"ERROR: %s of arguments to %s do not match",
						(reply == BADSTRUC) ?
						"structure shape" : "dimensions",
						FUNCNAME);
			}			
		}
		else if(iop == IROUND)
		{						/* second argument must be integer */
			param[0] = DATAVALUE(arg2,0);
			if(!isScalar(arg2) || isMissing(param[0]) ||
			   param[0] != floor(param[0]))
			{
				sprintf(OUTSTR,
						"ERROR: 2nd argument to %s must be single integer",
						FUNCNAME);
			}
		}
		else					/* should not be reached */
		{
			sprintf(OUTSTR,"ERROR: program error in Transform()");
		}
			
		if(*OUTSTR)
		{
			goto errorExit;
		}
	}
	else
	{
		param[0] = 0.0;
		arg2 = (Symbolhandle) 0;
	}
	if(isscratch(NAME(arg1)))
	{
		control |= REUSELEFT;
		COMPVALUE(list,0) = (Symbolhandle) 0;
	}
	else if((iop == IATAN2 || iop == IHYPOT) &&
			isscratch(NAME(arg2)))
	{
		control |= REUSERIGHT;
		COMPVALUE(list,1) = (Symbolhandle) 0;
	}
	
	status = 0;
	result = doRecur2(doTrans, arg1, arg2, param, control | iop, &status);

	if(status & FOUNDMISSING)
	{
		sprintf(OUTSTR,"WARNING: missing values in argument(s) to %s",FUNCNAME);
		putOUTSTR();
	}
	
	if(status & FOUNDOVERFLOW)
	{
		sprintf(OUTSTR,"WARNING: possible overflow on %s",FUNCNAME);
		putOUTSTR();
	}
	
	if(status & FOUNDNONPOS)
	{
		sprintf(OUTSTR,"WARNING: %s of non-positive argument set to MISSING",
				FUNCNAME);
	}
	else if(status & FOUNDNEG)
	{
		sprintf(OUTSTR,"WARNING: %s of negative argument set to MISSING",
				FUNCNAME);
	}
	else if(status & FOUNDDOMERR)
	{
		sprintf(OUTSTR,
				"WARNING: %s of %s set to MISSING", FUNCNAME,
				(iop == INBITS) ? "illegal value" : "x with |x| > 1");
	}
	putOUTSTR();

	return (result);
	
  errorExit:
	putOUTSTR();

	return (0);
}

double commonLog(x)
double      x;
{
	return (log10(x));
}

double atanh1(x)
double      x;
{
	return(0.5*log((1 + x)/(1 - x)));
}

double round(x,param)
double      x,param;
{
	double      sgn = (x > 0.0) ? 1.0 : -1.0;
	WHERE("round");

	return(sgn * floor(param * fabs(x) + .5) / param);
}

#ifdef UNDEFINED__
double nbits(Double x)
{
	register unsigned long     ulx = (unsigned long) x;
	register int               n = 0;

	while(ulx != 0)
	{
		n += 1 & ulx;
		ulx >>= 1;
	}
	return ((double) n);
}
#else /*UNDEFINED__*/
/* slightly faster version of preceding */
static char bitcounts[] = 
{
	0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4
};

double nbits(double x)
{
	register unsigned long     ulx;
	register char              n = '\0';
	register char             *bc = bitcounts;
	
	for(ulx = (unsigned long) x; ulx != 0; ulx >>= 4)
	{
		n += bc[ulx & 0x0f];
	}
	return ((double) n);
}
	
#endif /*UNDEFINED__*/
