/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Transpose
#else
#define __SEG__ Transpose
#endif /*MPW3*/
#endif /*MPW*/
#include "globals.h"
#include "dbug.h"

extern Symbolhandle   Install();
extern void           badNargs();
extern long           argOK();
extern void           badType();
extern long           dupDims();

static Symbolhandle    doTranspose(Symbolhandle arg, double *params,
								   unsigned long control, unsigned long *status)
{
	Symbolhandle result = (Symbolhandle) 0;
	long         type = TYPE(arg);
	long         size, j, rplace, aplace;
	long         kr;
	long         ndims;
	long         dim[MAXDIMS+1],ii[MAXDIMS+1], incs[MAXDIMS+1];
	long         place[MAXDIMS+1];
	double     **axH, **rxH;
	double      *ax, *rx;
	char       **asH, **rsH;
	char        *as, *rs;
	WHERE("doTranspose");
	
	*status = 0;
	size = dupDims(dim,arg);

	result = Install(SCRATCH,type);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	if(type != CHAR)
	{
		axH = DATA(arg);
		TMPHANDLE = mygethandle(myhandlelength((char **) axH));
		setDATA(result,(double **) TMPHANDLE);
		rxH = DATA(result);
	}
	else
	{		
		asH = STRING(arg);
		TMPHANDLE = mygethandle(myhandlelength(asH));
		setSTRING(result,TMPHANDLE);
		rsH = STRING(result);
	}
	if(TMPHANDLE == (char **) 0)
	{
		goto errorExit;
	}

	if(type != CHAR)
	{
		ax = *axH;
		rx = *rxH;
	}
	else
	{
		as = *asH;
		rs = *rsH;
	}
	
	if (NDIMS(arg) == 1)
	{
		dim[0] = 2;
		dim[2] = 1;
	}
	ndims = dim[0];
	setNDIMS(result,ndims);

	incs[0] = size;
	for(j = 1;j<=ndims;j++)
	{
		ii[j-1] = 0;
		place[j-1] = 0;
		setDIM(result,j,dim[ndims-j+1]);
		incs[j] = incs[j-1]/dim[ndims-j+1];
	}
	
	/* rearrange as transpose */

	rplace = aplace = 0;
	for(kr=0;kr<size;kr++)
	{
		if(type != CHAR)
		{
			rx[rplace++] = ax[aplace];
		}
		else
		{/* string equivalent of rplace++*/
			strcpy(rs + rplace, as + aplace);
			rplace += strlen(rs + rplace) + 1;
		}
		
/* the following is essentially equivalent to ndims nested for loops */
		for (j = 1; j <= ndims; j++)
		{
			if(ii[j-1] == 0)
			{
				place[j-1] = aplace;
			}
			if((ii[j-1] += incs[j]) < incs[j-1])
			{
				if(type != CHAR)
				{
					aplace += incs[j];
				}
				else
				{ /* string equivalent of aplace += incs[j] */
					aplace = skipStrings(as + aplace,incs[j]) - as;
				}
				break;
			} /*if((ii[j-1] += incs[j]) < incs[j-1])*/
			ii[j-1] = 0;
			aplace = place[j-1];
		} /*for (j = 1; j <= ndims; j++)*/
	} /*for(kr=0;kr<size;kr++)*/
	
	return(result);
	
  errorExit:
	Removesymbol(result);
	return (0);
}

/*
  returns non-zero if and only if all non-structure components are non-null
   and have type REAL, LOGIC, or CHAR
*/
static checktypes(Symbolhandle arg)
{
	long        type;
	long        i, ncomps;
	long        answer = 0;

	if(!isNull(arg))
	{
		ncomps = NCOMPS(arg);
		
		if(isStruc(arg))
		{
			for(i=0;i<ncomps;i++)
			{
				answer = checktypes(COMPVALUE(arg,i));
				if(!answer)
				{
					break;
				}
			}
		} /*if(isStruc(arg))*/
		else if((type = TYPE(arg)) == REAL || type == LOGIC || type == CHAR)
		{
			answer = 1;
		} /*if(isStruc(arg)){}else{}*/
	} /*if(!isNull(arg))*/

	return (answer);
	
}

Symbolhandle    transpose(list)
Symbolhandle    list;
{

	/* make a transpose of a matrix */

	Symbolhandle    result, arg;
	unsigned long   control = 0, status;
	
	WHERE("transpose");

	*OUTSTR = '\0';
	
	if(NARGS(list) != 1)
	{
		badNargs(FUNCNAME,1);
		goto errorExit;
	}
	
	arg = COMPVALUE(list, 0);
	if(!argOK(arg,(long) 0, 0))
	{
		goto errorExit;
	}

	if (!checktypes(arg))
	{
		if(isStruc(arg))
		{
			sprintf(OUTSTR,
					"ERROR: all components of structure must be REAL, LOGICAL, or CHARACTER for %s",
					FUNCNAME);
		}
		else
		{
			badType(FUNCNAME,-TYPE(arg),0);
		}
		goto errorExit;
	}
	if(isStruc(arg))
	{
		control |= LEFTRECUR;
	}
	
	result = doRecur1(doTranspose,arg,(double *) 0,control,&status);

	if(result == 0)
	{
		goto errorExit;
	}
	
	return (result);

  errorExit:
	putOUTSTR();

	return (0);
}
