/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Tserops
#else
#define __SEG__ Tserops
#endif				/*MPW3*/
#endif				/*MPW*/

#include "globals.h"
#include "tser.h"

typedef struct opEntry
{
	char *name;
	short iop;
	short nargs;
} opEntry;

static opEntry Ops[] =
{ /* name        iop      nargs */
	"hconj",    IHCONJ,     1, /* hconj(hx) */
	"cconj",    ICCONJ,     1, /* cconj(cx) */
	"hreal",    IHREAL,     1, /* hreal(hx) */
	"creal",    ICREAL,     1, /* creal(cx) */
	"himag",    IHIMAG,     1, /* himag(hx) */
	"cimag",    ICIMAG,     1, /* cimag(cx) */
	"hcopyc",   ICTOH,      1, /* hcopyc(hx) (obsolete)*/
	"ccopyh",   IHTOC,      1, /* ccopyh(cx) (obsolete) */
	"htoc",     IHTOC,      1, /* htoc(hx) */
	"ctoh",     ICTOH,      1, /* ctoh(cx) */
	"hprdh",    IHPRDH,    -2, /* hprdh(hx1 [,hx2]) */
	"hprdhj",   IHPRDHJ,   -2, /* hprdhj(hx1 [,hx2]) */
	"cprdc",    ICPRDC,    -2, /* cprdc(cx1 [,cx2]) */
	"cprdcj",   ICPRDCJ,   -2, /* cprdcj(cx1 [,cx2]) */
	"cmplx",    ICMPLX,    -2, /* cmplx(rx1, rx2) */
	"hpolar",   IHPOLAR,   -3, /* hpolar(hx [,unwind:F] [,crit:val]) */
	"cpolar",   ICPOLAR,   -3, /* cpolar(cx [,unwind:F] [,crit:val]) */
	"hrect",    IHRECT,     1, /* hrect(hx) */
	"crect",    ICRECT,     1, /* crect(cx) */
	"unwind",   IUNWIND,    -2, /* unwind(theta [,crit:val]) */
	"convolve", ICONVOLVE,  -3, /* convolve(coefs,rx[,reverse:T]) */
	"rotate",   IROTATE,    2, /* rotate(rx,lag) */
	"reverse",  IREVERSE,   1, /* reverse(rx) */
	"padto",    IPADTO,     2, /* padto(rx,nrows) */
	(char *) 0, 0,          0
};

#define CRITERION .75 /* default for keyword 'crit' */


Symbolhandle tserops(list)
Symbolhandle          list;
{
	Symbolhandle      arg[3], result = (Symbolhandle) 0;
	opEntry          *op;
	long              nargs = NARGS(list);
	long              margs = 1, type;
	long              logicOK;
	long              iop = 0, i, istart;
	long              dims[2][2];
	long              ncoefs, nrowsIn1, ncolsIn1, ncolsIn2, ncolsOut, nrowsOut;
	long              krot;
	long              xprd = 0, rect = 0, polar = 0;
	long              unwind = 1; /*default for keyword unwind*/
	long              reverseit = 0; /* default on convolve */
	double            criterion = CRITERION;
	char             *keyword = (char *) 0, usage[80];
	long              reuseArg1, reuseArg2;
	WHERE("tserops");
	
	*OUTSTR = '\0';
	
	for(op = Ops; op->iop != 0 && strcmp(op->name,FUNCNAME) != 0; op++)
	{
		;
	}
	
	if(op->iop == 0)
	{ /* should not occur unless programming error */
		sprintf(OUTSTR,"ERROR: (internal) unrecognized operation %s",FUNCNAME);
		goto errorExit;
	}

	margs = op->nargs;
	iop = op->iop;

	xprd = (iop == ICPRDC) || (iop == ICPRDCJ) || (iop == IHPRDH) ||
		(iop == IHPRDHJ);
	polar = (iop == IHPOLAR) || (iop == ICPOLAR) || (iop == IUNWIND);
	rect = (iop == IHRECT) || (iop == ICRECT);
	logicOK = (iop == IROTATE || iop == IREVERSE);
	
	if(nargs > labs(margs) || margs > 0 && nargs != margs)
	{
		badNargs(FUNCNAME,margs);
		goto errorExit;
	}
	
	for(i = 0;i<nargs;i++)
	{
		if(!argOK(arg[i] = COMPVALUE(list,i),0,i+1))
		{
			goto errorExit;
		}
		
		if(!logicOK && strucAnyMissing(arg[i]))
		{
			sprintf(OUTSTR,
					"ERROR: MISSING values not allowed in arguments to %s",
					FUNCNAME);
			goto errorExit;
		}
		type = TYPE(arg[i]);
		
		if((i == 0 && logicOK && type != REAL && type != LOGIC) ||
		   (i == 0 && !logicOK || (i == 1 && !polar) ||
			(i == 2 && !polar && iop != ICONVOLVE)) && TYPE(arg[i]) != REAL)
		{
			badType(FUNCNAME,-TYPE(arg[i]),i+1);
			goto errorExit;
		}
	} /*for(i = 0;i<nargs;i++)*/

	/* might it be possible to reuse scratch arguments; further checks later*/
	reuseArg1 = (iop != IPADTO && iop != IROTATE && iop != ICONVOLVE &&
				 isscratch(NAME(arg[0])));
	reuseArg2 = (xprd && nargs > 1 && isscratch(NAME(arg[1])));

	*OUTSTR = '\0';
/*	all operations except convolve() may have matrix for arg 1 */
	if(iop == ICONVOLVE)
	{
		if(!isVector(arg[0]))
		{
			sprintf(OUTSTR,"ERROR: first argument to %s is not vector",
					FUNCNAME);
		}
		else
		{
			(void) isMatrix(arg[0],dims[0]); /* set dims[0] */
			if(nargs == 3)
			{
				if((keyword = isKeyword(arg[2])) && 
				   strcmp(keyword,"reverse") != 0)
				{
					badKeyword(FUNCNAME,keyword);
					goto errorExit;
				}
				
				if(TYPE(arg[2]) != LOGIC || !isScalar(arg[2]))
				{
					sprintf(OUTSTR,
							"ERROR: value for %s must be T or F for %s",
							(keyword) ? keyword : "argument 3",
							FUNCNAME);
				}
				else
				{
					reverseit = (long) DATAVALUE(arg[2],0);
					nargs = 2;
				}
			}
		}
	} /*if(iop == ICONVOLVE)*/
	else
	{
		if(!isMatrix(arg[0],dims[0]))
		{
			sprintf(OUTSTR,
					"ERROR: first argument to %s is not a vector or matrix",
					FUNCNAME);
		}
	} /* if(iop == ICONVOLVE){}else{}*/
	if(*OUTSTR)
	{
		goto errorExit;
	}
	
	if(nargs == 2 && !polar)
	{
		if(iop == IROTATE || iop == IPADTO)
		{ /* these require scalar arg 2 */
			if(!isScalar(arg[1]) ||
			   floor(DATAVALUE(arg[1],0)) != DATAVALUE(arg[1],0))
			{
				sprintf(OUTSTR,
						"ERROR: second argument to %s is not single integer",
						FUNCNAME);				
			}
		}
		else if(!isMatrix(arg[1], dims[1]))
		{
			sprintf(OUTSTR,
					"ERROR: second argument to %s is not vector or matrix",
					FUNCNAME);
		}
	} /*if(nargs == 2 && !polar)*/
	else if(polar)
	{ /*
		hpolar(hx [,unwind:F], [crit:val])
		cpolar(cx [,unwind:F], [crit:val])
		unwind(rx, [,crit:val])
		*/

		if(iop == IUNWIND)
		{
			sprintf(usage,"usage is %s(rx [,crit:value])",FUNCNAME);
		}
		else
		{
			sprintf(usage,"usage is %s(%s [,unwind:F] [,crit:value])",
					FUNCNAME,(iop == IHPOLAR) ? "hx" : "cx");
		}
		istart = 1;
		if(iop != IUNWIND && nargs > 1 && !isKeyword(arg[1]))
		{ /* allow {h,c}polar(arg,F) */
			if(TYPE(arg[1]) != LOGIC || !isScalar(arg[1]))
			{
				sprintf(OUTSTR,
						"ERROR: second argument to %s not T or F",FUNCNAME);
				goto errorExit;
			}
			else
			{
				unwind = (long) DATAVALUE(arg[1],0);
				istart = 2;
			}
		}
		for(i=istart;i<nargs;i++)
		{
			if(!(keyword = isKeyword(arg[i])))
			{
				sprintf(OUTSTR,"ERROR: %s",usage);
			}
			if(strcmp(keyword,"crit") == 0)
			{
				if(TYPE(arg[i]) != REAL || !isScalar(arg[i]) ||
				   (criterion = DATAVALUE(arg[i],0)) < 0.5 ||
					   criterion > 1)
				{
						sprintf(OUTSTR,
								"ERROR: value for %s must be scalar between 0.5 and 1",
								keyword);
				}
			}
			else if(strcmp(keyword,"unwind") == 0 && iop != IUNWIND)
			{ /* keyword 'unwind' */
				if(TYPE(arg[i]) != LOGIC || !isScalar(arg[i]))
				{
					sprintf(OUTSTR,
							"ERROR: value for %s must be T or F",keyword);
				}
				else
				{
					unwind = (DATAVALUE(arg[i],0) != 0.0);
				}
			}
			else
			{
				badKeyword(FUNCNAME,keyword);
				goto errorExit;
			}
		} /*for(i=istart;i<nargs;i++)*/
	} /*if(nargs == 2 && !polar){...} else if(polar){...}*/
	
	if(*OUTSTR)
	{
		goto errorExit;
	}
	
		
/* determine size of nrowsOut by ncolsOut output matrix */

	nrowsOut = nrowsIn1 = dims[0][0];
	ncolsIn1 = dims[0][1];
	
	if(margs == 1 || iop == IHPOLAR || iop == ICPOLAR || iop == IUNWIND)
	{/*
	   hconj, cconj, hreal, creal, himag, cimag, htoc, ctoh,
	   hrect, crect, hpolar, cpolar, unwind
	   */
		if(iop == ICCONJ || iop == ICRECT || iop == ICPOLAR)
		{
			ncolsOut = ncolsIn1 + (ncolsIn1 & 1);
		}
		else if(iop == ICREAL || iop == ICIMAG || iop == ICTOH)
		{
			ncolsOut = (ncolsIn1 + 1)/2;
		}
		else if(iop == IHTOC)
		{
			ncolsOut = 2*ncolsIn1;
		}
		else
		{ /* hconj, hreal, himag, hrect, unwind, reverse */
			ncolsOut = ncolsIn1;
		}
	} /*if(margs == 1 || iop == IHPOLAR || iop == ICPOLAR || iop == IUNWIND)*/
	else if(xprd)
	{/* hprdh, hprdhj, cprdc, cprdcj*/
		if(nargs == 1)
		{
			arg[1] = arg[0];
			dims[1][0] = dims[0][0];
			dims[1][1] = dims[0][1];
		}
		ncolsIn2 = dims[1][1];
		if(dims[1][0] != nrowsIn1)
		{
			sprintf(OUTSTR,
					"ERROR: arguments to %s have different numbers of rows",
					FUNCNAME);
		}
		else
		{
			if(iop == ICPRDC || iop == ICPRDCJ)
			{
				if(ncolsIn1 > 2 && ncolsIn2 > 2 && labs(ncolsIn1-ncolsIn2) > 1)
				{
					sprintf(OUTSTR,
							"ERROR: number columns of arguments to %s do not match",
							FUNCNAME);
				}
				else
				{
					ncolsOut = (ncolsIn1 > ncolsIn2) ? ncolsIn1 : ncolsIn2;
					ncolsOut += ncolsOut & 1; /* increase to even */
				}
			} /*if(iop == ICPRDC || iop == ICPRDCJ)*/
			else
			{/* hprdh or hprdhj */
				if(ncolsIn1 > 1 && ncolsIn2 > 1 && ncolsIn1 != ncolsIn2)
				{
					sprintf(OUTSTR,
							"ERROR: arguments to %s have different numbers of columns",
							FUNCNAME);
				}
				else
				{
					ncolsOut = (ncolsIn1 > ncolsIn2) ? ncolsIn1 : ncolsIn2;
				}
			} /*if(iop == ICPRDC || iop == ICPRDCJ) */
		} /*if(dims[1][0] != nrowsIn1)*/
	}
 /* margs == 2: cmplx, convolve, rotate, padto */
	else if(iop == ICMPLX)
	{
		if(nargs > 1 && (dims[1][0] != nrowsIn1 || dims[1][1] != ncolsIn1))
		{
			sprintf(OUTSTR,
					"ERROR: arguments to %s not the same size and shape",
					FUNCNAME);
		}
		else
		{
			ncolsOut = 2*ncolsIn1;
		}
	}
	else if(iop == ICONVOLVE)
	{
		ncoefs = nrowsIn1;
		nrowsOut = nrowsIn1 = dims[1][0];
		ncolsOut = ncolsIn1 = dims[1][1];
	}
	else if(iop == IROTATE || iop == IPADTO)
	{ /* padto, rotate */
		krot = (long) DATAVALUE(arg[1],0);
		if(iop == IPADTO && krot <= 0)
		{
			sprintf(OUTSTR,"ERROR: 2nd argument to %s must be integer > 0",
					FUNCNAME);
		}
		ncolsOut = ncolsIn1;
		nrowsOut = (iop == IROTATE) ? nrowsIn1 : krot;
	}
	else
	{
		sprintf(OUTSTR,"ERROR:****programming problem 1 for %s",FUNCNAME);
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}


	if(reuseArg1 && nrowsOut == nrowsIn1 && ncolsOut == ncolsIn1)
	{
		reuseArg2 = 0;
	}
	else
	{
		reuseArg1 = 0;
	}
	if(reuseArg1 ||
	   reuseArg2 && (nrowsOut != nrowsIn1 || ncolsOut != ncolsIn2))
	{
		reuseArg2 = 0;
	}
	
	if(reuseArg1)
	{
		result = arg[0];
		COMPVALUE(list, 0) = (Symbolhandle) 0;
	}
	else if(reuseArg2)
	{
		result = arg[1];
		COMPVALUE(list, 1) = (Symbolhandle) 0;
	}
	else
	{
		result = RInstall(SCRATCH,nrowsOut*ncolsOut);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
	}
	
	setNCLASS(result, -1);

	if(ncolsOut > 1)
	{
		setNDIMS(result,2);
		setDIM(result,2,ncolsOut);
	}
	setDIM(result,1,nrowsOut);
	if(logicOK)
	{
		setTYPE(result, TYPE(arg[0]));
	}
	if(margs == 1 && !rect)
	{
		qonerg(DATAPTR(arg[0]), DATAPTR(result),
			   nrowsIn1, ncolsIn1, ncolsOut,iop);
	}
	else if(xprd || iop == ICMPLX)
	{
		qtworg(DATAPTR(arg[0]),
			   (iop == ICMPLX && nargs == 1) ? (double *) 0 : DATAPTR(arg[1]),
			   DATAPTR(result),
			  nrowsIn1, ncolsIn1, ncolsIn2, iop);
	}
	else if(polar || rect)
	{
		qrect(DATAPTR(arg[0]), DATAPTR(result),
			  nrowsIn1, ncolsIn1, unwind, criterion, iop);
	}
	else if(iop == IPADTO)
	{
		qpadto(DATAPTR(arg[0]),DATAPTR(result),nrowsIn1,ncolsIn1,nrowsOut);
	}
	else if(iop == IROTATE)
	{
		qrota(DATAPTR(arg[0]),DATAPTR(result),nrowsIn1,ncolsIn1,krot);
	}
	else if(iop == ICONVOLVE)
	{
		qconv(DATAPTR(arg[0]),DATAPTR(arg[1]),DATAPTR(result),
			  ncoefs,nrowsIn1,ncolsIn1,reverseit);
	}
	else	
	{ /* should not happen */
		sprintf(OUTSTR,"ERROR:****programming problem 2 for %s",FUNCNAME);
		goto errorExit;
	}
	
	return (result);

  errorExit:
	putOUTSTR();
	Removesymbol(result);
	return(0);
	
}
