/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Tsersubs
#else
#define __SEG__ Tsersubs
#endif				/*MPW3*/
#endif				/*MPW*/

#include "globals.h"
#include "tser.h"

/*
   This includes to following functions:
   qonerg	dispatcher for of hconj, cconj, hreal, himag, creal, cimag,
			ctoh, htoc, revers
   qtworg	dispatcher for hprdh, hprdhj, cprdc, cprdcj, cmplx
   qconj	computation of hconj, cconj
   qxprd	computation of hprdh, hprdhj, cprdc, cprdcj
   qconv    computation of convolve
   qreal	computation of creal, cimag, hreal, himag
   qcmplx	computation of cmplx
   qrect	computation of hpolar, cpolar, hrect, crect and dispatch unwind
   qhcopy	computation of ctoh, htoc
   qpadto	computation of padto
   qrota	computation of rotate
   qreverse	computation of reverse
   qunwind  computation of unwind

*/
/* qconj.c */


/* qonerg.c */



/*
 	function to control various one-argument fft-related routines
 	iop	    op		iop	    op
 	IHCONJ hconj	IHIMAG himag
 	ICCONJ cconj	ICIMAG cimag
 	IHREAL hreal	IHTOC  htoc
 	ICREAL creal	ICTOC  ctoh

 	version of 940915
*/

void qonerg(a, b, m, n, n1, iop)
long                 m, n, n1, iop;
double              *a, *b;
/*double a[m][n], b[m][n1];*/
{
	WHERE("qonerg");
	
	switch (iop)
	{
	  case IHCONJ:
	  case ICCONJ:
		qconj(a, b, m, n, iop);
		break;
		
	  case IHREAL:
	  case IHIMAG:
	  case ICREAL:
	  case ICIMAG:
		qreal(a, b, m, n, iop);
		break;
		
	  case ICTOH:
	  case IHTOC:
		qhcopy(a, b, m, n, iop);
		break;

	  case IREVERSE:
		qreverse(a, b, m, n);
		break;
		
	} /*switch (iop)*/
} /*qonerg()*/

/* qtworg.c */

/*
 	function to control various two-argument fft-related routines
 	iop      op    iop	    op
 	IHPRDH  hprdh  ICPRDCJ cprdcj
 	IHPRDHJ hprdhj ICMPLX  cmplx
 	ICPRCJ  cprdc

 	version of 940915
*/

void                 qxprd(), qcplx();

void qtworg(a, b, c, m, n1, n2, iop)
long                 m, n1, n2, iop;
double              *a, *b, *c;
/*double a[m][n1], b[m][n2], c[m][n1]; */
{
	switch (iop)
	{
	  case IHPRDH:
	  case IHPRDHJ:
	  case ICPRDC:
	  case ICPRDCJ:
		qxprd(a, b, c, m, n1, n2, iop);
		break;
		
	  case ICMPLX:
		qcmplx(a, b, c, m, n1);
		break;
	}
} /*qtworg()*/

/*
  function to compute complex conjugate of hermitian or complex
  matrix according as iop = IHCONJ or ICCONJ.

*/

void            qconj(a, b, m, n, iop)
long            m, n, iop;
double         *a, *b;
{
	double         *ajre, *ajim;
	double         *bjre, *bjim;
	double         *aj, *bj;
	long            i, ir, j, m1, nn;
	long            different = (a != b);
	
	if (iop == ICCONJ)		/* cconj */
	{
		ajre = a;
		bjre = b;
		nn = n / 2;
		for (j = 0; j < nn; j++)
		{
			ajim = ajre + m;
			bjim = bjre + m;
			if (different)
			{
				for (i = 0; i < m; i++)
				{
					bjre[i] = ajre[i];
				}
			}
			for (i = 0; i < m; i++)
			{
				bjim[i] = -ajim[i];
			}
			ajre += 2*m;
			bjre += 2*m;
		} /*for (j = 0; j < nn; j++)*/

		if (n % 2 != 0)
		{ /* missing imaginary part for last column; treat as if zero */
			bjim = bjre + m;
			for (i = 0; i < m; i++)
			{ /* must be different */
				bjre[i] = ajre[i];
				bjim[i] = 0.0;
			}
		} /*if (n % 2 != 0)*/
	} /*if (iop == ICCONJ)*/
	else
	{ 		/* hconj */
		m1 = m / 2;
		
		aj = a;
		bj = b;
		for (j = 0; j < n; j++)
		{
			if (different)
			{
				for (i = 0; i <= m1; i++)
				{
					bj[i] = aj[i];
				}
			}
			for(ir = m-1;ir > m1; ir--)
			{
				bj[ir] = -aj[ir];
			}
			aj += m;
			bj += m;
		} /*for (j = 0; j < n; j++)*/
	} /*if (iop == ICCONJ){}else{}*/
	return;
} /*qconj()*/

/* qxprd.c */

#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define MAX(a,b) (((a) > (b)) ? (a) : (b))

/*
    function to carry out elementwise multiplication of
    complex and hermitian matrices. iop == IHPRDHJ + 0,1,2,3 means
    hprdh, hprdhj, cprdc, cprdcj.
*/
void            qxprd(a, b, c, m, n1, n2, iop)
long            m, n1, n2, iop;
double         *a, *b, *c;
{
	long            m1, m2, i, j, ir, nn, iwhich, twom;
	long            n, inc1, inc2, n1odd, n2odd;
	long            odd;
	double          amult, t1, t2;
	double         *aj, *bj, *cj;
	double         *ajre, *bjre, *cjre;
	double         *ajim, *bjim, *cjim;
	double          temp;
	
	odd = m & 1;
	amult = (iop == IHPRDHJ || iop == ICPRDCJ) ? -1.0 : 1.0;
	
	if (iop == IHPRDH || iop == IHPRDHJ)
	{/* hprdh (amult == 1), hprdhj (amult == -1) */
		inc1 = (n1 == 1) ?  0 : m;
		inc2 = (n2 == 1) ?  0 : m;
		m1 = m / 2;
		m2 = (m - 1) / 2;
		n = MAX(n1, n2);

		aj = a;
		bj = b;
		cj = c;
		for (j = 0; j < n; j++)
		{
			cj[0] = aj[0] * bj[0];
			for (i = 1, ir = m-1; i <= m2; i++, ir--)
			{
				temp = aj[i] * bj[i] - amult * aj[ir] * bj[ir];
				cj[ir] = amult * aj[i] * bj[ir] + aj[ir] * bj[i];
				cj[i] = temp; /*insurance in case storage re-used*/
			}
			if (!odd)
			{
				cj[m1] = aj[m1] * bj[m1];
			}
			aj += inc1;
			bj += inc2;
			cj += m;
		} /*for (j = 0; j < n1; j++)*/
	} /*if (iop == IHPRDH || iop == IHPRDHJ)*/
	else
	{/* cprdc (amult == 1), cprdcj (amult == -1) */
		
		n1odd = (n1 & 1) != 0;
		n2odd = (n2 & 1) != 0;
		nn = MAX(n1, n2);
		if((n1odd || n2odd) && (nn & 1) == 0)
		{
			nn--;
		}
		nn /= 2;
		
		twom = 2*m;
		inc1 = (n1 <= 2) ? 0 : twom;
		inc2 = (n2 <= 2) ? 0 : twom;
		ajre = a;
		bjre = b;
		cjre = c;
		for (j = 0; j < nn; j++)
		{
			ajim = ajre + m;
			bjim = bjre + m;
			cjim = cjre + m;

			for (i = 0; i < m; i++)
			{
				temp = ajre[i] * bjre[i] - amult * ajim[i] * bjim[i];
				cjim[i] = amult * ajre[i] * bjim[i] + ajim[i] * bjre[i];
				cjre[i] = temp; /*insurance in case storage re-used*/
			}
			ajre += inc1;
			bjre += inc2;
			cjre += twom;
		} /*for (j = 0; j < nn; j++)*/

		if (n1odd || n2odd)
		{
			/*
			  iwhich =1, 2, or 3 according as a is short and b ok,
			  a is ok and b short, or both a and b are short.
			  */
			iwhich = (n1odd && !n2odd) ? 1 : ((!n1odd && n2odd) ? 2 : 3);
			t1 = t2 = 0.0;
			for (i = 0; i < m; i++)
			{
				if (iwhich == 1)
				{
					t2 = bjim[i];
				}
				else if (iwhich == 2)
				{
					t1 = ajim[i];
				}

				temp = ajre[i] * bjre[i];
				cjim[i] = amult * ajre[i] * t2 + t1 * bjre[i];
				cjre[i] = temp; /*insurance in case storage re-used*/
			} /*for (i = 0; i < m; i++)*/
		} /*if (n1odd || n2odd)*/
	} /*if (iop == IHPRDH || iop == IHPRDHJ){}else{}*/
} /*qxprd()*/

/* qconv.c */


/*
  function to perform circular convolution of a with columns of b.
  reverse == 0    ordinary convolution
  reverse != 0    sums of lagged products
*/

void qconv(a, b, c, m1, m2, n, reverse)
long               m1, m2, n, reverse;
double            *a, *b, *c;
/*double a[m1], b[m2][n], c[m2][n];*/
{
	long           inc, is, it, j, k, l;
	long           start, end;
	double        *cj, *bj;
	double         s;
	WHERE("qconv");
	
	if(!reverse)
	{ /* regular convolution, subscripts move in opposite directions */
		inc = -1;
		end = -1;
		start = m2-1;
	}
	else
	{ /* reverse convolution, subscripts move in same direction */
		inc = 1;
		start = 0;
		end = m2;
	}

	bj = b;
	cj = c;
	
	for(j=0;j<n;j++)
	{
		l = 0;
		for(it=0;it<m2;it++)
		{
			k = l;
			s = 0.0;
			for(is=0;is<m1;is++)
			{
				s += a[is] * bj[k];
				k += inc;
				if(k == end)
				{
					k = start;
				}
			} /*for(is=0;is<m1;is++)*/
			cj[it] = s;
			l -= inc;
			if(it==0 && reverse)
			{
				l += m2;
			}
		} /*for(it=0;it<m2;it++)*/
		bj += m2;
		cj += m2;
	} /*for(j=0;j<n;j++)*/
} /*qconv()*/

/* qreal.c */


/*
  function to take real and imaginary parts of hermitian and
  complex matrices. iop = IHREAL, IHIMAG, ICREAL, ICIMAG mean
  hreal,himag,creal,cimag
*/

void qreal(a, b, m, n, iop)
long                m, n, iop;
double             *a, *b;
/*double a[m][n], b[m][n];*/
{
	long            m1, m2, i, j, ir, nn;
	double         *aj, *bj;
	long            nOdd, mOdd, realOp;

	realOp = (iop == IHREAL) || (iop == ICREAL);

	if (iop == ICREAL || iop == ICIMAG)
	{
		nOdd = n & 1;
		aj = (realOp) ? a : a + m;
		bj = b;
		nn = n/2;
		if(realOp && nOdd)
		{
			nn++;
		}
		
		for(j=0;j<nn ;j++)
		{ 
			for(i=0;i<m;i++)
			{
				bj[i] = aj[i];
			}
			aj += 2*m;
			bj += m;
		} /*for(j=0;j<nn ;j++)*/
		if (!realOp && nOdd) 
		{ /* fill out last column with 0's */
			for(i=0;i<m;i++)
			{
				bj[i] = 0.0;
			}
		}
	} /*if (iop == ICREAL || iop == ICIMAG)*/
	else 
	{ /* hreal, himag */
		m1 = m/2;
		m2 = (m+1)/2;
		mOdd = m & 1;
		aj = a;
		bj = b;
		for(j=0;j<n ;j++)
		{		
			bj[0] = (realOp) ? aj[0] : 0.0;
			if(!mOdd)
			{
				bj[m1] = (realOp) ? aj[m1] : 0.0;
			}
			for(i=1, ir=m-1;i<m2;i++, ir--)
			{
				bj[i] = (realOp) ? aj[i] : aj[ir];
				bj[ir] = (realOp) ? bj[i] : -bj[i];
			}
			aj += m;
			bj += m;
		} /*for(j=0;j<n ;j++)*/
	} /*if (iop == ICREAL || iop == ICIMAG){}else{}*/
} /*qreal()*/

/* qcmplx.c */


/*
  function to create complex matrix from real and imaginary
  parts in a and b.
*/

void qcmplx(a, b, c, m, n)
long            m, n;
double         *a, *b, *c;
/*double a[m][n], b[m][n], c[m][n];*/
{
	long        i, j;
	double     *aj = a, *bj = b, *cjre, *cjim;
	long        breal = (b == (double *) 0);
	
	
	cjre = c;
	for(j=0;j<n;j++)
	{
		cjim = cjre + m;
		for(i=0;i<m;i++)
		{
			cjre[i] = aj[i];
			cjim[i] = (breal) ? 0.0 : bj[i];
		} /*for(i=0;i<m;i++)*/
		aj += m;
		bj += m;
		cjre += 2*m;
	} /*for(j=0;j<n;j++)*/
} /*qcmplx()*/

/* qrect.c */

/*
    subroutine to convert hermitian and complex series from
    rectangular (normal) form to polar and vice versa,
    and "unwind" phase functions.
      iop   operation    units  angles
      IHPOLAR hpolar        1    radians
      IHRECT  hrect         2    degrees
      ICPOLAR cpolar        3    cycles
      ICRECT  crect
      IUNWIND unwind
  if cpolar or crect b is m by 2*((n+1)/2)
*/
/* 
  MacAnova note: There is no argument 'units' since that is taken from global
  ANGLES2RADIANS
*/

void qrect(a, b, m, n, unwind, crit, iop)
long                 m, n, iop;
long                 unwind;
double              *a, *b, crit;
/*double a[m][n], b[m][n], crit;*/
{
	long             opPolar;
	long             m1, m2, i, ir, j, nn;
	long             mOdd, nOdd;
	double           factor, twopi, fullcycle, halfcycle, s, t;
	double          *aj, *bj;
	double          *ajre, *ajim;
	double          *bjre, *bjim;
	WHERE("qrect");

	opPolar = (iop == IHPOLAR||iop == ICPOLAR);
	nOdd = n & 1;
	mOdd = m & 1;

	twopi = 8.0*atan(1.0);
	factor = ANGLES2RADIANS;
	fullcycle = twopi/factor;
	
	if (iop == IHRECT || iop == IHPOLAR)
	{
		m1 = m/2;
		m2 = (m+1)/2;
		aj = a;
		bj = b;
		for(j=0;j<n ;j++)
		{
			bj[0] = aj[0];
			if(!mOdd)
			{
				bj[m1] = aj[m1];
			}
			
			for(i=1, ir = m-1;i<m2;i++, ir--)
			{
				s = aj[i];
				t = aj[ir];
				if(!opPolar)
				{
					t *= factor;
					bj[i] = s*cos(t);
					bj[ir] = s*sin(t);
				}
				else
				{
					bj[i] = pythag(s, t);
					bj[ir] = atan2(t, s)/factor;
				}
			} /*for(i=1, ir = m-1;i<m2;i++, ir--)*/
			if (unwind && opPolar)
			{
				qunwnd(bj, m, fullcycle, crit, IHERMITIAN);
			}
			aj += m;
			bj += m;
		} /*for(j=0;j<n ;j++)*/
	}
	else if (iop == ICPOLAR || iop == ICRECT)
	{
		/* ****cpolar, crect */
		nn = n/2;
		ajre = a;
		bjre = b;
		for(j=0;j<nn ;j++)
		{
			ajim = ajre + m;
			bjim = bjre + m;
			for(i=0;i<m ;i++)
			{
				s = ajre[i];
				t = ajim[i];
				if (!opPolar)
				{
					t *= factor;
					bjre[i] = cos(t)*s;
					bjim[i] = sin(t)*s;
				}
				else
				{
					bjre[i] = pythag(s, t);
					bjim[i] = atan2(t, s)/factor;
				}
			} /*for(i=0;i<m ;i++)*/
			if (unwind && opPolar)
			{
				qunwnd(bjim, m, fullcycle, crit, IREAL);
			}
			ajre += 2*m;
			bjre += 2*m;
		} /*for(j=0;j<nn ;j++)*/

		if (nOdd)
		{
			/* ****add final column if n is not even. */
			halfcycle = 0.5*fullcycle;
			for(i = 0;i<m;i++)
			{
				if(!opPolar)
				{
					bjre[i] = ajre[i];
					bjim[i] = 0.0;
				}
				else
				{
					s = ajre[i];
					bjre[i] = fabs(s);
					bjim[i] = (s >= 0.0) ? 0.0 : halfcycle;
				}
			} /*for(i = 0;i<m;i++)*/
		} /*if (nOdd)*/
	}
	else
	{ /* unwind */
		aj = a;
		bj = b;
		for(j = 0;j < n;j++)
		{
			if(b != a)
			{
				for(i=0;i<m;i++)
				{
					bj[i] = aj[i];
				}
			}
			qunwnd(bj, m, fullcycle, crit, IREAL);
			aj += m;
			bj += m;
		} /*for(j = 0;j < n;j++)*/
	}
} /*qrect()*/

/* qhcopy.c */


/*
  function to convert from hermitian to full complex and
  vice versa.  iop = ICTOH, IHTOC means ctoh, htoc
*/

void qhcopy(a, b, m, n, iop)
long                  m, n, iop;
double               *a, *b;
/*double a[m][n], b[m][n];*/
{
	long              m1, m2, j, i, ir, nn;
	long              mOdd, nOdd;
	double           *aj, *bj;
	double           *ajre, *ajim, *bjre, *bjim;
	
	mOdd = m & 1;
	m1 = m/2;
	m2 = (m+1)/2;
	if (iop == IHTOC)
	{
		/* **** htoc */
		
		aj = a;
		bjre = b;
		for(j=0;j<n ;j++)
		{
			bjim = bjre + m;
			bjre[0] = aj[0];
			bjim[0] = 0.0;
			
			if(!mOdd)
			{
				bjre[m1] = aj[m1];
				bjim[m1] = 0.0;
			}
			
			for(i = 1, ir = m-1;i<m2;i++, ir--)
			{
				bjre[ir] = bjre[i] = aj[i];
				bjim[ir] = -(bjim[i] = aj[ir]);
			}
			aj += m;
			bjre += 2*m;
		} /*for(j=0;j<n ;j++)*/
	} /*if (iop == IHTOC)*/
	else
	{ /* ctoh */
		nn = n/2;
		nOdd = n & 1;
		ajre = a;
		bj = b;
		for(j=0;j<nn ;j++)
		{
			ajim = ajre + m;
			bj[0] = ajre[0];
			
			if(!mOdd)
			{
				bj[m1] = ajre[m1];
			}
			
			for(i = 1, ir = m-1;i < m2;i++, ir--)
			{ /* hermitian symmetrization of output matrix */
				bj[i] = 0.5*(ajre[i] + ajre[ir]);
				bj[ir] = 0.5*(ajim[i] - ajim[ir]);
			}
			ajre += 2*m;
			bj += m;
		} /*for(j=0;j<nn ;j++)*/
		if (nOdd)
		{ /* a has odd number of columns */
			bj[0] = ajre[0];
			if(!mOdd)
			{
				bj[m1] = ajre[m1];
			}
			for(i = 1, ir = m-1;i < m2; i++, ir--)
			{
				bj[i] = 0.5*(ajre[i] + ajre[ir]);
				bj[ir] = 0.0;
			} /*for(i = 1;i < m2; i++)*/
		} /*if (nOdd)*/
	} /*if (iop == IHTOC){}else{}*/
} /*qhcopy()*/

/* qpadto.c */


/*
  function to add rows zeros to in or delete rows from in so as to make
  out m2 by n.
  Adapted from Fortran 920420
  Version of 940915
 */

void qpadto(in, out, m1, n, m2)
long                 m1, m2, n;
double              *in, *out;
/*double in[m1][n], out[m2][n];*/
{
	long             i, j;
	double          *inj = in;
	double          *outj = out;

	for(j = 0;j < n; j++)
	{
		for(i=0;i<m2 ;i++)
		{
			outj[i] = (i < m1) ? inj[i] : 0.0;
		}
		inj += m1;
		outj += m2;
	} /*for(j = 0;j < n; j++)*/
} /*qpadto()*/

/* qrota.c */



/*
  subroutine to 'rotate' columns of a into b, offsetting
  them circularly by krot.
*/
void qrota(a, b, m, n, krot)
long            m, n, krot;
double         *a, *b;
/*double a[m][n], b[m][n];*/
{
	long        i, is, j;
	double     *aj = a, *bj = b;
	WHERE("qrota");
	
	krot %= m;

	if (krot < 0)
	{
		krot += m;
	}
	
	for(j=0;j<n;j++)
	{
		for(i = 0, is = krot;i< m; i++)
		{
			bj[is++] = aj[i];
			if(is == m)
			{
				is = 0;
			}
		} /*for(i = 0, is = krot;i< m; i++)*/
		aj += m;
		bj += m;
	} /*for(j=0;j<n;j++)*/
} /*qrota()*/

/*
  function to reverse the rows of a into b.
*/
void qreverse(a, b, m, n)
long            m, n;
double         *a, *b;
{
	long        i, ir, j, m2 = m/2, odd = (2*m2 != m);
	double     *aj = a, *bj = b;
	double      temp;
	WHERE("qreverse");
	
	for(j=0;j<n;j++)
	{
		for(i = 0, ir = m-1;i< m2; i++, ir--)
		{ /* coded so it is safe for a to be the same as b*/
			temp = aj[i];
			bj[i] = aj[ir];
			bj[ir] = temp;
		} /*for(i = 0;i< m2; i++)*/
		if(odd)
		{
			bj[m2] = aj[m2];
		}
		aj += m;
		bj += m;
	} /*for(j=0;j<n;j++)*/
} /*qreverse()*/

/* qunwnd.c */


/*
  Function to 'unwind' phase functions.
   iop    meaning
    1     phase is to be real (e.g., imag part of cpolar output)
    2     phase is hermitian (hpolar output), and only 2nd half is modified.
   if the phase changes by more than +-crit cycles (crit should be
   between 0.5 and 1.0), it is assumed to be 'wrapping around'
   and a multiple of twopi  (2*pi, 360, or 1 depending on units)
   is added or subtracted, in an attempt to eliminate meaningless
   jumps in phase functions.
*/

void qunwnd(phase, m, oneCycle, crit, iop)
long                m, iop;
double             *phase, oneCycle, crit;
/*double phase[m], oneCycle, crit;*/
{
	long            i, k, inc, m2;
	double          phlast, diffph;
	WHERE("qunwnd");

	crit = oneCycle*crit;

	if (iop==IREAL) 
	{
		/* ****initialize for iop eq 1 (full real form) */
		phlast = phase[0];
		inc = 1;
		k = 0;
		m2 = m;
	}
	else 
	{
		/* ****initialize for iop eq 2 (hermitian form) */
		phlast = (phase[0] >= 0) ? 0.0 : oneCycle/2.0;
		inc = -1;
		k = m;
		m2 = (m+1)/2;
	}
	for(i=1;i<m2 ;i++)
	{
		k += inc;
		diffph = phase[k] - phlast;
		phlast = phase[k];
		while(diffph > crit)
		{
			diffph -= oneCycle;
			phase[k] -= oneCycle;
		}
		while (diffph < -crit)
		{
			diffph += oneCycle;
			phase[k] += oneCycle;
		}
		phlast = phase[k];
	} /*for(i=1;i<m ;i++)*/
} /*qunwnd()*/

