/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Unbalano
#else
#define __SEG__ Unbalano
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <math.h>

/*
#define x(i,j) ((*REGX)[ (long)i + (long)j* ndata ])
*/

/*
  routine for computing ss and df for unbalanced designs

  RESIDUALS	handle to response vector (which will be changed)
  residuals	pointer to *RESIDUALS
  X		array of handles to class variables
  NVARS	numberof variables (columns of x)
  NCLASSES	number of classes in each variable
  NDATA	number of data values
  MODEL	handle giving terms in order to be used in model
	  term i contains the variables in the binary expansion of i
	  0 is constant; 3 is AB; 5 is AC etc
  SS	handle to term sums of squares
  DF	handle of term df
*/

long             unbalanova(void)
{
	long            i, icurr, j, k, k1, k2;
	long            step1, step2, step3, step4;
	long            ny = (long) NY, ndata = (long) NDATA, nregx = (long) NREGX;
	long            nterms = (long) NTERMS;
	long            ik1, ik1a, ik1k2, ik1k2a;
	long            maxdfi, dfi;
	long            control = MODELTYPE & DOCOEFS;
	double          errordf;
	double         *hii;
	double         *allwts, *residuals, *wtdresiduals, *misswts;
	double         *regx, *ss, *regss, *xij;
	WHERE("unbalanova");

	SETUPINT(errorExit);
	step1 = nterms + 1;
	step2 = ny * step1;
	step3 = nregx + 1;
	step4 = ny * step3;
	
	if (MODELTYPE & ITERWEIGHTS)
	{
		goto errorExit;	/* can't do anything yet */
	}
	/* setup xtxinv if required */
	if(control & DOCOEFS)
	{
		xtxinvInit();
		control |= DORINV;/* compute R^(-1) when computing coefficients */
	}
	
	/* notify of singular columns only if there are no factors */
	control |= ((NFACTORS > 0) ? DONTNOTIFY : DONOTIFY) | DOSWEEP | DOSS;

	if (MODELTYPE & USEWEIGHTS)
	{
		ALLWTS = (double **) mygethandle(ndata * sizeof(double));
		if (ALLWTS == (double **) 0)
		{
			goto errorExit;
		}
		
		allwts = *ALLWTS;
		regx = *REGX;
		residuals = *RESIDUALS;

		for (i = 0; i < ndata; i++)
		{
			allwts[i] = 1.0;
			if (MODELTYPE & MISSWEIGHTS)
			{
				allwts[i] *= (*MISSWTS)[i];
			}
			if (MODELTYPE & CASEWEIGHTS && allwts[i])
			{
				allwts[i] *= sqrt((*CASEWTS)[i]);
			}
		} /*for (i = 0; i < ndata; i++)*/
		
		xij = regx;
		for(j = 0;j < nregx; j++)
		{
			for(i = 0;i < ndata; i++)
			{
				xij[i] *= allwts[i];
			}
			xij += ndata;
		}
		for (j = 0; j < ny; j++)
		{
			for (i = 0; i < ndata; i++)
			{
				residuals[i] *= allwts[i];
			}
			residuals += ndata;
		} /*for (j = 0; j < ny; j++)*/
	} /*if (MODELTYPE & USEWEIGHTS)*/

	gramschmidt(control);

	if (INTERRUPT != INTNOTSET)
	{
		goto errorExit;
	}

	/* combine REGSS into SS */
	ss = *SS;
	regss = *REGSS;
	icurr = 0;
	errordf = (double) NOTMISSING;

	doubleFill(ss, 0.0, (nterms+1)*ny*ny); /* zero *SS */
	for (i = 0; i < nterms; i++)
	{
		dfi = 0;
		maxdfi = (*DF)[i];
		for (j = 0; j < maxdfi; j++)
		{/* go through the cols for each term */
			if (regss[icurr] > -0.5)
			{
				if(ny == 1)
				{
					ss[i] += regss[icurr];
				} /*if(ny == 1)*/
				else
				{
					ik1 = i;	/* i + k1*step1 */
					ik1a = icurr; /* icurr + k1 * (nregx + 1) */
					for (k1 = 0; k1 < ny; k1++)
					{
						ik1k2 = ik1; /* i + k1 * step1 + k2 * step2 */
						ik1k2a = ik1a; /* icurr + k1 * (nregx + 1) + k2 * (ny * (nregx + 1)) */
						for (k2 = 0; k2 < ny; k2++)
						{
							ss[ik1k2] += regss[ik1k2a];
							/*
							   (*SS)[i + k1 * step1 + k2 * step2] +=
							   (*REGSS)[icurr + k1 * (nregx + 1) +
							   k2 * (ny * (nregx + 1))];
							   */
							ik1k2 += step2;
							ik1k2a += step4;
						} /*for (k2 = 0; k2 < ny; k2++)*/
						ik1 += step1;
						ik1a += step3;
					} /*for (k1 = 0; k1 < ny; k1++)*/
				} /*if(ny == 1){}else{}*/
				dfi++;
			} /*if (regss[icurr] > -0.5)*/
			icurr++;
		} /*for (j = 0; j < maxdfi; j++)*/
		errordf -= dfi;
		(*DF)[i] = (double) dfi;
	} /*for (i = 0; i < nterms; i++)*/

	(*DF)[nterms] = errordf;

	if(ny == 1)
	{
		if(errordf == 0)
		{
			regss[nregx] = 0.0;
		}
		ss[nterms] = regss[nregx];
	} /*if(ny == 1)*/
	else
	{
		ik1 = nterms; /* nterms + k1 * step1 */
		ik1a = nregx; /* nregx + k1 * (nregx + 1) */
		for (k1 = 0; k1 < ny; k1++)
		{
			ik1k2 = ik1; /* nterms + k1 * step1 + k2 * (ny * (nregx + 1)) */
			ik1k2a = ik1a; /* nregx + k1 * (nregx + 1) + k2 * (ny * (nregx + 1)) */
			for (k2 = 0; k2 < ny; k2++)
			{
				if(errordf == 0)
				{
					regss[ik1k2a] = 0.0;
				}
				ss[ik1k2] = regss[ik1k2a]; /* error ss */
				/*
				   (*SS)[nterms + k1 * step1 + k2 * step2] =
				   (*REGSS)[nregx + k1 * (nregx + 1) + k2 * (ny * (nregx + 1))];
				   */
				ik1k2 += step2;
				ik1k2a += step4;
			} /*for (k2 = 0; k2 < ny; k2++)*/
			ik1 += step1;
			ik1a += step3;
		} /*for (k1 = 0; k1 < ny; k1++)*/
	} /*if(ny == 1){}else{}*/
	
	
	/* fix up HII */
	if(MODELTYPE & MISSWEIGHTS)
	{
		misswts = *MISSWTS;
	}
	
	hii = *HII;

	for (i = 0; i < ndata; i++)
	{
		if (!(MODELTYPE & MISSWEIGHTS) || misswts[i] != 0.)
		{
			hii[i] = 0.0;
		}
		else
		{
			setMissing(hii[i]);
		}
	} /*for (i = 0; i < ndata; i++)*/
	
	xij = *REGX;
	
	for(j = 0; j < nregx; j++)
	{
		if (regss[j] > -0.5)
		{
			for (i = 0;i < ndata; i++)
			{
				if(!isMissing(hii[i]))
				{
					hii[i] += xij[i] * xij[i];
				}
			}
		} /*if (regss[j] > -0.5)*/		
		xij += ndata;
	} /*for(j = 0; j < nregx; j++)*/
	
	/* fix up RESIDUALS if needed */

	/* fix up real residuals if case or iter weights used */
	if (MODELTYPE & CASEWEIGHTS)
	{
		TMPHANDLE = (char **) WTDRESIDUALS;
		WTDRESIDUALS = RESIDUALS;	/* our residuals were actually weighted */
		RESIDUALS = (double **) TMPHANDLE;
		allwts = *ALLWTS;
		residuals = *RESIDUALS;
		wtdresiduals = *WTDRESIDUALS;
		for(k = 0; k < ny; k++)
		{
			for (i = 0; i < ndata; i++)
			{
				if (allwts[i] > 0.0)
				{
					residuals[i] = wtdresiduals[i] / allwts[i];
				}
				else
				{
					/*
					   set residuals and wtdresiduals MISSING for cases with zero case weight
					   or missing data
					   */
					setMissing(residuals[i]);
					setMissing(wtdresiduals[i]);
				}
			} /*for (i = 0; i < ndata; i++)*/
			residuals += ndata;
			wtdresiduals += ndata;
		} /*for(k = 0; k < ny; k++)*/
	} /*if (MODELTYPE & CASEWEIGHTS)*/
	else if (MODELTYPE & MISSWEIGHTS)
	{
		misswts = *MISSWTS;
		residuals = *RESIDUALS;
		for(k = 0; k < ny; k++)
		{
			for (i = 0; i < ndata; i++)
			{
				if (misswts[i] == 0.0)
				{
					setMissing(residuals[i]);
				}
			}
			residuals += ndata;
		}
	}

	RETURN (2);

  errorExit:
	RETURN (0);
	
}
