/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#include "matdat.h"
#include "globals.h"

#include <setjmp.h>
#include <signal.h>
#include <stdlib.h>

#ifdef HASPWDH
#include <pwd.h>
#endif /*HASPWDH*/

#ifdef BCPP
#include <conio.h> /* for kbhit()*/
#endif /*BCPP*/

void            fmyprint();
void            fmyeol();
void            putprompt();
long            isfilename();

#define BEL '\007'

static             UpdateActive = -1;
#ifdef ACTIVEUPDATE

#define MAXSAVEDLINES    60
#define DEFAULTLENGTH    80
#define SAVEDCHUNK       20
#define lineI(I) (*(SavedLines[i]))

static char      **SavedLines[MAXSAVEDLINES];
static long        NSavedLines = 0;
static             LinePlace = 0;

static void   shiftSavedLines(void)
{
	short       i;
	char      **temp;
	
	if (NSavedLines > 0)
	{
		temp = SavedLines[NSavedLines-1];
		for (i = NSavedLines - 1; i > 0; i--)
		{
			SavedLines[i] = SavedLines[i - 1];
		}
		SavedLines[0] = temp;
		(*temp)[0] = '\0';
		LinePlace = 0;
	}
} /*shiftSaveLines()*/

void     saveForUpdate(char *msg)
{
	char      **line;
	short       i;
	short       msgLength;
	short       iseol;
	
	if(UpdateActive > 0 && *msg)
	{
		line = SavedLines[0];

		msgLength = strlen(msg);
		
		iseol = (msg[msgLength-1] == '\n');
		if (LinePlace + msgLength + 1 > myhandlelength(line))
		{
			line = mygrowhandle(line, LinePlace + msgLength + SAVEDCHUNK);
			SavedLines[0] = line;
			if(line == (char **) 0)
			{
				goto errorExit;
			}
		}
		strcat(*line + LinePlace, msg);
		LinePlace += msgLength;
		(*line)[LinePlace] = '\0';
		if(iseol)
		{
			shiftSavedLines();
		}
	} /*if(UpdateActive > 0 && *msg)*/
	return;

  errorExit:
/* don't know anything better to do that stop doing updating */
	for(i=0;i<NSavedLines;i++)
	{
		mydisphandle(SavedLines[i]);
		SavedLines[i] = (char **) 0;
	}
	NSavedLines = 0;
} /*saveForUpdate()*/

void     updateConsole(void)
{
	long     i, j, nlines= NLINES;
	char    *line;
	
	if(UpdateActive > 0)
	{
		UpdateActive = 0;
		
		for (i=NSavedLines - 1; i >= 0; i--)
		{
			line = lineI(i);			
			if(*line != '\0')
			{
				NLINES = 0; /* make sure it doesn't trigger pause */
				myprint(line);
			}
		} /*for (i=NSavedLines - 1; i >= 0; i--)*/
		NLINES = nlines;
		UpdateActive = 1;
	} /*if(UpdateActive > 0)*/
} /*updateConsole()*/
	
void activeUpdateInit(void)
{
	long        nlines = (SCREENHEIGHT) ? SCREENHEIGHT : 24;
	long        i;
	
	nlines = (nlines > MAXSAVEDLINES) ? MAXSAVEDLINES : nlines;
	for (i = NSavedLines; i < nlines; i++)
	{
		SavedLines[i] = mygethandle(DEFAULTLENGTH+1);
		if(SavedLines[i] == (char **) 0)
		{
			break;
		}
		(*SavedLines[i])[0] = '\0';
		NSavedLines++;
	}
	UpdateActive = (NSavedLines != 0) ? 1 : -1;

} /*activeUpdateInit*/
#else /*ACTIVEUPDATE*/
static void   shiftSavedLines(void)
{
} /*shiftSaveLines()*/

void     saveForUpdate(char *msg)
{
} /*saveForUpdate()*/

void     updateConsole(void)
{
} /*updateConsole()*/
	
void activeUpdateInit(void)
{
} /*activeUpdateInit*/

#endif /*ACTIVEUPDATE*/

/*
   Reads a line, including the terminating '\n', from fp into *LINE.
   If the line is longer than MAXLINE, it reads MAXLINE-1 characters,
   adds a '\n'  and skips to just past the next '\n'.
   If an EOF is encountered during the read, it returns EOF; otherwise
   it returns 1.  The calling function can determine how many characters
   were read by strlen(*LINE).
*/
static int fileToLINE(FILE * fp)
{
	int         c;
	long        n = 0, nc = MAXLINE;
	char       *buffer = *LINE;
	WHERE("fileToLINE");

	*buffer = '\0';
	c = getc(fp);
	if(c != EOF)
	{
		do
		{
#ifdef BINARYHELPFILE
			if(fp == HELPFILE && c == CR)
			{ /* discard CR when help file is opened in binary mode*/
				c = getc(fp);
				if (c == EOF)
				{
					break;
				}
			}
#endif /*BINARYHELPFILE*/
			*buffer++ = c;
			nc--;
			n++;
		} while(nc > 0 && c != CR && c != NL && (c = getc(fp)) != EOF);
	
		if(c != EOF)
		{
			buffer--;
		}
		*buffer++ = '\n';
		*buffer = '\0';
		
		while(c != CR && c != NL && c != EOF)
		{/* read remainder of the line */
			c = getc(fp);
		}
	} /*if(c != EOF)*/
	
	if(n > MAXLINE)
	{
		sprintf(OUTSTR,
				"WARNING: line longer than %ld characters; last %ld discarded",
				(long) MAXLINE, n - MAXLINE);
		putOUTSTR();
	}
	return ((c == EOF) ? EOF : 1);
}

static short    JustPaused = 0;

void incrementNLINES(char *msg)
{
	int           lines = 0;
	int           c, c1;
#ifdef USECBREAK
	char          c2;
#endif /*USECBREAK*/

	while (*msg != '\0')
	{
		if (*msg++ == '\n')
		{
			lines++;
		}
	}
	NLINES += lines;

#ifndef SCROLLABLEWINDOW
	if(ISATTY && SCREENHEIGHT > 0 && NLINES >= SCREENHEIGHT-1)
	{

		NLINES = 0;
		if (!JustPaused)
		{ /* don't pause twice in a row */
#ifdef USECBREAK
			fprintf(stderr,
					"Hit any key but 'q' to continue ");
			cbreak();
			read(0, &c2, 1);
			nocbreak();
			c = c2;
#else /*USECBREAK*/
			fprintf(stderr,
					"Hit RETURN to continue or q RETURN to go to next command line: ");
			c = getchar();
			if(c != '\n' && c != '\r')
			{
				while((c1 = getchar()) != '\n' && c1 != '\r' && c1 != EOF)
				{
					;
				}
			}
		
#endif /*USECBREAK*/
			JustPaused = 1;
			if(c == 'q' || c == 'Q')
			{                               
				INTERRUPT = PRINTABORT;
#ifdef SIGNALARG
				intRoutine(0);
#else /*SIGNALARG*/
				intRoutine();
#endif /*SIGNALARG*/
				/* no return */
			}
		} /*if (!JustPaused)*/
	} /*if(ISATTY && SCREENHEIGHT > 0 && NLINES >= SCREENHEIGHT-1)*/
	else
	{
		JustPaused = 0;
	}
	
#endif /*SCROLLABLEWINDOW*/
} /*incrementNLINES()*/

/*
  Write line of output to file fp
  If writing to console, interpret leading '\007's as BEL's
  If spooling is active, write line to spool file
  If line starts with "ERROR:", ring bell and count error
*/
  
void fmyprint(msg,fp)
char           *msg;
FILE           *fp;
{
	WHERE("fmyprint");

	if(fp == STDOUT)
	{
		while(*msg == BEL)
		{ /* turn leading BEL's into beeps */
			mybeep(1);
			msg++;
		}
		if (UpdateActive && SPOOLFILE != (FILE *) 0)
		{
			fputs(msg, SPOOLFILE);
		}

		if(UpdateActive && strncmp("ERROR:",msg, (int) 6) == 0)
		{
			if(MAXERRORS1)
			{
				NERRORS++;
			}
			mybeep(1);
		}
		if(PRINTWARNINGS || strncmp("WARNING:",msg, (int) 8) != 0)
		{
			fputs(msg, stdout);
			incrementNLINES(msg);
		}
#ifdef ACTIVEUPDATE
		saveForUpdate (msg);
#endif /*ACTIVEUPDATE*/
	} /*if(fp == STDOUT)*/
	else
	{
		fputs(msg, fp);
	}
} /*fmyprint()*/

/*
  terminate output line and pause if writing the console and screen is full
  (no pausing on Mac)
*/


void fmyeol(fp)
FILE          *fp;
{
	WHERE("fmyeol");
	
	fmyprint("\n",fp);
	
} /*fmyeol()*/

void fmyflush(fp)
FILE      *fp;
{
	fflush((fp == STDOUT) ? stdout : fp);
}

/*
  sound n beeps, but only if not reading batch file and output is to console
*/

#define BELL "\007"

void mybeep(n)
long        n;
{
	if(BDEPTH == 0 && ISATTY)
	{
		while(n-- > 0)
		{
			fputs(BELL,stdout);
		}
	}
}

/*
  Fill global LINE with line from FILE *fn
  On Dos and Unix, fn should always be non-null.
  On Mac, STDIN is (FILE *) 0, and LINE is filled from dialog window when 
  fn == STDIN
*/

int fillLINE(fn)
FILE        *fn;
{
	if(fn != (FILE *) 0)
	{
		return (fileToLINE(fn));
	}
	else
	{ /* should not happen on UNIX (on Mac this means keyboard) */
		(*LINE)[0] = '\0';
		return 0;
	}
}

/* 
  If argument prompt is NULL, use global PROMPT (BDEPTH == 0)
  or batch file name (BDEPTH > 0), preceded by EOL.  This should only
  happen at first prompt for a command line.
  If prompt != NULL (probably "more> "), print it with no preceding EOL
*/

void putprompt(prompt)
char        *prompt;
{
	char    *prmpt;
	WHERE("putprompt");

	/* write out the prompt */
	prmpt = (prompt == (char *) 0) ? PROMPT : prompt;

	if(BDEPTH == 0)
	{
		if(prompt == (char *) 0)
		{
			myeol();
		}
		myprint(prmpt);
	}
	else if(BATCHECHO[BDEPTH])
	{
		if(prompt == (char *) 0)
		{
			myeol();
			myprint(*INPUTFILENAMES[BDEPTH]);
			myprint("> ");
		}
		else
		{
			myprint(prmpt);
		}
	}
}

#ifdef MSDOS
/*  '/' equivalent to '\' on MSDOS*/
#define isSlash(C) ((C) == DIRSEPARATOR[0] || (C) == '/')
#else /*MSDOS*/
#define isSlash(C) ((C) == DIRSEPARATOR[0])
#endif /*MSDOS*/

/* 
  Expand leading '~' in a filename.  A name of the form ~/path (or ~\path
  under MSDOS) is expanded to home/path or home\path, where home is the value
  of MacAnova CHARACTER variable HOME.  Under Unix, HOME is predefined to
  have value taken from environmental variable $HOME.  Under MSDOS, HOME is
  predefined to contain the name of the directory containing MACANOVA.EXE.
  Under Unix, a name of form ~username/path is expanded to homedir/path where
  getpwname() is used to get the home directory homedir for user username.
  Such a file name is not expanded under MSDOS and results in an error.

  950829 Fixed bug in DJGPP version
         Under MSDOS, any embedded '/' are changed to DIRSEPARATOR[0]
         (backslash))
*/

static char    FullPathName[PATHSIZE];

char *expandFilename(fname)
char         *fname;
{
	Symbolhandle    home;
	char           *pc, c;
	char           *userdir = (char *) 0;
	long            dirNameLength;
#ifdef HASPWDH
	struct passwd  *passwd;
#endif /*HASPWDH*/
		
	if(fname[0] == TILDE)
	{
		if (isSlash(fname[1]))
		{/* treat ~/... as $HOME/... */
			home = Lookup("HOME"); /* should be initialized at startup */
			if(home == (Symbolhandle) 0 || TYPE(home) != CHAR ||
			   !isScalar(home))
			{
				userdir = getenv("HOME");
			}
			else
			{
				userdir = STRINGPTR(home);
			}
			
			pc = fname+2;
		} /*if (isSlash(fname[1]))*/
		else
		{
#ifndef HASPWDH
			/* can't expand ~name on except on Unix */
			userdir = (char *) 0;
#else /*HASPWDH*/
			for (pc = fname+1;*pc && !isSlash(*pc);pc++)
			{
				;
			}
			c = *pc;
			*pc = '\0';
			passwd = getpwnam(fname+1);
			*pc = c;
			if(c != '\0')
			{
				pc++;
			}
				
			userdir = (passwd != (struct passwd *) 0) ?
				passwd->pw_dir :  (char *) 0;
#endif /*HASPWDH*/
		} /* if (isSlash(fname[1])){}else{}*/
		
		if(userdir == (char *) 0)
		{
			putPieces("ERROR: unable to expand ", fname,
					  (char *) 0, (char *) 0);
			goto errorExit;
		}
			
		dirNameLength = strlen(userdir);
/* allow for the possiblity that HOME ends in '/' */
		if (isSlash(userdir[dirNameLength-1]))
		{
			dirNameLength--;
		}
		
		if(dirNameLength + strlen(pc) + 1 >= PATHSIZE)
		{
			putPieces("ERROR: expanded file name ",userdir,pc," is too long");
			goto errorExit;
		}
		
		strcpy(FullPathName,userdir);
		FullPathName[dirNameLength++] = DIRSEPARATOR[0];
		FullPathName[dirNameLength] = '\0';
		strcat(FullPathName,pc);
	} /*if(fname[0] == TILDE)*/
	else
	{
		if(strlen(fname) >= PATHSIZE)
		{
			putPieces("ERROR: file name ", fname, " is too long", (char *) 0);
			goto errorExit;
		}
		
		strcpy(FullPathName, fname);
	}
	
#ifdef MSDOS
	for (pc = FullPathName; *pc; pc++)
	{
		if (isSlash(*pc))
		{
			*pc = DIRSEPARATOR[0];
		}
	} /*for (pc = FullPathName; *pc; pc++)*/
#endif /*MSDOS*/
	
	return ((char *) FullPathName);
		
  errorExit:

	return ((char *) 0);

} /*expandFilename()*/
	
/*
  Check file names for legality on platform
*/

#define isDot(C) ((C) == '.')

#ifdef MSDOS

/*
  For MSDOS accept names of the form
  pathElement/.../pathElement, or /pathElement/.../pathElement,
  where pathElement is of form '.', '..', Name or Name.Ext, and
  Name has from 1 to 8 each characters and Ext has from 0 to three
  characters
  '/' may be replaced by '\'.
  Either form may be prefixed by a drive designator of the form [a-zA-Z]:
*/
static int filenameFormOK(char *fname)
{
	int        dotFound = 0;
	char      *start = fname, *pc;
	WHERE("filenameFormOK");

	for(pc = start;*pc != '\0';pc++)
	{
		if(isDot(*pc))
		{
			if(pc == start && !dotFound)
			{
				if(isSlash(pc[1]))
				{ /* pathElement is '.' */
					pc++;
				}
				else if(isDot(pc[1]) && isSlash(pc[2]))
				{ /* pathElement is '..' */
					pc += 2;
				}
				else
				{ /* no Name; error */
					return (0);
				}
			} /*if(pc == start && !dotFound)*/
			else
			{ /* Name completed */
				dotFound = 1;
			}
			start = pc + 1;
		} /*if(isDot(*pc))*/
		else if(isSlash(*pc))
		{ /* end of pathElement */
			dotFound = 0;
			start = pc + 1;
		}
		else if(*pc == ':')
		{ /* allow [a-z]: at start of filename */
			if(pc != fname + 1 || !isalpha(*fname))
			{
				return (0);
			}
			start = pc + 1;
		}
		else if(pc >= start + 8 || dotFound && pc >= start + 3)
		{ /* Name or Ext too long */
			return (0);
		}
	} /*for(pc = start;*pc;pc++)*/
	return (1);
} /* filenameFormOK()*/

#else /*MSDOS*/

/*
  On Unix accept non-empty names of the form
       pathElement or pathElement/pathElement or pathElement/.../pathElement
  where a pathElement does not contain '/' but can be empty/  Note that
  this includes paths with repeated '/' (e.g., "foo//bar") or
  of the form "../bar"
*/
int filenameFormOK(char * fname)
{
	/*
	   only current restriction is that it can't end in '/';
	   or start with '-'
	   repeated '/'s allowed
	*/
	return (fname[0] != '-' && !isSlash(fname[strlen(fname)-1]));

} /* filenameFormOK()*/

#endif /*MSDOS*/

/*
   Check for illegal characters in fname
   All alphanumerics are legal plus the characters in okNonAlnum.
   Since fname has been checked for form before being checked
   here, special characters like '/' are included in okNonAlnum
*/
static int filenameCharsOK(char * fname)
{
#ifdef MSDOS
	char      *okNonAlnum = ".\\/:_-`!@#$%^&(){}'";
#else /*MSDOS*/
	/*don't allow (){}[]`'";|\~ even though that are technically legal */
	char      *okNonAlnum = "./_,-+!@#$%^&";
#endif /*MSDOS*/
	char       c;
	
	while((c = *fname++) != '\0')
	{
		if (!isalnum(c) && !strchr(okNonAlnum, c))
		{
			return (0);
		}
	} /*while((c = *fname++) != '\0')*/
	return (1);
} /*filenameCharsOK()*/

long isfilename(char *fname)
{
	char     *pc;
	
	if(*fname == '\0')
	{
		myerrorout("ERROR: zero length file name");
		goto errorExit;
	}
	
	/* check for legal form for file name (does not check for legal chars) */

	if (!filenameFormOK(fname) || !filenameCharsOK(fname))
	{
		goto badName;
	} /*if (!filenameFormOK(fname))*/
	
	return (1);

  badName:
	putPieces("ERROR: improper file name ", fname, (char *) 0, (char *) 0);
	
  errorExit:
	return (0);
} /*isfilename()*/


#ifdef UNDEFINED__

static char Path[PATHSIZE] = ".\\";

char *get_dataPath()
{
	char *envPath;
	envPath = getenv("MACANOVA");
	if( envPath == (char *) 0) 
	{
		myerrorout("WARNING: MACANOVA DOS environment variable not set.");
		return( Path );
	}
	if( strlen(envPath) > PATHSIZE - 2)
	{
		sprintf(OUTSTR,
		   "ERROR: MACANOVA DOS environment variable too long (> %ld)",
			(long) PATHSIZE);
		putOUTSTR();
		return( Path );
	}
	strcpy(Path,envPath);
	Path[strlen(Path)+1] = 0;
	Path[strlen(Path)] = '\\';
	return(Path);
} /*get_dataPath()*/

#endif /*UNDEFINED__*/

#ifdef MSDOS

#ifdef BCPP
#include <dos.h>
#define ArgV _argv
void mypause(void)
{
	fprintf(stderr,"Hit any key to go on");
	while(!kbhit())
	{
		;
	}
} /*mypause()*/
#endif /* BCPP */

extern char      **ArgV;
static char        Path[PATHSIZE];
/* instead of string.h */
extern char *strrchr();

char *get_dataPath()
{
	char    *pathp;

	if( strlen( ArgV[0] ) > PATHSIZE )
	{
		myerrorout("ERROR: Path to MacAnova.exe too long");
		Path[0] = '\0';
		return(Path);
	}
	(void) strcpy(Path,ArgV[0]);
	if( (pathp = strrchr(Path,'\\')) == (char *) 0 )
	{
		pathp = strrchr(Path,'/');
	}
	if(pathp)
	{
		*(++pathp) = 0x00; 
	}
	else
	{
		Path[0] = '\0';
	}
	
	return(Path);
} /*get_dataPath()*/
#endif /*MSDOS*/
