/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Utils
#else
#define __SEG__ Utils
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "plot.h"

long                 argOK();
long                 isNull(),isDefined();
long                 isScalar(), isVector(), isMatrix();
long                 isFactor();
long                 symbolSize(), checkSymmetry();
void                 installFormat();
void                 saveFormat();
void                 restoreFormat();
long                 matchKey();
long                 anyMissing();
char                *isKeyword();
char               **getKeyword();
char                *typeName();
char                *opName();
char                *tokenName();

long argOK(arg,type,argno)
Symbolhandle         arg;
long                 type, argno;
{
	long             reply = 0;

	if(isNull(arg))
	{
		noData(FUNCNAME, labs(argno));
	}
	else if(!isDefined(arg))
	{
		undefArg(FUNCNAME,arg,labs(argno));
	}
	else if(type > 0 && TYPE(arg) != type || type < 0 && TYPE(arg) == -type)
	{ /* do not check type if type == 0 */
		badType(FUNCNAME,type,argno);
	}
	else
	{
		reply = 1;
	}
	return (reply);
} /*argOK()*/

long isNull(arg)
Symbolhandle          arg;
{
	/* NOTE calling program should check separately for UNDEF */
	return (arg == (Symbolhandle) 0 || (DIMVAL(arg,1) == 0 && isDefined(arg)));
} /*isNull()*/

/*
   Returns 1 if and only if the handle is non-null and the type is neither
   UNDEF or ASSIGNED (an ASSIGNED is an UNDEF that was added to the
   regular symbol table because it appeared on the l.h.s. of an assignment
*/

long isDefined(arg)
Symbolhandle          arg;
{
	return (arg != (Symbolhandle) 0 && TYPE(arg) != UNDEF &&
			TYPE(arg) != ASSIGNED);
} /*isDefined()*/

/*
   return non-zero if and only if arg is a non-null REAL, CHARACTER or
   LOGICAL variable of length 1
*/

long isScalar(arg)
Symbolhandle          arg;
{
	long       type;
	
	return (symbolSize(arg) == 1 &&
		((type = TYPE(arg)) == REAL || type == LOGIC || type == CHAR));
} /*isScalar()*/

/*
  Anything whose only dimension > 1 is its first is to be considered a vector
*/
long isVector(arg)
Symbolhandle          arg;
{
	long           type;
	
	return (isDefined(arg) && !isNull(arg) &&
			symbolSize(arg) == DIMVAL(arg,1) &&
			((type = TYPE(arg)) == REAL || type == LOGIC || type == CHAR));

} /*isVector()*/

/*
  Anything with at most 2 dimensions > 1 is to be considered a matrix.

  An object with length 1 is considered 1 by 1.

  An vector of length n is considered to be a n by 1 matrix.

  A [1 x ] ... [x 1 x] m x [1 x ] ... [x 1 x] n [x 1 x ... x 1] object is
  considered to be an m by n matrix, where m>1,n>1.

  A m x 1 [x 1 ... x 1] object is considered to be a m by 1 object.

  A 1 x n [x 1 x . . . x 1] object is considered to be a 1 by n object

  The row and column dimensions are returned in dim[].
*/

long isMatrix(arg,dim)
Symbolhandle          arg;
long                 *dim;
{
	long              ndims, i;
	WHERE("isMatrix");


	if(isNull(arg) ||
	   (TYPE(arg) != REAL && TYPE(arg) != LOGIC && TYPE(arg) != CHAR))
	{
		if(dim != (long *) 0)
		{
			dim[0] = dim[1] = 0;
		}

		return (0);
	}

	if(symbolSize(arg) == 1)
	{
		if(dim != (long *) 0)
		{
			dim[0] = dim[1] = 1;
		}

		ndims = 2;
	}
	else
	{
		ndims = 0;
		for(i=1;i<=NDIMS(arg);i++)
		{
			if(DIMVAL(arg,i) > 1)
			{
				if(ndims >= 2)
				{
					return (0);
				}
				if(dim != (long *) 0)
				{
					dim[ndims] = DIMVAL(arg,i);
				}
				ndims++;
			}
		}
	}

	if(ndims == 1 && dim != (long *) 0)
	{
		if(DIMVAL(arg,1) == 1)
		{
			dim[1] = dim[0];
			dim[0] = 1;
		}
		else
		{
			dim[1] = 1;
		}
	}
	return(1);

} /*isMatrix()*/

/*
  returns 0 if all MISSING or if there are non positive integer values
  returns NCLASS(symh) if NCLASS(symh) > 0
  returns -(max value) otherwise, i.e., -NCLASS(symh) if symh were strict
  factor
  It treats logical variable as having levels 1 (F) and 2(T)
*/

long isFactor(symh)
Symbolhandle        symh;
{
	long            nclass = -1;
	long            n, i, type;
	double         *in;
	double          y;

	if (symh != (Symbolhandle) 0 && isVector(symh) &&
		((type = TYPE(symh)) == REAL || type == LOGIC))
	{
		if((nclass = NCLASS(symh)) < 0)
		{ /* not a 'factor'; check to see if it could be */
			nclass = 0;
			n = DIMVAL(symh,1);
			in = DATAPTR(symh);
			for (i = 0; i < n; i++)
			{
				y = in[i];
				if (!isMissing(y))
				{
					if(type == REAL)
					{
						if(y != floor(y) || y < 1.0)
						{		/* not positive integer */
							nclass = 0;
							break;
						}
						if (y > nclass)
						{
							nclass = (long) y;
						}
					}
					else
					{ /* at least 1 non-missing */
						nclass = 2;
						break;
					}
				}/*if (!isMissing(y))*/
			} /* for i */
			if(nclass == -1)
			{ /* all were missing */
				nclass = 0;
			}
			else
			{
				nclass = -nclass; /* value < 0 => not true factor */
			}
		} /*if((nclass = NCLASS(symh)) < 0)*/
	}
	else
	{
		nclass = 0;
	}

	return (nclass);
} /*isFactor()*/

#define FUZZ  1e-7
/*
   Check the symmetry of a square matrix.  It returns non-zero if and only
   if |a[i,j]-a[j,i]| < FUZZ * average(|a[*,*]|) all i and j
   It is assumed there are no MISSING values (caller must screen)
*/

long checkSymmetry(a,nrows)
double               *a;
long                  nrows;
{
	double            sum, *aj = a, *aii, fuzz;
	long              i, j, k;
	
	sum = 0.0;

	for (j = 0; j < nrows; j++)
	{
		for (i = 0; i < nrows; i++)
		{
			sum += fabs(aj[i]);
		}
		aj += nrows;
	}

	if(sum == 0.0)
	{
		return (1);
	}

	fuzz = FUZZ * (sum / (double) (nrows * nrows));
	aii = aj = a;
	for (j = 0; j < nrows; j++)
	{
		k = nrows;
		for (i = j+1; i < nrows; i++)
		{
			if (fabs(aj[i] - aii[k]) > fuzz)
			{
				return (0);
			}
			k += nrows;
		}
		aii += nrows+1;
		aj += nrows;
	}
	return (1);
} /*checkSymmetry()*/

/*
   Returns the product of the dimensions.  For a STRUC this will be the
   number of components
*/
long symbolSize(arg)
Symbolhandle arg;
{
	long totl;
	long i,n;

	if(!isDefined(arg))
	{
		return (0);
	}

	n = NDIMS(arg);
	totl = DIMVAL(arg,1);
	for(i=2;i<=n;i++)
	{
		totl *= DIMVAL(arg,i);
	}
	return totl;
} /*symbolSize()*/

/*
   copy the dimensions of a symbol and return the product.  The number
   of dimensions is returned in dims[0] and the dimensions themselves
   in dims[1], dims[2], ...
*/
long dupDims(dims,arg)
long          *dims;
Symbolhandle   arg;
{
	long       totl = 1;
	long       i, n, m;

	if(isNull(arg))
	{
		if(dims)
		{
			*dims = 0;
		}
		return (0);
	}

	n = NDIMS(arg);
	for(i=0;i<=n;i++)
	{
		m = DIMVAL(arg,i);
		if(dims)
		{
			*dims++ = m;
		}

		if(i != 0)
		{
			totl *= m;
		}
	}
	return (totl);
} /*dupDims()*/

#undef UNDEFINED__
#ifdef UNDEFINED__ /* getKeyword no longer used.  It was too complicated */
/*
  Function to decode keywords and return handle to value.  specs should be
  of the form shape+type where shape is one of SCALAR, VECTOR, MATRIX, or ARRAY
  and type is one of REAL, LOGIC, CHAR, sor STRUC.

  If severity != 0, error messages start with "ERROR"; otherwise they start
  with "WARNING" and end with "ignored".

  If arg is not a keyword or if its name does not match name, (char **) 0
  is returned.  If TYPE(arg) is wrong, BADTYPE is returned; if the shape is
  wrong BADSPECS is returned.
*/

char **getKeyword(arg,name,specs, severity)
Symbolhandle      arg;
char             *name;
long              specs,severity;
{
	long          type,dim[2];
	char        **result;
	char         *keywordName;
	char         *error, *ignored;

	WHERE("getKeyword");

	OUTSTR[0] = '\0';
	error = (severity) ? "ERROR" : "WARNING";
	ignored = (severity) ? "" : "; ignored";

	if(arg == (Symbolhandle) 0 ||
	   (keywordName = isKeyword(arg)) == (char *) NULL ||
	   strcmp(keywordName,name) != 0)
	{
		result = (char **) 0;
	}
	else if(!isDefined(arg))
	{
		sprintf(OUTSTR,
				"%s: value for keyword '%s'is undefined%s",
				error,name,ignored);
		result = (char **) BADTYPE;
	}
	else
	{
		type = TYPEMASK & specs;

		if(TYPE(arg) != type)
		{
			sprintf(OUTSTR,"%s: keyword '%s' not type %s%s",
					error, name, typeName(type),ignored);
			result = (char **) BADTYPE;
		}
		else
		{
			switch ((int) type)
			{
			  case REAL:
			  case LOGIC:
			  case LIST:
				result = (char **) DATA(arg);
				break;
			  case CHAR:
			  case MACRO:
			  case MACRO1:
				result = STRING(arg);
				break;
			  case PLOTINFO:
				result = (char **) GRAPH(arg);
				break;
			  case STRUC:
				result = (char **) COMP(arg);
			}
			if(specs & SPECMASK)
			{
				if(specs & SCALAR && !isScalar(arg))
				{
					sprintf(OUTSTR,"%s: keyword '%s' not scalar%s",
							error, name, ignored);
				}
				else if(specs & VECTOR && !isVector(arg))
				{
					sprintf(OUTSTR,"%s: keyword '%s' not scalar%s",
							error, name, ignored);
				}
				else if(specs & MATRIX && !isMatrix(arg,dim))
				{
					sprintf(OUTSTR,"%s: keyword '%s' not matrix%s",
							error, name, ignored);
				}
			}
			if(*OUTSTR)
			{
				result = (char **) BADSPECS;
			}
		}
	}
	putOUTSTR();

	return (result);

} /*getKeyword()*/
#endif /*UNDEFINED__*/

static char      TempKeywordName[NAMELENGTH+1];
/*
  returns a pointer to the keyword name Name if NAME(arg) is of the form
  '@@Name'.  If NAME(arg) is not of this form it return (char *) 0.
*/

char *isKeyword(arg)
Symbolhandle     arg;
{
	char        *name;
	WHERE("isKeyword");

	if(arg == (Symbolhandle) 0 ||
	   *(name = NAME(arg)) != KEYPREFIX1 || name[1] != KEYPREFIX2)
	{
		return (char *) 0;
	}

	strncpy(TempKeywordName, name+2, NAMELENGTH);
	TempKeywordName[NAMELENGTH] == '\0';

	return ((char *) TempKeywordName);
} /*isKeyword()*/

/*
  search argument list for presence of specified keyword.  Return index in list
  handle if the first strlen(word) characters of the keyword match word,
  -1 otherwise.
*/

long findKeyword(list,word,start)
Symbolhandle        list;
char               *word;
long                start;
{
	long            nargs = NARGS(list);
	long            i,length = strlen(word);
	char           *keyword;

	for(i=start;i<nargs;i++)
	{
		if((keyword = isKeyword(COMPVALUE(list,i))) &&
		   strncmp(keyword,word,(size_t) length) == 0)
		{
			return (i);
		}
	}
	return (-1);
} /*findKeyword()*/

/*
  attempt to match name in charvals.  If matched, return the corresponding
  value from codes; otherwiser return 0.
  If codes is NULL, return index+1
  Useful for translating character value to defined code for use in switch
*/

long matchKey(name, charvals, codes)
char              *name, *charvals[];
long               codes[];
{
	long           i;

	for(i=0;charvals[i];i++)
	{
		if(strcmp(charvals[i],name) == 0)
		{
			return ( (codes != (long *) 0) ? codes[i] : i+1);
		}
	}
	return (0);
} /*matchKey()*/

/* comparison of strings ignoring case */
long mystrncmp(char *s1, char *s2, int n)
{
	int c1, c2;
	int i;
	WHERE("mystrncmp");

	for(i=0;i < n;i++)
	{
		c1 = s1[i];
		c2 = s2[i];
		c1 = (isupper(c1)) ? tolower(c1) : c1;
		c2 = (isupper(c2)) ? tolower(c2) : c2;
		if(c1 != c2)
		{
			return c1-c2;
		}
		else if(c1 == '\0')
		{
			return 0;
		}
	}
	c1 = s1[n];
	c2 = s2[n];
	return (((c1 == '\0' || isspace(c1)) && (c2 == '\0' || isspace(c2))) ?
			0 : 1);
	
} /*mystrncmp()*/

/*
	Decode string and check whether it is a number
*/
#define EMPTYSTRING     1
#define NOTANUMBER      2
#define TRAILINGSTUFF   3

double decodeString(char * string, char sep, long * error)
{
	char        *eptr, c;
	double       value = mystrtod(string, &eptr);
	WHERE("decodeString");

	*error = (*string == '\0') ? EMPTYSTRING :
		((eptr == string) ? NOTANUMBER : 0);

	if(!*error && *eptr != '\0' && *eptr != '\n')
	{
		if(sep == '\0')
		{
			*error = (*eptr == ' ' || *eptr == '\t') ? 0 : TRAILINGSTUFF;
		}
		else
		{
			while((c = *eptr) != sep && (c == ' ' || c == '\t'))
			{ /*skip trailing blanks && tabs*/
				eptr++;
			}
			*error = (*eptr != '\0') ? TRAILINGSTUFF : 0;
		}
	}
	return (value);
} /*decodeString()*/

long setFormat(format, fmt)
char            *format;
long             fmt[]; /* beforeDec, afterDec, fmtType */
{
	char        *legalChars = "fg";
	char        *pc,*pc1,*pdec;

	fmt[0] = fmt[1] = fmt[2] = 0;
	for(pc = format;isdigit(*pc);pc++)
	{
		;
	}
	if(pc == format)
	{
		fmt[0] = -1;
	}

	if(*pc == '.')
	{
		pc++;
	}
	pdec = pc;
	for(pc = pdec;isdigit(*pc);pc++)
	{
		;
	}
	if(pc == pdec)
	{
		fmt[1] = 6;
	}

	for(pc1 = legalChars;*pc1;pc1++)
	{
		if(*pc1 == *pc)
		{
			break;
		}
	}

	pc++;
	if(*pc1 == '\0' || *pc != '\0')
	{
		return (0);
	}

	fmt[2] = (long) *pc1;
	if(fmt[1] == 0)
	{
		sscanf(pdec,"%ld",&fmt[1]);
	}
	if(fmt[0] == 0)
	{
		sscanf(format,"%ld",&fmt[0]);
	}
	else if(fmt[0] < 0)
	{
		fmt[0] = fmt[1] + 7;
	}
	if (fmt[0] >= BUFFERLENGTH)
	{
		fmt[0] = BUFFERLENGTH - 10;
	}
	
	return (1);
} /*setFormat()*/

void installFormat(beforeDec, afterDec, fmtType)
long             beforeDec, afterDec;
long             fmtType;
{
	WHERE("installFormat");
	sprintf(DATAFMT, " %%%ld.%ld%c", beforeDec, afterDec, (char) fmtType);
	sprintf(MISSINGFMT, " %%%lds", beforeDec);
	FIELDWIDTH = beforeDec + 1;
	BEFOREDEC = beforeDec;
	AFTERDEC = afterDec;
	FMTTYPE = fmtType;
} /*installFormat()*/

static char      saveDATAFMT[20], saveFMTTYPE;
static char      saveMISSINGFMT[20];
static long      saveFIELDWIDTH,  saveAFTERDEC,  saveBEFOREDEC;

void saveFormat(void)
{
	strcpy(saveMISSINGFMT, MISSINGFMT);
	strcpy(saveDATAFMT, DATAFMT);
	saveFIELDWIDTH = FIELDWIDTH;
	saveFMTTYPE = FMTTYPE;
	saveAFTERDEC = AFTERDEC;
	saveBEFOREDEC = BEFOREDEC;
} /*saveFormat()*/

void restoreFormat(void)
{
	strcpy(MISSINGFMT, saveMISSINGFMT);
	strcpy(DATAFMT, saveDATAFMT);
	FIELDWIDTH = saveFIELDWIDTH;
	FMTTYPE = saveFMTTYPE;
	AFTERDEC = saveAFTERDEC;
	BEFOREDEC = saveBEFOREDEC;
} /*restoreFormat()*/

/*
  function to check for missing values in argument
*/

long anyMissing(a)
Symbolhandle         a;
{
	register long       n;
	register double    *x;
	register char      *string;
	long                dupDIms();

	if(!isNull(a))
	{
		n = symbolSize(a);
		if(TYPE(a) == REAL || TYPE(a) == LOGIC)
		{
			x = DATAPTR(a);
			while(n-- > 0)
			{
				if(isMissing(*x++))
				{
					return (1);
				}
			}
		}
		else if(TYPE(a) == CHAR)
		{
			string = STRINGPTR(a);
			while(n-- > 0)
			{
				if(*string == '\0')
				{
					return (1);
				}
				string += strlen(string) + 1;
			}
		}
	} /*if(!isNull(a))*/

	return (0);
} /*anyMissing()*/


void setScratchName(symh)
Symbolhandle       symh;
{
	char          *name;
	int            type;

	if(symh != (Symbolhandle) 0 && isDefined(symh) && isscratch(NAME(symh)))
	{
		type = TYPE(symh);
		switch (NDIMS(symh))
		{
		  case 1:
			if(type == STRUC)
			{
				name = STRUCSCRATCH;
			}
			else if(type == MACRO)
			{
				name = MACROSCRATCH;
			}
			else if(type == PLOTINFO)
			{
				name = GRAPHSCRATCH;
			}
			else if(DIMVAL(symh,1) == 1)
			{
				switch (type)
				{
				  case CHAR:
					name = STRINGSCRATCH;
					break;

				  case REAL:
					name = NUMSCRATCH;
					break;

				  case LOGIC:
					name = LOGICSCRATCH;
					break;
				} /* switch(type) */
			}
			else
			{
				name = VECSCRATCH;
			}
			break;

		  case 2:
			name = MATSCRATCH;
			break;
		  default:
			name = ARRAYSCRATCH;
		} /* switch(ndims) */
		setNAME(symh,name);
	}
} /*setScratchName()*/

/* set component names */
void setCompName(symh,name)
Symbolhandle       symh;
char              *name;
{
	int            type;
	WHERE("setCompName");

	if(symh != (Symbolhandle) 0)
	{
		if(strcmp(name, LOGICSCRATCH + 2) == 0 ||
		   strcmp(name, NUMSCRATCH + 2) == 0 ||
		   strcmp(name, STRINGSCRATCH + 2) == 0 ||
		   strcmp(name, SCALSCRATCH + 2) == 0 ||
		   strcmp(name, VECSCRATCH + 2) == 0 ||
		   strcmp(name, MATSCRATCH + 2) == 0 ||
		   strcmp(name, ARRAYSCRATCH + 2) == 0 ||
		   strcmp(name, MACROSCRATCH + 2) == 0 ||
		   strcmp(name, GRAPHSCRATCH + 2) == 0 ||
		   strcmp(name, STRUCSCRATCH + 2) == 0 ||
		   strcmp(name, "NULL") == 0)
		{
			name = SCRATCH;
		}

		if(isscratch(name))
		{
			type = TYPE(symh);
			switch (type)
			{
			  case STRUC:
				name = STRUCSCRATCH+2;
				break;

			  case MACRO:
				name = MACROSCRATCH+2;
				break;

			  case PLOTINFO:
				name = GRAPHSCRATCH+2;
				break;

			  default:
				if(NDIMS(symh) == 0 || DIMVAL(symh,1) == 0)
				{
					name = "NULL";
				}
				else if(NDIMS(symh) == 1)
				{
					if(DIMVAL(symh,1) == 1)
					{
						switch (type)
						{
						  case CHAR:
							name = STRINGSCRATCH+2;
							break;

						  case REAL:
							name = NUMSCRATCH+2;
							break;

						  case LOGIC:
							name = LOGICSCRATCH+2;
							break;
						}		/* switch(type) */
					}
					else
					{
						name = VECSCRATCH+2;
					}
				}
				else if(NDIMS(symh) == 2)
				{
					name = MATSCRATCH+2;
				}
				else
				{
					name = ARRAYSCRATCH+2;
				}
			} /* switch(type) */
		} /*if(isscratch(name))*/
		else if(iskeyname(name))
		{
			name += 2;
		}
		else if(istempname(name))
		{
			name++;
		}
		setNAME(symh,name);
	} /*if(symh != (Symbolhandle) 0)*/
} /*setCompName()*/

static char TempName[36];

char *typeName(type)
long         type;
{
	switch ((int) type)
	{
	  case REAL:
		return ("REAL");
	  case LOGIC:
		return ("LOGIC");
	  case CHAR:
		return ("CHAR");
	  case LONG:
		return ("LONG");
	  case GARB:
		return ("GARBAGE");
	  case LIST:
		return("LIST");
	  case BLTIN:
		return ("BLTIN");
	  case MACRO:
		return ("MACRO");
	  case MACRO1:
		return ("MACRO1");
	  case STRUC:
		return ("STRUC");
	  case PLOTINFO:
		return ("GRAPH");
	  case UNDEF:
		return ("UNDEF");
	  case ASSIGNED:
		return ("ASSIGNED");
	  default:
		sprintf(TempName,"unknown type = %ld",type);
		return ((char *) TempName);
	}
} /*typeName()*/

static char    TempOpName[36];

char *opName(op)
long         op;
{
	switch ((int) op)
	{
	  case ADD:
	  case '+':
		return ("+");
	  case SUB:
	  case '-':
		return ("-");
	  case MULT:
		return ("*");
	  case MATMULT:
		return("%*%");
	  case TRMATMULT:
		return ("%c%");
	  case MATMULTTR:
		return ("%C%");
#ifdef BITOPS
	  case BITNOT:
		return ("%!");
	  case BITAND:
		return ("%&");
	  case BITXOR:
		return ("%^");
	  case BITOR:
		return ("%|");
#endif /*BITOPS*/
	  case DIV:
		return ("/");
	  case MOD:
		return ("%%");
	  case EXP:
		return ("^");
	  case EQ:
		return ("==");
	  case NE:
		return ("!=");
	  case LE:
		return ("<=");
	  case GE:
		return (">=");
	  case LT:
		return ("<");
	  case GT:
		return (">");
	  case AND:
		return ("&&");
	  case OR:
		return ("||");
	  case NOT:
		return ("!");
	  case ASSIGNADD:
		return ("<-+");
	  case ASSIGNSUB:
		return ("<--");
	  case ASSIGNMULT:
		return ("<-+*");
	  case ASSIGNDIV:
		return ("<-/");
	  case ASSIGNPOW:
		return ("<-^");
	  case ASSIGNMOD:
		return ("<-%%");
	  default:
		sprintf(TempOpName,"unknown operation %ld",op);
		return ((char *) TempOpName);
	}
} /*opName()*/


char *tokenName(token)
long         token;
{
	switch ((int) token)
	{
	  case REAL:
	  case LOGIC:
	  case CHAR:
	  case LIST:
	  case BLTIN:
	  case MACRO:
	  case MACRO1:
	  case STRUC:
	  case UNDEF:
	  case ASSIGNED:
		return (typeName(token));
	  case VAR:
		return("VAR");
	  case END:
		return("END");
	  case NUMBER:
		return("NUMBER");
	  case IF:
		return("IF");
	  case ELSE:
		return("ELSE");
	  case ELSEIF:
		return("ELSEIF");
	  case WHILE:
		return("WHILE");
	  case FOR:
		return("FOR");
	  case BREAK:
		return("BREAK");
	  case BREAKALL:
		return("BREAKALL");
	  case BATCH:
		return("BATCH");
	  case ERROR:
		return("ERROR");
	  case FATALERROR:
		return("FATALERROR");
	  case LEFTANGLE:
		return("LEFTANGLE");
	  case RIGHTANGLE:
		return("RIGHTANGLE");
	  case LEFTBRACK2:
		return("LEFTBRACK2");
	  case RIGHTBRACK2:
		return("RIGHTBRACK2");
	  case MULT:
		return("MULT");
	  case ADD:
		return("ADD");
	  case DIV:
		return("DIV");
	  case SUB:
		return("SUB");
	  case MOD:
		return ("MOD");
#ifdef BITOPS
	  case BITNOT:
		return ("BITNOT");
	  case BITAND:
		return ("BITAND");
	  case BITXOR:
		return ("BITXOR");
	  case BITOR:
		return ("BITOR");
#endif /*BITOPS*/
	  case EXP:
		return("EXP");
	  case GE:
		return("GE");
	  case LE:
		return("LE");
	  case GT:
		return("GT");
	  case LT:
		return("LT");
	  case EQ:
		return("EQ");
	  case NE:
		return("NE");
	  case AND:
		return("AND");
	  case OR:
		return("OR");
	  case NOT:
		return("NOT");
	  case ENDOFL:
		return("ENDOFL");
	  case POW:
		return("POW");
	  case MATMULT:
		return("MATMULT");
	  case TRMATMULT:
		return("TRMATMULT");
	  case MATMULTTR:
		return("MATMULTTR");
	  case ASSIGN:
		return("ASSIGN");
	  case ASSIGN1:
		return("ASSIGN1");
	  case ASSIGNADD:
		return ("ASSIGNADD");
	  case ASSIGNSUB:
		return ("ASSIGNSUB");
	  case ASSIGNMULT:
		return ("ASSIGNMULT");
	  case ASSIGNDIV:
		return ("ASSIGNDIV");
	  case ASSIGNPOW:
		return ("ASSIGNPOW");
	  case ASSIGNMOD:
		return ("ASSIGNMOD");
	  case BADTOKEN:
		return("BADTOKEN");

	  default:
		if(token > ' ' && token <= '~')
		{
			sprintf(TempName,"'%c'",token);
		}
		else
		{
			sprintf(TempName,"%02x",token);
		}

		return ((char *) TempName );
	}
} /*tokenName()*/

/* compute the start of the l-th string after ch */
char *skipStrings(ch, l)
char             *ch;
long              l;
{
	long          i;

	for(i=0;i<l;i++)
	{
		ch += strlen(ch) + 1;
	}
	return (ch);
} /*skipStrings()*/

#define STARTMATCH    1
#define ENDMATCH      2
#define EXACTMATCH    4
#define MIDDLEMATCH   (STARTMATCH | ENDMATCH)
#define ANYMATCH      8


/*
  Decode pattern in string into pattern.  Note there is no check for
  length of string;  do that before calling.
  Allowable patterns are
     "part", "*part", "part*" and "*part*"
  where "part" must consist of name characters.  "part" and "*part*"
  are equivalent.  If part is empty, any name will be matched.
*/

short scanPat(char *string,long *matchType, char *pattern)
{
	int              patlen = strlen(string);
	char             c;

	*matchType = 0;
	if(patlen > 0 && string[patlen-1] == '*')
	{ /* terminal '*' found */
		patlen--;
		*matchType = STARTMATCH;
	}
	if(patlen > 0 && string[0] == '*')
	{				/* initial '*' found */
		*matchType |= ENDMATCH;
		string++;
		patlen--;
	}
	if(patlen > 0)
	{
		strncpy(pattern,string,(size_t) patlen);
	}
	else
	{
		*matchType = ANYMATCH;
	}
	pattern[patlen] = '\0';
	if(*matchType == 0)
	{
		*matchType = EXACTMATCH;
	}

	while((c = *pattern++))
	{
		if(!isnamechar(c) && c != TEMPPREFIX)
		{
			*matchType = 0;
			return (0);
		}
	}
	return (1);
} /*scanPat()*/

short matchName(char *name, long matchType, char *pattern)
{
	int       i, n1, n2, reply;

	if(matchType < 0)
	{
		return (0);
	}
	else if(matchType == ANYMATCH)
	{
		return (1);
	}
	else if(matchType == EXACTMATCH)
	{
		return (strcmp(name,pattern) == 0);
	}
	else
	{
		n1 = strlen(pattern);
		n2 = strlen(name);
		if(n1 <= n2)
		{
			if(matchType == STARTMATCH)
			{
				reply = (strncmp(pattern,name,(size_t) n1) == 0);
			}
			else if(matchType == ENDMATCH)
			{
				reply = (strcmp(pattern, name + n2 - n1) == 0);
			}
			else
			{
				for(i=n1;i<=n2;i++)
				{
					if(strncmp(pattern,name++,(size_t) n1) == 0)
					{
						break;
					}
				}
				reply = (i <= n2);
			}
		}
		else
		{
			reply = 0;
		}
	}
	return (reply);
} /*matchName()*/

#ifdef HASTIMEH
#ifndef EPX
#include <time.h>
#else /*EPX*/
#include <bsd43/time.h>
typedef long time_t;
#endif /*EPX*/
#endif /*HASTIMEH*/

char        *getTimeAndDate(void)
{
	char           *timeAndDate;
	int             length;
#ifdef HASTIMEH
	time_t          now;

	now = time((time_t *) 0);
	timeAndDate = asctime(localtime(&now));
#else /*HASTIMEH*/
	timeAndDate = "unknown date and time";
#endif /*HASTIMEH*/
	length = strlen(timeAndDate);
	if(timeAndDate[length-1] == '\n' || timeAndDate[length-1] == '\r')
	{
		timeAndDate[length-1] = '\0';
	}

	return (timeAndDate);
} /*getTimeAndDate)*/

void symExtremes(Symbolhandle symh, double vec[2])
{
	long      n;
	double    *x;

	vec[1] = -(vec[0] = HUGE);
	if(!isNull(symh))
	{
		n = symbolSize(symh);
		x = DATAPTR(symh);

		while(n-- > 0)
		{
			if(!isMissing(*x))
			{
				if(*x < vec[0])
				{
					vec[0] = *x;
				}
				else if(*x > vec[1])
				{
					vec[1] = *x;
				}
			}
			x++;
		}
	} /*if(!isNull(symh))*/
} /*symExtremes()*/

void symFill(Symbolhandle symh, double value)
{
	doubleFill(DATAPTR(symh), value, symbolSize(symh));
} /*symFill()*/

void doubleFill(double *x, double value, long length)
{
	while(length-- > 0)
	{
		*x++ = value;
	}
} /*doubleFill()*/
