/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment GetNames
#else
#define __SEG__ GetNames
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

/*
  names <- varnames([model]), where model is a CHARACTER variable (default
  STRMODEL) returns a CHARACTER vector with the names of the dependent
  variable and the factors and  variates in the model.  No syntax checking
  is done except names a checked for length.  Thus varnames("y a b a.b.c") and
  varnames("y=a+b+a.b.c") all return cat("y","a","b","c").
*/

#define   TAB       '\t'
#define   NL        10
#define   CR        13
#define   SP        ' '
static long         getVariableNames(char *model,
									 char names[MAXVARS+1][NAMELENGTH+1])
{ 
	char            c;
	char            token[2*NAMELENGTH+1];
	long            length, nfactors = 0, i;
	WHERE("getVariableNames");

	while(1)
	{
		while((c = *model) != '\0' && c != '#' && !isnamestart(c) &&
			  c != TEMPPREFIX)
		{ /* look for start of name */
			model++;
		}
		if(c == '\0' || c == '#')
		{
			break;
		}
		token[0] = c;
		
		for(length=1;length<2*NAMELENGTH;length++)
		{
			if(!isnamechar(model[length]))
			{
				break;
			}
			token[length] = model[length];
		}
		token[length] = '\0';
		if(length > NAMELENGTH)
		{
			sprintf(OUTSTR,
					"ERROR: name %s in model is longer than %ld characters",
					token, (long) NAMELENGTH);
			goto errorExit;
		}
		if(strcmp(token,"E") == 0)
		{ /* allow for error terms */
			for(i=1;(c = model[i]) != '\0';i++)
			{ /* skip white space looking for '(' */
				if(c != SP && c != TAB && c != NL && c != CR)
				{
					break;
				}
			}
		}
		else
		{
			c = '\0';
		}
		if(c != '(')
		{ /* not error term */
			for(i=0;i<nfactors;i++)
			{ /* have we seen name before? */
				if(strlen(names[i]) == length && strcmp(token,names[i]) == 0)
				{ /* yes */
					break;
				}
			}
			if(i == nfactors)
			{					/* new name */
				if(nfactors > MAXVARS)
				{
					sprintf(OUTSTR,
							"ERROR: more than %ld names in model",
							(long) MAXVARS);
					goto errorExit;
				}
				strcpy(names[nfactors],token);
				nfactors++;
			}
		}
		model += length;
	} /* while(1) */

	return (nfactors);
	
  errorExit:
	putOUTSTR();
	return (-1);
	
}


Symbolhandle    varnames(list)
Symbolhandle    list;
{
	long            nargs = NARGS(list);
	Symbolhandle    symh;
	Symbolhandle    result = (Symbolhandle) 0;
	char            names[MAXVARS+1][NAMELENGTH+1];
	char           *model, *place;
	long            nvars = (long) NVARS;
	long            i, nfactors = 0, needed = 0;
	WHERE("factornames");
	
	*OUTSTR = '\0';
	if(nargs > 1)
	{
		badNargs(FUNCNAME,-1);
		goto errorExit;
	}
	symh = COMPVALUE(list,0);
	if(symh != (Symbolhandle) 0)
	{
		nvars = -1;
		if(TYPE(symh) != CHAR || !isScalar(symh))
		{
			sprintf(OUTSTR,
					"ERROR: argument to %s must be quoted string or CHARACTER variable",
					FUNCNAME);
			goto errorExit;
		}
		model = STRINGPTR(symh);
	}
	else if(nvars >= 0)
	{
		nfactors = nvars+1;
		for(i=0;i<nfactors;i++)
		{
			strcpy(names[i],VARNAMES[i]);
		}
	}
	else if((symh = Lookup("STRMODEL")) != (Symbolhandle) 0 &&
			TYPE(symh) == CHAR && isScalar(symh))
	{
		model = STRINGPTR(symh);
		nvars = -1;
	}
	else
	{
		sprintf(OUTSTR,
				"ERROR: no argument for %s and no current model",
				FUNCNAME);
		goto errorExit;
	}
	if(nvars < 0 && (nfactors = getVariableNames(model,names)) < 0)
	{
		goto errorExit;
	}

	if(nfactors == 0)
	{
		sprintf(OUTSTR,"ERROR: no legal variate or factor names in %s",model);
		goto errorExit;
	}
	
	for(i=0;i<nfactors;i++)
	{
		needed += strlen(names[i]) + 1;
	}
	
	result = Install(SCRATCH,CHAR);
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	TMPHANDLE = mygethandle(needed);
	setSTRING(result,TMPHANDLE);
	if(TMPHANDLE == (char **) 0)
	{
		goto errorExit;
	}
	place = STRINGPTR(result);
	for(i=0;i<nfactors;i++)
	{
		strcpy(place,names[i]);
		place += strlen(place) + 1;
	}
	setNDIMS(result,1);
	setDIM(result,1,nfactors);
	
	return (result);
	
  errorExit:
	putOUTSTR();
	Removesymbol(result);
	
	return(0);
	
}

#define YONLY          0
#define XONLY         -1
#define ALL           -2
#define VARIATESONLY  -3
#define FACTORSONLY   -4

Symbolhandle    modelvars(Symbolhandle list)
{
	Symbolhandle        result = (Symbolhandle) 0, symh;
	Symbolhandle        symhModel = (Symbolhandle) 0;
	modelInfoPointer    info = (modelInfoPointer) 0;
	long                nargs = NARGS(list), nvars = (long) NVARS;
	long                mvars, ndata, nfactors, nmissing;
	long                varnos[MAXVARS+1], maxvarnos = 0;
	long                length, i, j;
	long                useCurrentModel = 0;
	double             *x, *y;
	char              **strmodel = (char **) 0;
	char              **strmodel1 = (char **) 0;
	char               *keyword = (char *) 0;
	WHERE("modelvar");
	TRASH(1,errorExit);
	
	if(nargs > 2)
	{
		badNargs(FUNCNAME,-2);
		goto errorExit;
	}

	symh = COMPVALUE(list,0);
	if(!argOK(symh,0,1))
	{
		goto errorExit;
	}

	if((keyword = isKeyword(symh)))
	{
		mvars = 1;
		if(strcmp(keyword,"all") == 0)
		{
			varnos[0] = ALL;
		}
		else if(strcmp(keyword,"x") == 0)
		{
			varnos[0] = XONLY;
		}
		else if(strcmp(keyword,"y") == 0)
		{
			varnos[0] = YONLY;
		}
		else if(strncmp(keyword,"variate",7) == 0)
		{
			varnos[0] = VARIATESONLY;
		}
		else if(strncmp(keyword,"factor",6) == 0)
		{
			varnos[0] = FACTORSONLY;
		}
		else
		{
			badKeyword(FUNCNAME,keyword);
			goto errorExit;
		}
		if(TYPE(symh) != LOGIC || !isScalar(symh) || DATAVALUE(symh,0) == 0.0)
		{
			sprintf(OUTSTR, "ERROR: value for keyword %s must be T",keyword);
			goto errorExit;
		}
	} /*if((keyword = isKeyword(symh)))*/
	else
	{
		if(TYPE(symh) != REAL || !isVector(symh))
		{
			sprintf(OUTSTR,
					"ERROR: 1st argument must be a REAL scalar or vector");
			goto errorExit;
		}
		mvars = symbolSize(symh);
		if(mvars > MAXVARS+1)
		{
			sprintf(OUTSTR,
					"ERROR: 1st argument has length greater than %ld",
					(long) MAXVARS+1);
			goto errorExit;
		}
	
		for(j=0;j<mvars;j++)
		{
			varnos[j] = (long) DATAVALUE(symh,j);
			if(DATAVALUE(symh,j) != (double) varnos[j] || varnos[j] < 0)
			{
				sprintf(OUTSTR,
						"ERROR: elements of 1st argument must be non-negative integers");
				goto errorExit;
			}
			if(varnos[j] > maxvarnos)
			{
				maxvarnos = varnos[j];
			}
		} /*for(j=0;j<mvars;j++)*/
	} /*if((keyword = isKeyword(symh))){...}else{...}*/	

	if(nargs == 1)
	{ /* no model specified since the model must be argument 2 */
		if(nvars >= 0)
		{ /* active model exists */
			if(maxvarnos > nvars)
			{
				sprintf(OUTSTR,
						"ERROR: fewer than %ld variables in model %s",
						maxvarnos, *STRMODEL);
			}
			else if(PREVMODELTYPE & FASTANOVA)
			{
				sprintf(OUTSTR,
						"ERROR: %s(varno) not valid after fastanova(); try %s(varno,STRMODEL)",
						FUNCNAME, FUNCNAME);
			}
			if(*OUTSTR)
			{
				goto errorExit;
			}
			useCurrentModel = 1;
		} /*if(nvars >= 0)*/
		else if(STRMODEL != (char **) 0)
		{
			strmodel = STRMODEL;
		}
		else
		{ /* use model in variable STRMODEL if available */
			symhModel = Lookup("STRMODEL");
			if(symhModel == (Symbolhandle) 0)
			{
				sprintf(OUTSTR,
						"ERROR: no active model and variable STRMODEL does not exist");
			}
			else if(TYPE(symhModel) != CHAR || !isScalar(symhModel) ||
					strlen(STRINGPTR(symhModel)) == 0)
			{
				sprintf(OUTSTR,
						"ERROR: STRMODEL is not scalar CHARACTER variable or is null string");
			}
			if(*OUTSTR)
			{
				goto errorExit;
			}
			strmodel = STRING(symhModel);
		}
	} /*if(nargs == 1)*/
	else
	{ /* model provided */
		symhModel = COMPVALUE(list,1);
		if(symhModel != (Symbolhandle) 0)
		{
			if(TYPE(symhModel) != CHAR || !isScalar(symhModel) ||
			   strlen(STRINGPTR(symhModel)) == 0)
			{
				sprintf(OUTSTR,
						"ERROR: model for %s must be non-null quoted string or CHARACTER variable",
						FUNCNAME);
				goto errorExit;
			}
		}
		strmodel = STRING(symhModel);
	} /*if(nargs == 1){...}else{...}*/

	if(!useCurrentModel)
	{
		if((strmodel1 = setStrmodel(strmodel,0)) == (char **) 0)
		{
			goto errorExit;
		}
		toTrash(strmodel1,0);
		
		initmodelparse(strmodel1);

		if (modelparse() != 0)
		{							/* parse it */
			goto errorExit;
		}
		info = getModelInfo(); /* retrieve model information */
		if((ndata = checkVariables(info)) < 0)
		{
			goto errorExit;
		}
		nmissing = countMissing(info, (double ***) 0);
		if(nmissing < 0)
		{
			goto errorExit;
		}
		nvars = info->nvars;
	} /*if(!useCurrentModel)*/
	else
	{
		strmodel = STRMODEL;
		info = MODELINFO;
		ndata = (long) NDATA;
	}
	nfactors = info->nfactors;

	if(mvars == 1 && varnos[0] < 0)
	{
		if(varnos[0] == ALL || varnos[0] == XONLY)
		{ /* all:T or x:T */
			if(varnos[0] == XONLY && nvars == 0)
			{
				sprintf(OUTSTR,
						"ERROR: there are no variates or factors or r.h.s. of %s",
						*strmodel);
				goto errorExit;
			}
			
			mvars = (varnos[0] == XONLY) ? nvars : nvars + 1;
			i = (varnos[0] == XONLY) ? 1 : 0;
			for(j=0;j<mvars;j++)
			{
				varnos[j] = j + i;
			}
		}
		else if(varnos[0] == VARIATESONLY)
		{ /* variates:T */
			if(nfactors == nvars)
			{
				sprintf(OUTSTR,
						"ERROR: there are no variates on r.h.s. of %s",
						*strmodel);
				goto errorExit;
			}
			mvars = 0;
			for(i=0;i<nvars;i++)
			{
				if(info->nclasses[i] < 0)
				{
					varnos[mvars++] = i;
				}
			}
		}
		else
		{ /* factors:T */
			if(nfactors == 0)
			{
				sprintf(OUTSTR,
						"ERROR: there are no factors on r.h.s. of %s",
						*strmodel);
				goto errorExit;
			}
			mvars = 0;
			for(i=0;i<nvars;i++)
			{
				if(info->nclasses[i] > 0)
				{
					varnos[mvars++] = i + 1;
				}
			}
		}
	} /*if(mvars == 1 && varnos[0] < 0)*/
	
	length = 0;
	for(j=0;j<mvars;j++)
	{
		if(varnos[j] > nvars)
		{
			sprintf(OUTSTR,
					"ERROR: model does not contain %ld variates or factors",
					varnos[j]);
			goto errorExit;
		}
		length += symbolSize(info->modelvars[varnos[j]]);
	} /*for(j=0;j<mvars;j++)*/
	
	result = RInstall(SCRATCH,length);
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	if(length > ndata)
	{
		setNDIMS(result,2);
		setDIM(result,1,ndata);
		setDIM(result,2,length/ndata);
	}
	else
	{
		setNCLASS(result,(varnos[0] == 0) ?
				  NCLASS(info->modelvars[0]) : info->nclasses[varnos[0]-1]);
	}
	
	y = DATAPTR(result);
	for(j=0;j<mvars;j++)
	{
		symh = info->modelvars[varnos[j]];
		length = symbolSize(symh);
		x = DATAPTR(symh);
		for(i=0;i<length;i++)
		{
			*y++ = *x++;
		}
	} /*for(j=0;j<mvars;j++)*/
	
	return (result);
	
  errorExit:
	putOUTSTR();
	Removesymbol(result);
	emptyTrash();
	
	return (0);
}

#define GSTRMODEL1    0
#define GMISSWTS      1
#define GTMPALPHA     2
#define GCOLPTR       3
#define NTRASH        4

/*
  xvariables() returns the X-variables (not the model variables) associated
  with the active model, or specified by STRMODEL if there is no active model
  xvariables(strmodel) returns the X-variables associated with strmodel, a
  CHARACTER variable that is parsed just as in the glm functions.

  modelinfo(xvars:T) is equivalent to xvariables()
  modelinfo(coefs:T) returns the coefficient vector (matrix) if after manova()
  associated with the design matrix.
  modelinfo(xtxinv:T) return (X'X)^-1 associated with the design matrix
  modelinfo(colcount:T) returns a vector containing the number of X-variables
  associated with each term
  modelinfo(weights:T) returns the vector of weights associated with the
  computation.
  modelinfo(strmodel:T) returns just the model provided or STRMODEL
  modelinfo(bitmodel:T) returns a vector of length NTERMS (NVAR==1)
    or a nterms by (2 or 3) matrix (depending on NVAR) of the bit 
    representation of the model
  modelinfo(all:T) is equivalent to modelinfo(xvars:T,coefs:T,xtxinv:T,
	colcount:T,weights:T,strmodel:T,bitmodel:T)
  If more than one keyword is specified a structure is returned with
  components, xvariables, coefs, xtxinv, colcounts, and/or weights
*/

Symbolhandle xvariables(list)
Symbolhandle      list;
{
	Symbolhandle     arg;
	Symbolhandle     result = (Symbolhandle) 0;
	modelInfoPointer info;
	Symbolhandle     symhxvars = (Symbolhandle) 0;
	Symbolhandle     symhyvar = (Symbolhandle) 0;
	Symbolhandle     symhcoefs = (Symbolhandle) 0;
	Symbolhandle     symhxtxinv = (Symbolhandle) 0;
	Symbolhandle     symhdfs = (Symbolhandle) 0;
	Symbolhandle     symhwts = (Symbolhandle) 0;
	Symbolhandle     symhstrmodel = (Symbolhandle) 0;
	Symbolhandle     symhbitmodel = (Symbolhandle) 0;
	Symbolhandle     symhtermnames = (Symbolhandle) 0;
	long             makextxinv = 0, makecoefs = 0, makexvars = 1, makedfs = 0;
	long             makeyvar = 0, makewts = 0;
	long             makestrmodel = 0, makebitmodel = 0, maketermnames = 0;
	long             nargs = NARGS(list);
	long             nvars = (long) NVARS, nregx;
	long             ndata, ny = (long) NY;
	long             nterms;
	long             ncomps = 1, icomps = 0, ncols = 0;
	long             nmissing = 0, ncells;
	long             colno = 0, i, j, iterm, itermj, logValue;
	double         **tmpalpha = (double **) 0;
	double         **coefs = (double **) 0;
	double         **dfs = (double **) 0;
	double          *alpha1, *alpha2;
	double         **regx = (double **) 0, *xij;
	double         **misswts = (double **) 0;
	double          *wts;
	double          *bitmodel;
	double           missval = 0;
	long           **colptr = (long **) 0;
	long             xvarsonly, margs, aliased = 0;
	char            *keyword, *prevfun = (char *) 0;
	char           **strmodel = (char **) 0, **strmodel1 = (char **) 0;
	WHERE("xvariables");
	TRASH(NTRASH,errorExit);
	
	xvarsonly = (strcmp(FUNCNAME,"xvariables") == 0);
	margs = (xvarsonly) ? 2 : 7;
	if(nargs > margs)
	{
		badNargs(FUNCNAME,-margs);
		goto errorExit;
	}

	arg = COMPVALUE(list,nargs-1);
	if((keyword = isKeyword(arg)) && strncmp(keyword,"miss",4) == 0)
	{
		if(TYPE(arg) != REAL || !isScalar(arg))
		{
			sprintf(OUTSTR,
					"ERROR: value for %s must be REAL scalar or '?'",
					keyword);
			goto errorExit;
		}
		if(isMissing(DATAVALUE(arg,0)))
		{
			setMissing(missval);
		}
		else
		{
			missval = DATAVALUE(arg,0);
		}
		Removesymbol(arg);
		COMPVALUE(list,nargs-1) = (Symbolhandle) 0;
		nargs -= (nargs > 1) ? 1 : 0;
	} /*if((keyword = isKeyword(arg)) && strncmp(keyword,"miss",4) == 0)*/
	
	arg = COMPVALUE(list,0);
	if(!xvarsonly)
	{ /* modelinfo() call */
		makexvars = 0;

		for(i=0;i<nargs;i++)
		{
			arg = COMPVALUE(list,i);
			if(!argOK(arg,0,i+1))
			{
				goto errorExit;
			}
			if(!(keyword = isKeyword(arg)))
			{
				sprintf(OUTSTR,
						"ERROR: all arguments to %s must be keywords",
						FUNCNAME);
			}
			else if(TYPE(arg) != LOGIC || !isScalar(arg))
			{
				sprintf(OUTSTR,
						"ERROR: value for %s must be T or F",keyword);
			}
			if(*OUTSTR)
			{
				goto errorExit;
			}
			logValue = (DATAVALUE(arg,0) != 0.0) ? 1 : 0;
			if(strncmp(keyword,"xvar",4) == 0)
			{
				makexvars = logValue;
			}
			else if(strncmp(keyword,"y",1) == 0)
			{
				makeyvar = logValue;
			}
			else if(strcmp(keyword,"xtxinv") == 0)
			{
				makextxinv = logValue;
			}
			else if(strncmp(keyword,"coef",4) == 0)
			{
				makecoefs = logValue;
			}
			else if(strncmp(keyword,"df",2) == 0 ||
					strncmp(keyword,"colc",4) == 0)
			{
				makedfs = logValue;
			}
			else if(strncmp(keyword,"wt",2) == 0 || 
					strncmp(keyword,"weight",6) == 0)
			{
				makewts = logValue;
			}
			else if(strncmp(keyword,"strm",4) == 0)
			{
				makestrmodel = logValue;
			}
			else if(strncmp(keyword,"bitm",4) == 0)
			{
				makebitmodel = logValue;
			}
			else if(strncmp(keyword,"term",4) == 0)
			{
				maketermnames = logValue;
			}
			else if(strcmp(keyword,"all") == 0)
			{
				if(logValue == 0)
				{
					sprintf(OUTSTR,
							"ERROR: value for %s must be T", keyword);
					goto errorExit;
				}
				makexvars = makeyvar = makecoefs = makextxinv = makedfs =
					makewts = makestrmodel = makebitmodel = maketermnames = 1;
				if(XTXINV == (double **) 0)
				{
					makecoefs = makextxinv = 0;
				}
			}
			else
			{
				badKeyword(FUNCNAME,keyword);
				goto errorExit;
			}
		} /*for(i=0;i<nargs;i++)*/

		ncomps = makexvars + makeyvar + makextxinv + makecoefs +
			makedfs + makewts + makestrmodel + makebitmodel + maketermnames;

		if(ncomps == 0)
		{
			sprintf(OUTSTR,
					"ERROR: you must specify T for at least one %s keyword",
					FUNCNAME);
		}
		else if(ncomps == 1 && makexvars)
		{
			Removesymbol(arg);
			arg = COMPVALUE(list,0) = (Symbolhandle) 0;
			xvarsonly = 1;
		}
		else if(nvars < 0)
		{
			sprintf(OUTSTR,
					"ERROR: no active model for %s", FUNCNAME);
		}
		else if(makecoefs || makextxinv || makewts)
		{
			if(PREVMODELTYPE & BALANOVA)
			{
				prevfun = "anova() with complete balanced design";
			}
			else if(PREVMODELTYPE & IPF)
			{
				prevfun = "ipf()";
			}
			else if(PREVMODELTYPE & FASTANOVA)
			{
				prevfun = "fastanova()";
			}
			if(prevfun)
			{
				sprintf(OUTSTR,
						"ERROR: improper use of %s after %s",
						FUNCNAME, prevfun);
			}
			else if((makecoefs || makextxinv) && XTXINV == (double **) 0)
			{
				sprintf(OUTSTR,
						"ERROR: coefficients and xtxinv not available");
			}
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
	} /*if(!xvarsonly)*/

	if(xvarsonly)
	{ /* xvariables(), no keywords except 'missing' allowed */
		if(arg != (Symbolhandle) 0)
		{
			if((keyword = isKeyword(arg)))
			{
				badKeyword(FUNCNAME,keyword);
				goto errorExit;
			}
			/* argument must be a model */
			if(TYPE(arg) != CHAR || !isScalar(arg))
			{
				sprintf(OUTSTR,
						"ERROR: model for %s must be scalar CHARACTER variable or quoted string",
						FUNCNAME);
				goto errorExit;
			}
			strmodel = STRING(arg);
		} /*if(arg != (Symbolhandle) 0)*/
		else if(nvars < 0)
		{ /* no model specified and no active model */
			arg = Lookup("STRMODEL");
			if(arg == (Symbolhandle) 0)
			{
				sprintf(OUTSTR,
						"ERROR: no active model and STRMODEL is not defined");
			}
			else if(TYPE(arg) != CHAR || !isScalar(arg))
			{
				sprintf(OUTSTR,
						"ERROR: STRMODEL is not scalar CHARACTER variable",
						FUNCNAME);
			}
			else
			{
				strmodel = STRING(arg);
			}
		}
		else if(PREVMODELTYPE & FASTANOVA)
		{
			strmodel = STRMODEL;
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
	} /*if(xvarsonly)*/
	
	if(strmodel != (char **) 0)
	{ /* we need to parse the model and generate X-variables */
		if((strmodel1 = setStrmodel(strmodel,0)) == (char **) 0)
		{
			goto errorExit;
		}
		toTrash(strmodel1,GSTRMODEL1);
		
		initmodelparse(strmodel1);

		if (modelparse() != 0)
		{							/* parse it */
			goto errorExit;
		}
		info = getModelInfo(); /* retrieve model information */
		if((ndata = checkVariables(info)) < 0)
		{
			goto errorExit;
		}
		nmissing = countMissing(info, &misswts);
		if(nmissing < 0)
		{ /* error */
			goto errorExit;
		}
		toTrash(misswts,GMISSWTS);

		nvars = info->nvars;
		nterms = info->nterms;
		
		nregx = 0;
		for(iterm = 0;iterm < nterms;iterm++)
		{
			nregx += buildXmatrix(iterm, (double *) 0, info);
		}
	} /*if(strmodel != (char **) 0)*/
	else
	{ /* use information from current model */
		info = MODELINFO;
		strmodel = STRMODEL;
		misswts = MISSWTS;
		ndata = (long) NDATA;
		nregx = (long) NREGX;
		nterms = (long) NTERMS;
		nmissing = ndata - (long) NOTMISSING;
	} /*if(strmodel != (char **) 0){...}else{...}*/

	discardScratch(strmodel1,GSTRMODEL1,char);

	if(ncomps == 1)
	{
		result = (makestrmodel || maketermnames) ?
			Install(SCRATCH, CHAR) : RInstall(SCRATCH,0);
	}
	else
	{
		result  = StrucInstall(SCRATCH,ncomps);
	}
	
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	
	if(makexvars)
	{
		if(ncomps == 1)
		{
			symhxvars = result;
		}
		else
		{
			symhxvars = COMPVALUE(result,icomps++) = Makereal(0);
			if(symhxvars == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symhxvars,"xvars");
		}
		setDATA(symhxvars,(double **) mygethandle(ndata*nregx*sizeof(double)));
		regx = DATA(symhxvars);

		if(regx == (double **) 0)
		{
			goto errorExit;
		}
		setDIM(symhxvars,1,ndata);
		if(nregx > 1)
		{
			setNDIMS(symhxvars ,2);
			setDIM(symhxvars ,1,ndata);
			setDIM(symhxvars ,2,nregx);
		}	
		for(iterm = 0;iterm < nterms;iterm++)
		{
			colno += buildXmatrix(iterm, *regx + colno*ndata, info);
		}

		if(misswts != (double **) 0)
		{
			if(!isMissing(missval))
			{
				sprintf(OUTSTR,
						"WARNING: X-variables for cases with missing data set to %g",
						missval);
			}
			else
			{
				sprintf(OUTSTR,
						"WARNING: X-variables for cases with missing data set to MISSING");
			}
			putOUTSTR();
			
			for(i=0;i<ndata;i++)
			{
				if((*misswts)[i] == 0.0)
				{
					xij = *regx + i;
					for(j=0;j<nregx;j++)
					{
						if(isMissing(missval))
						{
							setMissing(*xij);
						}
						else
						{
							*xij = missval;
						}
						xij += ndata;
					}
				}
			} /*for(i=0;i<ndata;i++)*/

			if(misswts != MISSWTS)
			{
				discardScratch(misswts,GMISSWTS,double);
			}
		} /*if(misswts != (double **) 0)*/
	} /*if(makexvars)*/

	if(makeyvar)
	{
		if(ncomps == 1)
		{
			symhyvar = result;
		}
		else
		{
			symhyvar = COMPVALUE(result,icomps++) = Makereal(0);
			if(symhyvar == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symhyvar,"y");
		}
		setDATA(symhyvar, (double **) myduphandle((char **) Y));
		if(DATA(symhyvar) == (double **) 0)
		{
			goto errorExit;
		}
		setNDIMS(symhyvar,(ny == 1) ? 1 : 2);
		setDIM(symhyvar,1,ndata);
		if(ny > 1)
		{
			setDIM(symhyvar,2,ny);
		}
	}/*if(makeyvar)*/
	
	if(makextxinv)
	{
		if(ncomps == 1)
		{
			symhxtxinv = result;
		}
		else
		{
			symhxtxinv = COMPVALUE(result,icomps++) = Makereal(0);
			if(symhxtxinv == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symhxtxinv,"xtxinv");
		}
		setDATA(symhxtxinv,(double **) myduphandle((char **) XTXINV));
		if(DATA(symhxtxinv) == (double **) 0)
		{
			goto errorExit;
		}
		setNDIMS(symhxtxinv,2);
		setDIM(symhxtxinv,1,nregx);
		setDIM(symhxtxinv,2,nregx);
	} /*if(makextxinv)*/

	if(makecoefs)
	{
		if(ncomps == 1)
		{
			symhcoefs = result;
		}
		else
		{
			symhcoefs = COMPVALUE(result,icomps++) = Makereal(0);
			if(symhcoefs == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symhcoefs,"coefs");
		}
		DATA(symhcoefs) = (double **) mygethandle(ny * nregx * sizeof(double));
		coefs = DATA(symhcoefs);
		if(coefs == (double **) 0)
		{
			goto errorExit;
		}
		setNDIMS(symhcoefs,(ny > 1) ? 2 : 1);
		setDIM(symhcoefs,1,nregx);
		if(ny > 0)
		{
			setDIM(symhcoefs,2,ny);
		}
		
		colno = 0;
		for(iterm = 0;iterm < nterms; iterm++)
		{
			if(getAlpha(iterm,colno, &tmpalpha, &colptr) < 0)
			{
				goto errorExit;
			}
			toTrash(tmpalpha,GTMPALPHA);
			toTrash(colptr,GCOLPTR);
			ncells = (*colptr)[0];

			for(i=1;i<=ncells;i++)
			{
				if((*colptr)[i] != -1)
				{ /* not intrinsically aliased */
					alpha1 = *tmpalpha + i - 1;
					alpha2 = *coefs + colno++;
					for(j=0;j<ny;j++)
					{
						*alpha2 = *alpha1;
						alpha1 += ncells;
						alpha2 += nregx;
					}
				} /*if((*colptr)[i] != -1)*/

				if((*colptr)[i] == -2)
				{
					aliased++;
				}
			} /*for(i=1;i<=ncells;i++)*/
			discardScratch(tmpalpha,GTMPALPHA,double);
			discardScratch(colptr,GCOLPTR,long);
		} /*for(iterm = 0;i< nterms; i++)*/
		if(aliased > 0)
		{
			sprintf(OUTSTR,
					"WARNING: coefficient%s of aliased X-variable%s set to 0",
					(aliased > 1) ? "s" : "", (aliased > 1) ? "s" : "");
			putOUTSTR();
		}		
	} /*if(makecoefs)*/

	if(makedfs)
	{
		if(ncomps == 1)
		{
			symhdfs = result;
		}
		else
		{
			symhdfs = COMPVALUE(result,icomps++) = Makereal(0);
			if(symhdfs == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symhdfs,"colcounts");
		}
		setDATA(symhdfs,(double **) mygethandle(nterms * sizeof(double)));
		dfs = DATA(symhdfs);
		if(dfs == (double **) 0)
		{
			goto errorExit;
		}
		setDIM(symhdfs,1,nterms);
		for(iterm = 0;iterm < nterms;iterm++)
		{
			(*dfs)[iterm] = (double) buildXmatrix(iterm, (double *) 0, info);
		}
	} /*if(makedfs)*/

	if(makewts)
	{
		if(ncomps == 1)
		{
			symhwts = result;
		}
		else
		{
			symhwts = COMPVALUE(result,icomps++) = Makereal(0);
			if(symhwts == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symhwts,"weights");
		}
		if(PREVMODELTYPE & USEWEIGHTS)
		{
			setDATA(symhwts,(double **) myduphandle((char **) ALLWTS));
		}
		else
		{
			setDATA(symhwts,(double **) mygethandle(ndata * sizeof(double)));
		}
		
		if(DATA(symhwts) == (double **) 0)
		{
			goto errorExit;
		}
		wts = DATAPTR(symhwts);
		for(i=0;i<ndata;i++)
		{
			*wts = (PREVMODELTYPE & USEWEIGHTS) ? (*wts)*(*wts) : 1.0;
			wts++;
		}		
		setDIM(symhwts,1,ndata);
	} /*if(makewts)*/

	if(makestrmodel)
	{
		if(ncomps == 1)
		{
			symhstrmodel = result;
		}
		else
		{
			symhstrmodel = COMPVALUE(result,icomps++) = Makesymbol(CHAR);
			if(symhstrmodel == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symhstrmodel,"strmodel");
		}
		TMPHANDLE = myduphandle(STRMODEL);
		setSTRING(symhstrmodel, TMPHANDLE);
		if(TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
		setNDIMS(symhstrmodel,1);
		setDIM(symhstrmodel,1,1);
	} /*if(makestrmodel)*/

	if(makebitmodel)
	{
		if(ncomps == 1)
		{
			symhbitmodel = result;
		}
		else
		{
			symhbitmodel = COMPVALUE(result,icomps++) = Makereal(0);
			if(symhbitmodel == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symhbitmodel,"bitmodel");
		}
		ncols = (nvars <= 32) ? 1 : ((nvars <= 64) ? 2 : 3);
		TMPHANDLE = mygethandle(nterms * ncols * sizeof(double));
		setDATA(symhbitmodel, (double **) TMPHANDLE);
		if(TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}

		setNDIMS(symhbitmodel, (ncols == 1) ? 1 : 2);
		setDIM(symhbitmodel,1,nterms);
		if(ncols > 1)
		{
			setDIM(symhbitmodel,2,ncols);
		}
		
		bitmodel = DATAPTR(symhbitmodel);
		for(iterm = 0;iterm < nterms;iterm++)
		{
			itermj = iterm; /* iterm + j * nterms */
			for(j=0;j<ncols;j++)
			{
				bitmodel[itermj] = (double) (*info->model)[iterm][j];
				itermj += nterms;
			}
		} /*for(iterm = 0;iterm < nterms;iterm++)*/
	} /*if(makebitmodel)*/

	if(maketermnames)
	{
		if(ncomps == 1)
		{
			symhtermnames = result;
		}
		else
		{
			symhtermnames = COMPVALUE(result,icomps++) = Makesymbol(CHAR);
			if(symhtermnames == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			setNAME(symhtermnames,"termnames");
		}
		TMPHANDLE = myduphandle(TERMNAMES);
		setSTRING(symhtermnames, TMPHANDLE);
		if(TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
		setNDIMS(symhtermnames,1);
		setDIM(symhtermnames,1,nterms+1);
	} /*if(maketermnames)*/

	emptyTrash();
	
	return (result);
	
  errorExit:
	putOUTSTR();
	emptyTrash();
	
	return (0);
}
