/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Yates
#else
#define __SEG__ Yates
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

extern long           argOK();
extern long           isVector();
extern void           badNargs();
extern Symbolhandle   RInstall();

#define GXOLD   0
#define GXNEW   1
#define NTRASH  2

Symbolhandle    yates(list)
Symbolhandle    list;

{
	Symbolhandle    symh, result = (Symbolhandle) 0;
	double        **xold = (double **) 0, **xnew = (double **) 0;
	double        **xtmp;
	double         *xin[2], *xout[2];
	long            n, ncols, i, j, k, halfn, nfact, dim[2];
	WHERE("yates");
	TRASH(NTRASH,errorExit);
	
	*OUTSTR = '\0';
	if(NARGS(list) > 1)
	{
		badNargs(FUNCNAME,1);
		goto errorExit;
	}
	
	symh = COMPVALUE(list, 0);
	if(!argOK(symh,0,1))
	{
		goto errorExit;
	}
	
	if (TYPE(symh) != REAL || !isMatrix(symh,dim))
	{
		sprintf(OUTSTR,"ERROR: argument to %s must be REAL vector or matrix",
				FUNCNAME);
		goto errorExit;
	}

	if(anyMissing(symh))
	{
		sprintf(OUTSTR,
				"ERROR: missing values not allowed in argument to %s",
				FUNCNAME);
		goto errorExit;
	}
	
	n = dim[0];
	ncols = dim[1];
	halfn = 1;
	for(nfact = 0; halfn < n; nfact++)
	{
		halfn *= 2;
	}

	if (n != halfn)
	{
		sprintf(OUTSTR,
				"ERROR: %s argument to %s must be a power of two",
				(ncols == 1) ? "length of" : "number of rows in",
				FUNCNAME);
		goto errorExit;
	}

	halfn = n / 2;

	result = RInstall(SCRATCH,(n - 1) * ncols);
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	setNDIMS(result,(ncols > 1) ? 2 : 1);
	setDIM(result,1,n-1);
	if(ncols > 1)
	{
		setDIM(result,2,ncols);
	}
	
	if(!getScratch(xold,GXOLD,n,double) ||
	   !getScratch(xnew,GXNEW,n,double))
	{
		goto errorExit;
	}

	for(j=0;j<ncols;j++)
	{
		xin[0] = DATAPTR(symh) + j*n;
		xout[0] = *xold;
	
		for (i = 0; i < n; i++)
		{
			*xout[0]++ = *xin[0]++;
		}
	
		for (k = 0;k < nfact;k++)
		{
			xin[0] = *xold;
			xin[1] = xin[0] + 1;
			xout[0] = *xnew;
			xout[1] = xout[0] + halfn;
			for (i = 0; i < halfn; i++)
			{
				*xout[0]++ = *xin[0] + *xin[1];
				*xout[1]++ = -*xin[0] + *xin[1];
				xin[0] += 2;
				xin[1] += 2;
			}
			/* swap xnew and xold */
			xtmp = xnew;
			xnew = xold;
			xold = xtmp;
		}

		xin[0] = (*xtmp) + 1;
		xout[0] = DATAPTR(result) + j*(n-1);
	
		for (i = 1; i < n; i++)
		{
			*xout[0]++ = *xin[0]++/(double) halfn;
		}
	}							/*for(j=0;j<ncols;j++)*/
	
	emptyTrash();

	return (result);

  errorExit:
	putOUTSTR();

	emptyTrash();
	Removesymbol(result);

	return (0);
	
}
