/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Yylex
#else
#define __SEG__ Yylex
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"

/* these must be defined identically in mainpars.y */
#define SKIPBRACKETBLOCK  3
#define SKIPPARENBLOCK    4

/*
  File contains routines related to lexical scanning.  The main entrance is
  yylex().  Most of the work is done in yylex1().

  930506: Dereferenced handle INPUTSTRING to inputString = *INPUTSTRING.
  Note this must be redone after every call that might cause memory to move,
  notably any .*Install(), mygethandle(), etc.

*/

#define Yylval       yylval

Symbolhandle         Install(), RInstall(), Lookup();
extern Symbolhandle  Yylval;
extern 	char       **mygethandle();
extern  char        *typeName();
extern  char        *tokenName();

extern long     PARSEMACROARGS;	/* defined in mainpars.c */
extern long     YYLEXMESSAGE ; /* defined in mainpars.c */

extern int     *ALTTOKENPTR;   /* defined in mainpars.c */
extern int      ALTINPUTTOKENS[];   /* defined in mainpars.c */
				/* if non-zero, replace next char if not ';'*/

extern long     MACROSTART; /* defined in Macrosetup */

extern long     LASTCH;

/*
  skip white space, allowing for line continuations
  it advances ISTRCHAR if necessary and returns (*INPUTSTRING)[ISTRCHAR]
*/

static unsigned int skipWhiteSpace(void)
{
	register unsigned char  c, c1;
	register unsigned char *inputString = *INPUTSTRING;
	
	while((c = inputString[ISTRCHAR]) == ' ' || c == '\t' ||
		  (c == '\\' &&
		   ((c1  = inputString[ISTRCHAR+1]) == '\n' || c1 == '\r')))
	{
		ISTRCHAR++;
		if(c == '\\')
		{
			ISTRCHAR++;
		}
	}
	return ((unsigned int) c);
} /*skipWhiteSpace()*/

/*
  Scan *INPUTSTRING for match of s, i.e., match all characters in s and
  determine that next character in *INPUTSTRING is not alpha-numeric.
  If match is found, return non-zero, otherwise zero.
*/

static long matchKeyword(char *s)
{
	unsigned char   *inputString = *INPUTSTRING + ISTRCHAR - 1;
	
	while(*s)
	{
		if(*s++ != *inputString++)
		{
			return (0);
		}
	}

	return (!isnamechar(*inputString));

} /*matchKeyword()*/

static char skipBrackets(long goalLevel, char leftbracket)
{
	unsigned char   *inputString = *INPUTSTRING;
	char             c;
	char             rightbracket = (leftbracket == '{') ? '}' : ')';
	long             level = (leftbracket == '{') ? BRACKETLEV : PARENLEV;
	long             inQuotes = 0, foundComment = 0;
	WHERE("skipBrackets");

	c = inputString[ISTRCHAR];
	if(c != '\0')
	{
		while(level > goalLevel)
		{
			c = inputString[ISTRCHAR];
			if(c == '\0')
			{
				break;
			}
			ISTRCHAR++;
			if(!foundComment)
			{
				if(c == '"')
				{
					inQuotes = !inQuotes;
				}
				if(!inQuotes)
				{
					if(c == rightbracket)
					{
						level--;
					}
					else if(c == leftbracket)
					{
						level++;
					}
					else if(c == '#')
					{
						foundComment = 1;
					}
					else if(c != ' ' && c != '\t')
					{
						LASTCH = c;
					}
				}
			} /*if(!foundComment)*/
			if(c == '\n' || c == '\r')
			{
				foundComment = 0;
			}
		} /*while(level > goalLevel)*/
	} /*if(c != '\0')*/
	if(leftbracket == '{')
	{
		BRACKETLEV = level;
	}
	else
	{
		PARENLEV = level;
	}
	return (c);  /* return the last character read */
} /*skipBrackets()*.


/*
  YYLEXMESSAGE == SKIPBRACKETBLOCK : skip compound statements on if, elseif,
  or while with condition F
  YYLEXMESSAGE == SKIPPARENBLOCK : skip '(...)' after elseif when condition
  from preceding if or elseif is T
  YYLEXMESSAGE > MAXWDEPTH : fakeout parser to get out of breaks or breakall
*/

static unsigned int processBreak(void)
{
	unsigned char         *inputString = *INPUTSTRING;
	unsigned int           c;
	long                   currentBracketLev = BRACKETLEV;
	long                   currentParenLev = PARENLEV;
	long                   wantedBracketLev, extraBrackets;
	long                   i;
	WHERE("processBreak");

	ALTTOKENPTR = ALTINPUTTOKENS;
	*ALTTOKENPTR = 0;

	if(YYLEXMESSAGE == SKIPBRACKETBLOCK)
	{
		/*
		  Just got F condition on if or while, skip to end.
		  Should be positioned immediately before '{' or erroneous
		  character.
		  */

		YYLEXMESSAGE = 0;
		if((c = inputString[ISTRCHAR++]) != '{')
		{						/* must be erroneous character; return it */
			ALTTOKENPTR = (int *) 0;
			return (c);
		}
 		BRACKETLEV++;
		/* move ISTRCHAR to point after real compound statement */
		c = skipBrackets(currentBracketLev, c);
		BRACKETLEV = currentBracketLev + 1;
		/* create empty compound statement */
		if(c == '}')
		{
			i = 0;
			ALTINPUTTOKENS[i++] = (int) ';';
			ALTINPUTTOKENS[i++] = (int) '}';
			ALTINPUTTOKENS[i] = 0;
		}
		else
		{			/* end of line before matching '}' */
			sprintf(OUTSTR,
					"ERROR: need matching '{' and '}' after 'if', 'while', and 'for'");
			putOUTSTR();
			ALTTOKENPTR = (int *) 0;
		}
		return ((unsigned int) '{');
	}

	if(YYLEXMESSAGE == SKIPPARENBLOCK)
	{
		/*
		  Pevious T condition on elsif; skip (...).
		  Should be positioned immediately before '('
		  */

		YYLEXMESSAGE = 0;
		if((c = inputString[ISTRCHAR++]) != '(')
		{/* must be erroneous character; return it (should not happen)*/
			ALTTOKENPTR = (int *) 0;
			return (c);
		}
 		PARENLEV++;
		/* move ISTRCHAR to point after real compound statement */
		c = skipBrackets(currentParenLev, c);
		PARENLEV = currentParenLev + 1;
		/* create empty parent list */
		if(c == ')')
		{
			i = 0;
			ALTINPUTTOKENS[i++] = (int) ')';
			ALTINPUTTOKENS[i] = 0;
		}
		else
		{			/* end of line before matching ')' */
			ALTTOKENPTR = (int *) 0;
		}
		return ((unsigned int) '(');
	}
	/*
	  If YYLEXMESSAGE > MAXWDEPTH it should be of form (wanted
	  bracketlevel) + MAXWDEPTH + 1  and we are just past
	  "break" or "breakall".  We need to set up ALTINPUTTOKENS
	  with enough ';}' to fake out the parser to think
	  that we are getting out of any enclosing if's, as well
	  as enough ';}' to convince it that we are getting out of
	  the enclosing while (break) or all the enclosing whiles
	  (breakall)
	  In every case the character we return now will be ';'
	  */

	if(YYLEXMESSAGE > MAXWDEPTH)
	{							/* must be break or breakall */
		wantedBracketLev = YYLEXMESSAGE - MAXWDEPTH - 1;
	}

	YYLEXMESSAGE = 0;
	*ALTTOKENPTR++ = (int) '}';
	extraBrackets = currentBracketLev - wantedBracketLev - 1;
	for(i = 0;i< extraBrackets;i++)
	{
		*ALTTOKENPTR++ = (int) ';';
		*ALTTOKENPTR++ = (int) '}';
	}
	c = skipBrackets(wantedBracketLev, '{');
	if(c != (int) '}')				/* end of line before matching '}' */
	{
		sprintf(OUTSTR,"ERROR: missing '}'");
		putOUTSTR();
		ALTTOKENPTR--;					/* wipe out one '}' to force error in parser */
	}
	*ALTTOKENPTR = 0;
	ALTTOKENPTR = ALTINPUTTOKENS;
	BRACKETLEV = currentBracketLev;
	return ((unsigned int) ';');

} /*processBreak()*/

/*
  create CHAR symbols containing macro arguments.  Called by
  yylex when PARSEMACROARGS != 0.
  If PARSEMACROARGS == 1, the next character must be '(', ',' , ')', or '\n'
  Otherwise, it encapsulates INPUTSTRING up to next one of these characters
  outside of quotes and at the same bracket level as an argument
*/
static unsigned int parseMacro(void)
{
	char            tempString[MAXSTRING+1];
	unsigned char  *inputString = *INPUTSTRING;
	unsigned int    c, c1;
	long            macroParenLevel = 0, inQuotes;
	long            empty = 1;
	long            place;
	WHERE("parseMacro");

	/* PARSEMACROARGS == 1 means next character must be '(', ',', ')' or '\n'*/
	if(PARSEMACROARGS == 1)
	{
		c = inputString[ISTRCHAR++];
		if(c == '(' || c == ',')
		{
			/* expect macro argument next time in */
			PARSEMACROARGS = 2;
			if(c == '(')
			{
				PARENLEV++;
			}
		}
		else					/* ')' or end of line */
		{
			PARSEMACROARGS = 0;
			if(c == ')')
			{
				PARENLEV--;
			}
			else
			{
				c = ENDOFL;
			}
		}
		return (c);
	} /*if(PARSEMACROARGS == 1)*/
	/*
	  PARSEMACROARGS must be 2.
	  Convert everything up to ',' or ')' with macroParenLevel = 0 to CHAR
	  */

	Yylval = (Symbolhandle) 0;

	/* skip leading white space, allowing for line continuations */
	c = skipWhiteSpace();
	if(c == '\0')
	{
		PARSEMACROARGS = 0;
		return (ENDOFL);
	}

	macroParenLevel = 0;
	inQuotes = 0;
	place = 0;
	tempString[0] = '\0';
	while(place < MAXSTRING)
	{
		ISTRCHAR++;
		while(!inQuotes && c == '\\' &&
			  ((c1 = inputString[ISTRCHAR]) == '\n' || c1 == '\r'))
		{						/* allow for continuation lines */
			ISTRCHAR++;
			c = inputString[ISTRCHAR++];
		}
		if(c == '\0')
		{
			break;
		}

		tempString[place++] = c;
		if (c == '"')
		{
			inQuotes = !inQuotes;
		}
		else if(!inQuotes &&
				((macroParenLevel <= 0 && (c == ',' || c == ')')) ||
				 c == '\n' || c == '\r'))
		{						/* found end of argument */
			/* put back character for next call */
			ISTRCHAR--;
			place--;
			PARSEMACROARGS = (c == ',') ? 1 : 0;
			break;
		}
		if(empty && !isspace(c))
		{
			empty = 0;
		}
		if(!inQuotes)
		{
			if(c == '(' || c == '[')
			{
				macroParenLevel++;
			}
			else if(c == ')' || c == ']')
			{
				macroParenLevel--;
			}
		}
		c = inputString[ISTRCHAR];
	} /*while(place < MAXSTRING)*/

	if (place == MAXSTRING)
	{
		sprintf(OUTSTR,"ERROR: macro argument too long (>%ld)",
				(long) MAXSTRING);
		putOUTSTR();
		return (BADTOKEN);
	}

	while(place > 0 && ((c1 = tempString[place-1]) == ' ' || c1 == '\t'))
	{ /* trim off trailing white space */
		place--;
	}

	if(place == 0 || empty)
	{
		/*
		  On empty macro argument, '()', '(,...)', '(...,,...)', or
		  '(...,)' , return null symbol, and token VAR.
		  White argument (e.g., ', ,') is also considered null
		  */
		Yylval = (Symbolhandle) 0;
		if(c != '\0')
		{
			c = VAR;
		}
	} /*if(place == 0 || empty)*/
	else if(place > 0 && c != '\0')
	{
		Yylval = Install(SCRATCH, CHAR);
		if(Yylval == (Symbolhandle) 0)
		{
			return (FATALERROR);
		}

		setNDIMS(Yylval,1);
		setDIM(Yylval,1,1);
		tempString[place] = '\0';
		TMPHANDLE = mygethandle((long) place + 1);
		setSTRING(Yylval,TMPHANDLE);
		if(TMPHANDLE == (char **) 0)
		{
			return (FATALERROR);
		}
		strcpy(STRINGPTR(Yylval), tempString);
		c = VAR; /* macro argument has been transformed to CHAR var */
	} /*if(place == 0 || empty){}else if(place > 0 && c != 0){}*/

	if(c == '\0')
	{
		PARSEMACROARGS = 0;
	}

	return (c);
} /*parseMacro()*/

static unsigned int scanNumber(void)
{
	unsigned char      *inputString = *INPUTSTRING, *endptr;
	double              y;
	
	y = mystrtod((char *) inputString+ISTRCHAR, (char **) &endptr);
	if(y == 0 && endptr == inputString+ISTRCHAR)
	{
		ISTRCHAR++;
		sprintf(OUTSTR,"ERROR: nonnumeric where numeric expected");
		return (ERROR);
	}
	ISTRCHAR = endptr - inputString;

	Yylval = RInstall(NUMSCRATCH, (long) 1);
	if(Yylval == (Symbolhandle) 0)
	{
		return (FATALERROR);
	}
	DATAVALUE(Yylval,0) = y;
	return (NUMBER);
} /*scanNumber()*/

#define MAXTOKENSCAN 25

 /* normal scan */
static unsigned int yylex1(void)
{
	unsigned char       *inputString;
	char                 tempString[MAXSTRING+1];
	unsigned int         c, c1, c2;
	double               y = -1.0;
	long                 i, j, nameStart,tokenLength, place;
	char                 token[MAXTOKENSCAN+1];
	unsigned char       *endptr;
	WHERE("yylex1");

  start:
	inputString = *INPUTSTRING;
	while ((c = (unsigned int) inputString[ISTRCHAR]) == ' ' || c == '\t')
	{
		ISTRCHAR++;				/* skip whitespace */
	}

	if(c == '\0')
	{
		return (BRACKETLEV != 0) ? BADTOKEN : 0;
	}

	if(c == '}' && LASTCH != ';' && LASTCH != ENDOFL)
	{							/* force ';' before '}' */
		c = ';';
		return(c);
	}

	ISTRCHAR++;

	switch (c)
	{
	  case '\n':
#ifndef MPW
	  case '\r':
#endif							/*MPW*/
		if(BRACKETLEV > 0 && (LASTCH == ';' || LASTCH == ENDOFL))
		{
			goto start;
		}
		if(BRACKETLEV == 0 && (c = skipWhiteSpace()) != '\0')
		{
			yyerror("ERROR: probably too many }'s in macro prior to or");
			return (ERROR);
		}
		
		return(ENDOFL);

	  case ';':
		return (c);

		/* simply pass through certain math symbols */
	  case '+':
	  case '-':
	  case '/':
	  case '$':
	  case ',':
	  case '\'':
		return (c);

	  case '(':
		PARENLEV++;
		return (c);
	  case ')':
		PARENLEV--;
		if (PARENLEV < 0)
		{
			yyerror("ERROR: too many )'s");
			c = ERROR;
		}
		return (c);
	  case '[':
	  case ']':
		if(inputString[ISTRCHAR] == c)
		{
			if(c == '[')
			{/* signal that following character is also '[' */
				c = LEFTBRACK2;
			}
			else
			{/* signal that following character is also ']' */
				c = RIGHTBRACK2;
			}
		}
		return (c);

	  case '*':
		/* "**" is equivalent to "^" */
		if((unsigned int) inputString[ISTRCHAR] != '*')
		{
			return (c);
		}
		ISTRCHAR++;
		/* fall through */
	  case '^':
		return (POW);
	  case '<':
	  case '>':
		if(inputString[ISTRCHAR] == c)
		{
			if(c == '>')
			{
				ISTRCHAR++;
				return (RIGHTANGLE);
			}

			for(i=2;inputString[ISTRCHAR-1+i] == '<';i++)
			{
				;				/* count number of successive '<'s */
			}

			if((i & 1) == 0)
			{					/* even number of '<'s */
				MACROSTART = ISTRCHAR-1;
				ISTRCHAR++;
				return (LEFTANGLE);
			}
			else
			{					/* probably of form expr <<<name>> parsed as expr < <<name>> */
				return ('<');
			}

		}
		/* fall through */
	  case '=':
	  case '!':
		/* '<', '>', '='. '!' */
		if ((c2 = (unsigned int) inputString[ISTRCHAR++]) == '=')
		{
			if(c == '=')
			{
				c = EQ;
			}
			else if(c == '>')
			{
				c = GE;
			}
			else if(c == '<')
			{
				c = LE;
			}
			else
			{
				c = NE;
			}
		}
		else if(c == '<' && c2 == '-')
		{
			c = ASSIGN;
			c1 = (unsigned int) inputString[ISTRCHAR];
			if(strchr("+*-/^%",(int) c1))
			{
				c2 = (unsigned int) inputString[ISTRCHAR+1];
				if(c1 == '+' || c1 == '-')
				{ /* must have white space after <-+ and <-- */
					if(c2 == '\\')
					{
						place = ISTRCHAR++;
						(void) skipWhiteSpace();
						for(i=place;i<ISTRCHAR;i++)
						{
							c2 = inputString[i];
							if(c2 == ' ' || c2 == '\t')
							{
								break;
							}
						}
						ISTRCHAR = place;
					}
					if(c2 != ' ' && c2 != '\t')
					{
						c2 = '\0';
					}					
				} /*if(c1 == '+' || c1 == '-')*/
				
				if(c2 != '\0')
				{
					ISTRCHAR++;
					switch (c1)
					{
					  case '+':
						c = ASSIGNADD;
						break;

					  case '-':
						c = ASSIGNSUB;
						break;

					  case '*':
						if(c2 == '*')
						{
							c = ASSIGNPOW;
							ISTRCHAR++;
						}
						else
						{
							c = ASSIGNMULT;
						}
						break;

					  case '/':
						c = ASSIGNDIV;
						break;

					  case '%':
						if(c2 == '%')
						{
							c = ASSIGNMOD;
							ISTRCHAR++;
						}
						else
						{
							ISTRCHAR--;
						}
						break;

					  case '^':
						c = ASSIGNPOW;
						break;
					} /*switch (c1)*/
				} /*if(c2 != '\0') */
			} /*if(strchr("+*-/^%",(int) c1))*/
		}/*else if(c == '<' && c2 == '-')*/
		else
		{
			ISTRCHAR--;
		}
		return (c);
#ifdef MACINTOSH
	  case 0xb2: /* Macintosh <= */
		return (LE);
	  case 0xb3: /* Macintosh >= */
		return (GE);
	  case 0xad: /* Macintosh != */
		return (NE);
#endif /*MACINTOSH*/
	  case ':':
		return (ASSIGN1);
	  case '&':					/* '&&' equivalent to '&' and '||' equivalent to '|' */
	  case '|':
		if (inputString[ISTRCHAR] == c)
		{
			ISTRCHAR++;
			c = (c == '&') ? AND : OR;
		}
		return (c);

	  case '%':
		c2 = inputString[ISTRCHAR++];
		switch (c2)
		{
		  case '%':
			c = MOD; /* '%%' */
			break;
#ifdef BITOPS
		  case '!':
			c = BITNOT; /* '%!' */
			break;
		  case '&':
			c = BITAND; /* '%&' */
			break;
		  case '^':
			c = BITXOR; /* '%^' */
			break;
		  case '|':
			c = BITOR; /* '%|' */
			break;
#endif /*BITOPS*/
		  default:
			if(inputString[ISTRCHAR++] == '%')
			{
				switch (c2)
				{
				  case '*':
					c = MATMULT;   /* '%*%' */
					break;
				  case 'c':
					c = TRMATMULT;   /* '%c%' */
					break;
				  case 'C':
					c = MATMULTTR;   /* '%C%' */
					break;
				  default:
					ISTRCHAR--;
					c = BADTOKEN;
				} /*switch (c2)*/
			} /*if(inputString[ISTRCHAR++] == '%') */
			else 
			{
				ISTRCHAR--;
				c = BADTOKEN;
			}
			break;
		} /*switch (c2)*/

		return (c);

	  case '?':					/* missing value */
		Yylval = RInstall(NUMSCRATCH, (long) 1);
		if(Yylval == (Symbolhandle) 0)
		{
			return (FATALERROR);
		}

		setNDIMS(Yylval,1);
		setDIM(Yylval,1,1);
		setMissing(DATAVALUE(Yylval,0));

		return (NUMBER);		/* missing value */

	  case '"':					/* start of string */
		Yylval = Install(STRINGSCRATCH, CHAR);
		inputString = *INPUTSTRING;
		if(Yylval == (Symbolhandle) 0)
		{
			return (FATALERROR);
		}
		setNDIMS(Yylval,1);
		setDIM(Yylval,1,1);
		for (i = 0; i < MAXSTRING; i++)
		{
			c = inputString[ISTRCHAR++];
			if(c == '\0')
			{
				ISTRCHAR--;
				c = '"';
			}
			if (c == '"')
			{
				break;
			}

			if(c == '\\')
			{					/* inside a string, so always accept */
				/*
				  diad			  result
				  '\n'				NL
				  '\r'				CR
				  '\t'				TAB
				  '\ijk', i < 4		decoded ascii character
				  '\EOL'			'\EOL'
				  '\somethingElse'	'somethingElse'
				  */
				c2 = inputString[ISTRCHAR++];
				if(c2 == '\0')
				{
					c = '"';
					ISTRCHAR--;
				}
				else if(c2 == '\n' || c2 == '\r')
				{
					tempString[i++] = c;
					if(i >= MAXSTRING)
					{
						break;
					}
					c = c2;
				}
				else if(c2 == 'n')
				{
					c = '\n';
				}
				else if(c2 == 'r')
				{
					c = '\r';
				}
				else if(c2 == 't')
				{
					c = '\t';
				}
				else if(isdigit(c2) && c2 < '8')
				{				/* octal characters, e.g., \012 */
					place = ISTRCHAR;
					c = c2 - '0';
					for(j=1;j<3 ;j++)
					{
						c2 = inputString[ISTRCHAR++];
						if(!isdigit(c2) || c2 > '7')
						{
							break;
						}
						c = 8*c + c2 - '0';
					} /*for(j=1;j<3 ;j++)*/
					if(j < 3)
					{
						ISTRCHAR--;
					}
					if(c == '\0')
					{
						yyerror("ERROR: illegal octal null character");
						return(BADTOKEN);
					}
					else if(c > 0377)
					{ /* use only first two digits */
						c /= 8;
						ISTRCHAR--;
					}
				}
				else
				{
					c = c2;
				}
			}
			tempString[i] = c;
		} /*for (i = 0; i < MAXSTRING; i++)*/

		if (i == MAXSTRING)
		{
			sprintf(OUTSTR,"ERROR: string too long (>%ld)",(long) MAXSTRING);
			yyerror(OUTSTR);
			return (ERROR);
		}
		tempString[i] = '\0';
		TMPHANDLE = mygethandle((long) i + 1);
		setSTRING(Yylval,TMPHANDLE);
		if(TMPHANDLE == (char **) 0)
		{
			return (FATALERROR);
		}
		strcpy(STRINGPTR(Yylval), tempString);
		return (VAR);			/* quoted string*/

	  case '{':					/* start of compound statement */
		BRACKETLEV++;
		return (c);

	  case '}':					/* end of one level of compound statement */
		BRACKETLEV--;
		if (BRACKETLEV < 0)
		{
			yyerror("ERROR: too many }'s");
			c = ERROR;
		}
		return (c);

	  case '#':					/* start of comment; skip rest of line */
		while((c = inputString[ISTRCHAR]) && c != '\n' && c != '\r')
		{
			ISTRCHAR++;
		}
		if (c != '\0')
		{
			ISTRCHAR++;
		}

		return (ENDOFL);

	  case '\\':				/* may be continuation line */
		if((c2 = inputString[ISTRCHAR]) == '\n' || c2 == '\r')
		{
			ISTRCHAR++;
			goto start;
		}
		/* else fall through */

	  default:					/* see if number or variable name*/
		c2 = inputString[ISTRCHAR];
		if((c == 'T' || c == 'F') && !isnamechar(c2))
		{ /* treat T and F special so they don't need to be in symbol table */
			Yylval = RInstall(LOGICSCRATCH,(long) 1);
			if(Yylval == (Symbolhandle) 0)
			{
				return (FATALERROR);
			}
			setTYPE(Yylval,LOGIC);
			DATAVALUE(Yylval,0) = (c == 'T') ? 1.0 : 0.0;
			return (VAR);
		}

		if (!isnamestart(c) && !isdigit(c) && c != '.' && c != TEMPPREFIX)
		{						/*illegal character*/
			yyerror("ERROR: illegal character");
			return (ERROR);
		}

		/* c is alpha or '@' or '.' or digit */
		nameStart = ISTRCHAR-1;	/* save loc of 1st char of name */
		if (isnamestart(c) || c == TEMPPREFIX)
		{/* may be keyword, name, or temporary name */
			/*  TEMPPREFIX starts temp name */
			if(strchr("befiqws",(int) c))
			{/* check for key words; befiqw are possible 1st characters */
				if (c == 'i' && matchKeyword("if"))
				{
					IFBRACKETLEV[IDEPTH] = BRACKETLEV;
					ISTRCHAR++;
					if(!findLParen("if"))
					{
						return (ERROR);
					}
					return (IF);
				}
				if (c == 'e')
				{
					if(matchKeyword("else"))
					{
						IFBRACKETLEV[IDEPTH] = BRACKETLEV;
						ISTRCHAR += 3;
						return (ELSE);
					}
					
					if(matchKeyword("elseif"))
					{
						IFBRACKETLEV[IDEPTH] = BRACKETLEV;
						ISTRCHAR += 5;
						if(!findLParen("elseif"))
						{
							return (ERROR);
						}
						return (ELSEIF);
					}
				}

				if(c == 'f' && matchKeyword("for"))
				{
					WHILESTARTS[WDEPTH] = ISTRCHAR - 1;
					WHILEBRACKETLEV[WDEPTH] = BRACKETLEV;
					ISTRCHAR += 2;
					if(!findLParen("for"))
					{
						return (ERROR);
					}
					return (FOR);
				}
				if(c == 'w' && matchKeyword("while"))
				{
					WHILESTARTS[WDEPTH] = ISTRCHAR - 1;
					WHILEBRACKETLEV[WDEPTH] = BRACKETLEV;
					ISTRCHAR += 4;
					if(!findLParen("while"))
					{
						return (ERROR);
					}
					return (WHILE);
				}
				if(c == 'b')
				{
					if(matchKeyword("break"))
					{
						ISTRCHAR += 4;
						Yylval = (Symbolhandle) 0;
						if(WDEPTH <= 0)
						{
							yyerror("ERROR: break without for or while");
							return (BREAK);
						}
						place = ISTRCHAR;
						c = skipWhiteSpace();
						if(isdigit(c) || c == '-' || c == '+')
						{ /* break n */
							y = mystrtod((char *) inputString+ISTRCHAR,
										 (char **) &endptr);
							ISTRCHAR = endptr - inputString;
							c2 = skipWhiteSpace();
							if(c2 != ' ' && c2 != '\t' &&
							   c2 != ';' && c2 != '}')
							{
								ISTRCHAR++;
								c = BADTOKEN;
							}
							else
							{
								ISTRCHAR = endptr - inputString;
								if(y <= 0 || y != floor(y))
								{
									yyerror("ERROR: break n, with n not integer > 0");
									y = -1;
								}
								else if((int) y > WDEPTH)
								{
									yyerror("ERROR: break n where n > loop depth");
									y = -1;
								}
								c = BREAK;
							}
						} /*if(isdigit(c) || c == '-' || c == '+')*/
						else
						{ /* break */
							c2 = inputString[ISTRCHAR];
							if(c2 != ' ' && c2 != '\t' &&
							   c2 != '\n' &&c2 != '\r' &&
							   c2 != ';' && c2 != '}')
							{
								ISTRCHAR++;
								c = BADTOKEN;
							}
							else
							{
								ISTRCHAR = place;
								c = BREAK;
								y = 1.0;
							}
						}
						if(y > 0.0)
						{
							Yylval = RInstall(NUMSCRATCH, (long) 1);
							if(Yylval == (Symbolhandle) 0)
							{
								return (FATALERROR);
							}
							DATAVALUE(Yylval,0) = y;
						}
						return (c);
					}
					else if(matchKeyword("breakall"))
					{			/* must check for breakall before break */
						ISTRCHAR += 7;
						if(WDEPTH <= 0)
						{
							yyerror("ERROR: breakall without for or while");
						}
						else
						{
							place = ISTRCHAR;
							c = skipWhiteSpace();
							if((c2 = inputString[ISTRCHAR]) != ';' &&
							   c2 != '}')
							{
								ISTRCHAR++;
								c = BADTOKEN;
							}
							else
							{
								ISTRCHAR = place;
								y = (double) WDEPTH;
								Yylval = RInstall(NUMSCRATCH, (long) 1);
								if(Yylval == (Symbolhandle) 0)
								{
									return (FATALERROR);
								}
								DATAVALUE(Yylval,0) = y;
							}
						}
						return (BREAK);
					}
				} /* if(c == 'b') */


				if (c == 'q' && matchKeyword("quit") ||
					c == 'b' && matchKeyword("bye") ||
					c == 'e' && matchKeyword("end") ||
					c == 's' && matchKeyword("stop"))
				{				/* it's quit or end */
					ISTRCHAR += (c == 'q' || c == 's') ? 3 : 2;
					if(inputString[ISTRCHAR] == '(' &&
					   inputString[ISTRCHAR+1] == ')')
					{/* allow quit() */
						ISTRCHAR += 2;
					}
					while((c2 = inputString[ISTRCHAR++]) != '\0' &&
						  (c2 == ';' || c2 == ' ' || c2 == '\t' ||
						   c2 == '\n' || c2 == '\r'))
					{
						;
					}

					if(c2 == '\0' || (BRACKETLEV == 0 && c2 == '#'))
					{
						return (END);
					}
					else
					{
						yyerror("ERROR: 'quit', 'bye', or 'end' not last command in line");
						return (ERROR);
					}
				}
			} /*if(strchr("befiqw",c))*/
			/* must be name or temporary name */

			tokenLength = 0;
			token[0] = c;
			while (isnamechar(c = inputString[ISTRCHAR]))
			{
				ISTRCHAR++;
				if (tokenLength < MAXTOKENSCAN-1)
				{
					token[++tokenLength] = c;
				}
				else
				{
					break;
				}
			} /*while (isnamechar(c = inputString[ISTRCHAR]))*/
			token[++tokenLength] = '\0';
			if(c == '.')
			{
				ISTRCHAR++;
				yyerror("ERROR:  do not use . in variable names");
				return (ERROR);
			}
				
			if (tokenLength > NAMELENGTH)
			{
				/* variable name too long */
				myprint("ERROR: variable name too long: ");
				myerrorout(token);
				ISTRCHAR -= tokenLength - NAMELENGTH - 1;
				return (ERROR);
			} /*if (tokenLength > NAMELENGTH)*/

			if(token[0] == TEMPPREFIX && !isnamestart(token[1]))
			{
				if(isnamechar(token[1]))
				{
					yyerror("ERROR: illegal temporary name");
				}
				else
				{
					ISTRCHAR++;
					myprint("ERROR: ");
					myprint(token);
					yyerror(" not followed by alphabetic character");
				}
				return(ERROR);
			}/*if(token[0] == TEMPPREFIX && !isnamestart(token[1]))*/

			Yylval = Lookup(token);
			if (Yylval == (Symbolhandle) 0)
			{
#ifdef MACINTOSH
				if(token[0] == 'C' && strcmp(token,"CLIPBOARD") == 0)
				{
					return (FATALERROR);
				}
				else
#endif /*MACINTOSH*/
				{
					Yylval = Install(token, UNDEF);
					if(Yylval == (Symbolhandle) 0)
					{
						return (FATALERROR);
					}
					setNDIMS(Yylval,1);
				}
			}
/* 920623 Return VAR for all names except BLTINs or MACROs followed by '(' */
			switch ((int) TYPE(Yylval))
			{
			  case BLTIN:
				if(findLParen((char *) 0))
				{
					if(tokenLength == 5 && strcmp(token, "batch") == 0)
					{
						return (BATCH);
					}
					else
					{
						return (BLTIN);
					}
				}
				else
				{
					return (VAR);
				}

			  case UNDEF:
			  case ASSIGNED:
			  case STRUC:
			  case PLOTINFO:
			  case REAL:
			  case CHAR:
			  case LIST:
			  case LOGIC:
				return (VAR);	/* symbol of type REAL, CHAR, LIST, LOGIC */
			  case MACRO:
				if(findLParen((char *) 0))
				{
					MACROSTART = nameStart;
					PARSEMACROARGS = 1;
					return (MACRO);
				}
				else
				{
					return (VAR);
				}


			  default:			/* unknown type */
				myerrorout("ERROR: symbol of unknown type");

				return (BADTOKEN);

			} /*switch ((int) TYPE(Yylval))*/
		} /*if (isnamestart(c) || c == TEMPPREFIX)*/

		/* if not alpha, ought to be number unless it is of form \.[a-zA-z] */
		if (c == '.' &&
			!isdigit(c1 = inputString[ISTRCHAR]))
		{
			ISTRCHAR++;
			if(c1 == TEMPPREFIX || isnamestart(c1) || isnamechar(c1))
			{
				yyerror("ERROR:  do not use . in variable names");
				return (ERROR);
			}
			return (BADTOKEN);
		}
		ISTRCHAR--;
		return (scanNumber());
	} /*switch (c)*/
} /*yylex1()*/

/*
  Called after token matching BLTIN or MACRO name or after IF, ELSEIF, WHILE,
  and FOR.
  If next non-white character is '(' it resets ISTRCHAR & returns 1
  it returns 0.
  If name != (char *) it prints an error message if '(' is not found
*/

long findLParen(char *name)  /* also called from Symbol.c */
{
	long           savePlace = ISTRCHAR;
	unsigned int   c;
	WHERE("findLParen");

	/* skip white space including \-protected newlines */
	c = skipWhiteSpace();
	if(c != '(')
	{
		ISTRCHAR = savePlace;
		if(name != (char *) 0)
		{
			sprintf(OUTSTR,
					"ERROR: need '(' after '%s'", name);
			yyerror(OUTSTR);
		}
		return (0);
	}
	else
	{
		return (1);
	}
} /*findLParen(*/

int yylex()
{				/* lexical analyzer */

	unsigned int    c = 0;

	WHERE("yylex");

	if(ALTTOKENPTR)
	{
		/*
		  first see if we have stashed stuff away to get out of while's
		  or if's or to provide a fake 'for()'
		  */
		c = (unsigned int) *ALTTOKENPTR++;
		if(c != 0)
		{
			if(c == '}')
			{
				BRACKETLEV--;
			}
			else if(c == '{')
			{
				BRACKETLEV++;
			}
		}
		else
		{
			ALTTOKENPTR = (int *) 0;
		}
		LASTCH  = '\0';
	}

	if(c == 0)
	{
		if(PARSEMACROARGS != 0)
		{
			c = parseMacro();
			LASTCH  = '\0';
		}
		else if(YYLEXMESSAGE > 0)
		{
			c = processBreak();
			LASTCH  = '\0';
		}
		else if(-YYLEXMESSAGE > ' ')
		{
			/*
			  -YYLEXMESSAGE is token to return next
			  After while(){} and if(){}, -YYLEXMESSAGE should be ';'
			  After <<name>> -YYLEXMESSAGE should be appropriate token
			  */
			c = -YYLEXMESSAGE;
			YYLEXMESSAGE = 0;
			LASTCH  = c;
		}
		else
		{
			/* normal scan of input stream */
			c = yylex1();
			LASTCH = c;
		}
	}

	if(GUBED & 1)
	{
		sprintf(OUTSTR,"yylex: token = %s",tokenName(c));
		
		if(c == VAR || c == BLTIN || c == MACRO || c == MACRO1 || c == UNDEF ||
		   c == ASSIGNED || c == NUMBER)
		{
			if(Yylval != (Symbolhandle) 0)
			{
				char    *name = NAME(Yylval);
				
				if(c == VAR)
				{
					sprintf(OUTSTR+strlen(OUTSTR),
							", type = %s", typeName(TYPE(Yylval)));
					if(strcmp(name,LOGICSCRATCH) == 0)
					{
						sprintf(OUTSTR+strlen(OUTSTR), ", value = %s",
								(DATAVALUE(Yylval,0) != 0.0) ? "T" : "F");
					}
				}
				else if(c == NUMBER)
				{
					if(isMissing(DATAVALUE(Yylval,0)))
					{
						sprintf(OUTSTR+strlen(OUTSTR),", value = MISSING");
					}
					else
					{
						sprintf(OUTSTR+strlen(OUTSTR),
								", value = %-12.5g",DATAVALUE(Yylval,0));
					}
				}
				if(!isscratch(NAME(Yylval)))
				{
					sprintf(OUTSTR+strlen(OUTSTR),", name = '%s'", name);
				}
			} /*if(Yylval != (Symbolhandle) 0)*/
		}
		else if(c == BREAK)
		{
			if(Yylval != (Symbolhandle) 0)
			{
				sprintf(OUTSTR+strlen(OUTSTR)," %g", DATAVALUE(Yylval,0));
			}
			else
			{
				sprintf(OUTSTR+strlen(OUTSTR)," ERROR");
			}
		}
		putOUTSTR();
	}

#ifdef YYLEXDB

	printf("token = %s\n",tokenName(c));

#endif /*YYLEXDB*/
	if(c == NUMBER)
	{
		c = VAR;
	}

	if(c == FATALERROR)
	{
		FatalError = 1;
	}
	
	return (c ? c : ENDOFL);

} /*yylex()*/

